/* Generated by Cython 3.0.12 */

/* BEGIN: Cython Metadata
{
    "distutils": {
        "define_macros": [
            [
                "NPY_NO_DEPRECATED_API",
                "NPY_1_7_API_VERSION"
            ]
        ],
        "depends": [],
        "extra_compile_args": [
            "-Wno-unreachable-code"
        ],
        "include_dirs": [
            "/tmp/build-env-82p3tyx6/lib/python3.12/site-packages/numpy/_core/include"
        ],
        "name": "serializor.binary.ser",
        "sources": [
            "src/serializor/binary/ser.py"
        ]
    },
    "module_name": "serializor.binary.ser"
}
END: Cython Metadata */

#ifndef PY_SSIZE_T_CLEAN
#define PY_SSIZE_T_CLEAN
#endif /* PY_SSIZE_T_CLEAN */
#if defined(CYTHON_LIMITED_API) && 0
  #ifndef Py_LIMITED_API
    #if CYTHON_LIMITED_API+0 > 0x03030000
      #define Py_LIMITED_API CYTHON_LIMITED_API
    #else
      #define Py_LIMITED_API 0x03030000
    #endif
  #endif
#endif

#include "Python.h"

    #if PY_MAJOR_VERSION >= 3
      #define __Pyx_PyFloat_FromString(obj)  PyFloat_FromString(obj)
    #else
      #define __Pyx_PyFloat_FromString(obj)  PyFloat_FromString(obj, NULL)
    #endif
    

    #if PY_MAJOR_VERSION <= 2
    #define PyDict_GetItemWithError _PyDict_GetItemWithError
    #endif
    

    #if (PY_VERSION_HEX < 0x030700b1 || (CYTHON_COMPILING_IN_PYPY && PYPY_VERSION_NUM < 0x07030600)) && !defined(PyContextVar_Get)
    #define PyContextVar_Get(var, d, v)         ((d) ?             ((void)(var), Py_INCREF(d), (v)[0] = (d), 0) :             ((v)[0] = NULL, 0)         )
    #endif
    
#ifndef Py_PYTHON_H
    #error Python headers needed to compile C extensions, please install development version of Python.
#elif PY_VERSION_HEX < 0x02070000 || (0x03000000 <= PY_VERSION_HEX && PY_VERSION_HEX < 0x03030000)
    #error Cython requires Python 2.7+ or Python 3.3+.
#else
#if defined(CYTHON_LIMITED_API) && CYTHON_LIMITED_API
#define __PYX_EXTRA_ABI_MODULE_NAME "limited"
#else
#define __PYX_EXTRA_ABI_MODULE_NAME ""
#endif
#define CYTHON_ABI "3_0_12" __PYX_EXTRA_ABI_MODULE_NAME
#define __PYX_ABI_MODULE_NAME "_cython_" CYTHON_ABI
#define __PYX_TYPE_MODULE_PREFIX __PYX_ABI_MODULE_NAME "."
#define CYTHON_HEX_VERSION 0x03000CF0
#define CYTHON_FUTURE_DIVISION 1
#include <stddef.h>
#ifndef offsetof
  #define offsetof(type, member) ( (size_t) & ((type*)0) -> member )
#endif
#if !defined(_WIN32) && !defined(WIN32) && !defined(MS_WINDOWS)
  #ifndef __stdcall
    #define __stdcall
  #endif
  #ifndef __cdecl
    #define __cdecl
  #endif
  #ifndef __fastcall
    #define __fastcall
  #endif
#endif
#ifndef DL_IMPORT
  #define DL_IMPORT(t) t
#endif
#ifndef DL_EXPORT
  #define DL_EXPORT(t) t
#endif
#define __PYX_COMMA ,
#ifndef HAVE_LONG_LONG
  #define HAVE_LONG_LONG
#endif
#ifndef PY_LONG_LONG
  #define PY_LONG_LONG LONG_LONG
#endif
#ifndef Py_HUGE_VAL
  #define Py_HUGE_VAL HUGE_VAL
#endif
#define __PYX_LIMITED_VERSION_HEX PY_VERSION_HEX
#if defined(GRAALVM_PYTHON)
  /* For very preliminary testing purposes. Most variables are set the same as PyPy.
     The existence of this section does not imply that anything works or is even tested */
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_LIMITED_API 0
  #define CYTHON_COMPILING_IN_GRAAL 1
  #define CYTHON_COMPILING_IN_NOGIL 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #undef CYTHON_USE_TYPE_SPECS
  #define CYTHON_USE_TYPE_SPECS 0
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #if PY_VERSION_HEX < 0x03050000
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #undef CYTHON_AVOID_BORROWED_REFS
  #define CYTHON_AVOID_BORROWED_REFS 1
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_GIL
  #define CYTHON_FAST_GIL 0
  #undef CYTHON_METH_FASTCALL
  #define CYTHON_METH_FASTCALL 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #ifndef CYTHON_PEP487_INIT_SUBCLASS
    #define CYTHON_PEP487_INIT_SUBCLASS (PY_MAJOR_VERSION >= 3)
  #endif
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 1
  #undef CYTHON_USE_MODULE_STATE
  #define CYTHON_USE_MODULE_STATE 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 0
  #endif
  #undef CYTHON_USE_FREELISTS
  #define CYTHON_USE_FREELISTS 0
#elif defined(PYPY_VERSION)
  #define CYTHON_COMPILING_IN_PYPY 1
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_LIMITED_API 0
  #define CYTHON_COMPILING_IN_GRAAL 0
  #define CYTHON_COMPILING_IN_NOGIL 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #ifndef CYTHON_USE_TYPE_SPECS
    #define CYTHON_USE_TYPE_SPECS 0
  #endif
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #if PY_VERSION_HEX < 0x03050000
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #undef CYTHON_AVOID_BORROWED_REFS
  #define CYTHON_AVOID_BORROWED_REFS 1
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_GIL
  #define CYTHON_FAST_GIL 0
  #undef CYTHON_METH_FASTCALL
  #define CYTHON_METH_FASTCALL 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #ifndef CYTHON_PEP487_INIT_SUBCLASS
    #define CYTHON_PEP487_INIT_SUBCLASS (PY_MAJOR_VERSION >= 3)
  #endif
  #if PY_VERSION_HEX < 0x03090000
    #undef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #elif !defined(CYTHON_PEP489_MULTI_PHASE_INIT)
    #define CYTHON_PEP489_MULTI_PHASE_INIT 1
  #endif
  #undef CYTHON_USE_MODULE_STATE
  #define CYTHON_USE_MODULE_STATE 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE (PY_VERSION_HEX >= 0x030400a1 && PYPY_VERSION_NUM >= 0x07030C00)
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 0
  #endif
  #undef CYTHON_USE_FREELISTS
  #define CYTHON_USE_FREELISTS 0
#elif defined(CYTHON_LIMITED_API)
  #ifdef Py_LIMITED_API
    #undef __PYX_LIMITED_VERSION_HEX
    #define __PYX_LIMITED_VERSION_HEX Py_LIMITED_API
  #endif
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_LIMITED_API 1
  #define CYTHON_COMPILING_IN_GRAAL 0
  #define CYTHON_COMPILING_IN_NOGIL 0
  #undef CYTHON_CLINE_IN_TRACEBACK
  #define CYTHON_CLINE_IN_TRACEBACK 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #undef CYTHON_USE_TYPE_SPECS
  #define CYTHON_USE_TYPE_SPECS 1
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #undef CYTHON_USE_ASYNC_SLOTS
  #define CYTHON_USE_ASYNC_SLOTS 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #ifndef CYTHON_USE_UNICODE_WRITER
    #define CYTHON_USE_UNICODE_WRITER 0
  #endif
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_GIL
  #define CYTHON_FAST_GIL 0
  #undef CYTHON_METH_FASTCALL
  #define CYTHON_METH_FASTCALL 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #ifndef CYTHON_PEP487_INIT_SUBCLASS
    #define CYTHON_PEP487_INIT_SUBCLASS 1
  #endif
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_MODULE_STATE
  #define CYTHON_USE_MODULE_STATE 1
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE 0
  #endif
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 0
  #endif
  #undef CYTHON_USE_FREELISTS
  #define CYTHON_USE_FREELISTS 0
#elif defined(Py_GIL_DISABLED) || defined(Py_NOGIL)
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_LIMITED_API 0
  #define CYTHON_COMPILING_IN_GRAAL 0
  #define CYTHON_COMPILING_IN_NOGIL 1
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #ifndef CYTHON_USE_TYPE_SPECS
    #define CYTHON_USE_TYPE_SPECS 0
  #endif
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #ifndef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #ifndef CYTHON_USE_PYLONG_INTERNALS
    #define CYTHON_USE_PYLONG_INTERNALS 0
  #endif
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_GIL
  #define CYTHON_FAST_GIL 0
  #ifndef CYTHON_METH_FASTCALL
    #define CYTHON_METH_FASTCALL 1
  #endif
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #ifndef CYTHON_PEP487_INIT_SUBCLASS
    #define CYTHON_PEP487_INIT_SUBCLASS 1
  #endif
  #ifndef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT 1
  #endif
  #ifndef CYTHON_USE_MODULE_STATE
    #define CYTHON_USE_MODULE_STATE 0
  #endif
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE 1
  #endif
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 1
  #endif
  #ifndef CYTHON_USE_FREELISTS
    #define CYTHON_USE_FREELISTS 0
  #endif
#else
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_CPYTHON 1
  #define CYTHON_COMPILING_IN_LIMITED_API 0
  #define CYTHON_COMPILING_IN_GRAAL 0
  #define CYTHON_COMPILING_IN_NOGIL 0
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #ifndef CYTHON_USE_TYPE_SPECS
    #define CYTHON_USE_TYPE_SPECS 0
  #endif
  #ifndef CYTHON_USE_PYTYPE_LOOKUP
    #define CYTHON_USE_PYTYPE_LOOKUP 1
  #endif
  #if PY_MAJOR_VERSION < 3
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #ifndef CYTHON_USE_PYLONG_INTERNALS
    #define CYTHON_USE_PYLONG_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_PYLIST_INTERNALS
    #define CYTHON_USE_PYLIST_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #if PY_VERSION_HEX < 0x030300F0 || PY_VERSION_HEX >= 0x030B00A2
    #undef CYTHON_USE_UNICODE_WRITER
    #define CYTHON_USE_UNICODE_WRITER 0
  #elif !defined(CYTHON_USE_UNICODE_WRITER)
    #define CYTHON_USE_UNICODE_WRITER 1
  #endif
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #ifndef CYTHON_FAST_THREAD_STATE
    #define CYTHON_FAST_THREAD_STATE 1
  #endif
  #ifndef CYTHON_FAST_GIL
    #define CYTHON_FAST_GIL (PY_MAJOR_VERSION < 3 || PY_VERSION_HEX >= 0x03060000 && PY_VERSION_HEX < 0x030C00A6)
  #endif
  #ifndef CYTHON_METH_FASTCALL
    #define CYTHON_METH_FASTCALL (PY_VERSION_HEX >= 0x030700A1)
  #endif
  #ifndef CYTHON_FAST_PYCALL
    #define CYTHON_FAST_PYCALL 1
  #endif
  #ifndef CYTHON_PEP487_INIT_SUBCLASS
    #define CYTHON_PEP487_INIT_SUBCLASS 1
  #endif
  #if PY_VERSION_HEX < 0x03050000
    #undef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #elif !defined(CYTHON_PEP489_MULTI_PHASE_INIT)
    #define CYTHON_PEP489_MULTI_PHASE_INIT 1
  #endif
  #ifndef CYTHON_USE_MODULE_STATE
    #define CYTHON_USE_MODULE_STATE 0
  #endif
  #if PY_VERSION_HEX < 0x030400a1
    #undef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE 0
  #elif !defined(CYTHON_USE_TP_FINALIZE)
    #define CYTHON_USE_TP_FINALIZE 1
  #endif
  #if PY_VERSION_HEX < 0x030600B1
    #undef CYTHON_USE_DICT_VERSIONS
    #define CYTHON_USE_DICT_VERSIONS 0
  #elif !defined(CYTHON_USE_DICT_VERSIONS)
    #define CYTHON_USE_DICT_VERSIONS  (PY_VERSION_HEX < 0x030C00A5)
  #endif
  #if PY_VERSION_HEX < 0x030700A3
    #undef CYTHON_USE_EXC_INFO_STACK
    #define CYTHON_USE_EXC_INFO_STACK 0
  #elif !defined(CYTHON_USE_EXC_INFO_STACK)
    #define CYTHON_USE_EXC_INFO_STACK 1
  #endif
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 1
  #endif
  #ifndef CYTHON_USE_FREELISTS
    #define CYTHON_USE_FREELISTS 1
  #endif
#endif
#if !defined(CYTHON_FAST_PYCCALL)
#define CYTHON_FAST_PYCCALL  (CYTHON_FAST_PYCALL && PY_VERSION_HEX >= 0x030600B1)
#endif
#if !defined(CYTHON_VECTORCALL)
#define CYTHON_VECTORCALL  (CYTHON_FAST_PYCCALL && PY_VERSION_HEX >= 0x030800B1)
#endif
#define CYTHON_BACKPORT_VECTORCALL (CYTHON_METH_FASTCALL && PY_VERSION_HEX < 0x030800B1)
#if CYTHON_USE_PYLONG_INTERNALS
  #if PY_MAJOR_VERSION < 3
    #include "longintrepr.h"
  #endif
  #undef SHIFT
  #undef BASE
  #undef MASK
  #ifdef SIZEOF_VOID_P
    enum { __pyx_check_sizeof_voidp = 1 / (int)(SIZEOF_VOID_P == sizeof(void*)) };
  #endif
#endif
#ifndef __has_attribute
  #define __has_attribute(x) 0
#endif
#ifndef __has_cpp_attribute
  #define __has_cpp_attribute(x) 0
#endif
#ifndef CYTHON_RESTRICT
  #if defined(__GNUC__)
    #define CYTHON_RESTRICT __restrict__
  #elif defined(_MSC_VER) && _MSC_VER >= 1400
    #define CYTHON_RESTRICT __restrict
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_RESTRICT restrict
  #else
    #define CYTHON_RESTRICT
  #endif
#endif
#ifndef CYTHON_UNUSED
  #if defined(__cplusplus)
    /* for clang __has_cpp_attribute(maybe_unused) is true even before C++17
     * but leads to warnings with -pedantic, since it is a C++17 feature */
    #if ((defined(_MSVC_LANG) && _MSVC_LANG >= 201703L) || __cplusplus >= 201703L)
      #if __has_cpp_attribute(maybe_unused)
        #define CYTHON_UNUSED [[maybe_unused]]
      #endif
    #endif
  #endif
#endif
#ifndef CYTHON_UNUSED
# if defined(__GNUC__)
#   if !(defined(__cplusplus)) || (__GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4))
#     define CYTHON_UNUSED __attribute__ ((__unused__))
#   else
#     define CYTHON_UNUSED
#   endif
# elif defined(__ICC) || (defined(__INTEL_COMPILER) && !defined(_MSC_VER))
#   define CYTHON_UNUSED __attribute__ ((__unused__))
# else
#   define CYTHON_UNUSED
# endif
#endif
#ifndef CYTHON_UNUSED_VAR
#  if defined(__cplusplus)
     template<class T> void CYTHON_UNUSED_VAR( const T& ) { }
#  else
#    define CYTHON_UNUSED_VAR(x) (void)(x)
#  endif
#endif
#ifndef CYTHON_MAYBE_UNUSED_VAR
  #define CYTHON_MAYBE_UNUSED_VAR(x) CYTHON_UNUSED_VAR(x)
#endif
#ifndef CYTHON_NCP_UNUSED
# if CYTHON_COMPILING_IN_CPYTHON
#  define CYTHON_NCP_UNUSED
# else
#  define CYTHON_NCP_UNUSED CYTHON_UNUSED
# endif
#endif
#ifndef CYTHON_USE_CPP_STD_MOVE
  #if defined(__cplusplus) && (\
    __cplusplus >= 201103L || (defined(_MSC_VER) && _MSC_VER >= 1600))
    #define CYTHON_USE_CPP_STD_MOVE 1
  #else
    #define CYTHON_USE_CPP_STD_MOVE 0
  #endif
#endif
#define __Pyx_void_to_None(void_result) ((void)(void_result), Py_INCREF(Py_None), Py_None)
#ifdef _MSC_VER
    #ifndef _MSC_STDINT_H_
        #if _MSC_VER < 1300
            typedef unsigned char     uint8_t;
            typedef unsigned short    uint16_t;
            typedef unsigned int      uint32_t;
        #else
            typedef unsigned __int8   uint8_t;
            typedef unsigned __int16  uint16_t;
            typedef unsigned __int32  uint32_t;
        #endif
    #endif
    #if _MSC_VER < 1300
        #ifdef _WIN64
            typedef unsigned long long  __pyx_uintptr_t;
        #else
            typedef unsigned int        __pyx_uintptr_t;
        #endif
    #else
        #ifdef _WIN64
            typedef unsigned __int64    __pyx_uintptr_t;
        #else
            typedef unsigned __int32    __pyx_uintptr_t;
        #endif
    #endif
#else
    #include <stdint.h>
    typedef uintptr_t  __pyx_uintptr_t;
#endif
#ifndef CYTHON_FALLTHROUGH
  #if defined(__cplusplus)
    /* for clang __has_cpp_attribute(fallthrough) is true even before C++17
     * but leads to warnings with -pedantic, since it is a C++17 feature */
    #if ((defined(_MSVC_LANG) && _MSVC_LANG >= 201703L) || __cplusplus >= 201703L)
      #if __has_cpp_attribute(fallthrough)
        #define CYTHON_FALLTHROUGH [[fallthrough]]
      #endif
    #endif
    #ifndef CYTHON_FALLTHROUGH
      #if __has_cpp_attribute(clang::fallthrough)
        #define CYTHON_FALLTHROUGH [[clang::fallthrough]]
      #elif __has_cpp_attribute(gnu::fallthrough)
        #define CYTHON_FALLTHROUGH [[gnu::fallthrough]]
      #endif
    #endif
  #endif
  #ifndef CYTHON_FALLTHROUGH
    #if __has_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH __attribute__((fallthrough))
    #else
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
  #if defined(__clang__) && defined(__apple_build_version__)
    #if __apple_build_version__ < 7000000
      #undef  CYTHON_FALLTHROUGH
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
#endif
#ifdef __cplusplus
  template <typename T>
  struct __PYX_IS_UNSIGNED_IMPL {static const bool value = T(0) < T(-1);};
  #define __PYX_IS_UNSIGNED(type) (__PYX_IS_UNSIGNED_IMPL<type>::value)
#else
  #define __PYX_IS_UNSIGNED(type) (((type)-1) > 0)
#endif
#if CYTHON_COMPILING_IN_PYPY == 1
  #define __PYX_NEED_TP_PRINT_SLOT  (PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x030A0000)
#else
  #define __PYX_NEED_TP_PRINT_SLOT  (PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000)
#endif
#define __PYX_REINTERPRET_FUNCION(func_pointer, other_pointer) ((func_pointer)(void(*)(void))(other_pointer))

#ifndef CYTHON_INLINE
  #if defined(__clang__)
    #define CYTHON_INLINE __inline__ __attribute__ ((__unused__))
  #elif defined(__GNUC__)
    #define CYTHON_INLINE __inline__
  #elif defined(_MSC_VER)
    #define CYTHON_INLINE __inline
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_INLINE inline
  #else
    #define CYTHON_INLINE
  #endif
#endif

#define __PYX_BUILD_PY_SSIZE_T "n"
#define CYTHON_FORMAT_SSIZE_T "z"
#if PY_MAJOR_VERSION < 3
  #define __Pyx_BUILTIN_MODULE_NAME "__builtin__"
  #define __Pyx_DefaultClassType PyClass_Type
  #define __Pyx_PyCode_New(a, p, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a+k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#else
  #define __Pyx_BUILTIN_MODULE_NAME "builtins"
  #define __Pyx_DefaultClassType PyType_Type
#if CYTHON_COMPILING_IN_LIMITED_API
    static CYTHON_INLINE PyObject* __Pyx_PyCode_New(int a, int p, int k, int l, int s, int f,
                                                    PyObject *code, PyObject *c, PyObject* n, PyObject *v,
                                                    PyObject *fv, PyObject *cell, PyObject* fn,
                                                    PyObject *name, int fline, PyObject *lnos) {
        PyObject *exception_table = NULL;
        PyObject *types_module=NULL, *code_type=NULL, *result=NULL;
        #if __PYX_LIMITED_VERSION_HEX < 0x030B0000
        PyObject *version_info;
        PyObject *py_minor_version = NULL;
        #endif
        long minor_version = 0;
        PyObject *type, *value, *traceback;
        PyErr_Fetch(&type, &value, &traceback);
        #if __PYX_LIMITED_VERSION_HEX >= 0x030B0000
        minor_version = 11;
        #else
        if (!(version_info = PySys_GetObject("version_info"))) goto end;
        if (!(py_minor_version = PySequence_GetItem(version_info, 1))) goto end;
        minor_version = PyLong_AsLong(py_minor_version);
        Py_DECREF(py_minor_version);
        if (minor_version == -1 && PyErr_Occurred()) goto end;
        #endif
        if (!(types_module = PyImport_ImportModule("types"))) goto end;
        if (!(code_type = PyObject_GetAttrString(types_module, "CodeType"))) goto end;
        if (minor_version <= 7) {
            (void)p;
            result = PyObject_CallFunction(code_type, "iiiiiOOOOOOiOO", a, k, l, s, f, code,
                          c, n, v, fn, name, fline, lnos, fv, cell);
        } else if (minor_version <= 10) {
            result = PyObject_CallFunction(code_type, "iiiiiiOOOOOOiOO", a,p, k, l, s, f, code,
                          c, n, v, fn, name, fline, lnos, fv, cell);
        } else {
            if (!(exception_table = PyBytes_FromStringAndSize(NULL, 0))) goto end;
            result = PyObject_CallFunction(code_type, "iiiiiiOOOOOOOiOO", a,p, k, l, s, f, code,
                          c, n, v, fn, name, name, fline, lnos, exception_table, fv, cell);
        }
    end:
        Py_XDECREF(code_type);
        Py_XDECREF(exception_table);
        Py_XDECREF(types_module);
        if (type) {
            PyErr_Restore(type, value, traceback);
        }
        return result;
    }
    #ifndef CO_OPTIMIZED
    #define CO_OPTIMIZED 0x0001
    #endif
    #ifndef CO_NEWLOCALS
    #define CO_NEWLOCALS 0x0002
    #endif
    #ifndef CO_VARARGS
    #define CO_VARARGS 0x0004
    #endif
    #ifndef CO_VARKEYWORDS
    #define CO_VARKEYWORDS 0x0008
    #endif
    #ifndef CO_ASYNC_GENERATOR
    #define CO_ASYNC_GENERATOR 0x0200
    #endif
    #ifndef CO_GENERATOR
    #define CO_GENERATOR 0x0020
    #endif
    #ifndef CO_COROUTINE
    #define CO_COROUTINE 0x0080
    #endif
#elif PY_VERSION_HEX >= 0x030B0000
  static CYTHON_INLINE PyCodeObject* __Pyx_PyCode_New(int a, int p, int k, int l, int s, int f,
                                                    PyObject *code, PyObject *c, PyObject* n, PyObject *v,
                                                    PyObject *fv, PyObject *cell, PyObject* fn,
                                                    PyObject *name, int fline, PyObject *lnos) {
    PyCodeObject *result;
    PyObject *empty_bytes = PyBytes_FromStringAndSize("", 0);
    if (!empty_bytes) return NULL;
    result =
      #if PY_VERSION_HEX >= 0x030C0000
        PyUnstable_Code_NewWithPosOnlyArgs
      #else
        PyCode_NewWithPosOnlyArgs
      #endif
        (a, p, k, l, s, f, code, c, n, v, fv, cell, fn, name, name, fline, lnos, empty_bytes);
    Py_DECREF(empty_bytes);
    return result;
  }
#elif PY_VERSION_HEX >= 0x030800B2 && !CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyCode_New(a, p, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_NewWithPosOnlyArgs(a, p, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#else
  #define __Pyx_PyCode_New(a, p, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#endif
#endif
#if PY_VERSION_HEX >= 0x030900A4 || defined(Py_IS_TYPE)
  #define __Pyx_IS_TYPE(ob, type) Py_IS_TYPE(ob, type)
#else
  #define __Pyx_IS_TYPE(ob, type) (((const PyObject*)ob)->ob_type == (type))
#endif
#if PY_VERSION_HEX >= 0x030A00B1 || defined(Py_Is)
  #define __Pyx_Py_Is(x, y)  Py_Is(x, y)
#else
  #define __Pyx_Py_Is(x, y) ((x) == (y))
#endif
#if PY_VERSION_HEX >= 0x030A00B1 || defined(Py_IsNone)
  #define __Pyx_Py_IsNone(ob) Py_IsNone(ob)
#else
  #define __Pyx_Py_IsNone(ob) __Pyx_Py_Is((ob), Py_None)
#endif
#if PY_VERSION_HEX >= 0x030A00B1 || defined(Py_IsTrue)
  #define __Pyx_Py_IsTrue(ob) Py_IsTrue(ob)
#else
  #define __Pyx_Py_IsTrue(ob) __Pyx_Py_Is((ob), Py_True)
#endif
#if PY_VERSION_HEX >= 0x030A00B1 || defined(Py_IsFalse)
  #define __Pyx_Py_IsFalse(ob) Py_IsFalse(ob)
#else
  #define __Pyx_Py_IsFalse(ob) __Pyx_Py_Is((ob), Py_False)
#endif
#define __Pyx_NoneAsNull(obj)  (__Pyx_Py_IsNone(obj) ? NULL : (obj))
#if PY_VERSION_HEX >= 0x030900F0 && !CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyObject_GC_IsFinalized(o) PyObject_GC_IsFinalized(o)
#else
  #define __Pyx_PyObject_GC_IsFinalized(o) _PyGC_FINALIZED(o)
#endif
#ifndef CO_COROUTINE
  #define CO_COROUTINE 0x80
#endif
#ifndef CO_ASYNC_GENERATOR
  #define CO_ASYNC_GENERATOR 0x200
#endif
#ifndef Py_TPFLAGS_CHECKTYPES
  #define Py_TPFLAGS_CHECKTYPES 0
#endif
#ifndef Py_TPFLAGS_HAVE_INDEX
  #define Py_TPFLAGS_HAVE_INDEX 0
#endif
#ifndef Py_TPFLAGS_HAVE_NEWBUFFER
  #define Py_TPFLAGS_HAVE_NEWBUFFER 0
#endif
#ifndef Py_TPFLAGS_HAVE_FINALIZE
  #define Py_TPFLAGS_HAVE_FINALIZE 0
#endif
#ifndef Py_TPFLAGS_SEQUENCE
  #define Py_TPFLAGS_SEQUENCE 0
#endif
#ifndef Py_TPFLAGS_MAPPING
  #define Py_TPFLAGS_MAPPING 0
#endif
#ifndef METH_STACKLESS
  #define METH_STACKLESS 0
#endif
#if PY_VERSION_HEX <= 0x030700A3 || !defined(METH_FASTCALL)
  #ifndef METH_FASTCALL
     #define METH_FASTCALL 0x80
  #endif
  typedef PyObject *(*__Pyx_PyCFunctionFast) (PyObject *self, PyObject *const *args, Py_ssize_t nargs);
  typedef PyObject *(*__Pyx_PyCFunctionFastWithKeywords) (PyObject *self, PyObject *const *args,
                                                          Py_ssize_t nargs, PyObject *kwnames);
#else
  #if PY_VERSION_HEX >= 0x030d00A4
  #  define __Pyx_PyCFunctionFast PyCFunctionFast
  #  define __Pyx_PyCFunctionFastWithKeywords PyCFunctionFastWithKeywords
  #else
  #  define __Pyx_PyCFunctionFast _PyCFunctionFast
  #  define __Pyx_PyCFunctionFastWithKeywords _PyCFunctionFastWithKeywords
  #endif
#endif
#if CYTHON_METH_FASTCALL
  #define __Pyx_METH_FASTCALL METH_FASTCALL
  #define __Pyx_PyCFunction_FastCall __Pyx_PyCFunctionFast
  #define __Pyx_PyCFunction_FastCallWithKeywords __Pyx_PyCFunctionFastWithKeywords
#else
  #define __Pyx_METH_FASTCALL METH_VARARGS
  #define __Pyx_PyCFunction_FastCall PyCFunction
  #define __Pyx_PyCFunction_FastCallWithKeywords PyCFunctionWithKeywords
#endif
#if CYTHON_VECTORCALL
  #define __pyx_vectorcallfunc vectorcallfunc
  #define __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET  PY_VECTORCALL_ARGUMENTS_OFFSET
  #define __Pyx_PyVectorcall_NARGS(n)  PyVectorcall_NARGS((size_t)(n))
#elif CYTHON_BACKPORT_VECTORCALL
  typedef PyObject *(*__pyx_vectorcallfunc)(PyObject *callable, PyObject *const *args,
                                            size_t nargsf, PyObject *kwnames);
  #define __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET  ((size_t)1 << (8 * sizeof(size_t) - 1))
  #define __Pyx_PyVectorcall_NARGS(n)  ((Py_ssize_t)(((size_t)(n)) & ~__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET))
#else
  #define __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET  0
  #define __Pyx_PyVectorcall_NARGS(n)  ((Py_ssize_t)(n))
#endif
#if PY_MAJOR_VERSION >= 0x030900B1
#define __Pyx_PyCFunction_CheckExact(func)  PyCFunction_CheckExact(func)
#else
#define __Pyx_PyCFunction_CheckExact(func)  PyCFunction_Check(func)
#endif
#define __Pyx_CyOrPyCFunction_Check(func)  PyCFunction_Check(func)
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_CyOrPyCFunction_GET_FUNCTION(func)  (((PyCFunctionObject*)(func))->m_ml->ml_meth)
#elif !CYTHON_COMPILING_IN_LIMITED_API
#define __Pyx_CyOrPyCFunction_GET_FUNCTION(func)  PyCFunction_GET_FUNCTION(func)
#endif
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_CyOrPyCFunction_GET_FLAGS(func)  (((PyCFunctionObject*)(func))->m_ml->ml_flags)
static CYTHON_INLINE PyObject* __Pyx_CyOrPyCFunction_GET_SELF(PyObject *func) {
    return (__Pyx_CyOrPyCFunction_GET_FLAGS(func) & METH_STATIC) ? NULL : ((PyCFunctionObject*)func)->m_self;
}
#endif
static CYTHON_INLINE int __Pyx__IsSameCFunction(PyObject *func, void *cfunc) {
#if CYTHON_COMPILING_IN_LIMITED_API
    return PyCFunction_Check(func) && PyCFunction_GetFunction(func) == (PyCFunction) cfunc;
#else
    return PyCFunction_Check(func) && PyCFunction_GET_FUNCTION(func) == (PyCFunction) cfunc;
#endif
}
#define __Pyx_IsSameCFunction(func, cfunc)   __Pyx__IsSameCFunction(func, cfunc)
#if __PYX_LIMITED_VERSION_HEX < 0x030900B1
  #define __Pyx_PyType_FromModuleAndSpec(m, s, b)  ((void)m, PyType_FromSpecWithBases(s, b))
  typedef PyObject *(*__Pyx_PyCMethod)(PyObject *, PyTypeObject *, PyObject *const *, size_t, PyObject *);
#else
  #define __Pyx_PyType_FromModuleAndSpec(m, s, b)  PyType_FromModuleAndSpec(m, s, b)
  #define __Pyx_PyCMethod  PyCMethod
#endif
#ifndef METH_METHOD
  #define METH_METHOD 0x200
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Malloc)
  #define PyObject_Malloc(s)   PyMem_Malloc(s)
  #define PyObject_Free(p)     PyMem_Free(p)
  #define PyObject_Realloc(p)  PyMem_Realloc(p)
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno)
#else
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno)  (frame)->f_lineno = (lineno)
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
  #define __Pyx_PyThreadState_Current PyThreadState_Get()
#elif !CYTHON_FAST_THREAD_STATE
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#elif PY_VERSION_HEX >= 0x030d00A1
  #define __Pyx_PyThreadState_Current PyThreadState_GetUnchecked()
#elif PY_VERSION_HEX >= 0x03060000
  #define __Pyx_PyThreadState_Current _PyThreadState_UncheckedGet()
#elif PY_VERSION_HEX >= 0x03000000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#else
  #define __Pyx_PyThreadState_Current _PyThreadState_Current
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
static CYTHON_INLINE void *__Pyx_PyModule_GetState(PyObject *op)
{
    void *result;
    result = PyModule_GetState(op);
    if (!result)
        Py_FatalError("Couldn't find the module state");
    return result;
}
#endif
#define __Pyx_PyObject_GetSlot(obj, name, func_ctype)  __Pyx_PyType_GetSlot(Py_TYPE(obj), name, func_ctype)
#if CYTHON_COMPILING_IN_LIMITED_API
  #define __Pyx_PyType_GetSlot(type, name, func_ctype)  ((func_ctype) PyType_GetSlot((type), Py_##name))
#else
  #define __Pyx_PyType_GetSlot(type, name, func_ctype)  ((type)->name)
#endif
#if PY_VERSION_HEX < 0x030700A2 && !defined(PyThread_tss_create) && !defined(Py_tss_NEEDS_INIT)
#include "pythread.h"
#define Py_tss_NEEDS_INIT 0
typedef int Py_tss_t;
static CYTHON_INLINE int PyThread_tss_create(Py_tss_t *key) {
  *key = PyThread_create_key();
  return 0;
}
static CYTHON_INLINE Py_tss_t * PyThread_tss_alloc(void) {
  Py_tss_t *key = (Py_tss_t *)PyObject_Malloc(sizeof(Py_tss_t));
  *key = Py_tss_NEEDS_INIT;
  return key;
}
static CYTHON_INLINE void PyThread_tss_free(Py_tss_t *key) {
  PyObject_Free(key);
}
static CYTHON_INLINE int PyThread_tss_is_created(Py_tss_t *key) {
  return *key != Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE void PyThread_tss_delete(Py_tss_t *key) {
  PyThread_delete_key(*key);
  *key = Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE int PyThread_tss_set(Py_tss_t *key, void *value) {
  return PyThread_set_key_value(*key, value);
}
static CYTHON_INLINE void * PyThread_tss_get(Py_tss_t *key) {
  return PyThread_get_key_value(*key);
}
#endif
#if PY_MAJOR_VERSION < 3
    #if CYTHON_COMPILING_IN_PYPY
        #if PYPY_VERSION_NUM < 0x07030600
            #if defined(__cplusplus) && __cplusplus >= 201402L
                [[deprecated("`with nogil:` inside a nogil function will not release the GIL in PyPy2 < 7.3.6")]]
            #elif defined(__GNUC__) || defined(__clang__)
                __attribute__ ((__deprecated__("`with nogil:` inside a nogil function will not release the GIL in PyPy2 < 7.3.6")))
            #elif defined(_MSC_VER)
                __declspec(deprecated("`with nogil:` inside a nogil function will not release the GIL in PyPy2 < 7.3.6"))
            #endif
            static CYTHON_INLINE int PyGILState_Check(void) {
                return 0;
            }
        #else  // PYPY_VERSION_NUM < 0x07030600
        #endif  // PYPY_VERSION_NUM < 0x07030600
    #else
        static CYTHON_INLINE int PyGILState_Check(void) {
            PyThreadState * tstate = _PyThreadState_Current;
            return tstate && (tstate == PyGILState_GetThisThreadState());
        }
    #endif
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030d0000 || defined(_PyDict_NewPresized)
#define __Pyx_PyDict_NewPresized(n)  ((n <= 8) ? PyDict_New() : _PyDict_NewPresized(n))
#else
#define __Pyx_PyDict_NewPresized(n)  PyDict_New()
#endif
#if PY_MAJOR_VERSION >= 3 || CYTHON_FUTURE_DIVISION
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_TrueDivide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceTrueDivide(x,y)
#else
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_Divide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceDivide(x,y)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX > 0x030600B4 && PY_VERSION_HEX < 0x030d0000 && CYTHON_USE_UNICODE_INTERNALS
#define __Pyx_PyDict_GetItemStrWithError(dict, name)  _PyDict_GetItem_KnownHash(dict, name, ((PyASCIIObject *) name)->hash)
static CYTHON_INLINE PyObject * __Pyx_PyDict_GetItemStr(PyObject *dict, PyObject *name) {
    PyObject *res = __Pyx_PyDict_GetItemStrWithError(dict, name);
    if (res == NULL) PyErr_Clear();
    return res;
}
#elif PY_MAJOR_VERSION >= 3 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07020000)
#define __Pyx_PyDict_GetItemStrWithError  PyDict_GetItemWithError
#define __Pyx_PyDict_GetItemStr           PyDict_GetItem
#else
static CYTHON_INLINE PyObject * __Pyx_PyDict_GetItemStrWithError(PyObject *dict, PyObject *name) {
#if CYTHON_COMPILING_IN_PYPY
    return PyDict_GetItem(dict, name);
#else
    PyDictEntry *ep;
    PyDictObject *mp = (PyDictObject*) dict;
    long hash = ((PyStringObject *) name)->ob_shash;
    assert(hash != -1);
    ep = (mp->ma_lookup)(mp, name, hash);
    if (ep == NULL) {
        return NULL;
    }
    return ep->me_value;
#endif
}
#define __Pyx_PyDict_GetItemStr           PyDict_GetItem
#endif
#if CYTHON_USE_TYPE_SLOTS
  #define __Pyx_PyType_GetFlags(tp)   (((PyTypeObject *)tp)->tp_flags)
  #define __Pyx_PyType_HasFeature(type, feature)  ((__Pyx_PyType_GetFlags(type) & (feature)) != 0)
  #define __Pyx_PyObject_GetIterNextFunc(obj)  (Py_TYPE(obj)->tp_iternext)
#else
  #define __Pyx_PyType_GetFlags(tp)   (PyType_GetFlags((PyTypeObject *)tp))
  #define __Pyx_PyType_HasFeature(type, feature)  PyType_HasFeature(type, feature)
  #define __Pyx_PyObject_GetIterNextFunc(obj)  PyIter_Next
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
  #define __Pyx_SetItemOnTypeDict(tp, k, v) PyObject_GenericSetAttr((PyObject*)tp, k, v)
#else
  #define __Pyx_SetItemOnTypeDict(tp, k, v) PyDict_SetItem(tp->tp_dict, k, v)
#endif
#if CYTHON_USE_TYPE_SPECS && PY_VERSION_HEX >= 0x03080000
#define __Pyx_PyHeapTypeObject_GC_Del(obj)  {\
    PyTypeObject *type = Py_TYPE((PyObject*)obj);\
    assert(__Pyx_PyType_HasFeature(type, Py_TPFLAGS_HEAPTYPE));\
    PyObject_GC_Del(obj);\
    Py_DECREF(type);\
}
#else
#define __Pyx_PyHeapTypeObject_GC_Del(obj)  PyObject_GC_Del(obj)
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
  #define CYTHON_PEP393_ENABLED 1
  #define __Pyx_PyUnicode_READY(op)       (0)
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GetLength(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_ReadChar(u, i)
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   ((void)u, 1114111U)
  #define __Pyx_PyUnicode_KIND(u)         ((void)u, (0))
  #define __Pyx_PyUnicode_DATA(u)         ((void*)u)
  #define __Pyx_PyUnicode_READ(k, d, i)   ((void)k, PyUnicode_ReadChar((PyObject*)(d), i))
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GetLength(u))
#elif PY_VERSION_HEX > 0x03030000 && defined(PyUnicode_KIND)
  #define CYTHON_PEP393_ENABLED 1
  #if PY_VERSION_HEX >= 0x030C0000
    #define __Pyx_PyUnicode_READY(op)       (0)
  #else
    #define __Pyx_PyUnicode_READY(op)       (likely(PyUnicode_IS_READY(op)) ?\
                                                0 : _PyUnicode_Ready((PyObject *)(op)))
  #endif
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_LENGTH(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_READ_CHAR(u, i)
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   PyUnicode_MAX_CHAR_VALUE(u)
  #define __Pyx_PyUnicode_KIND(u)         ((int)PyUnicode_KIND(u))
  #define __Pyx_PyUnicode_DATA(u)         PyUnicode_DATA(u)
  #define __Pyx_PyUnicode_READ(k, d, i)   PyUnicode_READ(k, d, i)
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  PyUnicode_WRITE(k, d, i, (Py_UCS4) ch)
  #if PY_VERSION_HEX >= 0x030C0000
    #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_LENGTH(u))
  #else
    #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03090000
    #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : ((PyCompactUnicodeObject *)(u))->wstr_length))
    #else
    #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : PyUnicode_GET_SIZE(u)))
    #endif
  #endif
#else
  #define CYTHON_PEP393_ENABLED 0
  #define PyUnicode_1BYTE_KIND  1
  #define PyUnicode_2BYTE_KIND  2
  #define PyUnicode_4BYTE_KIND  4
  #define __Pyx_PyUnicode_READY(op)       (0)
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_SIZE(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) ((Py_UCS4)(PyUnicode_AS_UNICODE(u)[i]))
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   ((sizeof(Py_UNICODE) == 2) ? 65535U : 1114111U)
  #define __Pyx_PyUnicode_KIND(u)         ((int)sizeof(Py_UNICODE))
  #define __Pyx_PyUnicode_DATA(u)         ((void*)PyUnicode_AS_UNICODE(u))
  #define __Pyx_PyUnicode_READ(k, d, i)   ((void)(k), (Py_UCS4)(((Py_UNICODE*)d)[i]))
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  (((void)(k)), ((Py_UNICODE*)d)[i] = (Py_UNICODE) ch)
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_SIZE(u))
#endif
#if CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyUnicode_Concat(a, b)      PyNumber_Add(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  PyNumber_Add(a, b)
#else
  #define __Pyx_PyUnicode_Concat(a, b)      PyUnicode_Concat(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  ((unlikely((a) == Py_None) || unlikely((b) == Py_None)) ?\
      PyNumber_Add(a, b) : __Pyx_PyUnicode_Concat(a, b))
#endif
#if CYTHON_COMPILING_IN_PYPY
  #if !defined(PyUnicode_DecodeUnicodeEscape)
    #define PyUnicode_DecodeUnicodeEscape(s, size, errors)  PyUnicode_Decode(s, size, "unicode_escape", errors)
  #endif
  #if !defined(PyUnicode_Contains) || (PY_MAJOR_VERSION == 2 && PYPY_VERSION_NUM < 0x07030500)
    #undef PyUnicode_Contains
    #define PyUnicode_Contains(u, s)  PySequence_Contains(u, s)
  #endif
  #if !defined(PyByteArray_Check)
    #define PyByteArray_Check(obj)  PyObject_TypeCheck(obj, &PyByteArray_Type)
  #endif
  #if !defined(PyObject_Format)
    #define PyObject_Format(obj, fmt)  PyObject_CallMethod(obj, "__format__", "O", fmt)
  #endif
#endif
#define __Pyx_PyString_FormatSafe(a, b)   ((unlikely((a) == Py_None || (PyString_Check(b) && !PyString_CheckExact(b)))) ? PyNumber_Remainder(a, b) : __Pyx_PyString_Format(a, b))
#define __Pyx_PyUnicode_FormatSafe(a, b)  ((unlikely((a) == Py_None || (PyUnicode_Check(b) && !PyUnicode_CheckExact(b)))) ? PyNumber_Remainder(a, b) : PyUnicode_Format(a, b))
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyString_Format(a, b)  PyUnicode_Format(a, b)
#else
  #define __Pyx_PyString_Format(a, b)  PyString_Format(a, b)
#endif
#if PY_MAJOR_VERSION < 3 && !defined(PyObject_ASCII)
  #define PyObject_ASCII(o)            PyObject_Repr(o)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBaseString_Type            PyUnicode_Type
  #define PyStringObject               PyUnicodeObject
  #define PyString_Type                PyUnicode_Type
  #define PyString_Check               PyUnicode_Check
  #define PyString_CheckExact          PyUnicode_CheckExact
#ifndef PyObject_Unicode
  #define PyObject_Unicode             PyObject_Str
#endif
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyBaseString_Check(obj) PyUnicode_Check(obj)
  #define __Pyx_PyBaseString_CheckExact(obj) PyUnicode_CheckExact(obj)
#else
  #define __Pyx_PyBaseString_Check(obj) (PyString_Check(obj) || PyUnicode_Check(obj))
  #define __Pyx_PyBaseString_CheckExact(obj) (PyString_CheckExact(obj) || PyUnicode_CheckExact(obj))
#endif
#if CYTHON_COMPILING_IN_CPYTHON
  #define __Pyx_PySequence_ListKeepNew(obj)\
    (likely(PyList_CheckExact(obj) && Py_REFCNT(obj) == 1) ? __Pyx_NewRef(obj) : PySequence_List(obj))
#else
  #define __Pyx_PySequence_ListKeepNew(obj)  PySequence_List(obj)
#endif
#ifndef PySet_CheckExact
  #define PySet_CheckExact(obj)        __Pyx_IS_TYPE(obj, &PySet_Type)
#endif
#if PY_VERSION_HEX >= 0x030900A4
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_SET_REFCNT(obj, refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SET_SIZE(obj, size)
#else
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_REFCNT(obj) = (refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SIZE(obj) = (size)
#endif
#if CYTHON_ASSUME_SAFE_MACROS
  #define __Pyx_PySequence_ITEM(o, i) PySequence_ITEM(o, i)
  #define __Pyx_PySequence_SIZE(seq)  Py_SIZE(seq)
  #define __Pyx_PyTuple_SET_ITEM(o, i, v) (PyTuple_SET_ITEM(o, i, v), (0))
  #define __Pyx_PyList_SET_ITEM(o, i, v) (PyList_SET_ITEM(o, i, v), (0))
  #define __Pyx_PyTuple_GET_SIZE(o) PyTuple_GET_SIZE(o)
  #define __Pyx_PyList_GET_SIZE(o) PyList_GET_SIZE(o)
  #define __Pyx_PySet_GET_SIZE(o) PySet_GET_SIZE(o)
  #define __Pyx_PyBytes_GET_SIZE(o) PyBytes_GET_SIZE(o)
  #define __Pyx_PyByteArray_GET_SIZE(o) PyByteArray_GET_SIZE(o)
#else
  #define __Pyx_PySequence_ITEM(o, i) PySequence_GetItem(o, i)
  #define __Pyx_PySequence_SIZE(seq)  PySequence_Size(seq)
  #define __Pyx_PyTuple_SET_ITEM(o, i, v) PyTuple_SetItem(o, i, v)
  #define __Pyx_PyList_SET_ITEM(o, i, v) PyList_SetItem(o, i, v)
  #define __Pyx_PyTuple_GET_SIZE(o) PyTuple_Size(o)
  #define __Pyx_PyList_GET_SIZE(o) PyList_Size(o)
  #define __Pyx_PySet_GET_SIZE(o) PySet_Size(o)
  #define __Pyx_PyBytes_GET_SIZE(o) PyBytes_Size(o)
  #define __Pyx_PyByteArray_GET_SIZE(o) PyByteArray_Size(o)
#endif
#if __PYX_LIMITED_VERSION_HEX >= 0x030d00A1
  #define __Pyx_PyImport_AddModuleRef(name) PyImport_AddModuleRef(name)
#else
  static CYTHON_INLINE PyObject *__Pyx_PyImport_AddModuleRef(const char *name) {
      PyObject *module = PyImport_AddModule(name);
      Py_XINCREF(module);
      return module;
  }
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyIntObject                  PyLongObject
  #define PyInt_Type                   PyLong_Type
  #define PyInt_Check(op)              PyLong_Check(op)
  #define PyInt_CheckExact(op)         PyLong_CheckExact(op)
  #define __Pyx_Py3Int_Check(op)       PyLong_Check(op)
  #define __Pyx_Py3Int_CheckExact(op)  PyLong_CheckExact(op)
  #define PyInt_FromString             PyLong_FromString
  #define PyInt_FromUnicode            PyLong_FromUnicode
  #define PyInt_FromLong               PyLong_FromLong
  #define PyInt_FromSize_t             PyLong_FromSize_t
  #define PyInt_FromSsize_t            PyLong_FromSsize_t
  #define PyInt_AsLong                 PyLong_AsLong
  #define PyInt_AS_LONG                PyLong_AS_LONG
  #define PyInt_AsSsize_t              PyLong_AsSsize_t
  #define PyInt_AsUnsignedLongMask     PyLong_AsUnsignedLongMask
  #define PyInt_AsUnsignedLongLongMask PyLong_AsUnsignedLongLongMask
  #define PyNumber_Int                 PyNumber_Long
#else
  #define __Pyx_Py3Int_Check(op)       (PyLong_Check(op) || PyInt_Check(op))
  #define __Pyx_Py3Int_CheckExact(op)  (PyLong_CheckExact(op) || PyInt_CheckExact(op))
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBoolObject                 PyLongObject
#endif
#if PY_MAJOR_VERSION >= 3 && CYTHON_COMPILING_IN_PYPY
  #ifndef PyUnicode_InternFromString
    #define PyUnicode_InternFromString(s) PyUnicode_FromString(s)
  #endif
#endif
#if PY_VERSION_HEX < 0x030200A4
  typedef long Py_hash_t;
  #define __Pyx_PyInt_FromHash_t PyInt_FromLong
  #define __Pyx_PyInt_AsHash_t   __Pyx_PyIndex_AsHash_t
#else
  #define __Pyx_PyInt_FromHash_t PyInt_FromSsize_t
  #define __Pyx_PyInt_AsHash_t   __Pyx_PyIndex_AsSsize_t
#endif
#if CYTHON_USE_ASYNC_SLOTS
  #if PY_VERSION_HEX >= 0x030500B1
    #define __Pyx_PyAsyncMethodsStruct PyAsyncMethods
    #define __Pyx_PyType_AsAsync(obj) (Py_TYPE(obj)->tp_as_async)
  #else
    #define __Pyx_PyType_AsAsync(obj) ((__Pyx_PyAsyncMethodsStruct*) (Py_TYPE(obj)->tp_reserved))
  #endif
#else
  #define __Pyx_PyType_AsAsync(obj) NULL
#endif
#ifndef __Pyx_PyAsyncMethodsStruct
    typedef struct {
        unaryfunc am_await;
        unaryfunc am_aiter;
        unaryfunc am_anext;
    } __Pyx_PyAsyncMethodsStruct;
#endif

#if defined(_WIN32) || defined(WIN32) || defined(MS_WINDOWS)
  #if !defined(_USE_MATH_DEFINES)
    #define _USE_MATH_DEFINES
  #endif
#endif
#include <math.h>
#ifdef NAN
#define __PYX_NAN() ((float) NAN)
#else
static CYTHON_INLINE float __PYX_NAN() {
  float value;
  memset(&value, 0xFF, sizeof(value));
  return value;
}
#endif
#if defined(__CYGWIN__) && defined(_LDBL_EQ_DBL)
#define __Pyx_truncl trunc
#else
#define __Pyx_truncl truncl
#endif

#define __PYX_MARK_ERR_POS(f_index, lineno) \
    { __pyx_filename = __pyx_f[f_index]; (void)__pyx_filename; __pyx_lineno = lineno; (void)__pyx_lineno; __pyx_clineno = __LINE__;  (void)__pyx_clineno; }
#define __PYX_ERR(f_index, lineno, Ln_error) \
    { __PYX_MARK_ERR_POS(f_index, lineno) goto Ln_error; }

#ifdef CYTHON_EXTERN_C
    #undef __PYX_EXTERN_C
    #define __PYX_EXTERN_C CYTHON_EXTERN_C
#elif defined(__PYX_EXTERN_C)
    #ifdef _MSC_VER
    #pragma message ("Please do not define the '__PYX_EXTERN_C' macro externally. Use 'CYTHON_EXTERN_C' instead.")
    #else
    #warning Please do not define the '__PYX_EXTERN_C' macro externally. Use 'CYTHON_EXTERN_C' instead.
    #endif
#else
  #ifdef __cplusplus
    #define __PYX_EXTERN_C extern "C"
  #else
    #define __PYX_EXTERN_C extern
  #endif
#endif

#define __PYX_HAVE__serializor__binary__ser
#define __PYX_HAVE_API__serializor__binary__ser
/* Early includes */
#include <limits.h>
#include <string.h>
#include <stdio.h>
#include <math.h>

    /* Using NumPy API declarations from "numpy/__init__.cython-30.pxd" */
    
#include "numpy/arrayobject.h"
#include "numpy/ndarrayobject.h"
#include "numpy/ndarraytypes.h"
#include "numpy/arrayscalars.h"
#include "numpy/ufuncobject.h"
#include <stddef.h>
#include "pythread.h"
#include "datetime.h"

    /* Backport for Python 2.x */
    #if PY_MAJOR_VERSION < 3
        #ifndef PyDateTime_DELTA_GET_DAYS
            #define PyDateTime_DELTA_GET_DAYS(o) (((PyDateTime_Delta*)o)->days)
        #endif
        #ifndef PyDateTime_DELTA_GET_SECONDS
            #define PyDateTime_DELTA_GET_SECONDS(o) (((PyDateTime_Delta*)o)->seconds)
        #endif
        #ifndef PyDateTime_DELTA_GET_MICROSECONDS
            #define PyDateTime_DELTA_GET_MICROSECONDS(o) (((PyDateTime_Delta*)o)->microseconds)
        #endif
    #endif

    /* Backport for Python < 3.6 */
    #if PY_VERSION_HEX < 0x030600a4
        #ifndef PyDateTime_TIME_GET_FOLD
            #define PyDateTime_TIME_GET_FOLD(o) ((void)(o), 0)
        #endif
        #ifndef PyDateTime_DATE_GET_FOLD
            #define PyDateTime_DATE_GET_FOLD(o) ((void)(o), 0)
        #endif
    #endif

    /* Backport for Python < 3.6 */
    #if PY_VERSION_HEX < 0x030600a4
        #define __Pyx_DateTime_DateTimeWithFold(year, month, day, hour, minute, second, microsecond, tz, fold)             ((void)(fold), PyDateTimeAPI->DateTime_FromDateAndTime(year, month, day, hour, minute, second,                 microsecond, tz, PyDateTimeAPI->DateTimeType))
        #define __Pyx_DateTime_TimeWithFold(hour, minute, second, microsecond, tz, fold)             ((void)(fold), PyDateTimeAPI->Time_FromTime(hour, minute, second, microsecond, tz, PyDateTimeAPI->TimeType))
    #else /* For Python 3.6+ so that we can pass tz */
        #define __Pyx_DateTime_DateTimeWithFold(year, month, day, hour, minute, second, microsecond, tz, fold)             PyDateTimeAPI->DateTime_FromDateAndTimeAndFold(year, month, day, hour, minute, second,                 microsecond, tz, fold, PyDateTimeAPI->DateTimeType)
        #define __Pyx_DateTime_TimeWithFold(hour, minute, second, microsecond, tz, fold)             PyDateTimeAPI->Time_FromTimeAndFold(hour, minute, second, microsecond, tz, fold, PyDateTimeAPI->TimeType)
    #endif

    /* Backport for Python < 3.7 */
    #if PY_VERSION_HEX < 0x030700b1
        #define __Pyx_TimeZone_UTC NULL
        #define __Pyx_TimeZone_FromOffsetAndName(offset, name) ((void)(offset), (void)(name), (PyObject*)NULL)
    #else
        #define __Pyx_TimeZone_UTC PyDateTime_TimeZone_UTC
        #define __Pyx_TimeZone_FromOffsetAndName(offset, name) PyTimeZone_FromOffsetAndName(offset, name)
    #endif

    /* Backport for Python < 3.10 */
    #if PY_VERSION_HEX < 0x030a00a1
        #ifndef PyDateTime_TIME_GET_TZINFO
            #define PyDateTime_TIME_GET_TZINFO(o)                 ((((PyDateTime_Time*)o)->hastzinfo) ? ((PyDateTime_Time*)o)->tzinfo : Py_None)
        #endif
        #ifndef PyDateTime_DATE_GET_TZINFO
            #define PyDateTime_DATE_GET_TZINFO(o)                 ((((PyDateTime_DateTime*)o)->hastzinfo) ? ((PyDateTime_DateTime*)o)->tzinfo : Py_None)
        #endif
    #endif
    
#ifdef _OPENMP
#include <omp.h>
#endif /* _OPENMP */

#if defined(PYREX_WITHOUT_ASSERTIONS) && !defined(CYTHON_WITHOUT_ASSERTIONS)
#define CYTHON_WITHOUT_ASSERTIONS
#endif

typedef struct {PyObject **p; const char *s; const Py_ssize_t n; const char* encoding;
                const char is_unicode; const char is_str; const char intern; } __Pyx_StringTabEntry;

#define __PYX_DEFAULT_STRING_ENCODING_IS_ASCII 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_UTF8 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT (PY_MAJOR_VERSION >= 3 && __PYX_DEFAULT_STRING_ENCODING_IS_UTF8)
#define __PYX_DEFAULT_STRING_ENCODING ""
#define __Pyx_PyObject_FromString __Pyx_PyBytes_FromString
#define __Pyx_PyObject_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#define __Pyx_uchar_cast(c) ((unsigned char)c)
#define __Pyx_long_cast(x) ((long)x)
#define __Pyx_fits_Py_ssize_t(v, type, is_signed)  (\
    (sizeof(type) < sizeof(Py_ssize_t))  ||\
    (sizeof(type) > sizeof(Py_ssize_t) &&\
          likely(v < (type)PY_SSIZE_T_MAX ||\
                 v == (type)PY_SSIZE_T_MAX)  &&\
          (!is_signed || likely(v > (type)PY_SSIZE_T_MIN ||\
                                v == (type)PY_SSIZE_T_MIN)))  ||\
    (sizeof(type) == sizeof(Py_ssize_t) &&\
          (is_signed || likely(v < (type)PY_SSIZE_T_MAX ||\
                               v == (type)PY_SSIZE_T_MAX)))  )
static CYTHON_INLINE int __Pyx_is_valid_index(Py_ssize_t i, Py_ssize_t limit) {
    return (size_t) i < (size_t) limit;
}
#if defined (__cplusplus) && __cplusplus >= 201103L
    #include <cstdlib>
    #define __Pyx_sst_abs(value) std::abs(value)
#elif SIZEOF_INT >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) abs(value)
#elif SIZEOF_LONG >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) labs(value)
#elif defined (_MSC_VER)
    #define __Pyx_sst_abs(value) ((Py_ssize_t)_abs64(value))
#elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define __Pyx_sst_abs(value) llabs(value)
#elif defined (__GNUC__)
    #define __Pyx_sst_abs(value) __builtin_llabs(value)
#else
    #define __Pyx_sst_abs(value) ((value<0) ? -value : value)
#endif
static CYTHON_INLINE Py_ssize_t __Pyx_ssize_strlen(const char *s);
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject*);
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject*, Py_ssize_t* length);
static CYTHON_INLINE PyObject* __Pyx_PyByteArray_FromString(const char*);
#define __Pyx_PyByteArray_FromStringAndSize(s, l) PyByteArray_FromStringAndSize((const char*)s, l)
#define __Pyx_PyBytes_FromString        PyBytes_FromString
#define __Pyx_PyBytes_FromStringAndSize PyBytes_FromStringAndSize
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char*);
#if PY_MAJOR_VERSION < 3
    #define __Pyx_PyStr_FromString        __Pyx_PyBytes_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#else
    #define __Pyx_PyStr_FromString        __Pyx_PyUnicode_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyUnicode_FromStringAndSize
#endif
#define __Pyx_PyBytes_AsWritableString(s)     ((char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableSString(s)    ((signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableUString(s)    ((unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsString(s)     ((const char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsSString(s)    ((const signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsUString(s)    ((const unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyObject_AsWritableString(s)    ((char*)(__pyx_uintptr_t) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableSString(s)    ((signed char*)(__pyx_uintptr_t) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableUString(s)    ((unsigned char*)(__pyx_uintptr_t) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsSString(s)    ((const signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsUString(s)    ((const unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_FromCString(s)  __Pyx_PyObject_FromString((const char*)s)
#define __Pyx_PyBytes_FromCString(s)   __Pyx_PyBytes_FromString((const char*)s)
#define __Pyx_PyByteArray_FromCString(s)   __Pyx_PyByteArray_FromString((const char*)s)
#define __Pyx_PyStr_FromCString(s)     __Pyx_PyStr_FromString((const char*)s)
#define __Pyx_PyUnicode_FromCString(s) __Pyx_PyUnicode_FromString((const char*)s)
#define __Pyx_PyUnicode_FromOrdinal(o)       PyUnicode_FromOrdinal((int)o)
#define __Pyx_PyUnicode_AsUnicode            PyUnicode_AsUnicode
#define __Pyx_NewRef(obj) (Py_INCREF(obj), obj)
#define __Pyx_Owned_Py_None(b) __Pyx_NewRef(Py_None)
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b);
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject*);
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject*);
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x);
#define __Pyx_PySequence_Tuple(obj)\
    (likely(PyTuple_CheckExact(obj)) ? __Pyx_NewRef(obj) : PySequence_Tuple(obj))
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject*);
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t);
static CYTHON_INLINE Py_hash_t __Pyx_PyIndex_AsHash_t(PyObject*);
#if CYTHON_ASSUME_SAFE_MACROS
#define __pyx_PyFloat_AsDouble(x) (PyFloat_CheckExact(x) ? PyFloat_AS_DOUBLE(x) : PyFloat_AsDouble(x))
#else
#define __pyx_PyFloat_AsDouble(x) PyFloat_AsDouble(x)
#endif
#define __pyx_PyFloat_AsFloat(x) ((float) __pyx_PyFloat_AsDouble(x))
#if PY_MAJOR_VERSION >= 3
#define __Pyx_PyNumber_Int(x) (PyLong_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Long(x))
#else
#define __Pyx_PyNumber_Int(x) (PyInt_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Int(x))
#endif
#if CYTHON_USE_PYLONG_INTERNALS
  #if PY_VERSION_HEX >= 0x030C00A7
  #ifndef _PyLong_SIGN_MASK
    #define _PyLong_SIGN_MASK 3
  #endif
  #ifndef _PyLong_NON_SIZE_BITS
    #define _PyLong_NON_SIZE_BITS 3
  #endif
  #define __Pyx_PyLong_Sign(x)  (((PyLongObject*)x)->long_value.lv_tag & _PyLong_SIGN_MASK)
  #define __Pyx_PyLong_IsNeg(x)  ((__Pyx_PyLong_Sign(x) & 2) != 0)
  #define __Pyx_PyLong_IsNonNeg(x)  (!__Pyx_PyLong_IsNeg(x))
  #define __Pyx_PyLong_IsZero(x)  (__Pyx_PyLong_Sign(x) & 1)
  #define __Pyx_PyLong_IsPos(x)  (__Pyx_PyLong_Sign(x) == 0)
  #define __Pyx_PyLong_CompactValueUnsigned(x)  (__Pyx_PyLong_Digits(x)[0])
  #define __Pyx_PyLong_DigitCount(x)  ((Py_ssize_t) (((PyLongObject*)x)->long_value.lv_tag >> _PyLong_NON_SIZE_BITS))
  #define __Pyx_PyLong_SignedDigitCount(x)\
        ((1 - (Py_ssize_t) __Pyx_PyLong_Sign(x)) * __Pyx_PyLong_DigitCount(x))
  #if defined(PyUnstable_Long_IsCompact) && defined(PyUnstable_Long_CompactValue)
    #define __Pyx_PyLong_IsCompact(x)     PyUnstable_Long_IsCompact((PyLongObject*) x)
    #define __Pyx_PyLong_CompactValue(x)  PyUnstable_Long_CompactValue((PyLongObject*) x)
  #else
    #define __Pyx_PyLong_IsCompact(x)     (((PyLongObject*)x)->long_value.lv_tag < (2 << _PyLong_NON_SIZE_BITS))
    #define __Pyx_PyLong_CompactValue(x)  ((1 - (Py_ssize_t) __Pyx_PyLong_Sign(x)) * (Py_ssize_t) __Pyx_PyLong_Digits(x)[0])
  #endif
  typedef Py_ssize_t  __Pyx_compact_pylong;
  typedef size_t  __Pyx_compact_upylong;
  #else
  #define __Pyx_PyLong_IsNeg(x)  (Py_SIZE(x) < 0)
  #define __Pyx_PyLong_IsNonNeg(x)  (Py_SIZE(x) >= 0)
  #define __Pyx_PyLong_IsZero(x)  (Py_SIZE(x) == 0)
  #define __Pyx_PyLong_IsPos(x)  (Py_SIZE(x) > 0)
  #define __Pyx_PyLong_CompactValueUnsigned(x)  ((Py_SIZE(x) == 0) ? 0 : __Pyx_PyLong_Digits(x)[0])
  #define __Pyx_PyLong_DigitCount(x)  __Pyx_sst_abs(Py_SIZE(x))
  #define __Pyx_PyLong_SignedDigitCount(x)  Py_SIZE(x)
  #define __Pyx_PyLong_IsCompact(x)  (Py_SIZE(x) == 0 || Py_SIZE(x) == 1 || Py_SIZE(x) == -1)
  #define __Pyx_PyLong_CompactValue(x)\
        ((Py_SIZE(x) == 0) ? (sdigit) 0 : ((Py_SIZE(x) < 0) ? -(sdigit)__Pyx_PyLong_Digits(x)[0] : (sdigit)__Pyx_PyLong_Digits(x)[0]))
  typedef sdigit  __Pyx_compact_pylong;
  typedef digit  __Pyx_compact_upylong;
  #endif
  #if PY_VERSION_HEX >= 0x030C00A5
  #define __Pyx_PyLong_Digits(x)  (((PyLongObject*)x)->long_value.ob_digit)
  #else
  #define __Pyx_PyLong_Digits(x)  (((PyLongObject*)x)->ob_digit)
  #endif
#endif
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
#include <string.h>
static int __Pyx_sys_getdefaultencoding_not_ascii;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    PyObject* ascii_chars_u = NULL;
    PyObject* ascii_chars_b = NULL;
    const char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    if (strcmp(default_encoding_c, "ascii") == 0) {
        __Pyx_sys_getdefaultencoding_not_ascii = 0;
    } else {
        char ascii_chars[128];
        int c;
        for (c = 0; c < 128; c++) {
            ascii_chars[c] = (char) c;
        }
        __Pyx_sys_getdefaultencoding_not_ascii = 1;
        ascii_chars_u = PyUnicode_DecodeASCII(ascii_chars, 128, NULL);
        if (!ascii_chars_u) goto bad;
        ascii_chars_b = PyUnicode_AsEncodedString(ascii_chars_u, default_encoding_c, NULL);
        if (!ascii_chars_b || !PyBytes_Check(ascii_chars_b) || memcmp(ascii_chars, PyBytes_AS_STRING(ascii_chars_b), 128) != 0) {
            PyErr_Format(
                PyExc_ValueError,
                "This module compiled with c_string_encoding=ascii, but default encoding '%.200s' is not a superset of ascii.",
                default_encoding_c);
            goto bad;
        }
        Py_DECREF(ascii_chars_u);
        Py_DECREF(ascii_chars_b);
    }
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    Py_XDECREF(ascii_chars_u);
    Py_XDECREF(ascii_chars_b);
    return -1;
}
#endif
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT && PY_MAJOR_VERSION >= 3
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_DecodeUTF8(c_str, size, NULL)
#else
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_Decode(c_str, size, __PYX_DEFAULT_STRING_ENCODING, NULL)
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
#include <string.h>
static char* __PYX_DEFAULT_STRING_ENCODING;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) (const char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    __PYX_DEFAULT_STRING_ENCODING = (char*) malloc(strlen(default_encoding_c) + 1);
    if (!__PYX_DEFAULT_STRING_ENCODING) goto bad;
    strcpy(__PYX_DEFAULT_STRING_ENCODING, default_encoding_c);
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    return -1;
}
#endif
#endif


/* Test for GCC > 2.95 */
#if defined(__GNUC__)     && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))
  #define likely(x)   __builtin_expect(!!(x), 1)
  #define unlikely(x) __builtin_expect(!!(x), 0)
#else /* !__GNUC__ or GCC < 2.95 */
  #define likely(x)   (x)
  #define unlikely(x) (x)
#endif /* __GNUC__ */
static CYTHON_INLINE void __Pyx_pretend_to_initialize(void* ptr) { (void)ptr; }

#if !CYTHON_USE_MODULE_STATE
static PyObject *__pyx_m = NULL;
#endif
static int __pyx_lineno;
static int __pyx_clineno = 0;
static const char * __pyx_cfilenm = __FILE__;
static const char *__pyx_filename;

/* Header.proto */
#if !defined(CYTHON_CCOMPLEX)
  #if defined(__cplusplus)
    #define CYTHON_CCOMPLEX 1
  #elif (defined(_Complex_I) && !defined(_MSC_VER)) || ((defined (__STDC_VERSION__) && __STDC_VERSION__ >= 201112L) && !defined(__STDC_NO_COMPLEX__) && !defined(_MSC_VER))
    #define CYTHON_CCOMPLEX 1
  #else
    #define CYTHON_CCOMPLEX 0
  #endif
#endif
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    #include <complex>
  #else
    #include <complex.h>
  #endif
#endif
#if CYTHON_CCOMPLEX && !defined(__cplusplus) && defined(__sun__) && defined(__GNUC__)
  #undef _Complex_I
  #define _Complex_I 1.0fj
#endif

/* #### Code section: filename_table ### */

static const char *__pyx_f[] = {
  "src/serializor/binary/ser.py",
  "src/serializor/binary/ser.pxd",
  "__init__.cython-30.pxd",
  "contextvars.pxd",
  "datetime.pxd",
  "src/serializor/utils.pxd",
  "type.pxd",
  "bool.pxd",
  "complex.pxd",
};
/* #### Code section: utility_code_proto_before_types ### */
/* ForceInitThreads.proto */
#ifndef __PYX_FORCE_INIT_THREADS
  #define __PYX_FORCE_INIT_THREADS 0
#endif

/* #### Code section: numeric_typedefs ### */

/* "../../../../../tmp/build-env-82p3tyx6/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":743
 * # in Cython to enable them only on the right systems.
 * 
 * ctypedef npy_int8       int8_t             # <<<<<<<<<<<<<<
 * ctypedef npy_int16      int16_t
 * ctypedef npy_int32      int32_t
 */
typedef npy_int8 __pyx_t_5numpy_int8_t;

/* "../../../../../tmp/build-env-82p3tyx6/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":744
 * 
 * ctypedef npy_int8       int8_t
 * ctypedef npy_int16      int16_t             # <<<<<<<<<<<<<<
 * ctypedef npy_int32      int32_t
 * ctypedef npy_int64      int64_t
 */
typedef npy_int16 __pyx_t_5numpy_int16_t;

/* "../../../../../tmp/build-env-82p3tyx6/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":745
 * ctypedef npy_int8       int8_t
 * ctypedef npy_int16      int16_t
 * ctypedef npy_int32      int32_t             # <<<<<<<<<<<<<<
 * ctypedef npy_int64      int64_t
 * 
 */
typedef npy_int32 __pyx_t_5numpy_int32_t;

/* "../../../../../tmp/build-env-82p3tyx6/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":746
 * ctypedef npy_int16      int16_t
 * ctypedef npy_int32      int32_t
 * ctypedef npy_int64      int64_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_uint8      uint8_t
 */
typedef npy_int64 __pyx_t_5numpy_int64_t;

/* "../../../../../tmp/build-env-82p3tyx6/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":748
 * ctypedef npy_int64      int64_t
 * 
 * ctypedef npy_uint8      uint8_t             # <<<<<<<<<<<<<<
 * ctypedef npy_uint16     uint16_t
 * ctypedef npy_uint32     uint32_t
 */
typedef npy_uint8 __pyx_t_5numpy_uint8_t;

/* "../../../../../tmp/build-env-82p3tyx6/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":749
 * 
 * ctypedef npy_uint8      uint8_t
 * ctypedef npy_uint16     uint16_t             # <<<<<<<<<<<<<<
 * ctypedef npy_uint32     uint32_t
 * ctypedef npy_uint64     uint64_t
 */
typedef npy_uint16 __pyx_t_5numpy_uint16_t;

/* "../../../../../tmp/build-env-82p3tyx6/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":750
 * ctypedef npy_uint8      uint8_t
 * ctypedef npy_uint16     uint16_t
 * ctypedef npy_uint32     uint32_t             # <<<<<<<<<<<<<<
 * ctypedef npy_uint64     uint64_t
 * 
 */
typedef npy_uint32 __pyx_t_5numpy_uint32_t;

/* "../../../../../tmp/build-env-82p3tyx6/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":751
 * ctypedef npy_uint16     uint16_t
 * ctypedef npy_uint32     uint32_t
 * ctypedef npy_uint64     uint64_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_float32    float32_t
 */
typedef npy_uint64 __pyx_t_5numpy_uint64_t;

/* "../../../../../tmp/build-env-82p3tyx6/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":753
 * ctypedef npy_uint64     uint64_t
 * 
 * ctypedef npy_float32    float32_t             # <<<<<<<<<<<<<<
 * ctypedef npy_float64    float64_t
 * #ctypedef npy_float80    float80_t
 */
typedef npy_float32 __pyx_t_5numpy_float32_t;

/* "../../../../../tmp/build-env-82p3tyx6/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":754
 * 
 * ctypedef npy_float32    float32_t
 * ctypedef npy_float64    float64_t             # <<<<<<<<<<<<<<
 * #ctypedef npy_float80    float80_t
 * #ctypedef npy_float128   float128_t
 */
typedef npy_float64 __pyx_t_5numpy_float64_t;

/* "../../../../../tmp/build-env-82p3tyx6/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":761
 * ctypedef double complex complex128_t
 * 
 * ctypedef npy_longlong   longlong_t             # <<<<<<<<<<<<<<
 * ctypedef npy_ulonglong  ulonglong_t
 * 
 */
typedef npy_longlong __pyx_t_5numpy_longlong_t;

/* "../../../../../tmp/build-env-82p3tyx6/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":762
 * 
 * ctypedef npy_longlong   longlong_t
 * ctypedef npy_ulonglong  ulonglong_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_intp       intp_t
 */
typedef npy_ulonglong __pyx_t_5numpy_ulonglong_t;

/* "../../../../../tmp/build-env-82p3tyx6/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":764
 * ctypedef npy_ulonglong  ulonglong_t
 * 
 * ctypedef npy_intp       intp_t             # <<<<<<<<<<<<<<
 * ctypedef npy_uintp      uintp_t
 * 
 */
typedef npy_intp __pyx_t_5numpy_intp_t;

/* "../../../../../tmp/build-env-82p3tyx6/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":765
 * 
 * ctypedef npy_intp       intp_t
 * ctypedef npy_uintp      uintp_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_double     float_t
 */
typedef npy_uintp __pyx_t_5numpy_uintp_t;

/* "../../../../../tmp/build-env-82p3tyx6/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":767
 * ctypedef npy_uintp      uintp_t
 * 
 * ctypedef npy_double     float_t             # <<<<<<<<<<<<<<
 * ctypedef npy_double     double_t
 * ctypedef npy_longdouble longdouble_t
 */
typedef npy_double __pyx_t_5numpy_float_t;

/* "../../../../../tmp/build-env-82p3tyx6/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":768
 * 
 * ctypedef npy_double     float_t
 * ctypedef npy_double     double_t             # <<<<<<<<<<<<<<
 * ctypedef npy_longdouble longdouble_t
 * 
 */
typedef npy_double __pyx_t_5numpy_double_t;

/* "../../../../../tmp/build-env-82p3tyx6/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":769
 * ctypedef npy_double     float_t
 * ctypedef npy_double     double_t
 * ctypedef npy_longdouble longdouble_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef float complex       cfloat_t
 */
typedef npy_longdouble __pyx_t_5numpy_longdouble_t;
/* #### Code section: complex_type_declarations ### */
/* Declarations.proto */
#if CYTHON_CCOMPLEX && (1) && (!0 || __cplusplus)
  #ifdef __cplusplus
    typedef ::std::complex< float > __pyx_t_float_complex;
  #else
    typedef float _Complex __pyx_t_float_complex;
  #endif
#else
    typedef struct { float real, imag; } __pyx_t_float_complex;
#endif
static CYTHON_INLINE __pyx_t_float_complex __pyx_t_float_complex_from_parts(float, float);

/* Declarations.proto */
#if CYTHON_CCOMPLEX && (1) && (!0 || __cplusplus)
  #ifdef __cplusplus
    typedef ::std::complex< double > __pyx_t_double_complex;
  #else
    typedef double _Complex __pyx_t_double_complex;
  #endif
#else
    typedef struct { double real, imag; } __pyx_t_double_complex;
#endif
static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double, double);

/* Declarations.proto */
#if CYTHON_CCOMPLEX && (1) && (!0 || __cplusplus)
  #ifdef __cplusplus
    typedef ::std::complex< long double > __pyx_t_long_double_complex;
  #else
    typedef long double _Complex __pyx_t_long_double_complex;
  #endif
#else
    typedef struct { long double real, imag; } __pyx_t_long_double_complex;
#endif
static CYTHON_INLINE __pyx_t_long_double_complex __pyx_t_long_double_complex_from_parts(long double, long double);

/* #### Code section: type_declarations ### */

/*--- Type declarations ---*/
struct __pyx_opt_args_7cpython_11contextvars_get_value;
struct __pyx_opt_args_7cpython_11contextvars_get_value_no_default;

/* "cpython/contextvars.pxd":112
 * 
 * 
 * cdef inline object get_value(var, default_value=None):             # <<<<<<<<<<<<<<
 *     """Return a new reference to the value of the context variable,
 *     or the default value of the context variable,
 */
struct __pyx_opt_args_7cpython_11contextvars_get_value {
  int __pyx_n;
  PyObject *default_value;
};

/* "cpython/contextvars.pxd":129
 * 
 * 
 * cdef inline object get_value_no_default(var, default_value=None):             # <<<<<<<<<<<<<<
 *     """Return a new reference to the value of the context variable,
 *     or the provided default value if no such value was found.
 */
struct __pyx_opt_args_7cpython_11contextvars_get_value_no_default {
  int __pyx_n;
  PyObject *default_value;
};
struct __pyx_opt_args_7cpython_8datetime_time_new;
struct __pyx_opt_args_7cpython_8datetime_datetime_new;
struct __pyx_opt_args_7cpython_8datetime_timezone_new;
struct __pyx_opt_args_7cpython_8datetime_datetime_from_timestamp;

/* "datetime.pxd":292
 * # Create time object using DateTime CAPI factory function
 * # Note, there are no range checks for any of the arguments.
 * cdef inline time time_new(int hour, int minute, int second, int microsecond, object tz, int fold=0):             # <<<<<<<<<<<<<<
 *     return __Pyx_DateTime_TimeWithFold(hour, minute, second, microsecond, tz, fold)
 * 
 */
struct __pyx_opt_args_7cpython_8datetime_time_new {
  int __pyx_n;
  int fold;
};

/* "datetime.pxd":297
 * # Create datetime object using DateTime CAPI factory function.
 * # Note, there are no range checks for any of the arguments.
 * cdef inline datetime datetime_new(int year, int month, int day, int hour, int minute, int second, int microsecond, object tz, int fold=0):             # <<<<<<<<<<<<<<
 *     return __Pyx_DateTime_DateTimeWithFold(year, month, day, hour, minute, second, microsecond, tz, fold)
 * 
 */
struct __pyx_opt_args_7cpython_8datetime_datetime_new {
  int __pyx_n;
  int fold;
};

/* "datetime.pxd":306
 * 
 * # Create timedelta object using DateTime CAPI factory function.
 * cdef inline object timezone_new(object offset, object name=None):             # <<<<<<<<<<<<<<
 *     if PY_VERSION_HEX < 0x030700b1:
 *         raise RuntimeError('Time zones are not available from the C-API.')
 */
struct __pyx_opt_args_7cpython_8datetime_timezone_new {
  int __pyx_n;
  PyObject *name;
};

/* "datetime.pxd":312
 * 
 * # Create datetime object using DB API constructor.
 * cdef inline datetime datetime_from_timestamp(timestamp, tz=None):             # <<<<<<<<<<<<<<
 *     return PyDateTimeAPI.DateTime_FromTimestamp(
 *         <PyObject*>PyDateTimeAPI.DateTimeType, (timestamp, tz) if tz is not None else (timestamp,), NULL)
 */
struct __pyx_opt_args_7cpython_8datetime_datetime_from_timestamp {
  int __pyx_n;
  PyObject *tz;
};
/* #### Code section: utility_code_proto ### */

/* --- Runtime support code (head) --- */
/* Refnanny.proto */
#ifndef CYTHON_REFNANNY
  #define CYTHON_REFNANNY 0
#endif
#if CYTHON_REFNANNY
  typedef struct {
    void (*INCREF)(void*, PyObject*, Py_ssize_t);
    void (*DECREF)(void*, PyObject*, Py_ssize_t);
    void (*GOTREF)(void*, PyObject*, Py_ssize_t);
    void (*GIVEREF)(void*, PyObject*, Py_ssize_t);
    void* (*SetupContext)(const char*, Py_ssize_t, const char*);
    void (*FinishContext)(void**);
  } __Pyx_RefNannyAPIStruct;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNanny = NULL;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname);
  #define __Pyx_RefNannyDeclarations void *__pyx_refnanny = NULL;
#ifdef WITH_THREAD
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          if (acquire_gil) {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), (__LINE__), (__FILE__));\
              PyGILState_Release(__pyx_gilstate_save);\
          } else {\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), (__LINE__), (__FILE__));\
          }
  #define __Pyx_RefNannyFinishContextNogil() {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __Pyx_RefNannyFinishContext();\
              PyGILState_Release(__pyx_gilstate_save);\
          }
#else
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), (__LINE__), (__FILE__))
  #define __Pyx_RefNannyFinishContextNogil() __Pyx_RefNannyFinishContext()
#endif
  #define __Pyx_RefNannyFinishContextNogil() {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __Pyx_RefNannyFinishContext();\
              PyGILState_Release(__pyx_gilstate_save);\
          }
  #define __Pyx_RefNannyFinishContext()\
          __Pyx_RefNanny->FinishContext(&__pyx_refnanny)
  #define __Pyx_INCREF(r)  __Pyx_RefNanny->INCREF(__pyx_refnanny, (PyObject *)(r), (__LINE__))
  #define __Pyx_DECREF(r)  __Pyx_RefNanny->DECREF(__pyx_refnanny, (PyObject *)(r), (__LINE__))
  #define __Pyx_GOTREF(r)  __Pyx_RefNanny->GOTREF(__pyx_refnanny, (PyObject *)(r), (__LINE__))
  #define __Pyx_GIVEREF(r) __Pyx_RefNanny->GIVEREF(__pyx_refnanny, (PyObject *)(r), (__LINE__))
  #define __Pyx_XINCREF(r)  do { if((r) == NULL); else {__Pyx_INCREF(r); }} while(0)
  #define __Pyx_XDECREF(r)  do { if((r) == NULL); else {__Pyx_DECREF(r); }} while(0)
  #define __Pyx_XGOTREF(r)  do { if((r) == NULL); else {__Pyx_GOTREF(r); }} while(0)
  #define __Pyx_XGIVEREF(r) do { if((r) == NULL); else {__Pyx_GIVEREF(r);}} while(0)
#else
  #define __Pyx_RefNannyDeclarations
  #define __Pyx_RefNannySetupContext(name, acquire_gil)
  #define __Pyx_RefNannyFinishContextNogil()
  #define __Pyx_RefNannyFinishContext()
  #define __Pyx_INCREF(r) Py_INCREF(r)
  #define __Pyx_DECREF(r) Py_DECREF(r)
  #define __Pyx_GOTREF(r)
  #define __Pyx_GIVEREF(r)
  #define __Pyx_XINCREF(r) Py_XINCREF(r)
  #define __Pyx_XDECREF(r) Py_XDECREF(r)
  #define __Pyx_XGOTREF(r)
  #define __Pyx_XGIVEREF(r)
#endif
#define __Pyx_Py_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; Py_XDECREF(tmp);\
    } while (0)
#define __Pyx_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_XDECREF(tmp);\
    } while (0)
#define __Pyx_DECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_DECREF(tmp);\
    } while (0)
#define __Pyx_CLEAR(r)    do { PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);} while(0)
#define __Pyx_XCLEAR(r)   do { if((r) != NULL) {PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);}} while(0)

/* PyErrExceptionMatches.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_ExceptionMatches(err) __Pyx_PyErr_ExceptionMatchesInState(__pyx_tstate, err)
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err);
#else
#define __Pyx_PyErr_ExceptionMatches(err)  PyErr_ExceptionMatches(err)
#endif

/* PyThreadStateGet.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyThreadState_declare  PyThreadState *__pyx_tstate;
#define __Pyx_PyThreadState_assign  __pyx_tstate = __Pyx_PyThreadState_Current;
#if PY_VERSION_HEX >= 0x030C00A6
#define __Pyx_PyErr_Occurred()  (__pyx_tstate->current_exception != NULL)
#define __Pyx_PyErr_CurrentExceptionType()  (__pyx_tstate->current_exception ? (PyObject*) Py_TYPE(__pyx_tstate->current_exception) : (PyObject*) NULL)
#else
#define __Pyx_PyErr_Occurred()  (__pyx_tstate->curexc_type != NULL)
#define __Pyx_PyErr_CurrentExceptionType()  (__pyx_tstate->curexc_type)
#endif
#else
#define __Pyx_PyThreadState_declare
#define __Pyx_PyThreadState_assign
#define __Pyx_PyErr_Occurred()  (PyErr_Occurred() != NULL)
#define __Pyx_PyErr_CurrentExceptionType()  PyErr_Occurred()
#endif

/* PyErrFetchRestore.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_Clear() __Pyx_ErrRestore(NULL, NULL, NULL)
#define __Pyx_ErrRestoreWithState(type, value, tb)  __Pyx_ErrRestoreInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)    __Pyx_ErrFetchInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  __Pyx_ErrRestoreInState(__pyx_tstate, type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)    __Pyx_ErrFetchInState(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A6
#define __Pyx_PyErr_SetNone(exc) (Py_INCREF(exc), __Pyx_ErrRestore((exc), NULL, NULL))
#else
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#endif
#else
#define __Pyx_PyErr_Clear() PyErr_Clear()
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#define __Pyx_ErrRestoreWithState(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestoreInState(tstate, type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchInState(tstate, type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)  PyErr_Fetch(type, value, tb)
#endif

/* PyObjectGetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GetAttrStr(o,n) PyObject_GetAttr(o,n)
#endif

/* PyObjectGetAttrStrNoError.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStrNoError(PyObject* obj, PyObject* attr_name);

/* GetBuiltinName.proto */
static PyObject *__Pyx_GetBuiltinName(PyObject *name);

/* GetTopmostException.proto */
#if CYTHON_USE_EXC_INFO_STACK && CYTHON_FAST_THREAD_STATE
static _PyErr_StackItem * __Pyx_PyErr_GetTopmostException(PyThreadState *tstate);
#endif

/* SaveResetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ExceptionSave(type, value, tb)  __Pyx__ExceptionSave(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#define __Pyx_ExceptionReset(type, value, tb)  __Pyx__ExceptionReset(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
#else
#define __Pyx_ExceptionSave(type, value, tb)   PyErr_GetExcInfo(type, value, tb)
#define __Pyx_ExceptionReset(type, value, tb)  PyErr_SetExcInfo(type, value, tb)
#endif

/* GetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_GetException(type, value, tb)  __Pyx__GetException(__pyx_tstate, type, value, tb)
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb);
#endif

/* PyObjectCall.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw);
#else
#define __Pyx_PyObject_Call(func, arg, kw) PyObject_Call(func, arg, kw)
#endif

/* RaiseException.proto */
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause);

/* ExtTypeTest.proto */
static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type);

/* RaiseUnexpectedTypeError.proto */
static int __Pyx_RaiseUnexpectedTypeError(const char *expected, PyObject *obj);

/* PyFunctionFastCall.proto */
#if CYTHON_FAST_PYCALL
#if !CYTHON_VECTORCALL
#define __Pyx_PyFunction_FastCall(func, args, nargs)\
    __Pyx_PyFunction_FastCallDict((func), (args), (nargs), NULL)
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, Py_ssize_t nargs, PyObject *kwargs);
#endif
#define __Pyx_BUILD_ASSERT_EXPR(cond)\
    (sizeof(char [1 - 2*!(cond)]) - 1)
#ifndef Py_MEMBER_SIZE
#define Py_MEMBER_SIZE(type, member) sizeof(((type *)0)->member)
#endif
#if !CYTHON_VECTORCALL
#if PY_VERSION_HEX >= 0x03080000
  #include "frameobject.h"
#if PY_VERSION_HEX >= 0x030b00a6 && !CYTHON_COMPILING_IN_LIMITED_API && !defined(PYPY_VERSION)
  #ifndef Py_BUILD_CORE
    #define Py_BUILD_CORE 1
  #endif
  #include "internal/pycore_frame.h"
#endif
  #define __Pxy_PyFrame_Initialize_Offsets()
  #define __Pyx_PyFrame_GetLocalsplus(frame)  ((frame)->f_localsplus)
#else
  static size_t __pyx_pyframe_localsplus_offset = 0;
  #include "frameobject.h"
  #define __Pxy_PyFrame_Initialize_Offsets()\
    ((void)__Pyx_BUILD_ASSERT_EXPR(sizeof(PyFrameObject) == offsetof(PyFrameObject, f_localsplus) + Py_MEMBER_SIZE(PyFrameObject, f_localsplus)),\
     (void)(__pyx_pyframe_localsplus_offset = ((size_t)PyFrame_Type.tp_basicsize) - Py_MEMBER_SIZE(PyFrameObject, f_localsplus)))
  #define __Pyx_PyFrame_GetLocalsplus(frame)\
    (assert(__pyx_pyframe_localsplus_offset), (PyObject **)(((char *)(frame)) + __pyx_pyframe_localsplus_offset))
#endif
#endif
#endif

/* PyObjectCallMethO.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg);
#endif

/* PyObjectFastCall.proto */
#define __Pyx_PyObject_FastCall(func, args, nargs)  __Pyx_PyObject_FastCallDict(func, args, (size_t)(nargs), NULL)
static CYTHON_INLINE PyObject* __Pyx_PyObject_FastCallDict(PyObject *func, PyObject **args, size_t nargs, PyObject *kwargs);

/* PyObjectCallOneArg.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg);

/* SwapException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ExceptionSwap(type, value, tb)  __Pyx__ExceptionSwap(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSwap(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static CYTHON_INLINE void __Pyx_ExceptionSwap(PyObject **type, PyObject **value, PyObject **tb);
#endif

/* PyDictVersioning.proto */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
#define __PYX_DICT_VERSION_INIT  ((PY_UINT64_T) -1)
#define __PYX_GET_DICT_VERSION(dict)  (((PyDictObject*)(dict))->ma_version_tag)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)\
    (version_var) = __PYX_GET_DICT_VERSION(dict);\
    (cache_var) = (value);
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP) {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    if (likely(__PYX_GET_DICT_VERSION(DICT) == __pyx_dict_version)) {\
        (VAR) = __pyx_dict_cached_value;\
    } else {\
        (VAR) = __pyx_dict_cached_value = (LOOKUP);\
        __pyx_dict_version = __PYX_GET_DICT_VERSION(DICT);\
    }\
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj);
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj);
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version);
#else
#define __PYX_GET_DICT_VERSION(dict)  (0)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP)  (VAR) = (LOOKUP);
#endif

/* GetModuleGlobalName.proto */
#if CYTHON_USE_DICT_VERSIONS
#define __Pyx_GetModuleGlobalName(var, name)  do {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    (var) = (likely(__pyx_dict_version == __PYX_GET_DICT_VERSION(__pyx_d))) ?\
        (likely(__pyx_dict_cached_value) ? __Pyx_NewRef(__pyx_dict_cached_value) : __Pyx_GetBuiltinName(name)) :\
        __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
} while(0)
#define __Pyx_GetModuleGlobalNameUncached(var, name)  do {\
    PY_UINT64_T __pyx_dict_version;\
    PyObject *__pyx_dict_cached_value;\
    (var) = __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
} while(0)
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value);
#else
#define __Pyx_GetModuleGlobalName(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
#define __Pyx_GetModuleGlobalNameUncached(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name);
#endif

/* PyObject_Unicode.proto */
#if PY_MAJOR_VERSION >= 3
#define __Pyx_PyObject_Unicode(obj)\
    (likely(PyUnicode_CheckExact(obj)) ? __Pyx_NewRef(obj) : PyObject_Str(obj))
#else
#define __Pyx_PyObject_Unicode(obj)\
    (likely(PyUnicode_CheckExact(obj)) ? __Pyx_NewRef(obj) : PyObject_Unicode(obj))
#endif

/* PyObjectCall2Args.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2);

/* PyObjectGetMethod.proto */
static int __Pyx_PyObject_GetMethod(PyObject *obj, PyObject *name, PyObject **method);

/* PyObjectCallMethod1.proto */
static PyObject* __Pyx_PyObject_CallMethod1(PyObject* obj, PyObject* method_name, PyObject* arg);

/* StringJoin.proto */
#if PY_MAJOR_VERSION < 3
#define __Pyx_PyString_Join __Pyx_PyBytes_Join
#define __Pyx_PyBaseString_Join(s, v) (PyUnicode_CheckExact(s) ? PyUnicode_Join(s, v) : __Pyx_PyBytes_Join(s, v))
#else
#define __Pyx_PyString_Join PyUnicode_Join
#define __Pyx_PyBaseString_Join PyUnicode_Join
#endif
static CYTHON_INLINE PyObject* __Pyx_PyBytes_Join(PyObject* sep, PyObject* values);

/* ListAppend.proto */
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static CYTHON_INLINE int __Pyx_PyList_Append(PyObject* list, PyObject* x) {
    PyListObject* L = (PyListObject*) list;
    Py_ssize_t len = Py_SIZE(list);
    if (likely(L->allocated > len) & likely(len > (L->allocated >> 1))) {
        Py_INCREF(x);
        #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030d0000
        L->ob_item[len] = x;
        #else
        PyList_SET_ITEM(list, len, x);
        #endif
        __Pyx_SET_SIZE(list, len + 1);
        return 0;
    }
    return PyList_Append(list, x);
}
#else
#define __Pyx_PyList_Append(L,x) PyList_Append(L,x)
#endif

/* UnpackUnboundCMethod.proto */
typedef struct {
    PyObject *type;
    PyObject **method_name;
    PyCFunction func;
    PyObject *method;
    int flag;
} __Pyx_CachedCFunction;

/* CallUnboundCMethod0.proto */
static PyObject* __Pyx__CallUnboundCMethod0(__Pyx_CachedCFunction* cfunc, PyObject* self);
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_CallUnboundCMethod0(cfunc, self)\
    (likely((cfunc)->func) ?\
        (likely((cfunc)->flag == METH_NOARGS) ?  (*((cfunc)->func))(self, NULL) :\
         (PY_VERSION_HEX >= 0x030600B1 && likely((cfunc)->flag == METH_FASTCALL) ?\
            (PY_VERSION_HEX >= 0x030700A0 ?\
                (*(__Pyx_PyCFunctionFast)(void*)(PyCFunction)(cfunc)->func)(self, &__pyx_empty_tuple, 0) :\
                (*(__Pyx_PyCFunctionFastWithKeywords)(void*)(PyCFunction)(cfunc)->func)(self, &__pyx_empty_tuple, 0, NULL)) :\
          (PY_VERSION_HEX >= 0x030700A0 && (cfunc)->flag == (METH_FASTCALL | METH_KEYWORDS) ?\
            (*(__Pyx_PyCFunctionFastWithKeywords)(void*)(PyCFunction)(cfunc)->func)(self, &__pyx_empty_tuple, 0, NULL) :\
            (likely((cfunc)->flag == (METH_VARARGS | METH_KEYWORDS)) ?  ((*(PyCFunctionWithKeywords)(void*)(PyCFunction)(cfunc)->func)(self, __pyx_empty_tuple, NULL)) :\
               ((cfunc)->flag == METH_VARARGS ?  (*((cfunc)->func))(self, __pyx_empty_tuple) :\
               __Pyx__CallUnboundCMethod0(cfunc, self)))))) :\
        __Pyx__CallUnboundCMethod0(cfunc, self))
#else
#define __Pyx_CallUnboundCMethod0(cfunc, self)  __Pyx__CallUnboundCMethod0(cfunc, self)
#endif

/* IterFinish.proto */
static CYTHON_INLINE int __Pyx_IterFinish(void);

/* set_iter.proto */
static CYTHON_INLINE PyObject* __Pyx_set_iterator(PyObject* iterable, int is_set,
                                                  Py_ssize_t* p_orig_length, int* p_source_is_set);
static CYTHON_INLINE int __Pyx_set_iter_next(
        PyObject* iter_obj, Py_ssize_t orig_length,
        Py_ssize_t* ppos, PyObject **value,
        int source_is_set);

/* PyObjectCallNoArg.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func);

/* PyObjectCallMethod0.proto */
static PyObject* __Pyx_PyObject_CallMethod0(PyObject* obj, PyObject* method_name);

/* RaiseNeedMoreValuesToUnpack.proto */
static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index);

/* RaiseTooManyValuesToUnpack.proto */
static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected);

/* UnpackItemEndCheck.proto */
static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected);

/* RaiseNoneIterError.proto */
static CYTHON_INLINE void __Pyx_RaiseNoneNotIterableError(void);

/* UnpackTupleError.proto */
static void __Pyx_UnpackTupleError(PyObject *, Py_ssize_t index);

/* UnpackTuple2.proto */
#define __Pyx_unpack_tuple2(tuple, value1, value2, is_tuple, has_known_size, decref_tuple)\
    (likely(is_tuple || PyTuple_Check(tuple)) ?\
        (likely(has_known_size || PyTuple_GET_SIZE(tuple) == 2) ?\
            __Pyx_unpack_tuple2_exact(tuple, value1, value2, decref_tuple) :\
            (__Pyx_UnpackTupleError(tuple, 2), -1)) :\
        __Pyx_unpack_tuple2_generic(tuple, value1, value2, has_known_size, decref_tuple))
static CYTHON_INLINE int __Pyx_unpack_tuple2_exact(
    PyObject* tuple, PyObject** value1, PyObject** value2, int decref_tuple);
static int __Pyx_unpack_tuple2_generic(
    PyObject* tuple, PyObject** value1, PyObject** value2, int has_known_size, int decref_tuple);

/* dict_iter.proto */
static CYTHON_INLINE PyObject* __Pyx_dict_iterator(PyObject* dict, int is_dict, PyObject* method_name,
                                                   Py_ssize_t* p_orig_length, int* p_is_dict);
static CYTHON_INLINE int __Pyx_dict_iter_next(PyObject* dict_or_iter, Py_ssize_t orig_length, Py_ssize_t* ppos,
                                              PyObject** pkey, PyObject** pvalue, PyObject** pitem, int is_dict);

/* pybytes_as_double.proto */
static double __Pyx_SlowPyString_AsDouble(PyObject *obj);
static double __Pyx__PyBytes_AsDouble(PyObject *obj, const char* start, Py_ssize_t length);
static CYTHON_INLINE double __Pyx_PyBytes_AsDouble(PyObject *obj) {
    char* as_c_string;
    Py_ssize_t size;
#if CYTHON_ASSUME_SAFE_MACROS
    as_c_string = PyBytes_AS_STRING(obj);
    size = PyBytes_GET_SIZE(obj);
#else
    if (PyBytes_AsStringAndSize(obj, &as_c_string, &size) < 0) {
        return (double)-1;
    }
#endif
    return __Pyx__PyBytes_AsDouble(obj, as_c_string, size);
}
static CYTHON_INLINE double __Pyx_PyByteArray_AsDouble(PyObject *obj) {
    char* as_c_string;
    Py_ssize_t size;
#if CYTHON_ASSUME_SAFE_MACROS
    as_c_string = PyByteArray_AS_STRING(obj);
    size = PyByteArray_GET_SIZE(obj);
#else
    as_c_string = PyByteArray_AsString(obj);
    if (as_c_string == NULL) {
        return (double)-1;
    }
    size = PyByteArray_Size(obj);
#endif
    return __Pyx__PyBytes_AsDouble(obj, as_c_string, size);
}

/* pyunicode_as_double.proto */
#if PY_MAJOR_VERSION >= 3 && !CYTHON_COMPILING_IN_PYPY && CYTHON_ASSUME_SAFE_MACROS
static const char* __Pyx__PyUnicode_AsDouble_Copy(const void* data, const int kind, char* buffer, Py_ssize_t start, Py_ssize_t end) {
    int last_was_punctuation;
    Py_ssize_t i;
    last_was_punctuation = 1;
    for (i=start; i <= end; i++) {
        Py_UCS4 chr = PyUnicode_READ(kind, data, i);
        int is_punctuation = (chr == '_') | (chr == '.');
        *buffer = (char)chr;
        buffer += (chr != '_');
        if (unlikely(chr > 127)) goto parse_failure;
        if (unlikely(last_was_punctuation & is_punctuation)) goto parse_failure;
        last_was_punctuation = is_punctuation;
    }
    if (unlikely(last_was_punctuation)) goto parse_failure;
    *buffer = '\0';
    return buffer;
parse_failure:
    return NULL;
}
static double __Pyx__PyUnicode_AsDouble_inf_nan(const void* data, int kind, Py_ssize_t start, Py_ssize_t length) {
    int matches = 1;
    Py_UCS4 chr;
    Py_UCS4 sign = PyUnicode_READ(kind, data, start);
    int is_signed = (sign == '-') | (sign == '+');
    start += is_signed;
    length -= is_signed;
    switch (PyUnicode_READ(kind, data, start)) {
        #ifdef Py_NAN
        case 'n':
        case 'N':
            if (unlikely(length != 3)) goto parse_failure;
            chr = PyUnicode_READ(kind, data, start+1);
            matches &= (chr == 'a') | (chr == 'A');
            chr = PyUnicode_READ(kind, data, start+2);
            matches &= (chr == 'n') | (chr == 'N');
            if (unlikely(!matches)) goto parse_failure;
            return (sign == '-') ? -Py_NAN : Py_NAN;
        #endif
        case 'i':
        case 'I':
            if (unlikely(length < 3)) goto parse_failure;
            chr = PyUnicode_READ(kind, data, start+1);
            matches &= (chr == 'n') | (chr == 'N');
            chr = PyUnicode_READ(kind, data, start+2);
            matches &= (chr == 'f') | (chr == 'F');
            if (likely(length == 3 && matches))
                return (sign == '-') ? -Py_HUGE_VAL : Py_HUGE_VAL;
            if (unlikely(length != 8)) goto parse_failure;
            chr = PyUnicode_READ(kind, data, start+3);
            matches &= (chr == 'i') | (chr == 'I');
            chr = PyUnicode_READ(kind, data, start+4);
            matches &= (chr == 'n') | (chr == 'N');
            chr = PyUnicode_READ(kind, data, start+5);
            matches &= (chr == 'i') | (chr == 'I');
            chr = PyUnicode_READ(kind, data, start+6);
            matches &= (chr == 't') | (chr == 'T');
            chr = PyUnicode_READ(kind, data, start+7);
            matches &= (chr == 'y') | (chr == 'Y');
            if (unlikely(!matches)) goto parse_failure;
            return (sign == '-') ? -Py_HUGE_VAL : Py_HUGE_VAL;
        case '.': case '0': case '1': case '2': case '3': case '4': case '5': case '6': case '7': case '8': case '9':
            break;
        default:
            goto parse_failure;
    }
    return 0.0;
parse_failure:
    return -1.0;
}
static double __Pyx_PyUnicode_AsDouble_WithSpaces(PyObject *obj) {
    double value;
    const char *last;
    char *end;
    Py_ssize_t start, length = PyUnicode_GET_LENGTH(obj);
    const int kind = PyUnicode_KIND(obj);
    const void* data = PyUnicode_DATA(obj);
    start = 0;
    while (Py_UNICODE_ISSPACE(PyUnicode_READ(kind, data, start)))
        start++;
    while (start < length - 1 && Py_UNICODE_ISSPACE(PyUnicode_READ(kind, data, length - 1)))
        length--;
    length -= start;
    if (unlikely(length <= 0)) goto fallback;
    value = __Pyx__PyUnicode_AsDouble_inf_nan(data, kind, start, length);
    if (unlikely(value == -1.0)) goto fallback;
    if (value != 0.0) return value;
    if (length < 40) {
        char number[40];
        last = __Pyx__PyUnicode_AsDouble_Copy(data, kind, number, start, start + length);
        if (unlikely(!last)) goto fallback;
        value = PyOS_string_to_double(number, &end, NULL);
    } else {
        char *number = (char*) PyMem_Malloc((length + 1) * sizeof(char));
        if (unlikely(!number)) goto fallback;
        last = __Pyx__PyUnicode_AsDouble_Copy(data, kind, number, start, start + length);
        if (unlikely(!last)) {
            PyMem_Free(number);
            goto fallback;
        }
        value = PyOS_string_to_double(number, &end, NULL);
        PyMem_Free(number);
    }
    if (likely(end == last) || (value == (double)-1 && PyErr_Occurred())) {
        return value;
    }
fallback:
    return __Pyx_SlowPyString_AsDouble(obj);
}
#endif
static CYTHON_INLINE double __Pyx_PyUnicode_AsDouble(PyObject *obj) {
#if PY_MAJOR_VERSION >= 3 && !CYTHON_COMPILING_IN_PYPY && CYTHON_ASSUME_SAFE_MACROS
    if (unlikely(__Pyx_PyUnicode_READY(obj) == -1))
        return (double)-1;
    if (likely(PyUnicode_IS_ASCII(obj))) {
        const char *s;
        Py_ssize_t length;
        s = PyUnicode_AsUTF8AndSize(obj, &length);
        return __Pyx__PyBytes_AsDouble(obj, s, length);
    }
    return __Pyx_PyUnicode_AsDouble_WithSpaces(obj);
#else
    return __Pyx_SlowPyString_AsDouble(obj);
#endif
}

/* pynumber_float.proto */
static CYTHON_INLINE PyObject* __Pyx__PyNumber_Float(PyObject* obj);
#define __Pyx_PyNumber_Float(x) (PyFloat_CheckExact(x) ? __Pyx_NewRef(x) : __Pyx__PyNumber_Float(x))

/* GetItemInt.proto */
#define __Pyx_GetItemInt(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Fast(o, (Py_ssize_t)i, is_list, wraparound, boundscheck) :\
    (is_list ? (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL) :\
               __Pyx_GetItemInt_Generic(o, to_py_func(i))))
#define __Pyx_GetItemInt_List(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_List_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
#define __Pyx_GetItemInt_Tuple(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Tuple_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "tuple index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j);
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i,
                                                     int is_list, int wraparound, int boundscheck);

/* ObjectGetItem.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject *__Pyx_PyObject_GetItem(PyObject *obj, PyObject *key);
#else
#define __Pyx_PyObject_GetItem(obj, key)  PyObject_GetItem(obj, key)
#endif

/* pyfrozenset_new.proto */
static CYTHON_INLINE PyObject* __Pyx_PyFrozenSet_New(PyObject* it);

/* PyObjectFormatAndDecref.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_FormatSimpleAndDecref(PyObject* s, PyObject* f);
static CYTHON_INLINE PyObject* __Pyx_PyObject_FormatAndDecref(PyObject* s, PyObject* f);

/* IncludeStringH.proto */
#include <string.h>

/* JoinPyUnicode.proto */
static PyObject* __Pyx_PyUnicode_Join(PyObject* value_tuple, Py_ssize_t value_count, Py_ssize_t result_ulength,
                                      Py_UCS4 max_char);

/* TupleAndListFromArray.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyList_FromArray(PyObject *const *src, Py_ssize_t n);
static CYTHON_INLINE PyObject* __Pyx_PyTuple_FromArray(PyObject *const *src, Py_ssize_t n);
#endif

/* BytesEquals.proto */
static CYTHON_INLINE int __Pyx_PyBytes_Equals(PyObject* s1, PyObject* s2, int equals);

/* UnicodeEquals.proto */
static CYTHON_INLINE int __Pyx_PyUnicode_Equals(PyObject* s1, PyObject* s2, int equals);

/* fastcall.proto */
#if CYTHON_AVOID_BORROWED_REFS
    #define __Pyx_Arg_VARARGS(args, i) PySequence_GetItem(args, i)
#elif CYTHON_ASSUME_SAFE_MACROS
    #define __Pyx_Arg_VARARGS(args, i) PyTuple_GET_ITEM(args, i)
#else
    #define __Pyx_Arg_VARARGS(args, i) PyTuple_GetItem(args, i)
#endif
#if CYTHON_AVOID_BORROWED_REFS
    #define __Pyx_Arg_NewRef_VARARGS(arg) __Pyx_NewRef(arg)
    #define __Pyx_Arg_XDECREF_VARARGS(arg) Py_XDECREF(arg)
#else
    #define __Pyx_Arg_NewRef_VARARGS(arg) arg
    #define __Pyx_Arg_XDECREF_VARARGS(arg)
#endif
#define __Pyx_NumKwargs_VARARGS(kwds) PyDict_Size(kwds)
#define __Pyx_KwValues_VARARGS(args, nargs) NULL
#define __Pyx_GetKwValue_VARARGS(kw, kwvalues, s) __Pyx_PyDict_GetItemStrWithError(kw, s)
#define __Pyx_KwargsAsDict_VARARGS(kw, kwvalues) PyDict_Copy(kw)
#if CYTHON_METH_FASTCALL
    #define __Pyx_Arg_FASTCALL(args, i) args[i]
    #define __Pyx_NumKwargs_FASTCALL(kwds) PyTuple_GET_SIZE(kwds)
    #define __Pyx_KwValues_FASTCALL(args, nargs) ((args) + (nargs))
    static CYTHON_INLINE PyObject * __Pyx_GetKwValue_FASTCALL(PyObject *kwnames, PyObject *const *kwvalues, PyObject *s);
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030d0000
    CYTHON_UNUSED static PyObject *__Pyx_KwargsAsDict_FASTCALL(PyObject *kwnames, PyObject *const *kwvalues);
  #else
    #define __Pyx_KwargsAsDict_FASTCALL(kw, kwvalues) _PyStack_AsDict(kwvalues, kw)
  #endif
    #define __Pyx_Arg_NewRef_FASTCALL(arg) arg  /* no-op, __Pyx_Arg_FASTCALL is direct and this needs
                                                   to have the same reference counting */
    #define __Pyx_Arg_XDECREF_FASTCALL(arg)
#else
    #define __Pyx_Arg_FASTCALL __Pyx_Arg_VARARGS
    #define __Pyx_NumKwargs_FASTCALL __Pyx_NumKwargs_VARARGS
    #define __Pyx_KwValues_FASTCALL __Pyx_KwValues_VARARGS
    #define __Pyx_GetKwValue_FASTCALL __Pyx_GetKwValue_VARARGS
    #define __Pyx_KwargsAsDict_FASTCALL __Pyx_KwargsAsDict_VARARGS
    #define __Pyx_Arg_NewRef_FASTCALL(arg) __Pyx_Arg_NewRef_VARARGS(arg)
    #define __Pyx_Arg_XDECREF_FASTCALL(arg) __Pyx_Arg_XDECREF_VARARGS(arg)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
#define __Pyx_ArgsSlice_VARARGS(args, start, stop) __Pyx_PyTuple_FromArray(&__Pyx_Arg_VARARGS(args, start), stop - start)
#define __Pyx_ArgsSlice_FASTCALL(args, start, stop) __Pyx_PyTuple_FromArray(&__Pyx_Arg_FASTCALL(args, start), stop - start)
#else
#define __Pyx_ArgsSlice_VARARGS(args, start, stop) PyTuple_GetSlice(args, start, stop)
#define __Pyx_ArgsSlice_FASTCALL(args, start, stop) PyTuple_GetSlice(args, start, stop)
#endif

/* RaiseDoubleKeywords.proto */
static void __Pyx_RaiseDoubleKeywordsError(const char* func_name, PyObject* kw_name);

/* ParseKeywords.proto */
static int __Pyx_ParseOptionalKeywords(PyObject *kwds, PyObject *const *kwvalues,
    PyObject **argnames[],
    PyObject *kwds2, PyObject *values[], Py_ssize_t num_pos_args,
    const char* function_name);

/* RaiseArgTupleInvalid.proto */
static void __Pyx_RaiseArgtupleInvalid(const char* func_name, int exact,
    Py_ssize_t num_min, Py_ssize_t num_max, Py_ssize_t num_found);

/* Import.proto */
static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level);

/* ImportDottedModule.proto */
static PyObject *__Pyx_ImportDottedModule(PyObject *name, PyObject *parts_tuple);
#if PY_MAJOR_VERSION >= 3
static PyObject *__Pyx_ImportDottedModule_WalkParts(PyObject *module, PyObject *name, PyObject *parts_tuple);
#endif

/* AssertionsEnabled.proto */
#if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x02070600 && !defined(Py_OptimizeFlag)
  #define __Pyx_init_assertions_enabled()  (0)
  #define __pyx_assertions_enabled()  (1)
#elif CYTHON_COMPILING_IN_LIMITED_API  ||  (CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030C0000)
  static int __pyx_assertions_enabled_flag;
  #define __pyx_assertions_enabled() (__pyx_assertions_enabled_flag)
  static int __Pyx_init_assertions_enabled(void) {
    PyObject *builtins, *debug, *debug_str;
    int flag;
    builtins = PyEval_GetBuiltins();
    if (!builtins) goto bad;
    debug_str = PyUnicode_FromStringAndSize("__debug__", 9);
    if (!debug_str) goto bad;
    debug = PyObject_GetItem(builtins, debug_str);
    Py_DECREF(debug_str);
    if (!debug) goto bad;
    flag = PyObject_IsTrue(debug);
    Py_DECREF(debug);
    if (flag == -1) goto bad;
    __pyx_assertions_enabled_flag = flag;
    return 0;
  bad:
    __pyx_assertions_enabled_flag = 1;
    return -1;
  }
#else
  #define __Pyx_init_assertions_enabled()  (0)
  #define __pyx_assertions_enabled()  (!Py_OptimizeFlag)
#endif

/* PyObjectFormatSimple.proto */
#if CYTHON_COMPILING_IN_PYPY
    #define __Pyx_PyObject_FormatSimple(s, f) (\
        likely(PyUnicode_CheckExact(s)) ? (Py_INCREF(s), s) :\
        PyObject_Format(s, f))
#elif PY_MAJOR_VERSION < 3
    #define __Pyx_PyObject_FormatSimple(s, f) (\
        likely(PyUnicode_CheckExact(s)) ? (Py_INCREF(s), s) :\
        likely(PyString_CheckExact(s)) ? PyUnicode_FromEncodedObject(s, NULL, "strict") :\
        PyObject_Format(s, f))
#elif CYTHON_USE_TYPE_SLOTS
    #define __Pyx_PyObject_FormatSimple(s, f) (\
        likely(PyUnicode_CheckExact(s)) ? (Py_INCREF(s), s) :\
        likely(PyLong_CheckExact(s)) ? PyLong_Type.tp_repr(s) :\
        likely(PyFloat_CheckExact(s)) ? PyFloat_Type.tp_repr(s) :\
        PyObject_Format(s, f))
#else
    #define __Pyx_PyObject_FormatSimple(s, f) (\
        likely(PyUnicode_CheckExact(s)) ? (Py_INCREF(s), s) :\
        PyObject_Format(s, f))
#endif

/* SliceObject.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetSlice(
        PyObject* obj, Py_ssize_t cstart, Py_ssize_t cstop,
        PyObject** py_start, PyObject** py_stop, PyObject** py_slice,
        int has_cstart, int has_cstop, int wraparound);

/* TypeImport.proto */
#ifndef __PYX_HAVE_RT_ImportType_proto_3_0_12
#define __PYX_HAVE_RT_ImportType_proto_3_0_12
#if defined (__STDC_VERSION__) && __STDC_VERSION__ >= 201112L
#include <stdalign.h>
#endif
#if (defined (__STDC_VERSION__) && __STDC_VERSION__ >= 201112L) || __cplusplus >= 201103L
#define __PYX_GET_STRUCT_ALIGNMENT_3_0_12(s) alignof(s)
#else
#define __PYX_GET_STRUCT_ALIGNMENT_3_0_12(s) sizeof(void*)
#endif
enum __Pyx_ImportType_CheckSize_3_0_12 {
   __Pyx_ImportType_CheckSize_Error_3_0_12 = 0,
   __Pyx_ImportType_CheckSize_Warn_3_0_12 = 1,
   __Pyx_ImportType_CheckSize_Ignore_3_0_12 = 2
};
static PyTypeObject *__Pyx_ImportType_3_0_12(PyObject* module, const char *module_name, const char *class_name, size_t size, size_t alignment, enum __Pyx_ImportType_CheckSize_3_0_12 check_size);
#endif

/* ImportFrom.proto */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name);

/* IncludeStructmemberH.proto */
#include <structmember.h>

/* FixUpExtensionType.proto */
#if CYTHON_USE_TYPE_SPECS
static int __Pyx_fix_up_extension_type_from_spec(PyType_Spec *spec, PyTypeObject *type);
#endif

/* FetchSharedCythonModule.proto */
static PyObject *__Pyx_FetchSharedCythonABIModule(void);

/* FetchCommonType.proto */
#if !CYTHON_USE_TYPE_SPECS
static PyTypeObject* __Pyx_FetchCommonType(PyTypeObject* type);
#else
static PyTypeObject* __Pyx_FetchCommonTypeFromSpec(PyObject *module, PyType_Spec *spec, PyObject *bases);
#endif

/* PyMethodNew.proto */
#if CYTHON_COMPILING_IN_LIMITED_API
static PyObject *__Pyx_PyMethod_New(PyObject *func, PyObject *self, PyObject *typ) {
    PyObject *typesModule=NULL, *methodType=NULL, *result=NULL;
    CYTHON_UNUSED_VAR(typ);
    if (!self)
        return __Pyx_NewRef(func);
    typesModule = PyImport_ImportModule("types");
    if (!typesModule) return NULL;
    methodType = PyObject_GetAttrString(typesModule, "MethodType");
    Py_DECREF(typesModule);
    if (!methodType) return NULL;
    result = PyObject_CallFunctionObjArgs(methodType, func, self, NULL);
    Py_DECREF(methodType);
    return result;
}
#elif PY_MAJOR_VERSION >= 3
static PyObject *__Pyx_PyMethod_New(PyObject *func, PyObject *self, PyObject *typ) {
    CYTHON_UNUSED_VAR(typ);
    if (!self)
        return __Pyx_NewRef(func);
    return PyMethod_New(func, self);
}
#else
    #define __Pyx_PyMethod_New PyMethod_New
#endif

/* PyVectorcallFastCallDict.proto */
#if CYTHON_METH_FASTCALL
static CYTHON_INLINE PyObject *__Pyx_PyVectorcall_FastCallDict(PyObject *func, __pyx_vectorcallfunc vc, PyObject *const *args, size_t nargs, PyObject *kw);
#endif

/* CythonFunctionShared.proto */
#define __Pyx_CyFunction_USED
#define __Pyx_CYFUNCTION_STATICMETHOD  0x01
#define __Pyx_CYFUNCTION_CLASSMETHOD   0x02
#define __Pyx_CYFUNCTION_CCLASS        0x04
#define __Pyx_CYFUNCTION_COROUTINE     0x08
#define __Pyx_CyFunction_GetClosure(f)\
    (((__pyx_CyFunctionObject *) (f))->func_closure)
#if PY_VERSION_HEX < 0x030900B1 || CYTHON_COMPILING_IN_LIMITED_API
  #define __Pyx_CyFunction_GetClassObj(f)\
      (((__pyx_CyFunctionObject *) (f))->func_classobj)
#else
  #define __Pyx_CyFunction_GetClassObj(f)\
      ((PyObject*) ((PyCMethodObject *) (f))->mm_class)
#endif
#define __Pyx_CyFunction_SetClassObj(f, classobj)\
    __Pyx__CyFunction_SetClassObj((__pyx_CyFunctionObject *) (f), (classobj))
#define __Pyx_CyFunction_Defaults(type, f)\
    ((type *)(((__pyx_CyFunctionObject *) (f))->defaults))
#define __Pyx_CyFunction_SetDefaultsGetter(f, g)\
    ((__pyx_CyFunctionObject *) (f))->defaults_getter = (g)
typedef struct {
#if CYTHON_COMPILING_IN_LIMITED_API
    PyObject_HEAD
    PyObject *func;
#elif PY_VERSION_HEX < 0x030900B1
    PyCFunctionObject func;
#else
    PyCMethodObject func;
#endif
#if CYTHON_BACKPORT_VECTORCALL
    __pyx_vectorcallfunc func_vectorcall;
#endif
#if PY_VERSION_HEX < 0x030500A0 || CYTHON_COMPILING_IN_LIMITED_API
    PyObject *func_weakreflist;
#endif
    PyObject *func_dict;
    PyObject *func_name;
    PyObject *func_qualname;
    PyObject *func_doc;
    PyObject *func_globals;
    PyObject *func_code;
    PyObject *func_closure;
#if PY_VERSION_HEX < 0x030900B1 || CYTHON_COMPILING_IN_LIMITED_API
    PyObject *func_classobj;
#endif
    void *defaults;
    int defaults_pyobjects;
    size_t defaults_size;
    int flags;
    PyObject *defaults_tuple;
    PyObject *defaults_kwdict;
    PyObject *(*defaults_getter)(PyObject *);
    PyObject *func_annotations;
    PyObject *func_is_coroutine;
} __pyx_CyFunctionObject;
#undef __Pyx_CyOrPyCFunction_Check
#define __Pyx_CyFunction_Check(obj)  __Pyx_TypeCheck(obj, __pyx_CyFunctionType)
#define __Pyx_CyOrPyCFunction_Check(obj)  __Pyx_TypeCheck2(obj, __pyx_CyFunctionType, &PyCFunction_Type)
#define __Pyx_CyFunction_CheckExact(obj)  __Pyx_IS_TYPE(obj, __pyx_CyFunctionType)
static CYTHON_INLINE int __Pyx__IsSameCyOrCFunction(PyObject *func, void *cfunc);
#undef __Pyx_IsSameCFunction
#define __Pyx_IsSameCFunction(func, cfunc)   __Pyx__IsSameCyOrCFunction(func, cfunc)
static PyObject *__Pyx_CyFunction_Init(__pyx_CyFunctionObject* op, PyMethodDef *ml,
                                      int flags, PyObject* qualname,
                                      PyObject *closure,
                                      PyObject *module, PyObject *globals,
                                      PyObject* code);
static CYTHON_INLINE void __Pyx__CyFunction_SetClassObj(__pyx_CyFunctionObject* f, PyObject* classobj);
static CYTHON_INLINE void *__Pyx_CyFunction_InitDefaults(PyObject *m,
                                                         size_t size,
                                                         int pyobjects);
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsTuple(PyObject *m,
                                                            PyObject *tuple);
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsKwDict(PyObject *m,
                                                             PyObject *dict);
static CYTHON_INLINE void __Pyx_CyFunction_SetAnnotationsDict(PyObject *m,
                                                              PyObject *dict);
static int __pyx_CyFunction_init(PyObject *module);
#if CYTHON_METH_FASTCALL
static PyObject * __Pyx_CyFunction_Vectorcall_NOARGS(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames);
static PyObject * __Pyx_CyFunction_Vectorcall_O(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames);
static PyObject * __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames);
static PyObject * __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS_METHOD(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames);
#if CYTHON_BACKPORT_VECTORCALL
#define __Pyx_CyFunction_func_vectorcall(f) (((__pyx_CyFunctionObject*)f)->func_vectorcall)
#else
#define __Pyx_CyFunction_func_vectorcall(f) (((PyCFunctionObject*)f)->vectorcall)
#endif
#endif

/* CythonFunction.proto */
static PyObject *__Pyx_CyFunction_New(PyMethodDef *ml,
                                      int flags, PyObject* qualname,
                                      PyObject *closure,
                                      PyObject *module, PyObject *globals,
                                      PyObject* code);

/* CLineInTraceback.proto */
#ifdef CYTHON_CLINE_IN_TRACEBACK
#define __Pyx_CLineForTraceback(tstate, c_line)  (((CYTHON_CLINE_IN_TRACEBACK)) ? c_line : 0)
#else
static int __Pyx_CLineForTraceback(PyThreadState *tstate, int c_line);
#endif

/* CodeObjectCache.proto */
#if !CYTHON_COMPILING_IN_LIMITED_API
typedef struct {
    PyCodeObject* code_object;
    int code_line;
} __Pyx_CodeObjectCacheEntry;
struct __Pyx_CodeObjectCache {
    int count;
    int max_count;
    __Pyx_CodeObjectCacheEntry* entries;
};
static struct __Pyx_CodeObjectCache __pyx_code_cache = {0,0,NULL};
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line);
static PyCodeObject *__pyx_find_code_object(int code_line);
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object);
#endif

/* AddTraceback.proto */
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename);

/* GCCDiagnostics.proto */
#if !defined(__INTEL_COMPILER) && defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 6))
#define __Pyx_HAS_GCC_DIAGNOSTIC
#endif

/* RealImag.proto */
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    #define __Pyx_CREAL(z) ((z).real())
    #define __Pyx_CIMAG(z) ((z).imag())
  #else
    #define __Pyx_CREAL(z) (__real__(z))
    #define __Pyx_CIMAG(z) (__imag__(z))
  #endif
#else
    #define __Pyx_CREAL(z) ((z).real)
    #define __Pyx_CIMAG(z) ((z).imag)
#endif
#if defined(__cplusplus) && CYTHON_CCOMPLEX\
        && (defined(_WIN32) || defined(__clang__) || (defined(__GNUC__) && (__GNUC__ >= 5 || __GNUC__ == 4 && __GNUC_MINOR__ >= 4 )) || __cplusplus >= 201103)
    #define __Pyx_SET_CREAL(z,x) ((z).real(x))
    #define __Pyx_SET_CIMAG(z,y) ((z).imag(y))
#else
    #define __Pyx_SET_CREAL(z,x) __Pyx_CREAL(z) = (x)
    #define __Pyx_SET_CIMAG(z,y) __Pyx_CIMAG(z) = (y)
#endif

/* Arithmetic.proto */
#if CYTHON_CCOMPLEX && (1) && (!0 || __cplusplus)
    #define __Pyx_c_eq_float(a, b)   ((a)==(b))
    #define __Pyx_c_sum_float(a, b)  ((a)+(b))
    #define __Pyx_c_diff_float(a, b) ((a)-(b))
    #define __Pyx_c_prod_float(a, b) ((a)*(b))
    #define __Pyx_c_quot_float(a, b) ((a)/(b))
    #define __Pyx_c_neg_float(a)     (-(a))
  #ifdef __cplusplus
    #define __Pyx_c_is_zero_float(z) ((z)==(float)0)
    #define __Pyx_c_conj_float(z)    (::std::conj(z))
    #if 1
        #define __Pyx_c_abs_float(z)     (::std::abs(z))
        #define __Pyx_c_pow_float(a, b)  (::std::pow(a, b))
    #endif
  #else
    #define __Pyx_c_is_zero_float(z) ((z)==0)
    #define __Pyx_c_conj_float(z)    (conjf(z))
    #if 1
        #define __Pyx_c_abs_float(z)     (cabsf(z))
        #define __Pyx_c_pow_float(a, b)  (cpowf(a, b))
    #endif
 #endif
#else
    static CYTHON_INLINE int __Pyx_c_eq_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_sum_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_diff_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_prod_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_quot_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_neg_float(__pyx_t_float_complex);
    static CYTHON_INLINE int __Pyx_c_is_zero_float(__pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_conj_float(__pyx_t_float_complex);
    #if 1
        static CYTHON_INLINE float __Pyx_c_abs_float(__pyx_t_float_complex);
        static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_pow_float(__pyx_t_float_complex, __pyx_t_float_complex);
    #endif
#endif

/* Arithmetic.proto */
#if CYTHON_CCOMPLEX && (1) && (!0 || __cplusplus)
    #define __Pyx_c_eq_double(a, b)   ((a)==(b))
    #define __Pyx_c_sum_double(a, b)  ((a)+(b))
    #define __Pyx_c_diff_double(a, b) ((a)-(b))
    #define __Pyx_c_prod_double(a, b) ((a)*(b))
    #define __Pyx_c_quot_double(a, b) ((a)/(b))
    #define __Pyx_c_neg_double(a)     (-(a))
  #ifdef __cplusplus
    #define __Pyx_c_is_zero_double(z) ((z)==(double)0)
    #define __Pyx_c_conj_double(z)    (::std::conj(z))
    #if 1
        #define __Pyx_c_abs_double(z)     (::std::abs(z))
        #define __Pyx_c_pow_double(a, b)  (::std::pow(a, b))
    #endif
  #else
    #define __Pyx_c_is_zero_double(z) ((z)==0)
    #define __Pyx_c_conj_double(z)    (conj(z))
    #if 1
        #define __Pyx_c_abs_double(z)     (cabs(z))
        #define __Pyx_c_pow_double(a, b)  (cpow(a, b))
    #endif
 #endif
#else
    static CYTHON_INLINE int __Pyx_c_eq_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_sum_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_diff_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_prod_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_quot_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_neg_double(__pyx_t_double_complex);
    static CYTHON_INLINE int __Pyx_c_is_zero_double(__pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_conj_double(__pyx_t_double_complex);
    #if 1
        static CYTHON_INLINE double __Pyx_c_abs_double(__pyx_t_double_complex);
        static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_pow_double(__pyx_t_double_complex, __pyx_t_double_complex);
    #endif
#endif

/* Arithmetic.proto */
#if CYTHON_CCOMPLEX && (1) && (!0 || __cplusplus)
    #define __Pyx_c_eq_long__double(a, b)   ((a)==(b))
    #define __Pyx_c_sum_long__double(a, b)  ((a)+(b))
    #define __Pyx_c_diff_long__double(a, b) ((a)-(b))
    #define __Pyx_c_prod_long__double(a, b) ((a)*(b))
    #define __Pyx_c_quot_long__double(a, b) ((a)/(b))
    #define __Pyx_c_neg_long__double(a)     (-(a))
  #ifdef __cplusplus
    #define __Pyx_c_is_zero_long__double(z) ((z)==(long double)0)
    #define __Pyx_c_conj_long__double(z)    (::std::conj(z))
    #if 1
        #define __Pyx_c_abs_long__double(z)     (::std::abs(z))
        #define __Pyx_c_pow_long__double(a, b)  (::std::pow(a, b))
    #endif
  #else
    #define __Pyx_c_is_zero_long__double(z) ((z)==0)
    #define __Pyx_c_conj_long__double(z)    (conjl(z))
    #if 1
        #define __Pyx_c_abs_long__double(z)     (cabsl(z))
        #define __Pyx_c_pow_long__double(a, b)  (cpowl(a, b))
    #endif
 #endif
#else
    static CYTHON_INLINE int __Pyx_c_eq_long__double(__pyx_t_long_double_complex, __pyx_t_long_double_complex);
    static CYTHON_INLINE __pyx_t_long_double_complex __Pyx_c_sum_long__double(__pyx_t_long_double_complex, __pyx_t_long_double_complex);
    static CYTHON_INLINE __pyx_t_long_double_complex __Pyx_c_diff_long__double(__pyx_t_long_double_complex, __pyx_t_long_double_complex);
    static CYTHON_INLINE __pyx_t_long_double_complex __Pyx_c_prod_long__double(__pyx_t_long_double_complex, __pyx_t_long_double_complex);
    static CYTHON_INLINE __pyx_t_long_double_complex __Pyx_c_quot_long__double(__pyx_t_long_double_complex, __pyx_t_long_double_complex);
    static CYTHON_INLINE __pyx_t_long_double_complex __Pyx_c_neg_long__double(__pyx_t_long_double_complex);
    static CYTHON_INLINE int __Pyx_c_is_zero_long__double(__pyx_t_long_double_complex);
    static CYTHON_INLINE __pyx_t_long_double_complex __Pyx_c_conj_long__double(__pyx_t_long_double_complex);
    #if 1
        static CYTHON_INLINE long double __Pyx_c_abs_long__double(__pyx_t_long_double_complex);
        static CYTHON_INLINE __pyx_t_long_double_complex __Pyx_c_pow_long__double(__pyx_t_long_double_complex, __pyx_t_long_double_complex);
    #endif
#endif

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value);

/* CIntFromPy.proto */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE PY_LONG_LONG __Pyx_PyInt_As_PY_LONG_LONG(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE unsigned int __Pyx_PyInt_As_unsigned_int(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE unsigned PY_LONG_LONG __Pyx_PyInt_As_unsigned_PY_LONG_LONG(PyObject *);

/* FormatTypeName.proto */
#if CYTHON_COMPILING_IN_LIMITED_API
typedef PyObject *__Pyx_TypeName;
#define __Pyx_FMT_TYPENAME "%U"
static __Pyx_TypeName __Pyx_PyType_GetName(PyTypeObject* tp);
#define __Pyx_DECREF_TypeName(obj) Py_XDECREF(obj)
#else
typedef const char *__Pyx_TypeName;
#define __Pyx_FMT_TYPENAME "%.200s"
#define __Pyx_PyType_GetName(tp) ((tp)->tp_name)
#define __Pyx_DECREF_TypeName(obj)
#endif

/* CIntFromPy.proto */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *);

/* FastTypeChecks.proto */
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_TypeCheck(obj, type) __Pyx_IsSubtype(Py_TYPE(obj), (PyTypeObject *)type)
#define __Pyx_TypeCheck2(obj, type1, type2) __Pyx_IsAnySubtype2(Py_TYPE(obj), (PyTypeObject *)type1, (PyTypeObject *)type2)
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b);
static CYTHON_INLINE int __Pyx_IsAnySubtype2(PyTypeObject *cls, PyTypeObject *a, PyTypeObject *b);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject *type);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *type1, PyObject *type2);
#else
#define __Pyx_TypeCheck(obj, type) PyObject_TypeCheck(obj, (PyTypeObject *)type)
#define __Pyx_TypeCheck2(obj, type1, type2) (PyObject_TypeCheck(obj, (PyTypeObject *)type1) || PyObject_TypeCheck(obj, (PyTypeObject *)type2))
#define __Pyx_PyErr_GivenExceptionMatches(err, type) PyErr_GivenExceptionMatches(err, type)
#define __Pyx_PyErr_GivenExceptionMatches2(err, type1, type2) (PyErr_GivenExceptionMatches(err, type1) || PyErr_GivenExceptionMatches(err, type2))
#endif
#define __Pyx_PyErr_ExceptionMatches2(err1, err2)  __Pyx_PyErr_GivenExceptionMatches2(__Pyx_PyErr_CurrentExceptionType(), err1, err2)
#define __Pyx_PyException_Check(obj) __Pyx_TypeCheck(obj, PyExc_Exception)

/* CheckBinaryVersion.proto */
static unsigned long __Pyx_get_runtime_version(void);
static int __Pyx_check_binary_version(unsigned long ct_version, unsigned long rt_version, int allow_newer);

/* PyObjectSetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
#define __Pyx_PyObject_DelAttrStr(o,n) __Pyx_PyObject_SetAttrStr(o, n, NULL)
static CYTHON_INLINE int __Pyx_PyObject_SetAttrStr(PyObject* obj, PyObject* attr_name, PyObject* value);
#else
#define __Pyx_PyObject_DelAttrStr(o,n)   PyObject_DelAttr(o,n)
#define __Pyx_PyObject_SetAttrStr(o,n,v) PyObject_SetAttr(o,n,v)
#endif

/* VoidPtrExport.proto */
static int __Pyx_ExportVoidPtr(PyObject *name, void *p, const char *sig);

/* FunctionExport.proto */
static int __Pyx_ExportFunction(const char *name, void (*f)(void), const char *sig);

/* VoidPtrImport.proto */
static int __Pyx_ImportVoidPtr_3_0_12(PyObject *module, const char *name, void **p, const char *sig);

/* InitStrings.proto */
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t);

/* #### Code section: module_declarations ### */
static CYTHON_INLINE npy_intp __pyx_f_5numpy_5dtype_8itemsize_itemsize(PyArray_Descr *__pyx_v_self); /* proto*/
static CYTHON_INLINE npy_intp __pyx_f_5numpy_5dtype_9alignment_alignment(PyArray_Descr *__pyx_v_self); /* proto*/
static CYTHON_INLINE PyObject *__pyx_f_5numpy_5dtype_6fields_fields(PyArray_Descr *__pyx_v_self); /* proto*/
static CYTHON_INLINE PyObject *__pyx_f_5numpy_5dtype_5names_names(PyArray_Descr *__pyx_v_self); /* proto*/
static CYTHON_INLINE PyArray_ArrayDescr *__pyx_f_5numpy_5dtype_8subarray_subarray(PyArray_Descr *__pyx_v_self); /* proto*/
static CYTHON_INLINE npy_uint64 __pyx_f_5numpy_5dtype_5flags_flags(PyArray_Descr *__pyx_v_self); /* proto*/
static CYTHON_INLINE int __pyx_f_5numpy_9broadcast_7numiter_numiter(PyArrayMultiIterObject *__pyx_v_self); /* proto*/
static CYTHON_INLINE npy_intp __pyx_f_5numpy_9broadcast_4size_size(PyArrayMultiIterObject *__pyx_v_self); /* proto*/
static CYTHON_INLINE npy_intp __pyx_f_5numpy_9broadcast_5index_index(PyArrayMultiIterObject *__pyx_v_self); /* proto*/
static CYTHON_INLINE int __pyx_f_5numpy_9broadcast_2nd_nd(PyArrayMultiIterObject *__pyx_v_self); /* proto*/
static CYTHON_INLINE npy_intp *__pyx_f_5numpy_9broadcast_10dimensions_dimensions(PyArrayMultiIterObject *__pyx_v_self); /* proto*/
static CYTHON_INLINE void **__pyx_f_5numpy_9broadcast_5iters_iters(PyArrayMultiIterObject *__pyx_v_self); /* proto*/
static CYTHON_INLINE PyObject *__pyx_f_5numpy_7ndarray_4base_base(PyArrayObject *__pyx_v_self); /* proto*/
static CYTHON_INLINE PyArray_Descr *__pyx_f_5numpy_7ndarray_5descr_descr(PyArrayObject *__pyx_v_self); /* proto*/
static CYTHON_INLINE int __pyx_f_5numpy_7ndarray_4ndim_ndim(PyArrayObject *__pyx_v_self); /* proto*/
static CYTHON_INLINE npy_intp *__pyx_f_5numpy_7ndarray_5shape_shape(PyArrayObject *__pyx_v_self); /* proto*/
static CYTHON_INLINE npy_intp *__pyx_f_5numpy_7ndarray_7strides_strides(PyArrayObject *__pyx_v_self); /* proto*/
static CYTHON_INLINE npy_intp __pyx_f_5numpy_7ndarray_4size_size(PyArrayObject *__pyx_v_self); /* proto*/
static CYTHON_INLINE char *__pyx_f_5numpy_7ndarray_4data_data(PyArrayObject *__pyx_v_self); /* proto*/
static CYTHON_INLINE double __pyx_f_7cpython_7complex_7complex_4real_real(PyComplexObject *__pyx_v_self); /* proto*/
static CYTHON_INLINE double __pyx_f_7cpython_7complex_7complex_4imag_imag(PyComplexObject *__pyx_v_self); /* proto*/
static CYTHON_INLINE int __pyx_f_7cpython_8datetime_4date_4year_year(PyDateTime_Date *__pyx_v_self); /* proto*/
static CYTHON_INLINE int __pyx_f_7cpython_8datetime_4date_5month_month(PyDateTime_Date *__pyx_v_self); /* proto*/
static CYTHON_INLINE int __pyx_f_7cpython_8datetime_4date_3day_day(PyDateTime_Date *__pyx_v_self); /* proto*/
static CYTHON_INLINE int __pyx_f_7cpython_8datetime_4time_4hour_hour(PyDateTime_Time *__pyx_v_self); /* proto*/
static CYTHON_INLINE int __pyx_f_7cpython_8datetime_4time_6minute_minute(PyDateTime_Time *__pyx_v_self); /* proto*/
static CYTHON_INLINE int __pyx_f_7cpython_8datetime_4time_6second_second(PyDateTime_Time *__pyx_v_self); /* proto*/
static CYTHON_INLINE int __pyx_f_7cpython_8datetime_4time_11microsecond_microsecond(PyDateTime_Time *__pyx_v_self); /* proto*/
static CYTHON_INLINE PyObject *__pyx_f_7cpython_8datetime_4time_6tzinfo_tzinfo(PyDateTime_Time *__pyx_v_self); /* proto*/
static CYTHON_INLINE int __pyx_f_7cpython_8datetime_4time_4fold_fold(PyDateTime_Time *__pyx_v_self); /* proto*/
static CYTHON_INLINE int __pyx_f_7cpython_8datetime_8datetime_4year_year(PyDateTime_DateTime *__pyx_v_self); /* proto*/
static CYTHON_INLINE int __pyx_f_7cpython_8datetime_8datetime_5month_month(PyDateTime_DateTime *__pyx_v_self); /* proto*/
static CYTHON_INLINE int __pyx_f_7cpython_8datetime_8datetime_3day_day(PyDateTime_DateTime *__pyx_v_self); /* proto*/
static CYTHON_INLINE int __pyx_f_7cpython_8datetime_8datetime_4hour_hour(PyDateTime_DateTime *__pyx_v_self); /* proto*/
static CYTHON_INLINE int __pyx_f_7cpython_8datetime_8datetime_6minute_minute(PyDateTime_DateTime *__pyx_v_self); /* proto*/
static CYTHON_INLINE int __pyx_f_7cpython_8datetime_8datetime_6second_second(PyDateTime_DateTime *__pyx_v_self); /* proto*/
static CYTHON_INLINE int __pyx_f_7cpython_8datetime_8datetime_11microsecond_microsecond(PyDateTime_DateTime *__pyx_v_self); /* proto*/
static CYTHON_INLINE PyObject *__pyx_f_7cpython_8datetime_8datetime_6tzinfo_tzinfo(PyDateTime_DateTime *__pyx_v_self); /* proto*/
static CYTHON_INLINE int __pyx_f_7cpython_8datetime_8datetime_4fold_fold(PyDateTime_DateTime *__pyx_v_self); /* proto*/
static CYTHON_INLINE int __pyx_f_7cpython_8datetime_9timedelta_3day_day(PyDateTime_Delta *__pyx_v_self); /* proto*/
static CYTHON_INLINE int __pyx_f_7cpython_8datetime_9timedelta_6second_second(PyDateTime_Delta *__pyx_v_self); /* proto*/
static CYTHON_INLINE int __pyx_f_7cpython_8datetime_9timedelta_11microsecond_microsecond(PyDateTime_Delta *__pyx_v_self); /* proto*/

/* Module declarations from "libc.limits" */

/* Module declarations from "libc.string" */

/* Module declarations from "libc.stdio" */

/* Module declarations from "__builtin__" */

/* Module declarations from "cpython.type" */

/* Module declarations from "cpython.version" */

/* Module declarations from "cpython.ref" */

/* Module declarations from "cpython.exc" */

/* Module declarations from "cpython.module" */

/* Module declarations from "cpython.mem" */

/* Module declarations from "cpython.tuple" */

/* Module declarations from "cpython.list" */

/* Module declarations from "cpython.sequence" */

/* Module declarations from "cpython.mapping" */

/* Module declarations from "cpython.iterator" */

/* Module declarations from "cpython.number" */

/* Module declarations from "cpython.int" */

/* Module declarations from "__builtin__" */

/* Module declarations from "cpython.bool" */

/* Module declarations from "cpython.long" */

/* Module declarations from "cpython.float" */

/* Module declarations from "__builtin__" */

/* Module declarations from "cpython.complex" */

/* Module declarations from "cpython.string" */

/* Module declarations from "libc.stddef" */

/* Module declarations from "cpython.unicode" */

/* Module declarations from "cpython.pyport" */

/* Module declarations from "cpython.dict" */

/* Module declarations from "cpython.instance" */

/* Module declarations from "cpython.function" */

/* Module declarations from "cpython.method" */

/* Module declarations from "cpython.weakref" */

/* Module declarations from "cpython.getargs" */

/* Module declarations from "cpython.pythread" */

/* Module declarations from "cpython.pystate" */

/* Module declarations from "cpython.cobject" */

/* Module declarations from "cpython.oldbuffer" */

/* Module declarations from "cpython.set" */

/* Module declarations from "cpython.buffer" */

/* Module declarations from "cpython.pycapsule" */

/* Module declarations from "cpython.contextvars" */

/* Module declarations from "cpython" */

/* Module declarations from "cpython.object" */

/* Module declarations from "cpython.bytes" */

/* Module declarations from "cython" */

/* Module declarations from "libc.math" */

/* Module declarations from "numpy" */

/* Module declarations from "numpy" */
static CYTHON_INLINE int __pyx_f_5numpy_import_array(void); /*proto*/
static CYTHON_INLINE int __pyx_f_5numpy_import_umath(void); /*proto*/
static CYTHON_INLINE npy_datetime __pyx_f_5numpy_get_datetime64_value(PyObject *); /*proto*/
static CYTHON_INLINE npy_timedelta __pyx_f_5numpy_get_timedelta64_value(PyObject *); /*proto*/
static CYTHON_INLINE NPY_DATETIMEUNIT __pyx_f_5numpy_get_datetime64_unit(PyObject *); /*proto*/

/* Module declarations from "datetime" */

/* Module declarations from "cpython.datetime" */
static CYTHON_INLINE void __pyx_f_7cpython_8datetime_import_datetime(void); /*proto*/
static CYTHON_INLINE PyObject *__pyx_f_7cpython_8datetime_time_tzinfo(PyObject *); /*proto*/
static CYTHON_INLINE PyObject *__pyx_f_7cpython_8datetime_datetime_tzinfo(PyObject *); /*proto*/
static CYTHON_INLINE int __pyx_f_7cpython_8datetime_date_year(PyObject *); /*proto*/
static CYTHON_INLINE int __pyx_f_7cpython_8datetime_date_month(PyObject *); /*proto*/
static CYTHON_INLINE int __pyx_f_7cpython_8datetime_date_day(PyObject *); /*proto*/
static CYTHON_INLINE int __pyx_f_7cpython_8datetime_datetime_year(PyObject *); /*proto*/
static CYTHON_INLINE int __pyx_f_7cpython_8datetime_datetime_month(PyObject *); /*proto*/
static CYTHON_INLINE int __pyx_f_7cpython_8datetime_datetime_day(PyObject *); /*proto*/
static CYTHON_INLINE int __pyx_f_7cpython_8datetime_time_hour(PyObject *); /*proto*/
static CYTHON_INLINE int __pyx_f_7cpython_8datetime_time_minute(PyObject *); /*proto*/
static CYTHON_INLINE int __pyx_f_7cpython_8datetime_time_second(PyObject *); /*proto*/
static CYTHON_INLINE int __pyx_f_7cpython_8datetime_time_microsecond(PyObject *); /*proto*/
static CYTHON_INLINE int __pyx_f_7cpython_8datetime_datetime_hour(PyObject *); /*proto*/
static CYTHON_INLINE int __pyx_f_7cpython_8datetime_datetime_minute(PyObject *); /*proto*/
static CYTHON_INLINE int __pyx_f_7cpython_8datetime_datetime_second(PyObject *); /*proto*/
static CYTHON_INLINE int __pyx_f_7cpython_8datetime_datetime_microsecond(PyObject *); /*proto*/
static CYTHON_INLINE int __pyx_f_7cpython_8datetime_timedelta_days(PyObject *); /*proto*/
static CYTHON_INLINE int __pyx_f_7cpython_8datetime_timedelta_seconds(PyObject *); /*proto*/
static CYTHON_INLINE int __pyx_f_7cpython_8datetime_timedelta_microseconds(PyObject *); /*proto*/

/* Module declarations from "serializor" */

/* Module declarations from "serializor.identifier" */
static char *__pyx_vp_10serializor_10identifier_STR = 0;
#define __pyx_v_10serializor_10identifier_STR (*__pyx_vp_10serializor_10identifier_STR)
static char *__pyx_vp_10serializor_10identifier_BOOL = 0;
#define __pyx_v_10serializor_10identifier_BOOL (*__pyx_vp_10serializor_10identifier_BOOL)
static char *__pyx_vp_10serializor_10identifier_FLOAT = 0;
#define __pyx_v_10serializor_10identifier_FLOAT (*__pyx_vp_10serializor_10identifier_FLOAT)
static char *__pyx_vp_10serializor_10identifier_INT = 0;
#define __pyx_v_10serializor_10identifier_INT (*__pyx_vp_10serializor_10identifier_INT)
static char *__pyx_vp_10serializor_10identifier_NONE = 0;
#define __pyx_v_10serializor_10identifier_NONE (*__pyx_vp_10serializor_10identifier_NONE)
static char *__pyx_vp_10serializor_10identifier_DATE = 0;
#define __pyx_v_10serializor_10identifier_DATE (*__pyx_vp_10serializor_10identifier_DATE)
static char *__pyx_vp_10serializor_10identifier_TIME = 0;
#define __pyx_v_10serializor_10identifier_TIME (*__pyx_vp_10serializor_10identifier_TIME)
static char *__pyx_vp_10serializor_10identifier_DATETIME = 0;
#define __pyx_v_10serializor_10identifier_DATETIME (*__pyx_vp_10serializor_10identifier_DATETIME)
static char *__pyx_vp_10serializor_10identifier_TIMEDELTA = 0;
#define __pyx_v_10serializor_10identifier_TIMEDELTA (*__pyx_vp_10serializor_10identifier_TIMEDELTA)
static char *__pyx_vp_10serializor_10identifier_STRUCT_TIME = 0;
#define __pyx_v_10serializor_10identifier_STRUCT_TIME (*__pyx_vp_10serializor_10identifier_STRUCT_TIME)
static char *__pyx_vp_10serializor_10identifier_DECIMAL = 0;
#define __pyx_v_10serializor_10identifier_DECIMAL (*__pyx_vp_10serializor_10identifier_DECIMAL)
static char *__pyx_vp_10serializor_10identifier_COMPLEX = 0;
#define __pyx_v_10serializor_10identifier_COMPLEX (*__pyx_vp_10serializor_10identifier_COMPLEX)
static char *__pyx_vp_10serializor_10identifier_BYTES = 0;
#define __pyx_v_10serializor_10identifier_BYTES (*__pyx_vp_10serializor_10identifier_BYTES)
static char *__pyx_vp_10serializor_10identifier_BYTEARRAY = 0;
#define __pyx_v_10serializor_10identifier_BYTEARRAY (*__pyx_vp_10serializor_10identifier_BYTEARRAY)
static char *__pyx_vp_10serializor_10identifier_LIST = 0;
#define __pyx_v_10serializor_10identifier_LIST (*__pyx_vp_10serializor_10identifier_LIST)
static char *__pyx_vp_10serializor_10identifier_TUPLE = 0;
#define __pyx_v_10serializor_10identifier_TUPLE (*__pyx_vp_10serializor_10identifier_TUPLE)
static char *__pyx_vp_10serializor_10identifier_SET = 0;
#define __pyx_v_10serializor_10identifier_SET (*__pyx_vp_10serializor_10identifier_SET)
static char *__pyx_vp_10serializor_10identifier_FROZENSET = 0;
#define __pyx_v_10serializor_10identifier_FROZENSET (*__pyx_vp_10serializor_10identifier_FROZENSET)
static char *__pyx_vp_10serializor_10identifier_RANGE = 0;
#define __pyx_v_10serializor_10identifier_RANGE (*__pyx_vp_10serializor_10identifier_RANGE)
static char *__pyx_vp_10serializor_10identifier_DICT = 0;
#define __pyx_v_10serializor_10identifier_DICT (*__pyx_vp_10serializor_10identifier_DICT)
static char *__pyx_vp_10serializor_10identifier_DATETIME64 = 0;
#define __pyx_v_10serializor_10identifier_DATETIME64 (*__pyx_vp_10serializor_10identifier_DATETIME64)
static char *__pyx_vp_10serializor_10identifier_TIMEDELTA64 = 0;
#define __pyx_v_10serializor_10identifier_TIMEDELTA64 (*__pyx_vp_10serializor_10identifier_TIMEDELTA64)
static char *__pyx_vp_10serializor_10identifier_NDARRAY = 0;
#define __pyx_v_10serializor_10identifier_NDARRAY (*__pyx_vp_10serializor_10identifier_NDARRAY)
static char *__pyx_vp_10serializor_10identifier_NDARRAY_OBJECT = 0;
#define __pyx_v_10serializor_10identifier_NDARRAY_OBJECT (*__pyx_vp_10serializor_10identifier_NDARRAY_OBJECT)
static char *__pyx_vp_10serializor_10identifier_NDARRAY_INT = 0;
#define __pyx_v_10serializor_10identifier_NDARRAY_INT (*__pyx_vp_10serializor_10identifier_NDARRAY_INT)
static char *__pyx_vp_10serializor_10identifier_NDARRAY_UINT = 0;
#define __pyx_v_10serializor_10identifier_NDARRAY_UINT (*__pyx_vp_10serializor_10identifier_NDARRAY_UINT)
static char *__pyx_vp_10serializor_10identifier_NDARRAY_FLOAT = 0;
#define __pyx_v_10serializor_10identifier_NDARRAY_FLOAT (*__pyx_vp_10serializor_10identifier_NDARRAY_FLOAT)
static char *__pyx_vp_10serializor_10identifier_NDARRAY_BOOL = 0;
#define __pyx_v_10serializor_10identifier_NDARRAY_BOOL (*__pyx_vp_10serializor_10identifier_NDARRAY_BOOL)
static char *__pyx_vp_10serializor_10identifier_NDARRAY_DT64 = 0;
#define __pyx_v_10serializor_10identifier_NDARRAY_DT64 (*__pyx_vp_10serializor_10identifier_NDARRAY_DT64)
static char *__pyx_vp_10serializor_10identifier_NDARRAY_TD64 = 0;
#define __pyx_v_10serializor_10identifier_NDARRAY_TD64 (*__pyx_vp_10serializor_10identifier_NDARRAY_TD64)
static char *__pyx_vp_10serializor_10identifier_NDARRAY_COMPLEX = 0;
#define __pyx_v_10serializor_10identifier_NDARRAY_COMPLEX (*__pyx_vp_10serializor_10identifier_NDARRAY_COMPLEX)
static char *__pyx_vp_10serializor_10identifier_NDARRAY_BYTES = 0;
#define __pyx_v_10serializor_10identifier_NDARRAY_BYTES (*__pyx_vp_10serializor_10identifier_NDARRAY_BYTES)
static char *__pyx_vp_10serializor_10identifier_NDARRAY_UNICODE = 0;
#define __pyx_v_10serializor_10identifier_NDARRAY_UNICODE (*__pyx_vp_10serializor_10identifier_NDARRAY_UNICODE)
static char *__pyx_vp_10serializor_10identifier_SERIES = 0;
#define __pyx_v_10serializor_10identifier_SERIES (*__pyx_vp_10serializor_10identifier_SERIES)
static char *__pyx_vp_10serializor_10identifier_DATAFRAME = 0;
#define __pyx_v_10serializor_10identifier_DATAFRAME (*__pyx_vp_10serializor_10identifier_DATAFRAME)
static char *__pyx_vp_10serializor_10identifier_PD_TIMESTAMP = 0;
#define __pyx_v_10serializor_10identifier_PD_TIMESTAMP (*__pyx_vp_10serializor_10identifier_PD_TIMESTAMP)
static char *__pyx_vp_10serializor_10identifier_PD_TIMEDELTA = 0;
#define __pyx_v_10serializor_10identifier_PD_TIMEDELTA (*__pyx_vp_10serializor_10identifier_PD_TIMEDELTA)
static char *__pyx_vp_10serializor_10identifier_DATETIMEINDEX = 0;
#define __pyx_v_10serializor_10identifier_DATETIMEINDEX (*__pyx_vp_10serializor_10identifier_DATETIMEINDEX)
static char *__pyx_vp_10serializor_10identifier_TIMEDELTAINDEX = 0;
#define __pyx_v_10serializor_10identifier_TIMEDELTAINDEX (*__pyx_vp_10serializor_10identifier_TIMEDELTAINDEX)
static int *__pyx_vp_10serializor_10identifier__duplicates_test = 0;
#define __pyx_v_10serializor_10identifier__duplicates_test (*__pyx_vp_10serializor_10identifier__duplicates_test)

/* Module declarations from "serializor.typeref" */
static PyObject **__pyx_vp_10serializor_7typeref_NONE = 0;
#define __pyx_v_10serializor_7typeref_NONE (*__pyx_vp_10serializor_7typeref_NONE)
static PyObject **__pyx_vp_10serializor_7typeref_DICT_KEYS = 0;
#define __pyx_v_10serializor_7typeref_DICT_KEYS (*__pyx_vp_10serializor_7typeref_DICT_KEYS)
static PyObject **__pyx_vp_10serializor_7typeref_DICT_VALUES = 0;
#define __pyx_v_10serializor_7typeref_DICT_VALUES (*__pyx_vp_10serializor_7typeref_DICT_VALUES)
static PyObject **__pyx_vp_10serializor_7typeref_DICT_ITEMS = 0;
#define __pyx_v_10serializor_7typeref_DICT_ITEMS (*__pyx_vp_10serializor_7typeref_DICT_ITEMS)
static PyObject **__pyx_vp_10serializor_7typeref_DECIMAL = 0;
#define __pyx_v_10serializor_7typeref_DECIMAL (*__pyx_vp_10serializor_7typeref_DECIMAL)
static PyObject **__pyx_vp_10serializor_7typeref_STRUCT_TIME = 0;
#define __pyx_v_10serializor_7typeref_STRUCT_TIME (*__pyx_vp_10serializor_7typeref_STRUCT_TIME)
static PyObject **__pyx_vp_10serializor_7typeref_DEQUE = 0;
#define __pyx_v_10serializor_7typeref_DEQUE (*__pyx_vp_10serializor_7typeref_DEQUE)
static PyObject **__pyx_vp_10serializor_7typeref_FLOAT16 = 0;
#define __pyx_v_10serializor_7typeref_FLOAT16 (*__pyx_vp_10serializor_7typeref_FLOAT16)
static PyObject **__pyx_vp_10serializor_7typeref_FLOAT32 = 0;
#define __pyx_v_10serializor_7typeref_FLOAT32 (*__pyx_vp_10serializor_7typeref_FLOAT32)
static PyObject **__pyx_vp_10serializor_7typeref_FLOAT64 = 0;
#define __pyx_v_10serializor_7typeref_FLOAT64 (*__pyx_vp_10serializor_7typeref_FLOAT64)
static PyObject **__pyx_vp_10serializor_7typeref_INT8 = 0;
#define __pyx_v_10serializor_7typeref_INT8 (*__pyx_vp_10serializor_7typeref_INT8)
static PyObject **__pyx_vp_10serializor_7typeref_INT16 = 0;
#define __pyx_v_10serializor_7typeref_INT16 (*__pyx_vp_10serializor_7typeref_INT16)
static PyObject **__pyx_vp_10serializor_7typeref_INT32 = 0;
#define __pyx_v_10serializor_7typeref_INT32 (*__pyx_vp_10serializor_7typeref_INT32)
static PyObject **__pyx_vp_10serializor_7typeref_INT64 = 0;
#define __pyx_v_10serializor_7typeref_INT64 (*__pyx_vp_10serializor_7typeref_INT64)
static PyObject **__pyx_vp_10serializor_7typeref_UINT8 = 0;
#define __pyx_v_10serializor_7typeref_UINT8 (*__pyx_vp_10serializor_7typeref_UINT8)
static PyObject **__pyx_vp_10serializor_7typeref_UINT16 = 0;
#define __pyx_v_10serializor_7typeref_UINT16 (*__pyx_vp_10serializor_7typeref_UINT16)
static PyObject **__pyx_vp_10serializor_7typeref_UINT32 = 0;
#define __pyx_v_10serializor_7typeref_UINT32 (*__pyx_vp_10serializor_7typeref_UINT32)
static PyObject **__pyx_vp_10serializor_7typeref_UINT64 = 0;
#define __pyx_v_10serializor_7typeref_UINT64 (*__pyx_vp_10serializor_7typeref_UINT64)
static PyObject **__pyx_vp_10serializor_7typeref_COMPLEX64 = 0;
#define __pyx_v_10serializor_7typeref_COMPLEX64 (*__pyx_vp_10serializor_7typeref_COMPLEX64)
static PyObject **__pyx_vp_10serializor_7typeref_COMPLEX128 = 0;
#define __pyx_v_10serializor_7typeref_COMPLEX128 (*__pyx_vp_10serializor_7typeref_COMPLEX128)
static PyObject **__pyx_vp_10serializor_7typeref_DATETIME64 = 0;
#define __pyx_v_10serializor_7typeref_DATETIME64 (*__pyx_vp_10serializor_7typeref_DATETIME64)
static PyObject **__pyx_vp_10serializor_7typeref_TIMEDELTA64 = 0;
#define __pyx_v_10serializor_7typeref_TIMEDELTA64 (*__pyx_vp_10serializor_7typeref_TIMEDELTA64)
static PyObject **__pyx_vp_10serializor_7typeref_STR_ = 0;
#define __pyx_v_10serializor_7typeref_STR_ (*__pyx_vp_10serializor_7typeref_STR_)
static PyObject **__pyx_vp_10serializor_7typeref_BOOL_ = 0;
#define __pyx_v_10serializor_7typeref_BOOL_ (*__pyx_vp_10serializor_7typeref_BOOL_)
static PyObject **__pyx_vp_10serializor_7typeref_BYTES_ = 0;
#define __pyx_v_10serializor_7typeref_BYTES_ (*__pyx_vp_10serializor_7typeref_BYTES_)
static PyObject **__pyx_vp_10serializor_7typeref_RECORD = 0;
#define __pyx_v_10serializor_7typeref_RECORD (*__pyx_vp_10serializor_7typeref_RECORD)
static PyObject **__pyx_vp_10serializor_7typeref_SERIES = 0;
#define __pyx_v_10serializor_7typeref_SERIES (*__pyx_vp_10serializor_7typeref_SERIES)
static PyObject **__pyx_vp_10serializor_7typeref_DATAFRAME = 0;
#define __pyx_v_10serializor_7typeref_DATAFRAME (*__pyx_vp_10serializor_7typeref_DATAFRAME)
static PyObject **__pyx_vp_10serializor_7typeref_PD_TIMESTAMP = 0;
#define __pyx_v_10serializor_7typeref_PD_TIMESTAMP (*__pyx_vp_10serializor_7typeref_PD_TIMESTAMP)
static PyObject **__pyx_vp_10serializor_7typeref_PD_TIMEDELTA = 0;
#define __pyx_v_10serializor_7typeref_PD_TIMEDELTA (*__pyx_vp_10serializor_7typeref_PD_TIMEDELTA)
static PyObject **__pyx_vp_10serializor_7typeref_DATETIMEINDEX = 0;
#define __pyx_v_10serializor_7typeref_DATETIMEINDEX (*__pyx_vp_10serializor_7typeref_DATETIMEINDEX)
static PyObject **__pyx_vp_10serializor_7typeref_TIMEDELTAINDEX = 0;
#define __pyx_v_10serializor_7typeref_TIMEDELTAINDEX (*__pyx_vp_10serializor_7typeref_TIMEDELTAINDEX)
static int *__pyx_vp_10serializor_7typeref_CYTIMES_AVAILABLE = 0;
#define __pyx_v_10serializor_7typeref_CYTIMES_AVAILABLE (*__pyx_vp_10serializor_7typeref_CYTIMES_AVAILABLE)
static PyObject **__pyx_vp_10serializor_7typeref_PYDT = 0;
#define __pyx_v_10serializor_7typeref_PYDT (*__pyx_vp_10serializor_7typeref_PYDT)
static PyObject **__pyx_vp_10serializor_7typeref_PDDT = 0;
#define __pyx_v_10serializor_7typeref_PDDT (*__pyx_vp_10serializor_7typeref_PDDT)

/* Module declarations from "serializor.utils" */
static PyObject **__pyx_vp_10serializor_5utils_AVAILABLE_TIMEZONES = 0;
#define __pyx_v_10serializor_5utils_AVAILABLE_TIMEZONES (*__pyx_vp_10serializor_5utils_AVAILABLE_TIMEZONES)
static CYTHON_INLINE PyObject *__pyx_f_10serializor_5utils_encode_str(PyObject *, char *); /*proto*/
static CYTHON_INLINE PyObject *__pyx_f_10serializor_5utils_arr_getitem_1d(PyArrayObject *, npy_intp); /*proto*/
static CYTHON_INLINE PyObject *__pyx_f_10serializor_5utils_arr_getitem_2d(PyArrayObject *, npy_intp, npy_intp); /*proto*/
static CYTHON_INLINE PyObject *__pyx_f_10serializor_5utils_arr_getitem_3d(PyArrayObject *, npy_intp, npy_intp, npy_intp); /*proto*/
static CYTHON_INLINE PyObject *__pyx_f_10serializor_5utils_arr_getitem_4d(PyArrayObject *, npy_intp, npy_intp, npy_intp, npy_intp); /*proto*/
static CYTHON_INLINE PyArrayObject *__pyx_f_10serializor_5utils_arr_flatten(PyArrayObject *); /*proto*/
static CYTHON_INLINE int __pyx_f_10serializor_5utils_map_nptime_unit_str2int(PyObject *); /*proto*/
static CYTHON_INLINE int __pyx_f_10serializor_5utils_parse_arr_nptime_unit(PyArrayObject *); /*proto*/

/* Module declarations from "serializor.binary.ser" */
static PyObject *__pyx_v_10serializor_6binary_3ser_OPT_SERIALIZE_NUMPY = 0;
static unsigned char __pyx_v_10serializor_6binary_3ser_UINT8_ENCODE_VALUE;
static unsigned char __pyx_v_10serializor_6binary_3ser_UINT16_ENCODE_VALUE;
static unsigned char __pyx_v_10serializor_6binary_3ser_UINT32_ENCODE_VALUE;
static unsigned char __pyx_v_10serializor_6binary_3ser_UINT64_ENCODE_VALUE;
static PyObject *__pyx_v_10serializor_6binary_3ser_DATETIME_ID = 0;
static PyObject *__pyx_v_10serializor_6binary_3ser_DATE_ID = 0;
static PyObject *__pyx_v_10serializor_6binary_3ser_TIME_ID = 0;
static PyObject *__pyx_v_10serializor_6binary_3ser_TIMEDELTA_ID = 0;
static PyObject *__pyx_v_10serializor_6binary_3ser_STRUCT_TIME_ID = 0;
static PyObject *__pyx_v_10serializor_6binary_3ser_COMPLEX_ID = 0;
static PyObject *__pyx_v_10serializor_6binary_3ser_LIST_ID = 0;
static PyObject *__pyx_v_10serializor_6binary_3ser_TUPLE_ID = 0;
static PyObject *__pyx_v_10serializor_6binary_3ser_SET_ID = 0;
static PyObject *__pyx_v_10serializor_6binary_3ser_FROZENSET_ID = 0;
static PyObject *__pyx_v_10serializor_6binary_3ser_RANGE_ID = 0;
static PyObject *__pyx_v_10serializor_6binary_3ser_DICT_ID = 0;
static PyObject *__pyx_v_10serializor_6binary_3ser_DT64_ID = 0;
static PyObject *__pyx_v_10serializor_6binary_3ser_TD64_ID = 0;
static PyObject *__pyx_v_10serializor_6binary_3ser_NDARRAY_ID = 0;
static PyObject *__pyx_v_10serializor_6binary_3ser_NDARRAY_OBJECT_DT = 0;
static PyObject *__pyx_v_10serializor_6binary_3ser_NDARRAY_INT_DT = 0;
static PyObject *__pyx_v_10serializor_6binary_3ser_NDARRAY_UINT_DT = 0;
static PyObject *__pyx_v_10serializor_6binary_3ser_NDARRAY_FLOAT_DT = 0;
static PyObject *__pyx_v_10serializor_6binary_3ser_NDARRAY_BOOL_DT = 0;
static PyObject *__pyx_v_10serializor_6binary_3ser_NDARRAY_DT64_DT = 0;
static PyObject *__pyx_v_10serializor_6binary_3ser_NDARRAY_TD64_DT = 0;
static PyObject *__pyx_v_10serializor_6binary_3ser_NDARRAY_COMPLEX_DT = 0;
static PyObject *__pyx_v_10serializor_6binary_3ser_NDARRAY_BYTES_DT = 0;
static PyObject *__pyx_v_10serializor_6binary_3ser_NDARRAY_UNICODE_DT = 0;
static PyObject *__pyx_v_10serializor_6binary_3ser_NDARRAY_OBJECT_IDDT = 0;
static PyObject *__pyx_v_10serializor_6binary_3ser_NDARRAY_INT_IDDT = 0;
static PyObject *__pyx_v_10serializor_6binary_3ser_NDARRAY_UINT_IDDT = 0;
static PyObject *__pyx_v_10serializor_6binary_3ser_NDARRAY_FLOAT_IDDT = 0;
static PyObject *__pyx_v_10serializor_6binary_3ser_NDARRAY_BOOL_IDDT = 0;
static PyObject *__pyx_v_10serializor_6binary_3ser_NDARRAY_DT64_IDDT = 0;
static PyObject *__pyx_v_10serializor_6binary_3ser_NDARRAY_TD64_IDDT = 0;
static PyObject *__pyx_v_10serializor_6binary_3ser_NDARRAY_COMPLEX_IDDT = 0;
static PyObject *__pyx_v_10serializor_6binary_3ser_NDARRAY_BYTES_IDDT = 0;
static PyObject *__pyx_v_10serializor_6binary_3ser_NDARRAY_UNICODE_IDDT = 0;
static PyObject *__pyx_v_10serializor_6binary_3ser_PD_TIMESTAMP_ID = 0;
static PyObject *__pyx_v_10serializor_6binary_3ser_PD_TIMEDELTA_ID = 0;
static PyObject *__pyx_v_10serializor_6binary_3ser_SERIES_ID = 0;
static PyObject *__pyx_v_10serializor_6binary_3ser_DATETIMEINDEX_ID = 0;
static PyObject *__pyx_v_10serializor_6binary_3ser_TIMEDELTAINDEX_ID = 0;
static PyObject *__pyx_v_10serializor_6binary_3ser_DATAFRAME_ID = 0;
static PyObject *__pyx_v_10serializor_6binary_3ser_TRUE_VALUE = 0;
static PyObject *__pyx_v_10serializor_6binary_3ser_FALSE_VALUE = 0;
static PyObject *__pyx_v_10serializor_6binary_3ser_NONE_VALUE = 0;
static CYTHON_INLINE PyObject *__pyx_f_10serializor_6binary_3ser_pack_uint8(unsigned int); /*proto*/
static CYTHON_INLINE PyObject *__pyx_f_10serializor_6binary_3ser_pack_uint16(unsigned int); /*proto*/
static CYTHON_INLINE PyObject *__pyx_f_10serializor_6binary_3ser_pack_uint24(unsigned int); /*proto*/
static CYTHON_INLINE PyObject *__pyx_f_10serializor_6binary_3ser_pack_uint32(unsigned PY_LONG_LONG); /*proto*/
static CYTHON_INLINE PyObject *__pyx_f_10serializor_6binary_3ser_pack_uint64(unsigned PY_LONG_LONG); /*proto*/
static CYTHON_INLINE PyObject *__pyx_f_10serializor_6binary_3ser_pack_int8(int); /*proto*/
static CYTHON_INLINE PyObject *__pyx_f_10serializor_6binary_3ser_pack_int16(int); /*proto*/
static CYTHON_INLINE PyObject *__pyx_f_10serializor_6binary_3ser_pack_int24(int); /*proto*/
static CYTHON_INLINE PyObject *__pyx_f_10serializor_6binary_3ser_pack_int32(PY_LONG_LONG); /*proto*/
static CYTHON_INLINE PyObject *__pyx_f_10serializor_6binary_3ser_pack_int64(PY_LONG_LONG); /*proto*/
static CYTHON_INLINE PyObject *__pyx_f_10serializor_6binary_3ser_gen_encoded_int(unsigned PY_LONG_LONG); /*proto*/
static CYTHON_INLINE PyObject *__pyx_f_10serializor_6binary_3ser_gen_header(unsigned char, unsigned PY_LONG_LONG); /*proto*/
static PyObject *__pyx_f_10serializor_6binary_3ser_serialize(PyObject *, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_10serializor_6binary_3ser__orjson_dumps(PyObject *); /*proto*/
static PyObject *__pyx_f_10serializor_6binary_3ser__orjson_dumps_numpy(PyObject *); /*proto*/
static PyObject *__pyx_f_10serializor_6binary_3ser__ser_str(PyObject *); /*proto*/
static PyObject *__pyx_f_10serializor_6binary_3ser__ser_int(PyObject *); /*proto*/
static PyObject *__pyx_f_10serializor_6binary_3ser__ser_float(PyObject *); /*proto*/
static PyObject *__pyx_f_10serializor_6binary_3ser__ser_bool(PyObject *); /*proto*/
static PyObject *__pyx_f_10serializor_6binary_3ser__ser_none(PyObject *); /*proto*/
static PyObject *__pyx_f_10serializor_6binary_3ser__ser_datetime(PyObject *); /*proto*/
static PyObject *__pyx_f_10serializor_6binary_3ser__ser_date(PyObject *); /*proto*/
static PyObject *__pyx_f_10serializor_6binary_3ser__ser_time(PyObject *); /*proto*/
static PyObject *__pyx_f_10serializor_6binary_3ser__ser_timedelta(PyObject *); /*proto*/
static PyObject *__pyx_f_10serializor_6binary_3ser__ser_struct_time(PyObject *); /*proto*/
static PyObject *__pyx_f_10serializor_6binary_3ser__ser_decimal(PyObject *); /*proto*/
static PyObject *__pyx_f_10serializor_6binary_3ser__ser_complex(PyObject *); /*proto*/
static PyObject *__pyx_f_10serializor_6binary_3ser__ser_bytes(PyObject *); /*proto*/
static PyObject *__pyx_f_10serializor_6binary_3ser__ser_bytearray(PyObject *); /*proto*/
static PyObject *__pyx_f_10serializor_6binary_3ser__ser_memoryview(PyObject *); /*proto*/
static PyObject *__pyx_f_10serializor_6binary_3ser__ser_list(PyObject *); /*proto*/
static PyObject *__pyx_f_10serializor_6binary_3ser__ser_tuple(PyObject *); /*proto*/
static PyObject *__pyx_f_10serializor_6binary_3ser__ser_set(PyObject *); /*proto*/
static PyObject *__pyx_f_10serializor_6binary_3ser__ser_frozenset(PyObject *); /*proto*/
static PyObject *__pyx_f_10serializor_6binary_3ser__ser_range(PyObject *); /*proto*/
static PyObject *__pyx_f_10serializor_6binary_3ser__ser_sequence(PyObject *); /*proto*/
static PyObject *__pyx_f_10serializor_6binary_3ser__ser_dict(PyObject *); /*proto*/
static PyObject *__pyx_f_10serializor_6binary_3ser__ser_float64(PyObject *); /*proto*/
static PyObject *__pyx_f_10serializor_6binary_3ser__ser_datetime64(PyObject *); /*proto*/
static PyObject *__pyx_f_10serializor_6binary_3ser__ser_timedelta64(PyObject *); /*proto*/
static PyObject *__pyx_f_10serializor_6binary_3ser__ser_complex64(PyObject *); /*proto*/
static PyObject *__pyx_f_10serializor_6binary_3ser__ser_complex128(PyObject *); /*proto*/
static PyObject *__pyx_f_10serializor_6binary_3ser__ser_ndarray(PyArrayObject *); /*proto*/
static PyObject *__pyx_f_10serializor_6binary_3ser__ser_ndarray_object(PyArrayObject *, PyObject *); /*proto*/
static PyObject *__pyx_f_10serializor_6binary_3ser__ser_ndarray_numeric(PyArrayObject *, PyObject *, int); /*proto*/
static PyObject *__pyx_f_10serializor_6binary_3ser__ser_ndarray_nptime(PyArrayObject *, PyObject *); /*proto*/
static PyObject *__pyx_f_10serializor_6binary_3ser__ser_ndarray_complex(PyArrayObject *, PyObject *); /*proto*/
static PyObject *__pyx_f_10serializor_6binary_3ser__ser_ndarray_bytes(PyArrayObject *, PyObject *); /*proto*/
static PyObject *__pyx_f_10serializor_6binary_3ser__ser_ndarray_unicode(PyArrayObject *, PyObject *); /*proto*/
static PyObject *__pyx_f_10serializor_6binary_3ser__ser_pd_timestamp(PyObject *); /*proto*/
static PyObject *__pyx_f_10serializor_6binary_3ser__ser_pd_timedelta(PyObject *); /*proto*/
static PyObject *__pyx_f_10serializor_6binary_3ser__ser_series(PyObject *); /*proto*/
static PyObject *__pyx_f_10serializor_6binary_3ser__ser_datetime_index(PyObject *); /*proto*/
static PyObject *__pyx_f_10serializor_6binary_3ser__ser_timedelta_index(PyObject *); /*proto*/
static PyObject *__pyx_f_10serializor_6binary_3ser__ser_dataframe(PyObject *); /*proto*/
static PyObject *__pyx_f_10serializor_6binary_3ser__ser_common(PyObject *); /*proto*/
static PyObject *__pyx_f_10serializor_6binary_3ser__ser_uncommon(PyObject *, PyTypeObject *); /*proto*/
static PyObject *__pyx_f_10serializor_6binary_3ser__ser_subclass(PyObject *, PyTypeObject *); /*proto*/
/* #### Code section: typeinfo ### */
/* #### Code section: before_global_var ### */
#define __Pyx_MODULE_NAME "serializor.binary.ser"
extern int __pyx_module_is_main_serializor__binary__ser;
int __pyx_module_is_main_serializor__binary__ser = 0;

/* Implementation of "serializor.binary.ser" */
/* #### Code section: global_var ### */
static PyObject *__pyx_builtin_TypeError;
static PyObject *__pyx_builtin_range;
static PyObject *__pyx_builtin_ValueError;
static PyObject *__pyx_builtin_AssertionError;
static PyObject *__pyx_builtin_print;
static PyObject *__pyx_builtin_ImportError;
static PyObject *__pyx_builtin_RuntimeError;
/* #### Code section: string_decls ### */
static const char __pyx_k_B[] = "<B";
static const char __pyx_k_D[] = "D";
static const char __pyx_k_H[] = "<H";
static const char __pyx_k_I[] = "<I";
static const char __pyx_k_M[] = "M";
static const char __pyx_k_Q[] = "<Q";
static const char __pyx_k_W[] = "W";
static const char __pyx_k_Y[] = "Y";
static const char __pyx_k_b[] = "<b";
static const char __pyx_k_h[] = "h";
static const char __pyx_k_i[] = "<i";
static const char __pyx_k_m[] = "m";
static const char __pyx_k_n[] = "n";
static const char __pyx_k_q[] = "<q";
static const char __pyx_k_s[] = "s";
static const char __pyx_k_v[] = "v";
static const char __pyx_k__4[] = "";
static const char __pyx_k__5[] = "\001";
static const char __pyx_k__6[] = "\002";
static const char __pyx_k__7[] = "\000";
static const char __pyx_k__8[] = "\003";
static const char __pyx_k__9[] = "\004";
static const char __pyx_k_as[] = "as";
static const char __pyx_k_dt[] = "dt";
static const char __pyx_k_fs[] = "fs";
static const char __pyx_k_ms[] = "ms";
static const char __pyx_k_np[] = "np";
static const char __pyx_k_ns[] = "ns";
static const char __pyx_k_ps[] = "ps";
static const char __pyx_k_tz[] = "tz";
static const char __pyx_k_us[] = "us";
static const char __pyx_k__13[] = "\n";
static const char __pyx_k__14[] = "*";
static const char __pyx_k__15[] = " | ";
static const char __pyx_k__16[] = "\r";
static const char __pyx_k__27[] = ".";
static const char __pyx_k__46[] = "?";
static const char __pyx_k_all[] = "__all__";
static const char __pyx_k_b_2[] = "b";
static const char __pyx_k_end[] = "end";
static const char __pyx_k_h_2[] = "<h";
static const char __pyx_k_num[] = " - num: ";
static const char __pyx_k_obj[] = "obj";
static const char __pyx_k_str[] = "str";
static const char __pyx_k_val[] = "val";
static const char __pyx_k_None[] = "None";
static const char __pyx_k_imag[] = "imag";
static const char __pyx_k_join[] = "join";
static const char __pyx_k_main[] = "__main__";
static const char __pyx_k_name[] = "name";
static const char __pyx_k_pack[] = "pack";
static const char __pyx_k_real[] = "real";
static const char __pyx_k_spec[] = "__spec__";
static const char __pyx_k_step[] = "step";
static const char __pyx_k_stop[] = "stop";
static const char __pyx_k_test[] = "__test__";
static const char __pyx_k_time[] = "time";
static const char __pyx_k_Error[] = "\nError: ";
static const char __pyx_k_dchar[] = "dchar";
static const char __pyx_k_dtype[] = "dtype";
static const char __pyx_k_dumps[] = "_dumps";
static const char __pyx_k_index[] = "index";
static const char __pyx_k_items[] = "items";
static const char __pyx_k_ljust[] = "ljust";
static const char __pyx_k_num_2[] = "num";
static const char __pyx_k_numpy[] = "numpy";
static const char __pyx_k_print[] = "print";
static const char __pyx_k_range[] = "range";
static const char __pyx_k_start[] = "start";
static const char __pyx_k_utils[] = "utils";
static const char __pyx_k_value[] = "value";
static const char __pyx_k_SET_ID[] = "SET_ID";
static const char __pyx_k_Series[] = "Series";
static const char __pyx_k_errors[] = "errors";
static const char __pyx_k_import[] = "__import__";
static const char __pyx_k_name_2[] = "__name__";
static const char __pyx_k_object[] = "object";
static const char __pyx_k_option[] = "option";
static const char __pyx_k_orjson[] = "orjson";
static const char __pyx_k_pandas[] = "pandas";
static const char __pyx_k_pass_s[] = "- pass: %s";
static const char __pyx_k_return[] = "return";
static const char __pyx_k_struct[] = "struct";
static const char __pyx_k_typing[] = "typing";
static const char __pyx_k_values[] = "values";
static const char __pyx_k_DATE_ID[] = "DATE_ID";
static const char __pyx_k_DICT_ID[] = "DICT_ID";
static const char __pyx_k_DT64_ID[] = "DT64_ID";
static const char __pyx_k_LIST_ID[] = "LIST_ID";
static const char __pyx_k_TD64_ID[] = "TD64_ID";
static const char __pyx_k_TIME_ID[] = "TIME_ID";
static const char __pyx_k_columns[] = "columns";
static const char __pyx_k_dumps_2[] = "dumps";
static const char __pyx_k_tm_zone[] = "tm_zone";
static const char __pyx_k_tobytes[] = "tobytes";
static const char __pyx_k_typeref[] = "typeref";
static const char __pyx_k_Iterable[] = "Iterable";
static const char __pyx_k_RANGE_ID[] = "RANGE_ID";
static const char __pyx_k_TUPLE_ID[] = "TUPLE_ID";
static const char __pyx_k_datetime[] = "datetime";
static const char __pyx_k_pyx_capi[] = "__pyx_capi__";
static const char __pyx_k_DataFrame[] = "DataFrame";
static const char __pyx_k_SERIES_ID[] = "SERIES_ID";
static const char __pyx_k_Timedelta[] = "Timedelta";
static const char __pyx_k_Timestamp[] = "Timestamp";
static const char __pyx_k_TypeError[] = "TypeError";
static const char __pyx_k_serialize[] = "serialize";
static const char __pyx_k_tm_gmtoff[] = "tm_gmtoff";
static const char __pyx_k_utcoffset[] = "utcoffset";
static const char __pyx_k_COMPLEX_ID[] = "COMPLEX_ID";
static const char __pyx_k_NDARRAY_ID[] = "NDARRAY_ID";
static const char __pyx_k_NONE_VALUE[] = "NONE_VALUE";
static const char __pyx_k_TRUE_VALUE[] = "TRUE_VALUE";
static const char __pyx_k_ValueError[] = "ValueError";
static const char __pyx_k_identifier[] = "identifier";
static const char __pyx_k_serializor[] = "serializor";
static const char __pyx_k_test_utils[] = "_test_utils";
static const char __pyx_k_DATETIME_ID[] = "DATETIME_ID";
static const char __pyx_k_FALSE_VALUE[] = "FALSE_VALUE";
static const char __pyx_k_ImportError[] = "ImportError";
static const char __pyx_k_struct_time[] = "struct_time";
static const char __pyx_k_DATAFRAME_ID[] = "DATAFRAME_ID";
static const char __pyx_k_FROZENSET_ID[] = "FROZENSET_ID";
static const char __pyx_k_RuntimeError[] = "RuntimeError";
static const char __pyx_k_TIMEDELTA_ID[] = "TIMEDELTA_ID";
static const char __pyx_k_initializing[] = "_initializing";
static const char __pyx_k_is_coroutine[] = "_is_coroutine";
static const char __pyx_k_DatetimeIndex[] = "DatetimeIndex";
static const char __pyx_k_class_getitem[] = "__class_getitem__";
static const char __pyx_k_AssertionError[] = "AssertionError";
static const char __pyx_k_NDARRAY_INT_DT[] = "NDARRAY_INT_DT";
static const char __pyx_k_Pass_Pack_int8[] = "Pass Pack int8";
static const char __pyx_k_STRUCT_TIME_ID[] = "STRUCT_TIME_ID";
static const char __pyx_k_SerializeError[] = "SerializeError";
static const char __pyx_k_TimedeltaIndex[] = "TimedeltaIndex";
static const char __pyx_k_test_pack_int8[] = "_test_pack_int8";
static const char __pyx_k_NDARRAY_BOOL_DT[] = "NDARRAY_BOOL_DT";
static const char __pyx_k_NDARRAY_DT64_DT[] = "NDARRAY_DT64_DT";
static const char __pyx_k_NDARRAY_TD64_DT[] = "NDARRAY_TD64_DT";
static const char __pyx_k_NDARRAY_UINT_DT[] = "NDARRAY_UINT_DT";
static const char __pyx_k_PD_TIMEDELTA_ID[] = "PD_TIMEDELTA_ID";
static const char __pyx_k_PD_TIMESTAMP_ID[] = "PD_TIMESTAMP_ID";
static const char __pyx_k_Pass_Gen_Header[] = "Pass Gen Header";
static const char __pyx_k_Pass_Pack_int16[] = "Pass Pack int16";
static const char __pyx_k_Pass_Pack_int24[] = "Pass Pack int24";
static const char __pyx_k_Pass_Pack_int32[] = "Pass Pack int32";
static const char __pyx_k_Pass_Pack_int64[] = "Pass Pack int64";
static const char __pyx_k_Pass_Pack_uint8[] = "Pass Pack uint8";
static const char __pyx_k_test_gen_header[] = "_test_gen_header";
static const char __pyx_k_test_pack_int16[] = "_test_pack_int16";
static const char __pyx_k_test_pack_int24[] = "_test_pack_int24";
static const char __pyx_k_test_pack_int32[] = "_test_pack_int32";
static const char __pyx_k_test_pack_int64[] = "_test_pack_int64";
static const char __pyx_k_test_pack_uint8[] = "_test_pack_uint8";
static const char __pyx_k_DATETIMEINDEX_ID[] = "DATETIMEINDEX_ID";
static const char __pyx_k_NDARRAY_BYTES_DT[] = "NDARRAY_BYTES_DT";
static const char __pyx_k_NDARRAY_FLOAT_DT[] = "NDARRAY_FLOAT_DT";
static const char __pyx_k_NDARRAY_INT_IDDT[] = "NDARRAY_INT_IDDT";
static const char __pyx_k_Pass_Pack_uint16[] = "Pass Pack uint16";
static const char __pyx_k_Pass_Pack_uint24[] = "Pass Pack uint24";
static const char __pyx_k_Pass_Pack_uint32[] = "Pass Pack uint32";
static const char __pyx_k_Pass_Pack_uint64[] = "Pass Pack uint64";
static const char __pyx_k_test_pack_uint16[] = "_test_pack_uint16";
static const char __pyx_k_test_pack_uint24[] = "_test_pack_uint24";
static const char __pyx_k_test_pack_uint32[] = "_test_pack_uint32";
static const char __pyx_k_test_pack_uint64[] = "_test_pack_uint64";
static const char __pyx_k_NDARRAY_BOOL_IDDT[] = "NDARRAY_BOOL_IDDT";
static const char __pyx_k_NDARRAY_DT64_IDDT[] = "NDARRAY_DT64_IDDT";
static const char __pyx_k_NDARRAY_OBJECT_DT[] = "NDARRAY_OBJECT_DT";
static const char __pyx_k_NDARRAY_TD64_IDDT[] = "NDARRAY_TD64_IDDT";
static const char __pyx_k_NDARRAY_UINT_IDDT[] = "NDARRAY_UINT_IDDT";
static const char __pyx_k_TIMEDELTAINDEX_ID[] = "TIMEDELTAINDEX_ID";
static const char __pyx_k_NDARRAY_BYTES_IDDT[] = "NDARRAY_BYTES_IDDT";
static const char __pyx_k_NDARRAY_COMPLEX_DT[] = "NDARRAY_COMPLEX_DT";
static const char __pyx_k_NDARRAY_FLOAT_IDDT[] = "NDARRAY_FLOAT_IDDT";
static const char __pyx_k_NDARRAY_UNICODE_DT[] = "NDARRAY_UNICODE_DT";
static const char __pyx_k_UINT8_ENCODE_VALUE[] = "UINT8_ENCODE_VALUE";
static const char __pyx_k_asyncio_coroutines[] = "asyncio.coroutines";
static const char __pyx_k_cline_in_traceback[] = "cline_in_traceback";
static const char __pyx_k_NDARRAY_OBJECT_IDDT[] = "NDARRAY_OBJECT_IDDT";
static const char __pyx_k_OPT_SERIALIZE_NUMPY[] = "OPT_SERIALIZE_NUMPY";
static const char __pyx_k_UINT16_ENCODE_VALUE[] = "UINT16_ENCODE_VALUE";
static const char __pyx_k_UINT32_ENCODE_VALUE[] = "UINT32_ENCODE_VALUE";
static const char __pyx_k_UINT64_ENCODE_VALUE[] = "UINT64_ENCODE_VALUE";
static const char __pyx_k_NDARRAY_COMPLEX_IDDT[] = "NDARRAY_COMPLEX_IDDT";
static const char __pyx_k_NDARRAY_UNICODE_IDDT[] = "NDARRAY_UNICODE_IDDT";
static const char __pyx_k_test_gen_encoded_int[] = "_test_gen_encoded_int";
static const char __pyx_k_serializor_binary_ser[] = "serializor.binary.ser";
static const char __pyx_k_unsupported_data_type_s[] = "unsupported 'data' type %s.";
static const char __pyx_k_Pass_Gen_Encoded_Integer[] = "Pass Gen Encoded Integer";
static const char __pyx_k_expects_pandas_Series_got_s[] = "expects <'pandas.Series'>, got %s.";
static const char __pyx_k_invalid_pandas_Series_name_r[] = "invalid <'pandas.Series'> name %r.";
static const char __pyx_k_src_serializor_binary_ser_py[] = "src/serializor/binary/ser.py";
static const char __pyx_k_unsupported_numpy_time_unit_d[] = "unsupported numpy time unit: %d.";
static const char __pyx_k_unsupported_numpy_time_unit_s[] = "unsupported numpy time unit: %s.";
static const char __pyx_k_Serializor_Failed_to_serialize[] = "<'Serializor'>\nFailed to serialize to 'bytes': ";
static const char __pyx_k_expects_pandas_Timedelta_got_s[] = "expects <'pandas.Timedelta'>, got %s.";
static const char __pyx_k_expects_pandas_Timestamp_got_s[] = "expects <'pandas.Timestamp'>, got %s.";
static const char __pyx_k_invalid_pandas_DataFrame_cannot[] = "invalid <'pandas.DataFrame'>, cannot access column's underlying 'ndarray.values':\n%s";
static const char __pyx_k_unsupported_numpy_ndarray_dtype[] = "unsupported <'numpy.ndarray'> dtype '%s'.";
static const char __pyx_k_unsupported_pandas_Series_dtype[] = "unsupported <'pandas.Series'> dtype '%s'.";
static const char __pyx_k_Time_zones_are_not_available_fro[] = "Time zones are not available from the C-API.";
static const char __pyx_k_expects_np_ndarray_dtype_of_date[] = "expects np.ndarray dtype of 'datetime64/timedelta64', got '%s'.";
static const char __pyx_k_expects_pandas_DatetimeIndex_got[] = "expects <'pandas.DatetimeIndex'>, got %s.";
static const char __pyx_k_expects_pandas_TimedeltaIndex_go[] = "expects <'pandas.TimedeltaIndex'>, got %s.";
static const char __pyx_k_invalid_pandas_DataFrame_columns[] = "invalid <'pandas.DataFrame'> columns:\n%s";
static const char __pyx_k_invalid_pandas_DatetimeIndex_nam[] = "invalid <'pandas.DatetimeIndex'> name %r.";
static const char __pyx_k_invalid_pandas_TimedeltaIndex_na[] = "invalid <'pandas.TimedeltaIndex'> name %r.";
static const char __pyx_k_numpy__core_multiarray_failed_to[] = "numpy._core.multiarray failed to import";
static const char __pyx_k_numpy__core_umath_failed_to_impo[] = "numpy._core.umath failed to import";
static const char __pyx_k_unable_to_parse_ndarray_time_uni[] = "unable to parse ndarray time unit from '%s'.";
static const char __pyx_k_unsupported_numpy_ndarray_dimens[] = "unsupported <'numpy.ndarray'> dimension [%d].";
static const char __pyx_k_unsupported_pandas_DataFrame_col[] = "unsupported <'pandas.DataFrame'> column dtype '%s'.";
/* #### Code section: decls ### */
static PyObject *__pyx_pf_10serializor_6binary_3ser_serialize(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_obj); /* proto */
static PyObject *__pyx_pf_10serializor_6binary_3ser_2_test_utils(CYTHON_UNUSED PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_10serializor_6binary_3ser_4_test_pack_int8(CYTHON_UNUSED PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_10serializor_6binary_3ser_6_test_pack_uint8(CYTHON_UNUSED PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_10serializor_6binary_3ser_8_test_pack_int16(CYTHON_UNUSED PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_10serializor_6binary_3ser_10_test_pack_uint16(CYTHON_UNUSED PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_10serializor_6binary_3ser_12_test_pack_int24(CYTHON_UNUSED PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_10serializor_6binary_3ser_14_test_pack_uint24(CYTHON_UNUSED PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_10serializor_6binary_3ser_16_test_pack_int32(CYTHON_UNUSED PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_10serializor_6binary_3ser_18_test_pack_uint32(CYTHON_UNUSED PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_10serializor_6binary_3ser_20_test_pack_int64(CYTHON_UNUSED PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_10serializor_6binary_3ser_22_test_pack_uint64(CYTHON_UNUSED PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_10serializor_6binary_3ser_24_test_gen_encoded_int(CYTHON_UNUSED PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_10serializor_6binary_3ser_26_test_gen_header(CYTHON_UNUSED PyObject *__pyx_self); /* proto */
static __Pyx_CachedCFunction __pyx_umethod_PyMemoryView_Type_tobytes = {0, 0, 0, 0, 0};
/* #### Code section: late_includes ### */
/* #### Code section: module_state ### */
typedef struct {
  PyObject *__pyx_d;
  PyObject *__pyx_b;
  PyObject *__pyx_cython_runtime;
  PyObject *__pyx_empty_tuple;
  PyObject *__pyx_empty_bytes;
  PyObject *__pyx_empty_unicode;
  #ifdef __Pyx_CyFunction_USED
  PyTypeObject *__pyx_CyFunctionType;
  #endif
  #ifdef __Pyx_FusedFunction_USED
  PyTypeObject *__pyx_FusedFunctionType;
  #endif
  #ifdef __Pyx_Generator_USED
  PyTypeObject *__pyx_GeneratorType;
  #endif
  #ifdef __Pyx_IterableCoroutine_USED
  PyTypeObject *__pyx_IterableCoroutineType;
  #endif
  #ifdef __Pyx_Coroutine_USED
  PyTypeObject *__pyx_CoroutineAwaitType;
  #endif
  #ifdef __Pyx_Coroutine_USED
  PyTypeObject *__pyx_CoroutineType;
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  PyTypeObject *__pyx_ptype_7cpython_4type_type;
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  PyTypeObject *__pyx_ptype_7cpython_4bool_bool;
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  PyTypeObject *__pyx_ptype_7cpython_7complex_complex;
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  PyTypeObject *__pyx_ptype_5numpy_dtype;
  PyTypeObject *__pyx_ptype_5numpy_flatiter;
  PyTypeObject *__pyx_ptype_5numpy_broadcast;
  PyTypeObject *__pyx_ptype_5numpy_ndarray;
  PyTypeObject *__pyx_ptype_5numpy_generic;
  PyTypeObject *__pyx_ptype_5numpy_number;
  PyTypeObject *__pyx_ptype_5numpy_integer;
  PyTypeObject *__pyx_ptype_5numpy_signedinteger;
  PyTypeObject *__pyx_ptype_5numpy_unsignedinteger;
  PyTypeObject *__pyx_ptype_5numpy_inexact;
  PyTypeObject *__pyx_ptype_5numpy_floating;
  PyTypeObject *__pyx_ptype_5numpy_complexfloating;
  PyTypeObject *__pyx_ptype_5numpy_flexible;
  PyTypeObject *__pyx_ptype_5numpy_character;
  PyTypeObject *__pyx_ptype_5numpy_ufunc;
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  PyTypeObject *__pyx_ptype_7cpython_8datetime_date;
  PyTypeObject *__pyx_ptype_7cpython_8datetime_time;
  PyTypeObject *__pyx_ptype_7cpython_8datetime_datetime;
  PyTypeObject *__pyx_ptype_7cpython_8datetime_timedelta;
  PyTypeObject *__pyx_ptype_7cpython_8datetime_tzinfo;
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  PyObject *__pyx_n_s_AssertionError;
  PyObject *__pyx_kp_u_B;
  PyObject *__pyx_n_s_COMPLEX_ID;
  PyObject *__pyx_n_u_D;
  PyObject *__pyx_n_s_DATAFRAME_ID;
  PyObject *__pyx_n_s_DATETIMEINDEX_ID;
  PyObject *__pyx_n_s_DATETIME_ID;
  PyObject *__pyx_n_s_DATE_ID;
  PyObject *__pyx_n_s_DICT_ID;
  PyObject *__pyx_n_s_DT64_ID;
  PyObject *__pyx_n_s_DataFrame;
  PyObject *__pyx_n_s_DatetimeIndex;
  PyObject *__pyx_kp_u_Error;
  PyObject *__pyx_n_s_FALSE_VALUE;
  PyObject *__pyx_n_s_FROZENSET_ID;
  PyObject *__pyx_kp_u_H;
  PyObject *__pyx_kp_u_I;
  PyObject *__pyx_n_s_ImportError;
  PyObject *__pyx_n_s_Iterable;
  PyObject *__pyx_n_s_LIST_ID;
  PyObject *__pyx_n_u_M;
  PyObject *__pyx_n_s_NDARRAY_BOOL_DT;
  PyObject *__pyx_n_s_NDARRAY_BOOL_IDDT;
  PyObject *__pyx_n_s_NDARRAY_BYTES_DT;
  PyObject *__pyx_n_s_NDARRAY_BYTES_IDDT;
  PyObject *__pyx_n_s_NDARRAY_COMPLEX_DT;
  PyObject *__pyx_n_s_NDARRAY_COMPLEX_IDDT;
  PyObject *__pyx_n_s_NDARRAY_DT64_DT;
  PyObject *__pyx_n_s_NDARRAY_DT64_IDDT;
  PyObject *__pyx_n_s_NDARRAY_FLOAT_DT;
  PyObject *__pyx_n_s_NDARRAY_FLOAT_IDDT;
  PyObject *__pyx_n_s_NDARRAY_ID;
  PyObject *__pyx_n_s_NDARRAY_INT_DT;
  PyObject *__pyx_n_s_NDARRAY_INT_IDDT;
  PyObject *__pyx_n_s_NDARRAY_OBJECT_DT;
  PyObject *__pyx_n_s_NDARRAY_OBJECT_IDDT;
  PyObject *__pyx_n_s_NDARRAY_TD64_DT;
  PyObject *__pyx_n_s_NDARRAY_TD64_IDDT;
  PyObject *__pyx_n_s_NDARRAY_UINT_DT;
  PyObject *__pyx_n_s_NDARRAY_UINT_IDDT;
  PyObject *__pyx_n_s_NDARRAY_UNICODE_DT;
  PyObject *__pyx_n_s_NDARRAY_UNICODE_IDDT;
  PyObject *__pyx_n_s_NONE_VALUE;
  PyObject *__pyx_n_s_None;
  PyObject *__pyx_n_s_OPT_SERIALIZE_NUMPY;
  PyObject *__pyx_n_s_PD_TIMEDELTA_ID;
  PyObject *__pyx_n_s_PD_TIMESTAMP_ID;
  PyObject *__pyx_kp_u_Pass_Gen_Encoded_Integer;
  PyObject *__pyx_kp_u_Pass_Gen_Header;
  PyObject *__pyx_kp_u_Pass_Pack_int16;
  PyObject *__pyx_kp_u_Pass_Pack_int24;
  PyObject *__pyx_kp_u_Pass_Pack_int32;
  PyObject *__pyx_kp_u_Pass_Pack_int64;
  PyObject *__pyx_kp_u_Pass_Pack_int8;
  PyObject *__pyx_kp_u_Pass_Pack_uint16;
  PyObject *__pyx_kp_u_Pass_Pack_uint24;
  PyObject *__pyx_kp_u_Pass_Pack_uint32;
  PyObject *__pyx_kp_u_Pass_Pack_uint64;
  PyObject *__pyx_kp_u_Pass_Pack_uint8;
  PyObject *__pyx_kp_u_Q;
  PyObject *__pyx_n_s_RANGE_ID;
  PyObject *__pyx_n_s_RuntimeError;
  PyObject *__pyx_n_s_SERIES_ID;
  PyObject *__pyx_n_s_SET_ID;
  PyObject *__pyx_n_s_STRUCT_TIME_ID;
  PyObject *__pyx_n_s_SerializeError;
  PyObject *__pyx_kp_u_Serializor_Failed_to_serialize;
  PyObject *__pyx_n_s_Series;
  PyObject *__pyx_n_s_TD64_ID;
  PyObject *__pyx_n_s_TIMEDELTAINDEX_ID;
  PyObject *__pyx_n_s_TIMEDELTA_ID;
  PyObject *__pyx_n_s_TIME_ID;
  PyObject *__pyx_n_s_TRUE_VALUE;
  PyObject *__pyx_n_s_TUPLE_ID;
  PyObject *__pyx_kp_u_Time_zones_are_not_available_fro;
  PyObject *__pyx_n_s_Timedelta;
  PyObject *__pyx_n_s_TimedeltaIndex;
  PyObject *__pyx_n_s_Timestamp;
  PyObject *__pyx_n_s_TypeError;
  PyObject *__pyx_n_s_UINT16_ENCODE_VALUE;
  PyObject *__pyx_n_s_UINT32_ENCODE_VALUE;
  PyObject *__pyx_n_s_UINT64_ENCODE_VALUE;
  PyObject *__pyx_n_s_UINT8_ENCODE_VALUE;
  PyObject *__pyx_n_s_ValueError;
  PyObject *__pyx_n_u_W;
  PyObject *__pyx_n_u_Y;
  PyObject *__pyx_kp_u__13;
  PyObject *__pyx_n_s__14;
  PyObject *__pyx_kp_u__15;
  PyObject *__pyx_kp_u__16;
  PyObject *__pyx_kp_u__27;
  PyObject *__pyx_kp_b__4;
  PyObject *__pyx_n_s__46;
  PyObject *__pyx_kp_b__5;
  PyObject *__pyx_kp_b__6;
  PyObject *__pyx_kp_b__7;
  PyObject *__pyx_kp_b__8;
  PyObject *__pyx_kp_b__9;
  PyObject *__pyx_n_s_all;
  PyObject *__pyx_n_u_as;
  PyObject *__pyx_n_s_asyncio_coroutines;
  PyObject *__pyx_kp_u_b;
  PyObject *__pyx_n_s_b_2;
  PyObject *__pyx_n_s_class_getitem;
  PyObject *__pyx_n_s_cline_in_traceback;
  PyObject *__pyx_n_s_columns;
  PyObject *__pyx_n_s_datetime;
  PyObject *__pyx_n_s_dchar;
  PyObject *__pyx_n_s_dt;
  PyObject *__pyx_n_s_dtype;
  PyObject *__pyx_n_s_dumps;
  PyObject *__pyx_n_s_dumps_2;
  PyObject *__pyx_n_s_end;
  PyObject *__pyx_n_s_errors;
  PyObject *__pyx_kp_u_expects_np_ndarray_dtype_of_date;
  PyObject *__pyx_kp_u_expects_pandas_DatetimeIndex_got;
  PyObject *__pyx_kp_u_expects_pandas_Series_got_s;
  PyObject *__pyx_kp_u_expects_pandas_TimedeltaIndex_go;
  PyObject *__pyx_kp_u_expects_pandas_Timedelta_got_s;
  PyObject *__pyx_kp_u_expects_pandas_Timestamp_got_s;
  PyObject *__pyx_n_u_fs;
  PyObject *__pyx_n_u_h;
  PyObject *__pyx_kp_u_h_2;
  PyObject *__pyx_kp_u_i;
  PyObject *__pyx_n_s_identifier;
  PyObject *__pyx_n_s_imag;
  PyObject *__pyx_n_s_import;
  PyObject *__pyx_n_s_index;
  PyObject *__pyx_n_s_initializing;
  PyObject *__pyx_kp_u_invalid_pandas_DataFrame_cannot;
  PyObject *__pyx_kp_u_invalid_pandas_DataFrame_columns;
  PyObject *__pyx_kp_u_invalid_pandas_DatetimeIndex_nam;
  PyObject *__pyx_kp_u_invalid_pandas_Series_name_r;
  PyObject *__pyx_kp_u_invalid_pandas_TimedeltaIndex_na;
  PyObject *__pyx_n_s_is_coroutine;
  PyObject *__pyx_n_s_items;
  PyObject *__pyx_n_s_join;
  PyObject *__pyx_n_s_ljust;
  PyObject *__pyx_n_u_m;
  PyObject *__pyx_n_s_main;
  PyObject *__pyx_n_u_ms;
  PyObject *__pyx_n_s_n;
  PyObject *__pyx_n_s_name;
  PyObject *__pyx_n_s_name_2;
  PyObject *__pyx_n_s_np;
  PyObject *__pyx_n_u_ns;
  PyObject *__pyx_kp_u_num;
  PyObject *__pyx_n_s_num_2;
  PyObject *__pyx_n_s_numpy;
  PyObject *__pyx_kp_u_numpy__core_multiarray_failed_to;
  PyObject *__pyx_kp_u_numpy__core_umath_failed_to_impo;
  PyObject *__pyx_n_s_obj;
  PyObject *__pyx_n_s_object;
  PyObject *__pyx_n_s_option;
  PyObject *__pyx_n_s_orjson;
  PyObject *__pyx_n_s_pack;
  PyObject *__pyx_n_s_pandas;
  PyObject *__pyx_kp_u_pass_s;
  PyObject *__pyx_n_s_print;
  PyObject *__pyx_n_u_ps;
  PyObject *__pyx_n_s_pyx_capi;
  PyObject *__pyx_kp_u_q;
  PyObject *__pyx_n_s_range;
  PyObject *__pyx_n_s_real;
  PyObject *__pyx_n_s_return;
  PyObject *__pyx_n_s_s;
  PyObject *__pyx_n_u_s;
  PyObject *__pyx_n_s_serialize;
  PyObject *__pyx_n_u_serialize;
  PyObject *__pyx_n_s_serializor;
  PyObject *__pyx_n_s_serializor_binary_ser;
  PyObject *__pyx_n_s_spec;
  PyObject *__pyx_kp_s_src_serializor_binary_ser_py;
  PyObject *__pyx_n_s_start;
  PyObject *__pyx_n_s_step;
  PyObject *__pyx_n_s_stop;
  PyObject *__pyx_n_s_str;
  PyObject *__pyx_n_s_struct;
  PyObject *__pyx_n_s_struct_time;
  PyObject *__pyx_n_s_test;
  PyObject *__pyx_n_s_test_gen_encoded_int;
  PyObject *__pyx_n_s_test_gen_header;
  PyObject *__pyx_n_s_test_pack_int16;
  PyObject *__pyx_n_s_test_pack_int24;
  PyObject *__pyx_n_s_test_pack_int32;
  PyObject *__pyx_n_s_test_pack_int64;
  PyObject *__pyx_n_s_test_pack_int8;
  PyObject *__pyx_n_s_test_pack_uint16;
  PyObject *__pyx_n_s_test_pack_uint24;
  PyObject *__pyx_n_s_test_pack_uint32;
  PyObject *__pyx_n_s_test_pack_uint64;
  PyObject *__pyx_n_s_test_pack_uint8;
  PyObject *__pyx_n_s_test_utils;
  PyObject *__pyx_n_s_time;
  PyObject *__pyx_n_s_tm_gmtoff;
  PyObject *__pyx_n_s_tm_zone;
  PyObject *__pyx_n_s_tobytes;
  PyObject *__pyx_n_s_typeref;
  PyObject *__pyx_n_s_typing;
  PyObject *__pyx_n_s_tz;
  PyObject *__pyx_kp_u_unable_to_parse_ndarray_time_uni;
  PyObject *__pyx_kp_u_unsupported_data_type_s;
  PyObject *__pyx_kp_u_unsupported_numpy_ndarray_dimens;
  PyObject *__pyx_kp_u_unsupported_numpy_ndarray_dtype;
  PyObject *__pyx_kp_u_unsupported_numpy_time_unit_d;
  PyObject *__pyx_kp_u_unsupported_numpy_time_unit_s;
  PyObject *__pyx_kp_u_unsupported_pandas_DataFrame_col;
  PyObject *__pyx_kp_u_unsupported_pandas_Series_dtype;
  PyObject *__pyx_n_u_us;
  PyObject *__pyx_n_s_utcoffset;
  PyObject *__pyx_n_s_utils;
  PyObject *__pyx_n_s_v;
  PyObject *__pyx_n_s_val;
  PyObject *__pyx_n_s_value;
  PyObject *__pyx_n_s_values;
  PyObject *__pyx_int_0;
  PyObject *__pyx_int_1;
  PyObject *__pyx_int_3;
  PyObject *__pyx_int_80;
  PyObject *__pyx_int_251;
  PyObject *__pyx_int_252;
  PyObject *__pyx_int_65535;
  PyObject *__pyx_int_65536;
  PyObject *__pyx_int_8388606;
  PyObject *__pyx_int_8388607;
  PyObject *__pyx_int_16777213;
  PyObject *__pyx_int_16777214;
  PyObject *__pyx_int_16777215;
  PyObject *__pyx_int_2147483646;
  PyObject *__pyx_int_2147483647;
  PyObject *__pyx_int_4294967293;
  PyObject *__pyx_int_4294967294;
  PyObject *__pyx_int_4294967295;
  PyObject *__pyx_int_4294967296;
  PyObject *__pyx_int_8589934592;
  PyObject *__pyx_int_0x7ffffffffffffffe;
  PyObject *__pyx_int_0x7fffffffffffffff;
  PyObject *__pyx_int_0xfffffffffffffffd;
  PyObject *__pyx_int_0xfffffffffffffffe;
  PyObject *__pyx_int_0xffffffffffffffff;
  PyObject *__pyx_int_neg_8388607;
  PyObject *__pyx_int_neg_8388608;
  PyObject *__pyx_int_neg_2147483647;
  PyObject *__pyx_int_neg_2147483648;
  PyObject *__pyx_int_neg_9223372036854775807;
  PyObject *__pyx_int_neg_9223372036854775808;
  PyObject *__pyx_tuple_;
  PyObject *__pyx_tuple__2;
  PyObject *__pyx_tuple__3;
  PyObject *__pyx_slice__18;
  PyObject *__pyx_tuple__10;
  PyObject *__pyx_tuple__11;
  PyObject *__pyx_tuple__12;
  PyObject *__pyx_tuple__17;
  PyObject *__pyx_tuple__19;
  PyObject *__pyx_tuple__20;
  PyObject *__pyx_tuple__21;
  PyObject *__pyx_tuple__22;
  PyObject *__pyx_tuple__23;
  PyObject *__pyx_tuple__24;
  PyObject *__pyx_tuple__25;
  PyObject *__pyx_tuple__26;
  PyObject *__pyx_tuple__28;
  PyObject *__pyx_tuple__31;
  PyObject *__pyx_tuple__42;
  PyObject *__pyx_tuple__44;
  PyObject *__pyx_codeobj__29;
  PyObject *__pyx_codeobj__30;
  PyObject *__pyx_codeobj__32;
  PyObject *__pyx_codeobj__33;
  PyObject *__pyx_codeobj__34;
  PyObject *__pyx_codeobj__35;
  PyObject *__pyx_codeobj__36;
  PyObject *__pyx_codeobj__37;
  PyObject *__pyx_codeobj__38;
  PyObject *__pyx_codeobj__39;
  PyObject *__pyx_codeobj__40;
  PyObject *__pyx_codeobj__41;
  PyObject *__pyx_codeobj__43;
  PyObject *__pyx_codeobj__45;
} __pyx_mstate;

#if CYTHON_USE_MODULE_STATE
#ifdef __cplusplus
namespace {
  extern struct PyModuleDef __pyx_moduledef;
} /* anonymous namespace */
#else
static struct PyModuleDef __pyx_moduledef;
#endif

#define __pyx_mstate(o) ((__pyx_mstate *)__Pyx_PyModule_GetState(o))

#define __pyx_mstate_global (__pyx_mstate(PyState_FindModule(&__pyx_moduledef)))

#define __pyx_m (PyState_FindModule(&__pyx_moduledef))
#else
static __pyx_mstate __pyx_mstate_global_static =
#ifdef __cplusplus
    {};
#else
    {0};
#endif
static __pyx_mstate *__pyx_mstate_global = &__pyx_mstate_global_static;
#endif
/* #### Code section: module_state_clear ### */
#if CYTHON_USE_MODULE_STATE
static int __pyx_m_clear(PyObject *m) {
  __pyx_mstate *clear_module_state = __pyx_mstate(m);
  if (!clear_module_state) return 0;
  Py_CLEAR(clear_module_state->__pyx_d);
  Py_CLEAR(clear_module_state->__pyx_b);
  Py_CLEAR(clear_module_state->__pyx_cython_runtime);
  Py_CLEAR(clear_module_state->__pyx_empty_tuple);
  Py_CLEAR(clear_module_state->__pyx_empty_bytes);
  Py_CLEAR(clear_module_state->__pyx_empty_unicode);
  #ifdef __Pyx_CyFunction_USED
  Py_CLEAR(clear_module_state->__pyx_CyFunctionType);
  #endif
  #ifdef __Pyx_FusedFunction_USED
  Py_CLEAR(clear_module_state->__pyx_FusedFunctionType);
  #endif
  Py_CLEAR(clear_module_state->__pyx_ptype_7cpython_4type_type);
  Py_CLEAR(clear_module_state->__pyx_ptype_7cpython_4bool_bool);
  Py_CLEAR(clear_module_state->__pyx_ptype_7cpython_7complex_complex);
  Py_CLEAR(clear_module_state->__pyx_ptype_5numpy_dtype);
  Py_CLEAR(clear_module_state->__pyx_ptype_5numpy_flatiter);
  Py_CLEAR(clear_module_state->__pyx_ptype_5numpy_broadcast);
  Py_CLEAR(clear_module_state->__pyx_ptype_5numpy_ndarray);
  Py_CLEAR(clear_module_state->__pyx_ptype_5numpy_generic);
  Py_CLEAR(clear_module_state->__pyx_ptype_5numpy_number);
  Py_CLEAR(clear_module_state->__pyx_ptype_5numpy_integer);
  Py_CLEAR(clear_module_state->__pyx_ptype_5numpy_signedinteger);
  Py_CLEAR(clear_module_state->__pyx_ptype_5numpy_unsignedinteger);
  Py_CLEAR(clear_module_state->__pyx_ptype_5numpy_inexact);
  Py_CLEAR(clear_module_state->__pyx_ptype_5numpy_floating);
  Py_CLEAR(clear_module_state->__pyx_ptype_5numpy_complexfloating);
  Py_CLEAR(clear_module_state->__pyx_ptype_5numpy_flexible);
  Py_CLEAR(clear_module_state->__pyx_ptype_5numpy_character);
  Py_CLEAR(clear_module_state->__pyx_ptype_5numpy_ufunc);
  Py_CLEAR(clear_module_state->__pyx_ptype_7cpython_8datetime_date);
  Py_CLEAR(clear_module_state->__pyx_ptype_7cpython_8datetime_time);
  Py_CLEAR(clear_module_state->__pyx_ptype_7cpython_8datetime_datetime);
  Py_CLEAR(clear_module_state->__pyx_ptype_7cpython_8datetime_timedelta);
  Py_CLEAR(clear_module_state->__pyx_ptype_7cpython_8datetime_tzinfo);
  Py_CLEAR(clear_module_state->__pyx_n_s_AssertionError);
  Py_CLEAR(clear_module_state->__pyx_kp_u_B);
  Py_CLEAR(clear_module_state->__pyx_n_s_COMPLEX_ID);
  Py_CLEAR(clear_module_state->__pyx_n_u_D);
  Py_CLEAR(clear_module_state->__pyx_n_s_DATAFRAME_ID);
  Py_CLEAR(clear_module_state->__pyx_n_s_DATETIMEINDEX_ID);
  Py_CLEAR(clear_module_state->__pyx_n_s_DATETIME_ID);
  Py_CLEAR(clear_module_state->__pyx_n_s_DATE_ID);
  Py_CLEAR(clear_module_state->__pyx_n_s_DICT_ID);
  Py_CLEAR(clear_module_state->__pyx_n_s_DT64_ID);
  Py_CLEAR(clear_module_state->__pyx_n_s_DataFrame);
  Py_CLEAR(clear_module_state->__pyx_n_s_DatetimeIndex);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Error);
  Py_CLEAR(clear_module_state->__pyx_n_s_FALSE_VALUE);
  Py_CLEAR(clear_module_state->__pyx_n_s_FROZENSET_ID);
  Py_CLEAR(clear_module_state->__pyx_kp_u_H);
  Py_CLEAR(clear_module_state->__pyx_kp_u_I);
  Py_CLEAR(clear_module_state->__pyx_n_s_ImportError);
  Py_CLEAR(clear_module_state->__pyx_n_s_Iterable);
  Py_CLEAR(clear_module_state->__pyx_n_s_LIST_ID);
  Py_CLEAR(clear_module_state->__pyx_n_u_M);
  Py_CLEAR(clear_module_state->__pyx_n_s_NDARRAY_BOOL_DT);
  Py_CLEAR(clear_module_state->__pyx_n_s_NDARRAY_BOOL_IDDT);
  Py_CLEAR(clear_module_state->__pyx_n_s_NDARRAY_BYTES_DT);
  Py_CLEAR(clear_module_state->__pyx_n_s_NDARRAY_BYTES_IDDT);
  Py_CLEAR(clear_module_state->__pyx_n_s_NDARRAY_COMPLEX_DT);
  Py_CLEAR(clear_module_state->__pyx_n_s_NDARRAY_COMPLEX_IDDT);
  Py_CLEAR(clear_module_state->__pyx_n_s_NDARRAY_DT64_DT);
  Py_CLEAR(clear_module_state->__pyx_n_s_NDARRAY_DT64_IDDT);
  Py_CLEAR(clear_module_state->__pyx_n_s_NDARRAY_FLOAT_DT);
  Py_CLEAR(clear_module_state->__pyx_n_s_NDARRAY_FLOAT_IDDT);
  Py_CLEAR(clear_module_state->__pyx_n_s_NDARRAY_ID);
  Py_CLEAR(clear_module_state->__pyx_n_s_NDARRAY_INT_DT);
  Py_CLEAR(clear_module_state->__pyx_n_s_NDARRAY_INT_IDDT);
  Py_CLEAR(clear_module_state->__pyx_n_s_NDARRAY_OBJECT_DT);
  Py_CLEAR(clear_module_state->__pyx_n_s_NDARRAY_OBJECT_IDDT);
  Py_CLEAR(clear_module_state->__pyx_n_s_NDARRAY_TD64_DT);
  Py_CLEAR(clear_module_state->__pyx_n_s_NDARRAY_TD64_IDDT);
  Py_CLEAR(clear_module_state->__pyx_n_s_NDARRAY_UINT_DT);
  Py_CLEAR(clear_module_state->__pyx_n_s_NDARRAY_UINT_IDDT);
  Py_CLEAR(clear_module_state->__pyx_n_s_NDARRAY_UNICODE_DT);
  Py_CLEAR(clear_module_state->__pyx_n_s_NDARRAY_UNICODE_IDDT);
  Py_CLEAR(clear_module_state->__pyx_n_s_NONE_VALUE);
  Py_CLEAR(clear_module_state->__pyx_n_s_None);
  Py_CLEAR(clear_module_state->__pyx_n_s_OPT_SERIALIZE_NUMPY);
  Py_CLEAR(clear_module_state->__pyx_n_s_PD_TIMEDELTA_ID);
  Py_CLEAR(clear_module_state->__pyx_n_s_PD_TIMESTAMP_ID);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Pass_Gen_Encoded_Integer);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Pass_Gen_Header);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Pass_Pack_int16);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Pass_Pack_int24);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Pass_Pack_int32);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Pass_Pack_int64);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Pass_Pack_int8);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Pass_Pack_uint16);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Pass_Pack_uint24);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Pass_Pack_uint32);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Pass_Pack_uint64);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Pass_Pack_uint8);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Q);
  Py_CLEAR(clear_module_state->__pyx_n_s_RANGE_ID);
  Py_CLEAR(clear_module_state->__pyx_n_s_RuntimeError);
  Py_CLEAR(clear_module_state->__pyx_n_s_SERIES_ID);
  Py_CLEAR(clear_module_state->__pyx_n_s_SET_ID);
  Py_CLEAR(clear_module_state->__pyx_n_s_STRUCT_TIME_ID);
  Py_CLEAR(clear_module_state->__pyx_n_s_SerializeError);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Serializor_Failed_to_serialize);
  Py_CLEAR(clear_module_state->__pyx_n_s_Series);
  Py_CLEAR(clear_module_state->__pyx_n_s_TD64_ID);
  Py_CLEAR(clear_module_state->__pyx_n_s_TIMEDELTAINDEX_ID);
  Py_CLEAR(clear_module_state->__pyx_n_s_TIMEDELTA_ID);
  Py_CLEAR(clear_module_state->__pyx_n_s_TIME_ID);
  Py_CLEAR(clear_module_state->__pyx_n_s_TRUE_VALUE);
  Py_CLEAR(clear_module_state->__pyx_n_s_TUPLE_ID);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Time_zones_are_not_available_fro);
  Py_CLEAR(clear_module_state->__pyx_n_s_Timedelta);
  Py_CLEAR(clear_module_state->__pyx_n_s_TimedeltaIndex);
  Py_CLEAR(clear_module_state->__pyx_n_s_Timestamp);
  Py_CLEAR(clear_module_state->__pyx_n_s_TypeError);
  Py_CLEAR(clear_module_state->__pyx_n_s_UINT16_ENCODE_VALUE);
  Py_CLEAR(clear_module_state->__pyx_n_s_UINT32_ENCODE_VALUE);
  Py_CLEAR(clear_module_state->__pyx_n_s_UINT64_ENCODE_VALUE);
  Py_CLEAR(clear_module_state->__pyx_n_s_UINT8_ENCODE_VALUE);
  Py_CLEAR(clear_module_state->__pyx_n_s_ValueError);
  Py_CLEAR(clear_module_state->__pyx_n_u_W);
  Py_CLEAR(clear_module_state->__pyx_n_u_Y);
  Py_CLEAR(clear_module_state->__pyx_kp_u__13);
  Py_CLEAR(clear_module_state->__pyx_n_s__14);
  Py_CLEAR(clear_module_state->__pyx_kp_u__15);
  Py_CLEAR(clear_module_state->__pyx_kp_u__16);
  Py_CLEAR(clear_module_state->__pyx_kp_u__27);
  Py_CLEAR(clear_module_state->__pyx_kp_b__4);
  Py_CLEAR(clear_module_state->__pyx_n_s__46);
  Py_CLEAR(clear_module_state->__pyx_kp_b__5);
  Py_CLEAR(clear_module_state->__pyx_kp_b__6);
  Py_CLEAR(clear_module_state->__pyx_kp_b__7);
  Py_CLEAR(clear_module_state->__pyx_kp_b__8);
  Py_CLEAR(clear_module_state->__pyx_kp_b__9);
  Py_CLEAR(clear_module_state->__pyx_n_s_all);
  Py_CLEAR(clear_module_state->__pyx_n_u_as);
  Py_CLEAR(clear_module_state->__pyx_n_s_asyncio_coroutines);
  Py_CLEAR(clear_module_state->__pyx_kp_u_b);
  Py_CLEAR(clear_module_state->__pyx_n_s_b_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_class_getitem);
  Py_CLEAR(clear_module_state->__pyx_n_s_cline_in_traceback);
  Py_CLEAR(clear_module_state->__pyx_n_s_columns);
  Py_CLEAR(clear_module_state->__pyx_n_s_datetime);
  Py_CLEAR(clear_module_state->__pyx_n_s_dchar);
  Py_CLEAR(clear_module_state->__pyx_n_s_dt);
  Py_CLEAR(clear_module_state->__pyx_n_s_dtype);
  Py_CLEAR(clear_module_state->__pyx_n_s_dumps);
  Py_CLEAR(clear_module_state->__pyx_n_s_dumps_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_end);
  Py_CLEAR(clear_module_state->__pyx_n_s_errors);
  Py_CLEAR(clear_module_state->__pyx_kp_u_expects_np_ndarray_dtype_of_date);
  Py_CLEAR(clear_module_state->__pyx_kp_u_expects_pandas_DatetimeIndex_got);
  Py_CLEAR(clear_module_state->__pyx_kp_u_expects_pandas_Series_got_s);
  Py_CLEAR(clear_module_state->__pyx_kp_u_expects_pandas_TimedeltaIndex_go);
  Py_CLEAR(clear_module_state->__pyx_kp_u_expects_pandas_Timedelta_got_s);
  Py_CLEAR(clear_module_state->__pyx_kp_u_expects_pandas_Timestamp_got_s);
  Py_CLEAR(clear_module_state->__pyx_n_u_fs);
  Py_CLEAR(clear_module_state->__pyx_n_u_h);
  Py_CLEAR(clear_module_state->__pyx_kp_u_h_2);
  Py_CLEAR(clear_module_state->__pyx_kp_u_i);
  Py_CLEAR(clear_module_state->__pyx_n_s_identifier);
  Py_CLEAR(clear_module_state->__pyx_n_s_imag);
  Py_CLEAR(clear_module_state->__pyx_n_s_import);
  Py_CLEAR(clear_module_state->__pyx_n_s_index);
  Py_CLEAR(clear_module_state->__pyx_n_s_initializing);
  Py_CLEAR(clear_module_state->__pyx_kp_u_invalid_pandas_DataFrame_cannot);
  Py_CLEAR(clear_module_state->__pyx_kp_u_invalid_pandas_DataFrame_columns);
  Py_CLEAR(clear_module_state->__pyx_kp_u_invalid_pandas_DatetimeIndex_nam);
  Py_CLEAR(clear_module_state->__pyx_kp_u_invalid_pandas_Series_name_r);
  Py_CLEAR(clear_module_state->__pyx_kp_u_invalid_pandas_TimedeltaIndex_na);
  Py_CLEAR(clear_module_state->__pyx_n_s_is_coroutine);
  Py_CLEAR(clear_module_state->__pyx_n_s_items);
  Py_CLEAR(clear_module_state->__pyx_n_s_join);
  Py_CLEAR(clear_module_state->__pyx_n_s_ljust);
  Py_CLEAR(clear_module_state->__pyx_n_u_m);
  Py_CLEAR(clear_module_state->__pyx_n_s_main);
  Py_CLEAR(clear_module_state->__pyx_n_u_ms);
  Py_CLEAR(clear_module_state->__pyx_n_s_n);
  Py_CLEAR(clear_module_state->__pyx_n_s_name);
  Py_CLEAR(clear_module_state->__pyx_n_s_name_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_np);
  Py_CLEAR(clear_module_state->__pyx_n_u_ns);
  Py_CLEAR(clear_module_state->__pyx_kp_u_num);
  Py_CLEAR(clear_module_state->__pyx_n_s_num_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_numpy);
  Py_CLEAR(clear_module_state->__pyx_kp_u_numpy__core_multiarray_failed_to);
  Py_CLEAR(clear_module_state->__pyx_kp_u_numpy__core_umath_failed_to_impo);
  Py_CLEAR(clear_module_state->__pyx_n_s_obj);
  Py_CLEAR(clear_module_state->__pyx_n_s_object);
  Py_CLEAR(clear_module_state->__pyx_n_s_option);
  Py_CLEAR(clear_module_state->__pyx_n_s_orjson);
  Py_CLEAR(clear_module_state->__pyx_n_s_pack);
  Py_CLEAR(clear_module_state->__pyx_n_s_pandas);
  Py_CLEAR(clear_module_state->__pyx_kp_u_pass_s);
  Py_CLEAR(clear_module_state->__pyx_n_s_print);
  Py_CLEAR(clear_module_state->__pyx_n_u_ps);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_capi);
  Py_CLEAR(clear_module_state->__pyx_kp_u_q);
  Py_CLEAR(clear_module_state->__pyx_n_s_range);
  Py_CLEAR(clear_module_state->__pyx_n_s_real);
  Py_CLEAR(clear_module_state->__pyx_n_s_return);
  Py_CLEAR(clear_module_state->__pyx_n_s_s);
  Py_CLEAR(clear_module_state->__pyx_n_u_s);
  Py_CLEAR(clear_module_state->__pyx_n_s_serialize);
  Py_CLEAR(clear_module_state->__pyx_n_u_serialize);
  Py_CLEAR(clear_module_state->__pyx_n_s_serializor);
  Py_CLEAR(clear_module_state->__pyx_n_s_serializor_binary_ser);
  Py_CLEAR(clear_module_state->__pyx_n_s_spec);
  Py_CLEAR(clear_module_state->__pyx_kp_s_src_serializor_binary_ser_py);
  Py_CLEAR(clear_module_state->__pyx_n_s_start);
  Py_CLEAR(clear_module_state->__pyx_n_s_step);
  Py_CLEAR(clear_module_state->__pyx_n_s_stop);
  Py_CLEAR(clear_module_state->__pyx_n_s_str);
  Py_CLEAR(clear_module_state->__pyx_n_s_struct);
  Py_CLEAR(clear_module_state->__pyx_n_s_struct_time);
  Py_CLEAR(clear_module_state->__pyx_n_s_test);
  Py_CLEAR(clear_module_state->__pyx_n_s_test_gen_encoded_int);
  Py_CLEAR(clear_module_state->__pyx_n_s_test_gen_header);
  Py_CLEAR(clear_module_state->__pyx_n_s_test_pack_int16);
  Py_CLEAR(clear_module_state->__pyx_n_s_test_pack_int24);
  Py_CLEAR(clear_module_state->__pyx_n_s_test_pack_int32);
  Py_CLEAR(clear_module_state->__pyx_n_s_test_pack_int64);
  Py_CLEAR(clear_module_state->__pyx_n_s_test_pack_int8);
  Py_CLEAR(clear_module_state->__pyx_n_s_test_pack_uint16);
  Py_CLEAR(clear_module_state->__pyx_n_s_test_pack_uint24);
  Py_CLEAR(clear_module_state->__pyx_n_s_test_pack_uint32);
  Py_CLEAR(clear_module_state->__pyx_n_s_test_pack_uint64);
  Py_CLEAR(clear_module_state->__pyx_n_s_test_pack_uint8);
  Py_CLEAR(clear_module_state->__pyx_n_s_test_utils);
  Py_CLEAR(clear_module_state->__pyx_n_s_time);
  Py_CLEAR(clear_module_state->__pyx_n_s_tm_gmtoff);
  Py_CLEAR(clear_module_state->__pyx_n_s_tm_zone);
  Py_CLEAR(clear_module_state->__pyx_n_s_tobytes);
  Py_CLEAR(clear_module_state->__pyx_n_s_typeref);
  Py_CLEAR(clear_module_state->__pyx_n_s_typing);
  Py_CLEAR(clear_module_state->__pyx_n_s_tz);
  Py_CLEAR(clear_module_state->__pyx_kp_u_unable_to_parse_ndarray_time_uni);
  Py_CLEAR(clear_module_state->__pyx_kp_u_unsupported_data_type_s);
  Py_CLEAR(clear_module_state->__pyx_kp_u_unsupported_numpy_ndarray_dimens);
  Py_CLEAR(clear_module_state->__pyx_kp_u_unsupported_numpy_ndarray_dtype);
  Py_CLEAR(clear_module_state->__pyx_kp_u_unsupported_numpy_time_unit_d);
  Py_CLEAR(clear_module_state->__pyx_kp_u_unsupported_numpy_time_unit_s);
  Py_CLEAR(clear_module_state->__pyx_kp_u_unsupported_pandas_DataFrame_col);
  Py_CLEAR(clear_module_state->__pyx_kp_u_unsupported_pandas_Series_dtype);
  Py_CLEAR(clear_module_state->__pyx_n_u_us);
  Py_CLEAR(clear_module_state->__pyx_n_s_utcoffset);
  Py_CLEAR(clear_module_state->__pyx_n_s_utils);
  Py_CLEAR(clear_module_state->__pyx_n_s_v);
  Py_CLEAR(clear_module_state->__pyx_n_s_val);
  Py_CLEAR(clear_module_state->__pyx_n_s_value);
  Py_CLEAR(clear_module_state->__pyx_n_s_values);
  Py_CLEAR(clear_module_state->__pyx_int_0);
  Py_CLEAR(clear_module_state->__pyx_int_1);
  Py_CLEAR(clear_module_state->__pyx_int_3);
  Py_CLEAR(clear_module_state->__pyx_int_80);
  Py_CLEAR(clear_module_state->__pyx_int_251);
  Py_CLEAR(clear_module_state->__pyx_int_252);
  Py_CLEAR(clear_module_state->__pyx_int_65535);
  Py_CLEAR(clear_module_state->__pyx_int_65536);
  Py_CLEAR(clear_module_state->__pyx_int_8388606);
  Py_CLEAR(clear_module_state->__pyx_int_8388607);
  Py_CLEAR(clear_module_state->__pyx_int_16777213);
  Py_CLEAR(clear_module_state->__pyx_int_16777214);
  Py_CLEAR(clear_module_state->__pyx_int_16777215);
  Py_CLEAR(clear_module_state->__pyx_int_2147483646);
  Py_CLEAR(clear_module_state->__pyx_int_2147483647);
  Py_CLEAR(clear_module_state->__pyx_int_4294967293);
  Py_CLEAR(clear_module_state->__pyx_int_4294967294);
  Py_CLEAR(clear_module_state->__pyx_int_4294967295);
  Py_CLEAR(clear_module_state->__pyx_int_4294967296);
  Py_CLEAR(clear_module_state->__pyx_int_8589934592);
  Py_CLEAR(clear_module_state->__pyx_int_0x7ffffffffffffffe);
  Py_CLEAR(clear_module_state->__pyx_int_0x7fffffffffffffff);
  Py_CLEAR(clear_module_state->__pyx_int_0xfffffffffffffffd);
  Py_CLEAR(clear_module_state->__pyx_int_0xfffffffffffffffe);
  Py_CLEAR(clear_module_state->__pyx_int_0xffffffffffffffff);
  Py_CLEAR(clear_module_state->__pyx_int_neg_8388607);
  Py_CLEAR(clear_module_state->__pyx_int_neg_8388608);
  Py_CLEAR(clear_module_state->__pyx_int_neg_2147483647);
  Py_CLEAR(clear_module_state->__pyx_int_neg_2147483648);
  Py_CLEAR(clear_module_state->__pyx_int_neg_9223372036854775807);
  Py_CLEAR(clear_module_state->__pyx_int_neg_9223372036854775808);
  Py_CLEAR(clear_module_state->__pyx_tuple_);
  Py_CLEAR(clear_module_state->__pyx_tuple__2);
  Py_CLEAR(clear_module_state->__pyx_tuple__3);
  Py_CLEAR(clear_module_state->__pyx_slice__18);
  Py_CLEAR(clear_module_state->__pyx_tuple__10);
  Py_CLEAR(clear_module_state->__pyx_tuple__11);
  Py_CLEAR(clear_module_state->__pyx_tuple__12);
  Py_CLEAR(clear_module_state->__pyx_tuple__17);
  Py_CLEAR(clear_module_state->__pyx_tuple__19);
  Py_CLEAR(clear_module_state->__pyx_tuple__20);
  Py_CLEAR(clear_module_state->__pyx_tuple__21);
  Py_CLEAR(clear_module_state->__pyx_tuple__22);
  Py_CLEAR(clear_module_state->__pyx_tuple__23);
  Py_CLEAR(clear_module_state->__pyx_tuple__24);
  Py_CLEAR(clear_module_state->__pyx_tuple__25);
  Py_CLEAR(clear_module_state->__pyx_tuple__26);
  Py_CLEAR(clear_module_state->__pyx_tuple__28);
  Py_CLEAR(clear_module_state->__pyx_tuple__31);
  Py_CLEAR(clear_module_state->__pyx_tuple__42);
  Py_CLEAR(clear_module_state->__pyx_tuple__44);
  Py_CLEAR(clear_module_state->__pyx_codeobj__29);
  Py_CLEAR(clear_module_state->__pyx_codeobj__30);
  Py_CLEAR(clear_module_state->__pyx_codeobj__32);
  Py_CLEAR(clear_module_state->__pyx_codeobj__33);
  Py_CLEAR(clear_module_state->__pyx_codeobj__34);
  Py_CLEAR(clear_module_state->__pyx_codeobj__35);
  Py_CLEAR(clear_module_state->__pyx_codeobj__36);
  Py_CLEAR(clear_module_state->__pyx_codeobj__37);
  Py_CLEAR(clear_module_state->__pyx_codeobj__38);
  Py_CLEAR(clear_module_state->__pyx_codeobj__39);
  Py_CLEAR(clear_module_state->__pyx_codeobj__40);
  Py_CLEAR(clear_module_state->__pyx_codeobj__41);
  Py_CLEAR(clear_module_state->__pyx_codeobj__43);
  Py_CLEAR(clear_module_state->__pyx_codeobj__45);
  return 0;
}
#endif
/* #### Code section: module_state_traverse ### */
#if CYTHON_USE_MODULE_STATE
static int __pyx_m_traverse(PyObject *m, visitproc visit, void *arg) {
  __pyx_mstate *traverse_module_state = __pyx_mstate(m);
  if (!traverse_module_state) return 0;
  Py_VISIT(traverse_module_state->__pyx_d);
  Py_VISIT(traverse_module_state->__pyx_b);
  Py_VISIT(traverse_module_state->__pyx_cython_runtime);
  Py_VISIT(traverse_module_state->__pyx_empty_tuple);
  Py_VISIT(traverse_module_state->__pyx_empty_bytes);
  Py_VISIT(traverse_module_state->__pyx_empty_unicode);
  #ifdef __Pyx_CyFunction_USED
  Py_VISIT(traverse_module_state->__pyx_CyFunctionType);
  #endif
  #ifdef __Pyx_FusedFunction_USED
  Py_VISIT(traverse_module_state->__pyx_FusedFunctionType);
  #endif
  Py_VISIT(traverse_module_state->__pyx_ptype_7cpython_4type_type);
  Py_VISIT(traverse_module_state->__pyx_ptype_7cpython_4bool_bool);
  Py_VISIT(traverse_module_state->__pyx_ptype_7cpython_7complex_complex);
  Py_VISIT(traverse_module_state->__pyx_ptype_5numpy_dtype);
  Py_VISIT(traverse_module_state->__pyx_ptype_5numpy_flatiter);
  Py_VISIT(traverse_module_state->__pyx_ptype_5numpy_broadcast);
  Py_VISIT(traverse_module_state->__pyx_ptype_5numpy_ndarray);
  Py_VISIT(traverse_module_state->__pyx_ptype_5numpy_generic);
  Py_VISIT(traverse_module_state->__pyx_ptype_5numpy_number);
  Py_VISIT(traverse_module_state->__pyx_ptype_5numpy_integer);
  Py_VISIT(traverse_module_state->__pyx_ptype_5numpy_signedinteger);
  Py_VISIT(traverse_module_state->__pyx_ptype_5numpy_unsignedinteger);
  Py_VISIT(traverse_module_state->__pyx_ptype_5numpy_inexact);
  Py_VISIT(traverse_module_state->__pyx_ptype_5numpy_floating);
  Py_VISIT(traverse_module_state->__pyx_ptype_5numpy_complexfloating);
  Py_VISIT(traverse_module_state->__pyx_ptype_5numpy_flexible);
  Py_VISIT(traverse_module_state->__pyx_ptype_5numpy_character);
  Py_VISIT(traverse_module_state->__pyx_ptype_5numpy_ufunc);
  Py_VISIT(traverse_module_state->__pyx_ptype_7cpython_8datetime_date);
  Py_VISIT(traverse_module_state->__pyx_ptype_7cpython_8datetime_time);
  Py_VISIT(traverse_module_state->__pyx_ptype_7cpython_8datetime_datetime);
  Py_VISIT(traverse_module_state->__pyx_ptype_7cpython_8datetime_timedelta);
  Py_VISIT(traverse_module_state->__pyx_ptype_7cpython_8datetime_tzinfo);
  Py_VISIT(traverse_module_state->__pyx_n_s_AssertionError);
  Py_VISIT(traverse_module_state->__pyx_kp_u_B);
  Py_VISIT(traverse_module_state->__pyx_n_s_COMPLEX_ID);
  Py_VISIT(traverse_module_state->__pyx_n_u_D);
  Py_VISIT(traverse_module_state->__pyx_n_s_DATAFRAME_ID);
  Py_VISIT(traverse_module_state->__pyx_n_s_DATETIMEINDEX_ID);
  Py_VISIT(traverse_module_state->__pyx_n_s_DATETIME_ID);
  Py_VISIT(traverse_module_state->__pyx_n_s_DATE_ID);
  Py_VISIT(traverse_module_state->__pyx_n_s_DICT_ID);
  Py_VISIT(traverse_module_state->__pyx_n_s_DT64_ID);
  Py_VISIT(traverse_module_state->__pyx_n_s_DataFrame);
  Py_VISIT(traverse_module_state->__pyx_n_s_DatetimeIndex);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Error);
  Py_VISIT(traverse_module_state->__pyx_n_s_FALSE_VALUE);
  Py_VISIT(traverse_module_state->__pyx_n_s_FROZENSET_ID);
  Py_VISIT(traverse_module_state->__pyx_kp_u_H);
  Py_VISIT(traverse_module_state->__pyx_kp_u_I);
  Py_VISIT(traverse_module_state->__pyx_n_s_ImportError);
  Py_VISIT(traverse_module_state->__pyx_n_s_Iterable);
  Py_VISIT(traverse_module_state->__pyx_n_s_LIST_ID);
  Py_VISIT(traverse_module_state->__pyx_n_u_M);
  Py_VISIT(traverse_module_state->__pyx_n_s_NDARRAY_BOOL_DT);
  Py_VISIT(traverse_module_state->__pyx_n_s_NDARRAY_BOOL_IDDT);
  Py_VISIT(traverse_module_state->__pyx_n_s_NDARRAY_BYTES_DT);
  Py_VISIT(traverse_module_state->__pyx_n_s_NDARRAY_BYTES_IDDT);
  Py_VISIT(traverse_module_state->__pyx_n_s_NDARRAY_COMPLEX_DT);
  Py_VISIT(traverse_module_state->__pyx_n_s_NDARRAY_COMPLEX_IDDT);
  Py_VISIT(traverse_module_state->__pyx_n_s_NDARRAY_DT64_DT);
  Py_VISIT(traverse_module_state->__pyx_n_s_NDARRAY_DT64_IDDT);
  Py_VISIT(traverse_module_state->__pyx_n_s_NDARRAY_FLOAT_DT);
  Py_VISIT(traverse_module_state->__pyx_n_s_NDARRAY_FLOAT_IDDT);
  Py_VISIT(traverse_module_state->__pyx_n_s_NDARRAY_ID);
  Py_VISIT(traverse_module_state->__pyx_n_s_NDARRAY_INT_DT);
  Py_VISIT(traverse_module_state->__pyx_n_s_NDARRAY_INT_IDDT);
  Py_VISIT(traverse_module_state->__pyx_n_s_NDARRAY_OBJECT_DT);
  Py_VISIT(traverse_module_state->__pyx_n_s_NDARRAY_OBJECT_IDDT);
  Py_VISIT(traverse_module_state->__pyx_n_s_NDARRAY_TD64_DT);
  Py_VISIT(traverse_module_state->__pyx_n_s_NDARRAY_TD64_IDDT);
  Py_VISIT(traverse_module_state->__pyx_n_s_NDARRAY_UINT_DT);
  Py_VISIT(traverse_module_state->__pyx_n_s_NDARRAY_UINT_IDDT);
  Py_VISIT(traverse_module_state->__pyx_n_s_NDARRAY_UNICODE_DT);
  Py_VISIT(traverse_module_state->__pyx_n_s_NDARRAY_UNICODE_IDDT);
  Py_VISIT(traverse_module_state->__pyx_n_s_NONE_VALUE);
  Py_VISIT(traverse_module_state->__pyx_n_s_None);
  Py_VISIT(traverse_module_state->__pyx_n_s_OPT_SERIALIZE_NUMPY);
  Py_VISIT(traverse_module_state->__pyx_n_s_PD_TIMEDELTA_ID);
  Py_VISIT(traverse_module_state->__pyx_n_s_PD_TIMESTAMP_ID);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Pass_Gen_Encoded_Integer);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Pass_Gen_Header);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Pass_Pack_int16);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Pass_Pack_int24);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Pass_Pack_int32);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Pass_Pack_int64);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Pass_Pack_int8);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Pass_Pack_uint16);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Pass_Pack_uint24);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Pass_Pack_uint32);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Pass_Pack_uint64);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Pass_Pack_uint8);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Q);
  Py_VISIT(traverse_module_state->__pyx_n_s_RANGE_ID);
  Py_VISIT(traverse_module_state->__pyx_n_s_RuntimeError);
  Py_VISIT(traverse_module_state->__pyx_n_s_SERIES_ID);
  Py_VISIT(traverse_module_state->__pyx_n_s_SET_ID);
  Py_VISIT(traverse_module_state->__pyx_n_s_STRUCT_TIME_ID);
  Py_VISIT(traverse_module_state->__pyx_n_s_SerializeError);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Serializor_Failed_to_serialize);
  Py_VISIT(traverse_module_state->__pyx_n_s_Series);
  Py_VISIT(traverse_module_state->__pyx_n_s_TD64_ID);
  Py_VISIT(traverse_module_state->__pyx_n_s_TIMEDELTAINDEX_ID);
  Py_VISIT(traverse_module_state->__pyx_n_s_TIMEDELTA_ID);
  Py_VISIT(traverse_module_state->__pyx_n_s_TIME_ID);
  Py_VISIT(traverse_module_state->__pyx_n_s_TRUE_VALUE);
  Py_VISIT(traverse_module_state->__pyx_n_s_TUPLE_ID);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Time_zones_are_not_available_fro);
  Py_VISIT(traverse_module_state->__pyx_n_s_Timedelta);
  Py_VISIT(traverse_module_state->__pyx_n_s_TimedeltaIndex);
  Py_VISIT(traverse_module_state->__pyx_n_s_Timestamp);
  Py_VISIT(traverse_module_state->__pyx_n_s_TypeError);
  Py_VISIT(traverse_module_state->__pyx_n_s_UINT16_ENCODE_VALUE);
  Py_VISIT(traverse_module_state->__pyx_n_s_UINT32_ENCODE_VALUE);
  Py_VISIT(traverse_module_state->__pyx_n_s_UINT64_ENCODE_VALUE);
  Py_VISIT(traverse_module_state->__pyx_n_s_UINT8_ENCODE_VALUE);
  Py_VISIT(traverse_module_state->__pyx_n_s_ValueError);
  Py_VISIT(traverse_module_state->__pyx_n_u_W);
  Py_VISIT(traverse_module_state->__pyx_n_u_Y);
  Py_VISIT(traverse_module_state->__pyx_kp_u__13);
  Py_VISIT(traverse_module_state->__pyx_n_s__14);
  Py_VISIT(traverse_module_state->__pyx_kp_u__15);
  Py_VISIT(traverse_module_state->__pyx_kp_u__16);
  Py_VISIT(traverse_module_state->__pyx_kp_u__27);
  Py_VISIT(traverse_module_state->__pyx_kp_b__4);
  Py_VISIT(traverse_module_state->__pyx_n_s__46);
  Py_VISIT(traverse_module_state->__pyx_kp_b__5);
  Py_VISIT(traverse_module_state->__pyx_kp_b__6);
  Py_VISIT(traverse_module_state->__pyx_kp_b__7);
  Py_VISIT(traverse_module_state->__pyx_kp_b__8);
  Py_VISIT(traverse_module_state->__pyx_kp_b__9);
  Py_VISIT(traverse_module_state->__pyx_n_s_all);
  Py_VISIT(traverse_module_state->__pyx_n_u_as);
  Py_VISIT(traverse_module_state->__pyx_n_s_asyncio_coroutines);
  Py_VISIT(traverse_module_state->__pyx_kp_u_b);
  Py_VISIT(traverse_module_state->__pyx_n_s_b_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_class_getitem);
  Py_VISIT(traverse_module_state->__pyx_n_s_cline_in_traceback);
  Py_VISIT(traverse_module_state->__pyx_n_s_columns);
  Py_VISIT(traverse_module_state->__pyx_n_s_datetime);
  Py_VISIT(traverse_module_state->__pyx_n_s_dchar);
  Py_VISIT(traverse_module_state->__pyx_n_s_dt);
  Py_VISIT(traverse_module_state->__pyx_n_s_dtype);
  Py_VISIT(traverse_module_state->__pyx_n_s_dumps);
  Py_VISIT(traverse_module_state->__pyx_n_s_dumps_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_end);
  Py_VISIT(traverse_module_state->__pyx_n_s_errors);
  Py_VISIT(traverse_module_state->__pyx_kp_u_expects_np_ndarray_dtype_of_date);
  Py_VISIT(traverse_module_state->__pyx_kp_u_expects_pandas_DatetimeIndex_got);
  Py_VISIT(traverse_module_state->__pyx_kp_u_expects_pandas_Series_got_s);
  Py_VISIT(traverse_module_state->__pyx_kp_u_expects_pandas_TimedeltaIndex_go);
  Py_VISIT(traverse_module_state->__pyx_kp_u_expects_pandas_Timedelta_got_s);
  Py_VISIT(traverse_module_state->__pyx_kp_u_expects_pandas_Timestamp_got_s);
  Py_VISIT(traverse_module_state->__pyx_n_u_fs);
  Py_VISIT(traverse_module_state->__pyx_n_u_h);
  Py_VISIT(traverse_module_state->__pyx_kp_u_h_2);
  Py_VISIT(traverse_module_state->__pyx_kp_u_i);
  Py_VISIT(traverse_module_state->__pyx_n_s_identifier);
  Py_VISIT(traverse_module_state->__pyx_n_s_imag);
  Py_VISIT(traverse_module_state->__pyx_n_s_import);
  Py_VISIT(traverse_module_state->__pyx_n_s_index);
  Py_VISIT(traverse_module_state->__pyx_n_s_initializing);
  Py_VISIT(traverse_module_state->__pyx_kp_u_invalid_pandas_DataFrame_cannot);
  Py_VISIT(traverse_module_state->__pyx_kp_u_invalid_pandas_DataFrame_columns);
  Py_VISIT(traverse_module_state->__pyx_kp_u_invalid_pandas_DatetimeIndex_nam);
  Py_VISIT(traverse_module_state->__pyx_kp_u_invalid_pandas_Series_name_r);
  Py_VISIT(traverse_module_state->__pyx_kp_u_invalid_pandas_TimedeltaIndex_na);
  Py_VISIT(traverse_module_state->__pyx_n_s_is_coroutine);
  Py_VISIT(traverse_module_state->__pyx_n_s_items);
  Py_VISIT(traverse_module_state->__pyx_n_s_join);
  Py_VISIT(traverse_module_state->__pyx_n_s_ljust);
  Py_VISIT(traverse_module_state->__pyx_n_u_m);
  Py_VISIT(traverse_module_state->__pyx_n_s_main);
  Py_VISIT(traverse_module_state->__pyx_n_u_ms);
  Py_VISIT(traverse_module_state->__pyx_n_s_n);
  Py_VISIT(traverse_module_state->__pyx_n_s_name);
  Py_VISIT(traverse_module_state->__pyx_n_s_name_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_np);
  Py_VISIT(traverse_module_state->__pyx_n_u_ns);
  Py_VISIT(traverse_module_state->__pyx_kp_u_num);
  Py_VISIT(traverse_module_state->__pyx_n_s_num_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_numpy);
  Py_VISIT(traverse_module_state->__pyx_kp_u_numpy__core_multiarray_failed_to);
  Py_VISIT(traverse_module_state->__pyx_kp_u_numpy__core_umath_failed_to_impo);
  Py_VISIT(traverse_module_state->__pyx_n_s_obj);
  Py_VISIT(traverse_module_state->__pyx_n_s_object);
  Py_VISIT(traverse_module_state->__pyx_n_s_option);
  Py_VISIT(traverse_module_state->__pyx_n_s_orjson);
  Py_VISIT(traverse_module_state->__pyx_n_s_pack);
  Py_VISIT(traverse_module_state->__pyx_n_s_pandas);
  Py_VISIT(traverse_module_state->__pyx_kp_u_pass_s);
  Py_VISIT(traverse_module_state->__pyx_n_s_print);
  Py_VISIT(traverse_module_state->__pyx_n_u_ps);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_capi);
  Py_VISIT(traverse_module_state->__pyx_kp_u_q);
  Py_VISIT(traverse_module_state->__pyx_n_s_range);
  Py_VISIT(traverse_module_state->__pyx_n_s_real);
  Py_VISIT(traverse_module_state->__pyx_n_s_return);
  Py_VISIT(traverse_module_state->__pyx_n_s_s);
  Py_VISIT(traverse_module_state->__pyx_n_u_s);
  Py_VISIT(traverse_module_state->__pyx_n_s_serialize);
  Py_VISIT(traverse_module_state->__pyx_n_u_serialize);
  Py_VISIT(traverse_module_state->__pyx_n_s_serializor);
  Py_VISIT(traverse_module_state->__pyx_n_s_serializor_binary_ser);
  Py_VISIT(traverse_module_state->__pyx_n_s_spec);
  Py_VISIT(traverse_module_state->__pyx_kp_s_src_serializor_binary_ser_py);
  Py_VISIT(traverse_module_state->__pyx_n_s_start);
  Py_VISIT(traverse_module_state->__pyx_n_s_step);
  Py_VISIT(traverse_module_state->__pyx_n_s_stop);
  Py_VISIT(traverse_module_state->__pyx_n_s_str);
  Py_VISIT(traverse_module_state->__pyx_n_s_struct);
  Py_VISIT(traverse_module_state->__pyx_n_s_struct_time);
  Py_VISIT(traverse_module_state->__pyx_n_s_test);
  Py_VISIT(traverse_module_state->__pyx_n_s_test_gen_encoded_int);
  Py_VISIT(traverse_module_state->__pyx_n_s_test_gen_header);
  Py_VISIT(traverse_module_state->__pyx_n_s_test_pack_int16);
  Py_VISIT(traverse_module_state->__pyx_n_s_test_pack_int24);
  Py_VISIT(traverse_module_state->__pyx_n_s_test_pack_int32);
  Py_VISIT(traverse_module_state->__pyx_n_s_test_pack_int64);
  Py_VISIT(traverse_module_state->__pyx_n_s_test_pack_int8);
  Py_VISIT(traverse_module_state->__pyx_n_s_test_pack_uint16);
  Py_VISIT(traverse_module_state->__pyx_n_s_test_pack_uint24);
  Py_VISIT(traverse_module_state->__pyx_n_s_test_pack_uint32);
  Py_VISIT(traverse_module_state->__pyx_n_s_test_pack_uint64);
  Py_VISIT(traverse_module_state->__pyx_n_s_test_pack_uint8);
  Py_VISIT(traverse_module_state->__pyx_n_s_test_utils);
  Py_VISIT(traverse_module_state->__pyx_n_s_time);
  Py_VISIT(traverse_module_state->__pyx_n_s_tm_gmtoff);
  Py_VISIT(traverse_module_state->__pyx_n_s_tm_zone);
  Py_VISIT(traverse_module_state->__pyx_n_s_tobytes);
  Py_VISIT(traverse_module_state->__pyx_n_s_typeref);
  Py_VISIT(traverse_module_state->__pyx_n_s_typing);
  Py_VISIT(traverse_module_state->__pyx_n_s_tz);
  Py_VISIT(traverse_module_state->__pyx_kp_u_unable_to_parse_ndarray_time_uni);
  Py_VISIT(traverse_module_state->__pyx_kp_u_unsupported_data_type_s);
  Py_VISIT(traverse_module_state->__pyx_kp_u_unsupported_numpy_ndarray_dimens);
  Py_VISIT(traverse_module_state->__pyx_kp_u_unsupported_numpy_ndarray_dtype);
  Py_VISIT(traverse_module_state->__pyx_kp_u_unsupported_numpy_time_unit_d);
  Py_VISIT(traverse_module_state->__pyx_kp_u_unsupported_numpy_time_unit_s);
  Py_VISIT(traverse_module_state->__pyx_kp_u_unsupported_pandas_DataFrame_col);
  Py_VISIT(traverse_module_state->__pyx_kp_u_unsupported_pandas_Series_dtype);
  Py_VISIT(traverse_module_state->__pyx_n_u_us);
  Py_VISIT(traverse_module_state->__pyx_n_s_utcoffset);
  Py_VISIT(traverse_module_state->__pyx_n_s_utils);
  Py_VISIT(traverse_module_state->__pyx_n_s_v);
  Py_VISIT(traverse_module_state->__pyx_n_s_val);
  Py_VISIT(traverse_module_state->__pyx_n_s_value);
  Py_VISIT(traverse_module_state->__pyx_n_s_values);
  Py_VISIT(traverse_module_state->__pyx_int_0);
  Py_VISIT(traverse_module_state->__pyx_int_1);
  Py_VISIT(traverse_module_state->__pyx_int_3);
  Py_VISIT(traverse_module_state->__pyx_int_80);
  Py_VISIT(traverse_module_state->__pyx_int_251);
  Py_VISIT(traverse_module_state->__pyx_int_252);
  Py_VISIT(traverse_module_state->__pyx_int_65535);
  Py_VISIT(traverse_module_state->__pyx_int_65536);
  Py_VISIT(traverse_module_state->__pyx_int_8388606);
  Py_VISIT(traverse_module_state->__pyx_int_8388607);
  Py_VISIT(traverse_module_state->__pyx_int_16777213);
  Py_VISIT(traverse_module_state->__pyx_int_16777214);
  Py_VISIT(traverse_module_state->__pyx_int_16777215);
  Py_VISIT(traverse_module_state->__pyx_int_2147483646);
  Py_VISIT(traverse_module_state->__pyx_int_2147483647);
  Py_VISIT(traverse_module_state->__pyx_int_4294967293);
  Py_VISIT(traverse_module_state->__pyx_int_4294967294);
  Py_VISIT(traverse_module_state->__pyx_int_4294967295);
  Py_VISIT(traverse_module_state->__pyx_int_4294967296);
  Py_VISIT(traverse_module_state->__pyx_int_8589934592);
  Py_VISIT(traverse_module_state->__pyx_int_0x7ffffffffffffffe);
  Py_VISIT(traverse_module_state->__pyx_int_0x7fffffffffffffff);
  Py_VISIT(traverse_module_state->__pyx_int_0xfffffffffffffffd);
  Py_VISIT(traverse_module_state->__pyx_int_0xfffffffffffffffe);
  Py_VISIT(traverse_module_state->__pyx_int_0xffffffffffffffff);
  Py_VISIT(traverse_module_state->__pyx_int_neg_8388607);
  Py_VISIT(traverse_module_state->__pyx_int_neg_8388608);
  Py_VISIT(traverse_module_state->__pyx_int_neg_2147483647);
  Py_VISIT(traverse_module_state->__pyx_int_neg_2147483648);
  Py_VISIT(traverse_module_state->__pyx_int_neg_9223372036854775807);
  Py_VISIT(traverse_module_state->__pyx_int_neg_9223372036854775808);
  Py_VISIT(traverse_module_state->__pyx_tuple_);
  Py_VISIT(traverse_module_state->__pyx_tuple__2);
  Py_VISIT(traverse_module_state->__pyx_tuple__3);
  Py_VISIT(traverse_module_state->__pyx_slice__18);
  Py_VISIT(traverse_module_state->__pyx_tuple__10);
  Py_VISIT(traverse_module_state->__pyx_tuple__11);
  Py_VISIT(traverse_module_state->__pyx_tuple__12);
  Py_VISIT(traverse_module_state->__pyx_tuple__17);
  Py_VISIT(traverse_module_state->__pyx_tuple__19);
  Py_VISIT(traverse_module_state->__pyx_tuple__20);
  Py_VISIT(traverse_module_state->__pyx_tuple__21);
  Py_VISIT(traverse_module_state->__pyx_tuple__22);
  Py_VISIT(traverse_module_state->__pyx_tuple__23);
  Py_VISIT(traverse_module_state->__pyx_tuple__24);
  Py_VISIT(traverse_module_state->__pyx_tuple__25);
  Py_VISIT(traverse_module_state->__pyx_tuple__26);
  Py_VISIT(traverse_module_state->__pyx_tuple__28);
  Py_VISIT(traverse_module_state->__pyx_tuple__31);
  Py_VISIT(traverse_module_state->__pyx_tuple__42);
  Py_VISIT(traverse_module_state->__pyx_tuple__44);
  Py_VISIT(traverse_module_state->__pyx_codeobj__29);
  Py_VISIT(traverse_module_state->__pyx_codeobj__30);
  Py_VISIT(traverse_module_state->__pyx_codeobj__32);
  Py_VISIT(traverse_module_state->__pyx_codeobj__33);
  Py_VISIT(traverse_module_state->__pyx_codeobj__34);
  Py_VISIT(traverse_module_state->__pyx_codeobj__35);
  Py_VISIT(traverse_module_state->__pyx_codeobj__36);
  Py_VISIT(traverse_module_state->__pyx_codeobj__37);
  Py_VISIT(traverse_module_state->__pyx_codeobj__38);
  Py_VISIT(traverse_module_state->__pyx_codeobj__39);
  Py_VISIT(traverse_module_state->__pyx_codeobj__40);
  Py_VISIT(traverse_module_state->__pyx_codeobj__41);
  Py_VISIT(traverse_module_state->__pyx_codeobj__43);
  Py_VISIT(traverse_module_state->__pyx_codeobj__45);
  return 0;
}
#endif
/* #### Code section: module_state_defines ### */
#define __pyx_d __pyx_mstate_global->__pyx_d
#define __pyx_b __pyx_mstate_global->__pyx_b
#define __pyx_cython_runtime __pyx_mstate_global->__pyx_cython_runtime
#define __pyx_empty_tuple __pyx_mstate_global->__pyx_empty_tuple
#define __pyx_empty_bytes __pyx_mstate_global->__pyx_empty_bytes
#define __pyx_empty_unicode __pyx_mstate_global->__pyx_empty_unicode
#ifdef __Pyx_CyFunction_USED
#define __pyx_CyFunctionType __pyx_mstate_global->__pyx_CyFunctionType
#endif
#ifdef __Pyx_FusedFunction_USED
#define __pyx_FusedFunctionType __pyx_mstate_global->__pyx_FusedFunctionType
#endif
#ifdef __Pyx_Generator_USED
#define __pyx_GeneratorType __pyx_mstate_global->__pyx_GeneratorType
#endif
#ifdef __Pyx_IterableCoroutine_USED
#define __pyx_IterableCoroutineType __pyx_mstate_global->__pyx_IterableCoroutineType
#endif
#ifdef __Pyx_Coroutine_USED
#define __pyx_CoroutineAwaitType __pyx_mstate_global->__pyx_CoroutineAwaitType
#endif
#ifdef __Pyx_Coroutine_USED
#define __pyx_CoroutineType __pyx_mstate_global->__pyx_CoroutineType
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#define __pyx_ptype_7cpython_4type_type __pyx_mstate_global->__pyx_ptype_7cpython_4type_type
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#define __pyx_ptype_7cpython_4bool_bool __pyx_mstate_global->__pyx_ptype_7cpython_4bool_bool
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#define __pyx_ptype_7cpython_7complex_complex __pyx_mstate_global->__pyx_ptype_7cpython_7complex_complex
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#define __pyx_ptype_5numpy_dtype __pyx_mstate_global->__pyx_ptype_5numpy_dtype
#define __pyx_ptype_5numpy_flatiter __pyx_mstate_global->__pyx_ptype_5numpy_flatiter
#define __pyx_ptype_5numpy_broadcast __pyx_mstate_global->__pyx_ptype_5numpy_broadcast
#define __pyx_ptype_5numpy_ndarray __pyx_mstate_global->__pyx_ptype_5numpy_ndarray
#define __pyx_ptype_5numpy_generic __pyx_mstate_global->__pyx_ptype_5numpy_generic
#define __pyx_ptype_5numpy_number __pyx_mstate_global->__pyx_ptype_5numpy_number
#define __pyx_ptype_5numpy_integer __pyx_mstate_global->__pyx_ptype_5numpy_integer
#define __pyx_ptype_5numpy_signedinteger __pyx_mstate_global->__pyx_ptype_5numpy_signedinteger
#define __pyx_ptype_5numpy_unsignedinteger __pyx_mstate_global->__pyx_ptype_5numpy_unsignedinteger
#define __pyx_ptype_5numpy_inexact __pyx_mstate_global->__pyx_ptype_5numpy_inexact
#define __pyx_ptype_5numpy_floating __pyx_mstate_global->__pyx_ptype_5numpy_floating
#define __pyx_ptype_5numpy_complexfloating __pyx_mstate_global->__pyx_ptype_5numpy_complexfloating
#define __pyx_ptype_5numpy_flexible __pyx_mstate_global->__pyx_ptype_5numpy_flexible
#define __pyx_ptype_5numpy_character __pyx_mstate_global->__pyx_ptype_5numpy_character
#define __pyx_ptype_5numpy_ufunc __pyx_mstate_global->__pyx_ptype_5numpy_ufunc
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#define __pyx_ptype_7cpython_8datetime_date __pyx_mstate_global->__pyx_ptype_7cpython_8datetime_date
#define __pyx_ptype_7cpython_8datetime_time __pyx_mstate_global->__pyx_ptype_7cpython_8datetime_time
#define __pyx_ptype_7cpython_8datetime_datetime __pyx_mstate_global->__pyx_ptype_7cpython_8datetime_datetime
#define __pyx_ptype_7cpython_8datetime_timedelta __pyx_mstate_global->__pyx_ptype_7cpython_8datetime_timedelta
#define __pyx_ptype_7cpython_8datetime_tzinfo __pyx_mstate_global->__pyx_ptype_7cpython_8datetime_tzinfo
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#define __pyx_n_s_AssertionError __pyx_mstate_global->__pyx_n_s_AssertionError
#define __pyx_kp_u_B __pyx_mstate_global->__pyx_kp_u_B
#define __pyx_n_s_COMPLEX_ID __pyx_mstate_global->__pyx_n_s_COMPLEX_ID
#define __pyx_n_u_D __pyx_mstate_global->__pyx_n_u_D
#define __pyx_n_s_DATAFRAME_ID __pyx_mstate_global->__pyx_n_s_DATAFRAME_ID
#define __pyx_n_s_DATETIMEINDEX_ID __pyx_mstate_global->__pyx_n_s_DATETIMEINDEX_ID
#define __pyx_n_s_DATETIME_ID __pyx_mstate_global->__pyx_n_s_DATETIME_ID
#define __pyx_n_s_DATE_ID __pyx_mstate_global->__pyx_n_s_DATE_ID
#define __pyx_n_s_DICT_ID __pyx_mstate_global->__pyx_n_s_DICT_ID
#define __pyx_n_s_DT64_ID __pyx_mstate_global->__pyx_n_s_DT64_ID
#define __pyx_n_s_DataFrame __pyx_mstate_global->__pyx_n_s_DataFrame
#define __pyx_n_s_DatetimeIndex __pyx_mstate_global->__pyx_n_s_DatetimeIndex
#define __pyx_kp_u_Error __pyx_mstate_global->__pyx_kp_u_Error
#define __pyx_n_s_FALSE_VALUE __pyx_mstate_global->__pyx_n_s_FALSE_VALUE
#define __pyx_n_s_FROZENSET_ID __pyx_mstate_global->__pyx_n_s_FROZENSET_ID
#define __pyx_kp_u_H __pyx_mstate_global->__pyx_kp_u_H
#define __pyx_kp_u_I __pyx_mstate_global->__pyx_kp_u_I
#define __pyx_n_s_ImportError __pyx_mstate_global->__pyx_n_s_ImportError
#define __pyx_n_s_Iterable __pyx_mstate_global->__pyx_n_s_Iterable
#define __pyx_n_s_LIST_ID __pyx_mstate_global->__pyx_n_s_LIST_ID
#define __pyx_n_u_M __pyx_mstate_global->__pyx_n_u_M
#define __pyx_n_s_NDARRAY_BOOL_DT __pyx_mstate_global->__pyx_n_s_NDARRAY_BOOL_DT
#define __pyx_n_s_NDARRAY_BOOL_IDDT __pyx_mstate_global->__pyx_n_s_NDARRAY_BOOL_IDDT
#define __pyx_n_s_NDARRAY_BYTES_DT __pyx_mstate_global->__pyx_n_s_NDARRAY_BYTES_DT
#define __pyx_n_s_NDARRAY_BYTES_IDDT __pyx_mstate_global->__pyx_n_s_NDARRAY_BYTES_IDDT
#define __pyx_n_s_NDARRAY_COMPLEX_DT __pyx_mstate_global->__pyx_n_s_NDARRAY_COMPLEX_DT
#define __pyx_n_s_NDARRAY_COMPLEX_IDDT __pyx_mstate_global->__pyx_n_s_NDARRAY_COMPLEX_IDDT
#define __pyx_n_s_NDARRAY_DT64_DT __pyx_mstate_global->__pyx_n_s_NDARRAY_DT64_DT
#define __pyx_n_s_NDARRAY_DT64_IDDT __pyx_mstate_global->__pyx_n_s_NDARRAY_DT64_IDDT
#define __pyx_n_s_NDARRAY_FLOAT_DT __pyx_mstate_global->__pyx_n_s_NDARRAY_FLOAT_DT
#define __pyx_n_s_NDARRAY_FLOAT_IDDT __pyx_mstate_global->__pyx_n_s_NDARRAY_FLOAT_IDDT
#define __pyx_n_s_NDARRAY_ID __pyx_mstate_global->__pyx_n_s_NDARRAY_ID
#define __pyx_n_s_NDARRAY_INT_DT __pyx_mstate_global->__pyx_n_s_NDARRAY_INT_DT
#define __pyx_n_s_NDARRAY_INT_IDDT __pyx_mstate_global->__pyx_n_s_NDARRAY_INT_IDDT
#define __pyx_n_s_NDARRAY_OBJECT_DT __pyx_mstate_global->__pyx_n_s_NDARRAY_OBJECT_DT
#define __pyx_n_s_NDARRAY_OBJECT_IDDT __pyx_mstate_global->__pyx_n_s_NDARRAY_OBJECT_IDDT
#define __pyx_n_s_NDARRAY_TD64_DT __pyx_mstate_global->__pyx_n_s_NDARRAY_TD64_DT
#define __pyx_n_s_NDARRAY_TD64_IDDT __pyx_mstate_global->__pyx_n_s_NDARRAY_TD64_IDDT
#define __pyx_n_s_NDARRAY_UINT_DT __pyx_mstate_global->__pyx_n_s_NDARRAY_UINT_DT
#define __pyx_n_s_NDARRAY_UINT_IDDT __pyx_mstate_global->__pyx_n_s_NDARRAY_UINT_IDDT
#define __pyx_n_s_NDARRAY_UNICODE_DT __pyx_mstate_global->__pyx_n_s_NDARRAY_UNICODE_DT
#define __pyx_n_s_NDARRAY_UNICODE_IDDT __pyx_mstate_global->__pyx_n_s_NDARRAY_UNICODE_IDDT
#define __pyx_n_s_NONE_VALUE __pyx_mstate_global->__pyx_n_s_NONE_VALUE
#define __pyx_n_s_None __pyx_mstate_global->__pyx_n_s_None
#define __pyx_n_s_OPT_SERIALIZE_NUMPY __pyx_mstate_global->__pyx_n_s_OPT_SERIALIZE_NUMPY
#define __pyx_n_s_PD_TIMEDELTA_ID __pyx_mstate_global->__pyx_n_s_PD_TIMEDELTA_ID
#define __pyx_n_s_PD_TIMESTAMP_ID __pyx_mstate_global->__pyx_n_s_PD_TIMESTAMP_ID
#define __pyx_kp_u_Pass_Gen_Encoded_Integer __pyx_mstate_global->__pyx_kp_u_Pass_Gen_Encoded_Integer
#define __pyx_kp_u_Pass_Gen_Header __pyx_mstate_global->__pyx_kp_u_Pass_Gen_Header
#define __pyx_kp_u_Pass_Pack_int16 __pyx_mstate_global->__pyx_kp_u_Pass_Pack_int16
#define __pyx_kp_u_Pass_Pack_int24 __pyx_mstate_global->__pyx_kp_u_Pass_Pack_int24
#define __pyx_kp_u_Pass_Pack_int32 __pyx_mstate_global->__pyx_kp_u_Pass_Pack_int32
#define __pyx_kp_u_Pass_Pack_int64 __pyx_mstate_global->__pyx_kp_u_Pass_Pack_int64
#define __pyx_kp_u_Pass_Pack_int8 __pyx_mstate_global->__pyx_kp_u_Pass_Pack_int8
#define __pyx_kp_u_Pass_Pack_uint16 __pyx_mstate_global->__pyx_kp_u_Pass_Pack_uint16
#define __pyx_kp_u_Pass_Pack_uint24 __pyx_mstate_global->__pyx_kp_u_Pass_Pack_uint24
#define __pyx_kp_u_Pass_Pack_uint32 __pyx_mstate_global->__pyx_kp_u_Pass_Pack_uint32
#define __pyx_kp_u_Pass_Pack_uint64 __pyx_mstate_global->__pyx_kp_u_Pass_Pack_uint64
#define __pyx_kp_u_Pass_Pack_uint8 __pyx_mstate_global->__pyx_kp_u_Pass_Pack_uint8
#define __pyx_kp_u_Q __pyx_mstate_global->__pyx_kp_u_Q
#define __pyx_n_s_RANGE_ID __pyx_mstate_global->__pyx_n_s_RANGE_ID
#define __pyx_n_s_RuntimeError __pyx_mstate_global->__pyx_n_s_RuntimeError
#define __pyx_n_s_SERIES_ID __pyx_mstate_global->__pyx_n_s_SERIES_ID
#define __pyx_n_s_SET_ID __pyx_mstate_global->__pyx_n_s_SET_ID
#define __pyx_n_s_STRUCT_TIME_ID __pyx_mstate_global->__pyx_n_s_STRUCT_TIME_ID
#define __pyx_n_s_SerializeError __pyx_mstate_global->__pyx_n_s_SerializeError
#define __pyx_kp_u_Serializor_Failed_to_serialize __pyx_mstate_global->__pyx_kp_u_Serializor_Failed_to_serialize
#define __pyx_n_s_Series __pyx_mstate_global->__pyx_n_s_Series
#define __pyx_n_s_TD64_ID __pyx_mstate_global->__pyx_n_s_TD64_ID
#define __pyx_n_s_TIMEDELTAINDEX_ID __pyx_mstate_global->__pyx_n_s_TIMEDELTAINDEX_ID
#define __pyx_n_s_TIMEDELTA_ID __pyx_mstate_global->__pyx_n_s_TIMEDELTA_ID
#define __pyx_n_s_TIME_ID __pyx_mstate_global->__pyx_n_s_TIME_ID
#define __pyx_n_s_TRUE_VALUE __pyx_mstate_global->__pyx_n_s_TRUE_VALUE
#define __pyx_n_s_TUPLE_ID __pyx_mstate_global->__pyx_n_s_TUPLE_ID
#define __pyx_kp_u_Time_zones_are_not_available_fro __pyx_mstate_global->__pyx_kp_u_Time_zones_are_not_available_fro
#define __pyx_n_s_Timedelta __pyx_mstate_global->__pyx_n_s_Timedelta
#define __pyx_n_s_TimedeltaIndex __pyx_mstate_global->__pyx_n_s_TimedeltaIndex
#define __pyx_n_s_Timestamp __pyx_mstate_global->__pyx_n_s_Timestamp
#define __pyx_n_s_TypeError __pyx_mstate_global->__pyx_n_s_TypeError
#define __pyx_n_s_UINT16_ENCODE_VALUE __pyx_mstate_global->__pyx_n_s_UINT16_ENCODE_VALUE
#define __pyx_n_s_UINT32_ENCODE_VALUE __pyx_mstate_global->__pyx_n_s_UINT32_ENCODE_VALUE
#define __pyx_n_s_UINT64_ENCODE_VALUE __pyx_mstate_global->__pyx_n_s_UINT64_ENCODE_VALUE
#define __pyx_n_s_UINT8_ENCODE_VALUE __pyx_mstate_global->__pyx_n_s_UINT8_ENCODE_VALUE
#define __pyx_n_s_ValueError __pyx_mstate_global->__pyx_n_s_ValueError
#define __pyx_n_u_W __pyx_mstate_global->__pyx_n_u_W
#define __pyx_n_u_Y __pyx_mstate_global->__pyx_n_u_Y
#define __pyx_kp_u__13 __pyx_mstate_global->__pyx_kp_u__13
#define __pyx_n_s__14 __pyx_mstate_global->__pyx_n_s__14
#define __pyx_kp_u__15 __pyx_mstate_global->__pyx_kp_u__15
#define __pyx_kp_u__16 __pyx_mstate_global->__pyx_kp_u__16
#define __pyx_kp_u__27 __pyx_mstate_global->__pyx_kp_u__27
#define __pyx_kp_b__4 __pyx_mstate_global->__pyx_kp_b__4
#define __pyx_n_s__46 __pyx_mstate_global->__pyx_n_s__46
#define __pyx_kp_b__5 __pyx_mstate_global->__pyx_kp_b__5
#define __pyx_kp_b__6 __pyx_mstate_global->__pyx_kp_b__6
#define __pyx_kp_b__7 __pyx_mstate_global->__pyx_kp_b__7
#define __pyx_kp_b__8 __pyx_mstate_global->__pyx_kp_b__8
#define __pyx_kp_b__9 __pyx_mstate_global->__pyx_kp_b__9
#define __pyx_n_s_all __pyx_mstate_global->__pyx_n_s_all
#define __pyx_n_u_as __pyx_mstate_global->__pyx_n_u_as
#define __pyx_n_s_asyncio_coroutines __pyx_mstate_global->__pyx_n_s_asyncio_coroutines
#define __pyx_kp_u_b __pyx_mstate_global->__pyx_kp_u_b
#define __pyx_n_s_b_2 __pyx_mstate_global->__pyx_n_s_b_2
#define __pyx_n_s_class_getitem __pyx_mstate_global->__pyx_n_s_class_getitem
#define __pyx_n_s_cline_in_traceback __pyx_mstate_global->__pyx_n_s_cline_in_traceback
#define __pyx_n_s_columns __pyx_mstate_global->__pyx_n_s_columns
#define __pyx_n_s_datetime __pyx_mstate_global->__pyx_n_s_datetime
#define __pyx_n_s_dchar __pyx_mstate_global->__pyx_n_s_dchar
#define __pyx_n_s_dt __pyx_mstate_global->__pyx_n_s_dt
#define __pyx_n_s_dtype __pyx_mstate_global->__pyx_n_s_dtype
#define __pyx_n_s_dumps __pyx_mstate_global->__pyx_n_s_dumps
#define __pyx_n_s_dumps_2 __pyx_mstate_global->__pyx_n_s_dumps_2
#define __pyx_n_s_end __pyx_mstate_global->__pyx_n_s_end
#define __pyx_n_s_errors __pyx_mstate_global->__pyx_n_s_errors
#define __pyx_kp_u_expects_np_ndarray_dtype_of_date __pyx_mstate_global->__pyx_kp_u_expects_np_ndarray_dtype_of_date
#define __pyx_kp_u_expects_pandas_DatetimeIndex_got __pyx_mstate_global->__pyx_kp_u_expects_pandas_DatetimeIndex_got
#define __pyx_kp_u_expects_pandas_Series_got_s __pyx_mstate_global->__pyx_kp_u_expects_pandas_Series_got_s
#define __pyx_kp_u_expects_pandas_TimedeltaIndex_go __pyx_mstate_global->__pyx_kp_u_expects_pandas_TimedeltaIndex_go
#define __pyx_kp_u_expects_pandas_Timedelta_got_s __pyx_mstate_global->__pyx_kp_u_expects_pandas_Timedelta_got_s
#define __pyx_kp_u_expects_pandas_Timestamp_got_s __pyx_mstate_global->__pyx_kp_u_expects_pandas_Timestamp_got_s
#define __pyx_n_u_fs __pyx_mstate_global->__pyx_n_u_fs
#define __pyx_n_u_h __pyx_mstate_global->__pyx_n_u_h
#define __pyx_kp_u_h_2 __pyx_mstate_global->__pyx_kp_u_h_2
#define __pyx_kp_u_i __pyx_mstate_global->__pyx_kp_u_i
#define __pyx_n_s_identifier __pyx_mstate_global->__pyx_n_s_identifier
#define __pyx_n_s_imag __pyx_mstate_global->__pyx_n_s_imag
#define __pyx_n_s_import __pyx_mstate_global->__pyx_n_s_import
#define __pyx_n_s_index __pyx_mstate_global->__pyx_n_s_index
#define __pyx_n_s_initializing __pyx_mstate_global->__pyx_n_s_initializing
#define __pyx_kp_u_invalid_pandas_DataFrame_cannot __pyx_mstate_global->__pyx_kp_u_invalid_pandas_DataFrame_cannot
#define __pyx_kp_u_invalid_pandas_DataFrame_columns __pyx_mstate_global->__pyx_kp_u_invalid_pandas_DataFrame_columns
#define __pyx_kp_u_invalid_pandas_DatetimeIndex_nam __pyx_mstate_global->__pyx_kp_u_invalid_pandas_DatetimeIndex_nam
#define __pyx_kp_u_invalid_pandas_Series_name_r __pyx_mstate_global->__pyx_kp_u_invalid_pandas_Series_name_r
#define __pyx_kp_u_invalid_pandas_TimedeltaIndex_na __pyx_mstate_global->__pyx_kp_u_invalid_pandas_TimedeltaIndex_na
#define __pyx_n_s_is_coroutine __pyx_mstate_global->__pyx_n_s_is_coroutine
#define __pyx_n_s_items __pyx_mstate_global->__pyx_n_s_items
#define __pyx_n_s_join __pyx_mstate_global->__pyx_n_s_join
#define __pyx_n_s_ljust __pyx_mstate_global->__pyx_n_s_ljust
#define __pyx_n_u_m __pyx_mstate_global->__pyx_n_u_m
#define __pyx_n_s_main __pyx_mstate_global->__pyx_n_s_main
#define __pyx_n_u_ms __pyx_mstate_global->__pyx_n_u_ms
#define __pyx_n_s_n __pyx_mstate_global->__pyx_n_s_n
#define __pyx_n_s_name __pyx_mstate_global->__pyx_n_s_name
#define __pyx_n_s_name_2 __pyx_mstate_global->__pyx_n_s_name_2
#define __pyx_n_s_np __pyx_mstate_global->__pyx_n_s_np
#define __pyx_n_u_ns __pyx_mstate_global->__pyx_n_u_ns
#define __pyx_kp_u_num __pyx_mstate_global->__pyx_kp_u_num
#define __pyx_n_s_num_2 __pyx_mstate_global->__pyx_n_s_num_2
#define __pyx_n_s_numpy __pyx_mstate_global->__pyx_n_s_numpy
#define __pyx_kp_u_numpy__core_multiarray_failed_to __pyx_mstate_global->__pyx_kp_u_numpy__core_multiarray_failed_to
#define __pyx_kp_u_numpy__core_umath_failed_to_impo __pyx_mstate_global->__pyx_kp_u_numpy__core_umath_failed_to_impo
#define __pyx_n_s_obj __pyx_mstate_global->__pyx_n_s_obj
#define __pyx_n_s_object __pyx_mstate_global->__pyx_n_s_object
#define __pyx_n_s_option __pyx_mstate_global->__pyx_n_s_option
#define __pyx_n_s_orjson __pyx_mstate_global->__pyx_n_s_orjson
#define __pyx_n_s_pack __pyx_mstate_global->__pyx_n_s_pack
#define __pyx_n_s_pandas __pyx_mstate_global->__pyx_n_s_pandas
#define __pyx_kp_u_pass_s __pyx_mstate_global->__pyx_kp_u_pass_s
#define __pyx_n_s_print __pyx_mstate_global->__pyx_n_s_print
#define __pyx_n_u_ps __pyx_mstate_global->__pyx_n_u_ps
#define __pyx_n_s_pyx_capi __pyx_mstate_global->__pyx_n_s_pyx_capi
#define __pyx_kp_u_q __pyx_mstate_global->__pyx_kp_u_q
#define __pyx_n_s_range __pyx_mstate_global->__pyx_n_s_range
#define __pyx_n_s_real __pyx_mstate_global->__pyx_n_s_real
#define __pyx_n_s_return __pyx_mstate_global->__pyx_n_s_return
#define __pyx_n_s_s __pyx_mstate_global->__pyx_n_s_s
#define __pyx_n_u_s __pyx_mstate_global->__pyx_n_u_s
#define __pyx_n_s_serialize __pyx_mstate_global->__pyx_n_s_serialize
#define __pyx_n_u_serialize __pyx_mstate_global->__pyx_n_u_serialize
#define __pyx_n_s_serializor __pyx_mstate_global->__pyx_n_s_serializor
#define __pyx_n_s_serializor_binary_ser __pyx_mstate_global->__pyx_n_s_serializor_binary_ser
#define __pyx_n_s_spec __pyx_mstate_global->__pyx_n_s_spec
#define __pyx_kp_s_src_serializor_binary_ser_py __pyx_mstate_global->__pyx_kp_s_src_serializor_binary_ser_py
#define __pyx_n_s_start __pyx_mstate_global->__pyx_n_s_start
#define __pyx_n_s_step __pyx_mstate_global->__pyx_n_s_step
#define __pyx_n_s_stop __pyx_mstate_global->__pyx_n_s_stop
#define __pyx_n_s_str __pyx_mstate_global->__pyx_n_s_str
#define __pyx_n_s_struct __pyx_mstate_global->__pyx_n_s_struct
#define __pyx_n_s_struct_time __pyx_mstate_global->__pyx_n_s_struct_time
#define __pyx_n_s_test __pyx_mstate_global->__pyx_n_s_test
#define __pyx_n_s_test_gen_encoded_int __pyx_mstate_global->__pyx_n_s_test_gen_encoded_int
#define __pyx_n_s_test_gen_header __pyx_mstate_global->__pyx_n_s_test_gen_header
#define __pyx_n_s_test_pack_int16 __pyx_mstate_global->__pyx_n_s_test_pack_int16
#define __pyx_n_s_test_pack_int24 __pyx_mstate_global->__pyx_n_s_test_pack_int24
#define __pyx_n_s_test_pack_int32 __pyx_mstate_global->__pyx_n_s_test_pack_int32
#define __pyx_n_s_test_pack_int64 __pyx_mstate_global->__pyx_n_s_test_pack_int64
#define __pyx_n_s_test_pack_int8 __pyx_mstate_global->__pyx_n_s_test_pack_int8
#define __pyx_n_s_test_pack_uint16 __pyx_mstate_global->__pyx_n_s_test_pack_uint16
#define __pyx_n_s_test_pack_uint24 __pyx_mstate_global->__pyx_n_s_test_pack_uint24
#define __pyx_n_s_test_pack_uint32 __pyx_mstate_global->__pyx_n_s_test_pack_uint32
#define __pyx_n_s_test_pack_uint64 __pyx_mstate_global->__pyx_n_s_test_pack_uint64
#define __pyx_n_s_test_pack_uint8 __pyx_mstate_global->__pyx_n_s_test_pack_uint8
#define __pyx_n_s_test_utils __pyx_mstate_global->__pyx_n_s_test_utils
#define __pyx_n_s_time __pyx_mstate_global->__pyx_n_s_time
#define __pyx_n_s_tm_gmtoff __pyx_mstate_global->__pyx_n_s_tm_gmtoff
#define __pyx_n_s_tm_zone __pyx_mstate_global->__pyx_n_s_tm_zone
#define __pyx_n_s_tobytes __pyx_mstate_global->__pyx_n_s_tobytes
#define __pyx_n_s_typeref __pyx_mstate_global->__pyx_n_s_typeref
#define __pyx_n_s_typing __pyx_mstate_global->__pyx_n_s_typing
#define __pyx_n_s_tz __pyx_mstate_global->__pyx_n_s_tz
#define __pyx_kp_u_unable_to_parse_ndarray_time_uni __pyx_mstate_global->__pyx_kp_u_unable_to_parse_ndarray_time_uni
#define __pyx_kp_u_unsupported_data_type_s __pyx_mstate_global->__pyx_kp_u_unsupported_data_type_s
#define __pyx_kp_u_unsupported_numpy_ndarray_dimens __pyx_mstate_global->__pyx_kp_u_unsupported_numpy_ndarray_dimens
#define __pyx_kp_u_unsupported_numpy_ndarray_dtype __pyx_mstate_global->__pyx_kp_u_unsupported_numpy_ndarray_dtype
#define __pyx_kp_u_unsupported_numpy_time_unit_d __pyx_mstate_global->__pyx_kp_u_unsupported_numpy_time_unit_d
#define __pyx_kp_u_unsupported_numpy_time_unit_s __pyx_mstate_global->__pyx_kp_u_unsupported_numpy_time_unit_s
#define __pyx_kp_u_unsupported_pandas_DataFrame_col __pyx_mstate_global->__pyx_kp_u_unsupported_pandas_DataFrame_col
#define __pyx_kp_u_unsupported_pandas_Series_dtype __pyx_mstate_global->__pyx_kp_u_unsupported_pandas_Series_dtype
#define __pyx_n_u_us __pyx_mstate_global->__pyx_n_u_us
#define __pyx_n_s_utcoffset __pyx_mstate_global->__pyx_n_s_utcoffset
#define __pyx_n_s_utils __pyx_mstate_global->__pyx_n_s_utils
#define __pyx_n_s_v __pyx_mstate_global->__pyx_n_s_v
#define __pyx_n_s_val __pyx_mstate_global->__pyx_n_s_val
#define __pyx_n_s_value __pyx_mstate_global->__pyx_n_s_value
#define __pyx_n_s_values __pyx_mstate_global->__pyx_n_s_values
#define __pyx_int_0 __pyx_mstate_global->__pyx_int_0
#define __pyx_int_1 __pyx_mstate_global->__pyx_int_1
#define __pyx_int_3 __pyx_mstate_global->__pyx_int_3
#define __pyx_int_80 __pyx_mstate_global->__pyx_int_80
#define __pyx_int_251 __pyx_mstate_global->__pyx_int_251
#define __pyx_int_252 __pyx_mstate_global->__pyx_int_252
#define __pyx_int_65535 __pyx_mstate_global->__pyx_int_65535
#define __pyx_int_65536 __pyx_mstate_global->__pyx_int_65536
#define __pyx_int_8388606 __pyx_mstate_global->__pyx_int_8388606
#define __pyx_int_8388607 __pyx_mstate_global->__pyx_int_8388607
#define __pyx_int_16777213 __pyx_mstate_global->__pyx_int_16777213
#define __pyx_int_16777214 __pyx_mstate_global->__pyx_int_16777214
#define __pyx_int_16777215 __pyx_mstate_global->__pyx_int_16777215
#define __pyx_int_2147483646 __pyx_mstate_global->__pyx_int_2147483646
#define __pyx_int_2147483647 __pyx_mstate_global->__pyx_int_2147483647
#define __pyx_int_4294967293 __pyx_mstate_global->__pyx_int_4294967293
#define __pyx_int_4294967294 __pyx_mstate_global->__pyx_int_4294967294
#define __pyx_int_4294967295 __pyx_mstate_global->__pyx_int_4294967295
#define __pyx_int_4294967296 __pyx_mstate_global->__pyx_int_4294967296
#define __pyx_int_8589934592 __pyx_mstate_global->__pyx_int_8589934592
#define __pyx_int_0x7ffffffffffffffe __pyx_mstate_global->__pyx_int_0x7ffffffffffffffe
#define __pyx_int_0x7fffffffffffffff __pyx_mstate_global->__pyx_int_0x7fffffffffffffff
#define __pyx_int_0xfffffffffffffffd __pyx_mstate_global->__pyx_int_0xfffffffffffffffd
#define __pyx_int_0xfffffffffffffffe __pyx_mstate_global->__pyx_int_0xfffffffffffffffe
#define __pyx_int_0xffffffffffffffff __pyx_mstate_global->__pyx_int_0xffffffffffffffff
#define __pyx_int_neg_8388607 __pyx_mstate_global->__pyx_int_neg_8388607
#define __pyx_int_neg_8388608 __pyx_mstate_global->__pyx_int_neg_8388608
#define __pyx_int_neg_2147483647 __pyx_mstate_global->__pyx_int_neg_2147483647
#define __pyx_int_neg_2147483648 __pyx_mstate_global->__pyx_int_neg_2147483648
#define __pyx_int_neg_9223372036854775807 __pyx_mstate_global->__pyx_int_neg_9223372036854775807
#define __pyx_int_neg_9223372036854775808 __pyx_mstate_global->__pyx_int_neg_9223372036854775808
#define __pyx_tuple_ __pyx_mstate_global->__pyx_tuple_
#define __pyx_tuple__2 __pyx_mstate_global->__pyx_tuple__2
#define __pyx_tuple__3 __pyx_mstate_global->__pyx_tuple__3
#define __pyx_slice__18 __pyx_mstate_global->__pyx_slice__18
#define __pyx_tuple__10 __pyx_mstate_global->__pyx_tuple__10
#define __pyx_tuple__11 __pyx_mstate_global->__pyx_tuple__11
#define __pyx_tuple__12 __pyx_mstate_global->__pyx_tuple__12
#define __pyx_tuple__17 __pyx_mstate_global->__pyx_tuple__17
#define __pyx_tuple__19 __pyx_mstate_global->__pyx_tuple__19
#define __pyx_tuple__20 __pyx_mstate_global->__pyx_tuple__20
#define __pyx_tuple__21 __pyx_mstate_global->__pyx_tuple__21
#define __pyx_tuple__22 __pyx_mstate_global->__pyx_tuple__22
#define __pyx_tuple__23 __pyx_mstate_global->__pyx_tuple__23
#define __pyx_tuple__24 __pyx_mstate_global->__pyx_tuple__24
#define __pyx_tuple__25 __pyx_mstate_global->__pyx_tuple__25
#define __pyx_tuple__26 __pyx_mstate_global->__pyx_tuple__26
#define __pyx_tuple__28 __pyx_mstate_global->__pyx_tuple__28
#define __pyx_tuple__31 __pyx_mstate_global->__pyx_tuple__31
#define __pyx_tuple__42 __pyx_mstate_global->__pyx_tuple__42
#define __pyx_tuple__44 __pyx_mstate_global->__pyx_tuple__44
#define __pyx_codeobj__29 __pyx_mstate_global->__pyx_codeobj__29
#define __pyx_codeobj__30 __pyx_mstate_global->__pyx_codeobj__30
#define __pyx_codeobj__32 __pyx_mstate_global->__pyx_codeobj__32
#define __pyx_codeobj__33 __pyx_mstate_global->__pyx_codeobj__33
#define __pyx_codeobj__34 __pyx_mstate_global->__pyx_codeobj__34
#define __pyx_codeobj__35 __pyx_mstate_global->__pyx_codeobj__35
#define __pyx_codeobj__36 __pyx_mstate_global->__pyx_codeobj__36
#define __pyx_codeobj__37 __pyx_mstate_global->__pyx_codeobj__37
#define __pyx_codeobj__38 __pyx_mstate_global->__pyx_codeobj__38
#define __pyx_codeobj__39 __pyx_mstate_global->__pyx_codeobj__39
#define __pyx_codeobj__40 __pyx_mstate_global->__pyx_codeobj__40
#define __pyx_codeobj__41 __pyx_mstate_global->__pyx_codeobj__41
#define __pyx_codeobj__43 __pyx_mstate_global->__pyx_codeobj__43
#define __pyx_codeobj__45 __pyx_mstate_global->__pyx_codeobj__45
/* #### Code section: module_code ### */

/* "serializor/binary/ser.pxd":64
 * 
 * # Utils: Pack unsigned integers
 * cdef inline bytes pack_uint8(unsigned int value):             # <<<<<<<<<<<<<<
 *     """Pack `UNSIGNED` 8-bit integer in little-endian order to `<'bytes'>`.
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_10serializor_6binary_3ser_pack_uint8(unsigned int __pyx_v_value) {
  char __pyx_v_buffer[1];
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("pack_uint8", 1);

  /* "serializor/binary/ser.pxd":71
 *     """
 *     cdef char buffer[1]
 *     buffer[0] = value & 0xFF             # <<<<<<<<<<<<<<
 *     return PyBytes_FromStringAndSize(buffer, 1)
 * 
 */
  (__pyx_v_buffer[0]) = (__pyx_v_value & 0xFF);

  /* "serializor/binary/ser.pxd":72
 *     cdef char buffer[1]
 *     buffer[0] = value & 0xFF
 *     return PyBytes_FromStringAndSize(buffer, 1)             # <<<<<<<<<<<<<<
 * 
 * cdef inline bytes pack_uint16(unsigned int value):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyBytes_FromStringAndSize(__pyx_v_buffer, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 72, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "serializor/binary/ser.pxd":64
 * 
 * # Utils: Pack unsigned integers
 * cdef inline bytes pack_uint8(unsigned int value):             # <<<<<<<<<<<<<<
 *     """Pack `UNSIGNED` 8-bit integer in little-endian order to `<'bytes'>`.
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("serializor.binary.ser.pack_uint8", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "serializor/binary/ser.pxd":74
 *     return PyBytes_FromStringAndSize(buffer, 1)
 * 
 * cdef inline bytes pack_uint16(unsigned int value):             # <<<<<<<<<<<<<<
 *     """Pack `UNSIGNED` 16-bit integer in little-endian order to `<'bytes'>`.
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_10serializor_6binary_3ser_pack_uint16(unsigned int __pyx_v_value) {
  char __pyx_v_buffer[2];
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("pack_uint16", 1);

  /* "serializor/binary/ser.pxd":81
 *     """
 *     cdef char buffer[2]
 *     buffer[0] = value & 0xFF             # <<<<<<<<<<<<<<
 *     buffer[1] = (value >> 8) & 0xFF
 *     return PyBytes_FromStringAndSize(buffer, 2)
 */
  (__pyx_v_buffer[0]) = (__pyx_v_value & 0xFF);

  /* "serializor/binary/ser.pxd":82
 *     cdef char buffer[2]
 *     buffer[0] = value & 0xFF
 *     buffer[1] = (value >> 8) & 0xFF             # <<<<<<<<<<<<<<
 *     return PyBytes_FromStringAndSize(buffer, 2)
 * 
 */
  (__pyx_v_buffer[1]) = ((__pyx_v_value >> 8) & 0xFF);

  /* "serializor/binary/ser.pxd":83
 *     buffer[0] = value & 0xFF
 *     buffer[1] = (value >> 8) & 0xFF
 *     return PyBytes_FromStringAndSize(buffer, 2)             # <<<<<<<<<<<<<<
 * 
 * cdef inline bytes pack_uint24(unsigned int value):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyBytes_FromStringAndSize(__pyx_v_buffer, 2); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 83, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "serializor/binary/ser.pxd":74
 *     return PyBytes_FromStringAndSize(buffer, 1)
 * 
 * cdef inline bytes pack_uint16(unsigned int value):             # <<<<<<<<<<<<<<
 *     """Pack `UNSIGNED` 16-bit integer in little-endian order to `<'bytes'>`.
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("serializor.binary.ser.pack_uint16", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "serializor/binary/ser.pxd":85
 *     return PyBytes_FromStringAndSize(buffer, 2)
 * 
 * cdef inline bytes pack_uint24(unsigned int value):             # <<<<<<<<<<<<<<
 *     """Pack `UNSIGNED` 24-bit integer in little-endian order to `<'bytes'>`.
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_10serializor_6binary_3ser_pack_uint24(unsigned int __pyx_v_value) {
  char __pyx_v_buffer[3];
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("pack_uint24", 1);

  /* "serializor/binary/ser.pxd":92
 *     """
 *     cdef char buffer[3]
 *     buffer[0] = value & 0xFF             # <<<<<<<<<<<<<<
 *     buffer[1] = (value >> 8) & 0xFF
 *     buffer[2] = (value >> 16) & 0xFF
 */
  (__pyx_v_buffer[0]) = (__pyx_v_value & 0xFF);

  /* "serializor/binary/ser.pxd":93
 *     cdef char buffer[3]
 *     buffer[0] = value & 0xFF
 *     buffer[1] = (value >> 8) & 0xFF             # <<<<<<<<<<<<<<
 *     buffer[2] = (value >> 16) & 0xFF
 *     return PyBytes_FromStringAndSize(buffer, 3)
 */
  (__pyx_v_buffer[1]) = ((__pyx_v_value >> 8) & 0xFF);

  /* "serializor/binary/ser.pxd":94
 *     buffer[0] = value & 0xFF
 *     buffer[1] = (value >> 8) & 0xFF
 *     buffer[2] = (value >> 16) & 0xFF             # <<<<<<<<<<<<<<
 *     return PyBytes_FromStringAndSize(buffer, 3)
 * 
 */
  (__pyx_v_buffer[2]) = ((__pyx_v_value >> 16) & 0xFF);

  /* "serializor/binary/ser.pxd":95
 *     buffer[1] = (value >> 8) & 0xFF
 *     buffer[2] = (value >> 16) & 0xFF
 *     return PyBytes_FromStringAndSize(buffer, 3)             # <<<<<<<<<<<<<<
 * 
 * cdef inline bytes pack_uint32(unsigned long long value):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyBytes_FromStringAndSize(__pyx_v_buffer, 3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 95, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "serializor/binary/ser.pxd":85
 *     return PyBytes_FromStringAndSize(buffer, 2)
 * 
 * cdef inline bytes pack_uint24(unsigned int value):             # <<<<<<<<<<<<<<
 *     """Pack `UNSIGNED` 24-bit integer in little-endian order to `<'bytes'>`.
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("serializor.binary.ser.pack_uint24", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "serializor/binary/ser.pxd":97
 *     return PyBytes_FromStringAndSize(buffer, 3)
 * 
 * cdef inline bytes pack_uint32(unsigned long long value):             # <<<<<<<<<<<<<<
 *     """Pack `UNSIGNED` 32-bit integer in little-endian order to `<'bytes'>`.
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_10serializor_6binary_3ser_pack_uint32(unsigned PY_LONG_LONG __pyx_v_value) {
  char __pyx_v_buffer[4];
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("pack_uint32", 1);

  /* "serializor/binary/ser.pxd":104
 *     """
 *     cdef char buffer[4]
 *     buffer[0] = value & 0xFF             # <<<<<<<<<<<<<<
 *     buffer[1] = (value >> 8) & 0xFF
 *     buffer[2] = (value >> 16) & 0xFF
 */
  (__pyx_v_buffer[0]) = (__pyx_v_value & 0xFF);

  /* "serializor/binary/ser.pxd":105
 *     cdef char buffer[4]
 *     buffer[0] = value & 0xFF
 *     buffer[1] = (value >> 8) & 0xFF             # <<<<<<<<<<<<<<
 *     buffer[2] = (value >> 16) & 0xFF
 *     buffer[3] = (value >> 24) & 0xFF
 */
  (__pyx_v_buffer[1]) = ((__pyx_v_value >> 8) & 0xFF);

  /* "serializor/binary/ser.pxd":106
 *     buffer[0] = value & 0xFF
 *     buffer[1] = (value >> 8) & 0xFF
 *     buffer[2] = (value >> 16) & 0xFF             # <<<<<<<<<<<<<<
 *     buffer[3] = (value >> 24) & 0xFF
 *     return PyBytes_FromStringAndSize(buffer, 4)
 */
  (__pyx_v_buffer[2]) = ((__pyx_v_value >> 16) & 0xFF);

  /* "serializor/binary/ser.pxd":107
 *     buffer[1] = (value >> 8) & 0xFF
 *     buffer[2] = (value >> 16) & 0xFF
 *     buffer[3] = (value >> 24) & 0xFF             # <<<<<<<<<<<<<<
 *     return PyBytes_FromStringAndSize(buffer, 4)
 * 
 */
  (__pyx_v_buffer[3]) = ((__pyx_v_value >> 24) & 0xFF);

  /* "serializor/binary/ser.pxd":108
 *     buffer[2] = (value >> 16) & 0xFF
 *     buffer[3] = (value >> 24) & 0xFF
 *     return PyBytes_FromStringAndSize(buffer, 4)             # <<<<<<<<<<<<<<
 * 
 * cdef inline bytes pack_uint64(unsigned long long value):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyBytes_FromStringAndSize(__pyx_v_buffer, 4); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 108, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "serializor/binary/ser.pxd":97
 *     return PyBytes_FromStringAndSize(buffer, 3)
 * 
 * cdef inline bytes pack_uint32(unsigned long long value):             # <<<<<<<<<<<<<<
 *     """Pack `UNSIGNED` 32-bit integer in little-endian order to `<'bytes'>`.
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("serializor.binary.ser.pack_uint32", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "serializor/binary/ser.pxd":110
 *     return PyBytes_FromStringAndSize(buffer, 4)
 * 
 * cdef inline bytes pack_uint64(unsigned long long value):             # <<<<<<<<<<<<<<
 *     """Pack `UNSIGNED` 64-bit integer in little-endian order to `<'bytes'>`.
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_10serializor_6binary_3ser_pack_uint64(unsigned PY_LONG_LONG __pyx_v_value) {
  char __pyx_v_buffer[8];
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("pack_uint64", 1);

  /* "serializor/binary/ser.pxd":117
 *     """
 *     cdef char buffer[8]
 *     buffer[0] = value & 0xFF             # <<<<<<<<<<<<<<
 *     buffer[1] = (value >> 8) & 0xFF
 *     buffer[2] = (value >> 16) & 0xFF
 */
  (__pyx_v_buffer[0]) = (__pyx_v_value & 0xFF);

  /* "serializor/binary/ser.pxd":118
 *     cdef char buffer[8]
 *     buffer[0] = value & 0xFF
 *     buffer[1] = (value >> 8) & 0xFF             # <<<<<<<<<<<<<<
 *     buffer[2] = (value >> 16) & 0xFF
 *     buffer[3] = (value >> 24) & 0xFF
 */
  (__pyx_v_buffer[1]) = ((__pyx_v_value >> 8) & 0xFF);

  /* "serializor/binary/ser.pxd":119
 *     buffer[0] = value & 0xFF
 *     buffer[1] = (value >> 8) & 0xFF
 *     buffer[2] = (value >> 16) & 0xFF             # <<<<<<<<<<<<<<
 *     buffer[3] = (value >> 24) & 0xFF
 *     buffer[4] = (value >> 32) & 0xFF
 */
  (__pyx_v_buffer[2]) = ((__pyx_v_value >> 16) & 0xFF);

  /* "serializor/binary/ser.pxd":120
 *     buffer[1] = (value >> 8) & 0xFF
 *     buffer[2] = (value >> 16) & 0xFF
 *     buffer[3] = (value >> 24) & 0xFF             # <<<<<<<<<<<<<<
 *     buffer[4] = (value >> 32) & 0xFF
 *     buffer[5] = (value >> 40) & 0xFF
 */
  (__pyx_v_buffer[3]) = ((__pyx_v_value >> 24) & 0xFF);

  /* "serializor/binary/ser.pxd":121
 *     buffer[2] = (value >> 16) & 0xFF
 *     buffer[3] = (value >> 24) & 0xFF
 *     buffer[4] = (value >> 32) & 0xFF             # <<<<<<<<<<<<<<
 *     buffer[5] = (value >> 40) & 0xFF
 *     buffer[6] = (value >> 48) & 0xFF
 */
  (__pyx_v_buffer[4]) = ((__pyx_v_value >> 32) & 0xFF);

  /* "serializor/binary/ser.pxd":122
 *     buffer[3] = (value >> 24) & 0xFF
 *     buffer[4] = (value >> 32) & 0xFF
 *     buffer[5] = (value >> 40) & 0xFF             # <<<<<<<<<<<<<<
 *     buffer[6] = (value >> 48) & 0xFF
 *     buffer[7] = (value >> 56) & 0xFF
 */
  (__pyx_v_buffer[5]) = ((__pyx_v_value >> 40) & 0xFF);

  /* "serializor/binary/ser.pxd":123
 *     buffer[4] = (value >> 32) & 0xFF
 *     buffer[5] = (value >> 40) & 0xFF
 *     buffer[6] = (value >> 48) & 0xFF             # <<<<<<<<<<<<<<
 *     buffer[7] = (value >> 56) & 0xFF
 *     return PyBytes_FromStringAndSize(buffer, 8)
 */
  (__pyx_v_buffer[6]) = ((__pyx_v_value >> 48) & 0xFF);

  /* "serializor/binary/ser.pxd":124
 *     buffer[5] = (value >> 40) & 0xFF
 *     buffer[6] = (value >> 48) & 0xFF
 *     buffer[7] = (value >> 56) & 0xFF             # <<<<<<<<<<<<<<
 *     return PyBytes_FromStringAndSize(buffer, 8)
 * 
 */
  (__pyx_v_buffer[7]) = ((__pyx_v_value >> 56) & 0xFF);

  /* "serializor/binary/ser.pxd":125
 *     buffer[6] = (value >> 48) & 0xFF
 *     buffer[7] = (value >> 56) & 0xFF
 *     return PyBytes_FromStringAndSize(buffer, 8)             # <<<<<<<<<<<<<<
 * 
 * # Utils: Pack signed integer
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyBytes_FromStringAndSize(__pyx_v_buffer, 8); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 125, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "serializor/binary/ser.pxd":110
 *     return PyBytes_FromStringAndSize(buffer, 4)
 * 
 * cdef inline bytes pack_uint64(unsigned long long value):             # <<<<<<<<<<<<<<
 *     """Pack `UNSIGNED` 64-bit integer in little-endian order to `<'bytes'>`.
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("serializor.binary.ser.pack_uint64", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "serializor/binary/ser.pxd":128
 * 
 * # Utils: Pack signed integer
 * cdef inline bytes pack_int8(int value):             # <<<<<<<<<<<<<<
 *     """Pack `SIGNED` 8-bit integer in little-endian order to `<'bytes'>`.
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_10serializor_6binary_3ser_pack_int8(int __pyx_v_value) {
  char __pyx_v_buffer[1];
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("pack_int8", 1);

  /* "serializor/binary/ser.pxd":135
 *     """
 *     cdef char buffer[1]
 *     buffer[0] = value & 0xFF             # <<<<<<<<<<<<<<
 *     return PyBytes_FromStringAndSize(buffer, 1)
 * 
 */
  (__pyx_v_buffer[0]) = (__pyx_v_value & 0xFF);

  /* "serializor/binary/ser.pxd":136
 *     cdef char buffer[1]
 *     buffer[0] = value & 0xFF
 *     return PyBytes_FromStringAndSize(buffer, 1)             # <<<<<<<<<<<<<<
 * 
 * cdef inline bytes pack_int16(int value):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyBytes_FromStringAndSize(__pyx_v_buffer, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 136, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "serializor/binary/ser.pxd":128
 * 
 * # Utils: Pack signed integer
 * cdef inline bytes pack_int8(int value):             # <<<<<<<<<<<<<<
 *     """Pack `SIGNED` 8-bit integer in little-endian order to `<'bytes'>`.
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("serializor.binary.ser.pack_int8", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "serializor/binary/ser.pxd":138
 *     return PyBytes_FromStringAndSize(buffer, 1)
 * 
 * cdef inline bytes pack_int16(int value):             # <<<<<<<<<<<<<<
 *     """Pack `SIGNED` 16-bit integer in little-endian order to `<'bytes'>`.
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_10serializor_6binary_3ser_pack_int16(int __pyx_v_value) {
  char __pyx_v_buffer[2];
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("pack_int16", 1);

  /* "serializor/binary/ser.pxd":145
 *     """
 *     cdef char buffer[2]
 *     buffer[0] = value & 0xFF             # <<<<<<<<<<<<<<
 *     buffer[1] = (value >> 8) & 0xFF
 *     return PyBytes_FromStringAndSize(buffer, 2)
 */
  (__pyx_v_buffer[0]) = (__pyx_v_value & 0xFF);

  /* "serializor/binary/ser.pxd":146
 *     cdef char buffer[2]
 *     buffer[0] = value & 0xFF
 *     buffer[1] = (value >> 8) & 0xFF             # <<<<<<<<<<<<<<
 *     return PyBytes_FromStringAndSize(buffer, 2)
 * 
 */
  (__pyx_v_buffer[1]) = ((__pyx_v_value >> 8) & 0xFF);

  /* "serializor/binary/ser.pxd":147
 *     buffer[0] = value & 0xFF
 *     buffer[1] = (value >> 8) & 0xFF
 *     return PyBytes_FromStringAndSize(buffer, 2)             # <<<<<<<<<<<<<<
 * 
 * cdef inline bytes pack_int24(int value):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyBytes_FromStringAndSize(__pyx_v_buffer, 2); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 147, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "serializor/binary/ser.pxd":138
 *     return PyBytes_FromStringAndSize(buffer, 1)
 * 
 * cdef inline bytes pack_int16(int value):             # <<<<<<<<<<<<<<
 *     """Pack `SIGNED` 16-bit integer in little-endian order to `<'bytes'>`.
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("serializor.binary.ser.pack_int16", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "serializor/binary/ser.pxd":149
 *     return PyBytes_FromStringAndSize(buffer, 2)
 * 
 * cdef inline bytes pack_int24(int value):             # <<<<<<<<<<<<<<
 *     """Pack `SIGNED` 24-bit integer in little-endian order to `<'bytes'>`.
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_10serializor_6binary_3ser_pack_int24(int __pyx_v_value) {
  char __pyx_v_buffer[3];
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("pack_int24", 1);

  /* "serializor/binary/ser.pxd":156
 *     """
 *     cdef char buffer[3]
 *     buffer[0] = value & 0xFF             # <<<<<<<<<<<<<<
 *     buffer[1] = (value >> 8) & 0xFF
 *     buffer[2] = (value >> 16) & 0xFF
 */
  (__pyx_v_buffer[0]) = (__pyx_v_value & 0xFF);

  /* "serializor/binary/ser.pxd":157
 *     cdef char buffer[3]
 *     buffer[0] = value & 0xFF
 *     buffer[1] = (value >> 8) & 0xFF             # <<<<<<<<<<<<<<
 *     buffer[2] = (value >> 16) & 0xFF
 *     return PyBytes_FromStringAndSize(buffer, 3)
 */
  (__pyx_v_buffer[1]) = ((__pyx_v_value >> 8) & 0xFF);

  /* "serializor/binary/ser.pxd":158
 *     buffer[0] = value & 0xFF
 *     buffer[1] = (value >> 8) & 0xFF
 *     buffer[2] = (value >> 16) & 0xFF             # <<<<<<<<<<<<<<
 *     return PyBytes_FromStringAndSize(buffer, 3)
 * 
 */
  (__pyx_v_buffer[2]) = ((__pyx_v_value >> 16) & 0xFF);

  /* "serializor/binary/ser.pxd":159
 *     buffer[1] = (value >> 8) & 0xFF
 *     buffer[2] = (value >> 16) & 0xFF
 *     return PyBytes_FromStringAndSize(buffer, 3)             # <<<<<<<<<<<<<<
 * 
 * cdef inline bytes pack_int32(long long value):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyBytes_FromStringAndSize(__pyx_v_buffer, 3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 159, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "serializor/binary/ser.pxd":149
 *     return PyBytes_FromStringAndSize(buffer, 2)
 * 
 * cdef inline bytes pack_int24(int value):             # <<<<<<<<<<<<<<
 *     """Pack `SIGNED` 24-bit integer in little-endian order to `<'bytes'>`.
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("serializor.binary.ser.pack_int24", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "serializor/binary/ser.pxd":161
 *     return PyBytes_FromStringAndSize(buffer, 3)
 * 
 * cdef inline bytes pack_int32(long long value):             # <<<<<<<<<<<<<<
 *     """Pack `SIGNED` 32-bit integer in little-endian order to `<'bytes'>`.
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_10serializor_6binary_3ser_pack_int32(PY_LONG_LONG __pyx_v_value) {
  char __pyx_v_buffer[4];
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("pack_int32", 1);

  /* "serializor/binary/ser.pxd":168
 *     """
 *     cdef char buffer[4]
 *     buffer[0] = value & 0xFF             # <<<<<<<<<<<<<<
 *     buffer[1] = (value >> 8) & 0xFF
 *     buffer[2] = (value >> 16) & 0xFF
 */
  (__pyx_v_buffer[0]) = (__pyx_v_value & 0xFF);

  /* "serializor/binary/ser.pxd":169
 *     cdef char buffer[4]
 *     buffer[0] = value & 0xFF
 *     buffer[1] = (value >> 8) & 0xFF             # <<<<<<<<<<<<<<
 *     buffer[2] = (value >> 16) & 0xFF
 *     buffer[3] = (value >> 24) & 0xFF
 */
  (__pyx_v_buffer[1]) = ((__pyx_v_value >> 8) & 0xFF);

  /* "serializor/binary/ser.pxd":170
 *     buffer[0] = value & 0xFF
 *     buffer[1] = (value >> 8) & 0xFF
 *     buffer[2] = (value >> 16) & 0xFF             # <<<<<<<<<<<<<<
 *     buffer[3] = (value >> 24) & 0xFF
 *     return PyBytes_FromStringAndSize(buffer, 4)
 */
  (__pyx_v_buffer[2]) = ((__pyx_v_value >> 16) & 0xFF);

  /* "serializor/binary/ser.pxd":171
 *     buffer[1] = (value >> 8) & 0xFF
 *     buffer[2] = (value >> 16) & 0xFF
 *     buffer[3] = (value >> 24) & 0xFF             # <<<<<<<<<<<<<<
 *     return PyBytes_FromStringAndSize(buffer, 4)
 * 
 */
  (__pyx_v_buffer[3]) = ((__pyx_v_value >> 24) & 0xFF);

  /* "serializor/binary/ser.pxd":172
 *     buffer[2] = (value >> 16) & 0xFF
 *     buffer[3] = (value >> 24) & 0xFF
 *     return PyBytes_FromStringAndSize(buffer, 4)             # <<<<<<<<<<<<<<
 * 
 * cdef inline bytes pack_int64(long long value):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyBytes_FromStringAndSize(__pyx_v_buffer, 4); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 172, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "serializor/binary/ser.pxd":161
 *     return PyBytes_FromStringAndSize(buffer, 3)
 * 
 * cdef inline bytes pack_int32(long long value):             # <<<<<<<<<<<<<<
 *     """Pack `SIGNED` 32-bit integer in little-endian order to `<'bytes'>`.
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("serializor.binary.ser.pack_int32", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "serializor/binary/ser.pxd":174
 *     return PyBytes_FromStringAndSize(buffer, 4)
 * 
 * cdef inline bytes pack_int64(long long value):             # <<<<<<<<<<<<<<
 *     """Pack `SIGNED` 64-bit integer in little-endian order to `<'bytes'>`.
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_10serializor_6binary_3ser_pack_int64(PY_LONG_LONG __pyx_v_value) {
  char __pyx_v_buffer[8];
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("pack_int64", 1);

  /* "serializor/binary/ser.pxd":181
 *     """
 *     cdef char buffer[8]
 *     buffer[0] = value & 0xFF             # <<<<<<<<<<<<<<
 *     buffer[1] = (value >> 8) & 0xFF
 *     buffer[2] = (value >> 16) & 0xFF
 */
  (__pyx_v_buffer[0]) = (__pyx_v_value & 0xFF);

  /* "serializor/binary/ser.pxd":182
 *     cdef char buffer[8]
 *     buffer[0] = value & 0xFF
 *     buffer[1] = (value >> 8) & 0xFF             # <<<<<<<<<<<<<<
 *     buffer[2] = (value >> 16) & 0xFF
 *     buffer[3] = (value >> 24) & 0xFF
 */
  (__pyx_v_buffer[1]) = ((__pyx_v_value >> 8) & 0xFF);

  /* "serializor/binary/ser.pxd":183
 *     buffer[0] = value & 0xFF
 *     buffer[1] = (value >> 8) & 0xFF
 *     buffer[2] = (value >> 16) & 0xFF             # <<<<<<<<<<<<<<
 *     buffer[3] = (value >> 24) & 0xFF
 *     buffer[4] = (value >> 32) & 0xFF
 */
  (__pyx_v_buffer[2]) = ((__pyx_v_value >> 16) & 0xFF);

  /* "serializor/binary/ser.pxd":184
 *     buffer[1] = (value >> 8) & 0xFF
 *     buffer[2] = (value >> 16) & 0xFF
 *     buffer[3] = (value >> 24) & 0xFF             # <<<<<<<<<<<<<<
 *     buffer[4] = (value >> 32) & 0xFF
 *     buffer[5] = (value >> 40) & 0xFF
 */
  (__pyx_v_buffer[3]) = ((__pyx_v_value >> 24) & 0xFF);

  /* "serializor/binary/ser.pxd":185
 *     buffer[2] = (value >> 16) & 0xFF
 *     buffer[3] = (value >> 24) & 0xFF
 *     buffer[4] = (value >> 32) & 0xFF             # <<<<<<<<<<<<<<
 *     buffer[5] = (value >> 40) & 0xFF
 *     buffer[6] = (value >> 48) & 0xFF
 */
  (__pyx_v_buffer[4]) = ((__pyx_v_value >> 32) & 0xFF);

  /* "serializor/binary/ser.pxd":186
 *     buffer[3] = (value >> 24) & 0xFF
 *     buffer[4] = (value >> 32) & 0xFF
 *     buffer[5] = (value >> 40) & 0xFF             # <<<<<<<<<<<<<<
 *     buffer[6] = (value >> 48) & 0xFF
 *     buffer[7] = (value >> 56) & 0xFF
 */
  (__pyx_v_buffer[5]) = ((__pyx_v_value >> 40) & 0xFF);

  /* "serializor/binary/ser.pxd":187
 *     buffer[4] = (value >> 32) & 0xFF
 *     buffer[5] = (value >> 40) & 0xFF
 *     buffer[6] = (value >> 48) & 0xFF             # <<<<<<<<<<<<<<
 *     buffer[7] = (value >> 56) & 0xFF
 *     return PyBytes_FromStringAndSize(buffer, 8)
 */
  (__pyx_v_buffer[6]) = ((__pyx_v_value >> 48) & 0xFF);

  /* "serializor/binary/ser.pxd":188
 *     buffer[5] = (value >> 40) & 0xFF
 *     buffer[6] = (value >> 48) & 0xFF
 *     buffer[7] = (value >> 56) & 0xFF             # <<<<<<<<<<<<<<
 *     return PyBytes_FromStringAndSize(buffer, 8)
 * 
 */
  (__pyx_v_buffer[7]) = ((__pyx_v_value >> 56) & 0xFF);

  /* "serializor/binary/ser.pxd":189
 *     buffer[6] = (value >> 48) & 0xFF
 *     buffer[7] = (value >> 56) & 0xFF
 *     return PyBytes_FromStringAndSize(buffer, 8)             # <<<<<<<<<<<<<<
 * 
 * # Encoded integer
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyBytes_FromStringAndSize(__pyx_v_buffer, 8); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 189, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "serializor/binary/ser.pxd":174
 *     return PyBytes_FromStringAndSize(buffer, 4)
 * 
 * cdef inline bytes pack_int64(long long value):             # <<<<<<<<<<<<<<
 *     """Pack `SIGNED` 64-bit integer in little-endian order to `<'bytes'>`.
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("serializor.binary.ser.pack_int64", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "serializor/binary/ser.pxd":192
 * 
 * # Encoded integer
 * cdef inline bytes gen_encoded_int(unsigned long long i):             # <<<<<<<<<<<<<<
 *     """Generate encoded integer `<'bytes'>`."""
 *     cdef char buffer[9]
 */

static CYTHON_INLINE PyObject *__pyx_f_10serializor_6binary_3ser_gen_encoded_int(unsigned PY_LONG_LONG __pyx_v_i) {
  char __pyx_v_buffer[9];
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("gen_encoded_int", 1);

  /* "serializor/binary/ser.pxd":195
 *     """Generate encoded integer `<'bytes'>`."""
 *     cdef char buffer[9]
 *     if i <= UINT8_ENCODE_VALUE:             # <<<<<<<<<<<<<<
 *         buffer[0] = i & 0xFF
 *         return PyBytes_FromStringAndSize(buffer, 1)
 */
  __pyx_t_1 = (__pyx_v_i <= __pyx_v_10serializor_6binary_3ser_UINT8_ENCODE_VALUE);
  if (__pyx_t_1) {

    /* "serializor/binary/ser.pxd":196
 *     cdef char buffer[9]
 *     if i <= UINT8_ENCODE_VALUE:
 *         buffer[0] = i & 0xFF             # <<<<<<<<<<<<<<
 *         return PyBytes_FromStringAndSize(buffer, 1)
 *     elif i <= USHRT_MAX:
 */
    (__pyx_v_buffer[0]) = (__pyx_v_i & 0xFF);

    /* "serializor/binary/ser.pxd":197
 *     if i <= UINT8_ENCODE_VALUE:
 *         buffer[0] = i & 0xFF
 *         return PyBytes_FromStringAndSize(buffer, 1)             # <<<<<<<<<<<<<<
 *     elif i <= USHRT_MAX:
 *         buffer[0] = UINT16_ENCODE_VALUE
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = PyBytes_FromStringAndSize(__pyx_v_buffer, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 197, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_r = ((PyObject*)__pyx_t_2);
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "serializor/binary/ser.pxd":195
 *     """Generate encoded integer `<'bytes'>`."""
 *     cdef char buffer[9]
 *     if i <= UINT8_ENCODE_VALUE:             # <<<<<<<<<<<<<<
 *         buffer[0] = i & 0xFF
 *         return PyBytes_FromStringAndSize(buffer, 1)
 */
  }

  /* "serializor/binary/ser.pxd":198
 *         buffer[0] = i & 0xFF
 *         return PyBytes_FromStringAndSize(buffer, 1)
 *     elif i <= USHRT_MAX:             # <<<<<<<<<<<<<<
 *         buffer[0] = UINT16_ENCODE_VALUE
 *         buffer[1] = i & 0xFF
 */
  __pyx_t_1 = (__pyx_v_i <= USHRT_MAX);
  if (__pyx_t_1) {

    /* "serializor/binary/ser.pxd":199
 *         return PyBytes_FromStringAndSize(buffer, 1)
 *     elif i <= USHRT_MAX:
 *         buffer[0] = UINT16_ENCODE_VALUE             # <<<<<<<<<<<<<<
 *         buffer[1] = i & 0xFF
 *         buffer[2] = (i >> 8) & 0xFF
 */
    (__pyx_v_buffer[0]) = __pyx_v_10serializor_6binary_3ser_UINT16_ENCODE_VALUE;

    /* "serializor/binary/ser.pxd":200
 *     elif i <= USHRT_MAX:
 *         buffer[0] = UINT16_ENCODE_VALUE
 *         buffer[1] = i & 0xFF             # <<<<<<<<<<<<<<
 *         buffer[2] = (i >> 8) & 0xFF
 *         return PyBytes_FromStringAndSize(buffer, 3)
 */
    (__pyx_v_buffer[1]) = (__pyx_v_i & 0xFF);

    /* "serializor/binary/ser.pxd":201
 *         buffer[0] = UINT16_ENCODE_VALUE
 *         buffer[1] = i & 0xFF
 *         buffer[2] = (i >> 8) & 0xFF             # <<<<<<<<<<<<<<
 *         return PyBytes_FromStringAndSize(buffer, 3)
 *     elif i <= UINT_MAX:
 */
    (__pyx_v_buffer[2]) = ((__pyx_v_i >> 8) & 0xFF);

    /* "serializor/binary/ser.pxd":202
 *         buffer[1] = i & 0xFF
 *         buffer[2] = (i >> 8) & 0xFF
 *         return PyBytes_FromStringAndSize(buffer, 3)             # <<<<<<<<<<<<<<
 *     elif i <= UINT_MAX:
 *         buffer[0] = UINT32_ENCODE_VALUE
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = PyBytes_FromStringAndSize(__pyx_v_buffer, 3); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 202, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_r = ((PyObject*)__pyx_t_2);
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "serializor/binary/ser.pxd":198
 *         buffer[0] = i & 0xFF
 *         return PyBytes_FromStringAndSize(buffer, 1)
 *     elif i <= USHRT_MAX:             # <<<<<<<<<<<<<<
 *         buffer[0] = UINT16_ENCODE_VALUE
 *         buffer[1] = i & 0xFF
 */
  }

  /* "serializor/binary/ser.pxd":203
 *         buffer[2] = (i >> 8) & 0xFF
 *         return PyBytes_FromStringAndSize(buffer, 3)
 *     elif i <= UINT_MAX:             # <<<<<<<<<<<<<<
 *         buffer[0] = UINT32_ENCODE_VALUE
 *         buffer[1] = i & 0xFF
 */
  __pyx_t_1 = (__pyx_v_i <= UINT_MAX);
  if (__pyx_t_1) {

    /* "serializor/binary/ser.pxd":204
 *         return PyBytes_FromStringAndSize(buffer, 3)
 *     elif i <= UINT_MAX:
 *         buffer[0] = UINT32_ENCODE_VALUE             # <<<<<<<<<<<<<<
 *         buffer[1] = i & 0xFF
 *         buffer[2] = (i >> 8) & 0xFF
 */
    (__pyx_v_buffer[0]) = __pyx_v_10serializor_6binary_3ser_UINT32_ENCODE_VALUE;

    /* "serializor/binary/ser.pxd":205
 *     elif i <= UINT_MAX:
 *         buffer[0] = UINT32_ENCODE_VALUE
 *         buffer[1] = i & 0xFF             # <<<<<<<<<<<<<<
 *         buffer[2] = (i >> 8) & 0xFF
 *         buffer[3] = (i >> 16) & 0xFF
 */
    (__pyx_v_buffer[1]) = (__pyx_v_i & 0xFF);

    /* "serializor/binary/ser.pxd":206
 *         buffer[0] = UINT32_ENCODE_VALUE
 *         buffer[1] = i & 0xFF
 *         buffer[2] = (i >> 8) & 0xFF             # <<<<<<<<<<<<<<
 *         buffer[3] = (i >> 16) & 0xFF
 *         buffer[4] = (i >> 24) & 0xFF
 */
    (__pyx_v_buffer[2]) = ((__pyx_v_i >> 8) & 0xFF);

    /* "serializor/binary/ser.pxd":207
 *         buffer[1] = i & 0xFF
 *         buffer[2] = (i >> 8) & 0xFF
 *         buffer[3] = (i >> 16) & 0xFF             # <<<<<<<<<<<<<<
 *         buffer[4] = (i >> 24) & 0xFF
 *         return PyBytes_FromStringAndSize(buffer, 5)
 */
    (__pyx_v_buffer[3]) = ((__pyx_v_i >> 16) & 0xFF);

    /* "serializor/binary/ser.pxd":208
 *         buffer[2] = (i >> 8) & 0xFF
 *         buffer[3] = (i >> 16) & 0xFF
 *         buffer[4] = (i >> 24) & 0xFF             # <<<<<<<<<<<<<<
 *         return PyBytes_FromStringAndSize(buffer, 5)
 *     else:
 */
    (__pyx_v_buffer[4]) = ((__pyx_v_i >> 24) & 0xFF);

    /* "serializor/binary/ser.pxd":209
 *         buffer[3] = (i >> 16) & 0xFF
 *         buffer[4] = (i >> 24) & 0xFF
 *         return PyBytes_FromStringAndSize(buffer, 5)             # <<<<<<<<<<<<<<
 *     else:
 *         buffer[0] = UINT64_ENCODE_VALUE
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = PyBytes_FromStringAndSize(__pyx_v_buffer, 5); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 209, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_r = ((PyObject*)__pyx_t_2);
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "serializor/binary/ser.pxd":203
 *         buffer[2] = (i >> 8) & 0xFF
 *         return PyBytes_FromStringAndSize(buffer, 3)
 *     elif i <= UINT_MAX:             # <<<<<<<<<<<<<<
 *         buffer[0] = UINT32_ENCODE_VALUE
 *         buffer[1] = i & 0xFF
 */
  }

  /* "serializor/binary/ser.pxd":211
 *         return PyBytes_FromStringAndSize(buffer, 5)
 *     else:
 *         buffer[0] = UINT64_ENCODE_VALUE             # <<<<<<<<<<<<<<
 *         buffer[1] = i & 0xFF
 *         buffer[2] = (i >> 8) & 0xFF
 */
  /*else*/ {
    (__pyx_v_buffer[0]) = __pyx_v_10serializor_6binary_3ser_UINT64_ENCODE_VALUE;

    /* "serializor/binary/ser.pxd":212
 *     else:
 *         buffer[0] = UINT64_ENCODE_VALUE
 *         buffer[1] = i & 0xFF             # <<<<<<<<<<<<<<
 *         buffer[2] = (i >> 8) & 0xFF
 *         buffer[3] = (i >> 16) & 0xFF
 */
    (__pyx_v_buffer[1]) = (__pyx_v_i & 0xFF);

    /* "serializor/binary/ser.pxd":213
 *         buffer[0] = UINT64_ENCODE_VALUE
 *         buffer[1] = i & 0xFF
 *         buffer[2] = (i >> 8) & 0xFF             # <<<<<<<<<<<<<<
 *         buffer[3] = (i >> 16) & 0xFF
 *         buffer[4] = (i >> 24) & 0xFF
 */
    (__pyx_v_buffer[2]) = ((__pyx_v_i >> 8) & 0xFF);

    /* "serializor/binary/ser.pxd":214
 *         buffer[1] = i & 0xFF
 *         buffer[2] = (i >> 8) & 0xFF
 *         buffer[3] = (i >> 16) & 0xFF             # <<<<<<<<<<<<<<
 *         buffer[4] = (i >> 24) & 0xFF
 *         buffer[5] = (i >> 32) & 0xFF
 */
    (__pyx_v_buffer[3]) = ((__pyx_v_i >> 16) & 0xFF);

    /* "serializor/binary/ser.pxd":215
 *         buffer[2] = (i >> 8) & 0xFF
 *         buffer[3] = (i >> 16) & 0xFF
 *         buffer[4] = (i >> 24) & 0xFF             # <<<<<<<<<<<<<<
 *         buffer[5] = (i >> 32) & 0xFF
 *         buffer[6] = (i >> 40) & 0xFF
 */
    (__pyx_v_buffer[4]) = ((__pyx_v_i >> 24) & 0xFF);

    /* "serializor/binary/ser.pxd":216
 *         buffer[3] = (i >> 16) & 0xFF
 *         buffer[4] = (i >> 24) & 0xFF
 *         buffer[5] = (i >> 32) & 0xFF             # <<<<<<<<<<<<<<
 *         buffer[6] = (i >> 40) & 0xFF
 *         buffer[7] = (i >> 48) & 0xFF
 */
    (__pyx_v_buffer[5]) = ((__pyx_v_i >> 32) & 0xFF);

    /* "serializor/binary/ser.pxd":217
 *         buffer[4] = (i >> 24) & 0xFF
 *         buffer[5] = (i >> 32) & 0xFF
 *         buffer[6] = (i >> 40) & 0xFF             # <<<<<<<<<<<<<<
 *         buffer[7] = (i >> 48) & 0xFF
 *         buffer[8] = (i >> 56) & 0xFF
 */
    (__pyx_v_buffer[6]) = ((__pyx_v_i >> 40) & 0xFF);

    /* "serializor/binary/ser.pxd":218
 *         buffer[5] = (i >> 32) & 0xFF
 *         buffer[6] = (i >> 40) & 0xFF
 *         buffer[7] = (i >> 48) & 0xFF             # <<<<<<<<<<<<<<
 *         buffer[8] = (i >> 56) & 0xFF
 *         return PyBytes_FromStringAndSize(buffer, 9)
 */
    (__pyx_v_buffer[7]) = ((__pyx_v_i >> 48) & 0xFF);

    /* "serializor/binary/ser.pxd":219
 *         buffer[6] = (i >> 40) & 0xFF
 *         buffer[7] = (i >> 48) & 0xFF
 *         buffer[8] = (i >> 56) & 0xFF             # <<<<<<<<<<<<<<
 *         return PyBytes_FromStringAndSize(buffer, 9)
 * 
 */
    (__pyx_v_buffer[8]) = ((__pyx_v_i >> 56) & 0xFF);

    /* "serializor/binary/ser.pxd":220
 *         buffer[7] = (i >> 48) & 0xFF
 *         buffer[8] = (i >> 56) & 0xFF
 *         return PyBytes_FromStringAndSize(buffer, 9)             # <<<<<<<<<<<<<<
 * 
 * # Header
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = PyBytes_FromStringAndSize(__pyx_v_buffer, 9); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 220, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_r = ((PyObject*)__pyx_t_2);
    __pyx_t_2 = 0;
    goto __pyx_L0;
  }

  /* "serializor/binary/ser.pxd":192
 * 
 * # Encoded integer
 * cdef inline bytes gen_encoded_int(unsigned long long i):             # <<<<<<<<<<<<<<
 *     """Generate encoded integer `<'bytes'>`."""
 *     cdef char buffer[9]
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("serializor.binary.ser.gen_encoded_int", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "serializor/binary/ser.pxd":223
 * 
 * # Header
 * cdef inline bytes gen_header(unsigned char identifier, unsigned long long i):             # <<<<<<<<<<<<<<
 *     """Generate header '[identifier]+[encoded integer]' `<'bytes'>`."""
 *     cdef char buffer[10]
 */

static CYTHON_INLINE PyObject *__pyx_f_10serializor_6binary_3ser_gen_header(unsigned char __pyx_v_identifier, unsigned PY_LONG_LONG __pyx_v_i) {
  char __pyx_v_buffer[10];
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("gen_header", 1);

  /* "serializor/binary/ser.pxd":226
 *     """Generate header '[identifier]+[encoded integer]' `<'bytes'>`."""
 *     cdef char buffer[10]
 *     if i <= UINT8_ENCODE_VALUE:             # <<<<<<<<<<<<<<
 *         buffer[0] = identifier
 *         buffer[1] = i & 0xFF
 */
  __pyx_t_1 = (__pyx_v_i <= __pyx_v_10serializor_6binary_3ser_UINT8_ENCODE_VALUE);
  if (__pyx_t_1) {

    /* "serializor/binary/ser.pxd":227
 *     cdef char buffer[10]
 *     if i <= UINT8_ENCODE_VALUE:
 *         buffer[0] = identifier             # <<<<<<<<<<<<<<
 *         buffer[1] = i & 0xFF
 *         return PyBytes_FromStringAndSize(buffer, 2)
 */
    (__pyx_v_buffer[0]) = __pyx_v_identifier;

    /* "serializor/binary/ser.pxd":228
 *     if i <= UINT8_ENCODE_VALUE:
 *         buffer[0] = identifier
 *         buffer[1] = i & 0xFF             # <<<<<<<<<<<<<<
 *         return PyBytes_FromStringAndSize(buffer, 2)
 *     elif i <= USHRT_MAX:
 */
    (__pyx_v_buffer[1]) = (__pyx_v_i & 0xFF);

    /* "serializor/binary/ser.pxd":229
 *         buffer[0] = identifier
 *         buffer[1] = i & 0xFF
 *         return PyBytes_FromStringAndSize(buffer, 2)             # <<<<<<<<<<<<<<
 *     elif i <= USHRT_MAX:
 *         buffer[0] = identifier
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = PyBytes_FromStringAndSize(__pyx_v_buffer, 2); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 229, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_r = ((PyObject*)__pyx_t_2);
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "serializor/binary/ser.pxd":226
 *     """Generate header '[identifier]+[encoded integer]' `<'bytes'>`."""
 *     cdef char buffer[10]
 *     if i <= UINT8_ENCODE_VALUE:             # <<<<<<<<<<<<<<
 *         buffer[0] = identifier
 *         buffer[1] = i & 0xFF
 */
  }

  /* "serializor/binary/ser.pxd":230
 *         buffer[1] = i & 0xFF
 *         return PyBytes_FromStringAndSize(buffer, 2)
 *     elif i <= USHRT_MAX:             # <<<<<<<<<<<<<<
 *         buffer[0] = identifier
 *         buffer[1] = UINT16_ENCODE_VALUE
 */
  __pyx_t_1 = (__pyx_v_i <= USHRT_MAX);
  if (__pyx_t_1) {

    /* "serializor/binary/ser.pxd":231
 *         return PyBytes_FromStringAndSize(buffer, 2)
 *     elif i <= USHRT_MAX:
 *         buffer[0] = identifier             # <<<<<<<<<<<<<<
 *         buffer[1] = UINT16_ENCODE_VALUE
 *         buffer[2] = i & 0xFF
 */
    (__pyx_v_buffer[0]) = __pyx_v_identifier;

    /* "serializor/binary/ser.pxd":232
 *     elif i <= USHRT_MAX:
 *         buffer[0] = identifier
 *         buffer[1] = UINT16_ENCODE_VALUE             # <<<<<<<<<<<<<<
 *         buffer[2] = i & 0xFF
 *         buffer[3] = (i >> 8) & 0xFF
 */
    (__pyx_v_buffer[1]) = __pyx_v_10serializor_6binary_3ser_UINT16_ENCODE_VALUE;

    /* "serializor/binary/ser.pxd":233
 *         buffer[0] = identifier
 *         buffer[1] = UINT16_ENCODE_VALUE
 *         buffer[2] = i & 0xFF             # <<<<<<<<<<<<<<
 *         buffer[3] = (i >> 8) & 0xFF
 *         return PyBytes_FromStringAndSize(buffer, 4)
 */
    (__pyx_v_buffer[2]) = (__pyx_v_i & 0xFF);

    /* "serializor/binary/ser.pxd":234
 *         buffer[1] = UINT16_ENCODE_VALUE
 *         buffer[2] = i & 0xFF
 *         buffer[3] = (i >> 8) & 0xFF             # <<<<<<<<<<<<<<
 *         return PyBytes_FromStringAndSize(buffer, 4)
 *     elif i <= UINT_MAX:
 */
    (__pyx_v_buffer[3]) = ((__pyx_v_i >> 8) & 0xFF);

    /* "serializor/binary/ser.pxd":235
 *         buffer[2] = i & 0xFF
 *         buffer[3] = (i >> 8) & 0xFF
 *         return PyBytes_FromStringAndSize(buffer, 4)             # <<<<<<<<<<<<<<
 *     elif i <= UINT_MAX:
 *         buffer[0] = identifier
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = PyBytes_FromStringAndSize(__pyx_v_buffer, 4); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 235, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_r = ((PyObject*)__pyx_t_2);
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "serializor/binary/ser.pxd":230
 *         buffer[1] = i & 0xFF
 *         return PyBytes_FromStringAndSize(buffer, 2)
 *     elif i <= USHRT_MAX:             # <<<<<<<<<<<<<<
 *         buffer[0] = identifier
 *         buffer[1] = UINT16_ENCODE_VALUE
 */
  }

  /* "serializor/binary/ser.pxd":236
 *         buffer[3] = (i >> 8) & 0xFF
 *         return PyBytes_FromStringAndSize(buffer, 4)
 *     elif i <= UINT_MAX:             # <<<<<<<<<<<<<<
 *         buffer[0] = identifier
 *         buffer[1] = UINT32_ENCODE_VALUE
 */
  __pyx_t_1 = (__pyx_v_i <= UINT_MAX);
  if (__pyx_t_1) {

    /* "serializor/binary/ser.pxd":237
 *         return PyBytes_FromStringAndSize(buffer, 4)
 *     elif i <= UINT_MAX:
 *         buffer[0] = identifier             # <<<<<<<<<<<<<<
 *         buffer[1] = UINT32_ENCODE_VALUE
 *         buffer[2] = i & 0xFF
 */
    (__pyx_v_buffer[0]) = __pyx_v_identifier;

    /* "serializor/binary/ser.pxd":238
 *     elif i <= UINT_MAX:
 *         buffer[0] = identifier
 *         buffer[1] = UINT32_ENCODE_VALUE             # <<<<<<<<<<<<<<
 *         buffer[2] = i & 0xFF
 *         buffer[3] = (i >> 8) & 0xFF
 */
    (__pyx_v_buffer[1]) = __pyx_v_10serializor_6binary_3ser_UINT32_ENCODE_VALUE;

    /* "serializor/binary/ser.pxd":239
 *         buffer[0] = identifier
 *         buffer[1] = UINT32_ENCODE_VALUE
 *         buffer[2] = i & 0xFF             # <<<<<<<<<<<<<<
 *         buffer[3] = (i >> 8) & 0xFF
 *         buffer[4] = (i >> 16) & 0xFF
 */
    (__pyx_v_buffer[2]) = (__pyx_v_i & 0xFF);

    /* "serializor/binary/ser.pxd":240
 *         buffer[1] = UINT32_ENCODE_VALUE
 *         buffer[2] = i & 0xFF
 *         buffer[3] = (i >> 8) & 0xFF             # <<<<<<<<<<<<<<
 *         buffer[4] = (i >> 16) & 0xFF
 *         buffer[5] = (i >> 24) & 0xFF
 */
    (__pyx_v_buffer[3]) = ((__pyx_v_i >> 8) & 0xFF);

    /* "serializor/binary/ser.pxd":241
 *         buffer[2] = i & 0xFF
 *         buffer[3] = (i >> 8) & 0xFF
 *         buffer[4] = (i >> 16) & 0xFF             # <<<<<<<<<<<<<<
 *         buffer[5] = (i >> 24) & 0xFF
 *         return PyBytes_FromStringAndSize(buffer, 6)
 */
    (__pyx_v_buffer[4]) = ((__pyx_v_i >> 16) & 0xFF);

    /* "serializor/binary/ser.pxd":242
 *         buffer[3] = (i >> 8) & 0xFF
 *         buffer[4] = (i >> 16) & 0xFF
 *         buffer[5] = (i >> 24) & 0xFF             # <<<<<<<<<<<<<<
 *         return PyBytes_FromStringAndSize(buffer, 6)
 *     else:
 */
    (__pyx_v_buffer[5]) = ((__pyx_v_i >> 24) & 0xFF);

    /* "serializor/binary/ser.pxd":243
 *         buffer[4] = (i >> 16) & 0xFF
 *         buffer[5] = (i >> 24) & 0xFF
 *         return PyBytes_FromStringAndSize(buffer, 6)             # <<<<<<<<<<<<<<
 *     else:
 *         buffer[0] = identifier
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = PyBytes_FromStringAndSize(__pyx_v_buffer, 6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 243, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_r = ((PyObject*)__pyx_t_2);
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "serializor/binary/ser.pxd":236
 *         buffer[3] = (i >> 8) & 0xFF
 *         return PyBytes_FromStringAndSize(buffer, 4)
 *     elif i <= UINT_MAX:             # <<<<<<<<<<<<<<
 *         buffer[0] = identifier
 *         buffer[1] = UINT32_ENCODE_VALUE
 */
  }

  /* "serializor/binary/ser.pxd":245
 *         return PyBytes_FromStringAndSize(buffer, 6)
 *     else:
 *         buffer[0] = identifier             # <<<<<<<<<<<<<<
 *         buffer[1] = UINT64_ENCODE_VALUE
 *         buffer[2] = i & 0xFF
 */
  /*else*/ {
    (__pyx_v_buffer[0]) = __pyx_v_identifier;

    /* "serializor/binary/ser.pxd":246
 *     else:
 *         buffer[0] = identifier
 *         buffer[1] = UINT64_ENCODE_VALUE             # <<<<<<<<<<<<<<
 *         buffer[2] = i & 0xFF
 *         buffer[3] = (i >> 8) & 0xFF
 */
    (__pyx_v_buffer[1]) = __pyx_v_10serializor_6binary_3ser_UINT64_ENCODE_VALUE;

    /* "serializor/binary/ser.pxd":247
 *         buffer[0] = identifier
 *         buffer[1] = UINT64_ENCODE_VALUE
 *         buffer[2] = i & 0xFF             # <<<<<<<<<<<<<<
 *         buffer[3] = (i >> 8) & 0xFF
 *         buffer[4] = (i >> 16) & 0xFF
 */
    (__pyx_v_buffer[2]) = (__pyx_v_i & 0xFF);

    /* "serializor/binary/ser.pxd":248
 *         buffer[1] = UINT64_ENCODE_VALUE
 *         buffer[2] = i & 0xFF
 *         buffer[3] = (i >> 8) & 0xFF             # <<<<<<<<<<<<<<
 *         buffer[4] = (i >> 16) & 0xFF
 *         buffer[5] = (i >> 24) & 0xFF
 */
    (__pyx_v_buffer[3]) = ((__pyx_v_i >> 8) & 0xFF);

    /* "serializor/binary/ser.pxd":249
 *         buffer[2] = i & 0xFF
 *         buffer[3] = (i >> 8) & 0xFF
 *         buffer[4] = (i >> 16) & 0xFF             # <<<<<<<<<<<<<<
 *         buffer[5] = (i >> 24) & 0xFF
 *         buffer[6] = (i >> 32) & 0xFF
 */
    (__pyx_v_buffer[4]) = ((__pyx_v_i >> 16) & 0xFF);

    /* "serializor/binary/ser.pxd":250
 *         buffer[3] = (i >> 8) & 0xFF
 *         buffer[4] = (i >> 16) & 0xFF
 *         buffer[5] = (i >> 24) & 0xFF             # <<<<<<<<<<<<<<
 *         buffer[6] = (i >> 32) & 0xFF
 *         buffer[7] = (i >> 40) & 0xFF
 */
    (__pyx_v_buffer[5]) = ((__pyx_v_i >> 24) & 0xFF);

    /* "serializor/binary/ser.pxd":251
 *         buffer[4] = (i >> 16) & 0xFF
 *         buffer[5] = (i >> 24) & 0xFF
 *         buffer[6] = (i >> 32) & 0xFF             # <<<<<<<<<<<<<<
 *         buffer[7] = (i >> 40) & 0xFF
 *         buffer[8] = (i >> 48) & 0xFF
 */
    (__pyx_v_buffer[6]) = ((__pyx_v_i >> 32) & 0xFF);

    /* "serializor/binary/ser.pxd":252
 *         buffer[5] = (i >> 24) & 0xFF
 *         buffer[6] = (i >> 32) & 0xFF
 *         buffer[7] = (i >> 40) & 0xFF             # <<<<<<<<<<<<<<
 *         buffer[8] = (i >> 48) & 0xFF
 *         buffer[9] = (i >> 56) & 0xFF
 */
    (__pyx_v_buffer[7]) = ((__pyx_v_i >> 40) & 0xFF);

    /* "serializor/binary/ser.pxd":253
 *         buffer[6] = (i >> 32) & 0xFF
 *         buffer[7] = (i >> 40) & 0xFF
 *         buffer[8] = (i >> 48) & 0xFF             # <<<<<<<<<<<<<<
 *         buffer[9] = (i >> 56) & 0xFF
 *         return PyBytes_FromStringAndSize(buffer, 10)
 */
    (__pyx_v_buffer[8]) = ((__pyx_v_i >> 48) & 0xFF);

    /* "serializor/binary/ser.pxd":254
 *         buffer[7] = (i >> 40) & 0xFF
 *         buffer[8] = (i >> 48) & 0xFF
 *         buffer[9] = (i >> 56) & 0xFF             # <<<<<<<<<<<<<<
 *         return PyBytes_FromStringAndSize(buffer, 10)
 * 
 */
    (__pyx_v_buffer[9]) = ((__pyx_v_i >> 56) & 0xFF);

    /* "serializor/binary/ser.pxd":255
 *         buffer[8] = (i >> 48) & 0xFF
 *         buffer[9] = (i >> 56) & 0xFF
 *         return PyBytes_FromStringAndSize(buffer, 10)             # <<<<<<<<<<<<<<
 * 
 * # Serialize
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = PyBytes_FromStringAndSize(__pyx_v_buffer, 10); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 255, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_r = ((PyObject*)__pyx_t_2);
    __pyx_t_2 = 0;
    goto __pyx_L0;
  }

  /* "serializor/binary/ser.pxd":223
 * 
 * # Header
 * cdef inline bytes gen_header(unsigned char identifier, unsigned long long i):             # <<<<<<<<<<<<<<
 *     """Generate header '[identifier]+[encoded integer]' `<'bytes'>`."""
 *     cdef char buffer[10]
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("serializor.binary.ser.gen_header", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../../tmp/build-env-82p3tyx6/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":243
 * 
 *         @property
 *         cdef inline npy_intp itemsize(self) noexcept nogil:             # <<<<<<<<<<<<<<
 *             return PyDataType_ELSIZE(self)
 * 
 */

static CYTHON_INLINE npy_intp __pyx_f_5numpy_5dtype_8itemsize_itemsize(PyArray_Descr *__pyx_v_self) {
  npy_intp __pyx_r;

  /* "../../../../../tmp/build-env-82p3tyx6/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":244
 *         @property
 *         cdef inline npy_intp itemsize(self) noexcept nogil:
 *             return PyDataType_ELSIZE(self)             # <<<<<<<<<<<<<<
 * 
 *         @property
 */
  __pyx_r = PyDataType_ELSIZE(__pyx_v_self);
  goto __pyx_L0;

  /* "../../../../../tmp/build-env-82p3tyx6/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":243
 * 
 *         @property
 *         cdef inline npy_intp itemsize(self) noexcept nogil:             # <<<<<<<<<<<<<<
 *             return PyDataType_ELSIZE(self)
 * 
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "../../../../../tmp/build-env-82p3tyx6/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":247
 * 
 *         @property
 *         cdef inline npy_intp alignment(self) noexcept nogil:             # <<<<<<<<<<<<<<
 *             return PyDataType_ALIGNMENT(self)
 * 
 */

static CYTHON_INLINE npy_intp __pyx_f_5numpy_5dtype_9alignment_alignment(PyArray_Descr *__pyx_v_self) {
  npy_intp __pyx_r;

  /* "../../../../../tmp/build-env-82p3tyx6/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":248
 *         @property
 *         cdef inline npy_intp alignment(self) noexcept nogil:
 *             return PyDataType_ALIGNMENT(self)             # <<<<<<<<<<<<<<
 * 
 *         # Use fields/names with care as they may be NULL.  You must check
 */
  __pyx_r = PyDataType_ALIGNMENT(__pyx_v_self);
  goto __pyx_L0;

  /* "../../../../../tmp/build-env-82p3tyx6/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":247
 * 
 *         @property
 *         cdef inline npy_intp alignment(self) noexcept nogil:             # <<<<<<<<<<<<<<
 *             return PyDataType_ALIGNMENT(self)
 * 
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "../../../../../tmp/build-env-82p3tyx6/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":253
 *         # for this using PyDataType_HASFIELDS.
 *         @property
 *         cdef inline object fields(self):             # <<<<<<<<<<<<<<
 *             return <object>PyDataType_FIELDS(self)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_5dtype_6fields_fields(PyArray_Descr *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1;
  __Pyx_RefNannySetupContext("fields", 1);

  /* "../../../../../tmp/build-env-82p3tyx6/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":254
 *         @property
 *         cdef inline object fields(self):
 *             return <object>PyDataType_FIELDS(self)             # <<<<<<<<<<<<<<
 * 
 *         @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyDataType_FIELDS(__pyx_v_self);
  __Pyx_INCREF(((PyObject *)__pyx_t_1));
  __pyx_r = ((PyObject *)__pyx_t_1);
  goto __pyx_L0;

  /* "../../../../../tmp/build-env-82p3tyx6/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":253
 *         # for this using PyDataType_HASFIELDS.
 *         @property
 *         cdef inline object fields(self):             # <<<<<<<<<<<<<<
 *             return <object>PyDataType_FIELDS(self)
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../../tmp/build-env-82p3tyx6/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":257
 * 
 *         @property
 *         cdef inline tuple names(self):             # <<<<<<<<<<<<<<
 *             return <tuple>PyDataType_NAMES(self)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_5dtype_5names_names(PyArray_Descr *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1;
  __Pyx_RefNannySetupContext("names", 1);

  /* "../../../../../tmp/build-env-82p3tyx6/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":258
 *         @property
 *         cdef inline tuple names(self):
 *             return <tuple>PyDataType_NAMES(self)             # <<<<<<<<<<<<<<
 * 
 *         # Use PyDataType_HASSUBARRAY to test whether this field is
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyDataType_NAMES(__pyx_v_self);
  __Pyx_INCREF(((PyObject*)__pyx_t_1));
  __pyx_r = ((PyObject*)__pyx_t_1);
  goto __pyx_L0;

  /* "../../../../../tmp/build-env-82p3tyx6/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":257
 * 
 *         @property
 *         cdef inline tuple names(self):             # <<<<<<<<<<<<<<
 *             return <tuple>PyDataType_NAMES(self)
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../../tmp/build-env-82p3tyx6/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":264
 *         # this field via the inline helper method PyDataType_SHAPE.
 *         @property
 *         cdef inline PyArray_ArrayDescr* subarray(self) noexcept nogil:             # <<<<<<<<<<<<<<
 *             return PyDataType_SUBARRAY(self)
 * 
 */

static CYTHON_INLINE PyArray_ArrayDescr *__pyx_f_5numpy_5dtype_8subarray_subarray(PyArray_Descr *__pyx_v_self) {
  PyArray_ArrayDescr *__pyx_r;

  /* "../../../../../tmp/build-env-82p3tyx6/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":265
 *         @property
 *         cdef inline PyArray_ArrayDescr* subarray(self) noexcept nogil:
 *             return PyDataType_SUBARRAY(self)             # <<<<<<<<<<<<<<
 * 
 *         @property
 */
  __pyx_r = PyDataType_SUBARRAY(__pyx_v_self);
  goto __pyx_L0;

  /* "../../../../../tmp/build-env-82p3tyx6/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":264
 *         # this field via the inline helper method PyDataType_SHAPE.
 *         @property
 *         cdef inline PyArray_ArrayDescr* subarray(self) noexcept nogil:             # <<<<<<<<<<<<<<
 *             return PyDataType_SUBARRAY(self)
 * 
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "../../../../../tmp/build-env-82p3tyx6/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":268
 * 
 *         @property
 *         cdef inline npy_uint64 flags(self) noexcept nogil:             # <<<<<<<<<<<<<<
 *             """The data types flags."""
 *             return PyDataType_FLAGS(self)
 */

static CYTHON_INLINE npy_uint64 __pyx_f_5numpy_5dtype_5flags_flags(PyArray_Descr *__pyx_v_self) {
  npy_uint64 __pyx_r;

  /* "../../../../../tmp/build-env-82p3tyx6/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":270
 *         cdef inline npy_uint64 flags(self) noexcept nogil:
 *             """The data types flags."""
 *             return PyDataType_FLAGS(self)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = PyDataType_FLAGS(__pyx_v_self);
  goto __pyx_L0;

  /* "../../../../../tmp/build-env-82p3tyx6/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":268
 * 
 *         @property
 *         cdef inline npy_uint64 flags(self) noexcept nogil:             # <<<<<<<<<<<<<<
 *             """The data types flags."""
 *             return PyDataType_FLAGS(self)
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "../../../../../tmp/build-env-82p3tyx6/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":280
 * 
 *         @property
 *         cdef inline int numiter(self) noexcept nogil:             # <<<<<<<<<<<<<<
 *             """The number of arrays that need to be broadcast to the same shape."""
 *             return PyArray_MultiIter_NUMITER(self)
 */

static CYTHON_INLINE int __pyx_f_5numpy_9broadcast_7numiter_numiter(PyArrayMultiIterObject *__pyx_v_self) {
  int __pyx_r;

  /* "../../../../../tmp/build-env-82p3tyx6/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":282
 *         cdef inline int numiter(self) noexcept nogil:
 *             """The number of arrays that need to be broadcast to the same shape."""
 *             return PyArray_MultiIter_NUMITER(self)             # <<<<<<<<<<<<<<
 * 
 *         @property
 */
  __pyx_r = PyArray_MultiIter_NUMITER(__pyx_v_self);
  goto __pyx_L0;

  /* "../../../../../tmp/build-env-82p3tyx6/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":280
 * 
 *         @property
 *         cdef inline int numiter(self) noexcept nogil:             # <<<<<<<<<<<<<<
 *             """The number of arrays that need to be broadcast to the same shape."""
 *             return PyArray_MultiIter_NUMITER(self)
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "../../../../../tmp/build-env-82p3tyx6/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":285
 * 
 *         @property
 *         cdef inline npy_intp size(self) noexcept nogil:             # <<<<<<<<<<<<<<
 *             """The total broadcasted size."""
 *             return PyArray_MultiIter_SIZE(self)
 */

static CYTHON_INLINE npy_intp __pyx_f_5numpy_9broadcast_4size_size(PyArrayMultiIterObject *__pyx_v_self) {
  npy_intp __pyx_r;

  /* "../../../../../tmp/build-env-82p3tyx6/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":287
 *         cdef inline npy_intp size(self) noexcept nogil:
 *             """The total broadcasted size."""
 *             return PyArray_MultiIter_SIZE(self)             # <<<<<<<<<<<<<<
 * 
 *         @property
 */
  __pyx_r = PyArray_MultiIter_SIZE(__pyx_v_self);
  goto __pyx_L0;

  /* "../../../../../tmp/build-env-82p3tyx6/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":285
 * 
 *         @property
 *         cdef inline npy_intp size(self) noexcept nogil:             # <<<<<<<<<<<<<<
 *             """The total broadcasted size."""
 *             return PyArray_MultiIter_SIZE(self)
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "../../../../../tmp/build-env-82p3tyx6/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":290
 * 
 *         @property
 *         cdef inline npy_intp index(self) noexcept nogil:             # <<<<<<<<<<<<<<
 *             """The current (1-d) index into the broadcasted result."""
 *             return PyArray_MultiIter_INDEX(self)
 */

static CYTHON_INLINE npy_intp __pyx_f_5numpy_9broadcast_5index_index(PyArrayMultiIterObject *__pyx_v_self) {
  npy_intp __pyx_r;

  /* "../../../../../tmp/build-env-82p3tyx6/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":292
 *         cdef inline npy_intp index(self) noexcept nogil:
 *             """The current (1-d) index into the broadcasted result."""
 *             return PyArray_MultiIter_INDEX(self)             # <<<<<<<<<<<<<<
 * 
 *         @property
 */
  __pyx_r = PyArray_MultiIter_INDEX(__pyx_v_self);
  goto __pyx_L0;

  /* "../../../../../tmp/build-env-82p3tyx6/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":290
 * 
 *         @property
 *         cdef inline npy_intp index(self) noexcept nogil:             # <<<<<<<<<<<<<<
 *             """The current (1-d) index into the broadcasted result."""
 *             return PyArray_MultiIter_INDEX(self)
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "../../../../../tmp/build-env-82p3tyx6/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":295
 * 
 *         @property
 *         cdef inline int nd(self) noexcept nogil:             # <<<<<<<<<<<<<<
 *             """The number of dimensions in the broadcasted result."""
 *             return PyArray_MultiIter_NDIM(self)
 */

static CYTHON_INLINE int __pyx_f_5numpy_9broadcast_2nd_nd(PyArrayMultiIterObject *__pyx_v_self) {
  int __pyx_r;

  /* "../../../../../tmp/build-env-82p3tyx6/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":297
 *         cdef inline int nd(self) noexcept nogil:
 *             """The number of dimensions in the broadcasted result."""
 *             return PyArray_MultiIter_NDIM(self)             # <<<<<<<<<<<<<<
 * 
 *         @property
 */
  __pyx_r = PyArray_MultiIter_NDIM(__pyx_v_self);
  goto __pyx_L0;

  /* "../../../../../tmp/build-env-82p3tyx6/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":295
 * 
 *         @property
 *         cdef inline int nd(self) noexcept nogil:             # <<<<<<<<<<<<<<
 *             """The number of dimensions in the broadcasted result."""
 *             return PyArray_MultiIter_NDIM(self)
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "../../../../../tmp/build-env-82p3tyx6/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":300
 * 
 *         @property
 *         cdef inline npy_intp* dimensions(self) noexcept nogil:             # <<<<<<<<<<<<<<
 *             """The shape of the broadcasted result."""
 *             return PyArray_MultiIter_DIMS(self)
 */

static CYTHON_INLINE npy_intp *__pyx_f_5numpy_9broadcast_10dimensions_dimensions(PyArrayMultiIterObject *__pyx_v_self) {
  npy_intp *__pyx_r;

  /* "../../../../../tmp/build-env-82p3tyx6/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":302
 *         cdef inline npy_intp* dimensions(self) noexcept nogil:
 *             """The shape of the broadcasted result."""
 *             return PyArray_MultiIter_DIMS(self)             # <<<<<<<<<<<<<<
 * 
 *         @property
 */
  __pyx_r = PyArray_MultiIter_DIMS(__pyx_v_self);
  goto __pyx_L0;

  /* "../../../../../tmp/build-env-82p3tyx6/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":300
 * 
 *         @property
 *         cdef inline npy_intp* dimensions(self) noexcept nogil:             # <<<<<<<<<<<<<<
 *             """The shape of the broadcasted result."""
 *             return PyArray_MultiIter_DIMS(self)
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "../../../../../tmp/build-env-82p3tyx6/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":305
 * 
 *         @property
 *         cdef inline void** iters(self) noexcept nogil:             # <<<<<<<<<<<<<<
 *             """An array of iterator objects that holds the iterators for the arrays to be broadcast together.
 *             On return, the iterators are adjusted for broadcasting."""
 */

static CYTHON_INLINE void **__pyx_f_5numpy_9broadcast_5iters_iters(PyArrayMultiIterObject *__pyx_v_self) {
  void **__pyx_r;

  /* "../../../../../tmp/build-env-82p3tyx6/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":308
 *             """An array of iterator objects that holds the iterators for the arrays to be broadcast together.
 *             On return, the iterators are adjusted for broadcasting."""
 *             return PyArray_MultiIter_ITERS(self)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = PyArray_MultiIter_ITERS(__pyx_v_self);
  goto __pyx_L0;

  /* "../../../../../tmp/build-env-82p3tyx6/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":305
 * 
 *         @property
 *         cdef inline void** iters(self) noexcept nogil:             # <<<<<<<<<<<<<<
 *             """An array of iterator objects that holds the iterators for the arrays to be broadcast together.
 *             On return, the iterators are adjusted for broadcasting."""
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "../../../../../tmp/build-env-82p3tyx6/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":323
 * 
 *         @property
 *         cdef inline PyObject* base(self) noexcept nogil:             # <<<<<<<<<<<<<<
 *             """Returns a borrowed reference to the object owning the data/memory.
 *             """
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_7ndarray_4base_base(PyArrayObject *__pyx_v_self) {
  PyObject *__pyx_r;

  /* "../../../../../tmp/build-env-82p3tyx6/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":326
 *             """Returns a borrowed reference to the object owning the data/memory.
 *             """
 *             return PyArray_BASE(self)             # <<<<<<<<<<<<<<
 * 
 *         @property
 */
  __pyx_r = PyArray_BASE(__pyx_v_self);
  goto __pyx_L0;

  /* "../../../../../tmp/build-env-82p3tyx6/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":323
 * 
 *         @property
 *         cdef inline PyObject* base(self) noexcept nogil:             # <<<<<<<<<<<<<<
 *             """Returns a borrowed reference to the object owning the data/memory.
 *             """
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "../../../../../tmp/build-env-82p3tyx6/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":329
 * 
 *         @property
 *         cdef inline dtype descr(self):             # <<<<<<<<<<<<<<
 *             """Returns an owned reference to the dtype of the array.
 *             """
 */

static CYTHON_INLINE PyArray_Descr *__pyx_f_5numpy_7ndarray_5descr_descr(PyArrayObject *__pyx_v_self) {
  PyArray_Descr *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyArray_Descr *__pyx_t_1;
  __Pyx_RefNannySetupContext("descr", 1);

  /* "../../../../../tmp/build-env-82p3tyx6/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":332
 *             """Returns an owned reference to the dtype of the array.
 *             """
 *             return <dtype>PyArray_DESCR(self)             # <<<<<<<<<<<<<<
 * 
 *         @property
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __pyx_t_1 = PyArray_DESCR(__pyx_v_self);
  __Pyx_INCREF((PyObject *)((PyArray_Descr *)__pyx_t_1));
  __pyx_r = ((PyArray_Descr *)__pyx_t_1);
  goto __pyx_L0;

  /* "../../../../../tmp/build-env-82p3tyx6/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":329
 * 
 *         @property
 *         cdef inline dtype descr(self):             # <<<<<<<<<<<<<<
 *             """Returns an owned reference to the dtype of the array.
 *             """
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../../tmp/build-env-82p3tyx6/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":335
 * 
 *         @property
 *         cdef inline int ndim(self) noexcept nogil:             # <<<<<<<<<<<<<<
 *             """Returns the number of dimensions in the array.
 *             """
 */

static CYTHON_INLINE int __pyx_f_5numpy_7ndarray_4ndim_ndim(PyArrayObject *__pyx_v_self) {
  int __pyx_r;

  /* "../../../../../tmp/build-env-82p3tyx6/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":338
 *             """Returns the number of dimensions in the array.
 *             """
 *             return PyArray_NDIM(self)             # <<<<<<<<<<<<<<
 * 
 *         @property
 */
  __pyx_r = PyArray_NDIM(__pyx_v_self);
  goto __pyx_L0;

  /* "../../../../../tmp/build-env-82p3tyx6/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":335
 * 
 *         @property
 *         cdef inline int ndim(self) noexcept nogil:             # <<<<<<<<<<<<<<
 *             """Returns the number of dimensions in the array.
 *             """
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "../../../../../tmp/build-env-82p3tyx6/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":341
 * 
 *         @property
 *         cdef inline npy_intp *shape(self) noexcept nogil:             # <<<<<<<<<<<<<<
 *             """Returns a pointer to the dimensions/shape of the array.
 *             The number of elements matches the number of dimensions of the array (ndim).
 */

static CYTHON_INLINE npy_intp *__pyx_f_5numpy_7ndarray_5shape_shape(PyArrayObject *__pyx_v_self) {
  npy_intp *__pyx_r;

  /* "../../../../../tmp/build-env-82p3tyx6/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":346
 *             Can return NULL for 0-dimensional arrays.
 *             """
 *             return PyArray_DIMS(self)             # <<<<<<<<<<<<<<
 * 
 *         @property
 */
  __pyx_r = PyArray_DIMS(__pyx_v_self);
  goto __pyx_L0;

  /* "../../../../../tmp/build-env-82p3tyx6/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":341
 * 
 *         @property
 *         cdef inline npy_intp *shape(self) noexcept nogil:             # <<<<<<<<<<<<<<
 *             """Returns a pointer to the dimensions/shape of the array.
 *             The number of elements matches the number of dimensions of the array (ndim).
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "../../../../../tmp/build-env-82p3tyx6/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":349
 * 
 *         @property
 *         cdef inline npy_intp *strides(self) noexcept nogil:             # <<<<<<<<<<<<<<
 *             """Returns a pointer to the strides of the array.
 *             The number of elements matches the number of dimensions of the array (ndim).
 */

static CYTHON_INLINE npy_intp *__pyx_f_5numpy_7ndarray_7strides_strides(PyArrayObject *__pyx_v_self) {
  npy_intp *__pyx_r;

  /* "../../../../../tmp/build-env-82p3tyx6/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":353
 *             The number of elements matches the number of dimensions of the array (ndim).
 *             """
 *             return PyArray_STRIDES(self)             # <<<<<<<<<<<<<<
 * 
 *         @property
 */
  __pyx_r = PyArray_STRIDES(__pyx_v_self);
  goto __pyx_L0;

  /* "../../../../../tmp/build-env-82p3tyx6/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":349
 * 
 *         @property
 *         cdef inline npy_intp *strides(self) noexcept nogil:             # <<<<<<<<<<<<<<
 *             """Returns a pointer to the strides of the array.
 *             The number of elements matches the number of dimensions of the array (ndim).
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "../../../../../tmp/build-env-82p3tyx6/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":356
 * 
 *         @property
 *         cdef inline npy_intp size(self) noexcept nogil:             # <<<<<<<<<<<<<<
 *             """Returns the total size (in number of elements) of the array.
 *             """
 */

static CYTHON_INLINE npy_intp __pyx_f_5numpy_7ndarray_4size_size(PyArrayObject *__pyx_v_self) {
  npy_intp __pyx_r;

  /* "../../../../../tmp/build-env-82p3tyx6/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":359
 *             """Returns the total size (in number of elements) of the array.
 *             """
 *             return PyArray_SIZE(self)             # <<<<<<<<<<<<<<
 * 
 *         @property
 */
  __pyx_r = PyArray_SIZE(__pyx_v_self);
  goto __pyx_L0;

  /* "../../../../../tmp/build-env-82p3tyx6/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":356
 * 
 *         @property
 *         cdef inline npy_intp size(self) noexcept nogil:             # <<<<<<<<<<<<<<
 *             """Returns the total size (in number of elements) of the array.
 *             """
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "../../../../../tmp/build-env-82p3tyx6/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":362
 * 
 *         @property
 *         cdef inline char* data(self) noexcept nogil:             # <<<<<<<<<<<<<<
 *             """The pointer to the data buffer as a char*.
 *             This is provided for legacy reasons to avoid direct struct field access.
 */

static CYTHON_INLINE char *__pyx_f_5numpy_7ndarray_4data_data(PyArrayObject *__pyx_v_self) {
  char *__pyx_r;

  /* "../../../../../tmp/build-env-82p3tyx6/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":368
 *             of `PyArray_DATA()` instead, which returns a 'void*'.
 *             """
 *             return PyArray_BYTES(self)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = PyArray_BYTES(__pyx_v_self);
  goto __pyx_L0;

  /* "../../../../../tmp/build-env-82p3tyx6/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":362
 * 
 *         @property
 *         cdef inline char* data(self) noexcept nogil:             # <<<<<<<<<<<<<<
 *             """The pointer to the data buffer as a char*.
 *             This is provided for legacy reasons to avoid direct struct field access.
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "../../../../../tmp/build-env-82p3tyx6/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":776
 * ctypedef long double complex clongdouble_t
 * 
 * cdef inline object PyArray_MultiIterNew1(a):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(1, <void*>a)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew1(PyObject *__pyx_v_a) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew1", 1);

  /* "../../../../../tmp/build-env-82p3tyx6/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":777
 * 
 * cdef inline object PyArray_MultiIterNew1(a):
 *     return PyArray_MultiIterNew(1, <void*>a)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew2(a, b):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(1, ((void *)__pyx_v_a)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 777, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "../../../../../tmp/build-env-82p3tyx6/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":776
 * ctypedef long double complex clongdouble_t
 * 
 * cdef inline object PyArray_MultiIterNew1(a):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(1, <void*>a)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew1", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../../tmp/build-env-82p3tyx6/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":779
 *     return PyArray_MultiIterNew(1, <void*>a)
 * 
 * cdef inline object PyArray_MultiIterNew2(a, b):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew2(PyObject *__pyx_v_a, PyObject *__pyx_v_b) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew2", 1);

  /* "../../../../../tmp/build-env-82p3tyx6/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":780
 * 
 * cdef inline object PyArray_MultiIterNew2(a, b):
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew3(a, b, c):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(2, ((void *)__pyx_v_a), ((void *)__pyx_v_b)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 780, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "../../../../../tmp/build-env-82p3tyx6/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":779
 *     return PyArray_MultiIterNew(1, <void*>a)
 * 
 * cdef inline object PyArray_MultiIterNew2(a, b):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew2", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../../tmp/build-env-82p3tyx6/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":782
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)
 * 
 * cdef inline object PyArray_MultiIterNew3(a, b, c):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew3(PyObject *__pyx_v_a, PyObject *__pyx_v_b, PyObject *__pyx_v_c) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew3", 1);

  /* "../../../../../tmp/build-env-82p3tyx6/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":783
 * 
 * cdef inline object PyArray_MultiIterNew3(a, b, c):
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew4(a, b, c, d):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(3, ((void *)__pyx_v_a), ((void *)__pyx_v_b), ((void *)__pyx_v_c)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 783, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "../../../../../tmp/build-env-82p3tyx6/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":782
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)
 * 
 * cdef inline object PyArray_MultiIterNew3(a, b, c):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew3", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../../tmp/build-env-82p3tyx6/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":785
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)
 * 
 * cdef inline object PyArray_MultiIterNew4(a, b, c, d):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew4(PyObject *__pyx_v_a, PyObject *__pyx_v_b, PyObject *__pyx_v_c, PyObject *__pyx_v_d) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew4", 1);

  /* "../../../../../tmp/build-env-82p3tyx6/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":786
 * 
 * cdef inline object PyArray_MultiIterNew4(a, b, c, d):
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew5(a, b, c, d, e):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(4, ((void *)__pyx_v_a), ((void *)__pyx_v_b), ((void *)__pyx_v_c), ((void *)__pyx_v_d)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 786, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "../../../../../tmp/build-env-82p3tyx6/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":785
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)
 * 
 * cdef inline object PyArray_MultiIterNew4(a, b, c, d):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew4", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../../tmp/build-env-82p3tyx6/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":788
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)
 * 
 * cdef inline object PyArray_MultiIterNew5(a, b, c, d, e):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew5(PyObject *__pyx_v_a, PyObject *__pyx_v_b, PyObject *__pyx_v_c, PyObject *__pyx_v_d, PyObject *__pyx_v_e) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew5", 1);

  /* "../../../../../tmp/build-env-82p3tyx6/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":789
 * 
 * cdef inline object PyArray_MultiIterNew5(a, b, c, d, e):
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)             # <<<<<<<<<<<<<<
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(5, ((void *)__pyx_v_a), ((void *)__pyx_v_b), ((void *)__pyx_v_c), ((void *)__pyx_v_d), ((void *)__pyx_v_e)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 789, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "../../../../../tmp/build-env-82p3tyx6/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":788
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)
 * 
 * cdef inline object PyArray_MultiIterNew5(a, b, c, d, e):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew5", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../../tmp/build-env-82p3tyx6/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":791
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):             # <<<<<<<<<<<<<<
 *     if PyDataType_HASSUBARRAY(d):
 *         return <tuple>d.subarray.shape
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyDataType_SHAPE(PyArray_Descr *__pyx_v_d) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2;
  __Pyx_RefNannySetupContext("PyDataType_SHAPE", 1);

  /* "../../../../../tmp/build-env-82p3tyx6/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":792
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):
 *     if PyDataType_HASSUBARRAY(d):             # <<<<<<<<<<<<<<
 *         return <tuple>d.subarray.shape
 *     else:
 */
  __pyx_t_1 = PyDataType_HASSUBARRAY(__pyx_v_d);
  if (__pyx_t_1) {

    /* "../../../../../tmp/build-env-82p3tyx6/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":793
 * cdef inline tuple PyDataType_SHAPE(dtype d):
 *     if PyDataType_HASSUBARRAY(d):
 *         return <tuple>d.subarray.shape             # <<<<<<<<<<<<<<
 *     else:
 *         return ()
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = __pyx_f_5numpy_5dtype_8subarray_subarray(__pyx_v_d)->shape;
    __Pyx_INCREF(((PyObject*)__pyx_t_2));
    __pyx_r = ((PyObject*)__pyx_t_2);
    goto __pyx_L0;

    /* "../../../../../tmp/build-env-82p3tyx6/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":792
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):
 *     if PyDataType_HASSUBARRAY(d):             # <<<<<<<<<<<<<<
 *         return <tuple>d.subarray.shape
 *     else:
 */
  }

  /* "../../../../../tmp/build-env-82p3tyx6/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":795
 *         return <tuple>d.subarray.shape
 *     else:
 *         return ()             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_empty_tuple);
    __pyx_r = __pyx_empty_tuple;
    goto __pyx_L0;
  }

  /* "../../../../../tmp/build-env-82p3tyx6/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":791
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):             # <<<<<<<<<<<<<<
 *     if PyDataType_HASSUBARRAY(d):
 *         return <tuple>d.subarray.shape
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../../tmp/build-env-82p3tyx6/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":994
 *     int _import_umath() except -1
 * 
 * cdef inline void set_array_base(ndarray arr, object base) except *:             # <<<<<<<<<<<<<<
 *     Py_INCREF(base) # important to do this before stealing the reference below!
 *     PyArray_SetBaseObject(arr, base)
 */

static CYTHON_INLINE void __pyx_f_5numpy_set_array_base(PyArrayObject *__pyx_v_arr, PyObject *__pyx_v_base) {
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "../../../../../tmp/build-env-82p3tyx6/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":995
 * 
 * cdef inline void set_array_base(ndarray arr, object base) except *:
 *     Py_INCREF(base) # important to do this before stealing the reference below!             # <<<<<<<<<<<<<<
 *     PyArray_SetBaseObject(arr, base)
 * 
 */
  Py_INCREF(__pyx_v_base);

  /* "../../../../../tmp/build-env-82p3tyx6/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":996
 * cdef inline void set_array_base(ndarray arr, object base) except *:
 *     Py_INCREF(base) # important to do this before stealing the reference below!
 *     PyArray_SetBaseObject(arr, base)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object get_array_base(ndarray arr):
 */
  __pyx_t_1 = PyArray_SetBaseObject(__pyx_v_arr, __pyx_v_base); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(2, 996, __pyx_L1_error)

  /* "../../../../../tmp/build-env-82p3tyx6/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":994
 *     int _import_umath() except -1
 * 
 * cdef inline void set_array_base(ndarray arr, object base) except *:             # <<<<<<<<<<<<<<
 *     Py_INCREF(base) # important to do this before stealing the reference below!
 *     PyArray_SetBaseObject(arr, base)
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("numpy.set_array_base", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
}

/* "../../../../../tmp/build-env-82p3tyx6/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":998
 *     PyArray_SetBaseObject(arr, base)
 * 
 * cdef inline object get_array_base(ndarray arr):             # <<<<<<<<<<<<<<
 *     base = PyArray_BASE(arr)
 *     if base is NULL:
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_get_array_base(PyArrayObject *__pyx_v_arr) {
  PyObject *__pyx_v_base;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("get_array_base", 1);

  /* "../../../../../tmp/build-env-82p3tyx6/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":999
 * 
 * cdef inline object get_array_base(ndarray arr):
 *     base = PyArray_BASE(arr)             # <<<<<<<<<<<<<<
 *     if base is NULL:
 *         return None
 */
  __pyx_v_base = PyArray_BASE(__pyx_v_arr);

  /* "../../../../../tmp/build-env-82p3tyx6/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1000
 * cdef inline object get_array_base(ndarray arr):
 *     base = PyArray_BASE(arr)
 *     if base is NULL:             # <<<<<<<<<<<<<<
 *         return None
 *     return <object>base
 */
  __pyx_t_1 = (__pyx_v_base == NULL);
  if (__pyx_t_1) {

    /* "../../../../../tmp/build-env-82p3tyx6/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1001
 *     base = PyArray_BASE(arr)
 *     if base is NULL:
 *         return None             # <<<<<<<<<<<<<<
 *     return <object>base
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "../../../../../tmp/build-env-82p3tyx6/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1000
 * cdef inline object get_array_base(ndarray arr):
 *     base = PyArray_BASE(arr)
 *     if base is NULL:             # <<<<<<<<<<<<<<
 *         return None
 *     return <object>base
 */
  }

  /* "../../../../../tmp/build-env-82p3tyx6/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1002
 *     if base is NULL:
 *         return None
 *     return <object>base             # <<<<<<<<<<<<<<
 * 
 * # Versions of the import_* functions which are more suitable for
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_base));
  __pyx_r = ((PyObject *)__pyx_v_base);
  goto __pyx_L0;

  /* "../../../../../tmp/build-env-82p3tyx6/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":998
 *     PyArray_SetBaseObject(arr, base)
 * 
 * cdef inline object get_array_base(ndarray arr):             # <<<<<<<<<<<<<<
 *     base = PyArray_BASE(arr)
 *     if base is NULL:
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../../tmp/build-env-82p3tyx6/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1006
 * # Versions of the import_* functions which are more suitable for
 * # Cython code.
 * cdef inline int import_array() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         __pyx_import_array()
 */

static CYTHON_INLINE int __pyx_f_5numpy_import_array(void) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("import_array", 1);

  /* "../../../../../tmp/build-env-82p3tyx6/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1007
 * # Cython code.
 * cdef inline int import_array() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         __pyx_import_array()
 *     except Exception:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "../../../../../tmp/build-env-82p3tyx6/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1008
 * cdef inline int import_array() except -1:
 *     try:
 *         __pyx_import_array()             # <<<<<<<<<<<<<<
 *     except Exception:
 *         raise ImportError("numpy._core.multiarray failed to import")
 */
      __pyx_t_4 = _import_array(); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(2, 1008, __pyx_L3_error)

      /* "../../../../../tmp/build-env-82p3tyx6/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1007
 * # Cython code.
 * cdef inline int import_array() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         __pyx_import_array()
 *     except Exception:
 */
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;

    /* "../../../../../tmp/build-env-82p3tyx6/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1009
 *     try:
 *         __pyx_import_array()
 *     except Exception:             # <<<<<<<<<<<<<<
 *         raise ImportError("numpy._core.multiarray failed to import")
 * 
 */
    __pyx_t_4 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
    if (__pyx_t_4) {
      __Pyx_AddTraceback("numpy.import_array", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_6, &__pyx_t_7) < 0) __PYX_ERR(2, 1009, __pyx_L5_except_error)
      __Pyx_XGOTREF(__pyx_t_5);
      __Pyx_XGOTREF(__pyx_t_6);
      __Pyx_XGOTREF(__pyx_t_7);

      /* "../../../../../tmp/build-env-82p3tyx6/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1010
 *         __pyx_import_array()
 *     except Exception:
 *         raise ImportError("numpy._core.multiarray failed to import")             # <<<<<<<<<<<<<<
 * 
 * cdef inline int import_umath() except -1:
 */
      __pyx_t_8 = __Pyx_PyObject_Call(__pyx_builtin_ImportError, __pyx_tuple_, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(2, 1010, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_Raise(__pyx_t_8, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __PYX_ERR(2, 1010, __pyx_L5_except_error)
    }
    goto __pyx_L5_except_error;

    /* "../../../../../tmp/build-env-82p3tyx6/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1007
 * # Cython code.
 * cdef inline int import_array() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         __pyx_import_array()
 *     except Exception:
 */
    __pyx_L5_except_error:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L8_try_end:;
  }

  /* "../../../../../tmp/build-env-82p3tyx6/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1006
 * # Versions of the import_* functions which are more suitable for
 * # Cython code.
 * cdef inline int import_array() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         __pyx_import_array()
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("numpy.import_array", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../../tmp/build-env-82p3tyx6/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1012
 *         raise ImportError("numpy._core.multiarray failed to import")
 * 
 * cdef inline int import_umath() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_umath()
 */

static CYTHON_INLINE int __pyx_f_5numpy_import_umath(void) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("import_umath", 1);

  /* "../../../../../tmp/build-env-82p3tyx6/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1013
 * 
 * cdef inline int import_umath() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "../../../../../tmp/build-env-82p3tyx6/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1014
 * cdef inline int import_umath() except -1:
 *     try:
 *         _import_umath()             # <<<<<<<<<<<<<<
 *     except Exception:
 *         raise ImportError("numpy._core.umath failed to import")
 */
      __pyx_t_4 = _import_umath(); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(2, 1014, __pyx_L3_error)

      /* "../../../../../tmp/build-env-82p3tyx6/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1013
 * 
 * cdef inline int import_umath() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;

    /* "../../../../../tmp/build-env-82p3tyx6/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1015
 *     try:
 *         _import_umath()
 *     except Exception:             # <<<<<<<<<<<<<<
 *         raise ImportError("numpy._core.umath failed to import")
 * 
 */
    __pyx_t_4 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
    if (__pyx_t_4) {
      __Pyx_AddTraceback("numpy.import_umath", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_6, &__pyx_t_7) < 0) __PYX_ERR(2, 1015, __pyx_L5_except_error)
      __Pyx_XGOTREF(__pyx_t_5);
      __Pyx_XGOTREF(__pyx_t_6);
      __Pyx_XGOTREF(__pyx_t_7);

      /* "../../../../../tmp/build-env-82p3tyx6/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1016
 *         _import_umath()
 *     except Exception:
 *         raise ImportError("numpy._core.umath failed to import")             # <<<<<<<<<<<<<<
 * 
 * cdef inline int import_ufunc() except -1:
 */
      __pyx_t_8 = __Pyx_PyObject_Call(__pyx_builtin_ImportError, __pyx_tuple__2, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(2, 1016, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_Raise(__pyx_t_8, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __PYX_ERR(2, 1016, __pyx_L5_except_error)
    }
    goto __pyx_L5_except_error;

    /* "../../../../../tmp/build-env-82p3tyx6/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1013
 * 
 * cdef inline int import_umath() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
    __pyx_L5_except_error:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L8_try_end:;
  }

  /* "../../../../../tmp/build-env-82p3tyx6/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1012
 *         raise ImportError("numpy._core.multiarray failed to import")
 * 
 * cdef inline int import_umath() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_umath()
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("numpy.import_umath", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../../tmp/build-env-82p3tyx6/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1018
 *         raise ImportError("numpy._core.umath failed to import")
 * 
 * cdef inline int import_ufunc() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_umath()
 */

static CYTHON_INLINE int __pyx_f_5numpy_import_ufunc(void) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("import_ufunc", 1);

  /* "../../../../../tmp/build-env-82p3tyx6/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1019
 * 
 * cdef inline int import_ufunc() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "../../../../../tmp/build-env-82p3tyx6/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1020
 * cdef inline int import_ufunc() except -1:
 *     try:
 *         _import_umath()             # <<<<<<<<<<<<<<
 *     except Exception:
 *         raise ImportError("numpy._core.umath failed to import")
 */
      __pyx_t_4 = _import_umath(); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(2, 1020, __pyx_L3_error)

      /* "../../../../../tmp/build-env-82p3tyx6/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1019
 * 
 * cdef inline int import_ufunc() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;

    /* "../../../../../tmp/build-env-82p3tyx6/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1021
 *     try:
 *         _import_umath()
 *     except Exception:             # <<<<<<<<<<<<<<
 *         raise ImportError("numpy._core.umath failed to import")
 * 
 */
    __pyx_t_4 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
    if (__pyx_t_4) {
      __Pyx_AddTraceback("numpy.import_ufunc", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_6, &__pyx_t_7) < 0) __PYX_ERR(2, 1021, __pyx_L5_except_error)
      __Pyx_XGOTREF(__pyx_t_5);
      __Pyx_XGOTREF(__pyx_t_6);
      __Pyx_XGOTREF(__pyx_t_7);

      /* "../../../../../tmp/build-env-82p3tyx6/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1022
 *         _import_umath()
 *     except Exception:
 *         raise ImportError("numpy._core.umath failed to import")             # <<<<<<<<<<<<<<
 * 
 * 
 */
      __pyx_t_8 = __Pyx_PyObject_Call(__pyx_builtin_ImportError, __pyx_tuple__2, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(2, 1022, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_Raise(__pyx_t_8, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __PYX_ERR(2, 1022, __pyx_L5_except_error)
    }
    goto __pyx_L5_except_error;

    /* "../../../../../tmp/build-env-82p3tyx6/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1019
 * 
 * cdef inline int import_ufunc() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
    __pyx_L5_except_error:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L8_try_end:;
  }

  /* "../../../../../tmp/build-env-82p3tyx6/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1018
 *         raise ImportError("numpy._core.umath failed to import")
 * 
 * cdef inline int import_ufunc() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_umath()
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("numpy.import_ufunc", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../../tmp/build-env-82p3tyx6/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1025
 * 
 * 
 * cdef inline bint is_timedelta64_object(object obj) noexcept:             # <<<<<<<<<<<<<<
 *     """
 *     Cython equivalent of `isinstance(obj, np.timedelta64)`
 */

static CYTHON_INLINE int __pyx_f_5numpy_is_timedelta64_object(PyObject *__pyx_v_obj) {
  int __pyx_r;

  /* "../../../../../tmp/build-env-82p3tyx6/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1037
 *     bool
 *     """
 *     return PyObject_TypeCheck(obj, &PyTimedeltaArrType_Type)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = PyObject_TypeCheck(__pyx_v_obj, (&PyTimedeltaArrType_Type));
  goto __pyx_L0;

  /* "../../../../../tmp/build-env-82p3tyx6/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1025
 * 
 * 
 * cdef inline bint is_timedelta64_object(object obj) noexcept:             # <<<<<<<<<<<<<<
 *     """
 *     Cython equivalent of `isinstance(obj, np.timedelta64)`
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "../../../../../tmp/build-env-82p3tyx6/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1040
 * 
 * 
 * cdef inline bint is_datetime64_object(object obj) noexcept:             # <<<<<<<<<<<<<<
 *     """
 *     Cython equivalent of `isinstance(obj, np.datetime64)`
 */

static CYTHON_INLINE int __pyx_f_5numpy_is_datetime64_object(PyObject *__pyx_v_obj) {
  int __pyx_r;

  /* "../../../../../tmp/build-env-82p3tyx6/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1052
 *     bool
 *     """
 *     return PyObject_TypeCheck(obj, &PyDatetimeArrType_Type)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = PyObject_TypeCheck(__pyx_v_obj, (&PyDatetimeArrType_Type));
  goto __pyx_L0;

  /* "../../../../../tmp/build-env-82p3tyx6/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1040
 * 
 * 
 * cdef inline bint is_datetime64_object(object obj) noexcept:             # <<<<<<<<<<<<<<
 *     """
 *     Cython equivalent of `isinstance(obj, np.datetime64)`
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "../../../../../tmp/build-env-82p3tyx6/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1055
 * 
 * 
 * cdef inline npy_datetime get_datetime64_value(object obj) noexcept nogil:             # <<<<<<<<<<<<<<
 *     """
 *     returns the int64 value underlying scalar numpy datetime64 object
 */

static CYTHON_INLINE npy_datetime __pyx_f_5numpy_get_datetime64_value(PyObject *__pyx_v_obj) {
  npy_datetime __pyx_r;

  /* "../../../../../tmp/build-env-82p3tyx6/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1062
 *     also needed.  That can be found using `get_datetime64_unit`.
 *     """
 *     return (<PyDatetimeScalarObject*>obj).obval             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = ((PyDatetimeScalarObject *)__pyx_v_obj)->obval;
  goto __pyx_L0;

  /* "../../../../../tmp/build-env-82p3tyx6/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1055
 * 
 * 
 * cdef inline npy_datetime get_datetime64_value(object obj) noexcept nogil:             # <<<<<<<<<<<<<<
 *     """
 *     returns the int64 value underlying scalar numpy datetime64 object
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "../../../../../tmp/build-env-82p3tyx6/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1065
 * 
 * 
 * cdef inline npy_timedelta get_timedelta64_value(object obj) noexcept nogil:             # <<<<<<<<<<<<<<
 *     """
 *     returns the int64 value underlying scalar numpy timedelta64 object
 */

static CYTHON_INLINE npy_timedelta __pyx_f_5numpy_get_timedelta64_value(PyObject *__pyx_v_obj) {
  npy_timedelta __pyx_r;

  /* "../../../../../tmp/build-env-82p3tyx6/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1069
 *     returns the int64 value underlying scalar numpy timedelta64 object
 *     """
 *     return (<PyTimedeltaScalarObject*>obj).obval             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = ((PyTimedeltaScalarObject *)__pyx_v_obj)->obval;
  goto __pyx_L0;

  /* "../../../../../tmp/build-env-82p3tyx6/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1065
 * 
 * 
 * cdef inline npy_timedelta get_timedelta64_value(object obj) noexcept nogil:             # <<<<<<<<<<<<<<
 *     """
 *     returns the int64 value underlying scalar numpy timedelta64 object
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "../../../../../tmp/build-env-82p3tyx6/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1072
 * 
 * 
 * cdef inline NPY_DATETIMEUNIT get_datetime64_unit(object obj) noexcept nogil:             # <<<<<<<<<<<<<<
 *     """
 *     returns the unit part of the dtype for a numpy datetime64 object.
 */

static CYTHON_INLINE NPY_DATETIMEUNIT __pyx_f_5numpy_get_datetime64_unit(PyObject *__pyx_v_obj) {
  NPY_DATETIMEUNIT __pyx_r;

  /* "../../../../../tmp/build-env-82p3tyx6/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1076
 *     returns the unit part of the dtype for a numpy datetime64 object.
 *     """
 *     return <NPY_DATETIMEUNIT>(<PyDatetimeScalarObject*>obj).obmeta.base             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = ((NPY_DATETIMEUNIT)((PyDatetimeScalarObject *)__pyx_v_obj)->obmeta.base);
  goto __pyx_L0;

  /* "../../../../../tmp/build-env-82p3tyx6/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1072
 * 
 * 
 * cdef inline NPY_DATETIMEUNIT get_datetime64_unit(object obj) noexcept nogil:             # <<<<<<<<<<<<<<
 *     """
 *     returns the unit part of the dtype for a numpy datetime64 object.
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "cpython/complex.pxd":19
 * 
 *         @property
 *         cdef inline double real(self) noexcept:             # <<<<<<<<<<<<<<
 *             return self.cval.real
 * 
 */

static CYTHON_INLINE double __pyx_f_7cpython_7complex_7complex_4real_real(PyComplexObject *__pyx_v_self) {
  double __pyx_r;

  /* "cpython/complex.pxd":20
 *         @property
 *         cdef inline double real(self) noexcept:
 *             return self.cval.real             # <<<<<<<<<<<<<<
 * 
 *         @property
 */
  __pyx_r = __pyx_v_self->cval.real;
  goto __pyx_L0;

  /* "cpython/complex.pxd":19
 * 
 *         @property
 *         cdef inline double real(self) noexcept:             # <<<<<<<<<<<<<<
 *             return self.cval.real
 * 
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "cpython/complex.pxd":23
 * 
 *         @property
 *         cdef inline double imag(self) noexcept:             # <<<<<<<<<<<<<<
 *             return self.cval.imag
 * 
 */

static CYTHON_INLINE double __pyx_f_7cpython_7complex_7complex_4imag_imag(PyComplexObject *__pyx_v_self) {
  double __pyx_r;

  /* "cpython/complex.pxd":24
 *         @property
 *         cdef inline double imag(self) noexcept:
 *             return self.cval.imag             # <<<<<<<<<<<<<<
 * 
 *     # PyTypeObject PyComplex_Type
 */
  __pyx_r = __pyx_v_self->cval.imag;
  goto __pyx_L0;

  /* "cpython/complex.pxd":23
 * 
 *         @property
 *         cdef inline double imag(self) noexcept:             # <<<<<<<<<<<<<<
 *             return self.cval.imag
 * 
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "cpython/contextvars.pxd":112
 * 
 * 
 * cdef inline object get_value(var, default_value=None):             # <<<<<<<<<<<<<<
 *     """Return a new reference to the value of the context variable,
 *     or the default value of the context variable,
 */

static CYTHON_INLINE PyObject *__pyx_f_7cpython_11contextvars_get_value(PyObject *__pyx_v_var, struct __pyx_opt_args_7cpython_11contextvars_get_value *__pyx_optional_args) {
  PyObject *__pyx_v_default_value = ((PyObject *)Py_None);
  PyObject *__pyx_v_value;
  PyObject *__pyx_v_pyvalue = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_value", 1);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_default_value = __pyx_optional_args->default_value;
    }
  }

  /* "cpython/contextvars.pxd":117
 *     or None if no such value or default was found.
 *     """
 *     cdef PyObject *value = NULL             # <<<<<<<<<<<<<<
 *     PyContextVar_Get(var, NULL, &value)
 *     if value is NULL:
 */
  __pyx_v_value = NULL;

  /* "cpython/contextvars.pxd":118
 *     """
 *     cdef PyObject *value = NULL
 *     PyContextVar_Get(var, NULL, &value)             # <<<<<<<<<<<<<<
 *     if value is NULL:
 *         # context variable does not have a default
 */
  __pyx_t_1 = PyContextVar_Get(__pyx_v_var, NULL, (&__pyx_v_value)); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(3, 118, __pyx_L1_error)

  /* "cpython/contextvars.pxd":119
 *     cdef PyObject *value = NULL
 *     PyContextVar_Get(var, NULL, &value)
 *     if value is NULL:             # <<<<<<<<<<<<<<
 *         # context variable does not have a default
 *         pyvalue = default_value
 */
  __pyx_t_2 = (__pyx_v_value == NULL);
  if (__pyx_t_2) {

    /* "cpython/contextvars.pxd":121
 *     if value is NULL:
 *         # context variable does not have a default
 *         pyvalue = default_value             # <<<<<<<<<<<<<<
 *     else:
 *         # value or default value of context variable
 */
    __Pyx_INCREF(__pyx_v_default_value);
    __pyx_v_pyvalue = __pyx_v_default_value;

    /* "cpython/contextvars.pxd":119
 *     cdef PyObject *value = NULL
 *     PyContextVar_Get(var, NULL, &value)
 *     if value is NULL:             # <<<<<<<<<<<<<<
 *         # context variable does not have a default
 *         pyvalue = default_value
 */
    goto __pyx_L3;
  }

  /* "cpython/contextvars.pxd":124
 *     else:
 *         # value or default value of context variable
 *         pyvalue = <object>value             # <<<<<<<<<<<<<<
 *         Py_XDECREF(value)  # PyContextVar_Get() returned an owned reference as 'PyObject*'
 *     return pyvalue
 */
  /*else*/ {
    __pyx_t_3 = ((PyObject *)__pyx_v_value);
    __Pyx_INCREF(__pyx_t_3);
    __pyx_v_pyvalue = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "cpython/contextvars.pxd":125
 *         # value or default value of context variable
 *         pyvalue = <object>value
 *         Py_XDECREF(value)  # PyContextVar_Get() returned an owned reference as 'PyObject*'             # <<<<<<<<<<<<<<
 *     return pyvalue
 * 
 */
    Py_XDECREF(__pyx_v_value);
  }
  __pyx_L3:;

  /* "cpython/contextvars.pxd":126
 *         pyvalue = <object>value
 *         Py_XDECREF(value)  # PyContextVar_Get() returned an owned reference as 'PyObject*'
 *     return pyvalue             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_pyvalue);
  __pyx_r = __pyx_v_pyvalue;
  goto __pyx_L0;

  /* "cpython/contextvars.pxd":112
 * 
 * 
 * cdef inline object get_value(var, default_value=None):             # <<<<<<<<<<<<<<
 *     """Return a new reference to the value of the context variable,
 *     or the default value of the context variable,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("cpython.contextvars.get_value", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_pyvalue);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cpython/contextvars.pxd":129
 * 
 * 
 * cdef inline object get_value_no_default(var, default_value=None):             # <<<<<<<<<<<<<<
 *     """Return a new reference to the value of the context variable,
 *     or the provided default value if no such value was found.
 */

static CYTHON_INLINE PyObject *__pyx_f_7cpython_11contextvars_get_value_no_default(PyObject *__pyx_v_var, struct __pyx_opt_args_7cpython_11contextvars_get_value_no_default *__pyx_optional_args) {
  PyObject *__pyx_v_default_value = ((PyObject *)Py_None);
  PyObject *__pyx_v_value;
  PyObject *__pyx_v_pyvalue = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_value_no_default", 1);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_default_value = __pyx_optional_args->default_value;
    }
  }

  /* "cpython/contextvars.pxd":135
 *     Ignores the default value of the context variable, if any.
 *     """
 *     cdef PyObject *value = NULL             # <<<<<<<<<<<<<<
 *     PyContextVar_Get(var, <PyObject*>default_value, &value)
 *     # value of context variable or 'default_value'
 */
  __pyx_v_value = NULL;

  /* "cpython/contextvars.pxd":136
 *     """
 *     cdef PyObject *value = NULL
 *     PyContextVar_Get(var, <PyObject*>default_value, &value)             # <<<<<<<<<<<<<<
 *     # value of context variable or 'default_value'
 *     pyvalue = <object>value
 */
  __pyx_t_1 = PyContextVar_Get(__pyx_v_var, ((PyObject *)__pyx_v_default_value), (&__pyx_v_value)); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(3, 136, __pyx_L1_error)

  /* "cpython/contextvars.pxd":138
 *     PyContextVar_Get(var, <PyObject*>default_value, &value)
 *     # value of context variable or 'default_value'
 *     pyvalue = <object>value             # <<<<<<<<<<<<<<
 *     Py_XDECREF(value)  # PyContextVar_Get() returned an owned reference as 'PyObject*'
 *     return pyvalue
 */
  __pyx_t_2 = ((PyObject *)__pyx_v_value);
  __Pyx_INCREF(__pyx_t_2);
  __pyx_v_pyvalue = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "cpython/contextvars.pxd":139
 *     # value of context variable or 'default_value'
 *     pyvalue = <object>value
 *     Py_XDECREF(value)  # PyContextVar_Get() returned an owned reference as 'PyObject*'             # <<<<<<<<<<<<<<
 *     return pyvalue
 */
  Py_XDECREF(__pyx_v_value);

  /* "cpython/contextvars.pxd":140
 *     pyvalue = <object>value
 *     Py_XDECREF(value)  # PyContextVar_Get() returned an owned reference as 'PyObject*'
 *     return pyvalue             # <<<<<<<<<<<<<<
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_pyvalue);
  __pyx_r = __pyx_v_pyvalue;
  goto __pyx_L0;

  /* "cpython/contextvars.pxd":129
 * 
 * 
 * cdef inline object get_value_no_default(var, default_value=None):             # <<<<<<<<<<<<<<
 *     """Return a new reference to the value of the context variable,
 *     or the provided default value if no such value was found.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("cpython.contextvars.get_value_no_default", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_pyvalue);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "datetime.pxd":72
 *     ctypedef extern class datetime.date[object PyDateTime_Date]:
 *         @property
 *         cdef inline int year(self) noexcept:             # <<<<<<<<<<<<<<
 *             return PyDateTime_GET_YEAR(self)
 * 
 */

static CYTHON_INLINE int __pyx_f_7cpython_8datetime_4date_4year_year(PyDateTime_Date *__pyx_v_self) {
  int __pyx_r;

  /* "datetime.pxd":73
 *         @property
 *         cdef inline int year(self) noexcept:
 *             return PyDateTime_GET_YEAR(self)             # <<<<<<<<<<<<<<
 * 
 *         @property
 */
  __pyx_r = PyDateTime_GET_YEAR(((PyObject *)__pyx_v_self));
  goto __pyx_L0;

  /* "datetime.pxd":72
 *     ctypedef extern class datetime.date[object PyDateTime_Date]:
 *         @property
 *         cdef inline int year(self) noexcept:             # <<<<<<<<<<<<<<
 *             return PyDateTime_GET_YEAR(self)
 * 
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "datetime.pxd":76
 * 
 *         @property
 *         cdef inline int month(self) noexcept:             # <<<<<<<<<<<<<<
 *             return PyDateTime_GET_MONTH(self)
 * 
 */

static CYTHON_INLINE int __pyx_f_7cpython_8datetime_4date_5month_month(PyDateTime_Date *__pyx_v_self) {
  int __pyx_r;

  /* "datetime.pxd":77
 *         @property
 *         cdef inline int month(self) noexcept:
 *             return PyDateTime_GET_MONTH(self)             # <<<<<<<<<<<<<<
 * 
 *         @property
 */
  __pyx_r = PyDateTime_GET_MONTH(((PyObject *)__pyx_v_self));
  goto __pyx_L0;

  /* "datetime.pxd":76
 * 
 *         @property
 *         cdef inline int month(self) noexcept:             # <<<<<<<<<<<<<<
 *             return PyDateTime_GET_MONTH(self)
 * 
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "datetime.pxd":80
 * 
 *         @property
 *         cdef inline int day(self) noexcept:             # <<<<<<<<<<<<<<
 *             return PyDateTime_GET_DAY(self)
 * 
 */

static CYTHON_INLINE int __pyx_f_7cpython_8datetime_4date_3day_day(PyDateTime_Date *__pyx_v_self) {
  int __pyx_r;

  /* "datetime.pxd":81
 *         @property
 *         cdef inline int day(self) noexcept:
 *             return PyDateTime_GET_DAY(self)             # <<<<<<<<<<<<<<
 * 
 *     ctypedef extern class datetime.time[object PyDateTime_Time]:
 */
  __pyx_r = PyDateTime_GET_DAY(((PyObject *)__pyx_v_self));
  goto __pyx_L0;

  /* "datetime.pxd":80
 * 
 *         @property
 *         cdef inline int day(self) noexcept:             # <<<<<<<<<<<<<<
 *             return PyDateTime_GET_DAY(self)
 * 
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "datetime.pxd":85
 *     ctypedef extern class datetime.time[object PyDateTime_Time]:
 *         @property
 *         cdef inline int hour(self) noexcept:             # <<<<<<<<<<<<<<
 *             return PyDateTime_TIME_GET_HOUR(self)
 * 
 */

static CYTHON_INLINE int __pyx_f_7cpython_8datetime_4time_4hour_hour(PyDateTime_Time *__pyx_v_self) {
  int __pyx_r;

  /* "datetime.pxd":86
 *         @property
 *         cdef inline int hour(self) noexcept:
 *             return PyDateTime_TIME_GET_HOUR(self)             # <<<<<<<<<<<<<<
 * 
 *         @property
 */
  __pyx_r = PyDateTime_TIME_GET_HOUR(((PyObject *)__pyx_v_self));
  goto __pyx_L0;

  /* "datetime.pxd":85
 *     ctypedef extern class datetime.time[object PyDateTime_Time]:
 *         @property
 *         cdef inline int hour(self) noexcept:             # <<<<<<<<<<<<<<
 *             return PyDateTime_TIME_GET_HOUR(self)
 * 
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "datetime.pxd":89
 * 
 *         @property
 *         cdef inline int minute(self) noexcept:             # <<<<<<<<<<<<<<
 *             return PyDateTime_TIME_GET_MINUTE(self)
 * 
 */

static CYTHON_INLINE int __pyx_f_7cpython_8datetime_4time_6minute_minute(PyDateTime_Time *__pyx_v_self) {
  int __pyx_r;

  /* "datetime.pxd":90
 *         @property
 *         cdef inline int minute(self) noexcept:
 *             return PyDateTime_TIME_GET_MINUTE(self)             # <<<<<<<<<<<<<<
 * 
 *         @property
 */
  __pyx_r = PyDateTime_TIME_GET_MINUTE(((PyObject *)__pyx_v_self));
  goto __pyx_L0;

  /* "datetime.pxd":89
 * 
 *         @property
 *         cdef inline int minute(self) noexcept:             # <<<<<<<<<<<<<<
 *             return PyDateTime_TIME_GET_MINUTE(self)
 * 
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "datetime.pxd":93
 * 
 *         @property
 *         cdef inline int second(self) noexcept:             # <<<<<<<<<<<<<<
 *             return PyDateTime_TIME_GET_SECOND(self)
 * 
 */

static CYTHON_INLINE int __pyx_f_7cpython_8datetime_4time_6second_second(PyDateTime_Time *__pyx_v_self) {
  int __pyx_r;

  /* "datetime.pxd":94
 *         @property
 *         cdef inline int second(self) noexcept:
 *             return PyDateTime_TIME_GET_SECOND(self)             # <<<<<<<<<<<<<<
 * 
 *         @property
 */
  __pyx_r = PyDateTime_TIME_GET_SECOND(((PyObject *)__pyx_v_self));
  goto __pyx_L0;

  /* "datetime.pxd":93
 * 
 *         @property
 *         cdef inline int second(self) noexcept:             # <<<<<<<<<<<<<<
 *             return PyDateTime_TIME_GET_SECOND(self)
 * 
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "datetime.pxd":97
 * 
 *         @property
 *         cdef inline int microsecond(self) noexcept:             # <<<<<<<<<<<<<<
 *             return PyDateTime_TIME_GET_MICROSECOND(self)
 * 
 */

static CYTHON_INLINE int __pyx_f_7cpython_8datetime_4time_11microsecond_microsecond(PyDateTime_Time *__pyx_v_self) {
  int __pyx_r;

  /* "datetime.pxd":98
 *         @property
 *         cdef inline int microsecond(self) noexcept:
 *             return PyDateTime_TIME_GET_MICROSECOND(self)             # <<<<<<<<<<<<<<
 * 
 *         @property
 */
  __pyx_r = PyDateTime_TIME_GET_MICROSECOND(((PyObject *)__pyx_v_self));
  goto __pyx_L0;

  /* "datetime.pxd":97
 * 
 *         @property
 *         cdef inline int microsecond(self) noexcept:             # <<<<<<<<<<<<<<
 *             return PyDateTime_TIME_GET_MICROSECOND(self)
 * 
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "datetime.pxd":101
 * 
 *         @property
 *         cdef inline object tzinfo(self):             # <<<<<<<<<<<<<<
 *             return <object>PyDateTime_TIME_GET_TZINFO(self)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_7cpython_8datetime_4time_6tzinfo_tzinfo(PyDateTime_Time *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1;
  __Pyx_RefNannySetupContext("tzinfo", 1);

  /* "datetime.pxd":102
 *         @property
 *         cdef inline object tzinfo(self):
 *             return <object>PyDateTime_TIME_GET_TZINFO(self)             # <<<<<<<<<<<<<<
 * 
 *         @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyDateTime_TIME_GET_TZINFO(((PyObject *)__pyx_v_self));
  __Pyx_INCREF(((PyObject *)__pyx_t_1));
  __pyx_r = ((PyObject *)__pyx_t_1);
  goto __pyx_L0;

  /* "datetime.pxd":101
 * 
 *         @property
 *         cdef inline object tzinfo(self):             # <<<<<<<<<<<<<<
 *             return <object>PyDateTime_TIME_GET_TZINFO(self)
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "datetime.pxd":105
 * 
 *         @property
 *         cdef inline int fold(self) noexcept:             # <<<<<<<<<<<<<<
 *             # For Python < 3.6 this returns 0 no matter what
 *             return PyDateTime_TIME_GET_FOLD(self)
 */

static CYTHON_INLINE int __pyx_f_7cpython_8datetime_4time_4fold_fold(PyDateTime_Time *__pyx_v_self) {
  int __pyx_r;

  /* "datetime.pxd":107
 *         cdef inline int fold(self) noexcept:
 *             # For Python < 3.6 this returns 0 no matter what
 *             return PyDateTime_TIME_GET_FOLD(self)             # <<<<<<<<<<<<<<
 * 
 *     ctypedef extern class datetime.datetime[object PyDateTime_DateTime]:
 */
  __pyx_r = PyDateTime_TIME_GET_FOLD(((PyObject *)__pyx_v_self));
  goto __pyx_L0;

  /* "datetime.pxd":105
 * 
 *         @property
 *         cdef inline int fold(self) noexcept:             # <<<<<<<<<<<<<<
 *             # For Python < 3.6 this returns 0 no matter what
 *             return PyDateTime_TIME_GET_FOLD(self)
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "datetime.pxd":111
 *     ctypedef extern class datetime.datetime[object PyDateTime_DateTime]:
 *         @property
 *         cdef inline int year(self) noexcept:             # <<<<<<<<<<<<<<
 *             return PyDateTime_GET_YEAR(self)
 * 
 */

static CYTHON_INLINE int __pyx_f_7cpython_8datetime_8datetime_4year_year(PyDateTime_DateTime *__pyx_v_self) {
  int __pyx_r;

  /* "datetime.pxd":112
 *         @property
 *         cdef inline int year(self) noexcept:
 *             return PyDateTime_GET_YEAR(self)             # <<<<<<<<<<<<<<
 * 
 *         @property
 */
  __pyx_r = PyDateTime_GET_YEAR(((PyObject *)__pyx_v_self));
  goto __pyx_L0;

  /* "datetime.pxd":111
 *     ctypedef extern class datetime.datetime[object PyDateTime_DateTime]:
 *         @property
 *         cdef inline int year(self) noexcept:             # <<<<<<<<<<<<<<
 *             return PyDateTime_GET_YEAR(self)
 * 
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "datetime.pxd":115
 * 
 *         @property
 *         cdef inline int month(self) noexcept:             # <<<<<<<<<<<<<<
 *             return PyDateTime_GET_MONTH(self)
 * 
 */

static CYTHON_INLINE int __pyx_f_7cpython_8datetime_8datetime_5month_month(PyDateTime_DateTime *__pyx_v_self) {
  int __pyx_r;

  /* "datetime.pxd":116
 *         @property
 *         cdef inline int month(self) noexcept:
 *             return PyDateTime_GET_MONTH(self)             # <<<<<<<<<<<<<<
 * 
 *         @property
 */
  __pyx_r = PyDateTime_GET_MONTH(((PyObject *)__pyx_v_self));
  goto __pyx_L0;

  /* "datetime.pxd":115
 * 
 *         @property
 *         cdef inline int month(self) noexcept:             # <<<<<<<<<<<<<<
 *             return PyDateTime_GET_MONTH(self)
 * 
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "datetime.pxd":119
 * 
 *         @property
 *         cdef inline int day(self) noexcept:             # <<<<<<<<<<<<<<
 *             return PyDateTime_GET_DAY(self)
 * 
 */

static CYTHON_INLINE int __pyx_f_7cpython_8datetime_8datetime_3day_day(PyDateTime_DateTime *__pyx_v_self) {
  int __pyx_r;

  /* "datetime.pxd":120
 *         @property
 *         cdef inline int day(self) noexcept:
 *             return PyDateTime_GET_DAY(self)             # <<<<<<<<<<<<<<
 * 
 *         @property
 */
  __pyx_r = PyDateTime_GET_DAY(((PyObject *)__pyx_v_self));
  goto __pyx_L0;

  /* "datetime.pxd":119
 * 
 *         @property
 *         cdef inline int day(self) noexcept:             # <<<<<<<<<<<<<<
 *             return PyDateTime_GET_DAY(self)
 * 
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "datetime.pxd":123
 * 
 *         @property
 *         cdef inline int hour(self) noexcept:             # <<<<<<<<<<<<<<
 *             return PyDateTime_DATE_GET_HOUR(self)
 * 
 */

static CYTHON_INLINE int __pyx_f_7cpython_8datetime_8datetime_4hour_hour(PyDateTime_DateTime *__pyx_v_self) {
  int __pyx_r;

  /* "datetime.pxd":124
 *         @property
 *         cdef inline int hour(self) noexcept:
 *             return PyDateTime_DATE_GET_HOUR(self)             # <<<<<<<<<<<<<<
 * 
 *         @property
 */
  __pyx_r = PyDateTime_DATE_GET_HOUR(((PyObject *)__pyx_v_self));
  goto __pyx_L0;

  /* "datetime.pxd":123
 * 
 *         @property
 *         cdef inline int hour(self) noexcept:             # <<<<<<<<<<<<<<
 *             return PyDateTime_DATE_GET_HOUR(self)
 * 
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "datetime.pxd":127
 * 
 *         @property
 *         cdef inline int minute(self) noexcept:             # <<<<<<<<<<<<<<
 *             return PyDateTime_DATE_GET_MINUTE(self)
 * 
 */

static CYTHON_INLINE int __pyx_f_7cpython_8datetime_8datetime_6minute_minute(PyDateTime_DateTime *__pyx_v_self) {
  int __pyx_r;

  /* "datetime.pxd":128
 *         @property
 *         cdef inline int minute(self) noexcept:
 *             return PyDateTime_DATE_GET_MINUTE(self)             # <<<<<<<<<<<<<<
 * 
 *         @property
 */
  __pyx_r = PyDateTime_DATE_GET_MINUTE(((PyObject *)__pyx_v_self));
  goto __pyx_L0;

  /* "datetime.pxd":127
 * 
 *         @property
 *         cdef inline int minute(self) noexcept:             # <<<<<<<<<<<<<<
 *             return PyDateTime_DATE_GET_MINUTE(self)
 * 
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "datetime.pxd":131
 * 
 *         @property
 *         cdef inline int second(self) noexcept:             # <<<<<<<<<<<<<<
 *             return PyDateTime_DATE_GET_SECOND(self)
 * 
 */

static CYTHON_INLINE int __pyx_f_7cpython_8datetime_8datetime_6second_second(PyDateTime_DateTime *__pyx_v_self) {
  int __pyx_r;

  /* "datetime.pxd":132
 *         @property
 *         cdef inline int second(self) noexcept:
 *             return PyDateTime_DATE_GET_SECOND(self)             # <<<<<<<<<<<<<<
 * 
 *         @property
 */
  __pyx_r = PyDateTime_DATE_GET_SECOND(((PyObject *)__pyx_v_self));
  goto __pyx_L0;

  /* "datetime.pxd":131
 * 
 *         @property
 *         cdef inline int second(self) noexcept:             # <<<<<<<<<<<<<<
 *             return PyDateTime_DATE_GET_SECOND(self)
 * 
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "datetime.pxd":135
 * 
 *         @property
 *         cdef inline int microsecond(self) noexcept:             # <<<<<<<<<<<<<<
 *             return PyDateTime_DATE_GET_MICROSECOND(self)
 * 
 */

static CYTHON_INLINE int __pyx_f_7cpython_8datetime_8datetime_11microsecond_microsecond(PyDateTime_DateTime *__pyx_v_self) {
  int __pyx_r;

  /* "datetime.pxd":136
 *         @property
 *         cdef inline int microsecond(self) noexcept:
 *             return PyDateTime_DATE_GET_MICROSECOND(self)             # <<<<<<<<<<<<<<
 * 
 *         @property
 */
  __pyx_r = PyDateTime_DATE_GET_MICROSECOND(((PyObject *)__pyx_v_self));
  goto __pyx_L0;

  /* "datetime.pxd":135
 * 
 *         @property
 *         cdef inline int microsecond(self) noexcept:             # <<<<<<<<<<<<<<
 *             return PyDateTime_DATE_GET_MICROSECOND(self)
 * 
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "datetime.pxd":139
 * 
 *         @property
 *         cdef inline object tzinfo(self):             # <<<<<<<<<<<<<<
 *             return <object>PyDateTime_DATE_GET_TZINFO(self)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_7cpython_8datetime_8datetime_6tzinfo_tzinfo(PyDateTime_DateTime *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1;
  __Pyx_RefNannySetupContext("tzinfo", 1);

  /* "datetime.pxd":140
 *         @property
 *         cdef inline object tzinfo(self):
 *             return <object>PyDateTime_DATE_GET_TZINFO(self)             # <<<<<<<<<<<<<<
 * 
 *         @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyDateTime_DATE_GET_TZINFO(((PyObject *)__pyx_v_self));
  __Pyx_INCREF(((PyObject *)__pyx_t_1));
  __pyx_r = ((PyObject *)__pyx_t_1);
  goto __pyx_L0;

  /* "datetime.pxd":139
 * 
 *         @property
 *         cdef inline object tzinfo(self):             # <<<<<<<<<<<<<<
 *             return <object>PyDateTime_DATE_GET_TZINFO(self)
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "datetime.pxd":143
 * 
 *         @property
 *         cdef inline int fold(self) noexcept:             # <<<<<<<<<<<<<<
 *             # For Python < 3.6 this returns 0 no matter what
 *             return PyDateTime_DATE_GET_FOLD(self)
 */

static CYTHON_INLINE int __pyx_f_7cpython_8datetime_8datetime_4fold_fold(PyDateTime_DateTime *__pyx_v_self) {
  int __pyx_r;

  /* "datetime.pxd":145
 *         cdef inline int fold(self) noexcept:
 *             # For Python < 3.6 this returns 0 no matter what
 *             return PyDateTime_DATE_GET_FOLD(self)             # <<<<<<<<<<<<<<
 * 
 *     ctypedef extern class datetime.timedelta[object PyDateTime_Delta]:
 */
  __pyx_r = PyDateTime_DATE_GET_FOLD(((PyObject *)__pyx_v_self));
  goto __pyx_L0;

  /* "datetime.pxd":143
 * 
 *         @property
 *         cdef inline int fold(self) noexcept:             # <<<<<<<<<<<<<<
 *             # For Python < 3.6 this returns 0 no matter what
 *             return PyDateTime_DATE_GET_FOLD(self)
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "datetime.pxd":149
 *     ctypedef extern class datetime.timedelta[object PyDateTime_Delta]:
 *         @property
 *         cdef inline int day(self) noexcept:             # <<<<<<<<<<<<<<
 *             return PyDateTime_DELTA_GET_DAYS(self)
 * 
 */

static CYTHON_INLINE int __pyx_f_7cpython_8datetime_9timedelta_3day_day(PyDateTime_Delta *__pyx_v_self) {
  int __pyx_r;

  /* "datetime.pxd":150
 *         @property
 *         cdef inline int day(self) noexcept:
 *             return PyDateTime_DELTA_GET_DAYS(self)             # <<<<<<<<<<<<<<
 * 
 *         @property
 */
  __pyx_r = PyDateTime_DELTA_GET_DAYS(((PyObject *)__pyx_v_self));
  goto __pyx_L0;

  /* "datetime.pxd":149
 *     ctypedef extern class datetime.timedelta[object PyDateTime_Delta]:
 *         @property
 *         cdef inline int day(self) noexcept:             # <<<<<<<<<<<<<<
 *             return PyDateTime_DELTA_GET_DAYS(self)
 * 
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "datetime.pxd":153
 * 
 *         @property
 *         cdef inline int second(self) noexcept:             # <<<<<<<<<<<<<<
 *             return PyDateTime_DELTA_GET_SECONDS(self)
 * 
 */

static CYTHON_INLINE int __pyx_f_7cpython_8datetime_9timedelta_6second_second(PyDateTime_Delta *__pyx_v_self) {
  int __pyx_r;

  /* "datetime.pxd":154
 *         @property
 *         cdef inline int second(self) noexcept:
 *             return PyDateTime_DELTA_GET_SECONDS(self)             # <<<<<<<<<<<<<<
 * 
 *         @property
 */
  __pyx_r = PyDateTime_DELTA_GET_SECONDS(((PyObject *)__pyx_v_self));
  goto __pyx_L0;

  /* "datetime.pxd":153
 * 
 *         @property
 *         cdef inline int second(self) noexcept:             # <<<<<<<<<<<<<<
 *             return PyDateTime_DELTA_GET_SECONDS(self)
 * 
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "datetime.pxd":157
 * 
 *         @property
 *         cdef inline int microsecond(self) noexcept:             # <<<<<<<<<<<<<<
 *             return PyDateTime_DELTA_GET_MICROSECONDS(self)
 * 
 */

static CYTHON_INLINE int __pyx_f_7cpython_8datetime_9timedelta_11microsecond_microsecond(PyDateTime_Delta *__pyx_v_self) {
  int __pyx_r;

  /* "datetime.pxd":158
 *         @property
 *         cdef inline int microsecond(self) noexcept:
 *             return PyDateTime_DELTA_GET_MICROSECONDS(self)             # <<<<<<<<<<<<<<
 * 
 *     ctypedef extern class datetime.tzinfo[object PyDateTime_TZInfo]:
 */
  __pyx_r = PyDateTime_DELTA_GET_MICROSECONDS(((PyObject *)__pyx_v_self));
  goto __pyx_L0;

  /* "datetime.pxd":157
 * 
 *         @property
 *         cdef inline int microsecond(self) noexcept:             # <<<<<<<<<<<<<<
 *             return PyDateTime_DELTA_GET_MICROSECONDS(self)
 * 
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "datetime.pxd":282
 * # Datetime C API initialization function.
 * # You have to call it before any usage of DateTime CAPI functions.
 * cdef inline void import_datetime() noexcept:             # <<<<<<<<<<<<<<
 *     PyDateTime_IMPORT
 * 
 */

static CYTHON_INLINE void __pyx_f_7cpython_8datetime_import_datetime(void) {

  /* "datetime.pxd":283
 * # You have to call it before any usage of DateTime CAPI functions.
 * cdef inline void import_datetime() noexcept:
 *     PyDateTime_IMPORT             # <<<<<<<<<<<<<<
 * 
 * # Create date object using DateTime CAPI factory function.
 */
  (void)(PyDateTime_IMPORT);

  /* "datetime.pxd":282
 * # Datetime C API initialization function.
 * # You have to call it before any usage of DateTime CAPI functions.
 * cdef inline void import_datetime() noexcept:             # <<<<<<<<<<<<<<
 *     PyDateTime_IMPORT
 * 
 */

  /* function exit code */
}

/* "datetime.pxd":287
 * # Create date object using DateTime CAPI factory function.
 * # Note, there are no range checks for any of the arguments.
 * cdef inline date date_new(int year, int month, int day):             # <<<<<<<<<<<<<<
 *     return PyDateTimeAPI.Date_FromDate(year, month, day, PyDateTimeAPI.DateType)
 * 
 */

static CYTHON_INLINE PyDateTime_Date *__pyx_f_7cpython_8datetime_date_new(int __pyx_v_year, int __pyx_v_month, int __pyx_v_day) {
  PyDateTime_Date *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("date_new", 1);

  /* "datetime.pxd":288
 * # Note, there are no range checks for any of the arguments.
 * cdef inline date date_new(int year, int month, int day):
 *     return PyDateTimeAPI.Date_FromDate(year, month, day, PyDateTimeAPI.DateType)             # <<<<<<<<<<<<<<
 * 
 * # Create time object using DateTime CAPI factory function
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __pyx_t_1 = ((PyObject *)PyDateTimeAPI->Date_FromDate(__pyx_v_year, __pyx_v_month, __pyx_v_day, PyDateTimeAPI->DateType)); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 288, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = ((PyDateTime_Date *)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "datetime.pxd":287
 * # Create date object using DateTime CAPI factory function.
 * # Note, there are no range checks for any of the arguments.
 * cdef inline date date_new(int year, int month, int day):             # <<<<<<<<<<<<<<
 *     return PyDateTimeAPI.Date_FromDate(year, month, day, PyDateTimeAPI.DateType)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cpython.datetime.date_new", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "datetime.pxd":292
 * # Create time object using DateTime CAPI factory function
 * # Note, there are no range checks for any of the arguments.
 * cdef inline time time_new(int hour, int minute, int second, int microsecond, object tz, int fold=0):             # <<<<<<<<<<<<<<
 *     return __Pyx_DateTime_TimeWithFold(hour, minute, second, microsecond, tz, fold)
 * 
 */

static CYTHON_INLINE PyDateTime_Time *__pyx_f_7cpython_8datetime_time_new(int __pyx_v_hour, int __pyx_v_minute, int __pyx_v_second, int __pyx_v_microsecond, PyObject *__pyx_v_tz, struct __pyx_opt_args_7cpython_8datetime_time_new *__pyx_optional_args) {
  int __pyx_v_fold = ((int)0);
  PyDateTime_Time *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("time_new", 1);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_fold = __pyx_optional_args->fold;
    }
  }

  /* "datetime.pxd":293
 * # Note, there are no range checks for any of the arguments.
 * cdef inline time time_new(int hour, int minute, int second, int microsecond, object tz, int fold=0):
 *     return __Pyx_DateTime_TimeWithFold(hour, minute, second, microsecond, tz, fold)             # <<<<<<<<<<<<<<
 * 
 * # Create datetime object using DateTime CAPI factory function.
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __pyx_t_1 = ((PyObject *)__Pyx_DateTime_TimeWithFold(__pyx_v_hour, __pyx_v_minute, __pyx_v_second, __pyx_v_microsecond, __pyx_v_tz, __pyx_v_fold)); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 293, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_7cpython_8datetime_time))))) __PYX_ERR(4, 293, __pyx_L1_error)
  __pyx_r = ((PyDateTime_Time *)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "datetime.pxd":292
 * # Create time object using DateTime CAPI factory function
 * # Note, there are no range checks for any of the arguments.
 * cdef inline time time_new(int hour, int minute, int second, int microsecond, object tz, int fold=0):             # <<<<<<<<<<<<<<
 *     return __Pyx_DateTime_TimeWithFold(hour, minute, second, microsecond, tz, fold)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cpython.datetime.time_new", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "datetime.pxd":297
 * # Create datetime object using DateTime CAPI factory function.
 * # Note, there are no range checks for any of the arguments.
 * cdef inline datetime datetime_new(int year, int month, int day, int hour, int minute, int second, int microsecond, object tz, int fold=0):             # <<<<<<<<<<<<<<
 *     return __Pyx_DateTime_DateTimeWithFold(year, month, day, hour, minute, second, microsecond, tz, fold)
 * 
 */

static CYTHON_INLINE PyDateTime_DateTime *__pyx_f_7cpython_8datetime_datetime_new(int __pyx_v_year, int __pyx_v_month, int __pyx_v_day, int __pyx_v_hour, int __pyx_v_minute, int __pyx_v_second, int __pyx_v_microsecond, PyObject *__pyx_v_tz, struct __pyx_opt_args_7cpython_8datetime_datetime_new *__pyx_optional_args) {
  int __pyx_v_fold = ((int)0);
  PyDateTime_DateTime *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("datetime_new", 1);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_fold = __pyx_optional_args->fold;
    }
  }

  /* "datetime.pxd":298
 * # Note, there are no range checks for any of the arguments.
 * cdef inline datetime datetime_new(int year, int month, int day, int hour, int minute, int second, int microsecond, object tz, int fold=0):
 *     return __Pyx_DateTime_DateTimeWithFold(year, month, day, hour, minute, second, microsecond, tz, fold)             # <<<<<<<<<<<<<<
 * 
 * # Create timedelta object using DateTime CAPI factory function.
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __pyx_t_1 = ((PyObject *)__Pyx_DateTime_DateTimeWithFold(__pyx_v_year, __pyx_v_month, __pyx_v_day, __pyx_v_hour, __pyx_v_minute, __pyx_v_second, __pyx_v_microsecond, __pyx_v_tz, __pyx_v_fold)); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 298, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = ((PyDateTime_DateTime *)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "datetime.pxd":297
 * # Create datetime object using DateTime CAPI factory function.
 * # Note, there are no range checks for any of the arguments.
 * cdef inline datetime datetime_new(int year, int month, int day, int hour, int minute, int second, int microsecond, object tz, int fold=0):             # <<<<<<<<<<<<<<
 *     return __Pyx_DateTime_DateTimeWithFold(year, month, day, hour, minute, second, microsecond, tz, fold)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cpython.datetime.datetime_new", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "datetime.pxd":302
 * # Create timedelta object using DateTime CAPI factory function.
 * # Note, there are no range checks for any of the arguments.
 * cdef inline timedelta timedelta_new(int days, int seconds, int useconds):             # <<<<<<<<<<<<<<
 *     return PyDateTimeAPI.Delta_FromDelta(days, seconds, useconds, 1, PyDateTimeAPI.DeltaType)
 * 
 */

static CYTHON_INLINE PyDateTime_Delta *__pyx_f_7cpython_8datetime_timedelta_new(int __pyx_v_days, int __pyx_v_seconds, int __pyx_v_useconds) {
  PyDateTime_Delta *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("timedelta_new", 1);

  /* "datetime.pxd":303
 * # Note, there are no range checks for any of the arguments.
 * cdef inline timedelta timedelta_new(int days, int seconds, int useconds):
 *     return PyDateTimeAPI.Delta_FromDelta(days, seconds, useconds, 1, PyDateTimeAPI.DeltaType)             # <<<<<<<<<<<<<<
 * 
 * # Create timedelta object using DateTime CAPI factory function.
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __pyx_t_1 = ((PyObject *)PyDateTimeAPI->Delta_FromDelta(__pyx_v_days, __pyx_v_seconds, __pyx_v_useconds, 1, PyDateTimeAPI->DeltaType)); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 303, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = ((PyDateTime_Delta *)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "datetime.pxd":302
 * # Create timedelta object using DateTime CAPI factory function.
 * # Note, there are no range checks for any of the arguments.
 * cdef inline timedelta timedelta_new(int days, int seconds, int useconds):             # <<<<<<<<<<<<<<
 *     return PyDateTimeAPI.Delta_FromDelta(days, seconds, useconds, 1, PyDateTimeAPI.DeltaType)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cpython.datetime.timedelta_new", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "datetime.pxd":306
 * 
 * # Create timedelta object using DateTime CAPI factory function.
 * cdef inline object timezone_new(object offset, object name=None):             # <<<<<<<<<<<<<<
 *     if PY_VERSION_HEX < 0x030700b1:
 *         raise RuntimeError('Time zones are not available from the C-API.')
 */

static CYTHON_INLINE PyObject *__pyx_f_7cpython_8datetime_timezone_new(PyObject *__pyx_v_offset, struct __pyx_opt_args_7cpython_8datetime_timezone_new *__pyx_optional_args) {
  PyObject *__pyx_v_name = ((PyObject *)Py_None);
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("timezone_new", 1);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_name = __pyx_optional_args->name;
    }
  }

  /* "datetime.pxd":307
 * # Create timedelta object using DateTime CAPI factory function.
 * cdef inline object timezone_new(object offset, object name=None):
 *     if PY_VERSION_HEX < 0x030700b1:             # <<<<<<<<<<<<<<
 *         raise RuntimeError('Time zones are not available from the C-API.')
 *     return __Pyx_TimeZone_FromOffsetAndName(offset, <PyObject*>name if name is not None else NULL)
 */
  __pyx_t_1 = (PY_VERSION_HEX < 0x030700b1);
  if (unlikely(__pyx_t_1)) {

    /* "datetime.pxd":308
 * cdef inline object timezone_new(object offset, object name=None):
 *     if PY_VERSION_HEX < 0x030700b1:
 *         raise RuntimeError('Time zones are not available from the C-API.')             # <<<<<<<<<<<<<<
 *     return __Pyx_TimeZone_FromOffsetAndName(offset, <PyObject*>name if name is not None else NULL)
 * 
 */
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_RuntimeError, __pyx_tuple__3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 308, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(4, 308, __pyx_L1_error)

    /* "datetime.pxd":307
 * # Create timedelta object using DateTime CAPI factory function.
 * cdef inline object timezone_new(object offset, object name=None):
 *     if PY_VERSION_HEX < 0x030700b1:             # <<<<<<<<<<<<<<
 *         raise RuntimeError('Time zones are not available from the C-API.')
 *     return __Pyx_TimeZone_FromOffsetAndName(offset, <PyObject*>name if name is not None else NULL)
 */
  }

  /* "datetime.pxd":309
 *     if PY_VERSION_HEX < 0x030700b1:
 *         raise RuntimeError('Time zones are not available from the C-API.')
 *     return __Pyx_TimeZone_FromOffsetAndName(offset, <PyObject*>name if name is not None else NULL)             # <<<<<<<<<<<<<<
 * 
 * # Create datetime object using DB API constructor.
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = (__pyx_v_name != Py_None);
  if (__pyx_t_1) {
    __pyx_t_3 = ((PyObject *)__pyx_v_name);
  } else {
    __pyx_t_3 = NULL;
  }
  __pyx_t_2 = __Pyx_TimeZone_FromOffsetAndName(__pyx_v_offset, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 309, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "datetime.pxd":306
 * 
 * # Create timedelta object using DateTime CAPI factory function.
 * cdef inline object timezone_new(object offset, object name=None):             # <<<<<<<<<<<<<<
 *     if PY_VERSION_HEX < 0x030700b1:
 *         raise RuntimeError('Time zones are not available from the C-API.')
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("cpython.datetime.timezone_new", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "datetime.pxd":312
 * 
 * # Create datetime object using DB API constructor.
 * cdef inline datetime datetime_from_timestamp(timestamp, tz=None):             # <<<<<<<<<<<<<<
 *     return PyDateTimeAPI.DateTime_FromTimestamp(
 *         <PyObject*>PyDateTimeAPI.DateTimeType, (timestamp, tz) if tz is not None else (timestamp,), NULL)
 */

static CYTHON_INLINE PyDateTime_DateTime *__pyx_f_7cpython_8datetime_datetime_from_timestamp(PyObject *__pyx_v_timestamp, struct __pyx_opt_args_7cpython_8datetime_datetime_from_timestamp *__pyx_optional_args) {
  PyObject *__pyx_v_tz = ((PyObject *)Py_None);
  PyDateTime_DateTime *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("datetime_from_timestamp", 1);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_tz = __pyx_optional_args->tz;
    }
  }

  /* "datetime.pxd":313
 * # Create datetime object using DB API constructor.
 * cdef inline datetime datetime_from_timestamp(timestamp, tz=None):
 *     return PyDateTimeAPI.DateTime_FromTimestamp(             # <<<<<<<<<<<<<<
 *         <PyObject*>PyDateTimeAPI.DateTimeType, (timestamp, tz) if tz is not None else (timestamp,), NULL)
 * 
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);

  /* "datetime.pxd":314
 * cdef inline datetime datetime_from_timestamp(timestamp, tz=None):
 *     return PyDateTimeAPI.DateTime_FromTimestamp(
 *         <PyObject*>PyDateTimeAPI.DateTimeType, (timestamp, tz) if tz is not None else (timestamp,), NULL)             # <<<<<<<<<<<<<<
 * 
 * # Create date object using DB API constructor.
 */
  __pyx_t_2 = (__pyx_v_tz != Py_None);
  if (__pyx_t_2) {
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 314, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(__pyx_v_timestamp);
    __Pyx_GIVEREF(__pyx_v_timestamp);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_timestamp)) __PYX_ERR(4, 314, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_tz);
    __Pyx_GIVEREF(__pyx_v_tz);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_tz)) __PYX_ERR(4, 314, __pyx_L1_error);
    __pyx_t_1 = __pyx_t_3;
    __pyx_t_3 = 0;
  } else {
    __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 314, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(__pyx_v_timestamp);
    __Pyx_GIVEREF(__pyx_v_timestamp);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_timestamp)) __PYX_ERR(4, 314, __pyx_L1_error);
    __pyx_t_1 = __pyx_t_3;
    __pyx_t_3 = 0;
  }

  /* "datetime.pxd":313
 * # Create datetime object using DB API constructor.
 * cdef inline datetime datetime_from_timestamp(timestamp, tz=None):
 *     return PyDateTimeAPI.DateTime_FromTimestamp(             # <<<<<<<<<<<<<<
 *         <PyObject*>PyDateTimeAPI.DateTimeType, (timestamp, tz) if tz is not None else (timestamp,), NULL)
 * 
 */
  __pyx_t_3 = ((PyObject *)PyDateTimeAPI->DateTime_FromTimestamp(((PyObject *)PyDateTimeAPI->DateTimeType), __pyx_t_1, NULL)); if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 313, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = ((PyDateTime_DateTime *)__pyx_t_3);
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "datetime.pxd":312
 * 
 * # Create datetime object using DB API constructor.
 * cdef inline datetime datetime_from_timestamp(timestamp, tz=None):             # <<<<<<<<<<<<<<
 *     return PyDateTimeAPI.DateTime_FromTimestamp(
 *         <PyObject*>PyDateTimeAPI.DateTimeType, (timestamp, tz) if tz is not None else (timestamp,), NULL)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("cpython.datetime.datetime_from_timestamp", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "datetime.pxd":317
 * 
 * # Create date object using DB API constructor.
 * cdef inline date date_from_timestamp(timestamp):             # <<<<<<<<<<<<<<
 *     return PyDateTimeAPI.Date_FromTimestamp(<PyObject*>PyDateTimeAPI.DateType, (timestamp,))
 * 
 */

static CYTHON_INLINE PyDateTime_Date *__pyx_f_7cpython_8datetime_date_from_timestamp(PyObject *__pyx_v_timestamp) {
  PyDateTime_Date *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("date_from_timestamp", 1);

  /* "datetime.pxd":318
 * # Create date object using DB API constructor.
 * cdef inline date date_from_timestamp(timestamp):
 *     return PyDateTimeAPI.Date_FromTimestamp(<PyObject*>PyDateTimeAPI.DateType, (timestamp,))             # <<<<<<<<<<<<<<
 * 
 * # More recognizable getters for date/time/datetime/timedelta.
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 318, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_timestamp);
  __Pyx_GIVEREF(__pyx_v_timestamp);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_timestamp)) __PYX_ERR(4, 318, __pyx_L1_error);
  __pyx_t_2 = ((PyObject *)PyDateTimeAPI->Date_FromTimestamp(((PyObject *)PyDateTimeAPI->DateType), __pyx_t_1)); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 318, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = ((PyDateTime_Date *)__pyx_t_2);
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "datetime.pxd":317
 * 
 * # Create date object using DB API constructor.
 * cdef inline date date_from_timestamp(timestamp):             # <<<<<<<<<<<<<<
 *     return PyDateTimeAPI.Date_FromTimestamp(<PyObject*>PyDateTimeAPI.DateType, (timestamp,))
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("cpython.datetime.date_from_timestamp", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "datetime.pxd":326
 * 
 * # Get UTC singleton
 * cdef inline object get_utc():             # <<<<<<<<<<<<<<
 *     if PY_VERSION_HEX < 0x030700b1:
 *         raise RuntimeError('Time zones are not available from the C-API.')
 */

static CYTHON_INLINE PyObject *__pyx_f_7cpython_8datetime_get_utc(void) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_utc", 1);

  /* "datetime.pxd":327
 * # Get UTC singleton
 * cdef inline object get_utc():
 *     if PY_VERSION_HEX < 0x030700b1:             # <<<<<<<<<<<<<<
 *         raise RuntimeError('Time zones are not available from the C-API.')
 *     return <object>__Pyx_TimeZone_UTC
 */
  __pyx_t_1 = (PY_VERSION_HEX < 0x030700b1);
  if (unlikely(__pyx_t_1)) {

    /* "datetime.pxd":328
 * cdef inline object get_utc():
 *     if PY_VERSION_HEX < 0x030700b1:
 *         raise RuntimeError('Time zones are not available from the C-API.')             # <<<<<<<<<<<<<<
 *     return <object>__Pyx_TimeZone_UTC
 * 
 */
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_RuntimeError, __pyx_tuple__3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 328, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(4, 328, __pyx_L1_error)

    /* "datetime.pxd":327
 * # Get UTC singleton
 * cdef inline object get_utc():
 *     if PY_VERSION_HEX < 0x030700b1:             # <<<<<<<<<<<<<<
 *         raise RuntimeError('Time zones are not available from the C-API.')
 *     return <object>__Pyx_TimeZone_UTC
 */
  }

  /* "datetime.pxd":329
 *     if PY_VERSION_HEX < 0x030700b1:
 *         raise RuntimeError('Time zones are not available from the C-API.')
 *     return <object>__Pyx_TimeZone_UTC             # <<<<<<<<<<<<<<
 * 
 * # Get tzinfo of time
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__Pyx_TimeZone_UTC));
  __pyx_r = ((PyObject *)__Pyx_TimeZone_UTC);
  goto __pyx_L0;

  /* "datetime.pxd":326
 * 
 * # Get UTC singleton
 * cdef inline object get_utc():             # <<<<<<<<<<<<<<
 *     if PY_VERSION_HEX < 0x030700b1:
 *         raise RuntimeError('Time zones are not available from the C-API.')
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("cpython.datetime.get_utc", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "datetime.pxd":332
 * 
 * # Get tzinfo of time
 * cdef inline object time_tzinfo(object o):             # <<<<<<<<<<<<<<
 *     return <object>PyDateTime_TIME_GET_TZINFO(o)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_7cpython_8datetime_time_tzinfo(PyObject *__pyx_v_o) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1;
  __Pyx_RefNannySetupContext("time_tzinfo", 1);

  /* "datetime.pxd":333
 * # Get tzinfo of time
 * cdef inline object time_tzinfo(object o):
 *     return <object>PyDateTime_TIME_GET_TZINFO(o)             # <<<<<<<<<<<<<<
 * 
 * # Get tzinfo of datetime
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyDateTime_TIME_GET_TZINFO(__pyx_v_o);
  __Pyx_INCREF(((PyObject *)__pyx_t_1));
  __pyx_r = ((PyObject *)__pyx_t_1);
  goto __pyx_L0;

  /* "datetime.pxd":332
 * 
 * # Get tzinfo of time
 * cdef inline object time_tzinfo(object o):             # <<<<<<<<<<<<<<
 *     return <object>PyDateTime_TIME_GET_TZINFO(o)
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "datetime.pxd":336
 * 
 * # Get tzinfo of datetime
 * cdef inline object datetime_tzinfo(object o):             # <<<<<<<<<<<<<<
 *     return <object>PyDateTime_DATE_GET_TZINFO(o)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_7cpython_8datetime_datetime_tzinfo(PyObject *__pyx_v_o) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1;
  __Pyx_RefNannySetupContext("datetime_tzinfo", 1);

  /* "datetime.pxd":337
 * # Get tzinfo of datetime
 * cdef inline object datetime_tzinfo(object o):
 *     return <object>PyDateTime_DATE_GET_TZINFO(o)             # <<<<<<<<<<<<<<
 * 
 * # Get year of date
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyDateTime_DATE_GET_TZINFO(__pyx_v_o);
  __Pyx_INCREF(((PyObject *)__pyx_t_1));
  __pyx_r = ((PyObject *)__pyx_t_1);
  goto __pyx_L0;

  /* "datetime.pxd":336
 * 
 * # Get tzinfo of datetime
 * cdef inline object datetime_tzinfo(object o):             # <<<<<<<<<<<<<<
 *     return <object>PyDateTime_DATE_GET_TZINFO(o)
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "datetime.pxd":340
 * 
 * # Get year of date
 * cdef inline int date_year(object o) noexcept:             # <<<<<<<<<<<<<<
 *     return PyDateTime_GET_YEAR(o)
 * 
 */

static CYTHON_INLINE int __pyx_f_7cpython_8datetime_date_year(PyObject *__pyx_v_o) {
  int __pyx_r;

  /* "datetime.pxd":341
 * # Get year of date
 * cdef inline int date_year(object o) noexcept:
 *     return PyDateTime_GET_YEAR(o)             # <<<<<<<<<<<<<<
 * 
 * # Get month of date
 */
  __pyx_r = PyDateTime_GET_YEAR(__pyx_v_o);
  goto __pyx_L0;

  /* "datetime.pxd":340
 * 
 * # Get year of date
 * cdef inline int date_year(object o) noexcept:             # <<<<<<<<<<<<<<
 *     return PyDateTime_GET_YEAR(o)
 * 
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "datetime.pxd":344
 * 
 * # Get month of date
 * cdef inline int date_month(object o) noexcept:             # <<<<<<<<<<<<<<
 *     return PyDateTime_GET_MONTH(o)
 * 
 */

static CYTHON_INLINE int __pyx_f_7cpython_8datetime_date_month(PyObject *__pyx_v_o) {
  int __pyx_r;

  /* "datetime.pxd":345
 * # Get month of date
 * cdef inline int date_month(object o) noexcept:
 *     return PyDateTime_GET_MONTH(o)             # <<<<<<<<<<<<<<
 * 
 * # Get day of date
 */
  __pyx_r = PyDateTime_GET_MONTH(__pyx_v_o);
  goto __pyx_L0;

  /* "datetime.pxd":344
 * 
 * # Get month of date
 * cdef inline int date_month(object o) noexcept:             # <<<<<<<<<<<<<<
 *     return PyDateTime_GET_MONTH(o)
 * 
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "datetime.pxd":348
 * 
 * # Get day of date
 * cdef inline int date_day(object o) noexcept:             # <<<<<<<<<<<<<<
 *     return PyDateTime_GET_DAY(o)
 * 
 */

static CYTHON_INLINE int __pyx_f_7cpython_8datetime_date_day(PyObject *__pyx_v_o) {
  int __pyx_r;

  /* "datetime.pxd":349
 * # Get day of date
 * cdef inline int date_day(object o) noexcept:
 *     return PyDateTime_GET_DAY(o)             # <<<<<<<<<<<<<<
 * 
 * # Get year of datetime
 */
  __pyx_r = PyDateTime_GET_DAY(__pyx_v_o);
  goto __pyx_L0;

  /* "datetime.pxd":348
 * 
 * # Get day of date
 * cdef inline int date_day(object o) noexcept:             # <<<<<<<<<<<<<<
 *     return PyDateTime_GET_DAY(o)
 * 
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "datetime.pxd":352
 * 
 * # Get year of datetime
 * cdef inline int datetime_year(object o) noexcept:             # <<<<<<<<<<<<<<
 *     return PyDateTime_GET_YEAR(o)
 * 
 */

static CYTHON_INLINE int __pyx_f_7cpython_8datetime_datetime_year(PyObject *__pyx_v_o) {
  int __pyx_r;

  /* "datetime.pxd":353
 * # Get year of datetime
 * cdef inline int datetime_year(object o) noexcept:
 *     return PyDateTime_GET_YEAR(o)             # <<<<<<<<<<<<<<
 * 
 * # Get month of datetime
 */
  __pyx_r = PyDateTime_GET_YEAR(__pyx_v_o);
  goto __pyx_L0;

  /* "datetime.pxd":352
 * 
 * # Get year of datetime
 * cdef inline int datetime_year(object o) noexcept:             # <<<<<<<<<<<<<<
 *     return PyDateTime_GET_YEAR(o)
 * 
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "datetime.pxd":356
 * 
 * # Get month of datetime
 * cdef inline int datetime_month(object o) noexcept:             # <<<<<<<<<<<<<<
 *     return PyDateTime_GET_MONTH(o)
 * 
 */

static CYTHON_INLINE int __pyx_f_7cpython_8datetime_datetime_month(PyObject *__pyx_v_o) {
  int __pyx_r;

  /* "datetime.pxd":357
 * # Get month of datetime
 * cdef inline int datetime_month(object o) noexcept:
 *     return PyDateTime_GET_MONTH(o)             # <<<<<<<<<<<<<<
 * 
 * # Get day of datetime
 */
  __pyx_r = PyDateTime_GET_MONTH(__pyx_v_o);
  goto __pyx_L0;

  /* "datetime.pxd":356
 * 
 * # Get month of datetime
 * cdef inline int datetime_month(object o) noexcept:             # <<<<<<<<<<<<<<
 *     return PyDateTime_GET_MONTH(o)
 * 
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "datetime.pxd":360
 * 
 * # Get day of datetime
 * cdef inline int datetime_day(object o) noexcept:             # <<<<<<<<<<<<<<
 *     return PyDateTime_GET_DAY(o)
 * 
 */

static CYTHON_INLINE int __pyx_f_7cpython_8datetime_datetime_day(PyObject *__pyx_v_o) {
  int __pyx_r;

  /* "datetime.pxd":361
 * # Get day of datetime
 * cdef inline int datetime_day(object o) noexcept:
 *     return PyDateTime_GET_DAY(o)             # <<<<<<<<<<<<<<
 * 
 * # Get hour of time
 */
  __pyx_r = PyDateTime_GET_DAY(__pyx_v_o);
  goto __pyx_L0;

  /* "datetime.pxd":360
 * 
 * # Get day of datetime
 * cdef inline int datetime_day(object o) noexcept:             # <<<<<<<<<<<<<<
 *     return PyDateTime_GET_DAY(o)
 * 
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "datetime.pxd":364
 * 
 * # Get hour of time
 * cdef inline int time_hour(object o) noexcept:             # <<<<<<<<<<<<<<
 *     return PyDateTime_TIME_GET_HOUR(o)
 * 
 */

static CYTHON_INLINE int __pyx_f_7cpython_8datetime_time_hour(PyObject *__pyx_v_o) {
  int __pyx_r;

  /* "datetime.pxd":365
 * # Get hour of time
 * cdef inline int time_hour(object o) noexcept:
 *     return PyDateTime_TIME_GET_HOUR(o)             # <<<<<<<<<<<<<<
 * 
 * # Get minute of time
 */
  __pyx_r = PyDateTime_TIME_GET_HOUR(__pyx_v_o);
  goto __pyx_L0;

  /* "datetime.pxd":364
 * 
 * # Get hour of time
 * cdef inline int time_hour(object o) noexcept:             # <<<<<<<<<<<<<<
 *     return PyDateTime_TIME_GET_HOUR(o)
 * 
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "datetime.pxd":368
 * 
 * # Get minute of time
 * cdef inline int time_minute(object o) noexcept:             # <<<<<<<<<<<<<<
 *     return PyDateTime_TIME_GET_MINUTE(o)
 * 
 */

static CYTHON_INLINE int __pyx_f_7cpython_8datetime_time_minute(PyObject *__pyx_v_o) {
  int __pyx_r;

  /* "datetime.pxd":369
 * # Get minute of time
 * cdef inline int time_minute(object o) noexcept:
 *     return PyDateTime_TIME_GET_MINUTE(o)             # <<<<<<<<<<<<<<
 * 
 * # Get second of time
 */
  __pyx_r = PyDateTime_TIME_GET_MINUTE(__pyx_v_o);
  goto __pyx_L0;

  /* "datetime.pxd":368
 * 
 * # Get minute of time
 * cdef inline int time_minute(object o) noexcept:             # <<<<<<<<<<<<<<
 *     return PyDateTime_TIME_GET_MINUTE(o)
 * 
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "datetime.pxd":372
 * 
 * # Get second of time
 * cdef inline int time_second(object o) noexcept:             # <<<<<<<<<<<<<<
 *     return PyDateTime_TIME_GET_SECOND(o)
 * 
 */

static CYTHON_INLINE int __pyx_f_7cpython_8datetime_time_second(PyObject *__pyx_v_o) {
  int __pyx_r;

  /* "datetime.pxd":373
 * # Get second of time
 * cdef inline int time_second(object o) noexcept:
 *     return PyDateTime_TIME_GET_SECOND(o)             # <<<<<<<<<<<<<<
 * 
 * # Get microsecond of time
 */
  __pyx_r = PyDateTime_TIME_GET_SECOND(__pyx_v_o);
  goto __pyx_L0;

  /* "datetime.pxd":372
 * 
 * # Get second of time
 * cdef inline int time_second(object o) noexcept:             # <<<<<<<<<<<<<<
 *     return PyDateTime_TIME_GET_SECOND(o)
 * 
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "datetime.pxd":376
 * 
 * # Get microsecond of time
 * cdef inline int time_microsecond(object o) noexcept:             # <<<<<<<<<<<<<<
 *     return PyDateTime_TIME_GET_MICROSECOND(o)
 * 
 */

static CYTHON_INLINE int __pyx_f_7cpython_8datetime_time_microsecond(PyObject *__pyx_v_o) {
  int __pyx_r;

  /* "datetime.pxd":377
 * # Get microsecond of time
 * cdef inline int time_microsecond(object o) noexcept:
 *     return PyDateTime_TIME_GET_MICROSECOND(o)             # <<<<<<<<<<<<<<
 * 
 * # Get fold of time
 */
  __pyx_r = PyDateTime_TIME_GET_MICROSECOND(__pyx_v_o);
  goto __pyx_L0;

  /* "datetime.pxd":376
 * 
 * # Get microsecond of time
 * cdef inline int time_microsecond(object o) noexcept:             # <<<<<<<<<<<<<<
 *     return PyDateTime_TIME_GET_MICROSECOND(o)
 * 
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "datetime.pxd":380
 * 
 * # Get fold of time
 * cdef inline int time_fold(object o) noexcept:             # <<<<<<<<<<<<<<
 *     # For Python < 3.6 this returns 0 no matter what
 *     return PyDateTime_TIME_GET_FOLD(o)
 */

static CYTHON_INLINE int __pyx_f_7cpython_8datetime_time_fold(PyObject *__pyx_v_o) {
  int __pyx_r;

  /* "datetime.pxd":382
 * cdef inline int time_fold(object o) noexcept:
 *     # For Python < 3.6 this returns 0 no matter what
 *     return PyDateTime_TIME_GET_FOLD(o)             # <<<<<<<<<<<<<<
 * 
 * # Get hour of datetime
 */
  __pyx_r = PyDateTime_TIME_GET_FOLD(__pyx_v_o);
  goto __pyx_L0;

  /* "datetime.pxd":380
 * 
 * # Get fold of time
 * cdef inline int time_fold(object o) noexcept:             # <<<<<<<<<<<<<<
 *     # For Python < 3.6 this returns 0 no matter what
 *     return PyDateTime_TIME_GET_FOLD(o)
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "datetime.pxd":385
 * 
 * # Get hour of datetime
 * cdef inline int datetime_hour(object o) noexcept:             # <<<<<<<<<<<<<<
 *     return PyDateTime_DATE_GET_HOUR(o)
 * 
 */

static CYTHON_INLINE int __pyx_f_7cpython_8datetime_datetime_hour(PyObject *__pyx_v_o) {
  int __pyx_r;

  /* "datetime.pxd":386
 * # Get hour of datetime
 * cdef inline int datetime_hour(object o) noexcept:
 *     return PyDateTime_DATE_GET_HOUR(o)             # <<<<<<<<<<<<<<
 * 
 * # Get minute of datetime
 */
  __pyx_r = PyDateTime_DATE_GET_HOUR(__pyx_v_o);
  goto __pyx_L0;

  /* "datetime.pxd":385
 * 
 * # Get hour of datetime
 * cdef inline int datetime_hour(object o) noexcept:             # <<<<<<<<<<<<<<
 *     return PyDateTime_DATE_GET_HOUR(o)
 * 
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "datetime.pxd":389
 * 
 * # Get minute of datetime
 * cdef inline int datetime_minute(object o) noexcept:             # <<<<<<<<<<<<<<
 *     return PyDateTime_DATE_GET_MINUTE(o)
 * 
 */

static CYTHON_INLINE int __pyx_f_7cpython_8datetime_datetime_minute(PyObject *__pyx_v_o) {
  int __pyx_r;

  /* "datetime.pxd":390
 * # Get minute of datetime
 * cdef inline int datetime_minute(object o) noexcept:
 *     return PyDateTime_DATE_GET_MINUTE(o)             # <<<<<<<<<<<<<<
 * 
 * # Get second of datetime
 */
  __pyx_r = PyDateTime_DATE_GET_MINUTE(__pyx_v_o);
  goto __pyx_L0;

  /* "datetime.pxd":389
 * 
 * # Get minute of datetime
 * cdef inline int datetime_minute(object o) noexcept:             # <<<<<<<<<<<<<<
 *     return PyDateTime_DATE_GET_MINUTE(o)
 * 
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "datetime.pxd":393
 * 
 * # Get second of datetime
 * cdef inline int datetime_second(object o) noexcept:             # <<<<<<<<<<<<<<
 *     return PyDateTime_DATE_GET_SECOND(o)
 * 
 */

static CYTHON_INLINE int __pyx_f_7cpython_8datetime_datetime_second(PyObject *__pyx_v_o) {
  int __pyx_r;

  /* "datetime.pxd":394
 * # Get second of datetime
 * cdef inline int datetime_second(object o) noexcept:
 *     return PyDateTime_DATE_GET_SECOND(o)             # <<<<<<<<<<<<<<
 * 
 * # Get microsecond of datetime
 */
  __pyx_r = PyDateTime_DATE_GET_SECOND(__pyx_v_o);
  goto __pyx_L0;

  /* "datetime.pxd":393
 * 
 * # Get second of datetime
 * cdef inline int datetime_second(object o) noexcept:             # <<<<<<<<<<<<<<
 *     return PyDateTime_DATE_GET_SECOND(o)
 * 
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "datetime.pxd":397
 * 
 * # Get microsecond of datetime
 * cdef inline int datetime_microsecond(object o) noexcept:             # <<<<<<<<<<<<<<
 *     return PyDateTime_DATE_GET_MICROSECOND(o)
 * 
 */

static CYTHON_INLINE int __pyx_f_7cpython_8datetime_datetime_microsecond(PyObject *__pyx_v_o) {
  int __pyx_r;

  /* "datetime.pxd":398
 * # Get microsecond of datetime
 * cdef inline int datetime_microsecond(object o) noexcept:
 *     return PyDateTime_DATE_GET_MICROSECOND(o)             # <<<<<<<<<<<<<<
 * 
 * # Get fold of datetime
 */
  __pyx_r = PyDateTime_DATE_GET_MICROSECOND(__pyx_v_o);
  goto __pyx_L0;

  /* "datetime.pxd":397
 * 
 * # Get microsecond of datetime
 * cdef inline int datetime_microsecond(object o) noexcept:             # <<<<<<<<<<<<<<
 *     return PyDateTime_DATE_GET_MICROSECOND(o)
 * 
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "datetime.pxd":401
 * 
 * # Get fold of datetime
 * cdef inline int datetime_fold(object o) noexcept:             # <<<<<<<<<<<<<<
 *     # For Python < 3.6 this returns 0 no matter what
 *     return PyDateTime_DATE_GET_FOLD(o)
 */

static CYTHON_INLINE int __pyx_f_7cpython_8datetime_datetime_fold(PyObject *__pyx_v_o) {
  int __pyx_r;

  /* "datetime.pxd":403
 * cdef inline int datetime_fold(object o) noexcept:
 *     # For Python < 3.6 this returns 0 no matter what
 *     return PyDateTime_DATE_GET_FOLD(o)             # <<<<<<<<<<<<<<
 * 
 * # Get days of timedelta
 */
  __pyx_r = PyDateTime_DATE_GET_FOLD(__pyx_v_o);
  goto __pyx_L0;

  /* "datetime.pxd":401
 * 
 * # Get fold of datetime
 * cdef inline int datetime_fold(object o) noexcept:             # <<<<<<<<<<<<<<
 *     # For Python < 3.6 this returns 0 no matter what
 *     return PyDateTime_DATE_GET_FOLD(o)
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "datetime.pxd":406
 * 
 * # Get days of timedelta
 * cdef inline int timedelta_days(object o) noexcept:             # <<<<<<<<<<<<<<
 *     return (<PyDateTime_Delta*>o).days
 * 
 */

static CYTHON_INLINE int __pyx_f_7cpython_8datetime_timedelta_days(PyObject *__pyx_v_o) {
  int __pyx_r;

  /* "datetime.pxd":407
 * # Get days of timedelta
 * cdef inline int timedelta_days(object o) noexcept:
 *     return (<PyDateTime_Delta*>o).days             # <<<<<<<<<<<<<<
 * 
 * # Get seconds of timedelta
 */
  __pyx_r = ((PyDateTime_Delta *)__pyx_v_o)->days;
  goto __pyx_L0;

  /* "datetime.pxd":406
 * 
 * # Get days of timedelta
 * cdef inline int timedelta_days(object o) noexcept:             # <<<<<<<<<<<<<<
 *     return (<PyDateTime_Delta*>o).days
 * 
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "datetime.pxd":410
 * 
 * # Get seconds of timedelta
 * cdef inline int timedelta_seconds(object o) noexcept:             # <<<<<<<<<<<<<<
 *     return (<PyDateTime_Delta*>o).seconds
 * 
 */

static CYTHON_INLINE int __pyx_f_7cpython_8datetime_timedelta_seconds(PyObject *__pyx_v_o) {
  int __pyx_r;

  /* "datetime.pxd":411
 * # Get seconds of timedelta
 * cdef inline int timedelta_seconds(object o) noexcept:
 *     return (<PyDateTime_Delta*>o).seconds             # <<<<<<<<<<<<<<
 * 
 * # Get microseconds of timedelta
 */
  __pyx_r = ((PyDateTime_Delta *)__pyx_v_o)->seconds;
  goto __pyx_L0;

  /* "datetime.pxd":410
 * 
 * # Get seconds of timedelta
 * cdef inline int timedelta_seconds(object o) noexcept:             # <<<<<<<<<<<<<<
 *     return (<PyDateTime_Delta*>o).seconds
 * 
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "datetime.pxd":414
 * 
 * # Get microseconds of timedelta
 * cdef inline int timedelta_microseconds(object o) noexcept:             # <<<<<<<<<<<<<<
 *     return (<PyDateTime_Delta*>o).microseconds
 * 
 */

static CYTHON_INLINE int __pyx_f_7cpython_8datetime_timedelta_microseconds(PyObject *__pyx_v_o) {
  int __pyx_r;

  /* "datetime.pxd":415
 * # Get microseconds of timedelta
 * cdef inline int timedelta_microseconds(object o) noexcept:
 *     return (<PyDateTime_Delta*>o).microseconds             # <<<<<<<<<<<<<<
 * 
 * cdef inline double total_seconds(timedelta obj) noexcept:
 */
  __pyx_r = ((PyDateTime_Delta *)__pyx_v_o)->microseconds;
  goto __pyx_L0;

  /* "datetime.pxd":414
 * 
 * # Get microseconds of timedelta
 * cdef inline int timedelta_microseconds(object o) noexcept:             # <<<<<<<<<<<<<<
 *     return (<PyDateTime_Delta*>o).microseconds
 * 
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "datetime.pxd":417
 *     return (<PyDateTime_Delta*>o).microseconds
 * 
 * cdef inline double total_seconds(timedelta obj) noexcept:             # <<<<<<<<<<<<<<
 *     # Mirrors the "timedelta.total_seconds()" method.
 *     # Note that this implementation is not guaranteed to give *exactly* the same
 */

static CYTHON_INLINE double __pyx_f_7cpython_8datetime_total_seconds(PyDateTime_Delta *__pyx_v_obj) {
  double __pyx_v_days;
  double __pyx_v_seconds;
  double __pyx_v_micros;
  double __pyx_r;

  /* "datetime.pxd":423
 *     cdef:
 *         double days, seconds, micros
 *     days = <double>PyDateTime_DELTA_GET_DAYS(obj)             # <<<<<<<<<<<<<<
 *     seconds = <double>PyDateTime_DELTA_GET_SECONDS(obj)
 *     micros = <double>PyDateTime_DELTA_GET_MICROSECONDS(obj)
 */
  __pyx_v_days = ((double)PyDateTime_DELTA_GET_DAYS(((PyObject *)__pyx_v_obj)));

  /* "datetime.pxd":424
 *         double days, seconds, micros
 *     days = <double>PyDateTime_DELTA_GET_DAYS(obj)
 *     seconds = <double>PyDateTime_DELTA_GET_SECONDS(obj)             # <<<<<<<<<<<<<<
 *     micros = <double>PyDateTime_DELTA_GET_MICROSECONDS(obj)
 *     return days * 24 * 3600 + seconds + micros / 1_000_000
 */
  __pyx_v_seconds = ((double)PyDateTime_DELTA_GET_SECONDS(((PyObject *)__pyx_v_obj)));

  /* "datetime.pxd":425
 *     days = <double>PyDateTime_DELTA_GET_DAYS(obj)
 *     seconds = <double>PyDateTime_DELTA_GET_SECONDS(obj)
 *     micros = <double>PyDateTime_DELTA_GET_MICROSECONDS(obj)             # <<<<<<<<<<<<<<
 *     return days * 24 * 3600 + seconds + micros / 1_000_000
 */
  __pyx_v_micros = ((double)PyDateTime_DELTA_GET_MICROSECONDS(((PyObject *)__pyx_v_obj)));

  /* "datetime.pxd":426
 *     seconds = <double>PyDateTime_DELTA_GET_SECONDS(obj)
 *     micros = <double>PyDateTime_DELTA_GET_MICROSECONDS(obj)
 *     return days * 24 * 3600 + seconds + micros / 1_000_000             # <<<<<<<<<<<<<<
 */
  __pyx_r = ((((__pyx_v_days * 24.0) * 3600.0) + __pyx_v_seconds) + (__pyx_v_micros / 1000000.0));
  goto __pyx_L0;

  /* "datetime.pxd":417
 *     return (<PyDateTime_Delta*>o).microseconds
 * 
 * cdef inline double total_seconds(timedelta obj) noexcept:             # <<<<<<<<<<<<<<
 *     # Mirrors the "timedelta.total_seconds()" method.
 *     # Note that this implementation is not guaranteed to give *exactly* the same
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "utils.pxd":33
 * 
 * # Utils: encode / decode
 * cdef inline bytes encode_str(object obj, char* encoding):             # <<<<<<<<<<<<<<
 *     """Encode string to bytes using the 'encoding' with
 *     'surrogateescape' error handling `<'bytes'>`."""
 */

static CYTHON_INLINE PyObject *__pyx_f_10serializor_5utils_encode_str(PyObject *__pyx_v_obj, char *__pyx_v_encoding) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("encode_str", 1);

  /* "utils.pxd":36
 *     """Encode string to bytes using the 'encoding' with
 *     'surrogateescape' error handling `<'bytes'>`."""
 *     return PyUnicode_AsEncodedString(obj, encoding, b"surrogateescape")             # <<<<<<<<<<<<<<
 * 
 * cdef inline str decode_bytes(object data, char* encoding):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyUnicode_AsEncodedString(__pyx_v_obj, __pyx_v_encoding, ((char *)"surrogateescape")); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 36, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyBytes_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("bytes", __pyx_t_1))) __PYX_ERR(5, 36, __pyx_L1_error)
  __pyx_r = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "utils.pxd":33
 * 
 * # Utils: encode / decode
 * cdef inline bytes encode_str(object obj, char* encoding):             # <<<<<<<<<<<<<<
 *     """Encode string to bytes using the 'encoding' with
 *     'surrogateescape' error handling `<'bytes'>`."""
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("serializor.utils.encode_str", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "utils.pxd":38
 *     return PyUnicode_AsEncodedString(obj, encoding, b"surrogateescape")
 * 
 * cdef inline str decode_bytes(object data, char* encoding):             # <<<<<<<<<<<<<<
 *     """Decode bytes to string using the 'encoding' with
 *     "surrogateescape" error handling `<'str'>`."""
 */

static CYTHON_INLINE PyObject *__pyx_f_10serializor_5utils_decode_bytes(PyObject *__pyx_v_data, char *__pyx_v_encoding) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  char *__pyx_t_1;
  Py_ssize_t __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("decode_bytes", 1);

  /* "utils.pxd":41
 *     """Decode bytes to string using the 'encoding' with
 *     "surrogateescape" error handling `<'str'>`."""
 *     return PyUnicode_Decode(bytes_to_chars(data), bytes_len(data), encoding, b"surrogateescape")             # <<<<<<<<<<<<<<
 * 
 * cdef inline str decode_bytes_utf8(object data):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyBytes_AsString(__pyx_v_data); if (unlikely(__pyx_t_1 == ((char *)NULL))) __PYX_ERR(5, 41, __pyx_L1_error)
  __pyx_t_2 = PyBytes_Size(__pyx_v_data); if (unlikely(__pyx_t_2 == ((Py_ssize_t)-1L))) __PYX_ERR(5, 41, __pyx_L1_error)
  __pyx_t_3 = PyUnicode_Decode(__pyx_t_1, __pyx_t_2, __pyx_v_encoding, ((char *)"surrogateescape")); if (unlikely(!__pyx_t_3)) __PYX_ERR(5, 41, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (!(likely(PyUnicode_CheckExact(__pyx_t_3))||((__pyx_t_3) == Py_None) || __Pyx_RaiseUnexpectedTypeError("unicode", __pyx_t_3))) __PYX_ERR(5, 41, __pyx_L1_error)
  __pyx_r = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "utils.pxd":38
 *     return PyUnicode_AsEncodedString(obj, encoding, b"surrogateescape")
 * 
 * cdef inline str decode_bytes(object data, char* encoding):             # <<<<<<<<<<<<<<
 *     """Decode bytes to string using the 'encoding' with
 *     "surrogateescape" error handling `<'str'>`."""
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("serializor.utils.decode_bytes", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "utils.pxd":43
 *     return PyUnicode_Decode(bytes_to_chars(data), bytes_len(data), encoding, b"surrogateescape")
 * 
 * cdef inline str decode_bytes_utf8(object data):             # <<<<<<<<<<<<<<
 *     """Decode bytes to string using 'utf-8' encoding with
 *     'surrogateescape' error handling `<'str'>`."""
 */

static CYTHON_INLINE PyObject *__pyx_f_10serializor_5utils_decode_bytes_utf8(PyObject *__pyx_v_data) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  char *__pyx_t_1;
  Py_ssize_t __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("decode_bytes_utf8", 1);

  /* "utils.pxd":46
 *     """Decode bytes to string using 'utf-8' encoding with
 *     'surrogateescape' error handling `<'str'>`."""
 *     return PyUnicode_DecodeUTF8(bytes_to_chars(data), bytes_len(data), b"surrogateescape")             # <<<<<<<<<<<<<<
 * 
 * cdef inline str decode_bytes_ascii(object data):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyBytes_AsString(__pyx_v_data); if (unlikely(__pyx_t_1 == ((char *)NULL))) __PYX_ERR(5, 46, __pyx_L1_error)
  __pyx_t_2 = PyBytes_Size(__pyx_v_data); if (unlikely(__pyx_t_2 == ((Py_ssize_t)-1L))) __PYX_ERR(5, 46, __pyx_L1_error)
  __pyx_t_3 = PyUnicode_DecodeUTF8(__pyx_t_1, __pyx_t_2, ((char *)"surrogateescape")); if (unlikely(!__pyx_t_3)) __PYX_ERR(5, 46, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_r = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "utils.pxd":43
 *     return PyUnicode_Decode(bytes_to_chars(data), bytes_len(data), encoding, b"surrogateescape")
 * 
 * cdef inline str decode_bytes_utf8(object data):             # <<<<<<<<<<<<<<
 *     """Decode bytes to string using 'utf-8' encoding with
 *     'surrogateescape' error handling `<'str'>`."""
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("serializor.utils.decode_bytes_utf8", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "utils.pxd":48
 *     return PyUnicode_DecodeUTF8(bytes_to_chars(data), bytes_len(data), b"surrogateescape")
 * 
 * cdef inline str decode_bytes_ascii(object data):             # <<<<<<<<<<<<<<
 *     """Decode bytes to string using 'ascii' encoding with
 *     'surrogateescape' error handling `<'str'>`."""
 */

static CYTHON_INLINE PyObject *__pyx_f_10serializor_5utils_decode_bytes_ascii(PyObject *__pyx_v_data) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  char *__pyx_t_1;
  Py_ssize_t __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("decode_bytes_ascii", 1);

  /* "utils.pxd":51
 *     """Decode bytes to string using 'ascii' encoding with
 *     'surrogateescape' error handling `<'str'>`."""
 *     return PyUnicode_DecodeASCII(bytes_to_chars(data), bytes_len(data), b"surrogateescape")             # <<<<<<<<<<<<<<
 * 
 * # NumPy: ndarray get item
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyBytes_AsString(__pyx_v_data); if (unlikely(__pyx_t_1 == ((char *)NULL))) __PYX_ERR(5, 51, __pyx_L1_error)
  __pyx_t_2 = PyBytes_Size(__pyx_v_data); if (unlikely(__pyx_t_2 == ((Py_ssize_t)-1L))) __PYX_ERR(5, 51, __pyx_L1_error)
  __pyx_t_3 = PyUnicode_DecodeASCII(__pyx_t_1, __pyx_t_2, ((char *)"surrogateescape")); if (unlikely(!__pyx_t_3)) __PYX_ERR(5, 51, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_r = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "utils.pxd":48
 *     return PyUnicode_DecodeUTF8(bytes_to_chars(data), bytes_len(data), b"surrogateescape")
 * 
 * cdef inline str decode_bytes_ascii(object data):             # <<<<<<<<<<<<<<
 *     """Decode bytes to string using 'ascii' encoding with
 *     'surrogateescape' error handling `<'str'>`."""
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("serializor.utils.decode_bytes_ascii", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "utils.pxd":54
 * 
 * # NumPy: ndarray get item
 * cdef inline object arr_getitem_1d(np.ndarray arr, np.npy_intp i):             # <<<<<<<<<<<<<<
 *     """Get item from 1-dimensional numpy ndarray as `<'object'>`."""
 *     cdef void* itemptr = <void*>PyArray_GETPTR1(arr, i)
 */

static CYTHON_INLINE PyObject *__pyx_f_10serializor_5utils_arr_getitem_1d(PyArrayObject *__pyx_v_arr, npy_intp __pyx_v_i) {
  void *__pyx_v_itemptr;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("arr_getitem_1d", 1);

  /* "utils.pxd":56
 * cdef inline object arr_getitem_1d(np.ndarray arr, np.npy_intp i):
 *     """Get item from 1-dimensional numpy ndarray as `<'object'>`."""
 *     cdef void* itemptr = <void*>PyArray_GETPTR1(arr, i)             # <<<<<<<<<<<<<<
 *     return PyArray_GETITEM(arr, itemptr)
 * 
 */
  __pyx_v_itemptr = ((void *)PyArray_GETPTR1(__pyx_v_arr, __pyx_v_i));

  /* "utils.pxd":57
 *     """Get item from 1-dimensional numpy ndarray as `<'object'>`."""
 *     cdef void* itemptr = <void*>PyArray_GETPTR1(arr, i)
 *     return PyArray_GETITEM(arr, itemptr)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object arr_getitem_2d(np.ndarray arr, np.npy_intp i, np.npy_intp j):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_GETITEM(__pyx_v_arr, __pyx_v_itemptr); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 57, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "utils.pxd":54
 * 
 * # NumPy: ndarray get item
 * cdef inline object arr_getitem_1d(np.ndarray arr, np.npy_intp i):             # <<<<<<<<<<<<<<
 *     """Get item from 1-dimensional numpy ndarray as `<'object'>`."""
 *     cdef void* itemptr = <void*>PyArray_GETPTR1(arr, i)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("serializor.utils.arr_getitem_1d", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "utils.pxd":59
 *     return PyArray_GETITEM(arr, itemptr)
 * 
 * cdef inline object arr_getitem_2d(np.ndarray arr, np.npy_intp i, np.npy_intp j):             # <<<<<<<<<<<<<<
 *     """Get item from 2-dimensional numpy ndarray as `<'object'>`."""
 *     cdef void* itemptr = <void*>PyArray_GETPTR2(arr, i, j)
 */

static CYTHON_INLINE PyObject *__pyx_f_10serializor_5utils_arr_getitem_2d(PyArrayObject *__pyx_v_arr, npy_intp __pyx_v_i, npy_intp __pyx_v_j) {
  void *__pyx_v_itemptr;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("arr_getitem_2d", 1);

  /* "utils.pxd":61
 * cdef inline object arr_getitem_2d(np.ndarray arr, np.npy_intp i, np.npy_intp j):
 *     """Get item from 2-dimensional numpy ndarray as `<'object'>`."""
 *     cdef void* itemptr = <void*>PyArray_GETPTR2(arr, i, j)             # <<<<<<<<<<<<<<
 *     return PyArray_GETITEM(arr, itemptr)
 * 
 */
  __pyx_v_itemptr = ((void *)PyArray_GETPTR2(__pyx_v_arr, __pyx_v_i, __pyx_v_j));

  /* "utils.pxd":62
 *     """Get item from 2-dimensional numpy ndarray as `<'object'>`."""
 *     cdef void* itemptr = <void*>PyArray_GETPTR2(arr, i, j)
 *     return PyArray_GETITEM(arr, itemptr)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object arr_getitem_3d(np.ndarray arr, np.npy_intp i, np.npy_intp j, np.npy_intp k):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_GETITEM(__pyx_v_arr, __pyx_v_itemptr); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 62, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "utils.pxd":59
 *     return PyArray_GETITEM(arr, itemptr)
 * 
 * cdef inline object arr_getitem_2d(np.ndarray arr, np.npy_intp i, np.npy_intp j):             # <<<<<<<<<<<<<<
 *     """Get item from 2-dimensional numpy ndarray as `<'object'>`."""
 *     cdef void* itemptr = <void*>PyArray_GETPTR2(arr, i, j)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("serializor.utils.arr_getitem_2d", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "utils.pxd":64
 *     return PyArray_GETITEM(arr, itemptr)
 * 
 * cdef inline object arr_getitem_3d(np.ndarray arr, np.npy_intp i, np.npy_intp j, np.npy_intp k):             # <<<<<<<<<<<<<<
 *     """Get item from 3-dimensional numpy ndarray as `<'object'>`."""
 *     cdef void* itemptr = <void*>PyArray_GETPTR3(arr, i, j, k)
 */

static CYTHON_INLINE PyObject *__pyx_f_10serializor_5utils_arr_getitem_3d(PyArrayObject *__pyx_v_arr, npy_intp __pyx_v_i, npy_intp __pyx_v_j, npy_intp __pyx_v_k) {
  void *__pyx_v_itemptr;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("arr_getitem_3d", 1);

  /* "utils.pxd":66
 * cdef inline object arr_getitem_3d(np.ndarray arr, np.npy_intp i, np.npy_intp j, np.npy_intp k):
 *     """Get item from 3-dimensional numpy ndarray as `<'object'>`."""
 *     cdef void* itemptr = <void*>PyArray_GETPTR3(arr, i, j, k)             # <<<<<<<<<<<<<<
 *     return PyArray_GETITEM(arr, itemptr)
 * 
 */
  __pyx_v_itemptr = ((void *)PyArray_GETPTR3(__pyx_v_arr, __pyx_v_i, __pyx_v_j, __pyx_v_k));

  /* "utils.pxd":67
 *     """Get item from 3-dimensional numpy ndarray as `<'object'>`."""
 *     cdef void* itemptr = <void*>PyArray_GETPTR3(arr, i, j, k)
 *     return PyArray_GETITEM(arr, itemptr)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object arr_getitem_4d(np.ndarray arr, np.npy_intp i, np.npy_intp j, np.npy_intp k, np.npy_intp l):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_GETITEM(__pyx_v_arr, __pyx_v_itemptr); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 67, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "utils.pxd":64
 *     return PyArray_GETITEM(arr, itemptr)
 * 
 * cdef inline object arr_getitem_3d(np.ndarray arr, np.npy_intp i, np.npy_intp j, np.npy_intp k):             # <<<<<<<<<<<<<<
 *     """Get item from 3-dimensional numpy ndarray as `<'object'>`."""
 *     cdef void* itemptr = <void*>PyArray_GETPTR3(arr, i, j, k)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("serializor.utils.arr_getitem_3d", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "utils.pxd":69
 *     return PyArray_GETITEM(arr, itemptr)
 * 
 * cdef inline object arr_getitem_4d(np.ndarray arr, np.npy_intp i, np.npy_intp j, np.npy_intp k, np.npy_intp l):             # <<<<<<<<<<<<<<
 *     """Get item from 4-dimensional numpy ndarray as `<'object'>`."""
 *     cdef void* itemptr = <void*>PyArray_GETPTR4(arr, i, j, k, l)
 */

static CYTHON_INLINE PyObject *__pyx_f_10serializor_5utils_arr_getitem_4d(PyArrayObject *__pyx_v_arr, npy_intp __pyx_v_i, npy_intp __pyx_v_j, npy_intp __pyx_v_k, npy_intp __pyx_v_l) {
  void *__pyx_v_itemptr;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("arr_getitem_4d", 1);

  /* "utils.pxd":71
 * cdef inline object arr_getitem_4d(np.ndarray arr, np.npy_intp i, np.npy_intp j, np.npy_intp k, np.npy_intp l):
 *     """Get item from 4-dimensional numpy ndarray as `<'object'>`."""
 *     cdef void* itemptr = <void*>PyArray_GETPTR4(arr, i, j, k, l)             # <<<<<<<<<<<<<<
 *     return PyArray_GETITEM(arr, itemptr)
 * 
 */
  __pyx_v_itemptr = ((void *)PyArray_GETPTR4(__pyx_v_arr, __pyx_v_i, __pyx_v_j, __pyx_v_k, __pyx_v_l));

  /* "utils.pxd":72
 *     """Get item from 4-dimensional numpy ndarray as `<'object'>`."""
 *     cdef void* itemptr = <void*>PyArray_GETPTR4(arr, i, j, k, l)
 *     return PyArray_GETITEM(arr, itemptr)             # <<<<<<<<<<<<<<
 * 
 * # NumPy: ndarray set item
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_GETITEM(__pyx_v_arr, __pyx_v_itemptr); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 72, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "utils.pxd":69
 *     return PyArray_GETITEM(arr, itemptr)
 * 
 * cdef inline object arr_getitem_4d(np.ndarray arr, np.npy_intp i, np.npy_intp j, np.npy_intp k, np.npy_intp l):             # <<<<<<<<<<<<<<
 *     """Get item from 4-dimensional numpy ndarray as `<'object'>`."""
 *     cdef void* itemptr = <void*>PyArray_GETPTR4(arr, i, j, k, l)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("serializor.utils.arr_getitem_4d", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "utils.pxd":75
 * 
 * # NumPy: ndarray set item
 * cdef inline bint arr_setitem_1d(np.ndarray arr, object item, np.npy_intp i) except -1:             # <<<<<<<<<<<<<<
 *     """Set item for 1-dimensional numpy ndarray."""
 *     cdef void* itemptr = <void*>PyArray_GETPTR1(arr, i)
 */

static CYTHON_INLINE int __pyx_f_10serializor_5utils_arr_setitem_1d(PyArrayObject *__pyx_v_arr, PyObject *__pyx_v_item, npy_intp __pyx_v_i) {
  void *__pyx_v_itemptr;
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "utils.pxd":77
 * cdef inline bint arr_setitem_1d(np.ndarray arr, object item, np.npy_intp i) except -1:
 *     """Set item for 1-dimensional numpy ndarray."""
 *     cdef void* itemptr = <void*>PyArray_GETPTR1(arr, i)             # <<<<<<<<<<<<<<
 *     return PyArray_SETITEM(arr, itemptr, item)
 * 
 */
  __pyx_v_itemptr = ((void *)PyArray_GETPTR1(__pyx_v_arr, __pyx_v_i));

  /* "utils.pxd":78
 *     """Set item for 1-dimensional numpy ndarray."""
 *     cdef void* itemptr = <void*>PyArray_GETPTR1(arr, i)
 *     return PyArray_SETITEM(arr, itemptr, item)             # <<<<<<<<<<<<<<
 * 
 * cdef inline bint arr_setitem_2d(np.ndarray arr, object item, np.npy_intp i, np.npy_intp j) except -1:
 */
  __pyx_t_1 = PyArray_SETITEM(__pyx_v_arr, __pyx_v_itemptr, __pyx_v_item); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(5, 78, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "utils.pxd":75
 * 
 * # NumPy: ndarray set item
 * cdef inline bint arr_setitem_1d(np.ndarray arr, object item, np.npy_intp i) except -1:             # <<<<<<<<<<<<<<
 *     """Set item for 1-dimensional numpy ndarray."""
 *     cdef void* itemptr = <void*>PyArray_GETPTR1(arr, i)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("serializor.utils.arr_setitem_1d", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "utils.pxd":80
 *     return PyArray_SETITEM(arr, itemptr, item)
 * 
 * cdef inline bint arr_setitem_2d(np.ndarray arr, object item, np.npy_intp i, np.npy_intp j) except -1:             # <<<<<<<<<<<<<<
 *     """Set item for 2-dimensional numpy ndarray."""
 *     cdef void* itemptr = <void*>PyArray_GETPTR2(arr, i, j)
 */

static CYTHON_INLINE int __pyx_f_10serializor_5utils_arr_setitem_2d(PyArrayObject *__pyx_v_arr, PyObject *__pyx_v_item, npy_intp __pyx_v_i, npy_intp __pyx_v_j) {
  void *__pyx_v_itemptr;
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "utils.pxd":82
 * cdef inline bint arr_setitem_2d(np.ndarray arr, object item, np.npy_intp i, np.npy_intp j) except -1:
 *     """Set item for 2-dimensional numpy ndarray."""
 *     cdef void* itemptr = <void*>PyArray_GETPTR2(arr, i, j)             # <<<<<<<<<<<<<<
 *     return PyArray_SETITEM(arr, itemptr, item)
 * 
 */
  __pyx_v_itemptr = ((void *)PyArray_GETPTR2(__pyx_v_arr, __pyx_v_i, __pyx_v_j));

  /* "utils.pxd":83
 *     """Set item for 2-dimensional numpy ndarray."""
 *     cdef void* itemptr = <void*>PyArray_GETPTR2(arr, i, j)
 *     return PyArray_SETITEM(arr, itemptr, item)             # <<<<<<<<<<<<<<
 * 
 * cdef inline bint arr_setitem_3d(np.ndarray arr, object item, np.npy_intp i, np.npy_intp j, np.npy_intp k) except -1:
 */
  __pyx_t_1 = PyArray_SETITEM(__pyx_v_arr, __pyx_v_itemptr, __pyx_v_item); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(5, 83, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "utils.pxd":80
 *     return PyArray_SETITEM(arr, itemptr, item)
 * 
 * cdef inline bint arr_setitem_2d(np.ndarray arr, object item, np.npy_intp i, np.npy_intp j) except -1:             # <<<<<<<<<<<<<<
 *     """Set item for 2-dimensional numpy ndarray."""
 *     cdef void* itemptr = <void*>PyArray_GETPTR2(arr, i, j)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("serializor.utils.arr_setitem_2d", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "utils.pxd":85
 *     return PyArray_SETITEM(arr, itemptr, item)
 * 
 * cdef inline bint arr_setitem_3d(np.ndarray arr, object item, np.npy_intp i, np.npy_intp j, np.npy_intp k) except -1:             # <<<<<<<<<<<<<<
 *     """Set item for 3-dimensional numpy ndarray."""
 *     cdef void* itemptr = <void*>PyArray_GETPTR3(arr, i, j, k)
 */

static CYTHON_INLINE int __pyx_f_10serializor_5utils_arr_setitem_3d(PyArrayObject *__pyx_v_arr, PyObject *__pyx_v_item, npy_intp __pyx_v_i, npy_intp __pyx_v_j, npy_intp __pyx_v_k) {
  void *__pyx_v_itemptr;
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "utils.pxd":87
 * cdef inline bint arr_setitem_3d(np.ndarray arr, object item, np.npy_intp i, np.npy_intp j, np.npy_intp k) except -1:
 *     """Set item for 3-dimensional numpy ndarray."""
 *     cdef void* itemptr = <void*>PyArray_GETPTR3(arr, i, j, k)             # <<<<<<<<<<<<<<
 *     return PyArray_SETITEM(arr, itemptr, item)
 * 
 */
  __pyx_v_itemptr = ((void *)PyArray_GETPTR3(__pyx_v_arr, __pyx_v_i, __pyx_v_j, __pyx_v_k));

  /* "utils.pxd":88
 *     """Set item for 3-dimensional numpy ndarray."""
 *     cdef void* itemptr = <void*>PyArray_GETPTR3(arr, i, j, k)
 *     return PyArray_SETITEM(arr, itemptr, item)             # <<<<<<<<<<<<<<
 * 
 * cdef inline bint arr_setitem_4d(np.ndarray arr, object item, np.npy_intp i, np.npy_intp j, np.npy_intp k, np.npy_intp l) except -1:
 */
  __pyx_t_1 = PyArray_SETITEM(__pyx_v_arr, __pyx_v_itemptr, __pyx_v_item); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(5, 88, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "utils.pxd":85
 *     return PyArray_SETITEM(arr, itemptr, item)
 * 
 * cdef inline bint arr_setitem_3d(np.ndarray arr, object item, np.npy_intp i, np.npy_intp j, np.npy_intp k) except -1:             # <<<<<<<<<<<<<<
 *     """Set item for 3-dimensional numpy ndarray."""
 *     cdef void* itemptr = <void*>PyArray_GETPTR3(arr, i, j, k)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("serializor.utils.arr_setitem_3d", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "utils.pxd":90
 *     return PyArray_SETITEM(arr, itemptr, item)
 * 
 * cdef inline bint arr_setitem_4d(np.ndarray arr, object item, np.npy_intp i, np.npy_intp j, np.npy_intp k, np.npy_intp l) except -1:             # <<<<<<<<<<<<<<
 *     """Set item for 4-dimensional numpy ndarray."""
 *     cdef void* itemptr = <void*>PyArray_GETPTR4(arr, i, j, k, l)
 */

static CYTHON_INLINE int __pyx_f_10serializor_5utils_arr_setitem_4d(PyArrayObject *__pyx_v_arr, PyObject *__pyx_v_item, npy_intp __pyx_v_i, npy_intp __pyx_v_j, npy_intp __pyx_v_k, npy_intp __pyx_v_l) {
  void *__pyx_v_itemptr;
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "utils.pxd":92
 * cdef inline bint arr_setitem_4d(np.ndarray arr, object item, np.npy_intp i, np.npy_intp j, np.npy_intp k, np.npy_intp l) except -1:
 *     """Set item for 4-dimensional numpy ndarray."""
 *     cdef void* itemptr = <void*>PyArray_GETPTR4(arr, i, j, k, l)             # <<<<<<<<<<<<<<
 *     return PyArray_SETITEM(arr, itemptr, item)
 * 
 */
  __pyx_v_itemptr = ((void *)PyArray_GETPTR4(__pyx_v_arr, __pyx_v_i, __pyx_v_j, __pyx_v_k, __pyx_v_l));

  /* "utils.pxd":93
 *     """Set item for 4-dimensional numpy ndarray."""
 *     cdef void* itemptr = <void*>PyArray_GETPTR4(arr, i, j, k, l)
 *     return PyArray_SETITEM(arr, itemptr, item)             # <<<<<<<<<<<<<<
 * 
 * # NumPy: ndarray flatten
 */
  __pyx_t_1 = PyArray_SETITEM(__pyx_v_arr, __pyx_v_itemptr, __pyx_v_item); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(5, 93, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "utils.pxd":90
 *     return PyArray_SETITEM(arr, itemptr, item)
 * 
 * cdef inline bint arr_setitem_4d(np.ndarray arr, object item, np.npy_intp i, np.npy_intp j, np.npy_intp k, np.npy_intp l) except -1:             # <<<<<<<<<<<<<<
 *     """Set item for 4-dimensional numpy ndarray."""
 *     cdef void* itemptr = <void*>PyArray_GETPTR4(arr, i, j, k, l)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("serializor.utils.arr_setitem_4d", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "utils.pxd":96
 * 
 * # NumPy: ndarray flatten
 * cdef inline np.ndarray arr_flatten(np.ndarray arr):             # <<<<<<<<<<<<<<
 *     """Flatten multi-dimensional numpy ndarray into
 *     1-dimension with 'C-ORDER' `<'np.ndarray'>`."""
 */

static CYTHON_INLINE PyArrayObject *__pyx_f_10serializor_5utils_arr_flatten(PyArrayObject *__pyx_v_arr) {
  PyArrayObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("arr_flatten", 1);

  /* "utils.pxd":99
 *     """Flatten multi-dimensional numpy ndarray into
 *     1-dimension with 'C-ORDER' `<'np.ndarray'>`."""
 *     return PyArray_Flatten(arr, NPY_ORDER.NPY_CORDER)             # <<<<<<<<<<<<<<
 * 
 * # NumPy: nptime unit
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __pyx_t_1 = PyArray_Flatten(__pyx_v_arr, NPY_CORDER); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 99, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(5, 99, __pyx_L1_error)
  __pyx_r = ((PyArrayObject *)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "utils.pxd":96
 * 
 * # NumPy: ndarray flatten
 * cdef inline np.ndarray arr_flatten(np.ndarray arr):             # <<<<<<<<<<<<<<
 *     """Flatten multi-dimensional numpy ndarray into
 *     1-dimension with 'C-ORDER' `<'np.ndarray'>`."""
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("serializor.utils.arr_flatten", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "utils.pxd":102
 * 
 * # NumPy: nptime unit
 * cdef inline str map_nptime_unit_int2str(int unit):             # <<<<<<<<<<<<<<
 *     """Map numpy datetime64/timedelta64 unit from integer
 *     to the corresponding string representation `<'str'>`."""
 */

static CYTHON_INLINE PyObject *__pyx_f_10serializor_5utils_map_nptime_unit_int2str(int __pyx_v_unit) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("map_nptime_unit_int2str", 1);

  /* "utils.pxd":106
 *     to the corresponding string representation `<'str'>`."""
 *     # Common units
 *     if unit == np.NPY_DATETIMEUNIT.NPY_FR_ns:             # <<<<<<<<<<<<<<
 *         return "ns"
 *     if unit == np.NPY_DATETIMEUNIT.NPY_FR_us:
 */
  __pyx_t_1 = (__pyx_v_unit == NPY_FR_ns);
  if (__pyx_t_1) {

    /* "utils.pxd":107
 *     # Common units
 *     if unit == np.NPY_DATETIMEUNIT.NPY_FR_ns:
 *         return "ns"             # <<<<<<<<<<<<<<
 *     if unit == np.NPY_DATETIMEUNIT.NPY_FR_us:
 *         return "us"
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_n_u_ns);
    __pyx_r = __pyx_n_u_ns;
    goto __pyx_L0;

    /* "utils.pxd":106
 *     to the corresponding string representation `<'str'>`."""
 *     # Common units
 *     if unit == np.NPY_DATETIMEUNIT.NPY_FR_ns:             # <<<<<<<<<<<<<<
 *         return "ns"
 *     if unit == np.NPY_DATETIMEUNIT.NPY_FR_us:
 */
  }

  /* "utils.pxd":108
 *     if unit == np.NPY_DATETIMEUNIT.NPY_FR_ns:
 *         return "ns"
 *     if unit == np.NPY_DATETIMEUNIT.NPY_FR_us:             # <<<<<<<<<<<<<<
 *         return "us"
 *     if unit == np.NPY_DATETIMEUNIT.NPY_FR_ms:
 */
  __pyx_t_1 = (__pyx_v_unit == NPY_FR_us);
  if (__pyx_t_1) {

    /* "utils.pxd":109
 *         return "ns"
 *     if unit == np.NPY_DATETIMEUNIT.NPY_FR_us:
 *         return "us"             # <<<<<<<<<<<<<<
 *     if unit == np.NPY_DATETIMEUNIT.NPY_FR_ms:
 *         return "ms"
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_n_u_us);
    __pyx_r = __pyx_n_u_us;
    goto __pyx_L0;

    /* "utils.pxd":108
 *     if unit == np.NPY_DATETIMEUNIT.NPY_FR_ns:
 *         return "ns"
 *     if unit == np.NPY_DATETIMEUNIT.NPY_FR_us:             # <<<<<<<<<<<<<<
 *         return "us"
 *     if unit == np.NPY_DATETIMEUNIT.NPY_FR_ms:
 */
  }

  /* "utils.pxd":110
 *     if unit == np.NPY_DATETIMEUNIT.NPY_FR_us:
 *         return "us"
 *     if unit == np.NPY_DATETIMEUNIT.NPY_FR_ms:             # <<<<<<<<<<<<<<
 *         return "ms"
 *     if unit == np.NPY_DATETIMEUNIT.NPY_FR_s:
 */
  __pyx_t_1 = (__pyx_v_unit == NPY_FR_ms);
  if (__pyx_t_1) {

    /* "utils.pxd":111
 *         return "us"
 *     if unit == np.NPY_DATETIMEUNIT.NPY_FR_ms:
 *         return "ms"             # <<<<<<<<<<<<<<
 *     if unit == np.NPY_DATETIMEUNIT.NPY_FR_s:
 *         return "s"
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_n_u_ms);
    __pyx_r = __pyx_n_u_ms;
    goto __pyx_L0;

    /* "utils.pxd":110
 *     if unit == np.NPY_DATETIMEUNIT.NPY_FR_us:
 *         return "us"
 *     if unit == np.NPY_DATETIMEUNIT.NPY_FR_ms:             # <<<<<<<<<<<<<<
 *         return "ms"
 *     if unit == np.NPY_DATETIMEUNIT.NPY_FR_s:
 */
  }

  /* "utils.pxd":112
 *     if unit == np.NPY_DATETIMEUNIT.NPY_FR_ms:
 *         return "ms"
 *     if unit == np.NPY_DATETIMEUNIT.NPY_FR_s:             # <<<<<<<<<<<<<<
 *         return "s"
 *     if unit == np.NPY_DATETIMEUNIT.NPY_FR_m:
 */
  __pyx_t_1 = (__pyx_v_unit == NPY_FR_s);
  if (__pyx_t_1) {

    /* "utils.pxd":113
 *         return "ms"
 *     if unit == np.NPY_DATETIMEUNIT.NPY_FR_s:
 *         return "s"             # <<<<<<<<<<<<<<
 *     if unit == np.NPY_DATETIMEUNIT.NPY_FR_m:
 *         return "m"
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_n_u_s);
    __pyx_r = __pyx_n_u_s;
    goto __pyx_L0;

    /* "utils.pxd":112
 *     if unit == np.NPY_DATETIMEUNIT.NPY_FR_ms:
 *         return "ms"
 *     if unit == np.NPY_DATETIMEUNIT.NPY_FR_s:             # <<<<<<<<<<<<<<
 *         return "s"
 *     if unit == np.NPY_DATETIMEUNIT.NPY_FR_m:
 */
  }

  /* "utils.pxd":114
 *     if unit == np.NPY_DATETIMEUNIT.NPY_FR_s:
 *         return "s"
 *     if unit == np.NPY_DATETIMEUNIT.NPY_FR_m:             # <<<<<<<<<<<<<<
 *         return "m"
 *     if unit == np.NPY_DATETIMEUNIT.NPY_FR_h:
 */
  __pyx_t_1 = (__pyx_v_unit == NPY_FR_m);
  if (__pyx_t_1) {

    /* "utils.pxd":115
 *         return "s"
 *     if unit == np.NPY_DATETIMEUNIT.NPY_FR_m:
 *         return "m"             # <<<<<<<<<<<<<<
 *     if unit == np.NPY_DATETIMEUNIT.NPY_FR_h:
 *         return "h"
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_n_u_m);
    __pyx_r = __pyx_n_u_m;
    goto __pyx_L0;

    /* "utils.pxd":114
 *     if unit == np.NPY_DATETIMEUNIT.NPY_FR_s:
 *         return "s"
 *     if unit == np.NPY_DATETIMEUNIT.NPY_FR_m:             # <<<<<<<<<<<<<<
 *         return "m"
 *     if unit == np.NPY_DATETIMEUNIT.NPY_FR_h:
 */
  }

  /* "utils.pxd":116
 *     if unit == np.NPY_DATETIMEUNIT.NPY_FR_m:
 *         return "m"
 *     if unit == np.NPY_DATETIMEUNIT.NPY_FR_h:             # <<<<<<<<<<<<<<
 *         return "h"
 *     if unit == np.NPY_DATETIMEUNIT.NPY_FR_D:
 */
  __pyx_t_1 = (__pyx_v_unit == NPY_FR_h);
  if (__pyx_t_1) {

    /* "utils.pxd":117
 *         return "m"
 *     if unit == np.NPY_DATETIMEUNIT.NPY_FR_h:
 *         return "h"             # <<<<<<<<<<<<<<
 *     if unit == np.NPY_DATETIMEUNIT.NPY_FR_D:
 *         return "D"
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_n_u_h);
    __pyx_r = __pyx_n_u_h;
    goto __pyx_L0;

    /* "utils.pxd":116
 *     if unit == np.NPY_DATETIMEUNIT.NPY_FR_m:
 *         return "m"
 *     if unit == np.NPY_DATETIMEUNIT.NPY_FR_h:             # <<<<<<<<<<<<<<
 *         return "h"
 *     if unit == np.NPY_DATETIMEUNIT.NPY_FR_D:
 */
  }

  /* "utils.pxd":118
 *     if unit == np.NPY_DATETIMEUNIT.NPY_FR_h:
 *         return "h"
 *     if unit == np.NPY_DATETIMEUNIT.NPY_FR_D:             # <<<<<<<<<<<<<<
 *         return "D"
 * 
 */
  __pyx_t_1 = (__pyx_v_unit == NPY_FR_D);
  if (__pyx_t_1) {

    /* "utils.pxd":119
 *         return "h"
 *     if unit == np.NPY_DATETIMEUNIT.NPY_FR_D:
 *         return "D"             # <<<<<<<<<<<<<<
 * 
 *     # Uncommon units
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_n_u_D);
    __pyx_r = __pyx_n_u_D;
    goto __pyx_L0;

    /* "utils.pxd":118
 *     if unit == np.NPY_DATETIMEUNIT.NPY_FR_h:
 *         return "h"
 *     if unit == np.NPY_DATETIMEUNIT.NPY_FR_D:             # <<<<<<<<<<<<<<
 *         return "D"
 * 
 */
  }

  /* "utils.pxd":122
 * 
 *     # Uncommon units
 *     if unit == np.NPY_DATETIMEUNIT.NPY_FR_Y:             # <<<<<<<<<<<<<<
 *         return "Y"
 *     if unit == np.NPY_DATETIMEUNIT.NPY_FR_M:
 */
  __pyx_t_1 = (__pyx_v_unit == NPY_FR_Y);
  if (__pyx_t_1) {

    /* "utils.pxd":123
 *     # Uncommon units
 *     if unit == np.NPY_DATETIMEUNIT.NPY_FR_Y:
 *         return "Y"             # <<<<<<<<<<<<<<
 *     if unit == np.NPY_DATETIMEUNIT.NPY_FR_M:
 *         return "M"
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_n_u_Y);
    __pyx_r = __pyx_n_u_Y;
    goto __pyx_L0;

    /* "utils.pxd":122
 * 
 *     # Uncommon units
 *     if unit == np.NPY_DATETIMEUNIT.NPY_FR_Y:             # <<<<<<<<<<<<<<
 *         return "Y"
 *     if unit == np.NPY_DATETIMEUNIT.NPY_FR_M:
 */
  }

  /* "utils.pxd":124
 *     if unit == np.NPY_DATETIMEUNIT.NPY_FR_Y:
 *         return "Y"
 *     if unit == np.NPY_DATETIMEUNIT.NPY_FR_M:             # <<<<<<<<<<<<<<
 *         return "M"
 *     if unit == np.NPY_DATETIMEUNIT.NPY_FR_W:
 */
  __pyx_t_1 = (__pyx_v_unit == NPY_FR_M);
  if (__pyx_t_1) {

    /* "utils.pxd":125
 *         return "Y"
 *     if unit == np.NPY_DATETIMEUNIT.NPY_FR_M:
 *         return "M"             # <<<<<<<<<<<<<<
 *     if unit == np.NPY_DATETIMEUNIT.NPY_FR_W:
 *         return "W"
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_n_u_M);
    __pyx_r = __pyx_n_u_M;
    goto __pyx_L0;

    /* "utils.pxd":124
 *     if unit == np.NPY_DATETIMEUNIT.NPY_FR_Y:
 *         return "Y"
 *     if unit == np.NPY_DATETIMEUNIT.NPY_FR_M:             # <<<<<<<<<<<<<<
 *         return "M"
 *     if unit == np.NPY_DATETIMEUNIT.NPY_FR_W:
 */
  }

  /* "utils.pxd":126
 *     if unit == np.NPY_DATETIMEUNIT.NPY_FR_M:
 *         return "M"
 *     if unit == np.NPY_DATETIMEUNIT.NPY_FR_W:             # <<<<<<<<<<<<<<
 *         return "W"
 *     if unit == np.NPY_DATETIMEUNIT.NPY_FR_ps:
 */
  __pyx_t_1 = (__pyx_v_unit == NPY_FR_W);
  if (__pyx_t_1) {

    /* "utils.pxd":127
 *         return "M"
 *     if unit == np.NPY_DATETIMEUNIT.NPY_FR_W:
 *         return "W"             # <<<<<<<<<<<<<<
 *     if unit == np.NPY_DATETIMEUNIT.NPY_FR_ps:
 *         return "ps"
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_n_u_W);
    __pyx_r = __pyx_n_u_W;
    goto __pyx_L0;

    /* "utils.pxd":126
 *     if unit == np.NPY_DATETIMEUNIT.NPY_FR_M:
 *         return "M"
 *     if unit == np.NPY_DATETIMEUNIT.NPY_FR_W:             # <<<<<<<<<<<<<<
 *         return "W"
 *     if unit == np.NPY_DATETIMEUNIT.NPY_FR_ps:
 */
  }

  /* "utils.pxd":128
 *     if unit == np.NPY_DATETIMEUNIT.NPY_FR_W:
 *         return "W"
 *     if unit == np.NPY_DATETIMEUNIT.NPY_FR_ps:             # <<<<<<<<<<<<<<
 *         return "ps"
 *     if unit == np.NPY_DATETIMEUNIT.NPY_FR_fs:
 */
  __pyx_t_1 = (__pyx_v_unit == NPY_FR_ps);
  if (__pyx_t_1) {

    /* "utils.pxd":129
 *         return "W"
 *     if unit == np.NPY_DATETIMEUNIT.NPY_FR_ps:
 *         return "ps"             # <<<<<<<<<<<<<<
 *     if unit == np.NPY_DATETIMEUNIT.NPY_FR_fs:
 *         return "fs"
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_n_u_ps);
    __pyx_r = __pyx_n_u_ps;
    goto __pyx_L0;

    /* "utils.pxd":128
 *     if unit == np.NPY_DATETIMEUNIT.NPY_FR_W:
 *         return "W"
 *     if unit == np.NPY_DATETIMEUNIT.NPY_FR_ps:             # <<<<<<<<<<<<<<
 *         return "ps"
 *     if unit == np.NPY_DATETIMEUNIT.NPY_FR_fs:
 */
  }

  /* "utils.pxd":130
 *     if unit == np.NPY_DATETIMEUNIT.NPY_FR_ps:
 *         return "ps"
 *     if unit == np.NPY_DATETIMEUNIT.NPY_FR_fs:             # <<<<<<<<<<<<<<
 *         return "fs"
 *     if unit == np.NPY_DATETIMEUNIT.NPY_FR_as:
 */
  __pyx_t_1 = (__pyx_v_unit == NPY_FR_fs);
  if (__pyx_t_1) {

    /* "utils.pxd":131
 *         return "ps"
 *     if unit == np.NPY_DATETIMEUNIT.NPY_FR_fs:
 *         return "fs"             # <<<<<<<<<<<<<<
 *     if unit == np.NPY_DATETIMEUNIT.NPY_FR_as:
 *         return "as"
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_n_u_fs);
    __pyx_r = __pyx_n_u_fs;
    goto __pyx_L0;

    /* "utils.pxd":130
 *     if unit == np.NPY_DATETIMEUNIT.NPY_FR_ps:
 *         return "ps"
 *     if unit == np.NPY_DATETIMEUNIT.NPY_FR_fs:             # <<<<<<<<<<<<<<
 *         return "fs"
 *     if unit == np.NPY_DATETIMEUNIT.NPY_FR_as:
 */
  }

  /* "utils.pxd":132
 *     if unit == np.NPY_DATETIMEUNIT.NPY_FR_fs:
 *         return "fs"
 *     if unit == np.NPY_DATETIMEUNIT.NPY_FR_as:             # <<<<<<<<<<<<<<
 *         return "as"
 *     # if unit == np.NPY_DATETIMEUNIT.NPY_FR_B:
 */
  __pyx_t_1 = (__pyx_v_unit == NPY_FR_as);
  if (__pyx_t_1) {

    /* "utils.pxd":133
 *         return "fs"
 *     if unit == np.NPY_DATETIMEUNIT.NPY_FR_as:
 *         return "as"             # <<<<<<<<<<<<<<
 *     # if unit == np.NPY_DATETIMEUNIT.NPY_FR_B:
 *     #     return "B"
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_n_u_as);
    __pyx_r = __pyx_n_u_as;
    goto __pyx_L0;

    /* "utils.pxd":132
 *     if unit == np.NPY_DATETIMEUNIT.NPY_FR_fs:
 *         return "fs"
 *     if unit == np.NPY_DATETIMEUNIT.NPY_FR_as:             # <<<<<<<<<<<<<<
 *         return "as"
 *     # if unit == np.NPY_DATETIMEUNIT.NPY_FR_B:
 */
  }

  /* "utils.pxd":138
 * 
 *     # Unsupported unit
 *     raise ValueError("unsupported numpy time unit: %d." % unit)             # <<<<<<<<<<<<<<
 * 
 * cdef inline int map_nptime_unit_str2int(str unit) except -1:
 */
  __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_v_unit); if (unlikely(!__pyx_t_2)) __PYX_ERR(5, 138, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyUnicode_Format(__pyx_kp_u_unsupported_numpy_time_unit_d, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(5, 138, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(5, 138, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_Raise(__pyx_t_2, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __PYX_ERR(5, 138, __pyx_L1_error)

  /* "utils.pxd":102
 * 
 * # NumPy: nptime unit
 * cdef inline str map_nptime_unit_int2str(int unit):             # <<<<<<<<<<<<<<
 *     """Map numpy datetime64/timedelta64 unit from integer
 *     to the corresponding string representation `<'str'>`."""
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("serializor.utils.map_nptime_unit_int2str", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "utils.pxd":140
 *     raise ValueError("unsupported numpy time unit: %d." % unit)
 * 
 * cdef inline int map_nptime_unit_str2int(str unit) except -1:             # <<<<<<<<<<<<<<
 *     """Map numpy datetime64/timedelta64 unit from string
 *     representation to the corresponding integer `<'int'>`."""
 */

static CYTHON_INLINE int __pyx_f_10serializor_5utils_map_nptime_unit_str2int(PyObject *__pyx_v_unit) {
  Py_ssize_t __pyx_v_size;
  Py_UCS4 __pyx_v_ch;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  Py_UCS4 __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("map_nptime_unit_str2int", 1);

  /* "utils.pxd":144
 *     representation to the corresponding integer `<'int'>`."""
 *     cdef:
 *         Py_ssize_t size = str_len(unit)             # <<<<<<<<<<<<<<
 *         Py_UCS4 ch
 * 
 */
  __pyx_v_size = PyUnicode_GET_LENGTH(__pyx_v_unit);

  /* "utils.pxd":148
 * 
 *     # Unit: 'ns', 'us', 'ms', 'ps', 'fs', 'as'
 *     if size == 2:             # <<<<<<<<<<<<<<
 *         # Common units
 *         ch = str_read(unit, 0)
 */
  switch (__pyx_v_size) {
    case 2:

    /* "utils.pxd":150
 *     if size == 2:
 *         # Common units
 *         ch = str_read(unit, 0)             # <<<<<<<<<<<<<<
 *         if ch == "n":
 *             return np.NPY_DATETIMEUNIT.NPY_FR_ns
 */
    __pyx_t_1 = PyUnicode_ReadChar(__pyx_v_unit, 0); if (unlikely(__pyx_t_1 == ((Py_UCS4)-1))) __PYX_ERR(5, 150, __pyx_L1_error)
    __pyx_v_ch = __pyx_t_1;

    /* "utils.pxd":151
 *         # Common units
 *         ch = str_read(unit, 0)
 *         if ch == "n":             # <<<<<<<<<<<<<<
 *             return np.NPY_DATETIMEUNIT.NPY_FR_ns
 *         if ch == "u":
 */
    __pyx_t_2 = (__pyx_v_ch == 0x6E);
    if (__pyx_t_2) {

      /* "utils.pxd":152
 *         ch = str_read(unit, 0)
 *         if ch == "n":
 *             return np.NPY_DATETIMEUNIT.NPY_FR_ns             # <<<<<<<<<<<<<<
 *         if ch == "u":
 *             return np.NPY_DATETIMEUNIT.NPY_FR_us
 */
      __pyx_r = NPY_FR_ns;
      goto __pyx_L0;

      /* "utils.pxd":151
 *         # Common units
 *         ch = str_read(unit, 0)
 *         if ch == "n":             # <<<<<<<<<<<<<<
 *             return np.NPY_DATETIMEUNIT.NPY_FR_ns
 *         if ch == "u":
 */
    }

    /* "utils.pxd":153
 *         if ch == "n":
 *             return np.NPY_DATETIMEUNIT.NPY_FR_ns
 *         if ch == "u":             # <<<<<<<<<<<<<<
 *             return np.NPY_DATETIMEUNIT.NPY_FR_us
 *         if ch == "m":
 */
    __pyx_t_2 = (__pyx_v_ch == 0x75);
    if (__pyx_t_2) {

      /* "utils.pxd":154
 *             return np.NPY_DATETIMEUNIT.NPY_FR_ns
 *         if ch == "u":
 *             return np.NPY_DATETIMEUNIT.NPY_FR_us             # <<<<<<<<<<<<<<
 *         if ch == "m":
 *             return np.NPY_DATETIMEUNIT.NPY_FR_ms
 */
      __pyx_r = NPY_FR_us;
      goto __pyx_L0;

      /* "utils.pxd":153
 *         if ch == "n":
 *             return np.NPY_DATETIMEUNIT.NPY_FR_ns
 *         if ch == "u":             # <<<<<<<<<<<<<<
 *             return np.NPY_DATETIMEUNIT.NPY_FR_us
 *         if ch == "m":
 */
    }

    /* "utils.pxd":155
 *         if ch == "u":
 *             return np.NPY_DATETIMEUNIT.NPY_FR_us
 *         if ch == "m":             # <<<<<<<<<<<<<<
 *             return np.NPY_DATETIMEUNIT.NPY_FR_ms
 *         # Uncommon units
 */
    __pyx_t_2 = (__pyx_v_ch == 0x6D);
    if (__pyx_t_2) {

      /* "utils.pxd":156
 *             return np.NPY_DATETIMEUNIT.NPY_FR_us
 *         if ch == "m":
 *             return np.NPY_DATETIMEUNIT.NPY_FR_ms             # <<<<<<<<<<<<<<
 *         # Uncommon units
 *         if ch == "p":
 */
      __pyx_r = NPY_FR_ms;
      goto __pyx_L0;

      /* "utils.pxd":155
 *         if ch == "u":
 *             return np.NPY_DATETIMEUNIT.NPY_FR_us
 *         if ch == "m":             # <<<<<<<<<<<<<<
 *             return np.NPY_DATETIMEUNIT.NPY_FR_ms
 *         # Uncommon units
 */
    }

    /* "utils.pxd":158
 *             return np.NPY_DATETIMEUNIT.NPY_FR_ms
 *         # Uncommon units
 *         if ch == "p":             # <<<<<<<<<<<<<<
 *             return np.NPY_DATETIMEUNIT.NPY_FR_ps
 *         if ch == "f":
 */
    __pyx_t_2 = (__pyx_v_ch == 0x70);
    if (__pyx_t_2) {

      /* "utils.pxd":159
 *         # Uncommon units
 *         if ch == "p":
 *             return np.NPY_DATETIMEUNIT.NPY_FR_ps             # <<<<<<<<<<<<<<
 *         if ch == "f":
 *             return np.NPY_DATETIMEUNIT.NPY_FR_fs
 */
      __pyx_r = NPY_FR_ps;
      goto __pyx_L0;

      /* "utils.pxd":158
 *             return np.NPY_DATETIMEUNIT.NPY_FR_ms
 *         # Uncommon units
 *         if ch == "p":             # <<<<<<<<<<<<<<
 *             return np.NPY_DATETIMEUNIT.NPY_FR_ps
 *         if ch == "f":
 */
    }

    /* "utils.pxd":160
 *         if ch == "p":
 *             return np.NPY_DATETIMEUNIT.NPY_FR_ps
 *         if ch == "f":             # <<<<<<<<<<<<<<
 *             return np.NPY_DATETIMEUNIT.NPY_FR_fs
 *         if ch == "a":
 */
    __pyx_t_2 = (__pyx_v_ch == 0x66);
    if (__pyx_t_2) {

      /* "utils.pxd":161
 *             return np.NPY_DATETIMEUNIT.NPY_FR_ps
 *         if ch == "f":
 *             return np.NPY_DATETIMEUNIT.NPY_FR_fs             # <<<<<<<<<<<<<<
 *         if ch == "a":
 *             return np.NPY_DATETIMEUNIT.NPY_FR_as
 */
      __pyx_r = NPY_FR_fs;
      goto __pyx_L0;

      /* "utils.pxd":160
 *         if ch == "p":
 *             return np.NPY_DATETIMEUNIT.NPY_FR_ps
 *         if ch == "f":             # <<<<<<<<<<<<<<
 *             return np.NPY_DATETIMEUNIT.NPY_FR_fs
 *         if ch == "a":
 */
    }

    /* "utils.pxd":162
 *         if ch == "f":
 *             return np.NPY_DATETIMEUNIT.NPY_FR_fs
 *         if ch == "a":             # <<<<<<<<<<<<<<
 *             return np.NPY_DATETIMEUNIT.NPY_FR_as
 * 
 */
    __pyx_t_2 = (__pyx_v_ch == 97);
    if (__pyx_t_2) {

      /* "utils.pxd":163
 *             return np.NPY_DATETIMEUNIT.NPY_FR_fs
 *         if ch == "a":
 *             return np.NPY_DATETIMEUNIT.NPY_FR_as             # <<<<<<<<<<<<<<
 * 
 *     # Unit: 's', 'm', 'h', 'D', 'Y', 'M', 'W', 'B'
 */
      __pyx_r = NPY_FR_as;
      goto __pyx_L0;

      /* "utils.pxd":162
 *         if ch == "f":
 *             return np.NPY_DATETIMEUNIT.NPY_FR_fs
 *         if ch == "a":             # <<<<<<<<<<<<<<
 *             return np.NPY_DATETIMEUNIT.NPY_FR_as
 * 
 */
    }

    /* "utils.pxd":148
 * 
 *     # Unit: 'ns', 'us', 'ms', 'ps', 'fs', 'as'
 *     if size == 2:             # <<<<<<<<<<<<<<
 *         # Common units
 *         ch = str_read(unit, 0)
 */
    break;
    case 1:

    /* "utils.pxd":168
 *     elif size == 1:
 *         # Common units
 *         ch = str_read(unit, 0)             # <<<<<<<<<<<<<<
 *         if ch == "s":
 *             return np.NPY_DATETIMEUNIT.NPY_FR_s
 */
    __pyx_t_1 = PyUnicode_ReadChar(__pyx_v_unit, 0); if (unlikely(__pyx_t_1 == ((Py_UCS4)-1))) __PYX_ERR(5, 168, __pyx_L1_error)
    __pyx_v_ch = __pyx_t_1;

    /* "utils.pxd":169
 *         # Common units
 *         ch = str_read(unit, 0)
 *         if ch == "s":             # <<<<<<<<<<<<<<
 *             return np.NPY_DATETIMEUNIT.NPY_FR_s
 *         if ch == "m":
 */
    __pyx_t_2 = (__pyx_v_ch == 0x73);
    if (__pyx_t_2) {

      /* "utils.pxd":170
 *         ch = str_read(unit, 0)
 *         if ch == "s":
 *             return np.NPY_DATETIMEUNIT.NPY_FR_s             # <<<<<<<<<<<<<<
 *         if ch == "m":
 *             return np.NPY_DATETIMEUNIT.NPY_FR_m
 */
      __pyx_r = NPY_FR_s;
      goto __pyx_L0;

      /* "utils.pxd":169
 *         # Common units
 *         ch = str_read(unit, 0)
 *         if ch == "s":             # <<<<<<<<<<<<<<
 *             return np.NPY_DATETIMEUNIT.NPY_FR_s
 *         if ch == "m":
 */
    }

    /* "utils.pxd":171
 *         if ch == "s":
 *             return np.NPY_DATETIMEUNIT.NPY_FR_s
 *         if ch == "m":             # <<<<<<<<<<<<<<
 *             return np.NPY_DATETIMEUNIT.NPY_FR_m
 *         if ch == "h":
 */
    __pyx_t_2 = (__pyx_v_ch == 0x6D);
    if (__pyx_t_2) {

      /* "utils.pxd":172
 *             return np.NPY_DATETIMEUNIT.NPY_FR_s
 *         if ch == "m":
 *             return np.NPY_DATETIMEUNIT.NPY_FR_m             # <<<<<<<<<<<<<<
 *         if ch == "h":
 *             return np.NPY_DATETIMEUNIT.NPY_FR_h
 */
      __pyx_r = NPY_FR_m;
      goto __pyx_L0;

      /* "utils.pxd":171
 *         if ch == "s":
 *             return np.NPY_DATETIMEUNIT.NPY_FR_s
 *         if ch == "m":             # <<<<<<<<<<<<<<
 *             return np.NPY_DATETIMEUNIT.NPY_FR_m
 *         if ch == "h":
 */
    }

    /* "utils.pxd":173
 *         if ch == "m":
 *             return np.NPY_DATETIMEUNIT.NPY_FR_m
 *         if ch == "h":             # <<<<<<<<<<<<<<
 *             return np.NPY_DATETIMEUNIT.NPY_FR_h
 *         if ch == "D":
 */
    __pyx_t_2 = (__pyx_v_ch == 0x68);
    if (__pyx_t_2) {

      /* "utils.pxd":174
 *             return np.NPY_DATETIMEUNIT.NPY_FR_m
 *         if ch == "h":
 *             return np.NPY_DATETIMEUNIT.NPY_FR_h             # <<<<<<<<<<<<<<
 *         if ch == "D":
 *             return np.NPY_DATETIMEUNIT.NPY_FR_D
 */
      __pyx_r = NPY_FR_h;
      goto __pyx_L0;

      /* "utils.pxd":173
 *         if ch == "m":
 *             return np.NPY_DATETIMEUNIT.NPY_FR_m
 *         if ch == "h":             # <<<<<<<<<<<<<<
 *             return np.NPY_DATETIMEUNIT.NPY_FR_h
 *         if ch == "D":
 */
    }

    /* "utils.pxd":175
 *         if ch == "h":
 *             return np.NPY_DATETIMEUNIT.NPY_FR_h
 *         if ch == "D":             # <<<<<<<<<<<<<<
 *             return np.NPY_DATETIMEUNIT.NPY_FR_D
 *         # Uncommon units
 */
    __pyx_t_2 = (__pyx_v_ch == 68);
    if (__pyx_t_2) {

      /* "utils.pxd":176
 *             return np.NPY_DATETIMEUNIT.NPY_FR_h
 *         if ch == "D":
 *             return np.NPY_DATETIMEUNIT.NPY_FR_D             # <<<<<<<<<<<<<<
 *         # Uncommon units
 *         if ch == "Y":
 */
      __pyx_r = NPY_FR_D;
      goto __pyx_L0;

      /* "utils.pxd":175
 *         if ch == "h":
 *             return np.NPY_DATETIMEUNIT.NPY_FR_h
 *         if ch == "D":             # <<<<<<<<<<<<<<
 *             return np.NPY_DATETIMEUNIT.NPY_FR_D
 *         # Uncommon units
 */
    }

    /* "utils.pxd":178
 *             return np.NPY_DATETIMEUNIT.NPY_FR_D
 *         # Uncommon units
 *         if ch == "Y":             # <<<<<<<<<<<<<<
 *             return np.NPY_DATETIMEUNIT.NPY_FR_Y
 *         if ch == "M":
 */
    __pyx_t_2 = (__pyx_v_ch == 89);
    if (__pyx_t_2) {

      /* "utils.pxd":179
 *         # Uncommon units
 *         if ch == "Y":
 *             return np.NPY_DATETIMEUNIT.NPY_FR_Y             # <<<<<<<<<<<<<<
 *         if ch == "M":
 *             return np.NPY_DATETIMEUNIT.NPY_FR_M
 */
      __pyx_r = NPY_FR_Y;
      goto __pyx_L0;

      /* "utils.pxd":178
 *             return np.NPY_DATETIMEUNIT.NPY_FR_D
 *         # Uncommon units
 *         if ch == "Y":             # <<<<<<<<<<<<<<
 *             return np.NPY_DATETIMEUNIT.NPY_FR_Y
 *         if ch == "M":
 */
    }

    /* "utils.pxd":180
 *         if ch == "Y":
 *             return np.NPY_DATETIMEUNIT.NPY_FR_Y
 *         if ch == "M":             # <<<<<<<<<<<<<<
 *             return np.NPY_DATETIMEUNIT.NPY_FR_M
 *         if ch == "W":
 */
    __pyx_t_2 = (__pyx_v_ch == 77);
    if (__pyx_t_2) {

      /* "utils.pxd":181
 *             return np.NPY_DATETIMEUNIT.NPY_FR_Y
 *         if ch == "M":
 *             return np.NPY_DATETIMEUNIT.NPY_FR_M             # <<<<<<<<<<<<<<
 *         if ch == "W":
 *             return np.NPY_DATETIMEUNIT.NPY_FR_W
 */
      __pyx_r = NPY_FR_M;
      goto __pyx_L0;

      /* "utils.pxd":180
 *         if ch == "Y":
 *             return np.NPY_DATETIMEUNIT.NPY_FR_Y
 *         if ch == "M":             # <<<<<<<<<<<<<<
 *             return np.NPY_DATETIMEUNIT.NPY_FR_M
 *         if ch == "W":
 */
    }

    /* "utils.pxd":182
 *         if ch == "M":
 *             return np.NPY_DATETIMEUNIT.NPY_FR_M
 *         if ch == "W":             # <<<<<<<<<<<<<<
 *             return np.NPY_DATETIMEUNIT.NPY_FR_W
 *         # if ch == "B":
 */
    __pyx_t_2 = (__pyx_v_ch == 87);
    if (__pyx_t_2) {

      /* "utils.pxd":183
 *             return np.NPY_DATETIMEUNIT.NPY_FR_M
 *         if ch == "W":
 *             return np.NPY_DATETIMEUNIT.NPY_FR_W             # <<<<<<<<<<<<<<
 *         # if ch == "B":
 *         #     return np.NPY_DATETIMEUNIT.NPY_FR_B
 */
      __pyx_r = NPY_FR_W;
      goto __pyx_L0;

      /* "utils.pxd":182
 *         if ch == "M":
 *             return np.NPY_DATETIMEUNIT.NPY_FR_M
 *         if ch == "W":             # <<<<<<<<<<<<<<
 *             return np.NPY_DATETIMEUNIT.NPY_FR_W
 *         # if ch == "B":
 */
    }

    /* "utils.pxd":166
 * 
 *     # Unit: 's', 'm', 'h', 'D', 'Y', 'M', 'W', 'B'
 *     elif size == 1:             # <<<<<<<<<<<<<<
 *         # Common units
 *         ch = str_read(unit, 0)
 */
    break;
    default: break;
  }

  /* "utils.pxd":188
 * 
 *     # Unsupported unit
 *     raise ValueError("unsupported numpy time unit: %s." % unit)             # <<<<<<<<<<<<<<
 * 
 * cdef inline int parse_arr_nptime_unit(np.ndarray arr) except -1:
 */
  __pyx_t_3 = PyUnicode_Format(__pyx_kp_u_unsupported_numpy_time_unit_s, __pyx_v_unit); if (unlikely(!__pyx_t_3)) __PYX_ERR(5, 188, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(5, 188, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_Raise(__pyx_t_4, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __PYX_ERR(5, 188, __pyx_L1_error)

  /* "utils.pxd":140
 *     raise ValueError("unsupported numpy time unit: %d." % unit)
 * 
 * cdef inline int map_nptime_unit_str2int(str unit) except -1:             # <<<<<<<<<<<<<<
 *     """Map numpy datetime64/timedelta64 unit from string
 *     representation to the corresponding integer `<'int'>`."""
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("serializor.utils.map_nptime_unit_str2int", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "utils.pxd":190
 *     raise ValueError("unsupported numpy time unit: %s." % unit)
 * 
 * cdef inline int parse_arr_nptime_unit(np.ndarray arr) except -1:             # <<<<<<<<<<<<<<
 *     """Parse ndarray[datetime64/timedelta64] unit from the,
 *     returns the unit in `<'int'>`."""
 */

static CYTHON_INLINE int __pyx_f_10serializor_5utils_parse_arr_nptime_unit(PyArrayObject *__pyx_v_arr) {
  int __pyx_v_dtype;
  PyObject *__pyx_v_dtype_str = 0;
  Py_ssize_t __pyx_v_size;
  PyObject *__pyx_v_err = NULL;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  int __pyx_t_12;
  char const *__pyx_t_13;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("parse_arr_nptime_unit", 1);

  /* "utils.pxd":193
 *     """Parse ndarray[datetime64/timedelta64] unit from the,
 *     returns the unit in `<'int'>`."""
 *     cdef int dtype = np.PyArray_TYPE(arr)             # <<<<<<<<<<<<<<
 *     if dtype not in (np.NPY_TYPES.NPY_DATETIME, np.NPY_TYPES.NPY_TIMEDELTA):
 *         raise TypeError(
 */
  __pyx_v_dtype = PyArray_TYPE(__pyx_v_arr);

  /* "utils.pxd":194
 *     returns the unit in `<'int'>`."""
 *     cdef int dtype = np.PyArray_TYPE(arr)
 *     if dtype not in (np.NPY_TYPES.NPY_DATETIME, np.NPY_TYPES.NPY_TIMEDELTA):             # <<<<<<<<<<<<<<
 *         raise TypeError(
 *             "expects np.ndarray dtype of 'datetime64/"
 */
  switch (__pyx_v_dtype) {
    case NPY_DATETIME:
    case NPY_TIMEDELTA:
    __pyx_t_1 = 0;
    break;
    default:
    __pyx_t_1 = 1;
    break;
  }
  __pyx_t_2 = __pyx_t_1;
  if (unlikely(__pyx_t_2)) {

    /* "utils.pxd":197
 *         raise TypeError(
 *             "expects np.ndarray dtype of 'datetime64/"
 *             "timedelta64', got '%s'." % arr.dtype             # <<<<<<<<<<<<<<
 *         )
 * 
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_arr), __pyx_n_s_dtype); if (unlikely(!__pyx_t_3)) __PYX_ERR(5, 197, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = __Pyx_PyUnicode_FormatSafe(__pyx_kp_u_expects_np_ndarray_dtype_of_date, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(5, 197, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "utils.pxd":195
 *     cdef int dtype = np.PyArray_TYPE(arr)
 *     if dtype not in (np.NPY_TYPES.NPY_DATETIME, np.NPY_TYPES.NPY_TIMEDELTA):
 *         raise TypeError(             # <<<<<<<<<<<<<<
 *             "expects np.ndarray dtype of 'datetime64/"
 *             "timedelta64', got '%s'." % arr.dtype
 */
    __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_builtin_TypeError, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(5, 195, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(5, 195, __pyx_L1_error)

    /* "utils.pxd":194
 *     returns the unit in `<'int'>`."""
 *     cdef int dtype = np.PyArray_TYPE(arr)
 *     if dtype not in (np.NPY_TYPES.NPY_DATETIME, np.NPY_TYPES.NPY_TIMEDELTA):             # <<<<<<<<<<<<<<
 *         raise TypeError(
 *             "expects np.ndarray dtype of 'datetime64/"
 */
  }

  /* "utils.pxd":200
 *         )
 * 
 *     cdef str dtype_str = arr.dtype.str             # <<<<<<<<<<<<<<
 *     cdef Py_ssize_t size = str_len(dtype_str)
 *     if size < 6:
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_arr), __pyx_n_s_dtype); if (unlikely(!__pyx_t_3)) __PYX_ERR(5, 200, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_str); if (unlikely(!__pyx_t_4)) __PYX_ERR(5, 200, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (!(likely(PyUnicode_CheckExact(__pyx_t_4))||((__pyx_t_4) == Py_None) || __Pyx_RaiseUnexpectedTypeError("unicode", __pyx_t_4))) __PYX_ERR(5, 200, __pyx_L1_error)
  __pyx_v_dtype_str = ((PyObject*)__pyx_t_4);
  __pyx_t_4 = 0;

  /* "utils.pxd":201
 * 
 *     cdef str dtype_str = arr.dtype.str
 *     cdef Py_ssize_t size = str_len(dtype_str)             # <<<<<<<<<<<<<<
 *     if size < 6:
 *         raise ValueError("unable to parse ndarray time unit from '%s'." % dtype_str)
 */
  __pyx_v_size = PyUnicode_GET_LENGTH(__pyx_v_dtype_str);

  /* "utils.pxd":202
 *     cdef str dtype_str = arr.dtype.str
 *     cdef Py_ssize_t size = str_len(dtype_str)
 *     if size < 6:             # <<<<<<<<<<<<<<
 *         raise ValueError("unable to parse ndarray time unit from '%s'." % dtype_str)
 *     dtype_str = str_substr(dtype_str, 4, size - 1)
 */
  __pyx_t_2 = (__pyx_v_size < 6);
  if (unlikely(__pyx_t_2)) {

    /* "utils.pxd":203
 *     cdef Py_ssize_t size = str_len(dtype_str)
 *     if size < 6:
 *         raise ValueError("unable to parse ndarray time unit from '%s'." % dtype_str)             # <<<<<<<<<<<<<<
 *     dtype_str = str_substr(dtype_str, 4, size - 1)
 *     try:
 */
    __pyx_t_4 = PyUnicode_Format(__pyx_kp_u_unable_to_parse_ndarray_time_uni, __pyx_v_dtype_str); if (unlikely(!__pyx_t_4)) __PYX_ERR(5, 203, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(5, 203, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(5, 203, __pyx_L1_error)

    /* "utils.pxd":202
 *     cdef str dtype_str = arr.dtype.str
 *     cdef Py_ssize_t size = str_len(dtype_str)
 *     if size < 6:             # <<<<<<<<<<<<<<
 *         raise ValueError("unable to parse ndarray time unit from '%s'." % dtype_str)
 *     dtype_str = str_substr(dtype_str, 4, size - 1)
 */
  }

  /* "utils.pxd":204
 *     if size < 6:
 *         raise ValueError("unable to parse ndarray time unit from '%s'." % dtype_str)
 *     dtype_str = str_substr(dtype_str, 4, size - 1)             # <<<<<<<<<<<<<<
 *     try:
 *         return map_nptime_unit_str2int(dtype_str)
 */
  __pyx_t_3 = PyUnicode_Substring(__pyx_v_dtype_str, 4, (__pyx_v_size - 1)); if (unlikely(!__pyx_t_3)) __PYX_ERR(5, 204, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF_SET(__pyx_v_dtype_str, ((PyObject*)__pyx_t_3));
  __pyx_t_3 = 0;

  /* "utils.pxd":205
 *         raise ValueError("unable to parse ndarray time unit from '%s'." % dtype_str)
 *     dtype_str = str_substr(dtype_str, 4, size - 1)
 *     try:             # <<<<<<<<<<<<<<
 *         return map_nptime_unit_str2int(dtype_str)
 *     except ValueError as err:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_5, &__pyx_t_6, &__pyx_t_7);
    __Pyx_XGOTREF(__pyx_t_5);
    __Pyx_XGOTREF(__pyx_t_6);
    __Pyx_XGOTREF(__pyx_t_7);
    /*try:*/ {

      /* "utils.pxd":206
 *     dtype_str = str_substr(dtype_str, 4, size - 1)
 *     try:
 *         return map_nptime_unit_str2int(dtype_str)             # <<<<<<<<<<<<<<
 *     except ValueError as err:
 *         raise ValueError("unable to parse ndarray time unit from '%s'." % dtype_str) from err
 */
      __pyx_t_8 = __pyx_f_10serializor_5utils_map_nptime_unit_str2int(__pyx_v_dtype_str); if (unlikely(__pyx_t_8 == ((int)-1))) __PYX_ERR(5, 206, __pyx_L5_error)
      __pyx_r = __pyx_t_8;
      goto __pyx_L9_try_return;

      /* "utils.pxd":205
 *         raise ValueError("unable to parse ndarray time unit from '%s'." % dtype_str)
 *     dtype_str = str_substr(dtype_str, 4, size - 1)
 *     try:             # <<<<<<<<<<<<<<
 *         return map_nptime_unit_str2int(dtype_str)
 *     except ValueError as err:
 */
    }
    __pyx_L5_error:;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "utils.pxd":207
 *     try:
 *         return map_nptime_unit_str2int(dtype_str)
 *     except ValueError as err:             # <<<<<<<<<<<<<<
 *         raise ValueError("unable to parse ndarray time unit from '%s'." % dtype_str) from err
 */
    __pyx_t_8 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_ValueError);
    if (__pyx_t_8) {
      __Pyx_AddTraceback("serializor.utils.parse_arr_nptime_unit", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_3, &__pyx_t_4, &__pyx_t_9) < 0) __PYX_ERR(5, 207, __pyx_L7_except_error)
      __Pyx_XGOTREF(__pyx_t_3);
      __Pyx_XGOTREF(__pyx_t_4);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_INCREF(__pyx_t_4);
      __pyx_v_err = __pyx_t_4;
      /*try:*/ {

        /* "utils.pxd":208
 *         return map_nptime_unit_str2int(dtype_str)
 *     except ValueError as err:
 *         raise ValueError("unable to parse ndarray time unit from '%s'." % dtype_str) from err             # <<<<<<<<<<<<<<
 */
        __pyx_t_10 = PyUnicode_Format(__pyx_kp_u_unable_to_parse_ndarray_time_uni, __pyx_v_dtype_str); if (unlikely(!__pyx_t_10)) __PYX_ERR(5, 208, __pyx_L16_error)
        __Pyx_GOTREF(__pyx_t_10);
        __pyx_t_11 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_10); if (unlikely(!__pyx_t_11)) __PYX_ERR(5, 208, __pyx_L16_error)
        __Pyx_GOTREF(__pyx_t_11);
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        __Pyx_Raise(__pyx_t_11, 0, 0, __pyx_v_err);
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        __PYX_ERR(5, 208, __pyx_L16_error)
      }

      /* "utils.pxd":207
 *     try:
 *         return map_nptime_unit_str2int(dtype_str)
 *     except ValueError as err:             # <<<<<<<<<<<<<<
 *         raise ValueError("unable to parse ndarray time unit from '%s'." % dtype_str) from err
 */
      /*finally:*/ {
        __pyx_L16_error:;
        /*exception exit:*/{
          __Pyx_PyThreadState_declare
          __Pyx_PyThreadState_assign
          __pyx_t_14 = 0; __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0;
          __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
          __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
          if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_17, &__pyx_t_18, &__pyx_t_19);
          if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_14, &__pyx_t_15, &__pyx_t_16) < 0)) __Pyx_ErrFetch(&__pyx_t_14, &__pyx_t_15, &__pyx_t_16);
          __Pyx_XGOTREF(__pyx_t_14);
          __Pyx_XGOTREF(__pyx_t_15);
          __Pyx_XGOTREF(__pyx_t_16);
          __Pyx_XGOTREF(__pyx_t_17);
          __Pyx_XGOTREF(__pyx_t_18);
          __Pyx_XGOTREF(__pyx_t_19);
          __pyx_t_8 = __pyx_lineno; __pyx_t_12 = __pyx_clineno; __pyx_t_13 = __pyx_filename;
          {
            __Pyx_DECREF(__pyx_v_err); __pyx_v_err = 0;
          }
          if (PY_MAJOR_VERSION >= 3) {
            __Pyx_XGIVEREF(__pyx_t_17);
            __Pyx_XGIVEREF(__pyx_t_18);
            __Pyx_XGIVEREF(__pyx_t_19);
            __Pyx_ExceptionReset(__pyx_t_17, __pyx_t_18, __pyx_t_19);
          }
          __Pyx_XGIVEREF(__pyx_t_14);
          __Pyx_XGIVEREF(__pyx_t_15);
          __Pyx_XGIVEREF(__pyx_t_16);
          __Pyx_ErrRestore(__pyx_t_14, __pyx_t_15, __pyx_t_16);
          __pyx_t_14 = 0; __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0;
          __pyx_lineno = __pyx_t_8; __pyx_clineno = __pyx_t_12; __pyx_filename = __pyx_t_13;
          goto __pyx_L7_except_error;
        }
      }
    }
    goto __pyx_L7_except_error;

    /* "utils.pxd":205
 *         raise ValueError("unable to parse ndarray time unit from '%s'." % dtype_str)
 *     dtype_str = str_substr(dtype_str, 4, size - 1)
 *     try:             # <<<<<<<<<<<<<<
 *         return map_nptime_unit_str2int(dtype_str)
 *     except ValueError as err:
 */
    __pyx_L7_except_error:;
    __Pyx_XGIVEREF(__pyx_t_5);
    __Pyx_XGIVEREF(__pyx_t_6);
    __Pyx_XGIVEREF(__pyx_t_7);
    __Pyx_ExceptionReset(__pyx_t_5, __pyx_t_6, __pyx_t_7);
    goto __pyx_L1_error;
    __pyx_L9_try_return:;
    __Pyx_XGIVEREF(__pyx_t_5);
    __Pyx_XGIVEREF(__pyx_t_6);
    __Pyx_XGIVEREF(__pyx_t_7);
    __Pyx_ExceptionReset(__pyx_t_5, __pyx_t_6, __pyx_t_7);
    goto __pyx_L0;
  }

  /* "utils.pxd":190
 *     raise ValueError("unsupported numpy time unit: %s." % unit)
 * 
 * cdef inline int parse_arr_nptime_unit(np.ndarray arr) except -1:             # <<<<<<<<<<<<<<
 *     """Parse ndarray[datetime64/timedelta64] unit from the,
 *     returns the unit in `<'int'>`."""
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_AddTraceback("serializor.utils.parse_arr_nptime_unit", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_dtype_str);
  __Pyx_XDECREF(__pyx_v_err);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "serializor/binary/ser.py":99
 * 
 * # Orjson dumps --------------------------------------------------------------------------------
 * @cython.cfunc             # <<<<<<<<<<<<<<
 * @cython.inline(True)
 * def _orjson_dumps(obj: object) -> bytes:
 */

static PyObject *__pyx_f_10serializor_6binary_3ser__orjson_dumps(PyObject *__pyx_v_obj) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  unsigned int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_orjson_dumps", 1);

  /* "serializor/binary/ser.py":106
 *     Based on [orjson](https://github.com/ijl/orjson) `'dumps()'` function.
 *     """
 *     return _dumps(obj)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_dumps); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 106, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_v_obj};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 106, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  if (!(likely(PyBytes_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("bytes", __pyx_t_1))) __PYX_ERR(0, 106, __pyx_L1_error)
  __pyx_r = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "serializor/binary/ser.py":99
 * 
 * # Orjson dumps --------------------------------------------------------------------------------
 * @cython.cfunc             # <<<<<<<<<<<<<<
 * @cython.inline(True)
 * def _orjson_dumps(obj: object) -> bytes:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("serializor.binary.ser._orjson_dumps", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "serializor/binary/ser.py":109
 * 
 * 
 * @cython.cfunc             # <<<<<<<<<<<<<<
 * @cython.inline(True)
 * def _orjson_dumps_numpy(obj: object) -> bytes:
 */

static PyObject *__pyx_f_10serializor_6binary_3ser__orjson_dumps_numpy(PyObject *__pyx_v_obj) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_orjson_dumps_numpy", 1);

  /* "serializor/binary/ser.py":116
 *     Based on [orjson](https://github.com/ijl/orjson) `'dumps()'` function.
 *     """
 *     return _dumps(obj, option=OPT_SERIALIZE_NUMPY)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_dumps); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 116, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 116, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_v_obj);
  __Pyx_GIVEREF(__pyx_v_obj);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_obj)) __PYX_ERR(0, 116, __pyx_L1_error);
  __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 116, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_option, __pyx_v_10serializor_6binary_3ser_OPT_SERIALIZE_NUMPY) < 0) __PYX_ERR(0, 116, __pyx_L1_error)
  __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 116, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (!(likely(PyBytes_CheckExact(__pyx_t_4))||((__pyx_t_4) == Py_None) || __Pyx_RaiseUnexpectedTypeError("bytes", __pyx_t_4))) __PYX_ERR(0, 116, __pyx_L1_error)
  __pyx_r = ((PyObject*)__pyx_t_4);
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "serializor/binary/ser.py":109
 * 
 * 
 * @cython.cfunc             # <<<<<<<<<<<<<<
 * @cython.inline(True)
 * def _orjson_dumps_numpy(obj: object) -> bytes:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("serializor.binary.ser._orjson_dumps_numpy", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "serializor/binary/ser.py":120
 * 
 * # Basic Types ---------------------------------------------------------------------------------
 * @cython.cfunc             # <<<<<<<<<<<<<<
 * @cython.inline(True)
 * def _ser_str(obj: object) -> bytes:
 */

static PyObject *__pyx_f_10serializor_6binary_3ser__ser_str(PyObject *__pyx_v_obj) {
  PyObject *__pyx_v_val = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_ser_str", 1);

  /* "serializor/binary/ser.py":128
 *     >>> b'[id(1B)][size(ENC)][value]'
 *     """
 *     val: bytes = utils.encode_str(obj, b"utf-8")             # <<<<<<<<<<<<<<
 *     return gen_header(identifier.STR, bytes_len(val)) + val  # type: ignore
 * 
 */
  __pyx_t_1 = __pyx_f_10serializor_5utils_encode_str(__pyx_v_obj, ((char *)"utf-8")); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 128, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_val = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "serializor/binary/ser.py":129
 *     """
 *     val: bytes = utils.encode_str(obj, b"utf-8")
 *     return gen_header(identifier.STR, bytes_len(val)) + val  # type: ignore             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = PyBytes_Size(__pyx_v_val); if (unlikely(__pyx_t_2 == ((Py_ssize_t)-1L))) __PYX_ERR(0, 129, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_10serializor_6binary_3ser_gen_header(__pyx_v_10serializor_10identifier_STR, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 129, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = PyNumber_Add(__pyx_t_1, __pyx_v_val); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 129, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "serializor/binary/ser.py":120
 * 
 * # Basic Types ---------------------------------------------------------------------------------
 * @cython.cfunc             # <<<<<<<<<<<<<<
 * @cython.inline(True)
 * def _ser_str(obj: object) -> bytes:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("serializor.binary.ser._ser_str", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_val);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "serializor/binary/ser.py":132
 * 
 * 
 * @cython.cfunc             # <<<<<<<<<<<<<<
 * @cython.inline(True)
 * def _ser_int(obj: object) -> bytes:
 */

static PyObject *__pyx_f_10serializor_6binary_3ser__ser_int(PyObject *__pyx_v_obj) {
  PyObject *__pyx_v_val = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  Py_ssize_t __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_ser_int", 1);

  /* "serializor/binary/ser.py":140
 *     >>> b'[id(1B)][size(ENC)][value]'
 *     """
 *     val: bytes = utils.encode_str(str(obj), b"ascii")             # <<<<<<<<<<<<<<
 *     return gen_header(identifier.INT, bytes_len(val)) + val  # type: ignore
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_Unicode(__pyx_v_obj); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 140, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __pyx_f_10serializor_5utils_encode_str(__pyx_t_1, ((char *)"ascii")); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 140, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_val = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "serializor/binary/ser.py":141
 *     """
 *     val: bytes = utils.encode_str(str(obj), b"ascii")
 *     return gen_header(identifier.INT, bytes_len(val)) + val  # type: ignore             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_3 = PyBytes_Size(__pyx_v_val); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1L))) __PYX_ERR(0, 141, __pyx_L1_error)
  __pyx_t_2 = __pyx_f_10serializor_6binary_3ser_gen_header(__pyx_v_10serializor_10identifier_INT, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 141, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = PyNumber_Add(__pyx_t_2, __pyx_v_val); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 141, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "serializor/binary/ser.py":132
 * 
 * 
 * @cython.cfunc             # <<<<<<<<<<<<<<
 * @cython.inline(True)
 * def _ser_int(obj: object) -> bytes:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("serializor.binary.ser._ser_int", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_val);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "serializor/binary/ser.py":144
 * 
 * 
 * @cython.cfunc             # <<<<<<<<<<<<<<
 * @cython.inline(True)
 * def _ser_float(obj: object) -> bytes:
 */

static PyObject *__pyx_f_10serializor_6binary_3ser__ser_float(PyObject *__pyx_v_obj) {
  PyObject *__pyx_v_val = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  long double __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_ser_float", 1);

  /* "serializor/binary/ser.py":154
 *     # For normal float numbers, orjson performs
 *     # faster than Python built-in `str()` function.
 *     if isnormal(obj):             # <<<<<<<<<<<<<<
 *         val: bytes = _orjson_dumps(obj)
 *         return gen_header(identifier.FLOAT, bytes_len(val)) + val  # type: ignore
 */
  __pyx_t_1 = __pyx_PyFloat_AsDouble(__pyx_v_obj); if (unlikely((__pyx_t_1 == (long double)-1) && PyErr_Occurred())) __PYX_ERR(0, 154, __pyx_L1_error)
  __pyx_t_2 = isnormal(__pyx_t_1);
  if (__pyx_t_2) {

    /* "serializor/binary/ser.py":155
 *     # faster than Python built-in `str()` function.
 *     if isnormal(obj):
 *         val: bytes = _orjson_dumps(obj)             # <<<<<<<<<<<<<<
 *         return gen_header(identifier.FLOAT, bytes_len(val)) + val  # type: ignore
 *     # For other float objects, fallback to Python
 */
    __pyx_t_3 = __pyx_f_10serializor_6binary_3ser__orjson_dumps(__pyx_v_obj); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 155, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_v_val = ((PyObject*)__pyx_t_3);
    __pyx_t_3 = 0;

    /* "serializor/binary/ser.py":156
 *     if isnormal(obj):
 *         val: bytes = _orjson_dumps(obj)
 *         return gen_header(identifier.FLOAT, bytes_len(val)) + val  # type: ignore             # <<<<<<<<<<<<<<
 *     # For other float objects, fallback to Python
 *     # built-in `str()` approach.
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = PyBytes_Size(__pyx_v_val); if (unlikely(__pyx_t_4 == ((Py_ssize_t)-1L))) __PYX_ERR(0, 156, __pyx_L1_error)
    __pyx_t_3 = __pyx_f_10serializor_6binary_3ser_gen_header(__pyx_v_10serializor_10identifier_FLOAT, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 156, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_5 = PyNumber_Add(__pyx_t_3, __pyx_v_val); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 156, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_r = ((PyObject*)__pyx_t_5);
    __pyx_t_5 = 0;
    goto __pyx_L0;

    /* "serializor/binary/ser.py":154
 *     # For normal float numbers, orjson performs
 *     # faster than Python built-in `str()` function.
 *     if isnormal(obj):             # <<<<<<<<<<<<<<
 *         val: bytes = _orjson_dumps(obj)
 *         return gen_header(identifier.FLOAT, bytes_len(val)) + val  # type: ignore
 */
  }

  /* "serializor/binary/ser.py":159
 *     # For other float objects, fallback to Python
 *     # built-in `str()` approach.
 *     return _ser_float64(obj)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_5 = __pyx_f_10serializor_6binary_3ser__ser_float64(__pyx_v_obj); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 159, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_r = ((PyObject*)__pyx_t_5);
  __pyx_t_5 = 0;
  goto __pyx_L0;

  /* "serializor/binary/ser.py":144
 * 
 * 
 * @cython.cfunc             # <<<<<<<<<<<<<<
 * @cython.inline(True)
 * def _ser_float(obj: object) -> bytes:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("serializor.binary.ser._ser_float", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_val);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "serializor/binary/ser.py":162
 * 
 * 
 * @cython.cfunc             # <<<<<<<<<<<<<<
 * @cython.inline(True)
 * def _ser_bool(obj: object) -> bytes:
 */

static PyObject *__pyx_f_10serializor_6binary_3ser__ser_bool(PyObject *__pyx_v_obj) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_ser_bool", 1);

  /* "serializor/binary/ser.py":170
 *     >>> b'[id(1B)][value]'
 *     """
 *     return TRUE_VALUE if bool(obj) else FALSE_VALUE             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_v_obj); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 170, __pyx_L1_error)
  if ((!(!__pyx_t_2))) {
    __Pyx_INCREF(__pyx_v_10serializor_6binary_3ser_TRUE_VALUE);
    __pyx_t_1 = __pyx_v_10serializor_6binary_3ser_TRUE_VALUE;
  } else {
    __Pyx_INCREF(__pyx_v_10serializor_6binary_3ser_FALSE_VALUE);
    __pyx_t_1 = __pyx_v_10serializor_6binary_3ser_FALSE_VALUE;
  }
  __pyx_r = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "serializor/binary/ser.py":162
 * 
 * 
 * @cython.cfunc             # <<<<<<<<<<<<<<
 * @cython.inline(True)
 * def _ser_bool(obj: object) -> bytes:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("serializor.binary.ser._ser_bool", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "serializor/binary/ser.py":173
 * 
 * 
 * @cython.cfunc             # <<<<<<<<<<<<<<
 * @cython.inline(True)
 * def _ser_none(_: object) -> bytes:
 */

static PyObject *__pyx_f_10serializor_6binary_3ser__ser_none(CYTHON_UNUSED PyObject *__pyx_v__) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_ser_none", 1);

  /* "serializor/binary/ser.py":181
 *     >>> b'[id(1B)]'
 *     """
 *     return NONE_VALUE             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_10serializor_6binary_3ser_NONE_VALUE);
  __pyx_r = __pyx_v_10serializor_6binary_3ser_NONE_VALUE;
  goto __pyx_L0;

  /* "serializor/binary/ser.py":173
 * 
 * 
 * @cython.cfunc             # <<<<<<<<<<<<<<
 * @cython.inline(True)
 * def _ser_none(_: object) -> bytes:
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "serializor/binary/ser.py":185
 * 
 * # Date&Time Types -----------------------------------------------------------------------------
 * @cython.cfunc             # <<<<<<<<<<<<<<
 * @cython.inline(True)
 * def _ser_datetime(obj: object) -> bytes:
 */

static PyObject *__pyx_f_10serializor_6binary_3ser__ser_datetime(PyObject *__pyx_v_obj) {
  PyObject *__pyx_v_name = NULL;
  PyObject *__pyx_v_tz = NULL;
  PyObject *__pyx_v_tz_name = NULL;
  PyObject *__pyx_v_offset = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  Py_ssize_t __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  unsigned int __pyx_t_12;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_ser_datetime", 1);

  /* "serializor/binary/ser.py":205
 *     """
 *     # With timezone
 *     if (tz := datetime.datetime_tzinfo(obj)) is not None:             # <<<<<<<<<<<<<<
 *         # . approach 1: timezone name
 *         if set_contains(utils.AVAILABLE_TIMEZONES, tz_name := str(tz)):
 */
  __pyx_t_1 = __pyx_f_7cpython_8datetime_datetime_tzinfo(__pyx_v_obj); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 205, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_tz = __pyx_t_1;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = (__pyx_t_1 != Py_None);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "serializor/binary/ser.py":207
 *     if (tz := datetime.datetime_tzinfo(obj)) is not None:
 *         # . approach 1: timezone name
 *         if set_contains(utils.AVAILABLE_TIMEZONES, tz_name := str(tz)):             # <<<<<<<<<<<<<<
 *             name = utils.encode_str(tz_name, b"ascii")
 *             return b"".join(
 */
    __pyx_t_1 = __pyx_v_10serializor_5utils_AVAILABLE_TIMEZONES;
    __Pyx_INCREF(__pyx_t_1);
    __pyx_t_3 = __Pyx_PyObject_Unicode(__pyx_v_tz); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 207, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_v_tz_name = __pyx_t_3;
    __Pyx_INCREF(__pyx_t_3);
    __pyx_t_4 = __pyx_t_3;
    __Pyx_INCREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_2 = PySet_Contains(__pyx_t_1, __pyx_t_4); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(0, 207, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (__pyx_t_2) {

      /* "serializor/binary/ser.py":208
 *         # . approach 1: timezone name
 *         if set_contains(utils.AVAILABLE_TIMEZONES, tz_name := str(tz)):
 *             name = utils.encode_str(tz_name, b"ascii")             # <<<<<<<<<<<<<<
 *             return b"".join(
 *                 [
 */
      __pyx_t_4 = __pyx_f_10serializor_5utils_encode_str(__pyx_v_tz_name, ((char *)"ascii")); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 208, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_v_name = ((PyObject*)__pyx_t_4);
      __pyx_t_4 = 0;

      /* "serializor/binary/ser.py":209
 *         if set_contains(utils.AVAILABLE_TIMEZONES, tz_name := str(tz)):
 *             name = utils.encode_str(tz_name, b"ascii")
 *             return b"".join(             # <<<<<<<<<<<<<<
 *                 [
 *                     DATETIME_ID,  # identifier
 */
      __Pyx_XDECREF(__pyx_r);

      /* "serializor/binary/ser.py":213
 *                     DATETIME_ID,  # identifier
 *                     b"\x01",  # approach 1
 *                     pack_uint16(datetime.datetime_year(obj)),  # type: ignore             # <<<<<<<<<<<<<<
 *                     pack_uint8(datetime.datetime_month(obj)),  # type: ignore
 *                     pack_uint8(datetime.datetime_day(obj)),  # type: ignore
 */
      __pyx_t_4 = __pyx_f_10serializor_6binary_3ser_pack_uint16(__pyx_f_7cpython_8datetime_datetime_year(__pyx_v_obj)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 213, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);

      /* "serializor/binary/ser.py":214
 *                     b"\x01",  # approach 1
 *                     pack_uint16(datetime.datetime_year(obj)),  # type: ignore
 *                     pack_uint8(datetime.datetime_month(obj)),  # type: ignore             # <<<<<<<<<<<<<<
 *                     pack_uint8(datetime.datetime_day(obj)),  # type: ignore
 *                     pack_uint8(datetime.datetime_hour(obj)),  # type: ignore
 */
      __pyx_t_1 = __pyx_f_10serializor_6binary_3ser_pack_uint8(__pyx_f_7cpython_8datetime_datetime_month(__pyx_v_obj)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 214, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);

      /* "serializor/binary/ser.py":215
 *                     pack_uint16(datetime.datetime_year(obj)),  # type: ignore
 *                     pack_uint8(datetime.datetime_month(obj)),  # type: ignore
 *                     pack_uint8(datetime.datetime_day(obj)),  # type: ignore             # <<<<<<<<<<<<<<
 *                     pack_uint8(datetime.datetime_hour(obj)),  # type: ignore
 *                     pack_uint8(datetime.datetime_minute(obj)),  # type: ignore
 */
      __pyx_t_3 = __pyx_f_10serializor_6binary_3ser_pack_uint8(__pyx_f_7cpython_8datetime_datetime_day(__pyx_v_obj)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 215, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);

      /* "serializor/binary/ser.py":216
 *                     pack_uint8(datetime.datetime_month(obj)),  # type: ignore
 *                     pack_uint8(datetime.datetime_day(obj)),  # type: ignore
 *                     pack_uint8(datetime.datetime_hour(obj)),  # type: ignore             # <<<<<<<<<<<<<<
 *                     pack_uint8(datetime.datetime_minute(obj)),  # type: ignore
 *                     pack_uint8(datetime.datetime_second(obj)),  # type: ignore
 */
      __pyx_t_5 = __pyx_f_10serializor_6binary_3ser_pack_uint8(__pyx_f_7cpython_8datetime_datetime_hour(__pyx_v_obj)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 216, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);

      /* "serializor/binary/ser.py":217
 *                     pack_uint8(datetime.datetime_day(obj)),  # type: ignore
 *                     pack_uint8(datetime.datetime_hour(obj)),  # type: ignore
 *                     pack_uint8(datetime.datetime_minute(obj)),  # type: ignore             # <<<<<<<<<<<<<<
 *                     pack_uint8(datetime.datetime_second(obj)),  # type: ignore
 *                     pack_uint32(datetime.datetime_microsecond(obj)),  # type: ignore
 */
      __pyx_t_6 = __pyx_f_10serializor_6binary_3ser_pack_uint8(__pyx_f_7cpython_8datetime_datetime_minute(__pyx_v_obj)); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 217, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);

      /* "serializor/binary/ser.py":218
 *                     pack_uint8(datetime.datetime_hour(obj)),  # type: ignore
 *                     pack_uint8(datetime.datetime_minute(obj)),  # type: ignore
 *                     pack_uint8(datetime.datetime_second(obj)),  # type: ignore             # <<<<<<<<<<<<<<
 *                     pack_uint32(datetime.datetime_microsecond(obj)),  # type: ignore
 *                     gen_encoded_int(bytes_len(name)),  # type: ignore
 */
      __pyx_t_7 = __pyx_f_10serializor_6binary_3ser_pack_uint8(__pyx_f_7cpython_8datetime_datetime_second(__pyx_v_obj)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 218, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);

      /* "serializor/binary/ser.py":219
 *                     pack_uint8(datetime.datetime_minute(obj)),  # type: ignore
 *                     pack_uint8(datetime.datetime_second(obj)),  # type: ignore
 *                     pack_uint32(datetime.datetime_microsecond(obj)),  # type: ignore             # <<<<<<<<<<<<<<
 *                     gen_encoded_int(bytes_len(name)),  # type: ignore
 *                     name,
 */
      __pyx_t_8 = __pyx_f_10serializor_6binary_3ser_pack_uint32(__pyx_f_7cpython_8datetime_datetime_microsecond(__pyx_v_obj)); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 219, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);

      /* "serializor/binary/ser.py":220
 *                     pack_uint8(datetime.datetime_second(obj)),  # type: ignore
 *                     pack_uint32(datetime.datetime_microsecond(obj)),  # type: ignore
 *                     gen_encoded_int(bytes_len(name)),  # type: ignore             # <<<<<<<<<<<<<<
 *                     name,
 *                 ]
 */
      __pyx_t_9 = PyBytes_Size(__pyx_v_name); if (unlikely(__pyx_t_9 == ((Py_ssize_t)-1L))) __PYX_ERR(0, 220, __pyx_L1_error)
      __pyx_t_10 = __pyx_f_10serializor_6binary_3ser_gen_encoded_int(__pyx_t_9); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 220, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);

      /* "serializor/binary/ser.py":210
 *             name = utils.encode_str(tz_name, b"ascii")
 *             return b"".join(
 *                 [             # <<<<<<<<<<<<<<
 *                     DATETIME_ID,  # identifier
 *                     b"\x01",  # approach 1
 */
      __pyx_t_11 = PyList_New(11); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 210, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_11);
      __Pyx_INCREF(__pyx_v_10serializor_6binary_3ser_DATETIME_ID);
      __Pyx_GIVEREF(__pyx_v_10serializor_6binary_3ser_DATETIME_ID);
      if (__Pyx_PyList_SET_ITEM(__pyx_t_11, 0, __pyx_v_10serializor_6binary_3ser_DATETIME_ID)) __PYX_ERR(0, 210, __pyx_L1_error);
      __Pyx_INCREF(__pyx_kp_b__5);
      __Pyx_GIVEREF(__pyx_kp_b__5);
      if (__Pyx_PyList_SET_ITEM(__pyx_t_11, 1, __pyx_kp_b__5)) __PYX_ERR(0, 210, __pyx_L1_error);
      __Pyx_GIVEREF(__pyx_t_4);
      if (__Pyx_PyList_SET_ITEM(__pyx_t_11, 2, __pyx_t_4)) __PYX_ERR(0, 210, __pyx_L1_error);
      __Pyx_GIVEREF(__pyx_t_1);
      if (__Pyx_PyList_SET_ITEM(__pyx_t_11, 3, __pyx_t_1)) __PYX_ERR(0, 210, __pyx_L1_error);
      __Pyx_GIVEREF(__pyx_t_3);
      if (__Pyx_PyList_SET_ITEM(__pyx_t_11, 4, __pyx_t_3)) __PYX_ERR(0, 210, __pyx_L1_error);
      __Pyx_GIVEREF(__pyx_t_5);
      if (__Pyx_PyList_SET_ITEM(__pyx_t_11, 5, __pyx_t_5)) __PYX_ERR(0, 210, __pyx_L1_error);
      __Pyx_GIVEREF(__pyx_t_6);
      if (__Pyx_PyList_SET_ITEM(__pyx_t_11, 6, __pyx_t_6)) __PYX_ERR(0, 210, __pyx_L1_error);
      __Pyx_GIVEREF(__pyx_t_7);
      if (__Pyx_PyList_SET_ITEM(__pyx_t_11, 7, __pyx_t_7)) __PYX_ERR(0, 210, __pyx_L1_error);
      __Pyx_GIVEREF(__pyx_t_8);
      if (__Pyx_PyList_SET_ITEM(__pyx_t_11, 8, __pyx_t_8)) __PYX_ERR(0, 210, __pyx_L1_error);
      __Pyx_GIVEREF(__pyx_t_10);
      if (__Pyx_PyList_SET_ITEM(__pyx_t_11, 9, __pyx_t_10)) __PYX_ERR(0, 210, __pyx_L1_error);
      __Pyx_INCREF(__pyx_v_name);
      __Pyx_GIVEREF(__pyx_v_name);
      if (__Pyx_PyList_SET_ITEM(__pyx_t_11, 10, __pyx_v_name)) __PYX_ERR(0, 210, __pyx_L1_error);
      __pyx_t_4 = 0;
      __pyx_t_1 = 0;
      __pyx_t_3 = 0;
      __pyx_t_5 = 0;
      __pyx_t_6 = 0;
      __pyx_t_7 = 0;
      __pyx_t_8 = 0;
      __pyx_t_10 = 0;

      /* "serializor/binary/ser.py":209
 *         if set_contains(utils.AVAILABLE_TIMEZONES, tz_name := str(tz)):
 *             name = utils.encode_str(tz_name, b"ascii")
 *             return b"".join(             # <<<<<<<<<<<<<<
 *                 [
 *                     DATETIME_ID,  # identifier
 */
      __pyx_t_10 = __Pyx_PyBytes_Join(__pyx_kp_b__4, __pyx_t_11); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 209, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
      if (!(likely(PyBytes_CheckExact(__pyx_t_10))||((__pyx_t_10) == Py_None) || __Pyx_RaiseUnexpectedTypeError("bytes", __pyx_t_10))) __PYX_ERR(0, 209, __pyx_L1_error)
      __pyx_r = ((PyObject*)__pyx_t_10);
      __pyx_t_10 = 0;
      goto __pyx_L0;

      /* "serializor/binary/ser.py":207
 *     if (tz := datetime.datetime_tzinfo(obj)) is not None:
 *         # . approach 1: timezone name
 *         if set_contains(utils.AVAILABLE_TIMEZONES, tz_name := str(tz)):             # <<<<<<<<<<<<<<
 *             name = utils.encode_str(tz_name, b"ascii")
 *             return b"".join(
 */
    }

    /* "serializor/binary/ser.py":225
 *             )
 *         # . approach 2: timezone offset
 *         if (offset := tz.utcoffset(obj)) is not None:             # <<<<<<<<<<<<<<
 *             return b"".join(
 *                 [
 */
    __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_v_tz, __pyx_n_s_utcoffset); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 225, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __pyx_t_8 = NULL;
    __pyx_t_12 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_11))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_11);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_11, function);
        __pyx_t_12 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_8, __pyx_v_obj};
      __pyx_t_10 = __Pyx_PyObject_FastCall(__pyx_t_11, __pyx_callargs+1-__pyx_t_12, 1+__pyx_t_12);
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 225, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    }
    __pyx_v_offset = __pyx_t_10;
    __Pyx_INCREF(__pyx_t_10);
    __pyx_t_2 = (__pyx_t_10 != Py_None);
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    if (__pyx_t_2) {

      /* "serializor/binary/ser.py":226
 *         # . approach 2: timezone offset
 *         if (offset := tz.utcoffset(obj)) is not None:
 *             return b"".join(             # <<<<<<<<<<<<<<
 *                 [
 *                     DATETIME_ID,  # identifier
 */
      __Pyx_XDECREF(__pyx_r);

      /* "serializor/binary/ser.py":230
 *                     DATETIME_ID,  # identifier
 *                     b"\x02",  # approach 2
 *                     pack_uint16(datetime.datetime_year(obj)),  # type: ignore             # <<<<<<<<<<<<<<
 *                     pack_uint8(datetime.datetime_month(obj)),  # type: ignore
 *                     pack_uint8(datetime.datetime_day(obj)),  # type: ignore
 */
      __pyx_t_10 = __pyx_f_10serializor_6binary_3ser_pack_uint16(__pyx_f_7cpython_8datetime_datetime_year(__pyx_v_obj)); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 230, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);

      /* "serializor/binary/ser.py":231
 *                     b"\x02",  # approach 2
 *                     pack_uint16(datetime.datetime_year(obj)),  # type: ignore
 *                     pack_uint8(datetime.datetime_month(obj)),  # type: ignore             # <<<<<<<<<<<<<<
 *                     pack_uint8(datetime.datetime_day(obj)),  # type: ignore
 *                     pack_uint8(datetime.datetime_hour(obj)),  # type: ignore
 */
      __pyx_t_11 = __pyx_f_10serializor_6binary_3ser_pack_uint8(__pyx_f_7cpython_8datetime_datetime_month(__pyx_v_obj)); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 231, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_11);

      /* "serializor/binary/ser.py":232
 *                     pack_uint16(datetime.datetime_year(obj)),  # type: ignore
 *                     pack_uint8(datetime.datetime_month(obj)),  # type: ignore
 *                     pack_uint8(datetime.datetime_day(obj)),  # type: ignore             # <<<<<<<<<<<<<<
 *                     pack_uint8(datetime.datetime_hour(obj)),  # type: ignore
 *                     pack_uint8(datetime.datetime_minute(obj)),  # type: ignore
 */
      __pyx_t_8 = __pyx_f_10serializor_6binary_3ser_pack_uint8(__pyx_f_7cpython_8datetime_datetime_day(__pyx_v_obj)); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 232, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);

      /* "serializor/binary/ser.py":233
 *                     pack_uint8(datetime.datetime_month(obj)),  # type: ignore
 *                     pack_uint8(datetime.datetime_day(obj)),  # type: ignore
 *                     pack_uint8(datetime.datetime_hour(obj)),  # type: ignore             # <<<<<<<<<<<<<<
 *                     pack_uint8(datetime.datetime_minute(obj)),  # type: ignore
 *                     pack_uint8(datetime.datetime_second(obj)),  # type: ignore
 */
      __pyx_t_7 = __pyx_f_10serializor_6binary_3ser_pack_uint8(__pyx_f_7cpython_8datetime_datetime_hour(__pyx_v_obj)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 233, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);

      /* "serializor/binary/ser.py":234
 *                     pack_uint8(datetime.datetime_day(obj)),  # type: ignore
 *                     pack_uint8(datetime.datetime_hour(obj)),  # type: ignore
 *                     pack_uint8(datetime.datetime_minute(obj)),  # type: ignore             # <<<<<<<<<<<<<<
 *                     pack_uint8(datetime.datetime_second(obj)),  # type: ignore
 *                     pack_uint32(datetime.datetime_microsecond(obj)),  # type: ignore
 */
      __pyx_t_6 = __pyx_f_10serializor_6binary_3ser_pack_uint8(__pyx_f_7cpython_8datetime_datetime_minute(__pyx_v_obj)); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 234, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);

      /* "serializor/binary/ser.py":235
 *                     pack_uint8(datetime.datetime_hour(obj)),  # type: ignore
 *                     pack_uint8(datetime.datetime_minute(obj)),  # type: ignore
 *                     pack_uint8(datetime.datetime_second(obj)),  # type: ignore             # <<<<<<<<<<<<<<
 *                     pack_uint32(datetime.datetime_microsecond(obj)),  # type: ignore
 *                     pack_int32(datetime.timedelta_days(offset)),  # type: ignore
 */
      __pyx_t_5 = __pyx_f_10serializor_6binary_3ser_pack_uint8(__pyx_f_7cpython_8datetime_datetime_second(__pyx_v_obj)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 235, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);

      /* "serializor/binary/ser.py":236
 *                     pack_uint8(datetime.datetime_minute(obj)),  # type: ignore
 *                     pack_uint8(datetime.datetime_second(obj)),  # type: ignore
 *                     pack_uint32(datetime.datetime_microsecond(obj)),  # type: ignore             # <<<<<<<<<<<<<<
 *                     pack_int32(datetime.timedelta_days(offset)),  # type: ignore
 *                     pack_int32(datetime.timedelta_seconds(offset)),  # type: ignore
 */
      __pyx_t_3 = __pyx_f_10serializor_6binary_3ser_pack_uint32(__pyx_f_7cpython_8datetime_datetime_microsecond(__pyx_v_obj)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 236, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);

      /* "serializor/binary/ser.py":237
 *                     pack_uint8(datetime.datetime_second(obj)),  # type: ignore
 *                     pack_uint32(datetime.datetime_microsecond(obj)),  # type: ignore
 *                     pack_int32(datetime.timedelta_days(offset)),  # type: ignore             # <<<<<<<<<<<<<<
 *                     pack_int32(datetime.timedelta_seconds(offset)),  # type: ignore
 *                     pack_int32(datetime.timedelta_microseconds(offset)),  # type: ignore
 */
      __pyx_t_1 = __pyx_f_10serializor_6binary_3ser_pack_int32(__pyx_f_7cpython_8datetime_timedelta_days(__pyx_v_offset)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 237, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);

      /* "serializor/binary/ser.py":238
 *                     pack_uint32(datetime.datetime_microsecond(obj)),  # type: ignore
 *                     pack_int32(datetime.timedelta_days(offset)),  # type: ignore
 *                     pack_int32(datetime.timedelta_seconds(offset)),  # type: ignore             # <<<<<<<<<<<<<<
 *                     pack_int32(datetime.timedelta_microseconds(offset)),  # type: ignore
 *                 ]
 */
      __pyx_t_4 = __pyx_f_10serializor_6binary_3ser_pack_int32(__pyx_f_7cpython_8datetime_timedelta_seconds(__pyx_v_offset)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 238, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);

      /* "serializor/binary/ser.py":239
 *                     pack_int32(datetime.timedelta_days(offset)),  # type: ignore
 *                     pack_int32(datetime.timedelta_seconds(offset)),  # type: ignore
 *                     pack_int32(datetime.timedelta_microseconds(offset)),  # type: ignore             # <<<<<<<<<<<<<<
 *                 ]
 *             )
 */
      __pyx_t_13 = __pyx_f_10serializor_6binary_3ser_pack_int32(__pyx_f_7cpython_8datetime_timedelta_microseconds(__pyx_v_offset)); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 239, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_13);

      /* "serializor/binary/ser.py":227
 *         if (offset := tz.utcoffset(obj)) is not None:
 *             return b"".join(
 *                 [             # <<<<<<<<<<<<<<
 *                     DATETIME_ID,  # identifier
 *                     b"\x02",  # approach 2
 */
      __pyx_t_14 = PyList_New(12); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 227, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_14);
      __Pyx_INCREF(__pyx_v_10serializor_6binary_3ser_DATETIME_ID);
      __Pyx_GIVEREF(__pyx_v_10serializor_6binary_3ser_DATETIME_ID);
      if (__Pyx_PyList_SET_ITEM(__pyx_t_14, 0, __pyx_v_10serializor_6binary_3ser_DATETIME_ID)) __PYX_ERR(0, 227, __pyx_L1_error);
      __Pyx_INCREF(__pyx_kp_b__6);
      __Pyx_GIVEREF(__pyx_kp_b__6);
      if (__Pyx_PyList_SET_ITEM(__pyx_t_14, 1, __pyx_kp_b__6)) __PYX_ERR(0, 227, __pyx_L1_error);
      __Pyx_GIVEREF(__pyx_t_10);
      if (__Pyx_PyList_SET_ITEM(__pyx_t_14, 2, __pyx_t_10)) __PYX_ERR(0, 227, __pyx_L1_error);
      __Pyx_GIVEREF(__pyx_t_11);
      if (__Pyx_PyList_SET_ITEM(__pyx_t_14, 3, __pyx_t_11)) __PYX_ERR(0, 227, __pyx_L1_error);
      __Pyx_GIVEREF(__pyx_t_8);
      if (__Pyx_PyList_SET_ITEM(__pyx_t_14, 4, __pyx_t_8)) __PYX_ERR(0, 227, __pyx_L1_error);
      __Pyx_GIVEREF(__pyx_t_7);
      if (__Pyx_PyList_SET_ITEM(__pyx_t_14, 5, __pyx_t_7)) __PYX_ERR(0, 227, __pyx_L1_error);
      __Pyx_GIVEREF(__pyx_t_6);
      if (__Pyx_PyList_SET_ITEM(__pyx_t_14, 6, __pyx_t_6)) __PYX_ERR(0, 227, __pyx_L1_error);
      __Pyx_GIVEREF(__pyx_t_5);
      if (__Pyx_PyList_SET_ITEM(__pyx_t_14, 7, __pyx_t_5)) __PYX_ERR(0, 227, __pyx_L1_error);
      __Pyx_GIVEREF(__pyx_t_3);
      if (__Pyx_PyList_SET_ITEM(__pyx_t_14, 8, __pyx_t_3)) __PYX_ERR(0, 227, __pyx_L1_error);
      __Pyx_GIVEREF(__pyx_t_1);
      if (__Pyx_PyList_SET_ITEM(__pyx_t_14, 9, __pyx_t_1)) __PYX_ERR(0, 227, __pyx_L1_error);
      __Pyx_GIVEREF(__pyx_t_4);
      if (__Pyx_PyList_SET_ITEM(__pyx_t_14, 10, __pyx_t_4)) __PYX_ERR(0, 227, __pyx_L1_error);
      __Pyx_GIVEREF(__pyx_t_13);
      if (__Pyx_PyList_SET_ITEM(__pyx_t_14, 11, __pyx_t_13)) __PYX_ERR(0, 227, __pyx_L1_error);
      __pyx_t_10 = 0;
      __pyx_t_11 = 0;
      __pyx_t_8 = 0;
      __pyx_t_7 = 0;
      __pyx_t_6 = 0;
      __pyx_t_5 = 0;
      __pyx_t_3 = 0;
      __pyx_t_1 = 0;
      __pyx_t_4 = 0;
      __pyx_t_13 = 0;

      /* "serializor/binary/ser.py":226
 *         # . approach 2: timezone offset
 *         if (offset := tz.utcoffset(obj)) is not None:
 *             return b"".join(             # <<<<<<<<<<<<<<
 *                 [
 *                     DATETIME_ID,  # identifier
 */
      __pyx_t_13 = __Pyx_PyBytes_Join(__pyx_kp_b__4, __pyx_t_14); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 226, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_13);
      __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
      if (!(likely(PyBytes_CheckExact(__pyx_t_13))||((__pyx_t_13) == Py_None) || __Pyx_RaiseUnexpectedTypeError("bytes", __pyx_t_13))) __PYX_ERR(0, 226, __pyx_L1_error)
      __pyx_r = ((PyObject*)__pyx_t_13);
      __pyx_t_13 = 0;
      goto __pyx_L0;

      /* "serializor/binary/ser.py":225
 *             )
 *         # . approach 2: timezone offset
 *         if (offset := tz.utcoffset(obj)) is not None:             # <<<<<<<<<<<<<<
 *             return b"".join(
 *                 [
 */
    }

    /* "serializor/binary/ser.py":205
 *     """
 *     # With timezone
 *     if (tz := datetime.datetime_tzinfo(obj)) is not None:             # <<<<<<<<<<<<<<
 *         # . approach 1: timezone name
 *         if set_contains(utils.AVAILABLE_TIMEZONES, tz_name := str(tz)):
 */
  }

  /* "serializor/binary/ser.py":243
 *             )
 *     # Without timzone / Fallback
 *     return b"".join(             # <<<<<<<<<<<<<<
 *         [
 *             DATETIME_ID,  # identifier
 */
  __Pyx_XDECREF(__pyx_r);

  /* "serializor/binary/ser.py":247
 *             DATETIME_ID,  # identifier
 *             b"\x00",  # approach 0
 *             pack_uint16(datetime.datetime_year(obj)),  # type: ignore             # <<<<<<<<<<<<<<
 *             pack_uint8(datetime.datetime_month(obj)),  # type: ignore
 *             pack_uint8(datetime.datetime_day(obj)),  # type: ignore
 */
  __pyx_t_13 = __pyx_f_10serializor_6binary_3ser_pack_uint16(__pyx_f_7cpython_8datetime_datetime_year(__pyx_v_obj)); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 247, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);

  /* "serializor/binary/ser.py":248
 *             b"\x00",  # approach 0
 *             pack_uint16(datetime.datetime_year(obj)),  # type: ignore
 *             pack_uint8(datetime.datetime_month(obj)),  # type: ignore             # <<<<<<<<<<<<<<
 *             pack_uint8(datetime.datetime_day(obj)),  # type: ignore
 *             pack_uint8(datetime.datetime_hour(obj)),  # type: ignore
 */
  __pyx_t_14 = __pyx_f_10serializor_6binary_3ser_pack_uint8(__pyx_f_7cpython_8datetime_datetime_month(__pyx_v_obj)); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 248, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_14);

  /* "serializor/binary/ser.py":249
 *             pack_uint16(datetime.datetime_year(obj)),  # type: ignore
 *             pack_uint8(datetime.datetime_month(obj)),  # type: ignore
 *             pack_uint8(datetime.datetime_day(obj)),  # type: ignore             # <<<<<<<<<<<<<<
 *             pack_uint8(datetime.datetime_hour(obj)),  # type: ignore
 *             pack_uint8(datetime.datetime_minute(obj)),  # type: ignore
 */
  __pyx_t_4 = __pyx_f_10serializor_6binary_3ser_pack_uint8(__pyx_f_7cpython_8datetime_datetime_day(__pyx_v_obj)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 249, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);

  /* "serializor/binary/ser.py":250
 *             pack_uint8(datetime.datetime_month(obj)),  # type: ignore
 *             pack_uint8(datetime.datetime_day(obj)),  # type: ignore
 *             pack_uint8(datetime.datetime_hour(obj)),  # type: ignore             # <<<<<<<<<<<<<<
 *             pack_uint8(datetime.datetime_minute(obj)),  # type: ignore
 *             pack_uint8(datetime.datetime_second(obj)),  # type: ignore
 */
  __pyx_t_1 = __pyx_f_10serializor_6binary_3ser_pack_uint8(__pyx_f_7cpython_8datetime_datetime_hour(__pyx_v_obj)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 250, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "serializor/binary/ser.py":251
 *             pack_uint8(datetime.datetime_day(obj)),  # type: ignore
 *             pack_uint8(datetime.datetime_hour(obj)),  # type: ignore
 *             pack_uint8(datetime.datetime_minute(obj)),  # type: ignore             # <<<<<<<<<<<<<<
 *             pack_uint8(datetime.datetime_second(obj)),  # type: ignore
 *             pack_uint32(datetime.datetime_microsecond(obj)),  # type: ignore
 */
  __pyx_t_3 = __pyx_f_10serializor_6binary_3ser_pack_uint8(__pyx_f_7cpython_8datetime_datetime_minute(__pyx_v_obj)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 251, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);

  /* "serializor/binary/ser.py":252
 *             pack_uint8(datetime.datetime_hour(obj)),  # type: ignore
 *             pack_uint8(datetime.datetime_minute(obj)),  # type: ignore
 *             pack_uint8(datetime.datetime_second(obj)),  # type: ignore             # <<<<<<<<<<<<<<
 *             pack_uint32(datetime.datetime_microsecond(obj)),  # type: ignore
 *         ]
 */
  __pyx_t_5 = __pyx_f_10serializor_6binary_3ser_pack_uint8(__pyx_f_7cpython_8datetime_datetime_second(__pyx_v_obj)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 252, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);

  /* "serializor/binary/ser.py":253
 *             pack_uint8(datetime.datetime_minute(obj)),  # type: ignore
 *             pack_uint8(datetime.datetime_second(obj)),  # type: ignore
 *             pack_uint32(datetime.datetime_microsecond(obj)),  # type: ignore             # <<<<<<<<<<<<<<
 *         ]
 *     )
 */
  __pyx_t_6 = __pyx_f_10serializor_6binary_3ser_pack_uint32(__pyx_f_7cpython_8datetime_datetime_microsecond(__pyx_v_obj)); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 253, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);

  /* "serializor/binary/ser.py":244
 *     # Without timzone / Fallback
 *     return b"".join(
 *         [             # <<<<<<<<<<<<<<
 *             DATETIME_ID,  # identifier
 *             b"\x00",  # approach 0
 */
  __pyx_t_7 = PyList_New(9); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 244, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_INCREF(__pyx_v_10serializor_6binary_3ser_DATETIME_ID);
  __Pyx_GIVEREF(__pyx_v_10serializor_6binary_3ser_DATETIME_ID);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_7, 0, __pyx_v_10serializor_6binary_3ser_DATETIME_ID)) __PYX_ERR(0, 244, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_b__7);
  __Pyx_GIVEREF(__pyx_kp_b__7);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_7, 1, __pyx_kp_b__7)) __PYX_ERR(0, 244, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_13);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_7, 2, __pyx_t_13)) __PYX_ERR(0, 244, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_14);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_7, 3, __pyx_t_14)) __PYX_ERR(0, 244, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_4);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_7, 4, __pyx_t_4)) __PYX_ERR(0, 244, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_7, 5, __pyx_t_1)) __PYX_ERR(0, 244, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_3);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_7, 6, __pyx_t_3)) __PYX_ERR(0, 244, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_5);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_7, 7, __pyx_t_5)) __PYX_ERR(0, 244, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_6);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_7, 8, __pyx_t_6)) __PYX_ERR(0, 244, __pyx_L1_error);
  __pyx_t_13 = 0;
  __pyx_t_14 = 0;
  __pyx_t_4 = 0;
  __pyx_t_1 = 0;
  __pyx_t_3 = 0;
  __pyx_t_5 = 0;
  __pyx_t_6 = 0;

  /* "serializor/binary/ser.py":243
 *             )
 *     # Without timzone / Fallback
 *     return b"".join(             # <<<<<<<<<<<<<<
 *         [
 *             DATETIME_ID,  # identifier
 */
  __pyx_t_6 = __Pyx_PyBytes_Join(__pyx_kp_b__4, __pyx_t_7); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 243, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (!(likely(PyBytes_CheckExact(__pyx_t_6))||((__pyx_t_6) == Py_None) || __Pyx_RaiseUnexpectedTypeError("bytes", __pyx_t_6))) __PYX_ERR(0, 243, __pyx_L1_error)
  __pyx_r = ((PyObject*)__pyx_t_6);
  __pyx_t_6 = 0;
  goto __pyx_L0;

  /* "serializor/binary/ser.py":185
 * 
 * # Date&Time Types -----------------------------------------------------------------------------
 * @cython.cfunc             # <<<<<<<<<<<<<<
 * @cython.inline(True)
 * def _ser_datetime(obj: object) -> bytes:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_AddTraceback("serializor.binary.ser._ser_datetime", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_name);
  __Pyx_XDECREF(__pyx_v_tz);
  __Pyx_XDECREF(__pyx_v_tz_name);
  __Pyx_XDECREF(__pyx_v_offset);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "serializor/binary/ser.py":258
 * 
 * 
 * @cython.cfunc             # <<<<<<<<<<<<<<
 * @cython.inline(True)
 * def _ser_date(obj: object) -> bytes:
 */

static PyObject *__pyx_f_10serializor_6binary_3ser__ser_date(PyObject *__pyx_v_obj) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_ser_date", 1);

  /* "serializor/binary/ser.py":266
 *     >>> b'[id(1B)][YY(2B)][MM(1B)][DD(1B)]'
 *     """
 *     return b"".join(             # <<<<<<<<<<<<<<
 *         [
 *             DATE_ID,  # identifier
 */
  __Pyx_XDECREF(__pyx_r);

  /* "serializor/binary/ser.py":269
 *         [
 *             DATE_ID,  # identifier
 *             pack_uint16(datetime.date_year(obj)),  # type: ignore             # <<<<<<<<<<<<<<
 *             pack_uint8(datetime.date_month(obj)),  # type: ignore
 *             pack_uint8(datetime.date_day(obj)),  # type: ignore
 */
  __pyx_t_1 = __pyx_f_10serializor_6binary_3ser_pack_uint16(__pyx_f_7cpython_8datetime_date_year(__pyx_v_obj)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 269, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "serializor/binary/ser.py":270
 *             DATE_ID,  # identifier
 *             pack_uint16(datetime.date_year(obj)),  # type: ignore
 *             pack_uint8(datetime.date_month(obj)),  # type: ignore             # <<<<<<<<<<<<<<
 *             pack_uint8(datetime.date_day(obj)),  # type: ignore
 *         ]
 */
  __pyx_t_2 = __pyx_f_10serializor_6binary_3ser_pack_uint8(__pyx_f_7cpython_8datetime_date_month(__pyx_v_obj)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 270, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);

  /* "serializor/binary/ser.py":271
 *             pack_uint16(datetime.date_year(obj)),  # type: ignore
 *             pack_uint8(datetime.date_month(obj)),  # type: ignore
 *             pack_uint8(datetime.date_day(obj)),  # type: ignore             # <<<<<<<<<<<<<<
 *         ]
 *     )
 */
  __pyx_t_3 = __pyx_f_10serializor_6binary_3ser_pack_uint8(__pyx_f_7cpython_8datetime_date_day(__pyx_v_obj)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 271, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);

  /* "serializor/binary/ser.py":267
 *     """
 *     return b"".join(
 *         [             # <<<<<<<<<<<<<<
 *             DATE_ID,  # identifier
 *             pack_uint16(datetime.date_year(obj)),  # type: ignore
 */
  __pyx_t_4 = PyList_New(4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 267, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_INCREF(__pyx_v_10serializor_6binary_3ser_DATE_ID);
  __Pyx_GIVEREF(__pyx_v_10serializor_6binary_3ser_DATE_ID);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_4, 0, __pyx_v_10serializor_6binary_3ser_DATE_ID)) __PYX_ERR(0, 267, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_4, 1, __pyx_t_1)) __PYX_ERR(0, 267, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_4, 2, __pyx_t_2)) __PYX_ERR(0, 267, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_3);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_4, 3, __pyx_t_3)) __PYX_ERR(0, 267, __pyx_L1_error);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;

  /* "serializor/binary/ser.py":266
 *     >>> b'[id(1B)][YY(2B)][MM(1B)][DD(1B)]'
 *     """
 *     return b"".join(             # <<<<<<<<<<<<<<
 *         [
 *             DATE_ID,  # identifier
 */
  __pyx_t_3 = __Pyx_PyBytes_Join(__pyx_kp_b__4, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 266, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (!(likely(PyBytes_CheckExact(__pyx_t_3))||((__pyx_t_3) == Py_None) || __Pyx_RaiseUnexpectedTypeError("bytes", __pyx_t_3))) __PYX_ERR(0, 266, __pyx_L1_error)
  __pyx_r = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "serializor/binary/ser.py":258
 * 
 * 
 * @cython.cfunc             # <<<<<<<<<<<<<<
 * @cython.inline(True)
 * def _ser_date(obj: object) -> bytes:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("serializor.binary.ser._ser_date", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "serializor/binary/ser.py":276
 * 
 * 
 * @cython.cfunc             # <<<<<<<<<<<<<<
 * @cython.inline(True)
 * def _ser_time(obj: object) -> bytes:
 */

static PyObject *__pyx_f_10serializor_6binary_3ser__ser_time(PyObject *__pyx_v_obj) {
  PyObject *__pyx_v_name = NULL;
  PyObject *__pyx_v_tz = NULL;
  PyObject *__pyx_v_tz_name = NULL;
  PyObject *__pyx_v_offset = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  Py_ssize_t __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  unsigned int __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_ser_time", 1);

  /* "serializor/binary/ser.py":296
 *     """
 *     # With timezone
 *     if (tz := datetime.time_tzinfo(obj)) is not None:             # <<<<<<<<<<<<<<
 *         # . approach 1: timezone name
 *         if set_contains(utils.AVAILABLE_TIMEZONES, tz_name := str(tz)):
 */
  __pyx_t_1 = __pyx_f_7cpython_8datetime_time_tzinfo(__pyx_v_obj); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 296, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_tz = __pyx_t_1;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = (__pyx_t_1 != Py_None);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "serializor/binary/ser.py":298
 *     if (tz := datetime.time_tzinfo(obj)) is not None:
 *         # . approach 1: timezone name
 *         if set_contains(utils.AVAILABLE_TIMEZONES, tz_name := str(tz)):             # <<<<<<<<<<<<<<
 *             name = utils.encode_str(tz_name, b"ascii")
 *             return b"".join(
 */
    __pyx_t_1 = __pyx_v_10serializor_5utils_AVAILABLE_TIMEZONES;
    __Pyx_INCREF(__pyx_t_1);
    __pyx_t_3 = __Pyx_PyObject_Unicode(__pyx_v_tz); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 298, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_v_tz_name = __pyx_t_3;
    __Pyx_INCREF(__pyx_t_3);
    __pyx_t_4 = __pyx_t_3;
    __Pyx_INCREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_2 = PySet_Contains(__pyx_t_1, __pyx_t_4); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(0, 298, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (__pyx_t_2) {

      /* "serializor/binary/ser.py":299
 *         # . approach 1: timezone name
 *         if set_contains(utils.AVAILABLE_TIMEZONES, tz_name := str(tz)):
 *             name = utils.encode_str(tz_name, b"ascii")             # <<<<<<<<<<<<<<
 *             return b"".join(
 *                 [
 */
      __pyx_t_4 = __pyx_f_10serializor_5utils_encode_str(__pyx_v_tz_name, ((char *)"ascii")); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 299, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_v_name = ((PyObject*)__pyx_t_4);
      __pyx_t_4 = 0;

      /* "serializor/binary/ser.py":300
 *         if set_contains(utils.AVAILABLE_TIMEZONES, tz_name := str(tz)):
 *             name = utils.encode_str(tz_name, b"ascii")
 *             return b"".join(             # <<<<<<<<<<<<<<
 *                 [
 *                     TIME_ID,  # identifier
 */
      __Pyx_XDECREF(__pyx_r);

      /* "serializor/binary/ser.py":304
 *                     TIME_ID,  # identifier
 *                     b"\x01",  # approach 1
 *                     pack_uint8(datetime.time_hour(obj)),  # type: ignore             # <<<<<<<<<<<<<<
 *                     pack_uint8(datetime.time_minute(obj)),  # type: ignore
 *                     pack_uint8(datetime.time_second(obj)),  # type: ignore
 */
      __pyx_t_4 = __pyx_f_10serializor_6binary_3ser_pack_uint8(__pyx_f_7cpython_8datetime_time_hour(__pyx_v_obj)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 304, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);

      /* "serializor/binary/ser.py":305
 *                     b"\x01",  # approach 1
 *                     pack_uint8(datetime.time_hour(obj)),  # type: ignore
 *                     pack_uint8(datetime.time_minute(obj)),  # type: ignore             # <<<<<<<<<<<<<<
 *                     pack_uint8(datetime.time_second(obj)),  # type: ignore
 *                     pack_uint32(datetime.time_microsecond(obj)),  # type: ignore
 */
      __pyx_t_1 = __pyx_f_10serializor_6binary_3ser_pack_uint8(__pyx_f_7cpython_8datetime_time_minute(__pyx_v_obj)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 305, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);

      /* "serializor/binary/ser.py":306
 *                     pack_uint8(datetime.time_hour(obj)),  # type: ignore
 *                     pack_uint8(datetime.time_minute(obj)),  # type: ignore
 *                     pack_uint8(datetime.time_second(obj)),  # type: ignore             # <<<<<<<<<<<<<<
 *                     pack_uint32(datetime.time_microsecond(obj)),  # type: ignore
 *                     gen_encoded_int(bytes_len(name)),  # type: ignore
 */
      __pyx_t_3 = __pyx_f_10serializor_6binary_3ser_pack_uint8(__pyx_f_7cpython_8datetime_time_second(__pyx_v_obj)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 306, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);

      /* "serializor/binary/ser.py":307
 *                     pack_uint8(datetime.time_minute(obj)),  # type: ignore
 *                     pack_uint8(datetime.time_second(obj)),  # type: ignore
 *                     pack_uint32(datetime.time_microsecond(obj)),  # type: ignore             # <<<<<<<<<<<<<<
 *                     gen_encoded_int(bytes_len(name)),  # type: ignore
 *                     name,
 */
      __pyx_t_5 = __pyx_f_10serializor_6binary_3ser_pack_uint32(__pyx_f_7cpython_8datetime_time_microsecond(__pyx_v_obj)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 307, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);

      /* "serializor/binary/ser.py":308
 *                     pack_uint8(datetime.time_second(obj)),  # type: ignore
 *                     pack_uint32(datetime.time_microsecond(obj)),  # type: ignore
 *                     gen_encoded_int(bytes_len(name)),  # type: ignore             # <<<<<<<<<<<<<<
 *                     name,
 *                 ]
 */
      __pyx_t_6 = PyBytes_Size(__pyx_v_name); if (unlikely(__pyx_t_6 == ((Py_ssize_t)-1L))) __PYX_ERR(0, 308, __pyx_L1_error)
      __pyx_t_7 = __pyx_f_10serializor_6binary_3ser_gen_encoded_int(__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 308, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);

      /* "serializor/binary/ser.py":301
 *             name = utils.encode_str(tz_name, b"ascii")
 *             return b"".join(
 *                 [             # <<<<<<<<<<<<<<
 *                     TIME_ID,  # identifier
 *                     b"\x01",  # approach 1
 */
      __pyx_t_8 = PyList_New(8); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 301, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_INCREF(__pyx_v_10serializor_6binary_3ser_TIME_ID);
      __Pyx_GIVEREF(__pyx_v_10serializor_6binary_3ser_TIME_ID);
      if (__Pyx_PyList_SET_ITEM(__pyx_t_8, 0, __pyx_v_10serializor_6binary_3ser_TIME_ID)) __PYX_ERR(0, 301, __pyx_L1_error);
      __Pyx_INCREF(__pyx_kp_b__5);
      __Pyx_GIVEREF(__pyx_kp_b__5);
      if (__Pyx_PyList_SET_ITEM(__pyx_t_8, 1, __pyx_kp_b__5)) __PYX_ERR(0, 301, __pyx_L1_error);
      __Pyx_GIVEREF(__pyx_t_4);
      if (__Pyx_PyList_SET_ITEM(__pyx_t_8, 2, __pyx_t_4)) __PYX_ERR(0, 301, __pyx_L1_error);
      __Pyx_GIVEREF(__pyx_t_1);
      if (__Pyx_PyList_SET_ITEM(__pyx_t_8, 3, __pyx_t_1)) __PYX_ERR(0, 301, __pyx_L1_error);
      __Pyx_GIVEREF(__pyx_t_3);
      if (__Pyx_PyList_SET_ITEM(__pyx_t_8, 4, __pyx_t_3)) __PYX_ERR(0, 301, __pyx_L1_error);
      __Pyx_GIVEREF(__pyx_t_5);
      if (__Pyx_PyList_SET_ITEM(__pyx_t_8, 5, __pyx_t_5)) __PYX_ERR(0, 301, __pyx_L1_error);
      __Pyx_GIVEREF(__pyx_t_7);
      if (__Pyx_PyList_SET_ITEM(__pyx_t_8, 6, __pyx_t_7)) __PYX_ERR(0, 301, __pyx_L1_error);
      __Pyx_INCREF(__pyx_v_name);
      __Pyx_GIVEREF(__pyx_v_name);
      if (__Pyx_PyList_SET_ITEM(__pyx_t_8, 7, __pyx_v_name)) __PYX_ERR(0, 301, __pyx_L1_error);
      __pyx_t_4 = 0;
      __pyx_t_1 = 0;
      __pyx_t_3 = 0;
      __pyx_t_5 = 0;
      __pyx_t_7 = 0;

      /* "serializor/binary/ser.py":300
 *         if set_contains(utils.AVAILABLE_TIMEZONES, tz_name := str(tz)):
 *             name = utils.encode_str(tz_name, b"ascii")
 *             return b"".join(             # <<<<<<<<<<<<<<
 *                 [
 *                     TIME_ID,  # identifier
 */
      __pyx_t_7 = __Pyx_PyBytes_Join(__pyx_kp_b__4, __pyx_t_8); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 300, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      if (!(likely(PyBytes_CheckExact(__pyx_t_7))||((__pyx_t_7) == Py_None) || __Pyx_RaiseUnexpectedTypeError("bytes", __pyx_t_7))) __PYX_ERR(0, 300, __pyx_L1_error)
      __pyx_r = ((PyObject*)__pyx_t_7);
      __pyx_t_7 = 0;
      goto __pyx_L0;

      /* "serializor/binary/ser.py":298
 *     if (tz := datetime.time_tzinfo(obj)) is not None:
 *         # . approach 1: timezone name
 *         if set_contains(utils.AVAILABLE_TIMEZONES, tz_name := str(tz)):             # <<<<<<<<<<<<<<
 *             name = utils.encode_str(tz_name, b"ascii")
 *             return b"".join(
 */
    }

    /* "serializor/binary/ser.py":313
 *             )
 *         # . approach 2: timezone offset
 *         if (offset := tz.utcoffset(None)) is not None:             # <<<<<<<<<<<<<<
 *             return b"".join(
 *                 [
 */
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_v_tz, __pyx_n_s_utcoffset); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 313, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_5 = NULL;
    __pyx_t_9 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_8))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_8);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_8, function);
        __pyx_t_9 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_5, Py_None};
      __pyx_t_7 = __Pyx_PyObject_FastCall(__pyx_t_8, __pyx_callargs+1-__pyx_t_9, 1+__pyx_t_9);
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 313, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    }
    __pyx_v_offset = __pyx_t_7;
    __Pyx_INCREF(__pyx_t_7);
    __pyx_t_2 = (__pyx_t_7 != Py_None);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    if (__pyx_t_2) {

      /* "serializor/binary/ser.py":314
 *         # . approach 2: timezone offset
 *         if (offset := tz.utcoffset(None)) is not None:
 *             return b"".join(             # <<<<<<<<<<<<<<
 *                 [
 *                     TIME_ID,  # identifier
 */
      __Pyx_XDECREF(__pyx_r);

      /* "serializor/binary/ser.py":318
 *                     TIME_ID,  # identifier
 *                     b"\x02",  # approach 2
 *                     pack_uint8(datetime.time_hour(obj)),  # type: ignore             # <<<<<<<<<<<<<<
 *                     pack_uint8(datetime.time_minute(obj)),  # type: ignore
 *                     pack_uint8(datetime.time_second(obj)),  # type: ignore
 */
      __pyx_t_7 = __pyx_f_10serializor_6binary_3ser_pack_uint8(__pyx_f_7cpython_8datetime_time_hour(__pyx_v_obj)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 318, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);

      /* "serializor/binary/ser.py":319
 *                     b"\x02",  # approach 2
 *                     pack_uint8(datetime.time_hour(obj)),  # type: ignore
 *                     pack_uint8(datetime.time_minute(obj)),  # type: ignore             # <<<<<<<<<<<<<<
 *                     pack_uint8(datetime.time_second(obj)),  # type: ignore
 *                     pack_uint32(datetime.time_microsecond(obj)),  # type: ignore
 */
      __pyx_t_8 = __pyx_f_10serializor_6binary_3ser_pack_uint8(__pyx_f_7cpython_8datetime_time_minute(__pyx_v_obj)); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 319, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);

      /* "serializor/binary/ser.py":320
 *                     pack_uint8(datetime.time_hour(obj)),  # type: ignore
 *                     pack_uint8(datetime.time_minute(obj)),  # type: ignore
 *                     pack_uint8(datetime.time_second(obj)),  # type: ignore             # <<<<<<<<<<<<<<
 *                     pack_uint32(datetime.time_microsecond(obj)),  # type: ignore
 *                     pack_int32(datetime.timedelta_days(offset)),  # type: ignore
 */
      __pyx_t_5 = __pyx_f_10serializor_6binary_3ser_pack_uint8(__pyx_f_7cpython_8datetime_time_second(__pyx_v_obj)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 320, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);

      /* "serializor/binary/ser.py":321
 *                     pack_uint8(datetime.time_minute(obj)),  # type: ignore
 *                     pack_uint8(datetime.time_second(obj)),  # type: ignore
 *                     pack_uint32(datetime.time_microsecond(obj)),  # type: ignore             # <<<<<<<<<<<<<<
 *                     pack_int32(datetime.timedelta_days(offset)),  # type: ignore
 *                     pack_int32(datetime.timedelta_seconds(offset)),  # type: ignore
 */
      __pyx_t_3 = __pyx_f_10serializor_6binary_3ser_pack_uint32(__pyx_f_7cpython_8datetime_time_microsecond(__pyx_v_obj)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 321, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);

      /* "serializor/binary/ser.py":322
 *                     pack_uint8(datetime.time_second(obj)),  # type: ignore
 *                     pack_uint32(datetime.time_microsecond(obj)),  # type: ignore
 *                     pack_int32(datetime.timedelta_days(offset)),  # type: ignore             # <<<<<<<<<<<<<<
 *                     pack_int32(datetime.timedelta_seconds(offset)),  # type: ignore
 *                     pack_int32(datetime.timedelta_microseconds(offset)),  # type: ignore
 */
      __pyx_t_1 = __pyx_f_10serializor_6binary_3ser_pack_int32(__pyx_f_7cpython_8datetime_timedelta_days(__pyx_v_offset)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 322, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);

      /* "serializor/binary/ser.py":323
 *                     pack_uint32(datetime.time_microsecond(obj)),  # type: ignore
 *                     pack_int32(datetime.timedelta_days(offset)),  # type: ignore
 *                     pack_int32(datetime.timedelta_seconds(offset)),  # type: ignore             # <<<<<<<<<<<<<<
 *                     pack_int32(datetime.timedelta_microseconds(offset)),  # type: ignore
 *                 ]
 */
      __pyx_t_4 = __pyx_f_10serializor_6binary_3ser_pack_int32(__pyx_f_7cpython_8datetime_timedelta_seconds(__pyx_v_offset)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 323, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);

      /* "serializor/binary/ser.py":324
 *                     pack_int32(datetime.timedelta_days(offset)),  # type: ignore
 *                     pack_int32(datetime.timedelta_seconds(offset)),  # type: ignore
 *                     pack_int32(datetime.timedelta_microseconds(offset)),  # type: ignore             # <<<<<<<<<<<<<<
 *                 ]
 *             )
 */
      __pyx_t_10 = __pyx_f_10serializor_6binary_3ser_pack_int32(__pyx_f_7cpython_8datetime_timedelta_microseconds(__pyx_v_offset)); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 324, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);

      /* "serializor/binary/ser.py":315
 *         if (offset := tz.utcoffset(None)) is not None:
 *             return b"".join(
 *                 [             # <<<<<<<<<<<<<<
 *                     TIME_ID,  # identifier
 *                     b"\x02",  # approach 2
 */
      __pyx_t_11 = PyList_New(9); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 315, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_11);
      __Pyx_INCREF(__pyx_v_10serializor_6binary_3ser_TIME_ID);
      __Pyx_GIVEREF(__pyx_v_10serializor_6binary_3ser_TIME_ID);
      if (__Pyx_PyList_SET_ITEM(__pyx_t_11, 0, __pyx_v_10serializor_6binary_3ser_TIME_ID)) __PYX_ERR(0, 315, __pyx_L1_error);
      __Pyx_INCREF(__pyx_kp_b__6);
      __Pyx_GIVEREF(__pyx_kp_b__6);
      if (__Pyx_PyList_SET_ITEM(__pyx_t_11, 1, __pyx_kp_b__6)) __PYX_ERR(0, 315, __pyx_L1_error);
      __Pyx_GIVEREF(__pyx_t_7);
      if (__Pyx_PyList_SET_ITEM(__pyx_t_11, 2, __pyx_t_7)) __PYX_ERR(0, 315, __pyx_L1_error);
      __Pyx_GIVEREF(__pyx_t_8);
      if (__Pyx_PyList_SET_ITEM(__pyx_t_11, 3, __pyx_t_8)) __PYX_ERR(0, 315, __pyx_L1_error);
      __Pyx_GIVEREF(__pyx_t_5);
      if (__Pyx_PyList_SET_ITEM(__pyx_t_11, 4, __pyx_t_5)) __PYX_ERR(0, 315, __pyx_L1_error);
      __Pyx_GIVEREF(__pyx_t_3);
      if (__Pyx_PyList_SET_ITEM(__pyx_t_11, 5, __pyx_t_3)) __PYX_ERR(0, 315, __pyx_L1_error);
      __Pyx_GIVEREF(__pyx_t_1);
      if (__Pyx_PyList_SET_ITEM(__pyx_t_11, 6, __pyx_t_1)) __PYX_ERR(0, 315, __pyx_L1_error);
      __Pyx_GIVEREF(__pyx_t_4);
      if (__Pyx_PyList_SET_ITEM(__pyx_t_11, 7, __pyx_t_4)) __PYX_ERR(0, 315, __pyx_L1_error);
      __Pyx_GIVEREF(__pyx_t_10);
      if (__Pyx_PyList_SET_ITEM(__pyx_t_11, 8, __pyx_t_10)) __PYX_ERR(0, 315, __pyx_L1_error);
      __pyx_t_7 = 0;
      __pyx_t_8 = 0;
      __pyx_t_5 = 0;
      __pyx_t_3 = 0;
      __pyx_t_1 = 0;
      __pyx_t_4 = 0;
      __pyx_t_10 = 0;

      /* "serializor/binary/ser.py":314
 *         # . approach 2: timezone offset
 *         if (offset := tz.utcoffset(None)) is not None:
 *             return b"".join(             # <<<<<<<<<<<<<<
 *                 [
 *                     TIME_ID,  # identifier
 */
      __pyx_t_10 = __Pyx_PyBytes_Join(__pyx_kp_b__4, __pyx_t_11); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 314, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
      if (!(likely(PyBytes_CheckExact(__pyx_t_10))||((__pyx_t_10) == Py_None) || __Pyx_RaiseUnexpectedTypeError("bytes", __pyx_t_10))) __PYX_ERR(0, 314, __pyx_L1_error)
      __pyx_r = ((PyObject*)__pyx_t_10);
      __pyx_t_10 = 0;
      goto __pyx_L0;

      /* "serializor/binary/ser.py":313
 *             )
 *         # . approach 2: timezone offset
 *         if (offset := tz.utcoffset(None)) is not None:             # <<<<<<<<<<<<<<
 *             return b"".join(
 *                 [
 */
    }

    /* "serializor/binary/ser.py":296
 *     """
 *     # With timezone
 *     if (tz := datetime.time_tzinfo(obj)) is not None:             # <<<<<<<<<<<<<<
 *         # . approach 1: timezone name
 *         if set_contains(utils.AVAILABLE_TIMEZONES, tz_name := str(tz)):
 */
  }

  /* "serializor/binary/ser.py":328
 *             )
 *     # Without timzone / Fallback
 *     return b"".join(             # <<<<<<<<<<<<<<
 *         [
 *             TIME_ID,  # identifier
 */
  __Pyx_XDECREF(__pyx_r);

  /* "serializor/binary/ser.py":332
 *             TIME_ID,  # identifier
 *             b"\x00",  # approach 0
 *             pack_uint8(datetime.time_hour(obj)),  # type: ignore             # <<<<<<<<<<<<<<
 *             pack_uint8(datetime.time_minute(obj)),  # type: ignore
 *             pack_uint8(datetime.time_second(obj)),  # type: ignore
 */
  __pyx_t_10 = __pyx_f_10serializor_6binary_3ser_pack_uint8(__pyx_f_7cpython_8datetime_time_hour(__pyx_v_obj)); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 332, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);

  /* "serializor/binary/ser.py":333
 *             b"\x00",  # approach 0
 *             pack_uint8(datetime.time_hour(obj)),  # type: ignore
 *             pack_uint8(datetime.time_minute(obj)),  # type: ignore             # <<<<<<<<<<<<<<
 *             pack_uint8(datetime.time_second(obj)),  # type: ignore
 *             pack_uint32(datetime.time_microsecond(obj)),  # type: ignore
 */
  __pyx_t_11 = __pyx_f_10serializor_6binary_3ser_pack_uint8(__pyx_f_7cpython_8datetime_time_minute(__pyx_v_obj)); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 333, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);

  /* "serializor/binary/ser.py":334
 *             pack_uint8(datetime.time_hour(obj)),  # type: ignore
 *             pack_uint8(datetime.time_minute(obj)),  # type: ignore
 *             pack_uint8(datetime.time_second(obj)),  # type: ignore             # <<<<<<<<<<<<<<
 *             pack_uint32(datetime.time_microsecond(obj)),  # type: ignore
 *         ]
 */
  __pyx_t_4 = __pyx_f_10serializor_6binary_3ser_pack_uint8(__pyx_f_7cpython_8datetime_time_second(__pyx_v_obj)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 334, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);

  /* "serializor/binary/ser.py":335
 *             pack_uint8(datetime.time_minute(obj)),  # type: ignore
 *             pack_uint8(datetime.time_second(obj)),  # type: ignore
 *             pack_uint32(datetime.time_microsecond(obj)),  # type: ignore             # <<<<<<<<<<<<<<
 *         ]
 *     )
 */
  __pyx_t_1 = __pyx_f_10serializor_6binary_3ser_pack_uint32(__pyx_f_7cpython_8datetime_time_microsecond(__pyx_v_obj)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 335, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "serializor/binary/ser.py":329
 *     # Without timzone / Fallback
 *     return b"".join(
 *         [             # <<<<<<<<<<<<<<
 *             TIME_ID,  # identifier
 *             b"\x00",  # approach 0
 */
  __pyx_t_3 = PyList_New(6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 329, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_v_10serializor_6binary_3ser_TIME_ID);
  __Pyx_GIVEREF(__pyx_v_10serializor_6binary_3ser_TIME_ID);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_3, 0, __pyx_v_10serializor_6binary_3ser_TIME_ID)) __PYX_ERR(0, 329, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_b__7);
  __Pyx_GIVEREF(__pyx_kp_b__7);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_3, 1, __pyx_kp_b__7)) __PYX_ERR(0, 329, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_10);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_3, 2, __pyx_t_10)) __PYX_ERR(0, 329, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_11);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_3, 3, __pyx_t_11)) __PYX_ERR(0, 329, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_4);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_3, 4, __pyx_t_4)) __PYX_ERR(0, 329, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_3, 5, __pyx_t_1)) __PYX_ERR(0, 329, __pyx_L1_error);
  __pyx_t_10 = 0;
  __pyx_t_11 = 0;
  __pyx_t_4 = 0;
  __pyx_t_1 = 0;

  /* "serializor/binary/ser.py":328
 *             )
 *     # Without timzone / Fallback
 *     return b"".join(             # <<<<<<<<<<<<<<
 *         [
 *             TIME_ID,  # identifier
 */
  __pyx_t_1 = __Pyx_PyBytes_Join(__pyx_kp_b__4, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 328, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (!(likely(PyBytes_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("bytes", __pyx_t_1))) __PYX_ERR(0, 328, __pyx_L1_error)
  __pyx_r = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "serializor/binary/ser.py":276
 * 
 * 
 * @cython.cfunc             # <<<<<<<<<<<<<<
 * @cython.inline(True)
 * def _ser_time(obj: object) -> bytes:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_AddTraceback("serializor.binary.ser._ser_time", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_name);
  __Pyx_XDECREF(__pyx_v_tz);
  __Pyx_XDECREF(__pyx_v_tz_name);
  __Pyx_XDECREF(__pyx_v_offset);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "serializor/binary/ser.py":340
 * 
 * 
 * @cython.cfunc             # <<<<<<<<<<<<<<
 * @cython.inline(True)
 * def _ser_timedelta(obj: object) -> bytes:
 */

static PyObject *__pyx_f_10serializor_6binary_3ser__ser_timedelta(PyObject *__pyx_v_obj) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_ser_timedelta", 1);

  /* "serializor/binary/ser.py":348
 *     >>> b'[id(1B)][days(4B)][seconds(4B)][microseconds(4B)]'
 *     """
 *     return b"".join(             # <<<<<<<<<<<<<<
 *         [
 *             TIMEDELTA_ID,  # identifier
 */
  __Pyx_XDECREF(__pyx_r);

  /* "serializor/binary/ser.py":351
 *         [
 *             TIMEDELTA_ID,  # identifier
 *             pack_int32(datetime.timedelta_days(obj)),  # type: ignore             # <<<<<<<<<<<<<<
 *             pack_int32(datetime.timedelta_seconds(obj)),  # type: ignore
 *             pack_int32(datetime.timedelta_microseconds(obj)),  # type: ignore
 */
  __pyx_t_1 = __pyx_f_10serializor_6binary_3ser_pack_int32(__pyx_f_7cpython_8datetime_timedelta_days(__pyx_v_obj)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 351, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "serializor/binary/ser.py":352
 *             TIMEDELTA_ID,  # identifier
 *             pack_int32(datetime.timedelta_days(obj)),  # type: ignore
 *             pack_int32(datetime.timedelta_seconds(obj)),  # type: ignore             # <<<<<<<<<<<<<<
 *             pack_int32(datetime.timedelta_microseconds(obj)),  # type: ignore
 *         ]
 */
  __pyx_t_2 = __pyx_f_10serializor_6binary_3ser_pack_int32(__pyx_f_7cpython_8datetime_timedelta_seconds(__pyx_v_obj)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 352, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);

  /* "serializor/binary/ser.py":353
 *             pack_int32(datetime.timedelta_days(obj)),  # type: ignore
 *             pack_int32(datetime.timedelta_seconds(obj)),  # type: ignore
 *             pack_int32(datetime.timedelta_microseconds(obj)),  # type: ignore             # <<<<<<<<<<<<<<
 *         ]
 *     )
 */
  __pyx_t_3 = __pyx_f_10serializor_6binary_3ser_pack_int32(__pyx_f_7cpython_8datetime_timedelta_microseconds(__pyx_v_obj)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 353, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);

  /* "serializor/binary/ser.py":349
 *     """
 *     return b"".join(
 *         [             # <<<<<<<<<<<<<<
 *             TIMEDELTA_ID,  # identifier
 *             pack_int32(datetime.timedelta_days(obj)),  # type: ignore
 */
  __pyx_t_4 = PyList_New(4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 349, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_INCREF(__pyx_v_10serializor_6binary_3ser_TIMEDELTA_ID);
  __Pyx_GIVEREF(__pyx_v_10serializor_6binary_3ser_TIMEDELTA_ID);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_4, 0, __pyx_v_10serializor_6binary_3ser_TIMEDELTA_ID)) __PYX_ERR(0, 349, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_4, 1, __pyx_t_1)) __PYX_ERR(0, 349, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_4, 2, __pyx_t_2)) __PYX_ERR(0, 349, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_3);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_4, 3, __pyx_t_3)) __PYX_ERR(0, 349, __pyx_L1_error);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;

  /* "serializor/binary/ser.py":348
 *     >>> b'[id(1B)][days(4B)][seconds(4B)][microseconds(4B)]'
 *     """
 *     return b"".join(             # <<<<<<<<<<<<<<
 *         [
 *             TIMEDELTA_ID,  # identifier
 */
  __pyx_t_3 = __Pyx_PyBytes_Join(__pyx_kp_b__4, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 348, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (!(likely(PyBytes_CheckExact(__pyx_t_3))||((__pyx_t_3) == Py_None) || __Pyx_RaiseUnexpectedTypeError("bytes", __pyx_t_3))) __PYX_ERR(0, 348, __pyx_L1_error)
  __pyx_r = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "serializor/binary/ser.py":340
 * 
 * 
 * @cython.cfunc             # <<<<<<<<<<<<<<
 * @cython.inline(True)
 * def _ser_timedelta(obj: object) -> bytes:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("serializor.binary.ser._ser_timedelta", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "serializor/binary/ser.py":358
 * 
 * 
 * @cython.cfunc             # <<<<<<<<<<<<<<
 * @cython.inline(True)
 * def _ser_struct_time(obj: struct_time) -> bytes:
 */

static PyObject *__pyx_f_10serializor_6binary_3ser__ser_struct_time(PyObject *__pyx_v_obj) {
  PyObject *__pyx_v_items = NULL;
  PyObject *__pyx_v_tm_zone = NULL;
  PyObject *__pyx_v_tm_gmtoff = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_ser_struct_time", 1);

  /* "serializor/binary/ser.py":366
 *     >>> b'[id(1B)][LIST(SER)]'
 *     """
 *     items = list(obj)             # <<<<<<<<<<<<<<
 *     if (tm_zone := obj.tm_zone) is not None:
 *         items.append(tm_zone)
 */
  __pyx_t_1 = PySequence_List(__pyx_v_obj); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 366, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_items = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "serializor/binary/ser.py":367
 *     """
 *     items = list(obj)
 *     if (tm_zone := obj.tm_zone) is not None:             # <<<<<<<<<<<<<<
 *         items.append(tm_zone)
 *     if (tm_gmtoff := obj.tm_gmtoff) is not None:
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_obj, __pyx_n_s_tm_zone); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 367, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_tm_zone = __pyx_t_1;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = (__pyx_t_1 != Py_None);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "serializor/binary/ser.py":368
 *     items = list(obj)
 *     if (tm_zone := obj.tm_zone) is not None:
 *         items.append(tm_zone)             # <<<<<<<<<<<<<<
 *     if (tm_gmtoff := obj.tm_gmtoff) is not None:
 *         items.append(tm_gmtoff)
 */
    __pyx_t_3 = __Pyx_PyList_Append(__pyx_v_items, __pyx_v_tm_zone); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(0, 368, __pyx_L1_error)

    /* "serializor/binary/ser.py":367
 *     """
 *     items = list(obj)
 *     if (tm_zone := obj.tm_zone) is not None:             # <<<<<<<<<<<<<<
 *         items.append(tm_zone)
 *     if (tm_gmtoff := obj.tm_gmtoff) is not None:
 */
  }

  /* "serializor/binary/ser.py":369
 *     if (tm_zone := obj.tm_zone) is not None:
 *         items.append(tm_zone)
 *     if (tm_gmtoff := obj.tm_gmtoff) is not None:             # <<<<<<<<<<<<<<
 *         items.append(tm_gmtoff)
 *     return STRUCT_TIME_ID + _ser_list(items)
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_obj, __pyx_n_s_tm_gmtoff); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 369, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_tm_gmtoff = __pyx_t_1;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = (__pyx_t_1 != Py_None);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "serializor/binary/ser.py":370
 *         items.append(tm_zone)
 *     if (tm_gmtoff := obj.tm_gmtoff) is not None:
 *         items.append(tm_gmtoff)             # <<<<<<<<<<<<<<
 *     return STRUCT_TIME_ID + _ser_list(items)
 * 
 */
    __pyx_t_3 = __Pyx_PyList_Append(__pyx_v_items, __pyx_v_tm_gmtoff); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(0, 370, __pyx_L1_error)

    /* "serializor/binary/ser.py":369
 *     if (tm_zone := obj.tm_zone) is not None:
 *         items.append(tm_zone)
 *     if (tm_gmtoff := obj.tm_gmtoff) is not None:             # <<<<<<<<<<<<<<
 *         items.append(tm_gmtoff)
 *     return STRUCT_TIME_ID + _ser_list(items)
 */
  }

  /* "serializor/binary/ser.py":371
 *     if (tm_gmtoff := obj.tm_gmtoff) is not None:
 *         items.append(tm_gmtoff)
 *     return STRUCT_TIME_ID + _ser_list(items)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_10serializor_6binary_3ser__ser_list(__pyx_v_items); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 371, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = PyNumber_Add(__pyx_v_10serializor_6binary_3ser_STRUCT_TIME_ID, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 371, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = ((PyObject*)__pyx_t_4);
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "serializor/binary/ser.py":358
 * 
 * 
 * @cython.cfunc             # <<<<<<<<<<<<<<
 * @cython.inline(True)
 * def _ser_struct_time(obj: struct_time) -> bytes:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("serializor.binary.ser._ser_struct_time", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_items);
  __Pyx_XDECREF(__pyx_v_tm_zone);
  __Pyx_XDECREF(__pyx_v_tm_gmtoff);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "serializor/binary/ser.py":375
 * 
 * # Numeric Types -------------------------------------------------------------------------------
 * @cython.cfunc             # <<<<<<<<<<<<<<
 * @cython.inline(True)
 * def _ser_decimal(obj: object) -> bytes:
 */

static PyObject *__pyx_f_10serializor_6binary_3ser__ser_decimal(PyObject *__pyx_v_obj) {
  PyObject *__pyx_v_val = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  Py_ssize_t __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_ser_decimal", 1);

  /* "serializor/binary/ser.py":383
 *     >>> b'[id(1B)][size(ENC)][value]'
 *     """
 *     val: bytes = utils.encode_str(str(obj), b"ascii")             # <<<<<<<<<<<<<<
 *     return gen_header(identifier.DECIMAL, bytes_len(val)) + val  # type: ignore
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_Unicode(__pyx_v_obj); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 383, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __pyx_f_10serializor_5utils_encode_str(__pyx_t_1, ((char *)"ascii")); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 383, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_val = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "serializor/binary/ser.py":384
 *     """
 *     val: bytes = utils.encode_str(str(obj), b"ascii")
 *     return gen_header(identifier.DECIMAL, bytes_len(val)) + val  # type: ignore             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_3 = PyBytes_Size(__pyx_v_val); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1L))) __PYX_ERR(0, 384, __pyx_L1_error)
  __pyx_t_2 = __pyx_f_10serializor_6binary_3ser_gen_header(__pyx_v_10serializor_10identifier_DECIMAL, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 384, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = PyNumber_Add(__pyx_t_2, __pyx_v_val); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 384, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "serializor/binary/ser.py":375
 * 
 * # Numeric Types -------------------------------------------------------------------------------
 * @cython.cfunc             # <<<<<<<<<<<<<<
 * @cython.inline(True)
 * def _ser_decimal(obj: object) -> bytes:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("serializor.binary.ser._ser_decimal", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_val);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "serializor/binary/ser.py":387
 * 
 * 
 * @cython.cfunc             # <<<<<<<<<<<<<<
 * @cython.inline(True)
 * def _ser_complex(obj: object) -> bytes:
 */

static PyObject *__pyx_f_10serializor_6binary_3ser__ser_complex(PyObject *__pyx_v_obj) {
  PyObject *__pyx_v_real_val = 0;
  PyObject *__pyx_v_real_size = 0;
  PyObject *__pyx_v_imag_val = 0;
  PyObject *__pyx_v_imag_size = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  double __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_ser_complex", 1);

  /* "serializor/binary/ser.py":395
 *     >>> b'[id(1B)][real_size(ENC)][real][imag_size(ENC)][imag]'
 *     """
 *     real_val: bytes = _orjson_dumps(complex_getreal(obj))             # <<<<<<<<<<<<<<
 *     real_size: bytes = gen_encoded_int(bytes_len(real_val))  # type: ignore
 *     imag_val: bytes = _orjson_dumps(complex_getimag(obj))
 */
  __pyx_t_1 = PyComplex_RealAsDouble(__pyx_v_obj); if (unlikely(__pyx_t_1 == ((double)-1.0) && PyErr_Occurred())) __PYX_ERR(0, 395, __pyx_L1_error)
  __pyx_t_2 = PyFloat_FromDouble(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 395, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __pyx_f_10serializor_6binary_3ser__orjson_dumps(__pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 395, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_real_val = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "serializor/binary/ser.py":396
 *     """
 *     real_val: bytes = _orjson_dumps(complex_getreal(obj))
 *     real_size: bytes = gen_encoded_int(bytes_len(real_val))  # type: ignore             # <<<<<<<<<<<<<<
 *     imag_val: bytes = _orjson_dumps(complex_getimag(obj))
 *     imag_size: bytes = gen_encoded_int(bytes_len(imag_val))  # type: ignore
 */
  __pyx_t_4 = PyBytes_Size(__pyx_v_real_val); if (unlikely(__pyx_t_4 == ((Py_ssize_t)-1L))) __PYX_ERR(0, 396, __pyx_L1_error)
  __pyx_t_3 = __pyx_f_10serializor_6binary_3ser_gen_encoded_int(__pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 396, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_v_real_size = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "serializor/binary/ser.py":397
 *     real_val: bytes = _orjson_dumps(complex_getreal(obj))
 *     real_size: bytes = gen_encoded_int(bytes_len(real_val))  # type: ignore
 *     imag_val: bytes = _orjson_dumps(complex_getimag(obj))             # <<<<<<<<<<<<<<
 *     imag_size: bytes = gen_encoded_int(bytes_len(imag_val))  # type: ignore
 *     return b"".join([COMPLEX_ID, real_size, real_val, imag_size, imag_val])
 */
  __pyx_t_1 = PyComplex_ImagAsDouble(__pyx_v_obj); if (unlikely(__pyx_t_1 == ((double)-1.0) && PyErr_Occurred())) __PYX_ERR(0, 397, __pyx_L1_error)
  __pyx_t_3 = PyFloat_FromDouble(__pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 397, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __pyx_f_10serializor_6binary_3ser__orjson_dumps(__pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 397, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_imag_val = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "serializor/binary/ser.py":398
 *     real_size: bytes = gen_encoded_int(bytes_len(real_val))  # type: ignore
 *     imag_val: bytes = _orjson_dumps(complex_getimag(obj))
 *     imag_size: bytes = gen_encoded_int(bytes_len(imag_val))  # type: ignore             # <<<<<<<<<<<<<<
 *     return b"".join([COMPLEX_ID, real_size, real_val, imag_size, imag_val])
 * 
 */
  __pyx_t_4 = PyBytes_Size(__pyx_v_imag_val); if (unlikely(__pyx_t_4 == ((Py_ssize_t)-1L))) __PYX_ERR(0, 398, __pyx_L1_error)
  __pyx_t_2 = __pyx_f_10serializor_6binary_3ser_gen_encoded_int(__pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 398, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_imag_size = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "serializor/binary/ser.py":399
 *     imag_val: bytes = _orjson_dumps(complex_getimag(obj))
 *     imag_size: bytes = gen_encoded_int(bytes_len(imag_val))  # type: ignore
 *     return b"".join([COMPLEX_ID, real_size, real_val, imag_size, imag_val])             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = PyList_New(5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 399, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_v_10serializor_6binary_3ser_COMPLEX_ID);
  __Pyx_GIVEREF(__pyx_v_10serializor_6binary_3ser_COMPLEX_ID);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_2, 0, __pyx_v_10serializor_6binary_3ser_COMPLEX_ID)) __PYX_ERR(0, 399, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_real_size);
  __Pyx_GIVEREF(__pyx_v_real_size);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_2, 1, __pyx_v_real_size)) __PYX_ERR(0, 399, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_real_val);
  __Pyx_GIVEREF(__pyx_v_real_val);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_2, 2, __pyx_v_real_val)) __PYX_ERR(0, 399, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_imag_size);
  __Pyx_GIVEREF(__pyx_v_imag_size);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_2, 3, __pyx_v_imag_size)) __PYX_ERR(0, 399, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_imag_val);
  __Pyx_GIVEREF(__pyx_v_imag_val);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_2, 4, __pyx_v_imag_val)) __PYX_ERR(0, 399, __pyx_L1_error);
  __pyx_t_3 = __Pyx_PyBytes_Join(__pyx_kp_b__4, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 399, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (!(likely(PyBytes_CheckExact(__pyx_t_3))||((__pyx_t_3) == Py_None) || __Pyx_RaiseUnexpectedTypeError("bytes", __pyx_t_3))) __PYX_ERR(0, 399, __pyx_L1_error)
  __pyx_r = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "serializor/binary/ser.py":387
 * 
 * 
 * @cython.cfunc             # <<<<<<<<<<<<<<
 * @cython.inline(True)
 * def _ser_complex(obj: object) -> bytes:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("serializor.binary.ser._ser_complex", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_real_val);
  __Pyx_XDECREF(__pyx_v_real_size);
  __Pyx_XDECREF(__pyx_v_imag_val);
  __Pyx_XDECREF(__pyx_v_imag_size);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "serializor/binary/ser.py":403
 * 
 * # Bytes Types ---------------------------------------------------------------------------------
 * @cython.cfunc             # <<<<<<<<<<<<<<
 * @cython.inline(True)
 * def _ser_bytes(obj: object) -> bytes:
 */

static PyObject *__pyx_f_10serializor_6binary_3ser__ser_bytes(PyObject *__pyx_v_obj) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  Py_ssize_t __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_ser_bytes", 1);

  /* "serializor/binary/ser.py":411
 *     >>> b'[id(1B)][size(ENC)][value]'
 *     """
 *     return gen_header(identifier.BYTES, bytes_len(obj)) + obj  # type: ignore             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyBytes_Size(__pyx_v_obj); if (unlikely(__pyx_t_1 == ((Py_ssize_t)-1L))) __PYX_ERR(0, 411, __pyx_L1_error)
  __pyx_t_2 = __pyx_f_10serializor_6binary_3ser_gen_header(__pyx_v_10serializor_10identifier_BYTES, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 411, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyNumber_Add(__pyx_t_2, __pyx_v_obj); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 411, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (!(likely(PyBytes_CheckExact(__pyx_t_3))||((__pyx_t_3) == Py_None) || __Pyx_RaiseUnexpectedTypeError("bytes", __pyx_t_3))) __PYX_ERR(0, 411, __pyx_L1_error)
  __pyx_r = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "serializor/binary/ser.py":403
 * 
 * # Bytes Types ---------------------------------------------------------------------------------
 * @cython.cfunc             # <<<<<<<<<<<<<<
 * @cython.inline(True)
 * def _ser_bytes(obj: object) -> bytes:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("serializor.binary.ser._ser_bytes", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "serializor/binary/ser.py":414
 * 
 * 
 * @cython.cfunc             # <<<<<<<<<<<<<<
 * @cython.inline(True)
 * def _ser_bytearray(obj: object) -> bytes:
 */

static PyObject *__pyx_f_10serializor_6binary_3ser__ser_bytearray(PyObject *__pyx_v_obj) {
  PyObject *__pyx_v_val = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_ser_bytearray", 1);

  /* "serializor/binary/ser.py":422
 *     >>> b'[id(1B)][size(ENC)][value]'
 *     """
 *     val = bytes(obj)             # <<<<<<<<<<<<<<
 *     return gen_header(identifier.BYTEARRAY, bytes_len(val)) + val  # type: ignore
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyBytes_Type)), __pyx_v_obj); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 422, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_val = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "serializor/binary/ser.py":423
 *     """
 *     val = bytes(obj)
 *     return gen_header(identifier.BYTEARRAY, bytes_len(val)) + val  # type: ignore             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = PyBytes_Size(__pyx_v_val); if (unlikely(__pyx_t_2 == ((Py_ssize_t)-1L))) __PYX_ERR(0, 423, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_10serializor_6binary_3ser_gen_header(__pyx_v_10serializor_10identifier_BYTEARRAY, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 423, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = PyNumber_Add(__pyx_t_1, __pyx_v_val); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 423, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "serializor/binary/ser.py":414
 * 
 * 
 * @cython.cfunc             # <<<<<<<<<<<<<<
 * @cython.inline(True)
 * def _ser_bytearray(obj: object) -> bytes:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("serializor.binary.ser._ser_bytearray", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_val);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "serializor/binary/ser.py":426
 * 
 * 
 * @cython.cfunc             # <<<<<<<<<<<<<<
 * @cython.inline(True)
 * def _ser_memoryview(obj: memoryview) -> bytes:
 */

static PyObject *__pyx_f_10serializor_6binary_3ser__ser_memoryview(PyObject *__pyx_v_obj) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_ser_memoryview", 1);

  /* "serializor/binary/ser.py":434
 *     then uses '_ser_bytes()' function to serialize it.
 *     """
 *     return _ser_bytes(obj.tobytes())             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_CallUnboundCMethod0(&__pyx_umethod_PyMemoryView_Type_tobytes, __pyx_v_obj); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 434, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __pyx_f_10serializor_6binary_3ser__ser_bytes(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 434, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "serializor/binary/ser.py":426
 * 
 * 
 * @cython.cfunc             # <<<<<<<<<<<<<<
 * @cython.inline(True)
 * def _ser_memoryview(obj: memoryview) -> bytes:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("serializor.binary.ser._ser_memoryview", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "serializor/binary/ser.py":438
 * 
 * # Sequence Types ------------------------------------------------------------------------------
 * @cython.cfunc             # <<<<<<<<<<<<<<
 * @cython.inline(True)
 * def _ser_list(obj: list) -> bytes:
 */

static PyObject *__pyx_f_10serializor_6binary_3ser__ser_list(PyObject *__pyx_v_obj) {
  Py_ssize_t __pyx_v_size;
  PyObject *__pyx_v_items = NULL;
  PyObject *__pyx_v_i = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_ser_list", 1);

  /* "serializor/binary/ser.py":447
 *     """
 *     # Get sequence size
 *     size: cython.Py_ssize_t = list_len(obj)             # <<<<<<<<<<<<<<
 *     if size == 0:
 *         return gen_header(identifier.LIST, 0)  # type: ignore
 */
  __pyx_v_size = PyList_GET_SIZE(__pyx_v_obj);

  /* "serializor/binary/ser.py":448
 *     # Get sequence size
 *     size: cython.Py_ssize_t = list_len(obj)
 *     if size == 0:             # <<<<<<<<<<<<<<
 *         return gen_header(identifier.LIST, 0)  # type: ignore
 *     # Serialize items
 */
  __pyx_t_1 = (__pyx_v_size == 0);
  if (__pyx_t_1) {

    /* "serializor/binary/ser.py":449
 *     size: cython.Py_ssize_t = list_len(obj)
 *     if size == 0:
 *         return gen_header(identifier.LIST, 0)  # type: ignore             # <<<<<<<<<<<<<<
 *     # Serialize items
 *     items = [LIST_ID, gen_encoded_int(size)]  # type: ignore
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = __pyx_f_10serializor_6binary_3ser_gen_header(__pyx_v_10serializor_10identifier_LIST, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 449, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_r = ((PyObject*)__pyx_t_2);
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "serializor/binary/ser.py":448
 *     # Get sequence size
 *     size: cython.Py_ssize_t = list_len(obj)
 *     if size == 0:             # <<<<<<<<<<<<<<
 *         return gen_header(identifier.LIST, 0)  # type: ignore
 *     # Serialize items
 */
  }

  /* "serializor/binary/ser.py":451
 *         return gen_header(identifier.LIST, 0)  # type: ignore
 *     # Serialize items
 *     items = [LIST_ID, gen_encoded_int(size)]  # type: ignore             # <<<<<<<<<<<<<<
 *     for i in obj:
 *         items.append(_ser_common(i))
 */
  __pyx_t_2 = __pyx_f_10serializor_6binary_3ser_gen_encoded_int(__pyx_v_size); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 451, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyList_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 451, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_v_10serializor_6binary_3ser_LIST_ID);
  __Pyx_GIVEREF(__pyx_v_10serializor_6binary_3ser_LIST_ID);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_3, 0, __pyx_v_10serializor_6binary_3ser_LIST_ID)) __PYX_ERR(0, 451, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_3, 1, __pyx_t_2)) __PYX_ERR(0, 451, __pyx_L1_error);
  __pyx_t_2 = 0;
  __pyx_v_items = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "serializor/binary/ser.py":452
 *     # Serialize items
 *     items = [LIST_ID, gen_encoded_int(size)]  # type: ignore
 *     for i in obj:             # <<<<<<<<<<<<<<
 *         items.append(_ser_common(i))
 *     return b"".join(items)
 */
  __pyx_t_3 = __pyx_v_obj; __Pyx_INCREF(__pyx_t_3);
  __pyx_t_4 = 0;
  for (;;) {
    {
      Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_3);
      #if !CYTHON_ASSUME_SAFE_MACROS
      if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 452, __pyx_L1_error)
      #endif
      if (__pyx_t_4 >= __pyx_temp) break;
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_2 = PyList_GET_ITEM(__pyx_t_3, __pyx_t_4); __Pyx_INCREF(__pyx_t_2); __pyx_t_4++; if (unlikely((0 < 0))) __PYX_ERR(0, 452, __pyx_L1_error)
    #else
    __pyx_t_2 = __Pyx_PySequence_ITEM(__pyx_t_3, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 452, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    #endif
    __Pyx_XDECREF_SET(__pyx_v_i, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "serializor/binary/ser.py":453
 *     items = [LIST_ID, gen_encoded_int(size)]  # type: ignore
 *     for i in obj:
 *         items.append(_ser_common(i))             # <<<<<<<<<<<<<<
 *     return b"".join(items)
 * 
 */
    __pyx_t_2 = __pyx_f_10serializor_6binary_3ser__ser_common(__pyx_v_i); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 453, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = __Pyx_PyList_Append(__pyx_v_items, __pyx_t_2); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(0, 453, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "serializor/binary/ser.py":452
 *     # Serialize items
 *     items = [LIST_ID, gen_encoded_int(size)]  # type: ignore
 *     for i in obj:             # <<<<<<<<<<<<<<
 *         items.append(_ser_common(i))
 *     return b"".join(items)
 */
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "serializor/binary/ser.py":454
 *     for i in obj:
 *         items.append(_ser_common(i))
 *     return b"".join(items)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_3 = __Pyx_PyBytes_Join(__pyx_kp_b__4, __pyx_v_items); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 454, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (!(likely(PyBytes_CheckExact(__pyx_t_3))||((__pyx_t_3) == Py_None) || __Pyx_RaiseUnexpectedTypeError("bytes", __pyx_t_3))) __PYX_ERR(0, 454, __pyx_L1_error)
  __pyx_r = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "serializor/binary/ser.py":438
 * 
 * # Sequence Types ------------------------------------------------------------------------------
 * @cython.cfunc             # <<<<<<<<<<<<<<
 * @cython.inline(True)
 * def _ser_list(obj: list) -> bytes:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("serializor.binary.ser._ser_list", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_items);
  __Pyx_XDECREF(__pyx_v_i);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "serializor/binary/ser.py":457
 * 
 * 
 * @cython.cfunc             # <<<<<<<<<<<<<<
 * @cython.inline(True)
 * def _ser_tuple(obj: tuple) -> bytes:
 */

static PyObject *__pyx_f_10serializor_6binary_3ser__ser_tuple(PyObject *__pyx_v_obj) {
  Py_ssize_t __pyx_v_size;
  PyObject *__pyx_v_items = NULL;
  PyObject *__pyx_v_i = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_ser_tuple", 1);

  /* "serializor/binary/ser.py":466
 *     """
 *     # Get sequence size
 *     size: cython.Py_ssize_t = tuple_len(obj)             # <<<<<<<<<<<<<<
 *     if size == 0:
 *         return gen_header(identifier.TUPLE, 0)  # type: ignore
 */
  __pyx_v_size = PyTuple_GET_SIZE(__pyx_v_obj);

  /* "serializor/binary/ser.py":467
 *     # Get sequence size
 *     size: cython.Py_ssize_t = tuple_len(obj)
 *     if size == 0:             # <<<<<<<<<<<<<<
 *         return gen_header(identifier.TUPLE, 0)  # type: ignore
 *     # Serialize items
 */
  __pyx_t_1 = (__pyx_v_size == 0);
  if (__pyx_t_1) {

    /* "serializor/binary/ser.py":468
 *     size: cython.Py_ssize_t = tuple_len(obj)
 *     if size == 0:
 *         return gen_header(identifier.TUPLE, 0)  # type: ignore             # <<<<<<<<<<<<<<
 *     # Serialize items
 *     items = [TUPLE_ID, gen_encoded_int(size)]  # type: ignore
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = __pyx_f_10serializor_6binary_3ser_gen_header(__pyx_v_10serializor_10identifier_TUPLE, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 468, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_r = ((PyObject*)__pyx_t_2);
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "serializor/binary/ser.py":467
 *     # Get sequence size
 *     size: cython.Py_ssize_t = tuple_len(obj)
 *     if size == 0:             # <<<<<<<<<<<<<<
 *         return gen_header(identifier.TUPLE, 0)  # type: ignore
 *     # Serialize items
 */
  }

  /* "serializor/binary/ser.py":470
 *         return gen_header(identifier.TUPLE, 0)  # type: ignore
 *     # Serialize items
 *     items = [TUPLE_ID, gen_encoded_int(size)]  # type: ignore             # <<<<<<<<<<<<<<
 *     for i in obj:
 *         items.append(_ser_common(i))
 */
  __pyx_t_2 = __pyx_f_10serializor_6binary_3ser_gen_encoded_int(__pyx_v_size); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 470, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyList_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 470, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_v_10serializor_6binary_3ser_TUPLE_ID);
  __Pyx_GIVEREF(__pyx_v_10serializor_6binary_3ser_TUPLE_ID);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_3, 0, __pyx_v_10serializor_6binary_3ser_TUPLE_ID)) __PYX_ERR(0, 470, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_3, 1, __pyx_t_2)) __PYX_ERR(0, 470, __pyx_L1_error);
  __pyx_t_2 = 0;
  __pyx_v_items = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "serializor/binary/ser.py":471
 *     # Serialize items
 *     items = [TUPLE_ID, gen_encoded_int(size)]  # type: ignore
 *     for i in obj:             # <<<<<<<<<<<<<<
 *         items.append(_ser_common(i))
 *     return b"".join(items)
 */
  __pyx_t_3 = __pyx_v_obj; __Pyx_INCREF(__pyx_t_3);
  __pyx_t_4 = 0;
  for (;;) {
    {
      Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_3);
      #if !CYTHON_ASSUME_SAFE_MACROS
      if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 471, __pyx_L1_error)
      #endif
      if (__pyx_t_4 >= __pyx_temp) break;
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_2 = PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_4); __Pyx_INCREF(__pyx_t_2); __pyx_t_4++; if (unlikely((0 < 0))) __PYX_ERR(0, 471, __pyx_L1_error)
    #else
    __pyx_t_2 = __Pyx_PySequence_ITEM(__pyx_t_3, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 471, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    #endif
    __Pyx_XDECREF_SET(__pyx_v_i, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "serializor/binary/ser.py":472
 *     items = [TUPLE_ID, gen_encoded_int(size)]  # type: ignore
 *     for i in obj:
 *         items.append(_ser_common(i))             # <<<<<<<<<<<<<<
 *     return b"".join(items)
 * 
 */
    __pyx_t_2 = __pyx_f_10serializor_6binary_3ser__ser_common(__pyx_v_i); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 472, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = __Pyx_PyList_Append(__pyx_v_items, __pyx_t_2); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(0, 472, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "serializor/binary/ser.py":471
 *     # Serialize items
 *     items = [TUPLE_ID, gen_encoded_int(size)]  # type: ignore
 *     for i in obj:             # <<<<<<<<<<<<<<
 *         items.append(_ser_common(i))
 *     return b"".join(items)
 */
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "serializor/binary/ser.py":473
 *     for i in obj:
 *         items.append(_ser_common(i))
 *     return b"".join(items)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_3 = __Pyx_PyBytes_Join(__pyx_kp_b__4, __pyx_v_items); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 473, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (!(likely(PyBytes_CheckExact(__pyx_t_3))||((__pyx_t_3) == Py_None) || __Pyx_RaiseUnexpectedTypeError("bytes", __pyx_t_3))) __PYX_ERR(0, 473, __pyx_L1_error)
  __pyx_r = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "serializor/binary/ser.py":457
 * 
 * 
 * @cython.cfunc             # <<<<<<<<<<<<<<
 * @cython.inline(True)
 * def _ser_tuple(obj: tuple) -> bytes:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("serializor.binary.ser._ser_tuple", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_items);
  __Pyx_XDECREF(__pyx_v_i);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "serializor/binary/ser.py":476
 * 
 * 
 * @cython.cfunc             # <<<<<<<<<<<<<<
 * @cython.inline(True)
 * def _ser_set(obj: set) -> bytes:
 */

static PyObject *__pyx_f_10serializor_6binary_3ser__ser_set(PyObject *__pyx_v_obj) {
  Py_ssize_t __pyx_v_size;
  PyObject *__pyx_v_items = NULL;
  PyObject *__pyx_v_i = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  int __pyx_t_6;
  int __pyx_t_7;
  int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_ser_set", 1);

  /* "serializor/binary/ser.py":485
 *     """
 *     # Get sequence size
 *     size: cython.Py_ssize_t = set_len(obj)             # <<<<<<<<<<<<<<
 *     if size == 0:
 *         return gen_header(identifier.SET, 0)  # type: ignore
 */
  __pyx_v_size = PySet_GET_SIZE(__pyx_v_obj);

  /* "serializor/binary/ser.py":486
 *     # Get sequence size
 *     size: cython.Py_ssize_t = set_len(obj)
 *     if size == 0:             # <<<<<<<<<<<<<<
 *         return gen_header(identifier.SET, 0)  # type: ignore
 *     # Serialize items
 */
  __pyx_t_1 = (__pyx_v_size == 0);
  if (__pyx_t_1) {

    /* "serializor/binary/ser.py":487
 *     size: cython.Py_ssize_t = set_len(obj)
 *     if size == 0:
 *         return gen_header(identifier.SET, 0)  # type: ignore             # <<<<<<<<<<<<<<
 *     # Serialize items
 *     items = [SET_ID, gen_encoded_int(size)]  # type: ignore
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = __pyx_f_10serializor_6binary_3ser_gen_header(__pyx_v_10serializor_10identifier_SET, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 487, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_r = ((PyObject*)__pyx_t_2);
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "serializor/binary/ser.py":486
 *     # Get sequence size
 *     size: cython.Py_ssize_t = set_len(obj)
 *     if size == 0:             # <<<<<<<<<<<<<<
 *         return gen_header(identifier.SET, 0)  # type: ignore
 *     # Serialize items
 */
  }

  /* "serializor/binary/ser.py":489
 *         return gen_header(identifier.SET, 0)  # type: ignore
 *     # Serialize items
 *     items = [SET_ID, gen_encoded_int(size)]  # type: ignore             # <<<<<<<<<<<<<<
 *     for i in obj:
 *         items.append(_ser_common(i))
 */
  __pyx_t_2 = __pyx_f_10serializor_6binary_3ser_gen_encoded_int(__pyx_v_size); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 489, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyList_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 489, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_v_10serializor_6binary_3ser_SET_ID);
  __Pyx_GIVEREF(__pyx_v_10serializor_6binary_3ser_SET_ID);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_3, 0, __pyx_v_10serializor_6binary_3ser_SET_ID)) __PYX_ERR(0, 489, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_3, 1, __pyx_t_2)) __PYX_ERR(0, 489, __pyx_L1_error);
  __pyx_t_2 = 0;
  __pyx_v_items = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "serializor/binary/ser.py":490
 *     # Serialize items
 *     items = [SET_ID, gen_encoded_int(size)]  # type: ignore
 *     for i in obj:             # <<<<<<<<<<<<<<
 *         items.append(_ser_common(i))
 *     return b"".join(items)
 */
  __pyx_t_4 = 0;
  __pyx_t_2 = __Pyx_set_iterator(__pyx_v_obj, 1, (&__pyx_t_5), (&__pyx_t_6)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 490, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __pyx_t_3 = __pyx_t_2;
  __pyx_t_2 = 0;
  while (1) {
    __pyx_t_7 = __Pyx_set_iter_next(__pyx_t_3, __pyx_t_5, &__pyx_t_4, &__pyx_t_2, __pyx_t_6);
    if (unlikely(__pyx_t_7 == 0)) break;
    if (unlikely(__pyx_t_7 == -1)) __PYX_ERR(0, 490, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_XDECREF_SET(__pyx_v_i, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "serializor/binary/ser.py":491
 *     items = [SET_ID, gen_encoded_int(size)]  # type: ignore
 *     for i in obj:
 *         items.append(_ser_common(i))             # <<<<<<<<<<<<<<
 *     return b"".join(items)
 * 
 */
    __pyx_t_2 = __pyx_f_10serializor_6binary_3ser__ser_common(__pyx_v_i); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 491, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_8 = __Pyx_PyList_Append(__pyx_v_items, __pyx_t_2); if (unlikely(__pyx_t_8 == ((int)-1))) __PYX_ERR(0, 491, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "serializor/binary/ser.py":492
 *     for i in obj:
 *         items.append(_ser_common(i))
 *     return b"".join(items)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_3 = __Pyx_PyBytes_Join(__pyx_kp_b__4, __pyx_v_items); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 492, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (!(likely(PyBytes_CheckExact(__pyx_t_3))||((__pyx_t_3) == Py_None) || __Pyx_RaiseUnexpectedTypeError("bytes", __pyx_t_3))) __PYX_ERR(0, 492, __pyx_L1_error)
  __pyx_r = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "serializor/binary/ser.py":476
 * 
 * 
 * @cython.cfunc             # <<<<<<<<<<<<<<
 * @cython.inline(True)
 * def _ser_set(obj: set) -> bytes:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("serializor.binary.ser._ser_set", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_items);
  __Pyx_XDECREF(__pyx_v_i);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "serializor/binary/ser.py":495
 * 
 * 
 * @cython.cfunc             # <<<<<<<<<<<<<<
 * @cython.inline(True)
 * def _ser_frozenset(obj: frozenset) -> bytes:
 */

static PyObject *__pyx_f_10serializor_6binary_3ser__ser_frozenset(PyObject *__pyx_v_obj) {
  Py_ssize_t __pyx_v_size;
  PyObject *__pyx_v_items = NULL;
  PyObject *__pyx_v_i = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  int __pyx_t_6;
  int __pyx_t_7;
  int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_ser_frozenset", 1);

  /* "serializor/binary/ser.py":504
 *     """
 *     # Get sequence size
 *     size: cython.Py_ssize_t = set_len(obj)             # <<<<<<<<<<<<<<
 *     if size == 0:
 *         return gen_header(identifier.FROZENSET, 0)  # type: ignore
 */
  __pyx_v_size = PySet_GET_SIZE(__pyx_v_obj);

  /* "serializor/binary/ser.py":505
 *     # Get sequence size
 *     size: cython.Py_ssize_t = set_len(obj)
 *     if size == 0:             # <<<<<<<<<<<<<<
 *         return gen_header(identifier.FROZENSET, 0)  # type: ignore
 *     # Serialize items
 */
  __pyx_t_1 = (__pyx_v_size == 0);
  if (__pyx_t_1) {

    /* "serializor/binary/ser.py":506
 *     size: cython.Py_ssize_t = set_len(obj)
 *     if size == 0:
 *         return gen_header(identifier.FROZENSET, 0)  # type: ignore             # <<<<<<<<<<<<<<
 *     # Serialize items
 *     items = [FROZENSET_ID, gen_encoded_int(size)]  # type: ignore
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = __pyx_f_10serializor_6binary_3ser_gen_header(__pyx_v_10serializor_10identifier_FROZENSET, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 506, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_r = ((PyObject*)__pyx_t_2);
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "serializor/binary/ser.py":505
 *     # Get sequence size
 *     size: cython.Py_ssize_t = set_len(obj)
 *     if size == 0:             # <<<<<<<<<<<<<<
 *         return gen_header(identifier.FROZENSET, 0)  # type: ignore
 *     # Serialize items
 */
  }

  /* "serializor/binary/ser.py":508
 *         return gen_header(identifier.FROZENSET, 0)  # type: ignore
 *     # Serialize items
 *     items = [FROZENSET_ID, gen_encoded_int(size)]  # type: ignore             # <<<<<<<<<<<<<<
 *     for i in obj:
 *         items.append(_ser_common(i))
 */
  __pyx_t_2 = __pyx_f_10serializor_6binary_3ser_gen_encoded_int(__pyx_v_size); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 508, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyList_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 508, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_v_10serializor_6binary_3ser_FROZENSET_ID);
  __Pyx_GIVEREF(__pyx_v_10serializor_6binary_3ser_FROZENSET_ID);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_3, 0, __pyx_v_10serializor_6binary_3ser_FROZENSET_ID)) __PYX_ERR(0, 508, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_3, 1, __pyx_t_2)) __PYX_ERR(0, 508, __pyx_L1_error);
  __pyx_t_2 = 0;
  __pyx_v_items = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "serializor/binary/ser.py":509
 *     # Serialize items
 *     items = [FROZENSET_ID, gen_encoded_int(size)]  # type: ignore
 *     for i in obj:             # <<<<<<<<<<<<<<
 *         items.append(_ser_common(i))
 *     return b"".join(items)
 */
  __pyx_t_4 = 0;
  __pyx_t_2 = __Pyx_set_iterator(__pyx_v_obj, 0, (&__pyx_t_5), (&__pyx_t_6)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 509, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __pyx_t_3 = __pyx_t_2;
  __pyx_t_2 = 0;
  while (1) {
    __pyx_t_7 = __Pyx_set_iter_next(__pyx_t_3, __pyx_t_5, &__pyx_t_4, &__pyx_t_2, __pyx_t_6);
    if (unlikely(__pyx_t_7 == 0)) break;
    if (unlikely(__pyx_t_7 == -1)) __PYX_ERR(0, 509, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_XDECREF_SET(__pyx_v_i, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "serializor/binary/ser.py":510
 *     items = [FROZENSET_ID, gen_encoded_int(size)]  # type: ignore
 *     for i in obj:
 *         items.append(_ser_common(i))             # <<<<<<<<<<<<<<
 *     return b"".join(items)
 * 
 */
    __pyx_t_2 = __pyx_f_10serializor_6binary_3ser__ser_common(__pyx_v_i); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 510, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_8 = __Pyx_PyList_Append(__pyx_v_items, __pyx_t_2); if (unlikely(__pyx_t_8 == ((int)-1))) __PYX_ERR(0, 510, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "serializor/binary/ser.py":511
 *     for i in obj:
 *         items.append(_ser_common(i))
 *     return b"".join(items)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_3 = __Pyx_PyBytes_Join(__pyx_kp_b__4, __pyx_v_items); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 511, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (!(likely(PyBytes_CheckExact(__pyx_t_3))||((__pyx_t_3) == Py_None) || __Pyx_RaiseUnexpectedTypeError("bytes", __pyx_t_3))) __PYX_ERR(0, 511, __pyx_L1_error)
  __pyx_r = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "serializor/binary/ser.py":495
 * 
 * 
 * @cython.cfunc             # <<<<<<<<<<<<<<
 * @cython.inline(True)
 * def _ser_frozenset(obj: frozenset) -> bytes:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("serializor.binary.ser._ser_frozenset", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_items);
  __Pyx_XDECREF(__pyx_v_i);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "serializor/binary/ser.py":514
 * 
 * 
 * @cython.cfunc             # <<<<<<<<<<<<<<
 * @cython.inline(True)
 * def _ser_range(obj: object) -> bytes:
 */

static PyObject *__pyx_f_10serializor_6binary_3ser__ser_range(PyObject *__pyx_v_obj) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_ser_range", 1);

  /* "serializor/binary/ser.py":522
 *     >>> b'[id(1B)][start(SER)][stop(SER)][step(SER)]'
 *     """
 *     return b"".join(             # <<<<<<<<<<<<<<
 *         [
 *             RANGE_ID,  # identifier
 */
  __Pyx_XDECREF(__pyx_r);

  /* "serializor/binary/ser.py":525
 *         [
 *             RANGE_ID,  # identifier
 *             _ser_int(obj.start),  # start             # <<<<<<<<<<<<<<
 *             _ser_int(obj.stop),  # stop
 *             _ser_int(obj.step),  # step
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_obj, __pyx_n_s_start); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 525, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __pyx_f_10serializor_6binary_3ser__ser_int(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 525, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "serializor/binary/ser.py":526
 *             RANGE_ID,  # identifier
 *             _ser_int(obj.start),  # start
 *             _ser_int(obj.stop),  # stop             # <<<<<<<<<<<<<<
 *             _ser_int(obj.step),  # step
 *         ]
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_obj, __pyx_n_s_stop); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 526, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __pyx_f_10serializor_6binary_3ser__ser_int(__pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 526, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "serializor/binary/ser.py":527
 *             _ser_int(obj.start),  # start
 *             _ser_int(obj.stop),  # stop
 *             _ser_int(obj.step),  # step             # <<<<<<<<<<<<<<
 *         ]
 *     )
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_obj, __pyx_n_s_step); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 527, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = __pyx_f_10serializor_6binary_3ser__ser_int(__pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 527, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "serializor/binary/ser.py":523
 *     """
 *     return b"".join(
 *         [             # <<<<<<<<<<<<<<
 *             RANGE_ID,  # identifier
 *             _ser_int(obj.start),  # start
 */
  __pyx_t_1 = PyList_New(4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 523, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_10serializor_6binary_3ser_RANGE_ID);
  __Pyx_GIVEREF(__pyx_v_10serializor_6binary_3ser_RANGE_ID);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_v_10serializor_6binary_3ser_RANGE_ID)) __PYX_ERR(0, 523, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 1, __pyx_t_2)) __PYX_ERR(0, 523, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_3);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 2, __pyx_t_3)) __PYX_ERR(0, 523, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_4);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 3, __pyx_t_4)) __PYX_ERR(0, 523, __pyx_L1_error);
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;

  /* "serializor/binary/ser.py":522
 *     >>> b'[id(1B)][start(SER)][stop(SER)][step(SER)]'
 *     """
 *     return b"".join(             # <<<<<<<<<<<<<<
 *         [
 *             RANGE_ID,  # identifier
 */
  __pyx_t_4 = __Pyx_PyBytes_Join(__pyx_kp_b__4, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 522, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (!(likely(PyBytes_CheckExact(__pyx_t_4))||((__pyx_t_4) == Py_None) || __Pyx_RaiseUnexpectedTypeError("bytes", __pyx_t_4))) __PYX_ERR(0, 522, __pyx_L1_error)
  __pyx_r = ((PyObject*)__pyx_t_4);
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "serializor/binary/ser.py":514
 * 
 * 
 * @cython.cfunc             # <<<<<<<<<<<<<<
 * @cython.inline(True)
 * def _ser_range(obj: object) -> bytes:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("serializor.binary.ser._ser_range", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "serializor/binary/ser.py":532
 * 
 * 
 * @cython.cfunc             # <<<<<<<<<<<<<<
 * @cython.inline(True)
 * def _ser_sequence(obj: Iterable) -> bytes:
 */

static PyObject *__pyx_f_10serializor_6binary_3ser__ser_sequence(PyObject *__pyx_v_obj) {
  Py_ssize_t __pyx_v_size;
  PyObject *__pyx_v_items = NULL;
  PyObject *__pyx_v_i = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  Py_ssize_t __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *(*__pyx_t_5)(PyObject *);
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_ser_sequence", 1);

  /* "serializor/binary/ser.py":546
 *     """
 *     # Get sequence size
 *     size: cython.Py_ssize_t = len(obj)             # <<<<<<<<<<<<<<
 *     if size == 0:
 *         return gen_header(identifier.LIST, 0)  # type: ignore
 */
  __pyx_t_1 = PyObject_Length(__pyx_v_obj); if (unlikely(__pyx_t_1 == ((Py_ssize_t)-1))) __PYX_ERR(0, 546, __pyx_L1_error)
  __pyx_v_size = __pyx_t_1;

  /* "serializor/binary/ser.py":547
 *     # Get sequence size
 *     size: cython.Py_ssize_t = len(obj)
 *     if size == 0:             # <<<<<<<<<<<<<<
 *         return gen_header(identifier.LIST, 0)  # type: ignore
 *     # Serialize items
 */
  __pyx_t_2 = (__pyx_v_size == 0);
  if (__pyx_t_2) {

    /* "serializor/binary/ser.py":548
 *     size: cython.Py_ssize_t = len(obj)
 *     if size == 0:
 *         return gen_header(identifier.LIST, 0)  # type: ignore             # <<<<<<<<<<<<<<
 *     # Serialize items
 *     items = [LIST_ID, gen_encoded_int(size)]  # type: ignore
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_3 = __pyx_f_10serializor_6binary_3ser_gen_header(__pyx_v_10serializor_10identifier_LIST, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 548, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_r = ((PyObject*)__pyx_t_3);
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "serializor/binary/ser.py":547
 *     # Get sequence size
 *     size: cython.Py_ssize_t = len(obj)
 *     if size == 0:             # <<<<<<<<<<<<<<
 *         return gen_header(identifier.LIST, 0)  # type: ignore
 *     # Serialize items
 */
  }

  /* "serializor/binary/ser.py":550
 *         return gen_header(identifier.LIST, 0)  # type: ignore
 *     # Serialize items
 *     items = [LIST_ID, gen_encoded_int(size)]  # type: ignore             # <<<<<<<<<<<<<<
 *     for i in obj:
 *         items.append(_ser_common(i))
 */
  __pyx_t_3 = __pyx_f_10serializor_6binary_3ser_gen_encoded_int(__pyx_v_size); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 550, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyList_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 550, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_INCREF(__pyx_v_10serializor_6binary_3ser_LIST_ID);
  __Pyx_GIVEREF(__pyx_v_10serializor_6binary_3ser_LIST_ID);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_4, 0, __pyx_v_10serializor_6binary_3ser_LIST_ID)) __PYX_ERR(0, 550, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_3);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_4, 1, __pyx_t_3)) __PYX_ERR(0, 550, __pyx_L1_error);
  __pyx_t_3 = 0;
  __pyx_v_items = ((PyObject*)__pyx_t_4);
  __pyx_t_4 = 0;

  /* "serializor/binary/ser.py":551
 *     # Serialize items
 *     items = [LIST_ID, gen_encoded_int(size)]  # type: ignore
 *     for i in obj:             # <<<<<<<<<<<<<<
 *         items.append(_ser_common(i))
 *     return b"".join(items)
 */
  if (likely(PyList_CheckExact(__pyx_v_obj)) || PyTuple_CheckExact(__pyx_v_obj)) {
    __pyx_t_4 = __pyx_v_obj; __Pyx_INCREF(__pyx_t_4);
    __pyx_t_1 = 0;
    __pyx_t_5 = NULL;
  } else {
    __pyx_t_1 = -1; __pyx_t_4 = PyObject_GetIter(__pyx_v_obj); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 551, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 551, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_5)) {
      if (likely(PyList_CheckExact(__pyx_t_4))) {
        {
          Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_4);
          #if !CYTHON_ASSUME_SAFE_MACROS
          if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 551, __pyx_L1_error)
          #endif
          if (__pyx_t_1 >= __pyx_temp) break;
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_3 = PyList_GET_ITEM(__pyx_t_4, __pyx_t_1); __Pyx_INCREF(__pyx_t_3); __pyx_t_1++; if (unlikely((0 < 0))) __PYX_ERR(0, 551, __pyx_L1_error)
        #else
        __pyx_t_3 = __Pyx_PySequence_ITEM(__pyx_t_4, __pyx_t_1); __pyx_t_1++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 551, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        #endif
      } else {
        {
          Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_4);
          #if !CYTHON_ASSUME_SAFE_MACROS
          if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 551, __pyx_L1_error)
          #endif
          if (__pyx_t_1 >= __pyx_temp) break;
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_4, __pyx_t_1); __Pyx_INCREF(__pyx_t_3); __pyx_t_1++; if (unlikely((0 < 0))) __PYX_ERR(0, 551, __pyx_L1_error)
        #else
        __pyx_t_3 = __Pyx_PySequence_ITEM(__pyx_t_4, __pyx_t_1); __pyx_t_1++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 551, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        #endif
      }
    } else {
      __pyx_t_3 = __pyx_t_5(__pyx_t_4);
      if (unlikely(!__pyx_t_3)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 551, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_3);
    }
    __Pyx_XDECREF_SET(__pyx_v_i, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "serializor/binary/ser.py":552
 *     items = [LIST_ID, gen_encoded_int(size)]  # type: ignore
 *     for i in obj:
 *         items.append(_ser_common(i))             # <<<<<<<<<<<<<<
 *     return b"".join(items)
 * 
 */
    __pyx_t_3 = __pyx_f_10serializor_6binary_3ser__ser_common(__pyx_v_i); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 552, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_6 = __Pyx_PyList_Append(__pyx_v_items, __pyx_t_3); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(0, 552, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "serializor/binary/ser.py":551
 *     # Serialize items
 *     items = [LIST_ID, gen_encoded_int(size)]  # type: ignore
 *     for i in obj:             # <<<<<<<<<<<<<<
 *         items.append(_ser_common(i))
 *     return b"".join(items)
 */
  }
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "serializor/binary/ser.py":553
 *     for i in obj:
 *         items.append(_ser_common(i))
 *     return b"".join(items)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_4 = __Pyx_PyBytes_Join(__pyx_kp_b__4, __pyx_v_items); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 553, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (!(likely(PyBytes_CheckExact(__pyx_t_4))||((__pyx_t_4) == Py_None) || __Pyx_RaiseUnexpectedTypeError("bytes", __pyx_t_4))) __PYX_ERR(0, 553, __pyx_L1_error)
  __pyx_r = ((PyObject*)__pyx_t_4);
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "serializor/binary/ser.py":532
 * 
 * 
 * @cython.cfunc             # <<<<<<<<<<<<<<
 * @cython.inline(True)
 * def _ser_sequence(obj: Iterable) -> bytes:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("serializor.binary.ser._ser_sequence", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_items);
  __Pyx_XDECREF(__pyx_v_i);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "serializor/binary/ser.py":557
 * 
 * # Mapping Types -------------------------------------------------------------------------------
 * @cython.cfunc             # <<<<<<<<<<<<<<
 * @cython.inline(True)
 * def _ser_dict(obj: dict) -> bytes:
 */

static PyObject *__pyx_f_10serializor_6binary_3ser__ser_dict(PyObject *__pyx_v_obj) {
  Py_ssize_t __pyx_v_size;
  PyObject *__pyx_v_items = NULL;
  PyObject *__pyx_v_key = NULL;
  PyObject *__pyx_v_val = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  Py_ssize_t __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  Py_ssize_t __pyx_t_5;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  int __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_ser_dict", 1);

  /* "serializor/binary/ser.py":566
 *     """
 *     # Get dict size
 *     size: cython.Py_ssize_t = dict_len(obj)             # <<<<<<<<<<<<<<
 *     if size == 0:
 *         return gen_header(identifier.DICT, 0)  # type: ignore
 */
  __pyx_t_1 = PyDict_Size(__pyx_v_obj); if (unlikely(__pyx_t_1 == ((Py_ssize_t)-1L))) __PYX_ERR(0, 566, __pyx_L1_error)
  __pyx_v_size = __pyx_t_1;

  /* "serializor/binary/ser.py":567
 *     # Get dict size
 *     size: cython.Py_ssize_t = dict_len(obj)
 *     if size == 0:             # <<<<<<<<<<<<<<
 *         return gen_header(identifier.DICT, 0)  # type: ignore
 *     # Serialize dict keys & values
 */
  __pyx_t_2 = (__pyx_v_size == 0);
  if (__pyx_t_2) {

    /* "serializor/binary/ser.py":568
 *     size: cython.Py_ssize_t = dict_len(obj)
 *     if size == 0:
 *         return gen_header(identifier.DICT, 0)  # type: ignore             # <<<<<<<<<<<<<<
 *     # Serialize dict keys & values
 *     items = [DICT_ID, gen_encoded_int(size)]  # type: ignore
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_3 = __pyx_f_10serializor_6binary_3ser_gen_header(__pyx_v_10serializor_10identifier_DICT, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 568, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_r = ((PyObject*)__pyx_t_3);
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "serializor/binary/ser.py":567
 *     # Get dict size
 *     size: cython.Py_ssize_t = dict_len(obj)
 *     if size == 0:             # <<<<<<<<<<<<<<
 *         return gen_header(identifier.DICT, 0)  # type: ignore
 *     # Serialize dict keys & values
 */
  }

  /* "serializor/binary/ser.py":570
 *         return gen_header(identifier.DICT, 0)  # type: ignore
 *     # Serialize dict keys & values
 *     items = [DICT_ID, gen_encoded_int(size)]  # type: ignore             # <<<<<<<<<<<<<<
 *     for key, val in obj.items():
 *         items.append(_ser_common(key))
 */
  __pyx_t_3 = __pyx_f_10serializor_6binary_3ser_gen_encoded_int(__pyx_v_size); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 570, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyList_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 570, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_INCREF(__pyx_v_10serializor_6binary_3ser_DICT_ID);
  __Pyx_GIVEREF(__pyx_v_10serializor_6binary_3ser_DICT_ID);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_4, 0, __pyx_v_10serializor_6binary_3ser_DICT_ID)) __PYX_ERR(0, 570, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_3);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_4, 1, __pyx_t_3)) __PYX_ERR(0, 570, __pyx_L1_error);
  __pyx_t_3 = 0;
  __pyx_v_items = ((PyObject*)__pyx_t_4);
  __pyx_t_4 = 0;

  /* "serializor/binary/ser.py":571
 *     # Serialize dict keys & values
 *     items = [DICT_ID, gen_encoded_int(size)]  # type: ignore
 *     for key, val in obj.items():             # <<<<<<<<<<<<<<
 *         items.append(_ser_common(key))
 *         items.append(_ser_common(val))
 */
  __pyx_t_1 = 0;
  __pyx_t_3 = __Pyx_dict_iterator(__pyx_v_obj, 1, __pyx_n_s_items, (&__pyx_t_5), (&__pyx_t_6)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 571, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __pyx_t_4 = __pyx_t_3;
  __pyx_t_3 = 0;
  while (1) {
    __pyx_t_8 = __Pyx_dict_iter_next(__pyx_t_4, __pyx_t_5, &__pyx_t_1, &__pyx_t_3, &__pyx_t_7, NULL, __pyx_t_6);
    if (unlikely(__pyx_t_8 == 0)) break;
    if (unlikely(__pyx_t_8 == -1)) __PYX_ERR(0, 571, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_XDECREF_SET(__pyx_v_key, __pyx_t_3);
    __pyx_t_3 = 0;
    __Pyx_XDECREF_SET(__pyx_v_val, __pyx_t_7);
    __pyx_t_7 = 0;

    /* "serializor/binary/ser.py":572
 *     items = [DICT_ID, gen_encoded_int(size)]  # type: ignore
 *     for key, val in obj.items():
 *         items.append(_ser_common(key))             # <<<<<<<<<<<<<<
 *         items.append(_ser_common(val))
 *     return b"".join(items)
 */
    __pyx_t_7 = __pyx_f_10serializor_6binary_3ser__ser_common(__pyx_v_key); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 572, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_9 = __Pyx_PyList_Append(__pyx_v_items, __pyx_t_7); if (unlikely(__pyx_t_9 == ((int)-1))) __PYX_ERR(0, 572, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

    /* "serializor/binary/ser.py":573
 *     for key, val in obj.items():
 *         items.append(_ser_common(key))
 *         items.append(_ser_common(val))             # <<<<<<<<<<<<<<
 *     return b"".join(items)
 * 
 */
    __pyx_t_7 = __pyx_f_10serializor_6binary_3ser__ser_common(__pyx_v_val); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 573, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_9 = __Pyx_PyList_Append(__pyx_v_items, __pyx_t_7); if (unlikely(__pyx_t_9 == ((int)-1))) __PYX_ERR(0, 573, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  }
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "serializor/binary/ser.py":574
 *         items.append(_ser_common(key))
 *         items.append(_ser_common(val))
 *     return b"".join(items)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_4 = __Pyx_PyBytes_Join(__pyx_kp_b__4, __pyx_v_items); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 574, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (!(likely(PyBytes_CheckExact(__pyx_t_4))||((__pyx_t_4) == Py_None) || __Pyx_RaiseUnexpectedTypeError("bytes", __pyx_t_4))) __PYX_ERR(0, 574, __pyx_L1_error)
  __pyx_r = ((PyObject*)__pyx_t_4);
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "serializor/binary/ser.py":557
 * 
 * # Mapping Types -------------------------------------------------------------------------------
 * @cython.cfunc             # <<<<<<<<<<<<<<
 * @cython.inline(True)
 * def _ser_dict(obj: dict) -> bytes:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("serializor.binary.ser._ser_dict", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_items);
  __Pyx_XDECREF(__pyx_v_key);
  __Pyx_XDECREF(__pyx_v_val);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "serializor/binary/ser.py":578
 * 
 * # NumPy Types ---------------------------------------------------------------------------------
 * @cython.cfunc             # <<<<<<<<<<<<<<
 * @cython.inline(True)
 * def _ser_float64(obj: object) -> bytes:
 */

static PyObject *__pyx_f_10serializor_6binary_3ser__ser_float64(PyObject *__pyx_v_obj) {
  PyObject *__pyx_v_val = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  Py_ssize_t __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_ser_float64", 1);

  /* "serializor/binary/ser.py":586
 *     >>> b'[id(1B)][size(ENC)][value]'
 *     """
 *     val: bytes = utils.encode_str(str(obj), b"ascii")             # <<<<<<<<<<<<<<
 *     return gen_header(identifier.FLOAT, bytes_len(val)) + val  # type: ignore
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_Unicode(__pyx_v_obj); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 586, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __pyx_f_10serializor_5utils_encode_str(__pyx_t_1, ((char *)"ascii")); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 586, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_val = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "serializor/binary/ser.py":587
 *     """
 *     val: bytes = utils.encode_str(str(obj), b"ascii")
 *     return gen_header(identifier.FLOAT, bytes_len(val)) + val  # type: ignore             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_3 = PyBytes_Size(__pyx_v_val); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1L))) __PYX_ERR(0, 587, __pyx_L1_error)
  __pyx_t_2 = __pyx_f_10serializor_6binary_3ser_gen_header(__pyx_v_10serializor_10identifier_FLOAT, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 587, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = PyNumber_Add(__pyx_t_2, __pyx_v_val); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 587, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "serializor/binary/ser.py":578
 * 
 * # NumPy Types ---------------------------------------------------------------------------------
 * @cython.cfunc             # <<<<<<<<<<<<<<
 * @cython.inline(True)
 * def _ser_float64(obj: object) -> bytes:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("serializor.binary.ser._ser_float64", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_val);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "serializor/binary/ser.py":590
 * 
 * 
 * @cython.cfunc             # <<<<<<<<<<<<<<
 * @cython.inline(True)
 * def _ser_datetime64(obj: object) -> bytes:
 */

static PyObject *__pyx_f_10serializor_6binary_3ser__ser_datetime64(PyObject *__pyx_v_obj) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_ser_datetime64", 1);

  /* "serializor/binary/ser.py":598
 *     >>> b'[id(1B)][nptime_unit(1B)][value(8B)]'
 *     """
 *     return b"".join(             # <<<<<<<<<<<<<<
 *         [
 *             DT64_ID,  # identifier
 */
  __Pyx_XDECREF(__pyx_r);

  /* "serializor/binary/ser.py":601
 *         [
 *             DT64_ID,  # identifier
 *             pack_uint8(np.get_datetime64_unit(obj)),  # type: ignore             # <<<<<<<<<<<<<<
 *             pack_int64(np.get_datetime64_value(obj)),  # type: ignore
 *         ]
 */
  __pyx_t_1 = __pyx_f_10serializor_6binary_3ser_pack_uint8(__pyx_f_5numpy_get_datetime64_unit(__pyx_v_obj)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 601, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "serializor/binary/ser.py":602
 *             DT64_ID,  # identifier
 *             pack_uint8(np.get_datetime64_unit(obj)),  # type: ignore
 *             pack_int64(np.get_datetime64_value(obj)),  # type: ignore             # <<<<<<<<<<<<<<
 *         ]
 *     )
 */
  __pyx_t_2 = __pyx_f_10serializor_6binary_3ser_pack_int64(__pyx_f_5numpy_get_datetime64_value(__pyx_v_obj)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 602, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);

  /* "serializor/binary/ser.py":599
 *     """
 *     return b"".join(
 *         [             # <<<<<<<<<<<<<<
 *             DT64_ID,  # identifier
 *             pack_uint8(np.get_datetime64_unit(obj)),  # type: ignore
 */
  __pyx_t_3 = PyList_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 599, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_v_10serializor_6binary_3ser_DT64_ID);
  __Pyx_GIVEREF(__pyx_v_10serializor_6binary_3ser_DT64_ID);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_3, 0, __pyx_v_10serializor_6binary_3ser_DT64_ID)) __PYX_ERR(0, 599, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_3, 1, __pyx_t_1)) __PYX_ERR(0, 599, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_3, 2, __pyx_t_2)) __PYX_ERR(0, 599, __pyx_L1_error);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;

  /* "serializor/binary/ser.py":598
 *     >>> b'[id(1B)][nptime_unit(1B)][value(8B)]'
 *     """
 *     return b"".join(             # <<<<<<<<<<<<<<
 *         [
 *             DT64_ID,  # identifier
 */
  __pyx_t_2 = __Pyx_PyBytes_Join(__pyx_kp_b__4, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 598, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (!(likely(PyBytes_CheckExact(__pyx_t_2))||((__pyx_t_2) == Py_None) || __Pyx_RaiseUnexpectedTypeError("bytes", __pyx_t_2))) __PYX_ERR(0, 598, __pyx_L1_error)
  __pyx_r = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "serializor/binary/ser.py":590
 * 
 * 
 * @cython.cfunc             # <<<<<<<<<<<<<<
 * @cython.inline(True)
 * def _ser_datetime64(obj: object) -> bytes:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("serializor.binary.ser._ser_datetime64", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "serializor/binary/ser.py":607
 * 
 * 
 * @cython.cfunc             # <<<<<<<<<<<<<<
 * @cython.inline(True)
 * def _ser_timedelta64(obj: object) -> bytes:
 */

static PyObject *__pyx_f_10serializor_6binary_3ser__ser_timedelta64(PyObject *__pyx_v_obj) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_ser_timedelta64", 1);

  /* "serializor/binary/ser.py":615
 *     >>> b'[id(1B)][nptime_unit(1B)][value(8B)]'
 *     """
 *     return b"".join(             # <<<<<<<<<<<<<<
 *         [
 *             TD64_ID,  # identifier
 */
  __Pyx_XDECREF(__pyx_r);

  /* "serializor/binary/ser.py":618
 *         [
 *             TD64_ID,  # identifier
 *             pack_uint8(np.get_datetime64_unit(obj)),  # type: ignore             # <<<<<<<<<<<<<<
 *             pack_int64(np.get_timedelta64_value(obj)),  # type: ignore
 *         ]
 */
  __pyx_t_1 = __pyx_f_10serializor_6binary_3ser_pack_uint8(__pyx_f_5numpy_get_datetime64_unit(__pyx_v_obj)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 618, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "serializor/binary/ser.py":619
 *             TD64_ID,  # identifier
 *             pack_uint8(np.get_datetime64_unit(obj)),  # type: ignore
 *             pack_int64(np.get_timedelta64_value(obj)),  # type: ignore             # <<<<<<<<<<<<<<
 *         ]
 *     )
 */
  __pyx_t_2 = __pyx_f_10serializor_6binary_3ser_pack_int64(__pyx_f_5numpy_get_timedelta64_value(__pyx_v_obj)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 619, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);

  /* "serializor/binary/ser.py":616
 *     """
 *     return b"".join(
 *         [             # <<<<<<<<<<<<<<
 *             TD64_ID,  # identifier
 *             pack_uint8(np.get_datetime64_unit(obj)),  # type: ignore
 */
  __pyx_t_3 = PyList_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 616, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_v_10serializor_6binary_3ser_TD64_ID);
  __Pyx_GIVEREF(__pyx_v_10serializor_6binary_3ser_TD64_ID);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_3, 0, __pyx_v_10serializor_6binary_3ser_TD64_ID)) __PYX_ERR(0, 616, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_3, 1, __pyx_t_1)) __PYX_ERR(0, 616, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_3, 2, __pyx_t_2)) __PYX_ERR(0, 616, __pyx_L1_error);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;

  /* "serializor/binary/ser.py":615
 *     >>> b'[id(1B)][nptime_unit(1B)][value(8B)]'
 *     """
 *     return b"".join(             # <<<<<<<<<<<<<<
 *         [
 *             TD64_ID,  # identifier
 */
  __pyx_t_2 = __Pyx_PyBytes_Join(__pyx_kp_b__4, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 615, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (!(likely(PyBytes_CheckExact(__pyx_t_2))||((__pyx_t_2) == Py_None) || __Pyx_RaiseUnexpectedTypeError("bytes", __pyx_t_2))) __PYX_ERR(0, 615, __pyx_L1_error)
  __pyx_r = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "serializor/binary/ser.py":607
 * 
 * 
 * @cython.cfunc             # <<<<<<<<<<<<<<
 * @cython.inline(True)
 * def _ser_timedelta64(obj: object) -> bytes:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("serializor.binary.ser._ser_timedelta64", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "serializor/binary/ser.py":624
 * 
 * 
 * @cython.cfunc             # <<<<<<<<<<<<<<
 * @cython.inline(True)
 * def _ser_complex64(obj: object) -> bytes:
 */

static PyObject *__pyx_f_10serializor_6binary_3ser__ser_complex64(PyObject *__pyx_v_obj) {
  PyObject *__pyx_v_real_val = 0;
  PyObject *__pyx_v_real_size = 0;
  PyObject *__pyx_v_imag_val = 0;
  PyObject *__pyx_v_imag_size = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  Py_ssize_t __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_ser_complex64", 1);

  /* "serializor/binary/ser.py":632
 *     >>> b'[id(1B)][real_size(ENC)][real][imag_size(ENC)][imag]'
 *     """
 *     real_val: bytes = _orjson_dumps(float(obj.real))             # <<<<<<<<<<<<<<
 *     real_size: bytes = gen_encoded_int(bytes_len(real_val))  # type: ignore
 *     imag_val: bytes = _orjson_dumps(float(obj.imag))
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_obj, __pyx_n_s_real); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 632, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyNumber_Float(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 632, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __pyx_f_10serializor_6binary_3ser__orjson_dumps(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 632, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_real_val = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "serializor/binary/ser.py":633
 *     """
 *     real_val: bytes = _orjson_dumps(float(obj.real))
 *     real_size: bytes = gen_encoded_int(bytes_len(real_val))  # type: ignore             # <<<<<<<<<<<<<<
 *     imag_val: bytes = _orjson_dumps(float(obj.imag))
 *     imag_size: bytes = gen_encoded_int(bytes_len(imag_val))  # type: ignore
 */
  __pyx_t_3 = PyBytes_Size(__pyx_v_real_val); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1L))) __PYX_ERR(0, 633, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_10serializor_6binary_3ser_gen_encoded_int(__pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 633, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_real_size = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "serializor/binary/ser.py":634
 *     real_val: bytes = _orjson_dumps(float(obj.real))
 *     real_size: bytes = gen_encoded_int(bytes_len(real_val))  # type: ignore
 *     imag_val: bytes = _orjson_dumps(float(obj.imag))             # <<<<<<<<<<<<<<
 *     imag_size: bytes = gen_encoded_int(bytes_len(imag_val))  # type: ignore
 *     return b"".join([COMPLEX_ID, real_size, real_val, imag_size, imag_val])
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_obj, __pyx_n_s_imag); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 634, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyNumber_Float(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 634, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __pyx_f_10serializor_6binary_3ser__orjson_dumps(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 634, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_imag_val = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "serializor/binary/ser.py":635
 *     real_size: bytes = gen_encoded_int(bytes_len(real_val))  # type: ignore
 *     imag_val: bytes = _orjson_dumps(float(obj.imag))
 *     imag_size: bytes = gen_encoded_int(bytes_len(imag_val))  # type: ignore             # <<<<<<<<<<<<<<
 *     return b"".join([COMPLEX_ID, real_size, real_val, imag_size, imag_val])
 * 
 */
  __pyx_t_3 = PyBytes_Size(__pyx_v_imag_val); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1L))) __PYX_ERR(0, 635, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_10serializor_6binary_3ser_gen_encoded_int(__pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 635, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_imag_size = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "serializor/binary/ser.py":636
 *     imag_val: bytes = _orjson_dumps(float(obj.imag))
 *     imag_size: bytes = gen_encoded_int(bytes_len(imag_val))  # type: ignore
 *     return b"".join([COMPLEX_ID, real_size, real_val, imag_size, imag_val])             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyList_New(5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 636, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_10serializor_6binary_3ser_COMPLEX_ID);
  __Pyx_GIVEREF(__pyx_v_10serializor_6binary_3ser_COMPLEX_ID);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_v_10serializor_6binary_3ser_COMPLEX_ID)) __PYX_ERR(0, 636, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_real_size);
  __Pyx_GIVEREF(__pyx_v_real_size);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 1, __pyx_v_real_size)) __PYX_ERR(0, 636, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_real_val);
  __Pyx_GIVEREF(__pyx_v_real_val);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 2, __pyx_v_real_val)) __PYX_ERR(0, 636, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_imag_size);
  __Pyx_GIVEREF(__pyx_v_imag_size);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 3, __pyx_v_imag_size)) __PYX_ERR(0, 636, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_imag_val);
  __Pyx_GIVEREF(__pyx_v_imag_val);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 4, __pyx_v_imag_val)) __PYX_ERR(0, 636, __pyx_L1_error);
  __pyx_t_2 = __Pyx_PyBytes_Join(__pyx_kp_b__4, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 636, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (!(likely(PyBytes_CheckExact(__pyx_t_2))||((__pyx_t_2) == Py_None) || __Pyx_RaiseUnexpectedTypeError("bytes", __pyx_t_2))) __PYX_ERR(0, 636, __pyx_L1_error)
  __pyx_r = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "serializor/binary/ser.py":624
 * 
 * 
 * @cython.cfunc             # <<<<<<<<<<<<<<
 * @cython.inline(True)
 * def _ser_complex64(obj: object) -> bytes:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("serializor.binary.ser._ser_complex64", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_real_val);
  __Pyx_XDECREF(__pyx_v_real_size);
  __Pyx_XDECREF(__pyx_v_imag_val);
  __Pyx_XDECREF(__pyx_v_imag_size);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "serializor/binary/ser.py":639
 * 
 * 
 * @cython.cfunc             # <<<<<<<<<<<<<<
 * @cython.inline(True)
 * def _ser_complex128(obj: object) -> bytes:
 */

static PyObject *__pyx_f_10serializor_6binary_3ser__ser_complex128(PyObject *__pyx_v_obj) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_ser_complex128", 1);

  /* "serializor/binary/ser.py":647
 *     >>> b'[id(1B)][real_size(ENC)][real][imag_size(ENC)][imag]'
 *     """
 *     return _ser_complex(obj)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_10serializor_6binary_3ser__ser_complex(__pyx_v_obj); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 647, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "serializor/binary/ser.py":639
 * 
 * 
 * @cython.cfunc             # <<<<<<<<<<<<<<
 * @cython.inline(True)
 * def _ser_complex128(obj: object) -> bytes:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("serializor.binary.ser._ser_complex128", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "serializor/binary/ser.py":650
 * 
 * 
 * @cython.cfunc             # <<<<<<<<<<<<<<
 * @cython.inline(True)
 * def _ser_ndarray(obj: np.ndarray) -> bytes:
 */

static PyObject *__pyx_f_10serializor_6binary_3ser__ser_ndarray(PyArrayObject *__pyx_v_obj) {
  char __pyx_v_npy_kind;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  char __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_ser_ndarray", 1);

  /* "serializor/binary/ser.py":671
 *     """
 *     # Get ndarray kind
 *     npy_kind: cython.char = obj.descr.kind             # <<<<<<<<<<<<<<
 * 
 *     # . ndarray[object]
 */
  __pyx_t_1 = ((PyObject *)__pyx_f_5numpy_7ndarray_5descr_descr(__pyx_v_obj)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 671, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = ((PyArray_Descr *)__pyx_t_1)->kind;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_npy_kind = __pyx_t_2;

  /* "serializor/binary/ser.py":674
 * 
 *     # . ndarray[object]
 *     if npy_kind == identifier.NDARRAY_OBJECT:             # <<<<<<<<<<<<<<
 *         return _ser_ndarray_object(obj, NDARRAY_OBJECT_IDDT)
 *     # . ndarray[int]
 */
  __pyx_t_3 = (__pyx_v_npy_kind == __pyx_v_10serializor_10identifier_NDARRAY_OBJECT);
  if (__pyx_t_3) {

    /* "serializor/binary/ser.py":675
 *     # . ndarray[object]
 *     if npy_kind == identifier.NDARRAY_OBJECT:
 *         return _ser_ndarray_object(obj, NDARRAY_OBJECT_IDDT)             # <<<<<<<<<<<<<<
 *     # . ndarray[int]
 *     if npy_kind == identifier.NDARRAY_INT:
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_1 = __pyx_v_10serializor_6binary_3ser_NDARRAY_OBJECT_IDDT;
    __Pyx_INCREF(__pyx_t_1);
    __pyx_t_4 = __pyx_f_10serializor_6binary_3ser__ser_ndarray_object(__pyx_v_obj, ((PyObject*)__pyx_t_1)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 675, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_r = ((PyObject*)__pyx_t_4);
    __pyx_t_4 = 0;
    goto __pyx_L0;

    /* "serializor/binary/ser.py":674
 * 
 *     # . ndarray[object]
 *     if npy_kind == identifier.NDARRAY_OBJECT:             # <<<<<<<<<<<<<<
 *         return _ser_ndarray_object(obj, NDARRAY_OBJECT_IDDT)
 *     # . ndarray[int]
 */
  }

  /* "serializor/binary/ser.py":677
 *         return _ser_ndarray_object(obj, NDARRAY_OBJECT_IDDT)
 *     # . ndarray[int]
 *     if npy_kind == identifier.NDARRAY_INT:             # <<<<<<<<<<<<<<
 *         return _ser_ndarray_numeric(obj, NDARRAY_INT_IDDT, obj.descr.type_num)
 *     # . ndarray[uint]
 */
  __pyx_t_3 = (__pyx_v_npy_kind == __pyx_v_10serializor_10identifier_NDARRAY_INT);
  if (__pyx_t_3) {

    /* "serializor/binary/ser.py":678
 *     # . ndarray[int]
 *     if npy_kind == identifier.NDARRAY_INT:
 *         return _ser_ndarray_numeric(obj, NDARRAY_INT_IDDT, obj.descr.type_num)             # <<<<<<<<<<<<<<
 *     # . ndarray[uint]
 *     if npy_kind == identifier.NDARRAY_UINT:
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __pyx_v_10serializor_6binary_3ser_NDARRAY_INT_IDDT;
    __Pyx_INCREF(__pyx_t_4);
    __pyx_t_1 = ((PyObject *)__pyx_f_5numpy_7ndarray_5descr_descr(__pyx_v_obj)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 678, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_5 = __pyx_f_10serializor_6binary_3ser__ser_ndarray_numeric(__pyx_v_obj, ((PyObject*)__pyx_t_4), ((PyArray_Descr *)__pyx_t_1)->type_num); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 678, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_r = ((PyObject*)__pyx_t_5);
    __pyx_t_5 = 0;
    goto __pyx_L0;

    /* "serializor/binary/ser.py":677
 *         return _ser_ndarray_object(obj, NDARRAY_OBJECT_IDDT)
 *     # . ndarray[int]
 *     if npy_kind == identifier.NDARRAY_INT:             # <<<<<<<<<<<<<<
 *         return _ser_ndarray_numeric(obj, NDARRAY_INT_IDDT, obj.descr.type_num)
 *     # . ndarray[uint]
 */
  }

  /* "serializor/binary/ser.py":680
 *         return _ser_ndarray_numeric(obj, NDARRAY_INT_IDDT, obj.descr.type_num)
 *     # . ndarray[uint]
 *     if npy_kind == identifier.NDARRAY_UINT:             # <<<<<<<<<<<<<<
 *         return _ser_ndarray_numeric(obj, NDARRAY_UINT_IDDT, obj.descr.type_num)
 *     # . ndarray[float]
 */
  __pyx_t_3 = (__pyx_v_npy_kind == __pyx_v_10serializor_10identifier_NDARRAY_UINT);
  if (__pyx_t_3) {

    /* "serializor/binary/ser.py":681
 *     # . ndarray[uint]
 *     if npy_kind == identifier.NDARRAY_UINT:
 *         return _ser_ndarray_numeric(obj, NDARRAY_UINT_IDDT, obj.descr.type_num)             # <<<<<<<<<<<<<<
 *     # . ndarray[float]
 *     if npy_kind == identifier.NDARRAY_FLOAT:
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_5 = __pyx_v_10serializor_6binary_3ser_NDARRAY_UINT_IDDT;
    __Pyx_INCREF(__pyx_t_5);
    __pyx_t_1 = ((PyObject *)__pyx_f_5numpy_7ndarray_5descr_descr(__pyx_v_obj)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 681, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_4 = __pyx_f_10serializor_6binary_3ser__ser_ndarray_numeric(__pyx_v_obj, ((PyObject*)__pyx_t_5), ((PyArray_Descr *)__pyx_t_1)->type_num); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 681, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_r = ((PyObject*)__pyx_t_4);
    __pyx_t_4 = 0;
    goto __pyx_L0;

    /* "serializor/binary/ser.py":680
 *         return _ser_ndarray_numeric(obj, NDARRAY_INT_IDDT, obj.descr.type_num)
 *     # . ndarray[uint]
 *     if npy_kind == identifier.NDARRAY_UINT:             # <<<<<<<<<<<<<<
 *         return _ser_ndarray_numeric(obj, NDARRAY_UINT_IDDT, obj.descr.type_num)
 *     # . ndarray[float]
 */
  }

  /* "serializor/binary/ser.py":683
 *         return _ser_ndarray_numeric(obj, NDARRAY_UINT_IDDT, obj.descr.type_num)
 *     # . ndarray[float]
 *     if npy_kind == identifier.NDARRAY_FLOAT:             # <<<<<<<<<<<<<<
 *         return _ser_ndarray_numeric(obj, NDARRAY_FLOAT_IDDT, obj.descr.type_num)
 *     # . ndarray[bool]
 */
  __pyx_t_3 = (__pyx_v_npy_kind == __pyx_v_10serializor_10identifier_NDARRAY_FLOAT);
  if (__pyx_t_3) {

    /* "serializor/binary/ser.py":684
 *     # . ndarray[float]
 *     if npy_kind == identifier.NDARRAY_FLOAT:
 *         return _ser_ndarray_numeric(obj, NDARRAY_FLOAT_IDDT, obj.descr.type_num)             # <<<<<<<<<<<<<<
 *     # . ndarray[bool]
 *     if npy_kind == identifier.NDARRAY_BOOL:
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __pyx_v_10serializor_6binary_3ser_NDARRAY_FLOAT_IDDT;
    __Pyx_INCREF(__pyx_t_4);
    __pyx_t_1 = ((PyObject *)__pyx_f_5numpy_7ndarray_5descr_descr(__pyx_v_obj)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 684, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_5 = __pyx_f_10serializor_6binary_3ser__ser_ndarray_numeric(__pyx_v_obj, ((PyObject*)__pyx_t_4), ((PyArray_Descr *)__pyx_t_1)->type_num); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 684, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_r = ((PyObject*)__pyx_t_5);
    __pyx_t_5 = 0;
    goto __pyx_L0;

    /* "serializor/binary/ser.py":683
 *         return _ser_ndarray_numeric(obj, NDARRAY_UINT_IDDT, obj.descr.type_num)
 *     # . ndarray[float]
 *     if npy_kind == identifier.NDARRAY_FLOAT:             # <<<<<<<<<<<<<<
 *         return _ser_ndarray_numeric(obj, NDARRAY_FLOAT_IDDT, obj.descr.type_num)
 *     # . ndarray[bool]
 */
  }

  /* "serializor/binary/ser.py":686
 *         return _ser_ndarray_numeric(obj, NDARRAY_FLOAT_IDDT, obj.descr.type_num)
 *     # . ndarray[bool]
 *     if npy_kind == identifier.NDARRAY_BOOL:             # <<<<<<<<<<<<<<
 *         return _ser_ndarray_numeric(
 *             np.PyArray_Cast(obj, np.NPY_TYPES.NPY_INT64),  # cast to int64
 */
  __pyx_t_3 = (__pyx_v_npy_kind == __pyx_v_10serializor_10identifier_NDARRAY_BOOL);
  if (__pyx_t_3) {

    /* "serializor/binary/ser.py":687
 *     # . ndarray[bool]
 *     if npy_kind == identifier.NDARRAY_BOOL:
 *         return _ser_ndarray_numeric(             # <<<<<<<<<<<<<<
 *             np.PyArray_Cast(obj, np.NPY_TYPES.NPY_INT64),  # cast to int64
 *             NDARRAY_BOOL_IDDT,
 */
    __Pyx_XDECREF(__pyx_r);

    /* "serializor/binary/ser.py":688
 *     if npy_kind == identifier.NDARRAY_BOOL:
 *         return _ser_ndarray_numeric(
 *             np.PyArray_Cast(obj, np.NPY_TYPES.NPY_INT64),  # cast to int64             # <<<<<<<<<<<<<<
 *             NDARRAY_BOOL_IDDT,
 *             obj.descr.type_num,
 */
    __pyx_t_5 = PyArray_Cast(__pyx_v_obj, NPY_INT64); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 688, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (!(likely(((__pyx_t_5) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_5, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 688, __pyx_L1_error)

    /* "serializor/binary/ser.py":689
 *         return _ser_ndarray_numeric(
 *             np.PyArray_Cast(obj, np.NPY_TYPES.NPY_INT64),  # cast to int64
 *             NDARRAY_BOOL_IDDT,             # <<<<<<<<<<<<<<
 *             obj.descr.type_num,
 *         )
 */
    __pyx_t_1 = __pyx_v_10serializor_6binary_3ser_NDARRAY_BOOL_IDDT;
    __Pyx_INCREF(__pyx_t_1);

    /* "serializor/binary/ser.py":690
 *             np.PyArray_Cast(obj, np.NPY_TYPES.NPY_INT64),  # cast to int64
 *             NDARRAY_BOOL_IDDT,
 *             obj.descr.type_num,             # <<<<<<<<<<<<<<
 *         )
 *     # . ndarray[datetime64]
 */
    __pyx_t_4 = ((PyObject *)__pyx_f_5numpy_7ndarray_5descr_descr(__pyx_v_obj)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 690, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);

    /* "serializor/binary/ser.py":687
 *     # . ndarray[bool]
 *     if npy_kind == identifier.NDARRAY_BOOL:
 *         return _ser_ndarray_numeric(             # <<<<<<<<<<<<<<
 *             np.PyArray_Cast(obj, np.NPY_TYPES.NPY_INT64),  # cast to int64
 *             NDARRAY_BOOL_IDDT,
 */
    __pyx_t_6 = __pyx_f_10serializor_6binary_3ser__ser_ndarray_numeric(((PyArrayObject *)__pyx_t_5), ((PyObject*)__pyx_t_1), ((PyArray_Descr *)__pyx_t_4)->type_num); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 687, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = ((PyObject*)__pyx_t_6);
    __pyx_t_6 = 0;
    goto __pyx_L0;

    /* "serializor/binary/ser.py":686
 *         return _ser_ndarray_numeric(obj, NDARRAY_FLOAT_IDDT, obj.descr.type_num)
 *     # . ndarray[bool]
 *     if npy_kind == identifier.NDARRAY_BOOL:             # <<<<<<<<<<<<<<
 *         return _ser_ndarray_numeric(
 *             np.PyArray_Cast(obj, np.NPY_TYPES.NPY_INT64),  # cast to int64
 */
  }

  /* "serializor/binary/ser.py":693
 *         )
 *     # . ndarray[datetime64]
 *     if npy_kind == identifier.NDARRAY_DT64:             # <<<<<<<<<<<<<<
 *         return _ser_ndarray_nptime(obj, NDARRAY_DT64_IDDT)
 *     # . ndarray[timedelta64]
 */
  __pyx_t_3 = (__pyx_v_npy_kind == __pyx_v_10serializor_10identifier_NDARRAY_DT64);
  if (__pyx_t_3) {

    /* "serializor/binary/ser.py":694
 *     # . ndarray[datetime64]
 *     if npy_kind == identifier.NDARRAY_DT64:
 *         return _ser_ndarray_nptime(obj, NDARRAY_DT64_IDDT)             # <<<<<<<<<<<<<<
 *     # . ndarray[timedelta64]
 *     if npy_kind == identifier.NDARRAY_TD64:
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_6 = __pyx_v_10serializor_6binary_3ser_NDARRAY_DT64_IDDT;
    __Pyx_INCREF(__pyx_t_6);
    __pyx_t_4 = __pyx_f_10serializor_6binary_3ser__ser_ndarray_nptime(__pyx_v_obj, ((PyObject*)__pyx_t_6)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 694, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_r = ((PyObject*)__pyx_t_4);
    __pyx_t_4 = 0;
    goto __pyx_L0;

    /* "serializor/binary/ser.py":693
 *         )
 *     # . ndarray[datetime64]
 *     if npy_kind == identifier.NDARRAY_DT64:             # <<<<<<<<<<<<<<
 *         return _ser_ndarray_nptime(obj, NDARRAY_DT64_IDDT)
 *     # . ndarray[timedelta64]
 */
  }

  /* "serializor/binary/ser.py":696
 *         return _ser_ndarray_nptime(obj, NDARRAY_DT64_IDDT)
 *     # . ndarray[timedelta64]
 *     if npy_kind == identifier.NDARRAY_TD64:             # <<<<<<<<<<<<<<
 *         return _ser_ndarray_nptime(obj, NDARRAY_TD64_IDDT)
 *     # . ndarray[complex]
 */
  __pyx_t_3 = (__pyx_v_npy_kind == __pyx_v_10serializor_10identifier_NDARRAY_TD64);
  if (__pyx_t_3) {

    /* "serializor/binary/ser.py":697
 *     # . ndarray[timedelta64]
 *     if npy_kind == identifier.NDARRAY_TD64:
 *         return _ser_ndarray_nptime(obj, NDARRAY_TD64_IDDT)             # <<<<<<<<<<<<<<
 *     # . ndarray[complex]
 *     if npy_kind == identifier.NDARRAY_COMPLEX:
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __pyx_v_10serializor_6binary_3ser_NDARRAY_TD64_IDDT;
    __Pyx_INCREF(__pyx_t_4);
    __pyx_t_6 = __pyx_f_10serializor_6binary_3ser__ser_ndarray_nptime(__pyx_v_obj, ((PyObject*)__pyx_t_4)); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 697, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = ((PyObject*)__pyx_t_6);
    __pyx_t_6 = 0;
    goto __pyx_L0;

    /* "serializor/binary/ser.py":696
 *         return _ser_ndarray_nptime(obj, NDARRAY_DT64_IDDT)
 *     # . ndarray[timedelta64]
 *     if npy_kind == identifier.NDARRAY_TD64:             # <<<<<<<<<<<<<<
 *         return _ser_ndarray_nptime(obj, NDARRAY_TD64_IDDT)
 *     # . ndarray[complex]
 */
  }

  /* "serializor/binary/ser.py":699
 *         return _ser_ndarray_nptime(obj, NDARRAY_TD64_IDDT)
 *     # . ndarray[complex]
 *     if npy_kind == identifier.NDARRAY_COMPLEX:             # <<<<<<<<<<<<<<
 *         return _ser_ndarray_complex(obj, NDARRAY_COMPLEX_IDDT)
 *     # . ndarray[bytes]
 */
  __pyx_t_3 = (__pyx_v_npy_kind == __pyx_v_10serializor_10identifier_NDARRAY_COMPLEX);
  if (__pyx_t_3) {

    /* "serializor/binary/ser.py":700
 *     # . ndarray[complex]
 *     if npy_kind == identifier.NDARRAY_COMPLEX:
 *         return _ser_ndarray_complex(obj, NDARRAY_COMPLEX_IDDT)             # <<<<<<<<<<<<<<
 *     # . ndarray[bytes]
 *     if npy_kind == identifier.NDARRAY_BYTES:
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_6 = __pyx_v_10serializor_6binary_3ser_NDARRAY_COMPLEX_IDDT;
    __Pyx_INCREF(__pyx_t_6);
    __pyx_t_4 = __pyx_f_10serializor_6binary_3ser__ser_ndarray_complex(__pyx_v_obj, ((PyObject*)__pyx_t_6)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 700, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_r = ((PyObject*)__pyx_t_4);
    __pyx_t_4 = 0;
    goto __pyx_L0;

    /* "serializor/binary/ser.py":699
 *         return _ser_ndarray_nptime(obj, NDARRAY_TD64_IDDT)
 *     # . ndarray[complex]
 *     if npy_kind == identifier.NDARRAY_COMPLEX:             # <<<<<<<<<<<<<<
 *         return _ser_ndarray_complex(obj, NDARRAY_COMPLEX_IDDT)
 *     # . ndarray[bytes]
 */
  }

  /* "serializor/binary/ser.py":702
 *         return _ser_ndarray_complex(obj, NDARRAY_COMPLEX_IDDT)
 *     # . ndarray[bytes]
 *     if npy_kind == identifier.NDARRAY_BYTES:             # <<<<<<<<<<<<<<
 *         return _ser_ndarray_bytes(obj, NDARRAY_BYTES_IDDT)
 *     # . ndarray[str]
 */
  __pyx_t_3 = (__pyx_v_npy_kind == __pyx_v_10serializor_10identifier_NDARRAY_BYTES);
  if (__pyx_t_3) {

    /* "serializor/binary/ser.py":703
 *     # . ndarray[bytes]
 *     if npy_kind == identifier.NDARRAY_BYTES:
 *         return _ser_ndarray_bytes(obj, NDARRAY_BYTES_IDDT)             # <<<<<<<<<<<<<<
 *     # . ndarray[str]
 *     if npy_kind == identifier.NDARRAY_UNICODE:
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __pyx_v_10serializor_6binary_3ser_NDARRAY_BYTES_IDDT;
    __Pyx_INCREF(__pyx_t_4);
    __pyx_t_6 = __pyx_f_10serializor_6binary_3ser__ser_ndarray_bytes(__pyx_v_obj, ((PyObject*)__pyx_t_4)); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 703, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = ((PyObject*)__pyx_t_6);
    __pyx_t_6 = 0;
    goto __pyx_L0;

    /* "serializor/binary/ser.py":702
 *         return _ser_ndarray_complex(obj, NDARRAY_COMPLEX_IDDT)
 *     # . ndarray[bytes]
 *     if npy_kind == identifier.NDARRAY_BYTES:             # <<<<<<<<<<<<<<
 *         return _ser_ndarray_bytes(obj, NDARRAY_BYTES_IDDT)
 *     # . ndarray[str]
 */
  }

  /* "serializor/binary/ser.py":705
 *         return _ser_ndarray_bytes(obj, NDARRAY_BYTES_IDDT)
 *     # . ndarray[str]
 *     if npy_kind == identifier.NDARRAY_UNICODE:             # <<<<<<<<<<<<<<
 *         return _ser_ndarray_unicode(obj, NDARRAY_UNICODE_IDDT)
 *     # . invalid dtype
 */
  __pyx_t_3 = (__pyx_v_npy_kind == __pyx_v_10serializor_10identifier_NDARRAY_UNICODE);
  if (__pyx_t_3) {

    /* "serializor/binary/ser.py":706
 *     # . ndarray[str]
 *     if npy_kind == identifier.NDARRAY_UNICODE:
 *         return _ser_ndarray_unicode(obj, NDARRAY_UNICODE_IDDT)             # <<<<<<<<<<<<<<
 *     # . invalid dtype
 *     raise TypeError("unsupported <'numpy.ndarray'> dtype '%s'." % obj.dtype)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_6 = __pyx_v_10serializor_6binary_3ser_NDARRAY_UNICODE_IDDT;
    __Pyx_INCREF(__pyx_t_6);
    __pyx_t_4 = __pyx_f_10serializor_6binary_3ser__ser_ndarray_unicode(__pyx_v_obj, ((PyObject*)__pyx_t_6)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 706, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_r = ((PyObject*)__pyx_t_4);
    __pyx_t_4 = 0;
    goto __pyx_L0;

    /* "serializor/binary/ser.py":705
 *         return _ser_ndarray_bytes(obj, NDARRAY_BYTES_IDDT)
 *     # . ndarray[str]
 *     if npy_kind == identifier.NDARRAY_UNICODE:             # <<<<<<<<<<<<<<
 *         return _ser_ndarray_unicode(obj, NDARRAY_UNICODE_IDDT)
 *     # . invalid dtype
 */
  }

  /* "serializor/binary/ser.py":708
 *         return _ser_ndarray_unicode(obj, NDARRAY_UNICODE_IDDT)
 *     # . invalid dtype
 *     raise TypeError("unsupported <'numpy.ndarray'> dtype '%s'." % obj.dtype)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_obj), __pyx_n_s_dtype); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 708, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_6 = __Pyx_PyUnicode_FormatSafe(__pyx_kp_u_unsupported_numpy_ndarray_dtype, __pyx_t_4); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 708, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_builtin_TypeError, __pyx_t_6); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 708, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_Raise(__pyx_t_4, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __PYX_ERR(0, 708, __pyx_L1_error)

  /* "serializor/binary/ser.py":650
 * 
 * 
 * @cython.cfunc             # <<<<<<<<<<<<<<
 * @cython.inline(True)
 * def _ser_ndarray(obj: np.ndarray) -> bytes:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("serializor.binary.ser._ser_ndarray", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "serializor/binary/ser.py":711
 * 
 * 
 * @cython.cfunc             # <<<<<<<<<<<<<<
 * @cython.inline(True)
 * def _ser_ndarray_object(arr: np.ndarray, arr_dtype: bytes) -> bytes:
 */

static PyObject *__pyx_f_10serializor_6binary_3ser__ser_ndarray_object(PyArrayObject *__pyx_v_arr, PyObject *__pyx_v_arr_dtype) {
  int __pyx_v_npy_dtype;
  int __pyx_v_ndim;
  npy_intp *__pyx_v_shape;
  Py_ssize_t __pyx_v_s_i;
  PyObject *__pyx_v_items = NULL;
  Py_ssize_t __pyx_v_i;
  Py_ssize_t __pyx_v_s_j;
  Py_ssize_t __pyx_v_j;
  Py_ssize_t __pyx_v_s_k;
  Py_ssize_t __pyx_v_k;
  Py_ssize_t __pyx_v_s_l;
  Py_ssize_t __pyx_v_l;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  Py_ssize_t __pyx_t_6;
  Py_ssize_t __pyx_t_7;
  Py_ssize_t __pyx_t_8;
  int __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  Py_ssize_t __pyx_t_11;
  Py_ssize_t __pyx_t_12;
  Py_ssize_t __pyx_t_13;
  PyObject *__pyx_t_14 = NULL;
  Py_ssize_t __pyx_t_15;
  Py_ssize_t __pyx_t_16;
  Py_ssize_t __pyx_t_17;
  PyObject *__pyx_t_18 = NULL;
  Py_ssize_t __pyx_t_19;
  Py_ssize_t __pyx_t_20;
  Py_ssize_t __pyx_t_21;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_ser_ndarray_object", 1);

  /* "serializor/binary/ser.py":730
 *     >>> b'...[SHAPE][i1(SER)]...[iN(SER)]'
 *     """
 *     npy_dtype: cython.int = arr.descr.type_num             # <<<<<<<<<<<<<<
 *     ndim: cython.int = arr.ndim
 *     shape = arr.shape
 */
  __pyx_t_1 = ((PyObject *)__pyx_f_5numpy_7ndarray_5descr_descr(__pyx_v_arr)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 730, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = ((PyArray_Descr *)__pyx_t_1)->type_num;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_npy_dtype = __pyx_t_2;

  /* "serializor/binary/ser.py":731
 *     """
 *     npy_dtype: cython.int = arr.descr.type_num
 *     ndim: cython.int = arr.ndim             # <<<<<<<<<<<<<<
 *     shape = arr.shape
 *     # 1-dimension
 */
  __pyx_v_ndim = __pyx_f_5numpy_7ndarray_4ndim_ndim(__pyx_v_arr);

  /* "serializor/binary/ser.py":732
 *     npy_dtype: cython.int = arr.descr.type_num
 *     ndim: cython.int = arr.ndim
 *     shape = arr.shape             # <<<<<<<<<<<<<<
 *     # 1-dimension
 *     if ndim == 1:
 */
  __pyx_v_shape = __pyx_f_5numpy_7ndarray_5shape_shape(__pyx_v_arr);

  /* "serializor/binary/ser.py":734
 *     shape = arr.shape
 *     # 1-dimension
 *     if ndim == 1:             # <<<<<<<<<<<<<<
 *         # . array shape
 *         s_i: cython.Py_ssize_t = shape[0]
 */
  __pyx_t_3 = (__pyx_v_ndim == 1);
  if (__pyx_t_3) {

    /* "serializor/binary/ser.py":736
 *     if ndim == 1:
 *         # . array shape
 *         s_i: cython.Py_ssize_t = shape[0]             # <<<<<<<<<<<<<<
 *         items = [
 *             arr_dtype,  # identifier & dtype
 */
    __pyx_v_s_i = (__pyx_v_shape[0]);

    /* "serializor/binary/ser.py":739
 *         items = [
 *             arr_dtype,  # identifier & dtype
 *             pack_uint8(npy_dtype),  # type: ignore             # <<<<<<<<<<<<<<
 *             b"\x01",  # 1-dimension
 *             gen_encoded_int(s_i),  # type: ignore
 */
    __pyx_t_1 = __pyx_f_10serializor_6binary_3ser_pack_uint8(__pyx_v_npy_dtype); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 739, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);

    /* "serializor/binary/ser.py":741
 *             pack_uint8(npy_dtype),  # type: ignore
 *             b"\x01",  # 1-dimension
 *             gen_encoded_int(s_i),  # type: ignore             # <<<<<<<<<<<<<<
 *         ]
 *         # . serialize values
 */
    __pyx_t_4 = __pyx_f_10serializor_6binary_3ser_gen_encoded_int(__pyx_v_s_i); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 741, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);

    /* "serializor/binary/ser.py":737
 *         # . array shape
 *         s_i: cython.Py_ssize_t = shape[0]
 *         items = [             # <<<<<<<<<<<<<<
 *             arr_dtype,  # identifier & dtype
 *             pack_uint8(npy_dtype),  # type: ignore
 */
    __pyx_t_5 = PyList_New(4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 737, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_INCREF(__pyx_v_arr_dtype);
    __Pyx_GIVEREF(__pyx_v_arr_dtype);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 0, __pyx_v_arr_dtype)) __PYX_ERR(0, 737, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 1, __pyx_t_1)) __PYX_ERR(0, 737, __pyx_L1_error);
    __Pyx_INCREF(__pyx_kp_b__5);
    __Pyx_GIVEREF(__pyx_kp_b__5);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 2, __pyx_kp_b__5)) __PYX_ERR(0, 737, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_4);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 3, __pyx_t_4)) __PYX_ERR(0, 737, __pyx_L1_error);
    __pyx_t_1 = 0;
    __pyx_t_4 = 0;
    __pyx_v_items = ((PyObject*)__pyx_t_5);
    __pyx_t_5 = 0;

    /* "serializor/binary/ser.py":744
 *         ]
 *         # . serialize values
 *         if s_i != 0:             # <<<<<<<<<<<<<<
 *             for i in range(s_i):
 *                 items.append(_ser_common(utils.arr_getitem_1d(arr, i)))
 */
    __pyx_t_3 = (__pyx_v_s_i != 0);
    if (__pyx_t_3) {

      /* "serializor/binary/ser.py":745
 *         # . serialize values
 *         if s_i != 0:
 *             for i in range(s_i):             # <<<<<<<<<<<<<<
 *                 items.append(_ser_common(utils.arr_getitem_1d(arr, i)))
 *         return b"".join(items)
 */
      __pyx_t_6 = __pyx_v_s_i;
      __pyx_t_7 = __pyx_t_6;
      for (__pyx_t_8 = 0; __pyx_t_8 < __pyx_t_7; __pyx_t_8+=1) {
        __pyx_v_i = __pyx_t_8;

        /* "serializor/binary/ser.py":746
 *         if s_i != 0:
 *             for i in range(s_i):
 *                 items.append(_ser_common(utils.arr_getitem_1d(arr, i)))             # <<<<<<<<<<<<<<
 *         return b"".join(items)
 * 
 */
        __pyx_t_5 = __pyx_f_10serializor_5utils_arr_getitem_1d(__pyx_v_arr, __pyx_v_i); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 746, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_4 = __pyx_f_10serializor_6binary_3ser__ser_common(__pyx_t_5); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 746, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __pyx_t_9 = __Pyx_PyList_Append(__pyx_v_items, __pyx_t_4); if (unlikely(__pyx_t_9 == ((int)-1))) __PYX_ERR(0, 746, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      }

      /* "serializor/binary/ser.py":744
 *         ]
 *         # . serialize values
 *         if s_i != 0:             # <<<<<<<<<<<<<<
 *             for i in range(s_i):
 *                 items.append(_ser_common(utils.arr_getitem_1d(arr, i)))
 */
    }

    /* "serializor/binary/ser.py":747
 *             for i in range(s_i):
 *                 items.append(_ser_common(utils.arr_getitem_1d(arr, i)))
 *         return b"".join(items)             # <<<<<<<<<<<<<<
 * 
 *     # 2-dimension
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyBytes_Join(__pyx_kp_b__4, __pyx_v_items); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 747, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    if (!(likely(PyBytes_CheckExact(__pyx_t_4))||((__pyx_t_4) == Py_None) || __Pyx_RaiseUnexpectedTypeError("bytes", __pyx_t_4))) __PYX_ERR(0, 747, __pyx_L1_error)
    __pyx_r = ((PyObject*)__pyx_t_4);
    __pyx_t_4 = 0;
    goto __pyx_L0;

    /* "serializor/binary/ser.py":734
 *     shape = arr.shape
 *     # 1-dimension
 *     if ndim == 1:             # <<<<<<<<<<<<<<
 *         # . array shape
 *         s_i: cython.Py_ssize_t = shape[0]
 */
  }

  /* "serializor/binary/ser.py":750
 * 
 *     # 2-dimension
 *     if ndim == 2:             # <<<<<<<<<<<<<<
 *         # . array shape
 *         s_i: cython.Py_ssize_t = shape[0]
 */
  __pyx_t_3 = (__pyx_v_ndim == 2);
  if (__pyx_t_3) {

    /* "serializor/binary/ser.py":752
 *     if ndim == 2:
 *         # . array shape
 *         s_i: cython.Py_ssize_t = shape[0]             # <<<<<<<<<<<<<<
 *         s_j: cython.Py_ssize_t = shape[1]
 *         items = [
 */
    __pyx_v_s_i = (__pyx_v_shape[0]);

    /* "serializor/binary/ser.py":753
 *         # . array shape
 *         s_i: cython.Py_ssize_t = shape[0]
 *         s_j: cython.Py_ssize_t = shape[1]             # <<<<<<<<<<<<<<
 *         items = [
 *             arr_dtype,  # identifier & dtype
 */
    __pyx_v_s_j = (__pyx_v_shape[1]);

    /* "serializor/binary/ser.py":756
 *         items = [
 *             arr_dtype,  # identifier & dtype
 *             pack_uint8(npy_dtype),  # type: ignore             # <<<<<<<<<<<<<<
 *             b"\x02",  # 2-dimension
 *             gen_encoded_int(s_i),  # type: ignore
 */
    __pyx_t_4 = __pyx_f_10serializor_6binary_3ser_pack_uint8(__pyx_v_npy_dtype); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 756, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);

    /* "serializor/binary/ser.py":758
 *             pack_uint8(npy_dtype),  # type: ignore
 *             b"\x02",  # 2-dimension
 *             gen_encoded_int(s_i),  # type: ignore             # <<<<<<<<<<<<<<
 *             gen_encoded_int(s_j),  # type: ignore
 *         ]
 */
    __pyx_t_5 = __pyx_f_10serializor_6binary_3ser_gen_encoded_int(__pyx_v_s_i); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 758, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);

    /* "serializor/binary/ser.py":759
 *             b"\x02",  # 2-dimension
 *             gen_encoded_int(s_i),  # type: ignore
 *             gen_encoded_int(s_j),  # type: ignore             # <<<<<<<<<<<<<<
 *         ]
 *         # . serialize values
 */
    __pyx_t_1 = __pyx_f_10serializor_6binary_3ser_gen_encoded_int(__pyx_v_s_j); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 759, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);

    /* "serializor/binary/ser.py":754
 *         s_i: cython.Py_ssize_t = shape[0]
 *         s_j: cython.Py_ssize_t = shape[1]
 *         items = [             # <<<<<<<<<<<<<<
 *             arr_dtype,  # identifier & dtype
 *             pack_uint8(npy_dtype),  # type: ignore
 */
    __pyx_t_10 = PyList_New(5); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 754, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_INCREF(__pyx_v_arr_dtype);
    __Pyx_GIVEREF(__pyx_v_arr_dtype);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_10, 0, __pyx_v_arr_dtype)) __PYX_ERR(0, 754, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_4);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_10, 1, __pyx_t_4)) __PYX_ERR(0, 754, __pyx_L1_error);
    __Pyx_INCREF(__pyx_kp_b__6);
    __Pyx_GIVEREF(__pyx_kp_b__6);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_10, 2, __pyx_kp_b__6)) __PYX_ERR(0, 754, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_5);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_10, 3, __pyx_t_5)) __PYX_ERR(0, 754, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_10, 4, __pyx_t_1)) __PYX_ERR(0, 754, __pyx_L1_error);
    __pyx_t_4 = 0;
    __pyx_t_5 = 0;
    __pyx_t_1 = 0;
    __pyx_v_items = ((PyObject*)__pyx_t_10);
    __pyx_t_10 = 0;

    /* "serializor/binary/ser.py":762
 *         ]
 *         # . serialize values
 *         if s_j != 0:             # <<<<<<<<<<<<<<
 *             for i in range(s_i):
 *                 for j in range(s_j):
 */
    __pyx_t_3 = (__pyx_v_s_j != 0);
    if (__pyx_t_3) {

      /* "serializor/binary/ser.py":763
 *         # . serialize values
 *         if s_j != 0:
 *             for i in range(s_i):             # <<<<<<<<<<<<<<
 *                 for j in range(s_j):
 *                     items.append(_ser_common(utils.arr_getitem_2d(arr, i, j)))
 */
      __pyx_t_6 = __pyx_v_s_i;
      __pyx_t_7 = __pyx_t_6;
      for (__pyx_t_8 = 0; __pyx_t_8 < __pyx_t_7; __pyx_t_8+=1) {
        __pyx_v_i = __pyx_t_8;

        /* "serializor/binary/ser.py":764
 *         if s_j != 0:
 *             for i in range(s_i):
 *                 for j in range(s_j):             # <<<<<<<<<<<<<<
 *                     items.append(_ser_common(utils.arr_getitem_2d(arr, i, j)))
 *         return b"".join(items)
 */
        __pyx_t_11 = __pyx_v_s_j;
        __pyx_t_12 = __pyx_t_11;
        for (__pyx_t_13 = 0; __pyx_t_13 < __pyx_t_12; __pyx_t_13+=1) {
          __pyx_v_j = __pyx_t_13;

          /* "serializor/binary/ser.py":765
 *             for i in range(s_i):
 *                 for j in range(s_j):
 *                     items.append(_ser_common(utils.arr_getitem_2d(arr, i, j)))             # <<<<<<<<<<<<<<
 *         return b"".join(items)
 * 
 */
          __pyx_t_10 = __pyx_f_10serializor_5utils_arr_getitem_2d(__pyx_v_arr, __pyx_v_i, __pyx_v_j); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 765, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_10);
          __pyx_t_1 = __pyx_f_10serializor_6binary_3ser__ser_common(__pyx_t_10); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 765, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
          __pyx_t_9 = __Pyx_PyList_Append(__pyx_v_items, __pyx_t_1); if (unlikely(__pyx_t_9 == ((int)-1))) __PYX_ERR(0, 765, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        }
      }

      /* "serializor/binary/ser.py":762
 *         ]
 *         # . serialize values
 *         if s_j != 0:             # <<<<<<<<<<<<<<
 *             for i in range(s_i):
 *                 for j in range(s_j):
 */
    }

    /* "serializor/binary/ser.py":766
 *                 for j in range(s_j):
 *                     items.append(_ser_common(utils.arr_getitem_2d(arr, i, j)))
 *         return b"".join(items)             # <<<<<<<<<<<<<<
 * 
 *     # 3-dimension
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_1 = __Pyx_PyBytes_Join(__pyx_kp_b__4, __pyx_v_items); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 766, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!(likely(PyBytes_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("bytes", __pyx_t_1))) __PYX_ERR(0, 766, __pyx_L1_error)
    __pyx_r = ((PyObject*)__pyx_t_1);
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "serializor/binary/ser.py":750
 * 
 *     # 2-dimension
 *     if ndim == 2:             # <<<<<<<<<<<<<<
 *         # . array shape
 *         s_i: cython.Py_ssize_t = shape[0]
 */
  }

  /* "serializor/binary/ser.py":769
 * 
 *     # 3-dimension
 *     if ndim == 3:             # <<<<<<<<<<<<<<
 *         # . array shape
 *         s_i: cython.Py_ssize_t = shape[0]
 */
  __pyx_t_3 = (__pyx_v_ndim == 3);
  if (__pyx_t_3) {

    /* "serializor/binary/ser.py":771
 *     if ndim == 3:
 *         # . array shape
 *         s_i: cython.Py_ssize_t = shape[0]             # <<<<<<<<<<<<<<
 *         s_j: cython.Py_ssize_t = shape[1]
 *         s_k: cython.Py_ssize_t = shape[2]
 */
    __pyx_v_s_i = (__pyx_v_shape[0]);

    /* "serializor/binary/ser.py":772
 *         # . array shape
 *         s_i: cython.Py_ssize_t = shape[0]
 *         s_j: cython.Py_ssize_t = shape[1]             # <<<<<<<<<<<<<<
 *         s_k: cython.Py_ssize_t = shape[2]
 *         items = [
 */
    __pyx_v_s_j = (__pyx_v_shape[1]);

    /* "serializor/binary/ser.py":773
 *         s_i: cython.Py_ssize_t = shape[0]
 *         s_j: cython.Py_ssize_t = shape[1]
 *         s_k: cython.Py_ssize_t = shape[2]             # <<<<<<<<<<<<<<
 *         items = [
 *             arr_dtype,  # identifier & dtype
 */
    __pyx_v_s_k = (__pyx_v_shape[2]);

    /* "serializor/binary/ser.py":776
 *         items = [
 *             arr_dtype,  # identifier & dtype
 *             pack_uint8(npy_dtype),  # type: ignore             # <<<<<<<<<<<<<<
 *             b"\x03",  # 3-dimension
 *             gen_encoded_int(s_i),  # type: ignore
 */
    __pyx_t_1 = __pyx_f_10serializor_6binary_3ser_pack_uint8(__pyx_v_npy_dtype); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 776, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);

    /* "serializor/binary/ser.py":778
 *             pack_uint8(npy_dtype),  # type: ignore
 *             b"\x03",  # 3-dimension
 *             gen_encoded_int(s_i),  # type: ignore             # <<<<<<<<<<<<<<
 *             gen_encoded_int(s_j),  # type: ignore
 *             gen_encoded_int(s_k),  # type: ignore
 */
    __pyx_t_10 = __pyx_f_10serializor_6binary_3ser_gen_encoded_int(__pyx_v_s_i); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 778, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);

    /* "serializor/binary/ser.py":779
 *             b"\x03",  # 3-dimension
 *             gen_encoded_int(s_i),  # type: ignore
 *             gen_encoded_int(s_j),  # type: ignore             # <<<<<<<<<<<<<<
 *             gen_encoded_int(s_k),  # type: ignore
 *         ]
 */
    __pyx_t_5 = __pyx_f_10serializor_6binary_3ser_gen_encoded_int(__pyx_v_s_j); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 779, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);

    /* "serializor/binary/ser.py":780
 *             gen_encoded_int(s_i),  # type: ignore
 *             gen_encoded_int(s_j),  # type: ignore
 *             gen_encoded_int(s_k),  # type: ignore             # <<<<<<<<<<<<<<
 *         ]
 *         # . serialize values
 */
    __pyx_t_4 = __pyx_f_10serializor_6binary_3ser_gen_encoded_int(__pyx_v_s_k); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 780, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);

    /* "serializor/binary/ser.py":774
 *         s_j: cython.Py_ssize_t = shape[1]
 *         s_k: cython.Py_ssize_t = shape[2]
 *         items = [             # <<<<<<<<<<<<<<
 *             arr_dtype,  # identifier & dtype
 *             pack_uint8(npy_dtype),  # type: ignore
 */
    __pyx_t_14 = PyList_New(6); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 774, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __Pyx_INCREF(__pyx_v_arr_dtype);
    __Pyx_GIVEREF(__pyx_v_arr_dtype);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_14, 0, __pyx_v_arr_dtype)) __PYX_ERR(0, 774, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_14, 1, __pyx_t_1)) __PYX_ERR(0, 774, __pyx_L1_error);
    __Pyx_INCREF(__pyx_kp_b__8);
    __Pyx_GIVEREF(__pyx_kp_b__8);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_14, 2, __pyx_kp_b__8)) __PYX_ERR(0, 774, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_10);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_14, 3, __pyx_t_10)) __PYX_ERR(0, 774, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_5);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_14, 4, __pyx_t_5)) __PYX_ERR(0, 774, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_4);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_14, 5, __pyx_t_4)) __PYX_ERR(0, 774, __pyx_L1_error);
    __pyx_t_1 = 0;
    __pyx_t_10 = 0;
    __pyx_t_5 = 0;
    __pyx_t_4 = 0;
    __pyx_v_items = ((PyObject*)__pyx_t_14);
    __pyx_t_14 = 0;

    /* "serializor/binary/ser.py":783
 *         ]
 *         # . serialize values
 *         if s_k != 0:             # <<<<<<<<<<<<<<
 *             for i in range(s_i):
 *                 for j in range(s_j):
 */
    __pyx_t_3 = (__pyx_v_s_k != 0);
    if (__pyx_t_3) {

      /* "serializor/binary/ser.py":784
 *         # . serialize values
 *         if s_k != 0:
 *             for i in range(s_i):             # <<<<<<<<<<<<<<
 *                 for j in range(s_j):
 *                     for k in range(s_k):
 */
      __pyx_t_6 = __pyx_v_s_i;
      __pyx_t_7 = __pyx_t_6;
      for (__pyx_t_8 = 0; __pyx_t_8 < __pyx_t_7; __pyx_t_8+=1) {
        __pyx_v_i = __pyx_t_8;

        /* "serializor/binary/ser.py":785
 *         if s_k != 0:
 *             for i in range(s_i):
 *                 for j in range(s_j):             # <<<<<<<<<<<<<<
 *                     for k in range(s_k):
 *                         items.append(_ser_common(utils.arr_getitem_3d(arr, i, j, k)))
 */
        __pyx_t_11 = __pyx_v_s_j;
        __pyx_t_12 = __pyx_t_11;
        for (__pyx_t_13 = 0; __pyx_t_13 < __pyx_t_12; __pyx_t_13+=1) {
          __pyx_v_j = __pyx_t_13;

          /* "serializor/binary/ser.py":786
 *             for i in range(s_i):
 *                 for j in range(s_j):
 *                     for k in range(s_k):             # <<<<<<<<<<<<<<
 *                         items.append(_ser_common(utils.arr_getitem_3d(arr, i, j, k)))
 *         return b"".join(items)
 */
          __pyx_t_15 = __pyx_v_s_k;
          __pyx_t_16 = __pyx_t_15;
          for (__pyx_t_17 = 0; __pyx_t_17 < __pyx_t_16; __pyx_t_17+=1) {
            __pyx_v_k = __pyx_t_17;

            /* "serializor/binary/ser.py":787
 *                 for j in range(s_j):
 *                     for k in range(s_k):
 *                         items.append(_ser_common(utils.arr_getitem_3d(arr, i, j, k)))             # <<<<<<<<<<<<<<
 *         return b"".join(items)
 * 
 */
            __pyx_t_14 = __pyx_f_10serializor_5utils_arr_getitem_3d(__pyx_v_arr, __pyx_v_i, __pyx_v_j, __pyx_v_k); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 787, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_14);
            __pyx_t_4 = __pyx_f_10serializor_6binary_3ser__ser_common(__pyx_t_14); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 787, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_4);
            __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
            __pyx_t_9 = __Pyx_PyList_Append(__pyx_v_items, __pyx_t_4); if (unlikely(__pyx_t_9 == ((int)-1))) __PYX_ERR(0, 787, __pyx_L1_error)
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          }
        }
      }

      /* "serializor/binary/ser.py":783
 *         ]
 *         # . serialize values
 *         if s_k != 0:             # <<<<<<<<<<<<<<
 *             for i in range(s_i):
 *                 for j in range(s_j):
 */
    }

    /* "serializor/binary/ser.py":788
 *                     for k in range(s_k):
 *                         items.append(_ser_common(utils.arr_getitem_3d(arr, i, j, k)))
 *         return b"".join(items)             # <<<<<<<<<<<<<<
 * 
 *     # 4-dimension
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyBytes_Join(__pyx_kp_b__4, __pyx_v_items); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 788, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    if (!(likely(PyBytes_CheckExact(__pyx_t_4))||((__pyx_t_4) == Py_None) || __Pyx_RaiseUnexpectedTypeError("bytes", __pyx_t_4))) __PYX_ERR(0, 788, __pyx_L1_error)
    __pyx_r = ((PyObject*)__pyx_t_4);
    __pyx_t_4 = 0;
    goto __pyx_L0;

    /* "serializor/binary/ser.py":769
 * 
 *     # 3-dimension
 *     if ndim == 3:             # <<<<<<<<<<<<<<
 *         # . array shape
 *         s_i: cython.Py_ssize_t = shape[0]
 */
  }

  /* "serializor/binary/ser.py":791
 * 
 *     # 4-dimension
 *     if ndim == 4:             # <<<<<<<<<<<<<<
 *         # . array shape
 *         s_i: cython.Py_ssize_t = shape[0]
 */
  __pyx_t_3 = (__pyx_v_ndim == 4);
  if (__pyx_t_3) {

    /* "serializor/binary/ser.py":793
 *     if ndim == 4:
 *         # . array shape
 *         s_i: cython.Py_ssize_t = shape[0]             # <<<<<<<<<<<<<<
 *         s_j: cython.Py_ssize_t = shape[1]
 *         s_k: cython.Py_ssize_t = shape[2]
 */
    __pyx_v_s_i = (__pyx_v_shape[0]);

    /* "serializor/binary/ser.py":794
 *         # . array shape
 *         s_i: cython.Py_ssize_t = shape[0]
 *         s_j: cython.Py_ssize_t = shape[1]             # <<<<<<<<<<<<<<
 *         s_k: cython.Py_ssize_t = shape[2]
 *         s_l: cython.Py_ssize_t = shape[3]
 */
    __pyx_v_s_j = (__pyx_v_shape[1]);

    /* "serializor/binary/ser.py":795
 *         s_i: cython.Py_ssize_t = shape[0]
 *         s_j: cython.Py_ssize_t = shape[1]
 *         s_k: cython.Py_ssize_t = shape[2]             # <<<<<<<<<<<<<<
 *         s_l: cython.Py_ssize_t = shape[3]
 *         items = [
 */
    __pyx_v_s_k = (__pyx_v_shape[2]);

    /* "serializor/binary/ser.py":796
 *         s_j: cython.Py_ssize_t = shape[1]
 *         s_k: cython.Py_ssize_t = shape[2]
 *         s_l: cython.Py_ssize_t = shape[3]             # <<<<<<<<<<<<<<
 *         items = [
 *             arr_dtype,  # identifier & dtype
 */
    __pyx_v_s_l = (__pyx_v_shape[3]);

    /* "serializor/binary/ser.py":799
 *         items = [
 *             arr_dtype,  # identifier & dtype
 *             pack_uint8(npy_dtype),  # type: ignore             # <<<<<<<<<<<<<<
 *             b"\x04",  # 4-dimension
 *             gen_encoded_int(s_i),  # type: ignore
 */
    __pyx_t_4 = __pyx_f_10serializor_6binary_3ser_pack_uint8(__pyx_v_npy_dtype); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 799, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);

    /* "serializor/binary/ser.py":801
 *             pack_uint8(npy_dtype),  # type: ignore
 *             b"\x04",  # 4-dimension
 *             gen_encoded_int(s_i),  # type: ignore             # <<<<<<<<<<<<<<
 *             gen_encoded_int(s_j),  # type: ignore
 *             gen_encoded_int(s_k),  # type: ignore
 */
    __pyx_t_14 = __pyx_f_10serializor_6binary_3ser_gen_encoded_int(__pyx_v_s_i); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 801, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);

    /* "serializor/binary/ser.py":802
 *             b"\x04",  # 4-dimension
 *             gen_encoded_int(s_i),  # type: ignore
 *             gen_encoded_int(s_j),  # type: ignore             # <<<<<<<<<<<<<<
 *             gen_encoded_int(s_k),  # type: ignore
 *             gen_encoded_int(s_l),  # type: ignore
 */
    __pyx_t_5 = __pyx_f_10serializor_6binary_3ser_gen_encoded_int(__pyx_v_s_j); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 802, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);

    /* "serializor/binary/ser.py":803
 *             gen_encoded_int(s_i),  # type: ignore
 *             gen_encoded_int(s_j),  # type: ignore
 *             gen_encoded_int(s_k),  # type: ignore             # <<<<<<<<<<<<<<
 *             gen_encoded_int(s_l),  # type: ignore
 *         ]
 */
    __pyx_t_10 = __pyx_f_10serializor_6binary_3ser_gen_encoded_int(__pyx_v_s_k); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 803, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);

    /* "serializor/binary/ser.py":804
 *             gen_encoded_int(s_j),  # type: ignore
 *             gen_encoded_int(s_k),  # type: ignore
 *             gen_encoded_int(s_l),  # type: ignore             # <<<<<<<<<<<<<<
 *         ]
 *         # . serialize values
 */
    __pyx_t_1 = __pyx_f_10serializor_6binary_3ser_gen_encoded_int(__pyx_v_s_l); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 804, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);

    /* "serializor/binary/ser.py":797
 *         s_k: cython.Py_ssize_t = shape[2]
 *         s_l: cython.Py_ssize_t = shape[3]
 *         items = [             # <<<<<<<<<<<<<<
 *             arr_dtype,  # identifier & dtype
 *             pack_uint8(npy_dtype),  # type: ignore
 */
    __pyx_t_18 = PyList_New(7); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 797, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_18);
    __Pyx_INCREF(__pyx_v_arr_dtype);
    __Pyx_GIVEREF(__pyx_v_arr_dtype);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_18, 0, __pyx_v_arr_dtype)) __PYX_ERR(0, 797, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_4);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_18, 1, __pyx_t_4)) __PYX_ERR(0, 797, __pyx_L1_error);
    __Pyx_INCREF(__pyx_kp_b__9);
    __Pyx_GIVEREF(__pyx_kp_b__9);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_18, 2, __pyx_kp_b__9)) __PYX_ERR(0, 797, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_14);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_18, 3, __pyx_t_14)) __PYX_ERR(0, 797, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_5);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_18, 4, __pyx_t_5)) __PYX_ERR(0, 797, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_10);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_18, 5, __pyx_t_10)) __PYX_ERR(0, 797, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_18, 6, __pyx_t_1)) __PYX_ERR(0, 797, __pyx_L1_error);
    __pyx_t_4 = 0;
    __pyx_t_14 = 0;
    __pyx_t_5 = 0;
    __pyx_t_10 = 0;
    __pyx_t_1 = 0;
    __pyx_v_items = ((PyObject*)__pyx_t_18);
    __pyx_t_18 = 0;

    /* "serializor/binary/ser.py":807
 *         ]
 *         # . serialize values
 *         if s_l != 0:             # <<<<<<<<<<<<<<
 *             for i in range(s_i):
 *                 for j in range(s_j):
 */
    __pyx_t_3 = (__pyx_v_s_l != 0);
    if (__pyx_t_3) {

      /* "serializor/binary/ser.py":808
 *         # . serialize values
 *         if s_l != 0:
 *             for i in range(s_i):             # <<<<<<<<<<<<<<
 *                 for j in range(s_j):
 *                     for k in range(s_k):
 */
      __pyx_t_6 = __pyx_v_s_i;
      __pyx_t_7 = __pyx_t_6;
      for (__pyx_t_8 = 0; __pyx_t_8 < __pyx_t_7; __pyx_t_8+=1) {
        __pyx_v_i = __pyx_t_8;

        /* "serializor/binary/ser.py":809
 *         if s_l != 0:
 *             for i in range(s_i):
 *                 for j in range(s_j):             # <<<<<<<<<<<<<<
 *                     for k in range(s_k):
 *                         for l in range(s_l):
 */
        __pyx_t_11 = __pyx_v_s_j;
        __pyx_t_12 = __pyx_t_11;
        for (__pyx_t_13 = 0; __pyx_t_13 < __pyx_t_12; __pyx_t_13+=1) {
          __pyx_v_j = __pyx_t_13;

          /* "serializor/binary/ser.py":810
 *             for i in range(s_i):
 *                 for j in range(s_j):
 *                     for k in range(s_k):             # <<<<<<<<<<<<<<
 *                         for l in range(s_l):
 *                             items.append(_ser_common(utils.arr_getitem_4d(arr, i, j, k, l)))  # type: ignore
 */
          __pyx_t_15 = __pyx_v_s_k;
          __pyx_t_16 = __pyx_t_15;
          for (__pyx_t_17 = 0; __pyx_t_17 < __pyx_t_16; __pyx_t_17+=1) {
            __pyx_v_k = __pyx_t_17;

            /* "serializor/binary/ser.py":811
 *                 for j in range(s_j):
 *                     for k in range(s_k):
 *                         for l in range(s_l):             # <<<<<<<<<<<<<<
 *                             items.append(_ser_common(utils.arr_getitem_4d(arr, i, j, k, l)))  # type: ignore
 *         return b"".join(items)
 */
            __pyx_t_19 = __pyx_v_s_l;
            __pyx_t_20 = __pyx_t_19;
            for (__pyx_t_21 = 0; __pyx_t_21 < __pyx_t_20; __pyx_t_21+=1) {
              __pyx_v_l = __pyx_t_21;

              /* "serializor/binary/ser.py":812
 *                     for k in range(s_k):
 *                         for l in range(s_l):
 *                             items.append(_ser_common(utils.arr_getitem_4d(arr, i, j, k, l)))  # type: ignore             # <<<<<<<<<<<<<<
 *         return b"".join(items)
 * 
 */
              __pyx_t_18 = __pyx_f_10serializor_5utils_arr_getitem_4d(__pyx_v_arr, __pyx_v_i, __pyx_v_j, __pyx_v_k, __pyx_v_l); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 812, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_18);
              __pyx_t_1 = __pyx_f_10serializor_6binary_3ser__ser_common(__pyx_t_18); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 812, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
              __pyx_t_9 = __Pyx_PyList_Append(__pyx_v_items, __pyx_t_1); if (unlikely(__pyx_t_9 == ((int)-1))) __PYX_ERR(0, 812, __pyx_L1_error)
              __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            }
          }
        }
      }

      /* "serializor/binary/ser.py":807
 *         ]
 *         # . serialize values
 *         if s_l != 0:             # <<<<<<<<<<<<<<
 *             for i in range(s_i):
 *                 for j in range(s_j):
 */
    }

    /* "serializor/binary/ser.py":813
 *                         for l in range(s_l):
 *                             items.append(_ser_common(utils.arr_getitem_4d(arr, i, j, k, l)))  # type: ignore
 *         return b"".join(items)             # <<<<<<<<<<<<<<
 * 
 *     # invalid
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_1 = __Pyx_PyBytes_Join(__pyx_kp_b__4, __pyx_v_items); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 813, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!(likely(PyBytes_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("bytes", __pyx_t_1))) __PYX_ERR(0, 813, __pyx_L1_error)
    __pyx_r = ((PyObject*)__pyx_t_1);
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "serializor/binary/ser.py":791
 * 
 *     # 4-dimension
 *     if ndim == 4:             # <<<<<<<<<<<<<<
 *         # . array shape
 *         s_i: cython.Py_ssize_t = shape[0]
 */
  }

  /* "serializor/binary/ser.py":816
 * 
 *     # invalid
 *     raise ValueError("unsupported <'numpy.ndarray'> dimension [%d]." % ndim)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_ndim); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 816, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_18 = PyUnicode_Format(__pyx_kp_u_unsupported_numpy_ndarray_dimens, __pyx_t_1); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 816, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_18);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_18); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 816, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(0, 816, __pyx_L1_error)

  /* "serializor/binary/ser.py":711
 * 
 * 
 * @cython.cfunc             # <<<<<<<<<<<<<<
 * @cython.inline(True)
 * def _ser_ndarray_object(arr: np.ndarray, arr_dtype: bytes) -> bytes:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_XDECREF(__pyx_t_18);
  __Pyx_AddTraceback("serializor.binary.ser._ser_ndarray_object", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_items);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "serializor/binary/ser.py":819
 * 
 * 
 * @cython.cfunc             # <<<<<<<<<<<<<<
 * @cython.inline(True)
 * def _ser_ndarray_numeric(
 */

static PyObject *__pyx_f_10serializor_6binary_3ser__ser_ndarray_numeric(PyArrayObject *__pyx_v_arr, PyObject *__pyx_v_arr_dtype, int __pyx_v_npy_dtype) {
  int __pyx_v_ndim;
  npy_intp *__pyx_v_shape;
  Py_ssize_t __pyx_v_s_i;
  PyObject *__pyx_v_items = NULL;
  PyObject *__pyx_v_values_json = NULL;
  Py_ssize_t __pyx_v_s_j;
  Py_ssize_t __pyx_v_s_k;
  Py_ssize_t __pyx_v_s_l;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  Py_ssize_t __pyx_t_5;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_ser_ndarray_numeric", 1);

  /* "serializor/binary/ser.py":843
 *     >>> b'...[SHAPE][values_json_size(ENC)][values_json]'
 *     """
 *     ndim: cython.int = arr.ndim             # <<<<<<<<<<<<<<
 *     shape = arr.shape
 *     # 1-dimension
 */
  __pyx_v_ndim = __pyx_f_5numpy_7ndarray_4ndim_ndim(__pyx_v_arr);

  /* "serializor/binary/ser.py":844
 *     """
 *     ndim: cython.int = arr.ndim
 *     shape = arr.shape             # <<<<<<<<<<<<<<
 *     # 1-dimension
 *     if ndim == 1:
 */
  __pyx_v_shape = __pyx_f_5numpy_7ndarray_5shape_shape(__pyx_v_arr);

  /* "serializor/binary/ser.py":846
 *     shape = arr.shape
 *     # 1-dimension
 *     if ndim == 1:             # <<<<<<<<<<<<<<
 *         # . array shape
 *         s_i: cython.Py_ssize_t = shape[0]
 */
  __pyx_t_1 = (__pyx_v_ndim == 1);
  if (__pyx_t_1) {

    /* "serializor/binary/ser.py":848
 *     if ndim == 1:
 *         # . array shape
 *         s_i: cython.Py_ssize_t = shape[0]             # <<<<<<<<<<<<<<
 *         items = [
 *             arr_dtype,  # identifier & dtype
 */
    __pyx_v_s_i = (__pyx_v_shape[0]);

    /* "serializor/binary/ser.py":851
 *         items = [
 *             arr_dtype,  # identifier & dtype
 *             pack_uint8(npy_dtype),  # type: ignore             # <<<<<<<<<<<<<<
 *             b"\x01",  # 1-dimension
 *             gen_encoded_int(s_i),  # type: ignore
 */
    __pyx_t_2 = __pyx_f_10serializor_6binary_3ser_pack_uint8(__pyx_v_npy_dtype); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 851, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);

    /* "serializor/binary/ser.py":853
 *             pack_uint8(npy_dtype),  # type: ignore
 *             b"\x01",  # 1-dimension
 *             gen_encoded_int(s_i),  # type: ignore             # <<<<<<<<<<<<<<
 *         ]
 *         # . serialize values
 */
    __pyx_t_3 = __pyx_f_10serializor_6binary_3ser_gen_encoded_int(__pyx_v_s_i); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 853, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);

    /* "serializor/binary/ser.py":849
 *         # . array shape
 *         s_i: cython.Py_ssize_t = shape[0]
 *         items = [             # <<<<<<<<<<<<<<
 *             arr_dtype,  # identifier & dtype
 *             pack_uint8(npy_dtype),  # type: ignore
 */
    __pyx_t_4 = PyList_New(4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 849, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_INCREF(__pyx_v_arr_dtype);
    __Pyx_GIVEREF(__pyx_v_arr_dtype);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_4, 0, __pyx_v_arr_dtype)) __PYX_ERR(0, 849, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_2);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_4, 1, __pyx_t_2)) __PYX_ERR(0, 849, __pyx_L1_error);
    __Pyx_INCREF(__pyx_kp_b__5);
    __Pyx_GIVEREF(__pyx_kp_b__5);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_4, 2, __pyx_kp_b__5)) __PYX_ERR(0, 849, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_3);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_4, 3, __pyx_t_3)) __PYX_ERR(0, 849, __pyx_L1_error);
    __pyx_t_2 = 0;
    __pyx_t_3 = 0;
    __pyx_v_items = ((PyObject*)__pyx_t_4);
    __pyx_t_4 = 0;

    /* "serializor/binary/ser.py":856
 *         ]
 *         # . serialize values
 *         if s_i != 0:             # <<<<<<<<<<<<<<
 *             values_json = _orjson_dumps_numpy(arr)
 *             items.append(gen_encoded_int(bytes_len(values_json)))  # type: ignore
 */
    __pyx_t_1 = (__pyx_v_s_i != 0);
    if (__pyx_t_1) {

      /* "serializor/binary/ser.py":857
 *         # . serialize values
 *         if s_i != 0:
 *             values_json = _orjson_dumps_numpy(arr)             # <<<<<<<<<<<<<<
 *             items.append(gen_encoded_int(bytes_len(values_json)))  # type: ignore
 *             items.append(values_json)
 */
      __pyx_t_4 = __pyx_f_10serializor_6binary_3ser__orjson_dumps_numpy(((PyObject *)__pyx_v_arr)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 857, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_v_values_json = ((PyObject*)__pyx_t_4);
      __pyx_t_4 = 0;

      /* "serializor/binary/ser.py":858
 *         if s_i != 0:
 *             values_json = _orjson_dumps_numpy(arr)
 *             items.append(gen_encoded_int(bytes_len(values_json)))  # type: ignore             # <<<<<<<<<<<<<<
 *             items.append(values_json)
 *         return b"".join(items)
 */
      __pyx_t_5 = PyBytes_Size(__pyx_v_values_json); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1L))) __PYX_ERR(0, 858, __pyx_L1_error)
      __pyx_t_4 = __pyx_f_10serializor_6binary_3ser_gen_encoded_int(__pyx_t_5); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 858, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_6 = __Pyx_PyList_Append(__pyx_v_items, __pyx_t_4); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(0, 858, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

      /* "serializor/binary/ser.py":859
 *             values_json = _orjson_dumps_numpy(arr)
 *             items.append(gen_encoded_int(bytes_len(values_json)))  # type: ignore
 *             items.append(values_json)             # <<<<<<<<<<<<<<
 *         return b"".join(items)
 * 
 */
      __pyx_t_6 = __Pyx_PyList_Append(__pyx_v_items, __pyx_v_values_json); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(0, 859, __pyx_L1_error)

      /* "serializor/binary/ser.py":856
 *         ]
 *         # . serialize values
 *         if s_i != 0:             # <<<<<<<<<<<<<<
 *             values_json = _orjson_dumps_numpy(arr)
 *             items.append(gen_encoded_int(bytes_len(values_json)))  # type: ignore
 */
    }

    /* "serializor/binary/ser.py":860
 *             items.append(gen_encoded_int(bytes_len(values_json)))  # type: ignore
 *             items.append(values_json)
 *         return b"".join(items)             # <<<<<<<<<<<<<<
 * 
 *     # 2-dimension
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyBytes_Join(__pyx_kp_b__4, __pyx_v_items); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 860, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    if (!(likely(PyBytes_CheckExact(__pyx_t_4))||((__pyx_t_4) == Py_None) || __Pyx_RaiseUnexpectedTypeError("bytes", __pyx_t_4))) __PYX_ERR(0, 860, __pyx_L1_error)
    __pyx_r = ((PyObject*)__pyx_t_4);
    __pyx_t_4 = 0;
    goto __pyx_L0;

    /* "serializor/binary/ser.py":846
 *     shape = arr.shape
 *     # 1-dimension
 *     if ndim == 1:             # <<<<<<<<<<<<<<
 *         # . array shape
 *         s_i: cython.Py_ssize_t = shape[0]
 */
  }

  /* "serializor/binary/ser.py":863
 * 
 *     # 2-dimension
 *     if ndim == 2:             # <<<<<<<<<<<<<<
 *         s_i: cython.Py_ssize_t = shape[0]
 *         s_j: cython.Py_ssize_t = shape[1]
 */
  __pyx_t_1 = (__pyx_v_ndim == 2);
  if (__pyx_t_1) {

    /* "serializor/binary/ser.py":864
 *     # 2-dimension
 *     if ndim == 2:
 *         s_i: cython.Py_ssize_t = shape[0]             # <<<<<<<<<<<<<<
 *         s_j: cython.Py_ssize_t = shape[1]
 *         items = [
 */
    __pyx_v_s_i = (__pyx_v_shape[0]);

    /* "serializor/binary/ser.py":865
 *     if ndim == 2:
 *         s_i: cython.Py_ssize_t = shape[0]
 *         s_j: cython.Py_ssize_t = shape[1]             # <<<<<<<<<<<<<<
 *         items = [
 *             arr_dtype,  # identifier & dtype
 */
    __pyx_v_s_j = (__pyx_v_shape[1]);

    /* "serializor/binary/ser.py":868
 *         items = [
 *             arr_dtype,  # identifier & dtype
 *             pack_uint8(npy_dtype),  # type: ignore             # <<<<<<<<<<<<<<
 *             b"\x02",  # 2-dimension
 *             gen_encoded_int(s_i),  # type: ignore
 */
    __pyx_t_4 = __pyx_f_10serializor_6binary_3ser_pack_uint8(__pyx_v_npy_dtype); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 868, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);

    /* "serializor/binary/ser.py":870
 *             pack_uint8(npy_dtype),  # type: ignore
 *             b"\x02",  # 2-dimension
 *             gen_encoded_int(s_i),  # type: ignore             # <<<<<<<<<<<<<<
 *             gen_encoded_int(s_j),  # type: ignore
 *         ]
 */
    __pyx_t_3 = __pyx_f_10serializor_6binary_3ser_gen_encoded_int(__pyx_v_s_i); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 870, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);

    /* "serializor/binary/ser.py":871
 *             b"\x02",  # 2-dimension
 *             gen_encoded_int(s_i),  # type: ignore
 *             gen_encoded_int(s_j),  # type: ignore             # <<<<<<<<<<<<<<
 *         ]
 *         # . serialize values
 */
    __pyx_t_2 = __pyx_f_10serializor_6binary_3ser_gen_encoded_int(__pyx_v_s_j); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 871, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);

    /* "serializor/binary/ser.py":866
 *         s_i: cython.Py_ssize_t = shape[0]
 *         s_j: cython.Py_ssize_t = shape[1]
 *         items = [             # <<<<<<<<<<<<<<
 *             arr_dtype,  # identifier & dtype
 *             pack_uint8(npy_dtype),  # type: ignore
 */
    __pyx_t_7 = PyList_New(5); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 866, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_INCREF(__pyx_v_arr_dtype);
    __Pyx_GIVEREF(__pyx_v_arr_dtype);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_7, 0, __pyx_v_arr_dtype)) __PYX_ERR(0, 866, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_4);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_7, 1, __pyx_t_4)) __PYX_ERR(0, 866, __pyx_L1_error);
    __Pyx_INCREF(__pyx_kp_b__6);
    __Pyx_GIVEREF(__pyx_kp_b__6);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_7, 2, __pyx_kp_b__6)) __PYX_ERR(0, 866, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_3);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_7, 3, __pyx_t_3)) __PYX_ERR(0, 866, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_2);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_7, 4, __pyx_t_2)) __PYX_ERR(0, 866, __pyx_L1_error);
    __pyx_t_4 = 0;
    __pyx_t_3 = 0;
    __pyx_t_2 = 0;
    __pyx_v_items = ((PyObject*)__pyx_t_7);
    __pyx_t_7 = 0;

    /* "serializor/binary/ser.py":874
 *         ]
 *         # . serialize values
 *         if s_j != 0:             # <<<<<<<<<<<<<<
 *             values_json = _orjson_dumps_numpy(utils.arr_flatten(arr))
 *             items.append(gen_encoded_int(bytes_len(values_json)))  # type: ignore
 */
    __pyx_t_1 = (__pyx_v_s_j != 0);
    if (__pyx_t_1) {

      /* "serializor/binary/ser.py":875
 *         # . serialize values
 *         if s_j != 0:
 *             values_json = _orjson_dumps_numpy(utils.arr_flatten(arr))             # <<<<<<<<<<<<<<
 *             items.append(gen_encoded_int(bytes_len(values_json)))  # type: ignore
 *             items.append(values_json)
 */
      __pyx_t_7 = ((PyObject *)__pyx_f_10serializor_5utils_arr_flatten(__pyx_v_arr)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 875, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_2 = __pyx_f_10serializor_6binary_3ser__orjson_dumps_numpy(__pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 875, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __pyx_v_values_json = ((PyObject*)__pyx_t_2);
      __pyx_t_2 = 0;

      /* "serializor/binary/ser.py":876
 *         if s_j != 0:
 *             values_json = _orjson_dumps_numpy(utils.arr_flatten(arr))
 *             items.append(gen_encoded_int(bytes_len(values_json)))  # type: ignore             # <<<<<<<<<<<<<<
 *             items.append(values_json)
 *         return b"".join(items)
 */
      __pyx_t_5 = PyBytes_Size(__pyx_v_values_json); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1L))) __PYX_ERR(0, 876, __pyx_L1_error)
      __pyx_t_2 = __pyx_f_10serializor_6binary_3ser_gen_encoded_int(__pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 876, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_6 = __Pyx_PyList_Append(__pyx_v_items, __pyx_t_2); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(0, 876, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "serializor/binary/ser.py":877
 *             values_json = _orjson_dumps_numpy(utils.arr_flatten(arr))
 *             items.append(gen_encoded_int(bytes_len(values_json)))  # type: ignore
 *             items.append(values_json)             # <<<<<<<<<<<<<<
 *         return b"".join(items)
 * 
 */
      __pyx_t_6 = __Pyx_PyList_Append(__pyx_v_items, __pyx_v_values_json); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(0, 877, __pyx_L1_error)

      /* "serializor/binary/ser.py":874
 *         ]
 *         # . serialize values
 *         if s_j != 0:             # <<<<<<<<<<<<<<
 *             values_json = _orjson_dumps_numpy(utils.arr_flatten(arr))
 *             items.append(gen_encoded_int(bytes_len(values_json)))  # type: ignore
 */
    }

    /* "serializor/binary/ser.py":878
 *             items.append(gen_encoded_int(bytes_len(values_json)))  # type: ignore
 *             items.append(values_json)
 *         return b"".join(items)             # <<<<<<<<<<<<<<
 * 
 *     # 3-dimension
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = __Pyx_PyBytes_Join(__pyx_kp_b__4, __pyx_v_items); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 878, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (!(likely(PyBytes_CheckExact(__pyx_t_2))||((__pyx_t_2) == Py_None) || __Pyx_RaiseUnexpectedTypeError("bytes", __pyx_t_2))) __PYX_ERR(0, 878, __pyx_L1_error)
    __pyx_r = ((PyObject*)__pyx_t_2);
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "serializor/binary/ser.py":863
 * 
 *     # 2-dimension
 *     if ndim == 2:             # <<<<<<<<<<<<<<
 *         s_i: cython.Py_ssize_t = shape[0]
 *         s_j: cython.Py_ssize_t = shape[1]
 */
  }

  /* "serializor/binary/ser.py":881
 * 
 *     # 3-dimension
 *     if ndim == 3:             # <<<<<<<<<<<<<<
 *         s_i: cython.Py_ssize_t = shape[0]
 *         s_j: cython.Py_ssize_t = shape[1]
 */
  __pyx_t_1 = (__pyx_v_ndim == 3);
  if (__pyx_t_1) {

    /* "serializor/binary/ser.py":882
 *     # 3-dimension
 *     if ndim == 3:
 *         s_i: cython.Py_ssize_t = shape[0]             # <<<<<<<<<<<<<<
 *         s_j: cython.Py_ssize_t = shape[1]
 *         s_k: cython.Py_ssize_t = shape[2]
 */
    __pyx_v_s_i = (__pyx_v_shape[0]);

    /* "serializor/binary/ser.py":883
 *     if ndim == 3:
 *         s_i: cython.Py_ssize_t = shape[0]
 *         s_j: cython.Py_ssize_t = shape[1]             # <<<<<<<<<<<<<<
 *         s_k: cython.Py_ssize_t = shape[2]
 *         items = [
 */
    __pyx_v_s_j = (__pyx_v_shape[1]);

    /* "serializor/binary/ser.py":884
 *         s_i: cython.Py_ssize_t = shape[0]
 *         s_j: cython.Py_ssize_t = shape[1]
 *         s_k: cython.Py_ssize_t = shape[2]             # <<<<<<<<<<<<<<
 *         items = [
 *             arr_dtype,  # identifier & dtype
 */
    __pyx_v_s_k = (__pyx_v_shape[2]);

    /* "serializor/binary/ser.py":887
 *         items = [
 *             arr_dtype,  # identifier & dtype
 *             pack_uint8(npy_dtype),  # type: ignore             # <<<<<<<<<<<<<<
 *             b"\x03",  # 3-dimension
 *             gen_encoded_int(s_i),  # type: ignore
 */
    __pyx_t_2 = __pyx_f_10serializor_6binary_3ser_pack_uint8(__pyx_v_npy_dtype); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 887, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);

    /* "serializor/binary/ser.py":889
 *             pack_uint8(npy_dtype),  # type: ignore
 *             b"\x03",  # 3-dimension
 *             gen_encoded_int(s_i),  # type: ignore             # <<<<<<<<<<<<<<
 *             gen_encoded_int(s_j),  # type: ignore
 *             gen_encoded_int(s_k),  # type: ignore
 */
    __pyx_t_7 = __pyx_f_10serializor_6binary_3ser_gen_encoded_int(__pyx_v_s_i); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 889, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);

    /* "serializor/binary/ser.py":890
 *             b"\x03",  # 3-dimension
 *             gen_encoded_int(s_i),  # type: ignore
 *             gen_encoded_int(s_j),  # type: ignore             # <<<<<<<<<<<<<<
 *             gen_encoded_int(s_k),  # type: ignore
 *         ]
 */
    __pyx_t_3 = __pyx_f_10serializor_6binary_3ser_gen_encoded_int(__pyx_v_s_j); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 890, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);

    /* "serializor/binary/ser.py":891
 *             gen_encoded_int(s_i),  # type: ignore
 *             gen_encoded_int(s_j),  # type: ignore
 *             gen_encoded_int(s_k),  # type: ignore             # <<<<<<<<<<<<<<
 *         ]
 *         # . serialize values
 */
    __pyx_t_4 = __pyx_f_10serializor_6binary_3ser_gen_encoded_int(__pyx_v_s_k); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 891, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);

    /* "serializor/binary/ser.py":885
 *         s_j: cython.Py_ssize_t = shape[1]
 *         s_k: cython.Py_ssize_t = shape[2]
 *         items = [             # <<<<<<<<<<<<<<
 *             arr_dtype,  # identifier & dtype
 *             pack_uint8(npy_dtype),  # type: ignore
 */
    __pyx_t_8 = PyList_New(6); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 885, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_INCREF(__pyx_v_arr_dtype);
    __Pyx_GIVEREF(__pyx_v_arr_dtype);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_8, 0, __pyx_v_arr_dtype)) __PYX_ERR(0, 885, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_2);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_8, 1, __pyx_t_2)) __PYX_ERR(0, 885, __pyx_L1_error);
    __Pyx_INCREF(__pyx_kp_b__8);
    __Pyx_GIVEREF(__pyx_kp_b__8);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_8, 2, __pyx_kp_b__8)) __PYX_ERR(0, 885, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_7);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_8, 3, __pyx_t_7)) __PYX_ERR(0, 885, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_3);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_8, 4, __pyx_t_3)) __PYX_ERR(0, 885, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_4);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_8, 5, __pyx_t_4)) __PYX_ERR(0, 885, __pyx_L1_error);
    __pyx_t_2 = 0;
    __pyx_t_7 = 0;
    __pyx_t_3 = 0;
    __pyx_t_4 = 0;
    __pyx_v_items = ((PyObject*)__pyx_t_8);
    __pyx_t_8 = 0;

    /* "serializor/binary/ser.py":894
 *         ]
 *         # . serialize values
 *         if s_k != 0:             # <<<<<<<<<<<<<<
 *             values_json = _orjson_dumps_numpy(utils.arr_flatten(arr))
 *             items.append(gen_encoded_int(bytes_len(values_json)))  # type: ignore
 */
    __pyx_t_1 = (__pyx_v_s_k != 0);
    if (__pyx_t_1) {

      /* "serializor/binary/ser.py":895
 *         # . serialize values
 *         if s_k != 0:
 *             values_json = _orjson_dumps_numpy(utils.arr_flatten(arr))             # <<<<<<<<<<<<<<
 *             items.append(gen_encoded_int(bytes_len(values_json)))  # type: ignore
 *             items.append(values_json)
 */
      __pyx_t_8 = ((PyObject *)__pyx_f_10serializor_5utils_arr_flatten(__pyx_v_arr)); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 895, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_4 = __pyx_f_10serializor_6binary_3ser__orjson_dumps_numpy(__pyx_t_8); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 895, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_v_values_json = ((PyObject*)__pyx_t_4);
      __pyx_t_4 = 0;

      /* "serializor/binary/ser.py":896
 *         if s_k != 0:
 *             values_json = _orjson_dumps_numpy(utils.arr_flatten(arr))
 *             items.append(gen_encoded_int(bytes_len(values_json)))  # type: ignore             # <<<<<<<<<<<<<<
 *             items.append(values_json)
 *         return b"".join(items)
 */
      __pyx_t_5 = PyBytes_Size(__pyx_v_values_json); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1L))) __PYX_ERR(0, 896, __pyx_L1_error)
      __pyx_t_4 = __pyx_f_10serializor_6binary_3ser_gen_encoded_int(__pyx_t_5); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 896, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_6 = __Pyx_PyList_Append(__pyx_v_items, __pyx_t_4); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(0, 896, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

      /* "serializor/binary/ser.py":897
 *             values_json = _orjson_dumps_numpy(utils.arr_flatten(arr))
 *             items.append(gen_encoded_int(bytes_len(values_json)))  # type: ignore
 *             items.append(values_json)             # <<<<<<<<<<<<<<
 *         return b"".join(items)
 * 
 */
      __pyx_t_6 = __Pyx_PyList_Append(__pyx_v_items, __pyx_v_values_json); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(0, 897, __pyx_L1_error)

      /* "serializor/binary/ser.py":894
 *         ]
 *         # . serialize values
 *         if s_k != 0:             # <<<<<<<<<<<<<<
 *             values_json = _orjson_dumps_numpy(utils.arr_flatten(arr))
 *             items.append(gen_encoded_int(bytes_len(values_json)))  # type: ignore
 */
    }

    /* "serializor/binary/ser.py":898
 *             items.append(gen_encoded_int(bytes_len(values_json)))  # type: ignore
 *             items.append(values_json)
 *         return b"".join(items)             # <<<<<<<<<<<<<<
 * 
 *     # 4-dimension
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyBytes_Join(__pyx_kp_b__4, __pyx_v_items); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 898, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    if (!(likely(PyBytes_CheckExact(__pyx_t_4))||((__pyx_t_4) == Py_None) || __Pyx_RaiseUnexpectedTypeError("bytes", __pyx_t_4))) __PYX_ERR(0, 898, __pyx_L1_error)
    __pyx_r = ((PyObject*)__pyx_t_4);
    __pyx_t_4 = 0;
    goto __pyx_L0;

    /* "serializor/binary/ser.py":881
 * 
 *     # 3-dimension
 *     if ndim == 3:             # <<<<<<<<<<<<<<
 *         s_i: cython.Py_ssize_t = shape[0]
 *         s_j: cython.Py_ssize_t = shape[1]
 */
  }

  /* "serializor/binary/ser.py":901
 * 
 *     # 4-dimension
 *     if ndim == 4:             # <<<<<<<<<<<<<<
 *         s_i: cython.Py_ssize_t = shape[0]
 *         s_j: cython.Py_ssize_t = shape[1]
 */
  __pyx_t_1 = (__pyx_v_ndim == 4);
  if (__pyx_t_1) {

    /* "serializor/binary/ser.py":902
 *     # 4-dimension
 *     if ndim == 4:
 *         s_i: cython.Py_ssize_t = shape[0]             # <<<<<<<<<<<<<<
 *         s_j: cython.Py_ssize_t = shape[1]
 *         s_k: cython.Py_ssize_t = shape[2]
 */
    __pyx_v_s_i = (__pyx_v_shape[0]);

    /* "serializor/binary/ser.py":903
 *     if ndim == 4:
 *         s_i: cython.Py_ssize_t = shape[0]
 *         s_j: cython.Py_ssize_t = shape[1]             # <<<<<<<<<<<<<<
 *         s_k: cython.Py_ssize_t = shape[2]
 *         s_l: cython.Py_ssize_t = shape[3]
 */
    __pyx_v_s_j = (__pyx_v_shape[1]);

    /* "serializor/binary/ser.py":904
 *         s_i: cython.Py_ssize_t = shape[0]
 *         s_j: cython.Py_ssize_t = shape[1]
 *         s_k: cython.Py_ssize_t = shape[2]             # <<<<<<<<<<<<<<
 *         s_l: cython.Py_ssize_t = shape[3]
 *         items = [
 */
    __pyx_v_s_k = (__pyx_v_shape[2]);

    /* "serializor/binary/ser.py":905
 *         s_j: cython.Py_ssize_t = shape[1]
 *         s_k: cython.Py_ssize_t = shape[2]
 *         s_l: cython.Py_ssize_t = shape[3]             # <<<<<<<<<<<<<<
 *         items = [
 *             arr_dtype,  # identifier & dtype
 */
    __pyx_v_s_l = (__pyx_v_shape[3]);

    /* "serializor/binary/ser.py":908
 *         items = [
 *             arr_dtype,  # identifier & dtype
 *             pack_uint8(npy_dtype),  # type: ignore             # <<<<<<<<<<<<<<
 *             b"\x04",  # 4-dimension
 *             gen_encoded_int(s_i),  # type: ignore
 */
    __pyx_t_4 = __pyx_f_10serializor_6binary_3ser_pack_uint8(__pyx_v_npy_dtype); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 908, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);

    /* "serializor/binary/ser.py":910
 *             pack_uint8(npy_dtype),  # type: ignore
 *             b"\x04",  # 4-dimension
 *             gen_encoded_int(s_i),  # type: ignore             # <<<<<<<<<<<<<<
 *             gen_encoded_int(s_j),  # type: ignore
 *             gen_encoded_int(s_k),  # type: ignore
 */
    __pyx_t_8 = __pyx_f_10serializor_6binary_3ser_gen_encoded_int(__pyx_v_s_i); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 910, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);

    /* "serializor/binary/ser.py":911
 *             b"\x04",  # 4-dimension
 *             gen_encoded_int(s_i),  # type: ignore
 *             gen_encoded_int(s_j),  # type: ignore             # <<<<<<<<<<<<<<
 *             gen_encoded_int(s_k),  # type: ignore
 *             gen_encoded_int(s_l),  # type: ignore
 */
    __pyx_t_3 = __pyx_f_10serializor_6binary_3ser_gen_encoded_int(__pyx_v_s_j); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 911, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);

    /* "serializor/binary/ser.py":912
 *             gen_encoded_int(s_i),  # type: ignore
 *             gen_encoded_int(s_j),  # type: ignore
 *             gen_encoded_int(s_k),  # type: ignore             # <<<<<<<<<<<<<<
 *             gen_encoded_int(s_l),  # type: ignore
 *         ]
 */
    __pyx_t_7 = __pyx_f_10serializor_6binary_3ser_gen_encoded_int(__pyx_v_s_k); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 912, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);

    /* "serializor/binary/ser.py":913
 *             gen_encoded_int(s_j),  # type: ignore
 *             gen_encoded_int(s_k),  # type: ignore
 *             gen_encoded_int(s_l),  # type: ignore             # <<<<<<<<<<<<<<
 *         ]
 *         # . serialize values
 */
    __pyx_t_2 = __pyx_f_10serializor_6binary_3ser_gen_encoded_int(__pyx_v_s_l); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 913, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);

    /* "serializor/binary/ser.py":906
 *         s_k: cython.Py_ssize_t = shape[2]
 *         s_l: cython.Py_ssize_t = shape[3]
 *         items = [             # <<<<<<<<<<<<<<
 *             arr_dtype,  # identifier & dtype
 *             pack_uint8(npy_dtype),  # type: ignore
 */
    __pyx_t_9 = PyList_New(7); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 906, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_INCREF(__pyx_v_arr_dtype);
    __Pyx_GIVEREF(__pyx_v_arr_dtype);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_9, 0, __pyx_v_arr_dtype)) __PYX_ERR(0, 906, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_4);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_9, 1, __pyx_t_4)) __PYX_ERR(0, 906, __pyx_L1_error);
    __Pyx_INCREF(__pyx_kp_b__9);
    __Pyx_GIVEREF(__pyx_kp_b__9);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_9, 2, __pyx_kp_b__9)) __PYX_ERR(0, 906, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_8);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_9, 3, __pyx_t_8)) __PYX_ERR(0, 906, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_3);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_9, 4, __pyx_t_3)) __PYX_ERR(0, 906, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_7);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_9, 5, __pyx_t_7)) __PYX_ERR(0, 906, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_2);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_9, 6, __pyx_t_2)) __PYX_ERR(0, 906, __pyx_L1_error);
    __pyx_t_4 = 0;
    __pyx_t_8 = 0;
    __pyx_t_3 = 0;
    __pyx_t_7 = 0;
    __pyx_t_2 = 0;
    __pyx_v_items = ((PyObject*)__pyx_t_9);
    __pyx_t_9 = 0;

    /* "serializor/binary/ser.py":916
 *         ]
 *         # . serialize values
 *         if s_l != 0:             # <<<<<<<<<<<<<<
 *             values_json = _orjson_dumps_numpy(utils.arr_flatten(arr))
 *             items.append(gen_encoded_int(bytes_len(values_json)))  # type: ignore
 */
    __pyx_t_1 = (__pyx_v_s_l != 0);
    if (__pyx_t_1) {

      /* "serializor/binary/ser.py":917
 *         # . serialize values
 *         if s_l != 0:
 *             values_json = _orjson_dumps_numpy(utils.arr_flatten(arr))             # <<<<<<<<<<<<<<
 *             items.append(gen_encoded_int(bytes_len(values_json)))  # type: ignore
 *             items.append(values_json)
 */
      __pyx_t_9 = ((PyObject *)__pyx_f_10serializor_5utils_arr_flatten(__pyx_v_arr)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 917, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_2 = __pyx_f_10serializor_6binary_3ser__orjson_dumps_numpy(__pyx_t_9); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 917, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __pyx_v_values_json = ((PyObject*)__pyx_t_2);
      __pyx_t_2 = 0;

      /* "serializor/binary/ser.py":918
 *         if s_l != 0:
 *             values_json = _orjson_dumps_numpy(utils.arr_flatten(arr))
 *             items.append(gen_encoded_int(bytes_len(values_json)))  # type: ignore             # <<<<<<<<<<<<<<
 *             items.append(values_json)
 *         return b"".join(items)
 */
      __pyx_t_5 = PyBytes_Size(__pyx_v_values_json); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1L))) __PYX_ERR(0, 918, __pyx_L1_error)
      __pyx_t_2 = __pyx_f_10serializor_6binary_3ser_gen_encoded_int(__pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 918, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_6 = __Pyx_PyList_Append(__pyx_v_items, __pyx_t_2); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(0, 918, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "serializor/binary/ser.py":919
 *             values_json = _orjson_dumps_numpy(utils.arr_flatten(arr))
 *             items.append(gen_encoded_int(bytes_len(values_json)))  # type: ignore
 *             items.append(values_json)             # <<<<<<<<<<<<<<
 *         return b"".join(items)
 * 
 */
      __pyx_t_6 = __Pyx_PyList_Append(__pyx_v_items, __pyx_v_values_json); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(0, 919, __pyx_L1_error)

      /* "serializor/binary/ser.py":916
 *         ]
 *         # . serialize values
 *         if s_l != 0:             # <<<<<<<<<<<<<<
 *             values_json = _orjson_dumps_numpy(utils.arr_flatten(arr))
 *             items.append(gen_encoded_int(bytes_len(values_json)))  # type: ignore
 */
    }

    /* "serializor/binary/ser.py":920
 *             items.append(gen_encoded_int(bytes_len(values_json)))  # type: ignore
 *             items.append(values_json)
 *         return b"".join(items)             # <<<<<<<<<<<<<<
 * 
 *     # invalid
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = __Pyx_PyBytes_Join(__pyx_kp_b__4, __pyx_v_items); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 920, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (!(likely(PyBytes_CheckExact(__pyx_t_2))||((__pyx_t_2) == Py_None) || __Pyx_RaiseUnexpectedTypeError("bytes", __pyx_t_2))) __PYX_ERR(0, 920, __pyx_L1_error)
    __pyx_r = ((PyObject*)__pyx_t_2);
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "serializor/binary/ser.py":901
 * 
 *     # 4-dimension
 *     if ndim == 4:             # <<<<<<<<<<<<<<
 *         s_i: cython.Py_ssize_t = shape[0]
 *         s_j: cython.Py_ssize_t = shape[1]
 */
  }

  /* "serializor/binary/ser.py":923
 * 
 *     # invalid
 *     raise ValueError("unsupported <'numpy.ndarray'> dimension [%d]." % ndim)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_v_ndim); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 923, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_9 = PyUnicode_Format(__pyx_kp_u_unsupported_numpy_ndarray_dimens, __pyx_t_2); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 923, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_9); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 923, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __Pyx_Raise(__pyx_t_2, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __PYX_ERR(0, 923, __pyx_L1_error)

  /* "serializor/binary/ser.py":819
 * 
 * 
 * @cython.cfunc             # <<<<<<<<<<<<<<
 * @cython.inline(True)
 * def _ser_ndarray_numeric(
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("serializor.binary.ser._ser_ndarray_numeric", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_items);
  __Pyx_XDECREF(__pyx_v_values_json);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "serializor/binary/ser.py":926
 * 
 * 
 * @cython.cfunc             # <<<<<<<<<<<<<<
 * @cython.inline(True)
 * def _ser_ndarray_nptime(arr: np.ndarray, arr_dtype: bytes) -> bytes:
 */

static PyObject *__pyx_f_10serializor_6binary_3ser__ser_ndarray_nptime(PyArrayObject *__pyx_v_arr, PyObject *__pyx_v_arr_dtype) {
  int __pyx_v_npy_dtype;
  int __pyx_v_ndim;
  npy_intp *__pyx_v_shape;
  Py_ssize_t __pyx_v_s_i;
  PyObject *__pyx_v_items = NULL;
  PyObject *__pyx_v_values_json = NULL;
  Py_ssize_t __pyx_v_s_j;
  Py_ssize_t __pyx_v_s_k;
  Py_ssize_t __pyx_v_s_l;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  Py_ssize_t __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_ser_ndarray_nptime", 0);
  __Pyx_INCREF((PyObject *)__pyx_v_arr);

  /* "serializor/binary/ser.py":946
 *     >>> b'...[SHAPE][npyTimeUnit(1B)][values_json_size(ENC)][values_json]'
 *     """
 *     npy_dtype: cython.int = arr.descr.type_num             # <<<<<<<<<<<<<<
 *     ndim: cython.int = arr.ndim
 *     shape = arr.shape
 */
  __pyx_t_1 = ((PyObject *)__pyx_f_5numpy_7ndarray_5descr_descr(__pyx_v_arr)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 946, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = ((PyArray_Descr *)__pyx_t_1)->type_num;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_npy_dtype = __pyx_t_2;

  /* "serializor/binary/ser.py":947
 *     """
 *     npy_dtype: cython.int = arr.descr.type_num
 *     ndim: cython.int = arr.ndim             # <<<<<<<<<<<<<<
 *     shape = arr.shape
 *     # 1-dimension
 */
  __pyx_v_ndim = __pyx_f_5numpy_7ndarray_4ndim_ndim(__pyx_v_arr);

  /* "serializor/binary/ser.py":948
 *     npy_dtype: cython.int = arr.descr.type_num
 *     ndim: cython.int = arr.ndim
 *     shape = arr.shape             # <<<<<<<<<<<<<<
 *     # 1-dimension
 *     if ndim == 1:
 */
  __pyx_v_shape = __pyx_f_5numpy_7ndarray_5shape_shape(__pyx_v_arr);

  /* "serializor/binary/ser.py":950
 *     shape = arr.shape
 *     # 1-dimension
 *     if ndim == 1:             # <<<<<<<<<<<<<<
 *         # . array shape
 *         s_i: cython.Py_ssize_t = shape[0]
 */
  __pyx_t_3 = (__pyx_v_ndim == 1);
  if (__pyx_t_3) {

    /* "serializor/binary/ser.py":952
 *     if ndim == 1:
 *         # . array shape
 *         s_i: cython.Py_ssize_t = shape[0]             # <<<<<<<<<<<<<<
 *         items = [
 *             arr_dtype,  # identifier & dtype
 */
    __pyx_v_s_i = (__pyx_v_shape[0]);

    /* "serializor/binary/ser.py":955
 *         items = [
 *             arr_dtype,  # identifier & dtype
 *             pack_uint8(npy_dtype),  # type: ignore             # <<<<<<<<<<<<<<
 *             b"\x01",  # 1-dimension
 *             gen_encoded_int(s_i),  # type: ignore
 */
    __pyx_t_1 = __pyx_f_10serializor_6binary_3ser_pack_uint8(__pyx_v_npy_dtype); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 955, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);

    /* "serializor/binary/ser.py":957
 *             pack_uint8(npy_dtype),  # type: ignore
 *             b"\x01",  # 1-dimension
 *             gen_encoded_int(s_i),  # type: ignore             # <<<<<<<<<<<<<<
 *         ]
 *         # . serialize values
 */
    __pyx_t_4 = __pyx_f_10serializor_6binary_3ser_gen_encoded_int(__pyx_v_s_i); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 957, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);

    /* "serializor/binary/ser.py":953
 *         # . array shape
 *         s_i: cython.Py_ssize_t = shape[0]
 *         items = [             # <<<<<<<<<<<<<<
 *             arr_dtype,  # identifier & dtype
 *             pack_uint8(npy_dtype),  # type: ignore
 */
    __pyx_t_5 = PyList_New(4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 953, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_INCREF(__pyx_v_arr_dtype);
    __Pyx_GIVEREF(__pyx_v_arr_dtype);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 0, __pyx_v_arr_dtype)) __PYX_ERR(0, 953, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 1, __pyx_t_1)) __PYX_ERR(0, 953, __pyx_L1_error);
    __Pyx_INCREF(__pyx_kp_b__5);
    __Pyx_GIVEREF(__pyx_kp_b__5);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 2, __pyx_kp_b__5)) __PYX_ERR(0, 953, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_4);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 3, __pyx_t_4)) __PYX_ERR(0, 953, __pyx_L1_error);
    __pyx_t_1 = 0;
    __pyx_t_4 = 0;
    __pyx_v_items = ((PyObject*)__pyx_t_5);
    __pyx_t_5 = 0;

    /* "serializor/binary/ser.py":960
 *         ]
 *         # . serialize values
 *         if s_i == 0:  # empty             # <<<<<<<<<<<<<<
 *             items.append(pack_uint8(utils.parse_arr_nptime_unit(arr)))  # type: ignore
 *         else:
 */
    __pyx_t_3 = (__pyx_v_s_i == 0);
    if (__pyx_t_3) {

      /* "serializor/binary/ser.py":961
 *         # . serialize values
 *         if s_i == 0:  # empty
 *             items.append(pack_uint8(utils.parse_arr_nptime_unit(arr)))  # type: ignore             # <<<<<<<<<<<<<<
 *         else:
 *             items.append(pack_uint8(np.get_datetime64_unit(arr[0])))  # type: ignore
 */
      __pyx_t_2 = __pyx_f_10serializor_5utils_parse_arr_nptime_unit(__pyx_v_arr); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(0, 961, __pyx_L1_error)
      __pyx_t_5 = __pyx_f_10serializor_6binary_3ser_pack_uint8(__pyx_t_2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 961, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_6 = __Pyx_PyList_Append(__pyx_v_items, __pyx_t_5); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(0, 961, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

      /* "serializor/binary/ser.py":960
 *         ]
 *         # . serialize values
 *         if s_i == 0:  # empty             # <<<<<<<<<<<<<<
 *             items.append(pack_uint8(utils.parse_arr_nptime_unit(arr)))  # type: ignore
 *         else:
 */
      goto __pyx_L4;
    }

    /* "serializor/binary/ser.py":963
 *             items.append(pack_uint8(utils.parse_arr_nptime_unit(arr)))  # type: ignore
 *         else:
 *             items.append(pack_uint8(np.get_datetime64_unit(arr[0])))  # type: ignore             # <<<<<<<<<<<<<<
 *             arr = np.PyArray_Cast(arr, np.NPY_TYPES.NPY_INT64)  # cast to int64
 *             values_json = _orjson_dumps_numpy(arr)
 */
    /*else*/ {
      __pyx_t_5 = __Pyx_GetItemInt(((PyObject *)__pyx_v_arr), 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 963, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_4 = __pyx_f_10serializor_6binary_3ser_pack_uint8(__pyx_f_5numpy_get_datetime64_unit(__pyx_t_5)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 963, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_6 = __Pyx_PyList_Append(__pyx_v_items, __pyx_t_4); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(0, 963, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

      /* "serializor/binary/ser.py":964
 *         else:
 *             items.append(pack_uint8(np.get_datetime64_unit(arr[0])))  # type: ignore
 *             arr = np.PyArray_Cast(arr, np.NPY_TYPES.NPY_INT64)  # cast to int64             # <<<<<<<<<<<<<<
 *             values_json = _orjson_dumps_numpy(arr)
 *             items.append(gen_encoded_int(bytes_len(values_json)))  # type: ignore
 */
      __pyx_t_4 = PyArray_Cast(__pyx_v_arr, NPY_INT64); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 964, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      if (!(likely(((__pyx_t_4) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_4, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 964, __pyx_L1_error)
      __Pyx_DECREF_SET(__pyx_v_arr, ((PyArrayObject *)__pyx_t_4));
      __pyx_t_4 = 0;

      /* "serializor/binary/ser.py":965
 *             items.append(pack_uint8(np.get_datetime64_unit(arr[0])))  # type: ignore
 *             arr = np.PyArray_Cast(arr, np.NPY_TYPES.NPY_INT64)  # cast to int64
 *             values_json = _orjson_dumps_numpy(arr)             # <<<<<<<<<<<<<<
 *             items.append(gen_encoded_int(bytes_len(values_json)))  # type: ignore
 *             items.append(values_json)
 */
      __pyx_t_4 = __pyx_f_10serializor_6binary_3ser__orjson_dumps_numpy(((PyObject *)__pyx_v_arr)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 965, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_v_values_json = ((PyObject*)__pyx_t_4);
      __pyx_t_4 = 0;

      /* "serializor/binary/ser.py":966
 *             arr = np.PyArray_Cast(arr, np.NPY_TYPES.NPY_INT64)  # cast to int64
 *             values_json = _orjson_dumps_numpy(arr)
 *             items.append(gen_encoded_int(bytes_len(values_json)))  # type: ignore             # <<<<<<<<<<<<<<
 *             items.append(values_json)
 *         return b"".join(items)
 */
      __pyx_t_7 = PyBytes_Size(__pyx_v_values_json); if (unlikely(__pyx_t_7 == ((Py_ssize_t)-1L))) __PYX_ERR(0, 966, __pyx_L1_error)
      __pyx_t_4 = __pyx_f_10serializor_6binary_3ser_gen_encoded_int(__pyx_t_7); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 966, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_6 = __Pyx_PyList_Append(__pyx_v_items, __pyx_t_4); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(0, 966, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

      /* "serializor/binary/ser.py":967
 *             values_json = _orjson_dumps_numpy(arr)
 *             items.append(gen_encoded_int(bytes_len(values_json)))  # type: ignore
 *             items.append(values_json)             # <<<<<<<<<<<<<<
 *         return b"".join(items)
 * 
 */
      __pyx_t_6 = __Pyx_PyList_Append(__pyx_v_items, __pyx_v_values_json); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(0, 967, __pyx_L1_error)
    }
    __pyx_L4:;

    /* "serializor/binary/ser.py":968
 *             items.append(gen_encoded_int(bytes_len(values_json)))  # type: ignore
 *             items.append(values_json)
 *         return b"".join(items)             # <<<<<<<<<<<<<<
 * 
 *     # 2-dimension
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyBytes_Join(__pyx_kp_b__4, __pyx_v_items); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 968, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    if (!(likely(PyBytes_CheckExact(__pyx_t_4))||((__pyx_t_4) == Py_None) || __Pyx_RaiseUnexpectedTypeError("bytes", __pyx_t_4))) __PYX_ERR(0, 968, __pyx_L1_error)
    __pyx_r = ((PyObject*)__pyx_t_4);
    __pyx_t_4 = 0;
    goto __pyx_L0;

    /* "serializor/binary/ser.py":950
 *     shape = arr.shape
 *     # 1-dimension
 *     if ndim == 1:             # <<<<<<<<<<<<<<
 *         # . array shape
 *         s_i: cython.Py_ssize_t = shape[0]
 */
  }

  /* "serializor/binary/ser.py":971
 * 
 *     # 2-dimension
 *     if ndim == 2:             # <<<<<<<<<<<<<<
 *         s_i: cython.Py_ssize_t = shape[0]
 *         s_j: cython.Py_ssize_t = shape[1]
 */
  __pyx_t_3 = (__pyx_v_ndim == 2);
  if (__pyx_t_3) {

    /* "serializor/binary/ser.py":972
 *     # 2-dimension
 *     if ndim == 2:
 *         s_i: cython.Py_ssize_t = shape[0]             # <<<<<<<<<<<<<<
 *         s_j: cython.Py_ssize_t = shape[1]
 *         items = [
 */
    __pyx_v_s_i = (__pyx_v_shape[0]);

    /* "serializor/binary/ser.py":973
 *     if ndim == 2:
 *         s_i: cython.Py_ssize_t = shape[0]
 *         s_j: cython.Py_ssize_t = shape[1]             # <<<<<<<<<<<<<<
 *         items = [
 *             arr_dtype,  # identifier & dtype
 */
    __pyx_v_s_j = (__pyx_v_shape[1]);

    /* "serializor/binary/ser.py":976
 *         items = [
 *             arr_dtype,  # identifier & dtype
 *             pack_uint8(npy_dtype),  # type: ignore             # <<<<<<<<<<<<<<
 *             b"\x02",  # 2-dimension
 *             gen_encoded_int(s_i),  # type: ignore
 */
    __pyx_t_4 = __pyx_f_10serializor_6binary_3ser_pack_uint8(__pyx_v_npy_dtype); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 976, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);

    /* "serializor/binary/ser.py":978
 *             pack_uint8(npy_dtype),  # type: ignore
 *             b"\x02",  # 2-dimension
 *             gen_encoded_int(s_i),  # type: ignore             # <<<<<<<<<<<<<<
 *             gen_encoded_int(s_j),  # type: ignore
 *         ]
 */
    __pyx_t_5 = __pyx_f_10serializor_6binary_3ser_gen_encoded_int(__pyx_v_s_i); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 978, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);

    /* "serializor/binary/ser.py":979
 *             b"\x02",  # 2-dimension
 *             gen_encoded_int(s_i),  # type: ignore
 *             gen_encoded_int(s_j),  # type: ignore             # <<<<<<<<<<<<<<
 *         ]
 *         # . serialize values
 */
    __pyx_t_1 = __pyx_f_10serializor_6binary_3ser_gen_encoded_int(__pyx_v_s_j); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 979, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);

    /* "serializor/binary/ser.py":974
 *         s_i: cython.Py_ssize_t = shape[0]
 *         s_j: cython.Py_ssize_t = shape[1]
 *         items = [             # <<<<<<<<<<<<<<
 *             arr_dtype,  # identifier & dtype
 *             pack_uint8(npy_dtype),  # type: ignore
 */
    __pyx_t_8 = PyList_New(5); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 974, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_INCREF(__pyx_v_arr_dtype);
    __Pyx_GIVEREF(__pyx_v_arr_dtype);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_8, 0, __pyx_v_arr_dtype)) __PYX_ERR(0, 974, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_4);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_8, 1, __pyx_t_4)) __PYX_ERR(0, 974, __pyx_L1_error);
    __Pyx_INCREF(__pyx_kp_b__6);
    __Pyx_GIVEREF(__pyx_kp_b__6);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_8, 2, __pyx_kp_b__6)) __PYX_ERR(0, 974, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_5);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_8, 3, __pyx_t_5)) __PYX_ERR(0, 974, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_8, 4, __pyx_t_1)) __PYX_ERR(0, 974, __pyx_L1_error);
    __pyx_t_4 = 0;
    __pyx_t_5 = 0;
    __pyx_t_1 = 0;
    __pyx_v_items = ((PyObject*)__pyx_t_8);
    __pyx_t_8 = 0;

    /* "serializor/binary/ser.py":982
 *         ]
 *         # . serialize values
 *         if s_j == 0:  # empty             # <<<<<<<<<<<<<<
 *             items.append(pack_uint8(utils.parse_arr_nptime_unit(arr)))  # type: ignore
 *         else:
 */
    __pyx_t_3 = (__pyx_v_s_j == 0);
    if (__pyx_t_3) {

      /* "serializor/binary/ser.py":983
 *         # . serialize values
 *         if s_j == 0:  # empty
 *             items.append(pack_uint8(utils.parse_arr_nptime_unit(arr)))  # type: ignore             # <<<<<<<<<<<<<<
 *         else:
 *             items.append(pack_uint8(np.get_datetime64_unit(arr[0, 0])))  # type: ignore
 */
      __pyx_t_2 = __pyx_f_10serializor_5utils_parse_arr_nptime_unit(__pyx_v_arr); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(0, 983, __pyx_L1_error)
      __pyx_t_8 = __pyx_f_10serializor_6binary_3ser_pack_uint8(__pyx_t_2); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 983, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_6 = __Pyx_PyList_Append(__pyx_v_items, __pyx_t_8); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(0, 983, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

      /* "serializor/binary/ser.py":982
 *         ]
 *         # . serialize values
 *         if s_j == 0:  # empty             # <<<<<<<<<<<<<<
 *             items.append(pack_uint8(utils.parse_arr_nptime_unit(arr)))  # type: ignore
 *         else:
 */
      goto __pyx_L6;
    }

    /* "serializor/binary/ser.py":985
 *             items.append(pack_uint8(utils.parse_arr_nptime_unit(arr)))  # type: ignore
 *         else:
 *             items.append(pack_uint8(np.get_datetime64_unit(arr[0, 0])))  # type: ignore             # <<<<<<<<<<<<<<
 *             arr = np.PyArray_Cast(arr, np.NPY_TYPES.NPY_INT64)  # cast to int64
 *             values_json = _orjson_dumps_numpy(utils.arr_flatten(arr))
 */
    /*else*/ {
      __pyx_t_8 = __Pyx_PyObject_GetItem(((PyObject *)__pyx_v_arr), __pyx_tuple__10); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 985, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_1 = __pyx_f_10serializor_6binary_3ser_pack_uint8(__pyx_f_5numpy_get_datetime64_unit(__pyx_t_8)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 985, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_t_6 = __Pyx_PyList_Append(__pyx_v_items, __pyx_t_1); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(0, 985, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "serializor/binary/ser.py":986
 *         else:
 *             items.append(pack_uint8(np.get_datetime64_unit(arr[0, 0])))  # type: ignore
 *             arr = np.PyArray_Cast(arr, np.NPY_TYPES.NPY_INT64)  # cast to int64             # <<<<<<<<<<<<<<
 *             values_json = _orjson_dumps_numpy(utils.arr_flatten(arr))
 *             items.append(gen_encoded_int(bytes_len(values_json)))  # type: ignore
 */
      __pyx_t_1 = PyArray_Cast(__pyx_v_arr, NPY_INT64); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 986, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 986, __pyx_L1_error)
      __Pyx_DECREF_SET(__pyx_v_arr, ((PyArrayObject *)__pyx_t_1));
      __pyx_t_1 = 0;

      /* "serializor/binary/ser.py":987
 *             items.append(pack_uint8(np.get_datetime64_unit(arr[0, 0])))  # type: ignore
 *             arr = np.PyArray_Cast(arr, np.NPY_TYPES.NPY_INT64)  # cast to int64
 *             values_json = _orjson_dumps_numpy(utils.arr_flatten(arr))             # <<<<<<<<<<<<<<
 *             items.append(gen_encoded_int(bytes_len(values_json)))  # type: ignore
 *             items.append(values_json)
 */
      __pyx_t_1 = ((PyObject *)__pyx_f_10serializor_5utils_arr_flatten(__pyx_v_arr)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 987, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_8 = __pyx_f_10serializor_6binary_3ser__orjson_dumps_numpy(__pyx_t_1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 987, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_v_values_json = ((PyObject*)__pyx_t_8);
      __pyx_t_8 = 0;

      /* "serializor/binary/ser.py":988
 *             arr = np.PyArray_Cast(arr, np.NPY_TYPES.NPY_INT64)  # cast to int64
 *             values_json = _orjson_dumps_numpy(utils.arr_flatten(arr))
 *             items.append(gen_encoded_int(bytes_len(values_json)))  # type: ignore             # <<<<<<<<<<<<<<
 *             items.append(values_json)
 *         return b"".join(items)
 */
      __pyx_t_7 = PyBytes_Size(__pyx_v_values_json); if (unlikely(__pyx_t_7 == ((Py_ssize_t)-1L))) __PYX_ERR(0, 988, __pyx_L1_error)
      __pyx_t_8 = __pyx_f_10serializor_6binary_3ser_gen_encoded_int(__pyx_t_7); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 988, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_6 = __Pyx_PyList_Append(__pyx_v_items, __pyx_t_8); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(0, 988, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

      /* "serializor/binary/ser.py":989
 *             values_json = _orjson_dumps_numpy(utils.arr_flatten(arr))
 *             items.append(gen_encoded_int(bytes_len(values_json)))  # type: ignore
 *             items.append(values_json)             # <<<<<<<<<<<<<<
 *         return b"".join(items)
 * 
 */
      __pyx_t_6 = __Pyx_PyList_Append(__pyx_v_items, __pyx_v_values_json); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(0, 989, __pyx_L1_error)
    }
    __pyx_L6:;

    /* "serializor/binary/ser.py":990
 *             items.append(gen_encoded_int(bytes_len(values_json)))  # type: ignore
 *             items.append(values_json)
 *         return b"".join(items)             # <<<<<<<<<<<<<<
 * 
 *     # 3-dimension
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_8 = __Pyx_PyBytes_Join(__pyx_kp_b__4, __pyx_v_items); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 990, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    if (!(likely(PyBytes_CheckExact(__pyx_t_8))||((__pyx_t_8) == Py_None) || __Pyx_RaiseUnexpectedTypeError("bytes", __pyx_t_8))) __PYX_ERR(0, 990, __pyx_L1_error)
    __pyx_r = ((PyObject*)__pyx_t_8);
    __pyx_t_8 = 0;
    goto __pyx_L0;

    /* "serializor/binary/ser.py":971
 * 
 *     # 2-dimension
 *     if ndim == 2:             # <<<<<<<<<<<<<<
 *         s_i: cython.Py_ssize_t = shape[0]
 *         s_j: cython.Py_ssize_t = shape[1]
 */
  }

  /* "serializor/binary/ser.py":993
 * 
 *     # 3-dimension
 *     if ndim == 3:             # <<<<<<<<<<<<<<
 *         s_i: cython.Py_ssize_t = shape[0]
 *         s_j: cython.Py_ssize_t = shape[1]
 */
  __pyx_t_3 = (__pyx_v_ndim == 3);
  if (__pyx_t_3) {

    /* "serializor/binary/ser.py":994
 *     # 3-dimension
 *     if ndim == 3:
 *         s_i: cython.Py_ssize_t = shape[0]             # <<<<<<<<<<<<<<
 *         s_j: cython.Py_ssize_t = shape[1]
 *         s_k: cython.Py_ssize_t = shape[2]
 */
    __pyx_v_s_i = (__pyx_v_shape[0]);

    /* "serializor/binary/ser.py":995
 *     if ndim == 3:
 *         s_i: cython.Py_ssize_t = shape[0]
 *         s_j: cython.Py_ssize_t = shape[1]             # <<<<<<<<<<<<<<
 *         s_k: cython.Py_ssize_t = shape[2]
 *         items = [
 */
    __pyx_v_s_j = (__pyx_v_shape[1]);

    /* "serializor/binary/ser.py":996
 *         s_i: cython.Py_ssize_t = shape[0]
 *         s_j: cython.Py_ssize_t = shape[1]
 *         s_k: cython.Py_ssize_t = shape[2]             # <<<<<<<<<<<<<<
 *         items = [
 *             arr_dtype,  # identifier & dtype
 */
    __pyx_v_s_k = (__pyx_v_shape[2]);

    /* "serializor/binary/ser.py":999
 *         items = [
 *             arr_dtype,  # identifier & dtype
 *             pack_uint8(npy_dtype),  # type: ignore             # <<<<<<<<<<<<<<
 *             b"\x03",  # 3-dimension
 *             gen_encoded_int(s_i),  # type: ignore
 */
    __pyx_t_8 = __pyx_f_10serializor_6binary_3ser_pack_uint8(__pyx_v_npy_dtype); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 999, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);

    /* "serializor/binary/ser.py":1001
 *             pack_uint8(npy_dtype),  # type: ignore
 *             b"\x03",  # 3-dimension
 *             gen_encoded_int(s_i),  # type: ignore             # <<<<<<<<<<<<<<
 *             gen_encoded_int(s_j),  # type: ignore
 *             gen_encoded_int(s_k),  # type: ignore
 */
    __pyx_t_1 = __pyx_f_10serializor_6binary_3ser_gen_encoded_int(__pyx_v_s_i); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1001, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);

    /* "serializor/binary/ser.py":1002
 *             b"\x03",  # 3-dimension
 *             gen_encoded_int(s_i),  # type: ignore
 *             gen_encoded_int(s_j),  # type: ignore             # <<<<<<<<<<<<<<
 *             gen_encoded_int(s_k),  # type: ignore
 *         ]
 */
    __pyx_t_5 = __pyx_f_10serializor_6binary_3ser_gen_encoded_int(__pyx_v_s_j); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1002, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);

    /* "serializor/binary/ser.py":1003
 *             gen_encoded_int(s_i),  # type: ignore
 *             gen_encoded_int(s_j),  # type: ignore
 *             gen_encoded_int(s_k),  # type: ignore             # <<<<<<<<<<<<<<
 *         ]
 *         # . serialize values
 */
    __pyx_t_4 = __pyx_f_10serializor_6binary_3ser_gen_encoded_int(__pyx_v_s_k); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1003, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);

    /* "serializor/binary/ser.py":997
 *         s_j: cython.Py_ssize_t = shape[1]
 *         s_k: cython.Py_ssize_t = shape[2]
 *         items = [             # <<<<<<<<<<<<<<
 *             arr_dtype,  # identifier & dtype
 *             pack_uint8(npy_dtype),  # type: ignore
 */
    __pyx_t_9 = PyList_New(6); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 997, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_INCREF(__pyx_v_arr_dtype);
    __Pyx_GIVEREF(__pyx_v_arr_dtype);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_9, 0, __pyx_v_arr_dtype)) __PYX_ERR(0, 997, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_8);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_9, 1, __pyx_t_8)) __PYX_ERR(0, 997, __pyx_L1_error);
    __Pyx_INCREF(__pyx_kp_b__8);
    __Pyx_GIVEREF(__pyx_kp_b__8);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_9, 2, __pyx_kp_b__8)) __PYX_ERR(0, 997, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_9, 3, __pyx_t_1)) __PYX_ERR(0, 997, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_5);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_9, 4, __pyx_t_5)) __PYX_ERR(0, 997, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_4);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_9, 5, __pyx_t_4)) __PYX_ERR(0, 997, __pyx_L1_error);
    __pyx_t_8 = 0;
    __pyx_t_1 = 0;
    __pyx_t_5 = 0;
    __pyx_t_4 = 0;
    __pyx_v_items = ((PyObject*)__pyx_t_9);
    __pyx_t_9 = 0;

    /* "serializor/binary/ser.py":1006
 *         ]
 *         # . serialize values
 *         if s_k == 0:  # empty             # <<<<<<<<<<<<<<
 *             items.append(pack_uint8(utils.parse_arr_nptime_unit(arr)))  # type: ignore
 *         else:
 */
    __pyx_t_3 = (__pyx_v_s_k == 0);
    if (__pyx_t_3) {

      /* "serializor/binary/ser.py":1007
 *         # . serialize values
 *         if s_k == 0:  # empty
 *             items.append(pack_uint8(utils.parse_arr_nptime_unit(arr)))  # type: ignore             # <<<<<<<<<<<<<<
 *         else:
 *             items.append(pack_uint8(np.get_datetime64_unit(arr[0, 0, 0])))  # type: ignore
 */
      __pyx_t_2 = __pyx_f_10serializor_5utils_parse_arr_nptime_unit(__pyx_v_arr); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(0, 1007, __pyx_L1_error)
      __pyx_t_9 = __pyx_f_10serializor_6binary_3ser_pack_uint8(__pyx_t_2); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1007, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_6 = __Pyx_PyList_Append(__pyx_v_items, __pyx_t_9); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(0, 1007, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

      /* "serializor/binary/ser.py":1006
 *         ]
 *         # . serialize values
 *         if s_k == 0:  # empty             # <<<<<<<<<<<<<<
 *             items.append(pack_uint8(utils.parse_arr_nptime_unit(arr)))  # type: ignore
 *         else:
 */
      goto __pyx_L8;
    }

    /* "serializor/binary/ser.py":1009
 *             items.append(pack_uint8(utils.parse_arr_nptime_unit(arr)))  # type: ignore
 *         else:
 *             items.append(pack_uint8(np.get_datetime64_unit(arr[0, 0, 0])))  # type: ignore             # <<<<<<<<<<<<<<
 *             arr = np.PyArray_Cast(arr, np.NPY_TYPES.NPY_INT64)  # cast to int64
 *             values_json = _orjson_dumps_numpy(utils.arr_flatten(arr))
 */
    /*else*/ {
      __pyx_t_9 = __Pyx_PyObject_GetItem(((PyObject *)__pyx_v_arr), __pyx_tuple__11); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1009, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_4 = __pyx_f_10serializor_6binary_3ser_pack_uint8(__pyx_f_5numpy_get_datetime64_unit(__pyx_t_9)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1009, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __pyx_t_6 = __Pyx_PyList_Append(__pyx_v_items, __pyx_t_4); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(0, 1009, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

      /* "serializor/binary/ser.py":1010
 *         else:
 *             items.append(pack_uint8(np.get_datetime64_unit(arr[0, 0, 0])))  # type: ignore
 *             arr = np.PyArray_Cast(arr, np.NPY_TYPES.NPY_INT64)  # cast to int64             # <<<<<<<<<<<<<<
 *             values_json = _orjson_dumps_numpy(utils.arr_flatten(arr))
 *             items.append(gen_encoded_int(bytes_len(values_json)))  # type: ignore
 */
      __pyx_t_4 = PyArray_Cast(__pyx_v_arr, NPY_INT64); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1010, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      if (!(likely(((__pyx_t_4) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_4, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 1010, __pyx_L1_error)
      __Pyx_DECREF_SET(__pyx_v_arr, ((PyArrayObject *)__pyx_t_4));
      __pyx_t_4 = 0;

      /* "serializor/binary/ser.py":1011
 *             items.append(pack_uint8(np.get_datetime64_unit(arr[0, 0, 0])))  # type: ignore
 *             arr = np.PyArray_Cast(arr, np.NPY_TYPES.NPY_INT64)  # cast to int64
 *             values_json = _orjson_dumps_numpy(utils.arr_flatten(arr))             # <<<<<<<<<<<<<<
 *             items.append(gen_encoded_int(bytes_len(values_json)))  # type: ignore
 *             items.append(values_json)
 */
      __pyx_t_4 = ((PyObject *)__pyx_f_10serializor_5utils_arr_flatten(__pyx_v_arr)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1011, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_9 = __pyx_f_10serializor_6binary_3ser__orjson_dumps_numpy(__pyx_t_4); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1011, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_v_values_json = ((PyObject*)__pyx_t_9);
      __pyx_t_9 = 0;

      /* "serializor/binary/ser.py":1012
 *             arr = np.PyArray_Cast(arr, np.NPY_TYPES.NPY_INT64)  # cast to int64
 *             values_json = _orjson_dumps_numpy(utils.arr_flatten(arr))
 *             items.append(gen_encoded_int(bytes_len(values_json)))  # type: ignore             # <<<<<<<<<<<<<<
 *             items.append(values_json)
 *         return b"".join(items)
 */
      __pyx_t_7 = PyBytes_Size(__pyx_v_values_json); if (unlikely(__pyx_t_7 == ((Py_ssize_t)-1L))) __PYX_ERR(0, 1012, __pyx_L1_error)
      __pyx_t_9 = __pyx_f_10serializor_6binary_3ser_gen_encoded_int(__pyx_t_7); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1012, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_6 = __Pyx_PyList_Append(__pyx_v_items, __pyx_t_9); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(0, 1012, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

      /* "serializor/binary/ser.py":1013
 *             values_json = _orjson_dumps_numpy(utils.arr_flatten(arr))
 *             items.append(gen_encoded_int(bytes_len(values_json)))  # type: ignore
 *             items.append(values_json)             # <<<<<<<<<<<<<<
 *         return b"".join(items)
 * 
 */
      __pyx_t_6 = __Pyx_PyList_Append(__pyx_v_items, __pyx_v_values_json); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(0, 1013, __pyx_L1_error)
    }
    __pyx_L8:;

    /* "serializor/binary/ser.py":1014
 *             items.append(gen_encoded_int(bytes_len(values_json)))  # type: ignore
 *             items.append(values_json)
 *         return b"".join(items)             # <<<<<<<<<<<<<<
 * 
 *     # 4-dimension
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_9 = __Pyx_PyBytes_Join(__pyx_kp_b__4, __pyx_v_items); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1014, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    if (!(likely(PyBytes_CheckExact(__pyx_t_9))||((__pyx_t_9) == Py_None) || __Pyx_RaiseUnexpectedTypeError("bytes", __pyx_t_9))) __PYX_ERR(0, 1014, __pyx_L1_error)
    __pyx_r = ((PyObject*)__pyx_t_9);
    __pyx_t_9 = 0;
    goto __pyx_L0;

    /* "serializor/binary/ser.py":993
 * 
 *     # 3-dimension
 *     if ndim == 3:             # <<<<<<<<<<<<<<
 *         s_i: cython.Py_ssize_t = shape[0]
 *         s_j: cython.Py_ssize_t = shape[1]
 */
  }

  /* "serializor/binary/ser.py":1017
 * 
 *     # 4-dimension
 *     if ndim == 4:             # <<<<<<<<<<<<<<
 *         s_i: cython.Py_ssize_t = shape[0]
 *         s_j: cython.Py_ssize_t = shape[1]
 */
  __pyx_t_3 = (__pyx_v_ndim == 4);
  if (__pyx_t_3) {

    /* "serializor/binary/ser.py":1018
 *     # 4-dimension
 *     if ndim == 4:
 *         s_i: cython.Py_ssize_t = shape[0]             # <<<<<<<<<<<<<<
 *         s_j: cython.Py_ssize_t = shape[1]
 *         s_k: cython.Py_ssize_t = shape[2]
 */
    __pyx_v_s_i = (__pyx_v_shape[0]);

    /* "serializor/binary/ser.py":1019
 *     if ndim == 4:
 *         s_i: cython.Py_ssize_t = shape[0]
 *         s_j: cython.Py_ssize_t = shape[1]             # <<<<<<<<<<<<<<
 *         s_k: cython.Py_ssize_t = shape[2]
 *         s_l: cython.Py_ssize_t = shape[3]
 */
    __pyx_v_s_j = (__pyx_v_shape[1]);

    /* "serializor/binary/ser.py":1020
 *         s_i: cython.Py_ssize_t = shape[0]
 *         s_j: cython.Py_ssize_t = shape[1]
 *         s_k: cython.Py_ssize_t = shape[2]             # <<<<<<<<<<<<<<
 *         s_l: cython.Py_ssize_t = shape[3]
 *         items = [
 */
    __pyx_v_s_k = (__pyx_v_shape[2]);

    /* "serializor/binary/ser.py":1021
 *         s_j: cython.Py_ssize_t = shape[1]
 *         s_k: cython.Py_ssize_t = shape[2]
 *         s_l: cython.Py_ssize_t = shape[3]             # <<<<<<<<<<<<<<
 *         items = [
 *             arr_dtype,  # identifier & dtype
 */
    __pyx_v_s_l = (__pyx_v_shape[3]);

    /* "serializor/binary/ser.py":1024
 *         items = [
 *             arr_dtype,  # identifier & dtype
 *             pack_uint8(npy_dtype),  # type: ignore             # <<<<<<<<<<<<<<
 *             b"\x04",  # 4-dimension
 *             gen_encoded_int(s_i),  # type: ignore
 */
    __pyx_t_9 = __pyx_f_10serializor_6binary_3ser_pack_uint8(__pyx_v_npy_dtype); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1024, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);

    /* "serializor/binary/ser.py":1026
 *             pack_uint8(npy_dtype),  # type: ignore
 *             b"\x04",  # 4-dimension
 *             gen_encoded_int(s_i),  # type: ignore             # <<<<<<<<<<<<<<
 *             gen_encoded_int(s_j),  # type: ignore
 *             gen_encoded_int(s_k),  # type: ignore
 */
    __pyx_t_4 = __pyx_f_10serializor_6binary_3ser_gen_encoded_int(__pyx_v_s_i); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1026, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);

    /* "serializor/binary/ser.py":1027
 *             b"\x04",  # 4-dimension
 *             gen_encoded_int(s_i),  # type: ignore
 *             gen_encoded_int(s_j),  # type: ignore             # <<<<<<<<<<<<<<
 *             gen_encoded_int(s_k),  # type: ignore
 *             gen_encoded_int(s_l),  # type: ignore
 */
    __pyx_t_5 = __pyx_f_10serializor_6binary_3ser_gen_encoded_int(__pyx_v_s_j); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1027, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);

    /* "serializor/binary/ser.py":1028
 *             gen_encoded_int(s_i),  # type: ignore
 *             gen_encoded_int(s_j),  # type: ignore
 *             gen_encoded_int(s_k),  # type: ignore             # <<<<<<<<<<<<<<
 *             gen_encoded_int(s_l),  # type: ignore
 *         ]
 */
    __pyx_t_1 = __pyx_f_10serializor_6binary_3ser_gen_encoded_int(__pyx_v_s_k); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1028, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);

    /* "serializor/binary/ser.py":1029
 *             gen_encoded_int(s_j),  # type: ignore
 *             gen_encoded_int(s_k),  # type: ignore
 *             gen_encoded_int(s_l),  # type: ignore             # <<<<<<<<<<<<<<
 *         ]
 *         # . serialize values
 */
    __pyx_t_8 = __pyx_f_10serializor_6binary_3ser_gen_encoded_int(__pyx_v_s_l); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1029, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);

    /* "serializor/binary/ser.py":1022
 *         s_k: cython.Py_ssize_t = shape[2]
 *         s_l: cython.Py_ssize_t = shape[3]
 *         items = [             # <<<<<<<<<<<<<<
 *             arr_dtype,  # identifier & dtype
 *             pack_uint8(npy_dtype),  # type: ignore
 */
    __pyx_t_10 = PyList_New(7); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1022, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_INCREF(__pyx_v_arr_dtype);
    __Pyx_GIVEREF(__pyx_v_arr_dtype);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_10, 0, __pyx_v_arr_dtype)) __PYX_ERR(0, 1022, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_9);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_10, 1, __pyx_t_9)) __PYX_ERR(0, 1022, __pyx_L1_error);
    __Pyx_INCREF(__pyx_kp_b__9);
    __Pyx_GIVEREF(__pyx_kp_b__9);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_10, 2, __pyx_kp_b__9)) __PYX_ERR(0, 1022, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_4);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_10, 3, __pyx_t_4)) __PYX_ERR(0, 1022, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_5);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_10, 4, __pyx_t_5)) __PYX_ERR(0, 1022, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_10, 5, __pyx_t_1)) __PYX_ERR(0, 1022, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_8);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_10, 6, __pyx_t_8)) __PYX_ERR(0, 1022, __pyx_L1_error);
    __pyx_t_9 = 0;
    __pyx_t_4 = 0;
    __pyx_t_5 = 0;
    __pyx_t_1 = 0;
    __pyx_t_8 = 0;
    __pyx_v_items = ((PyObject*)__pyx_t_10);
    __pyx_t_10 = 0;

    /* "serializor/binary/ser.py":1032
 *         ]
 *         # . serialize values
 *         if s_l == 0:  # empty             # <<<<<<<<<<<<<<
 *             items.append(pack_uint8(utils.parse_arr_nptime_unit(arr)))  # type: ignore
 *         else:
 */
    __pyx_t_3 = (__pyx_v_s_l == 0);
    if (__pyx_t_3) {

      /* "serializor/binary/ser.py":1033
 *         # . serialize values
 *         if s_l == 0:  # empty
 *             items.append(pack_uint8(utils.parse_arr_nptime_unit(arr)))  # type: ignore             # <<<<<<<<<<<<<<
 *         else:
 *             items.append(pack_uint8(np.get_datetime64_unit(arr[0, 0, 0, 0])))  # type: ignore
 */
      __pyx_t_2 = __pyx_f_10serializor_5utils_parse_arr_nptime_unit(__pyx_v_arr); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(0, 1033, __pyx_L1_error)
      __pyx_t_10 = __pyx_f_10serializor_6binary_3ser_pack_uint8(__pyx_t_2); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1033, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __pyx_t_6 = __Pyx_PyList_Append(__pyx_v_items, __pyx_t_10); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(0, 1033, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

      /* "serializor/binary/ser.py":1032
 *         ]
 *         # . serialize values
 *         if s_l == 0:  # empty             # <<<<<<<<<<<<<<
 *             items.append(pack_uint8(utils.parse_arr_nptime_unit(arr)))  # type: ignore
 *         else:
 */
      goto __pyx_L10;
    }

    /* "serializor/binary/ser.py":1035
 *             items.append(pack_uint8(utils.parse_arr_nptime_unit(arr)))  # type: ignore
 *         else:
 *             items.append(pack_uint8(np.get_datetime64_unit(arr[0, 0, 0, 0])))  # type: ignore             # <<<<<<<<<<<<<<
 *             arr = np.PyArray_Cast(arr, np.NPY_TYPES.NPY_INT64)  # cast to int64
 *             values_json = _orjson_dumps_numpy(utils.arr_flatten(arr))
 */
    /*else*/ {
      __pyx_t_10 = __Pyx_PyObject_GetItem(((PyObject *)__pyx_v_arr), __pyx_tuple__12); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1035, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __pyx_t_8 = __pyx_f_10serializor_6binary_3ser_pack_uint8(__pyx_f_5numpy_get_datetime64_unit(__pyx_t_10)); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1035, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __pyx_t_6 = __Pyx_PyList_Append(__pyx_v_items, __pyx_t_8); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(0, 1035, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

      /* "serializor/binary/ser.py":1036
 *         else:
 *             items.append(pack_uint8(np.get_datetime64_unit(arr[0, 0, 0, 0])))  # type: ignore
 *             arr = np.PyArray_Cast(arr, np.NPY_TYPES.NPY_INT64)  # cast to int64             # <<<<<<<<<<<<<<
 *             values_json = _orjson_dumps_numpy(utils.arr_flatten(arr))
 *             items.append(gen_encoded_int(bytes_len(values_json)))  # type: ignore
 */
      __pyx_t_8 = PyArray_Cast(__pyx_v_arr, NPY_INT64); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1036, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      if (!(likely(((__pyx_t_8) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_8, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 1036, __pyx_L1_error)
      __Pyx_DECREF_SET(__pyx_v_arr, ((PyArrayObject *)__pyx_t_8));
      __pyx_t_8 = 0;

      /* "serializor/binary/ser.py":1037
 *             items.append(pack_uint8(np.get_datetime64_unit(arr[0, 0, 0, 0])))  # type: ignore
 *             arr = np.PyArray_Cast(arr, np.NPY_TYPES.NPY_INT64)  # cast to int64
 *             values_json = _orjson_dumps_numpy(utils.arr_flatten(arr))             # <<<<<<<<<<<<<<
 *             items.append(gen_encoded_int(bytes_len(values_json)))  # type: ignore
 *             items.append(values_json)
 */
      __pyx_t_8 = ((PyObject *)__pyx_f_10serializor_5utils_arr_flatten(__pyx_v_arr)); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1037, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_10 = __pyx_f_10serializor_6binary_3ser__orjson_dumps_numpy(__pyx_t_8); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1037, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_v_values_json = ((PyObject*)__pyx_t_10);
      __pyx_t_10 = 0;

      /* "serializor/binary/ser.py":1038
 *             arr = np.PyArray_Cast(arr, np.NPY_TYPES.NPY_INT64)  # cast to int64
 *             values_json = _orjson_dumps_numpy(utils.arr_flatten(arr))
 *             items.append(gen_encoded_int(bytes_len(values_json)))  # type: ignore             # <<<<<<<<<<<<<<
 *             items.append(values_json)
 *         return b"".join(items)
 */
      __pyx_t_7 = PyBytes_Size(__pyx_v_values_json); if (unlikely(__pyx_t_7 == ((Py_ssize_t)-1L))) __PYX_ERR(0, 1038, __pyx_L1_error)
      __pyx_t_10 = __pyx_f_10serializor_6binary_3ser_gen_encoded_int(__pyx_t_7); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1038, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __pyx_t_6 = __Pyx_PyList_Append(__pyx_v_items, __pyx_t_10); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(0, 1038, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

      /* "serializor/binary/ser.py":1039
 *             values_json = _orjson_dumps_numpy(utils.arr_flatten(arr))
 *             items.append(gen_encoded_int(bytes_len(values_json)))  # type: ignore
 *             items.append(values_json)             # <<<<<<<<<<<<<<
 *         return b"".join(items)
 * 
 */
      __pyx_t_6 = __Pyx_PyList_Append(__pyx_v_items, __pyx_v_values_json); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(0, 1039, __pyx_L1_error)
    }
    __pyx_L10:;

    /* "serializor/binary/ser.py":1040
 *             items.append(gen_encoded_int(bytes_len(values_json)))  # type: ignore
 *             items.append(values_json)
 *         return b"".join(items)             # <<<<<<<<<<<<<<
 * 
 *     # invalid
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyBytes_Join(__pyx_kp_b__4, __pyx_v_items); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1040, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    if (!(likely(PyBytes_CheckExact(__pyx_t_10))||((__pyx_t_10) == Py_None) || __Pyx_RaiseUnexpectedTypeError("bytes", __pyx_t_10))) __PYX_ERR(0, 1040, __pyx_L1_error)
    __pyx_r = ((PyObject*)__pyx_t_10);
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "serializor/binary/ser.py":1017
 * 
 *     # 4-dimension
 *     if ndim == 4:             # <<<<<<<<<<<<<<
 *         s_i: cython.Py_ssize_t = shape[0]
 *         s_j: cython.Py_ssize_t = shape[1]
 */
  }

  /* "serializor/binary/ser.py":1043
 * 
 *     # invalid
 *     raise ValueError("unsupported <'numpy.ndarray'> dimension [%d]." % ndim)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_10 = __Pyx_PyInt_From_int(__pyx_v_ndim); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1043, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __pyx_t_8 = PyUnicode_Format(__pyx_kp_u_unsupported_numpy_ndarray_dimens, __pyx_t_10); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1043, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  __pyx_t_10 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_8); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1043, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __Pyx_Raise(__pyx_t_10, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  __PYX_ERR(0, 1043, __pyx_L1_error)

  /* "serializor/binary/ser.py":926
 * 
 * 
 * @cython.cfunc             # <<<<<<<<<<<<<<
 * @cython.inline(True)
 * def _ser_ndarray_nptime(arr: np.ndarray, arr_dtype: bytes) -> bytes:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("serializor.binary.ser._ser_ndarray_nptime", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_items);
  __Pyx_XDECREF(__pyx_v_values_json);
  __Pyx_XDECREF((PyObject *)__pyx_v_arr);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "serializor/binary/ser.py":1046
 * 
 * 
 * @cython.cfunc             # <<<<<<<<<<<<<<
 * @cython.inline(True)
 * def _ser_ndarray_complex(arr: np.ndarray, arr_dtype: bytes) -> bytes:
 */

static PyObject *__pyx_f_10serializor_6binary_3ser__ser_ndarray_complex(PyArrayObject *__pyx_v_arr, PyObject *__pyx_v_arr_dtype) {
  int __pyx_v_npy_dtype;
  int __pyx_v_ndim;
  npy_intp *__pyx_v_shape;
  Py_ssize_t __pyx_v_s_i;
  PyObject *__pyx_v_items = NULL;
  PyObject *__pyx_v_values = 0;
  Py_ssize_t __pyx_v_i;
  PyObject *__pyx_v_item = NULL;
  PyObject *__pyx_v_values_json = NULL;
  Py_ssize_t __pyx_v_s_j;
  Py_ssize_t __pyx_v_j;
  Py_ssize_t __pyx_v_s_k;
  Py_ssize_t __pyx_v_k;
  Py_ssize_t __pyx_v_s_l;
  Py_ssize_t __pyx_v_l;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  Py_ssize_t __pyx_t_6;
  Py_ssize_t __pyx_t_7;
  Py_ssize_t __pyx_t_8;
  double __pyx_t_9;
  int __pyx_t_10;
  PyObject *__pyx_t_11 = NULL;
  Py_ssize_t __pyx_t_12;
  Py_ssize_t __pyx_t_13;
  Py_ssize_t __pyx_t_14;
  PyObject *__pyx_t_15 = NULL;
  Py_ssize_t __pyx_t_16;
  Py_ssize_t __pyx_t_17;
  Py_ssize_t __pyx_t_18;
  PyObject *__pyx_t_19 = NULL;
  Py_ssize_t __pyx_t_20;
  Py_ssize_t __pyx_t_21;
  Py_ssize_t __pyx_t_22;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_ser_ndarray_complex", 1);

  /* "serializor/binary/ser.py":1066
 *     >>> b'...[SHAPE][values_json_size(ENC)][values_json]'
 *     """
 *     npy_dtype: cython.int = arr.descr.type_num             # <<<<<<<<<<<<<<
 *     ndim: cython.int = arr.ndim
 *     shape = arr.shape
 */
  __pyx_t_1 = ((PyObject *)__pyx_f_5numpy_7ndarray_5descr_descr(__pyx_v_arr)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1066, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = ((PyArray_Descr *)__pyx_t_1)->type_num;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_npy_dtype = __pyx_t_2;

  /* "serializor/binary/ser.py":1067
 *     """
 *     npy_dtype: cython.int = arr.descr.type_num
 *     ndim: cython.int = arr.ndim             # <<<<<<<<<<<<<<
 *     shape = arr.shape
 *     # 1-dimension
 */
  __pyx_v_ndim = __pyx_f_5numpy_7ndarray_4ndim_ndim(__pyx_v_arr);

  /* "serializor/binary/ser.py":1068
 *     npy_dtype: cython.int = arr.descr.type_num
 *     ndim: cython.int = arr.ndim
 *     shape = arr.shape             # <<<<<<<<<<<<<<
 *     # 1-dimension
 *     if ndim == 1:
 */
  __pyx_v_shape = __pyx_f_5numpy_7ndarray_5shape_shape(__pyx_v_arr);

  /* "serializor/binary/ser.py":1070
 *     shape = arr.shape
 *     # 1-dimension
 *     if ndim == 1:             # <<<<<<<<<<<<<<
 *         # . array shape
 *         s_i: cython.Py_ssize_t = shape[0]
 */
  __pyx_t_3 = (__pyx_v_ndim == 1);
  if (__pyx_t_3) {

    /* "serializor/binary/ser.py":1072
 *     if ndim == 1:
 *         # . array shape
 *         s_i: cython.Py_ssize_t = shape[0]             # <<<<<<<<<<<<<<
 *         items = [
 *             arr_dtype,  # identifier & dtype
 */
    __pyx_v_s_i = (__pyx_v_shape[0]);

    /* "serializor/binary/ser.py":1075
 *         items = [
 *             arr_dtype,  # identifier & dtype
 *             pack_uint8(npy_dtype),  # type: ignore             # <<<<<<<<<<<<<<
 *             b"\x01",  # 1-dimension
 *             gen_encoded_int(s_i),  # type: ignore
 */
    __pyx_t_1 = __pyx_f_10serializor_6binary_3ser_pack_uint8(__pyx_v_npy_dtype); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1075, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);

    /* "serializor/binary/ser.py":1077
 *             pack_uint8(npy_dtype),  # type: ignore
 *             b"\x01",  # 1-dimension
 *             gen_encoded_int(s_i),  # type: ignore             # <<<<<<<<<<<<<<
 *         ]
 *         # . serialize values
 */
    __pyx_t_4 = __pyx_f_10serializor_6binary_3ser_gen_encoded_int(__pyx_v_s_i); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1077, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);

    /* "serializor/binary/ser.py":1073
 *         # . array shape
 *         s_i: cython.Py_ssize_t = shape[0]
 *         items = [             # <<<<<<<<<<<<<<
 *             arr_dtype,  # identifier & dtype
 *             pack_uint8(npy_dtype),  # type: ignore
 */
    __pyx_t_5 = PyList_New(4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1073, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_INCREF(__pyx_v_arr_dtype);
    __Pyx_GIVEREF(__pyx_v_arr_dtype);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 0, __pyx_v_arr_dtype)) __PYX_ERR(0, 1073, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 1, __pyx_t_1)) __PYX_ERR(0, 1073, __pyx_L1_error);
    __Pyx_INCREF(__pyx_kp_b__5);
    __Pyx_GIVEREF(__pyx_kp_b__5);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 2, __pyx_kp_b__5)) __PYX_ERR(0, 1073, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_4);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 3, __pyx_t_4)) __PYX_ERR(0, 1073, __pyx_L1_error);
    __pyx_t_1 = 0;
    __pyx_t_4 = 0;
    __pyx_v_items = ((PyObject*)__pyx_t_5);
    __pyx_t_5 = 0;

    /* "serializor/binary/ser.py":1080
 *         ]
 *         # . serialize values
 *         if s_i != 0:             # <<<<<<<<<<<<<<
 *             values: list = []
 *             for i in range(s_i):
 */
    __pyx_t_3 = (__pyx_v_s_i != 0);
    if (__pyx_t_3) {

      /* "serializor/binary/ser.py":1081
 *         # . serialize values
 *         if s_i != 0:
 *             values: list = []             # <<<<<<<<<<<<<<
 *             for i in range(s_i):
 *                 item = utils.arr_getitem_1d(arr, i)
 */
      __pyx_t_5 = PyList_New(0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1081, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_v_values = ((PyObject*)__pyx_t_5);
      __pyx_t_5 = 0;

      /* "serializor/binary/ser.py":1082
 *         if s_i != 0:
 *             values: list = []
 *             for i in range(s_i):             # <<<<<<<<<<<<<<
 *                 item = utils.arr_getitem_1d(arr, i)
 *                 values.append(complex_getreal(item))
 */
      __pyx_t_6 = __pyx_v_s_i;
      __pyx_t_7 = __pyx_t_6;
      for (__pyx_t_8 = 0; __pyx_t_8 < __pyx_t_7; __pyx_t_8+=1) {
        __pyx_v_i = __pyx_t_8;

        /* "serializor/binary/ser.py":1083
 *             values: list = []
 *             for i in range(s_i):
 *                 item = utils.arr_getitem_1d(arr, i)             # <<<<<<<<<<<<<<
 *                 values.append(complex_getreal(item))
 *                 values.append(complex_getimag(item))
 */
        __pyx_t_5 = __pyx_f_10serializor_5utils_arr_getitem_1d(__pyx_v_arr, __pyx_v_i); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1083, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_XDECREF_SET(__pyx_v_item, __pyx_t_5);
        __pyx_t_5 = 0;

        /* "serializor/binary/ser.py":1084
 *             for i in range(s_i):
 *                 item = utils.arr_getitem_1d(arr, i)
 *                 values.append(complex_getreal(item))             # <<<<<<<<<<<<<<
 *                 values.append(complex_getimag(item))
 *             values_json = _orjson_dumps(values)
 */
        __pyx_t_9 = PyComplex_RealAsDouble(__pyx_v_item); if (unlikely(__pyx_t_9 == ((double)-1.0) && PyErr_Occurred())) __PYX_ERR(0, 1084, __pyx_L1_error)
        __pyx_t_5 = PyFloat_FromDouble(__pyx_t_9); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1084, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_10 = __Pyx_PyList_Append(__pyx_v_values, __pyx_t_5); if (unlikely(__pyx_t_10 == ((int)-1))) __PYX_ERR(0, 1084, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

        /* "serializor/binary/ser.py":1085
 *                 item = utils.arr_getitem_1d(arr, i)
 *                 values.append(complex_getreal(item))
 *                 values.append(complex_getimag(item))             # <<<<<<<<<<<<<<
 *             values_json = _orjson_dumps(values)
 *             items.append(gen_encoded_int(bytes_len(values_json)))  # type: ignore
 */
        __pyx_t_9 = PyComplex_ImagAsDouble(__pyx_v_item); if (unlikely(__pyx_t_9 == ((double)-1.0) && PyErr_Occurred())) __PYX_ERR(0, 1085, __pyx_L1_error)
        __pyx_t_5 = PyFloat_FromDouble(__pyx_t_9); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1085, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_10 = __Pyx_PyList_Append(__pyx_v_values, __pyx_t_5); if (unlikely(__pyx_t_10 == ((int)-1))) __PYX_ERR(0, 1085, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }

      /* "serializor/binary/ser.py":1086
 *                 values.append(complex_getreal(item))
 *                 values.append(complex_getimag(item))
 *             values_json = _orjson_dumps(values)             # <<<<<<<<<<<<<<
 *             items.append(gen_encoded_int(bytes_len(values_json)))  # type: ignore
 *             items.append(values_json)
 */
      __pyx_t_5 = __pyx_f_10serializor_6binary_3ser__orjson_dumps(__pyx_v_values); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1086, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_v_values_json = ((PyObject*)__pyx_t_5);
      __pyx_t_5 = 0;

      /* "serializor/binary/ser.py":1087
 *                 values.append(complex_getimag(item))
 *             values_json = _orjson_dumps(values)
 *             items.append(gen_encoded_int(bytes_len(values_json)))  # type: ignore             # <<<<<<<<<<<<<<
 *             items.append(values_json)
 *         return b"".join(items)
 */
      __pyx_t_6 = PyBytes_Size(__pyx_v_values_json); if (unlikely(__pyx_t_6 == ((Py_ssize_t)-1L))) __PYX_ERR(0, 1087, __pyx_L1_error)
      __pyx_t_5 = __pyx_f_10serializor_6binary_3ser_gen_encoded_int(__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1087, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_10 = __Pyx_PyList_Append(__pyx_v_items, __pyx_t_5); if (unlikely(__pyx_t_10 == ((int)-1))) __PYX_ERR(0, 1087, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

      /* "serializor/binary/ser.py":1088
 *             values_json = _orjson_dumps(values)
 *             items.append(gen_encoded_int(bytes_len(values_json)))  # type: ignore
 *             items.append(values_json)             # <<<<<<<<<<<<<<
 *         return b"".join(items)
 * 
 */
      __pyx_t_10 = __Pyx_PyList_Append(__pyx_v_items, __pyx_v_values_json); if (unlikely(__pyx_t_10 == ((int)-1))) __PYX_ERR(0, 1088, __pyx_L1_error)

      /* "serializor/binary/ser.py":1080
 *         ]
 *         # . serialize values
 *         if s_i != 0:             # <<<<<<<<<<<<<<
 *             values: list = []
 *             for i in range(s_i):
 */
    }

    /* "serializor/binary/ser.py":1089
 *             items.append(gen_encoded_int(bytes_len(values_json)))  # type: ignore
 *             items.append(values_json)
 *         return b"".join(items)             # <<<<<<<<<<<<<<
 * 
 *     # 2-dimension
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_5 = __Pyx_PyBytes_Join(__pyx_kp_b__4, __pyx_v_items); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1089, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (!(likely(PyBytes_CheckExact(__pyx_t_5))||((__pyx_t_5) == Py_None) || __Pyx_RaiseUnexpectedTypeError("bytes", __pyx_t_5))) __PYX_ERR(0, 1089, __pyx_L1_error)
    __pyx_r = ((PyObject*)__pyx_t_5);
    __pyx_t_5 = 0;
    goto __pyx_L0;

    /* "serializor/binary/ser.py":1070
 *     shape = arr.shape
 *     # 1-dimension
 *     if ndim == 1:             # <<<<<<<<<<<<<<
 *         # . array shape
 *         s_i: cython.Py_ssize_t = shape[0]
 */
  }

  /* "serializor/binary/ser.py":1092
 * 
 *     # 2-dimension
 *     if ndim == 2:             # <<<<<<<<<<<<<<
 *         s_i: cython.Py_ssize_t = shape[0]
 *         s_j: cython.Py_ssize_t = shape[1]
 */
  __pyx_t_3 = (__pyx_v_ndim == 2);
  if (__pyx_t_3) {

    /* "serializor/binary/ser.py":1093
 *     # 2-dimension
 *     if ndim == 2:
 *         s_i: cython.Py_ssize_t = shape[0]             # <<<<<<<<<<<<<<
 *         s_j: cython.Py_ssize_t = shape[1]
 *         items = [
 */
    __pyx_v_s_i = (__pyx_v_shape[0]);

    /* "serializor/binary/ser.py":1094
 *     if ndim == 2:
 *         s_i: cython.Py_ssize_t = shape[0]
 *         s_j: cython.Py_ssize_t = shape[1]             # <<<<<<<<<<<<<<
 *         items = [
 *             arr_dtype,  # identifier & dtype
 */
    __pyx_v_s_j = (__pyx_v_shape[1]);

    /* "serializor/binary/ser.py":1097
 *         items = [
 *             arr_dtype,  # identifier & dtype
 *             pack_uint8(npy_dtype),  # type: ignore             # <<<<<<<<<<<<<<
 *             b"\x02",  # 2-dimension
 *             gen_encoded_int(s_i),  # type: ignore
 */
    __pyx_t_5 = __pyx_f_10serializor_6binary_3ser_pack_uint8(__pyx_v_npy_dtype); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1097, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);

    /* "serializor/binary/ser.py":1099
 *             pack_uint8(npy_dtype),  # type: ignore
 *             b"\x02",  # 2-dimension
 *             gen_encoded_int(s_i),  # type: ignore             # <<<<<<<<<<<<<<
 *             gen_encoded_int(s_j),  # type: ignore
 *         ]
 */
    __pyx_t_4 = __pyx_f_10serializor_6binary_3ser_gen_encoded_int(__pyx_v_s_i); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1099, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);

    /* "serializor/binary/ser.py":1100
 *             b"\x02",  # 2-dimension
 *             gen_encoded_int(s_i),  # type: ignore
 *             gen_encoded_int(s_j),  # type: ignore             # <<<<<<<<<<<<<<
 *         ]
 *         # . serialize values
 */
    __pyx_t_1 = __pyx_f_10serializor_6binary_3ser_gen_encoded_int(__pyx_v_s_j); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1100, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);

    /* "serializor/binary/ser.py":1095
 *         s_i: cython.Py_ssize_t = shape[0]
 *         s_j: cython.Py_ssize_t = shape[1]
 *         items = [             # <<<<<<<<<<<<<<
 *             arr_dtype,  # identifier & dtype
 *             pack_uint8(npy_dtype),  # type: ignore
 */
    __pyx_t_11 = PyList_New(5); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1095, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __Pyx_INCREF(__pyx_v_arr_dtype);
    __Pyx_GIVEREF(__pyx_v_arr_dtype);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_11, 0, __pyx_v_arr_dtype)) __PYX_ERR(0, 1095, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_5);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_11, 1, __pyx_t_5)) __PYX_ERR(0, 1095, __pyx_L1_error);
    __Pyx_INCREF(__pyx_kp_b__6);
    __Pyx_GIVEREF(__pyx_kp_b__6);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_11, 2, __pyx_kp_b__6)) __PYX_ERR(0, 1095, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_4);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_11, 3, __pyx_t_4)) __PYX_ERR(0, 1095, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_11, 4, __pyx_t_1)) __PYX_ERR(0, 1095, __pyx_L1_error);
    __pyx_t_5 = 0;
    __pyx_t_4 = 0;
    __pyx_t_1 = 0;
    __pyx_v_items = ((PyObject*)__pyx_t_11);
    __pyx_t_11 = 0;

    /* "serializor/binary/ser.py":1103
 *         ]
 *         # . serialize values
 *         if s_j != 0:             # <<<<<<<<<<<<<<
 *             values: list = []
 *             for i in range(s_i):
 */
    __pyx_t_3 = (__pyx_v_s_j != 0);
    if (__pyx_t_3) {

      /* "serializor/binary/ser.py":1104
 *         # . serialize values
 *         if s_j != 0:
 *             values: list = []             # <<<<<<<<<<<<<<
 *             for i in range(s_i):
 *                 for j in range(s_j):
 */
      __pyx_t_11 = PyList_New(0); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1104, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_11);
      __pyx_v_values = ((PyObject*)__pyx_t_11);
      __pyx_t_11 = 0;

      /* "serializor/binary/ser.py":1105
 *         if s_j != 0:
 *             values: list = []
 *             for i in range(s_i):             # <<<<<<<<<<<<<<
 *                 for j in range(s_j):
 *                     item = utils.arr_getitem_2d(arr, i, j)
 */
      __pyx_t_6 = __pyx_v_s_i;
      __pyx_t_7 = __pyx_t_6;
      for (__pyx_t_8 = 0; __pyx_t_8 < __pyx_t_7; __pyx_t_8+=1) {
        __pyx_v_i = __pyx_t_8;

        /* "serializor/binary/ser.py":1106
 *             values: list = []
 *             for i in range(s_i):
 *                 for j in range(s_j):             # <<<<<<<<<<<<<<
 *                     item = utils.arr_getitem_2d(arr, i, j)
 *                     values.append(complex_getreal(item))
 */
        __pyx_t_12 = __pyx_v_s_j;
        __pyx_t_13 = __pyx_t_12;
        for (__pyx_t_14 = 0; __pyx_t_14 < __pyx_t_13; __pyx_t_14+=1) {
          __pyx_v_j = __pyx_t_14;

          /* "serializor/binary/ser.py":1107
 *             for i in range(s_i):
 *                 for j in range(s_j):
 *                     item = utils.arr_getitem_2d(arr, i, j)             # <<<<<<<<<<<<<<
 *                     values.append(complex_getreal(item))
 *                     values.append(complex_getimag(item))
 */
          __pyx_t_11 = __pyx_f_10serializor_5utils_arr_getitem_2d(__pyx_v_arr, __pyx_v_i, __pyx_v_j); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1107, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_11);
          __Pyx_XDECREF_SET(__pyx_v_item, __pyx_t_11);
          __pyx_t_11 = 0;

          /* "serializor/binary/ser.py":1108
 *                 for j in range(s_j):
 *                     item = utils.arr_getitem_2d(arr, i, j)
 *                     values.append(complex_getreal(item))             # <<<<<<<<<<<<<<
 *                     values.append(complex_getimag(item))
 *             values_json = _orjson_dumps(values)
 */
          __pyx_t_9 = PyComplex_RealAsDouble(__pyx_v_item); if (unlikely(__pyx_t_9 == ((double)-1.0) && PyErr_Occurred())) __PYX_ERR(0, 1108, __pyx_L1_error)
          __pyx_t_11 = PyFloat_FromDouble(__pyx_t_9); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1108, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_11);
          __pyx_t_10 = __Pyx_PyList_Append(__pyx_v_values, __pyx_t_11); if (unlikely(__pyx_t_10 == ((int)-1))) __PYX_ERR(0, 1108, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

          /* "serializor/binary/ser.py":1109
 *                     item = utils.arr_getitem_2d(arr, i, j)
 *                     values.append(complex_getreal(item))
 *                     values.append(complex_getimag(item))             # <<<<<<<<<<<<<<
 *             values_json = _orjson_dumps(values)
 *             items.append(gen_encoded_int(bytes_len(values_json)))  # type: ignore
 */
          __pyx_t_9 = PyComplex_ImagAsDouble(__pyx_v_item); if (unlikely(__pyx_t_9 == ((double)-1.0) && PyErr_Occurred())) __PYX_ERR(0, 1109, __pyx_L1_error)
          __pyx_t_11 = PyFloat_FromDouble(__pyx_t_9); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1109, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_11);
          __pyx_t_10 = __Pyx_PyList_Append(__pyx_v_values, __pyx_t_11); if (unlikely(__pyx_t_10 == ((int)-1))) __PYX_ERR(0, 1109, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        }
      }

      /* "serializor/binary/ser.py":1110
 *                     values.append(complex_getreal(item))
 *                     values.append(complex_getimag(item))
 *             values_json = _orjson_dumps(values)             # <<<<<<<<<<<<<<
 *             items.append(gen_encoded_int(bytes_len(values_json)))  # type: ignore
 *             items.append(values_json)
 */
      __pyx_t_11 = __pyx_f_10serializor_6binary_3ser__orjson_dumps(__pyx_v_values); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1110, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_11);
      __pyx_v_values_json = ((PyObject*)__pyx_t_11);
      __pyx_t_11 = 0;

      /* "serializor/binary/ser.py":1111
 *                     values.append(complex_getimag(item))
 *             values_json = _orjson_dumps(values)
 *             items.append(gen_encoded_int(bytes_len(values_json)))  # type: ignore             # <<<<<<<<<<<<<<
 *             items.append(values_json)
 *         return b"".join(items)
 */
      __pyx_t_6 = PyBytes_Size(__pyx_v_values_json); if (unlikely(__pyx_t_6 == ((Py_ssize_t)-1L))) __PYX_ERR(0, 1111, __pyx_L1_error)
      __pyx_t_11 = __pyx_f_10serializor_6binary_3ser_gen_encoded_int(__pyx_t_6); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1111, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_11);
      __pyx_t_10 = __Pyx_PyList_Append(__pyx_v_items, __pyx_t_11); if (unlikely(__pyx_t_10 == ((int)-1))) __PYX_ERR(0, 1111, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

      /* "serializor/binary/ser.py":1112
 *             values_json = _orjson_dumps(values)
 *             items.append(gen_encoded_int(bytes_len(values_json)))  # type: ignore
 *             items.append(values_json)             # <<<<<<<<<<<<<<
 *         return b"".join(items)
 * 
 */
      __pyx_t_10 = __Pyx_PyList_Append(__pyx_v_items, __pyx_v_values_json); if (unlikely(__pyx_t_10 == ((int)-1))) __PYX_ERR(0, 1112, __pyx_L1_error)

      /* "serializor/binary/ser.py":1103
 *         ]
 *         # . serialize values
 *         if s_j != 0:             # <<<<<<<<<<<<<<
 *             values: list = []
 *             for i in range(s_i):
 */
    }

    /* "serializor/binary/ser.py":1113
 *             items.append(gen_encoded_int(bytes_len(values_json)))  # type: ignore
 *             items.append(values_json)
 *         return b"".join(items)             # <<<<<<<<<<<<<<
 * 
 *     # 3-dimension
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_11 = __Pyx_PyBytes_Join(__pyx_kp_b__4, __pyx_v_items); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1113, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    if (!(likely(PyBytes_CheckExact(__pyx_t_11))||((__pyx_t_11) == Py_None) || __Pyx_RaiseUnexpectedTypeError("bytes", __pyx_t_11))) __PYX_ERR(0, 1113, __pyx_L1_error)
    __pyx_r = ((PyObject*)__pyx_t_11);
    __pyx_t_11 = 0;
    goto __pyx_L0;

    /* "serializor/binary/ser.py":1092
 * 
 *     # 2-dimension
 *     if ndim == 2:             # <<<<<<<<<<<<<<
 *         s_i: cython.Py_ssize_t = shape[0]
 *         s_j: cython.Py_ssize_t = shape[1]
 */
  }

  /* "serializor/binary/ser.py":1116
 * 
 *     # 3-dimension
 *     if ndim == 3:             # <<<<<<<<<<<<<<
 *         s_i: cython.Py_ssize_t = shape[0]
 *         s_j: cython.Py_ssize_t = shape[1]
 */
  __pyx_t_3 = (__pyx_v_ndim == 3);
  if (__pyx_t_3) {

    /* "serializor/binary/ser.py":1117
 *     # 3-dimension
 *     if ndim == 3:
 *         s_i: cython.Py_ssize_t = shape[0]             # <<<<<<<<<<<<<<
 *         s_j: cython.Py_ssize_t = shape[1]
 *         s_k: cython.Py_ssize_t = shape[2]
 */
    __pyx_v_s_i = (__pyx_v_shape[0]);

    /* "serializor/binary/ser.py":1118
 *     if ndim == 3:
 *         s_i: cython.Py_ssize_t = shape[0]
 *         s_j: cython.Py_ssize_t = shape[1]             # <<<<<<<<<<<<<<
 *         s_k: cython.Py_ssize_t = shape[2]
 *         items = [
 */
    __pyx_v_s_j = (__pyx_v_shape[1]);

    /* "serializor/binary/ser.py":1119
 *         s_i: cython.Py_ssize_t = shape[0]
 *         s_j: cython.Py_ssize_t = shape[1]
 *         s_k: cython.Py_ssize_t = shape[2]             # <<<<<<<<<<<<<<
 *         items = [
 *             arr_dtype,  # identifier & dtype
 */
    __pyx_v_s_k = (__pyx_v_shape[2]);

    /* "serializor/binary/ser.py":1122
 *         items = [
 *             arr_dtype,  # identifier & dtype
 *             pack_uint8(npy_dtype),  # type: ignore             # <<<<<<<<<<<<<<
 *             b"\x03",  # 3-dimension
 *             gen_encoded_int(s_i),  # type: ignore
 */
    __pyx_t_11 = __pyx_f_10serializor_6binary_3ser_pack_uint8(__pyx_v_npy_dtype); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1122, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);

    /* "serializor/binary/ser.py":1124
 *             pack_uint8(npy_dtype),  # type: ignore
 *             b"\x03",  # 3-dimension
 *             gen_encoded_int(s_i),  # type: ignore             # <<<<<<<<<<<<<<
 *             gen_encoded_int(s_j),  # type: ignore
 *             gen_encoded_int(s_k),  # type: ignore
 */
    __pyx_t_1 = __pyx_f_10serializor_6binary_3ser_gen_encoded_int(__pyx_v_s_i); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1124, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);

    /* "serializor/binary/ser.py":1125
 *             b"\x03",  # 3-dimension
 *             gen_encoded_int(s_i),  # type: ignore
 *             gen_encoded_int(s_j),  # type: ignore             # <<<<<<<<<<<<<<
 *             gen_encoded_int(s_k),  # type: ignore
 *         ]
 */
    __pyx_t_4 = __pyx_f_10serializor_6binary_3ser_gen_encoded_int(__pyx_v_s_j); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1125, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);

    /* "serializor/binary/ser.py":1126
 *             gen_encoded_int(s_i),  # type: ignore
 *             gen_encoded_int(s_j),  # type: ignore
 *             gen_encoded_int(s_k),  # type: ignore             # <<<<<<<<<<<<<<
 *         ]
 *         # . serialize values
 */
    __pyx_t_5 = __pyx_f_10serializor_6binary_3ser_gen_encoded_int(__pyx_v_s_k); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1126, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);

    /* "serializor/binary/ser.py":1120
 *         s_j: cython.Py_ssize_t = shape[1]
 *         s_k: cython.Py_ssize_t = shape[2]
 *         items = [             # <<<<<<<<<<<<<<
 *             arr_dtype,  # identifier & dtype
 *             pack_uint8(npy_dtype),  # type: ignore
 */
    __pyx_t_15 = PyList_New(6); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1120, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    __Pyx_INCREF(__pyx_v_arr_dtype);
    __Pyx_GIVEREF(__pyx_v_arr_dtype);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_15, 0, __pyx_v_arr_dtype)) __PYX_ERR(0, 1120, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_11);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_15, 1, __pyx_t_11)) __PYX_ERR(0, 1120, __pyx_L1_error);
    __Pyx_INCREF(__pyx_kp_b__8);
    __Pyx_GIVEREF(__pyx_kp_b__8);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_15, 2, __pyx_kp_b__8)) __PYX_ERR(0, 1120, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_15, 3, __pyx_t_1)) __PYX_ERR(0, 1120, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_4);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_15, 4, __pyx_t_4)) __PYX_ERR(0, 1120, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_5);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_15, 5, __pyx_t_5)) __PYX_ERR(0, 1120, __pyx_L1_error);
    __pyx_t_11 = 0;
    __pyx_t_1 = 0;
    __pyx_t_4 = 0;
    __pyx_t_5 = 0;
    __pyx_v_items = ((PyObject*)__pyx_t_15);
    __pyx_t_15 = 0;

    /* "serializor/binary/ser.py":1129
 *         ]
 *         # . serialize values
 *         if s_k != 0:             # <<<<<<<<<<<<<<
 *             values: list = []
 *             for i in range(s_i):
 */
    __pyx_t_3 = (__pyx_v_s_k != 0);
    if (__pyx_t_3) {

      /* "serializor/binary/ser.py":1130
 *         # . serialize values
 *         if s_k != 0:
 *             values: list = []             # <<<<<<<<<<<<<<
 *             for i in range(s_i):
 *                 for j in range(s_j):
 */
      __pyx_t_15 = PyList_New(0); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1130, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_15);
      __pyx_v_values = ((PyObject*)__pyx_t_15);
      __pyx_t_15 = 0;

      /* "serializor/binary/ser.py":1131
 *         if s_k != 0:
 *             values: list = []
 *             for i in range(s_i):             # <<<<<<<<<<<<<<
 *                 for j in range(s_j):
 *                     for k in range(s_k):
 */
      __pyx_t_6 = __pyx_v_s_i;
      __pyx_t_7 = __pyx_t_6;
      for (__pyx_t_8 = 0; __pyx_t_8 < __pyx_t_7; __pyx_t_8+=1) {
        __pyx_v_i = __pyx_t_8;

        /* "serializor/binary/ser.py":1132
 *             values: list = []
 *             for i in range(s_i):
 *                 for j in range(s_j):             # <<<<<<<<<<<<<<
 *                     for k in range(s_k):
 *                         item = utils.arr_getitem_3d(arr, i, j, k)
 */
        __pyx_t_12 = __pyx_v_s_j;
        __pyx_t_13 = __pyx_t_12;
        for (__pyx_t_14 = 0; __pyx_t_14 < __pyx_t_13; __pyx_t_14+=1) {
          __pyx_v_j = __pyx_t_14;

          /* "serializor/binary/ser.py":1133
 *             for i in range(s_i):
 *                 for j in range(s_j):
 *                     for k in range(s_k):             # <<<<<<<<<<<<<<
 *                         item = utils.arr_getitem_3d(arr, i, j, k)
 *                         values.append(complex_getreal(item))
 */
          __pyx_t_16 = __pyx_v_s_k;
          __pyx_t_17 = __pyx_t_16;
          for (__pyx_t_18 = 0; __pyx_t_18 < __pyx_t_17; __pyx_t_18+=1) {
            __pyx_v_k = __pyx_t_18;

            /* "serializor/binary/ser.py":1134
 *                 for j in range(s_j):
 *                     for k in range(s_k):
 *                         item = utils.arr_getitem_3d(arr, i, j, k)             # <<<<<<<<<<<<<<
 *                         values.append(complex_getreal(item))
 *                         values.append(complex_getimag(item))
 */
            __pyx_t_15 = __pyx_f_10serializor_5utils_arr_getitem_3d(__pyx_v_arr, __pyx_v_i, __pyx_v_j, __pyx_v_k); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1134, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_15);
            __Pyx_XDECREF_SET(__pyx_v_item, __pyx_t_15);
            __pyx_t_15 = 0;

            /* "serializor/binary/ser.py":1135
 *                     for k in range(s_k):
 *                         item = utils.arr_getitem_3d(arr, i, j, k)
 *                         values.append(complex_getreal(item))             # <<<<<<<<<<<<<<
 *                         values.append(complex_getimag(item))
 *             values_json = _orjson_dumps(values)
 */
            __pyx_t_9 = PyComplex_RealAsDouble(__pyx_v_item); if (unlikely(__pyx_t_9 == ((double)-1.0) && PyErr_Occurred())) __PYX_ERR(0, 1135, __pyx_L1_error)
            __pyx_t_15 = PyFloat_FromDouble(__pyx_t_9); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1135, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_15);
            __pyx_t_10 = __Pyx_PyList_Append(__pyx_v_values, __pyx_t_15); if (unlikely(__pyx_t_10 == ((int)-1))) __PYX_ERR(0, 1135, __pyx_L1_error)
            __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;

            /* "serializor/binary/ser.py":1136
 *                         item = utils.arr_getitem_3d(arr, i, j, k)
 *                         values.append(complex_getreal(item))
 *                         values.append(complex_getimag(item))             # <<<<<<<<<<<<<<
 *             values_json = _orjson_dumps(values)
 *             items.append(gen_encoded_int(bytes_len(values_json)))  # type: ignore
 */
            __pyx_t_9 = PyComplex_ImagAsDouble(__pyx_v_item); if (unlikely(__pyx_t_9 == ((double)-1.0) && PyErr_Occurred())) __PYX_ERR(0, 1136, __pyx_L1_error)
            __pyx_t_15 = PyFloat_FromDouble(__pyx_t_9); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1136, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_15);
            __pyx_t_10 = __Pyx_PyList_Append(__pyx_v_values, __pyx_t_15); if (unlikely(__pyx_t_10 == ((int)-1))) __PYX_ERR(0, 1136, __pyx_L1_error)
            __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
          }
        }
      }

      /* "serializor/binary/ser.py":1137
 *                         values.append(complex_getreal(item))
 *                         values.append(complex_getimag(item))
 *             values_json = _orjson_dumps(values)             # <<<<<<<<<<<<<<
 *             items.append(gen_encoded_int(bytes_len(values_json)))  # type: ignore
 *             items.append(values_json)
 */
      __pyx_t_15 = __pyx_f_10serializor_6binary_3ser__orjson_dumps(__pyx_v_values); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1137, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_15);
      __pyx_v_values_json = ((PyObject*)__pyx_t_15);
      __pyx_t_15 = 0;

      /* "serializor/binary/ser.py":1138
 *                         values.append(complex_getimag(item))
 *             values_json = _orjson_dumps(values)
 *             items.append(gen_encoded_int(bytes_len(values_json)))  # type: ignore             # <<<<<<<<<<<<<<
 *             items.append(values_json)
 *         return b"".join(items)
 */
      __pyx_t_6 = PyBytes_Size(__pyx_v_values_json); if (unlikely(__pyx_t_6 == ((Py_ssize_t)-1L))) __PYX_ERR(0, 1138, __pyx_L1_error)
      __pyx_t_15 = __pyx_f_10serializor_6binary_3ser_gen_encoded_int(__pyx_t_6); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1138, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_15);
      __pyx_t_10 = __Pyx_PyList_Append(__pyx_v_items, __pyx_t_15); if (unlikely(__pyx_t_10 == ((int)-1))) __PYX_ERR(0, 1138, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;

      /* "serializor/binary/ser.py":1139
 *             values_json = _orjson_dumps(values)
 *             items.append(gen_encoded_int(bytes_len(values_json)))  # type: ignore
 *             items.append(values_json)             # <<<<<<<<<<<<<<
 *         return b"".join(items)
 * 
 */
      __pyx_t_10 = __Pyx_PyList_Append(__pyx_v_items, __pyx_v_values_json); if (unlikely(__pyx_t_10 == ((int)-1))) __PYX_ERR(0, 1139, __pyx_L1_error)

      /* "serializor/binary/ser.py":1129
 *         ]
 *         # . serialize values
 *         if s_k != 0:             # <<<<<<<<<<<<<<
 *             values: list = []
 *             for i in range(s_i):
 */
    }

    /* "serializor/binary/ser.py":1140
 *             items.append(gen_encoded_int(bytes_len(values_json)))  # type: ignore
 *             items.append(values_json)
 *         return b"".join(items)             # <<<<<<<<<<<<<<
 * 
 *     # 4-dimension
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_15 = __Pyx_PyBytes_Join(__pyx_kp_b__4, __pyx_v_items); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1140, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    if (!(likely(PyBytes_CheckExact(__pyx_t_15))||((__pyx_t_15) == Py_None) || __Pyx_RaiseUnexpectedTypeError("bytes", __pyx_t_15))) __PYX_ERR(0, 1140, __pyx_L1_error)
    __pyx_r = ((PyObject*)__pyx_t_15);
    __pyx_t_15 = 0;
    goto __pyx_L0;

    /* "serializor/binary/ser.py":1116
 * 
 *     # 3-dimension
 *     if ndim == 3:             # <<<<<<<<<<<<<<
 *         s_i: cython.Py_ssize_t = shape[0]
 *         s_j: cython.Py_ssize_t = shape[1]
 */
  }

  /* "serializor/binary/ser.py":1143
 * 
 *     # 4-dimension
 *     if ndim == 4:             # <<<<<<<<<<<<<<
 *         s_i: cython.Py_ssize_t = shape[0]
 *         s_j: cython.Py_ssize_t = shape[1]
 */
  __pyx_t_3 = (__pyx_v_ndim == 4);
  if (__pyx_t_3) {

    /* "serializor/binary/ser.py":1144
 *     # 4-dimension
 *     if ndim == 4:
 *         s_i: cython.Py_ssize_t = shape[0]             # <<<<<<<<<<<<<<
 *         s_j: cython.Py_ssize_t = shape[1]
 *         s_k: cython.Py_ssize_t = shape[2]
 */
    __pyx_v_s_i = (__pyx_v_shape[0]);

    /* "serializor/binary/ser.py":1145
 *     if ndim == 4:
 *         s_i: cython.Py_ssize_t = shape[0]
 *         s_j: cython.Py_ssize_t = shape[1]             # <<<<<<<<<<<<<<
 *         s_k: cython.Py_ssize_t = shape[2]
 *         s_l: cython.Py_ssize_t = shape[3]
 */
    __pyx_v_s_j = (__pyx_v_shape[1]);

    /* "serializor/binary/ser.py":1146
 *         s_i: cython.Py_ssize_t = shape[0]
 *         s_j: cython.Py_ssize_t = shape[1]
 *         s_k: cython.Py_ssize_t = shape[2]             # <<<<<<<<<<<<<<
 *         s_l: cython.Py_ssize_t = shape[3]
 *         items = [
 */
    __pyx_v_s_k = (__pyx_v_shape[2]);

    /* "serializor/binary/ser.py":1147
 *         s_j: cython.Py_ssize_t = shape[1]
 *         s_k: cython.Py_ssize_t = shape[2]
 *         s_l: cython.Py_ssize_t = shape[3]             # <<<<<<<<<<<<<<
 *         items = [
 *             arr_dtype,  # identifier & dtype
 */
    __pyx_v_s_l = (__pyx_v_shape[3]);

    /* "serializor/binary/ser.py":1150
 *         items = [
 *             arr_dtype,  # identifier & dtype
 *             pack_uint8(npy_dtype),  # type: ignore             # <<<<<<<<<<<<<<
 *             b"\x04",  # 4-dimension
 *             gen_encoded_int(s_i),  # type: ignore
 */
    __pyx_t_15 = __pyx_f_10serializor_6binary_3ser_pack_uint8(__pyx_v_npy_dtype); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1150, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);

    /* "serializor/binary/ser.py":1152
 *             pack_uint8(npy_dtype),  # type: ignore
 *             b"\x04",  # 4-dimension
 *             gen_encoded_int(s_i),  # type: ignore             # <<<<<<<<<<<<<<
 *             gen_encoded_int(s_j),  # type: ignore
 *             gen_encoded_int(s_k),  # type: ignore
 */
    __pyx_t_5 = __pyx_f_10serializor_6binary_3ser_gen_encoded_int(__pyx_v_s_i); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1152, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);

    /* "serializor/binary/ser.py":1153
 *             b"\x04",  # 4-dimension
 *             gen_encoded_int(s_i),  # type: ignore
 *             gen_encoded_int(s_j),  # type: ignore             # <<<<<<<<<<<<<<
 *             gen_encoded_int(s_k),  # type: ignore
 *             gen_encoded_int(s_l),  # type: ignore
 */
    __pyx_t_4 = __pyx_f_10serializor_6binary_3ser_gen_encoded_int(__pyx_v_s_j); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1153, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);

    /* "serializor/binary/ser.py":1154
 *             gen_encoded_int(s_i),  # type: ignore
 *             gen_encoded_int(s_j),  # type: ignore
 *             gen_encoded_int(s_k),  # type: ignore             # <<<<<<<<<<<<<<
 *             gen_encoded_int(s_l),  # type: ignore
 *         ]
 */
    __pyx_t_1 = __pyx_f_10serializor_6binary_3ser_gen_encoded_int(__pyx_v_s_k); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1154, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);

    /* "serializor/binary/ser.py":1155
 *             gen_encoded_int(s_j),  # type: ignore
 *             gen_encoded_int(s_k),  # type: ignore
 *             gen_encoded_int(s_l),  # type: ignore             # <<<<<<<<<<<<<<
 *         ]
 *         # . serialize values
 */
    __pyx_t_11 = __pyx_f_10serializor_6binary_3ser_gen_encoded_int(__pyx_v_s_l); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1155, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);

    /* "serializor/binary/ser.py":1148
 *         s_k: cython.Py_ssize_t = shape[2]
 *         s_l: cython.Py_ssize_t = shape[3]
 *         items = [             # <<<<<<<<<<<<<<
 *             arr_dtype,  # identifier & dtype
 *             pack_uint8(npy_dtype),  # type: ignore
 */
    __pyx_t_19 = PyList_New(7); if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 1148, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_19);
    __Pyx_INCREF(__pyx_v_arr_dtype);
    __Pyx_GIVEREF(__pyx_v_arr_dtype);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_19, 0, __pyx_v_arr_dtype)) __PYX_ERR(0, 1148, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_15);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_19, 1, __pyx_t_15)) __PYX_ERR(0, 1148, __pyx_L1_error);
    __Pyx_INCREF(__pyx_kp_b__9);
    __Pyx_GIVEREF(__pyx_kp_b__9);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_19, 2, __pyx_kp_b__9)) __PYX_ERR(0, 1148, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_5);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_19, 3, __pyx_t_5)) __PYX_ERR(0, 1148, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_4);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_19, 4, __pyx_t_4)) __PYX_ERR(0, 1148, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_19, 5, __pyx_t_1)) __PYX_ERR(0, 1148, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_11);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_19, 6, __pyx_t_11)) __PYX_ERR(0, 1148, __pyx_L1_error);
    __pyx_t_15 = 0;
    __pyx_t_5 = 0;
    __pyx_t_4 = 0;
    __pyx_t_1 = 0;
    __pyx_t_11 = 0;
    __pyx_v_items = ((PyObject*)__pyx_t_19);
    __pyx_t_19 = 0;

    /* "serializor/binary/ser.py":1158
 *         ]
 *         # . serialize values
 *         if s_l != 0:             # <<<<<<<<<<<<<<
 *             values: list = []
 *             for i in range(s_i):
 */
    __pyx_t_3 = (__pyx_v_s_l != 0);
    if (__pyx_t_3) {

      /* "serializor/binary/ser.py":1159
 *         # . serialize values
 *         if s_l != 0:
 *             values: list = []             # <<<<<<<<<<<<<<
 *             for i in range(s_i):
 *                 for j in range(s_j):
 */
      __pyx_t_19 = PyList_New(0); if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 1159, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_19);
      __pyx_v_values = ((PyObject*)__pyx_t_19);
      __pyx_t_19 = 0;

      /* "serializor/binary/ser.py":1160
 *         if s_l != 0:
 *             values: list = []
 *             for i in range(s_i):             # <<<<<<<<<<<<<<
 *                 for j in range(s_j):
 *                     for k in range(s_k):
 */
      __pyx_t_6 = __pyx_v_s_i;
      __pyx_t_7 = __pyx_t_6;
      for (__pyx_t_8 = 0; __pyx_t_8 < __pyx_t_7; __pyx_t_8+=1) {
        __pyx_v_i = __pyx_t_8;

        /* "serializor/binary/ser.py":1161
 *             values: list = []
 *             for i in range(s_i):
 *                 for j in range(s_j):             # <<<<<<<<<<<<<<
 *                     for k in range(s_k):
 *                         for l in range(s_l):
 */
        __pyx_t_12 = __pyx_v_s_j;
        __pyx_t_13 = __pyx_t_12;
        for (__pyx_t_14 = 0; __pyx_t_14 < __pyx_t_13; __pyx_t_14+=1) {
          __pyx_v_j = __pyx_t_14;

          /* "serializor/binary/ser.py":1162
 *             for i in range(s_i):
 *                 for j in range(s_j):
 *                     for k in range(s_k):             # <<<<<<<<<<<<<<
 *                         for l in range(s_l):
 *                             item = utils.arr_getitem_4d(arr, i, j, k, l)
 */
          __pyx_t_16 = __pyx_v_s_k;
          __pyx_t_17 = __pyx_t_16;
          for (__pyx_t_18 = 0; __pyx_t_18 < __pyx_t_17; __pyx_t_18+=1) {
            __pyx_v_k = __pyx_t_18;

            /* "serializor/binary/ser.py":1163
 *                 for j in range(s_j):
 *                     for k in range(s_k):
 *                         for l in range(s_l):             # <<<<<<<<<<<<<<
 *                             item = utils.arr_getitem_4d(arr, i, j, k, l)
 *                             values.append(complex_getreal(item))
 */
            __pyx_t_20 = __pyx_v_s_l;
            __pyx_t_21 = __pyx_t_20;
            for (__pyx_t_22 = 0; __pyx_t_22 < __pyx_t_21; __pyx_t_22+=1) {
              __pyx_v_l = __pyx_t_22;

              /* "serializor/binary/ser.py":1164
 *                     for k in range(s_k):
 *                         for l in range(s_l):
 *                             item = utils.arr_getitem_4d(arr, i, j, k, l)             # <<<<<<<<<<<<<<
 *                             values.append(complex_getreal(item))
 *                             values.append(complex_getimag(item))
 */
              __pyx_t_19 = __pyx_f_10serializor_5utils_arr_getitem_4d(__pyx_v_arr, __pyx_v_i, __pyx_v_j, __pyx_v_k, __pyx_v_l); if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 1164, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_19);
              __Pyx_XDECREF_SET(__pyx_v_item, __pyx_t_19);
              __pyx_t_19 = 0;

              /* "serializor/binary/ser.py":1165
 *                         for l in range(s_l):
 *                             item = utils.arr_getitem_4d(arr, i, j, k, l)
 *                             values.append(complex_getreal(item))             # <<<<<<<<<<<<<<
 *                             values.append(complex_getimag(item))
 *             values_json = _orjson_dumps(values)
 */
              __pyx_t_9 = PyComplex_RealAsDouble(__pyx_v_item); if (unlikely(__pyx_t_9 == ((double)-1.0) && PyErr_Occurred())) __PYX_ERR(0, 1165, __pyx_L1_error)
              __pyx_t_19 = PyFloat_FromDouble(__pyx_t_9); if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 1165, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_19);
              __pyx_t_10 = __Pyx_PyList_Append(__pyx_v_values, __pyx_t_19); if (unlikely(__pyx_t_10 == ((int)-1))) __PYX_ERR(0, 1165, __pyx_L1_error)
              __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;

              /* "serializor/binary/ser.py":1166
 *                             item = utils.arr_getitem_4d(arr, i, j, k, l)
 *                             values.append(complex_getreal(item))
 *                             values.append(complex_getimag(item))             # <<<<<<<<<<<<<<
 *             values_json = _orjson_dumps(values)
 *             items.append(gen_encoded_int(bytes_len(values_json)))  # type: ignore
 */
              __pyx_t_9 = PyComplex_ImagAsDouble(__pyx_v_item); if (unlikely(__pyx_t_9 == ((double)-1.0) && PyErr_Occurred())) __PYX_ERR(0, 1166, __pyx_L1_error)
              __pyx_t_19 = PyFloat_FromDouble(__pyx_t_9); if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 1166, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_19);
              __pyx_t_10 = __Pyx_PyList_Append(__pyx_v_values, __pyx_t_19); if (unlikely(__pyx_t_10 == ((int)-1))) __PYX_ERR(0, 1166, __pyx_L1_error)
              __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
            }
          }
        }
      }

      /* "serializor/binary/ser.py":1167
 *                             values.append(complex_getreal(item))
 *                             values.append(complex_getimag(item))
 *             values_json = _orjson_dumps(values)             # <<<<<<<<<<<<<<
 *             items.append(gen_encoded_int(bytes_len(values_json)))  # type: ignore
 *             items.append(values_json)
 */
      __pyx_t_19 = __pyx_f_10serializor_6binary_3ser__orjson_dumps(__pyx_v_values); if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 1167, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_19);
      __pyx_v_values_json = ((PyObject*)__pyx_t_19);
      __pyx_t_19 = 0;

      /* "serializor/binary/ser.py":1168
 *                             values.append(complex_getimag(item))
 *             values_json = _orjson_dumps(values)
 *             items.append(gen_encoded_int(bytes_len(values_json)))  # type: ignore             # <<<<<<<<<<<<<<
 *             items.append(values_json)
 *         return b"".join(items)
 */
      __pyx_t_6 = PyBytes_Size(__pyx_v_values_json); if (unlikely(__pyx_t_6 == ((Py_ssize_t)-1L))) __PYX_ERR(0, 1168, __pyx_L1_error)
      __pyx_t_19 = __pyx_f_10serializor_6binary_3ser_gen_encoded_int(__pyx_t_6); if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 1168, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_19);
      __pyx_t_10 = __Pyx_PyList_Append(__pyx_v_items, __pyx_t_19); if (unlikely(__pyx_t_10 == ((int)-1))) __PYX_ERR(0, 1168, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;

      /* "serializor/binary/ser.py":1169
 *             values_json = _orjson_dumps(values)
 *             items.append(gen_encoded_int(bytes_len(values_json)))  # type: ignore
 *             items.append(values_json)             # <<<<<<<<<<<<<<
 *         return b"".join(items)
 * 
 */
      __pyx_t_10 = __Pyx_PyList_Append(__pyx_v_items, __pyx_v_values_json); if (unlikely(__pyx_t_10 == ((int)-1))) __PYX_ERR(0, 1169, __pyx_L1_error)

      /* "serializor/binary/ser.py":1158
 *         ]
 *         # . serialize values
 *         if s_l != 0:             # <<<<<<<<<<<<<<
 *             values: list = []
 *             for i in range(s_i):
 */
    }

    /* "serializor/binary/ser.py":1170
 *             items.append(gen_encoded_int(bytes_len(values_json)))  # type: ignore
 *             items.append(values_json)
 *         return b"".join(items)             # <<<<<<<<<<<<<<
 * 
 *     # invalid
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_19 = __Pyx_PyBytes_Join(__pyx_kp_b__4, __pyx_v_items); if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 1170, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_19);
    if (!(likely(PyBytes_CheckExact(__pyx_t_19))||((__pyx_t_19) == Py_None) || __Pyx_RaiseUnexpectedTypeError("bytes", __pyx_t_19))) __PYX_ERR(0, 1170, __pyx_L1_error)
    __pyx_r = ((PyObject*)__pyx_t_19);
    __pyx_t_19 = 0;
    goto __pyx_L0;

    /* "serializor/binary/ser.py":1143
 * 
 *     # 4-dimension
 *     if ndim == 4:             # <<<<<<<<<<<<<<
 *         s_i: cython.Py_ssize_t = shape[0]
 *         s_j: cython.Py_ssize_t = shape[1]
 */
  }

  /* "serializor/binary/ser.py":1173
 * 
 *     # invalid
 *     raise ValueError("unsupported <'numpy.ndarray'> dimension [%d]." % ndim)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_19 = __Pyx_PyInt_From_int(__pyx_v_ndim); if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 1173, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_19);
  __pyx_t_11 = PyUnicode_Format(__pyx_kp_u_unsupported_numpy_ndarray_dimens, __pyx_t_19); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1173, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
  __pyx_t_19 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_11); if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 1173, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_19);
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
  __Pyx_Raise(__pyx_t_19, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
  __PYX_ERR(0, 1173, __pyx_L1_error)

  /* "serializor/binary/ser.py":1046
 * 
 * 
 * @cython.cfunc             # <<<<<<<<<<<<<<
 * @cython.inline(True)
 * def _ser_ndarray_complex(arr: np.ndarray, arr_dtype: bytes) -> bytes:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_15);
  __Pyx_XDECREF(__pyx_t_19);
  __Pyx_AddTraceback("serializor.binary.ser._ser_ndarray_complex", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_items);
  __Pyx_XDECREF(__pyx_v_values);
  __Pyx_XDECREF(__pyx_v_item);
  __Pyx_XDECREF(__pyx_v_values_json);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "serializor/binary/ser.py":1176
 * 
 * 
 * @cython.cfunc             # <<<<<<<<<<<<<<
 * @cython.inline(True)
 * def _ser_ndarray_bytes(arr: np.ndarray, arr_dtype: bytes) -> bytes:
 */

static PyObject *__pyx_f_10serializor_6binary_3ser__ser_ndarray_bytes(PyArrayObject *__pyx_v_arr, PyObject *__pyx_v_arr_dtype) {
  int __pyx_v_npy_dtype;
  int __pyx_v_ndim;
  npy_intp *__pyx_v_shape;
  Py_ssize_t __pyx_v_s_i;
  PyObject *__pyx_v_items = NULL;
  int __pyx_v_ch_size;
  Py_ssize_t __pyx_v_i;
  Py_ssize_t __pyx_v_s_j;
  Py_ssize_t __pyx_v_j;
  Py_ssize_t __pyx_v_s_k;
  Py_ssize_t __pyx_v_k;
  Py_ssize_t __pyx_v_s_l;
  Py_ssize_t __pyx_v_l;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  Py_ssize_t __pyx_t_7;
  Py_ssize_t __pyx_t_8;
  Py_ssize_t __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  Py_ssize_t __pyx_t_11;
  Py_ssize_t __pyx_t_12;
  Py_ssize_t __pyx_t_13;
  PyObject *__pyx_t_14 = NULL;
  Py_ssize_t __pyx_t_15;
  Py_ssize_t __pyx_t_16;
  Py_ssize_t __pyx_t_17;
  PyObject *__pyx_t_18 = NULL;
  Py_ssize_t __pyx_t_19;
  Py_ssize_t __pyx_t_20;
  Py_ssize_t __pyx_t_21;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_ser_ndarray_bytes", 1);

  /* "serializor/binary/ser.py":1195
 *     >>> b'...[SHAPE][i1(SER)]...[iN(SER)]'
 *     """
 *     npy_dtype: cython.int = arr.descr.type_num             # <<<<<<<<<<<<<<
 *     ndim: cython.int = arr.ndim
 *     shape = arr.shape
 */
  __pyx_t_1 = ((PyObject *)__pyx_f_5numpy_7ndarray_5descr_descr(__pyx_v_arr)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1195, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = ((PyArray_Descr *)__pyx_t_1)->type_num;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_npy_dtype = __pyx_t_2;

  /* "serializor/binary/ser.py":1196
 *     """
 *     npy_dtype: cython.int = arr.descr.type_num
 *     ndim: cython.int = arr.ndim             # <<<<<<<<<<<<<<
 *     shape = arr.shape
 *     # 1-dimension
 */
  __pyx_v_ndim = __pyx_f_5numpy_7ndarray_4ndim_ndim(__pyx_v_arr);

  /* "serializor/binary/ser.py":1197
 *     npy_dtype: cython.int = arr.descr.type_num
 *     ndim: cython.int = arr.ndim
 *     shape = arr.shape             # <<<<<<<<<<<<<<
 *     # 1-dimension
 *     if ndim == 1:
 */
  __pyx_v_shape = __pyx_f_5numpy_7ndarray_5shape_shape(__pyx_v_arr);

  /* "serializor/binary/ser.py":1199
 *     shape = arr.shape
 *     # 1-dimension
 *     if ndim == 1:             # <<<<<<<<<<<<<<
 *         # . array shape
 *         s_i: cython.Py_ssize_t = shape[0]
 */
  __pyx_t_3 = (__pyx_v_ndim == 1);
  if (__pyx_t_3) {

    /* "serializor/binary/ser.py":1201
 *     if ndim == 1:
 *         # . array shape
 *         s_i: cython.Py_ssize_t = shape[0]             # <<<<<<<<<<<<<<
 *         items = [
 *             arr_dtype,  # identifier & dtype
 */
    __pyx_v_s_i = (__pyx_v_shape[0]);

    /* "serializor/binary/ser.py":1204
 *         items = [
 *             arr_dtype,  # identifier & dtype
 *             pack_uint8(npy_dtype),  # type: ignore             # <<<<<<<<<<<<<<
 *             b"\x01",  # 1-dimension
 *             gen_encoded_int(s_i),  # type: ignore
 */
    __pyx_t_1 = __pyx_f_10serializor_6binary_3ser_pack_uint8(__pyx_v_npy_dtype); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1204, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);

    /* "serializor/binary/ser.py":1206
 *             pack_uint8(npy_dtype),  # type: ignore
 *             b"\x01",  # 1-dimension
 *             gen_encoded_int(s_i),  # type: ignore             # <<<<<<<<<<<<<<
 *         ]
 *         # . serialize values
 */
    __pyx_t_4 = __pyx_f_10serializor_6binary_3ser_gen_encoded_int(__pyx_v_s_i); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1206, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);

    /* "serializor/binary/ser.py":1202
 *         # . array shape
 *         s_i: cython.Py_ssize_t = shape[0]
 *         items = [             # <<<<<<<<<<<<<<
 *             arr_dtype,  # identifier & dtype
 *             pack_uint8(npy_dtype),  # type: ignore
 */
    __pyx_t_5 = PyList_New(4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1202, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_INCREF(__pyx_v_arr_dtype);
    __Pyx_GIVEREF(__pyx_v_arr_dtype);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 0, __pyx_v_arr_dtype)) __PYX_ERR(0, 1202, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 1, __pyx_t_1)) __PYX_ERR(0, 1202, __pyx_L1_error);
    __Pyx_INCREF(__pyx_kp_b__5);
    __Pyx_GIVEREF(__pyx_kp_b__5);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 2, __pyx_kp_b__5)) __PYX_ERR(0, 1202, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_4);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 3, __pyx_t_4)) __PYX_ERR(0, 1202, __pyx_L1_error);
    __pyx_t_1 = 0;
    __pyx_t_4 = 0;
    __pyx_v_items = ((PyObject*)__pyx_t_5);
    __pyx_t_5 = 0;

    /* "serializor/binary/ser.py":1209
 *         ]
 *         # . serialize values
 *         if s_i != 0:             # <<<<<<<<<<<<<<
 *             ch_size: cython.int = arr.descr.itemsize
 *             items.append(gen_encoded_int(ch_size))  # type: ignore
 */
    __pyx_t_3 = (__pyx_v_s_i != 0);
    if (__pyx_t_3) {

      /* "serializor/binary/ser.py":1210
 *         # . serialize values
 *         if s_i != 0:
 *             ch_size: cython.int = arr.descr.itemsize             # <<<<<<<<<<<<<<
 *             items.append(gen_encoded_int(ch_size))  # type: ignore
 *             for i in range(s_i):
 */
      __pyx_t_5 = ((PyObject *)__pyx_f_5numpy_7ndarray_5descr_descr(__pyx_v_arr)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1210, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_v_ch_size = __pyx_f_5numpy_5dtype_8itemsize_itemsize(((PyArray_Descr *)__pyx_t_5));
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

      /* "serializor/binary/ser.py":1211
 *         if s_i != 0:
 *             ch_size: cython.int = arr.descr.itemsize
 *             items.append(gen_encoded_int(ch_size))  # type: ignore             # <<<<<<<<<<<<<<
 *             for i in range(s_i):
 *                 items.append(_ser_bytes(utils.arr_getitem_1d(arr, i)))
 */
      __pyx_t_5 = __pyx_f_10serializor_6binary_3ser_gen_encoded_int(__pyx_v_ch_size); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1211, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_6 = __Pyx_PyList_Append(__pyx_v_items, __pyx_t_5); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(0, 1211, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

      /* "serializor/binary/ser.py":1212
 *             ch_size: cython.int = arr.descr.itemsize
 *             items.append(gen_encoded_int(ch_size))  # type: ignore
 *             for i in range(s_i):             # <<<<<<<<<<<<<<
 *                 items.append(_ser_bytes(utils.arr_getitem_1d(arr, i)))
 *         return b"".join(items)
 */
      __pyx_t_7 = __pyx_v_s_i;
      __pyx_t_8 = __pyx_t_7;
      for (__pyx_t_9 = 0; __pyx_t_9 < __pyx_t_8; __pyx_t_9+=1) {
        __pyx_v_i = __pyx_t_9;

        /* "serializor/binary/ser.py":1213
 *             items.append(gen_encoded_int(ch_size))  # type: ignore
 *             for i in range(s_i):
 *                 items.append(_ser_bytes(utils.arr_getitem_1d(arr, i)))             # <<<<<<<<<<<<<<
 *         return b"".join(items)
 * 
 */
        __pyx_t_5 = __pyx_f_10serializor_5utils_arr_getitem_1d(__pyx_v_arr, __pyx_v_i); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1213, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_4 = __pyx_f_10serializor_6binary_3ser__ser_bytes(__pyx_t_5); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1213, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __pyx_t_6 = __Pyx_PyList_Append(__pyx_v_items, __pyx_t_4); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(0, 1213, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      }

      /* "serializor/binary/ser.py":1209
 *         ]
 *         # . serialize values
 *         if s_i != 0:             # <<<<<<<<<<<<<<
 *             ch_size: cython.int = arr.descr.itemsize
 *             items.append(gen_encoded_int(ch_size))  # type: ignore
 */
    }

    /* "serializor/binary/ser.py":1214
 *             for i in range(s_i):
 *                 items.append(_ser_bytes(utils.arr_getitem_1d(arr, i)))
 *         return b"".join(items)             # <<<<<<<<<<<<<<
 * 
 *     # 2-dimension
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyBytes_Join(__pyx_kp_b__4, __pyx_v_items); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1214, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    if (!(likely(PyBytes_CheckExact(__pyx_t_4))||((__pyx_t_4) == Py_None) || __Pyx_RaiseUnexpectedTypeError("bytes", __pyx_t_4))) __PYX_ERR(0, 1214, __pyx_L1_error)
    __pyx_r = ((PyObject*)__pyx_t_4);
    __pyx_t_4 = 0;
    goto __pyx_L0;

    /* "serializor/binary/ser.py":1199
 *     shape = arr.shape
 *     # 1-dimension
 *     if ndim == 1:             # <<<<<<<<<<<<<<
 *         # . array shape
 *         s_i: cython.Py_ssize_t = shape[0]
 */
  }

  /* "serializor/binary/ser.py":1217
 * 
 *     # 2-dimension
 *     if ndim == 2:             # <<<<<<<<<<<<<<
 *         # . array shape
 *         s_i: cython.Py_ssize_t = shape[0]
 */
  __pyx_t_3 = (__pyx_v_ndim == 2);
  if (__pyx_t_3) {

    /* "serializor/binary/ser.py":1219
 *     if ndim == 2:
 *         # . array shape
 *         s_i: cython.Py_ssize_t = shape[0]             # <<<<<<<<<<<<<<
 *         s_j: cython.Py_ssize_t = shape[1]
 *         items = [
 */
    __pyx_v_s_i = (__pyx_v_shape[0]);

    /* "serializor/binary/ser.py":1220
 *         # . array shape
 *         s_i: cython.Py_ssize_t = shape[0]
 *         s_j: cython.Py_ssize_t = shape[1]             # <<<<<<<<<<<<<<
 *         items = [
 *             arr_dtype,  # identifier & dtype
 */
    __pyx_v_s_j = (__pyx_v_shape[1]);

    /* "serializor/binary/ser.py":1223
 *         items = [
 *             arr_dtype,  # identifier & dtype
 *             pack_uint8(npy_dtype),  # type: ignore             # <<<<<<<<<<<<<<
 *             b"\x02",  # 2-dimension
 *             gen_encoded_int(s_i),  # type: ignore
 */
    __pyx_t_4 = __pyx_f_10serializor_6binary_3ser_pack_uint8(__pyx_v_npy_dtype); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1223, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);

    /* "serializor/binary/ser.py":1225
 *             pack_uint8(npy_dtype),  # type: ignore
 *             b"\x02",  # 2-dimension
 *             gen_encoded_int(s_i),  # type: ignore             # <<<<<<<<<<<<<<
 *             gen_encoded_int(s_j),  # type: ignore
 *         ]
 */
    __pyx_t_5 = __pyx_f_10serializor_6binary_3ser_gen_encoded_int(__pyx_v_s_i); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1225, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);

    /* "serializor/binary/ser.py":1226
 *             b"\x02",  # 2-dimension
 *             gen_encoded_int(s_i),  # type: ignore
 *             gen_encoded_int(s_j),  # type: ignore             # <<<<<<<<<<<<<<
 *         ]
 *         # . serialize values
 */
    __pyx_t_1 = __pyx_f_10serializor_6binary_3ser_gen_encoded_int(__pyx_v_s_j); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1226, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);

    /* "serializor/binary/ser.py":1221
 *         s_i: cython.Py_ssize_t = shape[0]
 *         s_j: cython.Py_ssize_t = shape[1]
 *         items = [             # <<<<<<<<<<<<<<
 *             arr_dtype,  # identifier & dtype
 *             pack_uint8(npy_dtype),  # type: ignore
 */
    __pyx_t_10 = PyList_New(5); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1221, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_INCREF(__pyx_v_arr_dtype);
    __Pyx_GIVEREF(__pyx_v_arr_dtype);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_10, 0, __pyx_v_arr_dtype)) __PYX_ERR(0, 1221, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_4);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_10, 1, __pyx_t_4)) __PYX_ERR(0, 1221, __pyx_L1_error);
    __Pyx_INCREF(__pyx_kp_b__6);
    __Pyx_GIVEREF(__pyx_kp_b__6);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_10, 2, __pyx_kp_b__6)) __PYX_ERR(0, 1221, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_5);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_10, 3, __pyx_t_5)) __PYX_ERR(0, 1221, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_10, 4, __pyx_t_1)) __PYX_ERR(0, 1221, __pyx_L1_error);
    __pyx_t_4 = 0;
    __pyx_t_5 = 0;
    __pyx_t_1 = 0;
    __pyx_v_items = ((PyObject*)__pyx_t_10);
    __pyx_t_10 = 0;

    /* "serializor/binary/ser.py":1229
 *         ]
 *         # . serialize values
 *         if s_j != 0:             # <<<<<<<<<<<<<<
 *             ch_size: cython.int = arr.descr.itemsize
 *             items.append(gen_encoded_int(ch_size))  # type: ignore
 */
    __pyx_t_3 = (__pyx_v_s_j != 0);
    if (__pyx_t_3) {

      /* "serializor/binary/ser.py":1230
 *         # . serialize values
 *         if s_j != 0:
 *             ch_size: cython.int = arr.descr.itemsize             # <<<<<<<<<<<<<<
 *             items.append(gen_encoded_int(ch_size))  # type: ignore
 *             for i in range(s_i):
 */
      __pyx_t_10 = ((PyObject *)__pyx_f_5numpy_7ndarray_5descr_descr(__pyx_v_arr)); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1230, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __pyx_v_ch_size = __pyx_f_5numpy_5dtype_8itemsize_itemsize(((PyArray_Descr *)__pyx_t_10));
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

      /* "serializor/binary/ser.py":1231
 *         if s_j != 0:
 *             ch_size: cython.int = arr.descr.itemsize
 *             items.append(gen_encoded_int(ch_size))  # type: ignore             # <<<<<<<<<<<<<<
 *             for i in range(s_i):
 *                 for j in range(s_j):
 */
      __pyx_t_10 = __pyx_f_10serializor_6binary_3ser_gen_encoded_int(__pyx_v_ch_size); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1231, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __pyx_t_6 = __Pyx_PyList_Append(__pyx_v_items, __pyx_t_10); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(0, 1231, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

      /* "serializor/binary/ser.py":1232
 *             ch_size: cython.int = arr.descr.itemsize
 *             items.append(gen_encoded_int(ch_size))  # type: ignore
 *             for i in range(s_i):             # <<<<<<<<<<<<<<
 *                 for j in range(s_j):
 *                     items.append(_ser_bytes(utils.arr_getitem_2d(arr, i, j)))
 */
      __pyx_t_7 = __pyx_v_s_i;
      __pyx_t_8 = __pyx_t_7;
      for (__pyx_t_9 = 0; __pyx_t_9 < __pyx_t_8; __pyx_t_9+=1) {
        __pyx_v_i = __pyx_t_9;

        /* "serializor/binary/ser.py":1233
 *             items.append(gen_encoded_int(ch_size))  # type: ignore
 *             for i in range(s_i):
 *                 for j in range(s_j):             # <<<<<<<<<<<<<<
 *                     items.append(_ser_bytes(utils.arr_getitem_2d(arr, i, j)))
 *         return b"".join(items)
 */
        __pyx_t_11 = __pyx_v_s_j;
        __pyx_t_12 = __pyx_t_11;
        for (__pyx_t_13 = 0; __pyx_t_13 < __pyx_t_12; __pyx_t_13+=1) {
          __pyx_v_j = __pyx_t_13;

          /* "serializor/binary/ser.py":1234
 *             for i in range(s_i):
 *                 for j in range(s_j):
 *                     items.append(_ser_bytes(utils.arr_getitem_2d(arr, i, j)))             # <<<<<<<<<<<<<<
 *         return b"".join(items)
 * 
 */
          __pyx_t_10 = __pyx_f_10serializor_5utils_arr_getitem_2d(__pyx_v_arr, __pyx_v_i, __pyx_v_j); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1234, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_10);
          __pyx_t_1 = __pyx_f_10serializor_6binary_3ser__ser_bytes(__pyx_t_10); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1234, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
          __pyx_t_6 = __Pyx_PyList_Append(__pyx_v_items, __pyx_t_1); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(0, 1234, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        }
      }

      /* "serializor/binary/ser.py":1229
 *         ]
 *         # . serialize values
 *         if s_j != 0:             # <<<<<<<<<<<<<<
 *             ch_size: cython.int = arr.descr.itemsize
 *             items.append(gen_encoded_int(ch_size))  # type: ignore
 */
    }

    /* "serializor/binary/ser.py":1235
 *                 for j in range(s_j):
 *                     items.append(_ser_bytes(utils.arr_getitem_2d(arr, i, j)))
 *         return b"".join(items)             # <<<<<<<<<<<<<<
 * 
 *     # 3-dimension
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_1 = __Pyx_PyBytes_Join(__pyx_kp_b__4, __pyx_v_items); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1235, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!(likely(PyBytes_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("bytes", __pyx_t_1))) __PYX_ERR(0, 1235, __pyx_L1_error)
    __pyx_r = ((PyObject*)__pyx_t_1);
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "serializor/binary/ser.py":1217
 * 
 *     # 2-dimension
 *     if ndim == 2:             # <<<<<<<<<<<<<<
 *         # . array shape
 *         s_i: cython.Py_ssize_t = shape[0]
 */
  }

  /* "serializor/binary/ser.py":1238
 * 
 *     # 3-dimension
 *     if ndim == 3:             # <<<<<<<<<<<<<<
 *         # . array shape
 *         s_i: cython.Py_ssize_t = shape[0]
 */
  __pyx_t_3 = (__pyx_v_ndim == 3);
  if (__pyx_t_3) {

    /* "serializor/binary/ser.py":1240
 *     if ndim == 3:
 *         # . array shape
 *         s_i: cython.Py_ssize_t = shape[0]             # <<<<<<<<<<<<<<
 *         s_j: cython.Py_ssize_t = shape[1]
 *         s_k: cython.Py_ssize_t = shape[2]
 */
    __pyx_v_s_i = (__pyx_v_shape[0]);

    /* "serializor/binary/ser.py":1241
 *         # . array shape
 *         s_i: cython.Py_ssize_t = shape[0]
 *         s_j: cython.Py_ssize_t = shape[1]             # <<<<<<<<<<<<<<
 *         s_k: cython.Py_ssize_t = shape[2]
 *         items = [
 */
    __pyx_v_s_j = (__pyx_v_shape[1]);

    /* "serializor/binary/ser.py":1242
 *         s_i: cython.Py_ssize_t = shape[0]
 *         s_j: cython.Py_ssize_t = shape[1]
 *         s_k: cython.Py_ssize_t = shape[2]             # <<<<<<<<<<<<<<
 *         items = [
 *             arr_dtype,  # identifier & dtype
 */
    __pyx_v_s_k = (__pyx_v_shape[2]);

    /* "serializor/binary/ser.py":1245
 *         items = [
 *             arr_dtype,  # identifier & dtype
 *             pack_uint8(npy_dtype),  # type: ignore             # <<<<<<<<<<<<<<
 *             b"\x03",  # 3-dimension
 *             gen_encoded_int(s_i),  # type: ignore
 */
    __pyx_t_1 = __pyx_f_10serializor_6binary_3ser_pack_uint8(__pyx_v_npy_dtype); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1245, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);

    /* "serializor/binary/ser.py":1247
 *             pack_uint8(npy_dtype),  # type: ignore
 *             b"\x03",  # 3-dimension
 *             gen_encoded_int(s_i),  # type: ignore             # <<<<<<<<<<<<<<
 *             gen_encoded_int(s_j),  # type: ignore
 *             gen_encoded_int(s_k),  # type: ignore
 */
    __pyx_t_10 = __pyx_f_10serializor_6binary_3ser_gen_encoded_int(__pyx_v_s_i); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1247, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);

    /* "serializor/binary/ser.py":1248
 *             b"\x03",  # 3-dimension
 *             gen_encoded_int(s_i),  # type: ignore
 *             gen_encoded_int(s_j),  # type: ignore             # <<<<<<<<<<<<<<
 *             gen_encoded_int(s_k),  # type: ignore
 *         ]
 */
    __pyx_t_5 = __pyx_f_10serializor_6binary_3ser_gen_encoded_int(__pyx_v_s_j); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1248, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);

    /* "serializor/binary/ser.py":1249
 *             gen_encoded_int(s_i),  # type: ignore
 *             gen_encoded_int(s_j),  # type: ignore
 *             gen_encoded_int(s_k),  # type: ignore             # <<<<<<<<<<<<<<
 *         ]
 *         # . serialize values
 */
    __pyx_t_4 = __pyx_f_10serializor_6binary_3ser_gen_encoded_int(__pyx_v_s_k); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1249, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);

    /* "serializor/binary/ser.py":1243
 *         s_j: cython.Py_ssize_t = shape[1]
 *         s_k: cython.Py_ssize_t = shape[2]
 *         items = [             # <<<<<<<<<<<<<<
 *             arr_dtype,  # identifier & dtype
 *             pack_uint8(npy_dtype),  # type: ignore
 */
    __pyx_t_14 = PyList_New(6); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1243, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __Pyx_INCREF(__pyx_v_arr_dtype);
    __Pyx_GIVEREF(__pyx_v_arr_dtype);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_14, 0, __pyx_v_arr_dtype)) __PYX_ERR(0, 1243, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_14, 1, __pyx_t_1)) __PYX_ERR(0, 1243, __pyx_L1_error);
    __Pyx_INCREF(__pyx_kp_b__8);
    __Pyx_GIVEREF(__pyx_kp_b__8);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_14, 2, __pyx_kp_b__8)) __PYX_ERR(0, 1243, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_10);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_14, 3, __pyx_t_10)) __PYX_ERR(0, 1243, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_5);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_14, 4, __pyx_t_5)) __PYX_ERR(0, 1243, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_4);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_14, 5, __pyx_t_4)) __PYX_ERR(0, 1243, __pyx_L1_error);
    __pyx_t_1 = 0;
    __pyx_t_10 = 0;
    __pyx_t_5 = 0;
    __pyx_t_4 = 0;
    __pyx_v_items = ((PyObject*)__pyx_t_14);
    __pyx_t_14 = 0;

    /* "serializor/binary/ser.py":1252
 *         ]
 *         # . serialize values
 *         if s_k != 0:             # <<<<<<<<<<<<<<
 *             ch_size: cython.int = arr.descr.itemsize
 *             items.append(gen_encoded_int(ch_size))  # type: ignore
 */
    __pyx_t_3 = (__pyx_v_s_k != 0);
    if (__pyx_t_3) {

      /* "serializor/binary/ser.py":1253
 *         # . serialize values
 *         if s_k != 0:
 *             ch_size: cython.int = arr.descr.itemsize             # <<<<<<<<<<<<<<
 *             items.append(gen_encoded_int(ch_size))  # type: ignore
 *             for i in range(s_i):
 */
      __pyx_t_14 = ((PyObject *)__pyx_f_5numpy_7ndarray_5descr_descr(__pyx_v_arr)); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1253, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_14);
      __pyx_v_ch_size = __pyx_f_5numpy_5dtype_8itemsize_itemsize(((PyArray_Descr *)__pyx_t_14));
      __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;

      /* "serializor/binary/ser.py":1254
 *         if s_k != 0:
 *             ch_size: cython.int = arr.descr.itemsize
 *             items.append(gen_encoded_int(ch_size))  # type: ignore             # <<<<<<<<<<<<<<
 *             for i in range(s_i):
 *                 for j in range(s_j):
 */
      __pyx_t_14 = __pyx_f_10serializor_6binary_3ser_gen_encoded_int(__pyx_v_ch_size); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1254, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_14);
      __pyx_t_6 = __Pyx_PyList_Append(__pyx_v_items, __pyx_t_14); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(0, 1254, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;

      /* "serializor/binary/ser.py":1255
 *             ch_size: cython.int = arr.descr.itemsize
 *             items.append(gen_encoded_int(ch_size))  # type: ignore
 *             for i in range(s_i):             # <<<<<<<<<<<<<<
 *                 for j in range(s_j):
 *                     for k in range(s_k):
 */
      __pyx_t_7 = __pyx_v_s_i;
      __pyx_t_8 = __pyx_t_7;
      for (__pyx_t_9 = 0; __pyx_t_9 < __pyx_t_8; __pyx_t_9+=1) {
        __pyx_v_i = __pyx_t_9;

        /* "serializor/binary/ser.py":1256
 *             items.append(gen_encoded_int(ch_size))  # type: ignore
 *             for i in range(s_i):
 *                 for j in range(s_j):             # <<<<<<<<<<<<<<
 *                     for k in range(s_k):
 *                         items.append(_ser_bytes(utils.arr_getitem_3d(arr, i, j, k)))
 */
        __pyx_t_11 = __pyx_v_s_j;
        __pyx_t_12 = __pyx_t_11;
        for (__pyx_t_13 = 0; __pyx_t_13 < __pyx_t_12; __pyx_t_13+=1) {
          __pyx_v_j = __pyx_t_13;

          /* "serializor/binary/ser.py":1257
 *             for i in range(s_i):
 *                 for j in range(s_j):
 *                     for k in range(s_k):             # <<<<<<<<<<<<<<
 *                         items.append(_ser_bytes(utils.arr_getitem_3d(arr, i, j, k)))
 *         return b"".join(items)
 */
          __pyx_t_15 = __pyx_v_s_k;
          __pyx_t_16 = __pyx_t_15;
          for (__pyx_t_17 = 0; __pyx_t_17 < __pyx_t_16; __pyx_t_17+=1) {
            __pyx_v_k = __pyx_t_17;

            /* "serializor/binary/ser.py":1258
 *                 for j in range(s_j):
 *                     for k in range(s_k):
 *                         items.append(_ser_bytes(utils.arr_getitem_3d(arr, i, j, k)))             # <<<<<<<<<<<<<<
 *         return b"".join(items)
 * 
 */
            __pyx_t_14 = __pyx_f_10serializor_5utils_arr_getitem_3d(__pyx_v_arr, __pyx_v_i, __pyx_v_j, __pyx_v_k); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1258, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_14);
            __pyx_t_4 = __pyx_f_10serializor_6binary_3ser__ser_bytes(__pyx_t_14); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1258, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_4);
            __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
            __pyx_t_6 = __Pyx_PyList_Append(__pyx_v_items, __pyx_t_4); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(0, 1258, __pyx_L1_error)
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          }
        }
      }

      /* "serializor/binary/ser.py":1252
 *         ]
 *         # . serialize values
 *         if s_k != 0:             # <<<<<<<<<<<<<<
 *             ch_size: cython.int = arr.descr.itemsize
 *             items.append(gen_encoded_int(ch_size))  # type: ignore
 */
    }

    /* "serializor/binary/ser.py":1259
 *                     for k in range(s_k):
 *                         items.append(_ser_bytes(utils.arr_getitem_3d(arr, i, j, k)))
 *         return b"".join(items)             # <<<<<<<<<<<<<<
 * 
 *     # 4-dimension
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyBytes_Join(__pyx_kp_b__4, __pyx_v_items); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1259, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    if (!(likely(PyBytes_CheckExact(__pyx_t_4))||((__pyx_t_4) == Py_None) || __Pyx_RaiseUnexpectedTypeError("bytes", __pyx_t_4))) __PYX_ERR(0, 1259, __pyx_L1_error)
    __pyx_r = ((PyObject*)__pyx_t_4);
    __pyx_t_4 = 0;
    goto __pyx_L0;

    /* "serializor/binary/ser.py":1238
 * 
 *     # 3-dimension
 *     if ndim == 3:             # <<<<<<<<<<<<<<
 *         # . array shape
 *         s_i: cython.Py_ssize_t = shape[0]
 */
  }

  /* "serializor/binary/ser.py":1262
 * 
 *     # 4-dimension
 *     if ndim == 4:             # <<<<<<<<<<<<<<
 *         # . array shape
 *         s_i: cython.Py_ssize_t = shape[0]
 */
  __pyx_t_3 = (__pyx_v_ndim == 4);
  if (__pyx_t_3) {

    /* "serializor/binary/ser.py":1264
 *     if ndim == 4:
 *         # . array shape
 *         s_i: cython.Py_ssize_t = shape[0]             # <<<<<<<<<<<<<<
 *         s_j: cython.Py_ssize_t = shape[1]
 *         s_k: cython.Py_ssize_t = shape[2]
 */
    __pyx_v_s_i = (__pyx_v_shape[0]);

    /* "serializor/binary/ser.py":1265
 *         # . array shape
 *         s_i: cython.Py_ssize_t = shape[0]
 *         s_j: cython.Py_ssize_t = shape[1]             # <<<<<<<<<<<<<<
 *         s_k: cython.Py_ssize_t = shape[2]
 *         s_l: cython.Py_ssize_t = shape[3]
 */
    __pyx_v_s_j = (__pyx_v_shape[1]);

    /* "serializor/binary/ser.py":1266
 *         s_i: cython.Py_ssize_t = shape[0]
 *         s_j: cython.Py_ssize_t = shape[1]
 *         s_k: cython.Py_ssize_t = shape[2]             # <<<<<<<<<<<<<<
 *         s_l: cython.Py_ssize_t = shape[3]
 *         items = [
 */
    __pyx_v_s_k = (__pyx_v_shape[2]);

    /* "serializor/binary/ser.py":1267
 *         s_j: cython.Py_ssize_t = shape[1]
 *         s_k: cython.Py_ssize_t = shape[2]
 *         s_l: cython.Py_ssize_t = shape[3]             # <<<<<<<<<<<<<<
 *         items = [
 *             arr_dtype,  # identifier & dtype
 */
    __pyx_v_s_l = (__pyx_v_shape[3]);

    /* "serializor/binary/ser.py":1270
 *         items = [
 *             arr_dtype,  # identifier & dtype
 *             pack_uint8(npy_dtype),  # type: ignore             # <<<<<<<<<<<<<<
 *             b"\x04",  # 4-dimension
 *             gen_encoded_int(s_i),  # type: ignore
 */
    __pyx_t_4 = __pyx_f_10serializor_6binary_3ser_pack_uint8(__pyx_v_npy_dtype); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1270, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);

    /* "serializor/binary/ser.py":1272
 *             pack_uint8(npy_dtype),  # type: ignore
 *             b"\x04",  # 4-dimension
 *             gen_encoded_int(s_i),  # type: ignore             # <<<<<<<<<<<<<<
 *             gen_encoded_int(s_j),  # type: ignore
 *             gen_encoded_int(s_k),  # type: ignore
 */
    __pyx_t_14 = __pyx_f_10serializor_6binary_3ser_gen_encoded_int(__pyx_v_s_i); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1272, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);

    /* "serializor/binary/ser.py":1273
 *             b"\x04",  # 4-dimension
 *             gen_encoded_int(s_i),  # type: ignore
 *             gen_encoded_int(s_j),  # type: ignore             # <<<<<<<<<<<<<<
 *             gen_encoded_int(s_k),  # type: ignore
 *             gen_encoded_int(s_l),  # type: ignore
 */
    __pyx_t_5 = __pyx_f_10serializor_6binary_3ser_gen_encoded_int(__pyx_v_s_j); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1273, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);

    /* "serializor/binary/ser.py":1274
 *             gen_encoded_int(s_i),  # type: ignore
 *             gen_encoded_int(s_j),  # type: ignore
 *             gen_encoded_int(s_k),  # type: ignore             # <<<<<<<<<<<<<<
 *             gen_encoded_int(s_l),  # type: ignore
 *         ]
 */
    __pyx_t_10 = __pyx_f_10serializor_6binary_3ser_gen_encoded_int(__pyx_v_s_k); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1274, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);

    /* "serializor/binary/ser.py":1275
 *             gen_encoded_int(s_j),  # type: ignore
 *             gen_encoded_int(s_k),  # type: ignore
 *             gen_encoded_int(s_l),  # type: ignore             # <<<<<<<<<<<<<<
 *         ]
 *         # . serialize values
 */
    __pyx_t_1 = __pyx_f_10serializor_6binary_3ser_gen_encoded_int(__pyx_v_s_l); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1275, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);

    /* "serializor/binary/ser.py":1268
 *         s_k: cython.Py_ssize_t = shape[2]
 *         s_l: cython.Py_ssize_t = shape[3]
 *         items = [             # <<<<<<<<<<<<<<
 *             arr_dtype,  # identifier & dtype
 *             pack_uint8(npy_dtype),  # type: ignore
 */
    __pyx_t_18 = PyList_New(7); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 1268, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_18);
    __Pyx_INCREF(__pyx_v_arr_dtype);
    __Pyx_GIVEREF(__pyx_v_arr_dtype);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_18, 0, __pyx_v_arr_dtype)) __PYX_ERR(0, 1268, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_4);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_18, 1, __pyx_t_4)) __PYX_ERR(0, 1268, __pyx_L1_error);
    __Pyx_INCREF(__pyx_kp_b__9);
    __Pyx_GIVEREF(__pyx_kp_b__9);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_18, 2, __pyx_kp_b__9)) __PYX_ERR(0, 1268, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_14);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_18, 3, __pyx_t_14)) __PYX_ERR(0, 1268, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_5);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_18, 4, __pyx_t_5)) __PYX_ERR(0, 1268, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_10);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_18, 5, __pyx_t_10)) __PYX_ERR(0, 1268, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_18, 6, __pyx_t_1)) __PYX_ERR(0, 1268, __pyx_L1_error);
    __pyx_t_4 = 0;
    __pyx_t_14 = 0;
    __pyx_t_5 = 0;
    __pyx_t_10 = 0;
    __pyx_t_1 = 0;
    __pyx_v_items = ((PyObject*)__pyx_t_18);
    __pyx_t_18 = 0;

    /* "serializor/binary/ser.py":1278
 *         ]
 *         # . serialize values
 *         if s_l != 0:             # <<<<<<<<<<<<<<
 *             ch_size: cython.int = arr.descr.itemsize
 *             items.append(gen_encoded_int(ch_size))  # type: ignore
 */
    __pyx_t_3 = (__pyx_v_s_l != 0);
    if (__pyx_t_3) {

      /* "serializor/binary/ser.py":1279
 *         # . serialize values
 *         if s_l != 0:
 *             ch_size: cython.int = arr.descr.itemsize             # <<<<<<<<<<<<<<
 *             items.append(gen_encoded_int(ch_size))  # type: ignore
 *             for i in range(s_i):
 */
      __pyx_t_18 = ((PyObject *)__pyx_f_5numpy_7ndarray_5descr_descr(__pyx_v_arr)); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 1279, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_18);
      __pyx_v_ch_size = __pyx_f_5numpy_5dtype_8itemsize_itemsize(((PyArray_Descr *)__pyx_t_18));
      __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;

      /* "serializor/binary/ser.py":1280
 *         if s_l != 0:
 *             ch_size: cython.int = arr.descr.itemsize
 *             items.append(gen_encoded_int(ch_size))  # type: ignore             # <<<<<<<<<<<<<<
 *             for i in range(s_i):
 *                 for j in range(s_j):
 */
      __pyx_t_18 = __pyx_f_10serializor_6binary_3ser_gen_encoded_int(__pyx_v_ch_size); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 1280, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_18);
      __pyx_t_6 = __Pyx_PyList_Append(__pyx_v_items, __pyx_t_18); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(0, 1280, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;

      /* "serializor/binary/ser.py":1281
 *             ch_size: cython.int = arr.descr.itemsize
 *             items.append(gen_encoded_int(ch_size))  # type: ignore
 *             for i in range(s_i):             # <<<<<<<<<<<<<<
 *                 for j in range(s_j):
 *                     for k in range(s_k):
 */
      __pyx_t_7 = __pyx_v_s_i;
      __pyx_t_8 = __pyx_t_7;
      for (__pyx_t_9 = 0; __pyx_t_9 < __pyx_t_8; __pyx_t_9+=1) {
        __pyx_v_i = __pyx_t_9;

        /* "serializor/binary/ser.py":1282
 *             items.append(gen_encoded_int(ch_size))  # type: ignore
 *             for i in range(s_i):
 *                 for j in range(s_j):             # <<<<<<<<<<<<<<
 *                     for k in range(s_k):
 *                         for l in range(s_l):
 */
        __pyx_t_11 = __pyx_v_s_j;
        __pyx_t_12 = __pyx_t_11;
        for (__pyx_t_13 = 0; __pyx_t_13 < __pyx_t_12; __pyx_t_13+=1) {
          __pyx_v_j = __pyx_t_13;

          /* "serializor/binary/ser.py":1283
 *             for i in range(s_i):
 *                 for j in range(s_j):
 *                     for k in range(s_k):             # <<<<<<<<<<<<<<
 *                         for l in range(s_l):
 *                             items.append(_ser_bytes(utils.arr_getitem_4d(arr, i, j, k, l)))  # type: ignore
 */
          __pyx_t_15 = __pyx_v_s_k;
          __pyx_t_16 = __pyx_t_15;
          for (__pyx_t_17 = 0; __pyx_t_17 < __pyx_t_16; __pyx_t_17+=1) {
            __pyx_v_k = __pyx_t_17;

            /* "serializor/binary/ser.py":1284
 *                 for j in range(s_j):
 *                     for k in range(s_k):
 *                         for l in range(s_l):             # <<<<<<<<<<<<<<
 *                             items.append(_ser_bytes(utils.arr_getitem_4d(arr, i, j, k, l)))  # type: ignore
 *         return b"".join(items)
 */
            __pyx_t_19 = __pyx_v_s_l;
            __pyx_t_20 = __pyx_t_19;
            for (__pyx_t_21 = 0; __pyx_t_21 < __pyx_t_20; __pyx_t_21+=1) {
              __pyx_v_l = __pyx_t_21;

              /* "serializor/binary/ser.py":1285
 *                     for k in range(s_k):
 *                         for l in range(s_l):
 *                             items.append(_ser_bytes(utils.arr_getitem_4d(arr, i, j, k, l)))  # type: ignore             # <<<<<<<<<<<<<<
 *         return b"".join(items)
 * 
 */
              __pyx_t_18 = __pyx_f_10serializor_5utils_arr_getitem_4d(__pyx_v_arr, __pyx_v_i, __pyx_v_j, __pyx_v_k, __pyx_v_l); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 1285, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_18);
              __pyx_t_1 = __pyx_f_10serializor_6binary_3ser__ser_bytes(__pyx_t_18); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1285, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
              __pyx_t_6 = __Pyx_PyList_Append(__pyx_v_items, __pyx_t_1); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(0, 1285, __pyx_L1_error)
              __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            }
          }
        }
      }

      /* "serializor/binary/ser.py":1278
 *         ]
 *         # . serialize values
 *         if s_l != 0:             # <<<<<<<<<<<<<<
 *             ch_size: cython.int = arr.descr.itemsize
 *             items.append(gen_encoded_int(ch_size))  # type: ignore
 */
    }

    /* "serializor/binary/ser.py":1286
 *                         for l in range(s_l):
 *                             items.append(_ser_bytes(utils.arr_getitem_4d(arr, i, j, k, l)))  # type: ignore
 *         return b"".join(items)             # <<<<<<<<<<<<<<
 * 
 *     # invalid
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_1 = __Pyx_PyBytes_Join(__pyx_kp_b__4, __pyx_v_items); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1286, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!(likely(PyBytes_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("bytes", __pyx_t_1))) __PYX_ERR(0, 1286, __pyx_L1_error)
    __pyx_r = ((PyObject*)__pyx_t_1);
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "serializor/binary/ser.py":1262
 * 
 *     # 4-dimension
 *     if ndim == 4:             # <<<<<<<<<<<<<<
 *         # . array shape
 *         s_i: cython.Py_ssize_t = shape[0]
 */
  }

  /* "serializor/binary/ser.py":1289
 * 
 *     # invalid
 *     raise ValueError("unsupported <'numpy.ndarray'> dimension [%d]." % ndim)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_ndim); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1289, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_18 = PyUnicode_Format(__pyx_kp_u_unsupported_numpy_ndarray_dimens, __pyx_t_1); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 1289, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_18);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_18); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1289, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(0, 1289, __pyx_L1_error)

  /* "serializor/binary/ser.py":1176
 * 
 * 
 * @cython.cfunc             # <<<<<<<<<<<<<<
 * @cython.inline(True)
 * def _ser_ndarray_bytes(arr: np.ndarray, arr_dtype: bytes) -> bytes:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_XDECREF(__pyx_t_18);
  __Pyx_AddTraceback("serializor.binary.ser._ser_ndarray_bytes", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_items);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "serializor/binary/ser.py":1292
 * 
 * 
 * @cython.cfunc             # <<<<<<<<<<<<<<
 * @cython.inline(True)
 * def _ser_ndarray_unicode(arr: np.ndarray, arr_dtype: bytes) -> bytes:
 */

static PyObject *__pyx_f_10serializor_6binary_3ser__ser_ndarray_unicode(PyArrayObject *__pyx_v_arr, PyObject *__pyx_v_arr_dtype) {
  int __pyx_v_npy_dtype;
  int __pyx_v_ndim;
  npy_intp *__pyx_v_shape;
  Py_ssize_t __pyx_v_s_i;
  PyObject *__pyx_v_items = NULL;
  int __pyx_v_ch_size;
  PyObject *__pyx_v_values_json = NULL;
  Py_ssize_t __pyx_v_s_j;
  Py_ssize_t __pyx_v_s_k;
  Py_ssize_t __pyx_v_s_l;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  Py_ssize_t __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_ser_ndarray_unicode", 1);

  /* "serializor/binary/ser.py":1311
 *     >>> b'...[SHAPE][values_json_size(ENC)][values_json]'
 *     """
 *     npy_dtype: cython.int = arr.descr.type_num             # <<<<<<<<<<<<<<
 *     ndim: cython.int = arr.ndim
 *     shape = arr.shape
 */
  __pyx_t_1 = ((PyObject *)__pyx_f_5numpy_7ndarray_5descr_descr(__pyx_v_arr)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1311, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = ((PyArray_Descr *)__pyx_t_1)->type_num;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_npy_dtype = __pyx_t_2;

  /* "serializor/binary/ser.py":1312
 *     """
 *     npy_dtype: cython.int = arr.descr.type_num
 *     ndim: cython.int = arr.ndim             # <<<<<<<<<<<<<<
 *     shape = arr.shape
 *     # 1-dimension
 */
  __pyx_v_ndim = __pyx_f_5numpy_7ndarray_4ndim_ndim(__pyx_v_arr);

  /* "serializor/binary/ser.py":1313
 *     npy_dtype: cython.int = arr.descr.type_num
 *     ndim: cython.int = arr.ndim
 *     shape = arr.shape             # <<<<<<<<<<<<<<
 *     # 1-dimension
 *     if ndim == 1:
 */
  __pyx_v_shape = __pyx_f_5numpy_7ndarray_5shape_shape(__pyx_v_arr);

  /* "serializor/binary/ser.py":1315
 *     shape = arr.shape
 *     # 1-dimension
 *     if ndim == 1:             # <<<<<<<<<<<<<<
 *         # . array shape
 *         s_i: cython.Py_ssize_t = shape[0]
 */
  __pyx_t_3 = (__pyx_v_ndim == 1);
  if (__pyx_t_3) {

    /* "serializor/binary/ser.py":1317
 *     if ndim == 1:
 *         # . array shape
 *         s_i: cython.Py_ssize_t = shape[0]             # <<<<<<<<<<<<<<
 *         items = [
 *             arr_dtype,  # identifier & dtype
 */
    __pyx_v_s_i = (__pyx_v_shape[0]);

    /* "serializor/binary/ser.py":1320
 *         items = [
 *             arr_dtype,  # identifier & dtype
 *             pack_uint8(npy_dtype),  # type: ignore             # <<<<<<<<<<<<<<
 *             b"\x01",  # 1-dimension
 *             gen_encoded_int(s_i),  # type: ignore
 */
    __pyx_t_1 = __pyx_f_10serializor_6binary_3ser_pack_uint8(__pyx_v_npy_dtype); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1320, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);

    /* "serializor/binary/ser.py":1322
 *             pack_uint8(npy_dtype),  # type: ignore
 *             b"\x01",  # 1-dimension
 *             gen_encoded_int(s_i),  # type: ignore             # <<<<<<<<<<<<<<
 *         ]
 *         # . serialize values
 */
    __pyx_t_4 = __pyx_f_10serializor_6binary_3ser_gen_encoded_int(__pyx_v_s_i); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1322, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);

    /* "serializor/binary/ser.py":1318
 *         # . array shape
 *         s_i: cython.Py_ssize_t = shape[0]
 *         items = [             # <<<<<<<<<<<<<<
 *             arr_dtype,  # identifier & dtype
 *             pack_uint8(npy_dtype),  # type: ignore
 */
    __pyx_t_5 = PyList_New(4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1318, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_INCREF(__pyx_v_arr_dtype);
    __Pyx_GIVEREF(__pyx_v_arr_dtype);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 0, __pyx_v_arr_dtype)) __PYX_ERR(0, 1318, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 1, __pyx_t_1)) __PYX_ERR(0, 1318, __pyx_L1_error);
    __Pyx_INCREF(__pyx_kp_b__5);
    __Pyx_GIVEREF(__pyx_kp_b__5);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 2, __pyx_kp_b__5)) __PYX_ERR(0, 1318, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_4);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 3, __pyx_t_4)) __PYX_ERR(0, 1318, __pyx_L1_error);
    __pyx_t_1 = 0;
    __pyx_t_4 = 0;
    __pyx_v_items = ((PyObject*)__pyx_t_5);
    __pyx_t_5 = 0;

    /* "serializor/binary/ser.py":1325
 *         ]
 *         # . serialize values
 *         if s_i != 0:             # <<<<<<<<<<<<<<
 *             ch_size: cython.int = int(arr.descr.itemsize / 4)
 *             items.append(gen_encoded_int(ch_size))  # type: ignore
 */
    __pyx_t_3 = (__pyx_v_s_i != 0);
    if (__pyx_t_3) {

      /* "serializor/binary/ser.py":1326
 *         # . serialize values
 *         if s_i != 0:
 *             ch_size: cython.int = int(arr.descr.itemsize / 4)             # <<<<<<<<<<<<<<
 *             items.append(gen_encoded_int(ch_size))  # type: ignore
 *             values_json = _orjson_dumps(np.PyArray_ToList(arr))
 */
      __pyx_t_5 = ((PyObject *)__pyx_f_5numpy_7ndarray_5descr_descr(__pyx_v_arr)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1326, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_v_ch_size = ((int)(((double)__pyx_f_5numpy_5dtype_8itemsize_itemsize(((PyArray_Descr *)__pyx_t_5))) / 4.0));
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

      /* "serializor/binary/ser.py":1327
 *         if s_i != 0:
 *             ch_size: cython.int = int(arr.descr.itemsize / 4)
 *             items.append(gen_encoded_int(ch_size))  # type: ignore             # <<<<<<<<<<<<<<
 *             values_json = _orjson_dumps(np.PyArray_ToList(arr))
 *             items.append(gen_encoded_int(bytes_len(values_json)))  # type: ignore
 */
      __pyx_t_5 = __pyx_f_10serializor_6binary_3ser_gen_encoded_int(__pyx_v_ch_size); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1327, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_6 = __Pyx_PyList_Append(__pyx_v_items, __pyx_t_5); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(0, 1327, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

      /* "serializor/binary/ser.py":1328
 *             ch_size: cython.int = int(arr.descr.itemsize / 4)
 *             items.append(gen_encoded_int(ch_size))  # type: ignore
 *             values_json = _orjson_dumps(np.PyArray_ToList(arr))             # <<<<<<<<<<<<<<
 *             items.append(gen_encoded_int(bytes_len(values_json)))  # type: ignore
 *             items.append(values_json)
 */
      __pyx_t_5 = PyArray_ToList(__pyx_v_arr); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1328, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_4 = __pyx_f_10serializor_6binary_3ser__orjson_dumps(__pyx_t_5); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1328, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_v_values_json = ((PyObject*)__pyx_t_4);
      __pyx_t_4 = 0;

      /* "serializor/binary/ser.py":1329
 *             items.append(gen_encoded_int(ch_size))  # type: ignore
 *             values_json = _orjson_dumps(np.PyArray_ToList(arr))
 *             items.append(gen_encoded_int(bytes_len(values_json)))  # type: ignore             # <<<<<<<<<<<<<<
 *             items.append(values_json)
 *         return b"".join(items)
 */
      __pyx_t_7 = PyBytes_Size(__pyx_v_values_json); if (unlikely(__pyx_t_7 == ((Py_ssize_t)-1L))) __PYX_ERR(0, 1329, __pyx_L1_error)
      __pyx_t_4 = __pyx_f_10serializor_6binary_3ser_gen_encoded_int(__pyx_t_7); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1329, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_6 = __Pyx_PyList_Append(__pyx_v_items, __pyx_t_4); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(0, 1329, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

      /* "serializor/binary/ser.py":1330
 *             values_json = _orjson_dumps(np.PyArray_ToList(arr))
 *             items.append(gen_encoded_int(bytes_len(values_json)))  # type: ignore
 *             items.append(values_json)             # <<<<<<<<<<<<<<
 *         return b"".join(items)
 * 
 */
      __pyx_t_6 = __Pyx_PyList_Append(__pyx_v_items, __pyx_v_values_json); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(0, 1330, __pyx_L1_error)

      /* "serializor/binary/ser.py":1325
 *         ]
 *         # . serialize values
 *         if s_i != 0:             # <<<<<<<<<<<<<<
 *             ch_size: cython.int = int(arr.descr.itemsize / 4)
 *             items.append(gen_encoded_int(ch_size))  # type: ignore
 */
    }

    /* "serializor/binary/ser.py":1331
 *             items.append(gen_encoded_int(bytes_len(values_json)))  # type: ignore
 *             items.append(values_json)
 *         return b"".join(items)             # <<<<<<<<<<<<<<
 * 
 *     # 2-dimension
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyBytes_Join(__pyx_kp_b__4, __pyx_v_items); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1331, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    if (!(likely(PyBytes_CheckExact(__pyx_t_4))||((__pyx_t_4) == Py_None) || __Pyx_RaiseUnexpectedTypeError("bytes", __pyx_t_4))) __PYX_ERR(0, 1331, __pyx_L1_error)
    __pyx_r = ((PyObject*)__pyx_t_4);
    __pyx_t_4 = 0;
    goto __pyx_L0;

    /* "serializor/binary/ser.py":1315
 *     shape = arr.shape
 *     # 1-dimension
 *     if ndim == 1:             # <<<<<<<<<<<<<<
 *         # . array shape
 *         s_i: cython.Py_ssize_t = shape[0]
 */
  }

  /* "serializor/binary/ser.py":1334
 * 
 *     # 2-dimension
 *     if ndim == 2:             # <<<<<<<<<<<<<<
 *         s_i: cython.Py_ssize_t = shape[0]
 *         s_j: cython.Py_ssize_t = shape[1]
 */
  __pyx_t_3 = (__pyx_v_ndim == 2);
  if (__pyx_t_3) {

    /* "serializor/binary/ser.py":1335
 *     # 2-dimension
 *     if ndim == 2:
 *         s_i: cython.Py_ssize_t = shape[0]             # <<<<<<<<<<<<<<
 *         s_j: cython.Py_ssize_t = shape[1]
 *         items = [
 */
    __pyx_v_s_i = (__pyx_v_shape[0]);

    /* "serializor/binary/ser.py":1336
 *     if ndim == 2:
 *         s_i: cython.Py_ssize_t = shape[0]
 *         s_j: cython.Py_ssize_t = shape[1]             # <<<<<<<<<<<<<<
 *         items = [
 *             arr_dtype,  # identifier & dtype
 */
    __pyx_v_s_j = (__pyx_v_shape[1]);

    /* "serializor/binary/ser.py":1339
 *         items = [
 *             arr_dtype,  # identifier & dtype
 *             pack_uint8(npy_dtype),  # type: ignore             # <<<<<<<<<<<<<<
 *             b"\x02",  # 2-dimension
 *             gen_encoded_int(s_i),  # type: ignore
 */
    __pyx_t_4 = __pyx_f_10serializor_6binary_3ser_pack_uint8(__pyx_v_npy_dtype); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1339, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);

    /* "serializor/binary/ser.py":1341
 *             pack_uint8(npy_dtype),  # type: ignore
 *             b"\x02",  # 2-dimension
 *             gen_encoded_int(s_i),  # type: ignore             # <<<<<<<<<<<<<<
 *             gen_encoded_int(s_j),  # type: ignore
 *         ]
 */
    __pyx_t_5 = __pyx_f_10serializor_6binary_3ser_gen_encoded_int(__pyx_v_s_i); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1341, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);

    /* "serializor/binary/ser.py":1342
 *             b"\x02",  # 2-dimension
 *             gen_encoded_int(s_i),  # type: ignore
 *             gen_encoded_int(s_j),  # type: ignore             # <<<<<<<<<<<<<<
 *         ]
 *         # . serialize values
 */
    __pyx_t_1 = __pyx_f_10serializor_6binary_3ser_gen_encoded_int(__pyx_v_s_j); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1342, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);

    /* "serializor/binary/ser.py":1337
 *         s_i: cython.Py_ssize_t = shape[0]
 *         s_j: cython.Py_ssize_t = shape[1]
 *         items = [             # <<<<<<<<<<<<<<
 *             arr_dtype,  # identifier & dtype
 *             pack_uint8(npy_dtype),  # type: ignore
 */
    __pyx_t_8 = PyList_New(5); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1337, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_INCREF(__pyx_v_arr_dtype);
    __Pyx_GIVEREF(__pyx_v_arr_dtype);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_8, 0, __pyx_v_arr_dtype)) __PYX_ERR(0, 1337, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_4);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_8, 1, __pyx_t_4)) __PYX_ERR(0, 1337, __pyx_L1_error);
    __Pyx_INCREF(__pyx_kp_b__6);
    __Pyx_GIVEREF(__pyx_kp_b__6);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_8, 2, __pyx_kp_b__6)) __PYX_ERR(0, 1337, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_5);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_8, 3, __pyx_t_5)) __PYX_ERR(0, 1337, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_8, 4, __pyx_t_1)) __PYX_ERR(0, 1337, __pyx_L1_error);
    __pyx_t_4 = 0;
    __pyx_t_5 = 0;
    __pyx_t_1 = 0;
    __pyx_v_items = ((PyObject*)__pyx_t_8);
    __pyx_t_8 = 0;

    /* "serializor/binary/ser.py":1345
 *         ]
 *         # . serialize values
 *         if s_j != 0:             # <<<<<<<<<<<<<<
 *             ch_size: cython.int = int(arr.descr.itemsize / 4)
 *             items.append(gen_encoded_int(ch_size))  # type: ignore
 */
    __pyx_t_3 = (__pyx_v_s_j != 0);
    if (__pyx_t_3) {

      /* "serializor/binary/ser.py":1346
 *         # . serialize values
 *         if s_j != 0:
 *             ch_size: cython.int = int(arr.descr.itemsize / 4)             # <<<<<<<<<<<<<<
 *             items.append(gen_encoded_int(ch_size))  # type: ignore
 *             values_json = _orjson_dumps(np.PyArray_ToList(utils.arr_flatten(arr)))
 */
      __pyx_t_8 = ((PyObject *)__pyx_f_5numpy_7ndarray_5descr_descr(__pyx_v_arr)); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1346, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_v_ch_size = ((int)(((double)__pyx_f_5numpy_5dtype_8itemsize_itemsize(((PyArray_Descr *)__pyx_t_8))) / 4.0));
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

      /* "serializor/binary/ser.py":1347
 *         if s_j != 0:
 *             ch_size: cython.int = int(arr.descr.itemsize / 4)
 *             items.append(gen_encoded_int(ch_size))  # type: ignore             # <<<<<<<<<<<<<<
 *             values_json = _orjson_dumps(np.PyArray_ToList(utils.arr_flatten(arr)))
 *             items.append(gen_encoded_int(bytes_len(values_json)))  # type: ignore
 */
      __pyx_t_8 = __pyx_f_10serializor_6binary_3ser_gen_encoded_int(__pyx_v_ch_size); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1347, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_6 = __Pyx_PyList_Append(__pyx_v_items, __pyx_t_8); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(0, 1347, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

      /* "serializor/binary/ser.py":1348
 *             ch_size: cython.int = int(arr.descr.itemsize / 4)
 *             items.append(gen_encoded_int(ch_size))  # type: ignore
 *             values_json = _orjson_dumps(np.PyArray_ToList(utils.arr_flatten(arr)))             # <<<<<<<<<<<<<<
 *             items.append(gen_encoded_int(bytes_len(values_json)))  # type: ignore
 *             items.append(values_json)
 */
      __pyx_t_8 = ((PyObject *)__pyx_f_10serializor_5utils_arr_flatten(__pyx_v_arr)); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1348, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_1 = PyArray_ToList(((PyArrayObject *)__pyx_t_8)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1348, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_t_8 = __pyx_f_10serializor_6binary_3ser__orjson_dumps(__pyx_t_1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1348, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_v_values_json = ((PyObject*)__pyx_t_8);
      __pyx_t_8 = 0;

      /* "serializor/binary/ser.py":1349
 *             items.append(gen_encoded_int(ch_size))  # type: ignore
 *             values_json = _orjson_dumps(np.PyArray_ToList(utils.arr_flatten(arr)))
 *             items.append(gen_encoded_int(bytes_len(values_json)))  # type: ignore             # <<<<<<<<<<<<<<
 *             items.append(values_json)
 *         return b"".join(items)
 */
      __pyx_t_7 = PyBytes_Size(__pyx_v_values_json); if (unlikely(__pyx_t_7 == ((Py_ssize_t)-1L))) __PYX_ERR(0, 1349, __pyx_L1_error)
      __pyx_t_8 = __pyx_f_10serializor_6binary_3ser_gen_encoded_int(__pyx_t_7); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1349, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_6 = __Pyx_PyList_Append(__pyx_v_items, __pyx_t_8); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(0, 1349, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

      /* "serializor/binary/ser.py":1350
 *             values_json = _orjson_dumps(np.PyArray_ToList(utils.arr_flatten(arr)))
 *             items.append(gen_encoded_int(bytes_len(values_json)))  # type: ignore
 *             items.append(values_json)             # <<<<<<<<<<<<<<
 *         return b"".join(items)
 * 
 */
      __pyx_t_6 = __Pyx_PyList_Append(__pyx_v_items, __pyx_v_values_json); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(0, 1350, __pyx_L1_error)

      /* "serializor/binary/ser.py":1345
 *         ]
 *         # . serialize values
 *         if s_j != 0:             # <<<<<<<<<<<<<<
 *             ch_size: cython.int = int(arr.descr.itemsize / 4)
 *             items.append(gen_encoded_int(ch_size))  # type: ignore
 */
    }

    /* "serializor/binary/ser.py":1351
 *             items.append(gen_encoded_int(bytes_len(values_json)))  # type: ignore
 *             items.append(values_json)
 *         return b"".join(items)             # <<<<<<<<<<<<<<
 * 
 *     # 3-dimension
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_8 = __Pyx_PyBytes_Join(__pyx_kp_b__4, __pyx_v_items); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1351, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    if (!(likely(PyBytes_CheckExact(__pyx_t_8))||((__pyx_t_8) == Py_None) || __Pyx_RaiseUnexpectedTypeError("bytes", __pyx_t_8))) __PYX_ERR(0, 1351, __pyx_L1_error)
    __pyx_r = ((PyObject*)__pyx_t_8);
    __pyx_t_8 = 0;
    goto __pyx_L0;

    /* "serializor/binary/ser.py":1334
 * 
 *     # 2-dimension
 *     if ndim == 2:             # <<<<<<<<<<<<<<
 *         s_i: cython.Py_ssize_t = shape[0]
 *         s_j: cython.Py_ssize_t = shape[1]
 */
  }

  /* "serializor/binary/ser.py":1354
 * 
 *     # 3-dimension
 *     if ndim == 3:             # <<<<<<<<<<<<<<
 *         s_i: cython.Py_ssize_t = shape[0]
 *         s_j: cython.Py_ssize_t = shape[1]
 */
  __pyx_t_3 = (__pyx_v_ndim == 3);
  if (__pyx_t_3) {

    /* "serializor/binary/ser.py":1355
 *     # 3-dimension
 *     if ndim == 3:
 *         s_i: cython.Py_ssize_t = shape[0]             # <<<<<<<<<<<<<<
 *         s_j: cython.Py_ssize_t = shape[1]
 *         s_k: cython.Py_ssize_t = shape[2]
 */
    __pyx_v_s_i = (__pyx_v_shape[0]);

    /* "serializor/binary/ser.py":1356
 *     if ndim == 3:
 *         s_i: cython.Py_ssize_t = shape[0]
 *         s_j: cython.Py_ssize_t = shape[1]             # <<<<<<<<<<<<<<
 *         s_k: cython.Py_ssize_t = shape[2]
 *         items = [
 */
    __pyx_v_s_j = (__pyx_v_shape[1]);

    /* "serializor/binary/ser.py":1357
 *         s_i: cython.Py_ssize_t = shape[0]
 *         s_j: cython.Py_ssize_t = shape[1]
 *         s_k: cython.Py_ssize_t = shape[2]             # <<<<<<<<<<<<<<
 *         items = [
 *             arr_dtype,  # identifier & dtype
 */
    __pyx_v_s_k = (__pyx_v_shape[2]);

    /* "serializor/binary/ser.py":1360
 *         items = [
 *             arr_dtype,  # identifier & dtype
 *             pack_uint8(npy_dtype),  # type: ignore             # <<<<<<<<<<<<<<
 *             b"\x03",  # 3-dimension
 *             gen_encoded_int(s_i),  # type: ignore
 */
    __pyx_t_8 = __pyx_f_10serializor_6binary_3ser_pack_uint8(__pyx_v_npy_dtype); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1360, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);

    /* "serializor/binary/ser.py":1362
 *             pack_uint8(npy_dtype),  # type: ignore
 *             b"\x03",  # 3-dimension
 *             gen_encoded_int(s_i),  # type: ignore             # <<<<<<<<<<<<<<
 *             gen_encoded_int(s_j),  # type: ignore
 *             gen_encoded_int(s_k),  # type: ignore
 */
    __pyx_t_1 = __pyx_f_10serializor_6binary_3ser_gen_encoded_int(__pyx_v_s_i); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1362, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);

    /* "serializor/binary/ser.py":1363
 *             b"\x03",  # 3-dimension
 *             gen_encoded_int(s_i),  # type: ignore
 *             gen_encoded_int(s_j),  # type: ignore             # <<<<<<<<<<<<<<
 *             gen_encoded_int(s_k),  # type: ignore
 *         ]
 */
    __pyx_t_5 = __pyx_f_10serializor_6binary_3ser_gen_encoded_int(__pyx_v_s_j); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1363, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);

    /* "serializor/binary/ser.py":1364
 *             gen_encoded_int(s_i),  # type: ignore
 *             gen_encoded_int(s_j),  # type: ignore
 *             gen_encoded_int(s_k),  # type: ignore             # <<<<<<<<<<<<<<
 *         ]
 *         # . serialize values
 */
    __pyx_t_4 = __pyx_f_10serializor_6binary_3ser_gen_encoded_int(__pyx_v_s_k); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1364, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);

    /* "serializor/binary/ser.py":1358
 *         s_j: cython.Py_ssize_t = shape[1]
 *         s_k: cython.Py_ssize_t = shape[2]
 *         items = [             # <<<<<<<<<<<<<<
 *             arr_dtype,  # identifier & dtype
 *             pack_uint8(npy_dtype),  # type: ignore
 */
    __pyx_t_9 = PyList_New(6); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1358, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_INCREF(__pyx_v_arr_dtype);
    __Pyx_GIVEREF(__pyx_v_arr_dtype);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_9, 0, __pyx_v_arr_dtype)) __PYX_ERR(0, 1358, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_8);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_9, 1, __pyx_t_8)) __PYX_ERR(0, 1358, __pyx_L1_error);
    __Pyx_INCREF(__pyx_kp_b__8);
    __Pyx_GIVEREF(__pyx_kp_b__8);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_9, 2, __pyx_kp_b__8)) __PYX_ERR(0, 1358, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_9, 3, __pyx_t_1)) __PYX_ERR(0, 1358, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_5);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_9, 4, __pyx_t_5)) __PYX_ERR(0, 1358, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_4);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_9, 5, __pyx_t_4)) __PYX_ERR(0, 1358, __pyx_L1_error);
    __pyx_t_8 = 0;
    __pyx_t_1 = 0;
    __pyx_t_5 = 0;
    __pyx_t_4 = 0;
    __pyx_v_items = ((PyObject*)__pyx_t_9);
    __pyx_t_9 = 0;

    /* "serializor/binary/ser.py":1367
 *         ]
 *         # . serialize values
 *         if s_k != 0:             # <<<<<<<<<<<<<<
 *             ch_size: cython.int = int(arr.descr.itemsize / 4)
 *             items.append(gen_encoded_int(ch_size))  # type: ignore
 */
    __pyx_t_3 = (__pyx_v_s_k != 0);
    if (__pyx_t_3) {

      /* "serializor/binary/ser.py":1368
 *         # . serialize values
 *         if s_k != 0:
 *             ch_size: cython.int = int(arr.descr.itemsize / 4)             # <<<<<<<<<<<<<<
 *             items.append(gen_encoded_int(ch_size))  # type: ignore
 *             values_json = _orjson_dumps(np.PyArray_ToList(utils.arr_flatten(arr)))
 */
      __pyx_t_9 = ((PyObject *)__pyx_f_5numpy_7ndarray_5descr_descr(__pyx_v_arr)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1368, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_v_ch_size = ((int)(((double)__pyx_f_5numpy_5dtype_8itemsize_itemsize(((PyArray_Descr *)__pyx_t_9))) / 4.0));
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

      /* "serializor/binary/ser.py":1369
 *         if s_k != 0:
 *             ch_size: cython.int = int(arr.descr.itemsize / 4)
 *             items.append(gen_encoded_int(ch_size))  # type: ignore             # <<<<<<<<<<<<<<
 *             values_json = _orjson_dumps(np.PyArray_ToList(utils.arr_flatten(arr)))
 *             items.append(gen_encoded_int(bytes_len(values_json)))  # type: ignore
 */
      __pyx_t_9 = __pyx_f_10serializor_6binary_3ser_gen_encoded_int(__pyx_v_ch_size); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1369, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_6 = __Pyx_PyList_Append(__pyx_v_items, __pyx_t_9); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(0, 1369, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

      /* "serializor/binary/ser.py":1370
 *             ch_size: cython.int = int(arr.descr.itemsize / 4)
 *             items.append(gen_encoded_int(ch_size))  # type: ignore
 *             values_json = _orjson_dumps(np.PyArray_ToList(utils.arr_flatten(arr)))             # <<<<<<<<<<<<<<
 *             items.append(gen_encoded_int(bytes_len(values_json)))  # type: ignore
 *             items.append(values_json)
 */
      __pyx_t_9 = ((PyObject *)__pyx_f_10serializor_5utils_arr_flatten(__pyx_v_arr)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1370, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_4 = PyArray_ToList(((PyArrayObject *)__pyx_t_9)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1370, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __pyx_t_9 = __pyx_f_10serializor_6binary_3ser__orjson_dumps(__pyx_t_4); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1370, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_v_values_json = ((PyObject*)__pyx_t_9);
      __pyx_t_9 = 0;

      /* "serializor/binary/ser.py":1371
 *             items.append(gen_encoded_int(ch_size))  # type: ignore
 *             values_json = _orjson_dumps(np.PyArray_ToList(utils.arr_flatten(arr)))
 *             items.append(gen_encoded_int(bytes_len(values_json)))  # type: ignore             # <<<<<<<<<<<<<<
 *             items.append(values_json)
 *         return b"".join(items)
 */
      __pyx_t_7 = PyBytes_Size(__pyx_v_values_json); if (unlikely(__pyx_t_7 == ((Py_ssize_t)-1L))) __PYX_ERR(0, 1371, __pyx_L1_error)
      __pyx_t_9 = __pyx_f_10serializor_6binary_3ser_gen_encoded_int(__pyx_t_7); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1371, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_6 = __Pyx_PyList_Append(__pyx_v_items, __pyx_t_9); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(0, 1371, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

      /* "serializor/binary/ser.py":1372
 *             values_json = _orjson_dumps(np.PyArray_ToList(utils.arr_flatten(arr)))
 *             items.append(gen_encoded_int(bytes_len(values_json)))  # type: ignore
 *             items.append(values_json)             # <<<<<<<<<<<<<<
 *         return b"".join(items)
 * 
 */
      __pyx_t_6 = __Pyx_PyList_Append(__pyx_v_items, __pyx_v_values_json); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(0, 1372, __pyx_L1_error)

      /* "serializor/binary/ser.py":1367
 *         ]
 *         # . serialize values
 *         if s_k != 0:             # <<<<<<<<<<<<<<
 *             ch_size: cython.int = int(arr.descr.itemsize / 4)
 *             items.append(gen_encoded_int(ch_size))  # type: ignore
 */
    }

    /* "serializor/binary/ser.py":1373
 *             items.append(gen_encoded_int(bytes_len(values_json)))  # type: ignore
 *             items.append(values_json)
 *         return b"".join(items)             # <<<<<<<<<<<<<<
 * 
 *     # 4-dimension
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_9 = __Pyx_PyBytes_Join(__pyx_kp_b__4, __pyx_v_items); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1373, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    if (!(likely(PyBytes_CheckExact(__pyx_t_9))||((__pyx_t_9) == Py_None) || __Pyx_RaiseUnexpectedTypeError("bytes", __pyx_t_9))) __PYX_ERR(0, 1373, __pyx_L1_error)
    __pyx_r = ((PyObject*)__pyx_t_9);
    __pyx_t_9 = 0;
    goto __pyx_L0;

    /* "serializor/binary/ser.py":1354
 * 
 *     # 3-dimension
 *     if ndim == 3:             # <<<<<<<<<<<<<<
 *         s_i: cython.Py_ssize_t = shape[0]
 *         s_j: cython.Py_ssize_t = shape[1]
 */
  }

  /* "serializor/binary/ser.py":1376
 * 
 *     # 4-dimension
 *     if ndim == 4:             # <<<<<<<<<<<<<<
 *         s_i: cython.Py_ssize_t = shape[0]
 *         s_j: cython.Py_ssize_t = shape[1]
 */
  __pyx_t_3 = (__pyx_v_ndim == 4);
  if (__pyx_t_3) {

    /* "serializor/binary/ser.py":1377
 *     # 4-dimension
 *     if ndim == 4:
 *         s_i: cython.Py_ssize_t = shape[0]             # <<<<<<<<<<<<<<
 *         s_j: cython.Py_ssize_t = shape[1]
 *         s_k: cython.Py_ssize_t = shape[2]
 */
    __pyx_v_s_i = (__pyx_v_shape[0]);

    /* "serializor/binary/ser.py":1378
 *     if ndim == 4:
 *         s_i: cython.Py_ssize_t = shape[0]
 *         s_j: cython.Py_ssize_t = shape[1]             # <<<<<<<<<<<<<<
 *         s_k: cython.Py_ssize_t = shape[2]
 *         s_l: cython.Py_ssize_t = shape[3]
 */
    __pyx_v_s_j = (__pyx_v_shape[1]);

    /* "serializor/binary/ser.py":1379
 *         s_i: cython.Py_ssize_t = shape[0]
 *         s_j: cython.Py_ssize_t = shape[1]
 *         s_k: cython.Py_ssize_t = shape[2]             # <<<<<<<<<<<<<<
 *         s_l: cython.Py_ssize_t = shape[3]
 *         items = [
 */
    __pyx_v_s_k = (__pyx_v_shape[2]);

    /* "serializor/binary/ser.py":1380
 *         s_j: cython.Py_ssize_t = shape[1]
 *         s_k: cython.Py_ssize_t = shape[2]
 *         s_l: cython.Py_ssize_t = shape[3]             # <<<<<<<<<<<<<<
 *         items = [
 *             arr_dtype,  # identifier & dtype
 */
    __pyx_v_s_l = (__pyx_v_shape[3]);

    /* "serializor/binary/ser.py":1383
 *         items = [
 *             arr_dtype,  # identifier & dtype
 *             pack_uint8(npy_dtype),  # type: ignore             # <<<<<<<<<<<<<<
 *             b"\x04",  # 4-dimension
 *             gen_encoded_int(s_i),  # type: ignore
 */
    __pyx_t_9 = __pyx_f_10serializor_6binary_3ser_pack_uint8(__pyx_v_npy_dtype); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1383, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);

    /* "serializor/binary/ser.py":1385
 *             pack_uint8(npy_dtype),  # type: ignore
 *             b"\x04",  # 4-dimension
 *             gen_encoded_int(s_i),  # type: ignore             # <<<<<<<<<<<<<<
 *             gen_encoded_int(s_j),  # type: ignore
 *             gen_encoded_int(s_k),  # type: ignore
 */
    __pyx_t_4 = __pyx_f_10serializor_6binary_3ser_gen_encoded_int(__pyx_v_s_i); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1385, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);

    /* "serializor/binary/ser.py":1386
 *             b"\x04",  # 4-dimension
 *             gen_encoded_int(s_i),  # type: ignore
 *             gen_encoded_int(s_j),  # type: ignore             # <<<<<<<<<<<<<<
 *             gen_encoded_int(s_k),  # type: ignore
 *             gen_encoded_int(s_l),  # type: ignore
 */
    __pyx_t_5 = __pyx_f_10serializor_6binary_3ser_gen_encoded_int(__pyx_v_s_j); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1386, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);

    /* "serializor/binary/ser.py":1387
 *             gen_encoded_int(s_i),  # type: ignore
 *             gen_encoded_int(s_j),  # type: ignore
 *             gen_encoded_int(s_k),  # type: ignore             # <<<<<<<<<<<<<<
 *             gen_encoded_int(s_l),  # type: ignore
 *         ]
 */
    __pyx_t_1 = __pyx_f_10serializor_6binary_3ser_gen_encoded_int(__pyx_v_s_k); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1387, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);

    /* "serializor/binary/ser.py":1388
 *             gen_encoded_int(s_j),  # type: ignore
 *             gen_encoded_int(s_k),  # type: ignore
 *             gen_encoded_int(s_l),  # type: ignore             # <<<<<<<<<<<<<<
 *         ]
 *         # . serialize values
 */
    __pyx_t_8 = __pyx_f_10serializor_6binary_3ser_gen_encoded_int(__pyx_v_s_l); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1388, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);

    /* "serializor/binary/ser.py":1381
 *         s_k: cython.Py_ssize_t = shape[2]
 *         s_l: cython.Py_ssize_t = shape[3]
 *         items = [             # <<<<<<<<<<<<<<
 *             arr_dtype,  # identifier & dtype
 *             pack_uint8(npy_dtype),  # type: ignore
 */
    __pyx_t_10 = PyList_New(7); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1381, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_INCREF(__pyx_v_arr_dtype);
    __Pyx_GIVEREF(__pyx_v_arr_dtype);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_10, 0, __pyx_v_arr_dtype)) __PYX_ERR(0, 1381, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_9);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_10, 1, __pyx_t_9)) __PYX_ERR(0, 1381, __pyx_L1_error);
    __Pyx_INCREF(__pyx_kp_b__9);
    __Pyx_GIVEREF(__pyx_kp_b__9);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_10, 2, __pyx_kp_b__9)) __PYX_ERR(0, 1381, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_4);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_10, 3, __pyx_t_4)) __PYX_ERR(0, 1381, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_5);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_10, 4, __pyx_t_5)) __PYX_ERR(0, 1381, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_10, 5, __pyx_t_1)) __PYX_ERR(0, 1381, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_8);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_10, 6, __pyx_t_8)) __PYX_ERR(0, 1381, __pyx_L1_error);
    __pyx_t_9 = 0;
    __pyx_t_4 = 0;
    __pyx_t_5 = 0;
    __pyx_t_1 = 0;
    __pyx_t_8 = 0;
    __pyx_v_items = ((PyObject*)__pyx_t_10);
    __pyx_t_10 = 0;

    /* "serializor/binary/ser.py":1391
 *         ]
 *         # . serialize values
 *         if s_l != 0:             # <<<<<<<<<<<<<<
 *             ch_size: cython.int = int(arr.descr.itemsize / 4)
 *             items.append(gen_encoded_int(ch_size))  # type: ignore
 */
    __pyx_t_3 = (__pyx_v_s_l != 0);
    if (__pyx_t_3) {

      /* "serializor/binary/ser.py":1392
 *         # . serialize values
 *         if s_l != 0:
 *             ch_size: cython.int = int(arr.descr.itemsize / 4)             # <<<<<<<<<<<<<<
 *             items.append(gen_encoded_int(ch_size))  # type: ignore
 *             values_json = _orjson_dumps(np.PyArray_ToList(utils.arr_flatten(arr)))
 */
      __pyx_t_10 = ((PyObject *)__pyx_f_5numpy_7ndarray_5descr_descr(__pyx_v_arr)); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1392, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __pyx_v_ch_size = ((int)(((double)__pyx_f_5numpy_5dtype_8itemsize_itemsize(((PyArray_Descr *)__pyx_t_10))) / 4.0));
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

      /* "serializor/binary/ser.py":1393
 *         if s_l != 0:
 *             ch_size: cython.int = int(arr.descr.itemsize / 4)
 *             items.append(gen_encoded_int(ch_size))  # type: ignore             # <<<<<<<<<<<<<<
 *             values_json = _orjson_dumps(np.PyArray_ToList(utils.arr_flatten(arr)))
 *             items.append(gen_encoded_int(bytes_len(values_json)))  # type: ignore
 */
      __pyx_t_10 = __pyx_f_10serializor_6binary_3ser_gen_encoded_int(__pyx_v_ch_size); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1393, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __pyx_t_6 = __Pyx_PyList_Append(__pyx_v_items, __pyx_t_10); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(0, 1393, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

      /* "serializor/binary/ser.py":1394
 *             ch_size: cython.int = int(arr.descr.itemsize / 4)
 *             items.append(gen_encoded_int(ch_size))  # type: ignore
 *             values_json = _orjson_dumps(np.PyArray_ToList(utils.arr_flatten(arr)))             # <<<<<<<<<<<<<<
 *             items.append(gen_encoded_int(bytes_len(values_json)))  # type: ignore
 *             items.append(values_json)
 */
      __pyx_t_10 = ((PyObject *)__pyx_f_10serializor_5utils_arr_flatten(__pyx_v_arr)); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1394, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __pyx_t_8 = PyArray_ToList(((PyArrayObject *)__pyx_t_10)); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1394, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __pyx_t_10 = __pyx_f_10serializor_6binary_3ser__orjson_dumps(__pyx_t_8); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1394, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_v_values_json = ((PyObject*)__pyx_t_10);
      __pyx_t_10 = 0;

      /* "serializor/binary/ser.py":1395
 *             items.append(gen_encoded_int(ch_size))  # type: ignore
 *             values_json = _orjson_dumps(np.PyArray_ToList(utils.arr_flatten(arr)))
 *             items.append(gen_encoded_int(bytes_len(values_json)))  # type: ignore             # <<<<<<<<<<<<<<
 *             items.append(values_json)
 *         return b"".join(items)
 */
      __pyx_t_7 = PyBytes_Size(__pyx_v_values_json); if (unlikely(__pyx_t_7 == ((Py_ssize_t)-1L))) __PYX_ERR(0, 1395, __pyx_L1_error)
      __pyx_t_10 = __pyx_f_10serializor_6binary_3ser_gen_encoded_int(__pyx_t_7); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1395, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __pyx_t_6 = __Pyx_PyList_Append(__pyx_v_items, __pyx_t_10); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(0, 1395, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

      /* "serializor/binary/ser.py":1396
 *             values_json = _orjson_dumps(np.PyArray_ToList(utils.arr_flatten(arr)))
 *             items.append(gen_encoded_int(bytes_len(values_json)))  # type: ignore
 *             items.append(values_json)             # <<<<<<<<<<<<<<
 *         return b"".join(items)
 * 
 */
      __pyx_t_6 = __Pyx_PyList_Append(__pyx_v_items, __pyx_v_values_json); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(0, 1396, __pyx_L1_error)

      /* "serializor/binary/ser.py":1391
 *         ]
 *         # . serialize values
 *         if s_l != 0:             # <<<<<<<<<<<<<<
 *             ch_size: cython.int = int(arr.descr.itemsize / 4)
 *             items.append(gen_encoded_int(ch_size))  # type: ignore
 */
    }

    /* "serializor/binary/ser.py":1397
 *             items.append(gen_encoded_int(bytes_len(values_json)))  # type: ignore
 *             items.append(values_json)
 *         return b"".join(items)             # <<<<<<<<<<<<<<
 * 
 *     # invalid
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyBytes_Join(__pyx_kp_b__4, __pyx_v_items); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1397, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    if (!(likely(PyBytes_CheckExact(__pyx_t_10))||((__pyx_t_10) == Py_None) || __Pyx_RaiseUnexpectedTypeError("bytes", __pyx_t_10))) __PYX_ERR(0, 1397, __pyx_L1_error)
    __pyx_r = ((PyObject*)__pyx_t_10);
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "serializor/binary/ser.py":1376
 * 
 *     # 4-dimension
 *     if ndim == 4:             # <<<<<<<<<<<<<<
 *         s_i: cython.Py_ssize_t = shape[0]
 *         s_j: cython.Py_ssize_t = shape[1]
 */
  }

  /* "serializor/binary/ser.py":1400
 * 
 *     # invalid
 *     raise ValueError("unsupported <'numpy.ndarray'> dimension [%d]." % ndim)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_10 = __Pyx_PyInt_From_int(__pyx_v_ndim); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1400, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __pyx_t_8 = PyUnicode_Format(__pyx_kp_u_unsupported_numpy_ndarray_dimens, __pyx_t_10); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1400, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  __pyx_t_10 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_8); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1400, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __Pyx_Raise(__pyx_t_10, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  __PYX_ERR(0, 1400, __pyx_L1_error)

  /* "serializor/binary/ser.py":1292
 * 
 * 
 * @cython.cfunc             # <<<<<<<<<<<<<<
 * @cython.inline(True)
 * def _ser_ndarray_unicode(arr: np.ndarray, arr_dtype: bytes) -> bytes:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("serializor.binary.ser._ser_ndarray_unicode", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_items);
  __Pyx_XDECREF(__pyx_v_values_json);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "serializor/binary/ser.py":1404
 * 
 * # Pandas Types --------------------------------------------------------------------------------
 * @cython.cfunc             # <<<<<<<<<<<<<<
 * @cython.inline(True)
 * def _ser_pd_timestamp(obj: Timestamp) -> bytes:
 */

static PyObject *__pyx_f_10serializor_6binary_3ser__ser_pd_timestamp(PyObject *__pyx_v_obj) {
  PY_LONG_LONG __pyx_v_val;
  PyObject *__pyx_v_err = NULL;
  PyObject *__pyx_v_name = NULL;
  PyObject *__pyx_v_tz = NULL;
  PyObject *__pyx_v_tz_name = NULL;
  PyObject *__pyx_v_offset = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PY_LONG_LONG __pyx_t_5;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_t_11;
  char const *__pyx_t_12;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  int __pyx_t_19;
  Py_ssize_t __pyx_t_20;
  unsigned int __pyx_t_21;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_ser_pd_timestamp", 1);

  /* "serializor/binary/ser.py":1424
 *     """
 *     # Get timestamp value
 *     try:             # <<<<<<<<<<<<<<
 *         val: cython.longlong = obj.value
 *     except Exception as err:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "serializor/binary/ser.py":1425
 *     # Get timestamp value
 *     try:
 *         val: cython.longlong = obj.value             # <<<<<<<<<<<<<<
 *     except Exception as err:
 *         raise TypeError("expects <'pandas.Timestamp'>, got %s." % type(obj)) from err
 */
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_obj, __pyx_n_s_value); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1425, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_5 = __Pyx_PyInt_As_PY_LONG_LONG(__pyx_t_4); if (unlikely((__pyx_t_5 == (PY_LONG_LONG)-1) && PyErr_Occurred())) __PYX_ERR(0, 1425, __pyx_L3_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_v_val = __pyx_t_5;

      /* "serializor/binary/ser.py":1424
 *     """
 *     # Get timestamp value
 *     try:             # <<<<<<<<<<<<<<
 *         val: cython.longlong = obj.value
 *     except Exception as err:
 */
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "serializor/binary/ser.py":1426
 *     try:
 *         val: cython.longlong = obj.value
 *     except Exception as err:             # <<<<<<<<<<<<<<
 *         raise TypeError("expects <'pandas.Timestamp'>, got %s." % type(obj)) from err
 *     # With timezone
 */
    __pyx_t_6 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
    if (__pyx_t_6) {
      __Pyx_AddTraceback("serializor.binary.ser._ser_pd_timestamp", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_4, &__pyx_t_7, &__pyx_t_8) < 0) __PYX_ERR(0, 1426, __pyx_L5_except_error)
      __Pyx_XGOTREF(__pyx_t_4);
      __Pyx_XGOTREF(__pyx_t_7);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_INCREF(__pyx_t_7);
      __pyx_v_err = __pyx_t_7;
      /*try:*/ {

        /* "serializor/binary/ser.py":1427
 *         val: cython.longlong = obj.value
 *     except Exception as err:
 *         raise TypeError("expects <'pandas.Timestamp'>, got %s." % type(obj)) from err             # <<<<<<<<<<<<<<
 *     # With timezone
 *     if (tz := datetime.datetime_tzinfo(obj)) is not None:
 */
        __pyx_t_9 = __Pyx_PyUnicode_FormatSafe(__pyx_kp_u_expects_pandas_Timestamp_got_s, ((PyObject *)Py_TYPE(__pyx_v_obj))); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1427, __pyx_L14_error)
        __Pyx_GOTREF(__pyx_t_9);
        __pyx_t_10 = __Pyx_PyObject_CallOneArg(__pyx_builtin_TypeError, __pyx_t_9); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1427, __pyx_L14_error)
        __Pyx_GOTREF(__pyx_t_10);
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        __Pyx_Raise(__pyx_t_10, 0, 0, __pyx_v_err);
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        __PYX_ERR(0, 1427, __pyx_L14_error)
      }

      /* "serializor/binary/ser.py":1426
 *     try:
 *         val: cython.longlong = obj.value
 *     except Exception as err:             # <<<<<<<<<<<<<<
 *         raise TypeError("expects <'pandas.Timestamp'>, got %s." % type(obj)) from err
 *     # With timezone
 */
      /*finally:*/ {
        __pyx_L14_error:;
        /*exception exit:*/{
          __Pyx_PyThreadState_declare
          __Pyx_PyThreadState_assign
          __pyx_t_13 = 0; __pyx_t_14 = 0; __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0;
          __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
          __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
          if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_16, &__pyx_t_17, &__pyx_t_18);
          if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_13, &__pyx_t_14, &__pyx_t_15) < 0)) __Pyx_ErrFetch(&__pyx_t_13, &__pyx_t_14, &__pyx_t_15);
          __Pyx_XGOTREF(__pyx_t_13);
          __Pyx_XGOTREF(__pyx_t_14);
          __Pyx_XGOTREF(__pyx_t_15);
          __Pyx_XGOTREF(__pyx_t_16);
          __Pyx_XGOTREF(__pyx_t_17);
          __Pyx_XGOTREF(__pyx_t_18);
          __pyx_t_6 = __pyx_lineno; __pyx_t_11 = __pyx_clineno; __pyx_t_12 = __pyx_filename;
          {
            __Pyx_DECREF(__pyx_v_err); __pyx_v_err = 0;
          }
          if (PY_MAJOR_VERSION >= 3) {
            __Pyx_XGIVEREF(__pyx_t_16);
            __Pyx_XGIVEREF(__pyx_t_17);
            __Pyx_XGIVEREF(__pyx_t_18);
            __Pyx_ExceptionReset(__pyx_t_16, __pyx_t_17, __pyx_t_18);
          }
          __Pyx_XGIVEREF(__pyx_t_13);
          __Pyx_XGIVEREF(__pyx_t_14);
          __Pyx_XGIVEREF(__pyx_t_15);
          __Pyx_ErrRestore(__pyx_t_13, __pyx_t_14, __pyx_t_15);
          __pyx_t_13 = 0; __pyx_t_14 = 0; __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0;
          __pyx_lineno = __pyx_t_6; __pyx_clineno = __pyx_t_11; __pyx_filename = __pyx_t_12;
          goto __pyx_L5_except_error;
        }
      }
    }
    goto __pyx_L5_except_error;

    /* "serializor/binary/ser.py":1424
 *     """
 *     # Get timestamp value
 *     try:             # <<<<<<<<<<<<<<
 *         val: cython.longlong = obj.value
 *     except Exception as err:
 */
    __pyx_L5_except_error:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L8_try_end:;
  }

  /* "serializor/binary/ser.py":1429
 *         raise TypeError("expects <'pandas.Timestamp'>, got %s." % type(obj)) from err
 *     # With timezone
 *     if (tz := datetime.datetime_tzinfo(obj)) is not None:             # <<<<<<<<<<<<<<
 *         # . approach 1: timezone name
 *         if set_contains(utils.AVAILABLE_TIMEZONES, tz_name := str(tz)):
 */
  __pyx_t_8 = __pyx_f_7cpython_8datetime_datetime_tzinfo(__pyx_v_obj); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1429, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_v_tz = __pyx_t_8;
  __Pyx_INCREF(__pyx_t_8);
  __pyx_t_19 = (__pyx_t_8 != Py_None);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (__pyx_t_19) {

    /* "serializor/binary/ser.py":1431
 *     if (tz := datetime.datetime_tzinfo(obj)) is not None:
 *         # . approach 1: timezone name
 *         if set_contains(utils.AVAILABLE_TIMEZONES, tz_name := str(tz)):             # <<<<<<<<<<<<<<
 *             name = utils.encode_str(tz_name, b"ascii")
 *             return b"".join(
 */
    __pyx_t_8 = __pyx_v_10serializor_5utils_AVAILABLE_TIMEZONES;
    __Pyx_INCREF(__pyx_t_8);
    __pyx_t_7 = __Pyx_PyObject_Unicode(__pyx_v_tz); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1431, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_v_tz_name = __pyx_t_7;
    __Pyx_INCREF(__pyx_t_7);
    __pyx_t_4 = __pyx_t_7;
    __Pyx_INCREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_19 = PySet_Contains(__pyx_t_8, __pyx_t_4); if (unlikely(__pyx_t_19 == ((int)-1))) __PYX_ERR(0, 1431, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (__pyx_t_19) {

      /* "serializor/binary/ser.py":1432
 *         # . approach 1: timezone name
 *         if set_contains(utils.AVAILABLE_TIMEZONES, tz_name := str(tz)):
 *             name = utils.encode_str(tz_name, b"ascii")             # <<<<<<<<<<<<<<
 *             return b"".join(
 *                 [
 */
      __pyx_t_4 = __pyx_f_10serializor_5utils_encode_str(__pyx_v_tz_name, ((char *)"ascii")); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1432, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_v_name = ((PyObject*)__pyx_t_4);
      __pyx_t_4 = 0;

      /* "serializor/binary/ser.py":1433
 *         if set_contains(utils.AVAILABLE_TIMEZONES, tz_name := str(tz)):
 *             name = utils.encode_str(tz_name, b"ascii")
 *             return b"".join(             # <<<<<<<<<<<<<<
 *                 [
 *                     PD_TIMESTAMP_ID,  # identifier
 */
      __Pyx_XDECREF(__pyx_r);

      /* "serializor/binary/ser.py":1437
 *                     PD_TIMESTAMP_ID,  # identifier
 *                     b"\x01",  # approach 1
 *                     pack_int64(val),  # type: ignore             # <<<<<<<<<<<<<<
 *                     gen_encoded_int(bytes_len(name)),  # type: ignore
 *                     name,
 */
      __pyx_t_4 = __pyx_f_10serializor_6binary_3ser_pack_int64(__pyx_v_val); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1437, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);

      /* "serializor/binary/ser.py":1438
 *                     b"\x01",  # approach 1
 *                     pack_int64(val),  # type: ignore
 *                     gen_encoded_int(bytes_len(name)),  # type: ignore             # <<<<<<<<<<<<<<
 *                     name,
 *                 ]
 */
      __pyx_t_20 = PyBytes_Size(__pyx_v_name); if (unlikely(__pyx_t_20 == ((Py_ssize_t)-1L))) __PYX_ERR(0, 1438, __pyx_L1_error)
      __pyx_t_8 = __pyx_f_10serializor_6binary_3ser_gen_encoded_int(__pyx_t_20); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1438, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);

      /* "serializor/binary/ser.py":1434
 *             name = utils.encode_str(tz_name, b"ascii")
 *             return b"".join(
 *                 [             # <<<<<<<<<<<<<<
 *                     PD_TIMESTAMP_ID,  # identifier
 *                     b"\x01",  # approach 1
 */
      __pyx_t_7 = PyList_New(5); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1434, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_INCREF(__pyx_v_10serializor_6binary_3ser_PD_TIMESTAMP_ID);
      __Pyx_GIVEREF(__pyx_v_10serializor_6binary_3ser_PD_TIMESTAMP_ID);
      if (__Pyx_PyList_SET_ITEM(__pyx_t_7, 0, __pyx_v_10serializor_6binary_3ser_PD_TIMESTAMP_ID)) __PYX_ERR(0, 1434, __pyx_L1_error);
      __Pyx_INCREF(__pyx_kp_b__5);
      __Pyx_GIVEREF(__pyx_kp_b__5);
      if (__Pyx_PyList_SET_ITEM(__pyx_t_7, 1, __pyx_kp_b__5)) __PYX_ERR(0, 1434, __pyx_L1_error);
      __Pyx_GIVEREF(__pyx_t_4);
      if (__Pyx_PyList_SET_ITEM(__pyx_t_7, 2, __pyx_t_4)) __PYX_ERR(0, 1434, __pyx_L1_error);
      __Pyx_GIVEREF(__pyx_t_8);
      if (__Pyx_PyList_SET_ITEM(__pyx_t_7, 3, __pyx_t_8)) __PYX_ERR(0, 1434, __pyx_L1_error);
      __Pyx_INCREF(__pyx_v_name);
      __Pyx_GIVEREF(__pyx_v_name);
      if (__Pyx_PyList_SET_ITEM(__pyx_t_7, 4, __pyx_v_name)) __PYX_ERR(0, 1434, __pyx_L1_error);
      __pyx_t_4 = 0;
      __pyx_t_8 = 0;

      /* "serializor/binary/ser.py":1433
 *         if set_contains(utils.AVAILABLE_TIMEZONES, tz_name := str(tz)):
 *             name = utils.encode_str(tz_name, b"ascii")
 *             return b"".join(             # <<<<<<<<<<<<<<
 *                 [
 *                     PD_TIMESTAMP_ID,  # identifier
 */
      __pyx_t_8 = __Pyx_PyBytes_Join(__pyx_kp_b__4, __pyx_t_7); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1433, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      if (!(likely(PyBytes_CheckExact(__pyx_t_8))||((__pyx_t_8) == Py_None) || __Pyx_RaiseUnexpectedTypeError("bytes", __pyx_t_8))) __PYX_ERR(0, 1433, __pyx_L1_error)
      __pyx_r = ((PyObject*)__pyx_t_8);
      __pyx_t_8 = 0;
      goto __pyx_L0;

      /* "serializor/binary/ser.py":1431
 *     if (tz := datetime.datetime_tzinfo(obj)) is not None:
 *         # . approach 1: timezone name
 *         if set_contains(utils.AVAILABLE_TIMEZONES, tz_name := str(tz)):             # <<<<<<<<<<<<<<
 *             name = utils.encode_str(tz_name, b"ascii")
 *             return b"".join(
 */
    }

    /* "serializor/binary/ser.py":1443
 *             )
 *         # . approach 2: timezone offset
 *         if (offset := tz.utcoffset(obj)) is not None:             # <<<<<<<<<<<<<<
 *             return b"".join(
 *                 [
 */
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_tz, __pyx_n_s_utcoffset); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1443, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_4 = NULL;
    __pyx_t_21 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_7))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_7);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_7, function);
        __pyx_t_21 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v_obj};
      __pyx_t_8 = __Pyx_PyObject_FastCall(__pyx_t_7, __pyx_callargs+1-__pyx_t_21, 1+__pyx_t_21);
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1443, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    }
    __pyx_v_offset = __pyx_t_8;
    __Pyx_INCREF(__pyx_t_8);
    __pyx_t_19 = (__pyx_t_8 != Py_None);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    if (__pyx_t_19) {

      /* "serializor/binary/ser.py":1444
 *         # . approach 2: timezone offset
 *         if (offset := tz.utcoffset(obj)) is not None:
 *             return b"".join(             # <<<<<<<<<<<<<<
 *                 [
 *                     PD_TIMESTAMP_ID,  # identifier
 */
      __Pyx_XDECREF(__pyx_r);

      /* "serializor/binary/ser.py":1448
 *                     PD_TIMESTAMP_ID,  # identifier
 *                     b"\x02",  # approach 2
 *                     pack_int64(val),  # type: ignore             # <<<<<<<<<<<<<<
 *                     pack_int32(datetime.timedelta_days(offset)),  # type: ignore
 *                     pack_int32(datetime.timedelta_seconds(offset)),  # type: ignore
 */
      __pyx_t_8 = __pyx_f_10serializor_6binary_3ser_pack_int64(__pyx_v_val); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1448, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);

      /* "serializor/binary/ser.py":1449
 *                     b"\x02",  # approach 2
 *                     pack_int64(val),  # type: ignore
 *                     pack_int32(datetime.timedelta_days(offset)),  # type: ignore             # <<<<<<<<<<<<<<
 *                     pack_int32(datetime.timedelta_seconds(offset)),  # type: ignore
 *                     pack_int32(datetime.timedelta_microseconds(offset)),  # type: ignore
 */
      __pyx_t_7 = __pyx_f_10serializor_6binary_3ser_pack_int32(__pyx_f_7cpython_8datetime_timedelta_days(__pyx_v_offset)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1449, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);

      /* "serializor/binary/ser.py":1450
 *                     pack_int64(val),  # type: ignore
 *                     pack_int32(datetime.timedelta_days(offset)),  # type: ignore
 *                     pack_int32(datetime.timedelta_seconds(offset)),  # type: ignore             # <<<<<<<<<<<<<<
 *                     pack_int32(datetime.timedelta_microseconds(offset)),  # type: ignore
 *                 ]
 */
      __pyx_t_4 = __pyx_f_10serializor_6binary_3ser_pack_int32(__pyx_f_7cpython_8datetime_timedelta_seconds(__pyx_v_offset)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1450, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);

      /* "serializor/binary/ser.py":1451
 *                     pack_int32(datetime.timedelta_days(offset)),  # type: ignore
 *                     pack_int32(datetime.timedelta_seconds(offset)),  # type: ignore
 *                     pack_int32(datetime.timedelta_microseconds(offset)),  # type: ignore             # <<<<<<<<<<<<<<
 *                 ]
 *             )
 */
      __pyx_t_10 = __pyx_f_10serializor_6binary_3ser_pack_int32(__pyx_f_7cpython_8datetime_timedelta_microseconds(__pyx_v_offset)); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1451, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);

      /* "serializor/binary/ser.py":1445
 *         if (offset := tz.utcoffset(obj)) is not None:
 *             return b"".join(
 *                 [             # <<<<<<<<<<<<<<
 *                     PD_TIMESTAMP_ID,  # identifier
 *                     b"\x02",  # approach 2
 */
      __pyx_t_9 = PyList_New(6); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1445, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_INCREF(__pyx_v_10serializor_6binary_3ser_PD_TIMESTAMP_ID);
      __Pyx_GIVEREF(__pyx_v_10serializor_6binary_3ser_PD_TIMESTAMP_ID);
      if (__Pyx_PyList_SET_ITEM(__pyx_t_9, 0, __pyx_v_10serializor_6binary_3ser_PD_TIMESTAMP_ID)) __PYX_ERR(0, 1445, __pyx_L1_error);
      __Pyx_INCREF(__pyx_kp_b__6);
      __Pyx_GIVEREF(__pyx_kp_b__6);
      if (__Pyx_PyList_SET_ITEM(__pyx_t_9, 1, __pyx_kp_b__6)) __PYX_ERR(0, 1445, __pyx_L1_error);
      __Pyx_GIVEREF(__pyx_t_8);
      if (__Pyx_PyList_SET_ITEM(__pyx_t_9, 2, __pyx_t_8)) __PYX_ERR(0, 1445, __pyx_L1_error);
      __Pyx_GIVEREF(__pyx_t_7);
      if (__Pyx_PyList_SET_ITEM(__pyx_t_9, 3, __pyx_t_7)) __PYX_ERR(0, 1445, __pyx_L1_error);
      __Pyx_GIVEREF(__pyx_t_4);
      if (__Pyx_PyList_SET_ITEM(__pyx_t_9, 4, __pyx_t_4)) __PYX_ERR(0, 1445, __pyx_L1_error);
      __Pyx_GIVEREF(__pyx_t_10);
      if (__Pyx_PyList_SET_ITEM(__pyx_t_9, 5, __pyx_t_10)) __PYX_ERR(0, 1445, __pyx_L1_error);
      __pyx_t_8 = 0;
      __pyx_t_7 = 0;
      __pyx_t_4 = 0;
      __pyx_t_10 = 0;

      /* "serializor/binary/ser.py":1444
 *         # . approach 2: timezone offset
 *         if (offset := tz.utcoffset(obj)) is not None:
 *             return b"".join(             # <<<<<<<<<<<<<<
 *                 [
 *                     PD_TIMESTAMP_ID,  # identifier
 */
      __pyx_t_10 = __Pyx_PyBytes_Join(__pyx_kp_b__4, __pyx_t_9); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1444, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      if (!(likely(PyBytes_CheckExact(__pyx_t_10))||((__pyx_t_10) == Py_None) || __Pyx_RaiseUnexpectedTypeError("bytes", __pyx_t_10))) __PYX_ERR(0, 1444, __pyx_L1_error)
      __pyx_r = ((PyObject*)__pyx_t_10);
      __pyx_t_10 = 0;
      goto __pyx_L0;

      /* "serializor/binary/ser.py":1443
 *             )
 *         # . approach 2: timezone offset
 *         if (offset := tz.utcoffset(obj)) is not None:             # <<<<<<<<<<<<<<
 *             return b"".join(
 *                 [
 */
    }

    /* "serializor/binary/ser.py":1429
 *         raise TypeError("expects <'pandas.Timestamp'>, got %s." % type(obj)) from err
 *     # With timezone
 *     if (tz := datetime.datetime_tzinfo(obj)) is not None:             # <<<<<<<<<<<<<<
 *         # . approach 1: timezone name
 *         if set_contains(utils.AVAILABLE_TIMEZONES, tz_name := str(tz)):
 */
  }

  /* "serializor/binary/ser.py":1455
 *             )
 *     # Without timezone / Fallback
 *     return b"".join([PD_TIMESTAMP_ID, b"\x00", pack_int64(val)])  # type: ignore             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_10 = __pyx_f_10serializor_6binary_3ser_pack_int64(__pyx_v_val); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1455, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __pyx_t_9 = PyList_New(3); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1455, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_INCREF(__pyx_v_10serializor_6binary_3ser_PD_TIMESTAMP_ID);
  __Pyx_GIVEREF(__pyx_v_10serializor_6binary_3ser_PD_TIMESTAMP_ID);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_9, 0, __pyx_v_10serializor_6binary_3ser_PD_TIMESTAMP_ID)) __PYX_ERR(0, 1455, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_b__7);
  __Pyx_GIVEREF(__pyx_kp_b__7);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_9, 1, __pyx_kp_b__7)) __PYX_ERR(0, 1455, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_10);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_9, 2, __pyx_t_10)) __PYX_ERR(0, 1455, __pyx_L1_error);
  __pyx_t_10 = 0;
  __pyx_t_10 = __Pyx_PyBytes_Join(__pyx_kp_b__4, __pyx_t_9); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1455, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (!(likely(PyBytes_CheckExact(__pyx_t_10))||((__pyx_t_10) == Py_None) || __Pyx_RaiseUnexpectedTypeError("bytes", __pyx_t_10))) __PYX_ERR(0, 1455, __pyx_L1_error)
  __pyx_r = ((PyObject*)__pyx_t_10);
  __pyx_t_10 = 0;
  goto __pyx_L0;

  /* "serializor/binary/ser.py":1404
 * 
 * # Pandas Types --------------------------------------------------------------------------------
 * @cython.cfunc             # <<<<<<<<<<<<<<
 * @cython.inline(True)
 * def _ser_pd_timestamp(obj: Timestamp) -> bytes:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("serializor.binary.ser._ser_pd_timestamp", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_err);
  __Pyx_XDECREF(__pyx_v_name);
  __Pyx_XDECREF(__pyx_v_tz);
  __Pyx_XDECREF(__pyx_v_tz_name);
  __Pyx_XDECREF(__pyx_v_offset);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "serializor/binary/ser.py":1458
 * 
 * 
 * @cython.cfunc             # <<<<<<<<<<<<<<
 * @cython.inline(True)
 * def _ser_pd_timedelta(obj: Timedelta) -> bytes:
 */

static PyObject *__pyx_f_10serializor_6binary_3ser__ser_pd_timedelta(PyObject *__pyx_v_obj) {
  PY_LONG_LONG __pyx_v_val;
  PyObject *__pyx_v_err = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PY_LONG_LONG __pyx_t_5;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_t_11;
  char const *__pyx_t_12;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_ser_pd_timedelta", 1);

  /* "serializor/binary/ser.py":1466
 *     >>> b'[id(1B)][VALUE(8B)]'
 *     """
 *     try:             # <<<<<<<<<<<<<<
 *         val: cython.longlong = obj.value
 *     except Exception as err:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "serializor/binary/ser.py":1467
 *     """
 *     try:
 *         val: cython.longlong = obj.value             # <<<<<<<<<<<<<<
 *     except Exception as err:
 *         raise TypeError("expects <'pandas.Timedelta'>, got %s." % type(obj)) from err
 */
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_obj, __pyx_n_s_value); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1467, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_5 = __Pyx_PyInt_As_PY_LONG_LONG(__pyx_t_4); if (unlikely((__pyx_t_5 == (PY_LONG_LONG)-1) && PyErr_Occurred())) __PYX_ERR(0, 1467, __pyx_L3_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_v_val = __pyx_t_5;

      /* "serializor/binary/ser.py":1466
 *     >>> b'[id(1B)][VALUE(8B)]'
 *     """
 *     try:             # <<<<<<<<<<<<<<
 *         val: cython.longlong = obj.value
 *     except Exception as err:
 */
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "serializor/binary/ser.py":1468
 *     try:
 *         val: cython.longlong = obj.value
 *     except Exception as err:             # <<<<<<<<<<<<<<
 *         raise TypeError("expects <'pandas.Timedelta'>, got %s." % type(obj)) from err
 *     return PD_TIMEDELTA_ID + pack_int64(val)  # type: ignore
 */
    __pyx_t_6 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
    if (__pyx_t_6) {
      __Pyx_AddTraceback("serializor.binary.ser._ser_pd_timedelta", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_4, &__pyx_t_7, &__pyx_t_8) < 0) __PYX_ERR(0, 1468, __pyx_L5_except_error)
      __Pyx_XGOTREF(__pyx_t_4);
      __Pyx_XGOTREF(__pyx_t_7);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_INCREF(__pyx_t_7);
      __pyx_v_err = __pyx_t_7;
      /*try:*/ {

        /* "serializor/binary/ser.py":1469
 *         val: cython.longlong = obj.value
 *     except Exception as err:
 *         raise TypeError("expects <'pandas.Timedelta'>, got %s." % type(obj)) from err             # <<<<<<<<<<<<<<
 *     return PD_TIMEDELTA_ID + pack_int64(val)  # type: ignore
 * 
 */
        __pyx_t_9 = __Pyx_PyUnicode_FormatSafe(__pyx_kp_u_expects_pandas_Timedelta_got_s, ((PyObject *)Py_TYPE(__pyx_v_obj))); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1469, __pyx_L14_error)
        __Pyx_GOTREF(__pyx_t_9);
        __pyx_t_10 = __Pyx_PyObject_CallOneArg(__pyx_builtin_TypeError, __pyx_t_9); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1469, __pyx_L14_error)
        __Pyx_GOTREF(__pyx_t_10);
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        __Pyx_Raise(__pyx_t_10, 0, 0, __pyx_v_err);
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        __PYX_ERR(0, 1469, __pyx_L14_error)
      }

      /* "serializor/binary/ser.py":1468
 *     try:
 *         val: cython.longlong = obj.value
 *     except Exception as err:             # <<<<<<<<<<<<<<
 *         raise TypeError("expects <'pandas.Timedelta'>, got %s." % type(obj)) from err
 *     return PD_TIMEDELTA_ID + pack_int64(val)  # type: ignore
 */
      /*finally:*/ {
        __pyx_L14_error:;
        /*exception exit:*/{
          __Pyx_PyThreadState_declare
          __Pyx_PyThreadState_assign
          __pyx_t_13 = 0; __pyx_t_14 = 0; __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0;
          __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
          __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
          if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_16, &__pyx_t_17, &__pyx_t_18);
          if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_13, &__pyx_t_14, &__pyx_t_15) < 0)) __Pyx_ErrFetch(&__pyx_t_13, &__pyx_t_14, &__pyx_t_15);
          __Pyx_XGOTREF(__pyx_t_13);
          __Pyx_XGOTREF(__pyx_t_14);
          __Pyx_XGOTREF(__pyx_t_15);
          __Pyx_XGOTREF(__pyx_t_16);
          __Pyx_XGOTREF(__pyx_t_17);
          __Pyx_XGOTREF(__pyx_t_18);
          __pyx_t_6 = __pyx_lineno; __pyx_t_11 = __pyx_clineno; __pyx_t_12 = __pyx_filename;
          {
            __Pyx_DECREF(__pyx_v_err); __pyx_v_err = 0;
          }
          if (PY_MAJOR_VERSION >= 3) {
            __Pyx_XGIVEREF(__pyx_t_16);
            __Pyx_XGIVEREF(__pyx_t_17);
            __Pyx_XGIVEREF(__pyx_t_18);
            __Pyx_ExceptionReset(__pyx_t_16, __pyx_t_17, __pyx_t_18);
          }
          __Pyx_XGIVEREF(__pyx_t_13);
          __Pyx_XGIVEREF(__pyx_t_14);
          __Pyx_XGIVEREF(__pyx_t_15);
          __Pyx_ErrRestore(__pyx_t_13, __pyx_t_14, __pyx_t_15);
          __pyx_t_13 = 0; __pyx_t_14 = 0; __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0;
          __pyx_lineno = __pyx_t_6; __pyx_clineno = __pyx_t_11; __pyx_filename = __pyx_t_12;
          goto __pyx_L5_except_error;
        }
      }
    }
    goto __pyx_L5_except_error;

    /* "serializor/binary/ser.py":1466
 *     >>> b'[id(1B)][VALUE(8B)]'
 *     """
 *     try:             # <<<<<<<<<<<<<<
 *         val: cython.longlong = obj.value
 *     except Exception as err:
 */
    __pyx_L5_except_error:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L8_try_end:;
  }

  /* "serializor/binary/ser.py":1470
 *     except Exception as err:
 *         raise TypeError("expects <'pandas.Timedelta'>, got %s." % type(obj)) from err
 *     return PD_TIMEDELTA_ID + pack_int64(val)  # type: ignore             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_8 = __pyx_f_10serializor_6binary_3ser_pack_int64(__pyx_v_val); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1470, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_7 = PyNumber_Add(__pyx_v_10serializor_6binary_3ser_PD_TIMEDELTA_ID, __pyx_t_8); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1470, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_r = ((PyObject*)__pyx_t_7);
  __pyx_t_7 = 0;
  goto __pyx_L0;

  /* "serializor/binary/ser.py":1458
 * 
 * 
 * @cython.cfunc             # <<<<<<<<<<<<<<
 * @cython.inline(True)
 * def _ser_pd_timedelta(obj: Timedelta) -> bytes:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("serializor.binary.ser._ser_pd_timedelta", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_err);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "serializor/binary/ser.py":1473
 * 
 * 
 * @cython.cfunc             # <<<<<<<<<<<<<<
 * @cython.inline(True)
 * def _ser_series(obj: Series) -> bytes:
 */

static PyObject *__pyx_f_10serializor_6binary_3ser__ser_series(PyObject *__pyx_v_obj) {
  PyArrayObject *__pyx_v_arr = 0;
  PyObject *__pyx_v_name = 0;
  PyObject *__pyx_v_err = NULL;
  PyObject *__pyx_v_items = NULL;
  PyObject *__pyx_v_name_value = NULL;
  char __pyx_v_npy_kind;
  PyObject *__pyx_v_values = NULL;
  PyObject *__pyx_v_tz = NULL;
  PyObject *__pyx_v_tz_name = NULL;
  PyObject *__pyx_v_offset = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_t_10;
  char const *__pyx_t_11;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  int __pyx_t_18;
  char const *__pyx_t_19;
  char __pyx_t_20;
  int __pyx_t_21;
  Py_ssize_t __pyx_t_22;
  unsigned int __pyx_t_23;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_ser_series", 1);

  /* "serializor/binary/ser.py":1498
 *     """
 *     # Access values and name
 *     try:             # <<<<<<<<<<<<<<
 *         arr: np.ndarray = obj.values
 *         name: object = obj.name
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "serializor/binary/ser.py":1499
 *     # Access values and name
 *     try:
 *         arr: np.ndarray = obj.values             # <<<<<<<<<<<<<<
 *         name: object = obj.name
 *     except Exception as err:
 */
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_obj, __pyx_n_s_values); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1499, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_4);
      if (!(likely(((__pyx_t_4) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_4, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 1499, __pyx_L3_error)
      __pyx_v_arr = ((PyArrayObject *)__pyx_t_4);
      __pyx_t_4 = 0;

      /* "serializor/binary/ser.py":1500
 *     try:
 *         arr: np.ndarray = obj.values
 *         name: object = obj.name             # <<<<<<<<<<<<<<
 *     except Exception as err:
 *         raise TypeError("expects <'pandas.Series'>, got %s." % type(obj)) from err
 */
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_obj, __pyx_n_s_name); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1500, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_v_name = __pyx_t_4;
      __pyx_t_4 = 0;

      /* "serializor/binary/ser.py":1498
 *     """
 *     # Access values and name
 *     try:             # <<<<<<<<<<<<<<
 *         arr: np.ndarray = obj.values
 *         name: object = obj.name
 */
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "serializor/binary/ser.py":1501
 *         arr: np.ndarray = obj.values
 *         name: object = obj.name
 *     except Exception as err:             # <<<<<<<<<<<<<<
 *         raise TypeError("expects <'pandas.Series'>, got %s." % type(obj)) from err
 * 
 */
    __pyx_t_5 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
    if (__pyx_t_5) {
      __Pyx_AddTraceback("serializor.binary.ser._ser_series", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_4, &__pyx_t_6, &__pyx_t_7) < 0) __PYX_ERR(0, 1501, __pyx_L5_except_error)
      __Pyx_XGOTREF(__pyx_t_4);
      __Pyx_XGOTREF(__pyx_t_6);
      __Pyx_XGOTREF(__pyx_t_7);
      __Pyx_INCREF(__pyx_t_6);
      __pyx_v_err = __pyx_t_6;
      /*try:*/ {

        /* "serializor/binary/ser.py":1502
 *         name: object = obj.name
 *     except Exception as err:
 *         raise TypeError("expects <'pandas.Series'>, got %s." % type(obj)) from err             # <<<<<<<<<<<<<<
 * 
 *     # Serialize name
 */
        __pyx_t_8 = __Pyx_PyUnicode_FormatSafe(__pyx_kp_u_expects_pandas_Series_got_s, ((PyObject *)Py_TYPE(__pyx_v_obj))); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1502, __pyx_L14_error)
        __Pyx_GOTREF(__pyx_t_8);
        __pyx_t_9 = __Pyx_PyObject_CallOneArg(__pyx_builtin_TypeError, __pyx_t_8); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1502, __pyx_L14_error)
        __Pyx_GOTREF(__pyx_t_9);
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_Raise(__pyx_t_9, 0, 0, __pyx_v_err);
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        __PYX_ERR(0, 1502, __pyx_L14_error)
      }

      /* "serializor/binary/ser.py":1501
 *         arr: np.ndarray = obj.values
 *         name: object = obj.name
 *     except Exception as err:             # <<<<<<<<<<<<<<
 *         raise TypeError("expects <'pandas.Series'>, got %s." % type(obj)) from err
 * 
 */
      /*finally:*/ {
        __pyx_L14_error:;
        /*exception exit:*/{
          __Pyx_PyThreadState_declare
          __Pyx_PyThreadState_assign
          __pyx_t_12 = 0; __pyx_t_13 = 0; __pyx_t_14 = 0; __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0;
          __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
          __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
          if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_15, &__pyx_t_16, &__pyx_t_17);
          if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_12, &__pyx_t_13, &__pyx_t_14) < 0)) __Pyx_ErrFetch(&__pyx_t_12, &__pyx_t_13, &__pyx_t_14);
          __Pyx_XGOTREF(__pyx_t_12);
          __Pyx_XGOTREF(__pyx_t_13);
          __Pyx_XGOTREF(__pyx_t_14);
          __Pyx_XGOTREF(__pyx_t_15);
          __Pyx_XGOTREF(__pyx_t_16);
          __Pyx_XGOTREF(__pyx_t_17);
          __pyx_t_5 = __pyx_lineno; __pyx_t_10 = __pyx_clineno; __pyx_t_11 = __pyx_filename;
          {
            __Pyx_DECREF(__pyx_v_err); __pyx_v_err = 0;
          }
          if (PY_MAJOR_VERSION >= 3) {
            __Pyx_XGIVEREF(__pyx_t_15);
            __Pyx_XGIVEREF(__pyx_t_16);
            __Pyx_XGIVEREF(__pyx_t_17);
            __Pyx_ExceptionReset(__pyx_t_15, __pyx_t_16, __pyx_t_17);
          }
          __Pyx_XGIVEREF(__pyx_t_12);
          __Pyx_XGIVEREF(__pyx_t_13);
          __Pyx_XGIVEREF(__pyx_t_14);
          __Pyx_ErrRestore(__pyx_t_12, __pyx_t_13, __pyx_t_14);
          __pyx_t_12 = 0; __pyx_t_13 = 0; __pyx_t_14 = 0; __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0;
          __pyx_lineno = __pyx_t_5; __pyx_clineno = __pyx_t_10; __pyx_filename = __pyx_t_11;
          goto __pyx_L5_except_error;
        }
      }
    }
    goto __pyx_L5_except_error;

    /* "serializor/binary/ser.py":1498
 *     """
 *     # Access values and name
 *     try:             # <<<<<<<<<<<<<<
 *         arr: np.ndarray = obj.values
 *         name: object = obj.name
 */
    __pyx_L5_except_error:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L8_try_end:;
  }

  /* "serializor/binary/ser.py":1505
 * 
 *     # Serialize name
 *     if name is None:  # no name             # <<<<<<<<<<<<<<
 *         items = [SERIES_ID, b"\x00"]
 *     else:  # has name
 */
  __pyx_t_18 = (__pyx_v_name == Py_None);
  if (__pyx_t_18) {

    /* "serializor/binary/ser.py":1506
 *     # Serialize name
 *     if name is None:  # no name
 *         items = [SERIES_ID, b"\x00"]             # <<<<<<<<<<<<<<
 *     else:  # has name
 *         try:
 */
    __pyx_t_7 = PyList_New(2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1506, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_INCREF(__pyx_v_10serializor_6binary_3ser_SERIES_ID);
    __Pyx_GIVEREF(__pyx_v_10serializor_6binary_3ser_SERIES_ID);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_7, 0, __pyx_v_10serializor_6binary_3ser_SERIES_ID)) __PYX_ERR(0, 1506, __pyx_L1_error);
    __Pyx_INCREF(__pyx_kp_b__7);
    __Pyx_GIVEREF(__pyx_kp_b__7);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_7, 1, __pyx_kp_b__7)) __PYX_ERR(0, 1506, __pyx_L1_error);
    __pyx_v_items = ((PyObject*)__pyx_t_7);
    __pyx_t_7 = 0;

    /* "serializor/binary/ser.py":1505
 * 
 *     # Serialize name
 *     if name is None:  # no name             # <<<<<<<<<<<<<<
 *         items = [SERIES_ID, b"\x00"]
 *     else:  # has name
 */
    goto __pyx_L20;
  }

  /* "serializor/binary/ser.py":1508
 *         items = [SERIES_ID, b"\x00"]
 *     else:  # has name
 *         try:             # <<<<<<<<<<<<<<
 *             name_value = _ser_common(name)
 *         except Exception as err:
 */
  /*else*/ {
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_3, &__pyx_t_2, &__pyx_t_1);
      __Pyx_XGOTREF(__pyx_t_3);
      __Pyx_XGOTREF(__pyx_t_2);
      __Pyx_XGOTREF(__pyx_t_1);
      /*try:*/ {

        /* "serializor/binary/ser.py":1509
 *     else:  # has name
 *         try:
 *             name_value = _ser_common(name)             # <<<<<<<<<<<<<<
 *         except Exception as err:
 *             raise ValueError("invalid <'pandas.Series'> name %r." % name) from err
 */
        __pyx_t_7 = __pyx_f_10serializor_6binary_3ser__ser_common(__pyx_v_name); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1509, __pyx_L21_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_v_name_value = ((PyObject*)__pyx_t_7);
        __pyx_t_7 = 0;

        /* "serializor/binary/ser.py":1508
 *         items = [SERIES_ID, b"\x00"]
 *     else:  # has name
 *         try:             # <<<<<<<<<<<<<<
 *             name_value = _ser_common(name)
 *         except Exception as err:
 */
      }
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L26_try_end;
      __pyx_L21_error:;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;

      /* "serializor/binary/ser.py":1510
 *         try:
 *             name_value = _ser_common(name)
 *         except Exception as err:             # <<<<<<<<<<<<<<
 *             raise ValueError("invalid <'pandas.Series'> name %r." % name) from err
 *         items = [SERIES_ID, b"\x01", name_value]
 */
      __pyx_t_10 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
      if (__pyx_t_10) {
        __Pyx_AddTraceback("serializor.binary.ser._ser_series", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_7, &__pyx_t_6, &__pyx_t_4) < 0) __PYX_ERR(0, 1510, __pyx_L23_except_error)
        __Pyx_XGOTREF(__pyx_t_7);
        __Pyx_XGOTREF(__pyx_t_6);
        __Pyx_XGOTREF(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_6);
        __pyx_v_err = __pyx_t_6;
        /*try:*/ {

          /* "serializor/binary/ser.py":1511
 *             name_value = _ser_common(name)
 *         except Exception as err:
 *             raise ValueError("invalid <'pandas.Series'> name %r." % name) from err             # <<<<<<<<<<<<<<
 *         items = [SERIES_ID, b"\x01", name_value]
 * 
 */
          __pyx_t_9 = __Pyx_PyUnicode_FormatSafe(__pyx_kp_u_invalid_pandas_Series_name_r, __pyx_v_name); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1511, __pyx_L32_error)
          __Pyx_GOTREF(__pyx_t_9);
          __pyx_t_8 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_9); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1511, __pyx_L32_error)
          __Pyx_GOTREF(__pyx_t_8);
          __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
          __Pyx_Raise(__pyx_t_8, 0, 0, __pyx_v_err);
          __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
          __PYX_ERR(0, 1511, __pyx_L32_error)
        }

        /* "serializor/binary/ser.py":1510
 *         try:
 *             name_value = _ser_common(name)
 *         except Exception as err:             # <<<<<<<<<<<<<<
 *             raise ValueError("invalid <'pandas.Series'> name %r." % name) from err
 *         items = [SERIES_ID, b"\x01", name_value]
 */
        /*finally:*/ {
          __pyx_L32_error:;
          /*exception exit:*/{
            __Pyx_PyThreadState_declare
            __Pyx_PyThreadState_assign
            __pyx_t_17 = 0; __pyx_t_16 = 0; __pyx_t_15 = 0; __pyx_t_14 = 0; __pyx_t_13 = 0; __pyx_t_12 = 0;
            __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
            __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
            if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_14, &__pyx_t_13, &__pyx_t_12);
            if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_17, &__pyx_t_16, &__pyx_t_15) < 0)) __Pyx_ErrFetch(&__pyx_t_17, &__pyx_t_16, &__pyx_t_15);
            __Pyx_XGOTREF(__pyx_t_17);
            __Pyx_XGOTREF(__pyx_t_16);
            __Pyx_XGOTREF(__pyx_t_15);
            __Pyx_XGOTREF(__pyx_t_14);
            __Pyx_XGOTREF(__pyx_t_13);
            __Pyx_XGOTREF(__pyx_t_12);
            __pyx_t_10 = __pyx_lineno; __pyx_t_5 = __pyx_clineno; __pyx_t_19 = __pyx_filename;
            {
              __Pyx_DECREF(__pyx_v_err); __pyx_v_err = 0;
            }
            if (PY_MAJOR_VERSION >= 3) {
              __Pyx_XGIVEREF(__pyx_t_14);
              __Pyx_XGIVEREF(__pyx_t_13);
              __Pyx_XGIVEREF(__pyx_t_12);
              __Pyx_ExceptionReset(__pyx_t_14, __pyx_t_13, __pyx_t_12);
            }
            __Pyx_XGIVEREF(__pyx_t_17);
            __Pyx_XGIVEREF(__pyx_t_16);
            __Pyx_XGIVEREF(__pyx_t_15);
            __Pyx_ErrRestore(__pyx_t_17, __pyx_t_16, __pyx_t_15);
            __pyx_t_17 = 0; __pyx_t_16 = 0; __pyx_t_15 = 0; __pyx_t_14 = 0; __pyx_t_13 = 0; __pyx_t_12 = 0;
            __pyx_lineno = __pyx_t_10; __pyx_clineno = __pyx_t_5; __pyx_filename = __pyx_t_19;
            goto __pyx_L23_except_error;
          }
        }
      }
      goto __pyx_L23_except_error;

      /* "serializor/binary/ser.py":1508
 *         items = [SERIES_ID, b"\x00"]
 *     else:  # has name
 *         try:             # <<<<<<<<<<<<<<
 *             name_value = _ser_common(name)
 *         except Exception as err:
 */
      __pyx_L23_except_error:;
      __Pyx_XGIVEREF(__pyx_t_3);
      __Pyx_XGIVEREF(__pyx_t_2);
      __Pyx_XGIVEREF(__pyx_t_1);
      __Pyx_ExceptionReset(__pyx_t_3, __pyx_t_2, __pyx_t_1);
      goto __pyx_L1_error;
      __pyx_L26_try_end:;
    }

    /* "serializor/binary/ser.py":1512
 *         except Exception as err:
 *             raise ValueError("invalid <'pandas.Series'> name %r." % name) from err
 *         items = [SERIES_ID, b"\x01", name_value]             # <<<<<<<<<<<<<<
 * 
 *     # Serialize values
 */
    __pyx_t_4 = PyList_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1512, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_INCREF(__pyx_v_10serializor_6binary_3ser_SERIES_ID);
    __Pyx_GIVEREF(__pyx_v_10serializor_6binary_3ser_SERIES_ID);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_4, 0, __pyx_v_10serializor_6binary_3ser_SERIES_ID)) __PYX_ERR(0, 1512, __pyx_L1_error);
    __Pyx_INCREF(__pyx_kp_b__5);
    __Pyx_GIVEREF(__pyx_kp_b__5);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_4, 1, __pyx_kp_b__5)) __PYX_ERR(0, 1512, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_name_value);
    __Pyx_GIVEREF(__pyx_v_name_value);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_4, 2, __pyx_v_name_value)) __PYX_ERR(0, 1512, __pyx_L1_error);
    __pyx_v_items = ((PyObject*)__pyx_t_4);
    __pyx_t_4 = 0;
  }
  __pyx_L20:;

  /* "serializor/binary/ser.py":1515
 * 
 *     # Serialize values
 *     npy_kind: cython.char = arr.descr.kind             # <<<<<<<<<<<<<<
 *     # . ndarray[object]
 *     if npy_kind == identifier.NDARRAY_OBJECT:
 */
  __pyx_t_4 = ((PyObject *)__pyx_f_5numpy_7ndarray_5descr_descr(__pyx_v_arr)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1515, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_20 = ((PyArray_Descr *)__pyx_t_4)->kind;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_v_npy_kind = __pyx_t_20;

  /* "serializor/binary/ser.py":1517
 *     npy_kind: cython.char = arr.descr.kind
 *     # . ndarray[object]
 *     if npy_kind == identifier.NDARRAY_OBJECT:             # <<<<<<<<<<<<<<
 *         items.append(_ser_ndarray_object(arr, NDARRAY_OBJECT_DT))
 *     # . ndarray[int]
 */
  __pyx_t_18 = (__pyx_v_npy_kind == __pyx_v_10serializor_10identifier_NDARRAY_OBJECT);
  if (__pyx_t_18) {

    /* "serializor/binary/ser.py":1518
 *     # . ndarray[object]
 *     if npy_kind == identifier.NDARRAY_OBJECT:
 *         items.append(_ser_ndarray_object(arr, NDARRAY_OBJECT_DT))             # <<<<<<<<<<<<<<
 *     # . ndarray[int]
 *     elif npy_kind == identifier.NDARRAY_INT:
 */
    __pyx_t_4 = __pyx_v_10serializor_6binary_3ser_NDARRAY_OBJECT_DT;
    __Pyx_INCREF(__pyx_t_4);
    __pyx_t_6 = __pyx_f_10serializor_6binary_3ser__ser_ndarray_object(__pyx_v_arr, ((PyObject*)__pyx_t_4)); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1518, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_21 = __Pyx_PyList_Append(__pyx_v_items, __pyx_t_6); if (unlikely(__pyx_t_21 == ((int)-1))) __PYX_ERR(0, 1518, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

    /* "serializor/binary/ser.py":1517
 *     npy_kind: cython.char = arr.descr.kind
 *     # . ndarray[object]
 *     if npy_kind == identifier.NDARRAY_OBJECT:             # <<<<<<<<<<<<<<
 *         items.append(_ser_ndarray_object(arr, NDARRAY_OBJECT_DT))
 *     # . ndarray[int]
 */
    goto __pyx_L38;
  }

  /* "serializor/binary/ser.py":1520
 *         items.append(_ser_ndarray_object(arr, NDARRAY_OBJECT_DT))
 *     # . ndarray[int]
 *     elif npy_kind == identifier.NDARRAY_INT:             # <<<<<<<<<<<<<<
 *         items.append(_ser_ndarray_numeric(arr, NDARRAY_INT_DT, arr.descr.type_num))
 *     # . ndarray[uint]
 */
  __pyx_t_18 = (__pyx_v_npy_kind == __pyx_v_10serializor_10identifier_NDARRAY_INT);
  if (__pyx_t_18) {

    /* "serializor/binary/ser.py":1521
 *     # . ndarray[int]
 *     elif npy_kind == identifier.NDARRAY_INT:
 *         items.append(_ser_ndarray_numeric(arr, NDARRAY_INT_DT, arr.descr.type_num))             # <<<<<<<<<<<<<<
 *     # . ndarray[uint]
 *     elif npy_kind == identifier.NDARRAY_UINT:
 */
    __pyx_t_6 = __pyx_v_10serializor_6binary_3ser_NDARRAY_INT_DT;
    __Pyx_INCREF(__pyx_t_6);
    __pyx_t_4 = ((PyObject *)__pyx_f_5numpy_7ndarray_5descr_descr(__pyx_v_arr)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1521, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_7 = __pyx_f_10serializor_6binary_3ser__ser_ndarray_numeric(__pyx_v_arr, ((PyObject*)__pyx_t_6), ((PyArray_Descr *)__pyx_t_4)->type_num); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1521, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_21 = __Pyx_PyList_Append(__pyx_v_items, __pyx_t_7); if (unlikely(__pyx_t_21 == ((int)-1))) __PYX_ERR(0, 1521, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

    /* "serializor/binary/ser.py":1520
 *         items.append(_ser_ndarray_object(arr, NDARRAY_OBJECT_DT))
 *     # . ndarray[int]
 *     elif npy_kind == identifier.NDARRAY_INT:             # <<<<<<<<<<<<<<
 *         items.append(_ser_ndarray_numeric(arr, NDARRAY_INT_DT, arr.descr.type_num))
 *     # . ndarray[uint]
 */
    goto __pyx_L38;
  }

  /* "serializor/binary/ser.py":1523
 *         items.append(_ser_ndarray_numeric(arr, NDARRAY_INT_DT, arr.descr.type_num))
 *     # . ndarray[uint]
 *     elif npy_kind == identifier.NDARRAY_UINT:             # <<<<<<<<<<<<<<
 *         items.append(_ser_ndarray_numeric(arr, NDARRAY_UINT_DT, arr.descr.type_num))
 *     # . ndarray[float]
 */
  __pyx_t_18 = (__pyx_v_npy_kind == __pyx_v_10serializor_10identifier_NDARRAY_UINT);
  if (__pyx_t_18) {

    /* "serializor/binary/ser.py":1524
 *     # . ndarray[uint]
 *     elif npy_kind == identifier.NDARRAY_UINT:
 *         items.append(_ser_ndarray_numeric(arr, NDARRAY_UINT_DT, arr.descr.type_num))             # <<<<<<<<<<<<<<
 *     # . ndarray[float]
 *     elif npy_kind == identifier.NDARRAY_FLOAT:
 */
    __pyx_t_7 = __pyx_v_10serializor_6binary_3ser_NDARRAY_UINT_DT;
    __Pyx_INCREF(__pyx_t_7);
    __pyx_t_4 = ((PyObject *)__pyx_f_5numpy_7ndarray_5descr_descr(__pyx_v_arr)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1524, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = __pyx_f_10serializor_6binary_3ser__ser_ndarray_numeric(__pyx_v_arr, ((PyObject*)__pyx_t_7), ((PyArray_Descr *)__pyx_t_4)->type_num); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1524, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_21 = __Pyx_PyList_Append(__pyx_v_items, __pyx_t_6); if (unlikely(__pyx_t_21 == ((int)-1))) __PYX_ERR(0, 1524, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

    /* "serializor/binary/ser.py":1523
 *         items.append(_ser_ndarray_numeric(arr, NDARRAY_INT_DT, arr.descr.type_num))
 *     # . ndarray[uint]
 *     elif npy_kind == identifier.NDARRAY_UINT:             # <<<<<<<<<<<<<<
 *         items.append(_ser_ndarray_numeric(arr, NDARRAY_UINT_DT, arr.descr.type_num))
 *     # . ndarray[float]
 */
    goto __pyx_L38;
  }

  /* "serializor/binary/ser.py":1526
 *         items.append(_ser_ndarray_numeric(arr, NDARRAY_UINT_DT, arr.descr.type_num))
 *     # . ndarray[float]
 *     elif npy_kind == identifier.NDARRAY_FLOAT:             # <<<<<<<<<<<<<<
 *         items.append(_ser_ndarray_numeric(arr, NDARRAY_FLOAT_DT, arr.descr.type_num))
 *     # . ndarray[bool]
 */
  __pyx_t_18 = (__pyx_v_npy_kind == __pyx_v_10serializor_10identifier_NDARRAY_FLOAT);
  if (__pyx_t_18) {

    /* "serializor/binary/ser.py":1527
 *     # . ndarray[float]
 *     elif npy_kind == identifier.NDARRAY_FLOAT:
 *         items.append(_ser_ndarray_numeric(arr, NDARRAY_FLOAT_DT, arr.descr.type_num))             # <<<<<<<<<<<<<<
 *     # . ndarray[bool]
 *     elif npy_kind == identifier.NDARRAY_BOOL:
 */
    __pyx_t_6 = __pyx_v_10serializor_6binary_3ser_NDARRAY_FLOAT_DT;
    __Pyx_INCREF(__pyx_t_6);
    __pyx_t_4 = ((PyObject *)__pyx_f_5numpy_7ndarray_5descr_descr(__pyx_v_arr)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1527, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_7 = __pyx_f_10serializor_6binary_3ser__ser_ndarray_numeric(__pyx_v_arr, ((PyObject*)__pyx_t_6), ((PyArray_Descr *)__pyx_t_4)->type_num); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1527, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_21 = __Pyx_PyList_Append(__pyx_v_items, __pyx_t_7); if (unlikely(__pyx_t_21 == ((int)-1))) __PYX_ERR(0, 1527, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

    /* "serializor/binary/ser.py":1526
 *         items.append(_ser_ndarray_numeric(arr, NDARRAY_UINT_DT, arr.descr.type_num))
 *     # . ndarray[float]
 *     elif npy_kind == identifier.NDARRAY_FLOAT:             # <<<<<<<<<<<<<<
 *         items.append(_ser_ndarray_numeric(arr, NDARRAY_FLOAT_DT, arr.descr.type_num))
 *     # . ndarray[bool]
 */
    goto __pyx_L38;
  }

  /* "serializor/binary/ser.py":1529
 *         items.append(_ser_ndarray_numeric(arr, NDARRAY_FLOAT_DT, arr.descr.type_num))
 *     # . ndarray[bool]
 *     elif npy_kind == identifier.NDARRAY_BOOL:             # <<<<<<<<<<<<<<
 *         items.append(
 *             _ser_ndarray_numeric(
 */
  __pyx_t_18 = (__pyx_v_npy_kind == __pyx_v_10serializor_10identifier_NDARRAY_BOOL);
  if (__pyx_t_18) {

    /* "serializor/binary/ser.py":1532
 *         items.append(
 *             _ser_ndarray_numeric(
 *                 np.PyArray_Cast(arr, np.NPY_TYPES.NPY_INT64),  # cast to int64             # <<<<<<<<<<<<<<
 *                 NDARRAY_BOOL_DT,
 *                 arr.descr.type_num,
 */
    __pyx_t_7 = PyArray_Cast(__pyx_v_arr, NPY_INT64); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1532, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    if (!(likely(((__pyx_t_7) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_7, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 1532, __pyx_L1_error)

    /* "serializor/binary/ser.py":1533
 *             _ser_ndarray_numeric(
 *                 np.PyArray_Cast(arr, np.NPY_TYPES.NPY_INT64),  # cast to int64
 *                 NDARRAY_BOOL_DT,             # <<<<<<<<<<<<<<
 *                 arr.descr.type_num,
 *             )
 */
    __pyx_t_4 = __pyx_v_10serializor_6binary_3ser_NDARRAY_BOOL_DT;
    __Pyx_INCREF(__pyx_t_4);

    /* "serializor/binary/ser.py":1534
 *                 np.PyArray_Cast(arr, np.NPY_TYPES.NPY_INT64),  # cast to int64
 *                 NDARRAY_BOOL_DT,
 *                 arr.descr.type_num,             # <<<<<<<<<<<<<<
 *             )
 *         )
 */
    __pyx_t_6 = ((PyObject *)__pyx_f_5numpy_7ndarray_5descr_descr(__pyx_v_arr)); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1534, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);

    /* "serializor/binary/ser.py":1531
 *     elif npy_kind == identifier.NDARRAY_BOOL:
 *         items.append(
 *             _ser_ndarray_numeric(             # <<<<<<<<<<<<<<
 *                 np.PyArray_Cast(arr, np.NPY_TYPES.NPY_INT64),  # cast to int64
 *                 NDARRAY_BOOL_DT,
 */
    __pyx_t_8 = __pyx_f_10serializor_6binary_3ser__ser_ndarray_numeric(((PyArrayObject *)__pyx_t_7), ((PyObject*)__pyx_t_4), ((PyArray_Descr *)__pyx_t_6)->type_num); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1531, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

    /* "serializor/binary/ser.py":1530
 *     # . ndarray[bool]
 *     elif npy_kind == identifier.NDARRAY_BOOL:
 *         items.append(             # <<<<<<<<<<<<<<
 *             _ser_ndarray_numeric(
 *                 np.PyArray_Cast(arr, np.NPY_TYPES.NPY_INT64),  # cast to int64
 */
    __pyx_t_21 = __Pyx_PyList_Append(__pyx_v_items, __pyx_t_8); if (unlikely(__pyx_t_21 == ((int)-1))) __PYX_ERR(0, 1530, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

    /* "serializor/binary/ser.py":1529
 *         items.append(_ser_ndarray_numeric(arr, NDARRAY_FLOAT_DT, arr.descr.type_num))
 *     # . ndarray[bool]
 *     elif npy_kind == identifier.NDARRAY_BOOL:             # <<<<<<<<<<<<<<
 *         items.append(
 *             _ser_ndarray_numeric(
 */
    goto __pyx_L38;
  }

  /* "serializor/binary/ser.py":1538
 *         )
 *     # . ndarray[datetime64]
 *     elif npy_kind == identifier.NDARRAY_DT64:             # <<<<<<<<<<<<<<
 *         items.append(NDARRAY_DT64_DT)
 *         values = _ser_ndarray_nptime(arr, NDARRAY_DT64_DT)
 */
  __pyx_t_18 = (__pyx_v_npy_kind == __pyx_v_10serializor_10identifier_NDARRAY_DT64);
  if (__pyx_t_18) {

    /* "serializor/binary/ser.py":1539
 *     # . ndarray[datetime64]
 *     elif npy_kind == identifier.NDARRAY_DT64:
 *         items.append(NDARRAY_DT64_DT)             # <<<<<<<<<<<<<<
 *         values = _ser_ndarray_nptime(arr, NDARRAY_DT64_DT)
 *         # . With timezone
 */
    __pyx_t_8 = __pyx_v_10serializor_6binary_3ser_NDARRAY_DT64_DT;
    __Pyx_INCREF(__pyx_t_8);
    __pyx_t_21 = __Pyx_PyList_Append(__pyx_v_items, __pyx_t_8); if (unlikely(__pyx_t_21 == ((int)-1))) __PYX_ERR(0, 1539, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

    /* "serializor/binary/ser.py":1540
 *     elif npy_kind == identifier.NDARRAY_DT64:
 *         items.append(NDARRAY_DT64_DT)
 *         values = _ser_ndarray_nptime(arr, NDARRAY_DT64_DT)             # <<<<<<<<<<<<<<
 *         # . With timezone
 *         if (tz := obj.dt.tz) is not None:
 */
    __pyx_t_8 = __pyx_v_10serializor_6binary_3ser_NDARRAY_DT64_DT;
    __Pyx_INCREF(__pyx_t_8);
    __pyx_t_6 = __pyx_f_10serializor_6binary_3ser__ser_ndarray_nptime(__pyx_v_arr, ((PyObject*)__pyx_t_8)); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1540, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_v_values = ((PyObject*)__pyx_t_6);
    __pyx_t_6 = 0;

    /* "serializor/binary/ser.py":1542
 *         values = _ser_ndarray_nptime(arr, NDARRAY_DT64_DT)
 *         # . With timezone
 *         if (tz := obj.dt.tz) is not None:             # <<<<<<<<<<<<<<
 *             # . approach 1: timezone name
 *             if set_contains(utils.AVAILABLE_TIMEZONES, tz_name := str(tz)):
 */
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_obj, __pyx_n_s_dt); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1542, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_tz); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1542, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_v_tz = __pyx_t_8;
    __Pyx_INCREF(__pyx_t_8);
    __pyx_t_18 = (__pyx_t_8 != Py_None);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    if (__pyx_t_18) {

      /* "serializor/binary/ser.py":1544
 *         if (tz := obj.dt.tz) is not None:
 *             # . approach 1: timezone name
 *             if set_contains(utils.AVAILABLE_TIMEZONES, tz_name := str(tz)):             # <<<<<<<<<<<<<<
 *                 name = utils.encode_str(tz_name, b"ascii")
 *                 items.append(b"\x01")  # approach 1
 */
      __pyx_t_8 = __pyx_v_10serializor_5utils_AVAILABLE_TIMEZONES;
      __Pyx_INCREF(__pyx_t_8);
      __pyx_t_6 = __Pyx_PyObject_Unicode(__pyx_v_tz); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1544, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_v_tz_name = __pyx_t_6;
      __Pyx_INCREF(__pyx_t_6);
      __pyx_t_4 = __pyx_t_6;
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_18 = PySet_Contains(__pyx_t_8, __pyx_t_4); if (unlikely(__pyx_t_18 == ((int)-1))) __PYX_ERR(0, 1544, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (__pyx_t_18) {

        /* "serializor/binary/ser.py":1545
 *             # . approach 1: timezone name
 *             if set_contains(utils.AVAILABLE_TIMEZONES, tz_name := str(tz)):
 *                 name = utils.encode_str(tz_name, b"ascii")             # <<<<<<<<<<<<<<
 *                 items.append(b"\x01")  # approach 1
 *                 items.append(gen_encoded_int(bytes_len(name)))  # type: ignore
 */
        __pyx_t_4 = __pyx_f_10serializor_5utils_encode_str(__pyx_v_tz_name, ((char *)"ascii")); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1545, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF_SET(__pyx_v_name, __pyx_t_4);
        __pyx_t_4 = 0;

        /* "serializor/binary/ser.py":1546
 *             if set_contains(utils.AVAILABLE_TIMEZONES, tz_name := str(tz)):
 *                 name = utils.encode_str(tz_name, b"ascii")
 *                 items.append(b"\x01")  # approach 1             # <<<<<<<<<<<<<<
 *                 items.append(gen_encoded_int(bytes_len(name)))  # type: ignore
 *                 items.append(name)
 */
        __pyx_t_21 = __Pyx_PyList_Append(__pyx_v_items, __pyx_kp_b__5); if (unlikely(__pyx_t_21 == ((int)-1))) __PYX_ERR(0, 1546, __pyx_L1_error)

        /* "serializor/binary/ser.py":1547
 *                 name = utils.encode_str(tz_name, b"ascii")
 *                 items.append(b"\x01")  # approach 1
 *                 items.append(gen_encoded_int(bytes_len(name)))  # type: ignore             # <<<<<<<<<<<<<<
 *                 items.append(name)
 *                 items.append(values)
 */
        __pyx_t_22 = PyBytes_Size(__pyx_v_name); if (unlikely(__pyx_t_22 == ((Py_ssize_t)-1L))) __PYX_ERR(0, 1547, __pyx_L1_error)
        __pyx_t_4 = __pyx_f_10serializor_6binary_3ser_gen_encoded_int(__pyx_t_22); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1547, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_21 = __Pyx_PyList_Append(__pyx_v_items, __pyx_t_4); if (unlikely(__pyx_t_21 == ((int)-1))) __PYX_ERR(0, 1547, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

        /* "serializor/binary/ser.py":1548
 *                 items.append(b"\x01")  # approach 1
 *                 items.append(gen_encoded_int(bytes_len(name)))  # type: ignore
 *                 items.append(name)             # <<<<<<<<<<<<<<
 *                 items.append(values)
 *                 return b"".join(items)  # exit
 */
        __pyx_t_21 = __Pyx_PyList_Append(__pyx_v_items, __pyx_v_name); if (unlikely(__pyx_t_21 == ((int)-1))) __PYX_ERR(0, 1548, __pyx_L1_error)

        /* "serializor/binary/ser.py":1549
 *                 items.append(gen_encoded_int(bytes_len(name)))  # type: ignore
 *                 items.append(name)
 *                 items.append(values)             # <<<<<<<<<<<<<<
 *                 return b"".join(items)  # exit
 *             # . approach 2: timezone offset
 */
        __pyx_t_21 = __Pyx_PyList_Append(__pyx_v_items, __pyx_v_values); if (unlikely(__pyx_t_21 == ((int)-1))) __PYX_ERR(0, 1549, __pyx_L1_error)

        /* "serializor/binary/ser.py":1550
 *                 items.append(name)
 *                 items.append(values)
 *                 return b"".join(items)  # exit             # <<<<<<<<<<<<<<
 *             # . approach 2: timezone offset
 *             if (offset := tz.utcoffset(None)) is not None:
 */
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_4 = __Pyx_PyBytes_Join(__pyx_kp_b__4, __pyx_v_items); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1550, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        if (!(likely(PyBytes_CheckExact(__pyx_t_4))||((__pyx_t_4) == Py_None) || __Pyx_RaiseUnexpectedTypeError("bytes", __pyx_t_4))) __PYX_ERR(0, 1550, __pyx_L1_error)
        __pyx_r = ((PyObject*)__pyx_t_4);
        __pyx_t_4 = 0;
        goto __pyx_L0;

        /* "serializor/binary/ser.py":1544
 *         if (tz := obj.dt.tz) is not None:
 *             # . approach 1: timezone name
 *             if set_contains(utils.AVAILABLE_TIMEZONES, tz_name := str(tz)):             # <<<<<<<<<<<<<<
 *                 name = utils.encode_str(tz_name, b"ascii")
 *                 items.append(b"\x01")  # approach 1
 */
      }

      /* "serializor/binary/ser.py":1552
 *                 return b"".join(items)  # exit
 *             # . approach 2: timezone offset
 *             if (offset := tz.utcoffset(None)) is not None:             # <<<<<<<<<<<<<<
 *                 items.append(b"\x02")  # approach 2
 *                 items.append(pack_int32(datetime.timedelta_days(offset)))  # type: ignore
 */
      __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_v_tz, __pyx_n_s_utcoffset); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1552, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_6 = NULL;
      __pyx_t_23 = 0;
      #if CYTHON_UNPACK_METHODS
      if (likely(PyMethod_Check(__pyx_t_8))) {
        __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_8);
        if (likely(__pyx_t_6)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
          __Pyx_INCREF(__pyx_t_6);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_8, function);
          __pyx_t_23 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_6, Py_None};
        __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_8, __pyx_callargs+1-__pyx_t_23, 1+__pyx_t_23);
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1552, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      }
      __pyx_v_offset = __pyx_t_4;
      __Pyx_INCREF(__pyx_t_4);
      __pyx_t_18 = (__pyx_t_4 != Py_None);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (__pyx_t_18) {

        /* "serializor/binary/ser.py":1553
 *             # . approach 2: timezone offset
 *             if (offset := tz.utcoffset(None)) is not None:
 *                 items.append(b"\x02")  # approach 2             # <<<<<<<<<<<<<<
 *                 items.append(pack_int32(datetime.timedelta_days(offset)))  # type: ignore
 *                 items.append(pack_int32(datetime.timedelta_seconds(offset)))  # type: ignore
 */
        __pyx_t_21 = __Pyx_PyList_Append(__pyx_v_items, __pyx_kp_b__6); if (unlikely(__pyx_t_21 == ((int)-1))) __PYX_ERR(0, 1553, __pyx_L1_error)

        /* "serializor/binary/ser.py":1554
 *             if (offset := tz.utcoffset(None)) is not None:
 *                 items.append(b"\x02")  # approach 2
 *                 items.append(pack_int32(datetime.timedelta_days(offset)))  # type: ignore             # <<<<<<<<<<<<<<
 *                 items.append(pack_int32(datetime.timedelta_seconds(offset)))  # type: ignore
 *                 items.append(pack_int32(datetime.timedelta_microseconds(offset)))  # type: ignore
 */
        __pyx_t_4 = __pyx_f_10serializor_6binary_3ser_pack_int32(__pyx_f_7cpython_8datetime_timedelta_days(__pyx_v_offset)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1554, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_21 = __Pyx_PyList_Append(__pyx_v_items, __pyx_t_4); if (unlikely(__pyx_t_21 == ((int)-1))) __PYX_ERR(0, 1554, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

        /* "serializor/binary/ser.py":1555
 *                 items.append(b"\x02")  # approach 2
 *                 items.append(pack_int32(datetime.timedelta_days(offset)))  # type: ignore
 *                 items.append(pack_int32(datetime.timedelta_seconds(offset)))  # type: ignore             # <<<<<<<<<<<<<<
 *                 items.append(pack_int32(datetime.timedelta_microseconds(offset)))  # type: ignore
 *                 items.append(values)
 */
        __pyx_t_4 = __pyx_f_10serializor_6binary_3ser_pack_int32(__pyx_f_7cpython_8datetime_timedelta_seconds(__pyx_v_offset)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1555, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_21 = __Pyx_PyList_Append(__pyx_v_items, __pyx_t_4); if (unlikely(__pyx_t_21 == ((int)-1))) __PYX_ERR(0, 1555, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

        /* "serializor/binary/ser.py":1556
 *                 items.append(pack_int32(datetime.timedelta_days(offset)))  # type: ignore
 *                 items.append(pack_int32(datetime.timedelta_seconds(offset)))  # type: ignore
 *                 items.append(pack_int32(datetime.timedelta_microseconds(offset)))  # type: ignore             # <<<<<<<<<<<<<<
 *                 items.append(values)
 *                 return b"".join(items)  # exit
 */
        __pyx_t_4 = __pyx_f_10serializor_6binary_3ser_pack_int32(__pyx_f_7cpython_8datetime_timedelta_microseconds(__pyx_v_offset)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1556, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_21 = __Pyx_PyList_Append(__pyx_v_items, __pyx_t_4); if (unlikely(__pyx_t_21 == ((int)-1))) __PYX_ERR(0, 1556, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

        /* "serializor/binary/ser.py":1557
 *                 items.append(pack_int32(datetime.timedelta_seconds(offset)))  # type: ignore
 *                 items.append(pack_int32(datetime.timedelta_microseconds(offset)))  # type: ignore
 *                 items.append(values)             # <<<<<<<<<<<<<<
 *                 return b"".join(items)  # exit
 *         # . Without timezone / Fallback
 */
        __pyx_t_21 = __Pyx_PyList_Append(__pyx_v_items, __pyx_v_values); if (unlikely(__pyx_t_21 == ((int)-1))) __PYX_ERR(0, 1557, __pyx_L1_error)

        /* "serializor/binary/ser.py":1558
 *                 items.append(pack_int32(datetime.timedelta_microseconds(offset)))  # type: ignore
 *                 items.append(values)
 *                 return b"".join(items)  # exit             # <<<<<<<<<<<<<<
 *         # . Without timezone / Fallback
 *         items.append(b"\x00")  # approach 0
 */
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_4 = __Pyx_PyBytes_Join(__pyx_kp_b__4, __pyx_v_items); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1558, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        if (!(likely(PyBytes_CheckExact(__pyx_t_4))||((__pyx_t_4) == Py_None) || __Pyx_RaiseUnexpectedTypeError("bytes", __pyx_t_4))) __PYX_ERR(0, 1558, __pyx_L1_error)
        __pyx_r = ((PyObject*)__pyx_t_4);
        __pyx_t_4 = 0;
        goto __pyx_L0;

        /* "serializor/binary/ser.py":1552
 *                 return b"".join(items)  # exit
 *             # . approach 2: timezone offset
 *             if (offset := tz.utcoffset(None)) is not None:             # <<<<<<<<<<<<<<
 *                 items.append(b"\x02")  # approach 2
 *                 items.append(pack_int32(datetime.timedelta_days(offset)))  # type: ignore
 */
      }

      /* "serializor/binary/ser.py":1542
 *         values = _ser_ndarray_nptime(arr, NDARRAY_DT64_DT)
 *         # . With timezone
 *         if (tz := obj.dt.tz) is not None:             # <<<<<<<<<<<<<<
 *             # . approach 1: timezone name
 *             if set_contains(utils.AVAILABLE_TIMEZONES, tz_name := str(tz)):
 */
    }

    /* "serializor/binary/ser.py":1560
 *                 return b"".join(items)  # exit
 *         # . Without timezone / Fallback
 *         items.append(b"\x00")  # approach 0             # <<<<<<<<<<<<<<
 *         items.append(values)
 *     # . ndarray[timedelta64]
 */
    __pyx_t_21 = __Pyx_PyList_Append(__pyx_v_items, __pyx_kp_b__7); if (unlikely(__pyx_t_21 == ((int)-1))) __PYX_ERR(0, 1560, __pyx_L1_error)

    /* "serializor/binary/ser.py":1561
 *         # . Without timezone / Fallback
 *         items.append(b"\x00")  # approach 0
 *         items.append(values)             # <<<<<<<<<<<<<<
 *     # . ndarray[timedelta64]
 *     elif npy_kind == identifier.NDARRAY_TD64:
 */
    __pyx_t_21 = __Pyx_PyList_Append(__pyx_v_items, __pyx_v_values); if (unlikely(__pyx_t_21 == ((int)-1))) __PYX_ERR(0, 1561, __pyx_L1_error)

    /* "serializor/binary/ser.py":1538
 *         )
 *     # . ndarray[datetime64]
 *     elif npy_kind == identifier.NDARRAY_DT64:             # <<<<<<<<<<<<<<
 *         items.append(NDARRAY_DT64_DT)
 *         values = _ser_ndarray_nptime(arr, NDARRAY_DT64_DT)
 */
    goto __pyx_L38;
  }

  /* "serializor/binary/ser.py":1563
 *         items.append(values)
 *     # . ndarray[timedelta64]
 *     elif npy_kind == identifier.NDARRAY_TD64:             # <<<<<<<<<<<<<<
 *         items.append(_ser_ndarray_nptime(arr, NDARRAY_TD64_DT))
 *     # . ndarray[complex]
 */
  __pyx_t_18 = (__pyx_v_npy_kind == __pyx_v_10serializor_10identifier_NDARRAY_TD64);
  if (__pyx_t_18) {

    /* "serializor/binary/ser.py":1564
 *     # . ndarray[timedelta64]
 *     elif npy_kind == identifier.NDARRAY_TD64:
 *         items.append(_ser_ndarray_nptime(arr, NDARRAY_TD64_DT))             # <<<<<<<<<<<<<<
 *     # . ndarray[complex]
 *     elif npy_kind == identifier.NDARRAY_COMPLEX:
 */
    __pyx_t_4 = __pyx_v_10serializor_6binary_3ser_NDARRAY_TD64_DT;
    __Pyx_INCREF(__pyx_t_4);
    __pyx_t_8 = __pyx_f_10serializor_6binary_3ser__ser_ndarray_nptime(__pyx_v_arr, ((PyObject*)__pyx_t_4)); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1564, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_21 = __Pyx_PyList_Append(__pyx_v_items, __pyx_t_8); if (unlikely(__pyx_t_21 == ((int)-1))) __PYX_ERR(0, 1564, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

    /* "serializor/binary/ser.py":1563
 *         items.append(values)
 *     # . ndarray[timedelta64]
 *     elif npy_kind == identifier.NDARRAY_TD64:             # <<<<<<<<<<<<<<
 *         items.append(_ser_ndarray_nptime(arr, NDARRAY_TD64_DT))
 *     # . ndarray[complex]
 */
    goto __pyx_L38;
  }

  /* "serializor/binary/ser.py":1566
 *         items.append(_ser_ndarray_nptime(arr, NDARRAY_TD64_DT))
 *     # . ndarray[complex]
 *     elif npy_kind == identifier.NDARRAY_COMPLEX:             # <<<<<<<<<<<<<<
 *         items.append(_ser_ndarray_complex(arr, NDARRAY_COMPLEX_DT))
 *     # . ndarray[bytes]
 */
  __pyx_t_18 = (__pyx_v_npy_kind == __pyx_v_10serializor_10identifier_NDARRAY_COMPLEX);
  if (__pyx_t_18) {

    /* "serializor/binary/ser.py":1567
 *     # . ndarray[complex]
 *     elif npy_kind == identifier.NDARRAY_COMPLEX:
 *         items.append(_ser_ndarray_complex(arr, NDARRAY_COMPLEX_DT))             # <<<<<<<<<<<<<<
 *     # . ndarray[bytes]
 *     elif npy_kind == identifier.NDARRAY_BYTES:
 */
    __pyx_t_8 = __pyx_v_10serializor_6binary_3ser_NDARRAY_COMPLEX_DT;
    __Pyx_INCREF(__pyx_t_8);
    __pyx_t_4 = __pyx_f_10serializor_6binary_3ser__ser_ndarray_complex(__pyx_v_arr, ((PyObject*)__pyx_t_8)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1567, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_t_21 = __Pyx_PyList_Append(__pyx_v_items, __pyx_t_4); if (unlikely(__pyx_t_21 == ((int)-1))) __PYX_ERR(0, 1567, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "serializor/binary/ser.py":1566
 *         items.append(_ser_ndarray_nptime(arr, NDARRAY_TD64_DT))
 *     # . ndarray[complex]
 *     elif npy_kind == identifier.NDARRAY_COMPLEX:             # <<<<<<<<<<<<<<
 *         items.append(_ser_ndarray_complex(arr, NDARRAY_COMPLEX_DT))
 *     # . ndarray[bytes]
 */
    goto __pyx_L38;
  }

  /* "serializor/binary/ser.py":1569
 *         items.append(_ser_ndarray_complex(arr, NDARRAY_COMPLEX_DT))
 *     # . ndarray[bytes]
 *     elif npy_kind == identifier.NDARRAY_BYTES:             # <<<<<<<<<<<<<<
 *         items.append(_ser_ndarray_bytes(arr, NDARRAY_BYTES_DT))
 *     # . ndarray[str]
 */
  __pyx_t_18 = (__pyx_v_npy_kind == __pyx_v_10serializor_10identifier_NDARRAY_BYTES);
  if (__pyx_t_18) {

    /* "serializor/binary/ser.py":1570
 *     # . ndarray[bytes]
 *     elif npy_kind == identifier.NDARRAY_BYTES:
 *         items.append(_ser_ndarray_bytes(arr, NDARRAY_BYTES_DT))             # <<<<<<<<<<<<<<
 *     # . ndarray[str]
 *     elif npy_kind == identifier.NDARRAY_UNICODE:
 */
    __pyx_t_4 = __pyx_v_10serializor_6binary_3ser_NDARRAY_BYTES_DT;
    __Pyx_INCREF(__pyx_t_4);
    __pyx_t_8 = __pyx_f_10serializor_6binary_3ser__ser_ndarray_bytes(__pyx_v_arr, ((PyObject*)__pyx_t_4)); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1570, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_21 = __Pyx_PyList_Append(__pyx_v_items, __pyx_t_8); if (unlikely(__pyx_t_21 == ((int)-1))) __PYX_ERR(0, 1570, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

    /* "serializor/binary/ser.py":1569
 *         items.append(_ser_ndarray_complex(arr, NDARRAY_COMPLEX_DT))
 *     # . ndarray[bytes]
 *     elif npy_kind == identifier.NDARRAY_BYTES:             # <<<<<<<<<<<<<<
 *         items.append(_ser_ndarray_bytes(arr, NDARRAY_BYTES_DT))
 *     # . ndarray[str]
 */
    goto __pyx_L38;
  }

  /* "serializor/binary/ser.py":1572
 *         items.append(_ser_ndarray_bytes(arr, NDARRAY_BYTES_DT))
 *     # . ndarray[str]
 *     elif npy_kind == identifier.NDARRAY_UNICODE:             # <<<<<<<<<<<<<<
 *         items.append(_ser_ndarray_unicode(arr, NDARRAY_UNICODE_DT))
 *     # . invalid dtype
 */
  __pyx_t_18 = (__pyx_v_npy_kind == __pyx_v_10serializor_10identifier_NDARRAY_UNICODE);
  if (likely(__pyx_t_18)) {

    /* "serializor/binary/ser.py":1573
 *     # . ndarray[str]
 *     elif npy_kind == identifier.NDARRAY_UNICODE:
 *         items.append(_ser_ndarray_unicode(arr, NDARRAY_UNICODE_DT))             # <<<<<<<<<<<<<<
 *     # . invalid dtype
 *     else:
 */
    __pyx_t_8 = __pyx_v_10serializor_6binary_3ser_NDARRAY_UNICODE_DT;
    __Pyx_INCREF(__pyx_t_8);
    __pyx_t_4 = __pyx_f_10serializor_6binary_3ser__ser_ndarray_unicode(__pyx_v_arr, ((PyObject*)__pyx_t_8)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1573, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_t_21 = __Pyx_PyList_Append(__pyx_v_items, __pyx_t_4); if (unlikely(__pyx_t_21 == ((int)-1))) __PYX_ERR(0, 1573, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "serializor/binary/ser.py":1572
 *         items.append(_ser_ndarray_bytes(arr, NDARRAY_BYTES_DT))
 *     # . ndarray[str]
 *     elif npy_kind == identifier.NDARRAY_UNICODE:             # <<<<<<<<<<<<<<
 *         items.append(_ser_ndarray_unicode(arr, NDARRAY_UNICODE_DT))
 *     # . invalid dtype
 */
    goto __pyx_L38;
  }

  /* "serializor/binary/ser.py":1576
 *     # . invalid dtype
 *     else:
 *         raise TypeError("unsupported <'pandas.Series'> dtype '%s'." % obj.dtype)             # <<<<<<<<<<<<<<
 * 
 *     # Compose
 */
  /*else*/ {
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_obj, __pyx_n_s_dtype); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1576, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_8 = __Pyx_PyUnicode_FormatSafe(__pyx_kp_u_unsupported_pandas_Series_dtype, __pyx_t_4); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1576, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_builtin_TypeError, __pyx_t_8); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1576, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_Raise(__pyx_t_4, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __PYX_ERR(0, 1576, __pyx_L1_error)
  }
  __pyx_L38:;

  /* "serializor/binary/ser.py":1579
 * 
 *     # Compose
 *     return b"".join(items)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_4 = __Pyx_PyBytes_Join(__pyx_kp_b__4, __pyx_v_items); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1579, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (!(likely(PyBytes_CheckExact(__pyx_t_4))||((__pyx_t_4) == Py_None) || __Pyx_RaiseUnexpectedTypeError("bytes", __pyx_t_4))) __PYX_ERR(0, 1579, __pyx_L1_error)
  __pyx_r = ((PyObject*)__pyx_t_4);
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "serializor/binary/ser.py":1473
 * 
 * 
 * @cython.cfunc             # <<<<<<<<<<<<<<
 * @cython.inline(True)
 * def _ser_series(obj: Series) -> bytes:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("serializor.binary.ser._ser_series", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_arr);
  __Pyx_XDECREF(__pyx_v_name);
  __Pyx_XDECREF(__pyx_v_err);
  __Pyx_XDECREF(__pyx_v_items);
  __Pyx_XDECREF(__pyx_v_name_value);
  __Pyx_XDECREF(__pyx_v_values);
  __Pyx_XDECREF(__pyx_v_tz);
  __Pyx_XDECREF(__pyx_v_tz_name);
  __Pyx_XDECREF(__pyx_v_offset);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "serializor/binary/ser.py":1582
 * 
 * 
 * @cython.cfunc             # <<<<<<<<<<<<<<
 * @cython.inline(True)
 * def _ser_datetime_index(obj: DatetimeIndex) -> bytes:
 */

static PyObject *__pyx_f_10serializor_6binary_3ser__ser_datetime_index(PyObject *__pyx_v_obj) {
  PyArrayObject *__pyx_v_arr = 0;
  PyObject *__pyx_v_name = 0;
  PyObject *__pyx_v_err = NULL;
  PyObject *__pyx_v_items = NULL;
  PyObject *__pyx_v_name_value = NULL;
  PyObject *__pyx_v_values = NULL;
  PyObject *__pyx_v_tz = NULL;
  PyObject *__pyx_v_tz_name = NULL;
  PyObject *__pyx_v_offset = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_t_10;
  char const *__pyx_t_11;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  int __pyx_t_18;
  char const *__pyx_t_19;
  int __pyx_t_20;
  Py_ssize_t __pyx_t_21;
  unsigned int __pyx_t_22;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_ser_datetime_index", 1);

  /* "serializor/binary/ser.py":1605
 *     """
 *     # Access values and name
 *     try:             # <<<<<<<<<<<<<<
 *         arr: np.ndarray = obj.values
 *         name: object = obj.name
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "serializor/binary/ser.py":1606
 *     # Access values and name
 *     try:
 *         arr: np.ndarray = obj.values             # <<<<<<<<<<<<<<
 *         name: object = obj.name
 *     except Exception as err:
 */
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_obj, __pyx_n_s_values); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1606, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_4);
      if (!(likely(((__pyx_t_4) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_4, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 1606, __pyx_L3_error)
      __pyx_v_arr = ((PyArrayObject *)__pyx_t_4);
      __pyx_t_4 = 0;

      /* "serializor/binary/ser.py":1607
 *     try:
 *         arr: np.ndarray = obj.values
 *         name: object = obj.name             # <<<<<<<<<<<<<<
 *     except Exception as err:
 *         raise TypeError(
 */
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_obj, __pyx_n_s_name); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1607, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_v_name = __pyx_t_4;
      __pyx_t_4 = 0;

      /* "serializor/binary/ser.py":1605
 *     """
 *     # Access values and name
 *     try:             # <<<<<<<<<<<<<<
 *         arr: np.ndarray = obj.values
 *         name: object = obj.name
 */
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "serializor/binary/ser.py":1608
 *         arr: np.ndarray = obj.values
 *         name: object = obj.name
 *     except Exception as err:             # <<<<<<<<<<<<<<
 *         raise TypeError(
 *             "expects <'pandas.DatetimeIndex'>, got %s." % type(obj)
 */
    __pyx_t_5 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
    if (__pyx_t_5) {
      __Pyx_AddTraceback("serializor.binary.ser._ser_datetime_index", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_4, &__pyx_t_6, &__pyx_t_7) < 0) __PYX_ERR(0, 1608, __pyx_L5_except_error)
      __Pyx_XGOTREF(__pyx_t_4);
      __Pyx_XGOTREF(__pyx_t_6);
      __Pyx_XGOTREF(__pyx_t_7);
      __Pyx_INCREF(__pyx_t_6);
      __pyx_v_err = __pyx_t_6;
      /*try:*/ {

        /* "serializor/binary/ser.py":1610
 *     except Exception as err:
 *         raise TypeError(
 *             "expects <'pandas.DatetimeIndex'>, got %s." % type(obj)             # <<<<<<<<<<<<<<
 *         ) from err
 * 
 */
        __pyx_t_8 = __Pyx_PyUnicode_FormatSafe(__pyx_kp_u_expects_pandas_DatetimeIndex_got, ((PyObject *)Py_TYPE(__pyx_v_obj))); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1610, __pyx_L14_error)
        __Pyx_GOTREF(__pyx_t_8);

        /* "serializor/binary/ser.py":1609
 *         name: object = obj.name
 *     except Exception as err:
 *         raise TypeError(             # <<<<<<<<<<<<<<
 *             "expects <'pandas.DatetimeIndex'>, got %s." % type(obj)
 *         ) from err
 */
        __pyx_t_9 = __Pyx_PyObject_CallOneArg(__pyx_builtin_TypeError, __pyx_t_8); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1609, __pyx_L14_error)
        __Pyx_GOTREF(__pyx_t_9);
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

        /* "serializor/binary/ser.py":1611
 *         raise TypeError(
 *             "expects <'pandas.DatetimeIndex'>, got %s." % type(obj)
 *         ) from err             # <<<<<<<<<<<<<<
 * 
 *     # Serialize name
 */
        __Pyx_Raise(__pyx_t_9, 0, 0, __pyx_v_err);
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        __PYX_ERR(0, 1609, __pyx_L14_error)
      }

      /* "serializor/binary/ser.py":1608
 *         arr: np.ndarray = obj.values
 *         name: object = obj.name
 *     except Exception as err:             # <<<<<<<<<<<<<<
 *         raise TypeError(
 *             "expects <'pandas.DatetimeIndex'>, got %s." % type(obj)
 */
      /*finally:*/ {
        __pyx_L14_error:;
        /*exception exit:*/{
          __Pyx_PyThreadState_declare
          __Pyx_PyThreadState_assign
          __pyx_t_12 = 0; __pyx_t_13 = 0; __pyx_t_14 = 0; __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0;
          __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
          __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
          if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_15, &__pyx_t_16, &__pyx_t_17);
          if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_12, &__pyx_t_13, &__pyx_t_14) < 0)) __Pyx_ErrFetch(&__pyx_t_12, &__pyx_t_13, &__pyx_t_14);
          __Pyx_XGOTREF(__pyx_t_12);
          __Pyx_XGOTREF(__pyx_t_13);
          __Pyx_XGOTREF(__pyx_t_14);
          __Pyx_XGOTREF(__pyx_t_15);
          __Pyx_XGOTREF(__pyx_t_16);
          __Pyx_XGOTREF(__pyx_t_17);
          __pyx_t_5 = __pyx_lineno; __pyx_t_10 = __pyx_clineno; __pyx_t_11 = __pyx_filename;
          {
            __Pyx_DECREF(__pyx_v_err); __pyx_v_err = 0;
          }
          if (PY_MAJOR_VERSION >= 3) {
            __Pyx_XGIVEREF(__pyx_t_15);
            __Pyx_XGIVEREF(__pyx_t_16);
            __Pyx_XGIVEREF(__pyx_t_17);
            __Pyx_ExceptionReset(__pyx_t_15, __pyx_t_16, __pyx_t_17);
          }
          __Pyx_XGIVEREF(__pyx_t_12);
          __Pyx_XGIVEREF(__pyx_t_13);
          __Pyx_XGIVEREF(__pyx_t_14);
          __Pyx_ErrRestore(__pyx_t_12, __pyx_t_13, __pyx_t_14);
          __pyx_t_12 = 0; __pyx_t_13 = 0; __pyx_t_14 = 0; __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0;
          __pyx_lineno = __pyx_t_5; __pyx_clineno = __pyx_t_10; __pyx_filename = __pyx_t_11;
          goto __pyx_L5_except_error;
        }
      }
    }
    goto __pyx_L5_except_error;

    /* "serializor/binary/ser.py":1605
 *     """
 *     # Access values and name
 *     try:             # <<<<<<<<<<<<<<
 *         arr: np.ndarray = obj.values
 *         name: object = obj.name
 */
    __pyx_L5_except_error:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L8_try_end:;
  }

  /* "serializor/binary/ser.py":1614
 * 
 *     # Serialize name
 *     if name is None:  # no name             # <<<<<<<<<<<<<<
 *         items = [DATETIMEINDEX_ID, b"\x00"]
 *     else:  # has name
 */
  __pyx_t_18 = (__pyx_v_name == Py_None);
  if (__pyx_t_18) {

    /* "serializor/binary/ser.py":1615
 *     # Serialize name
 *     if name is None:  # no name
 *         items = [DATETIMEINDEX_ID, b"\x00"]             # <<<<<<<<<<<<<<
 *     else:  # has name
 *         try:
 */
    __pyx_t_7 = PyList_New(2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1615, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_INCREF(__pyx_v_10serializor_6binary_3ser_DATETIMEINDEX_ID);
    __Pyx_GIVEREF(__pyx_v_10serializor_6binary_3ser_DATETIMEINDEX_ID);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_7, 0, __pyx_v_10serializor_6binary_3ser_DATETIMEINDEX_ID)) __PYX_ERR(0, 1615, __pyx_L1_error);
    __Pyx_INCREF(__pyx_kp_b__7);
    __Pyx_GIVEREF(__pyx_kp_b__7);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_7, 1, __pyx_kp_b__7)) __PYX_ERR(0, 1615, __pyx_L1_error);
    __pyx_v_items = ((PyObject*)__pyx_t_7);
    __pyx_t_7 = 0;

    /* "serializor/binary/ser.py":1614
 * 
 *     # Serialize name
 *     if name is None:  # no name             # <<<<<<<<<<<<<<
 *         items = [DATETIMEINDEX_ID, b"\x00"]
 *     else:  # has name
 */
    goto __pyx_L20;
  }

  /* "serializor/binary/ser.py":1617
 *         items = [DATETIMEINDEX_ID, b"\x00"]
 *     else:  # has name
 *         try:             # <<<<<<<<<<<<<<
 *             name_value = _ser_common(name)
 *         except Exception as err:
 */
  /*else*/ {
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_3, &__pyx_t_2, &__pyx_t_1);
      __Pyx_XGOTREF(__pyx_t_3);
      __Pyx_XGOTREF(__pyx_t_2);
      __Pyx_XGOTREF(__pyx_t_1);
      /*try:*/ {

        /* "serializor/binary/ser.py":1618
 *     else:  # has name
 *         try:
 *             name_value = _ser_common(name)             # <<<<<<<<<<<<<<
 *         except Exception as err:
 *             raise ValueError(
 */
        __pyx_t_7 = __pyx_f_10serializor_6binary_3ser__ser_common(__pyx_v_name); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1618, __pyx_L21_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_v_name_value = ((PyObject*)__pyx_t_7);
        __pyx_t_7 = 0;

        /* "serializor/binary/ser.py":1617
 *         items = [DATETIMEINDEX_ID, b"\x00"]
 *     else:  # has name
 *         try:             # <<<<<<<<<<<<<<
 *             name_value = _ser_common(name)
 *         except Exception as err:
 */
      }
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L26_try_end;
      __pyx_L21_error:;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;

      /* "serializor/binary/ser.py":1619
 *         try:
 *             name_value = _ser_common(name)
 *         except Exception as err:             # <<<<<<<<<<<<<<
 *             raise ValueError(
 *                 "invalid <'pandas.DatetimeIndex'> name %r." % name
 */
      __pyx_t_10 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
      if (__pyx_t_10) {
        __Pyx_AddTraceback("serializor.binary.ser._ser_datetime_index", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_7, &__pyx_t_6, &__pyx_t_4) < 0) __PYX_ERR(0, 1619, __pyx_L23_except_error)
        __Pyx_XGOTREF(__pyx_t_7);
        __Pyx_XGOTREF(__pyx_t_6);
        __Pyx_XGOTREF(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_6);
        __pyx_v_err = __pyx_t_6;
        /*try:*/ {

          /* "serializor/binary/ser.py":1621
 *         except Exception as err:
 *             raise ValueError(
 *                 "invalid <'pandas.DatetimeIndex'> name %r." % name             # <<<<<<<<<<<<<<
 *             ) from err
 *         items = [DATETIMEINDEX_ID, b"\x01", name_value]
 */
          __pyx_t_9 = __Pyx_PyUnicode_FormatSafe(__pyx_kp_u_invalid_pandas_DatetimeIndex_nam, __pyx_v_name); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1621, __pyx_L32_error)
          __Pyx_GOTREF(__pyx_t_9);

          /* "serializor/binary/ser.py":1620
 *             name_value = _ser_common(name)
 *         except Exception as err:
 *             raise ValueError(             # <<<<<<<<<<<<<<
 *                 "invalid <'pandas.DatetimeIndex'> name %r." % name
 *             ) from err
 */
          __pyx_t_8 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_9); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1620, __pyx_L32_error)
          __Pyx_GOTREF(__pyx_t_8);
          __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

          /* "serializor/binary/ser.py":1622
 *             raise ValueError(
 *                 "invalid <'pandas.DatetimeIndex'> name %r." % name
 *             ) from err             # <<<<<<<<<<<<<<
 *         items = [DATETIMEINDEX_ID, b"\x01", name_value]
 * 
 */
          __Pyx_Raise(__pyx_t_8, 0, 0, __pyx_v_err);
          __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
          __PYX_ERR(0, 1620, __pyx_L32_error)
        }

        /* "serializor/binary/ser.py":1619
 *         try:
 *             name_value = _ser_common(name)
 *         except Exception as err:             # <<<<<<<<<<<<<<
 *             raise ValueError(
 *                 "invalid <'pandas.DatetimeIndex'> name %r." % name
 */
        /*finally:*/ {
          __pyx_L32_error:;
          /*exception exit:*/{
            __Pyx_PyThreadState_declare
            __Pyx_PyThreadState_assign
            __pyx_t_17 = 0; __pyx_t_16 = 0; __pyx_t_15 = 0; __pyx_t_14 = 0; __pyx_t_13 = 0; __pyx_t_12 = 0;
            __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
            __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
            if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_14, &__pyx_t_13, &__pyx_t_12);
            if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_17, &__pyx_t_16, &__pyx_t_15) < 0)) __Pyx_ErrFetch(&__pyx_t_17, &__pyx_t_16, &__pyx_t_15);
            __Pyx_XGOTREF(__pyx_t_17);
            __Pyx_XGOTREF(__pyx_t_16);
            __Pyx_XGOTREF(__pyx_t_15);
            __Pyx_XGOTREF(__pyx_t_14);
            __Pyx_XGOTREF(__pyx_t_13);
            __Pyx_XGOTREF(__pyx_t_12);
            __pyx_t_10 = __pyx_lineno; __pyx_t_5 = __pyx_clineno; __pyx_t_19 = __pyx_filename;
            {
              __Pyx_DECREF(__pyx_v_err); __pyx_v_err = 0;
            }
            if (PY_MAJOR_VERSION >= 3) {
              __Pyx_XGIVEREF(__pyx_t_14);
              __Pyx_XGIVEREF(__pyx_t_13);
              __Pyx_XGIVEREF(__pyx_t_12);
              __Pyx_ExceptionReset(__pyx_t_14, __pyx_t_13, __pyx_t_12);
            }
            __Pyx_XGIVEREF(__pyx_t_17);
            __Pyx_XGIVEREF(__pyx_t_16);
            __Pyx_XGIVEREF(__pyx_t_15);
            __Pyx_ErrRestore(__pyx_t_17, __pyx_t_16, __pyx_t_15);
            __pyx_t_17 = 0; __pyx_t_16 = 0; __pyx_t_15 = 0; __pyx_t_14 = 0; __pyx_t_13 = 0; __pyx_t_12 = 0;
            __pyx_lineno = __pyx_t_10; __pyx_clineno = __pyx_t_5; __pyx_filename = __pyx_t_19;
            goto __pyx_L23_except_error;
          }
        }
      }
      goto __pyx_L23_except_error;

      /* "serializor/binary/ser.py":1617
 *         items = [DATETIMEINDEX_ID, b"\x00"]
 *     else:  # has name
 *         try:             # <<<<<<<<<<<<<<
 *             name_value = _ser_common(name)
 *         except Exception as err:
 */
      __pyx_L23_except_error:;
      __Pyx_XGIVEREF(__pyx_t_3);
      __Pyx_XGIVEREF(__pyx_t_2);
      __Pyx_XGIVEREF(__pyx_t_1);
      __Pyx_ExceptionReset(__pyx_t_3, __pyx_t_2, __pyx_t_1);
      goto __pyx_L1_error;
      __pyx_L26_try_end:;
    }

    /* "serializor/binary/ser.py":1623
 *                 "invalid <'pandas.DatetimeIndex'> name %r." % name
 *             ) from err
 *         items = [DATETIMEINDEX_ID, b"\x01", name_value]             # <<<<<<<<<<<<<<
 * 
 *     # Serialize values
 */
    __pyx_t_4 = PyList_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1623, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_INCREF(__pyx_v_10serializor_6binary_3ser_DATETIMEINDEX_ID);
    __Pyx_GIVEREF(__pyx_v_10serializor_6binary_3ser_DATETIMEINDEX_ID);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_4, 0, __pyx_v_10serializor_6binary_3ser_DATETIMEINDEX_ID)) __PYX_ERR(0, 1623, __pyx_L1_error);
    __Pyx_INCREF(__pyx_kp_b__5);
    __Pyx_GIVEREF(__pyx_kp_b__5);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_4, 1, __pyx_kp_b__5)) __PYX_ERR(0, 1623, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_name_value);
    __Pyx_GIVEREF(__pyx_v_name_value);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_4, 2, __pyx_v_name_value)) __PYX_ERR(0, 1623, __pyx_L1_error);
    __pyx_v_items = ((PyObject*)__pyx_t_4);
    __pyx_t_4 = 0;
  }
  __pyx_L20:;

  /* "serializor/binary/ser.py":1626
 * 
 *     # Serialize values
 *     values = _ser_ndarray_nptime(arr, NDARRAY_DT64_DT)             # <<<<<<<<<<<<<<
 *     # . With timezone
 *     if (tz := obj.tz) is not None:
 */
  __pyx_t_4 = __pyx_v_10serializor_6binary_3ser_NDARRAY_DT64_DT;
  __Pyx_INCREF(__pyx_t_4);
  __pyx_t_6 = __pyx_f_10serializor_6binary_3ser__ser_ndarray_nptime(__pyx_v_arr, ((PyObject*)__pyx_t_4)); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1626, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_v_values = ((PyObject*)__pyx_t_6);
  __pyx_t_6 = 0;

  /* "serializor/binary/ser.py":1628
 *     values = _ser_ndarray_nptime(arr, NDARRAY_DT64_DT)
 *     # . With timezone
 *     if (tz := obj.tz) is not None:             # <<<<<<<<<<<<<<
 *         # . approach 1: timezone name
 *         if set_contains(utils.AVAILABLE_TIMEZONES, tz_name := str(tz)):
 */
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_obj, __pyx_n_s_tz); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1628, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_v_tz = __pyx_t_6;
  __Pyx_INCREF(__pyx_t_6);
  __pyx_t_18 = (__pyx_t_6 != Py_None);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (__pyx_t_18) {

    /* "serializor/binary/ser.py":1630
 *     if (tz := obj.tz) is not None:
 *         # . approach 1: timezone name
 *         if set_contains(utils.AVAILABLE_TIMEZONES, tz_name := str(tz)):             # <<<<<<<<<<<<<<
 *             name = utils.encode_str(tz_name, b"ascii")
 *             items.append(b"\x01")
 */
    __pyx_t_6 = __pyx_v_10serializor_5utils_AVAILABLE_TIMEZONES;
    __Pyx_INCREF(__pyx_t_6);
    __pyx_t_4 = __Pyx_PyObject_Unicode(__pyx_v_tz); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1630, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_v_tz_name = __pyx_t_4;
    __Pyx_INCREF(__pyx_t_4);
    __pyx_t_7 = __pyx_t_4;
    __Pyx_INCREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_18 = PySet_Contains(__pyx_t_6, __pyx_t_7); if (unlikely(__pyx_t_18 == ((int)-1))) __PYX_ERR(0, 1630, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    if (__pyx_t_18) {

      /* "serializor/binary/ser.py":1631
 *         # . approach 1: timezone name
 *         if set_contains(utils.AVAILABLE_TIMEZONES, tz_name := str(tz)):
 *             name = utils.encode_str(tz_name, b"ascii")             # <<<<<<<<<<<<<<
 *             items.append(b"\x01")
 *             items.append(gen_encoded_int(bytes_len(name)))  # type: ignore
 */
      __pyx_t_7 = __pyx_f_10serializor_5utils_encode_str(__pyx_v_tz_name, ((char *)"ascii")); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1631, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF_SET(__pyx_v_name, __pyx_t_7);
      __pyx_t_7 = 0;

      /* "serializor/binary/ser.py":1632
 *         if set_contains(utils.AVAILABLE_TIMEZONES, tz_name := str(tz)):
 *             name = utils.encode_str(tz_name, b"ascii")
 *             items.append(b"\x01")             # <<<<<<<<<<<<<<
 *             items.append(gen_encoded_int(bytes_len(name)))  # type: ignore
 *             items.append(name)
 */
      __pyx_t_20 = __Pyx_PyList_Append(__pyx_v_items, __pyx_kp_b__5); if (unlikely(__pyx_t_20 == ((int)-1))) __PYX_ERR(0, 1632, __pyx_L1_error)

      /* "serializor/binary/ser.py":1633
 *             name = utils.encode_str(tz_name, b"ascii")
 *             items.append(b"\x01")
 *             items.append(gen_encoded_int(bytes_len(name)))  # type: ignore             # <<<<<<<<<<<<<<
 *             items.append(name)
 *             items.append(values)
 */
      __pyx_t_21 = PyBytes_Size(__pyx_v_name); if (unlikely(__pyx_t_21 == ((Py_ssize_t)-1L))) __PYX_ERR(0, 1633, __pyx_L1_error)
      __pyx_t_7 = __pyx_f_10serializor_6binary_3ser_gen_encoded_int(__pyx_t_21); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1633, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_20 = __Pyx_PyList_Append(__pyx_v_items, __pyx_t_7); if (unlikely(__pyx_t_20 == ((int)-1))) __PYX_ERR(0, 1633, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

      /* "serializor/binary/ser.py":1634
 *             items.append(b"\x01")
 *             items.append(gen_encoded_int(bytes_len(name)))  # type: ignore
 *             items.append(name)             # <<<<<<<<<<<<<<
 *             items.append(values)
 *             return b"".join(items)  # exit
 */
      __pyx_t_20 = __Pyx_PyList_Append(__pyx_v_items, __pyx_v_name); if (unlikely(__pyx_t_20 == ((int)-1))) __PYX_ERR(0, 1634, __pyx_L1_error)

      /* "serializor/binary/ser.py":1635
 *             items.append(gen_encoded_int(bytes_len(name)))  # type: ignore
 *             items.append(name)
 *             items.append(values)             # <<<<<<<<<<<<<<
 *             return b"".join(items)  # exit
 *         # . approach 2: timezone offset
 */
      __pyx_t_20 = __Pyx_PyList_Append(__pyx_v_items, __pyx_v_values); if (unlikely(__pyx_t_20 == ((int)-1))) __PYX_ERR(0, 1635, __pyx_L1_error)

      /* "serializor/binary/ser.py":1636
 *             items.append(name)
 *             items.append(values)
 *             return b"".join(items)  # exit             # <<<<<<<<<<<<<<
 *         # . approach 2: timezone offset
 *         if (offset := tz.utcoffset(None)) is not None:
 */
      __Pyx_XDECREF(__pyx_r);
      __pyx_t_7 = __Pyx_PyBytes_Join(__pyx_kp_b__4, __pyx_v_items); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1636, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      if (!(likely(PyBytes_CheckExact(__pyx_t_7))||((__pyx_t_7) == Py_None) || __Pyx_RaiseUnexpectedTypeError("bytes", __pyx_t_7))) __PYX_ERR(0, 1636, __pyx_L1_error)
      __pyx_r = ((PyObject*)__pyx_t_7);
      __pyx_t_7 = 0;
      goto __pyx_L0;

      /* "serializor/binary/ser.py":1630
 *     if (tz := obj.tz) is not None:
 *         # . approach 1: timezone name
 *         if set_contains(utils.AVAILABLE_TIMEZONES, tz_name := str(tz)):             # <<<<<<<<<<<<<<
 *             name = utils.encode_str(tz_name, b"ascii")
 *             items.append(b"\x01")
 */
    }

    /* "serializor/binary/ser.py":1638
 *             return b"".join(items)  # exit
 *         # . approach 2: timezone offset
 *         if (offset := tz.utcoffset(None)) is not None:             # <<<<<<<<<<<<<<
 *             items.append(b"\x02")
 *             items.append(pack_int32(datetime.timedelta_days(offset)))  # type: ignore
 */
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_tz, __pyx_n_s_utcoffset); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1638, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_4 = NULL;
    __pyx_t_22 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
        __pyx_t_22 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_4, Py_None};
      __pyx_t_7 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_22, 1+__pyx_t_22);
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1638, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __pyx_v_offset = __pyx_t_7;
    __Pyx_INCREF(__pyx_t_7);
    __pyx_t_18 = (__pyx_t_7 != Py_None);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    if (__pyx_t_18) {

      /* "serializor/binary/ser.py":1639
 *         # . approach 2: timezone offset
 *         if (offset := tz.utcoffset(None)) is not None:
 *             items.append(b"\x02")             # <<<<<<<<<<<<<<
 *             items.append(pack_int32(datetime.timedelta_days(offset)))  # type: ignore
 *             items.append(pack_int32(datetime.timedelta_seconds(offset)))  # type: ignore
 */
      __pyx_t_20 = __Pyx_PyList_Append(__pyx_v_items, __pyx_kp_b__6); if (unlikely(__pyx_t_20 == ((int)-1))) __PYX_ERR(0, 1639, __pyx_L1_error)

      /* "serializor/binary/ser.py":1640
 *         if (offset := tz.utcoffset(None)) is not None:
 *             items.append(b"\x02")
 *             items.append(pack_int32(datetime.timedelta_days(offset)))  # type: ignore             # <<<<<<<<<<<<<<
 *             items.append(pack_int32(datetime.timedelta_seconds(offset)))  # type: ignore
 *             items.append(pack_int32(datetime.timedelta_microseconds(offset)))  # type: ignore
 */
      __pyx_t_7 = __pyx_f_10serializor_6binary_3ser_pack_int32(__pyx_f_7cpython_8datetime_timedelta_days(__pyx_v_offset)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1640, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_20 = __Pyx_PyList_Append(__pyx_v_items, __pyx_t_7); if (unlikely(__pyx_t_20 == ((int)-1))) __PYX_ERR(0, 1640, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

      /* "serializor/binary/ser.py":1641
 *             items.append(b"\x02")
 *             items.append(pack_int32(datetime.timedelta_days(offset)))  # type: ignore
 *             items.append(pack_int32(datetime.timedelta_seconds(offset)))  # type: ignore             # <<<<<<<<<<<<<<
 *             items.append(pack_int32(datetime.timedelta_microseconds(offset)))  # type: ignore
 *             items.append(values)
 */
      __pyx_t_7 = __pyx_f_10serializor_6binary_3ser_pack_int32(__pyx_f_7cpython_8datetime_timedelta_seconds(__pyx_v_offset)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1641, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_20 = __Pyx_PyList_Append(__pyx_v_items, __pyx_t_7); if (unlikely(__pyx_t_20 == ((int)-1))) __PYX_ERR(0, 1641, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

      /* "serializor/binary/ser.py":1642
 *             items.append(pack_int32(datetime.timedelta_days(offset)))  # type: ignore
 *             items.append(pack_int32(datetime.timedelta_seconds(offset)))  # type: ignore
 *             items.append(pack_int32(datetime.timedelta_microseconds(offset)))  # type: ignore             # <<<<<<<<<<<<<<
 *             items.append(values)
 *             return b"".join(items)  # exit
 */
      __pyx_t_7 = __pyx_f_10serializor_6binary_3ser_pack_int32(__pyx_f_7cpython_8datetime_timedelta_microseconds(__pyx_v_offset)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1642, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_20 = __Pyx_PyList_Append(__pyx_v_items, __pyx_t_7); if (unlikely(__pyx_t_20 == ((int)-1))) __PYX_ERR(0, 1642, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

      /* "serializor/binary/ser.py":1643
 *             items.append(pack_int32(datetime.timedelta_seconds(offset)))  # type: ignore
 *             items.append(pack_int32(datetime.timedelta_microseconds(offset)))  # type: ignore
 *             items.append(values)             # <<<<<<<<<<<<<<
 *             return b"".join(items)  # exit
 *     # . Without timezone / Fallback
 */
      __pyx_t_20 = __Pyx_PyList_Append(__pyx_v_items, __pyx_v_values); if (unlikely(__pyx_t_20 == ((int)-1))) __PYX_ERR(0, 1643, __pyx_L1_error)

      /* "serializor/binary/ser.py":1644
 *             items.append(pack_int32(datetime.timedelta_microseconds(offset)))  # type: ignore
 *             items.append(values)
 *             return b"".join(items)  # exit             # <<<<<<<<<<<<<<
 *     # . Without timezone / Fallback
 *     items.append(b"\x00")
 */
      __Pyx_XDECREF(__pyx_r);
      __pyx_t_7 = __Pyx_PyBytes_Join(__pyx_kp_b__4, __pyx_v_items); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1644, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      if (!(likely(PyBytes_CheckExact(__pyx_t_7))||((__pyx_t_7) == Py_None) || __Pyx_RaiseUnexpectedTypeError("bytes", __pyx_t_7))) __PYX_ERR(0, 1644, __pyx_L1_error)
      __pyx_r = ((PyObject*)__pyx_t_7);
      __pyx_t_7 = 0;
      goto __pyx_L0;

      /* "serializor/binary/ser.py":1638
 *             return b"".join(items)  # exit
 *         # . approach 2: timezone offset
 *         if (offset := tz.utcoffset(None)) is not None:             # <<<<<<<<<<<<<<
 *             items.append(b"\x02")
 *             items.append(pack_int32(datetime.timedelta_days(offset)))  # type: ignore
 */
    }

    /* "serializor/binary/ser.py":1628
 *     values = _ser_ndarray_nptime(arr, NDARRAY_DT64_DT)
 *     # . With timezone
 *     if (tz := obj.tz) is not None:             # <<<<<<<<<<<<<<
 *         # . approach 1: timezone name
 *         if set_contains(utils.AVAILABLE_TIMEZONES, tz_name := str(tz)):
 */
  }

  /* "serializor/binary/ser.py":1646
 *             return b"".join(items)  # exit
 *     # . Without timezone / Fallback
 *     items.append(b"\x00")             # <<<<<<<<<<<<<<
 *     items.append(values)
 *     return b"".join(items)
 */
  __pyx_t_20 = __Pyx_PyList_Append(__pyx_v_items, __pyx_kp_b__7); if (unlikely(__pyx_t_20 == ((int)-1))) __PYX_ERR(0, 1646, __pyx_L1_error)

  /* "serializor/binary/ser.py":1647
 *     # . Without timezone / Fallback
 *     items.append(b"\x00")
 *     items.append(values)             # <<<<<<<<<<<<<<
 *     return b"".join(items)
 * 
 */
  __pyx_t_20 = __Pyx_PyList_Append(__pyx_v_items, __pyx_v_values); if (unlikely(__pyx_t_20 == ((int)-1))) __PYX_ERR(0, 1647, __pyx_L1_error)

  /* "serializor/binary/ser.py":1648
 *     items.append(b"\x00")
 *     items.append(values)
 *     return b"".join(items)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_7 = __Pyx_PyBytes_Join(__pyx_kp_b__4, __pyx_v_items); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1648, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (!(likely(PyBytes_CheckExact(__pyx_t_7))||((__pyx_t_7) == Py_None) || __Pyx_RaiseUnexpectedTypeError("bytes", __pyx_t_7))) __PYX_ERR(0, 1648, __pyx_L1_error)
  __pyx_r = ((PyObject*)__pyx_t_7);
  __pyx_t_7 = 0;
  goto __pyx_L0;

  /* "serializor/binary/ser.py":1582
 * 
 * 
 * @cython.cfunc             # <<<<<<<<<<<<<<
 * @cython.inline(True)
 * def _ser_datetime_index(obj: DatetimeIndex) -> bytes:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("serializor.binary.ser._ser_datetime_index", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_arr);
  __Pyx_XDECREF(__pyx_v_name);
  __Pyx_XDECREF(__pyx_v_err);
  __Pyx_XDECREF(__pyx_v_items);
  __Pyx_XDECREF(__pyx_v_name_value);
  __Pyx_XDECREF(__pyx_v_values);
  __Pyx_XDECREF(__pyx_v_tz);
  __Pyx_XDECREF(__pyx_v_tz_name);
  __Pyx_XDECREF(__pyx_v_offset);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "serializor/binary/ser.py":1651
 * 
 * 
 * @cython.cfunc             # <<<<<<<<<<<<<<
 * @cython.inline(True)
 * def _ser_timedelta_index(obj: TimedeltaIndex) -> bytes:
 */

static PyObject *__pyx_f_10serializor_6binary_3ser__ser_timedelta_index(PyObject *__pyx_v_obj) {
  PyArrayObject *__pyx_v_arr = 0;
  PyObject *__pyx_v_name = 0;
  PyObject *__pyx_v_err = NULL;
  PyObject *__pyx_v_values = NULL;
  PyObject *__pyx_v_items = NULL;
  PyObject *__pyx_v_name_value = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_t_10;
  char const *__pyx_t_11;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  int __pyx_t_18;
  char const *__pyx_t_19;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_ser_timedelta_index", 1);

  /* "serializor/binary/ser.py":1666
 *     """
 *     # Access values and name
 *     try:             # <<<<<<<<<<<<<<
 *         arr: np.ndarray = obj.values
 *         name: object = obj.name
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "serializor/binary/ser.py":1667
 *     # Access values and name
 *     try:
 *         arr: np.ndarray = obj.values             # <<<<<<<<<<<<<<
 *         name: object = obj.name
 *     except Exception as err:
 */
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_obj, __pyx_n_s_values); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1667, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_4);
      if (!(likely(((__pyx_t_4) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_4, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 1667, __pyx_L3_error)
      __pyx_v_arr = ((PyArrayObject *)__pyx_t_4);
      __pyx_t_4 = 0;

      /* "serializor/binary/ser.py":1668
 *     try:
 *         arr: np.ndarray = obj.values
 *         name: object = obj.name             # <<<<<<<<<<<<<<
 *     except Exception as err:
 *         raise TypeError(
 */
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_obj, __pyx_n_s_name); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1668, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_v_name = __pyx_t_4;
      __pyx_t_4 = 0;

      /* "serializor/binary/ser.py":1666
 *     """
 *     # Access values and name
 *     try:             # <<<<<<<<<<<<<<
 *         arr: np.ndarray = obj.values
 *         name: object = obj.name
 */
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "serializor/binary/ser.py":1669
 *         arr: np.ndarray = obj.values
 *         name: object = obj.name
 *     except Exception as err:             # <<<<<<<<<<<<<<
 *         raise TypeError(
 *             "expects <'pandas.TimedeltaIndex'>, got %s." % type(obj)
 */
    __pyx_t_5 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
    if (__pyx_t_5) {
      __Pyx_AddTraceback("serializor.binary.ser._ser_timedelta_index", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_4, &__pyx_t_6, &__pyx_t_7) < 0) __PYX_ERR(0, 1669, __pyx_L5_except_error)
      __Pyx_XGOTREF(__pyx_t_4);
      __Pyx_XGOTREF(__pyx_t_6);
      __Pyx_XGOTREF(__pyx_t_7);
      __Pyx_INCREF(__pyx_t_6);
      __pyx_v_err = __pyx_t_6;
      /*try:*/ {

        /* "serializor/binary/ser.py":1671
 *     except Exception as err:
 *         raise TypeError(
 *             "expects <'pandas.TimedeltaIndex'>, got %s." % type(obj)             # <<<<<<<<<<<<<<
 *         ) from err
 * 
 */
        __pyx_t_8 = __Pyx_PyUnicode_FormatSafe(__pyx_kp_u_expects_pandas_TimedeltaIndex_go, ((PyObject *)Py_TYPE(__pyx_v_obj))); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1671, __pyx_L14_error)
        __Pyx_GOTREF(__pyx_t_8);

        /* "serializor/binary/ser.py":1670
 *         name: object = obj.name
 *     except Exception as err:
 *         raise TypeError(             # <<<<<<<<<<<<<<
 *             "expects <'pandas.TimedeltaIndex'>, got %s." % type(obj)
 *         ) from err
 */
        __pyx_t_9 = __Pyx_PyObject_CallOneArg(__pyx_builtin_TypeError, __pyx_t_8); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1670, __pyx_L14_error)
        __Pyx_GOTREF(__pyx_t_9);
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

        /* "serializor/binary/ser.py":1672
 *         raise TypeError(
 *             "expects <'pandas.TimedeltaIndex'>, got %s." % type(obj)
 *         ) from err             # <<<<<<<<<<<<<<
 * 
 *     # Serialize
 */
        __Pyx_Raise(__pyx_t_9, 0, 0, __pyx_v_err);
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        __PYX_ERR(0, 1670, __pyx_L14_error)
      }

      /* "serializor/binary/ser.py":1669
 *         arr: np.ndarray = obj.values
 *         name: object = obj.name
 *     except Exception as err:             # <<<<<<<<<<<<<<
 *         raise TypeError(
 *             "expects <'pandas.TimedeltaIndex'>, got %s." % type(obj)
 */
      /*finally:*/ {
        __pyx_L14_error:;
        /*exception exit:*/{
          __Pyx_PyThreadState_declare
          __Pyx_PyThreadState_assign
          __pyx_t_12 = 0; __pyx_t_13 = 0; __pyx_t_14 = 0; __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0;
          __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
          __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
          if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_15, &__pyx_t_16, &__pyx_t_17);
          if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_12, &__pyx_t_13, &__pyx_t_14) < 0)) __Pyx_ErrFetch(&__pyx_t_12, &__pyx_t_13, &__pyx_t_14);
          __Pyx_XGOTREF(__pyx_t_12);
          __Pyx_XGOTREF(__pyx_t_13);
          __Pyx_XGOTREF(__pyx_t_14);
          __Pyx_XGOTREF(__pyx_t_15);
          __Pyx_XGOTREF(__pyx_t_16);
          __Pyx_XGOTREF(__pyx_t_17);
          __pyx_t_5 = __pyx_lineno; __pyx_t_10 = __pyx_clineno; __pyx_t_11 = __pyx_filename;
          {
            __Pyx_DECREF(__pyx_v_err); __pyx_v_err = 0;
          }
          if (PY_MAJOR_VERSION >= 3) {
            __Pyx_XGIVEREF(__pyx_t_15);
            __Pyx_XGIVEREF(__pyx_t_16);
            __Pyx_XGIVEREF(__pyx_t_17);
            __Pyx_ExceptionReset(__pyx_t_15, __pyx_t_16, __pyx_t_17);
          }
          __Pyx_XGIVEREF(__pyx_t_12);
          __Pyx_XGIVEREF(__pyx_t_13);
          __Pyx_XGIVEREF(__pyx_t_14);
          __Pyx_ErrRestore(__pyx_t_12, __pyx_t_13, __pyx_t_14);
          __pyx_t_12 = 0; __pyx_t_13 = 0; __pyx_t_14 = 0; __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0;
          __pyx_lineno = __pyx_t_5; __pyx_clineno = __pyx_t_10; __pyx_filename = __pyx_t_11;
          goto __pyx_L5_except_error;
        }
      }
    }
    goto __pyx_L5_except_error;

    /* "serializor/binary/ser.py":1666
 *     """
 *     # Access values and name
 *     try:             # <<<<<<<<<<<<<<
 *         arr: np.ndarray = obj.values
 *         name: object = obj.name
 */
    __pyx_L5_except_error:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L8_try_end:;
  }

  /* "serializor/binary/ser.py":1675
 * 
 *     # Serialize
 *     values = _ser_ndarray_nptime(arr, NDARRAY_TD64_DT)             # <<<<<<<<<<<<<<
 *     if name is None:  # no name
 *         items = [TIMEDELTAINDEX_ID, b"\x00", values]
 */
  __pyx_t_7 = __pyx_v_10serializor_6binary_3ser_NDARRAY_TD64_DT;
  __Pyx_INCREF(__pyx_t_7);
  __pyx_t_6 = __pyx_f_10serializor_6binary_3ser__ser_ndarray_nptime(__pyx_v_arr, ((PyObject*)__pyx_t_7)); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1675, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_v_values = ((PyObject*)__pyx_t_6);
  __pyx_t_6 = 0;

  /* "serializor/binary/ser.py":1676
 *     # Serialize
 *     values = _ser_ndarray_nptime(arr, NDARRAY_TD64_DT)
 *     if name is None:  # no name             # <<<<<<<<<<<<<<
 *         items = [TIMEDELTAINDEX_ID, b"\x00", values]
 *     else:  # has name
 */
  __pyx_t_18 = (__pyx_v_name == Py_None);
  if (__pyx_t_18) {

    /* "serializor/binary/ser.py":1677
 *     values = _ser_ndarray_nptime(arr, NDARRAY_TD64_DT)
 *     if name is None:  # no name
 *         items = [TIMEDELTAINDEX_ID, b"\x00", values]             # <<<<<<<<<<<<<<
 *     else:  # has name
 *         try:
 */
    __pyx_t_6 = PyList_New(3); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1677, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_INCREF(__pyx_v_10serializor_6binary_3ser_TIMEDELTAINDEX_ID);
    __Pyx_GIVEREF(__pyx_v_10serializor_6binary_3ser_TIMEDELTAINDEX_ID);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_6, 0, __pyx_v_10serializor_6binary_3ser_TIMEDELTAINDEX_ID)) __PYX_ERR(0, 1677, __pyx_L1_error);
    __Pyx_INCREF(__pyx_kp_b__7);
    __Pyx_GIVEREF(__pyx_kp_b__7);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_6, 1, __pyx_kp_b__7)) __PYX_ERR(0, 1677, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_values);
    __Pyx_GIVEREF(__pyx_v_values);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_6, 2, __pyx_v_values)) __PYX_ERR(0, 1677, __pyx_L1_error);
    __pyx_v_items = ((PyObject*)__pyx_t_6);
    __pyx_t_6 = 0;

    /* "serializor/binary/ser.py":1676
 *     # Serialize
 *     values = _ser_ndarray_nptime(arr, NDARRAY_TD64_DT)
 *     if name is None:  # no name             # <<<<<<<<<<<<<<
 *         items = [TIMEDELTAINDEX_ID, b"\x00", values]
 *     else:  # has name
 */
    goto __pyx_L20;
  }

  /* "serializor/binary/ser.py":1679
 *         items = [TIMEDELTAINDEX_ID, b"\x00", values]
 *     else:  # has name
 *         try:             # <<<<<<<<<<<<<<
 *             name_value = _ser_common(name)
 *         except Exception as err:
 */
  /*else*/ {
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_3, &__pyx_t_2, &__pyx_t_1);
      __Pyx_XGOTREF(__pyx_t_3);
      __Pyx_XGOTREF(__pyx_t_2);
      __Pyx_XGOTREF(__pyx_t_1);
      /*try:*/ {

        /* "serializor/binary/ser.py":1680
 *     else:  # has name
 *         try:
 *             name_value = _ser_common(name)             # <<<<<<<<<<<<<<
 *         except Exception as err:
 *             raise ValueError(
 */
        __pyx_t_6 = __pyx_f_10serializor_6binary_3ser__ser_common(__pyx_v_name); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1680, __pyx_L21_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_v_name_value = ((PyObject*)__pyx_t_6);
        __pyx_t_6 = 0;

        /* "serializor/binary/ser.py":1679
 *         items = [TIMEDELTAINDEX_ID, b"\x00", values]
 *     else:  # has name
 *         try:             # <<<<<<<<<<<<<<
 *             name_value = _ser_common(name)
 *         except Exception as err:
 */
      }
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L26_try_end;
      __pyx_L21_error:;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;

      /* "serializor/binary/ser.py":1681
 *         try:
 *             name_value = _ser_common(name)
 *         except Exception as err:             # <<<<<<<<<<<<<<
 *             raise ValueError(
 *                 "invalid <'pandas.TimedeltaIndex'> name %r." % name
 */
      __pyx_t_10 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
      if (__pyx_t_10) {
        __Pyx_AddTraceback("serializor.binary.ser._ser_timedelta_index", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_6, &__pyx_t_7, &__pyx_t_4) < 0) __PYX_ERR(0, 1681, __pyx_L23_except_error)
        __Pyx_XGOTREF(__pyx_t_6);
        __Pyx_XGOTREF(__pyx_t_7);
        __Pyx_XGOTREF(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_7);
        __pyx_v_err = __pyx_t_7;
        /*try:*/ {

          /* "serializor/binary/ser.py":1683
 *         except Exception as err:
 *             raise ValueError(
 *                 "invalid <'pandas.TimedeltaIndex'> name %r." % name             # <<<<<<<<<<<<<<
 *             ) from err
 *         items = [TIMEDELTAINDEX_ID, b"\x01", name_value, values]
 */
          __pyx_t_9 = __Pyx_PyUnicode_FormatSafe(__pyx_kp_u_invalid_pandas_TimedeltaIndex_na, __pyx_v_name); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1683, __pyx_L32_error)
          __Pyx_GOTREF(__pyx_t_9);

          /* "serializor/binary/ser.py":1682
 *             name_value = _ser_common(name)
 *         except Exception as err:
 *             raise ValueError(             # <<<<<<<<<<<<<<
 *                 "invalid <'pandas.TimedeltaIndex'> name %r." % name
 *             ) from err
 */
          __pyx_t_8 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_9); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1682, __pyx_L32_error)
          __Pyx_GOTREF(__pyx_t_8);
          __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

          /* "serializor/binary/ser.py":1684
 *             raise ValueError(
 *                 "invalid <'pandas.TimedeltaIndex'> name %r." % name
 *             ) from err             # <<<<<<<<<<<<<<
 *         items = [TIMEDELTAINDEX_ID, b"\x01", name_value, values]
 *     return b"".join(items)
 */
          __Pyx_Raise(__pyx_t_8, 0, 0, __pyx_v_err);
          __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
          __PYX_ERR(0, 1682, __pyx_L32_error)
        }

        /* "serializor/binary/ser.py":1681
 *         try:
 *             name_value = _ser_common(name)
 *         except Exception as err:             # <<<<<<<<<<<<<<
 *             raise ValueError(
 *                 "invalid <'pandas.TimedeltaIndex'> name %r." % name
 */
        /*finally:*/ {
          __pyx_L32_error:;
          /*exception exit:*/{
            __Pyx_PyThreadState_declare
            __Pyx_PyThreadState_assign
            __pyx_t_17 = 0; __pyx_t_16 = 0; __pyx_t_15 = 0; __pyx_t_14 = 0; __pyx_t_13 = 0; __pyx_t_12 = 0;
            __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
            __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
            if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_14, &__pyx_t_13, &__pyx_t_12);
            if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_17, &__pyx_t_16, &__pyx_t_15) < 0)) __Pyx_ErrFetch(&__pyx_t_17, &__pyx_t_16, &__pyx_t_15);
            __Pyx_XGOTREF(__pyx_t_17);
            __Pyx_XGOTREF(__pyx_t_16);
            __Pyx_XGOTREF(__pyx_t_15);
            __Pyx_XGOTREF(__pyx_t_14);
            __Pyx_XGOTREF(__pyx_t_13);
            __Pyx_XGOTREF(__pyx_t_12);
            __pyx_t_10 = __pyx_lineno; __pyx_t_5 = __pyx_clineno; __pyx_t_19 = __pyx_filename;
            {
              __Pyx_DECREF(__pyx_v_err); __pyx_v_err = 0;
            }
            if (PY_MAJOR_VERSION >= 3) {
              __Pyx_XGIVEREF(__pyx_t_14);
              __Pyx_XGIVEREF(__pyx_t_13);
              __Pyx_XGIVEREF(__pyx_t_12);
              __Pyx_ExceptionReset(__pyx_t_14, __pyx_t_13, __pyx_t_12);
            }
            __Pyx_XGIVEREF(__pyx_t_17);
            __Pyx_XGIVEREF(__pyx_t_16);
            __Pyx_XGIVEREF(__pyx_t_15);
            __Pyx_ErrRestore(__pyx_t_17, __pyx_t_16, __pyx_t_15);
            __pyx_t_17 = 0; __pyx_t_16 = 0; __pyx_t_15 = 0; __pyx_t_14 = 0; __pyx_t_13 = 0; __pyx_t_12 = 0;
            __pyx_lineno = __pyx_t_10; __pyx_clineno = __pyx_t_5; __pyx_filename = __pyx_t_19;
            goto __pyx_L23_except_error;
          }
        }
      }
      goto __pyx_L23_except_error;

      /* "serializor/binary/ser.py":1679
 *         items = [TIMEDELTAINDEX_ID, b"\x00", values]
 *     else:  # has name
 *         try:             # <<<<<<<<<<<<<<
 *             name_value = _ser_common(name)
 *         except Exception as err:
 */
      __pyx_L23_except_error:;
      __Pyx_XGIVEREF(__pyx_t_3);
      __Pyx_XGIVEREF(__pyx_t_2);
      __Pyx_XGIVEREF(__pyx_t_1);
      __Pyx_ExceptionReset(__pyx_t_3, __pyx_t_2, __pyx_t_1);
      goto __pyx_L1_error;
      __pyx_L26_try_end:;
    }

    /* "serializor/binary/ser.py":1685
 *                 "invalid <'pandas.TimedeltaIndex'> name %r." % name
 *             ) from err
 *         items = [TIMEDELTAINDEX_ID, b"\x01", name_value, values]             # <<<<<<<<<<<<<<
 *     return b"".join(items)
 * 
 */
    __pyx_t_4 = PyList_New(4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1685, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_INCREF(__pyx_v_10serializor_6binary_3ser_TIMEDELTAINDEX_ID);
    __Pyx_GIVEREF(__pyx_v_10serializor_6binary_3ser_TIMEDELTAINDEX_ID);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_4, 0, __pyx_v_10serializor_6binary_3ser_TIMEDELTAINDEX_ID)) __PYX_ERR(0, 1685, __pyx_L1_error);
    __Pyx_INCREF(__pyx_kp_b__5);
    __Pyx_GIVEREF(__pyx_kp_b__5);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_4, 1, __pyx_kp_b__5)) __PYX_ERR(0, 1685, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_name_value);
    __Pyx_GIVEREF(__pyx_v_name_value);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_4, 2, __pyx_v_name_value)) __PYX_ERR(0, 1685, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_values);
    __Pyx_GIVEREF(__pyx_v_values);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_4, 3, __pyx_v_values)) __PYX_ERR(0, 1685, __pyx_L1_error);
    __pyx_v_items = ((PyObject*)__pyx_t_4);
    __pyx_t_4 = 0;
  }
  __pyx_L20:;

  /* "serializor/binary/ser.py":1686
 *             ) from err
 *         items = [TIMEDELTAINDEX_ID, b"\x01", name_value, values]
 *     return b"".join(items)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_4 = __Pyx_PyBytes_Join(__pyx_kp_b__4, __pyx_v_items); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1686, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (!(likely(PyBytes_CheckExact(__pyx_t_4))||((__pyx_t_4) == Py_None) || __Pyx_RaiseUnexpectedTypeError("bytes", __pyx_t_4))) __PYX_ERR(0, 1686, __pyx_L1_error)
  __pyx_r = ((PyObject*)__pyx_t_4);
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "serializor/binary/ser.py":1651
 * 
 * 
 * @cython.cfunc             # <<<<<<<<<<<<<<
 * @cython.inline(True)
 * def _ser_timedelta_index(obj: TimedeltaIndex) -> bytes:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("serializor.binary.ser._ser_timedelta_index", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_arr);
  __Pyx_XDECREF(__pyx_v_name);
  __Pyx_XDECREF(__pyx_v_err);
  __Pyx_XDECREF(__pyx_v_values);
  __Pyx_XDECREF(__pyx_v_items);
  __Pyx_XDECREF(__pyx_v_name_value);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "serializor/binary/ser.py":1689
 * 
 * 
 * @cython.cfunc             # <<<<<<<<<<<<<<
 * @cython.inline(True)
 * def _ser_dataframe(obj: DataFrame) -> bytes:
 */

static PyObject *__pyx_f_10serializor_6binary_3ser__ser_dataframe(PyObject *__pyx_v_obj) {
  PyObject *__pyx_v_cols = 0;
  Py_ssize_t __pyx_v_cols_num;
  PyObject *__pyx_v_cols_value = NULL;
  PyObject *__pyx_v_err = NULL;
  Py_ssize_t __pyx_v_rows_num;
  PyObject *__pyx_v_items = NULL;
  CYTHON_UNUSED PyObject *__pyx_v__ = NULL;
  PyObject *__pyx_v_col = NULL;
  PyArrayObject *__pyx_v_arr = 0;
  char __pyx_v_npy_kind;
  PyObject *__pyx_v_values = NULL;
  PyObject *__pyx_v_name = NULL;
  PyObject *__pyx_v_tz = NULL;
  PyObject *__pyx_v_tz_name = NULL;
  PyObject *__pyx_v_offset = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_t_11;
  char const *__pyx_t_12;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  Py_ssize_t __pyx_t_19;
  Py_ssize_t __pyx_t_20;
  PyObject *__pyx_t_21 = NULL;
  int __pyx_t_22;
  char const *__pyx_t_23;
  char __pyx_t_24;
  int __pyx_t_25;
  Py_ssize_t __pyx_t_26;
  unsigned int __pyx_t_27;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_ser_dataframe", 1);

  /* "serializor/binary/ser.py":1707
 *     """
 *     # Serialize columns name
 *     cols: list = list(obj.columns)             # <<<<<<<<<<<<<<
 *     cols_num: cython.Py_ssize_t = list_len(cols)
 *     # . empty w/o columns
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_obj, __pyx_n_s_columns); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1707, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PySequence_ListKeepNew(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1707, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_cols = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "serializor/binary/ser.py":1708
 *     # Serialize columns name
 *     cols: list = list(obj.columns)
 *     cols_num: cython.Py_ssize_t = list_len(cols)             # <<<<<<<<<<<<<<
 *     # . empty w/o columns
 *     if cols_num == 0:
 */
  __pyx_v_cols_num = PyList_GET_SIZE(__pyx_v_cols);

  /* "serializor/binary/ser.py":1710
 *     cols_num: cython.Py_ssize_t = list_len(cols)
 *     # . empty w/o columns
 *     if cols_num == 0:             # <<<<<<<<<<<<<<
 *         return gen_header(identifier.DATAFRAME, 0)  # type: ignore
 *     try:
 */
  __pyx_t_3 = (__pyx_v_cols_num == 0);
  if (__pyx_t_3) {

    /* "serializor/binary/ser.py":1711
 *     # . empty w/o columns
 *     if cols_num == 0:
 *         return gen_header(identifier.DATAFRAME, 0)  # type: ignore             # <<<<<<<<<<<<<<
 *     try:
 *         cols_value = _ser_list(cols)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = __pyx_f_10serializor_6binary_3ser_gen_header(__pyx_v_10serializor_10identifier_DATAFRAME, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1711, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_r = ((PyObject*)__pyx_t_2);
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "serializor/binary/ser.py":1710
 *     cols_num: cython.Py_ssize_t = list_len(cols)
 *     # . empty w/o columns
 *     if cols_num == 0:             # <<<<<<<<<<<<<<
 *         return gen_header(identifier.DATAFRAME, 0)  # type: ignore
 *     try:
 */
  }

  /* "serializor/binary/ser.py":1712
 *     if cols_num == 0:
 *         return gen_header(identifier.DATAFRAME, 0)  # type: ignore
 *     try:             # <<<<<<<<<<<<<<
 *         cols_value = _ser_list(cols)
 *     except Exception as err:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_4, &__pyx_t_5, &__pyx_t_6);
    __Pyx_XGOTREF(__pyx_t_4);
    __Pyx_XGOTREF(__pyx_t_5);
    __Pyx_XGOTREF(__pyx_t_6);
    /*try:*/ {

      /* "serializor/binary/ser.py":1713
 *         return gen_header(identifier.DATAFRAME, 0)  # type: ignore
 *     try:
 *         cols_value = _ser_list(cols)             # <<<<<<<<<<<<<<
 *     except Exception as err:
 *         raise ValueError("invalid <'pandas.DataFrame'> columns:\n%s" % cols) from err
 */
      __pyx_t_2 = __pyx_f_10serializor_6binary_3ser__ser_list(__pyx_v_cols); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1713, __pyx_L4_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_v_cols_value = ((PyObject*)__pyx_t_2);
      __pyx_t_2 = 0;

      /* "serializor/binary/ser.py":1712
 *     if cols_num == 0:
 *         return gen_header(identifier.DATAFRAME, 0)  # type: ignore
 *     try:             # <<<<<<<<<<<<<<
 *         cols_value = _ser_list(cols)
 *     except Exception as err:
 */
    }
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    goto __pyx_L9_try_end;
    __pyx_L4_error:;
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "serializor/binary/ser.py":1714
 *     try:
 *         cols_value = _ser_list(cols)
 *     except Exception as err:             # <<<<<<<<<<<<<<
 *         raise ValueError("invalid <'pandas.DataFrame'> columns:\n%s" % cols) from err
 *     rows_num: cython.Py_ssize_t = len(obj.index)
 */
    __pyx_t_7 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
    if (__pyx_t_7) {
      __Pyx_AddTraceback("serializor.binary.ser._ser_dataframe", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_2, &__pyx_t_1, &__pyx_t_8) < 0) __PYX_ERR(0, 1714, __pyx_L6_except_error)
      __Pyx_XGOTREF(__pyx_t_2);
      __Pyx_XGOTREF(__pyx_t_1);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_INCREF(__pyx_t_1);
      __pyx_v_err = __pyx_t_1;
      /*try:*/ {

        /* "serializor/binary/ser.py":1715
 *         cols_value = _ser_list(cols)
 *     except Exception as err:
 *         raise ValueError("invalid <'pandas.DataFrame'> columns:\n%s" % cols) from err             # <<<<<<<<<<<<<<
 *     rows_num: cython.Py_ssize_t = len(obj.index)
 *     # . empty w/o rows
 */
        __pyx_t_9 = PyUnicode_Format(__pyx_kp_u_invalid_pandas_DataFrame_columns, __pyx_v_cols); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1715, __pyx_L15_error)
        __Pyx_GOTREF(__pyx_t_9);
        __pyx_t_10 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_9); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1715, __pyx_L15_error)
        __Pyx_GOTREF(__pyx_t_10);
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        __Pyx_Raise(__pyx_t_10, 0, 0, __pyx_v_err);
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        __PYX_ERR(0, 1715, __pyx_L15_error)
      }

      /* "serializor/binary/ser.py":1714
 *     try:
 *         cols_value = _ser_list(cols)
 *     except Exception as err:             # <<<<<<<<<<<<<<
 *         raise ValueError("invalid <'pandas.DataFrame'> columns:\n%s" % cols) from err
 *     rows_num: cython.Py_ssize_t = len(obj.index)
 */
      /*finally:*/ {
        __pyx_L15_error:;
        /*exception exit:*/{
          __Pyx_PyThreadState_declare
          __Pyx_PyThreadState_assign
          __pyx_t_13 = 0; __pyx_t_14 = 0; __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0;
          __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
          __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
          if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_16, &__pyx_t_17, &__pyx_t_18);
          if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_13, &__pyx_t_14, &__pyx_t_15) < 0)) __Pyx_ErrFetch(&__pyx_t_13, &__pyx_t_14, &__pyx_t_15);
          __Pyx_XGOTREF(__pyx_t_13);
          __Pyx_XGOTREF(__pyx_t_14);
          __Pyx_XGOTREF(__pyx_t_15);
          __Pyx_XGOTREF(__pyx_t_16);
          __Pyx_XGOTREF(__pyx_t_17);
          __Pyx_XGOTREF(__pyx_t_18);
          __pyx_t_7 = __pyx_lineno; __pyx_t_11 = __pyx_clineno; __pyx_t_12 = __pyx_filename;
          {
            __Pyx_DECREF(__pyx_v_err); __pyx_v_err = 0;
          }
          if (PY_MAJOR_VERSION >= 3) {
            __Pyx_XGIVEREF(__pyx_t_16);
            __Pyx_XGIVEREF(__pyx_t_17);
            __Pyx_XGIVEREF(__pyx_t_18);
            __Pyx_ExceptionReset(__pyx_t_16, __pyx_t_17, __pyx_t_18);
          }
          __Pyx_XGIVEREF(__pyx_t_13);
          __Pyx_XGIVEREF(__pyx_t_14);
          __Pyx_XGIVEREF(__pyx_t_15);
          __Pyx_ErrRestore(__pyx_t_13, __pyx_t_14, __pyx_t_15);
          __pyx_t_13 = 0; __pyx_t_14 = 0; __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0;
          __pyx_lineno = __pyx_t_7; __pyx_clineno = __pyx_t_11; __pyx_filename = __pyx_t_12;
          goto __pyx_L6_except_error;
        }
      }
    }
    goto __pyx_L6_except_error;

    /* "serializor/binary/ser.py":1712
 *     if cols_num == 0:
 *         return gen_header(identifier.DATAFRAME, 0)  # type: ignore
 *     try:             # <<<<<<<<<<<<<<
 *         cols_value = _ser_list(cols)
 *     except Exception as err:
 */
    __pyx_L6_except_error:;
    __Pyx_XGIVEREF(__pyx_t_4);
    __Pyx_XGIVEREF(__pyx_t_5);
    __Pyx_XGIVEREF(__pyx_t_6);
    __Pyx_ExceptionReset(__pyx_t_4, __pyx_t_5, __pyx_t_6);
    goto __pyx_L1_error;
    __pyx_L9_try_end:;
  }

  /* "serializor/binary/ser.py":1716
 *     except Exception as err:
 *         raise ValueError("invalid <'pandas.DataFrame'> columns:\n%s" % cols) from err
 *     rows_num: cython.Py_ssize_t = len(obj.index)             # <<<<<<<<<<<<<<
 *     # . empty w/o rows
 *     if rows_num == 0:
 */
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_v_obj, __pyx_n_s_index); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1716, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_19 = PyObject_Length(__pyx_t_8); if (unlikely(__pyx_t_19 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1716, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_v_rows_num = __pyx_t_19;

  /* "serializor/binary/ser.py":1718
 *     rows_num: cython.Py_ssize_t = len(obj.index)
 *     # . empty w/o rows
 *     if rows_num == 0:             # <<<<<<<<<<<<<<
 *         return b"".join([DATAFRAME_ID, b"\x01", cols_value, b"\x00"])
 * 
 */
  __pyx_t_3 = (__pyx_v_rows_num == 0);
  if (__pyx_t_3) {

    /* "serializor/binary/ser.py":1719
 *     # . empty w/o rows
 *     if rows_num == 0:
 *         return b"".join([DATAFRAME_ID, b"\x01", cols_value, b"\x00"])             # <<<<<<<<<<<<<<
 * 
 *     # Serialize columns values
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_8 = PyList_New(4); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1719, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_INCREF(__pyx_v_10serializor_6binary_3ser_DATAFRAME_ID);
    __Pyx_GIVEREF(__pyx_v_10serializor_6binary_3ser_DATAFRAME_ID);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_8, 0, __pyx_v_10serializor_6binary_3ser_DATAFRAME_ID)) __PYX_ERR(0, 1719, __pyx_L1_error);
    __Pyx_INCREF(__pyx_kp_b__5);
    __Pyx_GIVEREF(__pyx_kp_b__5);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_8, 1, __pyx_kp_b__5)) __PYX_ERR(0, 1719, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_cols_value);
    __Pyx_GIVEREF(__pyx_v_cols_value);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_8, 2, __pyx_v_cols_value)) __PYX_ERR(0, 1719, __pyx_L1_error);
    __Pyx_INCREF(__pyx_kp_b__7);
    __Pyx_GIVEREF(__pyx_kp_b__7);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_8, 3, __pyx_kp_b__7)) __PYX_ERR(0, 1719, __pyx_L1_error);
    __pyx_t_1 = __Pyx_PyBytes_Join(__pyx_kp_b__4, __pyx_t_8); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1719, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    if (!(likely(PyBytes_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("bytes", __pyx_t_1))) __PYX_ERR(0, 1719, __pyx_L1_error)
    __pyx_r = ((PyObject*)__pyx_t_1);
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "serializor/binary/ser.py":1718
 *     rows_num: cython.Py_ssize_t = len(obj.index)
 *     # . empty w/o rows
 *     if rows_num == 0:             # <<<<<<<<<<<<<<
 *         return b"".join([DATAFRAME_ID, b"\x01", cols_value, b"\x00"])
 * 
 */
  }

  /* "serializor/binary/ser.py":1722
 * 
 *     # Serialize columns values
 *     items = [DATAFRAME_ID, b"\x01", cols_value, b"\x01"]             # <<<<<<<<<<<<<<
 *     for _, col in obj.items():
 *         try:
 */
  __pyx_t_1 = PyList_New(4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1722, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_10serializor_6binary_3ser_DATAFRAME_ID);
  __Pyx_GIVEREF(__pyx_v_10serializor_6binary_3ser_DATAFRAME_ID);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_v_10serializor_6binary_3ser_DATAFRAME_ID)) __PYX_ERR(0, 1722, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_b__5);
  __Pyx_GIVEREF(__pyx_kp_b__5);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 1, __pyx_kp_b__5)) __PYX_ERR(0, 1722, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_cols_value);
  __Pyx_GIVEREF(__pyx_v_cols_value);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 2, __pyx_v_cols_value)) __PYX_ERR(0, 1722, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_b__5);
  __Pyx_GIVEREF(__pyx_kp_b__5);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 3, __pyx_kp_b__5)) __PYX_ERR(0, 1722, __pyx_L1_error);
  __pyx_v_items = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "serializor/binary/ser.py":1723
 *     # Serialize columns values
 *     items = [DATAFRAME_ID, b"\x01", cols_value, b"\x01"]
 *     for _, col in obj.items():             # <<<<<<<<<<<<<<
 *         try:
 *             arr: np.ndarray = col.values
 */
  __pyx_t_19 = 0;
  if (unlikely(__pyx_v_obj == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "items");
    __PYX_ERR(0, 1723, __pyx_L1_error)
  }
  __pyx_t_8 = __Pyx_dict_iterator(__pyx_v_obj, 0, __pyx_n_s_items, (&__pyx_t_20), (&__pyx_t_11)); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1723, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_1);
  __pyx_t_1 = __pyx_t_8;
  __pyx_t_8 = 0;
  while (1) {
    __pyx_t_7 = __Pyx_dict_iter_next(__pyx_t_1, __pyx_t_20, &__pyx_t_19, &__pyx_t_8, &__pyx_t_2, NULL, __pyx_t_11);
    if (unlikely(__pyx_t_7 == 0)) break;
    if (unlikely(__pyx_t_7 == -1)) __PYX_ERR(0, 1723, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_XDECREF_SET(__pyx_v__, __pyx_t_8);
    __pyx_t_8 = 0;
    __Pyx_XDECREF_SET(__pyx_v_col, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "serializor/binary/ser.py":1724
 *     items = [DATAFRAME_ID, b"\x01", cols_value, b"\x01"]
 *     for _, col in obj.items():
 *         try:             # <<<<<<<<<<<<<<
 *             arr: np.ndarray = col.values
 *         except Exception as err:
 */
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_6, &__pyx_t_5, &__pyx_t_4);
      __Pyx_XGOTREF(__pyx_t_6);
      __Pyx_XGOTREF(__pyx_t_5);
      __Pyx_XGOTREF(__pyx_t_4);
      /*try:*/ {

        /* "serializor/binary/ser.py":1725
 *     for _, col in obj.items():
 *         try:
 *             arr: np.ndarray = col.values             # <<<<<<<<<<<<<<
 *         except Exception as err:
 *             raise ValueError(
 */
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_col, __pyx_n_s_values); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1725, __pyx_L24_error)
        __Pyx_GOTREF(__pyx_t_2);
        if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 1725, __pyx_L24_error)
        __Pyx_XDECREF_SET(__pyx_v_arr, ((PyArrayObject *)__pyx_t_2));
        __pyx_t_2 = 0;

        /* "serializor/binary/ser.py":1724
 *     items = [DATAFRAME_ID, b"\x01", cols_value, b"\x01"]
 *     for _, col in obj.items():
 *         try:             # <<<<<<<<<<<<<<
 *             arr: np.ndarray = col.values
 *         except Exception as err:
 */
      }
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      goto __pyx_L31_try_end;
      __pyx_L24_error:;
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;

      /* "serializor/binary/ser.py":1726
 *         try:
 *             arr: np.ndarray = col.values
 *         except Exception as err:             # <<<<<<<<<<<<<<
 *             raise ValueError(
 *                 "invalid <'pandas.DataFrame'>, cannot access "
 */
      __pyx_t_7 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
      if (__pyx_t_7) {
        __Pyx_AddTraceback("serializor.binary.ser._ser_dataframe", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_2, &__pyx_t_8, &__pyx_t_10) < 0) __PYX_ERR(0, 1726, __pyx_L26_except_error)
        __Pyx_XGOTREF(__pyx_t_2);
        __Pyx_XGOTREF(__pyx_t_8);
        __Pyx_XGOTREF(__pyx_t_10);
        __Pyx_INCREF(__pyx_t_8);
        __pyx_v_err = __pyx_t_8;
        /*try:*/ {

          /* "serializor/binary/ser.py":1729
 *             raise ValueError(
 *                 "invalid <'pandas.DataFrame'>, cannot access "
 *                 "column's underlying 'ndarray.values':\n%s" % col             # <<<<<<<<<<<<<<
 *             ) from err
 *         npy_kind: cython.char = arr.descr.kind
 */
          __pyx_t_9 = __Pyx_PyUnicode_FormatSafe(__pyx_kp_u_invalid_pandas_DataFrame_cannot, __pyx_v_col); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1729, __pyx_L37_error)
          __Pyx_GOTREF(__pyx_t_9);

          /* "serializor/binary/ser.py":1727
 *             arr: np.ndarray = col.values
 *         except Exception as err:
 *             raise ValueError(             # <<<<<<<<<<<<<<
 *                 "invalid <'pandas.DataFrame'>, cannot access "
 *                 "column's underlying 'ndarray.values':\n%s" % col
 */
          __pyx_t_21 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_9); if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 1727, __pyx_L37_error)
          __Pyx_GOTREF(__pyx_t_21);
          __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

          /* "serializor/binary/ser.py":1730
 *                 "invalid <'pandas.DataFrame'>, cannot access "
 *                 "column's underlying 'ndarray.values':\n%s" % col
 *             ) from err             # <<<<<<<<<<<<<<
 *         npy_kind: cython.char = arr.descr.kind
 *         # . ndarray[object]
 */
          __Pyx_Raise(__pyx_t_21, 0, 0, __pyx_v_err);
          __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
          __PYX_ERR(0, 1727, __pyx_L37_error)
        }

        /* "serializor/binary/ser.py":1726
 *         try:
 *             arr: np.ndarray = col.values
 *         except Exception as err:             # <<<<<<<<<<<<<<
 *             raise ValueError(
 *                 "invalid <'pandas.DataFrame'>, cannot access "
 */
        /*finally:*/ {
          __pyx_L37_error:;
          /*exception exit:*/{
            __Pyx_PyThreadState_declare
            __Pyx_PyThreadState_assign
            __pyx_t_18 = 0; __pyx_t_17 = 0; __pyx_t_16 = 0; __pyx_t_15 = 0; __pyx_t_14 = 0; __pyx_t_13 = 0;
            __Pyx_XDECREF(__pyx_t_21); __pyx_t_21 = 0;
            __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
            if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_15, &__pyx_t_14, &__pyx_t_13);
            if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_18, &__pyx_t_17, &__pyx_t_16) < 0)) __Pyx_ErrFetch(&__pyx_t_18, &__pyx_t_17, &__pyx_t_16);
            __Pyx_XGOTREF(__pyx_t_18);
            __Pyx_XGOTREF(__pyx_t_17);
            __Pyx_XGOTREF(__pyx_t_16);
            __Pyx_XGOTREF(__pyx_t_15);
            __Pyx_XGOTREF(__pyx_t_14);
            __Pyx_XGOTREF(__pyx_t_13);
            __pyx_t_7 = __pyx_lineno; __pyx_t_22 = __pyx_clineno; __pyx_t_23 = __pyx_filename;
            {
              __Pyx_DECREF(__pyx_v_err); __pyx_v_err = 0;
            }
            if (PY_MAJOR_VERSION >= 3) {
              __Pyx_XGIVEREF(__pyx_t_15);
              __Pyx_XGIVEREF(__pyx_t_14);
              __Pyx_XGIVEREF(__pyx_t_13);
              __Pyx_ExceptionReset(__pyx_t_15, __pyx_t_14, __pyx_t_13);
            }
            __Pyx_XGIVEREF(__pyx_t_18);
            __Pyx_XGIVEREF(__pyx_t_17);
            __Pyx_XGIVEREF(__pyx_t_16);
            __Pyx_ErrRestore(__pyx_t_18, __pyx_t_17, __pyx_t_16);
            __pyx_t_18 = 0; __pyx_t_17 = 0; __pyx_t_16 = 0; __pyx_t_15 = 0; __pyx_t_14 = 0; __pyx_t_13 = 0;
            __pyx_lineno = __pyx_t_7; __pyx_clineno = __pyx_t_22; __pyx_filename = __pyx_t_23;
            goto __pyx_L26_except_error;
          }
        }
      }
      goto __pyx_L26_except_error;

      /* "serializor/binary/ser.py":1724
 *     items = [DATAFRAME_ID, b"\x01", cols_value, b"\x01"]
 *     for _, col in obj.items():
 *         try:             # <<<<<<<<<<<<<<
 *             arr: np.ndarray = col.values
 *         except Exception as err:
 */
      __pyx_L26_except_error:;
      __Pyx_XGIVEREF(__pyx_t_6);
      __Pyx_XGIVEREF(__pyx_t_5);
      __Pyx_XGIVEREF(__pyx_t_4);
      __Pyx_ExceptionReset(__pyx_t_6, __pyx_t_5, __pyx_t_4);
      goto __pyx_L1_error;
      __pyx_L31_try_end:;
    }

    /* "serializor/binary/ser.py":1731
 *                 "column's underlying 'ndarray.values':\n%s" % col
 *             ) from err
 *         npy_kind: cython.char = arr.descr.kind             # <<<<<<<<<<<<<<
 *         # . ndarray[object]
 *         if npy_kind == identifier.NDARRAY_OBJECT:
 */
    __pyx_t_10 = ((PyObject *)__pyx_f_5numpy_7ndarray_5descr_descr(__pyx_v_arr)); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1731, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_t_24 = ((PyArray_Descr *)__pyx_t_10)->kind;
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __pyx_v_npy_kind = __pyx_t_24;

    /* "serializor/binary/ser.py":1733
 *         npy_kind: cython.char = arr.descr.kind
 *         # . ndarray[object]
 *         if npy_kind == identifier.NDARRAY_OBJECT:             # <<<<<<<<<<<<<<
 *             items.append(_ser_ndarray_object(arr, NDARRAY_OBJECT_DT))
 *         # . ndarray[int]
 */
    __pyx_t_3 = (__pyx_v_npy_kind == __pyx_v_10serializor_10identifier_NDARRAY_OBJECT);
    if (__pyx_t_3) {

      /* "serializor/binary/ser.py":1734
 *         # . ndarray[object]
 *         if npy_kind == identifier.NDARRAY_OBJECT:
 *             items.append(_ser_ndarray_object(arr, NDARRAY_OBJECT_DT))             # <<<<<<<<<<<<<<
 *         # . ndarray[int]
 *         elif npy_kind == identifier.NDARRAY_INT:
 */
      __pyx_t_10 = __pyx_v_10serializor_6binary_3ser_NDARRAY_OBJECT_DT;
      __Pyx_INCREF(__pyx_t_10);
      __pyx_t_8 = __pyx_f_10serializor_6binary_3ser__ser_ndarray_object(__pyx_v_arr, ((PyObject*)__pyx_t_10)); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1734, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __pyx_t_25 = __Pyx_PyList_Append(__pyx_v_items, __pyx_t_8); if (unlikely(__pyx_t_25 == ((int)-1))) __PYX_ERR(0, 1734, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

      /* "serializor/binary/ser.py":1733
 *         npy_kind: cython.char = arr.descr.kind
 *         # . ndarray[object]
 *         if npy_kind == identifier.NDARRAY_OBJECT:             # <<<<<<<<<<<<<<
 *             items.append(_ser_ndarray_object(arr, NDARRAY_OBJECT_DT))
 *         # . ndarray[int]
 */
      goto __pyx_L43;
    }

    /* "serializor/binary/ser.py":1736
 *             items.append(_ser_ndarray_object(arr, NDARRAY_OBJECT_DT))
 *         # . ndarray[int]
 *         elif npy_kind == identifier.NDARRAY_INT:             # <<<<<<<<<<<<<<
 *             items.append(_ser_ndarray_numeric(arr, NDARRAY_INT_DT, arr.descr.type_num))
 *         # . ndarray[uint]
 */
    __pyx_t_3 = (__pyx_v_npy_kind == __pyx_v_10serializor_10identifier_NDARRAY_INT);
    if (__pyx_t_3) {

      /* "serializor/binary/ser.py":1737
 *         # . ndarray[int]
 *         elif npy_kind == identifier.NDARRAY_INT:
 *             items.append(_ser_ndarray_numeric(arr, NDARRAY_INT_DT, arr.descr.type_num))             # <<<<<<<<<<<<<<
 *         # . ndarray[uint]
 *         elif npy_kind == identifier.NDARRAY_UINT:
 */
      __pyx_t_8 = __pyx_v_10serializor_6binary_3ser_NDARRAY_INT_DT;
      __Pyx_INCREF(__pyx_t_8);
      __pyx_t_10 = ((PyObject *)__pyx_f_5numpy_7ndarray_5descr_descr(__pyx_v_arr)); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1737, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __pyx_t_2 = __pyx_f_10serializor_6binary_3ser__ser_ndarray_numeric(__pyx_v_arr, ((PyObject*)__pyx_t_8), ((PyArray_Descr *)__pyx_t_10)->type_num); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1737, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __pyx_t_25 = __Pyx_PyList_Append(__pyx_v_items, __pyx_t_2); if (unlikely(__pyx_t_25 == ((int)-1))) __PYX_ERR(0, 1737, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "serializor/binary/ser.py":1736
 *             items.append(_ser_ndarray_object(arr, NDARRAY_OBJECT_DT))
 *         # . ndarray[int]
 *         elif npy_kind == identifier.NDARRAY_INT:             # <<<<<<<<<<<<<<
 *             items.append(_ser_ndarray_numeric(arr, NDARRAY_INT_DT, arr.descr.type_num))
 *         # . ndarray[uint]
 */
      goto __pyx_L43;
    }

    /* "serializor/binary/ser.py":1739
 *             items.append(_ser_ndarray_numeric(arr, NDARRAY_INT_DT, arr.descr.type_num))
 *         # . ndarray[uint]
 *         elif npy_kind == identifier.NDARRAY_UINT:             # <<<<<<<<<<<<<<
 *             items.append(_ser_ndarray_numeric(arr, NDARRAY_UINT_DT, arr.descr.type_num))
 *         # . ndarray[float]
 */
    __pyx_t_3 = (__pyx_v_npy_kind == __pyx_v_10serializor_10identifier_NDARRAY_UINT);
    if (__pyx_t_3) {

      /* "serializor/binary/ser.py":1740
 *         # . ndarray[uint]
 *         elif npy_kind == identifier.NDARRAY_UINT:
 *             items.append(_ser_ndarray_numeric(arr, NDARRAY_UINT_DT, arr.descr.type_num))             # <<<<<<<<<<<<<<
 *         # . ndarray[float]
 *         elif npy_kind == identifier.NDARRAY_FLOAT:
 */
      __pyx_t_2 = __pyx_v_10serializor_6binary_3ser_NDARRAY_UINT_DT;
      __Pyx_INCREF(__pyx_t_2);
      __pyx_t_10 = ((PyObject *)__pyx_f_5numpy_7ndarray_5descr_descr(__pyx_v_arr)); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1740, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __pyx_t_8 = __pyx_f_10serializor_6binary_3ser__ser_ndarray_numeric(__pyx_v_arr, ((PyObject*)__pyx_t_2), ((PyArray_Descr *)__pyx_t_10)->type_num); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1740, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __pyx_t_25 = __Pyx_PyList_Append(__pyx_v_items, __pyx_t_8); if (unlikely(__pyx_t_25 == ((int)-1))) __PYX_ERR(0, 1740, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

      /* "serializor/binary/ser.py":1739
 *             items.append(_ser_ndarray_numeric(arr, NDARRAY_INT_DT, arr.descr.type_num))
 *         # . ndarray[uint]
 *         elif npy_kind == identifier.NDARRAY_UINT:             # <<<<<<<<<<<<<<
 *             items.append(_ser_ndarray_numeric(arr, NDARRAY_UINT_DT, arr.descr.type_num))
 *         # . ndarray[float]
 */
      goto __pyx_L43;
    }

    /* "serializor/binary/ser.py":1742
 *             items.append(_ser_ndarray_numeric(arr, NDARRAY_UINT_DT, arr.descr.type_num))
 *         # . ndarray[float]
 *         elif npy_kind == identifier.NDARRAY_FLOAT:             # <<<<<<<<<<<<<<
 *             items.append(
 *                 _ser_ndarray_numeric(arr, NDARRAY_FLOAT_DT, arr.descr.type_num)
 */
    __pyx_t_3 = (__pyx_v_npy_kind == __pyx_v_10serializor_10identifier_NDARRAY_FLOAT);
    if (__pyx_t_3) {

      /* "serializor/binary/ser.py":1744
 *         elif npy_kind == identifier.NDARRAY_FLOAT:
 *             items.append(
 *                 _ser_ndarray_numeric(arr, NDARRAY_FLOAT_DT, arr.descr.type_num)             # <<<<<<<<<<<<<<
 *             )
 *         # . ndarray[bool]
 */
      __pyx_t_8 = __pyx_v_10serializor_6binary_3ser_NDARRAY_FLOAT_DT;
      __Pyx_INCREF(__pyx_t_8);
      __pyx_t_10 = ((PyObject *)__pyx_f_5numpy_7ndarray_5descr_descr(__pyx_v_arr)); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1744, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __pyx_t_2 = __pyx_f_10serializor_6binary_3ser__ser_ndarray_numeric(__pyx_v_arr, ((PyObject*)__pyx_t_8), ((PyArray_Descr *)__pyx_t_10)->type_num); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1744, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

      /* "serializor/binary/ser.py":1743
 *         # . ndarray[float]
 *         elif npy_kind == identifier.NDARRAY_FLOAT:
 *             items.append(             # <<<<<<<<<<<<<<
 *                 _ser_ndarray_numeric(arr, NDARRAY_FLOAT_DT, arr.descr.type_num)
 *             )
 */
      __pyx_t_25 = __Pyx_PyList_Append(__pyx_v_items, __pyx_t_2); if (unlikely(__pyx_t_25 == ((int)-1))) __PYX_ERR(0, 1743, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "serializor/binary/ser.py":1742
 *             items.append(_ser_ndarray_numeric(arr, NDARRAY_UINT_DT, arr.descr.type_num))
 *         # . ndarray[float]
 *         elif npy_kind == identifier.NDARRAY_FLOAT:             # <<<<<<<<<<<<<<
 *             items.append(
 *                 _ser_ndarray_numeric(arr, NDARRAY_FLOAT_DT, arr.descr.type_num)
 */
      goto __pyx_L43;
    }

    /* "serializor/binary/ser.py":1747
 *             )
 *         # . ndarray[bool]
 *         elif npy_kind == identifier.NDARRAY_BOOL:             # <<<<<<<<<<<<<<
 *             items.append(
 *                 _ser_ndarray_numeric(
 */
    __pyx_t_3 = (__pyx_v_npy_kind == __pyx_v_10serializor_10identifier_NDARRAY_BOOL);
    if (__pyx_t_3) {

      /* "serializor/binary/ser.py":1750
 *             items.append(
 *                 _ser_ndarray_numeric(
 *                     np.PyArray_Cast(arr, np.NPY_TYPES.NPY_INT64),  # cast to int64             # <<<<<<<<<<<<<<
 *                     NDARRAY_BOOL_DT,
 *                     arr.descr.type_num,
 */
      __pyx_t_2 = PyArray_Cast(__pyx_v_arr, NPY_INT64); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1750, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 1750, __pyx_L1_error)

      /* "serializor/binary/ser.py":1751
 *                 _ser_ndarray_numeric(
 *                     np.PyArray_Cast(arr, np.NPY_TYPES.NPY_INT64),  # cast to int64
 *                     NDARRAY_BOOL_DT,             # <<<<<<<<<<<<<<
 *                     arr.descr.type_num,
 *                 )
 */
      __pyx_t_10 = __pyx_v_10serializor_6binary_3ser_NDARRAY_BOOL_DT;
      __Pyx_INCREF(__pyx_t_10);

      /* "serializor/binary/ser.py":1752
 *                     np.PyArray_Cast(arr, np.NPY_TYPES.NPY_INT64),  # cast to int64
 *                     NDARRAY_BOOL_DT,
 *                     arr.descr.type_num,             # <<<<<<<<<<<<<<
 *                 )
 *             )
 */
      __pyx_t_8 = ((PyObject *)__pyx_f_5numpy_7ndarray_5descr_descr(__pyx_v_arr)); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1752, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);

      /* "serializor/binary/ser.py":1749
 *         elif npy_kind == identifier.NDARRAY_BOOL:
 *             items.append(
 *                 _ser_ndarray_numeric(             # <<<<<<<<<<<<<<
 *                     np.PyArray_Cast(arr, np.NPY_TYPES.NPY_INT64),  # cast to int64
 *                     NDARRAY_BOOL_DT,
 */
      __pyx_t_21 = __pyx_f_10serializor_6binary_3ser__ser_ndarray_numeric(((PyArrayObject *)__pyx_t_2), ((PyObject*)__pyx_t_10), ((PyArray_Descr *)__pyx_t_8)->type_num); if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 1749, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_21);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

      /* "serializor/binary/ser.py":1748
 *         # . ndarray[bool]
 *         elif npy_kind == identifier.NDARRAY_BOOL:
 *             items.append(             # <<<<<<<<<<<<<<
 *                 _ser_ndarray_numeric(
 *                     np.PyArray_Cast(arr, np.NPY_TYPES.NPY_INT64),  # cast to int64
 */
      __pyx_t_25 = __Pyx_PyList_Append(__pyx_v_items, __pyx_t_21); if (unlikely(__pyx_t_25 == ((int)-1))) __PYX_ERR(0, 1748, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;

      /* "serializor/binary/ser.py":1747
 *             )
 *         # . ndarray[bool]
 *         elif npy_kind == identifier.NDARRAY_BOOL:             # <<<<<<<<<<<<<<
 *             items.append(
 *                 _ser_ndarray_numeric(
 */
      goto __pyx_L43;
    }

    /* "serializor/binary/ser.py":1756
 *             )
 *         # . ndarray[datetime64]
 *         elif npy_kind == identifier.NDARRAY_DT64:             # <<<<<<<<<<<<<<
 *             items.append(NDARRAY_DT64_DT)
 *             values = _ser_ndarray_nptime(arr, NDARRAY_DT64_DT)
 */
    __pyx_t_3 = (__pyx_v_npy_kind == __pyx_v_10serializor_10identifier_NDARRAY_DT64);
    if (__pyx_t_3) {

      /* "serializor/binary/ser.py":1757
 *         # . ndarray[datetime64]
 *         elif npy_kind == identifier.NDARRAY_DT64:
 *             items.append(NDARRAY_DT64_DT)             # <<<<<<<<<<<<<<
 *             values = _ser_ndarray_nptime(arr, NDARRAY_DT64_DT)
 *             # . With timezone
 */
      __pyx_t_21 = __pyx_v_10serializor_6binary_3ser_NDARRAY_DT64_DT;
      __Pyx_INCREF(__pyx_t_21);
      __pyx_t_25 = __Pyx_PyList_Append(__pyx_v_items, __pyx_t_21); if (unlikely(__pyx_t_25 == ((int)-1))) __PYX_ERR(0, 1757, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;

      /* "serializor/binary/ser.py":1758
 *         elif npy_kind == identifier.NDARRAY_DT64:
 *             items.append(NDARRAY_DT64_DT)
 *             values = _ser_ndarray_nptime(arr, NDARRAY_DT64_DT)             # <<<<<<<<<<<<<<
 *             # . With timezone
 *             if (tz := col.dt.tz) is not None:
 */
      __pyx_t_21 = __pyx_v_10serializor_6binary_3ser_NDARRAY_DT64_DT;
      __Pyx_INCREF(__pyx_t_21);
      __pyx_t_8 = __pyx_f_10serializor_6binary_3ser__ser_ndarray_nptime(__pyx_v_arr, ((PyObject*)__pyx_t_21)); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1758, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
      __Pyx_XDECREF_SET(__pyx_v_values, ((PyObject*)__pyx_t_8));
      __pyx_t_8 = 0;

      /* "serializor/binary/ser.py":1760
 *             values = _ser_ndarray_nptime(arr, NDARRAY_DT64_DT)
 *             # . With timezone
 *             if (tz := col.dt.tz) is not None:             # <<<<<<<<<<<<<<
 *                 # . approach 1: timezone name
 *                 if set_contains(utils.AVAILABLE_TIMEZONES, tz_name := str(tz)):
 */
      __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_v_col, __pyx_n_s_dt); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1760, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_21 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_tz); if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 1760, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_21);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_XDECREF_SET(__pyx_v_tz, __pyx_t_21);
      __Pyx_INCREF(__pyx_t_21);
      __pyx_t_3 = (__pyx_t_21 != Py_None);
      __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
      if (__pyx_t_3) {

        /* "serializor/binary/ser.py":1762
 *             if (tz := col.dt.tz) is not None:
 *                 # . approach 1: timezone name
 *                 if set_contains(utils.AVAILABLE_TIMEZONES, tz_name := str(tz)):             # <<<<<<<<<<<<<<
 *                     name = utils.encode_str(tz_name, b"ascii")
 *                     items.append(b"\x01")  # approach 1
 */
        __pyx_t_21 = __pyx_v_10serializor_5utils_AVAILABLE_TIMEZONES;
        __Pyx_INCREF(__pyx_t_21);
        __pyx_t_8 = __Pyx_PyObject_Unicode(__pyx_v_tz); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1762, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        __Pyx_XDECREF_SET(__pyx_v_tz_name, __pyx_t_8);
        __Pyx_INCREF(__pyx_t_8);
        __pyx_t_10 = __pyx_t_8;
        __Pyx_INCREF(__pyx_t_10);
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        __pyx_t_3 = PySet_Contains(__pyx_t_21, __pyx_t_10); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(0, 1762, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        if (__pyx_t_3) {

          /* "serializor/binary/ser.py":1763
 *                 # . approach 1: timezone name
 *                 if set_contains(utils.AVAILABLE_TIMEZONES, tz_name := str(tz)):
 *                     name = utils.encode_str(tz_name, b"ascii")             # <<<<<<<<<<<<<<
 *                     items.append(b"\x01")  # approach 1
 *                     items.append(gen_encoded_int(bytes_len(name)))  # type: ignore
 */
          __pyx_t_10 = __pyx_f_10serializor_5utils_encode_str(__pyx_v_tz_name, ((char *)"ascii")); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1763, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_10);
          __Pyx_XDECREF_SET(__pyx_v_name, ((PyObject*)__pyx_t_10));
          __pyx_t_10 = 0;

          /* "serializor/binary/ser.py":1764
 *                 if set_contains(utils.AVAILABLE_TIMEZONES, tz_name := str(tz)):
 *                     name = utils.encode_str(tz_name, b"ascii")
 *                     items.append(b"\x01")  # approach 1             # <<<<<<<<<<<<<<
 *                     items.append(gen_encoded_int(bytes_len(name)))  # type: ignore
 *                     items.append(name)
 */
          __pyx_t_25 = __Pyx_PyList_Append(__pyx_v_items, __pyx_kp_b__5); if (unlikely(__pyx_t_25 == ((int)-1))) __PYX_ERR(0, 1764, __pyx_L1_error)

          /* "serializor/binary/ser.py":1765
 *                     name = utils.encode_str(tz_name, b"ascii")
 *                     items.append(b"\x01")  # approach 1
 *                     items.append(gen_encoded_int(bytes_len(name)))  # type: ignore             # <<<<<<<<<<<<<<
 *                     items.append(name)
 *                     items.append(values)
 */
          __pyx_t_26 = PyBytes_Size(__pyx_v_name); if (unlikely(__pyx_t_26 == ((Py_ssize_t)-1L))) __PYX_ERR(0, 1765, __pyx_L1_error)
          __pyx_t_10 = __pyx_f_10serializor_6binary_3ser_gen_encoded_int(__pyx_t_26); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1765, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_10);
          __pyx_t_25 = __Pyx_PyList_Append(__pyx_v_items, __pyx_t_10); if (unlikely(__pyx_t_25 == ((int)-1))) __PYX_ERR(0, 1765, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

          /* "serializor/binary/ser.py":1766
 *                     items.append(b"\x01")  # approach 1
 *                     items.append(gen_encoded_int(bytes_len(name)))  # type: ignore
 *                     items.append(name)             # <<<<<<<<<<<<<<
 *                     items.append(values)
 *                 # . approach 2: timezone offset
 */
          __pyx_t_25 = __Pyx_PyList_Append(__pyx_v_items, __pyx_v_name); if (unlikely(__pyx_t_25 == ((int)-1))) __PYX_ERR(0, 1766, __pyx_L1_error)

          /* "serializor/binary/ser.py":1767
 *                     items.append(gen_encoded_int(bytes_len(name)))  # type: ignore
 *                     items.append(name)
 *                     items.append(values)             # <<<<<<<<<<<<<<
 *                 # . approach 2: timezone offset
 *                 elif (offset := tz.utcoffset(None)) is not None:
 */
          __pyx_t_25 = __Pyx_PyList_Append(__pyx_v_items, __pyx_v_values); if (unlikely(__pyx_t_25 == ((int)-1))) __PYX_ERR(0, 1767, __pyx_L1_error)

          /* "serializor/binary/ser.py":1762
 *             if (tz := col.dt.tz) is not None:
 *                 # . approach 1: timezone name
 *                 if set_contains(utils.AVAILABLE_TIMEZONES, tz_name := str(tz)):             # <<<<<<<<<<<<<<
 *                     name = utils.encode_str(tz_name, b"ascii")
 *                     items.append(b"\x01")  # approach 1
 */
          goto __pyx_L45;
        }

        /* "serializor/binary/ser.py":1769
 *                     items.append(values)
 *                 # . approach 2: timezone offset
 *                 elif (offset := tz.utcoffset(None)) is not None:             # <<<<<<<<<<<<<<
 *                     items.append(b"\x02")  # approach 2
 *                     items.append(pack_int32(datetime.timedelta_days(offset)))  # type: ignore
 */
        __pyx_t_21 = __Pyx_PyObject_GetAttrStr(__pyx_v_tz, __pyx_n_s_utcoffset); if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 1769, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_21);
        __pyx_t_8 = NULL;
        __pyx_t_27 = 0;
        #if CYTHON_UNPACK_METHODS
        if (likely(PyMethod_Check(__pyx_t_21))) {
          __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_21);
          if (likely(__pyx_t_8)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_21);
            __Pyx_INCREF(__pyx_t_8);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_21, function);
            __pyx_t_27 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_8, Py_None};
          __pyx_t_10 = __Pyx_PyObject_FastCall(__pyx_t_21, __pyx_callargs+1-__pyx_t_27, 1+__pyx_t_27);
          __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
          if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1769, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_10);
          __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
        }
        __Pyx_XDECREF_SET(__pyx_v_offset, __pyx_t_10);
        __Pyx_INCREF(__pyx_t_10);
        __pyx_t_3 = (__pyx_t_10 != Py_None);
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        if (__pyx_t_3) {

          /* "serializor/binary/ser.py":1770
 *                 # . approach 2: timezone offset
 *                 elif (offset := tz.utcoffset(None)) is not None:
 *                     items.append(b"\x02")  # approach 2             # <<<<<<<<<<<<<<
 *                     items.append(pack_int32(datetime.timedelta_days(offset)))  # type: ignore
 *                     items.append(pack_int32(datetime.timedelta_seconds(offset)))  # type: ignore
 */
          __pyx_t_25 = __Pyx_PyList_Append(__pyx_v_items, __pyx_kp_b__6); if (unlikely(__pyx_t_25 == ((int)-1))) __PYX_ERR(0, 1770, __pyx_L1_error)

          /* "serializor/binary/ser.py":1771
 *                 elif (offset := tz.utcoffset(None)) is not None:
 *                     items.append(b"\x02")  # approach 2
 *                     items.append(pack_int32(datetime.timedelta_days(offset)))  # type: ignore             # <<<<<<<<<<<<<<
 *                     items.append(pack_int32(datetime.timedelta_seconds(offset)))  # type: ignore
 *                     items.append(pack_int32(datetime.timedelta_microseconds(offset)))  # type: ignore
 */
          __pyx_t_10 = __pyx_f_10serializor_6binary_3ser_pack_int32(__pyx_f_7cpython_8datetime_timedelta_days(__pyx_v_offset)); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1771, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_10);
          __pyx_t_25 = __Pyx_PyList_Append(__pyx_v_items, __pyx_t_10); if (unlikely(__pyx_t_25 == ((int)-1))) __PYX_ERR(0, 1771, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

          /* "serializor/binary/ser.py":1772
 *                     items.append(b"\x02")  # approach 2
 *                     items.append(pack_int32(datetime.timedelta_days(offset)))  # type: ignore
 *                     items.append(pack_int32(datetime.timedelta_seconds(offset)))  # type: ignore             # <<<<<<<<<<<<<<
 *                     items.append(pack_int32(datetime.timedelta_microseconds(offset)))  # type: ignore
 *                     items.append(values)
 */
          __pyx_t_10 = __pyx_f_10serializor_6binary_3ser_pack_int32(__pyx_f_7cpython_8datetime_timedelta_seconds(__pyx_v_offset)); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1772, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_10);
          __pyx_t_25 = __Pyx_PyList_Append(__pyx_v_items, __pyx_t_10); if (unlikely(__pyx_t_25 == ((int)-1))) __PYX_ERR(0, 1772, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

          /* "serializor/binary/ser.py":1773
 *                     items.append(pack_int32(datetime.timedelta_days(offset)))  # type: ignore
 *                     items.append(pack_int32(datetime.timedelta_seconds(offset)))  # type: ignore
 *                     items.append(pack_int32(datetime.timedelta_microseconds(offset)))  # type: ignore             # <<<<<<<<<<<<<<
 *                     items.append(values)
 *                 # . Fallback
 */
          __pyx_t_10 = __pyx_f_10serializor_6binary_3ser_pack_int32(__pyx_f_7cpython_8datetime_timedelta_microseconds(__pyx_v_offset)); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1773, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_10);
          __pyx_t_25 = __Pyx_PyList_Append(__pyx_v_items, __pyx_t_10); if (unlikely(__pyx_t_25 == ((int)-1))) __PYX_ERR(0, 1773, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

          /* "serializor/binary/ser.py":1774
 *                     items.append(pack_int32(datetime.timedelta_seconds(offset)))  # type: ignore
 *                     items.append(pack_int32(datetime.timedelta_microseconds(offset)))  # type: ignore
 *                     items.append(values)             # <<<<<<<<<<<<<<
 *                 # . Fallback
 *                 else:
 */
          __pyx_t_25 = __Pyx_PyList_Append(__pyx_v_items, __pyx_v_values); if (unlikely(__pyx_t_25 == ((int)-1))) __PYX_ERR(0, 1774, __pyx_L1_error)

          /* "serializor/binary/ser.py":1769
 *                     items.append(values)
 *                 # . approach 2: timezone offset
 *                 elif (offset := tz.utcoffset(None)) is not None:             # <<<<<<<<<<<<<<
 *                     items.append(b"\x02")  # approach 2
 *                     items.append(pack_int32(datetime.timedelta_days(offset)))  # type: ignore
 */
          goto __pyx_L45;
        }

        /* "serializor/binary/ser.py":1777
 *                 # . Fallback
 *                 else:
 *                     items.append(b"\x00")  # approach 0             # <<<<<<<<<<<<<<
 *                     items.append(values)
 *             else:
 */
        /*else*/ {
          __pyx_t_25 = __Pyx_PyList_Append(__pyx_v_items, __pyx_kp_b__7); if (unlikely(__pyx_t_25 == ((int)-1))) __PYX_ERR(0, 1777, __pyx_L1_error)

          /* "serializor/binary/ser.py":1778
 *                 else:
 *                     items.append(b"\x00")  # approach 0
 *                     items.append(values)             # <<<<<<<<<<<<<<
 *             else:
 *                 # . Without timezone
 */
          __pyx_t_25 = __Pyx_PyList_Append(__pyx_v_items, __pyx_v_values); if (unlikely(__pyx_t_25 == ((int)-1))) __PYX_ERR(0, 1778, __pyx_L1_error)
        }
        __pyx_L45:;

        /* "serializor/binary/ser.py":1760
 *             values = _ser_ndarray_nptime(arr, NDARRAY_DT64_DT)
 *             # . With timezone
 *             if (tz := col.dt.tz) is not None:             # <<<<<<<<<<<<<<
 *                 # . approach 1: timezone name
 *                 if set_contains(utils.AVAILABLE_TIMEZONES, tz_name := str(tz)):
 */
        goto __pyx_L44;
      }

      /* "serializor/binary/ser.py":1781
 *             else:
 *                 # . Without timezone
 *                 items.append(b"\x00")  # approach 0             # <<<<<<<<<<<<<<
 *                 items.append(values)
 *         # . ndarray[timedelta64]
 */
      /*else*/ {
        __pyx_t_25 = __Pyx_PyList_Append(__pyx_v_items, __pyx_kp_b__7); if (unlikely(__pyx_t_25 == ((int)-1))) __PYX_ERR(0, 1781, __pyx_L1_error)

        /* "serializor/binary/ser.py":1782
 *                 # . Without timezone
 *                 items.append(b"\x00")  # approach 0
 *                 items.append(values)             # <<<<<<<<<<<<<<
 *         # . ndarray[timedelta64]
 *         elif npy_kind == identifier.NDARRAY_TD64:
 */
        __pyx_t_25 = __Pyx_PyList_Append(__pyx_v_items, __pyx_v_values); if (unlikely(__pyx_t_25 == ((int)-1))) __PYX_ERR(0, 1782, __pyx_L1_error)
      }
      __pyx_L44:;

      /* "serializor/binary/ser.py":1756
 *             )
 *         # . ndarray[datetime64]
 *         elif npy_kind == identifier.NDARRAY_DT64:             # <<<<<<<<<<<<<<
 *             items.append(NDARRAY_DT64_DT)
 *             values = _ser_ndarray_nptime(arr, NDARRAY_DT64_DT)
 */
      goto __pyx_L43;
    }

    /* "serializor/binary/ser.py":1784
 *                 items.append(values)
 *         # . ndarray[timedelta64]
 *         elif npy_kind == identifier.NDARRAY_TD64:             # <<<<<<<<<<<<<<
 *             items.append(_ser_ndarray_nptime(arr, NDARRAY_TD64_DT))
 *         # . ndarray[complex]
 */
    __pyx_t_3 = (__pyx_v_npy_kind == __pyx_v_10serializor_10identifier_NDARRAY_TD64);
    if (__pyx_t_3) {

      /* "serializor/binary/ser.py":1785
 *         # . ndarray[timedelta64]
 *         elif npy_kind == identifier.NDARRAY_TD64:
 *             items.append(_ser_ndarray_nptime(arr, NDARRAY_TD64_DT))             # <<<<<<<<<<<<<<
 *         # . ndarray[complex]
 *         elif npy_kind == identifier.NDARRAY_COMPLEX:
 */
      __pyx_t_10 = __pyx_v_10serializor_6binary_3ser_NDARRAY_TD64_DT;
      __Pyx_INCREF(__pyx_t_10);
      __pyx_t_21 = __pyx_f_10serializor_6binary_3ser__ser_ndarray_nptime(__pyx_v_arr, ((PyObject*)__pyx_t_10)); if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 1785, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_21);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __pyx_t_25 = __Pyx_PyList_Append(__pyx_v_items, __pyx_t_21); if (unlikely(__pyx_t_25 == ((int)-1))) __PYX_ERR(0, 1785, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;

      /* "serializor/binary/ser.py":1784
 *                 items.append(values)
 *         # . ndarray[timedelta64]
 *         elif npy_kind == identifier.NDARRAY_TD64:             # <<<<<<<<<<<<<<
 *             items.append(_ser_ndarray_nptime(arr, NDARRAY_TD64_DT))
 *         # . ndarray[complex]
 */
      goto __pyx_L43;
    }

    /* "serializor/binary/ser.py":1787
 *             items.append(_ser_ndarray_nptime(arr, NDARRAY_TD64_DT))
 *         # . ndarray[complex]
 *         elif npy_kind == identifier.NDARRAY_COMPLEX:             # <<<<<<<<<<<<<<
 *             items.append(_ser_ndarray_complex(arr, NDARRAY_COMPLEX_DT))
 *         # . ndarray[bytes]
 */
    __pyx_t_3 = (__pyx_v_npy_kind == __pyx_v_10serializor_10identifier_NDARRAY_COMPLEX);
    if (__pyx_t_3) {

      /* "serializor/binary/ser.py":1788
 *         # . ndarray[complex]
 *         elif npy_kind == identifier.NDARRAY_COMPLEX:
 *             items.append(_ser_ndarray_complex(arr, NDARRAY_COMPLEX_DT))             # <<<<<<<<<<<<<<
 *         # . ndarray[bytes]
 *         elif npy_kind == identifier.NDARRAY_BYTES:
 */
      __pyx_t_21 = __pyx_v_10serializor_6binary_3ser_NDARRAY_COMPLEX_DT;
      __Pyx_INCREF(__pyx_t_21);
      __pyx_t_10 = __pyx_f_10serializor_6binary_3ser__ser_ndarray_complex(__pyx_v_arr, ((PyObject*)__pyx_t_21)); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1788, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
      __pyx_t_25 = __Pyx_PyList_Append(__pyx_v_items, __pyx_t_10); if (unlikely(__pyx_t_25 == ((int)-1))) __PYX_ERR(0, 1788, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

      /* "serializor/binary/ser.py":1787
 *             items.append(_ser_ndarray_nptime(arr, NDARRAY_TD64_DT))
 *         # . ndarray[complex]
 *         elif npy_kind == identifier.NDARRAY_COMPLEX:             # <<<<<<<<<<<<<<
 *             items.append(_ser_ndarray_complex(arr, NDARRAY_COMPLEX_DT))
 *         # . ndarray[bytes]
 */
      goto __pyx_L43;
    }

    /* "serializor/binary/ser.py":1790
 *             items.append(_ser_ndarray_complex(arr, NDARRAY_COMPLEX_DT))
 *         # . ndarray[bytes]
 *         elif npy_kind == identifier.NDARRAY_BYTES:             # <<<<<<<<<<<<<<
 *             items.append(_ser_ndarray_bytes(arr, NDARRAY_BYTES_DT))
 *         # . ndarray[str]
 */
    __pyx_t_3 = (__pyx_v_npy_kind == __pyx_v_10serializor_10identifier_NDARRAY_BYTES);
    if (__pyx_t_3) {

      /* "serializor/binary/ser.py":1791
 *         # . ndarray[bytes]
 *         elif npy_kind == identifier.NDARRAY_BYTES:
 *             items.append(_ser_ndarray_bytes(arr, NDARRAY_BYTES_DT))             # <<<<<<<<<<<<<<
 *         # . ndarray[str]
 *         elif npy_kind == identifier.NDARRAY_UNICODE:
 */
      __pyx_t_10 = __pyx_v_10serializor_6binary_3ser_NDARRAY_BYTES_DT;
      __Pyx_INCREF(__pyx_t_10);
      __pyx_t_21 = __pyx_f_10serializor_6binary_3ser__ser_ndarray_bytes(__pyx_v_arr, ((PyObject*)__pyx_t_10)); if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 1791, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_21);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __pyx_t_25 = __Pyx_PyList_Append(__pyx_v_items, __pyx_t_21); if (unlikely(__pyx_t_25 == ((int)-1))) __PYX_ERR(0, 1791, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;

      /* "serializor/binary/ser.py":1790
 *             items.append(_ser_ndarray_complex(arr, NDARRAY_COMPLEX_DT))
 *         # . ndarray[bytes]
 *         elif npy_kind == identifier.NDARRAY_BYTES:             # <<<<<<<<<<<<<<
 *             items.append(_ser_ndarray_bytes(arr, NDARRAY_BYTES_DT))
 *         # . ndarray[str]
 */
      goto __pyx_L43;
    }

    /* "serializor/binary/ser.py":1793
 *             items.append(_ser_ndarray_bytes(arr, NDARRAY_BYTES_DT))
 *         # . ndarray[str]
 *         elif npy_kind == identifier.NDARRAY_UNICODE:             # <<<<<<<<<<<<<<
 *             items.append(_ser_ndarray_unicode(arr, NDARRAY_UNICODE_DT))
 *         # . invalid dtype
 */
    __pyx_t_3 = (__pyx_v_npy_kind == __pyx_v_10serializor_10identifier_NDARRAY_UNICODE);
    if (likely(__pyx_t_3)) {

      /* "serializor/binary/ser.py":1794
 *         # . ndarray[str]
 *         elif npy_kind == identifier.NDARRAY_UNICODE:
 *             items.append(_ser_ndarray_unicode(arr, NDARRAY_UNICODE_DT))             # <<<<<<<<<<<<<<
 *         # . invalid dtype
 *         else:
 */
      __pyx_t_21 = __pyx_v_10serializor_6binary_3ser_NDARRAY_UNICODE_DT;
      __Pyx_INCREF(__pyx_t_21);
      __pyx_t_10 = __pyx_f_10serializor_6binary_3ser__ser_ndarray_unicode(__pyx_v_arr, ((PyObject*)__pyx_t_21)); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1794, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
      __pyx_t_25 = __Pyx_PyList_Append(__pyx_v_items, __pyx_t_10); if (unlikely(__pyx_t_25 == ((int)-1))) __PYX_ERR(0, 1794, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

      /* "serializor/binary/ser.py":1793
 *             items.append(_ser_ndarray_bytes(arr, NDARRAY_BYTES_DT))
 *         # . ndarray[str]
 *         elif npy_kind == identifier.NDARRAY_UNICODE:             # <<<<<<<<<<<<<<
 *             items.append(_ser_ndarray_unicode(arr, NDARRAY_UNICODE_DT))
 *         # . invalid dtype
 */
      goto __pyx_L43;
    }

    /* "serializor/binary/ser.py":1797
 *         # . invalid dtype
 *         else:
 *             raise TypeError(             # <<<<<<<<<<<<<<
 *                 "unsupported <'pandas.DataFrame'> column dtype '%s'." % col.dtype
 *             )
 */
    /*else*/ {

      /* "serializor/binary/ser.py":1798
 *         else:
 *             raise TypeError(
 *                 "unsupported <'pandas.DataFrame'> column dtype '%s'." % col.dtype             # <<<<<<<<<<<<<<
 *             )
 * 
 */
      __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v_col, __pyx_n_s_dtype); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1798, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __pyx_t_21 = __Pyx_PyUnicode_FormatSafe(__pyx_kp_u_unsupported_pandas_DataFrame_col, __pyx_t_10); if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 1798, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_21);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

      /* "serializor/binary/ser.py":1797
 *         # . invalid dtype
 *         else:
 *             raise TypeError(             # <<<<<<<<<<<<<<
 *                 "unsupported <'pandas.DataFrame'> column dtype '%s'." % col.dtype
 *             )
 */
      __pyx_t_10 = __Pyx_PyObject_CallOneArg(__pyx_builtin_TypeError, __pyx_t_21); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1797, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
      __Pyx_Raise(__pyx_t_10, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __PYX_ERR(0, 1797, __pyx_L1_error)
    }
    __pyx_L43:;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "serializor/binary/ser.py":1802
 * 
 *     # Compose
 *     return b"".join(items)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBytes_Join(__pyx_kp_b__4, __pyx_v_items); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1802, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyBytes_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("bytes", __pyx_t_1))) __PYX_ERR(0, 1802, __pyx_L1_error)
  __pyx_r = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "serializor/binary/ser.py":1689
 * 
 * 
 * @cython.cfunc             # <<<<<<<<<<<<<<
 * @cython.inline(True)
 * def _ser_dataframe(obj: DataFrame) -> bytes:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_21);
  __Pyx_AddTraceback("serializor.binary.ser._ser_dataframe", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_cols);
  __Pyx_XDECREF(__pyx_v_cols_value);
  __Pyx_XDECREF(__pyx_v_err);
  __Pyx_XDECREF(__pyx_v_items);
  __Pyx_XDECREF(__pyx_v__);
  __Pyx_XDECREF(__pyx_v_col);
  __Pyx_XDECREF((PyObject *)__pyx_v_arr);
  __Pyx_XDECREF(__pyx_v_values);
  __Pyx_XDECREF(__pyx_v_name);
  __Pyx_XDECREF(__pyx_v_tz);
  __Pyx_XDECREF(__pyx_v_tz_name);
  __Pyx_XDECREF(__pyx_v_offset);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "serializor/binary/ser.py":1806
 * 
 * # Serialize -----------------------------------------------------------------------------------
 * @cython.cfunc             # <<<<<<<<<<<<<<
 * @cython.inline(True)
 * def _ser_common(obj: object) -> bytes:
 */

static PyObject *__pyx_f_10serializor_6binary_3ser__ser_common(PyObject *__pyx_v_obj) {
  PyTypeObject *__pyx_v_obj_dtype = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_ser_common", 1);

  /* "serializor/binary/ser.py":1811
 *     """(cfunc) Serialize common object to `<'bytes'>`."""
 *     # Get data type
 *     obj_dtype = type(obj)             # <<<<<<<<<<<<<<
 * 
 *     ##### Common Types #####
 */
  __Pyx_INCREF(((PyObject *)Py_TYPE(__pyx_v_obj)));
  __pyx_v_obj_dtype = ((PyTypeObject*)((PyObject *)Py_TYPE(__pyx_v_obj)));

  /* "serializor/binary/ser.py":1816
 *     # Basic Types
 *     # . <'str'>
 *     if obj_dtype is str:             # <<<<<<<<<<<<<<
 *         return _ser_str(obj)
 *     # . <'int'>
 */
  __pyx_t_1 = (__pyx_v_obj_dtype == (&PyUnicode_Type));
  if (__pyx_t_1) {

    /* "serializor/binary/ser.py":1817
 *     # . <'str'>
 *     if obj_dtype is str:
 *         return _ser_str(obj)             # <<<<<<<<<<<<<<
 *     # . <'int'>
 *     if obj_dtype is int:
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = __pyx_f_10serializor_6binary_3ser__ser_str(__pyx_v_obj); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1817, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_r = ((PyObject*)__pyx_t_2);
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "serializor/binary/ser.py":1816
 *     # Basic Types
 *     # . <'str'>
 *     if obj_dtype is str:             # <<<<<<<<<<<<<<
 *         return _ser_str(obj)
 *     # . <'int'>
 */
  }

  /* "serializor/binary/ser.py":1819
 *         return _ser_str(obj)
 *     # . <'int'>
 *     if obj_dtype is int:             # <<<<<<<<<<<<<<
 *         return _ser_int(obj)
 *     # . <'float'>
 */
  __pyx_t_1 = (__pyx_v_obj_dtype == (&PyInt_Type));
  if (__pyx_t_1) {

    /* "serializor/binary/ser.py":1820
 *     # . <'int'>
 *     if obj_dtype is int:
 *         return _ser_int(obj)             # <<<<<<<<<<<<<<
 *     # . <'float'>
 *     if obj_dtype is float:
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = __pyx_f_10serializor_6binary_3ser__ser_int(__pyx_v_obj); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1820, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_r = ((PyObject*)__pyx_t_2);
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "serializor/binary/ser.py":1819
 *         return _ser_str(obj)
 *     # . <'int'>
 *     if obj_dtype is int:             # <<<<<<<<<<<<<<
 *         return _ser_int(obj)
 *     # . <'float'>
 */
  }

  /* "serializor/binary/ser.py":1822
 *         return _ser_int(obj)
 *     # . <'float'>
 *     if obj_dtype is float:             # <<<<<<<<<<<<<<
 *         return _ser_float(obj)
 *     # . <'bool'>
 */
  __pyx_t_1 = (__pyx_v_obj_dtype == (&PyFloat_Type));
  if (__pyx_t_1) {

    /* "serializor/binary/ser.py":1823
 *     # . <'float'>
 *     if obj_dtype is float:
 *         return _ser_float(obj)             # <<<<<<<<<<<<<<
 *     # . <'bool'>
 *     if obj_dtype is bool:
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = __pyx_f_10serializor_6binary_3ser__ser_float(__pyx_v_obj); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1823, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_r = ((PyObject*)__pyx_t_2);
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "serializor/binary/ser.py":1822
 *         return _ser_int(obj)
 *     # . <'float'>
 *     if obj_dtype is float:             # <<<<<<<<<<<<<<
 *         return _ser_float(obj)
 *     # . <'bool'>
 */
  }

  /* "serializor/binary/ser.py":1825
 *         return _ser_float(obj)
 *     # . <'bool'>
 *     if obj_dtype is bool:             # <<<<<<<<<<<<<<
 *         return _ser_bool(obj)
 *     # . <None>
 */
  __pyx_t_1 = (__pyx_v_obj_dtype == ((PyTypeObject*)((PyObject*)&PyBool_Type)));
  if (__pyx_t_1) {

    /* "serializor/binary/ser.py":1826
 *     # . <'bool'>
 *     if obj_dtype is bool:
 *         return _ser_bool(obj)             # <<<<<<<<<<<<<<
 *     # . <None>
 *     if obj_dtype is typeref.NONE:
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = __pyx_f_10serializor_6binary_3ser__ser_bool(__pyx_v_obj); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1826, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_r = ((PyObject*)__pyx_t_2);
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "serializor/binary/ser.py":1825
 *         return _ser_float(obj)
 *     # . <'bool'>
 *     if obj_dtype is bool:             # <<<<<<<<<<<<<<
 *         return _ser_bool(obj)
 *     # . <None>
 */
  }

  /* "serializor/binary/ser.py":1828
 *         return _ser_bool(obj)
 *     # . <None>
 *     if obj_dtype is typeref.NONE:             # <<<<<<<<<<<<<<
 *         return _ser_none(obj)
 * 
 */
  __pyx_t_1 = (__pyx_v_obj_dtype == ((PyTypeObject*)__pyx_v_10serializor_7typeref_NONE));
  if (__pyx_t_1) {

    /* "serializor/binary/ser.py":1829
 *     # . <None>
 *     if obj_dtype is typeref.NONE:
 *         return _ser_none(obj)             # <<<<<<<<<<<<<<
 * 
 *     # Date&Time Types
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = __pyx_f_10serializor_6binary_3ser__ser_none(__pyx_v_obj); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1829, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_r = ((PyObject*)__pyx_t_2);
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "serializor/binary/ser.py":1828
 *         return _ser_bool(obj)
 *     # . <None>
 *     if obj_dtype is typeref.NONE:             # <<<<<<<<<<<<<<
 *         return _ser_none(obj)
 * 
 */
  }

  /* "serializor/binary/ser.py":1833
 *     # Date&Time Types
 *     # . <'datetime.datetime'>
 *     if obj_dtype is datetime.datetime:             # <<<<<<<<<<<<<<
 *         return _ser_datetime(obj)
 *     # . <'datetime.date'>
 */
  __pyx_t_1 = (__pyx_v_obj_dtype == __pyx_ptype_7cpython_8datetime_datetime);
  if (__pyx_t_1) {

    /* "serializor/binary/ser.py":1834
 *     # . <'datetime.datetime'>
 *     if obj_dtype is datetime.datetime:
 *         return _ser_datetime(obj)             # <<<<<<<<<<<<<<
 *     # . <'datetime.date'>
 *     if obj_dtype is datetime.date:
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = __pyx_f_10serializor_6binary_3ser__ser_datetime(__pyx_v_obj); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1834, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_r = ((PyObject*)__pyx_t_2);
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "serializor/binary/ser.py":1833
 *     # Date&Time Types
 *     # . <'datetime.datetime'>
 *     if obj_dtype is datetime.datetime:             # <<<<<<<<<<<<<<
 *         return _ser_datetime(obj)
 *     # . <'datetime.date'>
 */
  }

  /* "serializor/binary/ser.py":1836
 *         return _ser_datetime(obj)
 *     # . <'datetime.date'>
 *     if obj_dtype is datetime.date:             # <<<<<<<<<<<<<<
 *         return _ser_date(obj)
 *     # . <'datetime.time'>
 */
  __pyx_t_1 = (__pyx_v_obj_dtype == __pyx_ptype_7cpython_8datetime_date);
  if (__pyx_t_1) {

    /* "serializor/binary/ser.py":1837
 *     # . <'datetime.date'>
 *     if obj_dtype is datetime.date:
 *         return _ser_date(obj)             # <<<<<<<<<<<<<<
 *     # . <'datetime.time'>
 *     if obj_dtype is datetime.time:
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = __pyx_f_10serializor_6binary_3ser__ser_date(__pyx_v_obj); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1837, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_r = ((PyObject*)__pyx_t_2);
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "serializor/binary/ser.py":1836
 *         return _ser_datetime(obj)
 *     # . <'datetime.date'>
 *     if obj_dtype is datetime.date:             # <<<<<<<<<<<<<<
 *         return _ser_date(obj)
 *     # . <'datetime.time'>
 */
  }

  /* "serializor/binary/ser.py":1839
 *         return _ser_date(obj)
 *     # . <'datetime.time'>
 *     if obj_dtype is datetime.time:             # <<<<<<<<<<<<<<
 *         return _ser_time(obj)
 *     # . <'datetime.timedelta'>
 */
  __pyx_t_1 = (__pyx_v_obj_dtype == __pyx_ptype_7cpython_8datetime_time);
  if (__pyx_t_1) {

    /* "serializor/binary/ser.py":1840
 *     # . <'datetime.time'>
 *     if obj_dtype is datetime.time:
 *         return _ser_time(obj)             # <<<<<<<<<<<<<<
 *     # . <'datetime.timedelta'>
 *     if obj_dtype is datetime.timedelta:
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = __pyx_f_10serializor_6binary_3ser__ser_time(__pyx_v_obj); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1840, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_r = ((PyObject*)__pyx_t_2);
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "serializor/binary/ser.py":1839
 *         return _ser_date(obj)
 *     # . <'datetime.time'>
 *     if obj_dtype is datetime.time:             # <<<<<<<<<<<<<<
 *         return _ser_time(obj)
 *     # . <'datetime.timedelta'>
 */
  }

  /* "serializor/binary/ser.py":1842
 *         return _ser_time(obj)
 *     # . <'datetime.timedelta'>
 *     if obj_dtype is datetime.timedelta:             # <<<<<<<<<<<<<<
 *         return _ser_timedelta(obj)
 * 
 */
  __pyx_t_1 = (__pyx_v_obj_dtype == __pyx_ptype_7cpython_8datetime_timedelta);
  if (__pyx_t_1) {

    /* "serializor/binary/ser.py":1843
 *     # . <'datetime.timedelta'>
 *     if obj_dtype is datetime.timedelta:
 *         return _ser_timedelta(obj)             # <<<<<<<<<<<<<<
 * 
 *     # Bytes Types
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = __pyx_f_10serializor_6binary_3ser__ser_timedelta(__pyx_v_obj); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1843, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_r = ((PyObject*)__pyx_t_2);
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "serializor/binary/ser.py":1842
 *         return _ser_time(obj)
 *     # . <'datetime.timedelta'>
 *     if obj_dtype is datetime.timedelta:             # <<<<<<<<<<<<<<
 *         return _ser_timedelta(obj)
 * 
 */
  }

  /* "serializor/binary/ser.py":1847
 *     # Bytes Types
 *     # . <'bytes'>
 *     if obj_dtype is bytes:             # <<<<<<<<<<<<<<
 *         return _ser_bytes(obj)
 * 
 */
  __pyx_t_1 = (__pyx_v_obj_dtype == (&PyBytes_Type));
  if (__pyx_t_1) {

    /* "serializor/binary/ser.py":1848
 *     # . <'bytes'>
 *     if obj_dtype is bytes:
 *         return _ser_bytes(obj)             # <<<<<<<<<<<<<<
 * 
 *     # Sequence Types
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = __pyx_f_10serializor_6binary_3ser__ser_bytes(__pyx_v_obj); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1848, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_r = ((PyObject*)__pyx_t_2);
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "serializor/binary/ser.py":1847
 *     # Bytes Types
 *     # . <'bytes'>
 *     if obj_dtype is bytes:             # <<<<<<<<<<<<<<
 *         return _ser_bytes(obj)
 * 
 */
  }

  /* "serializor/binary/ser.py":1852
 *     # Sequence Types
 *     # . <'list'>
 *     if obj_dtype is list:             # <<<<<<<<<<<<<<
 *         return _ser_list(obj)
 *     # . <'tuple'>
 */
  __pyx_t_1 = (__pyx_v_obj_dtype == (&PyList_Type));
  if (__pyx_t_1) {

    /* "serializor/binary/ser.py":1853
 *     # . <'list'>
 *     if obj_dtype is list:
 *         return _ser_list(obj)             # <<<<<<<<<<<<<<
 *     # . <'tuple'>
 *     if obj_dtype is tuple:
 */
    __Pyx_XDECREF(__pyx_r);
    if (!(likely(PyList_CheckExact(__pyx_v_obj))||((__pyx_v_obj) == Py_None) || __Pyx_RaiseUnexpectedTypeError("list", __pyx_v_obj))) __PYX_ERR(0, 1853, __pyx_L1_error)
    __pyx_t_2 = __pyx_f_10serializor_6binary_3ser__ser_list(((PyObject*)__pyx_v_obj)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1853, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_r = ((PyObject*)__pyx_t_2);
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "serializor/binary/ser.py":1852
 *     # Sequence Types
 *     # . <'list'>
 *     if obj_dtype is list:             # <<<<<<<<<<<<<<
 *         return _ser_list(obj)
 *     # . <'tuple'>
 */
  }

  /* "serializor/binary/ser.py":1855
 *         return _ser_list(obj)
 *     # . <'tuple'>
 *     if obj_dtype is tuple:             # <<<<<<<<<<<<<<
 *         return _ser_tuple(obj)
 * 
 */
  __pyx_t_1 = (__pyx_v_obj_dtype == (&PyTuple_Type));
  if (__pyx_t_1) {

    /* "serializor/binary/ser.py":1856
 *     # . <'tuple'>
 *     if obj_dtype is tuple:
 *         return _ser_tuple(obj)             # <<<<<<<<<<<<<<
 * 
 *     # Mapping Types
 */
    __Pyx_XDECREF(__pyx_r);
    if (!(likely(PyTuple_CheckExact(__pyx_v_obj))||((__pyx_v_obj) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v_obj))) __PYX_ERR(0, 1856, __pyx_L1_error)
    __pyx_t_2 = __pyx_f_10serializor_6binary_3ser__ser_tuple(((PyObject*)__pyx_v_obj)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1856, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_r = ((PyObject*)__pyx_t_2);
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "serializor/binary/ser.py":1855
 *         return _ser_list(obj)
 *     # . <'tuple'>
 *     if obj_dtype is tuple:             # <<<<<<<<<<<<<<
 *         return _ser_tuple(obj)
 * 
 */
  }

  /* "serializor/binary/ser.py":1860
 *     # Mapping Types
 *     # . <'dict'>
 *     if obj_dtype is dict:             # <<<<<<<<<<<<<<
 *         return _ser_dict(obj)
 * 
 */
  __pyx_t_1 = (__pyx_v_obj_dtype == (&PyDict_Type));
  if (__pyx_t_1) {

    /* "serializor/binary/ser.py":1861
 *     # . <'dict'>
 *     if obj_dtype is dict:
 *         return _ser_dict(obj)             # <<<<<<<<<<<<<<
 * 
 *     ##### Uncommon Types #####
 */
    __Pyx_XDECREF(__pyx_r);
    if (!(likely(PyDict_CheckExact(__pyx_v_obj))||((__pyx_v_obj) == Py_None) || __Pyx_RaiseUnexpectedTypeError("dict", __pyx_v_obj))) __PYX_ERR(0, 1861, __pyx_L1_error)
    __pyx_t_2 = __pyx_f_10serializor_6binary_3ser__ser_dict(((PyObject*)__pyx_v_obj)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1861, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_r = ((PyObject*)__pyx_t_2);
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "serializor/binary/ser.py":1860
 *     # Mapping Types
 *     # . <'dict'>
 *     if obj_dtype is dict:             # <<<<<<<<<<<<<<
 *         return _ser_dict(obj)
 * 
 */
  }

  /* "serializor/binary/ser.py":1864
 * 
 *     ##### Uncommon Types #####
 *     return _ser_uncommon(obj, obj_dtype)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __pyx_f_10serializor_6binary_3ser__ser_uncommon(__pyx_v_obj, __pyx_v_obj_dtype); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1864, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "serializor/binary/ser.py":1806
 * 
 * # Serialize -----------------------------------------------------------------------------------
 * @cython.cfunc             # <<<<<<<<<<<<<<
 * @cython.inline(True)
 * def _ser_common(obj: object) -> bytes:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("serializor.binary.ser._ser_common", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_obj_dtype);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "serializor/binary/ser.py":1867
 * 
 * 
 * @cython.cfunc             # <<<<<<<<<<<<<<
 * @cython.inline(True)
 * def _ser_uncommon(obj: object, obj_dtype: type) -> bytes:
 */

static PyObject *__pyx_f_10serializor_6binary_3ser__ser_uncommon(PyObject *__pyx_v_obj, PyTypeObject *__pyx_v_obj_dtype) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_ser_uncommon", 1);

  /* "serializor/binary/ser.py":1875
 *     # . <'numpy.float_'>
 *     if (
 *         obj_dtype is typeref.FLOAT64             # <<<<<<<<<<<<<<
 *         or obj_dtype is typeref.FLOAT32
 *         or obj_dtype is typeref.FLOAT16
 */
  __pyx_t_2 = (__pyx_v_obj_dtype == ((PyTypeObject*)__pyx_v_10serializor_7typeref_FLOAT64));
  if (!__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L4_bool_binop_done;
  }

  /* "serializor/binary/ser.py":1876
 *     if (
 *         obj_dtype is typeref.FLOAT64
 *         or obj_dtype is typeref.FLOAT32             # <<<<<<<<<<<<<<
 *         or obj_dtype is typeref.FLOAT16
 *     ):
 */
  __pyx_t_2 = (__pyx_v_obj_dtype == ((PyTypeObject*)__pyx_v_10serializor_7typeref_FLOAT32));
  if (!__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L4_bool_binop_done;
  }

  /* "serializor/binary/ser.py":1877
 *         obj_dtype is typeref.FLOAT64
 *         or obj_dtype is typeref.FLOAT32
 *         or obj_dtype is typeref.FLOAT16             # <<<<<<<<<<<<<<
 *     ):
 *         return _ser_float64(obj)
 */
  __pyx_t_2 = (__pyx_v_obj_dtype == ((PyTypeObject*)__pyx_v_10serializor_7typeref_FLOAT16));
  __pyx_t_1 = __pyx_t_2;
  __pyx_L4_bool_binop_done:;

  /* "serializor/binary/ser.py":1874
 *     # Basic Types
 *     # . <'numpy.float_'>
 *     if (             # <<<<<<<<<<<<<<
 *         obj_dtype is typeref.FLOAT64
 *         or obj_dtype is typeref.FLOAT32
 */
  if (__pyx_t_1) {

    /* "serializor/binary/ser.py":1879
 *         or obj_dtype is typeref.FLOAT16
 *     ):
 *         return _ser_float64(obj)             # <<<<<<<<<<<<<<
 *     # . <'numpy.int_'>
 *     if (
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_3 = __pyx_f_10serializor_6binary_3ser__ser_float64(__pyx_v_obj); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1879, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_r = ((PyObject*)__pyx_t_3);
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "serializor/binary/ser.py":1874
 *     # Basic Types
 *     # . <'numpy.float_'>
 *     if (             # <<<<<<<<<<<<<<
 *         obj_dtype is typeref.FLOAT64
 *         or obj_dtype is typeref.FLOAT32
 */
  }

  /* "serializor/binary/ser.py":1882
 *     # . <'numpy.int_'>
 *     if (
 *         obj_dtype is typeref.INT64             # <<<<<<<<<<<<<<
 *         or obj_dtype is typeref.INT32
 *         or obj_dtype is typeref.INT16
 */
  __pyx_t_2 = (__pyx_v_obj_dtype == ((PyTypeObject*)__pyx_v_10serializor_7typeref_INT64));
  if (!__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L8_bool_binop_done;
  }

  /* "serializor/binary/ser.py":1883
 *     if (
 *         obj_dtype is typeref.INT64
 *         or obj_dtype is typeref.INT32             # <<<<<<<<<<<<<<
 *         or obj_dtype is typeref.INT16
 *         or obj_dtype is typeref.INT8
 */
  __pyx_t_2 = (__pyx_v_obj_dtype == ((PyTypeObject*)__pyx_v_10serializor_7typeref_INT32));
  if (!__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L8_bool_binop_done;
  }

  /* "serializor/binary/ser.py":1884
 *         obj_dtype is typeref.INT64
 *         or obj_dtype is typeref.INT32
 *         or obj_dtype is typeref.INT16             # <<<<<<<<<<<<<<
 *         or obj_dtype is typeref.INT8
 *     ):
 */
  __pyx_t_2 = (__pyx_v_obj_dtype == ((PyTypeObject*)__pyx_v_10serializor_7typeref_INT16));
  if (!__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L8_bool_binop_done;
  }

  /* "serializor/binary/ser.py":1885
 *         or obj_dtype is typeref.INT32
 *         or obj_dtype is typeref.INT16
 *         or obj_dtype is typeref.INT8             # <<<<<<<<<<<<<<
 *     ):
 *         return _ser_int(obj)
 */
  __pyx_t_2 = (__pyx_v_obj_dtype == ((PyTypeObject*)__pyx_v_10serializor_7typeref_INT8));
  __pyx_t_1 = __pyx_t_2;
  __pyx_L8_bool_binop_done:;

  /* "serializor/binary/ser.py":1881
 *         return _ser_float64(obj)
 *     # . <'numpy.int_'>
 *     if (             # <<<<<<<<<<<<<<
 *         obj_dtype is typeref.INT64
 *         or obj_dtype is typeref.INT32
 */
  if (__pyx_t_1) {

    /* "serializor/binary/ser.py":1887
 *         or obj_dtype is typeref.INT8
 *     ):
 *         return _ser_int(obj)             # <<<<<<<<<<<<<<
 *     # . <'numpy.uint'>
 *     if (
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_3 = __pyx_f_10serializor_6binary_3ser__ser_int(__pyx_v_obj); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1887, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_r = ((PyObject*)__pyx_t_3);
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "serializor/binary/ser.py":1881
 *         return _ser_float64(obj)
 *     # . <'numpy.int_'>
 *     if (             # <<<<<<<<<<<<<<
 *         obj_dtype is typeref.INT64
 *         or obj_dtype is typeref.INT32
 */
  }

  /* "serializor/binary/ser.py":1890
 *     # . <'numpy.uint'>
 *     if (
 *         obj_dtype is typeref.UINT64             # <<<<<<<<<<<<<<
 *         or obj_dtype is typeref.UINT32
 *         or obj_dtype is typeref.UINT16
 */
  __pyx_t_2 = (__pyx_v_obj_dtype == ((PyTypeObject*)__pyx_v_10serializor_7typeref_UINT64));
  if (!__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L13_bool_binop_done;
  }

  /* "serializor/binary/ser.py":1891
 *     if (
 *         obj_dtype is typeref.UINT64
 *         or obj_dtype is typeref.UINT32             # <<<<<<<<<<<<<<
 *         or obj_dtype is typeref.UINT16
 *         or obj_dtype is typeref.UINT8
 */
  __pyx_t_2 = (__pyx_v_obj_dtype == ((PyTypeObject*)__pyx_v_10serializor_7typeref_UINT32));
  if (!__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L13_bool_binop_done;
  }

  /* "serializor/binary/ser.py":1892
 *         obj_dtype is typeref.UINT64
 *         or obj_dtype is typeref.UINT32
 *         or obj_dtype is typeref.UINT16             # <<<<<<<<<<<<<<
 *         or obj_dtype is typeref.UINT8
 *     ):
 */
  __pyx_t_2 = (__pyx_v_obj_dtype == ((PyTypeObject*)__pyx_v_10serializor_7typeref_UINT16));
  if (!__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L13_bool_binop_done;
  }

  /* "serializor/binary/ser.py":1893
 *         or obj_dtype is typeref.UINT32
 *         or obj_dtype is typeref.UINT16
 *         or obj_dtype is typeref.UINT8             # <<<<<<<<<<<<<<
 *     ):
 *         return _ser_int(obj)
 */
  __pyx_t_2 = (__pyx_v_obj_dtype == ((PyTypeObject*)__pyx_v_10serializor_7typeref_UINT8));
  __pyx_t_1 = __pyx_t_2;
  __pyx_L13_bool_binop_done:;

  /* "serializor/binary/ser.py":1889
 *         return _ser_int(obj)
 *     # . <'numpy.uint'>
 *     if (             # <<<<<<<<<<<<<<
 *         obj_dtype is typeref.UINT64
 *         or obj_dtype is typeref.UINT32
 */
  if (__pyx_t_1) {

    /* "serializor/binary/ser.py":1895
 *         or obj_dtype is typeref.UINT8
 *     ):
 *         return _ser_int(obj)             # <<<<<<<<<<<<<<
 *     # . <'numpy.bool_'>
 *     if obj_dtype is typeref.BOOL_:
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_3 = __pyx_f_10serializor_6binary_3ser__ser_int(__pyx_v_obj); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1895, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_r = ((PyObject*)__pyx_t_3);
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "serializor/binary/ser.py":1889
 *         return _ser_int(obj)
 *     # . <'numpy.uint'>
 *     if (             # <<<<<<<<<<<<<<
 *         obj_dtype is typeref.UINT64
 *         or obj_dtype is typeref.UINT32
 */
  }

  /* "serializor/binary/ser.py":1897
 *         return _ser_int(obj)
 *     # . <'numpy.bool_'>
 *     if obj_dtype is typeref.BOOL_:             # <<<<<<<<<<<<<<
 *         return _ser_bool(obj)
 * 
 */
  __pyx_t_1 = (__pyx_v_obj_dtype == ((PyTypeObject*)__pyx_v_10serializor_7typeref_BOOL_));
  if (__pyx_t_1) {

    /* "serializor/binary/ser.py":1898
 *     # . <'numpy.bool_'>
 *     if obj_dtype is typeref.BOOL_:
 *         return _ser_bool(obj)             # <<<<<<<<<<<<<<
 * 
 *     # Date&Time Types
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_3 = __pyx_f_10serializor_6binary_3ser__ser_bool(__pyx_v_obj); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1898, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_r = ((PyObject*)__pyx_t_3);
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "serializor/binary/ser.py":1897
 *         return _ser_int(obj)
 *     # . <'numpy.bool_'>
 *     if obj_dtype is typeref.BOOL_:             # <<<<<<<<<<<<<<
 *         return _ser_bool(obj)
 * 
 */
  }

  /* "serializor/binary/ser.py":1902
 *     # Date&Time Types
 *     # . <'numpy.datetime64'>
 *     if obj_dtype is typeref.DATETIME64:             # <<<<<<<<<<<<<<
 *         return _ser_datetime64(obj)
 *     # . <'numpy.timedelta64'>
 */
  __pyx_t_1 = (__pyx_v_obj_dtype == ((PyTypeObject*)__pyx_v_10serializor_7typeref_DATETIME64));
  if (__pyx_t_1) {

    /* "serializor/binary/ser.py":1903
 *     # . <'numpy.datetime64'>
 *     if obj_dtype is typeref.DATETIME64:
 *         return _ser_datetime64(obj)             # <<<<<<<<<<<<<<
 *     # . <'numpy.timedelta64'>
 *     if obj_dtype is typeref.TIMEDELTA64:
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_3 = __pyx_f_10serializor_6binary_3ser__ser_datetime64(__pyx_v_obj); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1903, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_r = ((PyObject*)__pyx_t_3);
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "serializor/binary/ser.py":1902
 *     # Date&Time Types
 *     # . <'numpy.datetime64'>
 *     if obj_dtype is typeref.DATETIME64:             # <<<<<<<<<<<<<<
 *         return _ser_datetime64(obj)
 *     # . <'numpy.timedelta64'>
 */
  }

  /* "serializor/binary/ser.py":1905
 *         return _ser_datetime64(obj)
 *     # . <'numpy.timedelta64'>
 *     if obj_dtype is typeref.TIMEDELTA64:             # <<<<<<<<<<<<<<
 *         return _ser_timedelta64(obj)
 *     # . <'pandas.Timestamp'>
 */
  __pyx_t_1 = (__pyx_v_obj_dtype == ((PyTypeObject*)__pyx_v_10serializor_7typeref_TIMEDELTA64));
  if (__pyx_t_1) {

    /* "serializor/binary/ser.py":1906
 *     # . <'numpy.timedelta64'>
 *     if obj_dtype is typeref.TIMEDELTA64:
 *         return _ser_timedelta64(obj)             # <<<<<<<<<<<<<<
 *     # . <'pandas.Timestamp'>
 *     if obj_dtype is typeref.PD_TIMESTAMP:
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_3 = __pyx_f_10serializor_6binary_3ser__ser_timedelta64(__pyx_v_obj); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1906, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_r = ((PyObject*)__pyx_t_3);
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "serializor/binary/ser.py":1905
 *         return _ser_datetime64(obj)
 *     # . <'numpy.timedelta64'>
 *     if obj_dtype is typeref.TIMEDELTA64:             # <<<<<<<<<<<<<<
 *         return _ser_timedelta64(obj)
 *     # . <'pandas.Timestamp'>
 */
  }

  /* "serializor/binary/ser.py":1908
 *         return _ser_timedelta64(obj)
 *     # . <'pandas.Timestamp'>
 *     if obj_dtype is typeref.PD_TIMESTAMP:             # <<<<<<<<<<<<<<
 *         return _ser_pd_timestamp(obj)
 *     # . <'pandas.Timedelta'>`
 */
  __pyx_t_1 = (__pyx_v_obj_dtype == ((PyTypeObject*)__pyx_v_10serializor_7typeref_PD_TIMESTAMP));
  if (__pyx_t_1) {

    /* "serializor/binary/ser.py":1909
 *     # . <'pandas.Timestamp'>
 *     if obj_dtype is typeref.PD_TIMESTAMP:
 *         return _ser_pd_timestamp(obj)             # <<<<<<<<<<<<<<
 *     # . <'pandas.Timedelta'>`
 *     if obj_dtype is typeref.PD_TIMEDELTA:
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_3 = __pyx_f_10serializor_6binary_3ser__ser_pd_timestamp(__pyx_v_obj); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1909, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_r = ((PyObject*)__pyx_t_3);
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "serializor/binary/ser.py":1908
 *         return _ser_timedelta64(obj)
 *     # . <'pandas.Timestamp'>
 *     if obj_dtype is typeref.PD_TIMESTAMP:             # <<<<<<<<<<<<<<
 *         return _ser_pd_timestamp(obj)
 *     # . <'pandas.Timedelta'>`
 */
  }

  /* "serializor/binary/ser.py":1911
 *         return _ser_pd_timestamp(obj)
 *     # . <'pandas.Timedelta'>`
 *     if obj_dtype is typeref.PD_TIMEDELTA:             # <<<<<<<<<<<<<<
 *         return _ser_pd_timedelta(obj)
 *     # . <'time.struct_time'>`
 */
  __pyx_t_1 = (__pyx_v_obj_dtype == ((PyTypeObject*)__pyx_v_10serializor_7typeref_PD_TIMEDELTA));
  if (__pyx_t_1) {

    /* "serializor/binary/ser.py":1912
 *     # . <'pandas.Timedelta'>`
 *     if obj_dtype is typeref.PD_TIMEDELTA:
 *         return _ser_pd_timedelta(obj)             # <<<<<<<<<<<<<<
 *     # . <'time.struct_time'>`
 *     if obj_dtype is typeref.STRUCT_TIME:
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_3 = __pyx_f_10serializor_6binary_3ser__ser_pd_timedelta(__pyx_v_obj); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1912, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_r = ((PyObject*)__pyx_t_3);
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "serializor/binary/ser.py":1911
 *         return _ser_pd_timestamp(obj)
 *     # . <'pandas.Timedelta'>`
 *     if obj_dtype is typeref.PD_TIMEDELTA:             # <<<<<<<<<<<<<<
 *         return _ser_pd_timedelta(obj)
 *     # . <'time.struct_time'>`
 */
  }

  /* "serializor/binary/ser.py":1914
 *         return _ser_pd_timedelta(obj)
 *     # . <'time.struct_time'>`
 *     if obj_dtype is typeref.STRUCT_TIME:             # <<<<<<<<<<<<<<
 *         return _ser_struct_time(obj)
 * 
 */
  __pyx_t_1 = (__pyx_v_obj_dtype == ((PyTypeObject*)__pyx_v_10serializor_7typeref_STRUCT_TIME));
  if (__pyx_t_1) {

    /* "serializor/binary/ser.py":1915
 *     # . <'time.struct_time'>`
 *     if obj_dtype is typeref.STRUCT_TIME:
 *         return _ser_struct_time(obj)             # <<<<<<<<<<<<<<
 * 
 *     # Numeric Types
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_3 = __pyx_f_10serializor_6binary_3ser__ser_struct_time(__pyx_v_obj); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1915, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_r = ((PyObject*)__pyx_t_3);
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "serializor/binary/ser.py":1914
 *         return _ser_pd_timedelta(obj)
 *     # . <'time.struct_time'>`
 *     if obj_dtype is typeref.STRUCT_TIME:             # <<<<<<<<<<<<<<
 *         return _ser_struct_time(obj)
 * 
 */
  }

  /* "serializor/binary/ser.py":1919
 *     # Numeric Types
 *     # . <'decimal.Decimal'>
 *     if obj_dtype is typeref.DECIMAL:             # <<<<<<<<<<<<<<
 *         return _ser_decimal(obj)
 *     # . <'complex'>
 */
  __pyx_t_1 = (__pyx_v_obj_dtype == ((PyTypeObject*)__pyx_v_10serializor_7typeref_DECIMAL));
  if (__pyx_t_1) {

    /* "serializor/binary/ser.py":1920
 *     # . <'decimal.Decimal'>
 *     if obj_dtype is typeref.DECIMAL:
 *         return _ser_decimal(obj)             # <<<<<<<<<<<<<<
 *     # . <'complex'>
 *     if obj_dtype is complex:
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_3 = __pyx_f_10serializor_6binary_3ser__ser_decimal(__pyx_v_obj); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1920, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_r = ((PyObject*)__pyx_t_3);
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "serializor/binary/ser.py":1919
 *     # Numeric Types
 *     # . <'decimal.Decimal'>
 *     if obj_dtype is typeref.DECIMAL:             # <<<<<<<<<<<<<<
 *         return _ser_decimal(obj)
 *     # . <'complex'>
 */
  }

  /* "serializor/binary/ser.py":1922
 *         return _ser_decimal(obj)
 *     # . <'complex'>
 *     if obj_dtype is complex:             # <<<<<<<<<<<<<<
 *         return _ser_complex(obj)
 *     # . <'numpy.complex128'>
 */
  __pyx_t_1 = (__pyx_v_obj_dtype == (&PyComplex_Type));
  if (__pyx_t_1) {

    /* "serializor/binary/ser.py":1923
 *     # . <'complex'>
 *     if obj_dtype is complex:
 *         return _ser_complex(obj)             # <<<<<<<<<<<<<<
 *     # . <'numpy.complex128'>
 *     if obj_dtype is typeref.COMPLEX128:
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_3 = __pyx_f_10serializor_6binary_3ser__ser_complex(__pyx_v_obj); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1923, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_r = ((PyObject*)__pyx_t_3);
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "serializor/binary/ser.py":1922
 *         return _ser_decimal(obj)
 *     # . <'complex'>
 *     if obj_dtype is complex:             # <<<<<<<<<<<<<<
 *         return _ser_complex(obj)
 *     # . <'numpy.complex128'>
 */
  }

  /* "serializor/binary/ser.py":1925
 *         return _ser_complex(obj)
 *     # . <'numpy.complex128'>
 *     if obj_dtype is typeref.COMPLEX128:             # <<<<<<<<<<<<<<
 *         return _ser_complex128(obj)
 *     # . <'numpy.complex64'>
 */
  __pyx_t_1 = (__pyx_v_obj_dtype == ((PyTypeObject*)__pyx_v_10serializor_7typeref_COMPLEX128));
  if (__pyx_t_1) {

    /* "serializor/binary/ser.py":1926
 *     # . <'numpy.complex128'>
 *     if obj_dtype is typeref.COMPLEX128:
 *         return _ser_complex128(obj)             # <<<<<<<<<<<<<<
 *     # . <'numpy.complex64'>
 *     if obj_dtype is typeref.COMPLEX64:
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_3 = __pyx_f_10serializor_6binary_3ser__ser_complex128(__pyx_v_obj); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1926, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_r = ((PyObject*)__pyx_t_3);
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "serializor/binary/ser.py":1925
 *         return _ser_complex(obj)
 *     # . <'numpy.complex128'>
 *     if obj_dtype is typeref.COMPLEX128:             # <<<<<<<<<<<<<<
 *         return _ser_complex128(obj)
 *     # . <'numpy.complex64'>
 */
  }

  /* "serializor/binary/ser.py":1928
 *         return _ser_complex128(obj)
 *     # . <'numpy.complex64'>
 *     if obj_dtype is typeref.COMPLEX64:             # <<<<<<<<<<<<<<
 *         return _ser_complex64(obj)
 * 
 */
  __pyx_t_1 = (__pyx_v_obj_dtype == ((PyTypeObject*)__pyx_v_10serializor_7typeref_COMPLEX64));
  if (__pyx_t_1) {

    /* "serializor/binary/ser.py":1929
 *     # . <'numpy.complex64'>
 *     if obj_dtype is typeref.COMPLEX64:
 *         return _ser_complex64(obj)             # <<<<<<<<<<<<<<
 * 
 *     # Bytes Types
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_3 = __pyx_f_10serializor_6binary_3ser__ser_complex64(__pyx_v_obj); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1929, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_r = ((PyObject*)__pyx_t_3);
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "serializor/binary/ser.py":1928
 *         return _ser_complex128(obj)
 *     # . <'numpy.complex64'>
 *     if obj_dtype is typeref.COMPLEX64:             # <<<<<<<<<<<<<<
 *         return _ser_complex64(obj)
 * 
 */
  }

  /* "serializor/binary/ser.py":1933
 *     # Bytes Types
 *     # . <'bytearray'>
 *     if obj_dtype is bytearray:             # <<<<<<<<<<<<<<
 *         return _ser_bytearray(obj)
 *     # . <'memoryview'>
 */
  __pyx_t_1 = (__pyx_v_obj_dtype == (&PyByteArray_Type));
  if (__pyx_t_1) {

    /* "serializor/binary/ser.py":1934
 *     # . <'bytearray'>
 *     if obj_dtype is bytearray:
 *         return _ser_bytearray(obj)             # <<<<<<<<<<<<<<
 *     # . <'memoryview'>
 *     if obj_dtype is memoryview:
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_3 = __pyx_f_10serializor_6binary_3ser__ser_bytearray(__pyx_v_obj); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1934, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_r = ((PyObject*)__pyx_t_3);
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "serializor/binary/ser.py":1933
 *     # Bytes Types
 *     # . <'bytearray'>
 *     if obj_dtype is bytearray:             # <<<<<<<<<<<<<<
 *         return _ser_bytearray(obj)
 *     # . <'memoryview'>
 */
  }

  /* "serializor/binary/ser.py":1936
 *         return _ser_bytearray(obj)
 *     # . <'memoryview'>
 *     if obj_dtype is memoryview:             # <<<<<<<<<<<<<<
 *         return _ser_memoryview(obj)
 *     # . <'numpy.bytes_'>
 */
  __pyx_t_1 = (__pyx_v_obj_dtype == (&PyMemoryView_Type));
  if (__pyx_t_1) {

    /* "serializor/binary/ser.py":1937
 *     # . <'memoryview'>
 *     if obj_dtype is memoryview:
 *         return _ser_memoryview(obj)             # <<<<<<<<<<<<<<
 *     # . <'numpy.bytes_'>
 *     if obj_dtype is typeref.BYTES_:
 */
    __Pyx_XDECREF(__pyx_r);
    if (!(likely(PyMemoryView_Check(__pyx_v_obj))||((__pyx_v_obj) == Py_None) || __Pyx_RaiseUnexpectedTypeError("memoryview", __pyx_v_obj))) __PYX_ERR(0, 1937, __pyx_L1_error)
    __pyx_t_3 = __pyx_f_10serializor_6binary_3ser__ser_memoryview(((PyObject*)__pyx_v_obj)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1937, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_r = ((PyObject*)__pyx_t_3);
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "serializor/binary/ser.py":1936
 *         return _ser_bytearray(obj)
 *     # . <'memoryview'>
 *     if obj_dtype is memoryview:             # <<<<<<<<<<<<<<
 *         return _ser_memoryview(obj)
 *     # . <'numpy.bytes_'>
 */
  }

  /* "serializor/binary/ser.py":1939
 *         return _ser_memoryview(obj)
 *     # . <'numpy.bytes_'>
 *     if obj_dtype is typeref.BYTES_:             # <<<<<<<<<<<<<<
 *         return _ser_bytes(obj)
 * 
 */
  __pyx_t_1 = (__pyx_v_obj_dtype == ((PyTypeObject*)__pyx_v_10serializor_7typeref_BYTES_));
  if (__pyx_t_1) {

    /* "serializor/binary/ser.py":1940
 *     # . <'numpy.bytes_'>
 *     if obj_dtype is typeref.BYTES_:
 *         return _ser_bytes(obj)             # <<<<<<<<<<<<<<
 * 
 *     # String Types:
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_3 = __pyx_f_10serializor_6binary_3ser__ser_bytes(__pyx_v_obj); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1940, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_r = ((PyObject*)__pyx_t_3);
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "serializor/binary/ser.py":1939
 *         return _ser_memoryview(obj)
 *     # . <'numpy.bytes_'>
 *     if obj_dtype is typeref.BYTES_:             # <<<<<<<<<<<<<<
 *         return _ser_bytes(obj)
 * 
 */
  }

  /* "serializor/binary/ser.py":1944
 *     # String Types:
 *     # . <'numpy.str_'>
 *     if obj_dtype is typeref.STR_:             # <<<<<<<<<<<<<<
 *         return _ser_str(obj)
 * 
 */
  __pyx_t_1 = (__pyx_v_obj_dtype == ((PyTypeObject*)__pyx_v_10serializor_7typeref_STR_));
  if (__pyx_t_1) {

    /* "serializor/binary/ser.py":1945
 *     # . <'numpy.str_'>
 *     if obj_dtype is typeref.STR_:
 *         return _ser_str(obj)             # <<<<<<<<<<<<<<
 * 
 *     # Sequence Types
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_3 = __pyx_f_10serializor_6binary_3ser__ser_str(__pyx_v_obj); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1945, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_r = ((PyObject*)__pyx_t_3);
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "serializor/binary/ser.py":1944
 *     # String Types:
 *     # . <'numpy.str_'>
 *     if obj_dtype is typeref.STR_:             # <<<<<<<<<<<<<<
 *         return _ser_str(obj)
 * 
 */
  }

  /* "serializor/binary/ser.py":1949
 *     # Sequence Types
 *     # . <'set'>
 *     if obj_dtype is set:             # <<<<<<<<<<<<<<
 *         return _ser_set(obj)
 *     # . <'frozenset'>
 */
  __pyx_t_1 = (__pyx_v_obj_dtype == (&PySet_Type));
  if (__pyx_t_1) {

    /* "serializor/binary/ser.py":1950
 *     # . <'set'>
 *     if obj_dtype is set:
 *         return _ser_set(obj)             # <<<<<<<<<<<<<<
 *     # . <'frozenset'>
 *     if obj_dtype is frozenset:
 */
    __Pyx_XDECREF(__pyx_r);
    if (!(likely(PySet_CheckExact(__pyx_v_obj))||((__pyx_v_obj) == Py_None) || __Pyx_RaiseUnexpectedTypeError("set", __pyx_v_obj))) __PYX_ERR(0, 1950, __pyx_L1_error)
    __pyx_t_3 = __pyx_f_10serializor_6binary_3ser__ser_set(((PyObject*)__pyx_v_obj)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1950, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_r = ((PyObject*)__pyx_t_3);
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "serializor/binary/ser.py":1949
 *     # Sequence Types
 *     # . <'set'>
 *     if obj_dtype is set:             # <<<<<<<<<<<<<<
 *         return _ser_set(obj)
 *     # . <'frozenset'>
 */
  }

  /* "serializor/binary/ser.py":1952
 *         return _ser_set(obj)
 *     # . <'frozenset'>
 *     if obj_dtype is frozenset:             # <<<<<<<<<<<<<<
 *         return _ser_frozenset(obj)
 *     # . <'range'>
 */
  __pyx_t_1 = (__pyx_v_obj_dtype == (&PyFrozenSet_Type));
  if (__pyx_t_1) {

    /* "serializor/binary/ser.py":1953
 *     # . <'frozenset'>
 *     if obj_dtype is frozenset:
 *         return _ser_frozenset(obj)             # <<<<<<<<<<<<<<
 *     # . <'range'>
 *     if obj_dtype is range:
 */
    __Pyx_XDECREF(__pyx_r);
    if (!(likely(PyFrozenSet_CheckExact(__pyx_v_obj))||((__pyx_v_obj) == Py_None) || __Pyx_RaiseUnexpectedTypeError("frozenset", __pyx_v_obj))) __PYX_ERR(0, 1953, __pyx_L1_error)
    __pyx_t_3 = __pyx_f_10serializor_6binary_3ser__ser_frozenset(((PyObject*)__pyx_v_obj)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1953, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_r = ((PyObject*)__pyx_t_3);
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "serializor/binary/ser.py":1952
 *         return _ser_set(obj)
 *     # . <'frozenset'>
 *     if obj_dtype is frozenset:             # <<<<<<<<<<<<<<
 *         return _ser_frozenset(obj)
 *     # . <'range'>
 */
  }

  /* "serializor/binary/ser.py":1955
 *         return _ser_frozenset(obj)
 *     # . <'range'>
 *     if obj_dtype is range:             # <<<<<<<<<<<<<<
 *         return _ser_range(obj)
 *     # . <'dict_keys'> & <'dict_values'> & <'collections.deque'>
 */
  __pyx_t_1 = (__pyx_v_obj_dtype == ((PyTypeObject*)__pyx_builtin_range));
  if (__pyx_t_1) {

    /* "serializor/binary/ser.py":1956
 *     # . <'range'>
 *     if obj_dtype is range:
 *         return _ser_range(obj)             # <<<<<<<<<<<<<<
 *     # . <'dict_keys'> & <'dict_values'> & <'collections.deque'>
 *     if (
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_3 = __pyx_f_10serializor_6binary_3ser__ser_range(__pyx_v_obj); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1956, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_r = ((PyObject*)__pyx_t_3);
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "serializor/binary/ser.py":1955
 *         return _ser_frozenset(obj)
 *     # . <'range'>
 *     if obj_dtype is range:             # <<<<<<<<<<<<<<
 *         return _ser_range(obj)
 *     # . <'dict_keys'> & <'dict_values'> & <'collections.deque'>
 */
  }

  /* "serializor/binary/ser.py":1959
 *     # . <'dict_keys'> & <'dict_values'> & <'collections.deque'>
 *     if (
 *         obj_dtype is typeref.DICT_KEYS             # <<<<<<<<<<<<<<
 *         or obj_dtype is typeref.DICT_VALUES
 *         or obj_dtype is typeref.DEQUE
 */
  __pyx_t_2 = (__pyx_v_obj_dtype == ((PyTypeObject*)__pyx_v_10serializor_7typeref_DICT_KEYS));
  if (!__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L35_bool_binop_done;
  }

  /* "serializor/binary/ser.py":1960
 *     if (
 *         obj_dtype is typeref.DICT_KEYS
 *         or obj_dtype is typeref.DICT_VALUES             # <<<<<<<<<<<<<<
 *         or obj_dtype is typeref.DEQUE
 *     ):
 */
  __pyx_t_2 = (__pyx_v_obj_dtype == ((PyTypeObject*)__pyx_v_10serializor_7typeref_DICT_VALUES));
  if (!__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L35_bool_binop_done;
  }

  /* "serializor/binary/ser.py":1961
 *         obj_dtype is typeref.DICT_KEYS
 *         or obj_dtype is typeref.DICT_VALUES
 *         or obj_dtype is typeref.DEQUE             # <<<<<<<<<<<<<<
 *     ):
 *         return _ser_sequence(obj)
 */
  __pyx_t_2 = (__pyx_v_obj_dtype == ((PyTypeObject*)__pyx_v_10serializor_7typeref_DEQUE));
  __pyx_t_1 = __pyx_t_2;
  __pyx_L35_bool_binop_done:;

  /* "serializor/binary/ser.py":1958
 *         return _ser_range(obj)
 *     # . <'dict_keys'> & <'dict_values'> & <'collections.deque'>
 *     if (             # <<<<<<<<<<<<<<
 *         obj_dtype is typeref.DICT_KEYS
 *         or obj_dtype is typeref.DICT_VALUES
 */
  if (__pyx_t_1) {

    /* "serializor/binary/ser.py":1963
 *         or obj_dtype is typeref.DEQUE
 *     ):
 *         return _ser_sequence(obj)             # <<<<<<<<<<<<<<
 * 
 *     # Mapping Types
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_3 = __pyx_f_10serializor_6binary_3ser__ser_sequence(__pyx_v_obj); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1963, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_r = ((PyObject*)__pyx_t_3);
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "serializor/binary/ser.py":1958
 *         return _ser_range(obj)
 *     # . <'dict_keys'> & <'dict_values'> & <'collections.deque'>
 *     if (             # <<<<<<<<<<<<<<
 *         obj_dtype is typeref.DICT_KEYS
 *         or obj_dtype is typeref.DICT_VALUES
 */
  }

  /* "serializor/binary/ser.py":1967
 *     # Mapping Types
 *     # . <'dict_items'>
 *     if obj_dtype is typeref.DICT_ITEMS:             # <<<<<<<<<<<<<<
 *         return _ser_dict(dict(obj))
 * 
 */
  __pyx_t_1 = (__pyx_v_obj_dtype == ((PyTypeObject*)__pyx_v_10serializor_7typeref_DICT_ITEMS));
  if (__pyx_t_1) {

    /* "serializor/binary/ser.py":1968
 *     # . <'dict_items'>
 *     if obj_dtype is typeref.DICT_ITEMS:
 *         return _ser_dict(dict(obj))             # <<<<<<<<<<<<<<
 * 
 *     # NumPy Types
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_3 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyDict_Type)), __pyx_v_obj); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1968, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = __pyx_f_10serializor_6binary_3ser__ser_dict(((PyObject*)__pyx_t_3)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1968, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_r = ((PyObject*)__pyx_t_4);
    __pyx_t_4 = 0;
    goto __pyx_L0;

    /* "serializor/binary/ser.py":1967
 *     # Mapping Types
 *     # . <'dict_items'>
 *     if obj_dtype is typeref.DICT_ITEMS:             # <<<<<<<<<<<<<<
 *         return _ser_dict(dict(obj))
 * 
 */
  }

  /* "serializor/binary/ser.py":1972
 *     # NumPy Types
 *     # . <'numpy.ndarray'>
 *     if obj_dtype is np.ndarray:             # <<<<<<<<<<<<<<
 *         return _ser_ndarray(obj)
 *     # . <'numpy.record'>
 */
  __pyx_t_1 = (__pyx_v_obj_dtype == __pyx_ptype_5numpy_ndarray);
  if (__pyx_t_1) {

    /* "serializor/binary/ser.py":1973
 *     # . <'numpy.ndarray'>
 *     if obj_dtype is np.ndarray:
 *         return _ser_ndarray(obj)             # <<<<<<<<<<<<<<
 *     # . <'numpy.record'>
 *     if obj_dtype is typeref.RECORD:
 */
    __Pyx_XDECREF(__pyx_r);
    if (!(likely(((__pyx_v_obj) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_obj, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 1973, __pyx_L1_error)
    __pyx_t_4 = __pyx_f_10serializor_6binary_3ser__ser_ndarray(((PyArrayObject *)__pyx_v_obj)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1973, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_r = ((PyObject*)__pyx_t_4);
    __pyx_t_4 = 0;
    goto __pyx_L0;

    /* "serializor/binary/ser.py":1972
 *     # NumPy Types
 *     # . <'numpy.ndarray'>
 *     if obj_dtype is np.ndarray:             # <<<<<<<<<<<<<<
 *         return _ser_ndarray(obj)
 *     # . <'numpy.record'>
 */
  }

  /* "serializor/binary/ser.py":1975
 *         return _ser_ndarray(obj)
 *     # . <'numpy.record'>
 *     if obj_dtype is typeref.RECORD:             # <<<<<<<<<<<<<<
 *         return _ser_sequence(obj)
 * 
 */
  __pyx_t_1 = (__pyx_v_obj_dtype == ((PyTypeObject*)__pyx_v_10serializor_7typeref_RECORD));
  if (__pyx_t_1) {

    /* "serializor/binary/ser.py":1976
 *     # . <'numpy.record'>
 *     if obj_dtype is typeref.RECORD:
 *         return _ser_sequence(obj)             # <<<<<<<<<<<<<<
 * 
 *     # Pandas Types
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __pyx_f_10serializor_6binary_3ser__ser_sequence(__pyx_v_obj); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1976, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_r = ((PyObject*)__pyx_t_4);
    __pyx_t_4 = 0;
    goto __pyx_L0;

    /* "serializor/binary/ser.py":1975
 *         return _ser_ndarray(obj)
 *     # . <'numpy.record'>
 *     if obj_dtype is typeref.RECORD:             # <<<<<<<<<<<<<<
 *         return _ser_sequence(obj)
 * 
 */
  }

  /* "serializor/binary/ser.py":1980
 *     # Pandas Types
 *     # . <'pandas.Series'>
 *     if obj_dtype is typeref.SERIES:             # <<<<<<<<<<<<<<
 *         return _ser_series(obj)
 *     # . <'pandas.DataFrame'>
 */
  __pyx_t_1 = (__pyx_v_obj_dtype == ((PyTypeObject*)__pyx_v_10serializor_7typeref_SERIES));
  if (__pyx_t_1) {

    /* "serializor/binary/ser.py":1981
 *     # . <'pandas.Series'>
 *     if obj_dtype is typeref.SERIES:
 *         return _ser_series(obj)             # <<<<<<<<<<<<<<
 *     # . <'pandas.DataFrame'>
 *     if obj_dtype is typeref.DATAFRAME:
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __pyx_f_10serializor_6binary_3ser__ser_series(__pyx_v_obj); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1981, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_r = ((PyObject*)__pyx_t_4);
    __pyx_t_4 = 0;
    goto __pyx_L0;

    /* "serializor/binary/ser.py":1980
 *     # Pandas Types
 *     # . <'pandas.Series'>
 *     if obj_dtype is typeref.SERIES:             # <<<<<<<<<<<<<<
 *         return _ser_series(obj)
 *     # . <'pandas.DataFrame'>
 */
  }

  /* "serializor/binary/ser.py":1983
 *         return _ser_series(obj)
 *     # . <'pandas.DataFrame'>
 *     if obj_dtype is typeref.DATAFRAME:             # <<<<<<<<<<<<<<
 *         return _ser_dataframe(obj)
 *     # . <'pandas.DatetimeIndex'>
 */
  __pyx_t_1 = (__pyx_v_obj_dtype == ((PyTypeObject*)__pyx_v_10serializor_7typeref_DATAFRAME));
  if (__pyx_t_1) {

    /* "serializor/binary/ser.py":1984
 *     # . <'pandas.DataFrame'>
 *     if obj_dtype is typeref.DATAFRAME:
 *         return _ser_dataframe(obj)             # <<<<<<<<<<<<<<
 *     # . <'pandas.DatetimeIndex'>
 *     if obj_dtype is typeref.DATETIMEINDEX:
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __pyx_f_10serializor_6binary_3ser__ser_dataframe(__pyx_v_obj); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1984, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_r = ((PyObject*)__pyx_t_4);
    __pyx_t_4 = 0;
    goto __pyx_L0;

    /* "serializor/binary/ser.py":1983
 *         return _ser_series(obj)
 *     # . <'pandas.DataFrame'>
 *     if obj_dtype is typeref.DATAFRAME:             # <<<<<<<<<<<<<<
 *         return _ser_dataframe(obj)
 *     # . <'pandas.DatetimeIndex'>
 */
  }

  /* "serializor/binary/ser.py":1986
 *         return _ser_dataframe(obj)
 *     # . <'pandas.DatetimeIndex'>
 *     if obj_dtype is typeref.DATETIMEINDEX:             # <<<<<<<<<<<<<<
 *         return _ser_datetime_index(obj)
 *     # . <'pandas.TimedeltaIndex'>
 */
  __pyx_t_1 = (__pyx_v_obj_dtype == ((PyTypeObject*)__pyx_v_10serializor_7typeref_DATETIMEINDEX));
  if (__pyx_t_1) {

    /* "serializor/binary/ser.py":1987
 *     # . <'pandas.DatetimeIndex'>
 *     if obj_dtype is typeref.DATETIMEINDEX:
 *         return _ser_datetime_index(obj)             # <<<<<<<<<<<<<<
 *     # . <'pandas.TimedeltaIndex'>
 *     if obj_dtype is typeref.TIMEDELTAINDEX:
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __pyx_f_10serializor_6binary_3ser__ser_datetime_index(__pyx_v_obj); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1987, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_r = ((PyObject*)__pyx_t_4);
    __pyx_t_4 = 0;
    goto __pyx_L0;

    /* "serializor/binary/ser.py":1986
 *         return _ser_dataframe(obj)
 *     # . <'pandas.DatetimeIndex'>
 *     if obj_dtype is typeref.DATETIMEINDEX:             # <<<<<<<<<<<<<<
 *         return _ser_datetime_index(obj)
 *     # . <'pandas.TimedeltaIndex'>
 */
  }

  /* "serializor/binary/ser.py":1989
 *         return _ser_datetime_index(obj)
 *     # . <'pandas.TimedeltaIndex'>
 *     if obj_dtype is typeref.TIMEDELTAINDEX:             # <<<<<<<<<<<<<<
 *         return _ser_timedelta_index(obj)
 * 
 */
  __pyx_t_1 = (__pyx_v_obj_dtype == ((PyTypeObject*)__pyx_v_10serializor_7typeref_TIMEDELTAINDEX));
  if (__pyx_t_1) {

    /* "serializor/binary/ser.py":1990
 *     # . <'pandas.TimedeltaIndex'>
 *     if obj_dtype is typeref.TIMEDELTAINDEX:
 *         return _ser_timedelta_index(obj)             # <<<<<<<<<<<<<<
 * 
 *     # Cytimes Types
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __pyx_f_10serializor_6binary_3ser__ser_timedelta_index(__pyx_v_obj); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1990, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_r = ((PyObject*)__pyx_t_4);
    __pyx_t_4 = 0;
    goto __pyx_L0;

    /* "serializor/binary/ser.py":1989
 *         return _ser_datetime_index(obj)
 *     # . <'pandas.TimedeltaIndex'>
 *     if obj_dtype is typeref.TIMEDELTAINDEX:             # <<<<<<<<<<<<<<
 *         return _ser_timedelta_index(obj)
 * 
 */
  }

  /* "serializor/binary/ser.py":1993
 * 
 *     # Cytimes Types
 *     if typeref.CYTIMES_AVAILABLE:             # <<<<<<<<<<<<<<
 *         # . <'cytimes.pydt'>
 *         if obj_dtype is typeref.PYDT:
 */
  if (__pyx_v_10serializor_7typeref_CYTIMES_AVAILABLE) {

    /* "serializor/binary/ser.py":1995
 *     if typeref.CYTIMES_AVAILABLE:
 *         # . <'cytimes.pydt'>
 *         if obj_dtype is typeref.PYDT:             # <<<<<<<<<<<<<<
 *             return _ser_datetime(obj.dt)
 *         # . <'cytimes.pddt'>
 */
    __pyx_t_1 = (__pyx_v_obj_dtype == ((PyTypeObject*)__pyx_v_10serializor_7typeref_PYDT));
    if (__pyx_t_1) {

      /* "serializor/binary/ser.py":1996
 *         # . <'cytimes.pydt'>
 *         if obj_dtype is typeref.PYDT:
 *             return _ser_datetime(obj.dt)             # <<<<<<<<<<<<<<
 *         # . <'cytimes.pddt'>
 *         if obj_dtype is typeref.PDDT:
 */
      __Pyx_XDECREF(__pyx_r);
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_obj, __pyx_n_s_dt); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1996, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = __pyx_f_10serializor_6binary_3ser__ser_datetime(__pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1996, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_r = ((PyObject*)__pyx_t_3);
      __pyx_t_3 = 0;
      goto __pyx_L0;

      /* "serializor/binary/ser.py":1995
 *     if typeref.CYTIMES_AVAILABLE:
 *         # . <'cytimes.pydt'>
 *         if obj_dtype is typeref.PYDT:             # <<<<<<<<<<<<<<
 *             return _ser_datetime(obj.dt)
 *         # . <'cytimes.pddt'>
 */
    }

    /* "serializor/binary/ser.py":1998
 *             return _ser_datetime(obj.dt)
 *         # . <'cytimes.pddt'>
 *         if obj_dtype is typeref.PDDT:             # <<<<<<<<<<<<<<
 *             return _ser_series(obj.dt)
 * 
 */
    __pyx_t_1 = (__pyx_v_obj_dtype == ((PyTypeObject*)__pyx_v_10serializor_7typeref_PDDT));
    if (__pyx_t_1) {

      /* "serializor/binary/ser.py":1999
 *         # . <'cytimes.pddt'>
 *         if obj_dtype is typeref.PDDT:
 *             return _ser_series(obj.dt)             # <<<<<<<<<<<<<<
 * 
 *     ##### Subclass Types #####
 */
      __Pyx_XDECREF(__pyx_r);
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_obj, __pyx_n_s_dt); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1999, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = __pyx_f_10serializor_6binary_3ser__ser_series(__pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1999, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_r = ((PyObject*)__pyx_t_4);
      __pyx_t_4 = 0;
      goto __pyx_L0;

      /* "serializor/binary/ser.py":1998
 *             return _ser_datetime(obj.dt)
 *         # . <'cytimes.pddt'>
 *         if obj_dtype is typeref.PDDT:             # <<<<<<<<<<<<<<
 *             return _ser_series(obj.dt)
 * 
 */
    }

    /* "serializor/binary/ser.py":1993
 * 
 *     # Cytimes Types
 *     if typeref.CYTIMES_AVAILABLE:             # <<<<<<<<<<<<<<
 *         # . <'cytimes.pydt'>
 *         if obj_dtype is typeref.PYDT:
 */
  }

  /* "serializor/binary/ser.py":2002
 * 
 *     ##### Subclass Types #####
 *     return _ser_subclass(obj, obj_dtype)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_4 = __pyx_f_10serializor_6binary_3ser__ser_subclass(__pyx_v_obj, __pyx_v_obj_dtype); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2002, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_r = ((PyObject*)__pyx_t_4);
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "serializor/binary/ser.py":1867
 * 
 * 
 * @cython.cfunc             # <<<<<<<<<<<<<<
 * @cython.inline(True)
 * def _ser_uncommon(obj: object, obj_dtype: type) -> bytes:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("serializor.binary.ser._ser_uncommon", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "serializor/binary/ser.py":2005
 * 
 * 
 * @cython.cfunc             # <<<<<<<<<<<<<<
 * @cython.inline(True)
 * def _ser_subclass(obj: object, obj_dtype: type) -> bytes:
 */

static PyObject *__pyx_f_10serializor_6binary_3ser__ser_subclass(PyObject *__pyx_v_obj, PyTypeObject *__pyx_v_obj_dtype) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_ser_subclass", 1);

  /* "serializor/binary/ser.py":2012
 *     # Basic Types
 *     # . subclass of <'str'>
 *     if isinstance(obj, str):             # <<<<<<<<<<<<<<
 *         return _ser_str(str(obj))
 *     # . subclass of <'int'>
 */
  __pyx_t_1 = PyUnicode_Check(__pyx_v_obj); 
  if (__pyx_t_1) {

    /* "serializor/binary/ser.py":2013
 *     # . subclass of <'str'>
 *     if isinstance(obj, str):
 *         return _ser_str(str(obj))             # <<<<<<<<<<<<<<
 *     # . subclass of <'int'>
 *     if isinstance(obj, int):
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = __Pyx_PyObject_Unicode(__pyx_v_obj); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2013, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __pyx_f_10serializor_6binary_3ser__ser_str(__pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2013, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_r = ((PyObject*)__pyx_t_3);
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "serializor/binary/ser.py":2012
 *     # Basic Types
 *     # . subclass of <'str'>
 *     if isinstance(obj, str):             # <<<<<<<<<<<<<<
 *         return _ser_str(str(obj))
 *     # . subclass of <'int'>
 */
  }

  /* "serializor/binary/ser.py":2015
 *         return _ser_str(str(obj))
 *     # . subclass of <'int'>
 *     if isinstance(obj, int):             # <<<<<<<<<<<<<<
 *         return _ser_int(int(obj))
 *     # . subclass of <'float'>
 */
  __pyx_t_1 = PyInt_Check(__pyx_v_obj); 
  if (__pyx_t_1) {

    /* "serializor/binary/ser.py":2016
 *     # . subclass of <'int'>
 *     if isinstance(obj, int):
 *         return _ser_int(int(obj))             # <<<<<<<<<<<<<<
 *     # . subclass of <'float'>
 *     if isinstance(obj, float):
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_3 = __Pyx_PyNumber_Int(__pyx_v_obj); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2016, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = __pyx_f_10serializor_6binary_3ser__ser_int(__pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2016, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_r = ((PyObject*)__pyx_t_2);
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "serializor/binary/ser.py":2015
 *         return _ser_str(str(obj))
 *     # . subclass of <'int'>
 *     if isinstance(obj, int):             # <<<<<<<<<<<<<<
 *         return _ser_int(int(obj))
 *     # . subclass of <'float'>
 */
  }

  /* "serializor/binary/ser.py":2018
 *         return _ser_int(int(obj))
 *     # . subclass of <'float'>
 *     if isinstance(obj, float):             # <<<<<<<<<<<<<<
 *         return _ser_float(float(obj))
 *     # . subclass of <'bool'>
 */
  __pyx_t_1 = PyFloat_Check(__pyx_v_obj); 
  if (__pyx_t_1) {

    /* "serializor/binary/ser.py":2019
 *     # . subclass of <'float'>
 *     if isinstance(obj, float):
 *         return _ser_float(float(obj))             # <<<<<<<<<<<<<<
 *     # . subclass of <'bool'>
 *     if isinstance(obj, bool):
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = __Pyx_PyNumber_Float(__pyx_v_obj); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2019, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __pyx_f_10serializor_6binary_3ser__ser_float(__pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2019, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_r = ((PyObject*)__pyx_t_3);
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "serializor/binary/ser.py":2018
 *         return _ser_int(int(obj))
 *     # . subclass of <'float'>
 *     if isinstance(obj, float):             # <<<<<<<<<<<<<<
 *         return _ser_float(float(obj))
 *     # . subclass of <'bool'>
 */
  }

  /* "serializor/binary/ser.py":2021
 *         return _ser_float(float(obj))
 *     # . subclass of <'bool'>
 *     if isinstance(obj, bool):             # <<<<<<<<<<<<<<
 *         return _ser_bool(bool(obj))
 * 
 */
  __pyx_t_3 = ((PyObject*)&PyBool_Type);
  __Pyx_INCREF(__pyx_t_3);
  __pyx_t_1 = PyObject_IsInstance(__pyx_v_obj, __pyx_t_3); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 2021, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__pyx_t_1) {

    /* "serializor/binary/ser.py":2022
 *     # . subclass of <'bool'>
 *     if isinstance(obj, bool):
 *         return _ser_bool(bool(obj))             # <<<<<<<<<<<<<<
 * 
 *     # Date&Time Types
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_obj); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(0, 2022, __pyx_L1_error)
    __pyx_t_3 = __Pyx_PyBool_FromLong((!(!__pyx_t_1))); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2022, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = __pyx_f_10serializor_6binary_3ser__ser_bool(__pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2022, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_r = ((PyObject*)__pyx_t_2);
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "serializor/binary/ser.py":2021
 *         return _ser_float(float(obj))
 *     # . subclass of <'bool'>
 *     if isinstance(obj, bool):             # <<<<<<<<<<<<<<
 *         return _ser_bool(bool(obj))
 * 
 */
  }

  /* "serializor/binary/ser.py":2026
 *     # Date&Time Types
 *     # . subclass of <'datetime.datetime'>
 *     if isinstance(obj, datetime.datetime):             # <<<<<<<<<<<<<<
 *         return _ser_datetime(obj)
 *     # . subclass of <'datetime.date'>
 */
  __pyx_t_1 = __Pyx_TypeCheck(__pyx_v_obj, __pyx_ptype_7cpython_8datetime_datetime); 
  if (__pyx_t_1) {

    /* "serializor/binary/ser.py":2027
 *     # . subclass of <'datetime.datetime'>
 *     if isinstance(obj, datetime.datetime):
 *         return _ser_datetime(obj)             # <<<<<<<<<<<<<<
 *     # . subclass of <'datetime.date'>
 *     if isinstance(obj, datetime.date):
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = __pyx_f_10serializor_6binary_3ser__ser_datetime(__pyx_v_obj); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2027, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_r = ((PyObject*)__pyx_t_2);
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "serializor/binary/ser.py":2026
 *     # Date&Time Types
 *     # . subclass of <'datetime.datetime'>
 *     if isinstance(obj, datetime.datetime):             # <<<<<<<<<<<<<<
 *         return _ser_datetime(obj)
 *     # . subclass of <'datetime.date'>
 */
  }

  /* "serializor/binary/ser.py":2029
 *         return _ser_datetime(obj)
 *     # . subclass of <'datetime.date'>
 *     if isinstance(obj, datetime.date):             # <<<<<<<<<<<<<<
 *         return _ser_date(obj)
 *     # . subclass of <'datetime.time'>
 */
  __pyx_t_1 = __Pyx_TypeCheck(__pyx_v_obj, __pyx_ptype_7cpython_8datetime_date); 
  if (__pyx_t_1) {

    /* "serializor/binary/ser.py":2030
 *     # . subclass of <'datetime.date'>
 *     if isinstance(obj, datetime.date):
 *         return _ser_date(obj)             # <<<<<<<<<<<<<<
 *     # . subclass of <'datetime.time'>
 *     if isinstance(obj, datetime.time):
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = __pyx_f_10serializor_6binary_3ser__ser_date(__pyx_v_obj); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2030, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_r = ((PyObject*)__pyx_t_2);
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "serializor/binary/ser.py":2029
 *         return _ser_datetime(obj)
 *     # . subclass of <'datetime.date'>
 *     if isinstance(obj, datetime.date):             # <<<<<<<<<<<<<<
 *         return _ser_date(obj)
 *     # . subclass of <'datetime.time'>
 */
  }

  /* "serializor/binary/ser.py":2032
 *         return _ser_date(obj)
 *     # . subclass of <'datetime.time'>
 *     if isinstance(obj, datetime.time):             # <<<<<<<<<<<<<<
 *         return _ser_time(obj)
 *     # . subclass of <'datetime.timedelta'>
 */
  __pyx_t_1 = __Pyx_TypeCheck(__pyx_v_obj, __pyx_ptype_7cpython_8datetime_time); 
  if (__pyx_t_1) {

    /* "serializor/binary/ser.py":2033
 *     # . subclass of <'datetime.time'>
 *     if isinstance(obj, datetime.time):
 *         return _ser_time(obj)             # <<<<<<<<<<<<<<
 *     # . subclass of <'datetime.timedelta'>
 *     if isinstance(obj, datetime.timedelta):
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = __pyx_f_10serializor_6binary_3ser__ser_time(__pyx_v_obj); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2033, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_r = ((PyObject*)__pyx_t_2);
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "serializor/binary/ser.py":2032
 *         return _ser_date(obj)
 *     # . subclass of <'datetime.time'>
 *     if isinstance(obj, datetime.time):             # <<<<<<<<<<<<<<
 *         return _ser_time(obj)
 *     # . subclass of <'datetime.timedelta'>
 */
  }

  /* "serializor/binary/ser.py":2035
 *         return _ser_time(obj)
 *     # . subclass of <'datetime.timedelta'>
 *     if isinstance(obj, datetime.timedelta):             # <<<<<<<<<<<<<<
 *         return _ser_timedelta(obj)
 * 
 */
  __pyx_t_1 = __Pyx_TypeCheck(__pyx_v_obj, __pyx_ptype_7cpython_8datetime_timedelta); 
  if (__pyx_t_1) {

    /* "serializor/binary/ser.py":2036
 *     # . subclass of <'datetime.timedelta'>
 *     if isinstance(obj, datetime.timedelta):
 *         return _ser_timedelta(obj)             # <<<<<<<<<<<<<<
 * 
 *     # Sequence Types
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = __pyx_f_10serializor_6binary_3ser__ser_timedelta(__pyx_v_obj); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2036, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_r = ((PyObject*)__pyx_t_2);
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "serializor/binary/ser.py":2035
 *         return _ser_time(obj)
 *     # . subclass of <'datetime.timedelta'>
 *     if isinstance(obj, datetime.timedelta):             # <<<<<<<<<<<<<<
 *         return _ser_timedelta(obj)
 * 
 */
  }

  /* "serializor/binary/ser.py":2040
 *     # Sequence Types
 *     # . subclass of <'list'>
 *     if isinstance(obj, list):             # <<<<<<<<<<<<<<
 *         return _ser_list(list(obj))
 *     # . subclass of <'tuple'>
 */
  __pyx_t_1 = PyList_Check(__pyx_v_obj); 
  if (__pyx_t_1) {

    /* "serializor/binary/ser.py":2041
 *     # . subclass of <'list'>
 *     if isinstance(obj, list):
 *         return _ser_list(list(obj))             # <<<<<<<<<<<<<<
 *     # . subclass of <'tuple'>
 *     if isinstance(obj, tuple):
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = PySequence_List(__pyx_v_obj); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2041, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __pyx_f_10serializor_6binary_3ser__ser_list(((PyObject*)__pyx_t_2)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2041, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_r = ((PyObject*)__pyx_t_3);
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "serializor/binary/ser.py":2040
 *     # Sequence Types
 *     # . subclass of <'list'>
 *     if isinstance(obj, list):             # <<<<<<<<<<<<<<
 *         return _ser_list(list(obj))
 *     # . subclass of <'tuple'>
 */
  }

  /* "serializor/binary/ser.py":2043
 *         return _ser_list(list(obj))
 *     # . subclass of <'tuple'>
 *     if isinstance(obj, tuple):             # <<<<<<<<<<<<<<
 *         return _ser_tuple(tuple(obj))
 *     # . subclass of <'set'>
 */
  __pyx_t_1 = PyTuple_Check(__pyx_v_obj); 
  if (__pyx_t_1) {

    /* "serializor/binary/ser.py":2044
 *     # . subclass of <'tuple'>
 *     if isinstance(obj, tuple):
 *         return _ser_tuple(tuple(obj))             # <<<<<<<<<<<<<<
 *     # . subclass of <'set'>
 *     if isinstance(obj, set):
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_3 = __Pyx_PySequence_Tuple(__pyx_v_obj); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2044, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = __pyx_f_10serializor_6binary_3ser__ser_tuple(((PyObject*)__pyx_t_3)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2044, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_r = ((PyObject*)__pyx_t_2);
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "serializor/binary/ser.py":2043
 *         return _ser_list(list(obj))
 *     # . subclass of <'tuple'>
 *     if isinstance(obj, tuple):             # <<<<<<<<<<<<<<
 *         return _ser_tuple(tuple(obj))
 *     # . subclass of <'set'>
 */
  }

  /* "serializor/binary/ser.py":2046
 *         return _ser_tuple(tuple(obj))
 *     # . subclass of <'set'>
 *     if isinstance(obj, set):             # <<<<<<<<<<<<<<
 *         return _ser_set(set(obj))
 *     # . subclass of <'frozenset'>
 */
  __pyx_t_1 = PySet_Check(__pyx_v_obj); 
  if (__pyx_t_1) {

    /* "serializor/binary/ser.py":2047
 *     # . subclass of <'set'>
 *     if isinstance(obj, set):
 *         return _ser_set(set(obj))             # <<<<<<<<<<<<<<
 *     # . subclass of <'frozenset'>
 *     if isinstance(obj, frozenset):
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = PySet_New(__pyx_v_obj); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2047, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __pyx_f_10serializor_6binary_3ser__ser_set(((PyObject*)__pyx_t_2)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2047, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_r = ((PyObject*)__pyx_t_3);
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "serializor/binary/ser.py":2046
 *         return _ser_tuple(tuple(obj))
 *     # . subclass of <'set'>
 *     if isinstance(obj, set):             # <<<<<<<<<<<<<<
 *         return _ser_set(set(obj))
 *     # . subclass of <'frozenset'>
 */
  }

  /* "serializor/binary/ser.py":2049
 *         return _ser_set(set(obj))
 *     # . subclass of <'frozenset'>
 *     if isinstance(obj, frozenset):             # <<<<<<<<<<<<<<
 *         return _ser_frozenset(frozenset(obj))
 * 
 */
  __pyx_t_1 = PyFrozenSet_Check(__pyx_v_obj); 
  if (__pyx_t_1) {

    /* "serializor/binary/ser.py":2050
 *     # . subclass of <'frozenset'>
 *     if isinstance(obj, frozenset):
 *         return _ser_frozenset(frozenset(obj))             # <<<<<<<<<<<<<<
 * 
 *     # Mapping Types
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_3 = __Pyx_PyFrozenSet_New(__pyx_v_obj); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2050, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = __pyx_f_10serializor_6binary_3ser__ser_frozenset(((PyObject*)__pyx_t_3)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2050, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_r = ((PyObject*)__pyx_t_2);
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "serializor/binary/ser.py":2049
 *         return _ser_set(set(obj))
 *     # . subclass of <'frozenset'>
 *     if isinstance(obj, frozenset):             # <<<<<<<<<<<<<<
 *         return _ser_frozenset(frozenset(obj))
 * 
 */
  }

  /* "serializor/binary/ser.py":2054
 *     # Mapping Types
 *     # . subclass of <'dict'>
 *     if isinstance(obj, dict):             # <<<<<<<<<<<<<<
 *         return _ser_dict(dict(obj))
 * 
 */
  __pyx_t_1 = PyDict_Check(__pyx_v_obj); 
  if (__pyx_t_1) {

    /* "serializor/binary/ser.py":2055
 *     # . subclass of <'dict'>
 *     if isinstance(obj, dict):
 *         return _ser_dict(dict(obj))             # <<<<<<<<<<<<<<
 * 
 *     # Invalid Data Type
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyDict_Type)), __pyx_v_obj); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2055, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __pyx_f_10serializor_6binary_3ser__ser_dict(((PyObject*)__pyx_t_2)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2055, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_r = ((PyObject*)__pyx_t_3);
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "serializor/binary/ser.py":2054
 *     # Mapping Types
 *     # . subclass of <'dict'>
 *     if isinstance(obj, dict):             # <<<<<<<<<<<<<<
 *         return _ser_dict(dict(obj))
 * 
 */
  }

  /* "serializor/binary/ser.py":2058
 * 
 *     # Invalid Data Type
 *     raise TypeError("unsupported 'data' type %s." % obj_dtype)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_3 = PyUnicode_Format(__pyx_kp_u_unsupported_data_type_s, ((PyObject *)__pyx_v_obj_dtype)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2058, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_TypeError, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2058, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_Raise(__pyx_t_2, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __PYX_ERR(0, 2058, __pyx_L1_error)

  /* "serializor/binary/ser.py":2005
 * 
 * 
 * @cython.cfunc             # <<<<<<<<<<<<<<
 * @cython.inline(True)
 * def _ser_subclass(obj: object, obj_dtype: type) -> bytes:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("serializor.binary.ser._ser_subclass", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "serializor/binary/ser.py":2061
 * 
 * 
 * @cython.ccall             # <<<<<<<<<<<<<<
 * def serialize(obj: object) -> bytes:
 *     """Serialize an object to `<'bytes'>`.
 */

static PyObject *__pyx_pw_10serializor_6binary_3ser_1serialize(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_10serializor_6binary_3ser_serialize(PyObject *__pyx_v_obj, CYTHON_UNUSED int __pyx_skip_dispatch) {
  PyObject *__pyx_v_err = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  Py_ssize_t __pyx_t_11;
  Py_UCS4 __pyx_t_12;
  PyObject *__pyx_t_13 = NULL;
  unsigned int __pyx_t_14;
  int __pyx_t_15;
  char const *__pyx_t_16;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  PyObject *__pyx_t_20 = NULL;
  PyObject *__pyx_t_21 = NULL;
  PyObject *__pyx_t_22 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("serialize", 1);

  /* "serializor/binary/ser.py":2107
 *         - Ser <'pd.DataFrame'> -> Des <'pd.DataFrame'>
 *     """
 *     try:             # <<<<<<<<<<<<<<
 *         return _ser_common(obj)
 *     except Exception as err:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "serializor/binary/ser.py":2108
 *     """
 *     try:
 *         return _ser_common(obj)             # <<<<<<<<<<<<<<
 *     except Exception as err:
 *         raise errors.SerializeError(
 */
      __Pyx_XDECREF(__pyx_r);
      __pyx_t_4 = __pyx_f_10serializor_6binary_3ser__ser_common(__pyx_v_obj); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2108, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_r = ((PyObject*)__pyx_t_4);
      __pyx_t_4 = 0;
      goto __pyx_L7_try_return;

      /* "serializor/binary/ser.py":2107
 *         - Ser <'pd.DataFrame'> -> Des <'pd.DataFrame'>
 *     """
 *     try:             # <<<<<<<<<<<<<<
 *         return _ser_common(obj)
 *     except Exception as err:
 */
    }
    __pyx_L3_error:;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "serializor/binary/ser.py":2109
 *     try:
 *         return _ser_common(obj)
 *     except Exception as err:             # <<<<<<<<<<<<<<
 *         raise errors.SerializeError(
 *             "<'Serializor'>\nFailed to serialize to 'bytes': "
 */
    __pyx_t_5 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
    if (__pyx_t_5) {
      __Pyx_AddTraceback("serializor.binary.ser.serialize", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_4, &__pyx_t_6, &__pyx_t_7) < 0) __PYX_ERR(0, 2109, __pyx_L5_except_error)
      __Pyx_XGOTREF(__pyx_t_4);
      __Pyx_XGOTREF(__pyx_t_6);
      __Pyx_XGOTREF(__pyx_t_7);
      __Pyx_INCREF(__pyx_t_6);
      __pyx_v_err = __pyx_t_6;
      /*try:*/ {

        /* "serializor/binary/ser.py":2110
 *         return _ser_common(obj)
 *     except Exception as err:
 *         raise errors.SerializeError(             # <<<<<<<<<<<<<<
 *             "<'Serializor'>\nFailed to serialize to 'bytes': "
 *             "%s\n%r\nError: %s" % (type(obj), obj, err)
 */
        __Pyx_GetModuleGlobalName(__pyx_t_9, __pyx_n_s_errors); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 2110, __pyx_L14_error)
        __Pyx_GOTREF(__pyx_t_9);
        __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_SerializeError); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 2110, __pyx_L14_error)
        __Pyx_GOTREF(__pyx_t_10);
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

        /* "serializor/binary/ser.py":2111
 *     except Exception as err:
 *         raise errors.SerializeError(
 *             "<'Serializor'>\nFailed to serialize to 'bytes': "             # <<<<<<<<<<<<<<
 *             "%s\n%r\nError: %s" % (type(obj), obj, err)
 *         ) from err
 */
        __pyx_t_9 = PyTuple_New(6); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 2111, __pyx_L14_error)
        __Pyx_GOTREF(__pyx_t_9);
        __pyx_t_11 = 0;
        __pyx_t_12 = 127;
        __Pyx_INCREF(__pyx_kp_u_Serializor_Failed_to_serialize);
        __pyx_t_11 += 47;
        __Pyx_GIVEREF(__pyx_kp_u_Serializor_Failed_to_serialize);
        PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_kp_u_Serializor_Failed_to_serialize);

        /* "serializor/binary/ser.py":2112
 *         raise errors.SerializeError(
 *             "<'Serializor'>\nFailed to serialize to 'bytes': "
 *             "%s\n%r\nError: %s" % (type(obj), obj, err)             # <<<<<<<<<<<<<<
 *         ) from err
 * 
 */
        __pyx_t_13 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Unicode(((PyObject *)Py_TYPE(__pyx_v_obj))), __pyx_empty_unicode); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 2112, __pyx_L14_error)
        __Pyx_GOTREF(__pyx_t_13);
        __pyx_t_12 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_13) > __pyx_t_12) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_13) : __pyx_t_12;
        __pyx_t_11 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_13);
        __Pyx_GIVEREF(__pyx_t_13);
        PyTuple_SET_ITEM(__pyx_t_9, 1, __pyx_t_13);
        __pyx_t_13 = 0;
        __Pyx_INCREF(__pyx_kp_u__13);
        __pyx_t_11 += 1;
        __Pyx_GIVEREF(__pyx_kp_u__13);
        PyTuple_SET_ITEM(__pyx_t_9, 2, __pyx_kp_u__13);
        __pyx_t_13 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Repr(__pyx_v_obj), __pyx_empty_unicode); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 2112, __pyx_L14_error)
        __Pyx_GOTREF(__pyx_t_13);
        __pyx_t_12 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_13) > __pyx_t_12) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_13) : __pyx_t_12;
        __pyx_t_11 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_13);
        __Pyx_GIVEREF(__pyx_t_13);
        PyTuple_SET_ITEM(__pyx_t_9, 3, __pyx_t_13);
        __pyx_t_13 = 0;
        __Pyx_INCREF(__pyx_kp_u_Error);
        __pyx_t_11 += 8;
        __Pyx_GIVEREF(__pyx_kp_u_Error);
        PyTuple_SET_ITEM(__pyx_t_9, 4, __pyx_kp_u_Error);
        __pyx_t_13 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Unicode(__pyx_v_err), __pyx_empty_unicode); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 2112, __pyx_L14_error)
        __Pyx_GOTREF(__pyx_t_13);
        __pyx_t_12 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_13) > __pyx_t_12) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_13) : __pyx_t_12;
        __pyx_t_11 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_13);
        __Pyx_GIVEREF(__pyx_t_13);
        PyTuple_SET_ITEM(__pyx_t_9, 5, __pyx_t_13);
        __pyx_t_13 = 0;

        /* "serializor/binary/ser.py":2111
 *     except Exception as err:
 *         raise errors.SerializeError(
 *             "<'Serializor'>\nFailed to serialize to 'bytes': "             # <<<<<<<<<<<<<<
 *             "%s\n%r\nError: %s" % (type(obj), obj, err)
 *         ) from err
 */
        __pyx_t_13 = __Pyx_PyUnicode_Join(__pyx_t_9, 6, __pyx_t_11, __pyx_t_12); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 2111, __pyx_L14_error)
        __Pyx_GOTREF(__pyx_t_13);
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        __pyx_t_9 = NULL;
        __pyx_t_14 = 0;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_10))) {
          __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_10);
          if (likely(__pyx_t_9)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_10);
            __Pyx_INCREF(__pyx_t_9);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_10, function);
            __pyx_t_14 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_9, __pyx_t_13};
          __pyx_t_8 = __Pyx_PyObject_FastCall(__pyx_t_10, __pyx_callargs+1-__pyx_t_14, 1+__pyx_t_14);
          __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
          __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
          if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 2110, __pyx_L14_error)
          __Pyx_GOTREF(__pyx_t_8);
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        }

        /* "serializor/binary/ser.py":2113
 *             "<'Serializor'>\nFailed to serialize to 'bytes': "
 *             "%s\n%r\nError: %s" % (type(obj), obj, err)
 *         ) from err             # <<<<<<<<<<<<<<
 * 
 * 
 */
        __Pyx_Raise(__pyx_t_8, 0, 0, __pyx_v_err);
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        __PYX_ERR(0, 2110, __pyx_L14_error)
      }

      /* "serializor/binary/ser.py":2109
 *     try:
 *         return _ser_common(obj)
 *     except Exception as err:             # <<<<<<<<<<<<<<
 *         raise errors.SerializeError(
 *             "<'Serializor'>\nFailed to serialize to 'bytes': "
 */
      /*finally:*/ {
        __pyx_L14_error:;
        /*exception exit:*/{
          __Pyx_PyThreadState_declare
          __Pyx_PyThreadState_assign
          __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0; __pyx_t_21 = 0; __pyx_t_22 = 0;
          __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
          __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
          __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
          __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
          if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_20, &__pyx_t_21, &__pyx_t_22);
          if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_17, &__pyx_t_18, &__pyx_t_19) < 0)) __Pyx_ErrFetch(&__pyx_t_17, &__pyx_t_18, &__pyx_t_19);
          __Pyx_XGOTREF(__pyx_t_17);
          __Pyx_XGOTREF(__pyx_t_18);
          __Pyx_XGOTREF(__pyx_t_19);
          __Pyx_XGOTREF(__pyx_t_20);
          __Pyx_XGOTREF(__pyx_t_21);
          __Pyx_XGOTREF(__pyx_t_22);
          __pyx_t_5 = __pyx_lineno; __pyx_t_15 = __pyx_clineno; __pyx_t_16 = __pyx_filename;
          {
            __Pyx_DECREF(__pyx_v_err); __pyx_v_err = 0;
          }
          if (PY_MAJOR_VERSION >= 3) {
            __Pyx_XGIVEREF(__pyx_t_20);
            __Pyx_XGIVEREF(__pyx_t_21);
            __Pyx_XGIVEREF(__pyx_t_22);
            __Pyx_ExceptionReset(__pyx_t_20, __pyx_t_21, __pyx_t_22);
          }
          __Pyx_XGIVEREF(__pyx_t_17);
          __Pyx_XGIVEREF(__pyx_t_18);
          __Pyx_XGIVEREF(__pyx_t_19);
          __Pyx_ErrRestore(__pyx_t_17, __pyx_t_18, __pyx_t_19);
          __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0; __pyx_t_21 = 0; __pyx_t_22 = 0;
          __pyx_lineno = __pyx_t_5; __pyx_clineno = __pyx_t_15; __pyx_filename = __pyx_t_16;
          goto __pyx_L5_except_error;
        }
      }
    }
    goto __pyx_L5_except_error;

    /* "serializor/binary/ser.py":2107
 *         - Ser <'pd.DataFrame'> -> Des <'pd.DataFrame'>
 *     """
 *     try:             # <<<<<<<<<<<<<<
 *         return _ser_common(obj)
 *     except Exception as err:
 */
    __pyx_L5_except_error:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L7_try_return:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L0;
  }

  /* "serializor/binary/ser.py":2061
 * 
 * 
 * @cython.ccall             # <<<<<<<<<<<<<<
 * def serialize(obj: object) -> bytes:
 *     """Serialize an object to `<'bytes'>`.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_AddTraceback("serializor.binary.ser.serialize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_err);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_10serializor_6binary_3ser_1serialize(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_10serializor_6binary_3ser_serialize, "Serialize an object to `<'bytes'>`.\n\n    Supports:\n    - Python natives:\n        - Ser <'str'> -> Des <'str'>\n        - Ser <'int'> -> Des <'int'>\n        - Ser <'float'> -> Des <'float'>\n        - Ser <'bool'> -> Des <'bool'>\n        - Ser <'datetime.datetime'> -> Des <'datetime.datetime'> `[Supports Timezone]`\n        - Ser <'datetime.date'> -> Des <'datetime.date'>\n        - Ser <'datetime.time'> -> Des <'datetime.time'> `[Supports Timezone]`\n        - Ser <'datetime.timedelta'> -> Des <'datetime.timedelta'>\n        - Ser <'time.struct_time'> -> Des <'time.struct_time'>\n        - Ser <'decimal.Decimal'> -> Des <'decimal.Decimal'>\n        - Ser <'complex'> -> Des <'comples'>\n        - Ser <'bytes'> -> Des <'bytes'>\n        - Ser <'bytearray'> -> Des <'bytearray'>\n        - Ser <'memoryview'> -> Des <'bytes'>\n        - Ser <'list'> -> Des <'list'>\n        - Ser <'tuple'> -> Des <'tuple'>\n        - Ser <'set'> -> Des <'set'>\n        - Ser <'frozenset'> -> Des <'frozenset'>\n        - Ser <'range'> -> Des <'range'>\n        - Ser <'deque'> -> Des <'list'>\n        - Ser <'dict'> -> Des <'dict'>\n    - NumPy objects:\n        - Ser <'np.str\\_'> -> Des <'str'>\n        - Ser <'np.int\\*'> -> Des <'int'>\n        - Ser <'np.uint\\*'> -> Des <'int'>\n        - Ser <'np.float\\*'> -> Des <'float'>\n        - Ser <'np.bool\\_'> -> Des <'bool'>\n        - Ser <'np.datetime64'> -> Des <'np.datetime64'>\n        - Ser <'np.timedelta64'> -> Des <'np.timedelta64'>\n        - Ser <'np.complex\\*'> -> Des <'complex'>\n        - Ser <'np.bytes\\_'> -> Des <'bytes'>\n        - Ser <'np.ndarray'> -> Des <'np.ndarray'> `[1-4 dimemsional]`\n    - Pandas objects:\n        - Ser <'pd.Timestamp'> -> Des <'pd.Timestamp'> `[Supports Timezone]`\n        - Ser <'pd.Timedelta'> -> Des <'pd.Timedelta'>\n        - Ser <'pd.Series'> -> Des <'pd.Series'>\n        - Ser <'pd.DatetimeIndex'> -> Des <'pd.DatetimeIndex'> `[Supports Timezone]`\n        - Ser <'pd.TimedeltaIndex'>"" -> Des <'pd.TimedeltaIndex'>\n        - Ser <'pd.DataFrame'> -> Des <'pd.DataFrame'>\n    ");
static PyMethodDef __pyx_mdef_10serializor_6binary_3ser_1serialize = {"serialize", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10serializor_6binary_3ser_1serialize, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_10serializor_6binary_3ser_serialize};
static PyObject *__pyx_pw_10serializor_6binary_3ser_1serialize(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_obj = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("serialize (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_obj,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_obj)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 2061, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "serialize") < 0)) __PYX_ERR(0, 2061, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_obj = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("serialize", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 2061, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("serializor.binary.ser.serialize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10serializor_6binary_3ser_serialize(__pyx_self, __pyx_v_obj);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10serializor_6binary_3ser_serialize(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_obj) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("serialize", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_10serializor_6binary_3ser_serialize(__pyx_v_obj, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2061, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("serializor.binary.ser.serialize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "serializor/binary/ser.py":2117
 * 
 * ########## The following functions are for testing purpose only ##########
 * def _test_utils() -> None:             # <<<<<<<<<<<<<<
 *     _test_pack_int8()
 *     _test_pack_uint8()
 */

/* Python wrapper */
static PyObject *__pyx_pw_10serializor_6binary_3ser_3_test_utils(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyMethodDef __pyx_mdef_10serializor_6binary_3ser_3_test_utils = {"_test_utils", (PyCFunction)__pyx_pw_10serializor_6binary_3ser_3_test_utils, METH_NOARGS, 0};
static PyObject *__pyx_pw_10serializor_6binary_3ser_3_test_utils(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_test_utils (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10serializor_6binary_3ser_2_test_utils(__pyx_self);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10serializor_6binary_3ser_2_test_utils(CYTHON_UNUSED PyObject *__pyx_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  unsigned int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_test_utils", 1);

  /* "serializor/binary/ser.py":2118
 * ########## The following functions are for testing purpose only ##########
 * def _test_utils() -> None:
 *     _test_pack_int8()             # <<<<<<<<<<<<<<
 *     _test_pack_uint8()
 *     _test_pack_int16()
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_test_pack_int8); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2118, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 0+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2118, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "serializor/binary/ser.py":2119
 * def _test_utils() -> None:
 *     _test_pack_int8()
 *     _test_pack_uint8()             # <<<<<<<<<<<<<<
 *     _test_pack_int16()
 *     _test_pack_uint16()
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_test_pack_uint8); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2119, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 0+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2119, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "serializor/binary/ser.py":2120
 *     _test_pack_int8()
 *     _test_pack_uint8()
 *     _test_pack_int16()             # <<<<<<<<<<<<<<
 *     _test_pack_uint16()
 *     _test_pack_int24()
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_test_pack_int16); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2120, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 0+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2120, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "serializor/binary/ser.py":2121
 *     _test_pack_uint8()
 *     _test_pack_int16()
 *     _test_pack_uint16()             # <<<<<<<<<<<<<<
 *     _test_pack_int24()
 *     _test_pack_uint24()
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_test_pack_uint16); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2121, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 0+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2121, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "serializor/binary/ser.py":2122
 *     _test_pack_int16()
 *     _test_pack_uint16()
 *     _test_pack_int24()             # <<<<<<<<<<<<<<
 *     _test_pack_uint24()
 *     _test_pack_int32()
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_test_pack_int24); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2122, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 0+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2122, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "serializor/binary/ser.py":2123
 *     _test_pack_uint16()
 *     _test_pack_int24()
 *     _test_pack_uint24()             # <<<<<<<<<<<<<<
 *     _test_pack_int32()
 *     _test_pack_uint32()
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_test_pack_uint24); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2123, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 0+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2123, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "serializor/binary/ser.py":2124
 *     _test_pack_int24()
 *     _test_pack_uint24()
 *     _test_pack_int32()             # <<<<<<<<<<<<<<
 *     _test_pack_uint32()
 *     _test_pack_int64()
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_test_pack_int32); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2124, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 0+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2124, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "serializor/binary/ser.py":2125
 *     _test_pack_uint24()
 *     _test_pack_int32()
 *     _test_pack_uint32()             # <<<<<<<<<<<<<<
 *     _test_pack_int64()
 *     _test_pack_uint64()
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_test_pack_uint32); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2125, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 0+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2125, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "serializor/binary/ser.py":2126
 *     _test_pack_int32()
 *     _test_pack_uint32()
 *     _test_pack_int64()             # <<<<<<<<<<<<<<
 *     _test_pack_uint64()
 *     _test_gen_encoded_int()
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_test_pack_int64); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2126, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 0+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2126, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "serializor/binary/ser.py":2127
 *     _test_pack_uint32()
 *     _test_pack_int64()
 *     _test_pack_uint64()             # <<<<<<<<<<<<<<
 *     _test_gen_encoded_int()
 *     _test_gen_header()
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_test_pack_uint64); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2127, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 0+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2127, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "serializor/binary/ser.py":2128
 *     _test_pack_int64()
 *     _test_pack_uint64()
 *     _test_gen_encoded_int()             # <<<<<<<<<<<<<<
 *     _test_gen_header()
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_test_gen_encoded_int); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2128, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 0+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2128, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "serializor/binary/ser.py":2129
 *     _test_pack_uint64()
 *     _test_gen_encoded_int()
 *     _test_gen_header()             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_test_gen_header); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2129, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 0+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2129, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "serializor/binary/ser.py":2117
 * 
 * ########## The following functions are for testing purpose only ##########
 * def _test_utils() -> None:             # <<<<<<<<<<<<<<
 *     _test_pack_int8()
 *     _test_pack_uint8()
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("serializor.binary.ser._test_utils", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "serializor/binary/ser.py":2132
 * 
 * 
 * def _test_pack_int8() -> None:             # <<<<<<<<<<<<<<
 *     import struct
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_10serializor_6binary_3ser_5_test_pack_int8(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyMethodDef __pyx_mdef_10serializor_6binary_3ser_5_test_pack_int8 = {"_test_pack_int8", (PyCFunction)__pyx_pw_10serializor_6binary_3ser_5_test_pack_int8, METH_NOARGS, 0};
static PyObject *__pyx_pw_10serializor_6binary_3ser_5_test_pack_int8(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_test_pack_int8 (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10serializor_6binary_3ser_4_test_pack_int8(__pyx_self);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10serializor_6binary_3ser_4_test_pack_int8(CYTHON_UNUSED PyObject *__pyx_self) {
  PyObject *__pyx_v_struct = NULL;
  PyObject *__pyx_v_val = NULL;
  PyObject *__pyx_v_s = NULL;
  PyObject *__pyx_v_b = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  long __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_t_7;
  Py_ssize_t __pyx_t_8;
  Py_UCS4 __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_test_pack_int8", 1);

  /* "serializor/binary/ser.py":2133
 * 
 * def _test_pack_int8() -> None:
 *     import struct             # <<<<<<<<<<<<<<
 * 
 *     for val in range(-128, 128):
 */
  __pyx_t_1 = __Pyx_ImportDottedModule(__pyx_n_s_struct, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2133, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_struct = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "serializor/binary/ser.py":2135
 *     import struct
 * 
 *     for val in range(-128, 128):             # <<<<<<<<<<<<<<
 *         s = struct.pack("<b", val)
 *         b = pack_int8(val)  # type: ignore
 */
  for (__pyx_t_2 = -128L; __pyx_t_2 < 0x80; __pyx_t_2+=1) {
    __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2135, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_XDECREF_SET(__pyx_v_val, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "serializor/binary/ser.py":2136
 * 
 *     for val in range(-128, 128):
 *         s = struct.pack("<b", val)             # <<<<<<<<<<<<<<
 *         b = pack_int8(val)  # type: ignore
 *         assert s == b, f"{s} | {b}"
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_struct, __pyx_n_s_pack); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2136, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = NULL;
    __pyx_t_5 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_5 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[3] = {__pyx_t_4, __pyx_kp_u_b, __pyx_v_val};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 2+__pyx_t_5);
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2136, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
    __Pyx_XDECREF_SET(__pyx_v_s, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "serializor/binary/ser.py":2137
 *     for val in range(-128, 128):
 *         s = struct.pack("<b", val)
 *         b = pack_int8(val)  # type: ignore             # <<<<<<<<<<<<<<
 *         assert s == b, f"{s} | {b}"
 *         print("- pass: %s" % val, end="\r")
 */
    __pyx_t_6 = __Pyx_PyInt_As_int(__pyx_v_val); if (unlikely((__pyx_t_6 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 2137, __pyx_L1_error)
    __pyx_t_1 = __pyx_f_10serializor_6binary_3ser_pack_int8(__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2137, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_XDECREF_SET(__pyx_v_b, ((PyObject*)__pyx_t_1));
    __pyx_t_1 = 0;

    /* "serializor/binary/ser.py":2138
 *         s = struct.pack("<b", val)
 *         b = pack_int8(val)  # type: ignore
 *         assert s == b, f"{s} | {b}"             # <<<<<<<<<<<<<<
 *         print("- pass: %s" % val, end="\r")
 *     print("Pass Pack int8".ljust(80))
 */
    #ifndef CYTHON_WITHOUT_ASSERTIONS
    if (unlikely(__pyx_assertions_enabled())) {
      __pyx_t_7 = (__Pyx_PyBytes_Equals(__pyx_v_s, __pyx_v_b, Py_EQ)); if (unlikely((__pyx_t_7 < 0))) __PYX_ERR(0, 2138, __pyx_L1_error)
      if (unlikely(!__pyx_t_7)) {
        __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2138, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_8 = 0;
        __pyx_t_9 = 127;
        __pyx_t_3 = __Pyx_PyObject_FormatSimple(__pyx_v_s, __pyx_empty_unicode); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2138, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_9 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_3) > __pyx_t_9) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_3) : __pyx_t_9;
        __pyx_t_8 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_3);
        __Pyx_GIVEREF(__pyx_t_3);
        PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_3);
        __pyx_t_3 = 0;
        __Pyx_INCREF(__pyx_kp_u__15);
        __pyx_t_8 += 3;
        __Pyx_GIVEREF(__pyx_kp_u__15);
        PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_kp_u__15);
        __pyx_t_3 = __Pyx_PyObject_FormatSimple(__pyx_v_b, __pyx_empty_unicode); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2138, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_9 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_3) > __pyx_t_9) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_3) : __pyx_t_9;
        __pyx_t_8 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_3);
        __Pyx_GIVEREF(__pyx_t_3);
        PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_t_3);
        __pyx_t_3 = 0;
        __pyx_t_3 = __Pyx_PyUnicode_Join(__pyx_t_1, 3, __pyx_t_8, __pyx_t_9); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2138, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_Raise(__pyx_builtin_AssertionError, __pyx_t_3, 0, 0);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __PYX_ERR(0, 2138, __pyx_L1_error)
      }
    }
    #else
    if ((1)); else __PYX_ERR(0, 2138, __pyx_L1_error)
    #endif

    /* "serializor/binary/ser.py":2139
 *         b = pack_int8(val)  # type: ignore
 *         assert s == b, f"{s} | {b}"
 *         print("- pass: %s" % val, end="\r")             # <<<<<<<<<<<<<<
 *     print("Pass Pack int8".ljust(80))
 * 
 */
    __pyx_t_3 = __Pyx_PyUnicode_FormatSafe(__pyx_kp_u_pass_s, __pyx_v_val); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2139, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2139, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_3);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_3)) __PYX_ERR(0, 2139, __pyx_L1_error);
    __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2139, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_end, __pyx_kp_u__16) < 0) __PYX_ERR(0, 2139, __pyx_L1_error)
    __pyx_t_4 = __Pyx_PyObject_Call(__pyx_builtin_print, __pyx_t_1, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2139, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  }

  /* "serializor/binary/ser.py":2140
 *         assert s == b, f"{s} | {b}"
 *         print("- pass: %s" % val, end="\r")
 *     print("Pass Pack int8".ljust(80))             # <<<<<<<<<<<<<<
 * 
 *     del struct
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_Pass_Pack_int8, __pyx_n_s_ljust); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2140, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_1 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_1)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_1, __pyx_int_80};
    __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2140, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_builtin_print, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2140, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "serializor/binary/ser.py":2142
 *     print("Pass Pack int8".ljust(80))
 * 
 *     del struct             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_DECREF(__pyx_v_struct); __pyx_v_struct = 0;

  /* "serializor/binary/ser.py":2132
 * 
 * 
 * def _test_pack_int8() -> None:             # <<<<<<<<<<<<<<
 *     import struct
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("serializor.binary.ser._test_pack_int8", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_struct);
  __Pyx_XDECREF(__pyx_v_val);
  __Pyx_XDECREF(__pyx_v_s);
  __Pyx_XDECREF(__pyx_v_b);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "serializor/binary/ser.py":2145
 * 
 * 
 * def _test_pack_uint8() -> None:             # <<<<<<<<<<<<<<
 *     import struct
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_10serializor_6binary_3ser_7_test_pack_uint8(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyMethodDef __pyx_mdef_10serializor_6binary_3ser_7_test_pack_uint8 = {"_test_pack_uint8", (PyCFunction)__pyx_pw_10serializor_6binary_3ser_7_test_pack_uint8, METH_NOARGS, 0};
static PyObject *__pyx_pw_10serializor_6binary_3ser_7_test_pack_uint8(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_test_pack_uint8 (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10serializor_6binary_3ser_6_test_pack_uint8(__pyx_self);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10serializor_6binary_3ser_6_test_pack_uint8(CYTHON_UNUSED PyObject *__pyx_self) {
  PyObject *__pyx_v_struct = NULL;
  PyObject *__pyx_v_val = NULL;
  PyObject *__pyx_v_s = NULL;
  PyObject *__pyx_v_b = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  long __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_t_6;
  Py_ssize_t __pyx_t_7;
  Py_UCS4 __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_test_pack_uint8", 1);

  /* "serializor/binary/ser.py":2146
 * 
 * def _test_pack_uint8() -> None:
 *     import struct             # <<<<<<<<<<<<<<
 * 
 *     for val in range(0, 256):
 */
  __pyx_t_1 = __Pyx_ImportDottedModule(__pyx_n_s_struct, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2146, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_struct = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "serializor/binary/ser.py":2148
 *     import struct
 * 
 *     for val in range(0, 256):             # <<<<<<<<<<<<<<
 *         s = struct.pack("<B", val)
 *         b = pack_uint8(val)  # type: ignore
 */
  for (__pyx_t_2 = 0; __pyx_t_2 < 0x100; __pyx_t_2+=1) {
    __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2148, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_XDECREF_SET(__pyx_v_val, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "serializor/binary/ser.py":2149
 * 
 *     for val in range(0, 256):
 *         s = struct.pack("<B", val)             # <<<<<<<<<<<<<<
 *         b = pack_uint8(val)  # type: ignore
 *         assert s == b, f"{s} | {b}"
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_struct, __pyx_n_s_pack); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2149, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = NULL;
    __pyx_t_5 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_5 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[3] = {__pyx_t_4, __pyx_kp_u_B, __pyx_v_val};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 2+__pyx_t_5);
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2149, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
    __Pyx_XDECREF_SET(__pyx_v_s, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "serializor/binary/ser.py":2150
 *     for val in range(0, 256):
 *         s = struct.pack("<B", val)
 *         b = pack_uint8(val)  # type: ignore             # <<<<<<<<<<<<<<
 *         assert s == b, f"{s} | {b}"
 *         print("- pass: %s" % val, end="\r")
 */
    __pyx_t_5 = __Pyx_PyInt_As_unsigned_int(__pyx_v_val); if (unlikely((__pyx_t_5 == (unsigned int)-1) && PyErr_Occurred())) __PYX_ERR(0, 2150, __pyx_L1_error)
    __pyx_t_1 = __pyx_f_10serializor_6binary_3ser_pack_uint8(__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2150, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_XDECREF_SET(__pyx_v_b, ((PyObject*)__pyx_t_1));
    __pyx_t_1 = 0;

    /* "serializor/binary/ser.py":2151
 *         s = struct.pack("<B", val)
 *         b = pack_uint8(val)  # type: ignore
 *         assert s == b, f"{s} | {b}"             # <<<<<<<<<<<<<<
 *         print("- pass: %s" % val, end="\r")
 *     print("Pass Pack uint8".ljust(80))
 */
    #ifndef CYTHON_WITHOUT_ASSERTIONS
    if (unlikely(__pyx_assertions_enabled())) {
      __pyx_t_6 = (__Pyx_PyBytes_Equals(__pyx_v_s, __pyx_v_b, Py_EQ)); if (unlikely((__pyx_t_6 < 0))) __PYX_ERR(0, 2151, __pyx_L1_error)
      if (unlikely(!__pyx_t_6)) {
        __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2151, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_7 = 0;
        __pyx_t_8 = 127;
        __pyx_t_3 = __Pyx_PyObject_FormatSimple(__pyx_v_s, __pyx_empty_unicode); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2151, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_8 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_3) > __pyx_t_8) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_3) : __pyx_t_8;
        __pyx_t_7 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_3);
        __Pyx_GIVEREF(__pyx_t_3);
        PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_3);
        __pyx_t_3 = 0;
        __Pyx_INCREF(__pyx_kp_u__15);
        __pyx_t_7 += 3;
        __Pyx_GIVEREF(__pyx_kp_u__15);
        PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_kp_u__15);
        __pyx_t_3 = __Pyx_PyObject_FormatSimple(__pyx_v_b, __pyx_empty_unicode); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2151, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_8 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_3) > __pyx_t_8) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_3) : __pyx_t_8;
        __pyx_t_7 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_3);
        __Pyx_GIVEREF(__pyx_t_3);
        PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_t_3);
        __pyx_t_3 = 0;
        __pyx_t_3 = __Pyx_PyUnicode_Join(__pyx_t_1, 3, __pyx_t_7, __pyx_t_8); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2151, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_Raise(__pyx_builtin_AssertionError, __pyx_t_3, 0, 0);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __PYX_ERR(0, 2151, __pyx_L1_error)
      }
    }
    #else
    if ((1)); else __PYX_ERR(0, 2151, __pyx_L1_error)
    #endif

    /* "serializor/binary/ser.py":2152
 *         b = pack_uint8(val)  # type: ignore
 *         assert s == b, f"{s} | {b}"
 *         print("- pass: %s" % val, end="\r")             # <<<<<<<<<<<<<<
 *     print("Pass Pack uint8".ljust(80))
 * 
 */
    __pyx_t_3 = __Pyx_PyUnicode_FormatSafe(__pyx_kp_u_pass_s, __pyx_v_val); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2152, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2152, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_3);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_3)) __PYX_ERR(0, 2152, __pyx_L1_error);
    __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2152, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_end, __pyx_kp_u__16) < 0) __PYX_ERR(0, 2152, __pyx_L1_error)
    __pyx_t_4 = __Pyx_PyObject_Call(__pyx_builtin_print, __pyx_t_1, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2152, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  }

  /* "serializor/binary/ser.py":2153
 *         assert s == b, f"{s} | {b}"
 *         print("- pass: %s" % val, end="\r")
 *     print("Pass Pack uint8".ljust(80))             # <<<<<<<<<<<<<<
 * 
 *     del struct
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_Pass_Pack_uint8, __pyx_n_s_ljust); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2153, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_1 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_1)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_1, __pyx_int_80};
    __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2153, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_builtin_print, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2153, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "serializor/binary/ser.py":2155
 *     print("Pass Pack uint8".ljust(80))
 * 
 *     del struct             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_DECREF(__pyx_v_struct); __pyx_v_struct = 0;

  /* "serializor/binary/ser.py":2145
 * 
 * 
 * def _test_pack_uint8() -> None:             # <<<<<<<<<<<<<<
 *     import struct
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("serializor.binary.ser._test_pack_uint8", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_struct);
  __Pyx_XDECREF(__pyx_v_val);
  __Pyx_XDECREF(__pyx_v_s);
  __Pyx_XDECREF(__pyx_v_b);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "serializor/binary/ser.py":2158
 * 
 * 
 * def _test_pack_int16() -> None:             # <<<<<<<<<<<<<<
 *     import struct
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_10serializor_6binary_3ser_9_test_pack_int16(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyMethodDef __pyx_mdef_10serializor_6binary_3ser_9_test_pack_int16 = {"_test_pack_int16", (PyCFunction)__pyx_pw_10serializor_6binary_3ser_9_test_pack_int16, METH_NOARGS, 0};
static PyObject *__pyx_pw_10serializor_6binary_3ser_9_test_pack_int16(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_test_pack_int16 (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10serializor_6binary_3ser_8_test_pack_int16(__pyx_self);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10serializor_6binary_3ser_8_test_pack_int16(CYTHON_UNUSED PyObject *__pyx_self) {
  PyObject *__pyx_v_struct = NULL;
  PyObject *__pyx_v_val = NULL;
  PyObject *__pyx_v_s = NULL;
  PyObject *__pyx_v_b = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  long __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_t_7;
  Py_ssize_t __pyx_t_8;
  Py_UCS4 __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_test_pack_int16", 1);

  /* "serializor/binary/ser.py":2159
 * 
 * def _test_pack_int16() -> None:
 *     import struct             # <<<<<<<<<<<<<<
 * 
 *     for val in range(-32768, 32768):
 */
  __pyx_t_1 = __Pyx_ImportDottedModule(__pyx_n_s_struct, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2159, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_struct = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "serializor/binary/ser.py":2161
 *     import struct
 * 
 *     for val in range(-32768, 32768):             # <<<<<<<<<<<<<<
 *         s = struct.pack("<h", val)
 *         b = pack_int16(val)  # type: ignore
 */
  for (__pyx_t_2 = -32768L; __pyx_t_2 < 0x8000; __pyx_t_2+=1) {
    __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2161, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_XDECREF_SET(__pyx_v_val, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "serializor/binary/ser.py":2162
 * 
 *     for val in range(-32768, 32768):
 *         s = struct.pack("<h", val)             # <<<<<<<<<<<<<<
 *         b = pack_int16(val)  # type: ignore
 *         assert s == b, f"{s} | {b}"
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_struct, __pyx_n_s_pack); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2162, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = NULL;
    __pyx_t_5 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_5 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[3] = {__pyx_t_4, __pyx_kp_u_h_2, __pyx_v_val};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 2+__pyx_t_5);
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2162, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
    __Pyx_XDECREF_SET(__pyx_v_s, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "serializor/binary/ser.py":2163
 *     for val in range(-32768, 32768):
 *         s = struct.pack("<h", val)
 *         b = pack_int16(val)  # type: ignore             # <<<<<<<<<<<<<<
 *         assert s == b, f"{s} | {b}"
 *         print("- pass: %s" % val, end="\r")
 */
    __pyx_t_6 = __Pyx_PyInt_As_int(__pyx_v_val); if (unlikely((__pyx_t_6 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 2163, __pyx_L1_error)
    __pyx_t_1 = __pyx_f_10serializor_6binary_3ser_pack_int16(__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2163, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_XDECREF_SET(__pyx_v_b, ((PyObject*)__pyx_t_1));
    __pyx_t_1 = 0;

    /* "serializor/binary/ser.py":2164
 *         s = struct.pack("<h", val)
 *         b = pack_int16(val)  # type: ignore
 *         assert s == b, f"{s} | {b}"             # <<<<<<<<<<<<<<
 *         print("- pass: %s" % val, end="\r")
 *     print("Pass Pack int16".ljust(80))
 */
    #ifndef CYTHON_WITHOUT_ASSERTIONS
    if (unlikely(__pyx_assertions_enabled())) {
      __pyx_t_7 = (__Pyx_PyBytes_Equals(__pyx_v_s, __pyx_v_b, Py_EQ)); if (unlikely((__pyx_t_7 < 0))) __PYX_ERR(0, 2164, __pyx_L1_error)
      if (unlikely(!__pyx_t_7)) {
        __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2164, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_8 = 0;
        __pyx_t_9 = 127;
        __pyx_t_3 = __Pyx_PyObject_FormatSimple(__pyx_v_s, __pyx_empty_unicode); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2164, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_9 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_3) > __pyx_t_9) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_3) : __pyx_t_9;
        __pyx_t_8 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_3);
        __Pyx_GIVEREF(__pyx_t_3);
        PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_3);
        __pyx_t_3 = 0;
        __Pyx_INCREF(__pyx_kp_u__15);
        __pyx_t_8 += 3;
        __Pyx_GIVEREF(__pyx_kp_u__15);
        PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_kp_u__15);
        __pyx_t_3 = __Pyx_PyObject_FormatSimple(__pyx_v_b, __pyx_empty_unicode); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2164, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_9 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_3) > __pyx_t_9) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_3) : __pyx_t_9;
        __pyx_t_8 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_3);
        __Pyx_GIVEREF(__pyx_t_3);
        PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_t_3);
        __pyx_t_3 = 0;
        __pyx_t_3 = __Pyx_PyUnicode_Join(__pyx_t_1, 3, __pyx_t_8, __pyx_t_9); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2164, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_Raise(__pyx_builtin_AssertionError, __pyx_t_3, 0, 0);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __PYX_ERR(0, 2164, __pyx_L1_error)
      }
    }
    #else
    if ((1)); else __PYX_ERR(0, 2164, __pyx_L1_error)
    #endif

    /* "serializor/binary/ser.py":2165
 *         b = pack_int16(val)  # type: ignore
 *         assert s == b, f"{s} | {b}"
 *         print("- pass: %s" % val, end="\r")             # <<<<<<<<<<<<<<
 *     print("Pass Pack int16".ljust(80))
 * 
 */
    __pyx_t_3 = __Pyx_PyUnicode_FormatSafe(__pyx_kp_u_pass_s, __pyx_v_val); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2165, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2165, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_3);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_3)) __PYX_ERR(0, 2165, __pyx_L1_error);
    __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2165, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_end, __pyx_kp_u__16) < 0) __PYX_ERR(0, 2165, __pyx_L1_error)
    __pyx_t_4 = __Pyx_PyObject_Call(__pyx_builtin_print, __pyx_t_1, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2165, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  }

  /* "serializor/binary/ser.py":2166
 *         assert s == b, f"{s} | {b}"
 *         print("- pass: %s" % val, end="\r")
 *     print("Pass Pack int16".ljust(80))             # <<<<<<<<<<<<<<
 * 
 *     del struct
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_Pass_Pack_int16, __pyx_n_s_ljust); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2166, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_1 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_1)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_1, __pyx_int_80};
    __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2166, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_builtin_print, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2166, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "serializor/binary/ser.py":2168
 *     print("Pass Pack int16".ljust(80))
 * 
 *     del struct             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_DECREF(__pyx_v_struct); __pyx_v_struct = 0;

  /* "serializor/binary/ser.py":2158
 * 
 * 
 * def _test_pack_int16() -> None:             # <<<<<<<<<<<<<<
 *     import struct
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("serializor.binary.ser._test_pack_int16", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_struct);
  __Pyx_XDECREF(__pyx_v_val);
  __Pyx_XDECREF(__pyx_v_s);
  __Pyx_XDECREF(__pyx_v_b);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "serializor/binary/ser.py":2171
 * 
 * 
 * def _test_pack_uint16() -> None:             # <<<<<<<<<<<<<<
 *     import struct
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_10serializor_6binary_3ser_11_test_pack_uint16(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyMethodDef __pyx_mdef_10serializor_6binary_3ser_11_test_pack_uint16 = {"_test_pack_uint16", (PyCFunction)__pyx_pw_10serializor_6binary_3ser_11_test_pack_uint16, METH_NOARGS, 0};
static PyObject *__pyx_pw_10serializor_6binary_3ser_11_test_pack_uint16(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_test_pack_uint16 (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10serializor_6binary_3ser_10_test_pack_uint16(__pyx_self);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10serializor_6binary_3ser_10_test_pack_uint16(CYTHON_UNUSED PyObject *__pyx_self) {
  PyObject *__pyx_v_struct = NULL;
  PyObject *__pyx_v_val = NULL;
  PyObject *__pyx_v_s = NULL;
  PyObject *__pyx_v_b = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  long __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_t_6;
  Py_ssize_t __pyx_t_7;
  Py_UCS4 __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_test_pack_uint16", 1);

  /* "serializor/binary/ser.py":2172
 * 
 * def _test_pack_uint16() -> None:
 *     import struct             # <<<<<<<<<<<<<<
 * 
 *     for val in range(0, 65536):
 */
  __pyx_t_1 = __Pyx_ImportDottedModule(__pyx_n_s_struct, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2172, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_struct = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "serializor/binary/ser.py":2174
 *     import struct
 * 
 *     for val in range(0, 65536):             # <<<<<<<<<<<<<<
 *         s = struct.pack("<H", val)
 *         b = pack_uint16(val)  # type: ignore
 */
  for (__pyx_t_2 = 0; __pyx_t_2 < 0x10000; __pyx_t_2+=1) {
    __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2174, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_XDECREF_SET(__pyx_v_val, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "serializor/binary/ser.py":2175
 * 
 *     for val in range(0, 65536):
 *         s = struct.pack("<H", val)             # <<<<<<<<<<<<<<
 *         b = pack_uint16(val)  # type: ignore
 *         assert s == b, f"{s} | {b}"
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_struct, __pyx_n_s_pack); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2175, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = NULL;
    __pyx_t_5 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_5 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[3] = {__pyx_t_4, __pyx_kp_u_H, __pyx_v_val};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 2+__pyx_t_5);
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2175, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
    __Pyx_XDECREF_SET(__pyx_v_s, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "serializor/binary/ser.py":2176
 *     for val in range(0, 65536):
 *         s = struct.pack("<H", val)
 *         b = pack_uint16(val)  # type: ignore             # <<<<<<<<<<<<<<
 *         assert s == b, f"{s} | {b}"
 *         print("- pass: %s" % val, end="\r")
 */
    __pyx_t_5 = __Pyx_PyInt_As_unsigned_int(__pyx_v_val); if (unlikely((__pyx_t_5 == (unsigned int)-1) && PyErr_Occurred())) __PYX_ERR(0, 2176, __pyx_L1_error)
    __pyx_t_1 = __pyx_f_10serializor_6binary_3ser_pack_uint16(__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2176, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_XDECREF_SET(__pyx_v_b, ((PyObject*)__pyx_t_1));
    __pyx_t_1 = 0;

    /* "serializor/binary/ser.py":2177
 *         s = struct.pack("<H", val)
 *         b = pack_uint16(val)  # type: ignore
 *         assert s == b, f"{s} | {b}"             # <<<<<<<<<<<<<<
 *         print("- pass: %s" % val, end="\r")
 *     print("Pass Pack uint16".ljust(80))
 */
    #ifndef CYTHON_WITHOUT_ASSERTIONS
    if (unlikely(__pyx_assertions_enabled())) {
      __pyx_t_6 = (__Pyx_PyBytes_Equals(__pyx_v_s, __pyx_v_b, Py_EQ)); if (unlikely((__pyx_t_6 < 0))) __PYX_ERR(0, 2177, __pyx_L1_error)
      if (unlikely(!__pyx_t_6)) {
        __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2177, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_7 = 0;
        __pyx_t_8 = 127;
        __pyx_t_3 = __Pyx_PyObject_FormatSimple(__pyx_v_s, __pyx_empty_unicode); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2177, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_8 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_3) > __pyx_t_8) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_3) : __pyx_t_8;
        __pyx_t_7 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_3);
        __Pyx_GIVEREF(__pyx_t_3);
        PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_3);
        __pyx_t_3 = 0;
        __Pyx_INCREF(__pyx_kp_u__15);
        __pyx_t_7 += 3;
        __Pyx_GIVEREF(__pyx_kp_u__15);
        PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_kp_u__15);
        __pyx_t_3 = __Pyx_PyObject_FormatSimple(__pyx_v_b, __pyx_empty_unicode); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2177, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_8 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_3) > __pyx_t_8) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_3) : __pyx_t_8;
        __pyx_t_7 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_3);
        __Pyx_GIVEREF(__pyx_t_3);
        PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_t_3);
        __pyx_t_3 = 0;
        __pyx_t_3 = __Pyx_PyUnicode_Join(__pyx_t_1, 3, __pyx_t_7, __pyx_t_8); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2177, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_Raise(__pyx_builtin_AssertionError, __pyx_t_3, 0, 0);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __PYX_ERR(0, 2177, __pyx_L1_error)
      }
    }
    #else
    if ((1)); else __PYX_ERR(0, 2177, __pyx_L1_error)
    #endif

    /* "serializor/binary/ser.py":2178
 *         b = pack_uint16(val)  # type: ignore
 *         assert s == b, f"{s} | {b}"
 *         print("- pass: %s" % val, end="\r")             # <<<<<<<<<<<<<<
 *     print("Pass Pack uint16".ljust(80))
 * 
 */
    __pyx_t_3 = __Pyx_PyUnicode_FormatSafe(__pyx_kp_u_pass_s, __pyx_v_val); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2178, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2178, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_3);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_3)) __PYX_ERR(0, 2178, __pyx_L1_error);
    __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2178, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_end, __pyx_kp_u__16) < 0) __PYX_ERR(0, 2178, __pyx_L1_error)
    __pyx_t_4 = __Pyx_PyObject_Call(__pyx_builtin_print, __pyx_t_1, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2178, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  }

  /* "serializor/binary/ser.py":2179
 *         assert s == b, f"{s} | {b}"
 *         print("- pass: %s" % val, end="\r")
 *     print("Pass Pack uint16".ljust(80))             # <<<<<<<<<<<<<<
 * 
 *     del struct
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_Pass_Pack_uint16, __pyx_n_s_ljust); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2179, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_1 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_1)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_1, __pyx_int_80};
    __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2179, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_builtin_print, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2179, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "serializor/binary/ser.py":2181
 *     print("Pass Pack uint16".ljust(80))
 * 
 *     del struct             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_DECREF(__pyx_v_struct); __pyx_v_struct = 0;

  /* "serializor/binary/ser.py":2171
 * 
 * 
 * def _test_pack_uint16() -> None:             # <<<<<<<<<<<<<<
 *     import struct
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("serializor.binary.ser._test_pack_uint16", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_struct);
  __Pyx_XDECREF(__pyx_v_val);
  __Pyx_XDECREF(__pyx_v_s);
  __Pyx_XDECREF(__pyx_v_b);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "serializor/binary/ser.py":2184
 * 
 * 
 * def _test_pack_int24() -> None:             # <<<<<<<<<<<<<<
 *     import struct
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_10serializor_6binary_3ser_13_test_pack_int24(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyMethodDef __pyx_mdef_10serializor_6binary_3ser_13_test_pack_int24 = {"_test_pack_int24", (PyCFunction)__pyx_pw_10serializor_6binary_3ser_13_test_pack_int24, METH_NOARGS, 0};
static PyObject *__pyx_pw_10serializor_6binary_3ser_13_test_pack_int24(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_test_pack_int24 (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10serializor_6binary_3ser_12_test_pack_int24(__pyx_self);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10serializor_6binary_3ser_12_test_pack_int24(CYTHON_UNUSED PyObject *__pyx_self) {
  PyObject *__pyx_v_struct = NULL;
  PyObject *__pyx_v_val = NULL;
  PyObject *__pyx_v_s = NULL;
  PyObject *__pyx_v_b = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  unsigned int __pyx_t_6;
  int __pyx_t_7;
  int __pyx_t_8;
  Py_ssize_t __pyx_t_9;
  Py_UCS4 __pyx_t_10;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_test_pack_int24", 1);

  /* "serializor/binary/ser.py":2185
 * 
 * def _test_pack_int24() -> None:
 *     import struct             # <<<<<<<<<<<<<<
 * 
 *     for val in (-8388608, -8388607, 0, 8388606, 8388607):
 */
  __pyx_t_1 = __Pyx_ImportDottedModule(__pyx_n_s_struct, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2185, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_struct = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "serializor/binary/ser.py":2187
 *     import struct
 * 
 *     for val in (-8388608, -8388607, 0, 8388606, 8388607):             # <<<<<<<<<<<<<<
 *         s = struct.pack("<i", val)[:3]
 *         b = pack_int24(val)  # type: ignore
 */
  __pyx_t_1 = __pyx_tuple__17; __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = 0;
  for (;;) {
    if (__pyx_t_2 >= 5) break;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_3); __pyx_t_2++; if (unlikely((0 < 0))) __PYX_ERR(0, 2187, __pyx_L1_error)
    #else
    __pyx_t_3 = __Pyx_PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2187, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    #endif
    __Pyx_XDECREF_SET(__pyx_v_val, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "serializor/binary/ser.py":2188
 * 
 *     for val in (-8388608, -8388607, 0, 8388606, 8388607):
 *         s = struct.pack("<i", val)[:3]             # <<<<<<<<<<<<<<
 *         b = pack_int24(val)  # type: ignore
 *         assert s == b, f"{s} | {b}"
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_struct, __pyx_n_s_pack); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2188, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    __pyx_t_6 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
        __pyx_t_6 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[3] = {__pyx_t_5, __pyx_kp_u_i, __pyx_v_val};
      __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_6, 2+__pyx_t_6);
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2188, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
    __pyx_t_4 = __Pyx_PyObject_GetSlice(__pyx_t_3, 0, 3, NULL, NULL, &__pyx_slice__18, 0, 1, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2188, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_XDECREF_SET(__pyx_v_s, __pyx_t_4);
    __pyx_t_4 = 0;

    /* "serializor/binary/ser.py":2189
 *     for val in (-8388608, -8388607, 0, 8388606, 8388607):
 *         s = struct.pack("<i", val)[:3]
 *         b = pack_int24(val)  # type: ignore             # <<<<<<<<<<<<<<
 *         assert s == b, f"{s} | {b}"
 *         print("- pass: %s" % val, end="\r")
 */
    __pyx_t_7 = __Pyx_PyInt_As_int(__pyx_v_val); if (unlikely((__pyx_t_7 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 2189, __pyx_L1_error)
    __pyx_t_4 = __pyx_f_10serializor_6binary_3ser_pack_int24(__pyx_t_7); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2189, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_XDECREF_SET(__pyx_v_b, ((PyObject*)__pyx_t_4));
    __pyx_t_4 = 0;

    /* "serializor/binary/ser.py":2190
 *         s = struct.pack("<i", val)[:3]
 *         b = pack_int24(val)  # type: ignore
 *         assert s == b, f"{s} | {b}"             # <<<<<<<<<<<<<<
 *         print("- pass: %s" % val, end="\r")
 *     print("Pass Pack int24".ljust(80))
 */
    #ifndef CYTHON_WITHOUT_ASSERTIONS
    if (unlikely(__pyx_assertions_enabled())) {
      __pyx_t_8 = (__Pyx_PyBytes_Equals(__pyx_v_s, __pyx_v_b, Py_EQ)); if (unlikely((__pyx_t_8 < 0))) __PYX_ERR(0, 2190, __pyx_L1_error)
      if (unlikely(!__pyx_t_8)) {
        __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2190, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_9 = 0;
        __pyx_t_10 = 127;
        __pyx_t_3 = __Pyx_PyObject_FormatSimple(__pyx_v_s, __pyx_empty_unicode); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2190, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_10 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_3) > __pyx_t_10) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_3) : __pyx_t_10;
        __pyx_t_9 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_3);
        __Pyx_GIVEREF(__pyx_t_3);
        PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_3);
        __pyx_t_3 = 0;
        __Pyx_INCREF(__pyx_kp_u__15);
        __pyx_t_9 += 3;
        __Pyx_GIVEREF(__pyx_kp_u__15);
        PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_kp_u__15);
        __pyx_t_3 = __Pyx_PyObject_FormatSimple(__pyx_v_b, __pyx_empty_unicode); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2190, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_10 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_3) > __pyx_t_10) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_3) : __pyx_t_10;
        __pyx_t_9 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_3);
        __Pyx_GIVEREF(__pyx_t_3);
        PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_t_3);
        __pyx_t_3 = 0;
        __pyx_t_3 = __Pyx_PyUnicode_Join(__pyx_t_4, 3, __pyx_t_9, __pyx_t_10); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2190, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_Raise(__pyx_builtin_AssertionError, __pyx_t_3, 0, 0);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __PYX_ERR(0, 2190, __pyx_L1_error)
      }
    }
    #else
    if ((1)); else __PYX_ERR(0, 2190, __pyx_L1_error)
    #endif

    /* "serializor/binary/ser.py":2191
 *         b = pack_int24(val)  # type: ignore
 *         assert s == b, f"{s} | {b}"
 *         print("- pass: %s" % val, end="\r")             # <<<<<<<<<<<<<<
 *     print("Pass Pack int24".ljust(80))
 * 
 */
    __pyx_t_3 = __Pyx_PyUnicode_FormatSafe(__pyx_kp_u_pass_s, __pyx_v_val); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2191, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2191, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_3);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_3)) __PYX_ERR(0, 2191, __pyx_L1_error);
    __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2191, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_end, __pyx_kp_u__16) < 0) __PYX_ERR(0, 2191, __pyx_L1_error)
    __pyx_t_5 = __Pyx_PyObject_Call(__pyx_builtin_print, __pyx_t_4, __pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2191, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "serializor/binary/ser.py":2187
 *     import struct
 * 
 *     for val in (-8388608, -8388607, 0, 8388606, 8388607):             # <<<<<<<<<<<<<<
 *         s = struct.pack("<i", val)[:3]
 *         b = pack_int24(val)  # type: ignore
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "serializor/binary/ser.py":2192
 *         assert s == b, f"{s} | {b}"
 *         print("- pass: %s" % val, end="\r")
 *     print("Pass Pack int24".ljust(80))             # <<<<<<<<<<<<<<
 * 
 *     del struct
 */
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_Pass_Pack_int24, __pyx_n_s_ljust); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2192, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_3 = NULL;
  __pyx_t_6 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
      __pyx_t_6 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_int_80};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2192, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __pyx_t_5 = __Pyx_PyObject_CallOneArg(__pyx_builtin_print, __pyx_t_1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2192, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "serializor/binary/ser.py":2194
 *     print("Pass Pack int24".ljust(80))
 * 
 *     del struct             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_DECREF(__pyx_v_struct); __pyx_v_struct = 0;

  /* "serializor/binary/ser.py":2184
 * 
 * 
 * def _test_pack_int24() -> None:             # <<<<<<<<<<<<<<
 *     import struct
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("serializor.binary.ser._test_pack_int24", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_struct);
  __Pyx_XDECREF(__pyx_v_val);
  __Pyx_XDECREF(__pyx_v_s);
  __Pyx_XDECREF(__pyx_v_b);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "serializor/binary/ser.py":2197
 * 
 * 
 * def _test_pack_uint24() -> None:             # <<<<<<<<<<<<<<
 *     import struct
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_10serializor_6binary_3ser_15_test_pack_uint24(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyMethodDef __pyx_mdef_10serializor_6binary_3ser_15_test_pack_uint24 = {"_test_pack_uint24", (PyCFunction)__pyx_pw_10serializor_6binary_3ser_15_test_pack_uint24, METH_NOARGS, 0};
static PyObject *__pyx_pw_10serializor_6binary_3ser_15_test_pack_uint24(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_test_pack_uint24 (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10serializor_6binary_3ser_14_test_pack_uint24(__pyx_self);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10serializor_6binary_3ser_14_test_pack_uint24(CYTHON_UNUSED PyObject *__pyx_self) {
  PyObject *__pyx_v_struct = NULL;
  PyObject *__pyx_v_val = NULL;
  PyObject *__pyx_v_s = NULL;
  PyObject *__pyx_v_b = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  unsigned int __pyx_t_6;
  int __pyx_t_7;
  Py_ssize_t __pyx_t_8;
  Py_UCS4 __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_test_pack_uint24", 1);

  /* "serializor/binary/ser.py":2198
 * 
 * def _test_pack_uint24() -> None:
 *     import struct             # <<<<<<<<<<<<<<
 * 
 *     for val in (0, 1, 16777213, 16777214, 16777215):
 */
  __pyx_t_1 = __Pyx_ImportDottedModule(__pyx_n_s_struct, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2198, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_struct = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "serializor/binary/ser.py":2200
 *     import struct
 * 
 *     for val in (0, 1, 16777213, 16777214, 16777215):             # <<<<<<<<<<<<<<
 *         s = struct.pack("<I", val)[:3]
 *         b = pack_uint24(val)  # type: ignore
 */
  __pyx_t_1 = __pyx_tuple__19; __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = 0;
  for (;;) {
    if (__pyx_t_2 >= 5) break;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_3); __pyx_t_2++; if (unlikely((0 < 0))) __PYX_ERR(0, 2200, __pyx_L1_error)
    #else
    __pyx_t_3 = __Pyx_PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2200, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    #endif
    __Pyx_XDECREF_SET(__pyx_v_val, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "serializor/binary/ser.py":2201
 * 
 *     for val in (0, 1, 16777213, 16777214, 16777215):
 *         s = struct.pack("<I", val)[:3]             # <<<<<<<<<<<<<<
 *         b = pack_uint24(val)  # type: ignore
 *         assert s == b, f"{s} | {b}"
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_struct, __pyx_n_s_pack); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2201, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    __pyx_t_6 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
        __pyx_t_6 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[3] = {__pyx_t_5, __pyx_kp_u_I, __pyx_v_val};
      __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_6, 2+__pyx_t_6);
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2201, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
    __pyx_t_4 = __Pyx_PyObject_GetSlice(__pyx_t_3, 0, 3, NULL, NULL, &__pyx_slice__18, 0, 1, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2201, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_XDECREF_SET(__pyx_v_s, __pyx_t_4);
    __pyx_t_4 = 0;

    /* "serializor/binary/ser.py":2202
 *     for val in (0, 1, 16777213, 16777214, 16777215):
 *         s = struct.pack("<I", val)[:3]
 *         b = pack_uint24(val)  # type: ignore             # <<<<<<<<<<<<<<
 *         assert s == b, f"{s} | {b}"
 *         print("- pass: %s" % val, end="\r")
 */
    __pyx_t_6 = __Pyx_PyInt_As_unsigned_int(__pyx_v_val); if (unlikely((__pyx_t_6 == (unsigned int)-1) && PyErr_Occurred())) __PYX_ERR(0, 2202, __pyx_L1_error)
    __pyx_t_4 = __pyx_f_10serializor_6binary_3ser_pack_uint24(__pyx_t_6); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2202, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_XDECREF_SET(__pyx_v_b, ((PyObject*)__pyx_t_4));
    __pyx_t_4 = 0;

    /* "serializor/binary/ser.py":2203
 *         s = struct.pack("<I", val)[:3]
 *         b = pack_uint24(val)  # type: ignore
 *         assert s == b, f"{s} | {b}"             # <<<<<<<<<<<<<<
 *         print("- pass: %s" % val, end="\r")
 *     print("Pass Pack uint24".ljust(80))
 */
    #ifndef CYTHON_WITHOUT_ASSERTIONS
    if (unlikely(__pyx_assertions_enabled())) {
      __pyx_t_7 = (__Pyx_PyBytes_Equals(__pyx_v_s, __pyx_v_b, Py_EQ)); if (unlikely((__pyx_t_7 < 0))) __PYX_ERR(0, 2203, __pyx_L1_error)
      if (unlikely(!__pyx_t_7)) {
        __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2203, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_8 = 0;
        __pyx_t_9 = 127;
        __pyx_t_3 = __Pyx_PyObject_FormatSimple(__pyx_v_s, __pyx_empty_unicode); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2203, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_9 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_3) > __pyx_t_9) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_3) : __pyx_t_9;
        __pyx_t_8 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_3);
        __Pyx_GIVEREF(__pyx_t_3);
        PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_3);
        __pyx_t_3 = 0;
        __Pyx_INCREF(__pyx_kp_u__15);
        __pyx_t_8 += 3;
        __Pyx_GIVEREF(__pyx_kp_u__15);
        PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_kp_u__15);
        __pyx_t_3 = __Pyx_PyObject_FormatSimple(__pyx_v_b, __pyx_empty_unicode); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2203, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_9 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_3) > __pyx_t_9) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_3) : __pyx_t_9;
        __pyx_t_8 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_3);
        __Pyx_GIVEREF(__pyx_t_3);
        PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_t_3);
        __pyx_t_3 = 0;
        __pyx_t_3 = __Pyx_PyUnicode_Join(__pyx_t_4, 3, __pyx_t_8, __pyx_t_9); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2203, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_Raise(__pyx_builtin_AssertionError, __pyx_t_3, 0, 0);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __PYX_ERR(0, 2203, __pyx_L1_error)
      }
    }
    #else
    if ((1)); else __PYX_ERR(0, 2203, __pyx_L1_error)
    #endif

    /* "serializor/binary/ser.py":2204
 *         b = pack_uint24(val)  # type: ignore
 *         assert s == b, f"{s} | {b}"
 *         print("- pass: %s" % val, end="\r")             # <<<<<<<<<<<<<<
 *     print("Pass Pack uint24".ljust(80))
 * 
 */
    __pyx_t_3 = __Pyx_PyUnicode_FormatSafe(__pyx_kp_u_pass_s, __pyx_v_val); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2204, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2204, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_3);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_3)) __PYX_ERR(0, 2204, __pyx_L1_error);
    __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2204, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_end, __pyx_kp_u__16) < 0) __PYX_ERR(0, 2204, __pyx_L1_error)
    __pyx_t_5 = __Pyx_PyObject_Call(__pyx_builtin_print, __pyx_t_4, __pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2204, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "serializor/binary/ser.py":2200
 *     import struct
 * 
 *     for val in (0, 1, 16777213, 16777214, 16777215):             # <<<<<<<<<<<<<<
 *         s = struct.pack("<I", val)[:3]
 *         b = pack_uint24(val)  # type: ignore
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "serializor/binary/ser.py":2205
 *         assert s == b, f"{s} | {b}"
 *         print("- pass: %s" % val, end="\r")
 *     print("Pass Pack uint24".ljust(80))             # <<<<<<<<<<<<<<
 * 
 *     del struct
 */
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_Pass_Pack_uint24, __pyx_n_s_ljust); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2205, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_3 = NULL;
  __pyx_t_6 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
      __pyx_t_6 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_int_80};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2205, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __pyx_t_5 = __Pyx_PyObject_CallOneArg(__pyx_builtin_print, __pyx_t_1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2205, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "serializor/binary/ser.py":2207
 *     print("Pass Pack uint24".ljust(80))
 * 
 *     del struct             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_DECREF(__pyx_v_struct); __pyx_v_struct = 0;

  /* "serializor/binary/ser.py":2197
 * 
 * 
 * def _test_pack_uint24() -> None:             # <<<<<<<<<<<<<<
 *     import struct
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("serializor.binary.ser._test_pack_uint24", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_struct);
  __Pyx_XDECREF(__pyx_v_val);
  __Pyx_XDECREF(__pyx_v_s);
  __Pyx_XDECREF(__pyx_v_b);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "serializor/binary/ser.py":2210
 * 
 * 
 * def _test_pack_int32() -> None:             # <<<<<<<<<<<<<<
 *     import struct
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_10serializor_6binary_3ser_17_test_pack_int32(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyMethodDef __pyx_mdef_10serializor_6binary_3ser_17_test_pack_int32 = {"_test_pack_int32", (PyCFunction)__pyx_pw_10serializor_6binary_3ser_17_test_pack_int32, METH_NOARGS, 0};
static PyObject *__pyx_pw_10serializor_6binary_3ser_17_test_pack_int32(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_test_pack_int32 (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10serializor_6binary_3ser_16_test_pack_int32(__pyx_self);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10serializor_6binary_3ser_16_test_pack_int32(CYTHON_UNUSED PyObject *__pyx_self) {
  PyObject *__pyx_v_struct = NULL;
  PyObject *__pyx_v_val = NULL;
  PyObject *__pyx_v_s = NULL;
  PyObject *__pyx_v_b = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  unsigned int __pyx_t_6;
  PY_LONG_LONG __pyx_t_7;
  int __pyx_t_8;
  Py_ssize_t __pyx_t_9;
  Py_UCS4 __pyx_t_10;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_test_pack_int32", 1);

  /* "serializor/binary/ser.py":2211
 * 
 * def _test_pack_int32() -> None:
 *     import struct             # <<<<<<<<<<<<<<
 * 
 *     for val in (-2147483648, -2147483647, 0, 2147483646, 2147483647):
 */
  __pyx_t_1 = __Pyx_ImportDottedModule(__pyx_n_s_struct, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2211, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_struct = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "serializor/binary/ser.py":2213
 *     import struct
 * 
 *     for val in (-2147483648, -2147483647, 0, 2147483646, 2147483647):             # <<<<<<<<<<<<<<
 *         s = struct.pack("<i", val)
 *         b = pack_int32(val)  # type: ignore
 */
  __pyx_t_1 = __pyx_tuple__20; __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = 0;
  for (;;) {
    if (__pyx_t_2 >= 5) break;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_3); __pyx_t_2++; if (unlikely((0 < 0))) __PYX_ERR(0, 2213, __pyx_L1_error)
    #else
    __pyx_t_3 = __Pyx_PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2213, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    #endif
    __Pyx_XDECREF_SET(__pyx_v_val, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "serializor/binary/ser.py":2214
 * 
 *     for val in (-2147483648, -2147483647, 0, 2147483646, 2147483647):
 *         s = struct.pack("<i", val)             # <<<<<<<<<<<<<<
 *         b = pack_int32(val)  # type: ignore
 *         assert s == b, f"{s} | {b}"
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_struct, __pyx_n_s_pack); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2214, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    __pyx_t_6 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
        __pyx_t_6 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[3] = {__pyx_t_5, __pyx_kp_u_i, __pyx_v_val};
      __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_6, 2+__pyx_t_6);
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2214, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
    __Pyx_XDECREF_SET(__pyx_v_s, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "serializor/binary/ser.py":2215
 *     for val in (-2147483648, -2147483647, 0, 2147483646, 2147483647):
 *         s = struct.pack("<i", val)
 *         b = pack_int32(val)  # type: ignore             # <<<<<<<<<<<<<<
 *         assert s == b, f"{s} | {b}"
 *         print("- pass: %s" % val, end="\r")
 */
    __pyx_t_7 = __Pyx_PyInt_As_PY_LONG_LONG(__pyx_v_val); if (unlikely((__pyx_t_7 == (PY_LONG_LONG)-1) && PyErr_Occurred())) __PYX_ERR(0, 2215, __pyx_L1_error)
    __pyx_t_3 = __pyx_f_10serializor_6binary_3ser_pack_int32(__pyx_t_7); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2215, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_XDECREF_SET(__pyx_v_b, ((PyObject*)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "serializor/binary/ser.py":2216
 *         s = struct.pack("<i", val)
 *         b = pack_int32(val)  # type: ignore
 *         assert s == b, f"{s} | {b}"             # <<<<<<<<<<<<<<
 *         print("- pass: %s" % val, end="\r")
 *     print("Pass Pack int32".ljust(80))
 */
    #ifndef CYTHON_WITHOUT_ASSERTIONS
    if (unlikely(__pyx_assertions_enabled())) {
      __pyx_t_8 = (__Pyx_PyBytes_Equals(__pyx_v_s, __pyx_v_b, Py_EQ)); if (unlikely((__pyx_t_8 < 0))) __PYX_ERR(0, 2216, __pyx_L1_error)
      if (unlikely(!__pyx_t_8)) {
        __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2216, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_9 = 0;
        __pyx_t_10 = 127;
        __pyx_t_4 = __Pyx_PyObject_FormatSimple(__pyx_v_s, __pyx_empty_unicode); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2216, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_10 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) > __pyx_t_10) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) : __pyx_t_10;
        __pyx_t_9 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_4);
        __Pyx_GIVEREF(__pyx_t_4);
        PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_4);
        __pyx_t_4 = 0;
        __Pyx_INCREF(__pyx_kp_u__15);
        __pyx_t_9 += 3;
        __Pyx_GIVEREF(__pyx_kp_u__15);
        PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_kp_u__15);
        __pyx_t_4 = __Pyx_PyObject_FormatSimple(__pyx_v_b, __pyx_empty_unicode); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2216, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_10 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) > __pyx_t_10) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) : __pyx_t_10;
        __pyx_t_9 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_4);
        __Pyx_GIVEREF(__pyx_t_4);
        PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_t_4);
        __pyx_t_4 = 0;
        __pyx_t_4 = __Pyx_PyUnicode_Join(__pyx_t_3, 3, __pyx_t_9, __pyx_t_10); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2216, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_Raise(__pyx_builtin_AssertionError, __pyx_t_4, 0, 0);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __PYX_ERR(0, 2216, __pyx_L1_error)
      }
    }
    #else
    if ((1)); else __PYX_ERR(0, 2216, __pyx_L1_error)
    #endif

    /* "serializor/binary/ser.py":2217
 *         b = pack_int32(val)  # type: ignore
 *         assert s == b, f"{s} | {b}"
 *         print("- pass: %s" % val, end="\r")             # <<<<<<<<<<<<<<
 *     print("Pass Pack int32".ljust(80))
 * 
 */
    __pyx_t_4 = __Pyx_PyUnicode_FormatSafe(__pyx_kp_u_pass_s, __pyx_v_val); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2217, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2217, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_4);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_4)) __PYX_ERR(0, 2217, __pyx_L1_error);
    __pyx_t_4 = 0;
    __pyx_t_4 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2217, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_end, __pyx_kp_u__16) < 0) __PYX_ERR(0, 2217, __pyx_L1_error)
    __pyx_t_5 = __Pyx_PyObject_Call(__pyx_builtin_print, __pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2217, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "serializor/binary/ser.py":2213
 *     import struct
 * 
 *     for val in (-2147483648, -2147483647, 0, 2147483646, 2147483647):             # <<<<<<<<<<<<<<
 *         s = struct.pack("<i", val)
 *         b = pack_int32(val)  # type: ignore
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "serializor/binary/ser.py":2218
 *         assert s == b, f"{s} | {b}"
 *         print("- pass: %s" % val, end="\r")
 *     print("Pass Pack int32".ljust(80))             # <<<<<<<<<<<<<<
 * 
 *     del struct
 */
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_Pass_Pack_int32, __pyx_n_s_ljust); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2218, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_4 = NULL;
  __pyx_t_6 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
      __pyx_t_6 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_int_80};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2218, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __pyx_t_5 = __Pyx_PyObject_CallOneArg(__pyx_builtin_print, __pyx_t_1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2218, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "serializor/binary/ser.py":2220
 *     print("Pass Pack int32".ljust(80))
 * 
 *     del struct             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_DECREF(__pyx_v_struct); __pyx_v_struct = 0;

  /* "serializor/binary/ser.py":2210
 * 
 * 
 * def _test_pack_int32() -> None:             # <<<<<<<<<<<<<<
 *     import struct
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("serializor.binary.ser._test_pack_int32", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_struct);
  __Pyx_XDECREF(__pyx_v_val);
  __Pyx_XDECREF(__pyx_v_s);
  __Pyx_XDECREF(__pyx_v_b);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "serializor/binary/ser.py":2223
 * 
 * 
 * def _test_pack_uint32() -> None:             # <<<<<<<<<<<<<<
 *     import struct
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_10serializor_6binary_3ser_19_test_pack_uint32(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyMethodDef __pyx_mdef_10serializor_6binary_3ser_19_test_pack_uint32 = {"_test_pack_uint32", (PyCFunction)__pyx_pw_10serializor_6binary_3ser_19_test_pack_uint32, METH_NOARGS, 0};
static PyObject *__pyx_pw_10serializor_6binary_3ser_19_test_pack_uint32(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_test_pack_uint32 (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10serializor_6binary_3ser_18_test_pack_uint32(__pyx_self);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10serializor_6binary_3ser_18_test_pack_uint32(CYTHON_UNUSED PyObject *__pyx_self) {
  PyObject *__pyx_v_struct = NULL;
  PyObject *__pyx_v_val = NULL;
  PyObject *__pyx_v_s = NULL;
  PyObject *__pyx_v_b = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  unsigned int __pyx_t_6;
  unsigned PY_LONG_LONG __pyx_t_7;
  int __pyx_t_8;
  Py_ssize_t __pyx_t_9;
  Py_UCS4 __pyx_t_10;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_test_pack_uint32", 1);

  /* "serializor/binary/ser.py":2224
 * 
 * def _test_pack_uint32() -> None:
 *     import struct             # <<<<<<<<<<<<<<
 * 
 *     for val in (0, 1, 4294967293, 4294967294, 4294967295):
 */
  __pyx_t_1 = __Pyx_ImportDottedModule(__pyx_n_s_struct, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2224, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_struct = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "serializor/binary/ser.py":2226
 *     import struct
 * 
 *     for val in (0, 1, 4294967293, 4294967294, 4294967295):             # <<<<<<<<<<<<<<
 *         s = struct.pack("<I", val)
 *         b = pack_uint32(val)  # type: ignore
 */
  __pyx_t_1 = __pyx_tuple__21; __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = 0;
  for (;;) {
    if (__pyx_t_2 >= 5) break;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_3); __pyx_t_2++; if (unlikely((0 < 0))) __PYX_ERR(0, 2226, __pyx_L1_error)
    #else
    __pyx_t_3 = __Pyx_PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2226, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    #endif
    __Pyx_XDECREF_SET(__pyx_v_val, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "serializor/binary/ser.py":2227
 * 
 *     for val in (0, 1, 4294967293, 4294967294, 4294967295):
 *         s = struct.pack("<I", val)             # <<<<<<<<<<<<<<
 *         b = pack_uint32(val)  # type: ignore
 *         assert s == b, f"{s} | {b}"
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_struct, __pyx_n_s_pack); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2227, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    __pyx_t_6 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
        __pyx_t_6 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[3] = {__pyx_t_5, __pyx_kp_u_I, __pyx_v_val};
      __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_6, 2+__pyx_t_6);
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2227, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
    __Pyx_XDECREF_SET(__pyx_v_s, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "serializor/binary/ser.py":2228
 *     for val in (0, 1, 4294967293, 4294967294, 4294967295):
 *         s = struct.pack("<I", val)
 *         b = pack_uint32(val)  # type: ignore             # <<<<<<<<<<<<<<
 *         assert s == b, f"{s} | {b}"
 *         print("- pass: %s" % val, end="\r")
 */
    __pyx_t_7 = __Pyx_PyInt_As_unsigned_PY_LONG_LONG(__pyx_v_val); if (unlikely((__pyx_t_7 == (unsigned PY_LONG_LONG)-1) && PyErr_Occurred())) __PYX_ERR(0, 2228, __pyx_L1_error)
    __pyx_t_3 = __pyx_f_10serializor_6binary_3ser_pack_uint32(__pyx_t_7); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2228, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_XDECREF_SET(__pyx_v_b, ((PyObject*)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "serializor/binary/ser.py":2229
 *         s = struct.pack("<I", val)
 *         b = pack_uint32(val)  # type: ignore
 *         assert s == b, f"{s} | {b}"             # <<<<<<<<<<<<<<
 *         print("- pass: %s" % val, end="\r")
 *     print("Pass Pack uint32".ljust(80))
 */
    #ifndef CYTHON_WITHOUT_ASSERTIONS
    if (unlikely(__pyx_assertions_enabled())) {
      __pyx_t_8 = (__Pyx_PyBytes_Equals(__pyx_v_s, __pyx_v_b, Py_EQ)); if (unlikely((__pyx_t_8 < 0))) __PYX_ERR(0, 2229, __pyx_L1_error)
      if (unlikely(!__pyx_t_8)) {
        __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2229, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_9 = 0;
        __pyx_t_10 = 127;
        __pyx_t_4 = __Pyx_PyObject_FormatSimple(__pyx_v_s, __pyx_empty_unicode); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2229, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_10 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) > __pyx_t_10) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) : __pyx_t_10;
        __pyx_t_9 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_4);
        __Pyx_GIVEREF(__pyx_t_4);
        PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_4);
        __pyx_t_4 = 0;
        __Pyx_INCREF(__pyx_kp_u__15);
        __pyx_t_9 += 3;
        __Pyx_GIVEREF(__pyx_kp_u__15);
        PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_kp_u__15);
        __pyx_t_4 = __Pyx_PyObject_FormatSimple(__pyx_v_b, __pyx_empty_unicode); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2229, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_10 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) > __pyx_t_10) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) : __pyx_t_10;
        __pyx_t_9 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_4);
        __Pyx_GIVEREF(__pyx_t_4);
        PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_t_4);
        __pyx_t_4 = 0;
        __pyx_t_4 = __Pyx_PyUnicode_Join(__pyx_t_3, 3, __pyx_t_9, __pyx_t_10); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2229, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_Raise(__pyx_builtin_AssertionError, __pyx_t_4, 0, 0);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __PYX_ERR(0, 2229, __pyx_L1_error)
      }
    }
    #else
    if ((1)); else __PYX_ERR(0, 2229, __pyx_L1_error)
    #endif

    /* "serializor/binary/ser.py":2230
 *         b = pack_uint32(val)  # type: ignore
 *         assert s == b, f"{s} | {b}"
 *         print("- pass: %s" % val, end="\r")             # <<<<<<<<<<<<<<
 *     print("Pass Pack uint32".ljust(80))
 * 
 */
    __pyx_t_4 = __Pyx_PyUnicode_FormatSafe(__pyx_kp_u_pass_s, __pyx_v_val); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2230, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2230, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_4);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_4)) __PYX_ERR(0, 2230, __pyx_L1_error);
    __pyx_t_4 = 0;
    __pyx_t_4 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2230, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_end, __pyx_kp_u__16) < 0) __PYX_ERR(0, 2230, __pyx_L1_error)
    __pyx_t_5 = __Pyx_PyObject_Call(__pyx_builtin_print, __pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2230, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "serializor/binary/ser.py":2226
 *     import struct
 * 
 *     for val in (0, 1, 4294967293, 4294967294, 4294967295):             # <<<<<<<<<<<<<<
 *         s = struct.pack("<I", val)
 *         b = pack_uint32(val)  # type: ignore
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "serializor/binary/ser.py":2231
 *         assert s == b, f"{s} | {b}"
 *         print("- pass: %s" % val, end="\r")
 *     print("Pass Pack uint32".ljust(80))             # <<<<<<<<<<<<<<
 * 
 *     del struct
 */
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_Pass_Pack_uint32, __pyx_n_s_ljust); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2231, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_4 = NULL;
  __pyx_t_6 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
      __pyx_t_6 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_int_80};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2231, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __pyx_t_5 = __Pyx_PyObject_CallOneArg(__pyx_builtin_print, __pyx_t_1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2231, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "serializor/binary/ser.py":2233
 *     print("Pass Pack uint32".ljust(80))
 * 
 *     del struct             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_DECREF(__pyx_v_struct); __pyx_v_struct = 0;

  /* "serializor/binary/ser.py":2223
 * 
 * 
 * def _test_pack_uint32() -> None:             # <<<<<<<<<<<<<<
 *     import struct
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("serializor.binary.ser._test_pack_uint32", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_struct);
  __Pyx_XDECREF(__pyx_v_val);
  __Pyx_XDECREF(__pyx_v_s);
  __Pyx_XDECREF(__pyx_v_b);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "serializor/binary/ser.py":2236
 * 
 * 
 * def _test_pack_int64() -> None:             # <<<<<<<<<<<<<<
 *     import struct
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_10serializor_6binary_3ser_21_test_pack_int64(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyMethodDef __pyx_mdef_10serializor_6binary_3ser_21_test_pack_int64 = {"_test_pack_int64", (PyCFunction)__pyx_pw_10serializor_6binary_3ser_21_test_pack_int64, METH_NOARGS, 0};
static PyObject *__pyx_pw_10serializor_6binary_3ser_21_test_pack_int64(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_test_pack_int64 (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10serializor_6binary_3ser_20_test_pack_int64(__pyx_self);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10serializor_6binary_3ser_20_test_pack_int64(CYTHON_UNUSED PyObject *__pyx_self) {
  PyObject *__pyx_v_struct = NULL;
  PyObject *__pyx_v_val = NULL;
  PyObject *__pyx_v_s = NULL;
  PyObject *__pyx_v_b = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  unsigned int __pyx_t_6;
  PY_LONG_LONG __pyx_t_7;
  int __pyx_t_8;
  Py_ssize_t __pyx_t_9;
  Py_UCS4 __pyx_t_10;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_test_pack_int64", 1);

  /* "serializor/binary/ser.py":2237
 * 
 * def _test_pack_int64() -> None:
 *     import struct             # <<<<<<<<<<<<<<
 * 
 *     for val in (
 */
  __pyx_t_1 = __Pyx_ImportDottedModule(__pyx_n_s_struct, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2237, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_struct = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "serializor/binary/ser.py":2239
 *     import struct
 * 
 *     for val in (             # <<<<<<<<<<<<<<
 *         -9223372036854775808,
 *         -9223372036854775807,
 */
  __pyx_t_1 = __pyx_tuple__22; __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = 0;
  for (;;) {
    if (__pyx_t_2 >= 5) break;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_3); __pyx_t_2++; if (unlikely((0 < 0))) __PYX_ERR(0, 2239, __pyx_L1_error)
    #else
    __pyx_t_3 = __Pyx_PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2239, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    #endif
    __Pyx_XDECREF_SET(__pyx_v_val, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "serializor/binary/ser.py":2246
 *         9223372036854775807,
 *     ):
 *         s = struct.pack("<q", val)             # <<<<<<<<<<<<<<
 *         b = pack_int64(val)  # type: ignore
 *         assert s == b, f"{s} | {b}"
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_struct, __pyx_n_s_pack); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2246, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    __pyx_t_6 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
        __pyx_t_6 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[3] = {__pyx_t_5, __pyx_kp_u_q, __pyx_v_val};
      __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_6, 2+__pyx_t_6);
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2246, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
    __Pyx_XDECREF_SET(__pyx_v_s, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "serializor/binary/ser.py":2247
 *     ):
 *         s = struct.pack("<q", val)
 *         b = pack_int64(val)  # type: ignore             # <<<<<<<<<<<<<<
 *         assert s == b, f"{s} | {b}"
 *         print("- pass: %s" % val, end="\r")
 */
    __pyx_t_7 = __Pyx_PyInt_As_PY_LONG_LONG(__pyx_v_val); if (unlikely((__pyx_t_7 == (PY_LONG_LONG)-1) && PyErr_Occurred())) __PYX_ERR(0, 2247, __pyx_L1_error)
    __pyx_t_3 = __pyx_f_10serializor_6binary_3ser_pack_int64(__pyx_t_7); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2247, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_XDECREF_SET(__pyx_v_b, ((PyObject*)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "serializor/binary/ser.py":2248
 *         s = struct.pack("<q", val)
 *         b = pack_int64(val)  # type: ignore
 *         assert s == b, f"{s} | {b}"             # <<<<<<<<<<<<<<
 *         print("- pass: %s" % val, end="\r")
 *     print("Pass Pack int64".ljust(80))
 */
    #ifndef CYTHON_WITHOUT_ASSERTIONS
    if (unlikely(__pyx_assertions_enabled())) {
      __pyx_t_8 = (__Pyx_PyBytes_Equals(__pyx_v_s, __pyx_v_b, Py_EQ)); if (unlikely((__pyx_t_8 < 0))) __PYX_ERR(0, 2248, __pyx_L1_error)
      if (unlikely(!__pyx_t_8)) {
        __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2248, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_9 = 0;
        __pyx_t_10 = 127;
        __pyx_t_4 = __Pyx_PyObject_FormatSimple(__pyx_v_s, __pyx_empty_unicode); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2248, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_10 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) > __pyx_t_10) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) : __pyx_t_10;
        __pyx_t_9 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_4);
        __Pyx_GIVEREF(__pyx_t_4);
        PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_4);
        __pyx_t_4 = 0;
        __Pyx_INCREF(__pyx_kp_u__15);
        __pyx_t_9 += 3;
        __Pyx_GIVEREF(__pyx_kp_u__15);
        PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_kp_u__15);
        __pyx_t_4 = __Pyx_PyObject_FormatSimple(__pyx_v_b, __pyx_empty_unicode); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2248, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_10 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) > __pyx_t_10) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) : __pyx_t_10;
        __pyx_t_9 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_4);
        __Pyx_GIVEREF(__pyx_t_4);
        PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_t_4);
        __pyx_t_4 = 0;
        __pyx_t_4 = __Pyx_PyUnicode_Join(__pyx_t_3, 3, __pyx_t_9, __pyx_t_10); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2248, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_Raise(__pyx_builtin_AssertionError, __pyx_t_4, 0, 0);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __PYX_ERR(0, 2248, __pyx_L1_error)
      }
    }
    #else
    if ((1)); else __PYX_ERR(0, 2248, __pyx_L1_error)
    #endif

    /* "serializor/binary/ser.py":2249
 *         b = pack_int64(val)  # type: ignore
 *         assert s == b, f"{s} | {b}"
 *         print("- pass: %s" % val, end="\r")             # <<<<<<<<<<<<<<
 *     print("Pass Pack int64".ljust(80))
 * 
 */
    __pyx_t_4 = __Pyx_PyUnicode_FormatSafe(__pyx_kp_u_pass_s, __pyx_v_val); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2249, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2249, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_4);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_4)) __PYX_ERR(0, 2249, __pyx_L1_error);
    __pyx_t_4 = 0;
    __pyx_t_4 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2249, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_end, __pyx_kp_u__16) < 0) __PYX_ERR(0, 2249, __pyx_L1_error)
    __pyx_t_5 = __Pyx_PyObject_Call(__pyx_builtin_print, __pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2249, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "serializor/binary/ser.py":2239
 *     import struct
 * 
 *     for val in (             # <<<<<<<<<<<<<<
 *         -9223372036854775808,
 *         -9223372036854775807,
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "serializor/binary/ser.py":2250
 *         assert s == b, f"{s} | {b}"
 *         print("- pass: %s" % val, end="\r")
 *     print("Pass Pack int64".ljust(80))             # <<<<<<<<<<<<<<
 * 
 *     del struct
 */
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_Pass_Pack_int64, __pyx_n_s_ljust); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2250, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_4 = NULL;
  __pyx_t_6 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
      __pyx_t_6 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_int_80};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2250, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __pyx_t_5 = __Pyx_PyObject_CallOneArg(__pyx_builtin_print, __pyx_t_1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2250, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "serializor/binary/ser.py":2252
 *     print("Pass Pack int64".ljust(80))
 * 
 *     del struct             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_DECREF(__pyx_v_struct); __pyx_v_struct = 0;

  /* "serializor/binary/ser.py":2236
 * 
 * 
 * def _test_pack_int64() -> None:             # <<<<<<<<<<<<<<
 *     import struct
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("serializor.binary.ser._test_pack_int64", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_struct);
  __Pyx_XDECREF(__pyx_v_val);
  __Pyx_XDECREF(__pyx_v_s);
  __Pyx_XDECREF(__pyx_v_b);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "serializor/binary/ser.py":2255
 * 
 * 
 * def _test_pack_uint64() -> None:             # <<<<<<<<<<<<<<
 *     import struct
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_10serializor_6binary_3ser_23_test_pack_uint64(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyMethodDef __pyx_mdef_10serializor_6binary_3ser_23_test_pack_uint64 = {"_test_pack_uint64", (PyCFunction)__pyx_pw_10serializor_6binary_3ser_23_test_pack_uint64, METH_NOARGS, 0};
static PyObject *__pyx_pw_10serializor_6binary_3ser_23_test_pack_uint64(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_test_pack_uint64 (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10serializor_6binary_3ser_22_test_pack_uint64(__pyx_self);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10serializor_6binary_3ser_22_test_pack_uint64(CYTHON_UNUSED PyObject *__pyx_self) {
  PyObject *__pyx_v_struct = NULL;
  PyObject *__pyx_v_val = NULL;
  PyObject *__pyx_v_s = NULL;
  PyObject *__pyx_v_b = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  unsigned int __pyx_t_6;
  unsigned PY_LONG_LONG __pyx_t_7;
  int __pyx_t_8;
  Py_ssize_t __pyx_t_9;
  Py_UCS4 __pyx_t_10;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_test_pack_uint64", 1);

  /* "serializor/binary/ser.py":2256
 * 
 * def _test_pack_uint64() -> None:
 *     import struct             # <<<<<<<<<<<<<<
 * 
 *     for val in (
 */
  __pyx_t_1 = __Pyx_ImportDottedModule(__pyx_n_s_struct, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2256, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_struct = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "serializor/binary/ser.py":2258
 *     import struct
 * 
 *     for val in (             # <<<<<<<<<<<<<<
 *         0,
 *         1,
 */
  __pyx_t_1 = __pyx_tuple__23; __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = 0;
  for (;;) {
    if (__pyx_t_2 >= 5) break;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_3); __pyx_t_2++; if (unlikely((0 < 0))) __PYX_ERR(0, 2258, __pyx_L1_error)
    #else
    __pyx_t_3 = __Pyx_PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2258, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    #endif
    __Pyx_XDECREF_SET(__pyx_v_val, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "serializor/binary/ser.py":2265
 *         18446744073709551615,
 *     ):
 *         s = struct.pack("<Q", val)             # <<<<<<<<<<<<<<
 *         b = pack_uint64(val)  # type: ignore
 *         assert s == b, f"{s} | {b}"
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_struct, __pyx_n_s_pack); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2265, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    __pyx_t_6 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
        __pyx_t_6 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[3] = {__pyx_t_5, __pyx_kp_u_Q, __pyx_v_val};
      __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_6, 2+__pyx_t_6);
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2265, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
    __Pyx_XDECREF_SET(__pyx_v_s, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "serializor/binary/ser.py":2266
 *     ):
 *         s = struct.pack("<Q", val)
 *         b = pack_uint64(val)  # type: ignore             # <<<<<<<<<<<<<<
 *         assert s == b, f"{s} | {b}"
 *         print("- pass: %s" % val, end="\r")
 */
    __pyx_t_7 = __Pyx_PyInt_As_unsigned_PY_LONG_LONG(__pyx_v_val); if (unlikely((__pyx_t_7 == (unsigned PY_LONG_LONG)-1) && PyErr_Occurred())) __PYX_ERR(0, 2266, __pyx_L1_error)
    __pyx_t_3 = __pyx_f_10serializor_6binary_3ser_pack_uint64(__pyx_t_7); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2266, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_XDECREF_SET(__pyx_v_b, ((PyObject*)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "serializor/binary/ser.py":2267
 *         s = struct.pack("<Q", val)
 *         b = pack_uint64(val)  # type: ignore
 *         assert s == b, f"{s} | {b}"             # <<<<<<<<<<<<<<
 *         print("- pass: %s" % val, end="\r")
 *     print("Pass Pack uint64".ljust(80))
 */
    #ifndef CYTHON_WITHOUT_ASSERTIONS
    if (unlikely(__pyx_assertions_enabled())) {
      __pyx_t_8 = (__Pyx_PyBytes_Equals(__pyx_v_s, __pyx_v_b, Py_EQ)); if (unlikely((__pyx_t_8 < 0))) __PYX_ERR(0, 2267, __pyx_L1_error)
      if (unlikely(!__pyx_t_8)) {
        __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2267, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_9 = 0;
        __pyx_t_10 = 127;
        __pyx_t_4 = __Pyx_PyObject_FormatSimple(__pyx_v_s, __pyx_empty_unicode); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2267, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_10 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) > __pyx_t_10) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) : __pyx_t_10;
        __pyx_t_9 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_4);
        __Pyx_GIVEREF(__pyx_t_4);
        PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_4);
        __pyx_t_4 = 0;
        __Pyx_INCREF(__pyx_kp_u__15);
        __pyx_t_9 += 3;
        __Pyx_GIVEREF(__pyx_kp_u__15);
        PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_kp_u__15);
        __pyx_t_4 = __Pyx_PyObject_FormatSimple(__pyx_v_b, __pyx_empty_unicode); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2267, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_10 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) > __pyx_t_10) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) : __pyx_t_10;
        __pyx_t_9 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_4);
        __Pyx_GIVEREF(__pyx_t_4);
        PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_t_4);
        __pyx_t_4 = 0;
        __pyx_t_4 = __Pyx_PyUnicode_Join(__pyx_t_3, 3, __pyx_t_9, __pyx_t_10); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2267, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_Raise(__pyx_builtin_AssertionError, __pyx_t_4, 0, 0);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __PYX_ERR(0, 2267, __pyx_L1_error)
      }
    }
    #else
    if ((1)); else __PYX_ERR(0, 2267, __pyx_L1_error)
    #endif

    /* "serializor/binary/ser.py":2268
 *         b = pack_uint64(val)  # type: ignore
 *         assert s == b, f"{s} | {b}"
 *         print("- pass: %s" % val, end="\r")             # <<<<<<<<<<<<<<
 *     print("Pass Pack uint64".ljust(80))
 * 
 */
    __pyx_t_4 = __Pyx_PyUnicode_FormatSafe(__pyx_kp_u_pass_s, __pyx_v_val); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2268, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2268, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_4);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_4)) __PYX_ERR(0, 2268, __pyx_L1_error);
    __pyx_t_4 = 0;
    __pyx_t_4 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2268, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_end, __pyx_kp_u__16) < 0) __PYX_ERR(0, 2268, __pyx_L1_error)
    __pyx_t_5 = __Pyx_PyObject_Call(__pyx_builtin_print, __pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2268, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "serializor/binary/ser.py":2258
 *     import struct
 * 
 *     for val in (             # <<<<<<<<<<<<<<
 *         0,
 *         1,
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "serializor/binary/ser.py":2269
 *         assert s == b, f"{s} | {b}"
 *         print("- pass: %s" % val, end="\r")
 *     print("Pass Pack uint64".ljust(80))             # <<<<<<<<<<<<<<
 * 
 *     del struct
 */
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_Pass_Pack_uint64, __pyx_n_s_ljust); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2269, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_4 = NULL;
  __pyx_t_6 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
      __pyx_t_6 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_int_80};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2269, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __pyx_t_5 = __Pyx_PyObject_CallOneArg(__pyx_builtin_print, __pyx_t_1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2269, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "serializor/binary/ser.py":2271
 *     print("Pass Pack uint64".ljust(80))
 * 
 *     del struct             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_DECREF(__pyx_v_struct); __pyx_v_struct = 0;

  /* "serializor/binary/ser.py":2255
 * 
 * 
 * def _test_pack_uint64() -> None:             # <<<<<<<<<<<<<<
 *     import struct
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("serializor.binary.ser._test_pack_uint64", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_struct);
  __Pyx_XDECREF(__pyx_v_val);
  __Pyx_XDECREF(__pyx_v_s);
  __Pyx_XDECREF(__pyx_v_b);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "serializor/binary/ser.py":2274
 * 
 * 
 * def _test_gen_encoded_int() -> None:             # <<<<<<<<<<<<<<
 *     # Test gen_encoded_int
 *     num = 251
 */

/* Python wrapper */
static PyObject *__pyx_pw_10serializor_6binary_3ser_25_test_gen_encoded_int(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyMethodDef __pyx_mdef_10serializor_6binary_3ser_25_test_gen_encoded_int = {"_test_gen_encoded_int", (PyCFunction)__pyx_pw_10serializor_6binary_3ser_25_test_gen_encoded_int, METH_NOARGS, 0};
static PyObject *__pyx_pw_10serializor_6binary_3ser_25_test_gen_encoded_int(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_test_gen_encoded_int (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10serializor_6binary_3ser_24_test_gen_encoded_int(__pyx_self);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10serializor_6binary_3ser_24_test_gen_encoded_int(CYTHON_UNUSED PyObject *__pyx_self) {
  PyObject *__pyx_v_num = NULL;
  PyObject *__pyx_v_v = NULL;
  PyObject *__pyx_v_n = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  unsigned PY_LONG_LONG __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  unsigned int __pyx_t_3;
  int __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  Py_UCS4 __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  Py_ssize_t __pyx_t_10;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_test_gen_encoded_int", 1);

  /* "serializor/binary/ser.py":2276
 * def _test_gen_encoded_int() -> None:
 *     # Test gen_encoded_int
 *     num = 251             # <<<<<<<<<<<<<<
 *     v = gen_encoded_int(num)  # type: ignore
 *     n = pack_uint8(num)  # type: ignore
 */
  __Pyx_INCREF(__pyx_int_251);
  __pyx_v_num = __pyx_int_251;

  /* "serializor/binary/ser.py":2277
 *     # Test gen_encoded_int
 *     num = 251
 *     v = gen_encoded_int(num)  # type: ignore             # <<<<<<<<<<<<<<
 *     n = pack_uint8(num)  # type: ignore
 *     assert v == n, f"{v} | {n} - num: {num}"
 */
  __pyx_t_1 = __Pyx_PyInt_As_unsigned_PY_LONG_LONG(__pyx_v_num); if (unlikely((__pyx_t_1 == (unsigned PY_LONG_LONG)-1) && PyErr_Occurred())) __PYX_ERR(0, 2277, __pyx_L1_error)
  __pyx_t_2 = __pyx_f_10serializor_6binary_3ser_gen_encoded_int(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2277, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_v = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "serializor/binary/ser.py":2278
 *     num = 251
 *     v = gen_encoded_int(num)  # type: ignore
 *     n = pack_uint8(num)  # type: ignore             # <<<<<<<<<<<<<<
 *     assert v == n, f"{v} | {n} - num: {num}"
 * 
 */
  __pyx_t_3 = __Pyx_PyInt_As_unsigned_int(__pyx_v_num); if (unlikely((__pyx_t_3 == (unsigned int)-1) && PyErr_Occurred())) __PYX_ERR(0, 2278, __pyx_L1_error)
  __pyx_t_2 = __pyx_f_10serializor_6binary_3ser_pack_uint8(__pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2278, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_n = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "serializor/binary/ser.py":2279
 *     v = gen_encoded_int(num)  # type: ignore
 *     n = pack_uint8(num)  # type: ignore
 *     assert v == n, f"{v} | {n} - num: {num}"             # <<<<<<<<<<<<<<
 * 
 *     for num in (252, 65_535):
 */
  #ifndef CYTHON_WITHOUT_ASSERTIONS
  if (unlikely(__pyx_assertions_enabled())) {
    __pyx_t_4 = (__Pyx_PyBytes_Equals(__pyx_v_v, __pyx_v_n, Py_EQ)); if (unlikely((__pyx_t_4 < 0))) __PYX_ERR(0, 2279, __pyx_L1_error)
    if (unlikely(!__pyx_t_4)) {
      __pyx_t_2 = PyTuple_New(5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2279, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_5 = 0;
      __pyx_t_6 = 127;
      __pyx_t_7 = __Pyx_PyObject_FormatSimple(__pyx_v_v, __pyx_empty_unicode); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2279, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_6 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_7) > __pyx_t_6) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_7) : __pyx_t_6;
      __pyx_t_5 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_7);
      __Pyx_GIVEREF(__pyx_t_7);
      PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_7);
      __pyx_t_7 = 0;
      __Pyx_INCREF(__pyx_kp_u__15);
      __pyx_t_5 += 3;
      __Pyx_GIVEREF(__pyx_kp_u__15);
      PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_kp_u__15);
      __pyx_t_7 = __Pyx_PyObject_FormatSimple(__pyx_v_n, __pyx_empty_unicode); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2279, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_6 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_7) > __pyx_t_6) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_7) : __pyx_t_6;
      __pyx_t_5 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_7);
      __Pyx_GIVEREF(__pyx_t_7);
      PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_t_7);
      __pyx_t_7 = 0;
      __Pyx_INCREF(__pyx_kp_u_num);
      __pyx_t_5 += 8;
      __Pyx_GIVEREF(__pyx_kp_u_num);
      PyTuple_SET_ITEM(__pyx_t_2, 3, __pyx_kp_u_num);
      __pyx_t_7 = __Pyx_PyObject_FormatSimple(__pyx_v_num, __pyx_empty_unicode); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2279, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_6 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_7) > __pyx_t_6) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_7) : __pyx_t_6;
      __pyx_t_5 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_7);
      __Pyx_GIVEREF(__pyx_t_7);
      PyTuple_SET_ITEM(__pyx_t_2, 4, __pyx_t_7);
      __pyx_t_7 = 0;
      __pyx_t_7 = __Pyx_PyUnicode_Join(__pyx_t_2, 5, __pyx_t_5, __pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2279, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_Raise(__pyx_builtin_AssertionError, __pyx_t_7, 0, 0);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __PYX_ERR(0, 2279, __pyx_L1_error)
    }
  }
  #else
  if ((1)); else __PYX_ERR(0, 2279, __pyx_L1_error)
  #endif

  /* "serializor/binary/ser.py":2281
 *     assert v == n, f"{v} | {n} - num: {num}"
 * 
 *     for num in (252, 65_535):             # <<<<<<<<<<<<<<
 *         v = gen_encoded_int(num)  # type: ignore
 *         n = pack_uint8(UINT16_ENCODE_VALUE) + pack_uint16(num)  # type: ignore
 */
  __pyx_t_7 = __pyx_tuple__24; __Pyx_INCREF(__pyx_t_7);
  __pyx_t_5 = 0;
  for (;;) {
    if (__pyx_t_5 >= 2) break;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_2 = PyTuple_GET_ITEM(__pyx_t_7, __pyx_t_5); __Pyx_INCREF(__pyx_t_2); __pyx_t_5++; if (unlikely((0 < 0))) __PYX_ERR(0, 2281, __pyx_L1_error)
    #else
    __pyx_t_2 = __Pyx_PySequence_ITEM(__pyx_t_7, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2281, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    #endif
    __Pyx_DECREF_SET(__pyx_v_num, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "serializor/binary/ser.py":2282
 * 
 *     for num in (252, 65_535):
 *         v = gen_encoded_int(num)  # type: ignore             # <<<<<<<<<<<<<<
 *         n = pack_uint8(UINT16_ENCODE_VALUE) + pack_uint16(num)  # type: ignore
 *         assert v == n, f"{v} | {n} - num: {num}"
 */
    __pyx_t_1 = __Pyx_PyInt_As_unsigned_PY_LONG_LONG(__pyx_v_num); if (unlikely((__pyx_t_1 == (unsigned PY_LONG_LONG)-1) && PyErr_Occurred())) __PYX_ERR(0, 2282, __pyx_L1_error)
    __pyx_t_2 = __pyx_f_10serializor_6binary_3ser_gen_encoded_int(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2282, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF_SET(__pyx_v_v, ((PyObject*)__pyx_t_2));
    __pyx_t_2 = 0;

    /* "serializor/binary/ser.py":2283
 *     for num in (252, 65_535):
 *         v = gen_encoded_int(num)  # type: ignore
 *         n = pack_uint8(UINT16_ENCODE_VALUE) + pack_uint16(num)  # type: ignore             # <<<<<<<<<<<<<<
 *         assert v == n, f"{v} | {n} - num: {num}"
 * 
 */
    __pyx_t_2 = __pyx_f_10serializor_6binary_3ser_pack_uint8(__pyx_v_10serializor_6binary_3ser_UINT16_ENCODE_VALUE); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2283, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyInt_As_unsigned_int(__pyx_v_num); if (unlikely((__pyx_t_3 == (unsigned int)-1) && PyErr_Occurred())) __PYX_ERR(0, 2283, __pyx_L1_error)
    __pyx_t_8 = __pyx_f_10serializor_6binary_3ser_pack_uint16(__pyx_t_3); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 2283, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_9 = PyNumber_Add(__pyx_t_2, __pyx_t_8); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 2283, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF_SET(__pyx_v_n, ((PyObject*)__pyx_t_9));
    __pyx_t_9 = 0;

    /* "serializor/binary/ser.py":2284
 *         v = gen_encoded_int(num)  # type: ignore
 *         n = pack_uint8(UINT16_ENCODE_VALUE) + pack_uint16(num)  # type: ignore
 *         assert v == n, f"{v} | {n} - num: {num}"             # <<<<<<<<<<<<<<
 * 
 *     for num in (65_536, 4_294_967_295):
 */
    #ifndef CYTHON_WITHOUT_ASSERTIONS
    if (unlikely(__pyx_assertions_enabled())) {
      __pyx_t_4 = (__Pyx_PyBytes_Equals(__pyx_v_v, __pyx_v_n, Py_EQ)); if (unlikely((__pyx_t_4 < 0))) __PYX_ERR(0, 2284, __pyx_L1_error)
      if (unlikely(!__pyx_t_4)) {
        __pyx_t_9 = PyTuple_New(5); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 2284, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        __pyx_t_10 = 0;
        __pyx_t_6 = 127;
        __pyx_t_8 = __Pyx_PyObject_FormatSimple(__pyx_v_v, __pyx_empty_unicode); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 2284, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        __pyx_t_6 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_8) > __pyx_t_6) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_8) : __pyx_t_6;
        __pyx_t_10 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_8);
        __Pyx_GIVEREF(__pyx_t_8);
        PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_8);
        __pyx_t_8 = 0;
        __Pyx_INCREF(__pyx_kp_u__15);
        __pyx_t_10 += 3;
        __Pyx_GIVEREF(__pyx_kp_u__15);
        PyTuple_SET_ITEM(__pyx_t_9, 1, __pyx_kp_u__15);
        __pyx_t_8 = __Pyx_PyObject_FormatSimple(__pyx_v_n, __pyx_empty_unicode); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 2284, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        __pyx_t_6 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_8) > __pyx_t_6) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_8) : __pyx_t_6;
        __pyx_t_10 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_8);
        __Pyx_GIVEREF(__pyx_t_8);
        PyTuple_SET_ITEM(__pyx_t_9, 2, __pyx_t_8);
        __pyx_t_8 = 0;
        __Pyx_INCREF(__pyx_kp_u_num);
        __pyx_t_10 += 8;
        __Pyx_GIVEREF(__pyx_kp_u_num);
        PyTuple_SET_ITEM(__pyx_t_9, 3, __pyx_kp_u_num);
        __pyx_t_8 = __Pyx_PyObject_FormatSimple(__pyx_v_num, __pyx_empty_unicode); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 2284, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        __pyx_t_6 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_8) > __pyx_t_6) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_8) : __pyx_t_6;
        __pyx_t_10 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_8);
        __Pyx_GIVEREF(__pyx_t_8);
        PyTuple_SET_ITEM(__pyx_t_9, 4, __pyx_t_8);
        __pyx_t_8 = 0;
        __pyx_t_8 = __Pyx_PyUnicode_Join(__pyx_t_9, 5, __pyx_t_10, __pyx_t_6); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 2284, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        __Pyx_Raise(__pyx_builtin_AssertionError, __pyx_t_8, 0, 0);
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        __PYX_ERR(0, 2284, __pyx_L1_error)
      }
    }
    #else
    if ((1)); else __PYX_ERR(0, 2284, __pyx_L1_error)
    #endif

    /* "serializor/binary/ser.py":2281
 *     assert v == n, f"{v} | {n} - num: {num}"
 * 
 *     for num in (252, 65_535):             # <<<<<<<<<<<<<<
 *         v = gen_encoded_int(num)  # type: ignore
 *         n = pack_uint8(UINT16_ENCODE_VALUE) + pack_uint16(num)  # type: ignore
 */
  }
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "serializor/binary/ser.py":2286
 *         assert v == n, f"{v} | {n} - num: {num}"
 * 
 *     for num in (65_536, 4_294_967_295):             # <<<<<<<<<<<<<<
 *         v = gen_encoded_int(num)  # type: ignore
 *         n = pack_uint8(UINT32_ENCODE_VALUE) + pack_uint32(num)  # type: ignore
 */
  __pyx_t_7 = __pyx_tuple__25; __Pyx_INCREF(__pyx_t_7);
  __pyx_t_5 = 0;
  for (;;) {
    if (__pyx_t_5 >= 2) break;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_8 = PyTuple_GET_ITEM(__pyx_t_7, __pyx_t_5); __Pyx_INCREF(__pyx_t_8); __pyx_t_5++; if (unlikely((0 < 0))) __PYX_ERR(0, 2286, __pyx_L1_error)
    #else
    __pyx_t_8 = __Pyx_PySequence_ITEM(__pyx_t_7, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 2286, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    #endif
    __Pyx_DECREF_SET(__pyx_v_num, __pyx_t_8);
    __pyx_t_8 = 0;

    /* "serializor/binary/ser.py":2287
 * 
 *     for num in (65_536, 4_294_967_295):
 *         v = gen_encoded_int(num)  # type: ignore             # <<<<<<<<<<<<<<
 *         n = pack_uint8(UINT32_ENCODE_VALUE) + pack_uint32(num)  # type: ignore
 *         assert v == n, f"{v} | {n} - num: {num}"
 */
    __pyx_t_1 = __Pyx_PyInt_As_unsigned_PY_LONG_LONG(__pyx_v_num); if (unlikely((__pyx_t_1 == (unsigned PY_LONG_LONG)-1) && PyErr_Occurred())) __PYX_ERR(0, 2287, __pyx_L1_error)
    __pyx_t_8 = __pyx_f_10serializor_6binary_3ser_gen_encoded_int(__pyx_t_1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 2287, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF_SET(__pyx_v_v, ((PyObject*)__pyx_t_8));
    __pyx_t_8 = 0;

    /* "serializor/binary/ser.py":2288
 *     for num in (65_536, 4_294_967_295):
 *         v = gen_encoded_int(num)  # type: ignore
 *         n = pack_uint8(UINT32_ENCODE_VALUE) + pack_uint32(num)  # type: ignore             # <<<<<<<<<<<<<<
 *         assert v == n, f"{v} | {n} - num: {num}"
 * 
 */
    __pyx_t_8 = __pyx_f_10serializor_6binary_3ser_pack_uint8(__pyx_v_10serializor_6binary_3ser_UINT32_ENCODE_VALUE); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 2288, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_1 = __Pyx_PyInt_As_unsigned_PY_LONG_LONG(__pyx_v_num); if (unlikely((__pyx_t_1 == (unsigned PY_LONG_LONG)-1) && PyErr_Occurred())) __PYX_ERR(0, 2288, __pyx_L1_error)
    __pyx_t_9 = __pyx_f_10serializor_6binary_3ser_pack_uint32(__pyx_t_1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 2288, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_2 = PyNumber_Add(__pyx_t_8, __pyx_t_9); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2288, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF_SET(__pyx_v_n, ((PyObject*)__pyx_t_2));
    __pyx_t_2 = 0;

    /* "serializor/binary/ser.py":2289
 *         v = gen_encoded_int(num)  # type: ignore
 *         n = pack_uint8(UINT32_ENCODE_VALUE) + pack_uint32(num)  # type: ignore
 *         assert v == n, f"{v} | {n} - num: {num}"             # <<<<<<<<<<<<<<
 * 
 *     for num in (4_294_967_296, 4_294_967_296 * 2):
 */
    #ifndef CYTHON_WITHOUT_ASSERTIONS
    if (unlikely(__pyx_assertions_enabled())) {
      __pyx_t_4 = (__Pyx_PyBytes_Equals(__pyx_v_v, __pyx_v_n, Py_EQ)); if (unlikely((__pyx_t_4 < 0))) __PYX_ERR(0, 2289, __pyx_L1_error)
      if (unlikely(!__pyx_t_4)) {
        __pyx_t_2 = PyTuple_New(5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2289, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_10 = 0;
        __pyx_t_6 = 127;
        __pyx_t_9 = __Pyx_PyObject_FormatSimple(__pyx_v_v, __pyx_empty_unicode); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 2289, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        __pyx_t_6 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_9) > __pyx_t_6) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_9) : __pyx_t_6;
        __pyx_t_10 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_9);
        __Pyx_GIVEREF(__pyx_t_9);
        PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_9);
        __pyx_t_9 = 0;
        __Pyx_INCREF(__pyx_kp_u__15);
        __pyx_t_10 += 3;
        __Pyx_GIVEREF(__pyx_kp_u__15);
        PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_kp_u__15);
        __pyx_t_9 = __Pyx_PyObject_FormatSimple(__pyx_v_n, __pyx_empty_unicode); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 2289, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        __pyx_t_6 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_9) > __pyx_t_6) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_9) : __pyx_t_6;
        __pyx_t_10 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_9);
        __Pyx_GIVEREF(__pyx_t_9);
        PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_t_9);
        __pyx_t_9 = 0;
        __Pyx_INCREF(__pyx_kp_u_num);
        __pyx_t_10 += 8;
        __Pyx_GIVEREF(__pyx_kp_u_num);
        PyTuple_SET_ITEM(__pyx_t_2, 3, __pyx_kp_u_num);
        __pyx_t_9 = __Pyx_PyObject_FormatSimple(__pyx_v_num, __pyx_empty_unicode); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 2289, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        __pyx_t_6 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_9) > __pyx_t_6) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_9) : __pyx_t_6;
        __pyx_t_10 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_9);
        __Pyx_GIVEREF(__pyx_t_9);
        PyTuple_SET_ITEM(__pyx_t_2, 4, __pyx_t_9);
        __pyx_t_9 = 0;
        __pyx_t_9 = __Pyx_PyUnicode_Join(__pyx_t_2, 5, __pyx_t_10, __pyx_t_6); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 2289, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_Raise(__pyx_builtin_AssertionError, __pyx_t_9, 0, 0);
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        __PYX_ERR(0, 2289, __pyx_L1_error)
      }
    }
    #else
    if ((1)); else __PYX_ERR(0, 2289, __pyx_L1_error)
    #endif

    /* "serializor/binary/ser.py":2286
 *         assert v == n, f"{v} | {n} - num: {num}"
 * 
 *     for num in (65_536, 4_294_967_295):             # <<<<<<<<<<<<<<
 *         v = gen_encoded_int(num)  # type: ignore
 *         n = pack_uint8(UINT32_ENCODE_VALUE) + pack_uint32(num)  # type: ignore
 */
  }
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "serializor/binary/ser.py":2291
 *         assert v == n, f"{v} | {n} - num: {num}"
 * 
 *     for num in (4_294_967_296, 4_294_967_296 * 2):             # <<<<<<<<<<<<<<
 *         v = gen_encoded_int(num)  # type: ignore
 *         n = pack_uint8(UINT64_ENCODE_VALUE) + pack_uint64(num)  # type: ignore
 */
  __pyx_t_7 = __pyx_tuple__26; __Pyx_INCREF(__pyx_t_7);
  __pyx_t_5 = 0;
  for (;;) {
    if (__pyx_t_5 >= 2) break;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_9 = PyTuple_GET_ITEM(__pyx_t_7, __pyx_t_5); __Pyx_INCREF(__pyx_t_9); __pyx_t_5++; if (unlikely((0 < 0))) __PYX_ERR(0, 2291, __pyx_L1_error)
    #else
    __pyx_t_9 = __Pyx_PySequence_ITEM(__pyx_t_7, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 2291, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    #endif
    __Pyx_DECREF_SET(__pyx_v_num, __pyx_t_9);
    __pyx_t_9 = 0;

    /* "serializor/binary/ser.py":2292
 * 
 *     for num in (4_294_967_296, 4_294_967_296 * 2):
 *         v = gen_encoded_int(num)  # type: ignore             # <<<<<<<<<<<<<<
 *         n = pack_uint8(UINT64_ENCODE_VALUE) + pack_uint64(num)  # type: ignore
 *         assert v == n, f"{v} | {n} - num: {num}"
 */
    __pyx_t_1 = __Pyx_PyInt_As_unsigned_PY_LONG_LONG(__pyx_v_num); if (unlikely((__pyx_t_1 == (unsigned PY_LONG_LONG)-1) && PyErr_Occurred())) __PYX_ERR(0, 2292, __pyx_L1_error)
    __pyx_t_9 = __pyx_f_10serializor_6binary_3ser_gen_encoded_int(__pyx_t_1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 2292, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF_SET(__pyx_v_v, ((PyObject*)__pyx_t_9));
    __pyx_t_9 = 0;

    /* "serializor/binary/ser.py":2293
 *     for num in (4_294_967_296, 4_294_967_296 * 2):
 *         v = gen_encoded_int(num)  # type: ignore
 *         n = pack_uint8(UINT64_ENCODE_VALUE) + pack_uint64(num)  # type: ignore             # <<<<<<<<<<<<<<
 *         assert v == n, f"{v} | {n} - num: {num}"
 *     print("Pass Gen Encoded Integer".ljust(80))
 */
    __pyx_t_9 = __pyx_f_10serializor_6binary_3ser_pack_uint8(__pyx_v_10serializor_6binary_3ser_UINT64_ENCODE_VALUE); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 2293, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_1 = __Pyx_PyInt_As_unsigned_PY_LONG_LONG(__pyx_v_num); if (unlikely((__pyx_t_1 == (unsigned PY_LONG_LONG)-1) && PyErr_Occurred())) __PYX_ERR(0, 2293, __pyx_L1_error)
    __pyx_t_2 = __pyx_f_10serializor_6binary_3ser_pack_uint64(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2293, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_8 = PyNumber_Add(__pyx_t_9, __pyx_t_2); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 2293, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF_SET(__pyx_v_n, ((PyObject*)__pyx_t_8));
    __pyx_t_8 = 0;

    /* "serializor/binary/ser.py":2294
 *         v = gen_encoded_int(num)  # type: ignore
 *         n = pack_uint8(UINT64_ENCODE_VALUE) + pack_uint64(num)  # type: ignore
 *         assert v == n, f"{v} | {n} - num: {num}"             # <<<<<<<<<<<<<<
 *     print("Pass Gen Encoded Integer".ljust(80))
 * 
 */
    #ifndef CYTHON_WITHOUT_ASSERTIONS
    if (unlikely(__pyx_assertions_enabled())) {
      __pyx_t_4 = (__Pyx_PyBytes_Equals(__pyx_v_v, __pyx_v_n, Py_EQ)); if (unlikely((__pyx_t_4 < 0))) __PYX_ERR(0, 2294, __pyx_L1_error)
      if (unlikely(!__pyx_t_4)) {
        __pyx_t_8 = PyTuple_New(5); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 2294, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        __pyx_t_10 = 0;
        __pyx_t_6 = 127;
        __pyx_t_2 = __Pyx_PyObject_FormatSimple(__pyx_v_v, __pyx_empty_unicode); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2294, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_6 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) > __pyx_t_6) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) : __pyx_t_6;
        __pyx_t_10 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2);
        __Pyx_GIVEREF(__pyx_t_2);
        PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_2);
        __pyx_t_2 = 0;
        __Pyx_INCREF(__pyx_kp_u__15);
        __pyx_t_10 += 3;
        __Pyx_GIVEREF(__pyx_kp_u__15);
        PyTuple_SET_ITEM(__pyx_t_8, 1, __pyx_kp_u__15);
        __pyx_t_2 = __Pyx_PyObject_FormatSimple(__pyx_v_n, __pyx_empty_unicode); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2294, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_6 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) > __pyx_t_6) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) : __pyx_t_6;
        __pyx_t_10 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2);
        __Pyx_GIVEREF(__pyx_t_2);
        PyTuple_SET_ITEM(__pyx_t_8, 2, __pyx_t_2);
        __pyx_t_2 = 0;
        __Pyx_INCREF(__pyx_kp_u_num);
        __pyx_t_10 += 8;
        __Pyx_GIVEREF(__pyx_kp_u_num);
        PyTuple_SET_ITEM(__pyx_t_8, 3, __pyx_kp_u_num);
        __pyx_t_2 = __Pyx_PyObject_FormatSimple(__pyx_v_num, __pyx_empty_unicode); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2294, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_6 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) > __pyx_t_6) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) : __pyx_t_6;
        __pyx_t_10 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2);
        __Pyx_GIVEREF(__pyx_t_2);
        PyTuple_SET_ITEM(__pyx_t_8, 4, __pyx_t_2);
        __pyx_t_2 = 0;
        __pyx_t_2 = __Pyx_PyUnicode_Join(__pyx_t_8, 5, __pyx_t_10, __pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2294, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_Raise(__pyx_builtin_AssertionError, __pyx_t_2, 0, 0);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __PYX_ERR(0, 2294, __pyx_L1_error)
      }
    }
    #else
    if ((1)); else __PYX_ERR(0, 2294, __pyx_L1_error)
    #endif

    /* "serializor/binary/ser.py":2291
 *         assert v == n, f"{v} | {n} - num: {num}"
 * 
 *     for num in (4_294_967_296, 4_294_967_296 * 2):             # <<<<<<<<<<<<<<
 *         v = gen_encoded_int(num)  # type: ignore
 *         n = pack_uint8(UINT64_ENCODE_VALUE) + pack_uint64(num)  # type: ignore
 */
  }
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "serializor/binary/ser.py":2295
 *         n = pack_uint8(UINT64_ENCODE_VALUE) + pack_uint64(num)  # type: ignore
 *         assert v == n, f"{v} | {n} - num: {num}"
 *     print("Pass Gen Encoded Integer".ljust(80))             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_Pass_Gen_Encoded_Integer, __pyx_n_s_ljust); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2295, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_8 = NULL;
  __pyx_t_3 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_8)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_8);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_3 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_8, __pyx_int_80};
    __pyx_t_7 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_3, 1+__pyx_t_3);
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2295, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_print, __pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2295, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "serializor/binary/ser.py":2274
 * 
 * 
 * def _test_gen_encoded_int() -> None:             # <<<<<<<<<<<<<<
 *     # Test gen_encoded_int
 *     num = 251
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("serializor.binary.ser._test_gen_encoded_int", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_num);
  __Pyx_XDECREF(__pyx_v_v);
  __Pyx_XDECREF(__pyx_v_n);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "serializor/binary/ser.py":2298
 * 
 * 
 * def _test_gen_header() -> None:             # <<<<<<<<<<<<<<
 *     dtype: bytes = b"\x01"
 *     dchar: cython.uchar = 1
 */

/* Python wrapper */
static PyObject *__pyx_pw_10serializor_6binary_3ser_27_test_gen_header(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyMethodDef __pyx_mdef_10serializor_6binary_3ser_27_test_gen_header = {"_test_gen_header", (PyCFunction)__pyx_pw_10serializor_6binary_3ser_27_test_gen_header, METH_NOARGS, 0};
static PyObject *__pyx_pw_10serializor_6binary_3ser_27_test_gen_header(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_test_gen_header (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10serializor_6binary_3ser_26_test_gen_header(__pyx_self);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10serializor_6binary_3ser_26_test_gen_header(CYTHON_UNUSED PyObject *__pyx_self) {
  PyObject *__pyx_v_dtype = 0;
  unsigned char __pyx_v_dchar;
  PyObject *__pyx_v_num = NULL;
  PyObject *__pyx_v_v = NULL;
  PyObject *__pyx_v_n = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  unsigned PY_LONG_LONG __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  Py_UCS4 __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  Py_ssize_t __pyx_t_8;
  unsigned int __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_test_gen_header", 1);

  /* "serializor/binary/ser.py":2299
 * 
 * def _test_gen_header() -> None:
 *     dtype: bytes = b"\x01"             # <<<<<<<<<<<<<<
 *     dchar: cython.uchar = 1
 *     # Test gen_header
 */
  __Pyx_INCREF(__pyx_kp_b__5);
  __pyx_v_dtype = __pyx_kp_b__5;

  /* "serializor/binary/ser.py":2300
 * def _test_gen_header() -> None:
 *     dtype: bytes = b"\x01"
 *     dchar: cython.uchar = 1             # <<<<<<<<<<<<<<
 *     # Test gen_header
 *     num = 251
 */
  __pyx_v_dchar = 1;

  /* "serializor/binary/ser.py":2302
 *     dchar: cython.uchar = 1
 *     # Test gen_header
 *     num = 251             # <<<<<<<<<<<<<<
 *     v = dtype + gen_encoded_int(num)  # type: ignore
 *     n = gen_header(dchar, num)  # type: ignore
 */
  __Pyx_INCREF(__pyx_int_251);
  __pyx_v_num = __pyx_int_251;

  /* "serializor/binary/ser.py":2303
 *     # Test gen_header
 *     num = 251
 *     v = dtype + gen_encoded_int(num)  # type: ignore             # <<<<<<<<<<<<<<
 *     n = gen_header(dchar, num)  # type: ignore
 *     assert v == n, f"{v} | {n} - num: {num}"
 */
  __pyx_t_1 = __Pyx_PyInt_As_unsigned_PY_LONG_LONG(__pyx_v_num); if (unlikely((__pyx_t_1 == (unsigned PY_LONG_LONG)-1) && PyErr_Occurred())) __PYX_ERR(0, 2303, __pyx_L1_error)
  __pyx_t_2 = __pyx_f_10serializor_6binary_3ser_gen_encoded_int(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2303, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyNumber_Add(__pyx_v_dtype, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2303, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_v = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "serializor/binary/ser.py":2304
 *     num = 251
 *     v = dtype + gen_encoded_int(num)  # type: ignore
 *     n = gen_header(dchar, num)  # type: ignore             # <<<<<<<<<<<<<<
 *     assert v == n, f"{v} | {n} - num: {num}"
 * 
 */
  __pyx_t_1 = __Pyx_PyInt_As_unsigned_PY_LONG_LONG(__pyx_v_num); if (unlikely((__pyx_t_1 == (unsigned PY_LONG_LONG)-1) && PyErr_Occurred())) __PYX_ERR(0, 2304, __pyx_L1_error)
  __pyx_t_3 = __pyx_f_10serializor_6binary_3ser_gen_header(__pyx_v_dchar, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2304, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_v_n = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "serializor/binary/ser.py":2305
 *     v = dtype + gen_encoded_int(num)  # type: ignore
 *     n = gen_header(dchar, num)  # type: ignore
 *     assert v == n, f"{v} | {n} - num: {num}"             # <<<<<<<<<<<<<<
 * 
 *     for num in (252, 65_535):
 */
  #ifndef CYTHON_WITHOUT_ASSERTIONS
  if (unlikely(__pyx_assertions_enabled())) {
    __pyx_t_4 = (__Pyx_PyBytes_Equals(__pyx_v_v, __pyx_v_n, Py_EQ)); if (unlikely((__pyx_t_4 < 0))) __PYX_ERR(0, 2305, __pyx_L1_error)
    if (unlikely(!__pyx_t_4)) {
      __pyx_t_3 = PyTuple_New(5); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2305, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_5 = 0;
      __pyx_t_6 = 127;
      __pyx_t_2 = __Pyx_PyObject_FormatSimple(__pyx_v_v, __pyx_empty_unicode); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2305, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_6 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) > __pyx_t_6) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) : __pyx_t_6;
      __pyx_t_5 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2);
      __Pyx_GIVEREF(__pyx_t_2);
      PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
      __pyx_t_2 = 0;
      __Pyx_INCREF(__pyx_kp_u__15);
      __pyx_t_5 += 3;
      __Pyx_GIVEREF(__pyx_kp_u__15);
      PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_kp_u__15);
      __pyx_t_2 = __Pyx_PyObject_FormatSimple(__pyx_v_n, __pyx_empty_unicode); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2305, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_6 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) > __pyx_t_6) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) : __pyx_t_6;
      __pyx_t_5 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2);
      __Pyx_GIVEREF(__pyx_t_2);
      PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_t_2);
      __pyx_t_2 = 0;
      __Pyx_INCREF(__pyx_kp_u_num);
      __pyx_t_5 += 8;
      __Pyx_GIVEREF(__pyx_kp_u_num);
      PyTuple_SET_ITEM(__pyx_t_3, 3, __pyx_kp_u_num);
      __pyx_t_2 = __Pyx_PyObject_FormatSimple(__pyx_v_num, __pyx_empty_unicode); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2305, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_6 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) > __pyx_t_6) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) : __pyx_t_6;
      __pyx_t_5 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2);
      __Pyx_GIVEREF(__pyx_t_2);
      PyTuple_SET_ITEM(__pyx_t_3, 4, __pyx_t_2);
      __pyx_t_2 = 0;
      __pyx_t_2 = __Pyx_PyUnicode_Join(__pyx_t_3, 5, __pyx_t_5, __pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2305, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_Raise(__pyx_builtin_AssertionError, __pyx_t_2, 0, 0);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __PYX_ERR(0, 2305, __pyx_L1_error)
    }
  }
  #else
  if ((1)); else __PYX_ERR(0, 2305, __pyx_L1_error)
  #endif

  /* "serializor/binary/ser.py":2307
 *     assert v == n, f"{v} | {n} - num: {num}"
 * 
 *     for num in (252, 65_535):             # <<<<<<<<<<<<<<
 *         v = dtype + gen_encoded_int(num)  # type: ignore
 *         n = gen_header(dchar, num)  # type: ignore
 */
  __pyx_t_2 = __pyx_tuple__24; __Pyx_INCREF(__pyx_t_2);
  __pyx_t_5 = 0;
  for (;;) {
    if (__pyx_t_5 >= 2) break;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_5); __Pyx_INCREF(__pyx_t_3); __pyx_t_5++; if (unlikely((0 < 0))) __PYX_ERR(0, 2307, __pyx_L1_error)
    #else
    __pyx_t_3 = __Pyx_PySequence_ITEM(__pyx_t_2, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2307, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    #endif
    __Pyx_DECREF_SET(__pyx_v_num, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "serializor/binary/ser.py":2308
 * 
 *     for num in (252, 65_535):
 *         v = dtype + gen_encoded_int(num)  # type: ignore             # <<<<<<<<<<<<<<
 *         n = gen_header(dchar, num)  # type: ignore
 *         assert v == n, f"{v} | {n} - num: {num}"
 */
    __pyx_t_1 = __Pyx_PyInt_As_unsigned_PY_LONG_LONG(__pyx_v_num); if (unlikely((__pyx_t_1 == (unsigned PY_LONG_LONG)-1) && PyErr_Occurred())) __PYX_ERR(0, 2308, __pyx_L1_error)
    __pyx_t_3 = __pyx_f_10serializor_6binary_3ser_gen_encoded_int(__pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2308, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_7 = PyNumber_Add(__pyx_v_dtype, __pyx_t_3); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2308, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF_SET(__pyx_v_v, ((PyObject*)__pyx_t_7));
    __pyx_t_7 = 0;

    /* "serializor/binary/ser.py":2309
 *     for num in (252, 65_535):
 *         v = dtype + gen_encoded_int(num)  # type: ignore
 *         n = gen_header(dchar, num)  # type: ignore             # <<<<<<<<<<<<<<
 *         assert v == n, f"{v} | {n} - num: {num}"
 * 
 */
    __pyx_t_1 = __Pyx_PyInt_As_unsigned_PY_LONG_LONG(__pyx_v_num); if (unlikely((__pyx_t_1 == (unsigned PY_LONG_LONG)-1) && PyErr_Occurred())) __PYX_ERR(0, 2309, __pyx_L1_error)
    __pyx_t_7 = __pyx_f_10serializor_6binary_3ser_gen_header(__pyx_v_dchar, __pyx_t_1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2309, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF_SET(__pyx_v_n, ((PyObject*)__pyx_t_7));
    __pyx_t_7 = 0;

    /* "serializor/binary/ser.py":2310
 *         v = dtype + gen_encoded_int(num)  # type: ignore
 *         n = gen_header(dchar, num)  # type: ignore
 *         assert v == n, f"{v} | {n} - num: {num}"             # <<<<<<<<<<<<<<
 * 
 *     for num in (65_536, 4_294_967_295):
 */
    #ifndef CYTHON_WITHOUT_ASSERTIONS
    if (unlikely(__pyx_assertions_enabled())) {
      __pyx_t_4 = (__Pyx_PyBytes_Equals(__pyx_v_v, __pyx_v_n, Py_EQ)); if (unlikely((__pyx_t_4 < 0))) __PYX_ERR(0, 2310, __pyx_L1_error)
      if (unlikely(!__pyx_t_4)) {
        __pyx_t_7 = PyTuple_New(5); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2310, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_8 = 0;
        __pyx_t_6 = 127;
        __pyx_t_3 = __Pyx_PyObject_FormatSimple(__pyx_v_v, __pyx_empty_unicode); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2310, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_6 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_3) > __pyx_t_6) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_3) : __pyx_t_6;
        __pyx_t_8 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_3);
        __Pyx_GIVEREF(__pyx_t_3);
        PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_3);
        __pyx_t_3 = 0;
        __Pyx_INCREF(__pyx_kp_u__15);
        __pyx_t_8 += 3;
        __Pyx_GIVEREF(__pyx_kp_u__15);
        PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_kp_u__15);
        __pyx_t_3 = __Pyx_PyObject_FormatSimple(__pyx_v_n, __pyx_empty_unicode); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2310, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_6 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_3) > __pyx_t_6) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_3) : __pyx_t_6;
        __pyx_t_8 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_3);
        __Pyx_GIVEREF(__pyx_t_3);
        PyTuple_SET_ITEM(__pyx_t_7, 2, __pyx_t_3);
        __pyx_t_3 = 0;
        __Pyx_INCREF(__pyx_kp_u_num);
        __pyx_t_8 += 8;
        __Pyx_GIVEREF(__pyx_kp_u_num);
        PyTuple_SET_ITEM(__pyx_t_7, 3, __pyx_kp_u_num);
        __pyx_t_3 = __Pyx_PyObject_FormatSimple(__pyx_v_num, __pyx_empty_unicode); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2310, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_6 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_3) > __pyx_t_6) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_3) : __pyx_t_6;
        __pyx_t_8 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_3);
        __Pyx_GIVEREF(__pyx_t_3);
        PyTuple_SET_ITEM(__pyx_t_7, 4, __pyx_t_3);
        __pyx_t_3 = 0;
        __pyx_t_3 = __Pyx_PyUnicode_Join(__pyx_t_7, 5, __pyx_t_8, __pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2310, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_Raise(__pyx_builtin_AssertionError, __pyx_t_3, 0, 0);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __PYX_ERR(0, 2310, __pyx_L1_error)
      }
    }
    #else
    if ((1)); else __PYX_ERR(0, 2310, __pyx_L1_error)
    #endif

    /* "serializor/binary/ser.py":2307
 *     assert v == n, f"{v} | {n} - num: {num}"
 * 
 *     for num in (252, 65_535):             # <<<<<<<<<<<<<<
 *         v = dtype + gen_encoded_int(num)  # type: ignore
 *         n = gen_header(dchar, num)  # type: ignore
 */
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "serializor/binary/ser.py":2312
 *         assert v == n, f"{v} | {n} - num: {num}"
 * 
 *     for num in (65_536, 4_294_967_295):             # <<<<<<<<<<<<<<
 *         v = dtype + gen_encoded_int(num)  # type: ignore
 *         n = gen_header(dchar, num)  # type: ignore
 */
  __pyx_t_2 = __pyx_tuple__25; __Pyx_INCREF(__pyx_t_2);
  __pyx_t_5 = 0;
  for (;;) {
    if (__pyx_t_5 >= 2) break;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_5); __Pyx_INCREF(__pyx_t_3); __pyx_t_5++; if (unlikely((0 < 0))) __PYX_ERR(0, 2312, __pyx_L1_error)
    #else
    __pyx_t_3 = __Pyx_PySequence_ITEM(__pyx_t_2, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2312, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    #endif
    __Pyx_DECREF_SET(__pyx_v_num, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "serializor/binary/ser.py":2313
 * 
 *     for num in (65_536, 4_294_967_295):
 *         v = dtype + gen_encoded_int(num)  # type: ignore             # <<<<<<<<<<<<<<
 *         n = gen_header(dchar, num)  # type: ignore
 *         assert v == n, f"{v} | {n} - num: {num}"
 */
    __pyx_t_1 = __Pyx_PyInt_As_unsigned_PY_LONG_LONG(__pyx_v_num); if (unlikely((__pyx_t_1 == (unsigned PY_LONG_LONG)-1) && PyErr_Occurred())) __PYX_ERR(0, 2313, __pyx_L1_error)
    __pyx_t_3 = __pyx_f_10serializor_6binary_3ser_gen_encoded_int(__pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2313, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_7 = PyNumber_Add(__pyx_v_dtype, __pyx_t_3); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2313, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF_SET(__pyx_v_v, ((PyObject*)__pyx_t_7));
    __pyx_t_7 = 0;

    /* "serializor/binary/ser.py":2314
 *     for num in (65_536, 4_294_967_295):
 *         v = dtype + gen_encoded_int(num)  # type: ignore
 *         n = gen_header(dchar, num)  # type: ignore             # <<<<<<<<<<<<<<
 *         assert v == n, f"{v} | {n} - num: {num}"
 * 
 */
    __pyx_t_1 = __Pyx_PyInt_As_unsigned_PY_LONG_LONG(__pyx_v_num); if (unlikely((__pyx_t_1 == (unsigned PY_LONG_LONG)-1) && PyErr_Occurred())) __PYX_ERR(0, 2314, __pyx_L1_error)
    __pyx_t_7 = __pyx_f_10serializor_6binary_3ser_gen_header(__pyx_v_dchar, __pyx_t_1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2314, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF_SET(__pyx_v_n, ((PyObject*)__pyx_t_7));
    __pyx_t_7 = 0;

    /* "serializor/binary/ser.py":2315
 *         v = dtype + gen_encoded_int(num)  # type: ignore
 *         n = gen_header(dchar, num)  # type: ignore
 *         assert v == n, f"{v} | {n} - num: {num}"             # <<<<<<<<<<<<<<
 * 
 *     for num in (4_294_967_296, 4_294_967_296 * 2):
 */
    #ifndef CYTHON_WITHOUT_ASSERTIONS
    if (unlikely(__pyx_assertions_enabled())) {
      __pyx_t_4 = (__Pyx_PyBytes_Equals(__pyx_v_v, __pyx_v_n, Py_EQ)); if (unlikely((__pyx_t_4 < 0))) __PYX_ERR(0, 2315, __pyx_L1_error)
      if (unlikely(!__pyx_t_4)) {
        __pyx_t_7 = PyTuple_New(5); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2315, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_8 = 0;
        __pyx_t_6 = 127;
        __pyx_t_3 = __Pyx_PyObject_FormatSimple(__pyx_v_v, __pyx_empty_unicode); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2315, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_6 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_3) > __pyx_t_6) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_3) : __pyx_t_6;
        __pyx_t_8 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_3);
        __Pyx_GIVEREF(__pyx_t_3);
        PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_3);
        __pyx_t_3 = 0;
        __Pyx_INCREF(__pyx_kp_u__15);
        __pyx_t_8 += 3;
        __Pyx_GIVEREF(__pyx_kp_u__15);
        PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_kp_u__15);
        __pyx_t_3 = __Pyx_PyObject_FormatSimple(__pyx_v_n, __pyx_empty_unicode); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2315, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_6 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_3) > __pyx_t_6) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_3) : __pyx_t_6;
        __pyx_t_8 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_3);
        __Pyx_GIVEREF(__pyx_t_3);
        PyTuple_SET_ITEM(__pyx_t_7, 2, __pyx_t_3);
        __pyx_t_3 = 0;
        __Pyx_INCREF(__pyx_kp_u_num);
        __pyx_t_8 += 8;
        __Pyx_GIVEREF(__pyx_kp_u_num);
        PyTuple_SET_ITEM(__pyx_t_7, 3, __pyx_kp_u_num);
        __pyx_t_3 = __Pyx_PyObject_FormatSimple(__pyx_v_num, __pyx_empty_unicode); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2315, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_6 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_3) > __pyx_t_6) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_3) : __pyx_t_6;
        __pyx_t_8 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_3);
        __Pyx_GIVEREF(__pyx_t_3);
        PyTuple_SET_ITEM(__pyx_t_7, 4, __pyx_t_3);
        __pyx_t_3 = 0;
        __pyx_t_3 = __Pyx_PyUnicode_Join(__pyx_t_7, 5, __pyx_t_8, __pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2315, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_Raise(__pyx_builtin_AssertionError, __pyx_t_3, 0, 0);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __PYX_ERR(0, 2315, __pyx_L1_error)
      }
    }
    #else
    if ((1)); else __PYX_ERR(0, 2315, __pyx_L1_error)
    #endif

    /* "serializor/binary/ser.py":2312
 *         assert v == n, f"{v} | {n} - num: {num}"
 * 
 *     for num in (65_536, 4_294_967_295):             # <<<<<<<<<<<<<<
 *         v = dtype + gen_encoded_int(num)  # type: ignore
 *         n = gen_header(dchar, num)  # type: ignore
 */
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "serializor/binary/ser.py":2317
 *         assert v == n, f"{v} | {n} - num: {num}"
 * 
 *     for num in (4_294_967_296, 4_294_967_296 * 2):             # <<<<<<<<<<<<<<
 *         v = dtype + gen_encoded_int(num)  # type: ignore
 *         n = gen_header(dchar, num)  # type: ignore
 */
  __pyx_t_2 = __pyx_tuple__26; __Pyx_INCREF(__pyx_t_2);
  __pyx_t_5 = 0;
  for (;;) {
    if (__pyx_t_5 >= 2) break;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_5); __Pyx_INCREF(__pyx_t_3); __pyx_t_5++; if (unlikely((0 < 0))) __PYX_ERR(0, 2317, __pyx_L1_error)
    #else
    __pyx_t_3 = __Pyx_PySequence_ITEM(__pyx_t_2, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2317, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    #endif
    __Pyx_DECREF_SET(__pyx_v_num, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "serializor/binary/ser.py":2318
 * 
 *     for num in (4_294_967_296, 4_294_967_296 * 2):
 *         v = dtype + gen_encoded_int(num)  # type: ignore             # <<<<<<<<<<<<<<
 *         n = gen_header(dchar, num)  # type: ignore
 *         assert v == n, f"{v} | {n} - num: {num}"
 */
    __pyx_t_1 = __Pyx_PyInt_As_unsigned_PY_LONG_LONG(__pyx_v_num); if (unlikely((__pyx_t_1 == (unsigned PY_LONG_LONG)-1) && PyErr_Occurred())) __PYX_ERR(0, 2318, __pyx_L1_error)
    __pyx_t_3 = __pyx_f_10serializor_6binary_3ser_gen_encoded_int(__pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2318, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_7 = PyNumber_Add(__pyx_v_dtype, __pyx_t_3); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2318, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF_SET(__pyx_v_v, ((PyObject*)__pyx_t_7));
    __pyx_t_7 = 0;

    /* "serializor/binary/ser.py":2319
 *     for num in (4_294_967_296, 4_294_967_296 * 2):
 *         v = dtype + gen_encoded_int(num)  # type: ignore
 *         n = gen_header(dchar, num)  # type: ignore             # <<<<<<<<<<<<<<
 *         assert v == n, f"{v} | {n} - num: {num}"
 *     print("Pass Gen Header".ljust(80))
 */
    __pyx_t_1 = __Pyx_PyInt_As_unsigned_PY_LONG_LONG(__pyx_v_num); if (unlikely((__pyx_t_1 == (unsigned PY_LONG_LONG)-1) && PyErr_Occurred())) __PYX_ERR(0, 2319, __pyx_L1_error)
    __pyx_t_7 = __pyx_f_10serializor_6binary_3ser_gen_header(__pyx_v_dchar, __pyx_t_1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2319, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF_SET(__pyx_v_n, ((PyObject*)__pyx_t_7));
    __pyx_t_7 = 0;

    /* "serializor/binary/ser.py":2320
 *         v = dtype + gen_encoded_int(num)  # type: ignore
 *         n = gen_header(dchar, num)  # type: ignore
 *         assert v == n, f"{v} | {n} - num: {num}"             # <<<<<<<<<<<<<<
 *     print("Pass Gen Header".ljust(80))
 */
    #ifndef CYTHON_WITHOUT_ASSERTIONS
    if (unlikely(__pyx_assertions_enabled())) {
      __pyx_t_4 = (__Pyx_PyBytes_Equals(__pyx_v_v, __pyx_v_n, Py_EQ)); if (unlikely((__pyx_t_4 < 0))) __PYX_ERR(0, 2320, __pyx_L1_error)
      if (unlikely(!__pyx_t_4)) {
        __pyx_t_7 = PyTuple_New(5); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2320, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_8 = 0;
        __pyx_t_6 = 127;
        __pyx_t_3 = __Pyx_PyObject_FormatSimple(__pyx_v_v, __pyx_empty_unicode); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2320, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_6 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_3) > __pyx_t_6) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_3) : __pyx_t_6;
        __pyx_t_8 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_3);
        __Pyx_GIVEREF(__pyx_t_3);
        PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_3);
        __pyx_t_3 = 0;
        __Pyx_INCREF(__pyx_kp_u__15);
        __pyx_t_8 += 3;
        __Pyx_GIVEREF(__pyx_kp_u__15);
        PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_kp_u__15);
        __pyx_t_3 = __Pyx_PyObject_FormatSimple(__pyx_v_n, __pyx_empty_unicode); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2320, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_6 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_3) > __pyx_t_6) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_3) : __pyx_t_6;
        __pyx_t_8 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_3);
        __Pyx_GIVEREF(__pyx_t_3);
        PyTuple_SET_ITEM(__pyx_t_7, 2, __pyx_t_3);
        __pyx_t_3 = 0;
        __Pyx_INCREF(__pyx_kp_u_num);
        __pyx_t_8 += 8;
        __Pyx_GIVEREF(__pyx_kp_u_num);
        PyTuple_SET_ITEM(__pyx_t_7, 3, __pyx_kp_u_num);
        __pyx_t_3 = __Pyx_PyObject_FormatSimple(__pyx_v_num, __pyx_empty_unicode); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2320, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_6 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_3) > __pyx_t_6) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_3) : __pyx_t_6;
        __pyx_t_8 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_3);
        __Pyx_GIVEREF(__pyx_t_3);
        PyTuple_SET_ITEM(__pyx_t_7, 4, __pyx_t_3);
        __pyx_t_3 = 0;
        __pyx_t_3 = __Pyx_PyUnicode_Join(__pyx_t_7, 5, __pyx_t_8, __pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2320, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_Raise(__pyx_builtin_AssertionError, __pyx_t_3, 0, 0);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __PYX_ERR(0, 2320, __pyx_L1_error)
      }
    }
    #else
    if ((1)); else __PYX_ERR(0, 2320, __pyx_L1_error)
    #endif

    /* "serializor/binary/ser.py":2317
 *         assert v == n, f"{v} | {n} - num: {num}"
 * 
 *     for num in (4_294_967_296, 4_294_967_296 * 2):             # <<<<<<<<<<<<<<
 *         v = dtype + gen_encoded_int(num)  # type: ignore
 *         n = gen_header(dchar, num)  # type: ignore
 */
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "serializor/binary/ser.py":2321
 *         n = gen_header(dchar, num)  # type: ignore
 *         assert v == n, f"{v} | {n} - num: {num}"
 *     print("Pass Gen Header".ljust(80))             # <<<<<<<<<<<<<<
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_Pass_Gen_Header, __pyx_n_s_ljust); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2321, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_7 = NULL;
  __pyx_t_9 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_7)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_7);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_9 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_int_80};
    __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_9, 1+__pyx_t_9);
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2321, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_builtin_print, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2321, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "serializor/binary/ser.py":2298
 * 
 * 
 * def _test_gen_header() -> None:             # <<<<<<<<<<<<<<
 *     dtype: bytes = b"\x01"
 *     dchar: cython.uchar = 1
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("serializor.binary.ser._test_gen_header", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_dtype);
  __Pyx_XDECREF(__pyx_v_num);
  __Pyx_XDECREF(__pyx_v_v);
  __Pyx_XDECREF(__pyx_v_n);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyMethodDef __pyx_methods[] = {
  {0, 0, 0, 0}
};
#ifndef CYTHON_SMALL_CODE
#if defined(__clang__)
    #define CYTHON_SMALL_CODE
#elif defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 3))
    #define CYTHON_SMALL_CODE __attribute__((cold))
#else
    #define CYTHON_SMALL_CODE
#endif
#endif
/* #### Code section: pystring_table ### */

static int __Pyx_CreateStringTabAndInitStrings(void) {
  __Pyx_StringTabEntry __pyx_string_tab[] = {
    {&__pyx_n_s_AssertionError, __pyx_k_AssertionError, sizeof(__pyx_k_AssertionError), 0, 0, 1, 1},
    {&__pyx_kp_u_B, __pyx_k_B, sizeof(__pyx_k_B), 0, 1, 0, 0},
    {&__pyx_n_s_COMPLEX_ID, __pyx_k_COMPLEX_ID, sizeof(__pyx_k_COMPLEX_ID), 0, 0, 1, 1},
    {&__pyx_n_u_D, __pyx_k_D, sizeof(__pyx_k_D), 0, 1, 0, 1},
    {&__pyx_n_s_DATAFRAME_ID, __pyx_k_DATAFRAME_ID, sizeof(__pyx_k_DATAFRAME_ID), 0, 0, 1, 1},
    {&__pyx_n_s_DATETIMEINDEX_ID, __pyx_k_DATETIMEINDEX_ID, sizeof(__pyx_k_DATETIMEINDEX_ID), 0, 0, 1, 1},
    {&__pyx_n_s_DATETIME_ID, __pyx_k_DATETIME_ID, sizeof(__pyx_k_DATETIME_ID), 0, 0, 1, 1},
    {&__pyx_n_s_DATE_ID, __pyx_k_DATE_ID, sizeof(__pyx_k_DATE_ID), 0, 0, 1, 1},
    {&__pyx_n_s_DICT_ID, __pyx_k_DICT_ID, sizeof(__pyx_k_DICT_ID), 0, 0, 1, 1},
    {&__pyx_n_s_DT64_ID, __pyx_k_DT64_ID, sizeof(__pyx_k_DT64_ID), 0, 0, 1, 1},
    {&__pyx_n_s_DataFrame, __pyx_k_DataFrame, sizeof(__pyx_k_DataFrame), 0, 0, 1, 1},
    {&__pyx_n_s_DatetimeIndex, __pyx_k_DatetimeIndex, sizeof(__pyx_k_DatetimeIndex), 0, 0, 1, 1},
    {&__pyx_kp_u_Error, __pyx_k_Error, sizeof(__pyx_k_Error), 0, 1, 0, 0},
    {&__pyx_n_s_FALSE_VALUE, __pyx_k_FALSE_VALUE, sizeof(__pyx_k_FALSE_VALUE), 0, 0, 1, 1},
    {&__pyx_n_s_FROZENSET_ID, __pyx_k_FROZENSET_ID, sizeof(__pyx_k_FROZENSET_ID), 0, 0, 1, 1},
    {&__pyx_kp_u_H, __pyx_k_H, sizeof(__pyx_k_H), 0, 1, 0, 0},
    {&__pyx_kp_u_I, __pyx_k_I, sizeof(__pyx_k_I), 0, 1, 0, 0},
    {&__pyx_n_s_ImportError, __pyx_k_ImportError, sizeof(__pyx_k_ImportError), 0, 0, 1, 1},
    {&__pyx_n_s_Iterable, __pyx_k_Iterable, sizeof(__pyx_k_Iterable), 0, 0, 1, 1},
    {&__pyx_n_s_LIST_ID, __pyx_k_LIST_ID, sizeof(__pyx_k_LIST_ID), 0, 0, 1, 1},
    {&__pyx_n_u_M, __pyx_k_M, sizeof(__pyx_k_M), 0, 1, 0, 1},
    {&__pyx_n_s_NDARRAY_BOOL_DT, __pyx_k_NDARRAY_BOOL_DT, sizeof(__pyx_k_NDARRAY_BOOL_DT), 0, 0, 1, 1},
    {&__pyx_n_s_NDARRAY_BOOL_IDDT, __pyx_k_NDARRAY_BOOL_IDDT, sizeof(__pyx_k_NDARRAY_BOOL_IDDT), 0, 0, 1, 1},
    {&__pyx_n_s_NDARRAY_BYTES_DT, __pyx_k_NDARRAY_BYTES_DT, sizeof(__pyx_k_NDARRAY_BYTES_DT), 0, 0, 1, 1},
    {&__pyx_n_s_NDARRAY_BYTES_IDDT, __pyx_k_NDARRAY_BYTES_IDDT, sizeof(__pyx_k_NDARRAY_BYTES_IDDT), 0, 0, 1, 1},
    {&__pyx_n_s_NDARRAY_COMPLEX_DT, __pyx_k_NDARRAY_COMPLEX_DT, sizeof(__pyx_k_NDARRAY_COMPLEX_DT), 0, 0, 1, 1},
    {&__pyx_n_s_NDARRAY_COMPLEX_IDDT, __pyx_k_NDARRAY_COMPLEX_IDDT, sizeof(__pyx_k_NDARRAY_COMPLEX_IDDT), 0, 0, 1, 1},
    {&__pyx_n_s_NDARRAY_DT64_DT, __pyx_k_NDARRAY_DT64_DT, sizeof(__pyx_k_NDARRAY_DT64_DT), 0, 0, 1, 1},
    {&__pyx_n_s_NDARRAY_DT64_IDDT, __pyx_k_NDARRAY_DT64_IDDT, sizeof(__pyx_k_NDARRAY_DT64_IDDT), 0, 0, 1, 1},
    {&__pyx_n_s_NDARRAY_FLOAT_DT, __pyx_k_NDARRAY_FLOAT_DT, sizeof(__pyx_k_NDARRAY_FLOAT_DT), 0, 0, 1, 1},
    {&__pyx_n_s_NDARRAY_FLOAT_IDDT, __pyx_k_NDARRAY_FLOAT_IDDT, sizeof(__pyx_k_NDARRAY_FLOAT_IDDT), 0, 0, 1, 1},
    {&__pyx_n_s_NDARRAY_ID, __pyx_k_NDARRAY_ID, sizeof(__pyx_k_NDARRAY_ID), 0, 0, 1, 1},
    {&__pyx_n_s_NDARRAY_INT_DT, __pyx_k_NDARRAY_INT_DT, sizeof(__pyx_k_NDARRAY_INT_DT), 0, 0, 1, 1},
    {&__pyx_n_s_NDARRAY_INT_IDDT, __pyx_k_NDARRAY_INT_IDDT, sizeof(__pyx_k_NDARRAY_INT_IDDT), 0, 0, 1, 1},
    {&__pyx_n_s_NDARRAY_OBJECT_DT, __pyx_k_NDARRAY_OBJECT_DT, sizeof(__pyx_k_NDARRAY_OBJECT_DT), 0, 0, 1, 1},
    {&__pyx_n_s_NDARRAY_OBJECT_IDDT, __pyx_k_NDARRAY_OBJECT_IDDT, sizeof(__pyx_k_NDARRAY_OBJECT_IDDT), 0, 0, 1, 1},
    {&__pyx_n_s_NDARRAY_TD64_DT, __pyx_k_NDARRAY_TD64_DT, sizeof(__pyx_k_NDARRAY_TD64_DT), 0, 0, 1, 1},
    {&__pyx_n_s_NDARRAY_TD64_IDDT, __pyx_k_NDARRAY_TD64_IDDT, sizeof(__pyx_k_NDARRAY_TD64_IDDT), 0, 0, 1, 1},
    {&__pyx_n_s_NDARRAY_UINT_DT, __pyx_k_NDARRAY_UINT_DT, sizeof(__pyx_k_NDARRAY_UINT_DT), 0, 0, 1, 1},
    {&__pyx_n_s_NDARRAY_UINT_IDDT, __pyx_k_NDARRAY_UINT_IDDT, sizeof(__pyx_k_NDARRAY_UINT_IDDT), 0, 0, 1, 1},
    {&__pyx_n_s_NDARRAY_UNICODE_DT, __pyx_k_NDARRAY_UNICODE_DT, sizeof(__pyx_k_NDARRAY_UNICODE_DT), 0, 0, 1, 1},
    {&__pyx_n_s_NDARRAY_UNICODE_IDDT, __pyx_k_NDARRAY_UNICODE_IDDT, sizeof(__pyx_k_NDARRAY_UNICODE_IDDT), 0, 0, 1, 1},
    {&__pyx_n_s_NONE_VALUE, __pyx_k_NONE_VALUE, sizeof(__pyx_k_NONE_VALUE), 0, 0, 1, 1},
    {&__pyx_n_s_None, __pyx_k_None, sizeof(__pyx_k_None), 0, 0, 1, 1},
    {&__pyx_n_s_OPT_SERIALIZE_NUMPY, __pyx_k_OPT_SERIALIZE_NUMPY, sizeof(__pyx_k_OPT_SERIALIZE_NUMPY), 0, 0, 1, 1},
    {&__pyx_n_s_PD_TIMEDELTA_ID, __pyx_k_PD_TIMEDELTA_ID, sizeof(__pyx_k_PD_TIMEDELTA_ID), 0, 0, 1, 1},
    {&__pyx_n_s_PD_TIMESTAMP_ID, __pyx_k_PD_TIMESTAMP_ID, sizeof(__pyx_k_PD_TIMESTAMP_ID), 0, 0, 1, 1},
    {&__pyx_kp_u_Pass_Gen_Encoded_Integer, __pyx_k_Pass_Gen_Encoded_Integer, sizeof(__pyx_k_Pass_Gen_Encoded_Integer), 0, 1, 0, 0},
    {&__pyx_kp_u_Pass_Gen_Header, __pyx_k_Pass_Gen_Header, sizeof(__pyx_k_Pass_Gen_Header), 0, 1, 0, 0},
    {&__pyx_kp_u_Pass_Pack_int16, __pyx_k_Pass_Pack_int16, sizeof(__pyx_k_Pass_Pack_int16), 0, 1, 0, 0},
    {&__pyx_kp_u_Pass_Pack_int24, __pyx_k_Pass_Pack_int24, sizeof(__pyx_k_Pass_Pack_int24), 0, 1, 0, 0},
    {&__pyx_kp_u_Pass_Pack_int32, __pyx_k_Pass_Pack_int32, sizeof(__pyx_k_Pass_Pack_int32), 0, 1, 0, 0},
    {&__pyx_kp_u_Pass_Pack_int64, __pyx_k_Pass_Pack_int64, sizeof(__pyx_k_Pass_Pack_int64), 0, 1, 0, 0},
    {&__pyx_kp_u_Pass_Pack_int8, __pyx_k_Pass_Pack_int8, sizeof(__pyx_k_Pass_Pack_int8), 0, 1, 0, 0},
    {&__pyx_kp_u_Pass_Pack_uint16, __pyx_k_Pass_Pack_uint16, sizeof(__pyx_k_Pass_Pack_uint16), 0, 1, 0, 0},
    {&__pyx_kp_u_Pass_Pack_uint24, __pyx_k_Pass_Pack_uint24, sizeof(__pyx_k_Pass_Pack_uint24), 0, 1, 0, 0},
    {&__pyx_kp_u_Pass_Pack_uint32, __pyx_k_Pass_Pack_uint32, sizeof(__pyx_k_Pass_Pack_uint32), 0, 1, 0, 0},
    {&__pyx_kp_u_Pass_Pack_uint64, __pyx_k_Pass_Pack_uint64, sizeof(__pyx_k_Pass_Pack_uint64), 0, 1, 0, 0},
    {&__pyx_kp_u_Pass_Pack_uint8, __pyx_k_Pass_Pack_uint8, sizeof(__pyx_k_Pass_Pack_uint8), 0, 1, 0, 0},
    {&__pyx_kp_u_Q, __pyx_k_Q, sizeof(__pyx_k_Q), 0, 1, 0, 0},
    {&__pyx_n_s_RANGE_ID, __pyx_k_RANGE_ID, sizeof(__pyx_k_RANGE_ID), 0, 0, 1, 1},
    {&__pyx_n_s_RuntimeError, __pyx_k_RuntimeError, sizeof(__pyx_k_RuntimeError), 0, 0, 1, 1},
    {&__pyx_n_s_SERIES_ID, __pyx_k_SERIES_ID, sizeof(__pyx_k_SERIES_ID), 0, 0, 1, 1},
    {&__pyx_n_s_SET_ID, __pyx_k_SET_ID, sizeof(__pyx_k_SET_ID), 0, 0, 1, 1},
    {&__pyx_n_s_STRUCT_TIME_ID, __pyx_k_STRUCT_TIME_ID, sizeof(__pyx_k_STRUCT_TIME_ID), 0, 0, 1, 1},
    {&__pyx_n_s_SerializeError, __pyx_k_SerializeError, sizeof(__pyx_k_SerializeError), 0, 0, 1, 1},
    {&__pyx_kp_u_Serializor_Failed_to_serialize, __pyx_k_Serializor_Failed_to_serialize, sizeof(__pyx_k_Serializor_Failed_to_serialize), 0, 1, 0, 0},
    {&__pyx_n_s_Series, __pyx_k_Series, sizeof(__pyx_k_Series), 0, 0, 1, 1},
    {&__pyx_n_s_TD64_ID, __pyx_k_TD64_ID, sizeof(__pyx_k_TD64_ID), 0, 0, 1, 1},
    {&__pyx_n_s_TIMEDELTAINDEX_ID, __pyx_k_TIMEDELTAINDEX_ID, sizeof(__pyx_k_TIMEDELTAINDEX_ID), 0, 0, 1, 1},
    {&__pyx_n_s_TIMEDELTA_ID, __pyx_k_TIMEDELTA_ID, sizeof(__pyx_k_TIMEDELTA_ID), 0, 0, 1, 1},
    {&__pyx_n_s_TIME_ID, __pyx_k_TIME_ID, sizeof(__pyx_k_TIME_ID), 0, 0, 1, 1},
    {&__pyx_n_s_TRUE_VALUE, __pyx_k_TRUE_VALUE, sizeof(__pyx_k_TRUE_VALUE), 0, 0, 1, 1},
    {&__pyx_n_s_TUPLE_ID, __pyx_k_TUPLE_ID, sizeof(__pyx_k_TUPLE_ID), 0, 0, 1, 1},
    {&__pyx_kp_u_Time_zones_are_not_available_fro, __pyx_k_Time_zones_are_not_available_fro, sizeof(__pyx_k_Time_zones_are_not_available_fro), 0, 1, 0, 0},
    {&__pyx_n_s_Timedelta, __pyx_k_Timedelta, sizeof(__pyx_k_Timedelta), 0, 0, 1, 1},
    {&__pyx_n_s_TimedeltaIndex, __pyx_k_TimedeltaIndex, sizeof(__pyx_k_TimedeltaIndex), 0, 0, 1, 1},
    {&__pyx_n_s_Timestamp, __pyx_k_Timestamp, sizeof(__pyx_k_Timestamp), 0, 0, 1, 1},
    {&__pyx_n_s_TypeError, __pyx_k_TypeError, sizeof(__pyx_k_TypeError), 0, 0, 1, 1},
    {&__pyx_n_s_UINT16_ENCODE_VALUE, __pyx_k_UINT16_ENCODE_VALUE, sizeof(__pyx_k_UINT16_ENCODE_VALUE), 0, 0, 1, 1},
    {&__pyx_n_s_UINT32_ENCODE_VALUE, __pyx_k_UINT32_ENCODE_VALUE, sizeof(__pyx_k_UINT32_ENCODE_VALUE), 0, 0, 1, 1},
    {&__pyx_n_s_UINT64_ENCODE_VALUE, __pyx_k_UINT64_ENCODE_VALUE, sizeof(__pyx_k_UINT64_ENCODE_VALUE), 0, 0, 1, 1},
    {&__pyx_n_s_UINT8_ENCODE_VALUE, __pyx_k_UINT8_ENCODE_VALUE, sizeof(__pyx_k_UINT8_ENCODE_VALUE), 0, 0, 1, 1},
    {&__pyx_n_s_ValueError, __pyx_k_ValueError, sizeof(__pyx_k_ValueError), 0, 0, 1, 1},
    {&__pyx_n_u_W, __pyx_k_W, sizeof(__pyx_k_W), 0, 1, 0, 1},
    {&__pyx_n_u_Y, __pyx_k_Y, sizeof(__pyx_k_Y), 0, 1, 0, 1},
    {&__pyx_kp_u__13, __pyx_k__13, sizeof(__pyx_k__13), 0, 1, 0, 0},
    {&__pyx_n_s__14, __pyx_k__14, sizeof(__pyx_k__14), 0, 0, 1, 1},
    {&__pyx_kp_u__15, __pyx_k__15, sizeof(__pyx_k__15), 0, 1, 0, 0},
    {&__pyx_kp_u__16, __pyx_k__16, sizeof(__pyx_k__16), 0, 1, 0, 0},
    {&__pyx_kp_u__27, __pyx_k__27, sizeof(__pyx_k__27), 0, 1, 0, 0},
    {&__pyx_kp_b__4, __pyx_k__4, sizeof(__pyx_k__4), 0, 0, 0, 0},
    {&__pyx_n_s__46, __pyx_k__46, sizeof(__pyx_k__46), 0, 0, 1, 1},
    {&__pyx_kp_b__5, __pyx_k__5, sizeof(__pyx_k__5), 0, 0, 0, 0},
    {&__pyx_kp_b__6, __pyx_k__6, sizeof(__pyx_k__6), 0, 0, 0, 0},
    {&__pyx_kp_b__7, __pyx_k__7, sizeof(__pyx_k__7), 0, 0, 0, 0},
    {&__pyx_kp_b__8, __pyx_k__8, sizeof(__pyx_k__8), 0, 0, 0, 0},
    {&__pyx_kp_b__9, __pyx_k__9, sizeof(__pyx_k__9), 0, 0, 0, 0},
    {&__pyx_n_s_all, __pyx_k_all, sizeof(__pyx_k_all), 0, 0, 1, 1},
    {&__pyx_n_u_as, __pyx_k_as, sizeof(__pyx_k_as), 0, 1, 0, 1},
    {&__pyx_n_s_asyncio_coroutines, __pyx_k_asyncio_coroutines, sizeof(__pyx_k_asyncio_coroutines), 0, 0, 1, 1},
    {&__pyx_kp_u_b, __pyx_k_b, sizeof(__pyx_k_b), 0, 1, 0, 0},
    {&__pyx_n_s_b_2, __pyx_k_b_2, sizeof(__pyx_k_b_2), 0, 0, 1, 1},
    {&__pyx_n_s_class_getitem, __pyx_k_class_getitem, sizeof(__pyx_k_class_getitem), 0, 0, 1, 1},
    {&__pyx_n_s_cline_in_traceback, __pyx_k_cline_in_traceback, sizeof(__pyx_k_cline_in_traceback), 0, 0, 1, 1},
    {&__pyx_n_s_columns, __pyx_k_columns, sizeof(__pyx_k_columns), 0, 0, 1, 1},
    {&__pyx_n_s_datetime, __pyx_k_datetime, sizeof(__pyx_k_datetime), 0, 0, 1, 1},
    {&__pyx_n_s_dchar, __pyx_k_dchar, sizeof(__pyx_k_dchar), 0, 0, 1, 1},
    {&__pyx_n_s_dt, __pyx_k_dt, sizeof(__pyx_k_dt), 0, 0, 1, 1},
    {&__pyx_n_s_dtype, __pyx_k_dtype, sizeof(__pyx_k_dtype), 0, 0, 1, 1},
    {&__pyx_n_s_dumps, __pyx_k_dumps, sizeof(__pyx_k_dumps), 0, 0, 1, 1},
    {&__pyx_n_s_dumps_2, __pyx_k_dumps_2, sizeof(__pyx_k_dumps_2), 0, 0, 1, 1},
    {&__pyx_n_s_end, __pyx_k_end, sizeof(__pyx_k_end), 0, 0, 1, 1},
    {&__pyx_n_s_errors, __pyx_k_errors, sizeof(__pyx_k_errors), 0, 0, 1, 1},
    {&__pyx_kp_u_expects_np_ndarray_dtype_of_date, __pyx_k_expects_np_ndarray_dtype_of_date, sizeof(__pyx_k_expects_np_ndarray_dtype_of_date), 0, 1, 0, 0},
    {&__pyx_kp_u_expects_pandas_DatetimeIndex_got, __pyx_k_expects_pandas_DatetimeIndex_got, sizeof(__pyx_k_expects_pandas_DatetimeIndex_got), 0, 1, 0, 0},
    {&__pyx_kp_u_expects_pandas_Series_got_s, __pyx_k_expects_pandas_Series_got_s, sizeof(__pyx_k_expects_pandas_Series_got_s), 0, 1, 0, 0},
    {&__pyx_kp_u_expects_pandas_TimedeltaIndex_go, __pyx_k_expects_pandas_TimedeltaIndex_go, sizeof(__pyx_k_expects_pandas_TimedeltaIndex_go), 0, 1, 0, 0},
    {&__pyx_kp_u_expects_pandas_Timedelta_got_s, __pyx_k_expects_pandas_Timedelta_got_s, sizeof(__pyx_k_expects_pandas_Timedelta_got_s), 0, 1, 0, 0},
    {&__pyx_kp_u_expects_pandas_Timestamp_got_s, __pyx_k_expects_pandas_Timestamp_got_s, sizeof(__pyx_k_expects_pandas_Timestamp_got_s), 0, 1, 0, 0},
    {&__pyx_n_u_fs, __pyx_k_fs, sizeof(__pyx_k_fs), 0, 1, 0, 1},
    {&__pyx_n_u_h, __pyx_k_h, sizeof(__pyx_k_h), 0, 1, 0, 1},
    {&__pyx_kp_u_h_2, __pyx_k_h_2, sizeof(__pyx_k_h_2), 0, 1, 0, 0},
    {&__pyx_kp_u_i, __pyx_k_i, sizeof(__pyx_k_i), 0, 1, 0, 0},
    {&__pyx_n_s_identifier, __pyx_k_identifier, sizeof(__pyx_k_identifier), 0, 0, 1, 1},
    {&__pyx_n_s_imag, __pyx_k_imag, sizeof(__pyx_k_imag), 0, 0, 1, 1},
    {&__pyx_n_s_import, __pyx_k_import, sizeof(__pyx_k_import), 0, 0, 1, 1},
    {&__pyx_n_s_index, __pyx_k_index, sizeof(__pyx_k_index), 0, 0, 1, 1},
    {&__pyx_n_s_initializing, __pyx_k_initializing, sizeof(__pyx_k_initializing), 0, 0, 1, 1},
    {&__pyx_kp_u_invalid_pandas_DataFrame_cannot, __pyx_k_invalid_pandas_DataFrame_cannot, sizeof(__pyx_k_invalid_pandas_DataFrame_cannot), 0, 1, 0, 0},
    {&__pyx_kp_u_invalid_pandas_DataFrame_columns, __pyx_k_invalid_pandas_DataFrame_columns, sizeof(__pyx_k_invalid_pandas_DataFrame_columns), 0, 1, 0, 0},
    {&__pyx_kp_u_invalid_pandas_DatetimeIndex_nam, __pyx_k_invalid_pandas_DatetimeIndex_nam, sizeof(__pyx_k_invalid_pandas_DatetimeIndex_nam), 0, 1, 0, 0},
    {&__pyx_kp_u_invalid_pandas_Series_name_r, __pyx_k_invalid_pandas_Series_name_r, sizeof(__pyx_k_invalid_pandas_Series_name_r), 0, 1, 0, 0},
    {&__pyx_kp_u_invalid_pandas_TimedeltaIndex_na, __pyx_k_invalid_pandas_TimedeltaIndex_na, sizeof(__pyx_k_invalid_pandas_TimedeltaIndex_na), 0, 1, 0, 0},
    {&__pyx_n_s_is_coroutine, __pyx_k_is_coroutine, sizeof(__pyx_k_is_coroutine), 0, 0, 1, 1},
    {&__pyx_n_s_items, __pyx_k_items, sizeof(__pyx_k_items), 0, 0, 1, 1},
    {&__pyx_n_s_join, __pyx_k_join, sizeof(__pyx_k_join), 0, 0, 1, 1},
    {&__pyx_n_s_ljust, __pyx_k_ljust, sizeof(__pyx_k_ljust), 0, 0, 1, 1},
    {&__pyx_n_u_m, __pyx_k_m, sizeof(__pyx_k_m), 0, 1, 0, 1},
    {&__pyx_n_s_main, __pyx_k_main, sizeof(__pyx_k_main), 0, 0, 1, 1},
    {&__pyx_n_u_ms, __pyx_k_ms, sizeof(__pyx_k_ms), 0, 1, 0, 1},
    {&__pyx_n_s_n, __pyx_k_n, sizeof(__pyx_k_n), 0, 0, 1, 1},
    {&__pyx_n_s_name, __pyx_k_name, sizeof(__pyx_k_name), 0, 0, 1, 1},
    {&__pyx_n_s_name_2, __pyx_k_name_2, sizeof(__pyx_k_name_2), 0, 0, 1, 1},
    {&__pyx_n_s_np, __pyx_k_np, sizeof(__pyx_k_np), 0, 0, 1, 1},
    {&__pyx_n_u_ns, __pyx_k_ns, sizeof(__pyx_k_ns), 0, 1, 0, 1},
    {&__pyx_kp_u_num, __pyx_k_num, sizeof(__pyx_k_num), 0, 1, 0, 0},
    {&__pyx_n_s_num_2, __pyx_k_num_2, sizeof(__pyx_k_num_2), 0, 0, 1, 1},
    {&__pyx_n_s_numpy, __pyx_k_numpy, sizeof(__pyx_k_numpy), 0, 0, 1, 1},
    {&__pyx_kp_u_numpy__core_multiarray_failed_to, __pyx_k_numpy__core_multiarray_failed_to, sizeof(__pyx_k_numpy__core_multiarray_failed_to), 0, 1, 0, 0},
    {&__pyx_kp_u_numpy__core_umath_failed_to_impo, __pyx_k_numpy__core_umath_failed_to_impo, sizeof(__pyx_k_numpy__core_umath_failed_to_impo), 0, 1, 0, 0},
    {&__pyx_n_s_obj, __pyx_k_obj, sizeof(__pyx_k_obj), 0, 0, 1, 1},
    {&__pyx_n_s_object, __pyx_k_object, sizeof(__pyx_k_object), 0, 0, 1, 1},
    {&__pyx_n_s_option, __pyx_k_option, sizeof(__pyx_k_option), 0, 0, 1, 1},
    {&__pyx_n_s_orjson, __pyx_k_orjson, sizeof(__pyx_k_orjson), 0, 0, 1, 1},
    {&__pyx_n_s_pack, __pyx_k_pack, sizeof(__pyx_k_pack), 0, 0, 1, 1},
    {&__pyx_n_s_pandas, __pyx_k_pandas, sizeof(__pyx_k_pandas), 0, 0, 1, 1},
    {&__pyx_kp_u_pass_s, __pyx_k_pass_s, sizeof(__pyx_k_pass_s), 0, 1, 0, 0},
    {&__pyx_n_s_print, __pyx_k_print, sizeof(__pyx_k_print), 0, 0, 1, 1},
    {&__pyx_n_u_ps, __pyx_k_ps, sizeof(__pyx_k_ps), 0, 1, 0, 1},
    {&__pyx_n_s_pyx_capi, __pyx_k_pyx_capi, sizeof(__pyx_k_pyx_capi), 0, 0, 1, 1},
    {&__pyx_kp_u_q, __pyx_k_q, sizeof(__pyx_k_q), 0, 1, 0, 0},
    {&__pyx_n_s_range, __pyx_k_range, sizeof(__pyx_k_range), 0, 0, 1, 1},
    {&__pyx_n_s_real, __pyx_k_real, sizeof(__pyx_k_real), 0, 0, 1, 1},
    {&__pyx_n_s_return, __pyx_k_return, sizeof(__pyx_k_return), 0, 0, 1, 1},
    {&__pyx_n_s_s, __pyx_k_s, sizeof(__pyx_k_s), 0, 0, 1, 1},
    {&__pyx_n_u_s, __pyx_k_s, sizeof(__pyx_k_s), 0, 1, 0, 1},
    {&__pyx_n_s_serialize, __pyx_k_serialize, sizeof(__pyx_k_serialize), 0, 0, 1, 1},
    {&__pyx_n_u_serialize, __pyx_k_serialize, sizeof(__pyx_k_serialize), 0, 1, 0, 1},
    {&__pyx_n_s_serializor, __pyx_k_serializor, sizeof(__pyx_k_serializor), 0, 0, 1, 1},
    {&__pyx_n_s_serializor_binary_ser, __pyx_k_serializor_binary_ser, sizeof(__pyx_k_serializor_binary_ser), 0, 0, 1, 1},
    {&__pyx_n_s_spec, __pyx_k_spec, sizeof(__pyx_k_spec), 0, 0, 1, 1},
    {&__pyx_kp_s_src_serializor_binary_ser_py, __pyx_k_src_serializor_binary_ser_py, sizeof(__pyx_k_src_serializor_binary_ser_py), 0, 0, 1, 0},
    {&__pyx_n_s_start, __pyx_k_start, sizeof(__pyx_k_start), 0, 0, 1, 1},
    {&__pyx_n_s_step, __pyx_k_step, sizeof(__pyx_k_step), 0, 0, 1, 1},
    {&__pyx_n_s_stop, __pyx_k_stop, sizeof(__pyx_k_stop), 0, 0, 1, 1},
    {&__pyx_n_s_str, __pyx_k_str, sizeof(__pyx_k_str), 0, 0, 1, 1},
    {&__pyx_n_s_struct, __pyx_k_struct, sizeof(__pyx_k_struct), 0, 0, 1, 1},
    {&__pyx_n_s_struct_time, __pyx_k_struct_time, sizeof(__pyx_k_struct_time), 0, 0, 1, 1},
    {&__pyx_n_s_test, __pyx_k_test, sizeof(__pyx_k_test), 0, 0, 1, 1},
    {&__pyx_n_s_test_gen_encoded_int, __pyx_k_test_gen_encoded_int, sizeof(__pyx_k_test_gen_encoded_int), 0, 0, 1, 1},
    {&__pyx_n_s_test_gen_header, __pyx_k_test_gen_header, sizeof(__pyx_k_test_gen_header), 0, 0, 1, 1},
    {&__pyx_n_s_test_pack_int16, __pyx_k_test_pack_int16, sizeof(__pyx_k_test_pack_int16), 0, 0, 1, 1},
    {&__pyx_n_s_test_pack_int24, __pyx_k_test_pack_int24, sizeof(__pyx_k_test_pack_int24), 0, 0, 1, 1},
    {&__pyx_n_s_test_pack_int32, __pyx_k_test_pack_int32, sizeof(__pyx_k_test_pack_int32), 0, 0, 1, 1},
    {&__pyx_n_s_test_pack_int64, __pyx_k_test_pack_int64, sizeof(__pyx_k_test_pack_int64), 0, 0, 1, 1},
    {&__pyx_n_s_test_pack_int8, __pyx_k_test_pack_int8, sizeof(__pyx_k_test_pack_int8), 0, 0, 1, 1},
    {&__pyx_n_s_test_pack_uint16, __pyx_k_test_pack_uint16, sizeof(__pyx_k_test_pack_uint16), 0, 0, 1, 1},
    {&__pyx_n_s_test_pack_uint24, __pyx_k_test_pack_uint24, sizeof(__pyx_k_test_pack_uint24), 0, 0, 1, 1},
    {&__pyx_n_s_test_pack_uint32, __pyx_k_test_pack_uint32, sizeof(__pyx_k_test_pack_uint32), 0, 0, 1, 1},
    {&__pyx_n_s_test_pack_uint64, __pyx_k_test_pack_uint64, sizeof(__pyx_k_test_pack_uint64), 0, 0, 1, 1},
    {&__pyx_n_s_test_pack_uint8, __pyx_k_test_pack_uint8, sizeof(__pyx_k_test_pack_uint8), 0, 0, 1, 1},
    {&__pyx_n_s_test_utils, __pyx_k_test_utils, sizeof(__pyx_k_test_utils), 0, 0, 1, 1},
    {&__pyx_n_s_time, __pyx_k_time, sizeof(__pyx_k_time), 0, 0, 1, 1},
    {&__pyx_n_s_tm_gmtoff, __pyx_k_tm_gmtoff, sizeof(__pyx_k_tm_gmtoff), 0, 0, 1, 1},
    {&__pyx_n_s_tm_zone, __pyx_k_tm_zone, sizeof(__pyx_k_tm_zone), 0, 0, 1, 1},
    {&__pyx_n_s_tobytes, __pyx_k_tobytes, sizeof(__pyx_k_tobytes), 0, 0, 1, 1},
    {&__pyx_n_s_typeref, __pyx_k_typeref, sizeof(__pyx_k_typeref), 0, 0, 1, 1},
    {&__pyx_n_s_typing, __pyx_k_typing, sizeof(__pyx_k_typing), 0, 0, 1, 1},
    {&__pyx_n_s_tz, __pyx_k_tz, sizeof(__pyx_k_tz), 0, 0, 1, 1},
    {&__pyx_kp_u_unable_to_parse_ndarray_time_uni, __pyx_k_unable_to_parse_ndarray_time_uni, sizeof(__pyx_k_unable_to_parse_ndarray_time_uni), 0, 1, 0, 0},
    {&__pyx_kp_u_unsupported_data_type_s, __pyx_k_unsupported_data_type_s, sizeof(__pyx_k_unsupported_data_type_s), 0, 1, 0, 0},
    {&__pyx_kp_u_unsupported_numpy_ndarray_dimens, __pyx_k_unsupported_numpy_ndarray_dimens, sizeof(__pyx_k_unsupported_numpy_ndarray_dimens), 0, 1, 0, 0},
    {&__pyx_kp_u_unsupported_numpy_ndarray_dtype, __pyx_k_unsupported_numpy_ndarray_dtype, sizeof(__pyx_k_unsupported_numpy_ndarray_dtype), 0, 1, 0, 0},
    {&__pyx_kp_u_unsupported_numpy_time_unit_d, __pyx_k_unsupported_numpy_time_unit_d, sizeof(__pyx_k_unsupported_numpy_time_unit_d), 0, 1, 0, 0},
    {&__pyx_kp_u_unsupported_numpy_time_unit_s, __pyx_k_unsupported_numpy_time_unit_s, sizeof(__pyx_k_unsupported_numpy_time_unit_s), 0, 1, 0, 0},
    {&__pyx_kp_u_unsupported_pandas_DataFrame_col, __pyx_k_unsupported_pandas_DataFrame_col, sizeof(__pyx_k_unsupported_pandas_DataFrame_col), 0, 1, 0, 0},
    {&__pyx_kp_u_unsupported_pandas_Series_dtype, __pyx_k_unsupported_pandas_Series_dtype, sizeof(__pyx_k_unsupported_pandas_Series_dtype), 0, 1, 0, 0},
    {&__pyx_n_u_us, __pyx_k_us, sizeof(__pyx_k_us), 0, 1, 0, 1},
    {&__pyx_n_s_utcoffset, __pyx_k_utcoffset, sizeof(__pyx_k_utcoffset), 0, 0, 1, 1},
    {&__pyx_n_s_utils, __pyx_k_utils, sizeof(__pyx_k_utils), 0, 0, 1, 1},
    {&__pyx_n_s_v, __pyx_k_v, sizeof(__pyx_k_v), 0, 0, 1, 1},
    {&__pyx_n_s_val, __pyx_k_val, sizeof(__pyx_k_val), 0, 0, 1, 1},
    {&__pyx_n_s_value, __pyx_k_value, sizeof(__pyx_k_value), 0, 0, 1, 1},
    {&__pyx_n_s_values, __pyx_k_values, sizeof(__pyx_k_values), 0, 0, 1, 1},
    {0, 0, 0, 0, 0, 0, 0}
  };
  return __Pyx_InitStrings(__pyx_string_tab);
}
/* #### Code section: cached_builtins ### */
static CYTHON_SMALL_CODE int __Pyx_InitCachedBuiltins(void) {
  __pyx_builtin_TypeError = __Pyx_GetBuiltinName(__pyx_n_s_TypeError); if (!__pyx_builtin_TypeError) __PYX_ERR(0, 708, __pyx_L1_error)
  __pyx_builtin_range = __Pyx_GetBuiltinName(__pyx_n_s_range); if (!__pyx_builtin_range) __PYX_ERR(0, 745, __pyx_L1_error)
  __pyx_builtin_ValueError = __Pyx_GetBuiltinName(__pyx_n_s_ValueError); if (!__pyx_builtin_ValueError) __PYX_ERR(0, 816, __pyx_L1_error)
  __pyx_builtin_AssertionError = __Pyx_GetBuiltinName(__pyx_n_s_AssertionError); if (!__pyx_builtin_AssertionError) __PYX_ERR(0, 2138, __pyx_L1_error)
  __pyx_builtin_print = __Pyx_GetBuiltinName(__pyx_n_s_print); if (!__pyx_builtin_print) __PYX_ERR(0, 2139, __pyx_L1_error)
  __pyx_builtin_ImportError = __Pyx_GetBuiltinName(__pyx_n_s_ImportError); if (!__pyx_builtin_ImportError) __PYX_ERR(2, 1010, __pyx_L1_error)
  __pyx_builtin_RuntimeError = __Pyx_GetBuiltinName(__pyx_n_s_RuntimeError); if (!__pyx_builtin_RuntimeError) __PYX_ERR(4, 308, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}
/* #### Code section: cached_constants ### */

static CYTHON_SMALL_CODE int __Pyx_InitCachedConstants(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_InitCachedConstants", 0);

  /* "../../../../../tmp/build-env-82p3tyx6/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1010
 *         __pyx_import_array()
 *     except Exception:
 *         raise ImportError("numpy._core.multiarray failed to import")             # <<<<<<<<<<<<<<
 * 
 * cdef inline int import_umath() except -1:
 */
  __pyx_tuple_ = PyTuple_Pack(1, __pyx_kp_u_numpy__core_multiarray_failed_to); if (unlikely(!__pyx_tuple_)) __PYX_ERR(2, 1010, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple_);
  __Pyx_GIVEREF(__pyx_tuple_);

  /* "../../../../../tmp/build-env-82p3tyx6/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1016
 *         _import_umath()
 *     except Exception:
 *         raise ImportError("numpy._core.umath failed to import")             # <<<<<<<<<<<<<<
 * 
 * cdef inline int import_ufunc() except -1:
 */
  __pyx_tuple__2 = PyTuple_Pack(1, __pyx_kp_u_numpy__core_umath_failed_to_impo); if (unlikely(!__pyx_tuple__2)) __PYX_ERR(2, 1016, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2);
  __Pyx_GIVEREF(__pyx_tuple__2);

  /* "datetime.pxd":308
 * cdef inline object timezone_new(object offset, object name=None):
 *     if PY_VERSION_HEX < 0x030700b1:
 *         raise RuntimeError('Time zones are not available from the C-API.')             # <<<<<<<<<<<<<<
 *     return __Pyx_TimeZone_FromOffsetAndName(offset, <PyObject*>name if name is not None else NULL)
 * 
 */
  __pyx_tuple__3 = PyTuple_Pack(1, __pyx_kp_u_Time_zones_are_not_available_fro); if (unlikely(!__pyx_tuple__3)) __PYX_ERR(4, 308, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__3);
  __Pyx_GIVEREF(__pyx_tuple__3);

  /* "serializor/binary/ser.py":985
 *             items.append(pack_uint8(utils.parse_arr_nptime_unit(arr)))  # type: ignore
 *         else:
 *             items.append(pack_uint8(np.get_datetime64_unit(arr[0, 0])))  # type: ignore             # <<<<<<<<<<<<<<
 *             arr = np.PyArray_Cast(arr, np.NPY_TYPES.NPY_INT64)  # cast to int64
 *             values_json = _orjson_dumps_numpy(utils.arr_flatten(arr))
 */
  __pyx_tuple__10 = PyTuple_Pack(2, __pyx_int_0, __pyx_int_0); if (unlikely(!__pyx_tuple__10)) __PYX_ERR(0, 985, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__10);
  __Pyx_GIVEREF(__pyx_tuple__10);

  /* "serializor/binary/ser.py":1009
 *             items.append(pack_uint8(utils.parse_arr_nptime_unit(arr)))  # type: ignore
 *         else:
 *             items.append(pack_uint8(np.get_datetime64_unit(arr[0, 0, 0])))  # type: ignore             # <<<<<<<<<<<<<<
 *             arr = np.PyArray_Cast(arr, np.NPY_TYPES.NPY_INT64)  # cast to int64
 *             values_json = _orjson_dumps_numpy(utils.arr_flatten(arr))
 */
  __pyx_tuple__11 = PyTuple_Pack(3, __pyx_int_0, __pyx_int_0, __pyx_int_0); if (unlikely(!__pyx_tuple__11)) __PYX_ERR(0, 1009, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__11);
  __Pyx_GIVEREF(__pyx_tuple__11);

  /* "serializor/binary/ser.py":1035
 *             items.append(pack_uint8(utils.parse_arr_nptime_unit(arr)))  # type: ignore
 *         else:
 *             items.append(pack_uint8(np.get_datetime64_unit(arr[0, 0, 0, 0])))  # type: ignore             # <<<<<<<<<<<<<<
 *             arr = np.PyArray_Cast(arr, np.NPY_TYPES.NPY_INT64)  # cast to int64
 *             values_json = _orjson_dumps_numpy(utils.arr_flatten(arr))
 */
  __pyx_tuple__12 = PyTuple_Pack(4, __pyx_int_0, __pyx_int_0, __pyx_int_0, __pyx_int_0); if (unlikely(!__pyx_tuple__12)) __PYX_ERR(0, 1035, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__12);
  __Pyx_GIVEREF(__pyx_tuple__12);

  /* "serializor/binary/ser.py":2187
 *     import struct
 * 
 *     for val in (-8388608, -8388607, 0, 8388606, 8388607):             # <<<<<<<<<<<<<<
 *         s = struct.pack("<i", val)[:3]
 *         b = pack_int24(val)  # type: ignore
 */
  __pyx_tuple__17 = PyTuple_Pack(5, __pyx_int_neg_8388608, __pyx_int_neg_8388607, __pyx_int_0, __pyx_int_8388606, __pyx_int_8388607); if (unlikely(!__pyx_tuple__17)) __PYX_ERR(0, 2187, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__17);
  __Pyx_GIVEREF(__pyx_tuple__17);

  /* "serializor/binary/ser.py":2188
 * 
 *     for val in (-8388608, -8388607, 0, 8388606, 8388607):
 *         s = struct.pack("<i", val)[:3]             # <<<<<<<<<<<<<<
 *         b = pack_int24(val)  # type: ignore
 *         assert s == b, f"{s} | {b}"
 */
  __pyx_slice__18 = PySlice_New(Py_None, __pyx_int_3, Py_None); if (unlikely(!__pyx_slice__18)) __PYX_ERR(0, 2188, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_slice__18);
  __Pyx_GIVEREF(__pyx_slice__18);

  /* "serializor/binary/ser.py":2200
 *     import struct
 * 
 *     for val in (0, 1, 16777213, 16777214, 16777215):             # <<<<<<<<<<<<<<
 *         s = struct.pack("<I", val)[:3]
 *         b = pack_uint24(val)  # type: ignore
 */
  __pyx_tuple__19 = PyTuple_Pack(5, __pyx_int_0, __pyx_int_1, __pyx_int_16777213, __pyx_int_16777214, __pyx_int_16777215); if (unlikely(!__pyx_tuple__19)) __PYX_ERR(0, 2200, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__19);
  __Pyx_GIVEREF(__pyx_tuple__19);

  /* "serializor/binary/ser.py":2213
 *     import struct
 * 
 *     for val in (-2147483648, -2147483647, 0, 2147483646, 2147483647):             # <<<<<<<<<<<<<<
 *         s = struct.pack("<i", val)
 *         b = pack_int32(val)  # type: ignore
 */
  __pyx_tuple__20 = PyTuple_Pack(5, __pyx_int_neg_2147483648, __pyx_int_neg_2147483647, __pyx_int_0, __pyx_int_2147483646, __pyx_int_2147483647); if (unlikely(!__pyx_tuple__20)) __PYX_ERR(0, 2213, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__20);
  __Pyx_GIVEREF(__pyx_tuple__20);

  /* "serializor/binary/ser.py":2226
 *     import struct
 * 
 *     for val in (0, 1, 4294967293, 4294967294, 4294967295):             # <<<<<<<<<<<<<<
 *         s = struct.pack("<I", val)
 *         b = pack_uint32(val)  # type: ignore
 */
  __pyx_tuple__21 = PyTuple_Pack(5, __pyx_int_0, __pyx_int_1, __pyx_int_4294967293, __pyx_int_4294967294, __pyx_int_4294967295); if (unlikely(!__pyx_tuple__21)) __PYX_ERR(0, 2226, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__21);
  __Pyx_GIVEREF(__pyx_tuple__21);

  /* "serializor/binary/ser.py":2240
 * 
 *     for val in (
 *         -9223372036854775808,             # <<<<<<<<<<<<<<
 *         -9223372036854775807,
 *         0,
 */
  __pyx_tuple__22 = PyTuple_Pack(5, __pyx_int_neg_9223372036854775808, __pyx_int_neg_9223372036854775807, __pyx_int_0, __pyx_int_0x7ffffffffffffffe, __pyx_int_0x7fffffffffffffff); if (unlikely(!__pyx_tuple__22)) __PYX_ERR(0, 2240, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__22);
  __Pyx_GIVEREF(__pyx_tuple__22);

  /* "serializor/binary/ser.py":2259
 * 
 *     for val in (
 *         0,             # <<<<<<<<<<<<<<
 *         1,
 *         18446744073709551613,
 */
  __pyx_tuple__23 = PyTuple_Pack(5, __pyx_int_0, __pyx_int_1, __pyx_int_0xfffffffffffffffd, __pyx_int_0xfffffffffffffffe, __pyx_int_0xffffffffffffffff); if (unlikely(!__pyx_tuple__23)) __PYX_ERR(0, 2259, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__23);
  __Pyx_GIVEREF(__pyx_tuple__23);

  /* "serializor/binary/ser.py":2281
 *     assert v == n, f"{v} | {n} - num: {num}"
 * 
 *     for num in (252, 65_535):             # <<<<<<<<<<<<<<
 *         v = gen_encoded_int(num)  # type: ignore
 *         n = pack_uint8(UINT16_ENCODE_VALUE) + pack_uint16(num)  # type: ignore
 */
  __pyx_tuple__24 = PyTuple_Pack(2, __pyx_int_252, __pyx_int_65535); if (unlikely(!__pyx_tuple__24)) __PYX_ERR(0, 2281, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__24);
  __Pyx_GIVEREF(__pyx_tuple__24);

  /* "serializor/binary/ser.py":2286
 *         assert v == n, f"{v} | {n} - num: {num}"
 * 
 *     for num in (65_536, 4_294_967_295):             # <<<<<<<<<<<<<<
 *         v = gen_encoded_int(num)  # type: ignore
 *         n = pack_uint8(UINT32_ENCODE_VALUE) + pack_uint32(num)  # type: ignore
 */
  __pyx_tuple__25 = PyTuple_Pack(2, __pyx_int_65536, __pyx_int_4294967295); if (unlikely(!__pyx_tuple__25)) __PYX_ERR(0, 2286, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__25);
  __Pyx_GIVEREF(__pyx_tuple__25);

  /* "serializor/binary/ser.py":2291
 *         assert v == n, f"{v} | {n} - num: {num}"
 * 
 *     for num in (4_294_967_296, 4_294_967_296 * 2):             # <<<<<<<<<<<<<<
 *         v = gen_encoded_int(num)  # type: ignore
 *         n = pack_uint8(UINT64_ENCODE_VALUE) + pack_uint64(num)  # type: ignore
 */
  __pyx_tuple__26 = PyTuple_Pack(2, __pyx_int_4294967296, __pyx_int_8589934592); if (unlikely(!__pyx_tuple__26)) __PYX_ERR(0, 2291, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__26);
  __Pyx_GIVEREF(__pyx_tuple__26);

  /* "serializor/binary/ser.py":2061
 * 
 * 
 * @cython.ccall             # <<<<<<<<<<<<<<
 * def serialize(obj: object) -> bytes:
 *     """Serialize an object to `<'bytes'>`.
 */
  __pyx_tuple__28 = PyTuple_Pack(1, __pyx_n_s_obj); if (unlikely(!__pyx_tuple__28)) __PYX_ERR(0, 2061, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__28);
  __Pyx_GIVEREF(__pyx_tuple__28);
  __pyx_codeobj__29 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__28, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_serializor_binary_ser_py, __pyx_n_s_serialize, 2061, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__29)) __PYX_ERR(0, 2061, __pyx_L1_error)

  /* "serializor/binary/ser.py":2117
 * 
 * ########## The following functions are for testing purpose only ##########
 * def _test_utils() -> None:             # <<<<<<<<<<<<<<
 *     _test_pack_int8()
 *     _test_pack_uint8()
 */
  __pyx_codeobj__30 = (PyObject*)__Pyx_PyCode_New(0, 0, 0, 0, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_serializor_binary_ser_py, __pyx_n_s_test_utils, 2117, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__30)) __PYX_ERR(0, 2117, __pyx_L1_error)

  /* "serializor/binary/ser.py":2132
 * 
 * 
 * def _test_pack_int8() -> None:             # <<<<<<<<<<<<<<
 *     import struct
 * 
 */
  __pyx_tuple__31 = PyTuple_Pack(4, __pyx_n_s_struct, __pyx_n_s_val, __pyx_n_s_s, __pyx_n_s_b_2); if (unlikely(!__pyx_tuple__31)) __PYX_ERR(0, 2132, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__31);
  __Pyx_GIVEREF(__pyx_tuple__31);
  __pyx_codeobj__32 = (PyObject*)__Pyx_PyCode_New(0, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__31, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_serializor_binary_ser_py, __pyx_n_s_test_pack_int8, 2132, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__32)) __PYX_ERR(0, 2132, __pyx_L1_error)

  /* "serializor/binary/ser.py":2145
 * 
 * 
 * def _test_pack_uint8() -> None:             # <<<<<<<<<<<<<<
 *     import struct
 * 
 */
  __pyx_codeobj__33 = (PyObject*)__Pyx_PyCode_New(0, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__31, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_serializor_binary_ser_py, __pyx_n_s_test_pack_uint8, 2145, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__33)) __PYX_ERR(0, 2145, __pyx_L1_error)

  /* "serializor/binary/ser.py":2158
 * 
 * 
 * def _test_pack_int16() -> None:             # <<<<<<<<<<<<<<
 *     import struct
 * 
 */
  __pyx_codeobj__34 = (PyObject*)__Pyx_PyCode_New(0, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__31, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_serializor_binary_ser_py, __pyx_n_s_test_pack_int16, 2158, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__34)) __PYX_ERR(0, 2158, __pyx_L1_error)

  /* "serializor/binary/ser.py":2171
 * 
 * 
 * def _test_pack_uint16() -> None:             # <<<<<<<<<<<<<<
 *     import struct
 * 
 */
  __pyx_codeobj__35 = (PyObject*)__Pyx_PyCode_New(0, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__31, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_serializor_binary_ser_py, __pyx_n_s_test_pack_uint16, 2171, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__35)) __PYX_ERR(0, 2171, __pyx_L1_error)

  /* "serializor/binary/ser.py":2184
 * 
 * 
 * def _test_pack_int24() -> None:             # <<<<<<<<<<<<<<
 *     import struct
 * 
 */
  __pyx_codeobj__36 = (PyObject*)__Pyx_PyCode_New(0, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__31, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_serializor_binary_ser_py, __pyx_n_s_test_pack_int24, 2184, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__36)) __PYX_ERR(0, 2184, __pyx_L1_error)

  /* "serializor/binary/ser.py":2197
 * 
 * 
 * def _test_pack_uint24() -> None:             # <<<<<<<<<<<<<<
 *     import struct
 * 
 */
  __pyx_codeobj__37 = (PyObject*)__Pyx_PyCode_New(0, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__31, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_serializor_binary_ser_py, __pyx_n_s_test_pack_uint24, 2197, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__37)) __PYX_ERR(0, 2197, __pyx_L1_error)

  /* "serializor/binary/ser.py":2210
 * 
 * 
 * def _test_pack_int32() -> None:             # <<<<<<<<<<<<<<
 *     import struct
 * 
 */
  __pyx_codeobj__38 = (PyObject*)__Pyx_PyCode_New(0, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__31, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_serializor_binary_ser_py, __pyx_n_s_test_pack_int32, 2210, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__38)) __PYX_ERR(0, 2210, __pyx_L1_error)

  /* "serializor/binary/ser.py":2223
 * 
 * 
 * def _test_pack_uint32() -> None:             # <<<<<<<<<<<<<<
 *     import struct
 * 
 */
  __pyx_codeobj__39 = (PyObject*)__Pyx_PyCode_New(0, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__31, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_serializor_binary_ser_py, __pyx_n_s_test_pack_uint32, 2223, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__39)) __PYX_ERR(0, 2223, __pyx_L1_error)

  /* "serializor/binary/ser.py":2236
 * 
 * 
 * def _test_pack_int64() -> None:             # <<<<<<<<<<<<<<
 *     import struct
 * 
 */
  __pyx_codeobj__40 = (PyObject*)__Pyx_PyCode_New(0, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__31, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_serializor_binary_ser_py, __pyx_n_s_test_pack_int64, 2236, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__40)) __PYX_ERR(0, 2236, __pyx_L1_error)

  /* "serializor/binary/ser.py":2255
 * 
 * 
 * def _test_pack_uint64() -> None:             # <<<<<<<<<<<<<<
 *     import struct
 * 
 */
  __pyx_codeobj__41 = (PyObject*)__Pyx_PyCode_New(0, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__31, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_serializor_binary_ser_py, __pyx_n_s_test_pack_uint64, 2255, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__41)) __PYX_ERR(0, 2255, __pyx_L1_error)

  /* "serializor/binary/ser.py":2274
 * 
 * 
 * def _test_gen_encoded_int() -> None:             # <<<<<<<<<<<<<<
 *     # Test gen_encoded_int
 *     num = 251
 */
  __pyx_tuple__42 = PyTuple_Pack(3, __pyx_n_s_num_2, __pyx_n_s_v, __pyx_n_s_n); if (unlikely(!__pyx_tuple__42)) __PYX_ERR(0, 2274, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__42);
  __Pyx_GIVEREF(__pyx_tuple__42);
  __pyx_codeobj__43 = (PyObject*)__Pyx_PyCode_New(0, 0, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__42, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_serializor_binary_ser_py, __pyx_n_s_test_gen_encoded_int, 2274, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__43)) __PYX_ERR(0, 2274, __pyx_L1_error)

  /* "serializor/binary/ser.py":2298
 * 
 * 
 * def _test_gen_header() -> None:             # <<<<<<<<<<<<<<
 *     dtype: bytes = b"\x01"
 *     dchar: cython.uchar = 1
 */
  __pyx_tuple__44 = PyTuple_Pack(5, __pyx_n_s_dtype, __pyx_n_s_dchar, __pyx_n_s_num_2, __pyx_n_s_v, __pyx_n_s_n); if (unlikely(!__pyx_tuple__44)) __PYX_ERR(0, 2298, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__44);
  __Pyx_GIVEREF(__pyx_tuple__44);
  __pyx_codeobj__45 = (PyObject*)__Pyx_PyCode_New(0, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__44, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_serializor_binary_ser_py, __pyx_n_s_test_gen_header, 2298, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__45)) __PYX_ERR(0, 2298, __pyx_L1_error)
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}
/* #### Code section: init_constants ### */

static CYTHON_SMALL_CODE int __Pyx_InitConstants(void) {
  __pyx_umethod_PyMemoryView_Type_tobytes.type = (PyObject*)&PyMemoryView_Type;
  __pyx_umethod_PyMemoryView_Type_tobytes.method_name = &__pyx_n_s_tobytes;
  if (__Pyx_CreateStringTabAndInitStrings() < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  __pyx_int_0 = PyInt_FromLong(0); if (unlikely(!__pyx_int_0)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_1 = PyInt_FromLong(1); if (unlikely(!__pyx_int_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_3 = PyInt_FromLong(3); if (unlikely(!__pyx_int_3)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_80 = PyInt_FromLong(80); if (unlikely(!__pyx_int_80)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_251 = PyInt_FromLong(251); if (unlikely(!__pyx_int_251)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_252 = PyInt_FromLong(252); if (unlikely(!__pyx_int_252)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_65535 = PyInt_FromLong(65535L); if (unlikely(!__pyx_int_65535)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_65536 = PyInt_FromLong(65536L); if (unlikely(!__pyx_int_65536)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_8388606 = PyInt_FromLong(8388606L); if (unlikely(!__pyx_int_8388606)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_8388607 = PyInt_FromLong(8388607L); if (unlikely(!__pyx_int_8388607)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_16777213 = PyInt_FromLong(16777213L); if (unlikely(!__pyx_int_16777213)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_16777214 = PyInt_FromLong(16777214L); if (unlikely(!__pyx_int_16777214)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_16777215 = PyInt_FromLong(16777215L); if (unlikely(!__pyx_int_16777215)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_2147483646 = PyInt_FromLong(2147483646L); if (unlikely(!__pyx_int_2147483646)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_2147483647 = PyInt_FromLong(2147483647L); if (unlikely(!__pyx_int_2147483647)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_4294967293 = PyInt_FromString((char *)"4294967293", 0, 0); if (unlikely(!__pyx_int_4294967293)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_4294967294 = PyInt_FromString((char *)"4294967294", 0, 0); if (unlikely(!__pyx_int_4294967294)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_4294967295 = PyInt_FromString((char *)"4294967295", 0, 0); if (unlikely(!__pyx_int_4294967295)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_4294967296 = PyInt_FromString((char *)"4294967296", 0, 0); if (unlikely(!__pyx_int_4294967296)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_8589934592 = PyInt_FromString((char *)"8589934592", 0, 0); if (unlikely(!__pyx_int_8589934592)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_0x7ffffffffffffffe = PyInt_FromString((char *)"0x7ffffffffffffffe", 0, 0); if (unlikely(!__pyx_int_0x7ffffffffffffffe)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_0x7fffffffffffffff = PyInt_FromString((char *)"0x7fffffffffffffff", 0, 0); if (unlikely(!__pyx_int_0x7fffffffffffffff)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_0xfffffffffffffffd = PyInt_FromString((char *)"0xfffffffffffffffd", 0, 0); if (unlikely(!__pyx_int_0xfffffffffffffffd)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_0xfffffffffffffffe = PyInt_FromString((char *)"0xfffffffffffffffe", 0, 0); if (unlikely(!__pyx_int_0xfffffffffffffffe)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_0xffffffffffffffff = PyInt_FromString((char *)"0xffffffffffffffff", 0, 0); if (unlikely(!__pyx_int_0xffffffffffffffff)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_neg_8388607 = PyInt_FromLong(-8388607L); if (unlikely(!__pyx_int_neg_8388607)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_neg_8388608 = PyInt_FromLong(-8388608L); if (unlikely(!__pyx_int_neg_8388608)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_neg_2147483647 = PyInt_FromLong(-2147483647L); if (unlikely(!__pyx_int_neg_2147483647)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_neg_2147483648 = PyInt_FromLong(-2147483648L); if (unlikely(!__pyx_int_neg_2147483648)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_neg_9223372036854775807 = PyInt_FromString((char *)"-9223372036854775807", 0, 0); if (unlikely(!__pyx_int_neg_9223372036854775807)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_neg_9223372036854775808 = PyInt_FromString((char *)"-9223372036854775808", 0, 0); if (unlikely(!__pyx_int_neg_9223372036854775808)) __PYX_ERR(0, 1, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}
/* #### Code section: init_globals ### */

static CYTHON_SMALL_CODE int __Pyx_InitGlobals(void) {
  /* AssertionsEnabled.init */
  if (likely(__Pyx_init_assertions_enabled() == 0)); else

if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1, __pyx_L1_error)

  return 0;
  __pyx_L1_error:;
  return -1;
}
/* #### Code section: init_module ### */

static CYTHON_SMALL_CODE int __Pyx_modinit_global_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_import_code(void); /*proto*/

static int __Pyx_modinit_global_init_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_global_init_code", 0);
  /*--- Global init code ---*/
  __pyx_v_10serializor_6binary_3ser_OPT_SERIALIZE_NUMPY = Py_None; Py_INCREF(Py_None);
  __pyx_v_10serializor_6binary_3ser_DATETIME_ID = ((PyObject*)Py_None); Py_INCREF(Py_None);
  __pyx_v_10serializor_6binary_3ser_DATE_ID = ((PyObject*)Py_None); Py_INCREF(Py_None);
  __pyx_v_10serializor_6binary_3ser_TIME_ID = ((PyObject*)Py_None); Py_INCREF(Py_None);
  __pyx_v_10serializor_6binary_3ser_TIMEDELTA_ID = ((PyObject*)Py_None); Py_INCREF(Py_None);
  __pyx_v_10serializor_6binary_3ser_STRUCT_TIME_ID = ((PyObject*)Py_None); Py_INCREF(Py_None);
  __pyx_v_10serializor_6binary_3ser_COMPLEX_ID = ((PyObject*)Py_None); Py_INCREF(Py_None);
  __pyx_v_10serializor_6binary_3ser_LIST_ID = ((PyObject*)Py_None); Py_INCREF(Py_None);
  __pyx_v_10serializor_6binary_3ser_TUPLE_ID = ((PyObject*)Py_None); Py_INCREF(Py_None);
  __pyx_v_10serializor_6binary_3ser_SET_ID = ((PyObject*)Py_None); Py_INCREF(Py_None);
  __pyx_v_10serializor_6binary_3ser_FROZENSET_ID = ((PyObject*)Py_None); Py_INCREF(Py_None);
  __pyx_v_10serializor_6binary_3ser_RANGE_ID = ((PyObject*)Py_None); Py_INCREF(Py_None);
  __pyx_v_10serializor_6binary_3ser_DICT_ID = ((PyObject*)Py_None); Py_INCREF(Py_None);
  __pyx_v_10serializor_6binary_3ser_DT64_ID = ((PyObject*)Py_None); Py_INCREF(Py_None);
  __pyx_v_10serializor_6binary_3ser_TD64_ID = ((PyObject*)Py_None); Py_INCREF(Py_None);
  __pyx_v_10serializor_6binary_3ser_NDARRAY_ID = ((PyObject*)Py_None); Py_INCREF(Py_None);
  __pyx_v_10serializor_6binary_3ser_NDARRAY_OBJECT_DT = ((PyObject*)Py_None); Py_INCREF(Py_None);
  __pyx_v_10serializor_6binary_3ser_NDARRAY_INT_DT = ((PyObject*)Py_None); Py_INCREF(Py_None);
  __pyx_v_10serializor_6binary_3ser_NDARRAY_UINT_DT = ((PyObject*)Py_None); Py_INCREF(Py_None);
  __pyx_v_10serializor_6binary_3ser_NDARRAY_FLOAT_DT = ((PyObject*)Py_None); Py_INCREF(Py_None);
  __pyx_v_10serializor_6binary_3ser_NDARRAY_BOOL_DT = ((PyObject*)Py_None); Py_INCREF(Py_None);
  __pyx_v_10serializor_6binary_3ser_NDARRAY_DT64_DT = ((PyObject*)Py_None); Py_INCREF(Py_None);
  __pyx_v_10serializor_6binary_3ser_NDARRAY_TD64_DT = ((PyObject*)Py_None); Py_INCREF(Py_None);
  __pyx_v_10serializor_6binary_3ser_NDARRAY_COMPLEX_DT = ((PyObject*)Py_None); Py_INCREF(Py_None);
  __pyx_v_10serializor_6binary_3ser_NDARRAY_BYTES_DT = ((PyObject*)Py_None); Py_INCREF(Py_None);
  __pyx_v_10serializor_6binary_3ser_NDARRAY_UNICODE_DT = ((PyObject*)Py_None); Py_INCREF(Py_None);
  __pyx_v_10serializor_6binary_3ser_NDARRAY_OBJECT_IDDT = ((PyObject*)Py_None); Py_INCREF(Py_None);
  __pyx_v_10serializor_6binary_3ser_NDARRAY_INT_IDDT = ((PyObject*)Py_None); Py_INCREF(Py_None);
  __pyx_v_10serializor_6binary_3ser_NDARRAY_UINT_IDDT = ((PyObject*)Py_None); Py_INCREF(Py_None);
  __pyx_v_10serializor_6binary_3ser_NDARRAY_FLOAT_IDDT = ((PyObject*)Py_None); Py_INCREF(Py_None);
  __pyx_v_10serializor_6binary_3ser_NDARRAY_BOOL_IDDT = ((PyObject*)Py_None); Py_INCREF(Py_None);
  __pyx_v_10serializor_6binary_3ser_NDARRAY_DT64_IDDT = ((PyObject*)Py_None); Py_INCREF(Py_None);
  __pyx_v_10serializor_6binary_3ser_NDARRAY_TD64_IDDT = ((PyObject*)Py_None); Py_INCREF(Py_None);
  __pyx_v_10serializor_6binary_3ser_NDARRAY_COMPLEX_IDDT = ((PyObject*)Py_None); Py_INCREF(Py_None);
  __pyx_v_10serializor_6binary_3ser_NDARRAY_BYTES_IDDT = ((PyObject*)Py_None); Py_INCREF(Py_None);
  __pyx_v_10serializor_6binary_3ser_NDARRAY_UNICODE_IDDT = ((PyObject*)Py_None); Py_INCREF(Py_None);
  __pyx_v_10serializor_6binary_3ser_PD_TIMESTAMP_ID = ((PyObject*)Py_None); Py_INCREF(Py_None);
  __pyx_v_10serializor_6binary_3ser_PD_TIMEDELTA_ID = ((PyObject*)Py_None); Py_INCREF(Py_None);
  __pyx_v_10serializor_6binary_3ser_SERIES_ID = ((PyObject*)Py_None); Py_INCREF(Py_None);
  __pyx_v_10serializor_6binary_3ser_DATETIMEINDEX_ID = ((PyObject*)Py_None); Py_INCREF(Py_None);
  __pyx_v_10serializor_6binary_3ser_TIMEDELTAINDEX_ID = ((PyObject*)Py_None); Py_INCREF(Py_None);
  __pyx_v_10serializor_6binary_3ser_DATAFRAME_ID = ((PyObject*)Py_None); Py_INCREF(Py_None);
  __pyx_v_10serializor_6binary_3ser_TRUE_VALUE = ((PyObject*)Py_None); Py_INCREF(Py_None);
  __pyx_v_10serializor_6binary_3ser_FALSE_VALUE = ((PyObject*)Py_None); Py_INCREF(Py_None);
  __pyx_v_10serializor_6binary_3ser_NONE_VALUE = ((PyObject*)Py_None); Py_INCREF(Py_None);
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_variable_export_code(void) {
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_export_code", 0);
  /*--- Variable export code ---*/
  if (__Pyx_ExportVoidPtr(__pyx_n_s_OPT_SERIALIZE_NUMPY, (void *)&__pyx_v_10serializor_6binary_3ser_OPT_SERIALIZE_NUMPY, "PyObject *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportVoidPtr(__pyx_n_s_UINT8_ENCODE_VALUE, (void *)&__pyx_v_10serializor_6binary_3ser_UINT8_ENCODE_VALUE, "unsigned char") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportVoidPtr(__pyx_n_s_UINT16_ENCODE_VALUE, (void *)&__pyx_v_10serializor_6binary_3ser_UINT16_ENCODE_VALUE, "unsigned char") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportVoidPtr(__pyx_n_s_UINT32_ENCODE_VALUE, (void *)&__pyx_v_10serializor_6binary_3ser_UINT32_ENCODE_VALUE, "unsigned char") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportVoidPtr(__pyx_n_s_UINT64_ENCODE_VALUE, (void *)&__pyx_v_10serializor_6binary_3ser_UINT64_ENCODE_VALUE, "unsigned char") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportVoidPtr(__pyx_n_s_DATETIME_ID, (void *)&__pyx_v_10serializor_6binary_3ser_DATETIME_ID, "PyObject *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportVoidPtr(__pyx_n_s_DATE_ID, (void *)&__pyx_v_10serializor_6binary_3ser_DATE_ID, "PyObject *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportVoidPtr(__pyx_n_s_TIME_ID, (void *)&__pyx_v_10serializor_6binary_3ser_TIME_ID, "PyObject *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportVoidPtr(__pyx_n_s_TIMEDELTA_ID, (void *)&__pyx_v_10serializor_6binary_3ser_TIMEDELTA_ID, "PyObject *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportVoidPtr(__pyx_n_s_STRUCT_TIME_ID, (void *)&__pyx_v_10serializor_6binary_3ser_STRUCT_TIME_ID, "PyObject *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportVoidPtr(__pyx_n_s_COMPLEX_ID, (void *)&__pyx_v_10serializor_6binary_3ser_COMPLEX_ID, "PyObject *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportVoidPtr(__pyx_n_s_LIST_ID, (void *)&__pyx_v_10serializor_6binary_3ser_LIST_ID, "PyObject *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportVoidPtr(__pyx_n_s_TUPLE_ID, (void *)&__pyx_v_10serializor_6binary_3ser_TUPLE_ID, "PyObject *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportVoidPtr(__pyx_n_s_SET_ID, (void *)&__pyx_v_10serializor_6binary_3ser_SET_ID, "PyObject *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportVoidPtr(__pyx_n_s_FROZENSET_ID, (void *)&__pyx_v_10serializor_6binary_3ser_FROZENSET_ID, "PyObject *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportVoidPtr(__pyx_n_s_RANGE_ID, (void *)&__pyx_v_10serializor_6binary_3ser_RANGE_ID, "PyObject *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportVoidPtr(__pyx_n_s_DICT_ID, (void *)&__pyx_v_10serializor_6binary_3ser_DICT_ID, "PyObject *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportVoidPtr(__pyx_n_s_DT64_ID, (void *)&__pyx_v_10serializor_6binary_3ser_DT64_ID, "PyObject *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportVoidPtr(__pyx_n_s_TD64_ID, (void *)&__pyx_v_10serializor_6binary_3ser_TD64_ID, "PyObject *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportVoidPtr(__pyx_n_s_NDARRAY_ID, (void *)&__pyx_v_10serializor_6binary_3ser_NDARRAY_ID, "PyObject *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportVoidPtr(__pyx_n_s_NDARRAY_OBJECT_DT, (void *)&__pyx_v_10serializor_6binary_3ser_NDARRAY_OBJECT_DT, "PyObject *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportVoidPtr(__pyx_n_s_NDARRAY_INT_DT, (void *)&__pyx_v_10serializor_6binary_3ser_NDARRAY_INT_DT, "PyObject *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportVoidPtr(__pyx_n_s_NDARRAY_UINT_DT, (void *)&__pyx_v_10serializor_6binary_3ser_NDARRAY_UINT_DT, "PyObject *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportVoidPtr(__pyx_n_s_NDARRAY_FLOAT_DT, (void *)&__pyx_v_10serializor_6binary_3ser_NDARRAY_FLOAT_DT, "PyObject *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportVoidPtr(__pyx_n_s_NDARRAY_BOOL_DT, (void *)&__pyx_v_10serializor_6binary_3ser_NDARRAY_BOOL_DT, "PyObject *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportVoidPtr(__pyx_n_s_NDARRAY_DT64_DT, (void *)&__pyx_v_10serializor_6binary_3ser_NDARRAY_DT64_DT, "PyObject *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportVoidPtr(__pyx_n_s_NDARRAY_TD64_DT, (void *)&__pyx_v_10serializor_6binary_3ser_NDARRAY_TD64_DT, "PyObject *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportVoidPtr(__pyx_n_s_NDARRAY_COMPLEX_DT, (void *)&__pyx_v_10serializor_6binary_3ser_NDARRAY_COMPLEX_DT, "PyObject *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportVoidPtr(__pyx_n_s_NDARRAY_BYTES_DT, (void *)&__pyx_v_10serializor_6binary_3ser_NDARRAY_BYTES_DT, "PyObject *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportVoidPtr(__pyx_n_s_NDARRAY_UNICODE_DT, (void *)&__pyx_v_10serializor_6binary_3ser_NDARRAY_UNICODE_DT, "PyObject *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportVoidPtr(__pyx_n_s_NDARRAY_OBJECT_IDDT, (void *)&__pyx_v_10serializor_6binary_3ser_NDARRAY_OBJECT_IDDT, "PyObject *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportVoidPtr(__pyx_n_s_NDARRAY_INT_IDDT, (void *)&__pyx_v_10serializor_6binary_3ser_NDARRAY_INT_IDDT, "PyObject *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportVoidPtr(__pyx_n_s_NDARRAY_UINT_IDDT, (void *)&__pyx_v_10serializor_6binary_3ser_NDARRAY_UINT_IDDT, "PyObject *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportVoidPtr(__pyx_n_s_NDARRAY_FLOAT_IDDT, (void *)&__pyx_v_10serializor_6binary_3ser_NDARRAY_FLOAT_IDDT, "PyObject *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportVoidPtr(__pyx_n_s_NDARRAY_BOOL_IDDT, (void *)&__pyx_v_10serializor_6binary_3ser_NDARRAY_BOOL_IDDT, "PyObject *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportVoidPtr(__pyx_n_s_NDARRAY_DT64_IDDT, (void *)&__pyx_v_10serializor_6binary_3ser_NDARRAY_DT64_IDDT, "PyObject *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportVoidPtr(__pyx_n_s_NDARRAY_TD64_IDDT, (void *)&__pyx_v_10serializor_6binary_3ser_NDARRAY_TD64_IDDT, "PyObject *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportVoidPtr(__pyx_n_s_NDARRAY_COMPLEX_IDDT, (void *)&__pyx_v_10serializor_6binary_3ser_NDARRAY_COMPLEX_IDDT, "PyObject *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportVoidPtr(__pyx_n_s_NDARRAY_BYTES_IDDT, (void *)&__pyx_v_10serializor_6binary_3ser_NDARRAY_BYTES_IDDT, "PyObject *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportVoidPtr(__pyx_n_s_NDARRAY_UNICODE_IDDT, (void *)&__pyx_v_10serializor_6binary_3ser_NDARRAY_UNICODE_IDDT, "PyObject *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportVoidPtr(__pyx_n_s_PD_TIMESTAMP_ID, (void *)&__pyx_v_10serializor_6binary_3ser_PD_TIMESTAMP_ID, "PyObject *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportVoidPtr(__pyx_n_s_PD_TIMEDELTA_ID, (void *)&__pyx_v_10serializor_6binary_3ser_PD_TIMEDELTA_ID, "PyObject *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportVoidPtr(__pyx_n_s_SERIES_ID, (void *)&__pyx_v_10serializor_6binary_3ser_SERIES_ID, "PyObject *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportVoidPtr(__pyx_n_s_DATETIMEINDEX_ID, (void *)&__pyx_v_10serializor_6binary_3ser_DATETIMEINDEX_ID, "PyObject *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportVoidPtr(__pyx_n_s_TIMEDELTAINDEX_ID, (void *)&__pyx_v_10serializor_6binary_3ser_TIMEDELTAINDEX_ID, "PyObject *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportVoidPtr(__pyx_n_s_DATAFRAME_ID, (void *)&__pyx_v_10serializor_6binary_3ser_DATAFRAME_ID, "PyObject *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportVoidPtr(__pyx_n_s_TRUE_VALUE, (void *)&__pyx_v_10serializor_6binary_3ser_TRUE_VALUE, "PyObject *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportVoidPtr(__pyx_n_s_FALSE_VALUE, (void *)&__pyx_v_10serializor_6binary_3ser_FALSE_VALUE, "PyObject *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportVoidPtr(__pyx_n_s_NONE_VALUE, (void *)&__pyx_v_10serializor_6binary_3ser_NONE_VALUE, "PyObject *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_function_export_code(void) {
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_export_code", 0);
  /*--- Function export code ---*/
  if (__Pyx_ExportFunction("serialize", (void (*)(void))__pyx_f_10serializor_6binary_3ser_serialize, "PyObject *(PyObject *, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_type_init_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_init_code", 0);
  /*--- Type init code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_type_import_code(void) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_import_code", 0);
  /*--- Type import code ---*/
  __pyx_t_1 = PyImport_ImportModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 9, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7cpython_4type_type = __Pyx_ImportType_3_0_12(__pyx_t_1, __Pyx_BUILTIN_MODULE_NAME, "type", 
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(PyTypeObject), __PYX_GET_STRUCT_ALIGNMENT_3_0_12(PyTypeObject),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(PyTypeObject), __PYX_GET_STRUCT_ALIGNMENT_3_0_12(PyTypeObject),
  #else
  sizeof(PyHeapTypeObject), __PYX_GET_STRUCT_ALIGNMENT_3_0_12(PyHeapTypeObject),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_0_12); if (!__pyx_ptype_7cpython_4type_type) __PYX_ERR(6, 9, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_t_1)) __PYX_ERR(7, 8, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7cpython_4bool_bool = __Pyx_ImportType_3_0_12(__pyx_t_1, __Pyx_BUILTIN_MODULE_NAME, "bool", sizeof(PyBoolObject), __PYX_GET_STRUCT_ALIGNMENT_3_0_12(PyBoolObject),__Pyx_ImportType_CheckSize_Warn_3_0_12); if (!__pyx_ptype_7cpython_4bool_bool) __PYX_ERR(7, 8, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_t_1)) __PYX_ERR(8, 15, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7cpython_7complex_complex = __Pyx_ImportType_3_0_12(__pyx_t_1, __Pyx_BUILTIN_MODULE_NAME, "complex", sizeof(PyComplexObject), __PYX_GET_STRUCT_ALIGNMENT_3_0_12(PyComplexObject),__Pyx_ImportType_CheckSize_Warn_3_0_12); if (!__pyx_ptype_7cpython_7complex_complex) __PYX_ERR(8, 15, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("numpy"); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 228, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_5numpy_dtype = __Pyx_ImportType_3_0_12(__pyx_t_1, "numpy", "dtype", sizeof(PyArray_Descr), __PYX_GET_STRUCT_ALIGNMENT_3_0_12(PyArray_Descr),__Pyx_ImportType_CheckSize_Ignore_3_0_12); if (!__pyx_ptype_5numpy_dtype) __PYX_ERR(2, 228, __pyx_L1_error)
  __pyx_ptype_5numpy_flatiter = __Pyx_ImportType_3_0_12(__pyx_t_1, "numpy", "flatiter", sizeof(PyArrayIterObject), __PYX_GET_STRUCT_ALIGNMENT_3_0_12(PyArrayIterObject),__Pyx_ImportType_CheckSize_Ignore_3_0_12); if (!__pyx_ptype_5numpy_flatiter) __PYX_ERR(2, 273, __pyx_L1_error)
  __pyx_ptype_5numpy_broadcast = __Pyx_ImportType_3_0_12(__pyx_t_1, "numpy", "broadcast", sizeof(PyArrayMultiIterObject), __PYX_GET_STRUCT_ALIGNMENT_3_0_12(PyArrayMultiIterObject),__Pyx_ImportType_CheckSize_Ignore_3_0_12); if (!__pyx_ptype_5numpy_broadcast) __PYX_ERR(2, 277, __pyx_L1_error)
  __pyx_ptype_5numpy_ndarray = __Pyx_ImportType_3_0_12(__pyx_t_1, "numpy", "ndarray", sizeof(PyArrayObject), __PYX_GET_STRUCT_ALIGNMENT_3_0_12(PyArrayObject),__Pyx_ImportType_CheckSize_Ignore_3_0_12); if (!__pyx_ptype_5numpy_ndarray) __PYX_ERR(2, 316, __pyx_L1_error)
  __pyx_ptype_5numpy_generic = __Pyx_ImportType_3_0_12(__pyx_t_1, "numpy", "generic", sizeof(PyObject), __PYX_GET_STRUCT_ALIGNMENT_3_0_12(PyObject),__Pyx_ImportType_CheckSize_Warn_3_0_12); if (!__pyx_ptype_5numpy_generic) __PYX_ERR(2, 825, __pyx_L1_error)
  __pyx_ptype_5numpy_number = __Pyx_ImportType_3_0_12(__pyx_t_1, "numpy", "number", sizeof(PyObject), __PYX_GET_STRUCT_ALIGNMENT_3_0_12(PyObject),__Pyx_ImportType_CheckSize_Warn_3_0_12); if (!__pyx_ptype_5numpy_number) __PYX_ERR(2, 827, __pyx_L1_error)
  __pyx_ptype_5numpy_integer = __Pyx_ImportType_3_0_12(__pyx_t_1, "numpy", "integer", sizeof(PyObject), __PYX_GET_STRUCT_ALIGNMENT_3_0_12(PyObject),__Pyx_ImportType_CheckSize_Warn_3_0_12); if (!__pyx_ptype_5numpy_integer) __PYX_ERR(2, 829, __pyx_L1_error)
  __pyx_ptype_5numpy_signedinteger = __Pyx_ImportType_3_0_12(__pyx_t_1, "numpy", "signedinteger", sizeof(PyObject), __PYX_GET_STRUCT_ALIGNMENT_3_0_12(PyObject),__Pyx_ImportType_CheckSize_Warn_3_0_12); if (!__pyx_ptype_5numpy_signedinteger) __PYX_ERR(2, 831, __pyx_L1_error)
  __pyx_ptype_5numpy_unsignedinteger = __Pyx_ImportType_3_0_12(__pyx_t_1, "numpy", "unsignedinteger", sizeof(PyObject), __PYX_GET_STRUCT_ALIGNMENT_3_0_12(PyObject),__Pyx_ImportType_CheckSize_Warn_3_0_12); if (!__pyx_ptype_5numpy_unsignedinteger) __PYX_ERR(2, 833, __pyx_L1_error)
  __pyx_ptype_5numpy_inexact = __Pyx_ImportType_3_0_12(__pyx_t_1, "numpy", "inexact", sizeof(PyObject), __PYX_GET_STRUCT_ALIGNMENT_3_0_12(PyObject),__Pyx_ImportType_CheckSize_Warn_3_0_12); if (!__pyx_ptype_5numpy_inexact) __PYX_ERR(2, 835, __pyx_L1_error)
  __pyx_ptype_5numpy_floating = __Pyx_ImportType_3_0_12(__pyx_t_1, "numpy", "floating", sizeof(PyObject), __PYX_GET_STRUCT_ALIGNMENT_3_0_12(PyObject),__Pyx_ImportType_CheckSize_Warn_3_0_12); if (!__pyx_ptype_5numpy_floating) __PYX_ERR(2, 837, __pyx_L1_error)
  __pyx_ptype_5numpy_complexfloating = __Pyx_ImportType_3_0_12(__pyx_t_1, "numpy", "complexfloating", sizeof(PyObject), __PYX_GET_STRUCT_ALIGNMENT_3_0_12(PyObject),__Pyx_ImportType_CheckSize_Warn_3_0_12); if (!__pyx_ptype_5numpy_complexfloating) __PYX_ERR(2, 839, __pyx_L1_error)
  __pyx_ptype_5numpy_flexible = __Pyx_ImportType_3_0_12(__pyx_t_1, "numpy", "flexible", sizeof(PyObject), __PYX_GET_STRUCT_ALIGNMENT_3_0_12(PyObject),__Pyx_ImportType_CheckSize_Warn_3_0_12); if (!__pyx_ptype_5numpy_flexible) __PYX_ERR(2, 841, __pyx_L1_error)
  __pyx_ptype_5numpy_character = __Pyx_ImportType_3_0_12(__pyx_t_1, "numpy", "character", sizeof(PyObject), __PYX_GET_STRUCT_ALIGNMENT_3_0_12(PyObject),__Pyx_ImportType_CheckSize_Warn_3_0_12); if (!__pyx_ptype_5numpy_character) __PYX_ERR(2, 843, __pyx_L1_error)
  __pyx_ptype_5numpy_ufunc = __Pyx_ImportType_3_0_12(__pyx_t_1, "numpy", "ufunc", sizeof(PyUFuncObject), __PYX_GET_STRUCT_ALIGNMENT_3_0_12(PyUFuncObject),__Pyx_ImportType_CheckSize_Ignore_3_0_12); if (!__pyx_ptype_5numpy_ufunc) __PYX_ERR(2, 907, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("datetime"); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 70, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7cpython_8datetime_date = __Pyx_ImportType_3_0_12(__pyx_t_1, "datetime", "date", sizeof(PyDateTime_Date), __PYX_GET_STRUCT_ALIGNMENT_3_0_12(PyDateTime_Date),__Pyx_ImportType_CheckSize_Warn_3_0_12); if (!__pyx_ptype_7cpython_8datetime_date) __PYX_ERR(4, 70, __pyx_L1_error)
  __pyx_ptype_7cpython_8datetime_time = __Pyx_ImportType_3_0_12(__pyx_t_1, "datetime", "time", sizeof(PyDateTime_Time), __PYX_GET_STRUCT_ALIGNMENT_3_0_12(PyDateTime_Time),__Pyx_ImportType_CheckSize_Warn_3_0_12); if (!__pyx_ptype_7cpython_8datetime_time) __PYX_ERR(4, 83, __pyx_L1_error)
  __pyx_ptype_7cpython_8datetime_datetime = __Pyx_ImportType_3_0_12(__pyx_t_1, "datetime", "datetime", sizeof(PyDateTime_DateTime), __PYX_GET_STRUCT_ALIGNMENT_3_0_12(PyDateTime_DateTime),__Pyx_ImportType_CheckSize_Warn_3_0_12); if (!__pyx_ptype_7cpython_8datetime_datetime) __PYX_ERR(4, 109, __pyx_L1_error)
  __pyx_ptype_7cpython_8datetime_timedelta = __Pyx_ImportType_3_0_12(__pyx_t_1, "datetime", "timedelta", sizeof(PyDateTime_Delta), __PYX_GET_STRUCT_ALIGNMENT_3_0_12(PyDateTime_Delta),__Pyx_ImportType_CheckSize_Warn_3_0_12); if (!__pyx_ptype_7cpython_8datetime_timedelta) __PYX_ERR(4, 147, __pyx_L1_error)
  __pyx_ptype_7cpython_8datetime_tzinfo = __Pyx_ImportType_3_0_12(__pyx_t_1, "datetime", "tzinfo", sizeof(PyDateTime_TZInfo), __PYX_GET_STRUCT_ALIGNMENT_3_0_12(PyDateTime_TZInfo),__Pyx_ImportType_CheckSize_Warn_3_0_12); if (!__pyx_ptype_7cpython_8datetime_tzinfo) __PYX_ERR(4, 160, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_variable_import_code(void) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_import_code", 0);
  /*--- Variable import code ---*/
  __pyx_t_1 = PyImport_ImportModule("serializor.identifier"); if (!__pyx_t_1) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_ImportVoidPtr_3_0_12(__pyx_t_1, "STR", (void **)&__pyx_vp_10serializor_10identifier_STR, "char") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr_3_0_12(__pyx_t_1, "BOOL", (void **)&__pyx_vp_10serializor_10identifier_BOOL, "char") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr_3_0_12(__pyx_t_1, "FLOAT", (void **)&__pyx_vp_10serializor_10identifier_FLOAT, "char") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr_3_0_12(__pyx_t_1, "INT", (void **)&__pyx_vp_10serializor_10identifier_INT, "char") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr_3_0_12(__pyx_t_1, "NONE", (void **)&__pyx_vp_10serializor_10identifier_NONE, "char") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr_3_0_12(__pyx_t_1, "DATE", (void **)&__pyx_vp_10serializor_10identifier_DATE, "char") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr_3_0_12(__pyx_t_1, "TIME", (void **)&__pyx_vp_10serializor_10identifier_TIME, "char") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr_3_0_12(__pyx_t_1, "DATETIME", (void **)&__pyx_vp_10serializor_10identifier_DATETIME, "char") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr_3_0_12(__pyx_t_1, "TIMEDELTA", (void **)&__pyx_vp_10serializor_10identifier_TIMEDELTA, "char") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr_3_0_12(__pyx_t_1, "STRUCT_TIME", (void **)&__pyx_vp_10serializor_10identifier_STRUCT_TIME, "char") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr_3_0_12(__pyx_t_1, "DECIMAL", (void **)&__pyx_vp_10serializor_10identifier_DECIMAL, "char") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr_3_0_12(__pyx_t_1, "COMPLEX", (void **)&__pyx_vp_10serializor_10identifier_COMPLEX, "char") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr_3_0_12(__pyx_t_1, "BYTES", (void **)&__pyx_vp_10serializor_10identifier_BYTES, "char") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr_3_0_12(__pyx_t_1, "BYTEARRAY", (void **)&__pyx_vp_10serializor_10identifier_BYTEARRAY, "char") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr_3_0_12(__pyx_t_1, "LIST", (void **)&__pyx_vp_10serializor_10identifier_LIST, "char") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr_3_0_12(__pyx_t_1, "TUPLE", (void **)&__pyx_vp_10serializor_10identifier_TUPLE, "char") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr_3_0_12(__pyx_t_1, "SET", (void **)&__pyx_vp_10serializor_10identifier_SET, "char") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr_3_0_12(__pyx_t_1, "FROZENSET", (void **)&__pyx_vp_10serializor_10identifier_FROZENSET, "char") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr_3_0_12(__pyx_t_1, "RANGE", (void **)&__pyx_vp_10serializor_10identifier_RANGE, "char") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr_3_0_12(__pyx_t_1, "DICT", (void **)&__pyx_vp_10serializor_10identifier_DICT, "char") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr_3_0_12(__pyx_t_1, "DATETIME64", (void **)&__pyx_vp_10serializor_10identifier_DATETIME64, "char") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr_3_0_12(__pyx_t_1, "TIMEDELTA64", (void **)&__pyx_vp_10serializor_10identifier_TIMEDELTA64, "char") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr_3_0_12(__pyx_t_1, "NDARRAY", (void **)&__pyx_vp_10serializor_10identifier_NDARRAY, "char") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr_3_0_12(__pyx_t_1, "NDARRAY_OBJECT", (void **)&__pyx_vp_10serializor_10identifier_NDARRAY_OBJECT, "char") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr_3_0_12(__pyx_t_1, "NDARRAY_INT", (void **)&__pyx_vp_10serializor_10identifier_NDARRAY_INT, "char") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr_3_0_12(__pyx_t_1, "NDARRAY_UINT", (void **)&__pyx_vp_10serializor_10identifier_NDARRAY_UINT, "char") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr_3_0_12(__pyx_t_1, "NDARRAY_FLOAT", (void **)&__pyx_vp_10serializor_10identifier_NDARRAY_FLOAT, "char") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr_3_0_12(__pyx_t_1, "NDARRAY_BOOL", (void **)&__pyx_vp_10serializor_10identifier_NDARRAY_BOOL, "char") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr_3_0_12(__pyx_t_1, "NDARRAY_DT64", (void **)&__pyx_vp_10serializor_10identifier_NDARRAY_DT64, "char") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr_3_0_12(__pyx_t_1, "NDARRAY_TD64", (void **)&__pyx_vp_10serializor_10identifier_NDARRAY_TD64, "char") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr_3_0_12(__pyx_t_1, "NDARRAY_COMPLEX", (void **)&__pyx_vp_10serializor_10identifier_NDARRAY_COMPLEX, "char") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr_3_0_12(__pyx_t_1, "NDARRAY_BYTES", (void **)&__pyx_vp_10serializor_10identifier_NDARRAY_BYTES, "char") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr_3_0_12(__pyx_t_1, "NDARRAY_UNICODE", (void **)&__pyx_vp_10serializor_10identifier_NDARRAY_UNICODE, "char") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr_3_0_12(__pyx_t_1, "SERIES", (void **)&__pyx_vp_10serializor_10identifier_SERIES, "char") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr_3_0_12(__pyx_t_1, "DATAFRAME", (void **)&__pyx_vp_10serializor_10identifier_DATAFRAME, "char") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr_3_0_12(__pyx_t_1, "PD_TIMESTAMP", (void **)&__pyx_vp_10serializor_10identifier_PD_TIMESTAMP, "char") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr_3_0_12(__pyx_t_1, "PD_TIMEDELTA", (void **)&__pyx_vp_10serializor_10identifier_PD_TIMEDELTA, "char") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr_3_0_12(__pyx_t_1, "DATETIMEINDEX", (void **)&__pyx_vp_10serializor_10identifier_DATETIMEINDEX, "char") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr_3_0_12(__pyx_t_1, "TIMEDELTAINDEX", (void **)&__pyx_vp_10serializor_10identifier_TIMEDELTAINDEX, "char") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr_3_0_12(__pyx_t_1, "_duplicates_test", (void **)&__pyx_vp_10serializor_10identifier__duplicates_test, "int") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("serializor.typeref"); if (!__pyx_t_1) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_ImportVoidPtr_3_0_12(__pyx_t_1, "NONE", (void **)&__pyx_vp_10serializor_7typeref_NONE, "PyObject *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr_3_0_12(__pyx_t_1, "DICT_KEYS", (void **)&__pyx_vp_10serializor_7typeref_DICT_KEYS, "PyObject *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr_3_0_12(__pyx_t_1, "DICT_VALUES", (void **)&__pyx_vp_10serializor_7typeref_DICT_VALUES, "PyObject *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr_3_0_12(__pyx_t_1, "DICT_ITEMS", (void **)&__pyx_vp_10serializor_7typeref_DICT_ITEMS, "PyObject *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr_3_0_12(__pyx_t_1, "DECIMAL", (void **)&__pyx_vp_10serializor_7typeref_DECIMAL, "PyObject *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr_3_0_12(__pyx_t_1, "STRUCT_TIME", (void **)&__pyx_vp_10serializor_7typeref_STRUCT_TIME, "PyObject *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr_3_0_12(__pyx_t_1, "DEQUE", (void **)&__pyx_vp_10serializor_7typeref_DEQUE, "PyObject *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr_3_0_12(__pyx_t_1, "FLOAT16", (void **)&__pyx_vp_10serializor_7typeref_FLOAT16, "PyObject *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr_3_0_12(__pyx_t_1, "FLOAT32", (void **)&__pyx_vp_10serializor_7typeref_FLOAT32, "PyObject *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr_3_0_12(__pyx_t_1, "FLOAT64", (void **)&__pyx_vp_10serializor_7typeref_FLOAT64, "PyObject *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr_3_0_12(__pyx_t_1, "INT8", (void **)&__pyx_vp_10serializor_7typeref_INT8, "PyObject *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr_3_0_12(__pyx_t_1, "INT16", (void **)&__pyx_vp_10serializor_7typeref_INT16, "PyObject *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr_3_0_12(__pyx_t_1, "INT32", (void **)&__pyx_vp_10serializor_7typeref_INT32, "PyObject *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr_3_0_12(__pyx_t_1, "INT64", (void **)&__pyx_vp_10serializor_7typeref_INT64, "PyObject *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr_3_0_12(__pyx_t_1, "UINT8", (void **)&__pyx_vp_10serializor_7typeref_UINT8, "PyObject *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr_3_0_12(__pyx_t_1, "UINT16", (void **)&__pyx_vp_10serializor_7typeref_UINT16, "PyObject *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr_3_0_12(__pyx_t_1, "UINT32", (void **)&__pyx_vp_10serializor_7typeref_UINT32, "PyObject *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr_3_0_12(__pyx_t_1, "UINT64", (void **)&__pyx_vp_10serializor_7typeref_UINT64, "PyObject *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr_3_0_12(__pyx_t_1, "COMPLEX64", (void **)&__pyx_vp_10serializor_7typeref_COMPLEX64, "PyObject *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr_3_0_12(__pyx_t_1, "COMPLEX128", (void **)&__pyx_vp_10serializor_7typeref_COMPLEX128, "PyObject *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr_3_0_12(__pyx_t_1, "DATETIME64", (void **)&__pyx_vp_10serializor_7typeref_DATETIME64, "PyObject *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr_3_0_12(__pyx_t_1, "TIMEDELTA64", (void **)&__pyx_vp_10serializor_7typeref_TIMEDELTA64, "PyObject *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr_3_0_12(__pyx_t_1, "STR_", (void **)&__pyx_vp_10serializor_7typeref_STR_, "PyObject *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr_3_0_12(__pyx_t_1, "BOOL_", (void **)&__pyx_vp_10serializor_7typeref_BOOL_, "PyObject *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr_3_0_12(__pyx_t_1, "BYTES_", (void **)&__pyx_vp_10serializor_7typeref_BYTES_, "PyObject *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr_3_0_12(__pyx_t_1, "RECORD", (void **)&__pyx_vp_10serializor_7typeref_RECORD, "PyObject *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr_3_0_12(__pyx_t_1, "SERIES", (void **)&__pyx_vp_10serializor_7typeref_SERIES, "PyObject *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr_3_0_12(__pyx_t_1, "DATAFRAME", (void **)&__pyx_vp_10serializor_7typeref_DATAFRAME, "PyObject *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr_3_0_12(__pyx_t_1, "PD_TIMESTAMP", (void **)&__pyx_vp_10serializor_7typeref_PD_TIMESTAMP, "PyObject *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr_3_0_12(__pyx_t_1, "PD_TIMEDELTA", (void **)&__pyx_vp_10serializor_7typeref_PD_TIMEDELTA, "PyObject *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr_3_0_12(__pyx_t_1, "DATETIMEINDEX", (void **)&__pyx_vp_10serializor_7typeref_DATETIMEINDEX, "PyObject *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr_3_0_12(__pyx_t_1, "TIMEDELTAINDEX", (void **)&__pyx_vp_10serializor_7typeref_TIMEDELTAINDEX, "PyObject *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr_3_0_12(__pyx_t_1, "CYTIMES_AVAILABLE", (void **)&__pyx_vp_10serializor_7typeref_CYTIMES_AVAILABLE, "int") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr_3_0_12(__pyx_t_1, "PYDT", (void **)&__pyx_vp_10serializor_7typeref_PYDT, "PyObject *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr_3_0_12(__pyx_t_1, "PDDT", (void **)&__pyx_vp_10serializor_7typeref_PDDT, "PyObject *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("serializor.utils"); if (!__pyx_t_1) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_ImportVoidPtr_3_0_12(__pyx_t_1, "AVAILABLE_TIMEZONES", (void **)&__pyx_vp_10serializor_5utils_AVAILABLE_TIMEZONES, "PyObject *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_function_import_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_import_code", 0);
  /*--- Function import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}


#if PY_MAJOR_VERSION >= 3
#if CYTHON_PEP489_MULTI_PHASE_INIT
static PyObject* __pyx_pymod_create(PyObject *spec, PyModuleDef *def); /*proto*/
static int __pyx_pymod_exec_ser(PyObject* module); /*proto*/
static PyModuleDef_Slot __pyx_moduledef_slots[] = {
  {Py_mod_create, (void*)__pyx_pymod_create},
  {Py_mod_exec, (void*)__pyx_pymod_exec_ser},
  {0, NULL}
};
#endif

#ifdef __cplusplus
namespace {
  struct PyModuleDef __pyx_moduledef =
  #else
  static struct PyModuleDef __pyx_moduledef =
  #endif
  {
      PyModuleDef_HEAD_INIT,
      "ser",
      0, /* m_doc */
    #if CYTHON_PEP489_MULTI_PHASE_INIT
      0, /* m_size */
    #elif CYTHON_USE_MODULE_STATE
      sizeof(__pyx_mstate), /* m_size */
    #else
      -1, /* m_size */
    #endif
      __pyx_methods /* m_methods */,
    #if CYTHON_PEP489_MULTI_PHASE_INIT
      __pyx_moduledef_slots, /* m_slots */
    #else
      NULL, /* m_reload */
    #endif
    #if CYTHON_USE_MODULE_STATE
      __pyx_m_traverse, /* m_traverse */
      __pyx_m_clear, /* m_clear */
      NULL /* m_free */
    #else
      NULL, /* m_traverse */
      NULL, /* m_clear */
      NULL /* m_free */
    #endif
  };
  #ifdef __cplusplus
} /* anonymous namespace */
#endif
#endif

#ifndef CYTHON_NO_PYINIT_EXPORT
#define __Pyx_PyMODINIT_FUNC PyMODINIT_FUNC
#elif PY_MAJOR_VERSION < 3
#ifdef __cplusplus
#define __Pyx_PyMODINIT_FUNC extern "C" void
#else
#define __Pyx_PyMODINIT_FUNC void
#endif
#else
#ifdef __cplusplus
#define __Pyx_PyMODINIT_FUNC extern "C" PyObject *
#else
#define __Pyx_PyMODINIT_FUNC PyObject *
#endif
#endif


#if PY_MAJOR_VERSION < 3
__Pyx_PyMODINIT_FUNC initser(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC initser(void)
#else
__Pyx_PyMODINIT_FUNC PyInit_ser(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC PyInit_ser(void)
#if CYTHON_PEP489_MULTI_PHASE_INIT
{
  return PyModuleDef_Init(&__pyx_moduledef);
}
static CYTHON_SMALL_CODE int __Pyx_check_single_interpreter(void) {
    #if PY_VERSION_HEX >= 0x030700A1
    static PY_INT64_T main_interpreter_id = -1;
    PY_INT64_T current_id = PyInterpreterState_GetID(PyThreadState_Get()->interp);
    if (main_interpreter_id == -1) {
        main_interpreter_id = current_id;
        return (unlikely(current_id == -1)) ? -1 : 0;
    } else if (unlikely(main_interpreter_id != current_id))
    #else
    static PyInterpreterState *main_interpreter = NULL;
    PyInterpreterState *current_interpreter = PyThreadState_Get()->interp;
    if (!main_interpreter) {
        main_interpreter = current_interpreter;
    } else if (unlikely(main_interpreter != current_interpreter))
    #endif
    {
        PyErr_SetString(
            PyExc_ImportError,
            "Interpreter change detected - this module can only be loaded into one interpreter per process.");
        return -1;
    }
    return 0;
}
#if CYTHON_COMPILING_IN_LIMITED_API
static CYTHON_SMALL_CODE int __Pyx_copy_spec_to_module(PyObject *spec, PyObject *module, const char* from_name, const char* to_name, int allow_none)
#else
static CYTHON_SMALL_CODE int __Pyx_copy_spec_to_module(PyObject *spec, PyObject *moddict, const char* from_name, const char* to_name, int allow_none)
#endif
{
    PyObject *value = PyObject_GetAttrString(spec, from_name);
    int result = 0;
    if (likely(value)) {
        if (allow_none || value != Py_None) {
#if CYTHON_COMPILING_IN_LIMITED_API
            result = PyModule_AddObject(module, to_name, value);
#else
            result = PyDict_SetItemString(moddict, to_name, value);
#endif
        }
        Py_DECREF(value);
    } else if (PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Clear();
    } else {
        result = -1;
    }
    return result;
}
static CYTHON_SMALL_CODE PyObject* __pyx_pymod_create(PyObject *spec, PyModuleDef *def) {
    PyObject *module = NULL, *moddict, *modname;
    CYTHON_UNUSED_VAR(def);
    if (__Pyx_check_single_interpreter())
        return NULL;
    if (__pyx_m)
        return __Pyx_NewRef(__pyx_m);
    modname = PyObject_GetAttrString(spec, "name");
    if (unlikely(!modname)) goto bad;
    module = PyModule_NewObject(modname);
    Py_DECREF(modname);
    if (unlikely(!module)) goto bad;
#if CYTHON_COMPILING_IN_LIMITED_API
    moddict = module;
#else
    moddict = PyModule_GetDict(module);
    if (unlikely(!moddict)) goto bad;
#endif
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "loader", "__loader__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "origin", "__file__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "parent", "__package__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "submodule_search_locations", "__path__", 0) < 0)) goto bad;
    return module;
bad:
    Py_XDECREF(module);
    return NULL;
}


static CYTHON_SMALL_CODE int __pyx_pymod_exec_ser(PyObject *__pyx_pyinit_module)
#endif
#endif
{
  int stringtab_initialized = 0;
  #if CYTHON_USE_MODULE_STATE
  int pystate_addmodule_run = 0;
  #endif
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  if (__pyx_m) {
    if (__pyx_m == __pyx_pyinit_module) return 0;
    PyErr_SetString(PyExc_RuntimeError, "Module 'ser' has already been imported. Re-initialisation is not supported.");
    return -1;
  }
  #elif PY_MAJOR_VERSION >= 3
  if (__pyx_m) return __Pyx_NewRef(__pyx_m);
  #endif
  /*--- Module creation code ---*/
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  __pyx_m = __pyx_pyinit_module;
  Py_INCREF(__pyx_m);
  #else
  #if PY_MAJOR_VERSION < 3
  __pyx_m = Py_InitModule4("ser", __pyx_methods, 0, 0, PYTHON_API_VERSION); Py_XINCREF(__pyx_m);
  if (unlikely(!__pyx_m)) __PYX_ERR(0, 1, __pyx_L1_error)
  #elif CYTHON_USE_MODULE_STATE
  __pyx_t_1 = PyModule_Create(&__pyx_moduledef); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  {
    int add_module_result = PyState_AddModule(__pyx_t_1, &__pyx_moduledef);
    __pyx_t_1 = 0; /* transfer ownership from __pyx_t_1 to "ser" pseudovariable */
    if (unlikely((add_module_result < 0))) __PYX_ERR(0, 1, __pyx_L1_error)
    pystate_addmodule_run = 1;
  }
  #else
  __pyx_m = PyModule_Create(&__pyx_moduledef);
  if (unlikely(!__pyx_m)) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #endif
  CYTHON_UNUSED_VAR(__pyx_t_1);
  __pyx_d = PyModule_GetDict(__pyx_m); if (unlikely(!__pyx_d)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_d);
  __pyx_b = __Pyx_PyImport_AddModuleRef(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_b)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_cython_runtime = __Pyx_PyImport_AddModuleRef((const char *) "cython_runtime"); if (unlikely(!__pyx_cython_runtime)) __PYX_ERR(0, 1, __pyx_L1_error)
  if (PyObject_SetAttrString(__pyx_m, "__builtins__", __pyx_b) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #if CYTHON_REFNANNY
__Pyx_RefNanny = __Pyx_RefNannyImportAPI("refnanny");
if (!__Pyx_RefNanny) {
  PyErr_Clear();
  __Pyx_RefNanny = __Pyx_RefNannyImportAPI("Cython.Runtime.refnanny");
  if (!__Pyx_RefNanny)
      Py_FatalError("failed to import 'refnanny' module");
}
#endif
  __Pyx_RefNannySetupContext("__Pyx_PyMODINIT_FUNC PyInit_ser(void)", 0);
  if (__Pyx_check_binary_version(__PYX_LIMITED_VERSION_HEX, __Pyx_get_runtime_version(), CYTHON_COMPILING_IN_LIMITED_API) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pxy_PyFrame_Initialize_Offsets
  __Pxy_PyFrame_Initialize_Offsets();
  #endif
  __pyx_empty_tuple = PyTuple_New(0); if (unlikely(!__pyx_empty_tuple)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_bytes = PyBytes_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_bytes)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_unicode = PyUnicode_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_unicode)) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pyx_CyFunction_USED
  if (__pyx_CyFunction_init(__pyx_m) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_FusedFunction_USED
  if (__pyx_FusedFunction_init(__pyx_m) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Coroutine_USED
  if (__pyx_Coroutine_init(__pyx_m) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Generator_USED
  if (__pyx_Generator_init(__pyx_m) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_AsyncGen_USED
  if (__pyx_AsyncGen_init(__pyx_m) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_StopAsyncIteration_USED
  if (__pyx_StopAsyncIteration_init(__pyx_m) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  /*--- Library function declarations ---*/
  /*--- Threads initialization code ---*/
  #if defined(WITH_THREAD) && PY_VERSION_HEX < 0x030700F0 && defined(__PYX_FORCE_INIT_THREADS) && __PYX_FORCE_INIT_THREADS
  PyEval_InitThreads();
  #endif
  /*--- Initialize various global constants etc. ---*/
  if (__Pyx_InitConstants() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  stringtab_initialized = 1;
  if (__Pyx_InitGlobals() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #if PY_MAJOR_VERSION < 3 && (__PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT)
  if (__Pyx_init_sys_getdefaultencoding_params() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  if (__pyx_module_is_main_serializor__binary__ser) {
    if (PyObject_SetAttr(__pyx_m, __pyx_n_s_name_2, __pyx_n_s_main) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  }
  #if PY_MAJOR_VERSION >= 3
  {
    PyObject *modules = PyImport_GetModuleDict(); if (unlikely(!modules)) __PYX_ERR(0, 1, __pyx_L1_error)
    if (!PyDict_GetItemString(modules, "serializor.binary.ser")) {
      if (unlikely((PyDict_SetItemString(modules, "serializor.binary.ser", __pyx_m) < 0))) __PYX_ERR(0, 1, __pyx_L1_error)
    }
  }
  #endif
  /*--- Builtin init code ---*/
  if (__Pyx_InitCachedBuiltins() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Constants init code ---*/
  if (__Pyx_InitCachedConstants() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Global type/function init code ---*/
  (void)__Pyx_modinit_global_init_code();
  if (unlikely((__Pyx_modinit_variable_export_code() < 0))) __PYX_ERR(0, 1, __pyx_L1_error)
  if (unlikely((__Pyx_modinit_function_export_code() < 0))) __PYX_ERR(0, 1, __pyx_L1_error)
  (void)__Pyx_modinit_type_init_code();
  if (unlikely((__Pyx_modinit_type_import_code() < 0))) __PYX_ERR(0, 1, __pyx_L1_error)
  if (unlikely((__Pyx_modinit_variable_import_code() < 0))) __PYX_ERR(0, 1, __pyx_L1_error)
  (void)__Pyx_modinit_function_import_code();
  /*--- Execution code ---*/
  #if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
  if (__Pyx_patch_abc() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif

  /* "serializor/binary/ser.py":20
 * from cython.cimports.serializor import identifier, typeref, utils  # type: ignore
 * 
 * np.import_array()             # <<<<<<<<<<<<<<
 * np.import_umath()
 * datetime.import_datetime()
 */
  __pyx_t_2 = __pyx_f_5numpy_import_array(); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(0, 20, __pyx_L1_error)

  /* "serializor/binary/ser.py":21
 * 
 * np.import_array()
 * np.import_umath()             # <<<<<<<<<<<<<<
 * datetime.import_datetime()
 * 
 */
  __pyx_t_2 = __pyx_f_5numpy_import_umath(); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(0, 21, __pyx_L1_error)

  /* "serializor/binary/ser.py":22
 * np.import_array()
 * np.import_umath()
 * datetime.import_datetime()             # <<<<<<<<<<<<<<
 * 
 * # Python imports
 */
  __pyx_f_7cpython_8datetime_import_datetime();

  /* "serializor/binary/ser.py":25
 * 
 * # Python imports
 * from typing import Iterable             # <<<<<<<<<<<<<<
 * import numpy as np, datetime
 * from time import struct_time
 */
  __pyx_t_3 = PyList_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_n_s_Iterable);
  __Pyx_GIVEREF(__pyx_n_s_Iterable);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_3, 0, __pyx_n_s_Iterable)) __PYX_ERR(0, 25, __pyx_L1_error);
  __pyx_t_4 = __Pyx_Import(__pyx_n_s_typing, __pyx_t_3, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_ImportFrom(__pyx_t_4, __pyx_n_s_Iterable); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_Iterable, __pyx_t_3) < 0) __PYX_ERR(0, 25, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "serializor/binary/ser.py":26
 * # Python imports
 * from typing import Iterable
 * import numpy as np, datetime             # <<<<<<<<<<<<<<
 * from time import struct_time
 * from orjson import dumps as _dumps, OPT_SERIALIZE_NUMPY
 */
  __pyx_t_4 = __Pyx_ImportDottedModule(__pyx_n_s_numpy, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_np, __pyx_t_4) < 0) __PYX_ERR(0, 26, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_ImportDottedModule(__pyx_n_s_datetime, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_datetime, __pyx_t_4) < 0) __PYX_ERR(0, 26, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "serializor/binary/ser.py":27
 * from typing import Iterable
 * import numpy as np, datetime
 * from time import struct_time             # <<<<<<<<<<<<<<
 * from orjson import dumps as _dumps, OPT_SERIALIZE_NUMPY
 * from pandas import (
 */
  __pyx_t_4 = PyList_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_INCREF(__pyx_n_s_struct_time);
  __Pyx_GIVEREF(__pyx_n_s_struct_time);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_4, 0, __pyx_n_s_struct_time)) __PYX_ERR(0, 27, __pyx_L1_error);
  __pyx_t_3 = __Pyx_Import(__pyx_n_s_time, __pyx_t_4, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_struct_time); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_struct_time, __pyx_t_4) < 0) __PYX_ERR(0, 27, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "serializor/binary/ser.py":28
 * import numpy as np, datetime
 * from time import struct_time
 * from orjson import dumps as _dumps, OPT_SERIALIZE_NUMPY             # <<<<<<<<<<<<<<
 * from pandas import (
 *     Timestamp,
 */
  __pyx_t_3 = PyList_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_n_s_dumps_2);
  __Pyx_GIVEREF(__pyx_n_s_dumps_2);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_3, 0, __pyx_n_s_dumps_2)) __PYX_ERR(0, 28, __pyx_L1_error);
  __Pyx_INCREF(__pyx_n_s_OPT_SERIALIZE_NUMPY);
  __Pyx_GIVEREF(__pyx_n_s_OPT_SERIALIZE_NUMPY);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_3, 1, __pyx_n_s_OPT_SERIALIZE_NUMPY)) __PYX_ERR(0, 28, __pyx_L1_error);
  __pyx_t_4 = __Pyx_Import(__pyx_n_s_orjson, __pyx_t_3, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_ImportFrom(__pyx_t_4, __pyx_n_s_dumps_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_dumps, __pyx_t_3) < 0) __PYX_ERR(0, 28, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_ImportFrom(__pyx_t_4, __pyx_n_s_OPT_SERIALIZE_NUMPY); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_t_3);
  __Pyx_XGOTREF(__pyx_v_10serializor_6binary_3ser_OPT_SERIALIZE_NUMPY);
  __Pyx_DECREF_SET(__pyx_v_10serializor_6binary_3ser_OPT_SERIALIZE_NUMPY, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "serializor/binary/ser.py":30
 * from orjson import dumps as _dumps, OPT_SERIALIZE_NUMPY
 * from pandas import (
 *     Timestamp,             # <<<<<<<<<<<<<<
 *     Timedelta,
 *     Series,
 */
  __pyx_t_4 = PyList_New(6); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 30, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_INCREF(__pyx_n_s_Timestamp);
  __Pyx_GIVEREF(__pyx_n_s_Timestamp);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_4, 0, __pyx_n_s_Timestamp)) __PYX_ERR(0, 30, __pyx_L1_error);
  __Pyx_INCREF(__pyx_n_s_Timedelta);
  __Pyx_GIVEREF(__pyx_n_s_Timedelta);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_4, 1, __pyx_n_s_Timedelta)) __PYX_ERR(0, 30, __pyx_L1_error);
  __Pyx_INCREF(__pyx_n_s_Series);
  __Pyx_GIVEREF(__pyx_n_s_Series);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_4, 2, __pyx_n_s_Series)) __PYX_ERR(0, 30, __pyx_L1_error);
  __Pyx_INCREF(__pyx_n_s_DatetimeIndex);
  __Pyx_GIVEREF(__pyx_n_s_DatetimeIndex);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_4, 3, __pyx_n_s_DatetimeIndex)) __PYX_ERR(0, 30, __pyx_L1_error);
  __Pyx_INCREF(__pyx_n_s_TimedeltaIndex);
  __Pyx_GIVEREF(__pyx_n_s_TimedeltaIndex);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_4, 4, __pyx_n_s_TimedeltaIndex)) __PYX_ERR(0, 30, __pyx_L1_error);
  __Pyx_INCREF(__pyx_n_s_DataFrame);
  __Pyx_GIVEREF(__pyx_n_s_DataFrame);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_4, 5, __pyx_n_s_DataFrame)) __PYX_ERR(0, 30, __pyx_L1_error);

  /* "serializor/binary/ser.py":29
 * from time import struct_time
 * from orjson import dumps as _dumps, OPT_SERIALIZE_NUMPY
 * from pandas import (             # <<<<<<<<<<<<<<
 *     Timestamp,
 *     Timedelta,
 */
  __pyx_t_3 = __Pyx_Import(__pyx_n_s_pandas, __pyx_t_4, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 29, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_Timestamp); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 29, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_Timestamp, __pyx_t_4) < 0) __PYX_ERR(0, 30, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_Timedelta); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 29, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_Timedelta, __pyx_t_4) < 0) __PYX_ERR(0, 31, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_Series); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 29, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_Series, __pyx_t_4) < 0) __PYX_ERR(0, 32, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_DatetimeIndex); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 29, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_DatetimeIndex, __pyx_t_4) < 0) __PYX_ERR(0, 33, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_TimedeltaIndex); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 29, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_TimedeltaIndex, __pyx_t_4) < 0) __PYX_ERR(0, 34, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_DataFrame); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 29, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_DataFrame, __pyx_t_4) < 0) __PYX_ERR(0, 35, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "serializor/binary/ser.py":37
 *     DataFrame,
 * )
 * from serializor import identifier, typeref, utils, errors             # <<<<<<<<<<<<<<
 * 
 * __all__ = ["serialize"]
 */
  __pyx_t_3 = PyList_New(4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 37, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_n_s_identifier);
  __Pyx_GIVEREF(__pyx_n_s_identifier);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_3, 0, __pyx_n_s_identifier)) __PYX_ERR(0, 37, __pyx_L1_error);
  __Pyx_INCREF(__pyx_n_s_typeref);
  __Pyx_GIVEREF(__pyx_n_s_typeref);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_3, 1, __pyx_n_s_typeref)) __PYX_ERR(0, 37, __pyx_L1_error);
  __Pyx_INCREF(__pyx_n_s_utils);
  __Pyx_GIVEREF(__pyx_n_s_utils);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_3, 2, __pyx_n_s_utils)) __PYX_ERR(0, 37, __pyx_L1_error);
  __Pyx_INCREF(__pyx_n_s_errors);
  __Pyx_GIVEREF(__pyx_n_s_errors);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_3, 3, __pyx_n_s_errors)) __PYX_ERR(0, 37, __pyx_L1_error);
  __pyx_t_4 = __Pyx_Import(__pyx_n_s_serializor, __pyx_t_3, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 37, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_ImportFrom(__pyx_t_4, __pyx_n_s_identifier); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 37, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_identifier, __pyx_t_3) < 0) __PYX_ERR(0, 37, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_ImportFrom(__pyx_t_4, __pyx_n_s_typeref); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 37, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_typeref, __pyx_t_3) < 0) __PYX_ERR(0, 37, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_ImportFrom(__pyx_t_4, __pyx_n_s_utils); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 37, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_utils, __pyx_t_3) < 0) __PYX_ERR(0, 37, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_ImportFrom(__pyx_t_4, __pyx_n_s_errors); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 37, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_errors, __pyx_t_3) < 0) __PYX_ERR(0, 37, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "serializor/binary/ser.py":39
 * from serializor import identifier, typeref, utils, errors
 * 
 * __all__ = ["serialize"]             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_4 = PyList_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 39, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_INCREF(__pyx_n_u_serialize);
  __Pyx_GIVEREF(__pyx_n_u_serialize);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_4, 0, __pyx_n_u_serialize)) __PYX_ERR(0, 39, __pyx_L1_error);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_all, __pyx_t_4) < 0) __PYX_ERR(0, 39, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "serializor/binary/ser.py":44
 * # Constants -----------------------------------------------------------------------------------
 * # . encoded integer
 * UINT8_ENCODE_VALUE: cython.uchar = 251             # <<<<<<<<<<<<<<
 * UINT16_ENCODE_VALUE: cython.uchar = 252
 * UINT32_ENCODE_VALUE: cython.uchar = 253
 */
  __pyx_v_10serializor_6binary_3ser_UINT8_ENCODE_VALUE = 0xFB;

  /* "serializor/binary/ser.py":45
 * # . encoded integer
 * UINT8_ENCODE_VALUE: cython.uchar = 251
 * UINT16_ENCODE_VALUE: cython.uchar = 252             # <<<<<<<<<<<<<<
 * UINT32_ENCODE_VALUE: cython.uchar = 253
 * UINT64_ENCODE_VALUE: cython.uchar = 254
 */
  __pyx_v_10serializor_6binary_3ser_UINT16_ENCODE_VALUE = 0xFC;

  /* "serializor/binary/ser.py":46
 * UINT8_ENCODE_VALUE: cython.uchar = 251
 * UINT16_ENCODE_VALUE: cython.uchar = 252
 * UINT32_ENCODE_VALUE: cython.uchar = 253             # <<<<<<<<<<<<<<
 * UINT64_ENCODE_VALUE: cython.uchar = 254
 * # . native types identifier
 */
  __pyx_v_10serializor_6binary_3ser_UINT32_ENCODE_VALUE = 0xFD;

  /* "serializor/binary/ser.py":47
 * UINT16_ENCODE_VALUE: cython.uchar = 252
 * UINT32_ENCODE_VALUE: cython.uchar = 253
 * UINT64_ENCODE_VALUE: cython.uchar = 254             # <<<<<<<<<<<<<<
 * # . native types identifier
 * DATETIME_ID: bytes = pack_uint8(identifier.DATETIME)  # type: ignore
 */
  __pyx_v_10serializor_6binary_3ser_UINT64_ENCODE_VALUE = 0xFE;

  /* "serializor/binary/ser.py":49
 * UINT64_ENCODE_VALUE: cython.uchar = 254
 * # . native types identifier
 * DATETIME_ID: bytes = pack_uint8(identifier.DATETIME)  # type: ignore             # <<<<<<<<<<<<<<
 * DATE_ID: bytes = pack_uint8(identifier.DATE)  # type: ignore
 * TIME_ID: bytes = pack_uint8(identifier.TIME)  # type: ignore
 */
  __pyx_t_4 = __pyx_f_10serializor_6binary_3ser_pack_uint8(__pyx_v_10serializor_10identifier_DATETIME); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 49, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_XGOTREF(__pyx_v_10serializor_6binary_3ser_DATETIME_ID);
  __Pyx_DECREF_SET(__pyx_v_10serializor_6binary_3ser_DATETIME_ID, ((PyObject*)__pyx_t_4));
  __Pyx_GIVEREF(__pyx_t_4);
  __pyx_t_4 = 0;

  /* "serializor/binary/ser.py":50
 * # . native types identifier
 * DATETIME_ID: bytes = pack_uint8(identifier.DATETIME)  # type: ignore
 * DATE_ID: bytes = pack_uint8(identifier.DATE)  # type: ignore             # <<<<<<<<<<<<<<
 * TIME_ID: bytes = pack_uint8(identifier.TIME)  # type: ignore
 * TIMEDELTA_ID: bytes = pack_uint8(identifier.TIMEDELTA)  # type: ignore
 */
  __pyx_t_4 = __pyx_f_10serializor_6binary_3ser_pack_uint8(__pyx_v_10serializor_10identifier_DATE); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 50, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_XGOTREF(__pyx_v_10serializor_6binary_3ser_DATE_ID);
  __Pyx_DECREF_SET(__pyx_v_10serializor_6binary_3ser_DATE_ID, ((PyObject*)__pyx_t_4));
  __Pyx_GIVEREF(__pyx_t_4);
  __pyx_t_4 = 0;

  /* "serializor/binary/ser.py":51
 * DATETIME_ID: bytes = pack_uint8(identifier.DATETIME)  # type: ignore
 * DATE_ID: bytes = pack_uint8(identifier.DATE)  # type: ignore
 * TIME_ID: bytes = pack_uint8(identifier.TIME)  # type: ignore             # <<<<<<<<<<<<<<
 * TIMEDELTA_ID: bytes = pack_uint8(identifier.TIMEDELTA)  # type: ignore
 * STRUCT_TIME_ID: bytes = pack_uint8(identifier.STRUCT_TIME)  # type: ignore
 */
  __pyx_t_4 = __pyx_f_10serializor_6binary_3ser_pack_uint8(__pyx_v_10serializor_10identifier_TIME); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 51, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_XGOTREF(__pyx_v_10serializor_6binary_3ser_TIME_ID);
  __Pyx_DECREF_SET(__pyx_v_10serializor_6binary_3ser_TIME_ID, ((PyObject*)__pyx_t_4));
  __Pyx_GIVEREF(__pyx_t_4);
  __pyx_t_4 = 0;

  /* "serializor/binary/ser.py":52
 * DATE_ID: bytes = pack_uint8(identifier.DATE)  # type: ignore
 * TIME_ID: bytes = pack_uint8(identifier.TIME)  # type: ignore
 * TIMEDELTA_ID: bytes = pack_uint8(identifier.TIMEDELTA)  # type: ignore             # <<<<<<<<<<<<<<
 * STRUCT_TIME_ID: bytes = pack_uint8(identifier.STRUCT_TIME)  # type: ignore
 * COMPLEX_ID: bytes = pack_uint8(identifier.COMPLEX)  # type: ignore
 */
  __pyx_t_4 = __pyx_f_10serializor_6binary_3ser_pack_uint8(__pyx_v_10serializor_10identifier_TIMEDELTA); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 52, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_XGOTREF(__pyx_v_10serializor_6binary_3ser_TIMEDELTA_ID);
  __Pyx_DECREF_SET(__pyx_v_10serializor_6binary_3ser_TIMEDELTA_ID, ((PyObject*)__pyx_t_4));
  __Pyx_GIVEREF(__pyx_t_4);
  __pyx_t_4 = 0;

  /* "serializor/binary/ser.py":53
 * TIME_ID: bytes = pack_uint8(identifier.TIME)  # type: ignore
 * TIMEDELTA_ID: bytes = pack_uint8(identifier.TIMEDELTA)  # type: ignore
 * STRUCT_TIME_ID: bytes = pack_uint8(identifier.STRUCT_TIME)  # type: ignore             # <<<<<<<<<<<<<<
 * COMPLEX_ID: bytes = pack_uint8(identifier.COMPLEX)  # type: ignore
 * LIST_ID: bytes = pack_uint8(identifier.LIST)  # type: ignore
 */
  __pyx_t_4 = __pyx_f_10serializor_6binary_3ser_pack_uint8(__pyx_v_10serializor_10identifier_STRUCT_TIME); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 53, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_XGOTREF(__pyx_v_10serializor_6binary_3ser_STRUCT_TIME_ID);
  __Pyx_DECREF_SET(__pyx_v_10serializor_6binary_3ser_STRUCT_TIME_ID, ((PyObject*)__pyx_t_4));
  __Pyx_GIVEREF(__pyx_t_4);
  __pyx_t_4 = 0;

  /* "serializor/binary/ser.py":54
 * TIMEDELTA_ID: bytes = pack_uint8(identifier.TIMEDELTA)  # type: ignore
 * STRUCT_TIME_ID: bytes = pack_uint8(identifier.STRUCT_TIME)  # type: ignore
 * COMPLEX_ID: bytes = pack_uint8(identifier.COMPLEX)  # type: ignore             # <<<<<<<<<<<<<<
 * LIST_ID: bytes = pack_uint8(identifier.LIST)  # type: ignore
 * TUPLE_ID: bytes = pack_uint8(identifier.TUPLE)  # type: ignore
 */
  __pyx_t_4 = __pyx_f_10serializor_6binary_3ser_pack_uint8(__pyx_v_10serializor_10identifier_COMPLEX); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 54, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_XGOTREF(__pyx_v_10serializor_6binary_3ser_COMPLEX_ID);
  __Pyx_DECREF_SET(__pyx_v_10serializor_6binary_3ser_COMPLEX_ID, ((PyObject*)__pyx_t_4));
  __Pyx_GIVEREF(__pyx_t_4);
  __pyx_t_4 = 0;

  /* "serializor/binary/ser.py":55
 * STRUCT_TIME_ID: bytes = pack_uint8(identifier.STRUCT_TIME)  # type: ignore
 * COMPLEX_ID: bytes = pack_uint8(identifier.COMPLEX)  # type: ignore
 * LIST_ID: bytes = pack_uint8(identifier.LIST)  # type: ignore             # <<<<<<<<<<<<<<
 * TUPLE_ID: bytes = pack_uint8(identifier.TUPLE)  # type: ignore
 * SET_ID: bytes = pack_uint8(identifier.SET)  # type: ignore
 */
  __pyx_t_4 = __pyx_f_10serializor_6binary_3ser_pack_uint8(__pyx_v_10serializor_10identifier_LIST); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 55, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_XGOTREF(__pyx_v_10serializor_6binary_3ser_LIST_ID);
  __Pyx_DECREF_SET(__pyx_v_10serializor_6binary_3ser_LIST_ID, ((PyObject*)__pyx_t_4));
  __Pyx_GIVEREF(__pyx_t_4);
  __pyx_t_4 = 0;

  /* "serializor/binary/ser.py":56
 * COMPLEX_ID: bytes = pack_uint8(identifier.COMPLEX)  # type: ignore
 * LIST_ID: bytes = pack_uint8(identifier.LIST)  # type: ignore
 * TUPLE_ID: bytes = pack_uint8(identifier.TUPLE)  # type: ignore             # <<<<<<<<<<<<<<
 * SET_ID: bytes = pack_uint8(identifier.SET)  # type: ignore
 * FROZENSET_ID: bytes = pack_uint8(identifier.FROZENSET)  # type: ignore
 */
  __pyx_t_4 = __pyx_f_10serializor_6binary_3ser_pack_uint8(__pyx_v_10serializor_10identifier_TUPLE); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 56, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_XGOTREF(__pyx_v_10serializor_6binary_3ser_TUPLE_ID);
  __Pyx_DECREF_SET(__pyx_v_10serializor_6binary_3ser_TUPLE_ID, ((PyObject*)__pyx_t_4));
  __Pyx_GIVEREF(__pyx_t_4);
  __pyx_t_4 = 0;

  /* "serializor/binary/ser.py":57
 * LIST_ID: bytes = pack_uint8(identifier.LIST)  # type: ignore
 * TUPLE_ID: bytes = pack_uint8(identifier.TUPLE)  # type: ignore
 * SET_ID: bytes = pack_uint8(identifier.SET)  # type: ignore             # <<<<<<<<<<<<<<
 * FROZENSET_ID: bytes = pack_uint8(identifier.FROZENSET)  # type: ignore
 * RANGE_ID: bytes = pack_uint8(identifier.RANGE)  # type: ignore
 */
  __pyx_t_4 = __pyx_f_10serializor_6binary_3ser_pack_uint8(__pyx_v_10serializor_10identifier_SET); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 57, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_XGOTREF(__pyx_v_10serializor_6binary_3ser_SET_ID);
  __Pyx_DECREF_SET(__pyx_v_10serializor_6binary_3ser_SET_ID, ((PyObject*)__pyx_t_4));
  __Pyx_GIVEREF(__pyx_t_4);
  __pyx_t_4 = 0;

  /* "serializor/binary/ser.py":58
 * TUPLE_ID: bytes = pack_uint8(identifier.TUPLE)  # type: ignore
 * SET_ID: bytes = pack_uint8(identifier.SET)  # type: ignore
 * FROZENSET_ID: bytes = pack_uint8(identifier.FROZENSET)  # type: ignore             # <<<<<<<<<<<<<<
 * RANGE_ID: bytes = pack_uint8(identifier.RANGE)  # type: ignore
 * DICT_ID: bytes = pack_uint8(identifier.DICT)  # type: ignore
 */
  __pyx_t_4 = __pyx_f_10serializor_6binary_3ser_pack_uint8(__pyx_v_10serializor_10identifier_FROZENSET); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 58, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_XGOTREF(__pyx_v_10serializor_6binary_3ser_FROZENSET_ID);
  __Pyx_DECREF_SET(__pyx_v_10serializor_6binary_3ser_FROZENSET_ID, ((PyObject*)__pyx_t_4));
  __Pyx_GIVEREF(__pyx_t_4);
  __pyx_t_4 = 0;

  /* "serializor/binary/ser.py":59
 * SET_ID: bytes = pack_uint8(identifier.SET)  # type: ignore
 * FROZENSET_ID: bytes = pack_uint8(identifier.FROZENSET)  # type: ignore
 * RANGE_ID: bytes = pack_uint8(identifier.RANGE)  # type: ignore             # <<<<<<<<<<<<<<
 * DICT_ID: bytes = pack_uint8(identifier.DICT)  # type: ignore
 * # . numpy identifider & dtype
 */
  __pyx_t_4 = __pyx_f_10serializor_6binary_3ser_pack_uint8(__pyx_v_10serializor_10identifier_RANGE); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 59, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_XGOTREF(__pyx_v_10serializor_6binary_3ser_RANGE_ID);
  __Pyx_DECREF_SET(__pyx_v_10serializor_6binary_3ser_RANGE_ID, ((PyObject*)__pyx_t_4));
  __Pyx_GIVEREF(__pyx_t_4);
  __pyx_t_4 = 0;

  /* "serializor/binary/ser.py":60
 * FROZENSET_ID: bytes = pack_uint8(identifier.FROZENSET)  # type: ignore
 * RANGE_ID: bytes = pack_uint8(identifier.RANGE)  # type: ignore
 * DICT_ID: bytes = pack_uint8(identifier.DICT)  # type: ignore             # <<<<<<<<<<<<<<
 * # . numpy identifider & dtype
 * DT64_ID: bytes = pack_uint8(identifier.DATETIME64)  # type: ignore
 */
  __pyx_t_4 = __pyx_f_10serializor_6binary_3ser_pack_uint8(__pyx_v_10serializor_10identifier_DICT); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 60, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_XGOTREF(__pyx_v_10serializor_6binary_3ser_DICT_ID);
  __Pyx_DECREF_SET(__pyx_v_10serializor_6binary_3ser_DICT_ID, ((PyObject*)__pyx_t_4));
  __Pyx_GIVEREF(__pyx_t_4);
  __pyx_t_4 = 0;

  /* "serializor/binary/ser.py":62
 * DICT_ID: bytes = pack_uint8(identifier.DICT)  # type: ignore
 * # . numpy identifider & dtype
 * DT64_ID: bytes = pack_uint8(identifier.DATETIME64)  # type: ignore             # <<<<<<<<<<<<<<
 * TD64_ID: bytes = pack_uint8(identifier.TIMEDELTA64)  # type: ignore
 * NDARRAY_ID: bytes = pack_uint8(identifier.NDARRAY)  # type: ignore
 */
  __pyx_t_4 = __pyx_f_10serializor_6binary_3ser_pack_uint8(__pyx_v_10serializor_10identifier_DATETIME64); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 62, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_XGOTREF(__pyx_v_10serializor_6binary_3ser_DT64_ID);
  __Pyx_DECREF_SET(__pyx_v_10serializor_6binary_3ser_DT64_ID, ((PyObject*)__pyx_t_4));
  __Pyx_GIVEREF(__pyx_t_4);
  __pyx_t_4 = 0;

  /* "serializor/binary/ser.py":63
 * # . numpy identifider & dtype
 * DT64_ID: bytes = pack_uint8(identifier.DATETIME64)  # type: ignore
 * TD64_ID: bytes = pack_uint8(identifier.TIMEDELTA64)  # type: ignore             # <<<<<<<<<<<<<<
 * NDARRAY_ID: bytes = pack_uint8(identifier.NDARRAY)  # type: ignore
 * NDARRAY_OBJECT_DT: bytes = pack_uint8(identifier.NDARRAY_OBJECT)  # type: ignore
 */
  __pyx_t_4 = __pyx_f_10serializor_6binary_3ser_pack_uint8(__pyx_v_10serializor_10identifier_TIMEDELTA64); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 63, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_XGOTREF(__pyx_v_10serializor_6binary_3ser_TD64_ID);
  __Pyx_DECREF_SET(__pyx_v_10serializor_6binary_3ser_TD64_ID, ((PyObject*)__pyx_t_4));
  __Pyx_GIVEREF(__pyx_t_4);
  __pyx_t_4 = 0;

  /* "serializor/binary/ser.py":64
 * DT64_ID: bytes = pack_uint8(identifier.DATETIME64)  # type: ignore
 * TD64_ID: bytes = pack_uint8(identifier.TIMEDELTA64)  # type: ignore
 * NDARRAY_ID: bytes = pack_uint8(identifier.NDARRAY)  # type: ignore             # <<<<<<<<<<<<<<
 * NDARRAY_OBJECT_DT: bytes = pack_uint8(identifier.NDARRAY_OBJECT)  # type: ignore
 * NDARRAY_INT_DT: bytes = pack_uint8(identifier.NDARRAY_INT)  # type: ignore
 */
  __pyx_t_4 = __pyx_f_10serializor_6binary_3ser_pack_uint8(__pyx_v_10serializor_10identifier_NDARRAY); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 64, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_XGOTREF(__pyx_v_10serializor_6binary_3ser_NDARRAY_ID);
  __Pyx_DECREF_SET(__pyx_v_10serializor_6binary_3ser_NDARRAY_ID, ((PyObject*)__pyx_t_4));
  __Pyx_GIVEREF(__pyx_t_4);
  __pyx_t_4 = 0;

  /* "serializor/binary/ser.py":65
 * TD64_ID: bytes = pack_uint8(identifier.TIMEDELTA64)  # type: ignore
 * NDARRAY_ID: bytes = pack_uint8(identifier.NDARRAY)  # type: ignore
 * NDARRAY_OBJECT_DT: bytes = pack_uint8(identifier.NDARRAY_OBJECT)  # type: ignore             # <<<<<<<<<<<<<<
 * NDARRAY_INT_DT: bytes = pack_uint8(identifier.NDARRAY_INT)  # type: ignore
 * NDARRAY_UINT_DT: bytes = pack_uint8(identifier.NDARRAY_UINT)  # type: ignore
 */
  __pyx_t_4 = __pyx_f_10serializor_6binary_3ser_pack_uint8(__pyx_v_10serializor_10identifier_NDARRAY_OBJECT); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 65, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_XGOTREF(__pyx_v_10serializor_6binary_3ser_NDARRAY_OBJECT_DT);
  __Pyx_DECREF_SET(__pyx_v_10serializor_6binary_3ser_NDARRAY_OBJECT_DT, ((PyObject*)__pyx_t_4));
  __Pyx_GIVEREF(__pyx_t_4);
  __pyx_t_4 = 0;

  /* "serializor/binary/ser.py":66
 * NDARRAY_ID: bytes = pack_uint8(identifier.NDARRAY)  # type: ignore
 * NDARRAY_OBJECT_DT: bytes = pack_uint8(identifier.NDARRAY_OBJECT)  # type: ignore
 * NDARRAY_INT_DT: bytes = pack_uint8(identifier.NDARRAY_INT)  # type: ignore             # <<<<<<<<<<<<<<
 * NDARRAY_UINT_DT: bytes = pack_uint8(identifier.NDARRAY_UINT)  # type: ignore
 * NDARRAY_FLOAT_DT: bytes = pack_uint8(identifier.NDARRAY_FLOAT)  # type: ignore
 */
  __pyx_t_4 = __pyx_f_10serializor_6binary_3ser_pack_uint8(__pyx_v_10serializor_10identifier_NDARRAY_INT); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 66, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_XGOTREF(__pyx_v_10serializor_6binary_3ser_NDARRAY_INT_DT);
  __Pyx_DECREF_SET(__pyx_v_10serializor_6binary_3ser_NDARRAY_INT_DT, ((PyObject*)__pyx_t_4));
  __Pyx_GIVEREF(__pyx_t_4);
  __pyx_t_4 = 0;

  /* "serializor/binary/ser.py":67
 * NDARRAY_OBJECT_DT: bytes = pack_uint8(identifier.NDARRAY_OBJECT)  # type: ignore
 * NDARRAY_INT_DT: bytes = pack_uint8(identifier.NDARRAY_INT)  # type: ignore
 * NDARRAY_UINT_DT: bytes = pack_uint8(identifier.NDARRAY_UINT)  # type: ignore             # <<<<<<<<<<<<<<
 * NDARRAY_FLOAT_DT: bytes = pack_uint8(identifier.NDARRAY_FLOAT)  # type: ignore
 * NDARRAY_BOOL_DT: bytes = pack_uint8(identifier.NDARRAY_BOOL)  # type: ignore
 */
  __pyx_t_4 = __pyx_f_10serializor_6binary_3ser_pack_uint8(__pyx_v_10serializor_10identifier_NDARRAY_UINT); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 67, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_XGOTREF(__pyx_v_10serializor_6binary_3ser_NDARRAY_UINT_DT);
  __Pyx_DECREF_SET(__pyx_v_10serializor_6binary_3ser_NDARRAY_UINT_DT, ((PyObject*)__pyx_t_4));
  __Pyx_GIVEREF(__pyx_t_4);
  __pyx_t_4 = 0;

  /* "serializor/binary/ser.py":68
 * NDARRAY_INT_DT: bytes = pack_uint8(identifier.NDARRAY_INT)  # type: ignore
 * NDARRAY_UINT_DT: bytes = pack_uint8(identifier.NDARRAY_UINT)  # type: ignore
 * NDARRAY_FLOAT_DT: bytes = pack_uint8(identifier.NDARRAY_FLOAT)  # type: ignore             # <<<<<<<<<<<<<<
 * NDARRAY_BOOL_DT: bytes = pack_uint8(identifier.NDARRAY_BOOL)  # type: ignore
 * NDARRAY_DT64_DT: bytes = pack_uint8(identifier.NDARRAY_DT64)  # type: ignore
 */
  __pyx_t_4 = __pyx_f_10serializor_6binary_3ser_pack_uint8(__pyx_v_10serializor_10identifier_NDARRAY_FLOAT); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 68, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_XGOTREF(__pyx_v_10serializor_6binary_3ser_NDARRAY_FLOAT_DT);
  __Pyx_DECREF_SET(__pyx_v_10serializor_6binary_3ser_NDARRAY_FLOAT_DT, ((PyObject*)__pyx_t_4));
  __Pyx_GIVEREF(__pyx_t_4);
  __pyx_t_4 = 0;

  /* "serializor/binary/ser.py":69
 * NDARRAY_UINT_DT: bytes = pack_uint8(identifier.NDARRAY_UINT)  # type: ignore
 * NDARRAY_FLOAT_DT: bytes = pack_uint8(identifier.NDARRAY_FLOAT)  # type: ignore
 * NDARRAY_BOOL_DT: bytes = pack_uint8(identifier.NDARRAY_BOOL)  # type: ignore             # <<<<<<<<<<<<<<
 * NDARRAY_DT64_DT: bytes = pack_uint8(identifier.NDARRAY_DT64)  # type: ignore
 * NDARRAY_TD64_DT: bytes = pack_uint8(identifier.NDARRAY_TD64)  # type: ignore
 */
  __pyx_t_4 = __pyx_f_10serializor_6binary_3ser_pack_uint8(__pyx_v_10serializor_10identifier_NDARRAY_BOOL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 69, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_XGOTREF(__pyx_v_10serializor_6binary_3ser_NDARRAY_BOOL_DT);
  __Pyx_DECREF_SET(__pyx_v_10serializor_6binary_3ser_NDARRAY_BOOL_DT, ((PyObject*)__pyx_t_4));
  __Pyx_GIVEREF(__pyx_t_4);
  __pyx_t_4 = 0;

  /* "serializor/binary/ser.py":70
 * NDARRAY_FLOAT_DT: bytes = pack_uint8(identifier.NDARRAY_FLOAT)  # type: ignore
 * NDARRAY_BOOL_DT: bytes = pack_uint8(identifier.NDARRAY_BOOL)  # type: ignore
 * NDARRAY_DT64_DT: bytes = pack_uint8(identifier.NDARRAY_DT64)  # type: ignore             # <<<<<<<<<<<<<<
 * NDARRAY_TD64_DT: bytes = pack_uint8(identifier.NDARRAY_TD64)  # type: ignore
 * NDARRAY_COMPLEX_DT: bytes = pack_uint8(identifier.NDARRAY_COMPLEX)  # type: ignore
 */
  __pyx_t_4 = __pyx_f_10serializor_6binary_3ser_pack_uint8(__pyx_v_10serializor_10identifier_NDARRAY_DT64); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 70, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_XGOTREF(__pyx_v_10serializor_6binary_3ser_NDARRAY_DT64_DT);
  __Pyx_DECREF_SET(__pyx_v_10serializor_6binary_3ser_NDARRAY_DT64_DT, ((PyObject*)__pyx_t_4));
  __Pyx_GIVEREF(__pyx_t_4);
  __pyx_t_4 = 0;

  /* "serializor/binary/ser.py":71
 * NDARRAY_BOOL_DT: bytes = pack_uint8(identifier.NDARRAY_BOOL)  # type: ignore
 * NDARRAY_DT64_DT: bytes = pack_uint8(identifier.NDARRAY_DT64)  # type: ignore
 * NDARRAY_TD64_DT: bytes = pack_uint8(identifier.NDARRAY_TD64)  # type: ignore             # <<<<<<<<<<<<<<
 * NDARRAY_COMPLEX_DT: bytes = pack_uint8(identifier.NDARRAY_COMPLEX)  # type: ignore
 * NDARRAY_BYTES_DT: bytes = pack_uint8(identifier.NDARRAY_BYTES)  # type: ignore
 */
  __pyx_t_4 = __pyx_f_10serializor_6binary_3ser_pack_uint8(__pyx_v_10serializor_10identifier_NDARRAY_TD64); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 71, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_XGOTREF(__pyx_v_10serializor_6binary_3ser_NDARRAY_TD64_DT);
  __Pyx_DECREF_SET(__pyx_v_10serializor_6binary_3ser_NDARRAY_TD64_DT, ((PyObject*)__pyx_t_4));
  __Pyx_GIVEREF(__pyx_t_4);
  __pyx_t_4 = 0;

  /* "serializor/binary/ser.py":72
 * NDARRAY_DT64_DT: bytes = pack_uint8(identifier.NDARRAY_DT64)  # type: ignore
 * NDARRAY_TD64_DT: bytes = pack_uint8(identifier.NDARRAY_TD64)  # type: ignore
 * NDARRAY_COMPLEX_DT: bytes = pack_uint8(identifier.NDARRAY_COMPLEX)  # type: ignore             # <<<<<<<<<<<<<<
 * NDARRAY_BYTES_DT: bytes = pack_uint8(identifier.NDARRAY_BYTES)  # type: ignore
 * NDARRAY_UNICODE_DT: bytes = pack_uint8(identifier.NDARRAY_UNICODE)  # type: ignore
 */
  __pyx_t_4 = __pyx_f_10serializor_6binary_3ser_pack_uint8(__pyx_v_10serializor_10identifier_NDARRAY_COMPLEX); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 72, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_XGOTREF(__pyx_v_10serializor_6binary_3ser_NDARRAY_COMPLEX_DT);
  __Pyx_DECREF_SET(__pyx_v_10serializor_6binary_3ser_NDARRAY_COMPLEX_DT, ((PyObject*)__pyx_t_4));
  __Pyx_GIVEREF(__pyx_t_4);
  __pyx_t_4 = 0;

  /* "serializor/binary/ser.py":73
 * NDARRAY_TD64_DT: bytes = pack_uint8(identifier.NDARRAY_TD64)  # type: ignore
 * NDARRAY_COMPLEX_DT: bytes = pack_uint8(identifier.NDARRAY_COMPLEX)  # type: ignore
 * NDARRAY_BYTES_DT: bytes = pack_uint8(identifier.NDARRAY_BYTES)  # type: ignore             # <<<<<<<<<<<<<<
 * NDARRAY_UNICODE_DT: bytes = pack_uint8(identifier.NDARRAY_UNICODE)  # type: ignore
 * NDARRAY_OBJECT_IDDT: bytes = NDARRAY_ID + NDARRAY_OBJECT_DT
 */
  __pyx_t_4 = __pyx_f_10serializor_6binary_3ser_pack_uint8(__pyx_v_10serializor_10identifier_NDARRAY_BYTES); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 73, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_XGOTREF(__pyx_v_10serializor_6binary_3ser_NDARRAY_BYTES_DT);
  __Pyx_DECREF_SET(__pyx_v_10serializor_6binary_3ser_NDARRAY_BYTES_DT, ((PyObject*)__pyx_t_4));
  __Pyx_GIVEREF(__pyx_t_4);
  __pyx_t_4 = 0;

  /* "serializor/binary/ser.py":74
 * NDARRAY_COMPLEX_DT: bytes = pack_uint8(identifier.NDARRAY_COMPLEX)  # type: ignore
 * NDARRAY_BYTES_DT: bytes = pack_uint8(identifier.NDARRAY_BYTES)  # type: ignore
 * NDARRAY_UNICODE_DT: bytes = pack_uint8(identifier.NDARRAY_UNICODE)  # type: ignore             # <<<<<<<<<<<<<<
 * NDARRAY_OBJECT_IDDT: bytes = NDARRAY_ID + NDARRAY_OBJECT_DT
 * NDARRAY_INT_IDDT: bytes = NDARRAY_ID + NDARRAY_INT_DT
 */
  __pyx_t_4 = __pyx_f_10serializor_6binary_3ser_pack_uint8(__pyx_v_10serializor_10identifier_NDARRAY_UNICODE); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 74, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_XGOTREF(__pyx_v_10serializor_6binary_3ser_NDARRAY_UNICODE_DT);
  __Pyx_DECREF_SET(__pyx_v_10serializor_6binary_3ser_NDARRAY_UNICODE_DT, ((PyObject*)__pyx_t_4));
  __Pyx_GIVEREF(__pyx_t_4);
  __pyx_t_4 = 0;

  /* "serializor/binary/ser.py":75
 * NDARRAY_BYTES_DT: bytes = pack_uint8(identifier.NDARRAY_BYTES)  # type: ignore
 * NDARRAY_UNICODE_DT: bytes = pack_uint8(identifier.NDARRAY_UNICODE)  # type: ignore
 * NDARRAY_OBJECT_IDDT: bytes = NDARRAY_ID + NDARRAY_OBJECT_DT             # <<<<<<<<<<<<<<
 * NDARRAY_INT_IDDT: bytes = NDARRAY_ID + NDARRAY_INT_DT
 * NDARRAY_UINT_IDDT: bytes = NDARRAY_ID + NDARRAY_UINT_DT
 */
  __pyx_t_4 = PyNumber_Add(__pyx_v_10serializor_6binary_3ser_NDARRAY_ID, __pyx_v_10serializor_6binary_3ser_NDARRAY_OBJECT_DT); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 75, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_XGOTREF(__pyx_v_10serializor_6binary_3ser_NDARRAY_OBJECT_IDDT);
  __Pyx_DECREF_SET(__pyx_v_10serializor_6binary_3ser_NDARRAY_OBJECT_IDDT, ((PyObject*)__pyx_t_4));
  __Pyx_GIVEREF(__pyx_t_4);
  __pyx_t_4 = 0;

  /* "serializor/binary/ser.py":76
 * NDARRAY_UNICODE_DT: bytes = pack_uint8(identifier.NDARRAY_UNICODE)  # type: ignore
 * NDARRAY_OBJECT_IDDT: bytes = NDARRAY_ID + NDARRAY_OBJECT_DT
 * NDARRAY_INT_IDDT: bytes = NDARRAY_ID + NDARRAY_INT_DT             # <<<<<<<<<<<<<<
 * NDARRAY_UINT_IDDT: bytes = NDARRAY_ID + NDARRAY_UINT_DT
 * NDARRAY_FLOAT_IDDT: bytes = NDARRAY_ID + NDARRAY_FLOAT_DT
 */
  __pyx_t_4 = PyNumber_Add(__pyx_v_10serializor_6binary_3ser_NDARRAY_ID, __pyx_v_10serializor_6binary_3ser_NDARRAY_INT_DT); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 76, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_XGOTREF(__pyx_v_10serializor_6binary_3ser_NDARRAY_INT_IDDT);
  __Pyx_DECREF_SET(__pyx_v_10serializor_6binary_3ser_NDARRAY_INT_IDDT, ((PyObject*)__pyx_t_4));
  __Pyx_GIVEREF(__pyx_t_4);
  __pyx_t_4 = 0;

  /* "serializor/binary/ser.py":77
 * NDARRAY_OBJECT_IDDT: bytes = NDARRAY_ID + NDARRAY_OBJECT_DT
 * NDARRAY_INT_IDDT: bytes = NDARRAY_ID + NDARRAY_INT_DT
 * NDARRAY_UINT_IDDT: bytes = NDARRAY_ID + NDARRAY_UINT_DT             # <<<<<<<<<<<<<<
 * NDARRAY_FLOAT_IDDT: bytes = NDARRAY_ID + NDARRAY_FLOAT_DT
 * NDARRAY_BOOL_IDDT: bytes = NDARRAY_ID + NDARRAY_BOOL_DT
 */
  __pyx_t_4 = PyNumber_Add(__pyx_v_10serializor_6binary_3ser_NDARRAY_ID, __pyx_v_10serializor_6binary_3ser_NDARRAY_UINT_DT); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 77, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_XGOTREF(__pyx_v_10serializor_6binary_3ser_NDARRAY_UINT_IDDT);
  __Pyx_DECREF_SET(__pyx_v_10serializor_6binary_3ser_NDARRAY_UINT_IDDT, ((PyObject*)__pyx_t_4));
  __Pyx_GIVEREF(__pyx_t_4);
  __pyx_t_4 = 0;

  /* "serializor/binary/ser.py":78
 * NDARRAY_INT_IDDT: bytes = NDARRAY_ID + NDARRAY_INT_DT
 * NDARRAY_UINT_IDDT: bytes = NDARRAY_ID + NDARRAY_UINT_DT
 * NDARRAY_FLOAT_IDDT: bytes = NDARRAY_ID + NDARRAY_FLOAT_DT             # <<<<<<<<<<<<<<
 * NDARRAY_BOOL_IDDT: bytes = NDARRAY_ID + NDARRAY_BOOL_DT
 * NDARRAY_DT64_IDDT: bytes = NDARRAY_ID + NDARRAY_DT64_DT
 */
  __pyx_t_4 = PyNumber_Add(__pyx_v_10serializor_6binary_3ser_NDARRAY_ID, __pyx_v_10serializor_6binary_3ser_NDARRAY_FLOAT_DT); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 78, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_XGOTREF(__pyx_v_10serializor_6binary_3ser_NDARRAY_FLOAT_IDDT);
  __Pyx_DECREF_SET(__pyx_v_10serializor_6binary_3ser_NDARRAY_FLOAT_IDDT, ((PyObject*)__pyx_t_4));
  __Pyx_GIVEREF(__pyx_t_4);
  __pyx_t_4 = 0;

  /* "serializor/binary/ser.py":79
 * NDARRAY_UINT_IDDT: bytes = NDARRAY_ID + NDARRAY_UINT_DT
 * NDARRAY_FLOAT_IDDT: bytes = NDARRAY_ID + NDARRAY_FLOAT_DT
 * NDARRAY_BOOL_IDDT: bytes = NDARRAY_ID + NDARRAY_BOOL_DT             # <<<<<<<<<<<<<<
 * NDARRAY_DT64_IDDT: bytes = NDARRAY_ID + NDARRAY_DT64_DT
 * NDARRAY_TD64_IDDT: bytes = NDARRAY_ID + NDARRAY_TD64_DT
 */
  __pyx_t_4 = PyNumber_Add(__pyx_v_10serializor_6binary_3ser_NDARRAY_ID, __pyx_v_10serializor_6binary_3ser_NDARRAY_BOOL_DT); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 79, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_XGOTREF(__pyx_v_10serializor_6binary_3ser_NDARRAY_BOOL_IDDT);
  __Pyx_DECREF_SET(__pyx_v_10serializor_6binary_3ser_NDARRAY_BOOL_IDDT, ((PyObject*)__pyx_t_4));
  __Pyx_GIVEREF(__pyx_t_4);
  __pyx_t_4 = 0;

  /* "serializor/binary/ser.py":80
 * NDARRAY_FLOAT_IDDT: bytes = NDARRAY_ID + NDARRAY_FLOAT_DT
 * NDARRAY_BOOL_IDDT: bytes = NDARRAY_ID + NDARRAY_BOOL_DT
 * NDARRAY_DT64_IDDT: bytes = NDARRAY_ID + NDARRAY_DT64_DT             # <<<<<<<<<<<<<<
 * NDARRAY_TD64_IDDT: bytes = NDARRAY_ID + NDARRAY_TD64_DT
 * NDARRAY_COMPLEX_IDDT: bytes = NDARRAY_ID + NDARRAY_COMPLEX_DT
 */
  __pyx_t_4 = PyNumber_Add(__pyx_v_10serializor_6binary_3ser_NDARRAY_ID, __pyx_v_10serializor_6binary_3ser_NDARRAY_DT64_DT); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 80, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_XGOTREF(__pyx_v_10serializor_6binary_3ser_NDARRAY_DT64_IDDT);
  __Pyx_DECREF_SET(__pyx_v_10serializor_6binary_3ser_NDARRAY_DT64_IDDT, ((PyObject*)__pyx_t_4));
  __Pyx_GIVEREF(__pyx_t_4);
  __pyx_t_4 = 0;

  /* "serializor/binary/ser.py":81
 * NDARRAY_BOOL_IDDT: bytes = NDARRAY_ID + NDARRAY_BOOL_DT
 * NDARRAY_DT64_IDDT: bytes = NDARRAY_ID + NDARRAY_DT64_DT
 * NDARRAY_TD64_IDDT: bytes = NDARRAY_ID + NDARRAY_TD64_DT             # <<<<<<<<<<<<<<
 * NDARRAY_COMPLEX_IDDT: bytes = NDARRAY_ID + NDARRAY_COMPLEX_DT
 * NDARRAY_BYTES_IDDT: bytes = NDARRAY_ID + NDARRAY_BYTES_DT
 */
  __pyx_t_4 = PyNumber_Add(__pyx_v_10serializor_6binary_3ser_NDARRAY_ID, __pyx_v_10serializor_6binary_3ser_NDARRAY_TD64_DT); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 81, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_XGOTREF(__pyx_v_10serializor_6binary_3ser_NDARRAY_TD64_IDDT);
  __Pyx_DECREF_SET(__pyx_v_10serializor_6binary_3ser_NDARRAY_TD64_IDDT, ((PyObject*)__pyx_t_4));
  __Pyx_GIVEREF(__pyx_t_4);
  __pyx_t_4 = 0;

  /* "serializor/binary/ser.py":82
 * NDARRAY_DT64_IDDT: bytes = NDARRAY_ID + NDARRAY_DT64_DT
 * NDARRAY_TD64_IDDT: bytes = NDARRAY_ID + NDARRAY_TD64_DT
 * NDARRAY_COMPLEX_IDDT: bytes = NDARRAY_ID + NDARRAY_COMPLEX_DT             # <<<<<<<<<<<<<<
 * NDARRAY_BYTES_IDDT: bytes = NDARRAY_ID + NDARRAY_BYTES_DT
 * NDARRAY_UNICODE_IDDT: bytes = NDARRAY_ID + NDARRAY_UNICODE_DT
 */
  __pyx_t_4 = PyNumber_Add(__pyx_v_10serializor_6binary_3ser_NDARRAY_ID, __pyx_v_10serializor_6binary_3ser_NDARRAY_COMPLEX_DT); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 82, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_XGOTREF(__pyx_v_10serializor_6binary_3ser_NDARRAY_COMPLEX_IDDT);
  __Pyx_DECREF_SET(__pyx_v_10serializor_6binary_3ser_NDARRAY_COMPLEX_IDDT, ((PyObject*)__pyx_t_4));
  __Pyx_GIVEREF(__pyx_t_4);
  __pyx_t_4 = 0;

  /* "serializor/binary/ser.py":83
 * NDARRAY_TD64_IDDT: bytes = NDARRAY_ID + NDARRAY_TD64_DT
 * NDARRAY_COMPLEX_IDDT: bytes = NDARRAY_ID + NDARRAY_COMPLEX_DT
 * NDARRAY_BYTES_IDDT: bytes = NDARRAY_ID + NDARRAY_BYTES_DT             # <<<<<<<<<<<<<<
 * NDARRAY_UNICODE_IDDT: bytes = NDARRAY_ID + NDARRAY_UNICODE_DT
 * # . pandas identifider
 */
  __pyx_t_4 = PyNumber_Add(__pyx_v_10serializor_6binary_3ser_NDARRAY_ID, __pyx_v_10serializor_6binary_3ser_NDARRAY_BYTES_DT); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 83, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_XGOTREF(__pyx_v_10serializor_6binary_3ser_NDARRAY_BYTES_IDDT);
  __Pyx_DECREF_SET(__pyx_v_10serializor_6binary_3ser_NDARRAY_BYTES_IDDT, ((PyObject*)__pyx_t_4));
  __Pyx_GIVEREF(__pyx_t_4);
  __pyx_t_4 = 0;

  /* "serializor/binary/ser.py":84
 * NDARRAY_COMPLEX_IDDT: bytes = NDARRAY_ID + NDARRAY_COMPLEX_DT
 * NDARRAY_BYTES_IDDT: bytes = NDARRAY_ID + NDARRAY_BYTES_DT
 * NDARRAY_UNICODE_IDDT: bytes = NDARRAY_ID + NDARRAY_UNICODE_DT             # <<<<<<<<<<<<<<
 * # . pandas identifider
 * PD_TIMESTAMP_ID: bytes = pack_uint8(identifier.PD_TIMESTAMP)  # type: ignore
 */
  __pyx_t_4 = PyNumber_Add(__pyx_v_10serializor_6binary_3ser_NDARRAY_ID, __pyx_v_10serializor_6binary_3ser_NDARRAY_UNICODE_DT); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 84, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_XGOTREF(__pyx_v_10serializor_6binary_3ser_NDARRAY_UNICODE_IDDT);
  __Pyx_DECREF_SET(__pyx_v_10serializor_6binary_3ser_NDARRAY_UNICODE_IDDT, ((PyObject*)__pyx_t_4));
  __Pyx_GIVEREF(__pyx_t_4);
  __pyx_t_4 = 0;

  /* "serializor/binary/ser.py":86
 * NDARRAY_UNICODE_IDDT: bytes = NDARRAY_ID + NDARRAY_UNICODE_DT
 * # . pandas identifider
 * PD_TIMESTAMP_ID: bytes = pack_uint8(identifier.PD_TIMESTAMP)  # type: ignore             # <<<<<<<<<<<<<<
 * PD_TIMEDELTA_ID: bytes = pack_uint8(identifier.PD_TIMEDELTA)  # type: ignore
 * SERIES_ID: bytes = pack_uint8(identifier.SERIES)  # type: ignore
 */
  __pyx_t_4 = __pyx_f_10serializor_6binary_3ser_pack_uint8(__pyx_v_10serializor_10identifier_PD_TIMESTAMP); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 86, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_XGOTREF(__pyx_v_10serializor_6binary_3ser_PD_TIMESTAMP_ID);
  __Pyx_DECREF_SET(__pyx_v_10serializor_6binary_3ser_PD_TIMESTAMP_ID, ((PyObject*)__pyx_t_4));
  __Pyx_GIVEREF(__pyx_t_4);
  __pyx_t_4 = 0;

  /* "serializor/binary/ser.py":87
 * # . pandas identifider
 * PD_TIMESTAMP_ID: bytes = pack_uint8(identifier.PD_TIMESTAMP)  # type: ignore
 * PD_TIMEDELTA_ID: bytes = pack_uint8(identifier.PD_TIMEDELTA)  # type: ignore             # <<<<<<<<<<<<<<
 * SERIES_ID: bytes = pack_uint8(identifier.SERIES)  # type: ignore
 * DATETIMEINDEX_ID: bytes = pack_uint8(identifier.DATETIMEINDEX)  # type: ignore
 */
  __pyx_t_4 = __pyx_f_10serializor_6binary_3ser_pack_uint8(__pyx_v_10serializor_10identifier_PD_TIMEDELTA); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 87, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_XGOTREF(__pyx_v_10serializor_6binary_3ser_PD_TIMEDELTA_ID);
  __Pyx_DECREF_SET(__pyx_v_10serializor_6binary_3ser_PD_TIMEDELTA_ID, ((PyObject*)__pyx_t_4));
  __Pyx_GIVEREF(__pyx_t_4);
  __pyx_t_4 = 0;

  /* "serializor/binary/ser.py":88
 * PD_TIMESTAMP_ID: bytes = pack_uint8(identifier.PD_TIMESTAMP)  # type: ignore
 * PD_TIMEDELTA_ID: bytes = pack_uint8(identifier.PD_TIMEDELTA)  # type: ignore
 * SERIES_ID: bytes = pack_uint8(identifier.SERIES)  # type: ignore             # <<<<<<<<<<<<<<
 * DATETIMEINDEX_ID: bytes = pack_uint8(identifier.DATETIMEINDEX)  # type: ignore
 * TIMEDELTAINDEX_ID: bytes = pack_uint8(identifier.TIMEDELTAINDEX)  # type: ignore
 */
  __pyx_t_4 = __pyx_f_10serializor_6binary_3ser_pack_uint8(__pyx_v_10serializor_10identifier_SERIES); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 88, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_XGOTREF(__pyx_v_10serializor_6binary_3ser_SERIES_ID);
  __Pyx_DECREF_SET(__pyx_v_10serializor_6binary_3ser_SERIES_ID, ((PyObject*)__pyx_t_4));
  __Pyx_GIVEREF(__pyx_t_4);
  __pyx_t_4 = 0;

  /* "serializor/binary/ser.py":89
 * PD_TIMEDELTA_ID: bytes = pack_uint8(identifier.PD_TIMEDELTA)  # type: ignore
 * SERIES_ID: bytes = pack_uint8(identifier.SERIES)  # type: ignore
 * DATETIMEINDEX_ID: bytes = pack_uint8(identifier.DATETIMEINDEX)  # type: ignore             # <<<<<<<<<<<<<<
 * TIMEDELTAINDEX_ID: bytes = pack_uint8(identifier.TIMEDELTAINDEX)  # type: ignore
 * DATAFRAME_ID: bytes = pack_uint8(identifier.DATAFRAME)  # type: ignore
 */
  __pyx_t_4 = __pyx_f_10serializor_6binary_3ser_pack_uint8(__pyx_v_10serializor_10identifier_DATETIMEINDEX); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_XGOTREF(__pyx_v_10serializor_6binary_3ser_DATETIMEINDEX_ID);
  __Pyx_DECREF_SET(__pyx_v_10serializor_6binary_3ser_DATETIMEINDEX_ID, ((PyObject*)__pyx_t_4));
  __Pyx_GIVEREF(__pyx_t_4);
  __pyx_t_4 = 0;

  /* "serializor/binary/ser.py":90
 * SERIES_ID: bytes = pack_uint8(identifier.SERIES)  # type: ignore
 * DATETIMEINDEX_ID: bytes = pack_uint8(identifier.DATETIMEINDEX)  # type: ignore
 * TIMEDELTAINDEX_ID: bytes = pack_uint8(identifier.TIMEDELTAINDEX)  # type: ignore             # <<<<<<<<<<<<<<
 * DATAFRAME_ID: bytes = pack_uint8(identifier.DATAFRAME)  # type: ignore
 * # . value
 */
  __pyx_t_4 = __pyx_f_10serializor_6binary_3ser_pack_uint8(__pyx_v_10serializor_10identifier_TIMEDELTAINDEX); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 90, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_XGOTREF(__pyx_v_10serializor_6binary_3ser_TIMEDELTAINDEX_ID);
  __Pyx_DECREF_SET(__pyx_v_10serializor_6binary_3ser_TIMEDELTAINDEX_ID, ((PyObject*)__pyx_t_4));
  __Pyx_GIVEREF(__pyx_t_4);
  __pyx_t_4 = 0;

  /* "serializor/binary/ser.py":91
 * DATETIMEINDEX_ID: bytes = pack_uint8(identifier.DATETIMEINDEX)  # type: ignore
 * TIMEDELTAINDEX_ID: bytes = pack_uint8(identifier.TIMEDELTAINDEX)  # type: ignore
 * DATAFRAME_ID: bytes = pack_uint8(identifier.DATAFRAME)  # type: ignore             # <<<<<<<<<<<<<<
 * # . value
 * TRUE_VALUE: bytes = gen_header(identifier.BOOL, 1)  # type: ignore
 */
  __pyx_t_4 = __pyx_f_10serializor_6binary_3ser_pack_uint8(__pyx_v_10serializor_10identifier_DATAFRAME); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 91, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_XGOTREF(__pyx_v_10serializor_6binary_3ser_DATAFRAME_ID);
  __Pyx_DECREF_SET(__pyx_v_10serializor_6binary_3ser_DATAFRAME_ID, ((PyObject*)__pyx_t_4));
  __Pyx_GIVEREF(__pyx_t_4);
  __pyx_t_4 = 0;

  /* "serializor/binary/ser.py":93
 * DATAFRAME_ID: bytes = pack_uint8(identifier.DATAFRAME)  # type: ignore
 * # . value
 * TRUE_VALUE: bytes = gen_header(identifier.BOOL, 1)  # type: ignore             # <<<<<<<<<<<<<<
 * FALSE_VALUE: bytes = gen_header(identifier.BOOL, 0)  # type: ignore
 * NONE_VALUE: bytes = pack_uint8(identifier.NONE)  # type: ignore
 */
  __pyx_t_4 = __pyx_f_10serializor_6binary_3ser_gen_header(__pyx_v_10serializor_10identifier_BOOL, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 93, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_XGOTREF(__pyx_v_10serializor_6binary_3ser_TRUE_VALUE);
  __Pyx_DECREF_SET(__pyx_v_10serializor_6binary_3ser_TRUE_VALUE, ((PyObject*)__pyx_t_4));
  __Pyx_GIVEREF(__pyx_t_4);
  __pyx_t_4 = 0;

  /* "serializor/binary/ser.py":94
 * # . value
 * TRUE_VALUE: bytes = gen_header(identifier.BOOL, 1)  # type: ignore
 * FALSE_VALUE: bytes = gen_header(identifier.BOOL, 0)  # type: ignore             # <<<<<<<<<<<<<<
 * NONE_VALUE: bytes = pack_uint8(identifier.NONE)  # type: ignore
 * 
 */
  __pyx_t_4 = __pyx_f_10serializor_6binary_3ser_gen_header(__pyx_v_10serializor_10identifier_BOOL, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 94, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_XGOTREF(__pyx_v_10serializor_6binary_3ser_FALSE_VALUE);
  __Pyx_DECREF_SET(__pyx_v_10serializor_6binary_3ser_FALSE_VALUE, ((PyObject*)__pyx_t_4));
  __Pyx_GIVEREF(__pyx_t_4);
  __pyx_t_4 = 0;

  /* "serializor/binary/ser.py":95
 * TRUE_VALUE: bytes = gen_header(identifier.BOOL, 1)  # type: ignore
 * FALSE_VALUE: bytes = gen_header(identifier.BOOL, 0)  # type: ignore
 * NONE_VALUE: bytes = pack_uint8(identifier.NONE)  # type: ignore             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_4 = __pyx_f_10serializor_6binary_3ser_pack_uint8(__pyx_v_10serializor_10identifier_NONE); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 95, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_XGOTREF(__pyx_v_10serializor_6binary_3ser_NONE_VALUE);
  __Pyx_DECREF_SET(__pyx_v_10serializor_6binary_3ser_NONE_VALUE, ((PyObject*)__pyx_t_4));
  __Pyx_GIVEREF(__pyx_t_4);
  __pyx_t_4 = 0;

  /* "serializor/binary/ser.py":2061
 * 
 * 
 * @cython.ccall             # <<<<<<<<<<<<<<
 * def serialize(obj: object) -> bytes:
 *     """Serialize an object to `<'bytes'>`.
 */
  __pyx_t_4 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2061, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_obj, __pyx_n_s_object) < 0) __PYX_ERR(0, 2061, __pyx_L1_error)
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_10serializor_6binary_3ser_1serialize, 0, __pyx_n_s_serialize, NULL, __pyx_n_s_serializor_binary_ser, __pyx_d, ((PyObject *)__pyx_codeobj__29)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2061, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_3, __pyx_t_4);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_serialize, __pyx_t_3) < 0) __PYX_ERR(0, 2061, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "serializor/binary/ser.py":2117
 * 
 * ########## The following functions are for testing purpose only ##########
 * def _test_utils() -> None:             # <<<<<<<<<<<<<<
 *     _test_pack_int8()
 *     _test_pack_uint8()
 */
  __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2117, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_return, __pyx_n_s_None) < 0) __PYX_ERR(0, 2117, __pyx_L1_error)
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_10serializor_6binary_3ser_3_test_utils, 0, __pyx_n_s_test_utils, NULL, __pyx_n_s_serializor_binary_ser, __pyx_d, ((PyObject *)__pyx_codeobj__30)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2117, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_4, __pyx_t_3);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_test_utils, __pyx_t_4) < 0) __PYX_ERR(0, 2117, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "serializor/binary/ser.py":2132
 * 
 * 
 * def _test_pack_int8() -> None:             # <<<<<<<<<<<<<<
 *     import struct
 * 
 */
  __pyx_t_4 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2132, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_return, __pyx_n_s_None) < 0) __PYX_ERR(0, 2132, __pyx_L1_error)
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_10serializor_6binary_3ser_5_test_pack_int8, 0, __pyx_n_s_test_pack_int8, NULL, __pyx_n_s_serializor_binary_ser, __pyx_d, ((PyObject *)__pyx_codeobj__32)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2132, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_3, __pyx_t_4);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_test_pack_int8, __pyx_t_3) < 0) __PYX_ERR(0, 2132, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "serializor/binary/ser.py":2145
 * 
 * 
 * def _test_pack_uint8() -> None:             # <<<<<<<<<<<<<<
 *     import struct
 * 
 */
  __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2145, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_return, __pyx_n_s_None) < 0) __PYX_ERR(0, 2145, __pyx_L1_error)
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_10serializor_6binary_3ser_7_test_pack_uint8, 0, __pyx_n_s_test_pack_uint8, NULL, __pyx_n_s_serializor_binary_ser, __pyx_d, ((PyObject *)__pyx_codeobj__33)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2145, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_4, __pyx_t_3);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_test_pack_uint8, __pyx_t_4) < 0) __PYX_ERR(0, 2145, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "serializor/binary/ser.py":2158
 * 
 * 
 * def _test_pack_int16() -> None:             # <<<<<<<<<<<<<<
 *     import struct
 * 
 */
  __pyx_t_4 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2158, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_return, __pyx_n_s_None) < 0) __PYX_ERR(0, 2158, __pyx_L1_error)
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_10serializor_6binary_3ser_9_test_pack_int16, 0, __pyx_n_s_test_pack_int16, NULL, __pyx_n_s_serializor_binary_ser, __pyx_d, ((PyObject *)__pyx_codeobj__34)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2158, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_3, __pyx_t_4);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_test_pack_int16, __pyx_t_3) < 0) __PYX_ERR(0, 2158, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "serializor/binary/ser.py":2171
 * 
 * 
 * def _test_pack_uint16() -> None:             # <<<<<<<<<<<<<<
 *     import struct
 * 
 */
  __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2171, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_return, __pyx_n_s_None) < 0) __PYX_ERR(0, 2171, __pyx_L1_error)
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_10serializor_6binary_3ser_11_test_pack_uint16, 0, __pyx_n_s_test_pack_uint16, NULL, __pyx_n_s_serializor_binary_ser, __pyx_d, ((PyObject *)__pyx_codeobj__35)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2171, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_4, __pyx_t_3);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_test_pack_uint16, __pyx_t_4) < 0) __PYX_ERR(0, 2171, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "serializor/binary/ser.py":2184
 * 
 * 
 * def _test_pack_int24() -> None:             # <<<<<<<<<<<<<<
 *     import struct
 * 
 */
  __pyx_t_4 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2184, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_return, __pyx_n_s_None) < 0) __PYX_ERR(0, 2184, __pyx_L1_error)
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_10serializor_6binary_3ser_13_test_pack_int24, 0, __pyx_n_s_test_pack_int24, NULL, __pyx_n_s_serializor_binary_ser, __pyx_d, ((PyObject *)__pyx_codeobj__36)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2184, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_3, __pyx_t_4);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_test_pack_int24, __pyx_t_3) < 0) __PYX_ERR(0, 2184, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "serializor/binary/ser.py":2197
 * 
 * 
 * def _test_pack_uint24() -> None:             # <<<<<<<<<<<<<<
 *     import struct
 * 
 */
  __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2197, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_return, __pyx_n_s_None) < 0) __PYX_ERR(0, 2197, __pyx_L1_error)
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_10serializor_6binary_3ser_15_test_pack_uint24, 0, __pyx_n_s_test_pack_uint24, NULL, __pyx_n_s_serializor_binary_ser, __pyx_d, ((PyObject *)__pyx_codeobj__37)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2197, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_4, __pyx_t_3);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_test_pack_uint24, __pyx_t_4) < 0) __PYX_ERR(0, 2197, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "serializor/binary/ser.py":2210
 * 
 * 
 * def _test_pack_int32() -> None:             # <<<<<<<<<<<<<<
 *     import struct
 * 
 */
  __pyx_t_4 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2210, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_return, __pyx_n_s_None) < 0) __PYX_ERR(0, 2210, __pyx_L1_error)
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_10serializor_6binary_3ser_17_test_pack_int32, 0, __pyx_n_s_test_pack_int32, NULL, __pyx_n_s_serializor_binary_ser, __pyx_d, ((PyObject *)__pyx_codeobj__38)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2210, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_3, __pyx_t_4);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_test_pack_int32, __pyx_t_3) < 0) __PYX_ERR(0, 2210, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "serializor/binary/ser.py":2223
 * 
 * 
 * def _test_pack_uint32() -> None:             # <<<<<<<<<<<<<<
 *     import struct
 * 
 */
  __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2223, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_return, __pyx_n_s_None) < 0) __PYX_ERR(0, 2223, __pyx_L1_error)
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_10serializor_6binary_3ser_19_test_pack_uint32, 0, __pyx_n_s_test_pack_uint32, NULL, __pyx_n_s_serializor_binary_ser, __pyx_d, ((PyObject *)__pyx_codeobj__39)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2223, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_4, __pyx_t_3);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_test_pack_uint32, __pyx_t_4) < 0) __PYX_ERR(0, 2223, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "serializor/binary/ser.py":2236
 * 
 * 
 * def _test_pack_int64() -> None:             # <<<<<<<<<<<<<<
 *     import struct
 * 
 */
  __pyx_t_4 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2236, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_return, __pyx_n_s_None) < 0) __PYX_ERR(0, 2236, __pyx_L1_error)
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_10serializor_6binary_3ser_21_test_pack_int64, 0, __pyx_n_s_test_pack_int64, NULL, __pyx_n_s_serializor_binary_ser, __pyx_d, ((PyObject *)__pyx_codeobj__40)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2236, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_3, __pyx_t_4);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_test_pack_int64, __pyx_t_3) < 0) __PYX_ERR(0, 2236, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "serializor/binary/ser.py":2255
 * 
 * 
 * def _test_pack_uint64() -> None:             # <<<<<<<<<<<<<<
 *     import struct
 * 
 */
  __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2255, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_return, __pyx_n_s_None) < 0) __PYX_ERR(0, 2255, __pyx_L1_error)
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_10serializor_6binary_3ser_23_test_pack_uint64, 0, __pyx_n_s_test_pack_uint64, NULL, __pyx_n_s_serializor_binary_ser, __pyx_d, ((PyObject *)__pyx_codeobj__41)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2255, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_4, __pyx_t_3);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_test_pack_uint64, __pyx_t_4) < 0) __PYX_ERR(0, 2255, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "serializor/binary/ser.py":2274
 * 
 * 
 * def _test_gen_encoded_int() -> None:             # <<<<<<<<<<<<<<
 *     # Test gen_encoded_int
 *     num = 251
 */
  __pyx_t_4 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2274, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_return, __pyx_n_s_None) < 0) __PYX_ERR(0, 2274, __pyx_L1_error)
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_10serializor_6binary_3ser_25_test_gen_encoded_int, 0, __pyx_n_s_test_gen_encoded_int, NULL, __pyx_n_s_serializor_binary_ser, __pyx_d, ((PyObject *)__pyx_codeobj__43)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2274, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_3, __pyx_t_4);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_test_gen_encoded_int, __pyx_t_3) < 0) __PYX_ERR(0, 2274, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "serializor/binary/ser.py":2298
 * 
 * 
 * def _test_gen_header() -> None:             # <<<<<<<<<<<<<<
 *     dtype: bytes = b"\x01"
 *     dchar: cython.uchar = 1
 */
  __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2298, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_return, __pyx_n_s_None) < 0) __PYX_ERR(0, 2298, __pyx_L1_error)
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_10serializor_6binary_3ser_27_test_gen_header, 0, __pyx_n_s_test_gen_header, NULL, __pyx_n_s_serializor_binary_ser, __pyx_d, ((PyObject *)__pyx_codeobj__45)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2298, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_4, __pyx_t_3);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_test_gen_header, __pyx_t_4) < 0) __PYX_ERR(0, 2298, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "serializor/binary/ser.py":1
 * # cython: language_level=3             # <<<<<<<<<<<<<<
 * # cython: wraparound=False
 * # cython: boundscheck=False
 */
  __pyx_t_4 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_test, __pyx_t_4) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /*--- Wrapped vars code ---*/

  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  if (__pyx_m) {
    if (__pyx_d && stringtab_initialized) {
      __Pyx_AddTraceback("init serializor.binary.ser", __pyx_clineno, __pyx_lineno, __pyx_filename);
    }
    #if !CYTHON_USE_MODULE_STATE
    Py_CLEAR(__pyx_m);
    #else
    Py_DECREF(__pyx_m);
    if (pystate_addmodule_run) {
      PyObject *tp, *value, *tb;
      PyErr_Fetch(&tp, &value, &tb);
      PyState_RemoveModule(&__pyx_moduledef);
      PyErr_Restore(tp, value, tb);
    }
    #endif
  } else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_ImportError, "init serializor.binary.ser");
  }
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  return (__pyx_m != NULL) ? 0 : -1;
  #elif PY_MAJOR_VERSION >= 3
  return __pyx_m;
  #else
  return;
  #endif
}
/* #### Code section: cleanup_globals ### */
/* #### Code section: cleanup_module ### */
/* #### Code section: main_method ### */
/* #### Code section: utility_code_pragmas ### */
#ifdef _MSC_VER
#pragma warning( push )
/* Warning 4127: conditional expression is constant
 * Cython uses constant conditional expressions to allow in inline functions to be optimized at
 * compile-time, so this warning is not useful
 */
#pragma warning( disable : 4127 )
#endif



/* #### Code section: utility_code_def ### */

/* --- Runtime support code --- */
/* Refnanny */
#if CYTHON_REFNANNY
static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname) {
    PyObject *m = NULL, *p = NULL;
    void *r = NULL;
    m = PyImport_ImportModule(modname);
    if (!m) goto end;
    p = PyObject_GetAttrString(m, "RefNannyAPI");
    if (!p) goto end;
    r = PyLong_AsVoidPtr(p);
end:
    Py_XDECREF(p);
    Py_XDECREF(m);
    return (__Pyx_RefNannyAPIStruct *)r;
}
#endif

/* PyErrExceptionMatches */
#if CYTHON_FAST_THREAD_STATE
static int __Pyx_PyErr_ExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        if (__Pyx_PyErr_GivenExceptionMatches(exc_type, PyTuple_GET_ITEM(tuple, i))) return 1;
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err) {
    int result;
    PyObject *exc_type;
#if PY_VERSION_HEX >= 0x030C00A6
    PyObject *current_exception = tstate->current_exception;
    if (unlikely(!current_exception)) return 0;
    exc_type = (PyObject*) Py_TYPE(current_exception);
    if (exc_type == err) return 1;
#else
    exc_type = tstate->curexc_type;
    if (exc_type == err) return 1;
    if (unlikely(!exc_type)) return 0;
#endif
    #if CYTHON_AVOID_BORROWED_REFS
    Py_INCREF(exc_type);
    #endif
    if (unlikely(PyTuple_Check(err))) {
        result = __Pyx_PyErr_ExceptionMatchesTuple(exc_type, err);
    } else {
        result = __Pyx_PyErr_GivenExceptionMatches(exc_type, err);
    }
    #if CYTHON_AVOID_BORROWED_REFS
    Py_DECREF(exc_type);
    #endif
    return result;
}
#endif

/* PyErrFetchRestore */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
#if PY_VERSION_HEX >= 0x030C00A6
    PyObject *tmp_value;
    assert(type == NULL || (value != NULL && type == (PyObject*) Py_TYPE(value)));
    if (value) {
        #if CYTHON_COMPILING_IN_CPYTHON
        if (unlikely(((PyBaseExceptionObject*) value)->traceback != tb))
        #endif
            PyException_SetTraceback(value, tb);
    }
    tmp_value = tstate->current_exception;
    tstate->current_exception = value;
    Py_XDECREF(tmp_value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
#else
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    tmp_type = tstate->curexc_type;
    tmp_value = tstate->curexc_value;
    tmp_tb = tstate->curexc_traceback;
    tstate->curexc_type = type;
    tstate->curexc_value = value;
    tstate->curexc_traceback = tb;
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
#endif
}
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
#if PY_VERSION_HEX >= 0x030C00A6
    PyObject* exc_value;
    exc_value = tstate->current_exception;
    tstate->current_exception = 0;
    *value = exc_value;
    *type = NULL;
    *tb = NULL;
    if (exc_value) {
        *type = (PyObject*) Py_TYPE(exc_value);
        Py_INCREF(*type);
        #if CYTHON_COMPILING_IN_CPYTHON
        *tb = ((PyBaseExceptionObject*) exc_value)->traceback;
        Py_XINCREF(*tb);
        #else
        *tb = PyException_GetTraceback(exc_value);
        #endif
    }
#else
    *type = tstate->curexc_type;
    *value = tstate->curexc_value;
    *tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
#endif
}
#endif

/* PyObjectGetAttrStr */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro))
        return tp->tp_getattro(obj, attr_name);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_getattr))
        return tp->tp_getattr(obj, PyString_AS_STRING(attr_name));
#endif
    return PyObject_GetAttr(obj, attr_name);
}
#endif

/* PyObjectGetAttrStrNoError */
#if __PYX_LIMITED_VERSION_HEX < 0x030d00A1
static void __Pyx_PyObject_GetAttrStr_ClearAttributeError(void) {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    if (likely(__Pyx_PyErr_ExceptionMatches(PyExc_AttributeError)))
        __Pyx_PyErr_Clear();
}
#endif
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStrNoError(PyObject* obj, PyObject* attr_name) {
    PyObject *result;
#if __PYX_LIMITED_VERSION_HEX >= 0x030d00A1
    (void) PyObject_GetOptionalAttr(obj, attr_name, &result);
    return result;
#else
#if CYTHON_COMPILING_IN_CPYTHON && CYTHON_USE_TYPE_SLOTS && PY_VERSION_HEX >= 0x030700B1
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro == PyObject_GenericGetAttr)) {
        return _PyObject_GenericGetAttrWithDict(obj, attr_name, NULL, 1);
    }
#endif
    result = __Pyx_PyObject_GetAttrStr(obj, attr_name);
    if (unlikely(!result)) {
        __Pyx_PyObject_GetAttrStr_ClearAttributeError();
    }
    return result;
#endif
}

/* GetBuiltinName */
static PyObject *__Pyx_GetBuiltinName(PyObject *name) {
    PyObject* result = __Pyx_PyObject_GetAttrStrNoError(__pyx_b, name);
    if (unlikely(!result) && !PyErr_Occurred()) {
        PyErr_Format(PyExc_NameError,
#if PY_MAJOR_VERSION >= 3
            "name '%U' is not defined", name);
#else
            "name '%.200s' is not defined", PyString_AS_STRING(name));
#endif
    }
    return result;
}

/* GetTopmostException */
#if CYTHON_USE_EXC_INFO_STACK && CYTHON_FAST_THREAD_STATE
static _PyErr_StackItem *
__Pyx_PyErr_GetTopmostException(PyThreadState *tstate)
{
    _PyErr_StackItem *exc_info = tstate->exc_info;
    while ((exc_info->exc_value == NULL || exc_info->exc_value == Py_None) &&
           exc_info->previous_item != NULL)
    {
        exc_info = exc_info->previous_item;
    }
    return exc_info;
}
#endif

/* SaveResetException */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
  #if CYTHON_USE_EXC_INFO_STACK && PY_VERSION_HEX >= 0x030B00a4
    _PyErr_StackItem *exc_info = __Pyx_PyErr_GetTopmostException(tstate);
    PyObject *exc_value = exc_info->exc_value;
    if (exc_value == NULL || exc_value == Py_None) {
        *value = NULL;
        *type = NULL;
        *tb = NULL;
    } else {
        *value = exc_value;
        Py_INCREF(*value);
        *type = (PyObject*) Py_TYPE(exc_value);
        Py_INCREF(*type);
        *tb = PyException_GetTraceback(exc_value);
    }
  #elif CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = __Pyx_PyErr_GetTopmostException(tstate);
    *type = exc_info->exc_type;
    *value = exc_info->exc_value;
    *tb = exc_info->exc_traceback;
    Py_XINCREF(*type);
    Py_XINCREF(*value);
    Py_XINCREF(*tb);
  #else
    *type = tstate->exc_type;
    *value = tstate->exc_value;
    *tb = tstate->exc_traceback;
    Py_XINCREF(*type);
    Py_XINCREF(*value);
    Py_XINCREF(*tb);
  #endif
}
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
  #if CYTHON_USE_EXC_INFO_STACK && PY_VERSION_HEX >= 0x030B00a4
    _PyErr_StackItem *exc_info = tstate->exc_info;
    PyObject *tmp_value = exc_info->exc_value;
    exc_info->exc_value = value;
    Py_XDECREF(tmp_value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
  #else
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = tstate->exc_info;
    tmp_type = exc_info->exc_type;
    tmp_value = exc_info->exc_value;
    tmp_tb = exc_info->exc_traceback;
    exc_info->exc_type = type;
    exc_info->exc_value = value;
    exc_info->exc_traceback = tb;
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = type;
    tstate->exc_value = value;
    tstate->exc_traceback = tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
  #endif
}
#endif

/* GetException */
#if CYTHON_FAST_THREAD_STATE
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb)
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb)
#endif
{
    PyObject *local_type = NULL, *local_value, *local_tb = NULL;
#if CYTHON_FAST_THREAD_STATE
    PyObject *tmp_type, *tmp_value, *tmp_tb;
  #if PY_VERSION_HEX >= 0x030C00A6
    local_value = tstate->current_exception;
    tstate->current_exception = 0;
    if (likely(local_value)) {
        local_type = (PyObject*) Py_TYPE(local_value);
        Py_INCREF(local_type);
        local_tb = PyException_GetTraceback(local_value);
    }
  #else
    local_type = tstate->curexc_type;
    local_value = tstate->curexc_value;
    local_tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
  #endif
#else
    PyErr_Fetch(&local_type, &local_value, &local_tb);
#endif
    PyErr_NormalizeException(&local_type, &local_value, &local_tb);
#if CYTHON_FAST_THREAD_STATE && PY_VERSION_HEX >= 0x030C00A6
    if (unlikely(tstate->current_exception))
#elif CYTHON_FAST_THREAD_STATE
    if (unlikely(tstate->curexc_type))
#else
    if (unlikely(PyErr_Occurred()))
#endif
        goto bad;
    #if PY_MAJOR_VERSION >= 3
    if (local_tb) {
        if (unlikely(PyException_SetTraceback(local_value, local_tb) < 0))
            goto bad;
    }
    #endif
    Py_XINCREF(local_tb);
    Py_XINCREF(local_type);
    Py_XINCREF(local_value);
    *type = local_type;
    *value = local_value;
    *tb = local_tb;
#if CYTHON_FAST_THREAD_STATE
    #if CYTHON_USE_EXC_INFO_STACK
    {
        _PyErr_StackItem *exc_info = tstate->exc_info;
      #if PY_VERSION_HEX >= 0x030B00a4
        tmp_value = exc_info->exc_value;
        exc_info->exc_value = local_value;
        tmp_type = NULL;
        tmp_tb = NULL;
        Py_XDECREF(local_type);
        Py_XDECREF(local_tb);
      #else
        tmp_type = exc_info->exc_type;
        tmp_value = exc_info->exc_value;
        tmp_tb = exc_info->exc_traceback;
        exc_info->exc_type = local_type;
        exc_info->exc_value = local_value;
        exc_info->exc_traceback = local_tb;
      #endif
    }
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = local_type;
    tstate->exc_value = local_value;
    tstate->exc_traceback = local_tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
#else
    PyErr_SetExcInfo(local_type, local_value, local_tb);
#endif
    return 0;
bad:
    *type = 0;
    *value = 0;
    *tb = 0;
    Py_XDECREF(local_type);
    Py_XDECREF(local_value);
    Py_XDECREF(local_tb);
    return -1;
}

/* PyObjectCall */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    PyObject *result;
    ternaryfunc call = Py_TYPE(func)->tp_call;
    if (unlikely(!call))
        return PyObject_Call(func, arg, kw);
    #if PY_MAJOR_VERSION < 3
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    #else
    if (unlikely(Py_EnterRecursiveCall(" while calling a Python object")))
        return NULL;
    #endif
    result = (*call)(func, arg, kw);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* RaiseException */
#if PY_MAJOR_VERSION < 3
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause) {
    __Pyx_PyThreadState_declare
    CYTHON_UNUSED_VAR(cause);
    Py_XINCREF(type);
    if (!value || value == Py_None)
        value = NULL;
    else
        Py_INCREF(value);
    if (!tb || tb == Py_None)
        tb = NULL;
    else {
        Py_INCREF(tb);
        if (!PyTraceBack_Check(tb)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: arg 3 must be a traceback or None");
            goto raise_error;
        }
    }
    if (PyType_Check(type)) {
#if CYTHON_COMPILING_IN_PYPY
        if (!value) {
            Py_INCREF(Py_None);
            value = Py_None;
        }
#endif
        PyErr_NormalizeException(&type, &value, &tb);
    } else {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto raise_error;
        }
        value = type;
        type = (PyObject*) Py_TYPE(type);
        Py_INCREF(type);
        if (!PyType_IsSubtype((PyTypeObject *)type, (PyTypeObject *)PyExc_BaseException)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: exception class must be a subclass of BaseException");
            goto raise_error;
        }
    }
    __Pyx_PyThreadState_assign
    __Pyx_ErrRestore(type, value, tb);
    return;
raise_error:
    Py_XDECREF(value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
    return;
}
#else
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause) {
    PyObject* owned_instance = NULL;
    if (tb == Py_None) {
        tb = 0;
    } else if (tb && !PyTraceBack_Check(tb)) {
        PyErr_SetString(PyExc_TypeError,
            "raise: arg 3 must be a traceback or None");
        goto bad;
    }
    if (value == Py_None)
        value = 0;
    if (PyExceptionInstance_Check(type)) {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto bad;
        }
        value = type;
        type = (PyObject*) Py_TYPE(value);
    } else if (PyExceptionClass_Check(type)) {
        PyObject *instance_class = NULL;
        if (value && PyExceptionInstance_Check(value)) {
            instance_class = (PyObject*) Py_TYPE(value);
            if (instance_class != type) {
                int is_subclass = PyObject_IsSubclass(instance_class, type);
                if (!is_subclass) {
                    instance_class = NULL;
                } else if (unlikely(is_subclass == -1)) {
                    goto bad;
                } else {
                    type = instance_class;
                }
            }
        }
        if (!instance_class) {
            PyObject *args;
            if (!value)
                args = PyTuple_New(0);
            else if (PyTuple_Check(value)) {
                Py_INCREF(value);
                args = value;
            } else
                args = PyTuple_Pack(1, value);
            if (!args)
                goto bad;
            owned_instance = PyObject_Call(type, args, NULL);
            Py_DECREF(args);
            if (!owned_instance)
                goto bad;
            value = owned_instance;
            if (!PyExceptionInstance_Check(value)) {
                PyErr_Format(PyExc_TypeError,
                             "calling %R should have returned an instance of "
                             "BaseException, not %R",
                             type, Py_TYPE(value));
                goto bad;
            }
        }
    } else {
        PyErr_SetString(PyExc_TypeError,
            "raise: exception class must be a subclass of BaseException");
        goto bad;
    }
    if (cause) {
        PyObject *fixed_cause;
        if (cause == Py_None) {
            fixed_cause = NULL;
        } else if (PyExceptionClass_Check(cause)) {
            fixed_cause = PyObject_CallObject(cause, NULL);
            if (fixed_cause == NULL)
                goto bad;
        } else if (PyExceptionInstance_Check(cause)) {
            fixed_cause = cause;
            Py_INCREF(fixed_cause);
        } else {
            PyErr_SetString(PyExc_TypeError,
                            "exception causes must derive from "
                            "BaseException");
            goto bad;
        }
        PyException_SetCause(value, fixed_cause);
    }
    PyErr_SetObject(type, value);
    if (tb) {
      #if PY_VERSION_HEX >= 0x030C00A6
        PyException_SetTraceback(value, tb);
      #elif CYTHON_FAST_THREAD_STATE
        PyThreadState *tstate = __Pyx_PyThreadState_Current;
        PyObject* tmp_tb = tstate->curexc_traceback;
        if (tb != tmp_tb) {
            Py_INCREF(tb);
            tstate->curexc_traceback = tb;
            Py_XDECREF(tmp_tb);
        }
#else
        PyObject *tmp_type, *tmp_value, *tmp_tb;
        PyErr_Fetch(&tmp_type, &tmp_value, &tmp_tb);
        Py_INCREF(tb);
        PyErr_Restore(tmp_type, tmp_value, tb);
        Py_XDECREF(tmp_tb);
#endif
    }
bad:
    Py_XDECREF(owned_instance);
    return;
}
#endif

/* ExtTypeTest */
static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type) {
    __Pyx_TypeName obj_type_name;
    __Pyx_TypeName type_name;
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    if (likely(__Pyx_TypeCheck(obj, type)))
        return 1;
    obj_type_name = __Pyx_PyType_GetName(Py_TYPE(obj));
    type_name = __Pyx_PyType_GetName(type);
    PyErr_Format(PyExc_TypeError,
                 "Cannot convert " __Pyx_FMT_TYPENAME " to " __Pyx_FMT_TYPENAME,
                 obj_type_name, type_name);
    __Pyx_DECREF_TypeName(obj_type_name);
    __Pyx_DECREF_TypeName(type_name);
    return 0;
}

/* RaiseUnexpectedTypeError */
static int
__Pyx_RaiseUnexpectedTypeError(const char *expected, PyObject *obj)
{
    __Pyx_TypeName obj_type_name = __Pyx_PyType_GetName(Py_TYPE(obj));
    PyErr_Format(PyExc_TypeError, "Expected %s, got " __Pyx_FMT_TYPENAME,
                 expected, obj_type_name);
    __Pyx_DECREF_TypeName(obj_type_name);
    return 0;
}

/* PyFunctionFastCall */
#if CYTHON_FAST_PYCALL && !CYTHON_VECTORCALL
static PyObject* __Pyx_PyFunction_FastCallNoKw(PyCodeObject *co, PyObject **args, Py_ssize_t na,
                                               PyObject *globals) {
    PyFrameObject *f;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject **fastlocals;
    Py_ssize_t i;
    PyObject *result;
    assert(globals != NULL);
    /* XXX Perhaps we should create a specialized
       PyFrame_New() that doesn't take locals, but does
       take builtins without sanity checking them.
       */
    assert(tstate != NULL);
    f = PyFrame_New(tstate, co, globals, NULL);
    if (f == NULL) {
        return NULL;
    }
    fastlocals = __Pyx_PyFrame_GetLocalsplus(f);
    for (i = 0; i < na; i++) {
        Py_INCREF(*args);
        fastlocals[i] = *args++;
    }
    result = PyEval_EvalFrameEx(f,0);
    ++tstate->recursion_depth;
    Py_DECREF(f);
    --tstate->recursion_depth;
    return result;
}
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, Py_ssize_t nargs, PyObject *kwargs) {
    PyCodeObject *co = (PyCodeObject *)PyFunction_GET_CODE(func);
    PyObject *globals = PyFunction_GET_GLOBALS(func);
    PyObject *argdefs = PyFunction_GET_DEFAULTS(func);
    PyObject *closure;
#if PY_MAJOR_VERSION >= 3
    PyObject *kwdefs;
#endif
    PyObject *kwtuple, **k;
    PyObject **d;
    Py_ssize_t nd;
    Py_ssize_t nk;
    PyObject *result;
    assert(kwargs == NULL || PyDict_Check(kwargs));
    nk = kwargs ? PyDict_Size(kwargs) : 0;
    #if PY_MAJOR_VERSION < 3
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object"))) {
        return NULL;
    }
    #else
    if (unlikely(Py_EnterRecursiveCall(" while calling a Python object"))) {
        return NULL;
    }
    #endif
    if (
#if PY_MAJOR_VERSION >= 3
            co->co_kwonlyargcount == 0 &&
#endif
            likely(kwargs == NULL || nk == 0) &&
            co->co_flags == (CO_OPTIMIZED | CO_NEWLOCALS | CO_NOFREE)) {
        if (argdefs == NULL && co->co_argcount == nargs) {
            result = __Pyx_PyFunction_FastCallNoKw(co, args, nargs, globals);
            goto done;
        }
        else if (nargs == 0 && argdefs != NULL
                 && co->co_argcount == Py_SIZE(argdefs)) {
            /* function called with no arguments, but all parameters have
               a default value: use default values as arguments .*/
            args = &PyTuple_GET_ITEM(argdefs, 0);
            result =__Pyx_PyFunction_FastCallNoKw(co, args, Py_SIZE(argdefs), globals);
            goto done;
        }
    }
    if (kwargs != NULL) {
        Py_ssize_t pos, i;
        kwtuple = PyTuple_New(2 * nk);
        if (kwtuple == NULL) {
            result = NULL;
            goto done;
        }
        k = &PyTuple_GET_ITEM(kwtuple, 0);
        pos = i = 0;
        while (PyDict_Next(kwargs, &pos, &k[i], &k[i+1])) {
            Py_INCREF(k[i]);
            Py_INCREF(k[i+1]);
            i += 2;
        }
        nk = i / 2;
    }
    else {
        kwtuple = NULL;
        k = NULL;
    }
    closure = PyFunction_GET_CLOSURE(func);
#if PY_MAJOR_VERSION >= 3
    kwdefs = PyFunction_GET_KW_DEFAULTS(func);
#endif
    if (argdefs != NULL) {
        d = &PyTuple_GET_ITEM(argdefs, 0);
        nd = Py_SIZE(argdefs);
    }
    else {
        d = NULL;
        nd = 0;
    }
#if PY_MAJOR_VERSION >= 3
    result = PyEval_EvalCodeEx((PyObject*)co, globals, (PyObject *)NULL,
                               args, (int)nargs,
                               k, (int)nk,
                               d, (int)nd, kwdefs, closure);
#else
    result = PyEval_EvalCodeEx(co, globals, (PyObject *)NULL,
                               args, (int)nargs,
                               k, (int)nk,
                               d, (int)nd, closure);
#endif
    Py_XDECREF(kwtuple);
done:
    Py_LeaveRecursiveCall();
    return result;
}
#endif

/* PyObjectCallMethO */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg) {
    PyObject *self, *result;
    PyCFunction cfunc;
    cfunc = __Pyx_CyOrPyCFunction_GET_FUNCTION(func);
    self = __Pyx_CyOrPyCFunction_GET_SELF(func);
    #if PY_MAJOR_VERSION < 3
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    #else
    if (unlikely(Py_EnterRecursiveCall(" while calling a Python object")))
        return NULL;
    #endif
    result = cfunc(self, arg);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectFastCall */
#if PY_VERSION_HEX < 0x03090000 || CYTHON_COMPILING_IN_LIMITED_API
static PyObject* __Pyx_PyObject_FastCall_fallback(PyObject *func, PyObject **args, size_t nargs, PyObject *kwargs) {
    PyObject *argstuple;
    PyObject *result = 0;
    size_t i;
    argstuple = PyTuple_New((Py_ssize_t)nargs);
    if (unlikely(!argstuple)) return NULL;
    for (i = 0; i < nargs; i++) {
        Py_INCREF(args[i]);
        if (__Pyx_PyTuple_SET_ITEM(argstuple, (Py_ssize_t)i, args[i]) < 0) goto bad;
    }
    result = __Pyx_PyObject_Call(func, argstuple, kwargs);
  bad:
    Py_DECREF(argstuple);
    return result;
}
#endif
static CYTHON_INLINE PyObject* __Pyx_PyObject_FastCallDict(PyObject *func, PyObject **args, size_t _nargs, PyObject *kwargs) {
    Py_ssize_t nargs = __Pyx_PyVectorcall_NARGS(_nargs);
#if CYTHON_COMPILING_IN_CPYTHON
    if (nargs == 0 && kwargs == NULL) {
        if (__Pyx_CyOrPyCFunction_Check(func) && likely( __Pyx_CyOrPyCFunction_GET_FLAGS(func) & METH_NOARGS))
            return __Pyx_PyObject_CallMethO(func, NULL);
    }
    else if (nargs == 1 && kwargs == NULL) {
        if (__Pyx_CyOrPyCFunction_Check(func) && likely( __Pyx_CyOrPyCFunction_GET_FLAGS(func) & METH_O))
            return __Pyx_PyObject_CallMethO(func, args[0]);
    }
#endif
    #if PY_VERSION_HEX < 0x030800B1
    #if CYTHON_FAST_PYCCALL
    if (PyCFunction_Check(func)) {
        if (kwargs) {
            return _PyCFunction_FastCallDict(func, args, nargs, kwargs);
        } else {
            return _PyCFunction_FastCallKeywords(func, args, nargs, NULL);
        }
    }
    #if PY_VERSION_HEX >= 0x030700A1
    if (!kwargs && __Pyx_IS_TYPE(func, &PyMethodDescr_Type)) {
        return _PyMethodDescr_FastCallKeywords(func, args, nargs, NULL);
    }
    #endif
    #endif
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCallDict(func, args, nargs, kwargs);
    }
    #endif
    #endif
    if (kwargs == NULL) {
        #if CYTHON_VECTORCALL
        #if PY_VERSION_HEX < 0x03090000
        vectorcallfunc f = _PyVectorcall_Function(func);
        #else
        vectorcallfunc f = PyVectorcall_Function(func);
        #endif
        if (f) {
            return f(func, args, (size_t)nargs, NULL);
        }
        #elif defined(__Pyx_CyFunction_USED) && CYTHON_BACKPORT_VECTORCALL
        if (__Pyx_CyFunction_CheckExact(func)) {
            __pyx_vectorcallfunc f = __Pyx_CyFunction_func_vectorcall(func);
            if (f) return f(func, args, (size_t)nargs, NULL);
        }
        #endif
    }
    if (nargs == 0) {
        return __Pyx_PyObject_Call(func, __pyx_empty_tuple, kwargs);
    }
    #if PY_VERSION_HEX >= 0x03090000 && !CYTHON_COMPILING_IN_LIMITED_API
    return PyObject_VectorcallDict(func, args, (size_t)nargs, kwargs);
    #else
    return __Pyx_PyObject_FastCall_fallback(func, args, (size_t)nargs, kwargs);
    #endif
}

/* PyObjectCallOneArg */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *args[2] = {NULL, arg};
    return __Pyx_PyObject_FastCall(func, args+1, 1 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET);
}

/* SwapException */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx__ExceptionSwap(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
  #if CYTHON_USE_EXC_INFO_STACK && PY_VERSION_HEX >= 0x030B00a4
    _PyErr_StackItem *exc_info = tstate->exc_info;
    tmp_value = exc_info->exc_value;
    exc_info->exc_value = *value;
    if (tmp_value == NULL || tmp_value == Py_None) {
        Py_XDECREF(tmp_value);
        tmp_value = NULL;
        tmp_type = NULL;
        tmp_tb = NULL;
    } else {
        tmp_type = (PyObject*) Py_TYPE(tmp_value);
        Py_INCREF(tmp_type);
        #if CYTHON_COMPILING_IN_CPYTHON
        tmp_tb = ((PyBaseExceptionObject*) tmp_value)->traceback;
        Py_XINCREF(tmp_tb);
        #else
        tmp_tb = PyException_GetTraceback(tmp_value);
        #endif
    }
  #elif CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = tstate->exc_info;
    tmp_type = exc_info->exc_type;
    tmp_value = exc_info->exc_value;
    tmp_tb = exc_info->exc_traceback;
    exc_info->exc_type = *type;
    exc_info->exc_value = *value;
    exc_info->exc_traceback = *tb;
  #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = *type;
    tstate->exc_value = *value;
    tstate->exc_traceback = *tb;
  #endif
    *type = tmp_type;
    *value = tmp_value;
    *tb = tmp_tb;
}
#else
static CYTHON_INLINE void __Pyx_ExceptionSwap(PyObject **type, PyObject **value, PyObject **tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    PyErr_GetExcInfo(&tmp_type, &tmp_value, &tmp_tb);
    PyErr_SetExcInfo(*type, *value, *tb);
    *type = tmp_type;
    *value = tmp_value;
    *tb = tmp_tb;
}
#endif

/* PyDictVersioning */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    return likely(dict) ? __PYX_GET_DICT_VERSION(dict) : 0;
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj) {
    PyObject **dictptr = NULL;
    Py_ssize_t offset = Py_TYPE(obj)->tp_dictoffset;
    if (offset) {
#if CYTHON_COMPILING_IN_CPYTHON
        dictptr = (likely(offset > 0)) ? (PyObject **) ((char *)obj + offset) : _PyObject_GetDictPtr(obj);
#else
        dictptr = _PyObject_GetDictPtr(obj);
#endif
    }
    return (dictptr && *dictptr) ? __PYX_GET_DICT_VERSION(*dictptr) : 0;
}
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    if (unlikely(!dict) || unlikely(tp_dict_version != __PYX_GET_DICT_VERSION(dict)))
        return 0;
    return obj_dict_version == __Pyx_get_object_dict_version(obj);
}
#endif

/* GetModuleGlobalName */
#if CYTHON_USE_DICT_VERSIONS
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value)
#else
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name)
#endif
{
    PyObject *result;
#if !CYTHON_AVOID_BORROWED_REFS
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1 && PY_VERSION_HEX < 0x030d0000
    result = _PyDict_GetItem_KnownHash(__pyx_d, name, ((PyASCIIObject *) name)->hash);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    } else if (unlikely(PyErr_Occurred())) {
        return NULL;
    }
#elif CYTHON_COMPILING_IN_LIMITED_API
    if (unlikely(!__pyx_m)) {
        return NULL;
    }
    result = PyObject_GetAttr(__pyx_m, name);
    if (likely(result)) {
        return result;
    }
#else
    result = PyDict_GetItem(__pyx_d, name);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
#endif
#else
    result = PyObject_GetItem(__pyx_d, name);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
    PyErr_Clear();
#endif
    return __Pyx_GetBuiltinName(name);
}

/* PyObjectCall2Args */
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2) {
    PyObject *args[3] = {NULL, arg1, arg2};
    return __Pyx_PyObject_FastCall(function, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET);
}

/* PyObjectGetMethod */
static int __Pyx_PyObject_GetMethod(PyObject *obj, PyObject *name, PyObject **method) {
    PyObject *attr;
#if CYTHON_UNPACK_METHODS && CYTHON_COMPILING_IN_CPYTHON && CYTHON_USE_PYTYPE_LOOKUP
    __Pyx_TypeName type_name;
    PyTypeObject *tp = Py_TYPE(obj);
    PyObject *descr;
    descrgetfunc f = NULL;
    PyObject **dictptr, *dict;
    int meth_found = 0;
    assert (*method == NULL);
    if (unlikely(tp->tp_getattro != PyObject_GenericGetAttr)) {
        attr = __Pyx_PyObject_GetAttrStr(obj, name);
        goto try_unpack;
    }
    if (unlikely(tp->tp_dict == NULL) && unlikely(PyType_Ready(tp) < 0)) {
        return 0;
    }
    descr = _PyType_Lookup(tp, name);
    if (likely(descr != NULL)) {
        Py_INCREF(descr);
#if defined(Py_TPFLAGS_METHOD_DESCRIPTOR) && Py_TPFLAGS_METHOD_DESCRIPTOR
        if (__Pyx_PyType_HasFeature(Py_TYPE(descr), Py_TPFLAGS_METHOD_DESCRIPTOR))
#elif PY_MAJOR_VERSION >= 3
        #ifdef __Pyx_CyFunction_USED
        if (likely(PyFunction_Check(descr) || __Pyx_IS_TYPE(descr, &PyMethodDescr_Type) || __Pyx_CyFunction_Check(descr)))
        #else
        if (likely(PyFunction_Check(descr) || __Pyx_IS_TYPE(descr, &PyMethodDescr_Type)))
        #endif
#else
        #ifdef __Pyx_CyFunction_USED
        if (likely(PyFunction_Check(descr) || __Pyx_CyFunction_Check(descr)))
        #else
        if (likely(PyFunction_Check(descr)))
        #endif
#endif
        {
            meth_found = 1;
        } else {
            f = Py_TYPE(descr)->tp_descr_get;
            if (f != NULL && PyDescr_IsData(descr)) {
                attr = f(descr, obj, (PyObject *)Py_TYPE(obj));
                Py_DECREF(descr);
                goto try_unpack;
            }
        }
    }
    dictptr = _PyObject_GetDictPtr(obj);
    if (dictptr != NULL && (dict = *dictptr) != NULL) {
        Py_INCREF(dict);
        attr = __Pyx_PyDict_GetItemStr(dict, name);
        if (attr != NULL) {
            Py_INCREF(attr);
            Py_DECREF(dict);
            Py_XDECREF(descr);
            goto try_unpack;
        }
        Py_DECREF(dict);
    }
    if (meth_found) {
        *method = descr;
        return 1;
    }
    if (f != NULL) {
        attr = f(descr, obj, (PyObject *)Py_TYPE(obj));
        Py_DECREF(descr);
        goto try_unpack;
    }
    if (likely(descr != NULL)) {
        *method = descr;
        return 0;
    }
    type_name = __Pyx_PyType_GetName(tp);
    PyErr_Format(PyExc_AttributeError,
#if PY_MAJOR_VERSION >= 3
                 "'" __Pyx_FMT_TYPENAME "' object has no attribute '%U'",
                 type_name, name);
#else
                 "'" __Pyx_FMT_TYPENAME "' object has no attribute '%.400s'",
                 type_name, PyString_AS_STRING(name));
#endif
    __Pyx_DECREF_TypeName(type_name);
    return 0;
#else
    attr = __Pyx_PyObject_GetAttrStr(obj, name);
    goto try_unpack;
#endif
try_unpack:
#if CYTHON_UNPACK_METHODS
    if (likely(attr) && PyMethod_Check(attr) && likely(PyMethod_GET_SELF(attr) == obj)) {
        PyObject *function = PyMethod_GET_FUNCTION(attr);
        Py_INCREF(function);
        Py_DECREF(attr);
        *method = function;
        return 1;
    }
#endif
    *method = attr;
    return 0;
}

/* PyObjectCallMethod1 */
#if !(CYTHON_VECTORCALL && __PYX_LIMITED_VERSION_HEX >= 0x030C00A2)
static PyObject* __Pyx__PyObject_CallMethod1(PyObject* method, PyObject* arg) {
    PyObject *result = __Pyx_PyObject_CallOneArg(method, arg);
    Py_DECREF(method);
    return result;
}
#endif
static PyObject* __Pyx_PyObject_CallMethod1(PyObject* obj, PyObject* method_name, PyObject* arg) {
#if CYTHON_VECTORCALL && __PYX_LIMITED_VERSION_HEX >= 0x030C00A2
    PyObject *args[2] = {obj, arg};
    (void) __Pyx_PyObject_GetMethod;
    (void) __Pyx_PyObject_CallOneArg;
    (void) __Pyx_PyObject_Call2Args;
    return PyObject_VectorcallMethod(method_name, args, 2 | PY_VECTORCALL_ARGUMENTS_OFFSET, NULL);
#else
    PyObject *method = NULL, *result;
    int is_method = __Pyx_PyObject_GetMethod(obj, method_name, &method);
    if (likely(is_method)) {
        result = __Pyx_PyObject_Call2Args(method, obj, arg);
        Py_DECREF(method);
        return result;
    }
    if (unlikely(!method)) return NULL;
    return __Pyx__PyObject_CallMethod1(method, arg);
#endif
}

/* StringJoin */
static CYTHON_INLINE PyObject* __Pyx_PyBytes_Join(PyObject* sep, PyObject* values) {
    (void) __Pyx_PyObject_CallMethod1;
#if CYTHON_COMPILING_IN_CPYTHON && PY_MAJOR_VERSION < 3
    return _PyString_Join(sep, values);
#elif CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030d0000
    return _PyBytes_Join(sep, values);
#else
    return __Pyx_PyObject_CallMethod1(sep, __pyx_n_s_join, values);
#endif
}

/* UnpackUnboundCMethod */
static PyObject *__Pyx_SelflessCall(PyObject *method, PyObject *args, PyObject *kwargs) {
    PyObject *result;
    PyObject *selfless_args = PyTuple_GetSlice(args, 1, PyTuple_Size(args));
    if (unlikely(!selfless_args)) return NULL;
    result = PyObject_Call(method, selfless_args, kwargs);
    Py_DECREF(selfless_args);
    return result;
}
static PyMethodDef __Pyx_UnboundCMethod_Def = {
     "CythonUnboundCMethod",
     __PYX_REINTERPRET_FUNCION(PyCFunction, __Pyx_SelflessCall),
     METH_VARARGS | METH_KEYWORDS,
     NULL
};
static int __Pyx_TryUnpackUnboundCMethod(__Pyx_CachedCFunction* target) {
    PyObject *method;
    method = __Pyx_PyObject_GetAttrStr(target->type, *target->method_name);
    if (unlikely(!method))
        return -1;
    target->method = method;
#if CYTHON_COMPILING_IN_CPYTHON
    #if PY_MAJOR_VERSION >= 3
    if (likely(__Pyx_TypeCheck(method, &PyMethodDescr_Type)))
    #else
    if (likely(!__Pyx_CyOrPyCFunction_Check(method)))
    #endif
    {
        PyMethodDescrObject *descr = (PyMethodDescrObject*) method;
        target->func = descr->d_method->ml_meth;
        target->flag = descr->d_method->ml_flags & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_STACKLESS);
    } else
#endif
#if CYTHON_COMPILING_IN_PYPY
#else
    if (PyCFunction_Check(method))
#endif
    {
        PyObject *self;
        int self_found;
#if CYTHON_COMPILING_IN_LIMITED_API || CYTHON_COMPILING_IN_PYPY
        self = PyObject_GetAttrString(method, "__self__");
        if (!self) {
            PyErr_Clear();
        }
#else
        self = PyCFunction_GET_SELF(method);
#endif
        self_found = (self && self != Py_None);
#if CYTHON_COMPILING_IN_LIMITED_API || CYTHON_COMPILING_IN_PYPY
        Py_XDECREF(self);
#endif
        if (self_found) {
            PyObject *unbound_method = PyCFunction_New(&__Pyx_UnboundCMethod_Def, method);
            if (unlikely(!unbound_method)) return -1;
            Py_DECREF(method);
            target->method = unbound_method;
        }
    }
    return 0;
}

/* CallUnboundCMethod0 */
static PyObject* __Pyx__CallUnboundCMethod0(__Pyx_CachedCFunction* cfunc, PyObject* self) {
    PyObject *args, *result = NULL;
    if (unlikely(!cfunc->method) && unlikely(__Pyx_TryUnpackUnboundCMethod(cfunc) < 0)) return NULL;
#if CYTHON_ASSUME_SAFE_MACROS
    args = PyTuple_New(1);
    if (unlikely(!args)) goto bad;
    Py_INCREF(self);
    PyTuple_SET_ITEM(args, 0, self);
#else
    args = PyTuple_Pack(1, self);
    if (unlikely(!args)) goto bad;
#endif
    result = __Pyx_PyObject_Call(cfunc->method, args, NULL);
    Py_DECREF(args);
bad:
    return result;
}

/* IterFinish */
static CYTHON_INLINE int __Pyx_IterFinish(void) {
    PyObject* exc_type;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    exc_type = __Pyx_PyErr_CurrentExceptionType();
    if (unlikely(exc_type)) {
        if (unlikely(!__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration)))
            return -1;
        __Pyx_PyErr_Clear();
        return 0;
    }
    return 0;
}

/* set_iter */
static CYTHON_INLINE PyObject* __Pyx_set_iterator(PyObject* iterable, int is_set,
                                                  Py_ssize_t* p_orig_length, int* p_source_is_set) {
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030d0000
    is_set = is_set || likely(PySet_CheckExact(iterable) || PyFrozenSet_CheckExact(iterable));
    *p_source_is_set = is_set;
    if (likely(is_set)) {
        *p_orig_length = PySet_Size(iterable);
        Py_INCREF(iterable);
        return iterable;
    }
#else
    CYTHON_UNUSED_VAR(is_set);
    *p_source_is_set = 0;
#endif
    *p_orig_length = 0;
    return PyObject_GetIter(iterable);
}
static CYTHON_INLINE int __Pyx_set_iter_next(
        PyObject* iter_obj, Py_ssize_t orig_length,
        Py_ssize_t* ppos, PyObject **value,
        int source_is_set) {
    if (!CYTHON_COMPILING_IN_CPYTHON || PY_VERSION_HEX >= 0x030d0000 || unlikely(!source_is_set)) {
        *value = PyIter_Next(iter_obj);
        if (unlikely(!*value)) {
            return __Pyx_IterFinish();
        }
        CYTHON_UNUSED_VAR(orig_length);
        CYTHON_UNUSED_VAR(ppos);
        return 1;
    }
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030d0000
    if (unlikely(PySet_GET_SIZE(iter_obj) != orig_length)) {
        PyErr_SetString(
            PyExc_RuntimeError,
            "set changed size during iteration");
        return -1;
    }
    {
        Py_hash_t hash;
        int ret = _PySet_NextEntry(iter_obj, ppos, value, &hash);
        assert (ret != -1);
        if (likely(ret)) {
            Py_INCREF(*value);
            return 1;
        }
    }
#endif
    return 0;
}

/* PyObjectCallNoArg */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func) {
    PyObject *arg[2] = {NULL, NULL};
    return __Pyx_PyObject_FastCall(func, arg + 1, 0 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET);
}

/* PyObjectCallMethod0 */
static PyObject* __Pyx_PyObject_CallMethod0(PyObject* obj, PyObject* method_name) {
    PyObject *method = NULL, *result = NULL;
    int is_method = __Pyx_PyObject_GetMethod(obj, method_name, &method);
    if (likely(is_method)) {
        result = __Pyx_PyObject_CallOneArg(method, obj);
        Py_DECREF(method);
        return result;
    }
    if (unlikely(!method)) goto bad;
    result = __Pyx_PyObject_CallNoArg(method);
    Py_DECREF(method);
bad:
    return result;
}

/* RaiseNeedMoreValuesToUnpack */
static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index) {
    PyErr_Format(PyExc_ValueError,
                 "need more than %" CYTHON_FORMAT_SSIZE_T "d value%.1s to unpack",
                 index, (index == 1) ? "" : "s");
}

/* RaiseTooManyValuesToUnpack */
static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected) {
    PyErr_Format(PyExc_ValueError,
                 "too many values to unpack (expected %" CYTHON_FORMAT_SSIZE_T "d)", expected);
}

/* UnpackItemEndCheck */
static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected) {
    if (unlikely(retval)) {
        Py_DECREF(retval);
        __Pyx_RaiseTooManyValuesError(expected);
        return -1;
    }
    return __Pyx_IterFinish();
}

/* RaiseNoneIterError */
static CYTHON_INLINE void __Pyx_RaiseNoneNotIterableError(void) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
}

/* UnpackTupleError */
static void __Pyx_UnpackTupleError(PyObject *t, Py_ssize_t index) {
    if (t == Py_None) {
      __Pyx_RaiseNoneNotIterableError();
    } else if (PyTuple_GET_SIZE(t) < index) {
      __Pyx_RaiseNeedMoreValuesError(PyTuple_GET_SIZE(t));
    } else {
      __Pyx_RaiseTooManyValuesError(index);
    }
}

/* UnpackTuple2 */
static CYTHON_INLINE int __Pyx_unpack_tuple2_exact(
        PyObject* tuple, PyObject** pvalue1, PyObject** pvalue2, int decref_tuple) {
    PyObject *value1 = NULL, *value2 = NULL;
#if CYTHON_COMPILING_IN_PYPY
    value1 = PySequence_ITEM(tuple, 0);  if (unlikely(!value1)) goto bad;
    value2 = PySequence_ITEM(tuple, 1);  if (unlikely(!value2)) goto bad;
#else
    value1 = PyTuple_GET_ITEM(tuple, 0);  Py_INCREF(value1);
    value2 = PyTuple_GET_ITEM(tuple, 1);  Py_INCREF(value2);
#endif
    if (decref_tuple) {
        Py_DECREF(tuple);
    }
    *pvalue1 = value1;
    *pvalue2 = value2;
    return 0;
#if CYTHON_COMPILING_IN_PYPY
bad:
    Py_XDECREF(value1);
    Py_XDECREF(value2);
    if (decref_tuple) { Py_XDECREF(tuple); }
    return -1;
#endif
}
static int __Pyx_unpack_tuple2_generic(PyObject* tuple, PyObject** pvalue1, PyObject** pvalue2,
                                       int has_known_size, int decref_tuple) {
    Py_ssize_t index;
    PyObject *value1 = NULL, *value2 = NULL, *iter = NULL;
    iternextfunc iternext;
    iter = PyObject_GetIter(tuple);
    if (unlikely(!iter)) goto bad;
    if (decref_tuple) { Py_DECREF(tuple); tuple = NULL; }
    iternext = __Pyx_PyObject_GetIterNextFunc(iter);
    value1 = iternext(iter); if (unlikely(!value1)) { index = 0; goto unpacking_failed; }
    value2 = iternext(iter); if (unlikely(!value2)) { index = 1; goto unpacking_failed; }
    if (!has_known_size && unlikely(__Pyx_IternextUnpackEndCheck(iternext(iter), 2))) goto bad;
    Py_DECREF(iter);
    *pvalue1 = value1;
    *pvalue2 = value2;
    return 0;
unpacking_failed:
    if (!has_known_size && __Pyx_IterFinish() == 0)
        __Pyx_RaiseNeedMoreValuesError(index);
bad:
    Py_XDECREF(iter);
    Py_XDECREF(value1);
    Py_XDECREF(value2);
    if (decref_tuple) { Py_XDECREF(tuple); }
    return -1;
}

/* dict_iter */
#if CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
#include <string.h>
#endif
static CYTHON_INLINE PyObject* __Pyx_dict_iterator(PyObject* iterable, int is_dict, PyObject* method_name,
                                                   Py_ssize_t* p_orig_length, int* p_source_is_dict) {
    is_dict = is_dict || likely(PyDict_CheckExact(iterable));
    *p_source_is_dict = is_dict;
    if (is_dict) {
#if !CYTHON_COMPILING_IN_PYPY
        *p_orig_length = PyDict_Size(iterable);
        Py_INCREF(iterable);
        return iterable;
#elif PY_MAJOR_VERSION >= 3
        static PyObject *py_items = NULL, *py_keys = NULL, *py_values = NULL;
        PyObject **pp = NULL;
        if (method_name) {
            const char *name = PyUnicode_AsUTF8(method_name);
            if (strcmp(name, "iteritems") == 0) pp = &py_items;
            else if (strcmp(name, "iterkeys") == 0) pp = &py_keys;
            else if (strcmp(name, "itervalues") == 0) pp = &py_values;
            if (pp) {
                if (!*pp) {
                    *pp = PyUnicode_FromString(name + 4);
                    if (!*pp)
                        return NULL;
                }
                method_name = *pp;
            }
        }
#endif
    }
    *p_orig_length = 0;
    if (method_name) {
        PyObject* iter;
        iterable = __Pyx_PyObject_CallMethod0(iterable, method_name);
        if (!iterable)
            return NULL;
#if !CYTHON_COMPILING_IN_PYPY
        if (PyTuple_CheckExact(iterable) || PyList_CheckExact(iterable))
            return iterable;
#endif
        iter = PyObject_GetIter(iterable);
        Py_DECREF(iterable);
        return iter;
    }
    return PyObject_GetIter(iterable);
}
static CYTHON_INLINE int __Pyx_dict_iter_next(
        PyObject* iter_obj, CYTHON_NCP_UNUSED Py_ssize_t orig_length, CYTHON_NCP_UNUSED Py_ssize_t* ppos,
        PyObject** pkey, PyObject** pvalue, PyObject** pitem, int source_is_dict) {
    PyObject* next_item;
#if !CYTHON_COMPILING_IN_PYPY
    if (source_is_dict) {
        PyObject *key, *value;
        if (unlikely(orig_length != PyDict_Size(iter_obj))) {
            PyErr_SetString(PyExc_RuntimeError, "dictionary changed size during iteration");
            return -1;
        }
        if (unlikely(!PyDict_Next(iter_obj, ppos, &key, &value))) {
            return 0;
        }
        if (pitem) {
            PyObject* tuple = PyTuple_New(2);
            if (unlikely(!tuple)) {
                return -1;
            }
            Py_INCREF(key);
            Py_INCREF(value);
            PyTuple_SET_ITEM(tuple, 0, key);
            PyTuple_SET_ITEM(tuple, 1, value);
            *pitem = tuple;
        } else {
            if (pkey) {
                Py_INCREF(key);
                *pkey = key;
            }
            if (pvalue) {
                Py_INCREF(value);
                *pvalue = value;
            }
        }
        return 1;
    } else if (PyTuple_CheckExact(iter_obj)) {
        Py_ssize_t pos = *ppos;
        if (unlikely(pos >= PyTuple_GET_SIZE(iter_obj))) return 0;
        *ppos = pos + 1;
        next_item = PyTuple_GET_ITEM(iter_obj, pos);
        Py_INCREF(next_item);
    } else if (PyList_CheckExact(iter_obj)) {
        Py_ssize_t pos = *ppos;
        if (unlikely(pos >= PyList_GET_SIZE(iter_obj))) return 0;
        *ppos = pos + 1;
        next_item = PyList_GET_ITEM(iter_obj, pos);
        Py_INCREF(next_item);
    } else
#endif
    {
        next_item = PyIter_Next(iter_obj);
        if (unlikely(!next_item)) {
            return __Pyx_IterFinish();
        }
    }
    if (pitem) {
        *pitem = next_item;
    } else if (pkey && pvalue) {
        if (__Pyx_unpack_tuple2(next_item, pkey, pvalue, source_is_dict, source_is_dict, 1))
            return -1;
    } else if (pkey) {
        *pkey = next_item;
    } else {
        *pvalue = next_item;
    }
    return 1;
}

/* pybytes_as_double */
static double __Pyx_SlowPyString_AsDouble(PyObject *obj) {
    PyObject *float_value;
#if PY_MAJOR_VERSION >= 3
    float_value = PyFloat_FromString(obj);
#else
    float_value = PyFloat_FromString(obj, 0);
#endif
    if (likely(float_value)) {
#if CYTHON_ASSUME_SAFE_MACROS
        double value = PyFloat_AS_DOUBLE(float_value);
#else
        double value = PyFloat_AsDouble(float_value);
#endif
        Py_DECREF(float_value);
        return value;
    }
    return (double)-1;
}
static const char* __Pyx__PyBytes_AsDouble_Copy(const char* start, char* buffer, Py_ssize_t length) {
    int last_was_punctuation = 1;
    Py_ssize_t i;
    for (i=0; i < length; i++) {
        char chr = start[i];
        int is_punctuation = (chr == '_') | (chr == '.') | (chr == 'e') | (chr == 'E');
        *buffer = chr;
        buffer += (chr != '_');
        if (unlikely(last_was_punctuation & is_punctuation)) goto parse_failure;
        last_was_punctuation = is_punctuation;
    }
    if (unlikely(last_was_punctuation)) goto parse_failure;
    *buffer = '\0';
    return buffer;
parse_failure:
    return NULL;
}
static double __Pyx__PyBytes_AsDouble_inf_nan(const char* start, Py_ssize_t length) {
    int matches = 1;
    char sign = start[0];
    int is_signed = (sign == '+') | (sign == '-');
    start += is_signed;
    length -= is_signed;
    switch (start[0]) {
        #ifdef Py_NAN
        case 'n':
        case 'N':
            if (unlikely(length != 3)) goto parse_failure;
            matches &= (start[1] == 'a' || start[1] == 'A');
            matches &= (start[2] == 'n' || start[2] == 'N');
            if (unlikely(!matches)) goto parse_failure;
            return (sign == '-') ? -Py_NAN : Py_NAN;
        #endif
        case 'i':
        case 'I':
            if (unlikely(length < 3)) goto parse_failure;
            matches &= (start[1] == 'n' || start[1] == 'N');
            matches &= (start[2] == 'f' || start[2] == 'F');
            if (likely(length == 3 && matches))
                return (sign == '-') ? -Py_HUGE_VAL : Py_HUGE_VAL;
            if (unlikely(length != 8)) goto parse_failure;
            matches &= (start[3] == 'i' || start[3] == 'I');
            matches &= (start[4] == 'n' || start[4] == 'N');
            matches &= (start[5] == 'i' || start[5] == 'I');
            matches &= (start[6] == 't' || start[6] == 'T');
            matches &= (start[7] == 'y' || start[7] == 'Y');
            if (unlikely(!matches)) goto parse_failure;
            return (sign == '-') ? -Py_HUGE_VAL : Py_HUGE_VAL;
        case '.': case '0': case '1': case '2': case '3': case '4': case '5': case '6': case '7': case '8': case '9':
            break;
        default:
            goto parse_failure;
    }
    return 0.0;
parse_failure:
    return -1.0;
}
static CYTHON_INLINE int __Pyx__PyBytes_AsDouble_IsSpace(char ch) {
    return (ch == 0x20) | !((ch < 0x9) | (ch > 0xd));
}
CYTHON_UNUSED static double __Pyx__PyBytes_AsDouble(PyObject *obj, const char* start, Py_ssize_t length) {
    double value;
    Py_ssize_t i, digits;
    const char *last = start + length;
    char *end;
    while (__Pyx__PyBytes_AsDouble_IsSpace(*start))
        start++;
    while (start < last - 1 && __Pyx__PyBytes_AsDouble_IsSpace(last[-1]))
        last--;
    length = last - start;
    if (unlikely(length <= 0)) goto fallback;
    value = __Pyx__PyBytes_AsDouble_inf_nan(start, length);
    if (unlikely(value == -1.0)) goto fallback;
    if (value != 0.0) return value;
    digits = 0;
    for (i=0; i < length; digits += start[i++] != '_');
    if (likely(digits == length)) {
        value = PyOS_string_to_double(start, &end, NULL);
    } else if (digits < 40) {
        char number[40];
        last = __Pyx__PyBytes_AsDouble_Copy(start, number, length);
        if (unlikely(!last)) goto fallback;
        value = PyOS_string_to_double(number, &end, NULL);
    } else {
        char *number = (char*) PyMem_Malloc((digits + 1) * sizeof(char));
        if (unlikely(!number)) goto fallback;
        last = __Pyx__PyBytes_AsDouble_Copy(start, number, length);
        if (unlikely(!last)) {
            PyMem_Free(number);
            goto fallback;
        }
        value = PyOS_string_to_double(number, &end, NULL);
        PyMem_Free(number);
    }
    if (likely(end == last) || (value == (double)-1 && PyErr_Occurred())) {
        return value;
    }
fallback:
    return __Pyx_SlowPyString_AsDouble(obj);
}

/* pynumber_float */
static CYTHON_INLINE PyObject* __Pyx__PyNumber_Float(PyObject* obj) {
    double val;
    if (PyLong_CheckExact(obj)) {
#if CYTHON_USE_PYLONG_INTERNALS
        if (likely(__Pyx_PyLong_IsCompact(obj))) {
            val = (double) __Pyx_PyLong_CompactValue(obj);
            goto no_error;
        }
#endif
        val = PyLong_AsDouble(obj);
    } else if (PyUnicode_CheckExact(obj)) {
        val = __Pyx_PyUnicode_AsDouble(obj);
    } else if (PyBytes_CheckExact(obj)) {
        val = __Pyx_PyBytes_AsDouble(obj);
    } else if (PyByteArray_CheckExact(obj)) {
        val = __Pyx_PyByteArray_AsDouble(obj);
    } else {
        return PyNumber_Float(obj);
    }
    if (unlikely(val == -1 && PyErr_Occurred())) {
        return NULL;
    }
#if CYTHON_USE_PYLONG_INTERNALS
no_error:
#endif
    return PyFloat_FromDouble(val);
}

/* GetItemInt */
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j) {
    PyObject *r;
    if (unlikely(!j)) return NULL;
    r = PyObject_GetItem(o, j);
    Py_DECREF(j);
    return r;
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyList_GET_SIZE(o);
    }
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyList_GET_SIZE(o)))) {
        PyObject *r = PyList_GET_ITEM(o, wrapped_i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyTuple_GET_SIZE(o);
    }
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyTuple_GET_SIZE(o)))) {
        PyObject *r = PyTuple_GET_ITEM(o, wrapped_i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i, int is_list,
                                                     CYTHON_NCP_UNUSED int wraparound,
                                                     CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS && CYTHON_USE_TYPE_SLOTS
    if (is_list || PyList_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyList_GET_SIZE(o);
        if ((!boundscheck) || (likely(__Pyx_is_valid_index(n, PyList_GET_SIZE(o))))) {
            PyObject *r = PyList_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    }
    else if (PyTuple_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyTuple_GET_SIZE(o);
        if ((!boundscheck) || likely(__Pyx_is_valid_index(n, PyTuple_GET_SIZE(o)))) {
            PyObject *r = PyTuple_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    } else {
        PyMappingMethods *mm = Py_TYPE(o)->tp_as_mapping;
        PySequenceMethods *sm = Py_TYPE(o)->tp_as_sequence;
        if (mm && mm->mp_subscript) {
            PyObject *r, *key = PyInt_FromSsize_t(i);
            if (unlikely(!key)) return NULL;
            r = mm->mp_subscript(o, key);
            Py_DECREF(key);
            return r;
        }
        if (likely(sm && sm->sq_item)) {
            if (wraparound && unlikely(i < 0) && likely(sm->sq_length)) {
                Py_ssize_t l = sm->sq_length(o);
                if (likely(l >= 0)) {
                    i += l;
                } else {
                    if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                        return NULL;
                    PyErr_Clear();
                }
            }
            return sm->sq_item(o, i);
        }
    }
#else
    if (is_list || !PyMapping_Check(o)) {
        return PySequence_GetItem(o, i);
    }
#endif
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
}

/* ObjectGetItem */
#if CYTHON_USE_TYPE_SLOTS
static PyObject *__Pyx_PyObject_GetIndex(PyObject *obj, PyObject *index) {
    PyObject *runerr = NULL;
    Py_ssize_t key_value;
    key_value = __Pyx_PyIndex_AsSsize_t(index);
    if (likely(key_value != -1 || !(runerr = PyErr_Occurred()))) {
        return __Pyx_GetItemInt_Fast(obj, key_value, 0, 1, 1);
    }
    if (PyErr_GivenExceptionMatches(runerr, PyExc_OverflowError)) {
        __Pyx_TypeName index_type_name = __Pyx_PyType_GetName(Py_TYPE(index));
        PyErr_Clear();
        PyErr_Format(PyExc_IndexError,
            "cannot fit '" __Pyx_FMT_TYPENAME "' into an index-sized integer", index_type_name);
        __Pyx_DECREF_TypeName(index_type_name);
    }
    return NULL;
}
static PyObject *__Pyx_PyObject_GetItem_Slow(PyObject *obj, PyObject *key) {
    __Pyx_TypeName obj_type_name;
    if (likely(PyType_Check(obj))) {
        PyObject *meth = __Pyx_PyObject_GetAttrStrNoError(obj, __pyx_n_s_class_getitem);
        if (!meth) {
            PyErr_Clear();
        } else {
            PyObject *result = __Pyx_PyObject_CallOneArg(meth, key);
            Py_DECREF(meth);
            return result;
        }
    }
    obj_type_name = __Pyx_PyType_GetName(Py_TYPE(obj));
    PyErr_Format(PyExc_TypeError,
        "'" __Pyx_FMT_TYPENAME "' object is not subscriptable", obj_type_name);
    __Pyx_DECREF_TypeName(obj_type_name);
    return NULL;
}
static PyObject *__Pyx_PyObject_GetItem(PyObject *obj, PyObject *key) {
    PyTypeObject *tp = Py_TYPE(obj);
    PyMappingMethods *mm = tp->tp_as_mapping;
    PySequenceMethods *sm = tp->tp_as_sequence;
    if (likely(mm && mm->mp_subscript)) {
        return mm->mp_subscript(obj, key);
    }
    if (likely(sm && sm->sq_item)) {
        return __Pyx_PyObject_GetIndex(obj, key);
    }
    return __Pyx_PyObject_GetItem_Slow(obj, key);
}
#endif

/* pyfrozenset_new */
static CYTHON_INLINE PyObject* __Pyx_PyFrozenSet_New(PyObject* it) {
    if (it) {
        PyObject* result;
#if CYTHON_COMPILING_IN_PYPY
        PyObject* args;
        args = PyTuple_Pack(1, it);
        if (unlikely(!args))
            return NULL;
        result = PyObject_Call((PyObject*)&PyFrozenSet_Type, args, NULL);
        Py_DECREF(args);
        return result;
#else
        if (PyFrozenSet_CheckExact(it)) {
            Py_INCREF(it);
            return it;
        }
        result = PyFrozenSet_New(it);
        if (unlikely(!result))
            return NULL;
        if ((PY_VERSION_HEX >= 0x031000A1) || likely(PySet_GET_SIZE(result)))
            return result;
        Py_DECREF(result);
#endif
    }
#if CYTHON_USE_TYPE_SLOTS
    return PyFrozenSet_Type.tp_new(&PyFrozenSet_Type, __pyx_empty_tuple, NULL);
#else
    return PyObject_Call((PyObject*)&PyFrozenSet_Type, __pyx_empty_tuple, NULL);
#endif
}

/* PyObjectFormatAndDecref */
static CYTHON_INLINE PyObject* __Pyx_PyObject_FormatSimpleAndDecref(PyObject* s, PyObject* f) {
    if (unlikely(!s)) return NULL;
    if (likely(PyUnicode_CheckExact(s))) return s;
    #if PY_MAJOR_VERSION < 3
    if (likely(PyString_CheckExact(s))) {
        PyObject *result = PyUnicode_FromEncodedObject(s, NULL, "strict");
        Py_DECREF(s);
        return result;
    }
    #endif
    return __Pyx_PyObject_FormatAndDecref(s, f);
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_FormatAndDecref(PyObject* s, PyObject* f) {
    PyObject *result;
    if (unlikely(!s)) return NULL;
    result = PyObject_Format(s, f);
    Py_DECREF(s);
    return result;
}

/* JoinPyUnicode */
static PyObject* __Pyx_PyUnicode_Join(PyObject* value_tuple, Py_ssize_t value_count, Py_ssize_t result_ulength,
                                      Py_UCS4 max_char) {
#if CYTHON_USE_UNICODE_INTERNALS && CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    PyObject *result_uval;
    int result_ukind, kind_shift;
    Py_ssize_t i, char_pos;
    void *result_udata;
    CYTHON_MAYBE_UNUSED_VAR(max_char);
#if CYTHON_PEP393_ENABLED
    result_uval = PyUnicode_New(result_ulength, max_char);
    if (unlikely(!result_uval)) return NULL;
    result_ukind = (max_char <= 255) ? PyUnicode_1BYTE_KIND : (max_char <= 65535) ? PyUnicode_2BYTE_KIND : PyUnicode_4BYTE_KIND;
    kind_shift = (result_ukind == PyUnicode_4BYTE_KIND) ? 2 : result_ukind - 1;
    result_udata = PyUnicode_DATA(result_uval);
#else
    result_uval = PyUnicode_FromUnicode(NULL, result_ulength);
    if (unlikely(!result_uval)) return NULL;
    result_ukind = sizeof(Py_UNICODE);
    kind_shift = (result_ukind == 4) ? 2 : result_ukind - 1;
    result_udata = PyUnicode_AS_UNICODE(result_uval);
#endif
    assert(kind_shift == 2 || kind_shift == 1 || kind_shift == 0);
    char_pos = 0;
    for (i=0; i < value_count; i++) {
        int ukind;
        Py_ssize_t ulength;
        void *udata;
        PyObject *uval = PyTuple_GET_ITEM(value_tuple, i);
        if (unlikely(__Pyx_PyUnicode_READY(uval)))
            goto bad;
        ulength = __Pyx_PyUnicode_GET_LENGTH(uval);
        if (unlikely(!ulength))
            continue;
        if (unlikely((PY_SSIZE_T_MAX >> kind_shift) - ulength < char_pos))
            goto overflow;
        ukind = __Pyx_PyUnicode_KIND(uval);
        udata = __Pyx_PyUnicode_DATA(uval);
        if (!CYTHON_PEP393_ENABLED || ukind == result_ukind) {
            memcpy((char *)result_udata + (char_pos << kind_shift), udata, (size_t) (ulength << kind_shift));
        } else {
            #if PY_VERSION_HEX >= 0x030d0000
            if (unlikely(PyUnicode_CopyCharacters(result_uval, char_pos, uval, 0, ulength) < 0)) goto bad;
            #elif CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030300F0 || defined(_PyUnicode_FastCopyCharacters)
            _PyUnicode_FastCopyCharacters(result_uval, char_pos, uval, 0, ulength);
            #else
            Py_ssize_t j;
            for (j=0; j < ulength; j++) {
                Py_UCS4 uchar = __Pyx_PyUnicode_READ(ukind, udata, j);
                __Pyx_PyUnicode_WRITE(result_ukind, result_udata, char_pos+j, uchar);
            }
            #endif
        }
        char_pos += ulength;
    }
    return result_uval;
overflow:
    PyErr_SetString(PyExc_OverflowError, "join() result is too long for a Python string");
bad:
    Py_DECREF(result_uval);
    return NULL;
#else
    CYTHON_UNUSED_VAR(max_char);
    CYTHON_UNUSED_VAR(result_ulength);
    CYTHON_UNUSED_VAR(value_count);
    return PyUnicode_Join(__pyx_empty_unicode, value_tuple);
#endif
}

/* TupleAndListFromArray */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE void __Pyx_copy_object_array(PyObject *const *CYTHON_RESTRICT src, PyObject** CYTHON_RESTRICT dest, Py_ssize_t length) {
    PyObject *v;
    Py_ssize_t i;
    for (i = 0; i < length; i++) {
        v = dest[i] = src[i];
        Py_INCREF(v);
    }
}
static CYTHON_INLINE PyObject *
__Pyx_PyTuple_FromArray(PyObject *const *src, Py_ssize_t n)
{
    PyObject *res;
    if (n <= 0) {
        Py_INCREF(__pyx_empty_tuple);
        return __pyx_empty_tuple;
    }
    res = PyTuple_New(n);
    if (unlikely(res == NULL)) return NULL;
    __Pyx_copy_object_array(src, ((PyTupleObject*)res)->ob_item, n);
    return res;
}
static CYTHON_INLINE PyObject *
__Pyx_PyList_FromArray(PyObject *const *src, Py_ssize_t n)
{
    PyObject *res;
    if (n <= 0) {
        return PyList_New(0);
    }
    res = PyList_New(n);
    if (unlikely(res == NULL)) return NULL;
    __Pyx_copy_object_array(src, ((PyListObject*)res)->ob_item, n);
    return res;
}
#endif

/* BytesEquals */
static CYTHON_INLINE int __Pyx_PyBytes_Equals(PyObject* s1, PyObject* s2, int equals) {
#if CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API
    return PyObject_RichCompareBool(s1, s2, equals);
#else
    if (s1 == s2) {
        return (equals == Py_EQ);
    } else if (PyBytes_CheckExact(s1) & PyBytes_CheckExact(s2)) {
        const char *ps1, *ps2;
        Py_ssize_t length = PyBytes_GET_SIZE(s1);
        if (length != PyBytes_GET_SIZE(s2))
            return (equals == Py_NE);
        ps1 = PyBytes_AS_STRING(s1);
        ps2 = PyBytes_AS_STRING(s2);
        if (ps1[0] != ps2[0]) {
            return (equals == Py_NE);
        } else if (length == 1) {
            return (equals == Py_EQ);
        } else {
            int result;
#if CYTHON_USE_UNICODE_INTERNALS && (PY_VERSION_HEX < 0x030B0000)
            Py_hash_t hash1, hash2;
            hash1 = ((PyBytesObject*)s1)->ob_shash;
            hash2 = ((PyBytesObject*)s2)->ob_shash;
            if (hash1 != hash2 && hash1 != -1 && hash2 != -1) {
                return (equals == Py_NE);
            }
#endif
            result = memcmp(ps1, ps2, (size_t)length);
            return (equals == Py_EQ) ? (result == 0) : (result != 0);
        }
    } else if ((s1 == Py_None) & PyBytes_CheckExact(s2)) {
        return (equals == Py_NE);
    } else if ((s2 == Py_None) & PyBytes_CheckExact(s1)) {
        return (equals == Py_NE);
    } else {
        int result;
        PyObject* py_result = PyObject_RichCompare(s1, s2, equals);
        if (!py_result)
            return -1;
        result = __Pyx_PyObject_IsTrue(py_result);
        Py_DECREF(py_result);
        return result;
    }
#endif
}

/* UnicodeEquals */
static CYTHON_INLINE int __Pyx_PyUnicode_Equals(PyObject* s1, PyObject* s2, int equals) {
#if CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API
    return PyObject_RichCompareBool(s1, s2, equals);
#else
#if PY_MAJOR_VERSION < 3
    PyObject* owned_ref = NULL;
#endif
    int s1_is_unicode, s2_is_unicode;
    if (s1 == s2) {
        goto return_eq;
    }
    s1_is_unicode = PyUnicode_CheckExact(s1);
    s2_is_unicode = PyUnicode_CheckExact(s2);
#if PY_MAJOR_VERSION < 3
    if ((s1_is_unicode & (!s2_is_unicode)) && PyString_CheckExact(s2)) {
        owned_ref = PyUnicode_FromObject(s2);
        if (unlikely(!owned_ref))
            return -1;
        s2 = owned_ref;
        s2_is_unicode = 1;
    } else if ((s2_is_unicode & (!s1_is_unicode)) && PyString_CheckExact(s1)) {
        owned_ref = PyUnicode_FromObject(s1);
        if (unlikely(!owned_ref))
            return -1;
        s1 = owned_ref;
        s1_is_unicode = 1;
    } else if (((!s2_is_unicode) & (!s1_is_unicode))) {
        return __Pyx_PyBytes_Equals(s1, s2, equals);
    }
#endif
    if (s1_is_unicode & s2_is_unicode) {
        Py_ssize_t length;
        int kind;
        void *data1, *data2;
        if (unlikely(__Pyx_PyUnicode_READY(s1) < 0) || unlikely(__Pyx_PyUnicode_READY(s2) < 0))
            return -1;
        length = __Pyx_PyUnicode_GET_LENGTH(s1);
        if (length != __Pyx_PyUnicode_GET_LENGTH(s2)) {
            goto return_ne;
        }
#if CYTHON_USE_UNICODE_INTERNALS
        {
            Py_hash_t hash1, hash2;
        #if CYTHON_PEP393_ENABLED
            hash1 = ((PyASCIIObject*)s1)->hash;
            hash2 = ((PyASCIIObject*)s2)->hash;
        #else
            hash1 = ((PyUnicodeObject*)s1)->hash;
            hash2 = ((PyUnicodeObject*)s2)->hash;
        #endif
            if (hash1 != hash2 && hash1 != -1 && hash2 != -1) {
                goto return_ne;
            }
        }
#endif
        kind = __Pyx_PyUnicode_KIND(s1);
        if (kind != __Pyx_PyUnicode_KIND(s2)) {
            goto return_ne;
        }
        data1 = __Pyx_PyUnicode_DATA(s1);
        data2 = __Pyx_PyUnicode_DATA(s2);
        if (__Pyx_PyUnicode_READ(kind, data1, 0) != __Pyx_PyUnicode_READ(kind, data2, 0)) {
            goto return_ne;
        } else if (length == 1) {
            goto return_eq;
        } else {
            int result = memcmp(data1, data2, (size_t)(length * kind));
            #if PY_MAJOR_VERSION < 3
            Py_XDECREF(owned_ref);
            #endif
            return (equals == Py_EQ) ? (result == 0) : (result != 0);
        }
    } else if ((s1 == Py_None) & s2_is_unicode) {
        goto return_ne;
    } else if ((s2 == Py_None) & s1_is_unicode) {
        goto return_ne;
    } else {
        int result;
        PyObject* py_result = PyObject_RichCompare(s1, s2, equals);
        #if PY_MAJOR_VERSION < 3
        Py_XDECREF(owned_ref);
        #endif
        if (!py_result)
            return -1;
        result = __Pyx_PyObject_IsTrue(py_result);
        Py_DECREF(py_result);
        return result;
    }
return_eq:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(owned_ref);
    #endif
    return (equals == Py_EQ);
return_ne:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(owned_ref);
    #endif
    return (equals == Py_NE);
#endif
}

/* fastcall */
#if CYTHON_METH_FASTCALL
static CYTHON_INLINE PyObject * __Pyx_GetKwValue_FASTCALL(PyObject *kwnames, PyObject *const *kwvalues, PyObject *s)
{
    Py_ssize_t i, n = PyTuple_GET_SIZE(kwnames);
    for (i = 0; i < n; i++)
    {
        if (s == PyTuple_GET_ITEM(kwnames, i)) return kwvalues[i];
    }
    for (i = 0; i < n; i++)
    {
        int eq = __Pyx_PyUnicode_Equals(s, PyTuple_GET_ITEM(kwnames, i), Py_EQ);
        if (unlikely(eq != 0)) {
            if (unlikely(eq < 0)) return NULL;
            return kwvalues[i];
        }
    }
    return NULL;
}
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030d0000
CYTHON_UNUSED static PyObject *__Pyx_KwargsAsDict_FASTCALL(PyObject *kwnames, PyObject *const *kwvalues) {
    Py_ssize_t i, nkwargs = PyTuple_GET_SIZE(kwnames);
    PyObject *dict;
    dict = PyDict_New();
    if (unlikely(!dict))
        return NULL;
    for (i=0; i<nkwargs; i++) {
        PyObject *key = PyTuple_GET_ITEM(kwnames, i);
        if (unlikely(PyDict_SetItem(dict, key, kwvalues[i]) < 0))
            goto bad;
    }
    return dict;
bad:
    Py_DECREF(dict);
    return NULL;
}
#endif
#endif

/* RaiseDoubleKeywords */
static void __Pyx_RaiseDoubleKeywordsError(
    const char* func_name,
    PyObject* kw_name)
{
    PyErr_Format(PyExc_TypeError,
        #if PY_MAJOR_VERSION >= 3
        "%s() got multiple values for keyword argument '%U'", func_name, kw_name);
        #else
        "%s() got multiple values for keyword argument '%s'", func_name,
        PyString_AsString(kw_name));
        #endif
}

/* ParseKeywords */
static int __Pyx_ParseOptionalKeywords(
    PyObject *kwds,
    PyObject *const *kwvalues,
    PyObject **argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    const char* function_name)
{
    PyObject *key = 0, *value = 0;
    Py_ssize_t pos = 0;
    PyObject*** name;
    PyObject*** first_kw_arg = argnames + num_pos_args;
    int kwds_is_tuple = CYTHON_METH_FASTCALL && likely(PyTuple_Check(kwds));
    while (1) {
        Py_XDECREF(key); key = NULL;
        Py_XDECREF(value); value = NULL;
        if (kwds_is_tuple) {
            Py_ssize_t size;
#if CYTHON_ASSUME_SAFE_MACROS
            size = PyTuple_GET_SIZE(kwds);
#else
            size = PyTuple_Size(kwds);
            if (size < 0) goto bad;
#endif
            if (pos >= size) break;
#if CYTHON_AVOID_BORROWED_REFS
            key = __Pyx_PySequence_ITEM(kwds, pos);
            if (!key) goto bad;
#elif CYTHON_ASSUME_SAFE_MACROS
            key = PyTuple_GET_ITEM(kwds, pos);
#else
            key = PyTuple_GetItem(kwds, pos);
            if (!key) goto bad;
#endif
            value = kwvalues[pos];
            pos++;
        }
        else
        {
            if (!PyDict_Next(kwds, &pos, &key, &value)) break;
#if CYTHON_AVOID_BORROWED_REFS
            Py_INCREF(key);
#endif
        }
        name = first_kw_arg;
        while (*name && (**name != key)) name++;
        if (*name) {
            values[name-argnames] = value;
#if CYTHON_AVOID_BORROWED_REFS
            Py_INCREF(value);
            Py_DECREF(key);
#endif
            key = NULL;
            value = NULL;
            continue;
        }
#if !CYTHON_AVOID_BORROWED_REFS
        Py_INCREF(key);
#endif
        Py_INCREF(value);
        name = first_kw_arg;
        #if PY_MAJOR_VERSION < 3
        if (likely(PyString_Check(key))) {
            while (*name) {
                if ((CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**name) == PyString_GET_SIZE(key))
                        && _PyString_Eq(**name, key)) {
                    values[name-argnames] = value;
#if CYTHON_AVOID_BORROWED_REFS
                    value = NULL;
#endif
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    if ((**argname == key) || (
                            (CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**argname) == PyString_GET_SIZE(key))
                             && _PyString_Eq(**argname, key))) {
                        goto arg_passed_twice;
                    }
                    argname++;
                }
            }
        } else
        #endif
        if (likely(PyUnicode_Check(key))) {
            while (*name) {
                int cmp = (
                #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                    (__Pyx_PyUnicode_GET_LENGTH(**name) != __Pyx_PyUnicode_GET_LENGTH(key)) ? 1 :
                #endif
                    PyUnicode_Compare(**name, key)
                );
                if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                if (cmp == 0) {
                    values[name-argnames] = value;
#if CYTHON_AVOID_BORROWED_REFS
                    value = NULL;
#endif
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    int cmp = (**argname == key) ? 0 :
                    #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                        (__Pyx_PyUnicode_GET_LENGTH(**argname) != __Pyx_PyUnicode_GET_LENGTH(key)) ? 1 :
                    #endif
                        PyUnicode_Compare(**argname, key);
                    if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                    if (cmp == 0) goto arg_passed_twice;
                    argname++;
                }
            }
        } else
            goto invalid_keyword_type;
        if (kwds2) {
            if (unlikely(PyDict_SetItem(kwds2, key, value))) goto bad;
        } else {
            goto invalid_keyword;
        }
    }
    Py_XDECREF(key);
    Py_XDECREF(value);
    return 0;
arg_passed_twice:
    __Pyx_RaiseDoubleKeywordsError(function_name, key);
    goto bad;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%.200s() keywords must be strings", function_name);
    goto bad;
invalid_keyword:
    #if PY_MAJOR_VERSION < 3
    PyErr_Format(PyExc_TypeError,
        "%.200s() got an unexpected keyword argument '%.200s'",
        function_name, PyString_AsString(key));
    #else
    PyErr_Format(PyExc_TypeError,
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    #endif
bad:
    Py_XDECREF(key);
    Py_XDECREF(value);
    return -1;
}

/* RaiseArgTupleInvalid */
static void __Pyx_RaiseArgtupleInvalid(
    const char* func_name,
    int exact,
    Py_ssize_t num_min,
    Py_ssize_t num_max,
    Py_ssize_t num_found)
{
    Py_ssize_t num_expected;
    const char *more_or_less;
    if (num_found < num_min) {
        num_expected = num_min;
        more_or_less = "at least";
    } else {
        num_expected = num_max;
        more_or_less = "at most";
    }
    if (exact) {
        more_or_less = "exactly";
    }
    PyErr_Format(PyExc_TypeError,
                 "%.200s() takes %.8s %" CYTHON_FORMAT_SSIZE_T "d positional argument%.1s (%" CYTHON_FORMAT_SSIZE_T "d given)",
                 func_name, more_or_less, num_expected,
                 (num_expected == 1) ? "" : "s", num_found);
}

/* Import */
static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level) {
    PyObject *module = 0;
    PyObject *empty_dict = 0;
    PyObject *empty_list = 0;
    #if PY_MAJOR_VERSION < 3
    PyObject *py_import;
    py_import = __Pyx_PyObject_GetAttrStr(__pyx_b, __pyx_n_s_import);
    if (unlikely(!py_import))
        goto bad;
    if (!from_list) {
        empty_list = PyList_New(0);
        if (unlikely(!empty_list))
            goto bad;
        from_list = empty_list;
    }
    #endif
    empty_dict = PyDict_New();
    if (unlikely(!empty_dict))
        goto bad;
    {
        #if PY_MAJOR_VERSION >= 3
        if (level == -1) {
            if (strchr(__Pyx_MODULE_NAME, '.') != NULL) {
                module = PyImport_ImportModuleLevelObject(
                    name, __pyx_d, empty_dict, from_list, 1);
                if (unlikely(!module)) {
                    if (unlikely(!PyErr_ExceptionMatches(PyExc_ImportError)))
                        goto bad;
                    PyErr_Clear();
                }
            }
            level = 0;
        }
        #endif
        if (!module) {
            #if PY_MAJOR_VERSION < 3
            PyObject *py_level = PyInt_FromLong(level);
            if (unlikely(!py_level))
                goto bad;
            module = PyObject_CallFunctionObjArgs(py_import,
                name, __pyx_d, empty_dict, from_list, py_level, (PyObject *)NULL);
            Py_DECREF(py_level);
            #else
            module = PyImport_ImportModuleLevelObject(
                name, __pyx_d, empty_dict, from_list, level);
            #endif
        }
    }
bad:
    Py_XDECREF(empty_dict);
    Py_XDECREF(empty_list);
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(py_import);
    #endif
    return module;
}

/* ImportDottedModule */
#if PY_MAJOR_VERSION >= 3
static PyObject *__Pyx__ImportDottedModule_Error(PyObject *name, PyObject *parts_tuple, Py_ssize_t count) {
    PyObject *partial_name = NULL, *slice = NULL, *sep = NULL;
    if (unlikely(PyErr_Occurred())) {
        PyErr_Clear();
    }
    if (likely(PyTuple_GET_SIZE(parts_tuple) == count)) {
        partial_name = name;
    } else {
        slice = PySequence_GetSlice(parts_tuple, 0, count);
        if (unlikely(!slice))
            goto bad;
        sep = PyUnicode_FromStringAndSize(".", 1);
        if (unlikely(!sep))
            goto bad;
        partial_name = PyUnicode_Join(sep, slice);
    }
    PyErr_Format(
#if PY_MAJOR_VERSION < 3
        PyExc_ImportError,
        "No module named '%s'", PyString_AS_STRING(partial_name));
#else
#if PY_VERSION_HEX >= 0x030600B1
        PyExc_ModuleNotFoundError,
#else
        PyExc_ImportError,
#endif
        "No module named '%U'", partial_name);
#endif
bad:
    Py_XDECREF(sep);
    Py_XDECREF(slice);
    Py_XDECREF(partial_name);
    return NULL;
}
#endif
#if PY_MAJOR_VERSION >= 3
static PyObject *__Pyx__ImportDottedModule_Lookup(PyObject *name) {
    PyObject *imported_module;
#if PY_VERSION_HEX < 0x030700A1 || (CYTHON_COMPILING_IN_PYPY && PYPY_VERSION_NUM  < 0x07030400)
    PyObject *modules = PyImport_GetModuleDict();
    if (unlikely(!modules))
        return NULL;
    imported_module = __Pyx_PyDict_GetItemStr(modules, name);
    Py_XINCREF(imported_module);
#else
    imported_module = PyImport_GetModule(name);
#endif
    return imported_module;
}
#endif
#if PY_MAJOR_VERSION >= 3
static PyObject *__Pyx_ImportDottedModule_WalkParts(PyObject *module, PyObject *name, PyObject *parts_tuple) {
    Py_ssize_t i, nparts;
    nparts = PyTuple_GET_SIZE(parts_tuple);
    for (i=1; i < nparts && module; i++) {
        PyObject *part, *submodule;
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        part = PyTuple_GET_ITEM(parts_tuple, i);
#else
        part = PySequence_ITEM(parts_tuple, i);
#endif
        submodule = __Pyx_PyObject_GetAttrStrNoError(module, part);
#if !(CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS)
        Py_DECREF(part);
#endif
        Py_DECREF(module);
        module = submodule;
    }
    if (unlikely(!module)) {
        return __Pyx__ImportDottedModule_Error(name, parts_tuple, i);
    }
    return module;
}
#endif
static PyObject *__Pyx__ImportDottedModule(PyObject *name, PyObject *parts_tuple) {
#if PY_MAJOR_VERSION < 3
    PyObject *module, *from_list, *star = __pyx_n_s__14;
    CYTHON_UNUSED_VAR(parts_tuple);
    from_list = PyList_New(1);
    if (unlikely(!from_list))
        return NULL;
    Py_INCREF(star);
    PyList_SET_ITEM(from_list, 0, star);
    module = __Pyx_Import(name, from_list, 0);
    Py_DECREF(from_list);
    return module;
#else
    PyObject *imported_module;
    PyObject *module = __Pyx_Import(name, NULL, 0);
    if (!parts_tuple || unlikely(!module))
        return module;
    imported_module = __Pyx__ImportDottedModule_Lookup(name);
    if (likely(imported_module)) {
        Py_DECREF(module);
        return imported_module;
    }
    PyErr_Clear();
    return __Pyx_ImportDottedModule_WalkParts(module, name, parts_tuple);
#endif
}
static PyObject *__Pyx_ImportDottedModule(PyObject *name, PyObject *parts_tuple) {
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030400B1
    PyObject *module = __Pyx__ImportDottedModule_Lookup(name);
    if (likely(module)) {
        PyObject *spec = __Pyx_PyObject_GetAttrStrNoError(module, __pyx_n_s_spec);
        if (likely(spec)) {
            PyObject *unsafe = __Pyx_PyObject_GetAttrStrNoError(spec, __pyx_n_s_initializing);
            if (likely(!unsafe || !__Pyx_PyObject_IsTrue(unsafe))) {
                Py_DECREF(spec);
                spec = NULL;
            }
            Py_XDECREF(unsafe);
        }
        if (likely(!spec)) {
            PyErr_Clear();
            return module;
        }
        Py_DECREF(spec);
        Py_DECREF(module);
    } else if (PyErr_Occurred()) {
        PyErr_Clear();
    }
#endif
    return __Pyx__ImportDottedModule(name, parts_tuple);
}

/* SliceObject */
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetSlice(PyObject* obj,
        Py_ssize_t cstart, Py_ssize_t cstop,
        PyObject** _py_start, PyObject** _py_stop, PyObject** _py_slice,
        int has_cstart, int has_cstop, int wraparound) {
    __Pyx_TypeName obj_type_name;
#if CYTHON_USE_TYPE_SLOTS
    PyMappingMethods* mp;
#if PY_MAJOR_VERSION < 3
    PySequenceMethods* ms = Py_TYPE(obj)->tp_as_sequence;
    if (likely(ms && ms->sq_slice)) {
        if (!has_cstart) {
            if (_py_start && (*_py_start != Py_None)) {
                cstart = __Pyx_PyIndex_AsSsize_t(*_py_start);
                if ((cstart == (Py_ssize_t)-1) && PyErr_Occurred()) goto bad;
            } else
                cstart = 0;
        }
        if (!has_cstop) {
            if (_py_stop && (*_py_stop != Py_None)) {
                cstop = __Pyx_PyIndex_AsSsize_t(*_py_stop);
                if ((cstop == (Py_ssize_t)-1) && PyErr_Occurred()) goto bad;
            } else
                cstop = PY_SSIZE_T_MAX;
        }
        if (wraparound && unlikely((cstart < 0) | (cstop < 0)) && likely(ms->sq_length)) {
            Py_ssize_t l = ms->sq_length(obj);
            if (likely(l >= 0)) {
                if (cstop < 0) {
                    cstop += l;
                    if (cstop < 0) cstop = 0;
                }
                if (cstart < 0) {
                    cstart += l;
                    if (cstart < 0) cstart = 0;
                }
            } else {
                if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                    goto bad;
                PyErr_Clear();
            }
        }
        return ms->sq_slice(obj, cstart, cstop);
    }
#else
    CYTHON_UNUSED_VAR(wraparound);
#endif
    mp = Py_TYPE(obj)->tp_as_mapping;
    if (likely(mp && mp->mp_subscript))
#else
    CYTHON_UNUSED_VAR(wraparound);
#endif
    {
        PyObject* result;
        PyObject *py_slice, *py_start, *py_stop;
        if (_py_slice) {
            py_slice = *_py_slice;
        } else {
            PyObject* owned_start = NULL;
            PyObject* owned_stop = NULL;
            if (_py_start) {
                py_start = *_py_start;
            } else {
                if (has_cstart) {
                    owned_start = py_start = PyInt_FromSsize_t(cstart);
                    if (unlikely(!py_start)) goto bad;
                } else
                    py_start = Py_None;
            }
            if (_py_stop) {
                py_stop = *_py_stop;
            } else {
                if (has_cstop) {
                    owned_stop = py_stop = PyInt_FromSsize_t(cstop);
                    if (unlikely(!py_stop)) {
                        Py_XDECREF(owned_start);
                        goto bad;
                    }
                } else
                    py_stop = Py_None;
            }
            py_slice = PySlice_New(py_start, py_stop, Py_None);
            Py_XDECREF(owned_start);
            Py_XDECREF(owned_stop);
            if (unlikely(!py_slice)) goto bad;
        }
#if CYTHON_USE_TYPE_SLOTS
        result = mp->mp_subscript(obj, py_slice);
#else
        result = PyObject_GetItem(obj, py_slice);
#endif
        if (!_py_slice) {
            Py_DECREF(py_slice);
        }
        return result;
    }
    obj_type_name = __Pyx_PyType_GetName(Py_TYPE(obj));
    PyErr_Format(PyExc_TypeError,
        "'" __Pyx_FMT_TYPENAME "' object is unsliceable", obj_type_name);
    __Pyx_DECREF_TypeName(obj_type_name);
bad:
    return NULL;
}

/* TypeImport */
#ifndef __PYX_HAVE_RT_ImportType_3_0_12
#define __PYX_HAVE_RT_ImportType_3_0_12
static PyTypeObject *__Pyx_ImportType_3_0_12(PyObject *module, const char *module_name, const char *class_name,
    size_t size, size_t alignment, enum __Pyx_ImportType_CheckSize_3_0_12 check_size)
{
    PyObject *result = 0;
    char warning[200];
    Py_ssize_t basicsize;
    Py_ssize_t itemsize;
#if CYTHON_COMPILING_IN_LIMITED_API
    PyObject *py_basicsize;
    PyObject *py_itemsize;
#endif
    result = PyObject_GetAttrString(module, class_name);
    if (!result)
        goto bad;
    if (!PyType_Check(result)) {
        PyErr_Format(PyExc_TypeError,
            "%.200s.%.200s is not a type object",
            module_name, class_name);
        goto bad;
    }
#if !CYTHON_COMPILING_IN_LIMITED_API
    basicsize = ((PyTypeObject *)result)->tp_basicsize;
    itemsize = ((PyTypeObject *)result)->tp_itemsize;
#else
    py_basicsize = PyObject_GetAttrString(result, "__basicsize__");
    if (!py_basicsize)
        goto bad;
    basicsize = PyLong_AsSsize_t(py_basicsize);
    Py_DECREF(py_basicsize);
    py_basicsize = 0;
    if (basicsize == (Py_ssize_t)-1 && PyErr_Occurred())
        goto bad;
    py_itemsize = PyObject_GetAttrString(result, "__itemsize__");
    if (!py_itemsize)
        goto bad;
    itemsize = PyLong_AsSsize_t(py_itemsize);
    Py_DECREF(py_itemsize);
    py_itemsize = 0;
    if (itemsize == (Py_ssize_t)-1 && PyErr_Occurred())
        goto bad;
#endif
    if (itemsize) {
        if (size % alignment) {
            alignment = size % alignment;
        }
        if (itemsize < (Py_ssize_t)alignment)
            itemsize = (Py_ssize_t)alignment;
    }
    if ((size_t)(basicsize + itemsize) < size) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize+itemsize);
        goto bad;
    }
    if (check_size == __Pyx_ImportType_CheckSize_Error_3_0_12 &&
            ((size_t)basicsize > size || (size_t)(basicsize + itemsize) < size)) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd-%zd from PyObject",
            module_name, class_name, size, basicsize, basicsize+itemsize);
        goto bad;
    }
    else if (check_size == __Pyx_ImportType_CheckSize_Warn_3_0_12 && (size_t)basicsize > size) {
        PyOS_snprintf(warning, sizeof(warning),
            "%s.%s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        if (PyErr_WarnEx(NULL, warning, 0) < 0) goto bad;
    }
    return (PyTypeObject *)result;
bad:
    Py_XDECREF(result);
    return NULL;
}
#endif

/* ImportFrom */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name) {
    PyObject* value = __Pyx_PyObject_GetAttrStr(module, name);
    if (unlikely(!value) && PyErr_ExceptionMatches(PyExc_AttributeError)) {
        const char* module_name_str = 0;
        PyObject* module_name = 0;
        PyObject* module_dot = 0;
        PyObject* full_name = 0;
        PyErr_Clear();
        module_name_str = PyModule_GetName(module);
        if (unlikely(!module_name_str)) { goto modbad; }
        module_name = PyUnicode_FromString(module_name_str);
        if (unlikely(!module_name)) { goto modbad; }
        module_dot = PyUnicode_Concat(module_name, __pyx_kp_u__27);
        if (unlikely(!module_dot)) { goto modbad; }
        full_name = PyUnicode_Concat(module_dot, name);
        if (unlikely(!full_name)) { goto modbad; }
        #if PY_VERSION_HEX < 0x030700A1 || (CYTHON_COMPILING_IN_PYPY && PYPY_VERSION_NUM  < 0x07030400)
        {
            PyObject *modules = PyImport_GetModuleDict();
            if (unlikely(!modules))
                goto modbad;
            value = PyObject_GetItem(modules, full_name);
        }
        #else
        value = PyImport_GetModule(full_name);
        #endif
      modbad:
        Py_XDECREF(full_name);
        Py_XDECREF(module_dot);
        Py_XDECREF(module_name);
    }
    if (unlikely(!value)) {
        PyErr_Format(PyExc_ImportError,
        #if PY_MAJOR_VERSION < 3
            "cannot import name %.230s", PyString_AS_STRING(name));
        #else
            "cannot import name %S", name);
        #endif
    }
    return value;
}

/* FixUpExtensionType */
#if CYTHON_USE_TYPE_SPECS
static int __Pyx_fix_up_extension_type_from_spec(PyType_Spec *spec, PyTypeObject *type) {
#if PY_VERSION_HEX > 0x030900B1 || CYTHON_COMPILING_IN_LIMITED_API
    CYTHON_UNUSED_VAR(spec);
    CYTHON_UNUSED_VAR(type);
#else
    const PyType_Slot *slot = spec->slots;
    while (slot && slot->slot && slot->slot != Py_tp_members)
        slot++;
    if (slot && slot->slot == Py_tp_members) {
        int changed = 0;
#if !(PY_VERSION_HEX <= 0x030900b1 && CYTHON_COMPILING_IN_CPYTHON)
        const
#endif
            PyMemberDef *memb = (PyMemberDef*) slot->pfunc;
        while (memb && memb->name) {
            if (memb->name[0] == '_' && memb->name[1] == '_') {
#if PY_VERSION_HEX < 0x030900b1
                if (strcmp(memb->name, "__weaklistoffset__") == 0) {
                    assert(memb->type == T_PYSSIZET);
                    assert(memb->flags == READONLY);
                    type->tp_weaklistoffset = memb->offset;
                    changed = 1;
                }
                else if (strcmp(memb->name, "__dictoffset__") == 0) {
                    assert(memb->type == T_PYSSIZET);
                    assert(memb->flags == READONLY);
                    type->tp_dictoffset = memb->offset;
                    changed = 1;
                }
#if CYTHON_METH_FASTCALL
                else if (strcmp(memb->name, "__vectorcalloffset__") == 0) {
                    assert(memb->type == T_PYSSIZET);
                    assert(memb->flags == READONLY);
#if PY_VERSION_HEX >= 0x030800b4
                    type->tp_vectorcall_offset = memb->offset;
#else
                    type->tp_print = (printfunc) memb->offset;
#endif
                    changed = 1;
                }
#endif
#else
                if ((0));
#endif
#if PY_VERSION_HEX <= 0x030900b1 && CYTHON_COMPILING_IN_CPYTHON
                else if (strcmp(memb->name, "__module__") == 0) {
                    PyObject *descr;
                    assert(memb->type == T_OBJECT);
                    assert(memb->flags == 0 || memb->flags == READONLY);
                    descr = PyDescr_NewMember(type, memb);
                    if (unlikely(!descr))
                        return -1;
                    if (unlikely(PyDict_SetItem(type->tp_dict, PyDescr_NAME(descr), descr) < 0)) {
                        Py_DECREF(descr);
                        return -1;
                    }
                    Py_DECREF(descr);
                    changed = 1;
                }
#endif
            }
            memb++;
        }
        if (changed)
            PyType_Modified(type);
    }
#endif
    return 0;
}
#endif

/* FetchSharedCythonModule */
static PyObject *__Pyx_FetchSharedCythonABIModule(void) {
    return __Pyx_PyImport_AddModuleRef((char*) __PYX_ABI_MODULE_NAME);
}

/* FetchCommonType */
static int __Pyx_VerifyCachedType(PyObject *cached_type,
                               const char *name,
                               Py_ssize_t basicsize,
                               Py_ssize_t expected_basicsize) {
    if (!PyType_Check(cached_type)) {
        PyErr_Format(PyExc_TypeError,
            "Shared Cython type %.200s is not a type object", name);
        return -1;
    }
    if (basicsize != expected_basicsize) {
        PyErr_Format(PyExc_TypeError,
            "Shared Cython type %.200s has the wrong size, try recompiling",
            name);
        return -1;
    }
    return 0;
}
#if !CYTHON_USE_TYPE_SPECS
static PyTypeObject* __Pyx_FetchCommonType(PyTypeObject* type) {
    PyObject* abi_module;
    const char* object_name;
    PyTypeObject *cached_type = NULL;
    abi_module = __Pyx_FetchSharedCythonABIModule();
    if (!abi_module) return NULL;
    object_name = strrchr(type->tp_name, '.');
    object_name = object_name ? object_name+1 : type->tp_name;
    cached_type = (PyTypeObject*) PyObject_GetAttrString(abi_module, object_name);
    if (cached_type) {
        if (__Pyx_VerifyCachedType(
              (PyObject *)cached_type,
              object_name,
              cached_type->tp_basicsize,
              type->tp_basicsize) < 0) {
            goto bad;
        }
        goto done;
    }
    if (!PyErr_ExceptionMatches(PyExc_AttributeError)) goto bad;
    PyErr_Clear();
    if (PyType_Ready(type) < 0) goto bad;
    if (PyObject_SetAttrString(abi_module, object_name, (PyObject *)type) < 0)
        goto bad;
    Py_INCREF(type);
    cached_type = type;
done:
    Py_DECREF(abi_module);
    return cached_type;
bad:
    Py_XDECREF(cached_type);
    cached_type = NULL;
    goto done;
}
#else
static PyTypeObject *__Pyx_FetchCommonTypeFromSpec(PyObject *module, PyType_Spec *spec, PyObject *bases) {
    PyObject *abi_module, *cached_type = NULL;
    const char* object_name = strrchr(spec->name, '.');
    object_name = object_name ? object_name+1 : spec->name;
    abi_module = __Pyx_FetchSharedCythonABIModule();
    if (!abi_module) return NULL;
    cached_type = PyObject_GetAttrString(abi_module, object_name);
    if (cached_type) {
        Py_ssize_t basicsize;
#if CYTHON_COMPILING_IN_LIMITED_API
        PyObject *py_basicsize;
        py_basicsize = PyObject_GetAttrString(cached_type, "__basicsize__");
        if (unlikely(!py_basicsize)) goto bad;
        basicsize = PyLong_AsSsize_t(py_basicsize);
        Py_DECREF(py_basicsize);
        py_basicsize = 0;
        if (unlikely(basicsize == (Py_ssize_t)-1) && PyErr_Occurred()) goto bad;
#else
        basicsize = likely(PyType_Check(cached_type)) ? ((PyTypeObject*) cached_type)->tp_basicsize : -1;
#endif
        if (__Pyx_VerifyCachedType(
              cached_type,
              object_name,
              basicsize,
              spec->basicsize) < 0) {
            goto bad;
        }
        goto done;
    }
    if (!PyErr_ExceptionMatches(PyExc_AttributeError)) goto bad;
    PyErr_Clear();
    CYTHON_UNUSED_VAR(module);
    cached_type = __Pyx_PyType_FromModuleAndSpec(abi_module, spec, bases);
    if (unlikely(!cached_type)) goto bad;
    if (unlikely(__Pyx_fix_up_extension_type_from_spec(spec, (PyTypeObject *) cached_type) < 0)) goto bad;
    if (PyObject_SetAttrString(abi_module, object_name, cached_type) < 0) goto bad;
done:
    Py_DECREF(abi_module);
    assert(cached_type == NULL || PyType_Check(cached_type));
    return (PyTypeObject *) cached_type;
bad:
    Py_XDECREF(cached_type);
    cached_type = NULL;
    goto done;
}
#endif

/* PyVectorcallFastCallDict */
#if CYTHON_METH_FASTCALL
static PyObject *__Pyx_PyVectorcall_FastCallDict_kw(PyObject *func, __pyx_vectorcallfunc vc, PyObject *const *args, size_t nargs, PyObject *kw)
{
    PyObject *res = NULL;
    PyObject *kwnames;
    PyObject **newargs;
    PyObject **kwvalues;
    Py_ssize_t i, pos;
    size_t j;
    PyObject *key, *value;
    unsigned long keys_are_strings;
    Py_ssize_t nkw = PyDict_GET_SIZE(kw);
    newargs = (PyObject **)PyMem_Malloc((nargs + (size_t)nkw) * sizeof(args[0]));
    if (unlikely(newargs == NULL)) {
        PyErr_NoMemory();
        return NULL;
    }
    for (j = 0; j < nargs; j++) newargs[j] = args[j];
    kwnames = PyTuple_New(nkw);
    if (unlikely(kwnames == NULL)) {
        PyMem_Free(newargs);
        return NULL;
    }
    kwvalues = newargs + nargs;
    pos = i = 0;
    keys_are_strings = Py_TPFLAGS_UNICODE_SUBCLASS;
    while (PyDict_Next(kw, &pos, &key, &value)) {
        keys_are_strings &= Py_TYPE(key)->tp_flags;
        Py_INCREF(key);
        Py_INCREF(value);
        PyTuple_SET_ITEM(kwnames, i, key);
        kwvalues[i] = value;
        i++;
    }
    if (unlikely(!keys_are_strings)) {
        PyErr_SetString(PyExc_TypeError, "keywords must be strings");
        goto cleanup;
    }
    res = vc(func, newargs, nargs, kwnames);
cleanup:
    Py_DECREF(kwnames);
    for (i = 0; i < nkw; i++)
        Py_DECREF(kwvalues[i]);
    PyMem_Free(newargs);
    return res;
}
static CYTHON_INLINE PyObject *__Pyx_PyVectorcall_FastCallDict(PyObject *func, __pyx_vectorcallfunc vc, PyObject *const *args, size_t nargs, PyObject *kw)
{
    if (likely(kw == NULL) || PyDict_GET_SIZE(kw) == 0) {
        return vc(func, args, nargs, NULL);
    }
    return __Pyx_PyVectorcall_FastCallDict_kw(func, vc, args, nargs, kw);
}
#endif

/* CythonFunctionShared */
#if CYTHON_COMPILING_IN_LIMITED_API
static CYTHON_INLINE int __Pyx__IsSameCyOrCFunction(PyObject *func, void *cfunc) {
    if (__Pyx_CyFunction_Check(func)) {
        return PyCFunction_GetFunction(((__pyx_CyFunctionObject*)func)->func) == (PyCFunction) cfunc;
    } else if (PyCFunction_Check(func)) {
        return PyCFunction_GetFunction(func) == (PyCFunction) cfunc;
    }
    return 0;
}
#else
static CYTHON_INLINE int __Pyx__IsSameCyOrCFunction(PyObject *func, void *cfunc) {
    return __Pyx_CyOrPyCFunction_Check(func) && __Pyx_CyOrPyCFunction_GET_FUNCTION(func) == (PyCFunction) cfunc;
}
#endif
static CYTHON_INLINE void __Pyx__CyFunction_SetClassObj(__pyx_CyFunctionObject* f, PyObject* classobj) {
#if PY_VERSION_HEX < 0x030900B1 || CYTHON_COMPILING_IN_LIMITED_API
    __Pyx_Py_XDECREF_SET(
        __Pyx_CyFunction_GetClassObj(f),
            ((classobj) ? __Pyx_NewRef(classobj) : NULL));
#else
    __Pyx_Py_XDECREF_SET(
        ((PyCMethodObject *) (f))->mm_class,
        (PyTypeObject*)((classobj) ? __Pyx_NewRef(classobj) : NULL));
#endif
}
static PyObject *
__Pyx_CyFunction_get_doc(__pyx_CyFunctionObject *op, void *closure)
{
    CYTHON_UNUSED_VAR(closure);
    if (unlikely(op->func_doc == NULL)) {
#if CYTHON_COMPILING_IN_LIMITED_API
        op->func_doc = PyObject_GetAttrString(op->func, "__doc__");
        if (unlikely(!op->func_doc)) return NULL;
#else
        if (((PyCFunctionObject*)op)->m_ml->ml_doc) {
#if PY_MAJOR_VERSION >= 3
            op->func_doc = PyUnicode_FromString(((PyCFunctionObject*)op)->m_ml->ml_doc);
#else
            op->func_doc = PyString_FromString(((PyCFunctionObject*)op)->m_ml->ml_doc);
#endif
            if (unlikely(op->func_doc == NULL))
                return NULL;
        } else {
            Py_INCREF(Py_None);
            return Py_None;
        }
#endif
    }
    Py_INCREF(op->func_doc);
    return op->func_doc;
}
static int
__Pyx_CyFunction_set_doc(__pyx_CyFunctionObject *op, PyObject *value, void *context)
{
    CYTHON_UNUSED_VAR(context);
    if (value == NULL) {
        value = Py_None;
    }
    Py_INCREF(value);
    __Pyx_Py_XDECREF_SET(op->func_doc, value);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_name(__pyx_CyFunctionObject *op, void *context)
{
    CYTHON_UNUSED_VAR(context);
    if (unlikely(op->func_name == NULL)) {
#if CYTHON_COMPILING_IN_LIMITED_API
        op->func_name = PyObject_GetAttrString(op->func, "__name__");
#elif PY_MAJOR_VERSION >= 3
        op->func_name = PyUnicode_InternFromString(((PyCFunctionObject*)op)->m_ml->ml_name);
#else
        op->func_name = PyString_InternFromString(((PyCFunctionObject*)op)->m_ml->ml_name);
#endif
        if (unlikely(op->func_name == NULL))
            return NULL;
    }
    Py_INCREF(op->func_name);
    return op->func_name;
}
static int
__Pyx_CyFunction_set_name(__pyx_CyFunctionObject *op, PyObject *value, void *context)
{
    CYTHON_UNUSED_VAR(context);
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value)))
#else
    if (unlikely(value == NULL || !PyString_Check(value)))
#endif
    {
        PyErr_SetString(PyExc_TypeError,
                        "__name__ must be set to a string object");
        return -1;
    }
    Py_INCREF(value);
    __Pyx_Py_XDECREF_SET(op->func_name, value);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_qualname(__pyx_CyFunctionObject *op, void *context)
{
    CYTHON_UNUSED_VAR(context);
    Py_INCREF(op->func_qualname);
    return op->func_qualname;
}
static int
__Pyx_CyFunction_set_qualname(__pyx_CyFunctionObject *op, PyObject *value, void *context)
{
    CYTHON_UNUSED_VAR(context);
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value)))
#else
    if (unlikely(value == NULL || !PyString_Check(value)))
#endif
    {
        PyErr_SetString(PyExc_TypeError,
                        "__qualname__ must be set to a string object");
        return -1;
    }
    Py_INCREF(value);
    __Pyx_Py_XDECREF_SET(op->func_qualname, value);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_dict(__pyx_CyFunctionObject *op, void *context)
{
    CYTHON_UNUSED_VAR(context);
    if (unlikely(op->func_dict == NULL)) {
        op->func_dict = PyDict_New();
        if (unlikely(op->func_dict == NULL))
            return NULL;
    }
    Py_INCREF(op->func_dict);
    return op->func_dict;
}
static int
__Pyx_CyFunction_set_dict(__pyx_CyFunctionObject *op, PyObject *value, void *context)
{
    CYTHON_UNUSED_VAR(context);
    if (unlikely(value == NULL)) {
        PyErr_SetString(PyExc_TypeError,
               "function's dictionary may not be deleted");
        return -1;
    }
    if (unlikely(!PyDict_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
               "setting function's dictionary to a non-dict");
        return -1;
    }
    Py_INCREF(value);
    __Pyx_Py_XDECREF_SET(op->func_dict, value);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_globals(__pyx_CyFunctionObject *op, void *context)
{
    CYTHON_UNUSED_VAR(context);
    Py_INCREF(op->func_globals);
    return op->func_globals;
}
static PyObject *
__Pyx_CyFunction_get_closure(__pyx_CyFunctionObject *op, void *context)
{
    CYTHON_UNUSED_VAR(op);
    CYTHON_UNUSED_VAR(context);
    Py_INCREF(Py_None);
    return Py_None;
}
static PyObject *
__Pyx_CyFunction_get_code(__pyx_CyFunctionObject *op, void *context)
{
    PyObject* result = (op->func_code) ? op->func_code : Py_None;
    CYTHON_UNUSED_VAR(context);
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_init_defaults(__pyx_CyFunctionObject *op) {
    int result = 0;
    PyObject *res = op->defaults_getter((PyObject *) op);
    if (unlikely(!res))
        return -1;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    op->defaults_tuple = PyTuple_GET_ITEM(res, 0);
    Py_INCREF(op->defaults_tuple);
    op->defaults_kwdict = PyTuple_GET_ITEM(res, 1);
    Py_INCREF(op->defaults_kwdict);
    #else
    op->defaults_tuple = __Pyx_PySequence_ITEM(res, 0);
    if (unlikely(!op->defaults_tuple)) result = -1;
    else {
        op->defaults_kwdict = __Pyx_PySequence_ITEM(res, 1);
        if (unlikely(!op->defaults_kwdict)) result = -1;
    }
    #endif
    Py_DECREF(res);
    return result;
}
static int
__Pyx_CyFunction_set_defaults(__pyx_CyFunctionObject *op, PyObject* value, void *context) {
    CYTHON_UNUSED_VAR(context);
    if (!value) {
        value = Py_None;
    } else if (unlikely(value != Py_None && !PyTuple_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
                        "__defaults__ must be set to a tuple object");
        return -1;
    }
    PyErr_WarnEx(PyExc_RuntimeWarning, "changes to cyfunction.__defaults__ will not "
                 "currently affect the values used in function calls", 1);
    Py_INCREF(value);
    __Pyx_Py_XDECREF_SET(op->defaults_tuple, value);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_defaults(__pyx_CyFunctionObject *op, void *context) {
    PyObject* result = op->defaults_tuple;
    CYTHON_UNUSED_VAR(context);
    if (unlikely(!result)) {
        if (op->defaults_getter) {
            if (unlikely(__Pyx_CyFunction_init_defaults(op) < 0)) return NULL;
            result = op->defaults_tuple;
        } else {
            result = Py_None;
        }
    }
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_set_kwdefaults(__pyx_CyFunctionObject *op, PyObject* value, void *context) {
    CYTHON_UNUSED_VAR(context);
    if (!value) {
        value = Py_None;
    } else if (unlikely(value != Py_None && !PyDict_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
                        "__kwdefaults__ must be set to a dict object");
        return -1;
    }
    PyErr_WarnEx(PyExc_RuntimeWarning, "changes to cyfunction.__kwdefaults__ will not "
                 "currently affect the values used in function calls", 1);
    Py_INCREF(value);
    __Pyx_Py_XDECREF_SET(op->defaults_kwdict, value);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_kwdefaults(__pyx_CyFunctionObject *op, void *context) {
    PyObject* result = op->defaults_kwdict;
    CYTHON_UNUSED_VAR(context);
    if (unlikely(!result)) {
        if (op->defaults_getter) {
            if (unlikely(__Pyx_CyFunction_init_defaults(op) < 0)) return NULL;
            result = op->defaults_kwdict;
        } else {
            result = Py_None;
        }
    }
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_set_annotations(__pyx_CyFunctionObject *op, PyObject* value, void *context) {
    CYTHON_UNUSED_VAR(context);
    if (!value || value == Py_None) {
        value = NULL;
    } else if (unlikely(!PyDict_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
                        "__annotations__ must be set to a dict object");
        return -1;
    }
    Py_XINCREF(value);
    __Pyx_Py_XDECREF_SET(op->func_annotations, value);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_annotations(__pyx_CyFunctionObject *op, void *context) {
    PyObject* result = op->func_annotations;
    CYTHON_UNUSED_VAR(context);
    if (unlikely(!result)) {
        result = PyDict_New();
        if (unlikely(!result)) return NULL;
        op->func_annotations = result;
    }
    Py_INCREF(result);
    return result;
}
static PyObject *
__Pyx_CyFunction_get_is_coroutine(__pyx_CyFunctionObject *op, void *context) {
    int is_coroutine;
    CYTHON_UNUSED_VAR(context);
    if (op->func_is_coroutine) {
        return __Pyx_NewRef(op->func_is_coroutine);
    }
    is_coroutine = op->flags & __Pyx_CYFUNCTION_COROUTINE;
#if PY_VERSION_HEX >= 0x03050000
    if (is_coroutine) {
        PyObject *module, *fromlist, *marker = __pyx_n_s_is_coroutine;
        fromlist = PyList_New(1);
        if (unlikely(!fromlist)) return NULL;
        Py_INCREF(marker);
#if CYTHON_ASSUME_SAFE_MACROS
        PyList_SET_ITEM(fromlist, 0, marker);
#else
        if (unlikely(PyList_SetItem(fromlist, 0, marker) < 0)) {
            Py_DECREF(marker);
            Py_DECREF(fromlist);
            return NULL;
        }
#endif
        module = PyImport_ImportModuleLevelObject(__pyx_n_s_asyncio_coroutines, NULL, NULL, fromlist, 0);
        Py_DECREF(fromlist);
        if (unlikely(!module)) goto ignore;
        op->func_is_coroutine = __Pyx_PyObject_GetAttrStr(module, marker);
        Py_DECREF(module);
        if (likely(op->func_is_coroutine)) {
            return __Pyx_NewRef(op->func_is_coroutine);
        }
ignore:
        PyErr_Clear();
    }
#endif
    op->func_is_coroutine = __Pyx_PyBool_FromLong(is_coroutine);
    return __Pyx_NewRef(op->func_is_coroutine);
}
#if CYTHON_COMPILING_IN_LIMITED_API
static PyObject *
__Pyx_CyFunction_get_module(__pyx_CyFunctionObject *op, void *context) {
    CYTHON_UNUSED_VAR(context);
    return PyObject_GetAttrString(op->func, "__module__");
}
static int
__Pyx_CyFunction_set_module(__pyx_CyFunctionObject *op, PyObject* value, void *context) {
    CYTHON_UNUSED_VAR(context);
    return PyObject_SetAttrString(op->func, "__module__", value);
}
#endif
static PyGetSetDef __pyx_CyFunction_getsets[] = {
    {(char *) "func_doc", (getter)__Pyx_CyFunction_get_doc, (setter)__Pyx_CyFunction_set_doc, 0, 0},
    {(char *) "__doc__",  (getter)__Pyx_CyFunction_get_doc, (setter)__Pyx_CyFunction_set_doc, 0, 0},
    {(char *) "func_name", (getter)__Pyx_CyFunction_get_name, (setter)__Pyx_CyFunction_set_name, 0, 0},
    {(char *) "__name__", (getter)__Pyx_CyFunction_get_name, (setter)__Pyx_CyFunction_set_name, 0, 0},
    {(char *) "__qualname__", (getter)__Pyx_CyFunction_get_qualname, (setter)__Pyx_CyFunction_set_qualname, 0, 0},
    {(char *) "func_dict", (getter)__Pyx_CyFunction_get_dict, (setter)__Pyx_CyFunction_set_dict, 0, 0},
    {(char *) "__dict__", (getter)__Pyx_CyFunction_get_dict, (setter)__Pyx_CyFunction_set_dict, 0, 0},
    {(char *) "func_globals", (getter)__Pyx_CyFunction_get_globals, 0, 0, 0},
    {(char *) "__globals__", (getter)__Pyx_CyFunction_get_globals, 0, 0, 0},
    {(char *) "func_closure", (getter)__Pyx_CyFunction_get_closure, 0, 0, 0},
    {(char *) "__closure__", (getter)__Pyx_CyFunction_get_closure, 0, 0, 0},
    {(char *) "func_code", (getter)__Pyx_CyFunction_get_code, 0, 0, 0},
    {(char *) "__code__", (getter)__Pyx_CyFunction_get_code, 0, 0, 0},
    {(char *) "func_defaults", (getter)__Pyx_CyFunction_get_defaults, (setter)__Pyx_CyFunction_set_defaults, 0, 0},
    {(char *) "__defaults__", (getter)__Pyx_CyFunction_get_defaults, (setter)__Pyx_CyFunction_set_defaults, 0, 0},
    {(char *) "__kwdefaults__", (getter)__Pyx_CyFunction_get_kwdefaults, (setter)__Pyx_CyFunction_set_kwdefaults, 0, 0},
    {(char *) "__annotations__", (getter)__Pyx_CyFunction_get_annotations, (setter)__Pyx_CyFunction_set_annotations, 0, 0},
    {(char *) "_is_coroutine", (getter)__Pyx_CyFunction_get_is_coroutine, 0, 0, 0},
#if CYTHON_COMPILING_IN_LIMITED_API
    {"__module__", (getter)__Pyx_CyFunction_get_module, (setter)__Pyx_CyFunction_set_module, 0, 0},
#endif
    {0, 0, 0, 0, 0}
};
static PyMemberDef __pyx_CyFunction_members[] = {
#if !CYTHON_COMPILING_IN_LIMITED_API
    {(char *) "__module__", T_OBJECT, offsetof(PyCFunctionObject, m_module), 0, 0},
#endif
#if CYTHON_USE_TYPE_SPECS
    {(char *) "__dictoffset__", T_PYSSIZET, offsetof(__pyx_CyFunctionObject, func_dict), READONLY, 0},
#if CYTHON_METH_FASTCALL
#if CYTHON_BACKPORT_VECTORCALL
    {(char *) "__vectorcalloffset__", T_PYSSIZET, offsetof(__pyx_CyFunctionObject, func_vectorcall), READONLY, 0},
#else
#if !CYTHON_COMPILING_IN_LIMITED_API
    {(char *) "__vectorcalloffset__", T_PYSSIZET, offsetof(PyCFunctionObject, vectorcall), READONLY, 0},
#endif
#endif
#endif
#if PY_VERSION_HEX < 0x030500A0 || CYTHON_COMPILING_IN_LIMITED_API
    {(char *) "__weaklistoffset__", T_PYSSIZET, offsetof(__pyx_CyFunctionObject, func_weakreflist), READONLY, 0},
#else
    {(char *) "__weaklistoffset__", T_PYSSIZET, offsetof(PyCFunctionObject, m_weakreflist), READONLY, 0},
#endif
#endif
    {0, 0, 0,  0, 0}
};
static PyObject *
__Pyx_CyFunction_reduce(__pyx_CyFunctionObject *m, PyObject *args)
{
    CYTHON_UNUSED_VAR(args);
#if PY_MAJOR_VERSION >= 3
    Py_INCREF(m->func_qualname);
    return m->func_qualname;
#else
    return PyString_FromString(((PyCFunctionObject*)m)->m_ml->ml_name);
#endif
}
static PyMethodDef __pyx_CyFunction_methods[] = {
    {"__reduce__", (PyCFunction)__Pyx_CyFunction_reduce, METH_VARARGS, 0},
    {0, 0, 0, 0}
};
#if PY_VERSION_HEX < 0x030500A0 || CYTHON_COMPILING_IN_LIMITED_API
#define __Pyx_CyFunction_weakreflist(cyfunc) ((cyfunc)->func_weakreflist)
#else
#define __Pyx_CyFunction_weakreflist(cyfunc) (((PyCFunctionObject*)cyfunc)->m_weakreflist)
#endif
static PyObject *__Pyx_CyFunction_Init(__pyx_CyFunctionObject *op, PyMethodDef *ml, int flags, PyObject* qualname,
                                       PyObject *closure, PyObject *module, PyObject* globals, PyObject* code) {
#if !CYTHON_COMPILING_IN_LIMITED_API
    PyCFunctionObject *cf = (PyCFunctionObject*) op;
#endif
    if (unlikely(op == NULL))
        return NULL;
#if CYTHON_COMPILING_IN_LIMITED_API
    op->func = PyCFunction_NewEx(ml, (PyObject*)op, module);
    if (unlikely(!op->func)) return NULL;
#endif
    op->flags = flags;
    __Pyx_CyFunction_weakreflist(op) = NULL;
#if !CYTHON_COMPILING_IN_LIMITED_API
    cf->m_ml = ml;
    cf->m_self = (PyObject *) op;
#endif
    Py_XINCREF(closure);
    op->func_closure = closure;
#if !CYTHON_COMPILING_IN_LIMITED_API
    Py_XINCREF(module);
    cf->m_module = module;
#endif
    op->func_dict = NULL;
    op->func_name = NULL;
    Py_INCREF(qualname);
    op->func_qualname = qualname;
    op->func_doc = NULL;
#if PY_VERSION_HEX < 0x030900B1 || CYTHON_COMPILING_IN_LIMITED_API
    op->func_classobj = NULL;
#else
    ((PyCMethodObject*)op)->mm_class = NULL;
#endif
    op->func_globals = globals;
    Py_INCREF(op->func_globals);
    Py_XINCREF(code);
    op->func_code = code;
    op->defaults_pyobjects = 0;
    op->defaults_size = 0;
    op->defaults = NULL;
    op->defaults_tuple = NULL;
    op->defaults_kwdict = NULL;
    op->defaults_getter = NULL;
    op->func_annotations = NULL;
    op->func_is_coroutine = NULL;
#if CYTHON_METH_FASTCALL
    switch (ml->ml_flags & (METH_VARARGS | METH_FASTCALL | METH_NOARGS | METH_O | METH_KEYWORDS | METH_METHOD)) {
    case METH_NOARGS:
        __Pyx_CyFunction_func_vectorcall(op) = __Pyx_CyFunction_Vectorcall_NOARGS;
        break;
    case METH_O:
        __Pyx_CyFunction_func_vectorcall(op) = __Pyx_CyFunction_Vectorcall_O;
        break;
    case METH_METHOD | METH_FASTCALL | METH_KEYWORDS:
        __Pyx_CyFunction_func_vectorcall(op) = __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS_METHOD;
        break;
    case METH_FASTCALL | METH_KEYWORDS:
        __Pyx_CyFunction_func_vectorcall(op) = __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS;
        break;
    case METH_VARARGS | METH_KEYWORDS:
        __Pyx_CyFunction_func_vectorcall(op) = NULL;
        break;
    default:
        PyErr_SetString(PyExc_SystemError, "Bad call flags for CyFunction");
        Py_DECREF(op);
        return NULL;
    }
#endif
    return (PyObject *) op;
}
static int
__Pyx_CyFunction_clear(__pyx_CyFunctionObject *m)
{
    Py_CLEAR(m->func_closure);
#if CYTHON_COMPILING_IN_LIMITED_API
    Py_CLEAR(m->func);
#else
    Py_CLEAR(((PyCFunctionObject*)m)->m_module);
#endif
    Py_CLEAR(m->func_dict);
    Py_CLEAR(m->func_name);
    Py_CLEAR(m->func_qualname);
    Py_CLEAR(m->func_doc);
    Py_CLEAR(m->func_globals);
    Py_CLEAR(m->func_code);
#if !CYTHON_COMPILING_IN_LIMITED_API
#if PY_VERSION_HEX < 0x030900B1
    Py_CLEAR(__Pyx_CyFunction_GetClassObj(m));
#else
    {
        PyObject *cls = (PyObject*) ((PyCMethodObject *) (m))->mm_class;
        ((PyCMethodObject *) (m))->mm_class = NULL;
        Py_XDECREF(cls);
    }
#endif
#endif
    Py_CLEAR(m->defaults_tuple);
    Py_CLEAR(m->defaults_kwdict);
    Py_CLEAR(m->func_annotations);
    Py_CLEAR(m->func_is_coroutine);
    if (m->defaults) {
        PyObject **pydefaults = __Pyx_CyFunction_Defaults(PyObject *, m);
        int i;
        for (i = 0; i < m->defaults_pyobjects; i++)
            Py_XDECREF(pydefaults[i]);
        PyObject_Free(m->defaults);
        m->defaults = NULL;
    }
    return 0;
}
static void __Pyx__CyFunction_dealloc(__pyx_CyFunctionObject *m)
{
    if (__Pyx_CyFunction_weakreflist(m) != NULL)
        PyObject_ClearWeakRefs((PyObject *) m);
    __Pyx_CyFunction_clear(m);
    __Pyx_PyHeapTypeObject_GC_Del(m);
}
static void __Pyx_CyFunction_dealloc(__pyx_CyFunctionObject *m)
{
    PyObject_GC_UnTrack(m);
    __Pyx__CyFunction_dealloc(m);
}
static int __Pyx_CyFunction_traverse(__pyx_CyFunctionObject *m, visitproc visit, void *arg)
{
    Py_VISIT(m->func_closure);
#if CYTHON_COMPILING_IN_LIMITED_API
    Py_VISIT(m->func);
#else
    Py_VISIT(((PyCFunctionObject*)m)->m_module);
#endif
    Py_VISIT(m->func_dict);
    Py_VISIT(m->func_name);
    Py_VISIT(m->func_qualname);
    Py_VISIT(m->func_doc);
    Py_VISIT(m->func_globals);
    Py_VISIT(m->func_code);
#if !CYTHON_COMPILING_IN_LIMITED_API
    Py_VISIT(__Pyx_CyFunction_GetClassObj(m));
#endif
    Py_VISIT(m->defaults_tuple);
    Py_VISIT(m->defaults_kwdict);
    Py_VISIT(m->func_is_coroutine);
    if (m->defaults) {
        PyObject **pydefaults = __Pyx_CyFunction_Defaults(PyObject *, m);
        int i;
        for (i = 0; i < m->defaults_pyobjects; i++)
            Py_VISIT(pydefaults[i]);
    }
    return 0;
}
static PyObject*
__Pyx_CyFunction_repr(__pyx_CyFunctionObject *op)
{
#if PY_MAJOR_VERSION >= 3
    return PyUnicode_FromFormat("<cyfunction %U at %p>",
                                op->func_qualname, (void *)op);
#else
    return PyString_FromFormat("<cyfunction %s at %p>",
                               PyString_AsString(op->func_qualname), (void *)op);
#endif
}
static PyObject * __Pyx_CyFunction_CallMethod(PyObject *func, PyObject *self, PyObject *arg, PyObject *kw) {
#if CYTHON_COMPILING_IN_LIMITED_API
    PyObject *f = ((__pyx_CyFunctionObject*)func)->func;
    PyObject *py_name = NULL;
    PyCFunction meth;
    int flags;
    meth = PyCFunction_GetFunction(f);
    if (unlikely(!meth)) return NULL;
    flags = PyCFunction_GetFlags(f);
    if (unlikely(flags < 0)) return NULL;
#else
    PyCFunctionObject* f = (PyCFunctionObject*)func;
    PyCFunction meth = f->m_ml->ml_meth;
    int flags = f->m_ml->ml_flags;
#endif
    Py_ssize_t size;
    switch (flags & (METH_VARARGS | METH_KEYWORDS | METH_NOARGS | METH_O)) {
    case METH_VARARGS:
        if (likely(kw == NULL || PyDict_Size(kw) == 0))
            return (*meth)(self, arg);
        break;
    case METH_VARARGS | METH_KEYWORDS:
        return (*(PyCFunctionWithKeywords)(void*)meth)(self, arg, kw);
    case METH_NOARGS:
        if (likely(kw == NULL || PyDict_Size(kw) == 0)) {
#if CYTHON_ASSUME_SAFE_MACROS
            size = PyTuple_GET_SIZE(arg);
#else
            size = PyTuple_Size(arg);
            if (unlikely(size < 0)) return NULL;
#endif
            if (likely(size == 0))
                return (*meth)(self, NULL);
#if CYTHON_COMPILING_IN_LIMITED_API
            py_name = __Pyx_CyFunction_get_name((__pyx_CyFunctionObject*)func, NULL);
            if (!py_name) return NULL;
            PyErr_Format(PyExc_TypeError,
                "%.200S() takes no arguments (%" CYTHON_FORMAT_SSIZE_T "d given)",
                py_name, size);
            Py_DECREF(py_name);
#else
            PyErr_Format(PyExc_TypeError,
                "%.200s() takes no arguments (%" CYTHON_FORMAT_SSIZE_T "d given)",
                f->m_ml->ml_name, size);
#endif
            return NULL;
        }
        break;
    case METH_O:
        if (likely(kw == NULL || PyDict_Size(kw) == 0)) {
#if CYTHON_ASSUME_SAFE_MACROS
            size = PyTuple_GET_SIZE(arg);
#else
            size = PyTuple_Size(arg);
            if (unlikely(size < 0)) return NULL;
#endif
            if (likely(size == 1)) {
                PyObject *result, *arg0;
                #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                arg0 = PyTuple_GET_ITEM(arg, 0);
                #else
                arg0 = __Pyx_PySequence_ITEM(arg, 0); if (unlikely(!arg0)) return NULL;
                #endif
                result = (*meth)(self, arg0);
                #if !(CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS)
                Py_DECREF(arg0);
                #endif
                return result;
            }
#if CYTHON_COMPILING_IN_LIMITED_API
            py_name = __Pyx_CyFunction_get_name((__pyx_CyFunctionObject*)func, NULL);
            if (!py_name) return NULL;
            PyErr_Format(PyExc_TypeError,
                "%.200S() takes exactly one argument (%" CYTHON_FORMAT_SSIZE_T "d given)",
                py_name, size);
            Py_DECREF(py_name);
#else
            PyErr_Format(PyExc_TypeError,
                "%.200s() takes exactly one argument (%" CYTHON_FORMAT_SSIZE_T "d given)",
                f->m_ml->ml_name, size);
#endif
            return NULL;
        }
        break;
    default:
        PyErr_SetString(PyExc_SystemError, "Bad call flags for CyFunction");
        return NULL;
    }
#if CYTHON_COMPILING_IN_LIMITED_API
    py_name = __Pyx_CyFunction_get_name((__pyx_CyFunctionObject*)func, NULL);
    if (!py_name) return NULL;
    PyErr_Format(PyExc_TypeError, "%.200S() takes no keyword arguments",
                 py_name);
    Py_DECREF(py_name);
#else
    PyErr_Format(PyExc_TypeError, "%.200s() takes no keyword arguments",
                 f->m_ml->ml_name);
#endif
    return NULL;
}
static CYTHON_INLINE PyObject *__Pyx_CyFunction_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    PyObject *self, *result;
#if CYTHON_COMPILING_IN_LIMITED_API
    self = PyCFunction_GetSelf(((__pyx_CyFunctionObject*)func)->func);
    if (unlikely(!self) && PyErr_Occurred()) return NULL;
#else
    self = ((PyCFunctionObject*)func)->m_self;
#endif
    result = __Pyx_CyFunction_CallMethod(func, self, arg, kw);
    return result;
}
static PyObject *__Pyx_CyFunction_CallAsMethod(PyObject *func, PyObject *args, PyObject *kw) {
    PyObject *result;
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *) func;
#if CYTHON_METH_FASTCALL
     __pyx_vectorcallfunc vc = __Pyx_CyFunction_func_vectorcall(cyfunc);
    if (vc) {
#if CYTHON_ASSUME_SAFE_MACROS
        return __Pyx_PyVectorcall_FastCallDict(func, vc, &PyTuple_GET_ITEM(args, 0), (size_t)PyTuple_GET_SIZE(args), kw);
#else
        (void) &__Pyx_PyVectorcall_FastCallDict;
        return PyVectorcall_Call(func, args, kw);
#endif
    }
#endif
    if ((cyfunc->flags & __Pyx_CYFUNCTION_CCLASS) && !(cyfunc->flags & __Pyx_CYFUNCTION_STATICMETHOD)) {
        Py_ssize_t argc;
        PyObject *new_args;
        PyObject *self;
#if CYTHON_ASSUME_SAFE_MACROS
        argc = PyTuple_GET_SIZE(args);
#else
        argc = PyTuple_Size(args);
        if (unlikely(!argc) < 0) return NULL;
#endif
        new_args = PyTuple_GetSlice(args, 1, argc);
        if (unlikely(!new_args))
            return NULL;
        self = PyTuple_GetItem(args, 0);
        if (unlikely(!self)) {
            Py_DECREF(new_args);
#if PY_MAJOR_VERSION > 2
            PyErr_Format(PyExc_TypeError,
                         "unbound method %.200S() needs an argument",
                         cyfunc->func_qualname);
#else
            PyErr_SetString(PyExc_TypeError,
                            "unbound method needs an argument");
#endif
            return NULL;
        }
        result = __Pyx_CyFunction_CallMethod(func, self, new_args, kw);
        Py_DECREF(new_args);
    } else {
        result = __Pyx_CyFunction_Call(func, args, kw);
    }
    return result;
}
#if CYTHON_METH_FASTCALL
static CYTHON_INLINE int __Pyx_CyFunction_Vectorcall_CheckArgs(__pyx_CyFunctionObject *cyfunc, Py_ssize_t nargs, PyObject *kwnames)
{
    int ret = 0;
    if ((cyfunc->flags & __Pyx_CYFUNCTION_CCLASS) && !(cyfunc->flags & __Pyx_CYFUNCTION_STATICMETHOD)) {
        if (unlikely(nargs < 1)) {
            PyErr_Format(PyExc_TypeError, "%.200s() needs an argument",
                         ((PyCFunctionObject*)cyfunc)->m_ml->ml_name);
            return -1;
        }
        ret = 1;
    }
    if (unlikely(kwnames) && unlikely(PyTuple_GET_SIZE(kwnames))) {
        PyErr_Format(PyExc_TypeError,
                     "%.200s() takes no keyword arguments", ((PyCFunctionObject*)cyfunc)->m_ml->ml_name);
        return -1;
    }
    return ret;
}
static PyObject * __Pyx_CyFunction_Vectorcall_NOARGS(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames)
{
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *)func;
    PyMethodDef* def = ((PyCFunctionObject*)cyfunc)->m_ml;
#if CYTHON_BACKPORT_VECTORCALL
    Py_ssize_t nargs = (Py_ssize_t)nargsf;
#else
    Py_ssize_t nargs = PyVectorcall_NARGS(nargsf);
#endif
    PyObject *self;
    switch (__Pyx_CyFunction_Vectorcall_CheckArgs(cyfunc, nargs, kwnames)) {
    case 1:
        self = args[0];
        args += 1;
        nargs -= 1;
        break;
    case 0:
        self = ((PyCFunctionObject*)cyfunc)->m_self;
        break;
    default:
        return NULL;
    }
    if (unlikely(nargs != 0)) {
        PyErr_Format(PyExc_TypeError,
            "%.200s() takes no arguments (%" CYTHON_FORMAT_SSIZE_T "d given)",
            def->ml_name, nargs);
        return NULL;
    }
    return def->ml_meth(self, NULL);
}
static PyObject * __Pyx_CyFunction_Vectorcall_O(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames)
{
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *)func;
    PyMethodDef* def = ((PyCFunctionObject*)cyfunc)->m_ml;
#if CYTHON_BACKPORT_VECTORCALL
    Py_ssize_t nargs = (Py_ssize_t)nargsf;
#else
    Py_ssize_t nargs = PyVectorcall_NARGS(nargsf);
#endif
    PyObject *self;
    switch (__Pyx_CyFunction_Vectorcall_CheckArgs(cyfunc, nargs, kwnames)) {
    case 1:
        self = args[0];
        args += 1;
        nargs -= 1;
        break;
    case 0:
        self = ((PyCFunctionObject*)cyfunc)->m_self;
        break;
    default:
        return NULL;
    }
    if (unlikely(nargs != 1)) {
        PyErr_Format(PyExc_TypeError,
            "%.200s() takes exactly one argument (%" CYTHON_FORMAT_SSIZE_T "d given)",
            def->ml_name, nargs);
        return NULL;
    }
    return def->ml_meth(self, args[0]);
}
static PyObject * __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames)
{
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *)func;
    PyMethodDef* def = ((PyCFunctionObject*)cyfunc)->m_ml;
#if CYTHON_BACKPORT_VECTORCALL
    Py_ssize_t nargs = (Py_ssize_t)nargsf;
#else
    Py_ssize_t nargs = PyVectorcall_NARGS(nargsf);
#endif
    PyObject *self;
    switch (__Pyx_CyFunction_Vectorcall_CheckArgs(cyfunc, nargs, NULL)) {
    case 1:
        self = args[0];
        args += 1;
        nargs -= 1;
        break;
    case 0:
        self = ((PyCFunctionObject*)cyfunc)->m_self;
        break;
    default:
        return NULL;
    }
    return ((__Pyx_PyCFunctionFastWithKeywords)(void(*)(void))def->ml_meth)(self, args, nargs, kwnames);
}
static PyObject * __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS_METHOD(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames)
{
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *)func;
    PyMethodDef* def = ((PyCFunctionObject*)cyfunc)->m_ml;
    PyTypeObject *cls = (PyTypeObject *) __Pyx_CyFunction_GetClassObj(cyfunc);
#if CYTHON_BACKPORT_VECTORCALL
    Py_ssize_t nargs = (Py_ssize_t)nargsf;
#else
    Py_ssize_t nargs = PyVectorcall_NARGS(nargsf);
#endif
    PyObject *self;
    switch (__Pyx_CyFunction_Vectorcall_CheckArgs(cyfunc, nargs, NULL)) {
    case 1:
        self = args[0];
        args += 1;
        nargs -= 1;
        break;
    case 0:
        self = ((PyCFunctionObject*)cyfunc)->m_self;
        break;
    default:
        return NULL;
    }
    return ((__Pyx_PyCMethod)(void(*)(void))def->ml_meth)(self, cls, args, (size_t)nargs, kwnames);
}
#endif
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_CyFunctionType_slots[] = {
    {Py_tp_dealloc, (void *)__Pyx_CyFunction_dealloc},
    {Py_tp_repr, (void *)__Pyx_CyFunction_repr},
    {Py_tp_call, (void *)__Pyx_CyFunction_CallAsMethod},
    {Py_tp_traverse, (void *)__Pyx_CyFunction_traverse},
    {Py_tp_clear, (void *)__Pyx_CyFunction_clear},
    {Py_tp_methods, (void *)__pyx_CyFunction_methods},
    {Py_tp_members, (void *)__pyx_CyFunction_members},
    {Py_tp_getset, (void *)__pyx_CyFunction_getsets},
    {Py_tp_descr_get, (void *)__Pyx_PyMethod_New},
    {0, 0},
};
static PyType_Spec __pyx_CyFunctionType_spec = {
    __PYX_TYPE_MODULE_PREFIX "cython_function_or_method",
    sizeof(__pyx_CyFunctionObject),
    0,
#ifdef Py_TPFLAGS_METHOD_DESCRIPTOR
    Py_TPFLAGS_METHOD_DESCRIPTOR |
#endif
#if (defined(_Py_TPFLAGS_HAVE_VECTORCALL) && CYTHON_METH_FASTCALL)
    _Py_TPFLAGS_HAVE_VECTORCALL |
#endif
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC | Py_TPFLAGS_BASETYPE,
    __pyx_CyFunctionType_slots
};
#else
static PyTypeObject __pyx_CyFunctionType_type = {
    PyVarObject_HEAD_INIT(0, 0)
    __PYX_TYPE_MODULE_PREFIX "cython_function_or_method",
    sizeof(__pyx_CyFunctionObject),
    0,
    (destructor) __Pyx_CyFunction_dealloc,
#if !CYTHON_METH_FASTCALL
    0,
#elif CYTHON_BACKPORT_VECTORCALL
    (printfunc)offsetof(__pyx_CyFunctionObject, func_vectorcall),
#else
    offsetof(PyCFunctionObject, vectorcall),
#endif
    0,
    0,
#if PY_MAJOR_VERSION < 3
    0,
#else
    0,
#endif
    (reprfunc) __Pyx_CyFunction_repr,
    0,
    0,
    0,
    0,
    __Pyx_CyFunction_CallAsMethod,
    0,
    0,
    0,
    0,
#ifdef Py_TPFLAGS_METHOD_DESCRIPTOR
    Py_TPFLAGS_METHOD_DESCRIPTOR |
#endif
#if defined(_Py_TPFLAGS_HAVE_VECTORCALL) && CYTHON_METH_FASTCALL
    _Py_TPFLAGS_HAVE_VECTORCALL |
#endif
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC | Py_TPFLAGS_BASETYPE,
    0,
    (traverseproc) __Pyx_CyFunction_traverse,
    (inquiry) __Pyx_CyFunction_clear,
    0,
#if PY_VERSION_HEX < 0x030500A0
    offsetof(__pyx_CyFunctionObject, func_weakreflist),
#else
    offsetof(PyCFunctionObject, m_weakreflist),
#endif
    0,
    0,
    __pyx_CyFunction_methods,
    __pyx_CyFunction_members,
    __pyx_CyFunction_getsets,
    0,
    0,
    __Pyx_PyMethod_New,
    0,
    offsetof(__pyx_CyFunctionObject, func_dict),
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
#if PY_VERSION_HEX >= 0x030400a1
    0,
#endif
#if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
    0,
#endif
#if __PYX_NEED_TP_PRINT_SLOT
    0,
#endif
#if PY_VERSION_HEX >= 0x030C0000
    0,
#endif
#if PY_VERSION_HEX >= 0x030d00A4
    0,
#endif
#if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
    0,
#endif
};
#endif
static int __pyx_CyFunction_init(PyObject *module) {
#if CYTHON_USE_TYPE_SPECS
    __pyx_CyFunctionType = __Pyx_FetchCommonTypeFromSpec(module, &__pyx_CyFunctionType_spec, NULL);
#else
    CYTHON_UNUSED_VAR(module);
    __pyx_CyFunctionType = __Pyx_FetchCommonType(&__pyx_CyFunctionType_type);
#endif
    if (unlikely(__pyx_CyFunctionType == NULL)) {
        return -1;
    }
    return 0;
}
static CYTHON_INLINE void *__Pyx_CyFunction_InitDefaults(PyObject *func, size_t size, int pyobjects) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults = PyObject_Malloc(size);
    if (unlikely(!m->defaults))
        return PyErr_NoMemory();
    memset(m->defaults, 0, size);
    m->defaults_pyobjects = pyobjects;
    m->defaults_size = size;
    return m->defaults;
}
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsTuple(PyObject *func, PyObject *tuple) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults_tuple = tuple;
    Py_INCREF(tuple);
}
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsKwDict(PyObject *func, PyObject *dict) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults_kwdict = dict;
    Py_INCREF(dict);
}
static CYTHON_INLINE void __Pyx_CyFunction_SetAnnotationsDict(PyObject *func, PyObject *dict) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->func_annotations = dict;
    Py_INCREF(dict);
}

/* CythonFunction */
static PyObject *__Pyx_CyFunction_New(PyMethodDef *ml, int flags, PyObject* qualname,
                                      PyObject *closure, PyObject *module, PyObject* globals, PyObject* code) {
    PyObject *op = __Pyx_CyFunction_Init(
        PyObject_GC_New(__pyx_CyFunctionObject, __pyx_CyFunctionType),
        ml, flags, qualname, closure, module, globals, code
    );
    if (likely(op)) {
        PyObject_GC_Track(op);
    }
    return op;
}

/* CLineInTraceback */
#ifndef CYTHON_CLINE_IN_TRACEBACK
static int __Pyx_CLineForTraceback(PyThreadState *tstate, int c_line) {
    PyObject *use_cline;
    PyObject *ptype, *pvalue, *ptraceback;
#if CYTHON_COMPILING_IN_CPYTHON
    PyObject **cython_runtime_dict;
#endif
    CYTHON_MAYBE_UNUSED_VAR(tstate);
    if (unlikely(!__pyx_cython_runtime)) {
        return c_line;
    }
    __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
#if CYTHON_COMPILING_IN_CPYTHON
    cython_runtime_dict = _PyObject_GetDictPtr(__pyx_cython_runtime);
    if (likely(cython_runtime_dict)) {
        __PYX_PY_DICT_LOOKUP_IF_MODIFIED(
            use_cline, *cython_runtime_dict,
            __Pyx_PyDict_GetItemStr(*cython_runtime_dict, __pyx_n_s_cline_in_traceback))
    } else
#endif
    {
      PyObject *use_cline_obj = __Pyx_PyObject_GetAttrStrNoError(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback);
      if (use_cline_obj) {
        use_cline = PyObject_Not(use_cline_obj) ? Py_False : Py_True;
        Py_DECREF(use_cline_obj);
      } else {
        PyErr_Clear();
        use_cline = NULL;
      }
    }
    if (!use_cline) {
        c_line = 0;
        (void) PyObject_SetAttr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback, Py_False);
    }
    else if (use_cline == Py_False || (use_cline != Py_True && PyObject_Not(use_cline) != 0)) {
        c_line = 0;
    }
    __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
    return c_line;
}
#endif

/* CodeObjectCache */
#if !CYTHON_COMPILING_IN_LIMITED_API
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line) {
    int start = 0, mid = 0, end = count - 1;
    if (end >= 0 && code_line > entries[end].code_line) {
        return count;
    }
    while (start < end) {
        mid = start + (end - start) / 2;
        if (code_line < entries[mid].code_line) {
            end = mid;
        } else if (code_line > entries[mid].code_line) {
             start = mid + 1;
        } else {
            return mid;
        }
    }
    if (code_line <= entries[mid].code_line) {
        return mid;
    } else {
        return mid + 1;
    }
}
static PyCodeObject *__pyx_find_code_object(int code_line) {
    PyCodeObject* code_object;
    int pos;
    if (unlikely(!code_line) || unlikely(!__pyx_code_cache.entries)) {
        return NULL;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if (unlikely(pos >= __pyx_code_cache.count) || unlikely(__pyx_code_cache.entries[pos].code_line != code_line)) {
        return NULL;
    }
    code_object = __pyx_code_cache.entries[pos].code_object;
    Py_INCREF(code_object);
    return code_object;
}
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object) {
    int pos, i;
    __Pyx_CodeObjectCacheEntry* entries = __pyx_code_cache.entries;
    if (unlikely(!code_line)) {
        return;
    }
    if (unlikely(!entries)) {
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Malloc(64*sizeof(__Pyx_CodeObjectCacheEntry));
        if (likely(entries)) {
            __pyx_code_cache.entries = entries;
            __pyx_code_cache.max_count = 64;
            __pyx_code_cache.count = 1;
            entries[0].code_line = code_line;
            entries[0].code_object = code_object;
            Py_INCREF(code_object);
        }
        return;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if ((pos < __pyx_code_cache.count) && unlikely(__pyx_code_cache.entries[pos].code_line == code_line)) {
        PyCodeObject* tmp = entries[pos].code_object;
        entries[pos].code_object = code_object;
        Py_DECREF(tmp);
        return;
    }
    if (__pyx_code_cache.count == __pyx_code_cache.max_count) {
        int new_max = __pyx_code_cache.max_count + 64;
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Realloc(
            __pyx_code_cache.entries, ((size_t)new_max) * sizeof(__Pyx_CodeObjectCacheEntry));
        if (unlikely(!entries)) {
            return;
        }
        __pyx_code_cache.entries = entries;
        __pyx_code_cache.max_count = new_max;
    }
    for (i=__pyx_code_cache.count; i>pos; i--) {
        entries[i] = entries[i-1];
    }
    entries[pos].code_line = code_line;
    entries[pos].code_object = code_object;
    __pyx_code_cache.count++;
    Py_INCREF(code_object);
}
#endif

/* AddTraceback */
#include "compile.h"
#include "frameobject.h"
#include "traceback.h"
#if PY_VERSION_HEX >= 0x030b00a6 && !CYTHON_COMPILING_IN_LIMITED_API && !defined(PYPY_VERSION)
  #ifndef Py_BUILD_CORE
    #define Py_BUILD_CORE 1
  #endif
  #include "internal/pycore_frame.h"
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
static PyObject *__Pyx_PyCode_Replace_For_AddTraceback(PyObject *code, PyObject *scratch_dict,
                                                       PyObject *firstlineno, PyObject *name) {
    PyObject *replace = NULL;
    if (unlikely(PyDict_SetItemString(scratch_dict, "co_firstlineno", firstlineno))) return NULL;
    if (unlikely(PyDict_SetItemString(scratch_dict, "co_name", name))) return NULL;
    replace = PyObject_GetAttrString(code, "replace");
    if (likely(replace)) {
        PyObject *result;
        result = PyObject_Call(replace, __pyx_empty_tuple, scratch_dict);
        Py_DECREF(replace);
        return result;
    }
    PyErr_Clear();
    #if __PYX_LIMITED_VERSION_HEX < 0x030780000
    {
        PyObject *compiled = NULL, *result = NULL;
        if (unlikely(PyDict_SetItemString(scratch_dict, "code", code))) return NULL;
        if (unlikely(PyDict_SetItemString(scratch_dict, "type", (PyObject*)(&PyType_Type)))) return NULL;
        compiled = Py_CompileString(
            "out = type(code)(\n"
            "  code.co_argcount, code.co_kwonlyargcount, code.co_nlocals, code.co_stacksize,\n"
            "  code.co_flags, code.co_code, code.co_consts, code.co_names,\n"
            "  code.co_varnames, code.co_filename, co_name, co_firstlineno,\n"
            "  code.co_lnotab)\n", "<dummy>", Py_file_input);
        if (!compiled) return NULL;
        result = PyEval_EvalCode(compiled, scratch_dict, scratch_dict);
        Py_DECREF(compiled);
        if (!result) PyErr_Print();
        Py_DECREF(result);
        result = PyDict_GetItemString(scratch_dict, "out");
        if (result) Py_INCREF(result);
        return result;
    }
    #else
    return NULL;
    #endif
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyObject *code_object = NULL, *py_py_line = NULL, *py_funcname = NULL, *dict = NULL;
    PyObject *replace = NULL, *getframe = NULL, *frame = NULL;
    PyObject *exc_type, *exc_value, *exc_traceback;
    int success = 0;
    if (c_line) {
        (void) __pyx_cfilenm;
        (void) __Pyx_CLineForTraceback(__Pyx_PyThreadState_Current, c_line);
    }
    PyErr_Fetch(&exc_type, &exc_value, &exc_traceback);
    code_object = Py_CompileString("_getframe()", filename, Py_eval_input);
    if (unlikely(!code_object)) goto bad;
    py_py_line = PyLong_FromLong(py_line);
    if (unlikely(!py_py_line)) goto bad;
    py_funcname = PyUnicode_FromString(funcname);
    if (unlikely(!py_funcname)) goto bad;
    dict = PyDict_New();
    if (unlikely(!dict)) goto bad;
    {
        PyObject *old_code_object = code_object;
        code_object = __Pyx_PyCode_Replace_For_AddTraceback(code_object, dict, py_py_line, py_funcname);
        Py_DECREF(old_code_object);
    }
    if (unlikely(!code_object)) goto bad;
    getframe = PySys_GetObject("_getframe");
    if (unlikely(!getframe)) goto bad;
    if (unlikely(PyDict_SetItemString(dict, "_getframe", getframe))) goto bad;
    frame = PyEval_EvalCode(code_object, dict, dict);
    if (unlikely(!frame) || frame == Py_None) goto bad;
    success = 1;
  bad:
    PyErr_Restore(exc_type, exc_value, exc_traceback);
    Py_XDECREF(code_object);
    Py_XDECREF(py_py_line);
    Py_XDECREF(py_funcname);
    Py_XDECREF(dict);
    Py_XDECREF(replace);
    if (success) {
        PyTraceBack_Here(
            (struct _frame*)frame);
    }
    Py_XDECREF(frame);
}
#else
static PyCodeObject* __Pyx_CreateCodeObjectForTraceback(
            const char *funcname, int c_line,
            int py_line, const char *filename) {
    PyCodeObject *py_code = NULL;
    PyObject *py_funcname = NULL;
    #if PY_MAJOR_VERSION < 3
    PyObject *py_srcfile = NULL;
    py_srcfile = PyString_FromString(filename);
    if (!py_srcfile) goto bad;
    #endif
    if (c_line) {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        if (!py_funcname) goto bad;
        #else
        py_funcname = PyUnicode_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        if (!py_funcname) goto bad;
        funcname = PyUnicode_AsUTF8(py_funcname);
        if (!funcname) goto bad;
        #endif
    }
    else {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromString(funcname);
        if (!py_funcname) goto bad;
        #endif
    }
    #if PY_MAJOR_VERSION < 3
    py_code = __Pyx_PyCode_New(
        0,
        0,
        0,
        0,
        0,
        0,
        __pyx_empty_bytes, /*PyObject *code,*/
        __pyx_empty_tuple, /*PyObject *consts,*/
        __pyx_empty_tuple, /*PyObject *names,*/
        __pyx_empty_tuple, /*PyObject *varnames,*/
        __pyx_empty_tuple, /*PyObject *freevars,*/
        __pyx_empty_tuple, /*PyObject *cellvars,*/
        py_srcfile,   /*PyObject *filename,*/
        py_funcname,  /*PyObject *name,*/
        py_line,
        __pyx_empty_bytes  /*PyObject *lnotab*/
    );
    Py_DECREF(py_srcfile);
    #else
    py_code = PyCode_NewEmpty(filename, funcname, py_line);
    #endif
    Py_XDECREF(py_funcname);
    return py_code;
bad:
    Py_XDECREF(py_funcname);
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(py_srcfile);
    #endif
    return NULL;
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyFrameObject *py_frame = 0;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject *ptype, *pvalue, *ptraceback;
    if (c_line) {
        c_line = __Pyx_CLineForTraceback(tstate, c_line);
    }
    py_code = __pyx_find_code_object(c_line ? -c_line : py_line);
    if (!py_code) {
        __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
        py_code = __Pyx_CreateCodeObjectForTraceback(
            funcname, c_line, py_line, filename);
        if (!py_code) {
            /* If the code object creation fails, then we should clear the
               fetched exception references and propagate the new exception */
            Py_XDECREF(ptype);
            Py_XDECREF(pvalue);
            Py_XDECREF(ptraceback);
            goto bad;
        }
        __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
        __pyx_insert_code_object(c_line ? -c_line : py_line, py_code);
    }
    py_frame = PyFrame_New(
        tstate,            /*PyThreadState *tstate,*/
        py_code,           /*PyCodeObject *code,*/
        __pyx_d,    /*PyObject *globals,*/
        0                  /*PyObject *locals*/
    );
    if (!py_frame) goto bad;
    __Pyx_PyFrame_SetLineNumber(py_frame, py_line);
    PyTraceBack_Here(py_frame);
bad:
    Py_XDECREF(py_code);
    Py_XDECREF(py_frame);
}
#endif

/* CIntFromPyVerify */
#define __PYX_VERIFY_RETURN_INT(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 0)
#define __PYX_VERIFY_RETURN_INT_EXC(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 1)
#define __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, exc)\
    {\
        func_type value = func_value;\
        if (sizeof(target_type) < sizeof(func_type)) {\
            if (unlikely(value != (func_type) (target_type) value)) {\
                func_type zero = 0;\
                if (exc && unlikely(value == (func_type)-1 && PyErr_Occurred()))\
                    return (target_type) -1;\
                if (is_unsigned && unlikely(value < zero))\
                    goto raise_neg_overflow;\
                else\
                    goto raise_overflow;\
            }\
        }\
        return (target_type) value;\
    }

/* Declarations */
#if CYTHON_CCOMPLEX && (1) && (!0 || __cplusplus)
  #ifdef __cplusplus
    static CYTHON_INLINE __pyx_t_float_complex __pyx_t_float_complex_from_parts(float x, float y) {
      return ::std::complex< float >(x, y);
    }
  #else
    static CYTHON_INLINE __pyx_t_float_complex __pyx_t_float_complex_from_parts(float x, float y) {
      return x + y*(__pyx_t_float_complex)_Complex_I;
    }
  #endif
#else
    static CYTHON_INLINE __pyx_t_float_complex __pyx_t_float_complex_from_parts(float x, float y) {
      __pyx_t_float_complex z;
      z.real = x;
      z.imag = y;
      return z;
    }
#endif

/* Arithmetic */
#if CYTHON_CCOMPLEX && (1) && (!0 || __cplusplus)
#else
    static CYTHON_INLINE int __Pyx_c_eq_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
       return (a.real == b.real) && (a.imag == b.imag);
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_sum_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        __pyx_t_float_complex z;
        z.real = a.real + b.real;
        z.imag = a.imag + b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_diff_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        __pyx_t_float_complex z;
        z.real = a.real - b.real;
        z.imag = a.imag - b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_prod_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        __pyx_t_float_complex z;
        z.real = a.real * b.real - a.imag * b.imag;
        z.imag = a.real * b.imag + a.imag * b.real;
        return z;
    }
    #if 1
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_quot_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        if (b.imag == 0) {
            return __pyx_t_float_complex_from_parts(a.real / b.real, a.imag / b.real);
        } else if (fabsf(b.real) >= fabsf(b.imag)) {
            if (b.real == 0 && b.imag == 0) {
                return __pyx_t_float_complex_from_parts(a.real / b.real, a.imag / b.imag);
            } else {
                float r = b.imag / b.real;
                float s = (float)(1.0) / (b.real + b.imag * r);
                return __pyx_t_float_complex_from_parts(
                    (a.real + a.imag * r) * s, (a.imag - a.real * r) * s);
            }
        } else {
            float r = b.real / b.imag;
            float s = (float)(1.0) / (b.imag + b.real * r);
            return __pyx_t_float_complex_from_parts(
                (a.real * r + a.imag) * s, (a.imag * r - a.real) * s);
        }
    }
    #else
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_quot_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        if (b.imag == 0) {
            return __pyx_t_float_complex_from_parts(a.real / b.real, a.imag / b.real);
        } else {
            float denom = b.real * b.real + b.imag * b.imag;
            return __pyx_t_float_complex_from_parts(
                (a.real * b.real + a.imag * b.imag) / denom,
                (a.imag * b.real - a.real * b.imag) / denom);
        }
    }
    #endif
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_neg_float(__pyx_t_float_complex a) {
        __pyx_t_float_complex z;
        z.real = -a.real;
        z.imag = -a.imag;
        return z;
    }
    static CYTHON_INLINE int __Pyx_c_is_zero_float(__pyx_t_float_complex a) {
       return (a.real == 0) && (a.imag == 0);
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_conj_float(__pyx_t_float_complex a) {
        __pyx_t_float_complex z;
        z.real =  a.real;
        z.imag = -a.imag;
        return z;
    }
    #if 1
        static CYTHON_INLINE float __Pyx_c_abs_float(__pyx_t_float_complex z) {
          #if !defined(HAVE_HYPOT) || defined(_MSC_VER)
            return sqrtf(z.real*z.real + z.imag*z.imag);
          #else
            return hypotf(z.real, z.imag);
          #endif
        }
        static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_pow_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
            __pyx_t_float_complex z;
            float r, lnr, theta, z_r, z_theta;
            if (b.imag == 0 && b.real == (int)b.real) {
                if (b.real < 0) {
                    float denom = a.real * a.real + a.imag * a.imag;
                    a.real = a.real / denom;
                    a.imag = -a.imag / denom;
                    b.real = -b.real;
                }
                switch ((int)b.real) {
                    case 0:
                        z.real = 1;
                        z.imag = 0;
                        return z;
                    case 1:
                        return a;
                    case 2:
                        return __Pyx_c_prod_float(a, a);
                    case 3:
                        z = __Pyx_c_prod_float(a, a);
                        return __Pyx_c_prod_float(z, a);
                    case 4:
                        z = __Pyx_c_prod_float(a, a);
                        return __Pyx_c_prod_float(z, z);
                }
            }
            if (a.imag == 0) {
                if (a.real == 0) {
                    return a;
                } else if ((b.imag == 0) && (a.real >= 0)) {
                    z.real = powf(a.real, b.real);
                    z.imag = 0;
                    return z;
                } else if (a.real > 0) {
                    r = a.real;
                    theta = 0;
                } else {
                    r = -a.real;
                    theta = atan2f(0.0, -1.0);
                }
            } else {
                r = __Pyx_c_abs_float(a);
                theta = atan2f(a.imag, a.real);
            }
            lnr = logf(r);
            z_r = expf(lnr * b.real - theta * b.imag);
            z_theta = theta * b.real + lnr * b.imag;
            z.real = z_r * cosf(z_theta);
            z.imag = z_r * sinf(z_theta);
            return z;
        }
    #endif
#endif

/* Declarations */
#if CYTHON_CCOMPLEX && (1) && (!0 || __cplusplus)
  #ifdef __cplusplus
    static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double x, double y) {
      return ::std::complex< double >(x, y);
    }
  #else
    static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double x, double y) {
      return x + y*(__pyx_t_double_complex)_Complex_I;
    }
  #endif
#else
    static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double x, double y) {
      __pyx_t_double_complex z;
      z.real = x;
      z.imag = y;
      return z;
    }
#endif

/* Arithmetic */
#if CYTHON_CCOMPLEX && (1) && (!0 || __cplusplus)
#else
    static CYTHON_INLINE int __Pyx_c_eq_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
       return (a.real == b.real) && (a.imag == b.imag);
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_sum_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        __pyx_t_double_complex z;
        z.real = a.real + b.real;
        z.imag = a.imag + b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_diff_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        __pyx_t_double_complex z;
        z.real = a.real - b.real;
        z.imag = a.imag - b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_prod_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        __pyx_t_double_complex z;
        z.real = a.real * b.real - a.imag * b.imag;
        z.imag = a.real * b.imag + a.imag * b.real;
        return z;
    }
    #if 1
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_quot_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        if (b.imag == 0) {
            return __pyx_t_double_complex_from_parts(a.real / b.real, a.imag / b.real);
        } else if (fabs(b.real) >= fabs(b.imag)) {
            if (b.real == 0 && b.imag == 0) {
                return __pyx_t_double_complex_from_parts(a.real / b.real, a.imag / b.imag);
            } else {
                double r = b.imag / b.real;
                double s = (double)(1.0) / (b.real + b.imag * r);
                return __pyx_t_double_complex_from_parts(
                    (a.real + a.imag * r) * s, (a.imag - a.real * r) * s);
            }
        } else {
            double r = b.real / b.imag;
            double s = (double)(1.0) / (b.imag + b.real * r);
            return __pyx_t_double_complex_from_parts(
                (a.real * r + a.imag) * s, (a.imag * r - a.real) * s);
        }
    }
    #else
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_quot_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        if (b.imag == 0) {
            return __pyx_t_double_complex_from_parts(a.real / b.real, a.imag / b.real);
        } else {
            double denom = b.real * b.real + b.imag * b.imag;
            return __pyx_t_double_complex_from_parts(
                (a.real * b.real + a.imag * b.imag) / denom,
                (a.imag * b.real - a.real * b.imag) / denom);
        }
    }
    #endif
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_neg_double(__pyx_t_double_complex a) {
        __pyx_t_double_complex z;
        z.real = -a.real;
        z.imag = -a.imag;
        return z;
    }
    static CYTHON_INLINE int __Pyx_c_is_zero_double(__pyx_t_double_complex a) {
       return (a.real == 0) && (a.imag == 0);
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_conj_double(__pyx_t_double_complex a) {
        __pyx_t_double_complex z;
        z.real =  a.real;
        z.imag = -a.imag;
        return z;
    }
    #if 1
        static CYTHON_INLINE double __Pyx_c_abs_double(__pyx_t_double_complex z) {
          #if !defined(HAVE_HYPOT) || defined(_MSC_VER)
            return sqrt(z.real*z.real + z.imag*z.imag);
          #else
            return hypot(z.real, z.imag);
          #endif
        }
        static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_pow_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
            __pyx_t_double_complex z;
            double r, lnr, theta, z_r, z_theta;
            if (b.imag == 0 && b.real == (int)b.real) {
                if (b.real < 0) {
                    double denom = a.real * a.real + a.imag * a.imag;
                    a.real = a.real / denom;
                    a.imag = -a.imag / denom;
                    b.real = -b.real;
                }
                switch ((int)b.real) {
                    case 0:
                        z.real = 1;
                        z.imag = 0;
                        return z;
                    case 1:
                        return a;
                    case 2:
                        return __Pyx_c_prod_double(a, a);
                    case 3:
                        z = __Pyx_c_prod_double(a, a);
                        return __Pyx_c_prod_double(z, a);
                    case 4:
                        z = __Pyx_c_prod_double(a, a);
                        return __Pyx_c_prod_double(z, z);
                }
            }
            if (a.imag == 0) {
                if (a.real == 0) {
                    return a;
                } else if ((b.imag == 0) && (a.real >= 0)) {
                    z.real = pow(a.real, b.real);
                    z.imag = 0;
                    return z;
                } else if (a.real > 0) {
                    r = a.real;
                    theta = 0;
                } else {
                    r = -a.real;
                    theta = atan2(0.0, -1.0);
                }
            } else {
                r = __Pyx_c_abs_double(a);
                theta = atan2(a.imag, a.real);
            }
            lnr = log(r);
            z_r = exp(lnr * b.real - theta * b.imag);
            z_theta = theta * b.real + lnr * b.imag;
            z.real = z_r * cos(z_theta);
            z.imag = z_r * sin(z_theta);
            return z;
        }
    #endif
#endif

/* Declarations */
#if CYTHON_CCOMPLEX && (1) && (!0 || __cplusplus)
  #ifdef __cplusplus
    static CYTHON_INLINE __pyx_t_long_double_complex __pyx_t_long_double_complex_from_parts(long double x, long double y) {
      return ::std::complex< long double >(x, y);
    }
  #else
    static CYTHON_INLINE __pyx_t_long_double_complex __pyx_t_long_double_complex_from_parts(long double x, long double y) {
      return x + y*(__pyx_t_long_double_complex)_Complex_I;
    }
  #endif
#else
    static CYTHON_INLINE __pyx_t_long_double_complex __pyx_t_long_double_complex_from_parts(long double x, long double y) {
      __pyx_t_long_double_complex z;
      z.real = x;
      z.imag = y;
      return z;
    }
#endif

/* Arithmetic */
#if CYTHON_CCOMPLEX && (1) && (!0 || __cplusplus)
#else
    static CYTHON_INLINE int __Pyx_c_eq_long__double(__pyx_t_long_double_complex a, __pyx_t_long_double_complex b) {
       return (a.real == b.real) && (a.imag == b.imag);
    }
    static CYTHON_INLINE __pyx_t_long_double_complex __Pyx_c_sum_long__double(__pyx_t_long_double_complex a, __pyx_t_long_double_complex b) {
        __pyx_t_long_double_complex z;
        z.real = a.real + b.real;
        z.imag = a.imag + b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_long_double_complex __Pyx_c_diff_long__double(__pyx_t_long_double_complex a, __pyx_t_long_double_complex b) {
        __pyx_t_long_double_complex z;
        z.real = a.real - b.real;
        z.imag = a.imag - b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_long_double_complex __Pyx_c_prod_long__double(__pyx_t_long_double_complex a, __pyx_t_long_double_complex b) {
        __pyx_t_long_double_complex z;
        z.real = a.real * b.real - a.imag * b.imag;
        z.imag = a.real * b.imag + a.imag * b.real;
        return z;
    }
    #if 1
    static CYTHON_INLINE __pyx_t_long_double_complex __Pyx_c_quot_long__double(__pyx_t_long_double_complex a, __pyx_t_long_double_complex b) {
        if (b.imag == 0) {
            return __pyx_t_long_double_complex_from_parts(a.real / b.real, a.imag / b.real);
        } else if (fabsl(b.real) >= fabsl(b.imag)) {
            if (b.real == 0 && b.imag == 0) {
                return __pyx_t_long_double_complex_from_parts(a.real / b.real, a.imag / b.imag);
            } else {
                long double r = b.imag / b.real;
                long double s = (long double)(1.0) / (b.real + b.imag * r);
                return __pyx_t_long_double_complex_from_parts(
                    (a.real + a.imag * r) * s, (a.imag - a.real * r) * s);
            }
        } else {
            long double r = b.real / b.imag;
            long double s = (long double)(1.0) / (b.imag + b.real * r);
            return __pyx_t_long_double_complex_from_parts(
                (a.real * r + a.imag) * s, (a.imag * r - a.real) * s);
        }
    }
    #else
    static CYTHON_INLINE __pyx_t_long_double_complex __Pyx_c_quot_long__double(__pyx_t_long_double_complex a, __pyx_t_long_double_complex b) {
        if (b.imag == 0) {
            return __pyx_t_long_double_complex_from_parts(a.real / b.real, a.imag / b.real);
        } else {
            long double denom = b.real * b.real + b.imag * b.imag;
            return __pyx_t_long_double_complex_from_parts(
                (a.real * b.real + a.imag * b.imag) / denom,
                (a.imag * b.real - a.real * b.imag) / denom);
        }
    }
    #endif
    static CYTHON_INLINE __pyx_t_long_double_complex __Pyx_c_neg_long__double(__pyx_t_long_double_complex a) {
        __pyx_t_long_double_complex z;
        z.real = -a.real;
        z.imag = -a.imag;
        return z;
    }
    static CYTHON_INLINE int __Pyx_c_is_zero_long__double(__pyx_t_long_double_complex a) {
       return (a.real == 0) && (a.imag == 0);
    }
    static CYTHON_INLINE __pyx_t_long_double_complex __Pyx_c_conj_long__double(__pyx_t_long_double_complex a) {
        __pyx_t_long_double_complex z;
        z.real =  a.real;
        z.imag = -a.imag;
        return z;
    }
    #if 1
        static CYTHON_INLINE long double __Pyx_c_abs_long__double(__pyx_t_long_double_complex z) {
          #if !defined(HAVE_HYPOT) || defined(_MSC_VER)
            return sqrtl(z.real*z.real + z.imag*z.imag);
          #else
            return hypotl(z.real, z.imag);
          #endif
        }
        static CYTHON_INLINE __pyx_t_long_double_complex __Pyx_c_pow_long__double(__pyx_t_long_double_complex a, __pyx_t_long_double_complex b) {
            __pyx_t_long_double_complex z;
            long double r, lnr, theta, z_r, z_theta;
            if (b.imag == 0 && b.real == (int)b.real) {
                if (b.real < 0) {
                    long double denom = a.real * a.real + a.imag * a.imag;
                    a.real = a.real / denom;
                    a.imag = -a.imag / denom;
                    b.real = -b.real;
                }
                switch ((int)b.real) {
                    case 0:
                        z.real = 1;
                        z.imag = 0;
                        return z;
                    case 1:
                        return a;
                    case 2:
                        return __Pyx_c_prod_long__double(a, a);
                    case 3:
                        z = __Pyx_c_prod_long__double(a, a);
                        return __Pyx_c_prod_long__double(z, a);
                    case 4:
                        z = __Pyx_c_prod_long__double(a, a);
                        return __Pyx_c_prod_long__double(z, z);
                }
            }
            if (a.imag == 0) {
                if (a.real == 0) {
                    return a;
                } else if ((b.imag == 0) && (a.real >= 0)) {
                    z.real = powl(a.real, b.real);
                    z.imag = 0;
                    return z;
                } else if (a.real > 0) {
                    r = a.real;
                    theta = 0;
                } else {
                    r = -a.real;
                    theta = atan2l(0.0, -1.0);
                }
            } else {
                r = __Pyx_c_abs_long__double(a);
                theta = atan2l(a.imag, a.real);
            }
            lnr = logl(r);
            z_r = expl(lnr * b.real - theta * b.imag);
            z_theta = theta * b.real + lnr * b.imag;
            z.real = z_r * cosl(z_theta);
            z.imag = z_r * sinl(z_theta);
            return z;
        }
    #endif
#endif

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int neg_one = (int) -1, const_zero = (int) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(int) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(int) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(int) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(int),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL;
        PyObject *py_bytes = NULL, *arg_tuple = NULL, *kwds = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(int));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        arg_tuple = PyTuple_Pack(2, py_bytes, order_str);
        if (!arg_tuple) goto limited_bad;
        if (!is_unsigned) {
            kwds = PyDict_New();
            if (!kwds) goto limited_bad;
            if (PyDict_SetItemString(kwds, "signed", __Pyx_NewRef(Py_True))) goto limited_bad;
        }
        result = PyObject_Call(from_bytes, arg_tuple, kwds);
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(arg_tuple);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const long neg_one = (long) -1, const_zero = (long) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(long) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(long) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(long) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(long),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL;
        PyObject *py_bytes = NULL, *arg_tuple = NULL, *kwds = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(long));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        arg_tuple = PyTuple_Pack(2, py_bytes, order_str);
        if (!arg_tuple) goto limited_bad;
        if (!is_unsigned) {
            kwds = PyDict_New();
            if (!kwds) goto limited_bad;
            if (PyDict_SetItemString(kwds, "signed", __Pyx_NewRef(Py_True))) goto limited_bad;
        }
        result = PyObject_Call(from_bytes, arg_tuple, kwds);
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(arg_tuple);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntFromPy */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int neg_one = (int) -1, const_zero = (int) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if ((sizeof(int) < sizeof(long))) {
            __PYX_VERIFY_RETURN_INT(int, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (int) val;
        }
    }
#endif
    if (unlikely(!PyLong_Check(x))) {
        int val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (int) -1;
        val = __Pyx_PyInt_As_int(tmp);
        Py_DECREF(tmp);
        return val;
    }
    if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
        if (unlikely(__Pyx_PyLong_IsNeg(x))) {
            goto raise_neg_overflow;
        } else if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(int, __Pyx_compact_upylong, __Pyx_PyLong_CompactValueUnsigned(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_DigitCount(x)) {
                case 2:
                    if ((8 * sizeof(int) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) >= 2 * PyLong_SHIFT)) {
                            return (int) (((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(int) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) >= 3 * PyLong_SHIFT)) {
                            return (int) (((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(int) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) >= 4 * PyLong_SHIFT)) {
                            return (int) (((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
            }
        }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
        if (unlikely(Py_SIZE(x) < 0)) {
            goto raise_neg_overflow;
        }
#else
        {
            int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
            if (unlikely(result < 0))
                return (int) -1;
            if (unlikely(result == 1))
                goto raise_neg_overflow;
        }
#endif
        if ((sizeof(int) <= sizeof(unsigned long))) {
            __PYX_VERIFY_RETURN_INT_EXC(int, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
        } else if ((sizeof(int) <= sizeof(unsigned PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(int, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
        }
    } else {
#if CYTHON_USE_PYLONG_INTERNALS
        if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(int, __Pyx_compact_pylong, __Pyx_PyLong_CompactValue(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_SignedDigitCount(x)) {
                case -2:
                    if ((8 * sizeof(int) - 1 > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 2 * PyLong_SHIFT)) {
                            return (int) (((int)-1)*(((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if ((8 * sizeof(int) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 2 * PyLong_SHIFT)) {
                            return (int) ((((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if ((8 * sizeof(int) - 1 > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 3 * PyLong_SHIFT)) {
                            return (int) (((int)-1)*(((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(int) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 3 * PyLong_SHIFT)) {
                            return (int) ((((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if ((8 * sizeof(int) - 1 > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 4 * PyLong_SHIFT)) {
                            return (int) (((int)-1)*(((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(int) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 4 * PyLong_SHIFT)) {
                            return (int) ((((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
            }
        }
#endif
        if ((sizeof(int) <= sizeof(long))) {
            __PYX_VERIFY_RETURN_INT_EXC(int, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
        } else if ((sizeof(int) <= sizeof(PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(int, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
        }
    }
    {
        int val;
        int ret = -1;
#if PY_VERSION_HEX >= 0x030d00A6 && !CYTHON_COMPILING_IN_LIMITED_API
        Py_ssize_t bytes_copied = PyLong_AsNativeBytes(
            x, &val, sizeof(val), Py_ASNATIVEBYTES_NATIVE_ENDIAN | (is_unsigned ? Py_ASNATIVEBYTES_UNSIGNED_BUFFER | Py_ASNATIVEBYTES_REJECT_NEGATIVE : 0));
        if (unlikely(bytes_copied == -1)) {
        } else if (unlikely(bytes_copied > (Py_ssize_t) sizeof(val))) {
            goto raise_overflow;
        } else {
            ret = 0;
        }
#elif PY_VERSION_HEX < 0x030d0000 && !(CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) || defined(_PyLong_AsByteArray)
        int one = 1; int is_little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&val;
        ret = _PyLong_AsByteArray((PyLongObject *)x,
                                    bytes, sizeof(val),
                                    is_little, !is_unsigned);
#else
        PyObject *v;
        PyObject *stepval = NULL, *mask = NULL, *shift = NULL;
        int bits, remaining_bits, is_negative = 0;
        int chunk_size = (sizeof(long) < 8) ? 30 : 62;
        if (likely(PyLong_CheckExact(x))) {
            v = __Pyx_NewRef(x);
        } else {
            v = PyNumber_Long(x);
            if (unlikely(!v)) return (int) -1;
            assert(PyLong_CheckExact(v));
        }
        {
            int result = PyObject_RichCompareBool(v, Py_False, Py_LT);
            if (unlikely(result < 0)) {
                Py_DECREF(v);
                return (int) -1;
            }
            is_negative = result == 1;
        }
        if (is_unsigned && unlikely(is_negative)) {
            Py_DECREF(v);
            goto raise_neg_overflow;
        } else if (is_negative) {
            stepval = PyNumber_Invert(v);
            Py_DECREF(v);
            if (unlikely(!stepval))
                return (int) -1;
        } else {
            stepval = v;
        }
        v = NULL;
        val = (int) 0;
        mask = PyLong_FromLong((1L << chunk_size) - 1); if (unlikely(!mask)) goto done;
        shift = PyLong_FromLong(chunk_size); if (unlikely(!shift)) goto done;
        for (bits = 0; bits < (int) sizeof(int) * 8 - chunk_size; bits += chunk_size) {
            PyObject *tmp, *digit;
            long idigit;
            digit = PyNumber_And(stepval, mask);
            if (unlikely(!digit)) goto done;
            idigit = PyLong_AsLong(digit);
            Py_DECREF(digit);
            if (unlikely(idigit < 0)) goto done;
            val |= ((int) idigit) << bits;
            tmp = PyNumber_Rshift(stepval, shift);
            if (unlikely(!tmp)) goto done;
            Py_DECREF(stepval); stepval = tmp;
        }
        Py_DECREF(shift); shift = NULL;
        Py_DECREF(mask); mask = NULL;
        {
            long idigit = PyLong_AsLong(stepval);
            if (unlikely(idigit < 0)) goto done;
            remaining_bits = ((int) sizeof(int) * 8) - bits - (is_unsigned ? 0 : 1);
            if (unlikely(idigit >= (1L << remaining_bits)))
                goto raise_overflow;
            val |= ((int) idigit) << bits;
        }
        if (!is_unsigned) {
            if (unlikely(val & (((int) 1) << (sizeof(int) * 8 - 1))))
                goto raise_overflow;
            if (is_negative)
                val = ~val;
        }
        ret = 0;
    done:
        Py_XDECREF(shift);
        Py_XDECREF(mask);
        Py_XDECREF(stepval);
#endif
        if (unlikely(ret))
            return (int) -1;
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to int");
    return (int) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to int");
    return (int) -1;
}

/* CIntFromPy */
static CYTHON_INLINE PY_LONG_LONG __Pyx_PyInt_As_PY_LONG_LONG(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const PY_LONG_LONG neg_one = (PY_LONG_LONG) -1, const_zero = (PY_LONG_LONG) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if ((sizeof(PY_LONG_LONG) < sizeof(long))) {
            __PYX_VERIFY_RETURN_INT(PY_LONG_LONG, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (PY_LONG_LONG) val;
        }
    }
#endif
    if (unlikely(!PyLong_Check(x))) {
        PY_LONG_LONG val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (PY_LONG_LONG) -1;
        val = __Pyx_PyInt_As_PY_LONG_LONG(tmp);
        Py_DECREF(tmp);
        return val;
    }
    if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
        if (unlikely(__Pyx_PyLong_IsNeg(x))) {
            goto raise_neg_overflow;
        } else if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(PY_LONG_LONG, __Pyx_compact_upylong, __Pyx_PyLong_CompactValueUnsigned(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_DigitCount(x)) {
                case 2:
                    if ((8 * sizeof(PY_LONG_LONG) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(PY_LONG_LONG, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(PY_LONG_LONG) >= 2 * PyLong_SHIFT)) {
                            return (PY_LONG_LONG) (((((PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (PY_LONG_LONG)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(PY_LONG_LONG) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(PY_LONG_LONG, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(PY_LONG_LONG) >= 3 * PyLong_SHIFT)) {
                            return (PY_LONG_LONG) (((((((PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (PY_LONG_LONG)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(PY_LONG_LONG) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(PY_LONG_LONG, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(PY_LONG_LONG) >= 4 * PyLong_SHIFT)) {
                            return (PY_LONG_LONG) (((((((((PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (PY_LONG_LONG)digits[0]));
                        }
                    }
                    break;
            }
        }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
        if (unlikely(Py_SIZE(x) < 0)) {
            goto raise_neg_overflow;
        }
#else
        {
            int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
            if (unlikely(result < 0))
                return (PY_LONG_LONG) -1;
            if (unlikely(result == 1))
                goto raise_neg_overflow;
        }
#endif
        if ((sizeof(PY_LONG_LONG) <= sizeof(unsigned long))) {
            __PYX_VERIFY_RETURN_INT_EXC(PY_LONG_LONG, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
        } else if ((sizeof(PY_LONG_LONG) <= sizeof(unsigned PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(PY_LONG_LONG, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
        }
    } else {
#if CYTHON_USE_PYLONG_INTERNALS
        if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(PY_LONG_LONG, __Pyx_compact_pylong, __Pyx_PyLong_CompactValue(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_SignedDigitCount(x)) {
                case -2:
                    if ((8 * sizeof(PY_LONG_LONG) - 1 > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(PY_LONG_LONG, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT)) {
                            return (PY_LONG_LONG) (((PY_LONG_LONG)-1)*(((((PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (PY_LONG_LONG)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if ((8 * sizeof(PY_LONG_LONG) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(PY_LONG_LONG, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT)) {
                            return (PY_LONG_LONG) ((((((PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (PY_LONG_LONG)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if ((8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(PY_LONG_LONG, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT)) {
                            return (PY_LONG_LONG) (((PY_LONG_LONG)-1)*(((((((PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (PY_LONG_LONG)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(PY_LONG_LONG) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(PY_LONG_LONG, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT)) {
                            return (PY_LONG_LONG) ((((((((PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (PY_LONG_LONG)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if ((8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(PY_LONG_LONG, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT)) {
                            return (PY_LONG_LONG) (((PY_LONG_LONG)-1)*(((((((((PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (PY_LONG_LONG)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(PY_LONG_LONG) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(PY_LONG_LONG, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT)) {
                            return (PY_LONG_LONG) ((((((((((PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (PY_LONG_LONG)digits[0])));
                        }
                    }
                    break;
            }
        }
#endif
        if ((sizeof(PY_LONG_LONG) <= sizeof(long))) {
            __PYX_VERIFY_RETURN_INT_EXC(PY_LONG_LONG, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
        } else if ((sizeof(PY_LONG_LONG) <= sizeof(PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(PY_LONG_LONG, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
        }
    }
    {
        PY_LONG_LONG val;
        int ret = -1;
#if PY_VERSION_HEX >= 0x030d00A6 && !CYTHON_COMPILING_IN_LIMITED_API
        Py_ssize_t bytes_copied = PyLong_AsNativeBytes(
            x, &val, sizeof(val), Py_ASNATIVEBYTES_NATIVE_ENDIAN | (is_unsigned ? Py_ASNATIVEBYTES_UNSIGNED_BUFFER | Py_ASNATIVEBYTES_REJECT_NEGATIVE : 0));
        if (unlikely(bytes_copied == -1)) {
        } else if (unlikely(bytes_copied > (Py_ssize_t) sizeof(val))) {
            goto raise_overflow;
        } else {
            ret = 0;
        }
#elif PY_VERSION_HEX < 0x030d0000 && !(CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) || defined(_PyLong_AsByteArray)
        int one = 1; int is_little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&val;
        ret = _PyLong_AsByteArray((PyLongObject *)x,
                                    bytes, sizeof(val),
                                    is_little, !is_unsigned);
#else
        PyObject *v;
        PyObject *stepval = NULL, *mask = NULL, *shift = NULL;
        int bits, remaining_bits, is_negative = 0;
        int chunk_size = (sizeof(long) < 8) ? 30 : 62;
        if (likely(PyLong_CheckExact(x))) {
            v = __Pyx_NewRef(x);
        } else {
            v = PyNumber_Long(x);
            if (unlikely(!v)) return (PY_LONG_LONG) -1;
            assert(PyLong_CheckExact(v));
        }
        {
            int result = PyObject_RichCompareBool(v, Py_False, Py_LT);
            if (unlikely(result < 0)) {
                Py_DECREF(v);
                return (PY_LONG_LONG) -1;
            }
            is_negative = result == 1;
        }
        if (is_unsigned && unlikely(is_negative)) {
            Py_DECREF(v);
            goto raise_neg_overflow;
        } else if (is_negative) {
            stepval = PyNumber_Invert(v);
            Py_DECREF(v);
            if (unlikely(!stepval))
                return (PY_LONG_LONG) -1;
        } else {
            stepval = v;
        }
        v = NULL;
        val = (PY_LONG_LONG) 0;
        mask = PyLong_FromLong((1L << chunk_size) - 1); if (unlikely(!mask)) goto done;
        shift = PyLong_FromLong(chunk_size); if (unlikely(!shift)) goto done;
        for (bits = 0; bits < (int) sizeof(PY_LONG_LONG) * 8 - chunk_size; bits += chunk_size) {
            PyObject *tmp, *digit;
            long idigit;
            digit = PyNumber_And(stepval, mask);
            if (unlikely(!digit)) goto done;
            idigit = PyLong_AsLong(digit);
            Py_DECREF(digit);
            if (unlikely(idigit < 0)) goto done;
            val |= ((PY_LONG_LONG) idigit) << bits;
            tmp = PyNumber_Rshift(stepval, shift);
            if (unlikely(!tmp)) goto done;
            Py_DECREF(stepval); stepval = tmp;
        }
        Py_DECREF(shift); shift = NULL;
        Py_DECREF(mask); mask = NULL;
        {
            long idigit = PyLong_AsLong(stepval);
            if (unlikely(idigit < 0)) goto done;
            remaining_bits = ((int) sizeof(PY_LONG_LONG) * 8) - bits - (is_unsigned ? 0 : 1);
            if (unlikely(idigit >= (1L << remaining_bits)))
                goto raise_overflow;
            val |= ((PY_LONG_LONG) idigit) << bits;
        }
        if (!is_unsigned) {
            if (unlikely(val & (((PY_LONG_LONG) 1) << (sizeof(PY_LONG_LONG) * 8 - 1))))
                goto raise_overflow;
            if (is_negative)
                val = ~val;
        }
        ret = 0;
    done:
        Py_XDECREF(shift);
        Py_XDECREF(mask);
        Py_XDECREF(stepval);
#endif
        if (unlikely(ret))
            return (PY_LONG_LONG) -1;
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to PY_LONG_LONG");
    return (PY_LONG_LONG) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to PY_LONG_LONG");
    return (PY_LONG_LONG) -1;
}

/* CIntFromPy */
static CYTHON_INLINE unsigned int __Pyx_PyInt_As_unsigned_int(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const unsigned int neg_one = (unsigned int) -1, const_zero = (unsigned int) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if ((sizeof(unsigned int) < sizeof(long))) {
            __PYX_VERIFY_RETURN_INT(unsigned int, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (unsigned int) val;
        }
    }
#endif
    if (unlikely(!PyLong_Check(x))) {
        unsigned int val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (unsigned int) -1;
        val = __Pyx_PyInt_As_unsigned_int(tmp);
        Py_DECREF(tmp);
        return val;
    }
    if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
        if (unlikely(__Pyx_PyLong_IsNeg(x))) {
            goto raise_neg_overflow;
        } else if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(unsigned int, __Pyx_compact_upylong, __Pyx_PyLong_CompactValueUnsigned(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_DigitCount(x)) {
                case 2:
                    if ((8 * sizeof(unsigned int) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(unsigned int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(unsigned int) >= 2 * PyLong_SHIFT)) {
                            return (unsigned int) (((((unsigned int)digits[1]) << PyLong_SHIFT) | (unsigned int)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(unsigned int) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(unsigned int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(unsigned int) >= 3 * PyLong_SHIFT)) {
                            return (unsigned int) (((((((unsigned int)digits[2]) << PyLong_SHIFT) | (unsigned int)digits[1]) << PyLong_SHIFT) | (unsigned int)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(unsigned int) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(unsigned int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(unsigned int) >= 4 * PyLong_SHIFT)) {
                            return (unsigned int) (((((((((unsigned int)digits[3]) << PyLong_SHIFT) | (unsigned int)digits[2]) << PyLong_SHIFT) | (unsigned int)digits[1]) << PyLong_SHIFT) | (unsigned int)digits[0]));
                        }
                    }
                    break;
            }
        }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
        if (unlikely(Py_SIZE(x) < 0)) {
            goto raise_neg_overflow;
        }
#else
        {
            int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
            if (unlikely(result < 0))
                return (unsigned int) -1;
            if (unlikely(result == 1))
                goto raise_neg_overflow;
        }
#endif
        if ((sizeof(unsigned int) <= sizeof(unsigned long))) {
            __PYX_VERIFY_RETURN_INT_EXC(unsigned int, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
        } else if ((sizeof(unsigned int) <= sizeof(unsigned PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(unsigned int, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
        }
    } else {
#if CYTHON_USE_PYLONG_INTERNALS
        if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(unsigned int, __Pyx_compact_pylong, __Pyx_PyLong_CompactValue(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_SignedDigitCount(x)) {
                case -2:
                    if ((8 * sizeof(unsigned int) - 1 > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(unsigned int, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(unsigned int) - 1 > 2 * PyLong_SHIFT)) {
                            return (unsigned int) (((unsigned int)-1)*(((((unsigned int)digits[1]) << PyLong_SHIFT) | (unsigned int)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if ((8 * sizeof(unsigned int) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(unsigned int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(unsigned int) - 1 > 2 * PyLong_SHIFT)) {
                            return (unsigned int) ((((((unsigned int)digits[1]) << PyLong_SHIFT) | (unsigned int)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if ((8 * sizeof(unsigned int) - 1 > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(unsigned int, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(unsigned int) - 1 > 3 * PyLong_SHIFT)) {
                            return (unsigned int) (((unsigned int)-1)*(((((((unsigned int)digits[2]) << PyLong_SHIFT) | (unsigned int)digits[1]) << PyLong_SHIFT) | (unsigned int)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(unsigned int) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(unsigned int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(unsigned int) - 1 > 3 * PyLong_SHIFT)) {
                            return (unsigned int) ((((((((unsigned int)digits[2]) << PyLong_SHIFT) | (unsigned int)digits[1]) << PyLong_SHIFT) | (unsigned int)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if ((8 * sizeof(unsigned int) - 1 > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(unsigned int, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(unsigned int) - 1 > 4 * PyLong_SHIFT)) {
                            return (unsigned int) (((unsigned int)-1)*(((((((((unsigned int)digits[3]) << PyLong_SHIFT) | (unsigned int)digits[2]) << PyLong_SHIFT) | (unsigned int)digits[1]) << PyLong_SHIFT) | (unsigned int)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(unsigned int) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(unsigned int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(unsigned int) - 1 > 4 * PyLong_SHIFT)) {
                            return (unsigned int) ((((((((((unsigned int)digits[3]) << PyLong_SHIFT) | (unsigned int)digits[2]) << PyLong_SHIFT) | (unsigned int)digits[1]) << PyLong_SHIFT) | (unsigned int)digits[0])));
                        }
                    }
                    break;
            }
        }
#endif
        if ((sizeof(unsigned int) <= sizeof(long))) {
            __PYX_VERIFY_RETURN_INT_EXC(unsigned int, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
        } else if ((sizeof(unsigned int) <= sizeof(PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(unsigned int, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
        }
    }
    {
        unsigned int val;
        int ret = -1;
#if PY_VERSION_HEX >= 0x030d00A6 && !CYTHON_COMPILING_IN_LIMITED_API
        Py_ssize_t bytes_copied = PyLong_AsNativeBytes(
            x, &val, sizeof(val), Py_ASNATIVEBYTES_NATIVE_ENDIAN | (is_unsigned ? Py_ASNATIVEBYTES_UNSIGNED_BUFFER | Py_ASNATIVEBYTES_REJECT_NEGATIVE : 0));
        if (unlikely(bytes_copied == -1)) {
        } else if (unlikely(bytes_copied > (Py_ssize_t) sizeof(val))) {
            goto raise_overflow;
        } else {
            ret = 0;
        }
#elif PY_VERSION_HEX < 0x030d0000 && !(CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) || defined(_PyLong_AsByteArray)
        int one = 1; int is_little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&val;
        ret = _PyLong_AsByteArray((PyLongObject *)x,
                                    bytes, sizeof(val),
                                    is_little, !is_unsigned);
#else
        PyObject *v;
        PyObject *stepval = NULL, *mask = NULL, *shift = NULL;
        int bits, remaining_bits, is_negative = 0;
        int chunk_size = (sizeof(long) < 8) ? 30 : 62;
        if (likely(PyLong_CheckExact(x))) {
            v = __Pyx_NewRef(x);
        } else {
            v = PyNumber_Long(x);
            if (unlikely(!v)) return (unsigned int) -1;
            assert(PyLong_CheckExact(v));
        }
        {
            int result = PyObject_RichCompareBool(v, Py_False, Py_LT);
            if (unlikely(result < 0)) {
                Py_DECREF(v);
                return (unsigned int) -1;
            }
            is_negative = result == 1;
        }
        if (is_unsigned && unlikely(is_negative)) {
            Py_DECREF(v);
            goto raise_neg_overflow;
        } else if (is_negative) {
            stepval = PyNumber_Invert(v);
            Py_DECREF(v);
            if (unlikely(!stepval))
                return (unsigned int) -1;
        } else {
            stepval = v;
        }
        v = NULL;
        val = (unsigned int) 0;
        mask = PyLong_FromLong((1L << chunk_size) - 1); if (unlikely(!mask)) goto done;
        shift = PyLong_FromLong(chunk_size); if (unlikely(!shift)) goto done;
        for (bits = 0; bits < (int) sizeof(unsigned int) * 8 - chunk_size; bits += chunk_size) {
            PyObject *tmp, *digit;
            long idigit;
            digit = PyNumber_And(stepval, mask);
            if (unlikely(!digit)) goto done;
            idigit = PyLong_AsLong(digit);
            Py_DECREF(digit);
            if (unlikely(idigit < 0)) goto done;
            val |= ((unsigned int) idigit) << bits;
            tmp = PyNumber_Rshift(stepval, shift);
            if (unlikely(!tmp)) goto done;
            Py_DECREF(stepval); stepval = tmp;
        }
        Py_DECREF(shift); shift = NULL;
        Py_DECREF(mask); mask = NULL;
        {
            long idigit = PyLong_AsLong(stepval);
            if (unlikely(idigit < 0)) goto done;
            remaining_bits = ((int) sizeof(unsigned int) * 8) - bits - (is_unsigned ? 0 : 1);
            if (unlikely(idigit >= (1L << remaining_bits)))
                goto raise_overflow;
            val |= ((unsigned int) idigit) << bits;
        }
        if (!is_unsigned) {
            if (unlikely(val & (((unsigned int) 1) << (sizeof(unsigned int) * 8 - 1))))
                goto raise_overflow;
            if (is_negative)
                val = ~val;
        }
        ret = 0;
    done:
        Py_XDECREF(shift);
        Py_XDECREF(mask);
        Py_XDECREF(stepval);
#endif
        if (unlikely(ret))
            return (unsigned int) -1;
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to unsigned int");
    return (unsigned int) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to unsigned int");
    return (unsigned int) -1;
}

/* CIntFromPy */
static CYTHON_INLINE unsigned PY_LONG_LONG __Pyx_PyInt_As_unsigned_PY_LONG_LONG(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const unsigned PY_LONG_LONG neg_one = (unsigned PY_LONG_LONG) -1, const_zero = (unsigned PY_LONG_LONG) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if ((sizeof(unsigned PY_LONG_LONG) < sizeof(long))) {
            __PYX_VERIFY_RETURN_INT(unsigned PY_LONG_LONG, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (unsigned PY_LONG_LONG) val;
        }
    }
#endif
    if (unlikely(!PyLong_Check(x))) {
        unsigned PY_LONG_LONG val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (unsigned PY_LONG_LONG) -1;
        val = __Pyx_PyInt_As_unsigned_PY_LONG_LONG(tmp);
        Py_DECREF(tmp);
        return val;
    }
    if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
        if (unlikely(__Pyx_PyLong_IsNeg(x))) {
            goto raise_neg_overflow;
        } else if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(unsigned PY_LONG_LONG, __Pyx_compact_upylong, __Pyx_PyLong_CompactValueUnsigned(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_DigitCount(x)) {
                case 2:
                    if ((8 * sizeof(unsigned PY_LONG_LONG) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(unsigned PY_LONG_LONG, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(unsigned PY_LONG_LONG) >= 2 * PyLong_SHIFT)) {
                            return (unsigned PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(unsigned PY_LONG_LONG) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(unsigned PY_LONG_LONG, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(unsigned PY_LONG_LONG) >= 3 * PyLong_SHIFT)) {
                            return (unsigned PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(unsigned PY_LONG_LONG) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(unsigned PY_LONG_LONG, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(unsigned PY_LONG_LONG) >= 4 * PyLong_SHIFT)) {
                            return (unsigned PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        }
                    }
                    break;
            }
        }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
        if (unlikely(Py_SIZE(x) < 0)) {
            goto raise_neg_overflow;
        }
#else
        {
            int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
            if (unlikely(result < 0))
                return (unsigned PY_LONG_LONG) -1;
            if (unlikely(result == 1))
                goto raise_neg_overflow;
        }
#endif
        if ((sizeof(unsigned PY_LONG_LONG) <= sizeof(unsigned long))) {
            __PYX_VERIFY_RETURN_INT_EXC(unsigned PY_LONG_LONG, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
        } else if ((sizeof(unsigned PY_LONG_LONG) <= sizeof(unsigned PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(unsigned PY_LONG_LONG, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
        }
    } else {
#if CYTHON_USE_PYLONG_INTERNALS
        if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(unsigned PY_LONG_LONG, __Pyx_compact_pylong, __Pyx_PyLong_CompactValue(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_SignedDigitCount(x)) {
                case -2:
                    if ((8 * sizeof(unsigned PY_LONG_LONG) - 1 > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(unsigned PY_LONG_LONG, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(unsigned PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT)) {
                            return (unsigned PY_LONG_LONG) (((unsigned PY_LONG_LONG)-1)*(((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if ((8 * sizeof(unsigned PY_LONG_LONG) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(unsigned PY_LONG_LONG, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(unsigned PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT)) {
                            return (unsigned PY_LONG_LONG) ((((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if ((8 * sizeof(unsigned PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(unsigned PY_LONG_LONG, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(unsigned PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT)) {
                            return (unsigned PY_LONG_LONG) (((unsigned PY_LONG_LONG)-1)*(((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(unsigned PY_LONG_LONG) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(unsigned PY_LONG_LONG, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(unsigned PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT)) {
                            return (unsigned PY_LONG_LONG) ((((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if ((8 * sizeof(unsigned PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(unsigned PY_LONG_LONG, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(unsigned PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT)) {
                            return (unsigned PY_LONG_LONG) (((unsigned PY_LONG_LONG)-1)*(((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(unsigned PY_LONG_LONG) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(unsigned PY_LONG_LONG, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(unsigned PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT)) {
                            return (unsigned PY_LONG_LONG) ((((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0])));
                        }
                    }
                    break;
            }
        }
#endif
        if ((sizeof(unsigned PY_LONG_LONG) <= sizeof(long))) {
            __PYX_VERIFY_RETURN_INT_EXC(unsigned PY_LONG_LONG, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
        } else if ((sizeof(unsigned PY_LONG_LONG) <= sizeof(PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(unsigned PY_LONG_LONG, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
        }
    }
    {
        unsigned PY_LONG_LONG val;
        int ret = -1;
#if PY_VERSION_HEX >= 0x030d00A6 && !CYTHON_COMPILING_IN_LIMITED_API
        Py_ssize_t bytes_copied = PyLong_AsNativeBytes(
            x, &val, sizeof(val), Py_ASNATIVEBYTES_NATIVE_ENDIAN | (is_unsigned ? Py_ASNATIVEBYTES_UNSIGNED_BUFFER | Py_ASNATIVEBYTES_REJECT_NEGATIVE : 0));
        if (unlikely(bytes_copied == -1)) {
        } else if (unlikely(bytes_copied > (Py_ssize_t) sizeof(val))) {
            goto raise_overflow;
        } else {
            ret = 0;
        }
#elif PY_VERSION_HEX < 0x030d0000 && !(CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) || defined(_PyLong_AsByteArray)
        int one = 1; int is_little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&val;
        ret = _PyLong_AsByteArray((PyLongObject *)x,
                                    bytes, sizeof(val),
                                    is_little, !is_unsigned);
#else
        PyObject *v;
        PyObject *stepval = NULL, *mask = NULL, *shift = NULL;
        int bits, remaining_bits, is_negative = 0;
        int chunk_size = (sizeof(long) < 8) ? 30 : 62;
        if (likely(PyLong_CheckExact(x))) {
            v = __Pyx_NewRef(x);
        } else {
            v = PyNumber_Long(x);
            if (unlikely(!v)) return (unsigned PY_LONG_LONG) -1;
            assert(PyLong_CheckExact(v));
        }
        {
            int result = PyObject_RichCompareBool(v, Py_False, Py_LT);
            if (unlikely(result < 0)) {
                Py_DECREF(v);
                return (unsigned PY_LONG_LONG) -1;
            }
            is_negative = result == 1;
        }
        if (is_unsigned && unlikely(is_negative)) {
            Py_DECREF(v);
            goto raise_neg_overflow;
        } else if (is_negative) {
            stepval = PyNumber_Invert(v);
            Py_DECREF(v);
            if (unlikely(!stepval))
                return (unsigned PY_LONG_LONG) -1;
        } else {
            stepval = v;
        }
        v = NULL;
        val = (unsigned PY_LONG_LONG) 0;
        mask = PyLong_FromLong((1L << chunk_size) - 1); if (unlikely(!mask)) goto done;
        shift = PyLong_FromLong(chunk_size); if (unlikely(!shift)) goto done;
        for (bits = 0; bits < (int) sizeof(unsigned PY_LONG_LONG) * 8 - chunk_size; bits += chunk_size) {
            PyObject *tmp, *digit;
            long idigit;
            digit = PyNumber_And(stepval, mask);
            if (unlikely(!digit)) goto done;
            idigit = PyLong_AsLong(digit);
            Py_DECREF(digit);
            if (unlikely(idigit < 0)) goto done;
            val |= ((unsigned PY_LONG_LONG) idigit) << bits;
            tmp = PyNumber_Rshift(stepval, shift);
            if (unlikely(!tmp)) goto done;
            Py_DECREF(stepval); stepval = tmp;
        }
        Py_DECREF(shift); shift = NULL;
        Py_DECREF(mask); mask = NULL;
        {
            long idigit = PyLong_AsLong(stepval);
            if (unlikely(idigit < 0)) goto done;
            remaining_bits = ((int) sizeof(unsigned PY_LONG_LONG) * 8) - bits - (is_unsigned ? 0 : 1);
            if (unlikely(idigit >= (1L << remaining_bits)))
                goto raise_overflow;
            val |= ((unsigned PY_LONG_LONG) idigit) << bits;
        }
        if (!is_unsigned) {
            if (unlikely(val & (((unsigned PY_LONG_LONG) 1) << (sizeof(unsigned PY_LONG_LONG) * 8 - 1))))
                goto raise_overflow;
            if (is_negative)
                val = ~val;
        }
        ret = 0;
    done:
        Py_XDECREF(shift);
        Py_XDECREF(mask);
        Py_XDECREF(stepval);
#endif
        if (unlikely(ret))
            return (unsigned PY_LONG_LONG) -1;
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to unsigned PY_LONG_LONG");
    return (unsigned PY_LONG_LONG) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to unsigned PY_LONG_LONG");
    return (unsigned PY_LONG_LONG) -1;
}

/* FormatTypeName */
#if CYTHON_COMPILING_IN_LIMITED_API
static __Pyx_TypeName
__Pyx_PyType_GetName(PyTypeObject* tp)
{
    PyObject *name = __Pyx_PyObject_GetAttrStr((PyObject *)tp,
                                               __pyx_n_s_name_2);
    if (unlikely(name == NULL) || unlikely(!PyUnicode_Check(name))) {
        PyErr_Clear();
        Py_XDECREF(name);
        name = __Pyx_NewRef(__pyx_n_s__46);
    }
    return name;
}
#endif

/* CIntFromPy */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const long neg_one = (long) -1, const_zero = (long) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if ((sizeof(long) < sizeof(long))) {
            __PYX_VERIFY_RETURN_INT(long, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (long) val;
        }
    }
#endif
    if (unlikely(!PyLong_Check(x))) {
        long val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (long) -1;
        val = __Pyx_PyInt_As_long(tmp);
        Py_DECREF(tmp);
        return val;
    }
    if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
        if (unlikely(__Pyx_PyLong_IsNeg(x))) {
            goto raise_neg_overflow;
        } else if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(long, __Pyx_compact_upylong, __Pyx_PyLong_CompactValueUnsigned(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_DigitCount(x)) {
                case 2:
                    if ((8 * sizeof(long) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) >= 2 * PyLong_SHIFT)) {
                            return (long) (((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(long) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) >= 3 * PyLong_SHIFT)) {
                            return (long) (((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(long) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) >= 4 * PyLong_SHIFT)) {
                            return (long) (((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
            }
        }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
        if (unlikely(Py_SIZE(x) < 0)) {
            goto raise_neg_overflow;
        }
#else
        {
            int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
            if (unlikely(result < 0))
                return (long) -1;
            if (unlikely(result == 1))
                goto raise_neg_overflow;
        }
#endif
        if ((sizeof(long) <= sizeof(unsigned long))) {
            __PYX_VERIFY_RETURN_INT_EXC(long, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
        } else if ((sizeof(long) <= sizeof(unsigned PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(long, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
        }
    } else {
#if CYTHON_USE_PYLONG_INTERNALS
        if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(long, __Pyx_compact_pylong, __Pyx_PyLong_CompactValue(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_SignedDigitCount(x)) {
                case -2:
                    if ((8 * sizeof(long) - 1 > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 2 * PyLong_SHIFT)) {
                            return (long) (((long)-1)*(((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if ((8 * sizeof(long) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 2 * PyLong_SHIFT)) {
                            return (long) ((((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if ((8 * sizeof(long) - 1 > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 3 * PyLong_SHIFT)) {
                            return (long) (((long)-1)*(((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(long) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 3 * PyLong_SHIFT)) {
                            return (long) ((((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if ((8 * sizeof(long) - 1 > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 4 * PyLong_SHIFT)) {
                            return (long) (((long)-1)*(((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(long) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 4 * PyLong_SHIFT)) {
                            return (long) ((((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
            }
        }
#endif
        if ((sizeof(long) <= sizeof(long))) {
            __PYX_VERIFY_RETURN_INT_EXC(long, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
        } else if ((sizeof(long) <= sizeof(PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(long, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
        }
    }
    {
        long val;
        int ret = -1;
#if PY_VERSION_HEX >= 0x030d00A6 && !CYTHON_COMPILING_IN_LIMITED_API
        Py_ssize_t bytes_copied = PyLong_AsNativeBytes(
            x, &val, sizeof(val), Py_ASNATIVEBYTES_NATIVE_ENDIAN | (is_unsigned ? Py_ASNATIVEBYTES_UNSIGNED_BUFFER | Py_ASNATIVEBYTES_REJECT_NEGATIVE : 0));
        if (unlikely(bytes_copied == -1)) {
        } else if (unlikely(bytes_copied > (Py_ssize_t) sizeof(val))) {
            goto raise_overflow;
        } else {
            ret = 0;
        }
#elif PY_VERSION_HEX < 0x030d0000 && !(CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) || defined(_PyLong_AsByteArray)
        int one = 1; int is_little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&val;
        ret = _PyLong_AsByteArray((PyLongObject *)x,
                                    bytes, sizeof(val),
                                    is_little, !is_unsigned);
#else
        PyObject *v;
        PyObject *stepval = NULL, *mask = NULL, *shift = NULL;
        int bits, remaining_bits, is_negative = 0;
        int chunk_size = (sizeof(long) < 8) ? 30 : 62;
        if (likely(PyLong_CheckExact(x))) {
            v = __Pyx_NewRef(x);
        } else {
            v = PyNumber_Long(x);
            if (unlikely(!v)) return (long) -1;
            assert(PyLong_CheckExact(v));
        }
        {
            int result = PyObject_RichCompareBool(v, Py_False, Py_LT);
            if (unlikely(result < 0)) {
                Py_DECREF(v);
                return (long) -1;
            }
            is_negative = result == 1;
        }
        if (is_unsigned && unlikely(is_negative)) {
            Py_DECREF(v);
            goto raise_neg_overflow;
        } else if (is_negative) {
            stepval = PyNumber_Invert(v);
            Py_DECREF(v);
            if (unlikely(!stepval))
                return (long) -1;
        } else {
            stepval = v;
        }
        v = NULL;
        val = (long) 0;
        mask = PyLong_FromLong((1L << chunk_size) - 1); if (unlikely(!mask)) goto done;
        shift = PyLong_FromLong(chunk_size); if (unlikely(!shift)) goto done;
        for (bits = 0; bits < (int) sizeof(long) * 8 - chunk_size; bits += chunk_size) {
            PyObject *tmp, *digit;
            long idigit;
            digit = PyNumber_And(stepval, mask);
            if (unlikely(!digit)) goto done;
            idigit = PyLong_AsLong(digit);
            Py_DECREF(digit);
            if (unlikely(idigit < 0)) goto done;
            val |= ((long) idigit) << bits;
            tmp = PyNumber_Rshift(stepval, shift);
            if (unlikely(!tmp)) goto done;
            Py_DECREF(stepval); stepval = tmp;
        }
        Py_DECREF(shift); shift = NULL;
        Py_DECREF(mask); mask = NULL;
        {
            long idigit = PyLong_AsLong(stepval);
            if (unlikely(idigit < 0)) goto done;
            remaining_bits = ((int) sizeof(long) * 8) - bits - (is_unsigned ? 0 : 1);
            if (unlikely(idigit >= (1L << remaining_bits)))
                goto raise_overflow;
            val |= ((long) idigit) << bits;
        }
        if (!is_unsigned) {
            if (unlikely(val & (((long) 1) << (sizeof(long) * 8 - 1))))
                goto raise_overflow;
            if (is_negative)
                val = ~val;
        }
        ret = 0;
    done:
        Py_XDECREF(shift);
        Py_XDECREF(mask);
        Py_XDECREF(stepval);
#endif
        if (unlikely(ret))
            return (long) -1;
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to long");
    return (long) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to long");
    return (long) -1;
}

/* FastTypeChecks */
#if CYTHON_COMPILING_IN_CPYTHON
static int __Pyx_InBases(PyTypeObject *a, PyTypeObject *b) {
    while (a) {
        a = __Pyx_PyType_GetSlot(a, tp_base, PyTypeObject*);
        if (a == b)
            return 1;
    }
    return b == &PyBaseObject_Type;
}
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b) {
    PyObject *mro;
    if (a == b) return 1;
    mro = a->tp_mro;
    if (likely(mro)) {
        Py_ssize_t i, n;
        n = PyTuple_GET_SIZE(mro);
        for (i = 0; i < n; i++) {
            if (PyTuple_GET_ITEM(mro, i) == (PyObject *)b)
                return 1;
        }
        return 0;
    }
    return __Pyx_InBases(a, b);
}
static CYTHON_INLINE int __Pyx_IsAnySubtype2(PyTypeObject *cls, PyTypeObject *a, PyTypeObject *b) {
    PyObject *mro;
    if (cls == a || cls == b) return 1;
    mro = cls->tp_mro;
    if (likely(mro)) {
        Py_ssize_t i, n;
        n = PyTuple_GET_SIZE(mro);
        for (i = 0; i < n; i++) {
            PyObject *base = PyTuple_GET_ITEM(mro, i);
            if (base == (PyObject *)a || base == (PyObject *)b)
                return 1;
        }
        return 0;
    }
    return __Pyx_InBases(cls, a) || __Pyx_InBases(cls, b);
}
#if PY_MAJOR_VERSION == 2
static int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject* exc_type2) {
    PyObject *exception, *value, *tb;
    int res;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&exception, &value, &tb);
    res = exc_type1 ? PyObject_IsSubclass(err, exc_type1) : 0;
    if (unlikely(res == -1)) {
        PyErr_WriteUnraisable(err);
        res = 0;
    }
    if (!res) {
        res = PyObject_IsSubclass(err, exc_type2);
        if (unlikely(res == -1)) {
            PyErr_WriteUnraisable(err);
            res = 0;
        }
    }
    __Pyx_ErrRestore(exception, value, tb);
    return res;
}
#else
static CYTHON_INLINE int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject *exc_type2) {
    if (exc_type1) {
        return __Pyx_IsAnySubtype2((PyTypeObject*)err, (PyTypeObject*)exc_type1, (PyTypeObject*)exc_type2);
    } else {
        return __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type2);
    }
}
#endif
static int __Pyx_PyErr_GivenExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    assert(PyExceptionClass_Check(exc_type));
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        PyObject *t = PyTuple_GET_ITEM(tuple, i);
        #if PY_MAJOR_VERSION < 3
        if (likely(exc_type == t)) return 1;
        #endif
        if (likely(PyExceptionClass_Check(t))) {
            if (__Pyx_inner_PyErr_GivenExceptionMatches2(exc_type, NULL, t)) return 1;
        } else {
        }
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject* exc_type) {
    if (likely(err == exc_type)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        if (likely(PyExceptionClass_Check(exc_type))) {
            return __Pyx_inner_PyErr_GivenExceptionMatches2(err, NULL, exc_type);
        } else if (likely(PyTuple_Check(exc_type))) {
            return __Pyx_PyErr_GivenExceptionMatchesTuple(err, exc_type);
        } else {
        }
    }
    return PyErr_GivenExceptionMatches(err, exc_type);
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *exc_type1, PyObject *exc_type2) {
    assert(PyExceptionClass_Check(exc_type1));
    assert(PyExceptionClass_Check(exc_type2));
    if (likely(err == exc_type1 || err == exc_type2)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        return __Pyx_inner_PyErr_GivenExceptionMatches2(err, exc_type1, exc_type2);
    }
    return (PyErr_GivenExceptionMatches(err, exc_type1) || PyErr_GivenExceptionMatches(err, exc_type2));
}
#endif

/* CheckBinaryVersion */
static unsigned long __Pyx_get_runtime_version(void) {
#if __PYX_LIMITED_VERSION_HEX >= 0x030B00A4
    return Py_Version & ~0xFFUL;
#else
    const char* rt_version = Py_GetVersion();
    unsigned long version = 0;
    unsigned long factor = 0x01000000UL;
    unsigned int digit = 0;
    int i = 0;
    while (factor) {
        while ('0' <= rt_version[i] && rt_version[i] <= '9') {
            digit = digit * 10 + (unsigned int) (rt_version[i] - '0');
            ++i;
        }
        version += factor * digit;
        if (rt_version[i] != '.')
            break;
        digit = 0;
        factor >>= 8;
        ++i;
    }
    return version;
#endif
}
static int __Pyx_check_binary_version(unsigned long ct_version, unsigned long rt_version, int allow_newer) {
    const unsigned long MAJOR_MINOR = 0xFFFF0000UL;
    if ((rt_version & MAJOR_MINOR) == (ct_version & MAJOR_MINOR))
        return 0;
    if (likely(allow_newer && (rt_version & MAJOR_MINOR) > (ct_version & MAJOR_MINOR)))
        return 1;
    {
        char message[200];
        PyOS_snprintf(message, sizeof(message),
                      "compile time Python version %d.%d "
                      "of module '%.100s' "
                      "%s "
                      "runtime version %d.%d",
                       (int) (ct_version >> 24), (int) ((ct_version >> 16) & 0xFF),
                       __Pyx_MODULE_NAME,
                       (allow_newer) ? "was newer than" : "does not match",
                       (int) (rt_version >> 24), (int) ((rt_version >> 16) & 0xFF)
       );
        return PyErr_WarnEx(NULL, message, 1);
    }
}

/* PyObjectSetAttrStr */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE int __Pyx_PyObject_SetAttrStr(PyObject* obj, PyObject* attr_name, PyObject* value) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_setattro))
        return tp->tp_setattro(obj, attr_name, value);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_setattr))
        return tp->tp_setattr(obj, PyString_AS_STRING(attr_name), value);
#endif
    return PyObject_SetAttr(obj, attr_name, value);
}
#endif

/* VoidPtrExport */
static int __Pyx_ExportVoidPtr(PyObject *name, void *p, const char *sig) {
    PyObject *d;
    PyObject *cobj = 0;
    d = PyDict_GetItem(__pyx_d, __pyx_n_s_pyx_capi);
    Py_XINCREF(d);
    if (!d) {
        d = PyDict_New();
        if (!d)
            goto bad;
        if (__Pyx_PyObject_SetAttrStr(__pyx_m, __pyx_n_s_pyx_capi, d) < 0)
            goto bad;
    }
    cobj = PyCapsule_New(p, sig, 0);
    if (!cobj)
        goto bad;
    if (PyDict_SetItem(d, name, cobj) < 0)
        goto bad;
    Py_DECREF(cobj);
    Py_DECREF(d);
    return 0;
bad:
    Py_XDECREF(cobj);
    Py_XDECREF(d);
    return -1;
}

/* FunctionExport */
static int __Pyx_ExportFunction(const char *name, void (*f)(void), const char *sig) {
    PyObject *d = 0;
    PyObject *cobj = 0;
    union {
        void (*fp)(void);
        void *p;
    } tmp;
    d = PyObject_GetAttrString(__pyx_m, (char *)"__pyx_capi__");
    if (!d) {
        PyErr_Clear();
        d = PyDict_New();
        if (!d)
            goto bad;
        Py_INCREF(d);
        if (PyModule_AddObject(__pyx_m, (char *)"__pyx_capi__", d) < 0)
            goto bad;
    }
    tmp.fp = f;
    cobj = PyCapsule_New(tmp.p, sig, 0);
    if (!cobj)
        goto bad;
    if (PyDict_SetItemString(d, name, cobj) < 0)
        goto bad;
    Py_DECREF(cobj);
    Py_DECREF(d);
    return 0;
bad:
    Py_XDECREF(cobj);
    Py_XDECREF(d);
    return -1;
}

/* VoidPtrImport */
#ifndef __PYX_HAVE_RT_ImportVoidPtr_3_0_12
#define __PYX_HAVE_RT_ImportVoidPtr_3_0_12
static int __Pyx_ImportVoidPtr_3_0_12(PyObject *module, const char *name, void **p, const char *sig) {
    PyObject *d = 0;
    PyObject *cobj = 0;
    d = PyObject_GetAttrString(module, (char *)"__pyx_capi__");
    if (!d)
        goto bad;
    cobj = PyDict_GetItemString(d, name);
    if (!cobj) {
        PyErr_Format(PyExc_ImportError,
            "%.200s does not export expected C variable %.200s",
                PyModule_GetName(module), name);
        goto bad;
    }
    if (!PyCapsule_IsValid(cobj, sig)) {
        PyErr_Format(PyExc_TypeError,
            "C variable %.200s.%.200s has wrong signature (expected %.500s, got %.500s)",
             PyModule_GetName(module), name, sig, PyCapsule_GetName(cobj));
        goto bad;
    }
    *p = PyCapsule_GetPointer(cobj, sig);
    if (!(*p))
        goto bad;
    Py_DECREF(d);
    return 0;
bad:
    Py_XDECREF(d);
    return -1;
}
#endif

/* InitStrings */
#if PY_MAJOR_VERSION >= 3
static int __Pyx_InitString(__Pyx_StringTabEntry t, PyObject **str) {
    if (t.is_unicode | t.is_str) {
        if (t.intern) {
            *str = PyUnicode_InternFromString(t.s);
        } else if (t.encoding) {
            *str = PyUnicode_Decode(t.s, t.n - 1, t.encoding, NULL);
        } else {
            *str = PyUnicode_FromStringAndSize(t.s, t.n - 1);
        }
    } else {
        *str = PyBytes_FromStringAndSize(t.s, t.n - 1);
    }
    if (!*str)
        return -1;
    if (PyObject_Hash(*str) == -1)
        return -1;
    return 0;
}
#endif
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t) {
    while (t->p) {
        #if PY_MAJOR_VERSION >= 3
        __Pyx_InitString(*t, t->p);
        #else
        if (t->is_unicode) {
            *t->p = PyUnicode_DecodeUTF8(t->s, t->n - 1, NULL);
        } else if (t->intern) {
            *t->p = PyString_InternFromString(t->s);
        } else {
            *t->p = PyString_FromStringAndSize(t->s, t->n - 1);
        }
        if (!*t->p)
            return -1;
        if (PyObject_Hash(*t->p) == -1)
            return -1;
        #endif
        ++t;
    }
    return 0;
}

#include <string.h>
static CYTHON_INLINE Py_ssize_t __Pyx_ssize_strlen(const char *s) {
    size_t len = strlen(s);
    if (unlikely(len > (size_t) PY_SSIZE_T_MAX)) {
        PyErr_SetString(PyExc_OverflowError, "byte string is too long");
        return -1;
    }
    return (Py_ssize_t) len;
}
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char* c_str) {
    Py_ssize_t len = __Pyx_ssize_strlen(c_str);
    if (unlikely(len < 0)) return NULL;
    return __Pyx_PyUnicode_FromStringAndSize(c_str, len);
}
static CYTHON_INLINE PyObject* __Pyx_PyByteArray_FromString(const char* c_str) {
    Py_ssize_t len = __Pyx_ssize_strlen(c_str);
    if (unlikely(len < 0)) return NULL;
    return PyByteArray_FromStringAndSize(c_str, len);
}
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject* o) {
    Py_ssize_t ignore;
    return __Pyx_PyObject_AsStringAndSize(o, &ignore);
}
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
#if !CYTHON_PEP393_ENABLED
static const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    char* defenc_c;
    PyObject* defenc = _PyUnicode_AsDefaultEncodedString(o, NULL);
    if (!defenc) return NULL;
    defenc_c = PyBytes_AS_STRING(defenc);
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    {
        char* end = defenc_c + PyBytes_GET_SIZE(defenc);
        char* c;
        for (c = defenc_c; c < end; c++) {
            if ((unsigned char) (*c) >= 128) {
                PyUnicode_AsASCIIString(o);
                return NULL;
            }
        }
    }
#endif
    *length = PyBytes_GET_SIZE(defenc);
    return defenc_c;
}
#else
static CYTHON_INLINE const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    if (unlikely(__Pyx_PyUnicode_READY(o) == -1)) return NULL;
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    if (likely(PyUnicode_IS_ASCII(o))) {
        *length = PyUnicode_GET_LENGTH(o);
        return PyUnicode_AsUTF8(o);
    } else {
        PyUnicode_AsASCIIString(o);
        return NULL;
    }
#else
    return PyUnicode_AsUTF8AndSize(o, length);
#endif
}
#endif
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
    if (
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
            __Pyx_sys_getdefaultencoding_not_ascii &&
#endif
            PyUnicode_Check(o)) {
        return __Pyx_PyUnicode_AsStringAndSize(o, length);
    } else
#endif
#if (!CYTHON_COMPILING_IN_PYPY && !CYTHON_COMPILING_IN_LIMITED_API) || (defined(PyByteArray_AS_STRING) && defined(PyByteArray_GET_SIZE))
    if (PyByteArray_Check(o)) {
        *length = PyByteArray_GET_SIZE(o);
        return PyByteArray_AS_STRING(o);
    } else
#endif
    {
        char* result;
        int r = PyBytes_AsStringAndSize(o, &result, length);
        if (unlikely(r < 0)) {
            return NULL;
        } else {
            return result;
        }
    }
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject* x) {
   int is_true = x == Py_True;
   if (is_true | (x == Py_False) | (x == Py_None)) return is_true;
   else return PyObject_IsTrue(x);
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject* x) {
    int retval;
    if (unlikely(!x)) return -1;
    retval = __Pyx_PyObject_IsTrue(x);
    Py_DECREF(x);
    return retval;
}
static PyObject* __Pyx_PyNumber_IntOrLongWrongResultType(PyObject* result, const char* type_name) {
    __Pyx_TypeName result_type_name = __Pyx_PyType_GetName(Py_TYPE(result));
#if PY_MAJOR_VERSION >= 3
    if (PyLong_Check(result)) {
        if (PyErr_WarnFormat(PyExc_DeprecationWarning, 1,
                "__int__ returned non-int (type " __Pyx_FMT_TYPENAME ").  "
                "The ability to return an instance of a strict subclass of int is deprecated, "
                "and may be removed in a future version of Python.",
                result_type_name)) {
            __Pyx_DECREF_TypeName(result_type_name);
            Py_DECREF(result);
            return NULL;
        }
        __Pyx_DECREF_TypeName(result_type_name);
        return result;
    }
#endif
    PyErr_Format(PyExc_TypeError,
                 "__%.4s__ returned non-%.4s (type " __Pyx_FMT_TYPENAME ")",
                 type_name, type_name, result_type_name);
    __Pyx_DECREF_TypeName(result_type_name);
    Py_DECREF(result);
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x) {
#if CYTHON_USE_TYPE_SLOTS
  PyNumberMethods *m;
#endif
  const char *name = NULL;
  PyObject *res = NULL;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_Check(x) || PyLong_Check(x)))
#else
  if (likely(PyLong_Check(x)))
#endif
    return __Pyx_NewRef(x);
#if CYTHON_USE_TYPE_SLOTS
  m = Py_TYPE(x)->tp_as_number;
  #if PY_MAJOR_VERSION < 3
  if (m && m->nb_int) {
    name = "int";
    res = m->nb_int(x);
  }
  else if (m && m->nb_long) {
    name = "long";
    res = m->nb_long(x);
  }
  #else
  if (likely(m && m->nb_int)) {
    name = "int";
    res = m->nb_int(x);
  }
  #endif
#else
  if (!PyBytes_CheckExact(x) && !PyUnicode_CheckExact(x)) {
    res = PyNumber_Int(x);
  }
#endif
  if (likely(res)) {
#if PY_MAJOR_VERSION < 3
    if (unlikely(!PyInt_Check(res) && !PyLong_Check(res))) {
#else
    if (unlikely(!PyLong_CheckExact(res))) {
#endif
        return __Pyx_PyNumber_IntOrLongWrongResultType(res, name);
    }
  }
  else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_TypeError,
                    "an integer is required");
  }
  return res;
}
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject* b) {
  Py_ssize_t ival;
  PyObject *x;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_CheckExact(b))) {
    if (sizeof(Py_ssize_t) >= sizeof(long))
        return PyInt_AS_LONG(b);
    else
        return PyInt_AsSsize_t(b);
  }
#endif
  if (likely(PyLong_CheckExact(b))) {
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(__Pyx_PyLong_IsCompact(b))) {
        return __Pyx_PyLong_CompactValue(b);
    } else {
      const digit* digits = __Pyx_PyLong_Digits(b);
      const Py_ssize_t size = __Pyx_PyLong_SignedDigitCount(b);
      switch (size) {
         case 2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
      }
    }
    #endif
    return PyLong_AsSsize_t(b);
  }
  x = PyNumber_Index(b);
  if (!x) return -1;
  ival = PyInt_AsSsize_t(x);
  Py_DECREF(x);
  return ival;
}
static CYTHON_INLINE Py_hash_t __Pyx_PyIndex_AsHash_t(PyObject* o) {
  if (sizeof(Py_hash_t) == sizeof(Py_ssize_t)) {
    return (Py_hash_t) __Pyx_PyIndex_AsSsize_t(o);
#if PY_MAJOR_VERSION < 3
  } else if (likely(PyInt_CheckExact(o))) {
    return PyInt_AS_LONG(o);
#endif
  } else {
    Py_ssize_t ival;
    PyObject *x;
    x = PyNumber_Index(o);
    if (!x) return -1;
    ival = PyInt_AsLong(x);
    Py_DECREF(x);
    return ival;
  }
}
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b) {
  return b ? __Pyx_NewRef(Py_True) : __Pyx_NewRef(Py_False);
}
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t ival) {
    return PyInt_FromSize_t(ival);
}


/* #### Code section: utility_code_pragmas_end ### */
#ifdef _MSC_VER
#pragma warning( pop )
#endif



/* #### Code section: end ### */
#endif /* Py_PYTHON_H */
